/*
** Module   : fwd-embedded-driver.p
** Abstract : FWD driver for running ADM windows in embedded mode.
**
** Copyright (c) 2017-2018, Golden Code Development Corporation.
**
** _#_ _I_ __Date__ ________________________________Description___________________________________
** 001 CA  20170825 Created initial version.
** 002 CA  20170928 Use destroyObject/destroy to clear a window resources and other improvements.
** 003 CA  20171004 Added APIs to expose legacy menus to the embedded driver (both definition and 
**                  event support).
** 004 CA  20171023 ADM version is kept on a per-window basis.
** 005 CA  20171109 Some improvements when initializing the embedded window's page.
** 006 CA  20171111 Allow default WINDOW:BGCOLOR-RGB to be set. Center ABL windows horizontally.
**                  Don't move the frame's column position, just row.
** 007 HC  20171206 Added fwd-user-logout shared variable that can be used as a logout condition.
** 008 CA  20180313 Added security support.  Added module and toolbar support.
** 009 CA  20180727 Fixes to better position ADM frames in embedded mode.
**                  Added notification that the business logic has switched the ADM page.
**                  Allow the JS side to hide the window status and/or message area for the 
**                  4GL windows handled via this driver.
** 010 CA  20180801 Allow multiple ADM windows to be shown at the same time, in embedded mode.
** 011 CA  20180814 Added fwd-embedded-mode var.
*/

/*
** This program is free software: you can redistribute it and/or modify
** it under the terms of the GNU Affero General Public License as
** published by the Free Software Foundation, either version 3 of the
** License, or (at your option) any later version.
**
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
** GNU Affero General Public License for more details.
**
** You may find a copy of the GNU Affero GPL version 3 at the following
** location: https://www.gnu.org/licenses/agpl-3.0.en.html
** 
** Additional terms under GNU Affero GPL version 3 section 7:
** 
**   Under Section 7 of the GNU Affero GPL version 3, the following additional
**   terms apply to the works covered under the License.  These additional terms
**   are non-permissive additional terms allowed under Section 7 of the GNU
**   Affero GPL version 3 and may not be removed by you.
** 
**   0. Attribution Requirement.
** 
**     You must preserve all legal notices or author attributions in the covered
**     work or Appropriate Legal Notices displayed by works containing the covered
**     work.  You may not remove from the covered work any author or developer
**     credit already included within the covered work.
** 
**   1. No License To Use Trademarks.
** 
**     This license does not grant any license or rights to use the trademarks
**     Golden Code, FWD, any Golden Code or FWD logo, or any other trademarks
**     of Golden Code Development Corporation. You are not authorized to use the
**     name Golden Code, FWD, or the names of any author or contributor, for
**     publicity purposes without written authorization.
** 
**   2. No Misrepresentation of Affiliation.
** 
**     You may not represent yourself as Golden Code Development Corporation or FWD.
** 
**     You may not represent yourself for publicity purposes as associated with
**     Golden Code Development Corporation, FWD, or any author or contributor to
**     the covered work, without written authorization.
** 
**   3. No Misrepresentation of Source or Origin.
** 
**     You may not represent the covered work as solely your work.  All modified
**     versions of the covered work must be marked in a reasonable way to make it
**     clear that the modified work is not originating from Golden Code Development
**     Corporation or FWD.  All modified versions must contain the notices of
**     attribution required in this license.
*/

/* allow both explicit and implicit (via p2j.embedded.js) run of this program. */
def new global shared var fwd-driver-handle as handle.
/* set this var in the embedded mode entry point's (i.e. emain.p), to true */
def new global shared var fwd-embedded-mode as log no-undo.

/* handle which refers to a procedure which (might) include custom APIs to be used for security,
   program initialization, etc */
def new global shared var fwd-custom-apis as handle.

/* indicates whether user has logged out, this variable can be used as a logout condition */
def new global shared var fwd-user-logout as log init false.

def new global shared temp-table fwd-embedded-window
   field pname as char
   field phandle as handle
   field menubar as handle
   field toolbar as handle
   field isAdm2 as logical
   field pages as logical extent 100
   field toolbars as char extent 100.

def temp-table page-toolbar
   field pname as char 
   field pnum as int
   field button as handle.

def var lastPage as int init ?.
def var lastWindow as char init ?.
def var lastWindowId as int init ?.
def var hpages as handle extent 100.
def var numPages as int.
def var fwdSelectPage as log init false.
def var admSelectedPage as int init ?.

// JS-side will configure the fwd-embedded-driver if window status-area and message-area are
// handled by JS.  For embedded mode clients, FWD will allow status-area and message-area to be 
// hidden, even if the window is realized (no error will be shown).
def var jsStatusArea as log init false.
def var jsMessageArea as log init false.

if valid-handle(fwd-driver-handle)
then do:
   /* if the driver was already loaded, don't run the external program code.  the APIs, even if
      invoked via this instance, it will re-use the globally shared temp-table */
   return.
end.

if not this-procedure:persistent
then do:
   // this program must be ran always in persistent mode...
   // TODO: raise error? abend?
end.

fwd-driver-handle = this-procedure.

function list-menu returns char private (input h as handle) forward.
function list-menu-tree returns char private (input h as handle, input lvl as int) forward.
function list-toolbar returns char private (input h as handle) forward.
function list-toolbar-button returns char private (input h as handle, input hx as int, input hy as int) forward.
function is-additional-toolbar returns log private (input h as handle, input hch as handle) forward.
function has-widget returns log private (input hparent as handle, input hchild as handle) forward.
function isPageFrame returns logical private (input hwid as handle) forward.
function resolveProgramName returns char private (input programName as char) forward.

procedure fwd-adm-page-changed.
   def input param p-caller as handle no-undo.
   def input param p-page# as int no-undo.

   def var pfile as char.
   def var windowId as int.
   def var pages as char.
   def var hwin as handle.
   
   pfile = resolveProgramName(p-caller:name).
   
   if substring(pfile, 1, length(pfile) - 1) <> substring(lastWindow, 1, length(lastWindow) - 1)
   then do:
      // some other ADM window has been selected, nothing to do.
   
      if can-do(p-caller:internal-entries, "fwdGetWindow")
         then hwin = dynamic-function("fwdGetWindow" in p-caller).
      
      if valid-handle(hwin)
      then do:
         windowId = hwin:hwnd.
         
         // notify the JS side that another page has been selected, to update the UI state
         P2J-REMOTE-CALL('~{ "api":"admPageChanged", "windowId":"' + string(windowId) + '", "page":"' + string(p-page#) + '" ~}').
      end.
      
      return.
   end.

   // a call from within the driver's page switch
   if fwdSelectPage 
   then do:
      // keep the page set by ADM
      admSelectedPage = p-page#.
      return.
   end.

   find first fwd-embedded-window where fwd-embedded-window.pname = lastWindow no-error.
   if avail fwd-embedded-window and not valid-handle(fwd-embedded-window.phandle)
   then do transaction: 
      delete fwd-embedded-window.
   end.

   if not avail fwd-embedded-window
   then do:
      return.
   end.
   
   // set this now
   lastPage = p-page#.
   
   run initEmbeddedPage(input p-caller, input p-page#, false, ?, output windowId, output pages).
   
   // notify the JS side that another page has been selected, to update the UI state
   P2J-REMOTE-CALL('~{ "api":"admPageChanged", "windowId":"' + string(windowId) + '", "page":"' + string(p-page#) + '" ~}').
end.

// publishing will be after business logic has selected the page:
// ADM: src/adm/method/brokcntn.i - broker-select-page - will have at line 409:
//     PUBLISH "fwd-adm-page-changed" (input p-caller, input p-page#).
// ADM2: src/adm2/containr.p - selectpage - will have at line 4592:
//     PUBLISH "fwd-adm-page-changed" (input target-procedure, input iCurrentPage).

subscribe procedure this-procedure to "fwd-adm-page-changed" anywhere.

procedure inMsg_embeddedStatusArea.
   def input param l as log.
   jsStatusArea = l.
end.

procedure inMsg_embeddedMessageArea.
   def input param l as log.
   jsMessageArea = l.
end.

procedure inMsg_logout.
   fwd-user-logout = true.

   if valid-handle(fwd-custom-apis) and
      can-do(fwd-custom-apis:internal-entries, "fwdLogout")
   then do:
      run fwdLogout in fwd-custom-apis.
   end.
   
   quit.
end.

function inMsg_fwdGetToolBar returns char.
   def var tbhandle as handle.

   find first fwd-embedded-window where fwd-embedded-window.pname = lastWindow  no-error.
   if not avail(fwd-embedded-window) then return ?.

   tbhandle = fwd-embedded-window.toolbar.
   if not valid-handle(tbhandle) then return ?.
   
   return fwd-embedded-window.toolbars[lastPage].
end.

function inMsg_fwdGetMenuBar returns char.
   find first fwd-embedded-window where fwd-embedded-window.pname = lastWindow no-error.
   
   if not avail(fwd-embedded-window) then return ?.
   
   def var mbhandle as handle.
   mbhandle = fwd-embedded-window.menubar.
   
   if not valid-handle(mbhandle) then return ?.
   
   return list-menu(mbhandle).
end.

function inMsg_fwdGetModules returns char.
   if valid-handle(fwd-custom-apis) and
      can-do(fwd-custom-apis:internal-entries, "fwdGetModules")
   then 
      return dynamic-function("fwdGetModules" in fwd-custom-apis).
   else
      return "".
end.

procedure inMsg_fwdMenuItemSelected.
   def input param hch as char.

   find first fwd-embedded-window where fwd-embedded-window.pname = lastWindow no-error.
   if not avail(fwd-embedded-window) then return.

   def var mbhandle as handle.
   mbhandle = fwd-embedded-window.menubar.
   
   if not valid-handle(mbhandle) then return.

   def var evt as char.
   def var h as handle.
   
   h = handle(hch).

   if not valid-handle(h) or not can-query(h, "TYPE") or not has-widget(mbhandle, h)
     then return.
   
   /*
   events:
   menu-drop: for sub-menu - only if it has children
   value-changed : for toggle-box
   choose : for everything else
   */
   if h:type = "SUB-MENU"
   then do:
      if h:first-child = ? then return.
      
      evt = "menu-drop".
   end.
   else if can-query(h, "toggle-box") and h:toggle-box
   then
      evt = "value-changed".
   else
      evt = "choose".
   
   apply evt to h.
end.

procedure inMsg_fwdToolbarItemSelected.
   def input param hch as char.

   find first fwd-embedded-window where fwd-embedded-window.pname = lastWindow no-error.
   if not avail(fwd-embedded-window) then return.

   def var tbhandle as handle.
   tbhandle = fwd-embedded-window.toolbar.
   
   if not valid-handle(tbhandle) then return.

   def var h as handle.
   
   /*
   events:
   choose: for button click
   */

   h = handle(hch).
   if not valid-handle(h) then return.
   
   if not can-find(page-toolbar where page-toolbar.pname = lastWindow and 
                                      page-toolbar.pnum = lastPage and page-toolbar.button = h) 
      then return.
   
   apply "choose" to h.
end.

procedure inMsg_fwdGetProgramModule.
   def input param module as char.
   def output param programName as char.

   if valid-handle(fwd-custom-apis) and
     can-do(fwd-custom-apis:internal-entries, "fwdGetProgramModule")
   then
   do: 
      run fwdGetProgramModule in fwd-custom-apis(input module, output programName).
      programName = resolveProgramName(programName).
   end.
   else
   do:
      programName = module.
   end.
end.

procedure inMsg_fwdDestroyWindow.
   def input param windowId as int.
   
   if windowId <> ? and windowId <> lastWindowId 
      then return.
      
   def var phandle as handle.
   
   do transaction:
      find first fwd-embedded-window where fwd-embedded-window.pname = lastWindow no-error.
      if not avail(fwd-embedded-window) then return.

      for each page-toolbar where page-toolbar.pname = lastWindow:
         delete page-toolbar.
      end.
   
      phandle = fwd-embedded-window.phandle.
      
      if valid-handle(phandle)
      then
      do on stop undo, leave:
         if fwd-embedded-window.isAdm2
         then
            run destroyObject in phandle no-error.
         else
            run dispatch in phandle ("destroy") no-error.
   
         delete object phandle no-error.
      end.
      
      delete fwd-embedded-window.
      
      lastWindowID = ?.
      lastWindow = ?.
      lastPage = ?.
   end.
end.

procedure inMsg_fwdSwitchWindow.
   def input param windowId as int.
   def input param pageNum as int.
   def output param windowProgram as char.
   def output param windowTitle as char.
   
   def var hwin as handle.
   def var hframe as handle.
   def var admVersion as char.
   def var pages as char.
   def var phandle as handle.
   def var tbhandle as handle.
   def var htab as handle.

   if lastWindowID = windowId
   then do:
      windowProgram = lastWindow.
      lastPage = pageNum.
      
      hwin = hwnd(windowId).
      windowTitle = hwin:title.
      return.
   end.
   else do:
      // check the previous window - if it has been destroyed, remove it...
      hwin = hwnd(lastWindowId).
      
      if not valid-handle(hwin) 
      then do:
         run inMsg_fwdDestroyWindow(lastWindowId).
      end.
      
      hwin = ?.
   end.

   hwin = hwnd(windowId).
   if not valid-handle(hwin)
   then do:
      // this window was already destroyed/not valid
      windowProgram = ?.
      windowTitle = ?.
      return.
   end.
   
   phandle = hwin:instantiating-procedure.
   
   run getFwdHandles(output hwin, output hframe, output admVersion, output pages, input phandle) no-error.
   
   if not valid-handle(hwin) or hwin:hwnd <> windowId or admVersion = ? or hframe = ?
   then do:
      // this window was not executed from an ADM-style program...
      windowProgram = ?.
      windowTitle = ?.
      return.
   end.
   
   windowProgram = resolveProgramName(phandle:name).
   windowTitle = hwin:title.
   // set this now, as later inMsg_fwdShowWindow will be called and we need to be set on this
   lastWindow = windowProgram.
   lastWindowId = windowId.
   lastPage = pageNum.
   
   find first fwd-embedded-window where fwd-embedded-window.pname = windowProgram no-error.
   if avail fwd-embedded-window and not valid-handle(fwd-embedded-window.phandle)
   then do transaction: 
      delete fwd-embedded-window.
   end.

   if not avail fwd-embedded-window
   then 
   do transaction:
      if can-do(phandle:internal-entries, "fwdGetToolbarFrame")
         then tbhandle = dynamic-function("fwdGetToolbarFrame" in phandle).
         else tbhandle = ?.
      
      /* do this before anything else */
      create fwd-embedded-window.
      fwd-embedded-window.pname = windowProgram.
      fwd-embedded-window.phandle = phandle.
      fwd-embedded-window.menubar = hwin:menubar.
      fwd-embedded-window.toolbar = tbhandle.
      fwd-embedded-window.isAdm2 = (admVersion >= "ADM2.2").
      fwd-embedded-window.pages = false.

      run initEmbeddedWindow(hwin).
      
      if can-do(phandle:internal-entries, "fwdGetTabHandle")
         then htab = dynamic-function("fwdGetTabHandle" in phandle).
         else htab = ?.

      if valid-handle(htab)
      then do:
         unsubscribe procedure htab to "changeFolderPage".
      end.
   end.
   else do:
      hwin:visible = true.
      hwin:move-to-top().
   end.
end.

procedure inMsg_fwdShowWindow.
   def input param pfile as char.
   def input param pnum as int.
   def input param module as char.
   def input param bgColorRgb as int.
   def output param windowId as int.
   def output param pages as char.
   
   windowId = ?.
   pages = ?.
   
   def var admVersion as char.
   def var phandle as handle.
   def var hframe as handle.
   def var tbhandle as handle.
   def var hwin as handle.
   def var hwid as handle.
   def var htab as handle.
   def var lastWin as handle.
   
   // standardize the program file name
   pfile = resolveProgramName(pfile).
   
   /*
   do not destroy the window, just create a new module and select it.
   if lastWindow <> pfile 
   then do:
      run inMsg_fwdDestroyWindow(lastWindowId).
   end.
   */
   lastWindow = pfile.
   lastPage = pnum.
   
   find first fwd-embedded-window where fwd-embedded-window.pname = pfile no-error.
   if avail fwd-embedded-window and not valid-handle(fwd-embedded-window.phandle)
   then do transaction: 
      delete fwd-embedded-window.
   end.

   if not avail fwd-embedded-window 
   then 
   do transaction:
      lastWin = session:last-child.

      if valid-handle(fwd-custom-apis) then do:
      
         if can-do(fwd-custom-apis:internal-entries, "fwdCheckAccess") then do:
            if not dynamic-function("fwdCheckAccess" in fwd-custom-apis, input pfile, input module)
               then return.
         end.
         else do:
            /* TODO: FWD security plugin */
         end.
         
         if can-do(fwd-custom-apis:internal-entries, "fwdRunModule") then do:
            phandle = dynamic-function("fwdRunModule" in fwd-custom-apis, input pfile, input module).
         end.
         else do:
            run value(pfile) persistent set phandle.
         end.
      end.
      else do:
         /* TODO: FWD security plugin */
         run value(pfile) persistent set phandle.
      end.
      
      if not valid-handle(phandle)
      then do:
         message "Module " module " can not be ran." view-as alert-box error.
         return.
      end.
      
      run getFwdHandles(output hwin, output hframe, output admVersion, output pages, input phandle) no-error.
      if can-do(phandle:internal-entries, "fwdGetTabHandle")
         then htab = dynamic-function("fwdGetTabHandle" in phandle).
         else htab = ?.

      // TODO: what if the business logic starts its own event-processing loop (i.e. WAIT-FOR),
      // with its main ADM window? this will never be reached until that window is discarded 

      if hwin <> ? then do:
         if not valid-handle(hwin) or (pages <> ? and pages <> "" and not valid-handle(htab))
         then do:
            // if the tab handle was not initialized but this window has pages, it means the 
            // module initialization/shown was aborted by some business logic.
            
            pages = ?.
            windowId = ?.
            delete object phandle no-error.
            return.
         end.
      
         windowId = hwin:hwnd.
      end.
      else do:
         // if business logic aborted window initialization, do nothing
         if lastWin = session:last-child
         then do:
            pages = ?.
            windowId = ?.
            delete object phandle no-error.
            return.
         end.
         
         /* the windowId is needed, but there are cases where the window is opened by another 
            program... use SESSION:LAST-CHILD instead to get last opened window. */
         hwin = session:last-child.
         windowId = hwin:hwnd.
         lastWindowId = windowId.
         
         if jsStatusArea then hwin:status-area = false.
         if jsMessageArea then hwin:message-area = false.

         if bgColorRgb <> ?
         then do:
            hwin:bgcolor-rgb = bgColorRgb.
         end.

         hwin:visible = true.
         hwin:move-to-top().
         
         hwin = ?.
      end.
      
      if admVersion = ? or hframe = ? or hwin = ? then do:
         /* TODO: show some error message ? */
         delete object phandle.
         return.
      end.

      hwin:disable-redraw = true.
      
      if can-do(phandle:internal-entries, "fwdGetToolbarFrame")
         then tbhandle = dynamic-function("fwdGetToolbarFrame" in phandle).
         else tbhandle = ?.
      
      /* do this before anything else */
      create fwd-embedded-window.
      fwd-embedded-window.pname = pfile.
      fwd-embedded-window.phandle = phandle.
      fwd-embedded-window.menubar = hwin:menubar.
      fwd-embedded-window.toolbar = tbhandle.
      fwd-embedded-window.isAdm2 = (admVersion >= "ADM2.2").
      fwd-embedded-window.pages = false.

      run initEmbeddedWindow(hwin).
      hwin:disable-redraw = true.
      
      if fwd-embedded-window.isAdm2
      then
         run initializeObject in phandle.
      else
         run dispatch in phandle ("initialize") no-error.

      if valid-handle(htab)
      then do:
         unsubscribe procedure htab to "changeFolderPage".
      end.
   end.
   else do:
      phandle = fwd-embedded-window.phandle.
      tbhandle = fwd-embedded-window.toolbar.
   end.
   
   run initEmbeddedPage(phandle, pnum, true, bgColorRgb, output windowId, output pages).
end.

procedure initEmbeddedPage private.
   def input param phandle as handle.
   def input param pnum as int.
   def input param selectPage as log.
   def input param bgColorRgb as int.
   def output param windowId as int.
   def output param pages as char.

   def var admVersion as char.
   def var hframe as handle.
   def var tbhandle as handle.
   def var hwin as handle.
   def var hwid as handle.

   tbhandle = fwd-embedded-window.toolbar.
   run getFwdHandles(output hwin, output hframe, output admVersion, output pages, input phandle) no-error.

   hwin:disable-redraw = true.

   def var wx as int.
   def var wy as int.
   wx = (session:width-pixels - hwin:width-pixels) / 2.
   wy = (session:height-pixels - hwin:height-pixels) / 2.
   wx = (if wx < 1 then 1 else wx).
   wy = (if wy < 1 then 1 else wy).
   if hwin:x <> wx then hwin:x = wx no-error.
   if hwin:y <> wy then hwin:y = wy no-error.
   
   if can-do(phandle:internal-entries, "fwdGetPageHandles")
   then do transaction:
      if selectPage 
      then do:
         fwdSelectPage = true.
         admSelectedPage = ?.
         
         if fwd-embedded-window.isAdm2
         then
            run selectPage in phandle (pnum) no-error.
         else
            run select-page in phandle (pnum) no-error.
         
         if admSelectedPage <> ? and admSelectedPage <> pnum
         then do:
            // the ADM code has selected another page - switch to it
            pnum = admSelectedPage.
            lastPage = admSelectedPage.
         end.
         else if admSelectedPage = ?
         then do:
            // TODO: the page switch has failed or was not needed - what to do?
         end.
         
         admSelectedPage = ?.
         fwdSelectPage = false.
      end.
      
      if not fwd-embedded-window.pages[pnum]
      then do:
         run fwdGetPageHandles in phandle (input pnum, output hpages, output numPages).

         for each page-toolbar where page-toolbar.pnum = pnum and page-toolbar.pname = lastWindow:
            delete page-toolbar.
         end.
         
         if tbhandle <> ? and not tbhandle:visible then tbhandle:visible = true.
         fwd-embedded-window.toolbars[pnum] = list-toolbar(tbhandle).

         hwid = hwin:first-child.
         do while hwid <> ?:
            if hwid:type = "FRAME" and not (hwid = hframe or isPageFrame(hwid))
            then do:
               hwid:visible = false.
            end.
            
            hwid = hwid:next-sibling.
         end.

         run initEmbeddedWidgets (input phandle, input hwin, input hframe, input tbhandle).
      
         fwd-embedded-window.pages[pnum] = true.
      end.
   end.

   if jsStatusArea then hwin:status-area = false.
   if jsMessageArea then hwin:message-area = false.

   hwin:window-state = WINDOW-NORMAL.
   hwin:visible = true.
   if bgColorRgb <> ?
   then do:
      hwin:bgcolor-rgb = bgColorRgb.
      if hframe:bgcolor = 15 then hframe:bgcolor = ?.
      hframe:bgcolor-rgb = bgColorRgb.
   end.
   hwin:disable-redraw = false.
   hwin:move-to-top().
   
   // let the embedded JS driver know this window's ID, so that it can track it if it gets
   // destroyed by the business logic
   windowId = hwin:hwnd.
   
   lastWindowId = windowId.
   lastPage = pnum.
end.

procedure getFwdHandles private.
   def output param hwin as handle.
   def output param hframe as handle.
   def output param admVersion as char.
   def output param pages as char.
   def input  param phandle as handle.
   
   hwin = ?.
   hframe = ?.
   admVersion = ?.
   
   if can-do(phandle:internal-entries, "fwdGetWindow")
      then hwin = dynamic-function("fwdGetWindow" in phandle).
   if can-do(phandle:internal-entries, "fwdGetMainFrame")
      then hframe = dynamic-function("fwdGetMainFrame" in phandle).
   if can-do(phandle:internal-entries, "fwdGetADMVersion")
      then admVersion = dynamic-function("fwdGetADMVersion" in phandle).
   if can-do(phandle:internal-entries, "fwdGetPages")
      then pages = dynamic-function("fwdGetPages" in phandle).
end.

procedure initEmbeddedWindow private.
   def input param hwin as handle.
   
   /* remove any menu attached to the window */
   hwin:menubar = ?.
   
   /* remove status-area and message-area */
   hwin:status-area = no no-error.
   hwin:message-area = no no-error.
   
   hwin:top-only = false.
   hwin:always-on-top = false.
   
   hwin:visible = true.
   hwin:move-to-top().
end.

function isPageFrame returns logical private (input hwid as handle).
   def var i as int.
   
   do i = 1 to numPages:
      if hpages[i] = hwid:instantiating-procedure
      then do:
         return true.
      end.
   end.
   
   return false.
end.

function resolveProgramName returns char private (input programName as char).
   programName = search(programName).
   
   if r-index(programName, ".r") = length(programName) - 1 then
   l1:
   do:
      programName = substring(programName, 1, length(programName) - 2).
      
      def var prog2 as char.
      prog2 = search(programName + ".w").
      if prog2 <> ? then
      do:
         programName = prog2.
         leave l1.
      end.
      
      prog2 = search(programName + ".p").
      if prog2 <> ? then
      do:
         programName = prog2.
         leave l1.
      end.
      
      message "Module " programName " can not be ran." view-as alert-box error.
      programName = ?.
   end.
   
   return programName.
end.

function screenRow returns dec private (input hwid as handle).
   def var hrow as dec init 0.
   
   do while hwid <> ?:
      hrow = hrow + hwid:y.
      if can-query(hwid, "border-top-pixels") then hrow = hrow + hwid:border-top-pixels.
      hwid = hwid:frame.
   end.
   
   return hrow.
end.

function screenCol returns dec private (input hwid as handle).
   def var hcol as dec init 0.
   
   do while hwid <> ?:
      hcol = hcol + hwid:x.
      if can-query(hwid, "border-left-pixels") then hcol = hcol + hwid:border-left-pixels.
      hwid = hwid:frame.
   end.
   
   return hcol.
end.

procedure initEmbeddedWidgets private.
   def input param phandle as handle.
   def input param hwin as handle.
   def input param hframe as handle.
   def input param htoolbar as handle.

   def var hiddenWidgets as char.
   def var hwid as handle.
   def var widname as char.
   def var idx as int.
   def var k as dec.

   if can-do(phandle:internal-entries, "fwdGetHiddenEmbeddedWidgets")
   then do:
      hiddenWidgets = dynamic-function("fwdGetHiddenEmbeddedWidgets" in phandle).
   end.
   else hiddenWidgets = "".
   
   /* hide all widgets which are not visible in embedded mode */
   idx = 1.
   do while idx <= num-entries(hiddenWidgets, ","):
      widname = entry(idx, hiddenWidgets, ",") no-error.
      if not error-status:error
      then do:
         hwid = hframe:first-child:first-child.
         do while hwid <> ?:
            if hwid:name = widname 
            then do:
               hwid:hidden = true.
            end.
            hwid = hwid:next-sibling.
         end.
      end.
      idx = idx + 1.
   end.

   hwid = hframe:first-child:first-child.
   do while hwid <> ?:
      // hwid:type = "FRAME" - this test was removed, as there may be non-FRAME widgets in the
      // main frame, which must be hidden
      if not isPageFrame(hwid) or is-additional-toolbar(htoolbar, hwid)
      then do:
         hwid:hidden = true.
      end.
      hwid = hwid:next-sibling.
   end.
   
   def var wrow as dec.
   def var wcol as dec.
   def var minRow as dec init ?.
   def var minCol as dec init ?.
   
   /* go through all widgets and compute min row/col from non-hidden ones */
   hwid = hframe:first-child:first-child.
   do while hwid <> ?:
      if not hwid:hidden
      then do:
         wrow = screenRow(hwid).
         wcol = screenCol(hwid).

         if minRow = ? or minRow > wrow then minRow = wrow.
         if minCol = ? or minCol > wcol then minCol = wcol.
      end.

      hwid = hwid:next-sibling.
   end.
   
   /* go through all page frames in the window and compute min row/col */
   hwid = hwin:first-child.
   do while hwid <> ?:
      if hwid:type = "FRAME" and isPageFrame(hwid)
      then do:
         wrow = screenRow(hwid).
         wcol = screenCol(hwid).

         if minRow = ? or minRow > wrow then minRow = wrow.
         if minCol = ? or minCol > wcol then minCol = wcol.
      end.

      hwid = hwid:next-sibling.
   end.
   
   /* re-adjust the widget location in the main frame */
   hwid = hframe:first-child:first-child.
   do while hwid <> ?:
      if not hwid:hidden
      then do:
         wrow = screenRow(hwid).
         wcol = screenCol(hwid).
         
         k = wrow - (minRow - 1).
         if hwid:y <> k then hwid:y = k no-error.
         
         /*
         k = wcol - (minCol - 1).
         if hwid:x <> k then hwid:x = k no-error.
         */
      end.

      hwid = hwid:next-sibling.
   end.
   
   /* re-adjust the location for non-nested page frames, relative to the widgets in hframe */
   hwid = hwin:first-child.
   do while hwid <> ?:
      if hwid:type = "FRAME" and isPageFrame(hwid)
      then do:
         wrow = screenRow(hwid).
         wcol = screenCol(hwid).
         
         k = wrow - (minRow - 1).
         if hwid:y <> k then hwid:y = k no-error.
         
         /*
         k = wcol - (minCol - 1).
         if hwid:x <> k then hwid:x = k no-error.
         */
      end.

      hwid = hwid:next-sibling.
   end.

   /* one more time, to move the main frame and non-nested page frames */
   minRow = ?.
   minCol = ?.

   /* go through all page frames in the window and compute min row/col */
   hwid = hwin:first-child.
   do while hwid <> ?:
      if hwid:type = "FRAME" and (hwid = hframe or isPageFrame(hwid))
      then do:
         wrow = screenRow(hwid).
         wcol = screenCol(hwid).

         if minRow = ? or minRow > wrow then minRow = wrow.
         if minCol = ? or minCol > wcol then minCol = wcol.
      end.

      hwid = hwid:next-sibling.
   end.
   
   /* re-adjust the location for non-nested page frames, relative to main frame */
   hwid = hwin:first-child.
   do while hwid <> ?:
      if hwid:type = "FRAME" and (hwid = hframe or isPageFrame(hwid))
      then do:
         wrow = screenRow(hwid).
         wcol = screenCol(hwid).
         
         k = wrow - (minRow - 1).
         if hwid:y <> k then hwid:y = k no-error.
         
         /*
         k = wcol - (minCol - 1).
         if hwid:x <> k then hwid:x = k no-error.
         */
      end.

      hwid = hwid:next-sibling.
   end.
   
   /* make all page-frames visible
   hwid = hwin:first-child.
   do while hwid <> ?:
      if hwid:type = "FRAME" and (hwid = hframe or isPageFrame(hwid))
      then do:
         if not hwid:visible then hwid:visible = true.
      end.

      hwid = hwid:next-sibling.
   end.
   */
end.

function list-menu-tree returns char private (input h as handle, input lvl as int).
/*
{
   "sensitive" : true/false, // if set, all menu-items are disabled (only for root)
   "visible"   : true/false, // if set, menubar is not visible (only for root)
   "menu-items" :
   [ 
      {
         "handle"      : "",
         "type"        : "MENU-ITEM/SUB-MENU",
         "subtype"     : "NORMAL/RULE/SKIP", // only for type=MENU-ITEM, default NORMAL
         "label"       : "", // this is the preprocessed 'label' attribute, with mnemonic resolved
         "accelerator" : "", // optional
         "mnemonic"    : 0,1,2,etc, // optional, default 0; this is the resolved mnemonic index, in the preprocessed label
         "read-only"   : true/false, // optional, default false; can't be chosen (same as sensitive=false?)
         "sensitive"   : true/false, // optional, default true
         // "visible"     : true/false, // optional, default true - not used, does nothing for menu-related stuff
         "toggle-box"  : true/false,  // optional, default false
         "checked"     : true/false, // optional, default false,  only for toggle-box
         "menu-items"  : // optional, only for sub-menus
         [
            {
               // recursive, again each element
            }
         ]
      },
      ...
   ]
}
*/

   def var res as char.
   def var hch as handle.
   
   res = "".

   res = res + fill(" " , lvl * 3) + '"menu-items" : ~n'.
   res = res + fill(" " , lvl * 3) + '[~n'.
   lvl = lvl + 1.
   
   hch = h:first-child.
   do while hch <> ?:
      res = res + fill(" " , lvl * 3) + "~{~n".
      lvl = lvl + 1.

      res = res + fill(" " , lvl * 3) + '"handle"      : "' + string(hch) + '",~n'.
      res = res + fill(" " , lvl * 3) + '"type"        : "' + hch:type + '",~n'.
      
      if (can-query(hch, "subtype") and hch:subtype <> ? and hch:subtype <> "NORMAL")
      then do:
         res = res + fill(" " , lvl * 3) + '"subtype"     : "' + hch:subtype  + '",~n'.
      end.
      
      if (can-query(hch, "preprocessed-label") and hch:preprocessed-label <> ?)
      then do:
         res = res + fill(" " , lvl * 3) + '"label"       : "' + hch:preprocessed-label + '",~n'.
      end.
      
      if (can-query(hch, "accelerator") and hch:accelerator <> ?)
      then do:
         res = res + fill(" " , lvl * 3) + '"accelerator" : "' + hch:accelerator + '",~n'.
      end.
      
      if (can-query(hch, "mnemonic") and hch:mnemonic <> ? and hch:mnemonic <> 0)
      then do:
         res = res + fill(" " , lvl * 3) + '"mnemonic"    : "' + string(hch:mnemonic) + '",~n'.
      end.
      
      if (can-query(hch, "sensitive") and hch:sensitive <> ? and not hch:sensitive)
      then do:
         res = res + fill(" " , lvl * 3) + '"sensitive"   : ' + string(hch:sensitive, "true/false") + ',~n'.
      end.
      
      if (can-query(hch, "read-only") and hch:read-only <> ? and hch:read-only)
      then do:
         res = res + fill(" " , lvl * 3) + '"read-only"   : ' + string(hch:read-only, "true/false") + ',~n'.
      end.
      
      /*
      if (can-query(hch, "visible") and hch:visible <> ? and not hch:visible)
      then do:
         res = res + fill(" " , lvl * 3) + '"visible"     : ' + string(hch:visible, "true/false") + ',~n'.
      end.
      */
      
      if (can-query(hch, "toggle-box") and hch:toggle-box <> ? and hch:toggle-box)
      then do:
         res = res + fill(" " , lvl * 3) + '"toggle-box"  : ' + string(hch:toggle-box, "true/false") + ',~n'.
      end.
      
      if (can-query(hch, "checked") and hch:checked <> ? and hch:checked)
      then do:
         res = res + fill(" " , lvl * 3) + '"checked"     : ' + string(hch:checked, "true/false") + ',~n'.
      end.

      if can-query(hch, "first-child")
      then do:
         res = res + list-menu-tree(hch, lvl).
      end.

      res = trim(res).
      if r-index(res, ",") = length(res)
      then do:
         res = substring(res, 1, length(res) - 1).
      end.

      lvl = lvl - 1.
      res = res + fill(" " , lvl * 3) + "}" + (if hch:next-sibling <> ? then "," else "") + "~n".
      
      hch = hch:next-sibling.
   end. 
   lvl = lvl - 1.
   res = res + fill(" " , lvl * 3) + ']~n'.

   return res.
end.

function list-menu returns char private (input h as handle).
   def var res as char.
   res = "".
   
   res = res + "~{~n".
   res = res + fill(" ", 3) + '"sensitive" : ' + string(h:sensitive, "true/false") + ",~n".
   res = res + fill(" ", 3) + '"visible" : ' + string(h:visible, "true/false") + ",~n".
   res = res + list-menu-tree(h, 1).
   res = res + "}~n".
   
   return res.
end.

function list-toolbar-button returns char private (input h as handle, input hx as int, input hy as int).
   def var res as char.
   def var x as int.
   def var y as int.
   
   if not h:visible or (not h:frame:visible and not isPageFrame(h:frame)) then return "".
   
   create page-toolbar.
   page-toolbar.pname = lastWindow.
   page-toolbar.pnum = lastPage.
   page-toolbar.button = h.
   
   x = h:x + hx.
   y = h:y + hy.
   
   res = "".
   
   res = res + "   " + "~{~n".
   res = res + "   " + '"handle": "'  + string(h) + '",~n'.
   if h:name <> ? then do:
      res = res + "      " + '"name": "'   + h:name + '",~n'.
   end.
   if h:label <> ? then do:
      res = res + "      " + '"label": "'   + h:label + '",~n'.
   end.
   if h:tooltip <> ? then do:
      res = res + "      " + '"tooltip": "'   + h:tooltip + '",~n'.
   end.
   if h:image-up <> "" and h:image-up <> ? then do:
      res = res + "      " + '"image": "'   + h:image-up + '",~n'.
   end.
   res = res + "      " + '"sensitive":' + string(h:sensitive, "true/false") + ',~n'.
   res = res + "      " + '"width":'     + string(h:width-pixels, ">>>>9") + ',~n'.
   res = res + "      " + '"height":'    + string(h:height-pixels, ">>>>9") + ',~n'.
   res = res + "      " + '"x":'         + string(x, ">>>>>9") + ',~n'.
   res = res + "      " + '"y":'         + string(y, ">>>>>9") + '~n'.
   res = res + "   " + "},~n".
   
   return res.
end.

function list-toolbar-buttons returns char private (input h as handle, input hx as int, input hy as int).
   def var res as char.
   def var hch as handle.
   
   res = "".
   
   hch = h:first-child.
   
   do while hch <> ?:
      if can-query(hch, "first-child") and hch:first-child <> ? then do:
         res = res + list-toolbar-buttons(hch, hx + hch:x, hy + h:y).
      end.
      else if hch:type = "BUTTON" then do:
         res = res + list-toolbar-button(hch, hx, hy).
      end.
      
      hch = hch:next-sibling.
   end.
   
   return res.
end.

function list-toolbar returns char private (input h as handle).
   def var res as char.
   def var hwid as handle.
   def var hpx as int.
   def var hpy as int.
   
   if not valid-handle(h) then return "".
   
   res = "".
   
   res = res + "[~n".
   
   res = res + list-toolbar-buttons(h, 0, 0).
   
   if h:parent:type = "WINDOW" then do:
      hpx = 0.
      hpy = 0.
   end.
   else do:
      hpx = h:parent:parent:x.
      hpy = h:parent:parent:y.
   end.
   
   hwid = h:parent:first-child.
   do while hwid <> ?:
      if hwid:type = "FRAME" and hwid <> h and hwid:y = h:y
      then do:
         res = res + list-toolbar-buttons(hwid, hpx + hwid:x, hpy + hwid:y).
      end.
      
      hwid = hwid:next-sibling.
   end.

   res = trim(res).
   if r-index(res, ",") = length(res)
   then do:
      res = substring(res, 1, length(res) - 1).
   end.

   res = res + "]~n".
   
   return res.
end.

function is-additional-toolbar returns log private (input h as handle, input hch as handle).
   def var hwid as handle.
   
   if not valid-handle(h) or hch:y <> h:y then return false.
   
   hwid = h:parent:first-child.
   do while hwid <> ?:
      if hwid:type = "FRAME" and hwid <> h and hwid = hch and hwid:y = h:y
        then return true.
      
      hwid = hwid:next-sibling.
   end.
   
   return false.
end.

function has-widget returns log private (input hparent as handle, input hchild as handle).
   def var hch as handle.
   
   hch = hparent:first-child.
   
   do while hch <> ?:
      if hch = hchild then return true.
   
      if can-query(hch, "first-child") and hch:first-child <> ? then do:
         if has-widget(hch, hchild) then return true.
      end.
      
      hch = hch:next-sibling.
   end.
   
   return false.
end.
