/*
** Module   : Aast.java
** Abstract : extension of AST interface to support post-processing features
**
** Copyright (c) 2005-2018, Golden Code Development Corporation.
**
** -#- -I- --Date-- --JPRM-- ----------------------------Description-----------------------------
** 001 ECF 20050218   @19888 Created initial version.
** 002 ECF 20050221   @19894 Added annotationKeys() method. Exposes keys
**                           for all annotations (if any) in this AST.
** 003 ECF 20050228   @20098 Added remove() method. Removes this node (and
**                           all descendant nodes) from containing tree.
** 004 GES 20050228   @20110 Added some helper methods to determine how
**                           many descendants (immediate and all) an AST
**                           has. Also added a token type symbol lookup.
** 005 ECF 20050303   @20160 Added two flavors of an iterator method. One
**                           walks the entire tree from the current node;
**                           the other walks only to a predetermined
**                           depth.
** 006 ECF 20050316   @20416 Changed iterator method signature to accept
**                           an additional argument. The version of this
**                           method which accepts arguments now accepts an
**                           AstWalkListener in addition to the number of
**                           nesting levels to walk.
** 007 GES 20050318   @20433 Added dumpTree() and getAncestor() methods.
** 008 GES 20050323   @20475 Added duplicate() method.
** 009 GES 20050323   @20478 Added token type/name conversion methods.
** 010 ECF 20050328   @20514 Added isRoot method.
** 011 GES 20050401   @20606 Added isAnnotation() method.
** 012 GES 20050406   @20652 Added support for array annotations.
** 013 ECF 20050408   @20657 Added graftCopyTo method. Grafts a copy of
**                           the current AST as a child of another AST,
**                           optionally pruning unwanted nodes from the
**                           copy.
** 014 ECF 20050408   @20659 Added setParent method.
** 015 GES 20050411   @20668 Added getImmediateChild() method to search
**                           for a child of a specific token type.
** 016 GES 20050428   @20925 Added support for getting the number of
**                           immediate children that have a specific
**                           token type.
** 017 GES 20050501   @20954 Added a variant of getAncestor() which
**                           handles a search up the tree for an ancestor
**                           matching a specified token type.
** 018 ECF 20050603   @21402 Added a variant of getImmediateChild() which
**                           uses a collection of types for its match.
** 019 GES 20050603   @21422 Added getChildAt() method for index based
**                           child lookups.
** 020 GES 20050620   @21524 Added graft() method which provides direct
**                           access to the part of the graftCopyTo()
**                           which is useful for adding a subtree without
**                           creating a copy first.
** 021 GES 20050627   @21553 Added graftAt() method which provides a way
**                           of inserting a child node at a specific
**                           child index.
** 022 GES 20050728   @21894 Added hidden mode support which changes
**                           iterator processing when requested.
** 023 ECF 20050817   @22157 Added a variant of putAnnotation() which adds
**                           a list of annotations at once.
** 024 NVS 20050825   @22248 Added an annotationCount() method that counts
**                           annotations in an array that equal an object.
** 025 GES 20050921   @22781 Added nearestEnclosing() method to search
**                           for a common ancestor node given 2 other
**                           nodes in the same tree.  Added
**                           getIndexPos() and getPrevSibling() to
**                           provide easy access to siblings (these
**                           compliment getChildAt().
** 026 GES 20050928   @22869 Added move() to ease sorting and moving
**                           nodes in a tree or between trees.
** 027 GES 20051014   @23047 Added clearIds().
** 028 GES 20051027   @23129 Added getLastImmediateChild().
** 029 GES 20051113   @23320 Added ancestor/descendant methods.
** 030 SIY 20051124   @23500 Added compareWith() method.
** 031 ECF 20060110   @23868 Added new variants of compareWith() and
**                           getAncestor(). New compareWith makes case
**                           sensitivity of text comparison optional. New
**                           getAncestor considers annotation values as
**                           well as type.
** 032 GES 20060129   @24135 Added new variant of duplicate and new
**                           convenience methods duplicateFresh.
** 033 GES 20060405   @25381 Added clearAnnotations().
** 034 ECF 20060615   @27352 Added isAncestorOf(). Tests whether this node
**                           is another's ancestor.
** 035 ECF 20060818   @28687 Added toStringVerbose(). Primarily for
**                           debugging purposes.
** 036 GES 20070531   @33948 Moved useful helper methods here from
**                           common AST support so that the code could
**                           be used in a wider variety of situations.
** 037 GES 20070629   @34332 Added setters for the line and column.
** 038 GES 20070629   @34342 Additional signature for graphCopyTo().
** 039 GES 20070720   @34646 Added getDescriptiveTokenText().     
** 040 SIY 20090119   @41180 Added commonParent() method.
** 041 SIY 20090126   @41217 Merged methods commonParent() and  
**                           nearestEnclosing().
** 042 GES 20090429   @42054 Match package and class name changes.
** 043 GES 20090515   @42224 Moved to AstManager from AstRegistry/AstPersister.
** 044 GES 20090518   @42366 Package change.
** 045 GES 20100927          Added support for tree-local data and shadow nodes.
** 046 ECF 20140404          Implemented generics.
** 047 OM  20150122          Added putAnnotation(Object) used by runtime evaluation.
** 048 OM  20150210          Added support for settable annotation output.
** 049 ECF 20150413          Added getAnnotationList method.
** 050 ECF 20160907          Added intern method.
** 051 ECF 20180409          Added matchAnnotation, new variants of descendant.
*/
/*
** This program is free software: you can redistribute it and/or modify
** it under the terms of the GNU Affero General Public License as
** published by the Free Software Foundation, either version 3 of the
** License, or (at your option) any later version.
**
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
** GNU Affero General Public License for more details.
**
** You may find a copy of the GNU Affero GPL version 3 at the following
** location: https://www.gnu.org/licenses/agpl-3.0.en.html
** 
** Additional terms under GNU Affero GPL version 3 section 7:
** 
**   Under Section 7 of the GNU Affero GPL version 3, the following additional
**   terms apply to the works covered under the License.  These additional terms
**   are non-permissive additional terms allowed under Section 7 of the GNU
**   Affero GPL version 3 and may not be removed by you.
** 
**   0. Attribution Requirement.
** 
**     You must preserve all legal notices or author attributions in the covered
**     work or Appropriate Legal Notices displayed by works containing the covered
**     work.  You may not remove from the covered work any author or developer
**     credit already included within the covered work.
** 
**   1. No License To Use Trademarks.
** 
**     This license does not grant any license or rights to use the trademarks
**     Golden Code, FWD, any Golden Code or FWD logo, or any other trademarks
**     of Golden Code Development Corporation. You are not authorized to use the
**     name Golden Code, FWD, or the names of any author or contributor, for
**     publicity purposes without written authorization.
** 
**   2. No Misrepresentation of Affiliation.
** 
**     You may not represent yourself as Golden Code Development Corporation or FWD.
** 
**     You may not represent yourself for publicity purposes as associated with
**     Golden Code Development Corporation, FWD, or any author or contributor to
**     the covered work, without written authorization.
** 
**   3. No Misrepresentation of Source or Origin.
** 
**     You may not represent the covered work as solely your work.  All modified
**     versions of the covered work must be marked in a reasonable way to make it
**     clear that the modified work is not originating from Golden Code Development
**     Corporation or FWD.  All modified versions must contain the notices of
**     attribution required in this license.
*/

package com.goldencode.ast;

import antlr.collections.AST;
import java.util.*;

/**
 * Implementations of this interface support additional features required by
 * the pattern recognition engine. These include:
 * <ul>
 * <li>a persistent, unique identifier;
 * <li>a link to a parent node, to permit leaf-to-root path traversal;
 * <li>an API to manage arbitrary annotations used in post-processing;
 * <li>access to source file name, line number and column, and nesting level
 *     information.
 * </ul>
 */
public interface Aast
extends AST
{
   /**
    * Set the source file line number for this node.
    *
    * @param    line
    *           New line number to associate with this node.
    */
   public void setLine(int line);
   
   /**
    * Set the source file column number for this node.
    *
    * @param    column
    *           New column number to associate with this node.
    */
   public void setColumn(int column);
   
   /**
    * Translates a text representation of a token type into the
    * actual integer value.
    *
    * @param    tokenName
    *           The text representation of the token type.
    *
    * @return   A valid integer token type value or -1 if no match was found.
    */
   public int lookupTokenType(String tokenName);
   
   /**
    * Translates an integer value of a token type into the associated 
    * human readable text representation.
    *
    * @param    type
    *           The integer token type.
    *
    * @return   A valid symbolic token type name or <code>null</code> if
    *           no match was found.
    */
   public String lookupTokenName(int type);
   
   /**
    * Grafts a copy of this AST node onto the specified parent AST. If a list
    * of valid token types is provided (i.e. <code>validTokens</code>), all
    * child nodes of the copy which <em>do not</em> match a valid token type
    * are removed from the copy node, before it is grafted onto the parent.
    * All necessary parent and ID fixups are performed on the copy to
    * integrate it with its new parent tree.
    *
    * @param   parent
    *          AST node which will become the direct parent of the grafted
    *          copy of this node.
    * @param   validTokens
    *          Optional array of valid token types. If not <code>null</code>,
    *          only those children of the copy node which match a valid token
    *          type are allowed to remain. All other child nodes are removed
    *          from the copy.
    * @param   preserve
    *          <code>true</code> to preserve line and column information in
    *          the duplicated tree.
    *
    * @return  The new AST node grafted into the tree.
    *
    * @throws  AstException
    *          if any error occurs registering the grafted node with the AST
    *          registry.
    */
   public Aast graftCopyTo(Aast parent, int[] validTokens, boolean preserve)
   throws AstException;
   
   /**
    * Grafts a copy of this AST node onto the specified parent AST. If a list
    * of valid token types is provided (i.e. <code>validTokens</code>), all
    * child nodes of the copy which <em>do not</em> match a valid token type
    * are removed from the copy node, before it is grafted onto the parent.
    *
    * @param   parent
    *          AST node which will become the direct parent of the grafted
    *          copy of this node.
    * @param   validTokens
    *          Optional array of valid token types. If not <code>null</code>,
    *          only those children of the copy node which match a valid token
    *          type are allowed to remain. All other child nodes are removed
    *          from the copy.
    *
    * @return  The new AST node grafted into the tree.
    *
    * @throws  AstException
    *          if any error occurs registering the grafted node with the AST
    *          registry.
    */
   public Aast graftCopyTo(Aast parent, int[] validTokens)
   throws AstException;
   
   /**
    * Grafts a new child subtree onto this AST. All necessary parent and ID 
    * fixups are performed to integrate the child tree with its new parent 
    * tree.
    *
    * @param   child
    *          AST subtree to add as a child of this node.
    *
    * @throws  AstException
    *          if any error occurs registering the grafted node with the AST
    *          registry.
    */
   public void graft(Aast child)
   throws AstException;
   
   /**
    * Grafts a new child subtree onto this AST at a specific 0-based child
    * index. All necessary parent and ID fixups are performed to integrate
    * the child tree with its new parent tree.
    *
    * @param   child
    *          AST subtree to add as a child of this node.
    * @param   index
    *          0 to insert as the first child, -1 to insert as the last
    *          child or a 0-based index to specify an exact child index
    *          at which to insert the new child.  All current children at
    *          that index or at a greater index, are moved right. If the
    *          index is greater than the current number of children, the
    *          new child will be added at the end.
    *
    * @throws  AstException
    *          if any error occurs registering the grafted node with the AST
    *          registry.
    */
   public void graftAt(Aast child, int index)
   throws AstException;
   
   /**
    * Creates a deep copy of the original, including all descendant ASTs. The
    * copy contains no cross-linked object references back to the original
    * AST. This method essentially results in recursive calls when
    * making copies of the AST's children.
    *
    * @return   The duplicated AST or <code>null</code> if there is any
    *           exception during the copy.
    */
   public Aast duplicate();
   
   /**
    * Creates a deep copy of the original, including all descendant ASTs. The
    * copy contains no cross-linked object references back to the original
    * AST. This method essentially results in recursive calls when
    * making copies of the AST's children. If an ID map is provided, map this
    * new node to its ID.
    *
    * @param    idMap
    *           Map of IDs to ASTs. May be <code>null</code>.
    *
    * @return   The duplicated AST.
    */   
   public Aast duplicate(Map<Long, Aast> idMap);
   
   /**
    * Creates a deep copy of the original, including all descendant ASTs. The
    * copy contains no cross-linked object references back to the original
    * AST. This method essentially results in recursive calls when
    * making copies of the AST's children.
    * <p>
    * All unique IDs of this new sub-tree are cleared before this method
    * returns (see {@link #clearIds}).
    *
    * @return   The duplicated AST or <code>null</code> if there is any
    *           exception during the copy.
    */
   public Aast duplicateFresh();
   
   /**
    * Creates a deep copy of the original, including all descendant ASTs. The
    * copy contains no cross-linked object references back to the original
    * AST. This method essentially results in recursive calls when
    * making copies of the AST's children. If an ID map is provided, map this
    * new node to its ID.
    * <p>
    * All unique IDs of this new sub-tree are cleared before this method
    * returns (see {@link #clearIds}).
    *
    * @param    idMap
    *           Map of IDs to ASTs. May be <code>null</code>.
    *
    * @return   The duplicated AST or <code>null</code> if there is any
    *           exception during the copy.
    */
   public Aast duplicateFresh(Map<Long, Aast> idMap);
   
   /**
    * Clears (sets to <code>null</code>) the unique ID of this node and all
    * descendant nodes using recursion.  Warning: this is very dangerous so
    * you have better know what you are doing!
    */
   public void clearIds();
   
   /**
    * Clear all IDs of an AST and register it under a new filename. The
    * AST is walked to fix-up all parent-child relationships as part of
    * the re-initialization.
    * <p>
    * This would be done in order to safely save a copy AST under a
    * different filename than the source AST. It is necessary in order to
    * avoid having two ASTs registered to the same filename and ID in the
    * AST registry.
    *
    * @param   filename
    *          New filename under which the AST is registered.
    *
    * @throws  AstException
    *          if any error occurs registering AST or applying fixups.
    */
   public void brainwash(String filename)
   throws AstException;
   
   /**
    * Compare subtree starting from <code>this</code> node as a root and
    * compare it with given subtree. Subtrees considered identical if both
    * contain identical nodes organized in the same order. Two nodes
    * considered identical if their type and text are identical. The order of
    * subtrees is defined by iterator provided by iterator() method.
    * <p>
    * The comparison of token text should be case sensitive.  For a case
    * insensitive comparison, use the {@link #compareWith(Aast,boolean)}
    * method variant.
    * 
    * @param   other
    *          The subtree to compare with.
    * 
    * @return  <code>true</code> if subtrees are identical and
    *          <code>false</code> otherwise.
    */
   public boolean compareWith(Aast other);
   
   /**
    * Compare subtree starting from <code>this</code> node as a root and
    * compare it with given subtree. Subtrees considered identical if both
    * contain identical nodes organized in the same order. Two nodes
    * considered identical if their type and text are identical. The order of
    * subtrees is defined by iterator provided by iterator() method.
    * 
    * @param   other
    *          The subtree to compare with.
    * @param   ignoreCase
    *          <code>true</code> to ignore case in the token text comparison;
    *          <code>false</code> for a case sensitive comparison.
    * 
    * @return  <code>true</code> if subtrees are identical and
    *          <code>false</code> otherwise.
    */
   public boolean compareWith(Aast other, boolean ignoreCase);
   
   /**
    * Retrieve this AST's unique, persistent identifier. An ID should be
    * assigned after AST creation but before external post-processing. Thus,
    * a <code>null</code> return from this method generally indicates an
    * invalid AST.
    *
    * @return  A persistent value that uniquely identifies this AST node, or
    *          <code>null</code> if no ID has been assigned.
    */
   public Long getId();
   
   /**
    * Set this AST's unique, persistent identifier. Once this value is set,
    * it is set "for life" and may not be changed.  An ID should be assigned
    * after AST creation but before external post-processing begins.
    * <p>
    * Normally, it would be better to set this value during construction.
    * However, we do not have control over the construction process when this
    * AST is instantiated by ANTLR during parsing, so it is done early in
    * post-processing instead.
    *
    * @param   id
    *          This AST's assigned ID.
    *
    * @return  <code>true</code> if the ID was set; <code>false</code> if it
    *          could not be set because it had previously been set.
    */
   public boolean setId(Long id);
   
   /**
    * Indicate whether this AST node represents the root of its hierarchy.
    *
    * @return   <code>true</code> if this node is the tree root, else
    *           <code>false</code>.
    */
   public boolean isRoot();
   
   /**
    * Indicate whether this AST node represents a leaf node in the tree (a
    * node with no children).
    *
    * @return   <code>true</code> if this node has no children, else
    *           <code>false</code>.
    */
   public boolean isLeaf();
   
   /**
    * Get this AST's parent node, if any.
    *
    * @return  Parent node, or <code>null</code> if this node represents the
    *          root of a hierarchy.
    */
   public Aast getParent();
   
   /**
    * Set this Aast's parent node.
    *
    * @param    parent
    *           Parent Aast node.
    */
   public void setParent(Aast parent);
   
   /**
    * Get this AST's hidden status.
    *
    * @return   <code>true</code> if this node should be hidden during
    *           iterator based walks (in which the iterator is specified to 
    *           honor the hidden mode).
    */
   public boolean isHidden();
   
   /**
    * Set this AST's hidden status.
    *
    * @param    hide
    *           <code>true</code> if this node should be hidden during
    *           iterator based walks (in which the iterator is specified to 
    *           honor the hidden mode).
    */
   public void setHidden(boolean hide);
   
   /**
    * Returns the index position of this instance in the array of children 
    * of its parent.  For example, if this instance is the first child of
    * its parent, then this method will return 0.  If this child is the
    * second child of its parent, this method will return 1, as so on.
    *
    * @return  The 0-based index position of this node OR 0 if this node has
    *          no parent (i.e. it is the tree root).
    */
   public int getIndexPos();
   
   /**
    * Returns the sibling to the left of this node in the tree or 
    * <code>null</code> if this instance is the first child of its parent.
    * <p>
    * This method is much slower than its counterpart 
    * <code>getNextSibling</code> since each node does not maintain a
    * reference back to the previous node.  Thus, the parent node must
    * be involved and this extra processing slows the resulting
    * implementation.
    *
    * @return  The previous sibling or <code>null</code> if this node is the
    *          first child (or has no parent).
    */
   public Aast getPrevSibling();
   
   /**
    * Get the ancestor node at the specified number of levels of ancestry,
    * if it exists.
    *
    * @param   level
    *          Number of generations up the tree to walk to obtain the 
    *          ancestor.  Set to <code>-1</code> to obtain the tree root, 
    *          <code>0</code> is the current node, <code>1</code> is the
    *          parent, <code>2</code> is the 'grandparent' and so forth.
    *
    * @return  The specified ancestor AST node if found or <code>null</code>
    *          if no such node exists (if <code>level</code> specified is 
    *          greater than depth of the tree).
    */
   public Aast getAncestor(int level);
   
   /**
    * Get the first ancestor node of a specified type found in a walk up to
    * the maximum specified number of levels of ancestry, if such a node
    * exists.
    *
    * @param   level
    *          Number of generations up the tree to walk to obtain the 
    *          ancestor.  Set to <code>-1</code> to obtain the tree root, 
    *          <code>0</code> is the current node, <code>1</code> is the
    *          parent, <code>2</code> is the 'grandparent' and so forth.
    * @param   type
    *          The token type of the ancestor to find.
    *
    * @return  The specified ancestor AST node if found or <code>null</code>
    *          if no such node exists (if <code>level</code> specified is 
    *          greater than depth of the tree).
    */
   public Aast getAncestor(int level, int type);
   
   /**
    * Get the first ancestor node of the specified type which also has a
    * particular annotation with the specified value.  If a <code>null</code>
    * value is provided, then the existence alone of an annotation with the
    * specified key is considered a match.  The search walks up the tree from
    * this node to the maximum specified number of levels of ancestry.
    *
    * @param   level
    *          Number of generations up the tree to walk to obtain the 
    *          ancestor.  Set to <code>-1</code> to obtain the tree root, 
    *          <code>0</code> is the current node, <code>1</code> is the
    *          parent, <code>2</code> is the 'grandparent' and so forth.
    * @param   type
    *          The token type of the ancestor to find.
    * @param   key
    *          Annotation key to match.
    * @param   value
    *          Annotation value to match.  <code>null</code> to match any
    *          node which has an annotation with the specified key,
    *          regardless of its value.
    *
    * @return  The specified ancestor AST node if found or <code>null</code>
    *          if no such node exists (if <code>level</code> specified is 
    *          greater than depth of the tree).
    */
   public Aast getAncestor(int level, int type, String key, Object value);
   
   /**
    * Searches the AST's ancestors for an exact token type match.  The
    * first match ends the search but in the case of no matches, all ancestor
    * nodes (up to the given maximum level) will have been searched.
    *
    * @param   level
    *          Maximum number of generations up the tree to search for an 
    *          ancestor.  Set to <code>-1</code> to search up to the tree 
    *          root, <code>0</code> for the current node, <code>1</code> for
    *          the parent, <code>2</code> for the 'grandparent' and so forth.
    * @param   type
    *          Token type to be matched.
    *
    * @return  <code>true</code> if <code>type</code> was found as a
    *          ancestor's token type within <code>level</code>
    *          generations, else <code>false</code>.
    */
   public boolean ancestor(int level, int type);
      
   /**
    * Searches the AST's ancestors for a token type that matches a given
    * range of token types (inclusively).  The first match ends the 
    * search but in the case of no matches, all ancestor nodes (up to the 
    * given maximum level) will have been searched.
    *
    * @param   level
    *          Maximum number of generations up the tree to search for an 
    *          ancestor.  Set to <code>-1</code> to search up to the tree 
    *          root, <code>0</code> for the current node, <code>1</code> for
    *          the parent, <code>2</code> for the 'grandparent' and so forth.
    * @param   lower
    *          Lower boundary of the token type range.
    * @param   upper
    *          Upper boundary of the token type range.
    *
    * @return  <code>true</code> if a ancestor's token type was found to
    *          match a range of types within <code>level</code> generations,
    *          else <code>false</code>.
    */
   public boolean ancestor(int level, int lower, int upper);
   
   /**
    * Searches the AST's ancestors for a token type that matches any of a
    * given list of token types.  The first match ends the search but in the 
    * case of no matches, all ancestor nodes (up to the given maximum level)
    * will have been searched.
    *
    * @param   level
    *          Maximum number of generations up the tree to search for an 
    *          ancestor.  Set to <code>-1</code> to search up to the tree 
    *          root, <code>0</code> for the current node, <code>1</code> for
    *          the parent, <code>2</code> for the 'grandparent' and so forth.
    * @param   types
    *          The list of token types for which to search.
    *
    * @return  <code>true</code> if a ancestor's token type was found to
    *          match any of the given list of types within <code>level</code>
    *          generations, else <code>false</code>.
    */
   public boolean ancestor(int level, int[] types);
   
   /**
    * Searches the AST's descendants for an exact token type match.  The
    * first match ends the search but in the case of no matches, all child
    * nodes (up to the given maximum level) will have been searched.
    *
    * @param   level
    *          Number of generations to search before abandoning the test.
    *          Set to <code>0</code> to search exhaustively.
    * @param   type
    *          Token type to be matched.
    *
    * @return  <code>true</code> if <code>type</code> was found as a
    *          descendant's token type within <code>level</code>
    *          generations, else <code>false</code>.
    */
   public boolean descendant(int level, int type);
      
   /**
    * Searches the AST's descendants for a token type that matches a given
    * range of token types (inclusively).  The first match ends the 
    * search but in the case of no matches, all child nodes (up to the given
    * maximum level) will have been searched.
    *
    * @param   level
    *          Number of generations to search before abandoning the test.
    *          Set to <code>0</code> to search exhaustively.
    * @param   lower
    *          Lower boundary of the token type range.
    * @param   upper
    *          Upper boundary of the token type range.
    *
    * @return  <code>true</code> if a descendant's token type was found to
    *          match a range of types within <code>level</code> generations,
    *          else <code>false</code>.
    */
   public boolean descendant(int level, int lower, int upper);
   
   /**
    * Searches the AST's descendants for a token type that matches a given range of token types
    * (inclusively), and for each found, matches the given value for an annotation.  The first
    * match ends the search but in the case of no matches, all child nodes (up to the given
    * maximum level) will have been searched.
    *
    * @param   level
    *          Number of generations to search before abandoning the test. Set to {@code 0} to
    *          search exhaustively.
    * @param   lower
    *          Lower boundary of the token type range.
    * @param   upper
    *          Upper boundary of the token type range.
    * @param   key
    *          Key of the annotation to match. If {@code null}, only the token types are used as
    *          match criteria.
    * @param   match
    *          Value of the annotation to match. Ignored if {@code key} is {@code null}. Must be
    *          the same data type as the annotation being searched.
    * @param   index
    *          Zero-based index of the annotation, if the annotation is part of a list.
    *          Set to {@code -1} to test all elements of an annotation list.
    *          Set to {@code null} for a scalar annotation.
    *
    * @return  {@code true} if a descendant's token type was found to match a range of types
    *          and an optional annotation was found to match within {@code level} generations,
    *          else {@code false}.
    */
   public boolean descendant(int level,
                             int lower,
                             int upper,
                             String key,
                             Object match,
                             Integer index);
   
   /**
    * Searches the AST's descendants for a token type that matches any of a
    * given list of token types.  The first match ends the search but in the 
    * case of no matches, all child nodes (up to the given maximum level)
    * will have been searched.
    *
    * @param   level
    *          Number of generations to search before abandoning the test.
    *          Set to <code>0</code> to search exhaustively.
    * @param   types
    *          The list of token types for which to search.
    *
    * @return  <code>true</code> if a descendant's token type was found to
    *          match any of the given list of types within <code>level</code>
    *          generations, else <code>false</code>.
    */
   public boolean descendant(int level, int[] types);
   
   /**
    * Searches the AST's descendants for a token type that matches any of a given list of token
    * types and for each found, matches the given value for an annotation.  The first match ends
    * the search but in the case of no matches, all child nodes (up to the given maximum level)
    * will have been searched.
    *
    * @param   level
    *          Number of generations to search before abandoning the test.
    *          Set to <code>0</code> to search exhaustively.
    * @param   types
    *          The list of token types for which to search.
    * @param   key
    *          Key of the annotation to match. If {@code null}, only the token types are used as
    *          match criteria.
    * @param   match
    *          Value of the annotation to match. Ignored if {@code key} is {@code null}. Must be
    *          the same data type as the annotation being searched.
    * @param   index
    *          Zero-based index of the annotation, if the annotation is part of a list.
    *          Set to {@code -1} to test all elements of an annotation list.
    *          Set to {@code null} for a scalar annotation.
    *
    * @return  {@code true} if a descendant's token type was found to match any of the given
    *          list of types and an optional annotation was matched within {@code level}
    *          generations, else {@code false}.
    */
   public boolean descendant(int level, int[] types, String key, Object match, Integer index);
   
   /**
    * Indicate whether this AST has an annotation which matches the specified conditions.
    * <p>
    * The match succeeds if an annotation exists with the given {@code key} and {@code match}
    * value. If {@code index} is not {@code null}, the annotation must exist in an annotation
    * list and the annotation element at the given {@code index} must match the given match
    * value.
    * <p>
    * The match fails under the following conditions without raising an error:
    * <ul>
    * <li>No annotation is found with the given {@code key};</li>
    * <li>An annotation is found with the given {@code key}, but it does not match the given
    *     {@code value};</li>
    * <li>{@code index} is a positive value that is out-of-bounds for the annotation list;</li>
    * <li>{@code index} is not {@code null} and the annotation is scalar.</li>
    * </ul>
    * 
    * @param   key
    *          The key under which to check for an annotation. May not be {@code null}.
    * @param   match
    *          An object to match as the annotation's value. Must be the same data type as the
    *          annotation, such that a comparison using {@code equals} will determine whether
    *          the annotation matches. May not be {@code null}.
    * @param   index
    *          Non-negative, zero-based index of the annotation, if it is part of a list.
    *          Set to {@code null} for a scalar annotation.
    *          Set to {@code -1} to check all elements of a list for a match.
    * 
    * @return  {@code true} if a matching annotation exists, else {@code false}.
    */
   public boolean matchAnnotation(String key, Object match, Integer index);
   
   /**
    * Test whether any of this node's siblings has the given token type.
    *
    * @param    type
    *           Token type to match.
    *
    * @return   <code>true</code> if any of this node's siblings match, else
    *           <code>false</code>.
    */
   public boolean sibling(long type);
   
   /**
    * Returns the number of siblings to this node's left (before this node
    * in a tree walk).  This is a synonym for {@link #getIndexPos}.
    *
    * @return   The number of left siblings or 0 if this node is the first
    *           child or has no parent (i.e. it is the tree root).
    */
   public int getNumLeftSiblings();
   
   /**
    * Returns the number of siblings to this node's right (after this node
    * in a tree walk).
    *
    * @return   The number of right siblings or 0 if this node is the last
    *           child or has no parent (i.e. it is the tree root).
    */
   public int getNumRightSiblings();
   
   /**
    * Report whether this node is an ancestor of the given node.
    *
    * @param   level
    *          The number of generations to search before abandoning the test.
    *          Set to <code>0</code> to search exhaustively.
    * @param   other
    *          The node whose ancestry will be tested.
    *
    * @return  <code>true</code> if this node is an ancestor of
    *          <code>other</code>.
    */
   public boolean isAncestorOf(int level, Aast other);
   
   /**
    * Compares this node's complete path from the root node (down to and
    * including the current node) with the specified path.
    *
    * @param    path
    *           Root-to-current path, formatted as described in 
    *           {@link #getPath}.
    *
    * @return   <code>true</code> if specified path matches this node's path
    *           exactly, else <code>false</code>.
    */
   public boolean absolutePath(String path);
   
   /**
    * Determines if this node's path from the root <b>ends with</b> the
    * specified path, on a case-insensitive basis. 
    * <p>
    * This node's type IS included in this path, which is different from the
    * {@link #upPath} which compares paths upward starting at the parent
    * node.
    *
    * @param    path
    *           Partial path from this node up to an arbitrary number of
    *           parent nodes, formatted as described in {@link #getPath}.
    *
    * @return   <code>true</code> if specified path matches the end of this 
    *           node's path exactly, else <code>false</code>.
    */
   public boolean relativePath(String path);
   
   /**
    * Determines if this node's path upward <b>matches</b> the specified
    * path, on a case-insensitive basis. The format of the 'segments' of the
    * up path starts with the parent's type and each subsequent segment must
    * be the type of the next parent up in the hierarchy.
    * <p>
    * The type of the current node is NOT included in the up path.  This
    * is the only difference between this method and {@link #relativePath}.
    *
    * @param    path
    *           Partial path from this node's parent upward to an arbitrary 
    *           number of ancestor nodes, formatted as described in
    *           {@link #getPath}.
    *
    * @return   <code>true</code> if specified path exactly matches the 
    *           upward path from this node's parent, else <code>false</code>.
    */
   public boolean upPath(String path);
   
   /**
    * Determines if this node's path downward <b>matches</b> the specified
    * path, on a case-insensitive basis. More than 1 down path may match this
    * specification since sibling nodes can have the same token type. If any
    * such match exists then this function will return <code>true</code>.
    * <p>
    * The format of the 'segments' of the down path starts with the
    * immediate child's type and each subsequent segment must be the type of
    * the next child down in the hierarchy.  The type of the current node is
    * NOT included in the down path.
    *
    * @param    path
    *           Partial path from this node down to an arbitrary number
    *           of descendant nodes, formatted as described in {@link #getPath}.
    *
    * @return   <code>true</code> if specified path exactly matches the 
    *           downward path from this node, else <code>false</code>.
    */
   public boolean downPath(String path);
   
   /**
    * Find and return the nearest ancestor of the current instance which also
    * encloses the node passed as a parameter.  By definition, if both nodes
    * exist in the same tree, there must be a common ancestor.  The only
    * exception to this rule is if one of the two nodes encloses the other
    * (e.g. it is the tree root or is itself the ancestor of the other node).
    * In this case, the enclosing node itself will be returned. If both
    * references are to the same node, then the parent node will be returned.
    *
    * @param    other
    *           Another reference node in the same tree.
    *
    * @return   The nearest enclosing ancestor or <code>null</code> if
    *           the two AST nodes are not in the same tree.
    */
   public Aast nearestEnclosing(Aast other);
   
   /**
    * Search the immediate children for one that matches a specific token 
    * type.  An optional starting child AST can be provided which allows 
    * a search to be made for matches subsequent to the first, if the start
    * AST is <code>null</code>, then the search will start with the first
    * child.
    *
    * @param    type
    *           The token type of the child for which to search.
    * @param    start
    *           The child from which to start the search or <code>null</code>
    *           if the search is to start with the first child.
    *
    * @return   The first child (to the right of the starting point) that
    *           matches or <code>null</code> if no match was found.
    */
   public Aast getImmediateChild(int type, Aast start);
   
   /**
    * Search the immediate children for one that matches any token type in
    * the specified Collection of types.  The first match is returned, as
    * determined by the iteration order of the <code>types</code> collection.
    * An optional starting child AST can be provided which allows a search
    * to be made for matches subsequent to the first, if the start AST is
    * <code>null</code>, then the search will start with the first child.
    *
    * @param    types
    *           Collection of token types of the child for which to search.
    * @param    start
    *           The child from which to start the search or <code>null</code>
    *           if the search is to start with the first child.
    *
    * @return   The first child (to the right of the starting point) that
    *           matches or <code>null</code> if no match was found.
    */
   public Aast getImmediateChild(Collection<Integer> types, Aast start);
   
   /**
    * Search the immediate children for the one at the specified index. If no
    * such child exists, return <code>null</code>.
    *
    * @param    index
    *           The 0-based index at which to search.
    *
    * @return   The child at the specified index or <code>null</code> if no
    *           such index exists.
    */
   public Aast getChildAt(int index);
   
   /**
    * Remove this AST (and the entire sub-tree rooted at this node) from its
    * parent's collection of children. Note that the implication of such
    * removal is that all descendants of this node remain parented by this
    * sub-tree but are removed from the original parent tree along with this
    * node.
    * <p>
    * This method has no effect if invoked on the root node of an AST.
    */
   public void remove();
   
   /**
    * Move the node from one location to another.  The move can occur between
    * index positions under a single parent, between 2 different parent nodes
    * in the same tree or between 2 different parent nodes in different
    * trees.  The type of the <code>Aast</code> nodes must be the same
    * between the current node and the parent node, otherwise the results 
    * will be undefined.
    * <p>
    * If the given parent node is <code>null</code> and the current parent
    * node of this instance is <code>null</code>, this method returns
    * immediately since there is nothing to do.
    * 
    * @param    parent
    *           The parent node which is the target of the move operation or
    *           <code>null</code> if the move is only affecting the index
    *           position of this node (a move of the child to a different
    *           index in the same parent).
    * @param    index
    *           0 to insert as the first child, -1 to insert as the last
    *           child or a 0-based index to specify an exact child index
    *           at which to insert the new child.  All current nodes at
    *           that index or at a greater index, are moved right. If the
    *           index is greater than the current number of nodes, the node
    *           will be added at the end.
    */
   public void move(Aast parent, int index)
   throws AstException;
   
   /**
    * Get the level of nesting of this AST from the root AST, where
    * <code>0</code> indicates the root, and a positive number indicates
    * generations from the root.
    *
    * @return  Nesting level.
    */
   public int getDepth();
   
   /**
    * Get the number of immediate child nodes of this AST.
    *
    * @return  Number of immediate child nodes.
    */
   public int getNumImmediateChildren();
   
   /**
    * Get the number of immediate child nodes of this AST that match a
    * specific token type.
    *
    * @param    type
    *           The token type of the child nodes to search for.
    *
    * @return  Number of immediate child nodes which have the specified
    *          token type.
    */
   public int getNumImmediateChildren(int type);
   
   /**
    * Search the immediate children for last child that matches a specific 
    * token type.
    *
    * @param    type
    *           The token type of the child for which to search.
    *
    * @return   The last child (rightmost) that matches or <code>null</code>
    *           if no match was found.
    */
   public Aast getLastImmediateChild(int type);
   
   /**
    * Get the total number of all descendants of this AST.
    *
    * @return  Total number of descendants.
    */
   public int getNumChildren();
   
   /**
    * Get the name of the source file associated with this AST
    *
    * @return  Name of source file, or <code>null</code> if no name has yet
    *          been assigned.
    */
   public String getFilename();
   
   /**
    * Return a string representing the path of this AST from the root AST,
    * in terms of the token type of each element from the root to this AST.
    * The token types are separated by a forward slash and a trailing slash
    * ends the path. For example:
    * <pre>
    *    BLOCK/STATEMENT/KW_IF/
    * </pre>
    * represents the following path:
    * <pre>
    *    ROOT [type=BLOCK]
    *     |
    *    NODE A [type=STATEMENT]
    *     |
    *    NODE B [type=KW_IF]
    * </pre>
    * The string uniquely identifies the path from this node back to the root,
    * but is not necessarily unique when traversed from the root to this node.
    *
    * @return  A string which, when read from right to left, indicates the
    *          token type of each node from this AST to the root AST.
    */
   public String getPath();
   
   /**
    * Returns the symbolic representation of this AST's token type.
    *
    * @return  Symbolic token name for this AST's token type.
    */
   public String getSymbolicTokenType();
   
   /**
    * Returns a descriptive string for the current node.  This string will
    * normally contain both the text and the symbolic representation of this
    * AST's token type.
    *
    * @return   A descriptive string for the token type of this node.
    */
   public String getDescriptiveTokenText();
   
   /**
    * Create a human readable text representation of the tree rooted at this
    * AST. This implementation walks the subtree and dumps the text, token
    * type, line and column of each node, with indents to show the nesting 
    * and each node on its own line.
    *
    * @return  The tree's text representation.
    */
   public String dumpTree();
   
   /**
    * Create a human readable text representation of the tree rooted at this AST. This
    * implementation walks the subtree and dumps the text, token type, line and column of each
    * node, with indents to show the nesting and each node on its own line.
    *
    * @param   annotations
    *          Verbose mode. If {@code true} dump the annotations along with node name and type.
    *
    * @return  The tree's text representation.
    */
   public String dumpTree(boolean annotations);
   
   /**
    * Access a shadow node that exists to the left of the current node.
    *
    * @return   The shadow node that is left of the current node or 
    *           <code>null</code> if no such node exists.
    */
   public Aast getLeftShadowNode();
   
   /**
    * Access a shadow node that exists to the right of the current node.
    *
    * @return   The shadow node that is right of the current node or 
    *           <code>null</code> if no such node exists.
    */
   public Aast getRightShadowNode();
   
   /**
    * Access the list of all shadow nodes.
    *
    * @return   All shadow nodes or <code>null</code> if none exist.
    */
   public Iterator<ShadowNode> getShadowNodeList();
   
   /**
    * Access the tree-local data directly.  WARNING: this data is mutable so
    * make sure you really want to change it. If the tree is subsequently
    * persisted, any changes will be saved.
    *
    * @return   The tree-local data or <code>null</code> if there is no
    *           tree-local data.
    */
   public TreeLocal getLocal();
   
   /**
    * Set the tree-local data directly.  WARNING: this data is mutable so
    * make sure you really want to change it. If the tree is subsequently
    * persisted, any changes will be saved.
    *
    * @param    local
    *           The tree-local data or <code>null</code> if there is no
    *           tree-local data.
    */
   public void setLocal(TreeLocal local);
   
   /**
    * Return an iterator to walk this AST recursively in a depth-first
    * traversal. "Depth-first" means that the iteration will traverse the
    * children of any given node before traversing that node's siblings.
    * <p>
    * This iterator will not support any "non-visiting" events (descent,
    * ascent and next-child).  In addition, all nodes will be returned,
    * regardless of hidden status.
    *
    * @return  An iterator which will walk this node and all of its
    *          descendants.
    */
   public Iterator<Aast> iterator();
   
   /**
    * Return an iterator to walk this AST recursively in a depth-first
    * traversal. "Depth-first" means that the iteration will traverse the
    * children of any given node before traversing that node's siblings.
    * <p>
    * The iterator will walk at most to a node depth of <code>maxLevels</code>
    * relative to this node. A <code>maxLevels</code> value of <code>0</code>
    * indicates that the full tree beneath this node should be walked, down
    * to every leaf, regardless of its depth.
    * <p>
    * If a <code>listener</code> is specified, it will be notified when the
    * tree walk transitions between nesting levels. The notifications occur
    * during processing of the iterator's <code>next</code> method. In each
    * case (whether the transition represents a descent to a lower tree level
    * or an ascent to a higher tree level), the parent of the parent/child
    * link being traversed will be sent with the notification. No more than
    * a single descent notification will ever be sent during a single
    * invocation of the iterator's <code>next</code> method. However,
    * multiple ascent notifications may be sent, since the tree walk may
    * proceed directly from a deeply nested node to the next sibling of a
    * distant ancestor node. In such a case, one call to the listener's
    * {@link AstWalkListener#ascent} method will be made for each upward
    * step from child to parent, until an ancestor with an unvisited sibling
    * is found, or the walk ends.
    * <p>
    * This iterator will return all nodes in the tree, regardless of hidden
    * status.
    *
    * @param   maxLevels
    *          Maximum number of levels or generations to visit relative to
    *          the root defined by <code>ast</code>. If the tree contains
    *          more levels than <code>maxLevels</code>, the levels beyond a
    *          relative depth of <code>maxLevels</code> will be ignored. Set
    *          to <code>0</code> to disable this restriction. Set to
    *          <code>1</code> for immediate children only, etc.
    * @param   listener
    *          An optional listener which is notified of significant events
    *          during the tree walk, such as descents to lower nesting levels,
    *          moves between children (next-child events) and ascents back up
    *          to higher nesting levels. This may be <code>null</code> if
    *          these "non-visiting" events are not needed.
    *
    * @return  An iterator which will behave as described above.
    */
   public Iterator<Aast> iterator(int maxLevels, AstWalkListener listener);
   
   /**
    * Return an iterator to walk this AST recursively in a depth-first
    * traversal. "Depth-first" means that the iteration will traverse the
    * children of any given node before traversing that node's siblings.
    * <p>
    * The iterator will walk at most to a node depth of <code>maxLevels</code>
    * relative to this node. A <code>maxLevels</code> value of <code>0</code>
    * indicates that the full tree beneath this node should be walked, down
    * to every leaf, regardless of its depth.
    * <p>
    * If a <code>listener</code> is specified, it will be notified when the
    * tree walk transitions between nesting levels. The notifications occur
    * during processing of the iterator's <code>next</code> method. In each
    * case (whether the transition represents a descent to a lower tree level
    * or an ascent to a higher tree level), the parent of the parent/child
    * link being traversed will be sent with the notification. No more than
    * a single descent notification will ever be sent during a single
    * invocation of the iterator's <code>next</code> method. However,
    * multiple ascent notifications may be sent, since the tree walk may
    * proceed directly from a deeply nested node to the next sibling of a
    * distant ancestor node. In such a case, one call to the listener's
    * {@link AstWalkListener#ascent} method will be made for each upward
    * step from child to parent, until an ancestor with an unvisited sibling
    * is found, or the walk ends.
    * <p>
    * Optionally, the returned iterator can omit nodes in which the hidden
    * status is enabled (see {@link #isHidden} and {@link #setHidden}).
    * Please note that if the root node (the instance upon which this method
    * is called) is set as <code>hidden</code>, this setting will
    * <code>not</code> be honored.  The <code>hidden</code> mode is only
    * honored on child nodes that have this mode set.
    *
    * @param   maxLevels
    *          Maximum number of levels or generations to visit relative to
    *          the root defined by <code>ast</code>. If the tree contains
    *          more levels than <code>maxLevels</code>, the levels beyond a
    *          relative depth of <code>maxLevels</code> will be ignored. Set
    *          to <code>0</code> to disable this restriction. Set to
    *          <code>1</code> for immediate children only, etc.
    * @param   listener
    *          An optional listener which is notified of significant events
    *          during the tree walk, such as descents to lower nesting levels,
    *          moves between children (next-child events) and ascents back up
    *          to higher nesting levels. This may be <code>null</code> if
    *          these "non-visiting" events are not needed.
    * @param   honorHidden
    *          If <code>true</code>, the <code>hidden</code> attribute of
    *          the node will be honored.  This means that the node will not
    *          appear in the iterator nor will any "non-visiting" events be
    *          triggered for such nodes.
    *
    * @return  An iterator which will behave as described above.
    */
   public Iterator<Aast> iterator(int             maxLevels,
                                  AstWalkListener listener,
                                  boolean         honorHidden);
   
   /**
    * Determines if an annotation exists of any <code>String</code>,
    * <code>Double</code>, <code>Long</code>, <code>Boolean</code> or
    * <code>List</code> value, indexed by an arbitrary key.
    *
    * @param   key
    *          Key under which the annotation is stored.
    *
    * @return  <code>true</code> if an annotation exists mapped by that key.
    */
   public boolean isAnnotation(String key);   
   
   /**
    * Returns the size of an array annotation as indexed by an arbitrary key.
    *
    * @param   key
    *          Key under which the annotation is stored.
    *
    * @return  The number of elements in an array annotation that exists
    *          mapped by that key or 0 if no such named array annotation
    *          exists.  A scalar annotation using that same name will also
    *          result in a return of 0.
    */
   public int annotationSize(String key);   
   
   /**
    * Returns the number of annotation objects of an array annotation that are
    * equal to the specified <code>Object</code>.
    *
    * @param   key
    *          Key under which the annotation is stored.
    *
    * @param   value
    *          <code>Object</code> to compare with.
    *
    * @return  The number of elements in an array annotation that equals
    *          the specified value or 0 if no such named array annotation
    *          exists.  A scalar annotation using that same name will also
    *          result in a return of 0.
    */
   public int annotationCount(String key, Object value);   

   /**
    * Retrieve an annotation. This may be any <code>String</code>,
    * <code>Double</code>, <code>Long</code>, or <code>Boolean</code> value,
    * indexed by an arbitrary key.  Note that it is possible to obtain the
    * entire array if the annotation is an array type.  This would be a
    * <code>List</code> of objects of the above mentioned types.
    *
    * @param   key
    *          Key under which the annotation is stored.
    *
    * @return  The annotation object stored under the specified key, or
    *          <code>null</code> if none exists.
    */
   public Object getAnnotation(String key);
   
   /**
    * Retrieve an element in an array annotation using the specified key as
    * a unique identifier and an index to specify the location in the array
    * from which to read the element.  The special index of -1 indicates the
    * last element in the array. The resulting data may be any
    * <code>String</code>, <code>Double</code>, <code>Long</code>, or
    * <code>Boolean</code> value.
    *
    * @param   key
    *          Key under which the annotation is stored.
    * @param   index
    *          The index of -1 indicates the last element of the array.
    *          Otherwise this is the explicit zero-based index at which to 
    *          find the annotation.
    *
    * @return  The annotation object stored under the specified key and at
    *          the specified index, or <code>null</code> if the annotation
    *          doesn't exist, if the annotation is not an array or if no
    *          element exists at the given index.
    */
   public Object getAnnotation(String key, int index);
   
   /**
    * Retrieve an unmodifiable list view of an array annotation.
    * 
    * @param   key
    *          Key under which the annotation is stored.
    * 
    * @return  Unmodifiable list view of the array annotation, or <code>null</code> if no such
    *          annotation exists.
    */
   public List<Object> getAnnotationList(String key);
   
   /**
    * Store a text annotation using the specified key as a unique identifier.
    *
    * @param   key
    *          Key under which to store the annotation.
    * @param   annotation
    *          Arbitrary text.
    */
   public void putAnnotation(String key, String annotation);
   
   /**
    * Store a text element in an array annotation using the specified key as
    * a unique identifier and an index to specify the location in which to
    * insert.  If an element already exists at that index, it and all
    * following elements will be located at indices that are 1 greater than
    * before.  The special index of -1 indicates an addition to the end of
    * the array.
    *
    * @param   key
    *          Key under which to store the annotation.
    * @param   index
    *          The index of -1 indicates an addition to the end of the array.
    *          Otherwise this is the explicit zero-based index at which to 
    *          add the annotation.
    * @param   annotation
    *          Arbitrary text.
    */
   public void putAnnotation(String key, String annotation, int index);
   
   /**
    * Store a floating point numeric annotation using the specified key as a
    * unique identifier.
    *
    * @param   key
    *          Key under which to store the annotation.
    * @param   annotation
    *          Arbitrary floating point numeric value.
    */
   public void putAnnotation(String key, Double annotation);
   
   /**
    * Store a floating point numeric element in an array annotation using the
    * specified key as a unique identifier and an index to specify the
    * location in which to insert.  If an element already exists at that
    * index, it and all following elements will be located at indices that
    * are 1 greater than before.  The special index of -1 indicates an
    * addition to the end of the array.
    *
    * @param   key
    *          Key under which to store the annotation.
    * @param   index
    *          The index of -1 indicates an addition to the end of the array.
    *          Otherwise this is the explicit zero-based index at which to 
    *          add the annotation.
    * @param   annotation
    *          Arbitrary floating point numeric value.
    */
   public void putAnnotation(String key, Double annotation, int index);
   
   /**
    * Store an integral numeric annotation using the specified key as a
    * unique identifier.
    *
    * @param   key
    *          Key under which to store the annotation.
    * @param   annotation
    *          Arbitrary integral numeric value.
    */
   public void putAnnotation(String key, Long annotation);
   
   /**
    * Store an integral numeric element in an array annotation using the
    * specified key as a unique identifier and an index to specify the
    * location in which to insert.  If an element already exists at that
    * index, it and all following elements will be located at indices that
    * are 1 greater than before.  The special index of -1 indicates an
    * addition to the end of the array.
    *
    * @param   key
    *          Key under which to store the annotation.
    * @param   index
    *          The index of -1 indicates an addition to the end of the array.
    *          Otherwise this is the explicit zero-based index at which to 
    *          add the annotation.
    * @param   annotation
    *          Arbitrary integral numeric value.
    */
   public void putAnnotation(String key, Long annotation, int index);
   
   /**
    * Store a boolean annotation using the specified key as a unique
    * identifier.
    *
    * @param   key
    *          Key under which to store the annotation.
    * @param   annotation
    *          Arbitrary boolean value.
    */
   public void putAnnotation(String key, Boolean annotation);
   
   /**
    * Store a boolean element in an array annotation using the
    * specified key as a unique identifier and an index to specify the
    * location in which to insert.  If an element already exists at that
    * index, it and all following elements will be located at indices that
    * are 1 greater than before.  The special index of -1 indicates an
    * addition to the end of the array.
    *
    * @param   key
    *          Key under which to store the annotation.
    * @param   index
    *          The index of -1 indicates an addition to the end of the array.
    *          Otherwise this is the explicit zero-based index at which to 
    *          add the annotation.
    * @param   annotation
    *          Arbitrary boolean value.
    */
   public void putAnnotation(String key, Boolean annotation, int index);
   
   /**
    * Store a list of annotations at once using the specified key as a unique
    * identifier.
    * The list should be made of objects of the same type, which, in turn,
    * should be one of the supported annotation types:
    * <ul>
    *    <li><code>String</code>
    *    <li><code>Double</code>
    *    <li><code>Long</code>
    *    <li><code>Boolean</code>
    * </ul>
    *
    * @param   key
    *          Key under which to store the annotation.
    * @param   annotation
    *          list of annotations of the same, supported type.
    */
   public void putAnnotation(String key, List<Object> annotation);
   
   /**
    * Store any object as annotation for this node. It will work only with 
    * <code>InMemoryRegistryPlugin</code> since this is the only plugin that does not truly
    * serialize the nodes.
    * 
    * @param   key
    *          The name of the annotation.
    *          
    * @param   annotation
    *          The value for the annotation. Any type of object can be stored as long it is not
    *          required to be saved.
    */
   public void putAnnotationObject(String key, Object annotation);
   
   /**
    * Remove a specific annotation from this AST.  Note that this annotation
    * may be an entire array (of type <code>List</code>) or it may be a
    * scalar value.
    *
    * @param   key
    *          Key under which the annotation is stored.
    *
    * @return  The annotation which was removed, or <code>null</code> if none
    *          was found.
    */
   public Object removeAnnotation(String key);
   
   /**
    * Remove a specific element from the array annotation in this AST.
    * The special index of -1 indicates the removal of the last element of
    * the array.
    *
    * @param   key
    *          Key under which the annotation is stored.
    * @param   index
    *          The index of -1 indicates an addition to the end of the array.
    *          Otherwise this is the explicit zero-based index at which to 
    *          add the annotation.
    *
    * @return  The element which was removed, or <code>null</code> if the
    *          annotation doesn't exist, if the annotation is not an array or
    *          if no element exists at the given index.
    */
   public Object removeAnnotation(String key, int index);
   
   /**
    * Removes ALL annotations from this AST.  BE VERY CAREFUL WITH THIS! 
    * There is no recovery once this is called.
    */
   public void clearAnnotations();
   
   /**
    * Get an iterator on the keys for all annotations (if any) stored in this
    * AST.
    *
    * @return  An iterator on all annotation keys. May be empty, but will not
    *          be <code>null</code>.
    */
   public Iterator<String> annotationKeys();
   
   /**
    * Walk the tree formed by this node and all of its descendants, interning all strings stored
    * in the AST's text and annotation values which are {@code String} objects.
    * <p>
    * Since all these objects must be retrieved and re-stored, this method will be expensive for
    * a large tree. It should only be invoked on trees which will remain in long term memory.
    */
   public void intern();
   
   /**
    * Create a verbose string representation of this AST.  This includes the
    * the token name associated with this AST's token type, its text, its
    * line and column number, and its ID if non-null.
    *
    * @return  Verbose description of this node.
    */
   public String toStringVerbose();
}