/*
** Module   : LFUAgingCache.java
** Abstract : LFU cache implementation which adjusts element importance by age of last access
**
** Copyright (c) 2014-2017, Golden Code Development Corporation.
**
** -#- -I- --Date-- --------------------------------Description----------------------------------
** 001 ECF 20140922 Created initial version.
** 002 ECF 20150715 Minor change to suppress compiler warning.
** 003 IAS 20160331 Fixed NodeImpl.equals().
*/
/*
** This program is free software: you can redistribute it and/or modify
** it under the terms of the GNU Affero General Public License as
** published by the Free Software Foundation, either version 3 of the
** License, or (at your option) any later version.
**
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
** GNU Affero General Public License for more details.
**
** You may find a copy of the GNU Affero GPL version 3 at the following
** location: https://www.gnu.org/licenses/agpl-3.0.en.html
** 
** Additional terms under GNU Affero GPL version 3 section 7:
** 
**   Under Section 7 of the GNU Affero GPL version 3, the following additional
**   terms apply to the works covered under the License.  These additional terms
**   are non-permissive additional terms allowed under Section 7 of the GNU
**   Affero GPL version 3 and may not be removed by you.
** 
**   0. Attribution Requirement.
** 
**     You must preserve all legal notices or author attributions in the covered
**     work or Appropriate Legal Notices displayed by works containing the covered
**     work.  You may not remove from the covered work any author or developer
**     credit already included within the covered work.
** 
**   1. No License To Use Trademarks.
** 
**     This license does not grant any license or rights to use the trademarks
**     Golden Code, FWD, any Golden Code or FWD logo, or any other trademarks
**     of Golden Code Development Corporation. You are not authorized to use the
**     name Golden Code, FWD, or the names of any author or contributor, for
**     publicity purposes without written authorization.
** 
**   2. No Misrepresentation of Affiliation.
** 
**     You may not represent yourself as Golden Code Development Corporation or FWD.
** 
**     You may not represent yourself for publicity purposes as associated with
**     Golden Code Development Corporation, FWD, or any author or contributor to
**     the covered work, without written authorization.
** 
**   3. No Misrepresentation of Source or Origin.
** 
**     You may not represent the covered work as solely your work.  All modified
**     versions of the covered work must be marked in a reasonable way to make it
**     clear that the modified work is not originating from Golden Code Development
**     Corporation or FWD.  All modified versions must contain the notices of
**     attribution required in this license.
*/

package com.goldencode.cache;

import java.util.*;
import java.util.logging.*;
import com.goldencode.p2j.util.*;

/**
 * An implementation of a least-frequently-used (LFU) cache which additionally adjusts the
 * importance of individual elements, based on the age of their last access. This will result in
 * nodes which may have been used frequently earlier in the history of the cache, followed by
 * periods of inactivity, eventually being aged out of the cache.
 * <p>
 * Registered listeners are notified of each expiry event.
 * <p>
 * <strong>This implementation is not synchronized</strong>. Any access to the methods of this
 * class which modify the internal state of the cache must be synchronized externally.
 * 
 * @param   K
 *          Type of cache element key.
 * @param   V
 *          Type of cache element value.
 */
public class LFUAgingCache<K, V>
extends ExpiryCache<K, V>
{
   /** Portion of cache limit to be expired when the cache is full and capacity is needed */
   private static final double CLEAN_RATIO = 0.125;
   
   /**
    * Constructor.
    * 
    * @param   limit
    *          Maximum number of elements allowed in the cache.
    * 
    * @throws  IllegalArgumentException
    *          if the specified limit is less than one.
    */
   public LFUAgingCache(int limit)
   {
      super(limit);
   }
   
   /**
    * Constructor.
    * 
    * @param   name
    *          Cache name, for debug and logging purposes.
    * @param   limit
    *          Maximum number of elements allowed in the cache.
    * 
    * @throws  IllegalArgumentException
    *          if the specified limit is less than one.
    */
   public LFUAgingCache(String name, int limit)
   {
      super(name, limit);
   }
   
   /**
    * Get the number of elements which will be expired from the cache each time it becomes full,
    * and capacity is required for a new element.
    * 
    * @return  Number of elements to expire on each clearing pass, in this case 1/8 of the cache's
    *          maximum capacity.
    */
   protected int getCleanLimit()
   {
      return (int) Math.ceil(limit * CLEAN_RATIO);
   }
   
   /**
    * Create a node of the appropriate type for the given key and value.
    * 
    * @param   key
    *          Hashable key.
    * @param   value
    *          Cached value.
    * 
    * @return  A sortable node to store in the cache.
    */
   protected Node<K, V> createNode(K key, V value)
   {
      return new NodeImpl<K, V>(key, value);
   }
   
   /**
    * Ensure there is sufficient capacity to add a new record to the cache, by expiring the least
    * desirable element(s), according to the policy of the cache. This cache's policy retains
    * nodes which have been most frequently used, however, the relevance of a node which has been
    * accessed previously degrades over time.
    * <p>
    * When elements are expired from the cache, a cache expiry event is sent to all registered
    * listeners.
    */
   protected void ensureCapacity()
   {
      super.ensureCapacity();
      
      if (cache.size() > 1)
      {
         ageNodes();
      }
   }
   
   /**
    * Recompute the relevance of nodes with older access ages. This may result in older nodes
    * having their hit counts reduced, to compensate for lack of recent activity.
    */
   private void ageNodes()
   {
      // collect ages of each each node's last access
      int size = sorted.size();
      double[] ages = new double[size];
      long now = System.currentTimeMillis();
      double total = 0;
      int i = 0;
      for (Node<K, V> n : sorted)
      {
         long age = now - ((NodeImpl<K, V>) n).lastAccess;
         total += age;
         ages[i++] = age;
      }
      
      // calculate standard deviation of last access ages
      double mean = total / size;
      total = 0;
      
      for (i = 0; i < size; i++)
      {
         double dist = ages[i] - mean;
         total += (dist * dist);
      }
      double sigma = Math.sqrt(total / size);
      
      // if any node's access age is beyond one standard deviation from the mean, bump its hit
      // count down by one to decrease that node's relevance
      List<NodeImpl<K, V>> reinserts = new ArrayList<>();
      Iterator<Node<K, V>> iter = sorted.descendingIterator();
      for (i = size - 1; iter.hasNext(); i--)
      {
         NodeImpl<K, V> n = (NodeImpl<K, V>) iter.next();
         if (n.hits == 0)
         {
            // current node cannot be demoted further and all remaining nodes will have 0 hits
            break;
         }
         
         // decrement hit count to demote element if last access age is more than one standard
         // deviation older than mean age
         if (ages[i] - mean > sigma)
         {
            // remove, demote, and (later) reinsert node, to ensure new sort order
            iter.remove();
            n.hits--;
            reinserts.add(n);
         }
      }
      
      // debug logging
      boolean debug = log.isLoggable(Level.FINE);
      String msg = (debug
                    ? String.format("[%s] Cache '%s' aged %d nodes:",
                                    Utils.describeContext(), name, reinserts.size())
                    : null);
      
      // reinsert nodes previously removed, to force re-sort; done with separate list to prevent
      // concurrent modification exception
      if (!reinserts.isEmpty())
      {
         for (NodeImpl<K, V> n: reinserts)
         {
            if (debug)
            {
               msg += (" " + n.hits);
            }
            sorted.add(n);
         }
      }
      
      if (debug)
      {
         log.log(Level.FINE, msg);
      }
   }
   
   /**
    * A sortable node containing the hashable key and value of a cache entry. The node's natural
    * order is determined by the number of times this node has been retrieved from the cache
    * (its hit count), and by how recently the last retrieval was. If these values match those of
    * another node exactly, a unique identifier is used as an arbitrary but consistent tie
    * breaker. This ensures each node is unique and none are excluded as duplicates due to set
    * semantics. Less desirable elements (those to be expired first) are sorted ahead of more
    * desirable elements (those to be retained longer).
    * 
    * @param <K>
    *        Hashable key
    * @param <V>
    *        Cached value
    */
   private static class NodeImpl<K, V>
   extends Node<K, V>
   {
      /** Next available node identifier */
      private static int nextID = 0;
      
      /** Internal identifier */
      private final int id = nextID++;
      
      /** Hit count -- number of times this node has been retrieved from the cache */
      private int hits = -1;
      
      /** Timestamp of last time this element was accessed (created or retrieved) */
      private long lastAccess;
      
      /**
       * Constructor. Sets initial access time to now, and hit count to 0.
       * 
       * @param   key
       *          Hashable key.
       * @param   value
       *          Cached value.
       */
      NodeImpl(K key, V value)
      {
         super(key, value);
      }
      
      /**
       * Compare this node to another. A node with a higher hit count sorts higher than that with
       * a lower hit count. If two nodes have the same hit count, that which has more recently
       * been accessed (created or retrieved) sorts higher. If two nodes have the same hit count
       * and last access time (should be quite a rare occurrence), a unique ID will be used to
       * force an arbitrary sort order. This last comparison is ensure uniqueness among nodes so
       * that the set does not exclude any nodes.
       * <p>
       * This algorithm generally leaves the least desirable element first in the sorted result.
       */
      public int compareTo(Node<K, V> n)
      {
         // don't catch ClassCastException, because we control the type of node in the cache
         NodeImpl<K, V> that = (NodeImpl<K, V>) n;
         
         int h0 = this.hits;
         int h1 = that.hits;
         
         if (h0 != h1)
         {
            return (h0 - h1);
         }
         
         long a0 = this.lastAccess;
         long a1 = that.lastAccess;
         
         if (a0 > a1)
         {
            return 1;
         }
         
         if (a0 < a1)
         {
            return -1;
         }
         
         return this.id - that.id;
      }
      
      /**
       * An implementation of equivalence checking consistent with <code>compareTo(Node)</code>.
       * 
       * @param   o
       *          Object with which to compare this instance.
       * 
       * @return  <code>true</code> if <code>o</code> is equivalent to this instance, else
       *          <code>false</code>.
       */
      public boolean equals(Object o)
      {
         if (!(o instanceof NodeImpl))
         {
            return false;
         }
         
         // don't catch ClassCastException, because we control the type of node in the cache
         @SuppressWarnings("unchecked")
         NodeImpl<K, V> that = (NodeImpl<K, V>) o;
         
         return (this.hits == that.hits &&
                 this.lastAccess == that.lastAccess &&
                 this.id == that.id);
      }
      
      /**
       * Increment the hit count for this node and record the timestamp of this access.
       */
      protected void access()
      {
         if (hits < Integer.MAX_VALUE)
         {
            hits++;
         }
         lastAccess = System.currentTimeMillis();
      }
   }
}