/*
** Module   : Utf8Constant.java
** Abstract : 
**
** Copyright (c) 2001-2017, Golden Code Development Corporation.
**
** -#- -I- --Date-- -T- --JPRM-- ----------------Description-----------------
** 001 ECF 20041130 ADD   @xxxxx Production version from JTA, reworked
**                               into a new package and with minor cleanups
**                               in javadoc.
*/
/*
** This program is free software: you can redistribute it and/or modify
** it under the terms of the GNU Affero General Public License as
** published by the Free Software Foundation, either version 3 of the
** License, or (at your option) any later version.
**
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
** GNU Affero General Public License for more details.
**
** You may find a copy of the GNU Affero GPL version 3 at the following
** location: https://www.gnu.org/licenses/agpl-3.0.en.html
** 
** Additional terms under GNU Affero GPL version 3 section 7:
** 
**   Under Section 7 of the GNU Affero GPL version 3, the following additional
**   terms apply to the works covered under the License.  These additional terms
**   are non-permissive additional terms allowed under Section 7 of the GNU
**   Affero GPL version 3 and may not be removed by you.
** 
**   0. Attribution Requirement.
** 
**     You must preserve all legal notices or author attributions in the covered
**     work or Appropriate Legal Notices displayed by works containing the covered
**     work.  You may not remove from the covered work any author or developer
**     credit already included within the covered work.
** 
**   1. No License To Use Trademarks.
** 
**     This license does not grant any license or rights to use the trademarks
**     Golden Code, FWD, any Golden Code or FWD logo, or any other trademarks
**     of Golden Code Development Corporation. You are not authorized to use the
**     name Golden Code, FWD, or the names of any author or contributor, for
**     publicity purposes without written authorization.
** 
**   2. No Misrepresentation of Affiliation.
** 
**     You may not represent yourself as Golden Code Development Corporation or FWD.
** 
**     You may not represent yourself for publicity purposes as associated with
**     Golden Code Development Corporation, FWD, or any author or contributor to
**     the covered work, without written authorization.
** 
**   3. No Misrepresentation of Source or Origin.
** 
**     You may not represent the covered work as solely your work.  All modified
**     versions of the covered work must be marked in a reasonable way to make it
**     clear that the modified work is not originating from Golden Code Development
**     Corporation or FWD.  All modified versions must contain the notices of
**     attribution required in this license.
*/

package com.goldencode.expr;

import java.io.*;

/**
 * This class represents a string constant in UTF-8 format, which is used
 * by the JVM to reference strings.
 */
public final class Utf8Constant extends Constant
{
   /** Text underlying this constant. */
   private String text;
   
   /**
    * Return an instance of this class, either newly created, or retrieved
    * from the specified constant pool.
    *
    * @param   pool
    *          Constant pool from which to retrieve instance of this class,
    *          if one with the same underlying text already has been created.
    * @param   text
    *          Text which should back this constant.
    *
    * @return  Instance of this class, either newly created, or retrieved
    *          from the specified constant pool.
    */
   static Utf8Constant create(ConstantPool pool, String text)
   {
      return (Utf8Constant) pool.get(new Utf8Constant(text));
   }
   
   /**
    * Create a new instance of this class, based on the specified text.
    *
    * @param   text
    *          Text which should back this constant.
    */
   private Utf8Constant(String text)
   {
      super(UTF8);
      
      this.text = text;
   }
   
   /**
    * Return the text underlying this constant.
    *
    * @return  Text underlying this constant.
    */
   public String getText()
   {
      return text;
   }
   
   /**
    * Return the unique hash code for this object.
    *
    * @return  Unique hash code for this object.
    */
   public int hashCode()
   {
      return (int) (hashBase + text.hashCode());
   }
   
   /**
    * This method returns true if this object is equal to the specified
    * object.  Equality is defined as both objects being instances of this
    * class and the text underlying this constant and the specified constant
    * being the same.
    *
    * @param   o
    *          Another Utf8Constant object to check for equality.
    *
    * @return  True if objects are equal, else false.
    */
   public boolean equals(Object o)
   {
      if (!(o instanceof Utf8Constant))
         return false;
      boolean res = text.equals(((Utf8Constant) o).text);
      
      return res;
   }
   
   /**
    * This method writes the text underlying this constant out to the
    * specified output stream in UTF-8 format.
    *
    * @param   out
    *          Output stream to which text is written.
    *
    * @throws  IOException
    *          If error occurs writing to stream.
    */
   protected void writeContent(DataOutput out)
   throws IOException
   {
      out.writeUTF(text);
   }
   
   /**
    * This method compares this object to the specified object to determine
    * the natural ordering among the two.  The superclass' <code>compareTo()
    * </code> method is first called, which sorts the objects according to
    * the constant types.  If this comparison indicates the objects compare
    * as equal, the underlying text is then compared.
    *
    * @param   o
    *          Another object to compare to this one, presumably another
    *          subclass of <code>Constant</code>.
    *
    * @return  0 if objects compare as equal, a negative number if this
    *          object compares as less than the specified object, or a
    *          positive number if this object compares as greater than the
    *          specified object.
    */
   public int compareTo(Object o)
   {
      int comp = super.compareTo(o);
      if (comp == 0)
      {
         comp = text.compareTo(((Utf8Constant) o).text);
      }
      
      return comp;
   }
   
   public String toString()
   {
      return text;
   }
}