/*
** Module   : expression.g
** Abstract : ANTLR grammar for the lexer, parser, and tree walker classes
**            which together provide the front-end to the expression compiler
**
** Copyright (c) 2004-2017, Golden Code Development Corporation.
**
** -#- -I- --Date-- -T- --JPRM-- ----------------Description-----------------
** 001 GES 20050126 ADD   @19653 First version implementing a complete lexer,
**                               parser and tree walker for infix expressions
**                               which can contain numeric and string
**                               constants, variables, and functions (which
**                               may include nested expressions as arguments).
**                               This implementation is based largely on the
**                               expression support built into the Progress
**                               language lexer and parser for P2J, with
**                               certain changes applied which make the
**                               expression support more Java-like in terms
**                               of syntax and operator precedence.
** 002 ECF 20050201 ADD   @19654 Added javadoc commentary.
** 003 ECF 20050207 CHG   @19655 Modified un_expr rule in parser to make the
**                               unary bitwise complement operator and the
**                               unary logical negation operator the roots of
**                               their respective ASTs.
** 004 ECF 20050215 CHG   @19820 Modified tree walker's next method to use
**                               token type rather than symbol to look up
**                               operator. This was necessary to allow for
**                               both binary and unary minus operators.
** 005 ECF 20050310 CHG   @20270 Modified parser to support inline resolution
**                               of constants to literals.
** 006 ECF 20050422 CHG   @20847 Modified parser's constant resolution to
**                               support namespaces. An optional namespace can
**                               now be prefixed to a constant symbol. This
**                               value is passed to the symbol resolver when
**                               it is asked to resolve the constant to a
**                               literal. Whether to use the namespace value
**                               is at the discretion of the symbol resolver
**                               implementation associated with the parser.
**                               Incidentally redefined the single equals sign
**                               (=) from an equality test token to an
**                               assignment token, in order to reserve this
**                               operator for a future assignment operation.
**                               Now only double equals (==) is a valid
**                               equality test.
** 007 ECF 20050428 CHG   @20927 Modified primary_expr parser rule to handle
**                               resolution to Integer and Float literals.
**                               Symbol resolvers may return these types in
**                               lieu of Long and Double values, respectively.
** 008 ECF 20050430 CHG   @20950 Extended namespace support to variables and
**                               user functions. Changed the namespace
**                               qualifier delimiter from colon to dot.
** 009 GES 20050509 CHG   @21144 Major modifications to the lexer and the
**                               parser to implement a more generic expression
**                               syntax which allows Java objects to be
**                               directly referenced in expressions.  This
**                               parser/lexer also implement associated
**                               support for chained methods, null literals
**                               and full expression tree verification.  All
**                               constant resolution still occurs in the
**                               parser, though a bit differently.  Variables
**                               are supported (including an assignment form
**                               in the top-level expression rule) but these
**                               are converted into a getter/setter method
**                               call and so there is no direct variable
**                               support in the resulting tree.  The last 
**                               vestiges of the abnormal expression language
**                               syntax (@ symbol in function names,
**                               bit/byte/off embedded function names...)
**                               have been removed.  The vestiges were from
**                               the original implementation and are
**                               unnecessary in the new parser.  A custom
**                               ExtraAst class is used to store related
**                               data with each AST node.  For METH_* node
**                               types, it stores the associated Function
**                               object that describes all data needed to
**                               invoke the method.  For all other types it
**                               stores the Class object of the resulting
**                               sub-expression rooted at that node EXCEPT
**                               for the NULL node for which this extra
**                               object will be null (since there is no such
**                               thing as a Class object that represents the
**                               null literal).
** 010 GES 20050527 CHG   @21320 Added a (unary) cast operator and updated 
**                               the method resolution processing to change
**                               the signature array accordingly so that the
**                               proper methods will be found. This is
**                               important since many methods may return a
**                               "wider" type (e.g. Object) and yet the
**                               method resolution would break in those
**                               cases.  In that case, the CAST operator is
**                               removed from the tree (it is not needed
**                               downstream) but in all other cases it is
**                               left behind.
** 011 ECF 20050615 CHG   @21497 Removed NEG_METH_BOOLEAN token type. This
**                               token type is no longer in use.
** 012 ECF 20071121 CHG   @39253 Added support for positional substitution
**                               parameters. This allows SQL-like expressions
**                               which use positional parameter substitution.
** 013 EVL 20160224              Javadoc fixes to make compatible with Oracle Java 8 for
**                               Solaris 10.
*/
/*
** This program is free software: you can redistribute it and/or modify
** it under the terms of the GNU Affero General Public License as
** published by the Free Software Foundation, either version 3 of the
** License, or (at your option) any later version.
**
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
** GNU Affero General Public License for more details.
**
** You may find a copy of the GNU Affero GPL version 3 at the following
** location: https://www.gnu.org/licenses/agpl-3.0.en.html
** 
** Additional terms under GNU Affero GPL version 3 section 7:
** 
**   Under Section 7 of the GNU Affero GPL version 3, the following additional
**   terms apply to the works covered under the License.  These additional terms
**   are non-permissive additional terms allowed under Section 7 of the GNU
**   Affero GPL version 3 and may not be removed by you.
** 
**   0. Attribution Requirement.
** 
**     You must preserve all legal notices or author attributions in the covered
**     work or Appropriate Legal Notices displayed by works containing the covered
**     work.  You may not remove from the covered work any author or developer
**     credit already included within the covered work.
** 
**   1. No License To Use Trademarks.
** 
**     This license does not grant any license or rights to use the trademarks
**     Golden Code, FWD, any Golden Code or FWD logo, or any other trademarks
**     of Golden Code Development Corporation. You are not authorized to use the
**     name Golden Code, FWD, or the names of any author or contributor, for
**     publicity purposes without written authorization.
** 
**   2. No Misrepresentation of Affiliation.
** 
**     You may not represent yourself as Golden Code Development Corporation or FWD.
** 
**     You may not represent yourself for publicity purposes as associated with
**     Golden Code Development Corporation, FWD, or any author or contributor to
**     the covered work, without written authorization.
** 
**   3. No Misrepresentation of Source or Origin.
** 
**     You may not represent the covered work as solely your work.  All modified
**     versions of the covered work must be marked in a reasonable way to make it
**     clear that the modified work is not originating from Golden Code Development
**     Corporation or FWD.  All modified versions must contain the notices of
**     attribution required in this license.
*/

header
{
/*
** Module   : ExpressionParser.java
**            ExpressionParserTokenTypes.java
**            ExpressionLexer.java
** Abstract : 
**
** Copyright (c) 2004-2016, Golden Code Development Corporation.
**
** -#- -I- --Date-- -T- --JPRM-- ----------------Description-----------------
** 001 ECF 20050126 NEW   @xxxxx WARNING, THIS IS A GENERATED FILE!!!
**                               DO NOT EDIT THIS FILE. The original source
**                               file is expression.g!
*/ 

package com.goldencode.expr;

import java.util.*;
import java.util.List;
import java.text.*;
import java.io.*;
import antlr.*;
}

/**
 * Creates an Abstract Syntax Tree (AST) representation of an expression
 * in infix notation from an input stream of tokens (provided by the
 * {@link ExpressionLexer}).  More specifically, the parser does the
 * following:
 * <ul>
 *    <li> Drives the lexer by calling its' <code>nextToken</code> method.
 *         This method tokenizes the lexer's input stream of characters and
 *         returns the next found <code>Token</code> object.
 *    <li> Provides the {@link #expression} method as an entry point.
 *    <li> Builds an AST (&quot;intermediate form representation&quot;) of
 *         all the recognized tokens. This AST is designed for subsequent
 *         processing by {@link Compiler} to compile the expression into a
 *         Java class.
 *    <li> Handles the context sensitive aspects of symbol resolution.
 *    <li> To the extent that is possible while still generating a well
 *         structured tree, the parser has been structured to enforce
 *         syntactic correctness.
 * </ul>
 * <p>
 * This grammar is highly specific to ANTLR 2.7.4 and it may not be valid
 * in future versions of ANTLR.  This is due to the fact that the generated
 * code and the grammar syntax of ANTLR can change between versions.  These
 * factors are the most important constraints defining the structure of the
 * resulting grammar.  As a result, it is possible that the grammar will
 * require changes to run in future versions of ANTLR.
 * <p> 
 * <b>Please note that this is a generated file using ANTLR 2.7.4 and a
 * grammar specified in <code>expression.g</code>.</b>
 *
 * @author GES
 * @author ECF
 */
class ExpressionParser extends Parser;

options
{   
   k = 4;
   buildAST = true;
   ASTLabelType = "ExtraAst";
}

tokens
{
   DOT;
   BOOL_FALSE;
   BOOL_TRUE;
   NULL;
   UN_MINUS;
   CAST;
   METH_OBJ;
   METH_BOOLEAN;
   METH_DOUBLE;
   METH_LONG;
   METH_INT;
   IS_NULL;
   NOT_NULL;
   HEX_LITERAL;
   DEC_LITERAL;
   IDENTITY;
   GOTO;
}
   
// Custom instance variables, constructors, methods...
{
   /** Name of the specialized AST class to use when parsing. */
   private static final String AST_CLASS = ExtraAst.class.getName();

   /** Object which resolves constants for the parser */
   private SymbolResolver resolver = null;
   
   /**
    * Create a new parser instance, given a lexer instance and a symbol
    * resolver.
    *
    * @param   lexer
    *          Expression lexer.
    * @param   resolver
    *          Object which resolves constants during lexing.
    *
    * @see     #setSymbolResolver
    */
   public ExpressionParser(ExpressionLexer lexer, SymbolResolver resolver)
   {
      this(lexer, resolver, false);
   }
   
   /**
    * Create a new parser instance, given a lexer instance, a symbol resolver,
    * and an indication of whether assignments are allowed.
    *
    * @param   lexer
    *          Expression lexer.
    * @param   resolver
    *          Object which resolves constants during lexing.
    * @param   noAssignment
    *          If <code>false</code>, assignment operations are permitted;
    *          if <code>true</code>, the ASSIGN (<code>=</code>) operator is
    *          converted to the EQUALS (<code>==</code>) operator.
    *
    * @see     #setSymbolResolver
    */
   public ExpressionParser(ExpressionLexer lexer,
                           SymbolResolver resolver,
                           boolean noAssignment)
   {
      this(lexer);
      
      setASTNodeClass(AST_CLASS);
      setSymbolResolver(resolver);
      if (noAssignment)
      {
         lexer.setNoAssignment();
      }
   }
   
   /**
    * Sets the object which will be used to resolve constants at parse time.
    * A string (without quotation marks) encountered in an expression will
    * represent a constant or a variable/function. The resolver first tries
    * to resolve a symbol to a constant. If it succeeds, the string
    * constant will be replaced with one of the following:
    * <ul>
    * <li>integral numeric literal
    * <li>floating point numeric literal
    * <li>string
    * <li>boolean
    * </ul>
    * Thus, if the replacement succeeds, the expression compiler will only
    * see one of the above objects, never the original string constant. If
    * the replacement fails, the original string value will be considered to
    * be a variable.
    *
    * @param   resolver
    *          Object which will resolve constants at parse time.
    */
   public void setSymbolResolver(SymbolResolver resolver)
   {
      this.resolver = resolver;
   }
   
   /**
    * Implements special processing for <code>NULL</code> equality and
    * inequality tests.  If a <code>NULL</code> token is detected on either
    * side of either of the operators, the result will be rooted at an
    * <code>IS_NULL</code> (equality) or <code>NOT_NULL</code> (inequality)
    * node.  The operand that is <code>NULL</code> will be dropped and the
    * other operand will be the only child.    
    * <p>
    * This is separated from the main processing and really is a kind of
    * post-processing of the entire tree.
    */
   public void convertNullTest(CommonAST ast)
   {
      int type = ast.getType();
   
      // process the current node, only equality and inequality are of
      // interest
      if (type == EQUALS || type == NOT_EQ)
      {
         CommonAST first  = (CommonAST) ast.getFirstChild();
         CommonAST second = (CommonAST) first.getNextSibling();
         
         int ftype = first.getType(); 
         int stype = second.getType();
      
         // detect comparisons with NULL and rewrite the sub-expression
         if (ftype == NULL || stype == NULL)
         {
            // special case that can be optimized easily
            if (ftype == NULL && stype == NULL)
            {
               ast.setType(type == EQUALS ? BOOL_TRUE : BOOL_FALSE);
               ast.removeChildren();
            }
            else
            {
               // normal case: rewrite the token as our artificial unary
               // operator
               ast.setType(type == EQUALS ? IS_NULL : NOT_NULL);
               
               // only leave the non-NULL child behind as the single operand
               if (first.getType() == NULL)
               {
                  ast.setFirstChild(second);
               }
               else
               {
                  first.setNextSibling(null);
               }
            }
         }
      }
      
      // either way, process all children using a recursive call 
      CommonAST child = (CommonAST) ast.getFirstChild();
      
      while (child != null)
      {
         convertNullTest(child);
         child = (CommonAST) child.getNextSibling();
      }
   }
   
   /**
    * Reads the return value (as a class) from a function and determines 
    * which kind of method call this represents.  The possibilities:
    * <pre>
    * Java Return                       Parser Type
    * ---------------------------       ---------------------
    * java.lang.Boolean                 METH_BOOLEAN
    * java.lang.Double                  METH_DOUBLE
    * java.lang.Long                    METH_LONG
    * java.lang.Integer                 METH_INT
    * anything else                     METH_OBJ
    * </pre>
    *
    * @param    func
    *           The object representing the function call.
    *
    * @return   The associated parser token type.
    */
   public int convertFunctionToTokenType(Function func)
   {
      Class rtype = func.getReturnType();
      int   type  = -1;
      
      if (java.lang.Boolean.TYPE.equals(rtype) ||
          java.lang.Boolean.class.equals(rtype))
      {
         type = METH_BOOLEAN;
      }
      else if (java.lang.Integer.TYPE.equals(rtype) ||
               java.lang.Integer.class.equals(rtype))
      {
         type = METH_INT;
      }
      else if (java.lang.Long.TYPE.equals(rtype) ||
               java.lang.Long.class.equals(rtype))
      {
         type = METH_LONG;
      }
      else if (java.lang.Double.TYPE.equals(rtype)||
               java.lang.Double.class.equals(rtype))
      {
         type = METH_DOUBLE;
      }
      else
      {
         // something else!
         type = METH_OBJ;
      }
      
      return type;
   }
   
   /**
    * Walks the children (optionally ignoring the first child if it
    * represents the target instance upon which the method is being
    * called) and for each one uses {@link Verifier#resolveDataType} to
    * determine the class of the resulting expression and builds the
    * corresponding method signature.
    * <p>
    * This rule also handles the special case of a parameter expression being
    * rooted at a <code>CAST</code> token which is really just a parse-time
    * unary operator that overrides the type of the resulting parameter's
    * entry in the signature array. Once this override is done, the
    * <code>CAST</code> AST is removed and the single child node is 
    * re-parented at the method call AST directly.  This allows a cast to
    * be used to force the correct match on the right method during method
    * resolution and the compiler handles any runtime casting that is needed. 
    *
    * @param    ast
    *           The AST node representing the method call.
    * @param    ignoreFirst
    *           If <code>true</code> the first child is the target instance
    *           upon which the method is being called and must be ignored.
    *           <code>false</code> if the first child is the first parameter
    *           to the method call.
    *
    * @return   The signature of the method (all class objects of the
    *           parameters in the proper number and order).
    */
   Class[] generateSignature(ExtraAst ast, boolean ignoreFirst)
   {
      ExtraAst child = (ExtraAst) ast.getFirstChild();
      ExtraAst last  = null;
      
      // the first parm may be an instance NOT the start of our signature
      if (ignoreFirst && child != null)
      {
         last  = child;
         child = (ExtraAst) child.getNextSibling();
      }
      
      // special case: no parameters (we might get this for free but since it
      // isn't explicitly documented in the ArrayList class, this is safer)
      if (child == null)
      {
         // empty signature
         return new Class[0];
      }
      
      ArrayList signature = new ArrayList();
      
      while (child != null)
      {
         if (child.getType() == CAST)
         {
            // obtain the real expression and verify it
            ExtraAst grandchild = (ExtraAst) child.getFirstChild();
            Verifier.resolveDataType(grandchild);
            
            // add our override type
            signature.add(resolveClass(child.getText()));
            
            // re-parent the expression at the method
            if (last == null)
            {
               // replace the first child of the mathod AST
               ast.setFirstChild(grandchild);               
            }
            else
            {
               last.setNextSibling(grandchild);
            }
            
            // set the downstream link to the next sibling, even if null
            grandchild.setNextSibling(child.getNextSibling());            
         }
         else
         {
            signature.add(Verifier.resolveDataType(child));
         }
         
         last  = child;         
         child = (ExtraAst) child.getNextSibling();
      }
      
      return (Class[])signature.toArray(new Class[0]);
   }
   
   /**
    * Convert a <code>String</code> (which is a class name or a primitive
    * type) into the corresponding Java <code>Class</code> object.
    *
    * @param    s
    *           The class or primitive type name.
    *
    * @return   The resulting class. 
    *
    * @throws   IllegalArgumentException
    *           If reflection cannot find the associated class.
    */
   public static Class resolveClass(String s)
   throws IllegalArgumentException
   {
      Class cls = null;
   
      // convert the primitive types (except void) into the corresponding
      // Class constants (because these cannot be looked up via reflection)
      if ("boolean".equals(s))
      {
         cls = Boolean.TYPE;
      }
      else if ("byte".equals(s))
      {
         cls = Byte.TYPE;
      }
      else if ("char".equals(s))
      {
         cls = Character.TYPE;
      }
      else if ("short".equals(s))
      {
         cls = Short.TYPE;
      }
      else if ("int".equals(s))
      {
         cls = Integer.TYPE;
      }
      else if ("long".equals(s))
      {
         cls = Long.TYPE;
      }
      else if ("float".equals(s))
      {
         cls = Float.TYPE;
      }
      else if ("double".equals(s))
      {
         cls = Double.TYPE;
      }
      
      // normal lookup processing
      if (cls == null)
      {
         try
         {
            cls = Class.forName(s);
         }
         
         catch (ClassNotFoundException cnf)
         {
            // give up!
            throw new IllegalArgumentException("Invalid class name in #(" +
                                               s + ").");
         }
      }
      
      return cls;
   }
   
   /**
    * Simple walker to print out the contents of an AST for debugging.
    *
    * @param    ast
    *           The current node to print.
    * @param    level
    *           The number of levels down from the root (0 == root node),
    *           which determines the indention level to print.
    */
   public static void visit(CommonAST ast, int level)
   {
      int          spaces = level * 3;
      StringBuffer sb     = new StringBuffer();
      
      // generate the indent based on the level from the root
      for (int i = 0; i < spaces; i++)
      {
         sb.append(' ');
      }
      
      // add our current node's text
      sb.append("[ ");
      sb.append(ast.getText());
      sb.append(" ] <");
      sb.append(_tokenNames[ast.getType()]);
      sb.append(">");
      
      // TEMPORARY FOR DEBUGGING...
      Object extra = ((ExtraAst) ast).getExtra();
      if (!(extra instanceof CodeUnit))
      {
         sb.append(" ");
         sb.append(((ExtraAst) ast).getExtra());
      }
      
      // print out the current node
      System.out.println(sb.toString());
   
      CommonAST child = (CommonAST) ast.getFirstChild();
      level++;
      
      // now visit all children
      while (child != null)
      {
         visit(child, level);
         child = (CommonAST) child.getNextSibling();
      }
   }
   
   /**
    * Provides a command line interface for an end user to drive and/or test
    * the ExpressionParser class.
    * <p>
    * Syntax:
    * <pre>
    *    java ExpressionParser &lt;expression&gt; [no-assignment]
    * </pre>
    *
    * @param   args 
    *          List of command line arguments.
    */
   public static void main(String[] args)  
   {
      int len = args.length;
      if (len < 1 || len > 2)
      {
         System.err.println(
            "Syntax: java ExpressionParser <expression> [no-assignment]");
         System.exit(-1);
      }
         
      try
      {
         StringReader     expr       = new StringReader(args[0]);
         ExpressionLexer  lexer      = new ExpressionLexer(expr);
         SymbolResolver   resolver   = SymbolResolver.getTestResolver();
         ExpressionParser parser     = new ExpressionParser(lexer, resolver);
         
         BaseAST.setVerboseStringConversion(true, _tokenNames);
         
         if (len == 2 && "true".equalsIgnoreCase(args[1]))
         {
            lexer.setNoAssignment();
         }
         
         Class resultType = parser.expression();         
         
         CommonAST result = (CommonAST) parser.getAST();
         
         // print our report
         visit(result, 0);
      }
       
      catch(Exception excpt)
      {
         excpt.printStackTrace();
      }
   }
}

/**
 * Parses an infix expression and creates an AST that is verified suitable 
 * for compilation.  This is the main "entry point" for expression parsing.  
 * <p>
 * The AST is structured with operator nodes as the root nodes and each
 * operator (unary or binary) has the matching number of child nodes (1 or
 * 2 respectively).  Each child node is an operand and the semantics of
 * left and right operands (for binary operators) is maintained.
 * <p>
 * Non-operator nodes in the tree can be literals or method calls (this
 * is how variables and functions are both implemented).  The list of
 * literals can be seen in the {@link #literal} rule.  Method resolution
 * supports full method chaining although all method calls (or chains
 * thereof) must start on an instance of a well known object which is 
 * a 'variable' (an object instance that can be resolved by a method call
 * and is exposed by the resolver).  All method support is implemented via
 * the {@link #method} rule.
 * <p>
 * All of the following operators are supported and the precedence order
 * follows that of the Java programming language. Alternative syntax, where
 * applicable, is shown in parentheses.
 * <pre>
 * lowest  conditional OR          :  or (||)
 *    |    conditional AND         :  and (&amp;&amp;)
 *    |    bitwise inclusive OR    :  |
 *    |    bitwise exclusive OR    :  ^
 *    |    bitwise AND             :  &amp;
 *    |    equality                :  ==, !=
 *    |    relational              :  &lt;, &gt;, &le;, &ge;
 *    |    shift                   :  &lt;&lt;, &gt;&gt;, &gt;&gt;&gt;
 *    |    additive                :  +, -
 *    |    multiplicative          :  *, /, %
 *    |    unary operators         :  -, !, ~, #()
 *    |    method call             :  method_name(...), object.method_name(...)
 * highest parentheses             :  (...)
 * </pre>
 * <p>
 * To handle the precedence issue, each precedence level (see the table
 * above) is handled in a separate rule.  Starting at in this top level
 * entry point, the lowest precedence rule is referenced. This reference is
 * made with an optional <code>OR</code> operator.  If this operator is
 * present, it is the root of the resulting tree.  This means it is the
 * last operator to be evaluated since its operands will be processed
 * first.  This is the proper definition of lowest precedence.  If this
 * operator is present, then a second operand is expected.
 * <p>
 * Each precedence level is similarly constructed, the left operand being
 * required, all operators being optional and each operator choice an equal 
 * alternative. If any of the operators exists, then it is the root of the
 * subtree and a right operand is expected.  This holds true for all rules
 * that process binary operators.  Please note that this left-mandatory
 * and optional right side (operator + right operand) is critical for
 * allowing an expression to be as simple as a single variable, function
 * call or literal. Alternatively, it can expand into a highly nested set of
 * recursively parsed expressions that all evaluate to a single scalar result.
 * <p>
 * At the bottom of the hierarchy of rules are unary operators, these
 * are implemented as an optional prefix operator and a mandatory right
 * operand.
 * <p>
 * The final level is the primary expression.  This is where the precedence
 * operator parenthesis is handled.  It also is where we allow a single
 * literal, variable, or function reference.  While the literals are fairly
 * easily defined, variables and functions are complicated by the fact that
 * they are used defined symbols and there is inherent ambiguity in the
 * definitions for each of these. See {@link #primary_expr} for more details
 * details on how this is resolved.
 * <p>
 * Ambiguity is primarily caused by the potential overlap in the rules
 * referenced from {@link #primary_expr}.  This is caused by the fact that
 * multiple rule references made there must all allow for a user-defined
 * symbol to appear in the token stream.  Multiple rules all requiring
 * the same definition to appear is the definition of <b>ambiguity</b>!
 * See that rule for more details on the solution.
 * <p>
 * The {@link #expr} is called as the only rule reference.
 * <p>
 * This rule implements the only assignment type syntax that is possible in
 * this parser.  In other words, assignments can only occur at the top-level
 * of an expression and cannot be placed inline.  Note that the assignment
 * operator <code>ASSIGN</code> is dropped from the resulting tree and
 * any variable setter method call is used as the true root node with
 * the required expression to the right of the assignment operator as its
 * only child.  If the variable and assignment operator aren't present, this
 * rule matches a single expression.
 * <p>
 * The resulting tree will have been completely verified for validity
 * using {@link Verifier#resolveDataType}.
 *
 * @return   The data type of the resulting, validated expression.
 */
expression returns [Class cls = null]
   :
      (
         (
              { LA(2) == ASSIGN || LA(4) == ASSIGN }?                  
              p:primary_expr[true] ASSIGN!
              
            | // the empty alternative when there is no assignment
         )            
         e:expr
      )
      {
         // make sure that the expression is the child of the method call
         // if this is an assignment
         if (#p != null)
         {
            // special case where a parameter to a method call would not
            // otherwise be verified, so we force it.
            Verifier.resolveDataType(#e);
            #p.addChild(#e);
         }
      
         // run the verifier and update our return data type
         cls = Verifier.resolveDataType(##);
      }
   ;

/**
 * Implements the lowest (1st) precedence level of expressions: 
 * <code>OR</code> (the logical OR operator).  Valid expressions can contain
 * sub-expressions and this entry point is properly recursive without causing
 * any ambiguity.
 * <p>
 * This rule is designed to allow recursion (the highest precedence levels
 * {@link #method} and {@link #primary_expr} reference this rule again,
 * allowing infinite nesting of sub-expressions, with proper parsing,
 * evaluation and tree building.
 * <p>
 * This is only called by the top level {@link #expression} rule.  This
 * method calls the {@link #log_and_expr} rule.
 * <p>
 * Implements special processing for <code>NULL</code> equality and inequality
 * tests.  If a <code>NULL</code> token is detected on either side of either
 * of the operators, the result will be rooted at an <code>IS_NULL</code>
 * (equality) or <code>NOT_NULL</code> (inequality) node.  The operand that
 * is <code>NULL</code> will be dropped and the other operand will be the
 * only child.  This must be implemented after the entire tree is created 
 * and the method {@link #convertNullTest} is used to implement this
 * processing.
 */
expr   
   :
      (
         log_and_expr (OR^ log_and_expr)*
      )
      {
         // convert all equality and inequality tests against the null
         // literal into a special operator
         convertNullTest((CommonAST) ##);         
      }
   ;
   
/**
 * Implements the 2nd lowest precedence level of expressions: <code>AND</code>
 * (the logical AND operator).
 * <p>
 * This is only called by the {@link #expr} rule.  This method calls the
 * {@link #bit_or_expr} rule.
 */
log_and_expr 
   :
      bit_or_expr (AND^ bit_or_expr)*
   ; 

/**
 * Implements the 3rd lowest precedence level of expressions: <code>|</code>
 * (the inclusive OR operator).
 * <p>
 * This is only called by the {@link #log_and_expr} rule.  This method
 * calls the {@link #bit_xor_expr} rule.
 */
bit_or_expr
   :
      bit_xor_expr (BIT_OR^ bit_xor_expr)*
   ;
   
/**
 * Implements the 4th lowest precedence level of expressions: <code>^</code>
 * (the exclusive OR operator).
 * <p>
 * This is only called by the {@link #bit_or_expr} rule.  This method
 * calls the {@link #bit_and_expr} rule.
 */
bit_xor_expr
   :
      bit_and_expr (BIT_XOR^ bit_and_expr)*
   ;    
   
/**
 * Implements the 5th lowest precedence level of expressions: <code>&amp;</code>
 * (the bitwise AND operator).
 * <p>
 * This is only called by the {@link #bit_xor_expr} rule.  This method
 * calls the {@link #equality_expr} rule.
 */
bit_and_expr 
   :
      equality_expr (BIT_AND^ equality_expr)*
   ; 
   
/**
 * Implements the 6th lowest precedence level of expressions - the equality
 * operators.
 * <p>
 * This is only called by the {@link #bit_and_expr} rule.  This method
 * calls the {@link #compare_expr} rule.
 * <p>
 * Provides the following operators:
 * <pre>
 *    ==       (equality)
 *    !=       (inequality)
 * </pre>
 */
equality_expr
   :
      compare_expr ((EQUALS^ | NOT_EQ^) compare_expr)*
   ;
   
/**
 * Implements the 7th lowest precedence level of expressions - the comparison
 * operators.
 * <p>
 * This is only called by the {@link #equality_expr} rule. This method calls
 * the {@link #shift_expr} rule.
 * <p>
 * Provides the following operators:
 * <pre>
 *    &gt;  (greater than)
 *    &lt;  (less than)
 *    &ge; (greater than or equal to)
 *    &le; (less than or equal to)
 * </pre>
 */
compare_expr
   :
      shift_expr ((GT^ | LT^ | GTE^ | LTE^) shift_expr)*
   ;
   
/**
 * Implements the 8th lowest precedence level of expressions - the shift
 * operators.
 * <p>
 * This is only called by the {@link #compare_expr} rule. This method calls
 * the {@link #sum_expr} rule.
 * <p>
 * Provides the following operators:
 * <pre>
 *    &gt;&gt;  (shift right with sign extension)
 *    &lt;&lt;  (shift left)
 *    &gt;&gt;&gt; (shift right with zero extension)
 * </pre>
 */
shift_expr
   :
      sum_expr ((RSHIFT^ | LSHIFT^ | ZRSHIFT^) sum_expr)*
   ;
   
/**
 * Implements the 9th lowest precedence level of expressions - the additive
 * operators.
 * <p>
 * This is only called by the {@link #shift_expr} rule. This method calls
 * the {@link #prod_expr} rule.
 * <p>
 * Provides the following operators:
 * <pre>
 *    + (binary plus)
 *    - (binary minus)
 * </pre>
 */
sum_expr
   :
      prod_expr ((PLUS^ | MINUS^) prod_expr)*
   ; 

/**
 * Implements the 10th lowest precedence level of expressions - the
 * multiplicative operators.
 * <p>
 * This is only called by the {@link #sum_expr} rule. This method calls
 * the {@link #un_expr} rule.
 * <p>
 * Provides the following operators:
 * <pre>
 *    * (multiply)
 *    / (divide)
 *    % (modulus)
 * </pre>
 */
prod_expr
   :
      un_expr ((MULTIPLY^ | DIVIDE^ | MODULO^) un_expr)* 
   ;

/**
 * Implements the 11th lowest precedence level of expressions - the unary
 * operators.
 * <p>
 * This is only called by the {@link #prod_expr} rule. This method calls
 * the {@link #primary_expr} rule.
 * <p>
 * Provides the following operators:
 * <pre>
 *    -                  unary minus
 *    ~                  bitwise complement
 *    !                  logical NOT
 *    #(classname)       cast operator
 * </pre>
 */
un_expr   
   :
      {
         String cls = null;
      }
      (
         (
              m:MINUS^ { #m.setType( UN_MINUS ); }
            | BIT_COMP^
            | NOT^
            | h:HASH^ cls=cast! { #h.setType(CAST); #h.setText(cls); } 
         )?
         primary_expr[false]
      )
   ;
   
/**
 * Implements a cast operator which consists of a valid Java class name
 * enclosed in parenthesis.  This rule is unusual in that it does NOT
 * build any part of the tree but rather it matches the tokens and
 * returns the resulting class name as a string.
 * <p>
 * Note that this operator is formatted similar to a Java cast but it is
 * prepended with a pound sign as in:
 * <pre>
 * #(java.lang.String)
 * </pre>
 * This was done to eliminate ambiguity since our lexer and parser were both 
 * designed to operate on multipart symbol names such that it expects
 * <code>DOT</code> and <code>SYMBOL</code> tokens to be separated.  Since
 * class names match this same form AND such names can be arbitrarily long,
 * this parser's design is not well suited to resolving this ambiguity
 * (we can't use syntactic predicates since we rely upon init rules for
 * data initialization and syntactic predicates make such init rules
 * conditional).  Though it is not optimal, using the hash/pound sign works 
 * well and eliminates all ambiguity.
 *
 * @return   The built-up class name.
 */
cast! returns [String cls = null]
   :
      {
         StringBuffer sb = new StringBuffer();
      }
      (
         LPARENS
         s1:SYMBOL      { sb.append(#s1.getText()); } 
         (
            DOT
            s2:SYMBOL { sb.append('.').append(#s2.getText()); } 
         )*
         RPARENS
      )
      {
         cls = sb.toString();
      }
   ;
   
/**
 * Implements the 12th and highest precendence level of expressions, the
 * parenthesis precedence operators, {@link #literal} and {@link #method}.
 * <p>
 * This rule is different in structure from the previous levels, as there is
 * just a set of alternatives and one of them allows direct recursion back
 * to the <code>expr</code> expression entry point.  In addition, the method
 * calls can be arbitrarily chained into long sequences of methods, each being
 * called on the result of the last method's returned object.
 * <p>
 * If a <code>SYMBOL</code> is matched in the first position, the symbol
 * resolver is asked to resolve the symbol to a literal as if it were a
 * constant. If it succeeds, we set the token's type to the appropriate
 * literal type, depending upon the type of object returned by the resolver.
 * We also set the token's text to the string value of the literal. If the
 * resolver returns <code>null</code>, we assume the symbol represents a
 * variable or method call and the real work is deferred to the rule 
 * references.  Constants are optionally preceded by a qualifier symbol. If
 * present, this token's string is passed to the symbol resolver to be used
 * during the resolution of the constant.  It is up to the symbol resolver
 * to derive meaning from the qualifier string within the context of its
 * implementation.
 * <p>
 * This is only called by the {@link #un_expr} rule.
 * <p>
 * See the rule {@link #expr} for more details.
 *
 * @param   topLevel
 *          <code>true</code> when we're called as an assignment;  otherwise
 *          <code>false</code>.
 */
primary_expr! [boolean topLevel]
   :
      {
         Token    tok     = null;
         Object   literal = null;
         String   space   = null;
         int      ttype   = -1;
         ExtraAst ast     = null;
         
         // try to resolve a constant
         if (LA(1) == SYMBOL  && 
             LA(2) != LPARENS &&
             (LA(2) != DOT || (LA(2) == DOT && LA(4) != LPARENS)))
         {
            if (LA(2) == DOT)
            {
               tok   = LT(3);
               space = LT(1).getText();
            }
            else
            {
               tok = LT(1);
            }
            
            literal = resolver.resolveConstant(space, tok.getText());
            
            if (literal != null)
            {
               if (literal instanceof Long ||
                   literal instanceof Integer)
               {
                  ttype = NUM_LITERAL;
               }
               else if (literal instanceof Double ||
                        literal instanceof Float)
               {
                  ttype = DEC_LITERAL;
               }
               else if (literal instanceof String)
               {
                  ttype = STRING;
               }
               else if (literal instanceof Boolean)
               {
                  Boolean b = (Boolean) literal;
                  ttype = (b.booleanValue() ? BOOL_TRUE : BOOL_FALSE);
               }
               else
               {
                  throw new NoViableAltException(tok, getFilename());
               }
               
               tok.setText(literal.toString());
               tok.setType(ttype);
            }
         }
      }
      (
           // constant support
           { !topLevel && (LA(1) != SYMBOL || literal != null) }? 
           (SYMBOL! DOT!)? l:literal
           { ## = #l; }
           
           // variables and method calls
         | m1:method[null] { ast = #m1; } 
           (DOT! m2:method[ast] { ast = #m2; })*
           { ## = ast; }
           
           // parenthesized sub-expressions (the precedence operator)
         | { !topLevel }?
           LPARENS! e:expr RPARENS!
            { ## = #e; }
      )
   ;
   
/**  
 * Matches all variations of expression literals (constants).  This includes:
 * <pre>
 *    NUM_LITERAL
 *    DEC_LITERAL
 *    STRING
 *    BOOL_TRUE
 *    BOOL_FALSE
 *    NULL
 * </pre>
 */
literal      
   :
      (
           NUM_LITERAL
         | HEX_LITERAL
         | DEC_LITERAL
         | STRING
         | BOOL_TRUE
         | BOOL_FALSE 
         | NULL
      )
   ;   
      
/**
 * Implements the logic for matching and resolving variables and function
 * calls embedded in expressions. This rule is designed to be called in
 * a standalone manner or in a chained manner (where a method call is
 * invoked on a target object returned from another method call).  Of
 * great importance is the fact that whether the result is a variable or
 * a function, it always resolves to a method call.
 * <p>
 * The following are the possible resulting root nodes:
 * <pre>
 * METH_OBJ
 * METH_BOOLEAN
 * METH_DOUBLE
 * METH_LONG
 * METH_INT
 * </pre>
 * The if an AST is passed in, it will be set as the first child and
 * represents the object upon which this method is to be called.  All other 
 * children represent the method parameters and are used to match the
 * signature of the method (number, order and type of parameters).
 * <p>
 * The matching portion of this rule comes in 2 parts:
 * <ul>
 *    <li> The leftmost token is the function name.  This is defined as any
 *         valid symbol. 
 *    <li> The next part is an optional set of 2 or more tokens where the
 *         first is always <code>LPARENS</code> and the last is always
 *         <code>RPARENS</code>. In between is an optional expression, 
 *         optionally followed by any number of [<code>COMMA</code>
 *         expression] constructs.  This is the parameter list and it can
 *         contain any kind of expression.  This recursive call to the top
 *         level expression entry point allows the full variety of all
 *         possible expressions to be specified as parameters.
 * </ul>
 * <p>
 * Note that in the resulting AST, the parentheses and commas used to define
 * the parameter list are discarded.  In other words, they do not appear in
 * the AST as nodes.  Instead, each parameter (which can be a complete
 * expression subtree or a simple single node like an lvalue) is created
 * as a separate child node of the function type token.  All parameters are 
 * maintained as child nodes in the exact order of the function call itself,
 * so this is essentially the signature of the function.  It it also the
 * proper AST definition to enable simple evaluation of functions with any
 * arbitrary signature.
 *
 * @param    ast
 *           The target object or <code>null</code> if no target object
 *           exists.
 */
method [ExtraAst ast]   
   :
      {
         Function instance = null;
         Function func     = null;
         boolean  setter   = (LA(2) == ASSIGN);
         
         if (ast != null)
         {
            instance = (Function) ast.getExtra();
         }
         
         if (LA(2) != LPARENS)
         {
            // variable form
            func = resolver.resolveFunction(instance,
                                            LT(1).getText(),
                                            setter);
            
            if (func == null)
            {
               // this should never happen!
               String errmsg = "No function resolved for instance (" + 
                               instance + ") and variable (setter == " +
                               setter + ", name == (" + LT(1).getText() +
                               "))";
               throw new UnresolvedSymbolException(errmsg);
            }
            
            // rewrite our token type
            LT(1).setType(convertFunctionToTokenType(func));
         }
      }
      (
         // SYMBOL is only here for "pull" purposes, it will never be left
         // behind in the final tree (however for the function/method form
         // it is not rewritten until the exit rule)
         (
              METH_OBJ^
            | METH_INT^
            | METH_BOOLEAN^
            | METH_LONG^
            | METH_DOUBLE^
            | SYMBOL^
         )
         {
            // set the object instance as our first child, if we have one
            if (ast != null) astFactory.addASTChild(currentAST, ast);            
         }
         
         // match the parameters for the method call form
         (LPARENS! ( expr (COMMA! expr)* )? RPARENS!)?
      )
      {
         // if this method call is not a variable form, it can only be  
         // resolved after all parameters are parsed (here) since we must
         // build a parameter list to use with the resolveFunction()
         if (##.getType() == SYMBOL)
         {
            // the method call form can only be resolved once we have the
            // data type of the parameters, which means we had to finish
            // parsing first
            Class[] signature = generateSignature(##, (ast != null));
            
            // lookup the method call
            func = resolver.resolveFunction(instance, ##.getText(), signature);
            
            if (func == null)
            {
               // this should never happen!
               String errmsg = "No function resolved for instance (" + 
                               instance + ") and method (name == " +
                               ##.getText() + ", signature == (" +
                               Arrays.asList(signature) + "))";
               throw new UnresolvedSymbolException(errmsg);
            }
            
            // rewrite our token type
            ##.setType(convertFunctionToTokenType(func));
         }
         
         // is this a global library reference?
         Function chained = func.getPrerequisite();
         
         // global library variable and method references won't have an
         // instance in the source expression but the compiler needs one
         if (ast == null && chained != null)
         {
            // determine the proper token type based on the return type
            int ctype = convertFunctionToTokenType(chained);
         
            // manufacture a new node
            ExtraAst prereq = #[ctype, chained.getSymbol()];
            
            // save off the method call info inside
            prereq.setExtra(chained);
            
            // make this new AST the first child of the method AST
            prereq.setNextSibling(##.getFirstChild());
            ##.setFirstChild(prereq);
         }
         
         // either way we must save off our function object into the root
         // method node as the "extra" object
         ##.setExtra(func);
      }
   ;
   
/**
 * Tokenizes an arithmetic or logical expression in infix notation (input as
 * a stream of characters) into a stream of tokens suitable for the
 * <code>ExpressionParser</code>.
 * <p>
 * Each token has an integer token type by which the parser references and
 * matches tokens.  This is done to make a high performance parser. To make
 * this work, the lexer must create a token object when it matches a top
 * level rule, this token object includes the token type and the actual text
 * found in the original character stream.  The lexer's main job is to
 * lookahead into the character stream, switch into the top level rule that
 * matches some set of characters and when this is complete, to create a
 * valid token object and return this to the parser.  The lexer thus
 * implements a token stream interface from the parser's perspective.
 * <p>
 * <b>Please note that this is a generated file using ANTLR 2.7.4 and a
 * grammar specified in <code>expression.g</code>.</b>
 * <p>
 * Symbol resolution is the process by which the proper token type is  
 * assigned to a token created from a match to the generic symbol rule.
 * In the lexer, the only symbol resolution that can be done is to match
 * with reserved keywords.  Reserved keywords are given preference over 
 * other symbol types in any case where there is a conflict.
 * <p>
 * Other key design points:
 * <ul>
 *    <li> This grammar has been designed to ignore case.  All symbol
 *         definitions use lowercase characters and ANTLR generates code to
 *         match both the uppercase and lowercase version of each character.
 *         This is enabled using the lexer options section.
 *    <li> The lexer is defined to create a single token of any string.  All
 *         logic for matching string literals is embedded in the lexer.
 *    <li> At the top level of the lexer, all tokens are one of the
 *         following types: 
 *       <ul>
 *          <li> string literals
 *          <li> whitespace (spaces, tabs, carriage returns, line feeds)
 *          <li> operators
 *          <li> integer and decimal literals
 *          <li> reserved keywords 
 *          <li> user-defined symbols
 *       </ul>
 *    <li> The lexer's top level entry point is <code>nextToken</code>. This
 *         method has a for-loop that uses up to 3 characters of lookahead
 *         to identify which of the top level token rules to call. To do
 *         this, each token rule's left-most match characters are "rolled up"
 *         and tested in this top level rule. 
 *    <li> Whitespace tokens are artificially set to the special "skip" token
 *         type, which drops them from the token stream seen by the parser.
 *         This is an important feature that simplifies the parser design.
 * </ul>
 * <p>
 * All token types are defined as integer constants in 
 * <code>{@link ExpressionParserTokenTypes}</code> which is an interface that
 * the parser, lexer and other related classes all implement.  This allows 
 * all of these classes to directly refer to the token types and share this
 * common set of definitions.  All top-level lexer rules generate a token
 * of the same name in the <code>ExpressionParserTokenTypes</code> interface. 
 * There is a tokens { } section in the parser where artificial tokens are
 * defined (tokens that are not backed by a lexer rule).  These tokens are
 * also added to the <code>ExpressionParserTokenTypes</code> interface and
 * can thus be referenced directly by the lexer and parser.
 *
 * @author GES
 */
class ExpressionLexer extends Lexer;

options
{
   k = 3;                                 // lookahead depth is 3 chars
   charVocabulary = '\3'..'\377';         // specifies range 0x03-0xFF
   testLiterals = false;                  // do our own keyword processing
   caseSensitive = false;                 // be case-insensitive
   caseSensitiveLiterals = false;   
}

// helper members and methods
{
   /** Map of keywords to symbol tokens */
   private final static Map keywords = new HashMap();
   
   /** Positional parameter counter */
   private int posParmCount = 0;
   
   /** If <code>true</code>, convert assignment to equality operator */
   private boolean noAssign = false;
   
   // Static initializer which populates a keyword map where the symbol text
   // (in the mSYMBOL method) can be submitted to the keyword map for a
   // lookup. If a match is found, then the token type of the token is set
   // based on the token type field stored in the map. This token type will
   // override the default token type of SYMBOL.  Since all of these keywords
   // are not encoded in the matching rules of the grammar itself, there are
   // no actual token definitions in the lexer that match these keywords.
   static
   {
      keywords.put( "and",   new Integer(AND) );
      keywords.put( "false", new Integer(BOOL_FALSE) );
      keywords.put( "not",   new Integer(NOT) );
      keywords.put( "null",  new Integer(NULL) );
      keywords.put( "or",    new Integer(OR) );
      keywords.put( "true",  new Integer(BOOL_TRUE) );
   }
   
   public void setNoAssignment()
   {
      this.noAssign = true;
   }
   
   /**
    * Provides a command line interface for an end user to drive and/or test
    * the ExpressionLexer class.
    * <p>
    * Syntax:
    * <pre>
    *    java ExpressionLexer &lt;expression&gt;
    * </pre>
    *
    * @param   args 
    *          List of command line arguments.
    */
   public static void main(String[] args)  
   {
      int len = args.length;
      if (len < 1 || len > 2)
      {
         System.err.println(
            "Syntax: java ExpressionLexer <expression> [no_assignment]");
         System.exit(-1);
      }
         
      try
      {
         StringReader    expr       = new StringReader(args[0]);
         ExpressionLexer lexer      = new ExpressionLexer(expr);
         int             maxSymSize = 20;
         NumberFormat    lfmt       = NumberFormat.getInstance();
         NumberFormat    cfmt       = NumberFormat.getInstance();
         
         CommonToken     next;
         String          decode;
         
         lfmt.setMinimumIntegerDigits(5);
         lfmt.setGroupingUsed(false);
         cfmt.setMinimumIntegerDigits(3);
         cfmt.setGroupingUsed(false);            
         
         if (len > 1 && "true".equalsIgnoreCase(args[1]))
         {
            lexer.setNoAssignment();
         }
         
         do
         {
            next = (CommonToken) lexer.nextToken();
            
            decode = ExpressionParser._tokenNames[next.getType()];
            
            StringBuffer sb = new StringBuffer(decode).append('>');
            int trailing = maxSymSize - decode.length();
            
            for (int i = 0; i < trailing; i++)
            {
               sb.append(' ');
            }
            
            System.out.println("["                               +
                               lfmt.format( next.getLine() )     +
                               ":"                               +
                               cfmt.format( next.getColumn() )   +
                               "] <"                             +
                               sb.toString()                     +
                               " "                               +
                               next.getText());
         } 
         while (next.getType() != Token.EOF_TYPE);
      }
       
      catch(Exception excpt)
      {
         excpt.printStackTrace();
      }
   }   
}

/**
 * Matches any amount of whitespace in a program and sets the token type to
 * &quot;skip&quot;.  In addition, each newline character causes the lexer's
 * line counter to be incremented in order to properly maintain each token's 
 * line information.
 * <p>
 * Spaces, tabs and carriage returns and line feeds (newlines) are all
 * matched.
 * <p>
 * This is a top level lexer rule which means that there is an associated
 * <code>WS</code> token.
 */
WS
   :
      ( ' ' | '\t' | '\r' | '\n' { newline(); } )+
      { $setType( Token.SKIP ); }
   ;

/**
 * Match a valid symbol name. All symbols must start with an alphabetic
 * character or the "@" character or the "#" character or the "$" character.
 * Subsequent characters can be alphanumeric or one of the special symbol
 * characters (<code> _ # $ : . </code>).  
 * <b>Symbols are matched case-insensitively.</b>
 * <p>
 * Once a symbol has been found, a keyword lookup occurs. If matched, the
 * the token's type is overridden from the default (<code>SYMBOL</code>)
 * to the artificial token type associated with the keyword.
 * <p>
 * This is a top level lexer rule which means that there is an associated 
 * <code>SYMBOL</code> token.
 */
SYMBOL
   :
         POSPARM
         {
            StringBuilder buf = new StringBuilder("$");
            buf.append(++posParmCount);
            
            $setText(buf.toString());
         }
         
      |  LETTER (VALID_SYM_CHAR)*      
         {
            int stype = SYMBOL;
            
            Integer found = (Integer) keywords.get(getText());
            
            if (found != null)
            {
               // Symbol is a keyword; set the token type accordingly.
               stype = found.intValue();
            }
            
            $setType(stype);
         }
   ;   
   
/**
 * Matches the "=" character. This is a top level rule, creating an
 * <code>ASSIGN</code> token type, unless {@link #setNoAssignment()} has been
 * invoked, in which case an <code>EQUALS</code> token type is created.
 */
ASSIGN
   :
      "="
      {
         if (noAssign)
            $setType(EQUALS);
      }
   ;

/**
 * Matches the ',' character. This is a top level rule, creating a
 * <code>COMMA</code> token type.
 */
COMMA    : ',' ;

/**
 * Matches the "==" string. This is a top level rule, creating an
 * <code>EQUALS</code> token type.
 */
EQUALS   : "==" ;

/**
 * Matches the "||" string. This is a top level rule, creating an
 * <code>OR</code> token type.
 */
OR       : "||" ;

/**
 * Matches the "&amp;&amp;" string. This is a top level rule, creating an
 * <code>AND</code> token type.
 */
AND      : "&&" ;

/**
 * Matches the logical negation character. This is a top level rule, creating
 * a <code>NOT</code> token type.
 */
NOT      : '!';

/**
 * Matches the "&gt;&gt;" string. This is a top level rule, creating a
 * <code>RSHIFT</code> token type.
 */
RSHIFT   : ">>" ;

/**
 * Matches the "&lt;&lt;" string. This is a top level rule, creating a
 * <code>LSHIFT</code> token type.
 */
LSHIFT   : "<<" ;

/**
 * Matches the "&gt;&gt;&gt;" string. This is a top level rule, creating a
 * <code>ZRSHIFT</code> token type.
 */
ZRSHIFT  : ">>>" ;

/**
 * Matches the '&amp;' character. This is a top level rule, creating an
 * <code>BIT_AND</code> token type.
 */
BIT_AND  : '&' ;

/**
 * Matches the '|' character. This is a top level rule, creating an
 * <code>BIT_OR</code> token type.
 */
BIT_OR   : '|' ;

/**
 * Matches the '^' character. This is a top level rule, creating an
 * <code>BIT_XOR</code> token type.
 */
BIT_XOR  : '^' ;

/**
 * Matches the '~' character. This is a top level rule, creating an
 * <code>BIT_COMP</code> token type.
 */
BIT_COMP : '~' ;

/**
 * Matches the inequality string. This is a top level rule, creating a
 * <code>NOT_EQ</code> token type.
 */
NOT_EQ   : "!=" ;

/**
 * Matches the '&gt;' character. This is a top level rule, creating a
 * <code>GT</code> token type.
 */
GT       : '>' ;

/**
 * Matches the '&lt;' character. This is a top level rule, creating a
 * <code>LT</code> token type.
 */
LT       : '<' ;

/**
 * Matches the '&gt;=' character sequence. This is a top level rule, creating
 * a <code>GTE</code> token type.
 */
GTE      : ">=" ;

/**
 * Matches the '&lt;=' character sequence. This is a top level rule, creating
 * a <code>LTE</code> token type.
 */
LTE      : "<=" ;

/**
 * Matches the '(' character. This is a top level rule, creating a
 * <code>LPARENS</code> token type.
 */
LPARENS  : '(' ;

/**
 * Matches the ')' character. This is a top level rule, creating a
 * <code>RPARENS</code> token type.
 */
RPARENS  : ')' ;

/**
 * Matches the '+' character. This is a top level rule, creating a
 * <code>PLUS</code> token type.
 */
PLUS     : '+' ;

/**
 * Matches the '-' character. This is a top level rule, creating a
 * <code>MINUS</code> token type.
 */
MINUS    : '-' ;

/**
 * Matches the '*' character. This is a top level rule, creating a
 * <code>MULTIPLY</code> token type.
 */
MULTIPLY : '*' ;

/**
 * Matches the '/' character. This is a top level rule, creating a
 * <code>DIVIDE</code> token type.
 */
DIVIDE   : '/' ;

/**
 * Matches the '%' character. This is a top level rule, creating a
 * <code>MODULO</code> token type.
 */
MODULO   : '%' ;

/**
 * Matches the '#' character. This is a top level rule, creating a
 * <code>HASH</code> token type.
 */
HASH     : '#' ;

/**
 * Matches all forms of valid integer literals and decimal literals.
 * Scenarios that are matched:
 * <ul>
 *    <li> Matches a sequence of one or more digits (without a decimal point
 *         and additional digits) as an integer literal with a
 *         <code>NUM_LITERAL</code> token type.
 *    <li> Matches a sequence of '0x' followed by one or more hexidecimal
 *         digits (0 through 9 and A through F) as a hexidecimal literal of
 *         <code>HEX_LITERAL</code> token type.
 *    <li> Matches a sequence of one or more digits followed by a '.' but
 *         no additional digits, as a decimal literal with a
 *         <code>DEC_LITERAL</code> token type. 
 *    <li> Matches one or more numeric digits followed by a decimal point
 *         and another sequence of numeric digits as a decimal constant 
 *         with a <code>DEC_LITERAL</code> token type.
 *    <li> Matches a decimal point followed by a sequence of numeric digits
 *         as a decimal constant with a <code>DEC_LITERAL</code> token type.
 * </ul>
 * <p>
 * This is a top level rule.  The token type defaults to 
 * <code>NUM_LITERAL</code> and is overridden by specific actions depending
 * on the scenario matched.
 */
NUM_LITERAL
   :
      {
         int ntype = NUM_LITERAL;
      }
      (  
           '.' (DIGIT)+              { ntype = DEC_LITERAL; }
         | '.'                       { ntype = DOT;         }
         | "0x"! (DIGIT | HEXDIGIT)+ { ntype = HEX_LITERAL; }
         | (DIGIT)+ ( '.' (DIGIT)*   { ntype = DEC_LITERAL; } )?
         
      )
      {
         $setType(ntype);
      }
   ;
   
/**
 * Matches any string literal which is enclosed in a pair of single quotes
 * or a pair of double quotes.  This is a top-level lexer rule.
 */
STRING   
   :
      (SSTRING | DSTRING)
   ;
   
/**   
 * Matches any single character that can appear in the 2nd or later position 
 * in a symbol (matches any <code>LETTER, DIGIT or SYM_CHAR</code>). This is
 * simply a helper rule to make references easier.
 */
protected   
VALID_SYM_CHAR
   :
      LETTER | DIGIT | SYM_CHAR
   ;
   
/**
 * Matches an opening single quote, arbitrary contents and an ending single
 * quote.  Two continguous single quote characters are used to specify a
 * one single quote in the output string (it does not terminate the
 * string).
 * <p>
 * Any newlines inside the string are identified and the lexer's internal
 * newline counter is properly maintained.  The newlines are not maintained
 * in the output string.
 * <p>
 * The greedy option does not need to be disabled here as the closure rule
 * termination is built into the subrule itself: it accepts anything that
 * isn't an unescaped single quote character (see above).
 * <p>
 * Tabs and spaces are maintained inside strings. Opening and closing single
 * quote characters are dropped.
 */
protected
SSTRING
   :  (
         '\''!                                // opening '
              (
                   "\'\'"!  { append('\''); } // allow '', convert to a '
                 | '\n'!    { newline(); }    // allow NL
                 | ~'\''                      // anything except closing '
              )*
         '\''!                                // closing '
      )
   ;

/**
 * Matches an opening double quote, arbitrary contents and an ending double
 * quote.  Two continguous double quote characters are used to specify a
 * single double quote in the output string (it does not terminate the
 * string).
 * <p>
 * Any newlines inside the string are identified and the lexer's internal
 * newline counter is properly maintained.  The newlines are not maintained
 * in the output string.
 * <p>
 * The greedy option does not need to be disabled here as the closure rule
 * termination is built into the subrule itself: it accepts anything that
 * isn't an unescaped double quote character (see above).
 * <p>
 * Tabs and spaces are maintained inside strings. Opening and closing double
 * quote characters are dropped.
 */
protected
DSTRING
   :  (
         '\"'!                                // opening "
              (
                   "\"\""!  { append('\"'); } // allow "", convert to a "
                 | '\n'!    { newline(); }    // allow NL
                 | ~'\"'                      // anything except closing "
              )*
         '\"'!                                // closing "
      )
   ;

/** Matches any alphabetic character: <code>a - z</code> */   
protected   
LETTER   : 'a'..'z';
   
/** Matches any numeric digit: <code>0 - 9</code> */   
protected
DIGIT    : '0'..'9';

/** Matches any valid letter that can be used in a hexidecimal number. */
protected
HEXDIGIT : 'a'..'f';

/**
 * Matches all characters that can be made part of a valid user-defined
 * symbol (except alphabetic and numeric characters).  These are:
 * <code>$ _</code>
 * <p>
 * Note that this intentionally matches the normal Java symbol rules.
 */
protected
SYM_CHAR : '_' | '$';

/**
 * Matches a positional parameter placeholder.
 */
protected
POSPARM  : '?' ;