/*
** Module   : UsersUIHandlers.java
** Abstract : UsersUIHandlers defines common business methods to manage user accounts.
**
** Copyright (c) 2017, Golden Code Development Corporation.
**
** -#- -I- --Date-- ----------------Description----------------------
** 001 SBI 20170601 Created initial version.
*/
/*
** This program is free software: you can redistribute it and/or modify
** it under the terms of the GNU Affero General Public License as
** published by the Free Software Foundation, either version 3 of the
** License, or (at your option) any later version.
**
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
** GNU Affero General Public License for more details.
**
** You may find a copy of the GNU Affero GPL version 3 at the following
** location: https://www.gnu.org/licenses/agpl-3.0.en.html
** 
** Additional terms under GNU Affero GPL version 3 section 7:
** 
**   Under Section 7 of the GNU Affero GPL version 3, the following additional
**   terms apply to the works covered under the License.  These additional terms
**   are non-permissive additional terms allowed under Section 7 of the GNU
**   Affero GPL version 3 and may not be removed by you.
** 
**   0. Attribution Requirement.
** 
**     You must preserve all legal notices or author attributions in the covered
**     work or Appropriate Legal Notices displayed by works containing the covered
**     work.  You may not remove from the covered work any author or developer
**     credit already included within the covered work.
** 
**   1. No License To Use Trademarks.
** 
**     This license does not grant any license or rights to use the trademarks
**     Golden Code, FWD, any Golden Code or FWD logo, or any other trademarks
**     of Golden Code Development Corporation. You are not authorized to use the
**     name Golden Code, FWD, or the names of any author or contributor, for
**     publicity purposes without written authorization.
** 
**   2. No Misrepresentation of Affiliation.
** 
**     You may not represent yourself as Golden Code Development Corporation or FWD.
** 
**     You may not represent yourself for publicity purposes as associated with
**     Golden Code Development Corporation, FWD, or any author or contributor to
**     the covered work, without written authorization.
** 
**   3. No Misrepresentation of Source or Origin.
** 
**     You may not represent the covered work as solely your work.  All modified
**     versions of the covered work must be marked in a reasonable way to make it
**     clear that the modified work is not originating from Golden Code Development
**     Corporation or FWD.  All modified versions must contain the notices of
**     attribution required in this license.
*/


package com.goldencode.p2j.admin.client.application.home.accounts.users;

import java.io.Serializable;
import java.util.Set;

import com.goldencode.p2j.admin.TaggedName;
import com.goldencode.p2j.admin.UserDef;
import com.goldencode.p2j.admin.client.application.home.accounts.AccountsUiHandlers;


/**
 * UsersUIHandlers defines common business methods to manage user accounts. This interface is
 * implemented by users controller UsersPresenter and is intended to be used by views.
 */
public interface UsersUIHandlers
extends AccountsUiHandlers
{
   /**
    * Fills and/or updates the all groups accounts table.
    */
   void populateAndRefreshAllGroups();
   
   /**
    * Triggers an action to refresh the all users table.
    */
   void onRefresh();
   
   /**
    * Triggers the users selection changed event.
    * 
    * @param    selectedUsers
    *           The selected users accounts
    */
   void onUserSelected(Set<TaggedName> selectedUsers);
   
   /**
    * Triggers the groups selection changed event.
    * 
    * @param    selectedGroups
    *           The selected groups accounts
    */
   void onGroupSelected(Set<TaggedName> selectedGroups);
   
   /**
    * Adds a new user.
    * 
    * @param    newUser
    *           The new user definition
    * @param    ext
    *           The extended account data.
    */
   void addUser(UserDef newUser, Serializable ext);
   
   /**
    * Updates a given user account.
    * 
    * @param    user
    *           The given user definition
    * @param    ext
    *           The given extended account data
    */
   void updateUser(UserDef user, Serializable ext);
   
   /**
    * Deletes the given selected users accounts.
    * 
    * @param    users
    *           The array of users accounts given by tagged names
    */
   void deleteSelectedUsers(TaggedName[] users);
   
   /**
    * Clones the target user account with its ACL if it is required preserving the same password
    * for a new created account.
    * 
    * @param    nameFrom
    *           The cloned user account name
    * @param    nameTo
    *           The new user account name
    * @param    acls
    *           True indicates that the ACL of the cloned user will be copied too, otherwise
    *           the ACL of the new created user account will be empty.
    */
   void cloneUserAccount(String nameFrom, String nameTo, boolean acls);
   
   /**
    * Clones the target user account with its ACL if it is required and having new password.
    * 
    * @param    nameFrom
    *           The cloned user account name
    * @param    nameTo
    *           The new user account name
    * @param    acls
    *           True indicates that the ACL of the cloned user will be copied too, otherwise
    *           the ACL of the new created user account will be empty.
    * @param    newPassword
    *           The password for the new created account
    */
   void cloneUserAccount(String nameFrom, String nameTo, boolean acls, String newPassword);
   
   /**
    * Makes a given user account to be a member of given groups accounts.
    * 
    * @param    user
    *           The target user account given by its tagged name
    * @param    groups
    *           The array of target groups accounts given by tagged names
    */
   void addUserToGroups(TaggedName user, TaggedName[] groups);
   
   /**
    * Removes a given user account from given groups accounts.
    * 
    * @param    user
    *           The target user account given by its tagged name
    * @param    groups
    *           The array of target groups accounts given by tagged names
    */
   void removeUserFromGroups(TaggedName user, TaggedName[] groups);
   
   /**
    * Makes a given user account to be a member of a given group account.
    * 
    * @param    user
    *           The target user account given by its tagged name
    * @param    groupName
    *           The target group account name
    */
   void addUserInNewGroup(TaggedName user, String groupName);
   
   /**
    * Displays only groups accounts that satisfy the given search criteria.
    * 
    * @param    filter
    *           The given search criteria
    */
   void setGroupsFilter(String filter);
   
   /**
    * Disables the target user account.
    * 
    * @param    user
    *           The target user account given by its tagged name
    */
   void disableUserAccount(TaggedName user);
   
   /**
    * Shows the Add User Account Dialog.
    */
   void showAddUserAccountDialog();
   
   /**
    * Setups the Find User Account Dialog.
    */
   void setupFindUserAccountDialog();
   
   /**
    * Setups the Deleting Users Alert.
    * 
    * @param    users
    *           The array of users accounts given by tagged names
    */
   void setupDeletingUsersAlert(TaggedName[] users);
   
   /**
    * Setups the Clone User Account Dialog.
    * 
    * @param    clonedUserAccount
    *           The cloned user account given by its tagged name
    */
   void setupCloneUserAccountDialog(TaggedName clonedUserAccount);
   
   /**
    * Shows the Edit User Account Dialog to edit the target user account.
    * 
    * @param    updatedUserAccount
    *           The target user account given by its tagged name
    */
   void showEditUserAccountDialog(TaggedName updatedUserAccount);
   
   /**
    * Setups the Removing Groups Alert for deleting a given user account from selected
    * groups accounts.
    * 
    * @param    user
    *           The given user account
    * @param    groups
    *           The selected groups accounts from which the given user is about to be deleted
    */
   void setupRemovingGroupsAlert(TaggedName user, TaggedName[] groups);

   /**
    * Cancels the Add User Account Dialog and rolls back any related changes.
    */
   void cancelAddUserAccountDialog();
   
   /**
    * Cancels the Deleting Users Alert and rolls back any related changes.
    */
   void cancelDeletingUsersAlert();
   
   /**
    * Cancels the Clone User Account Dialog and rolls back any related changes.
    */
   void cancelCloneUserAccountDialog();
   
   /**
    * Cancels the Edit User Account Dialog and rolls back any related changes.
    */
   void cancelEditUserAccountDialog();
   
   /**
    * Cancels the Find User Account Dialog and rolls back any related changes.
    */
   void cancelFindUserAccountDialog();
   
   /**
    * Cancels the Removing Groups Alert and rolls back any related changes.
    */
   void cancelRemovingGroupsAlert();
}
