/*
** Module   : AddRecordLockView.java
** Abstract : AddRecordLockView represents a dialog to add a new record to be acquired by a lock.
**
** Copyright (c) 2017, Golden Code Development Corporation.
**
** -#- -I- --Date-- ----------------Description----------------------
** 001 SBI 20170601 Created initial version.
*/
/*
** This program is free software: you can redistribute it and/or modify
** it under the terms of the GNU Affero General Public License as
** published by the Free Software Foundation, either version 3 of the
** License, or (at your option) any later version.
**
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
** GNU Affero General Public License for more details.
**
** You may find a copy of the GNU Affero GPL version 3 at the following
** location: https://www.gnu.org/licenses/agpl-3.0.en.html
** 
** Additional terms under GNU Affero GPL version 3 section 7:
** 
**   Under Section 7 of the GNU Affero GPL version 3, the following additional
**   terms apply to the works covered under the License.  These additional terms
**   are non-permissive additional terms allowed under Section 7 of the GNU
**   Affero GPL version 3 and may not be removed by you.
** 
**   0. Attribution Requirement.
** 
**     You must preserve all legal notices or author attributions in the covered
**     work or Appropriate Legal Notices displayed by works containing the covered
**     work.  You may not remove from the covered work any author or developer
**     credit already included within the covered work.
** 
**   1. No License To Use Trademarks.
** 
**     This license does not grant any license or rights to use the trademarks
**     Golden Code, FWD, any Golden Code or FWD logo, or any other trademarks
**     of Golden Code Development Corporation. You are not authorized to use the
**     name Golden Code, FWD, or the names of any author or contributor, for
**     publicity purposes without written authorization.
** 
**   2. No Misrepresentation of Affiliation.
** 
**     You may not represent yourself as Golden Code Development Corporation or FWD.
** 
**     You may not represent yourself for publicity purposes as associated with
**     Golden Code Development Corporation, FWD, or any author or contributor to
**     the covered work, without written authorization.
** 
**   3. No Misrepresentation of Source or Origin.
** 
**     You may not represent the covered work as solely your work.  All modified
**     versions of the covered work must be marked in a reasonable way to make it
**     clear that the modified work is not originating from Golden Code Development
**     Corporation or FWD.  All modified versions must contain the notices of
**     attribution required in this license.
*/


package com.goldencode.p2j.admin.client.application.home.console.acquiredlocks;

import java.util.ArrayList;
import java.util.List;
import java.util.logging.Level;

import org.gwtbootstrap3.client.ui.Button;
import org.gwtbootstrap3.client.ui.ListBox;
import org.gwtbootstrap3.client.ui.Modal;
import org.gwtbootstrap3.client.ui.TextBox;
import org.gwtbootstrap3.client.ui.form.error.BasicEditorError;
import org.gwtbootstrap3.client.ui.form.validator.Validator;

import com.goldencode.p2j.admin.client.widget.dialog.ModalViewWithUiHandlers;
import com.google.gwt.editor.client.Editor;
import com.google.gwt.editor.client.EditorError;
import com.google.gwt.event.dom.client.ChangeEvent;
import com.google.gwt.event.dom.client.ChangeHandler;
import com.google.gwt.event.dom.client.ClickEvent;
import com.google.gwt.uibinder.client.UiBinder;
import com.google.gwt.uibinder.client.UiField;
import com.google.gwt.uibinder.client.UiHandler;
import com.google.gwt.user.client.ui.Widget;
import com.google.inject.Inject;


/**
 * AddRecordLockView represents a dialog to add a new record to be acquired by a lock.
 */
public class AddRecordLockView extends ModalViewWithUiHandlers<AcquiredLocksUIHandler>
{
   /**
    * GWT UI creator
    */
   interface Binder
   extends UiBinder<Widget, AddRecordLockView>
   {}
   
   /** The underlined modal dialog */
   @UiField
   Modal modal;
   
   /** The record id text box */
   @UiField
   TextBox recordIdTextBox;
   
   /** The databases list box */
   @UiField
   ListBox databaseListBox;
   
   /** The tables list box */
   @UiField
   ListBox tableListBox;
   
   /** The Add button */
   @UiField
   Button addAction;
   
   /** The Cancel button */
   @UiField
   Button cancelAction;
   
   /**
    * Constructs this view, used by MPV gwtplatform of ArcBees Inc.
    * 
    * @param    binder
    *           UI creator
    */
   @Inject
   public AddRecordLockView(AddRecordLockView.Binder binder)
   {
      initWidget(binder.createAndBindUi(this));

      databaseListBox.addChangeHandler(new ChangeHandler()
      {
         @Override
         public void onChange(ChangeEvent event)
         {
            String database = databaseListBox.getSelectedValue();
            if (database != null)
            {
               getUiHandlers().refreshTables(database);
            }
         }
      });
      
      recordIdTextBox.addValidator(new Validator(){

         @Override
         public List<EditorError> validate(Editor editor, Object value)
         {
            List<EditorError> result = new ArrayList<EditorError>();
            try
            {
               Long.parseLong(value.toString());
            }
            catch(NumberFormatException ex)
            {
               result.add(new BasicEditorError(recordIdTextBox, value, "Must be the whole number."));
            }
            return result;
         }

         @Override
         public int getPriority()
         {
            return 0;
         }
      });
      
      addShowHandler(()-> {
         recordIdTextBox.clear();
      });
   }
   
   /**
    * The Add button click handler.
    * 
    * @param    e
    *           Click event
    */
   @UiHandler("addAction")
   void onSaveAction(ClickEvent e)
   {
      logger.log(Level.INFO, "onAddAction");
      if (recordIdTextBox.validate(true))
      {
         Long recID = Long.parseLong(recordIdTextBox.getValue());
         getUiHandlers().addRecordToLocks(databaseListBox.getSelectedValue(),
                  tableListBox.getSelectedValue(), recID);
      }
   }
   
   /**
    * The Cancel button click handler.
    * 
    * @param    e
    *           Click event
    */
   @UiHandler("cancelAction")
   void onCancelAction(ClickEvent e)
   {
      logger.log(Level.INFO, "onCancelAction");
   }
   
   /**
    * Sets the databases model.
    * 
    * @param    databases
    *           The array of databases
    */
   public void setDatabaseList(String[] databases)
   {
      if (databases == null)
      {
         return;
      }
      databaseListBox.clear();
      for (String databaseName : databases)
      {
         databaseListBox.addItem(databaseName, databaseName);
      }
      if (databases.length > 0)
      {
         databaseListBox.setItemSelected(0, true);
         // change event isn't fired here
         getUiHandlers().refreshTables(databases[0]);
      }
   }
   
   /**
    * Sets the tables model.
    * 
    * @param    tables
    *           The array of tables
    */
   public void setTableList(String[] tables)
   {
      if (tables == null)
      {
         return;
      }
      tableListBox.clear();
      for (String tableName : tables)
      {
         tableListBox.addItem(tableName, tableName);
      }
      if (tables.length > 0)
      {
         tableListBox.setItemSelected(0, true);
      }
   }
   
   /**
    * Provides an access to the target modal dialog.
    * 
    * @return   The underlined modal dialog
    */
   @Override
   protected Modal getModal()
   {
      return modal;
   }
}
