/*
** Module   : LoginPresenter.java
** Abstract : Login presenter.
**
** Copyright (c) 2017, Golden Code Development Corporation.
**
** -#- -I- --Date-- --------------------------------Description-----------------------------------
** 001 HC  20170612 Initial version.
*/
/*
** This program is free software: you can redistribute it and/or modify
** it under the terms of the GNU Affero General Public License as
** published by the Free Software Foundation, either version 3 of the
** License, or (at your option) any later version.
**
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
** GNU Affero General Public License for more details.
**
** You may find a copy of the GNU Affero GPL version 3 at the following
** location: https://www.gnu.org/licenses/agpl-3.0.en.html
**
** Additional terms under GNU Affero GPL version 3 section 7:
**
**   Under Section 7 of the GNU Affero GPL version 3, the following additional
**   terms apply to the works covered under the License.  These additional terms
**   are non-permissive additional terms allowed under Section 7 of the GNU
**   Affero GPL version 3 and may not be removed by you.
**
**   0. Attribution Requirement.
**
**     You must preserve all legal notices or author attributions in the covered
**     work or Appropriate Legal Notices displayed by works containing the covered
**     work.  You may not remove from the covered work any author or developer
**     credit already included within the covered work.
**
**   1. No License To Use Trademarks.
**
**     This license does not grant any license or rights to use the trademarks
**     Golden Code, FWD, any Golden Code or FWD logo, or any other trademarks
**     of Golden Code Development Corporation. You are not authorized to use the
**     name Golden Code, FWD, or the names of any author or contributor, for
**     publicity purposes without written authorization.
**
**   2. No Misrepresentation of Affiliation.
**
**     You may not represent yourself as Golden Code Development Corporation or FWD.
**
**     You may not represent yourself for publicity purposes as associated with
**     Golden Code Development Corporation, FWD, or any author or contributor to
**     the covered work, without written authorization.
**
**   3. No Misrepresentation of Source or Origin.
**
**     You may not represent the covered work as solely your work.  All modified
**     versions of the covered work must be marked in a reasonable way to make it
**     clear that the modified work is not originating from Golden Code Development
**     Corporation or FWD.  All modified versions must contain the notices of
**     attribution required in this license.
*/

package com.goldencode.p2j.admin.client.application.login;

import com.goldencode.p2j.admin.client.*;
import com.goldencode.p2j.admin.client.application.*;
import com.goldencode.p2j.admin.shared.*;
import com.goldencode.p2j.security.*;
import com.google.gwt.user.client.*;
import com.google.gwt.user.client.rpc.*;
import com.google.inject.*;
import com.google.web.bindery.event.shared.*;
import com.gwtplatform.mvp.client.*;
import com.gwtplatform.mvp.client.annotations.*;
import com.gwtplatform.mvp.client.proxy.*;
import com.gwtplatform.mvp.shared.proxy.*;

import java.util.logging.*;

/**
 * Login presenter.
 */
public class LoginPresenter
extends Presenter<LoginPresenter.MyView, LoginPresenter.MyProxy>
implements LoginUiHandlers
{
   /**
    * The presenter's proxy.
    */
   @ProxyStandard
   @NameToken(NameTokens.LOGIN)
   @NoGatekeeper
   public interface MyProxy extends ProxyPlace<LoginPresenter>
   {
   }

   /**
    * The presenter's view.
    */
   public interface MyView extends View, HasUiHandlers<LoginUiHandlers>
   {
      /**
       * Reports login error to the user.
       *
       * @param   code
       *          Login error code.
       */
      void reportLoginError(int code);

      /**
       * Reports login failure to the user.
       */
      void reportFailure();

      /**
       * Clears the input password.
       */
      void clearPassword();
   }

   /** Logger */
   private static Logger logger = Logger.getLogger(LoginPresenter.class.getSimpleName());

   /** Place manager reference */
   private final PlaceManager placeManager;

   /** Authentication service reference */
   private final AuthServiceAsync authService;

   /** Admin service reference */
   private final AdminServiceAsync adm;

   /** Alarm service */
   private final Alarm alarm;

   /**
    * Ctor.
    *
    * @param   eventBus
    *          Event bus reference.
    * @param   view
    *          Presenter's view.
    * @param   proxy
    *          Presenter's proxy.
    * @param   placeManager
    *          Place manager reference.
    * @param   authService
    *          Authentication service reference.
    * @param   adm
    *          Admin service reference.
    * @param   alarm
    *          Alarm UI service reference.
    */
   @Inject
   LoginPresenter(EventBus eventBus,
                  MyView view,
                  MyProxy proxy,
                  PlaceManager placeManager,
                  com.goldencode.p2j.admin.shared.AuthServiceAsync authService,
                  com.goldencode.p2j.admin.shared.AdminServiceAsync adm,
                  Alarm alarm)
   {
      super(eventBus, view, proxy, ApplicationPresenter.SLOT_MAIN);
      this.placeManager = placeManager;
      this.authService = authService;
      this.adm = adm;
      this.alarm = alarm;
      getView().setUiHandlers(this);
   }

   /**
    * Called when the user submits the authentication information.
    *
    * @param   userid
    *          User id.
    * @param   pwd
    *          Password.
    */
   @Override
   public void onSubmitLoginInfo(String userid, String pwd)
   {
      authService.login(userid, pwd, new AsyncCallback<Integer>()
      {
         @Override
         public void onFailure(Throwable throwable)
         {
            getView().clearPassword();
            getView().reportFailure();
         }

         @Override
         public void onSuccess(Integer res)
         {
            getView().clearPassword();

            // dump all messages to the output log, this also clears all the messages
            // so any subsequent ring will only show relevant messages
            alarm.getAllMessages(msg ->
            {
               logger.info("Server messages after login:");
               StringBuilder sb = new StringBuilder();
               for (String m : msg)
               {
                  if (sb.length() > 0)
                  {
                     sb.append("\n");
                  }
                  sb.append(m);
               }

               if (sb.length() > 0)
               {
                  logger.info(sb.toString());
               }
            });

            if (res == SecurityConstants.AUTH_RESULT_SUCCESS)
            {
               adm.getServerName(new AdminCallback<String>(adm, alarm)
               {
                  @Override
                  public void onDone(String result)
                  {
                     if (result == null)
                     {
                        return;
                     }

                     ClientContext.instance().reset(userid, result);

                     PlaceRequest placeRequest = new PlaceRequest.Builder()
                        .nameToken(NameTokens.HOME)
                        .build();
                     placeManager.revealPlace(placeRequest);
                  }
               });
            }
            else
            {
               getView().reportLoginError(res);
            }
         }
      });
   }
}
