/*
** Module   : ReportServlet.java
** Abstract : ReportServlet provides report service for external clients.
**
** Copyright (c) 2017, Golden Code Development Corporation.
**
** -#- -I- --Date-- ----------------Description----------------------
** 001 SBI 20170601 Created initial version.
**
*/
/*
** This program is free software: you can redistribute it and/or modify
** it under the terms of the GNU Affero General Public License as
** published by the Free Software Foundation, either version 3 of the
** License, or (at your option) any later version.
**
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
** GNU Affero General Public License for more details.
**
** You may find a copy of the GNU Affero GPL version 3 at the following
** location: https://www.gnu.org/licenses/agpl-3.0.en.html
** 
** Additional terms under GNU Affero GPL version 3 section 7:
** 
**   Under Section 7 of the GNU Affero GPL version 3, the following additional
**   terms apply to the works covered under the License.  These additional terms
**   are non-permissive additional terms allowed under Section 7 of the GNU
**   Affero GPL version 3 and may not be removed by you.
** 
**   0. Attribution Requirement.
** 
**     You must preserve all legal notices or author attributions in the covered
**     work or Appropriate Legal Notices displayed by works containing the covered
**     work.  You may not remove from the covered work any author or developer
**     credit already included within the covered work.
** 
**   1. No License To Use Trademarks.
** 
**     This license does not grant any license or rights to use the trademarks
**     Golden Code, FWD, any Golden Code or FWD logo, or any other trademarks
**     of Golden Code Development Corporation. You are not authorized to use the
**     name Golden Code, FWD, or the names of any author or contributor, for
**     publicity purposes without written authorization.
** 
**   2. No Misrepresentation of Affiliation.
** 
**     You may not represent yourself as Golden Code Development Corporation or FWD.
** 
**     You may not represent yourself for publicity purposes as associated with
**     Golden Code Development Corporation, FWD, or any author or contributor to
**     the covered work, without written authorization.
** 
**   3. No Misrepresentation of Source or Origin.
** 
**     You may not represent the covered work as solely your work.  All modified
**     versions of the covered work must be marked in a reasonable way to make it
**     clear that the modified work is not originating from Golden Code Development
**     Corporation or FWD.  All modified versions must contain the notices of
**     attribution required in this license.
*/


package com.goldencode.p2j.admin.server;

import java.io.*;
import java.util.*;
import java.util.zip.*;

import javax.servlet.*;
import javax.servlet.http.*;

import com.goldencode.p2j.admin.server.reports.ReportBuilder.*;
import com.goldencode.p2j.admin.server.reports.*;
import com.goldencode.p2j.admin.shared.*;

import de.intarsys.pdf.content.CSContent;
import de.intarsys.pdf.cos.COSObject;
import de.intarsys.pdf.parser.COSLoadException;
import de.intarsys.pdf.pd.PDDocument;
import de.intarsys.pdf.pd.PDPage;
import de.intarsys.pdf.pd.PDPageTree;
import de.intarsys.pdf.pd.PDResources;
import de.intarsys.tools.locator.ByteArrayLocator;
import de.intarsys.tools.locator.FileLocator;
import de.intarsys.tools.locator.ILocator;


/**
 * Report servlet handles print requests. Accepted report parameters are listed by
 * this enumeration ReportParameters.
 */
public class ReportServlet extends HttpServlet
{
   /** The buffer size to hold responses */
   private static final int BUFFER_SIZE = 102400;
   
   /**
    * Handles HTTP GET requests to build reports.
    * 
    * @param    req
    *           The http request
    * @param    resp
    *           The http response
    */
   @Override
   protected void doGet(HttpServletRequest req, HttpServletResponse resp)
   throws ServletException, IOException
   {
      
      Enumeration<String> parameterNames = req.getParameterNames();
      
      ReportRequest reportRequest = new ReportRequest();
      
      LinkedHashMap<String, String> reportParameters = new LinkedHashMap<>();
      
      LinkedHashSet<String> selectedRowsList = new LinkedHashSet<String>();
      
      boolean printAction = false;
      
      while(parameterNames.hasMoreElements())
      {
         String name = parameterNames.nextElement();
         String value = req.getParameter(name);
         if (ReportParameters.REPORT_ID.getParameter().equals(name))
         {
            reportRequest.setReportId(value);
         }
         else if (ReportParameters.FILTER.getParameter().equals(name))
         {
            reportRequest.setFilter(value);
         }
         else if (ReportParameters.PAPER_FORMAT.getParameter().equals(name) ||
                  ReportParameters.PAPER_ORIENT.getParameter().equals(name) ||
                  ReportParameters.DETAILS.getParameter().equals(name) ||
                  ReportParameters.REPORT_TYPE.getParameter().equals(name) ||
                  ReportParameters.ACTION.getParameter().equals(name)
                     && (/*set print action */ printAction = true))
         {
            reportParameters.put(name, value);
         }
         else if (name.startsWith(ReportParameters.SELECTION.getParameter()))
         {
            if (!value.isEmpty())
            {
               selectedRowsList.add(value);
            }
         }
         else if (name.startsWith(ReportParameters.EXTENSION.getParameter()))
         {
            reportParameters.put(name, value);
         }
      };

      reportRequest.setReportParameters(reportParameters);
      
      reportRequest.setSelectedRows(selectedRowsList.toArray(new String[selectedRowsList.size()]));
      
      try (Document doc = ReportsManager.buildReport(reportRequest))
      {
         int parts = doc.getSplittedContent().size();
         
         if (parts > 1)
         {
//            writeSplittedDocument(doc, resp);
            mergeSplitDocuments(doc, resp, printAction);
         }
         else
         {
            writeOriginalDocument(doc, resp, printAction);
         }
      }
   }

   /**
    * Writes split content of the target document in a one zip archive.
    * 
    * @param    doc
    *           The document with split content
    * @param    resp
    *           Provides output stream
    * 
    * @throws   IOException
    *           The exception if there is a failed output operation.
    */
   private void writeSplittedDocument(Document doc, HttpServletResponse resp)
   throws IOException
   {
      int parts = doc.getSplittedContent().size();
      String fileName = doc.getDocumentFileName();
      String ext = doc.getDocumentType().getExtension();
      int index = fileName.lastIndexOf("." + ext);
      if (index != -1)
      {
         fileName = fileName.substring(0, index);
      }
      resp.setContentType("Content-type: application/zip");
      resp.setHeader("Content-Disposition", "attachment; filename=" + fileName + ".zip");
      // zip content
      ServletOutputStream outputStream = resp.getOutputStream();
      ZipOutputStream zos = new ZipOutputStream(new BufferedOutputStream(outputStream));
      for (int i = 0; i < parts; i++)
      {
         zos.putNextEntry(new ZipEntry(fileName + "_part_" + i + "." + ext));
         ByteArrayInputStream bis = new ByteArrayInputStream(doc.getSplittedContent().get(i));
         BufferedInputStream bus = new BufferedInputStream(bis);

         // Write splitted content
         int data = 0;
         while ((data = bus.read()) != -1)
         {
            zos.write(data);
         }
         bus.close();

         zos.closeEntry();
      }
      
      zos.close();
   }
   
   /**
    * Writes the target document in the http output stream preserving its content type.
    * 
    * @param    doc
    *           The target document
    * @param    resp
    *           Provides output stream
    * @param    printAction
    *           The true value indicates a print request, otherwise a download file request.
    * 
    * @throws   IOException
    *           The exception if there is a failed output operation.
    */
   private void writeOriginalDocument(Document doc,
                                      HttpServletResponse resp,
                                      boolean printAction)
   throws IOException
   {
      byte[] binaryReport = doc.getBinaryContent();
      DocumentType docType = doc.getDocumentType();
      String fileName = doc.getDocumentFileName();
      if (binaryReport != null && docType != null && fileName != null)
      {
         resp.setContentType(docType.getMimeType());
         if (printAction)
         {
            resp.setHeader( "Content-Disposition:", "inline");
         }
         else
         {
            resp.setHeader( "Content-Disposition:", "attachment;filename=" + fileName);
         }
         resp.setContentLength(binaryReport.length);
         resp.setBufferSize(BUFFER_SIZE);
         ServletOutputStream outputStream = resp.getOutputStream();
         outputStream.write(binaryReport);
         outputStream.flush();
      }
   }
   
   /**
    * Writes split content of the target document in a one merged pdf file.
    * 
    * @param    doc
    *           The document with split content
    * @param    resp
    *           Provides output stream
    * @param    printAction
    *           The true value indicates a print request, otherwise a download file request.
    *           
    * @throws   IOException
    *           The exception if there is a failed output operation.
    */
   private void mergeSplitDocuments(Document doc, HttpServletResponse resp, boolean printAction)
   throws IOException
   {
      PDDocument desPDDoc = null;
      PDDocument partialContentDoc = null;
      File mergedDoc = null;
      
      int parts = doc.getSplittedContent().size();
      String fileName = doc.getDocumentFileName();
      DocumentType docType = doc.getDocumentType();
      String ext = docType.getExtension();
      int index = fileName.lastIndexOf("." + ext);
      if (index != -1)
      {
         fileName = fileName.substring(0, index);
      }

      //adding the source files
      try
      {
         desPDDoc = PDDocument.createNew();
         
         long startMerge = System.currentTimeMillis();
         
         for (int i = 0; i < parts; i++)
         {
            long startMergePart = System.currentTimeMillis();
            ILocator contentLocator = new ByteArrayLocator(doc.getSplittedContent().get(i),
                                                           fileName + "_part_" + i + "." + ext);
            partialContentDoc = PDDocument.createFromLocator(contentLocator);
            PDPageTree pageTree = partialContentDoc.getPageTree();
            int count = pageTree.getCount();
            for(int k = 0; k < count; k++)
            {
               PDPage mergedPage = (PDPage) PDPage.META.createNew();
               PDPage copiedPage = pageTree.getPageAt(k);
               CSContent content =  copiedPage.getContentStream();
               COSObject resources = copiedPage.getResources().cosGetObject().copyDeep();
               PDResources pdResources = (PDResources) PDResources.META.createFromCos(resources);
               mergedPage.setResources(pdResources);
               mergedPage.setContentStream(content);
               desPDDoc.addPageNode(mergedPage);
            }
            System.err.println(contentLocator.getFullName() + " took " + (System.currentTimeMillis() - startMergePart) + "ms");
            partialContentDoc.close();
         }
         
         System.err.println(fileName + " took " + (System.currentTimeMillis() - startMerge) + "ms");
         
         mergedDoc = File.createTempFile(fileName, ext);
         
         mergedDoc.deleteOnExit();
         
         desPDDoc.save(new FileLocator(mergedDoc));
         
         // Free the system resources before 
         doc.close();
         
         long startDeliver = System.currentTimeMillis();
         
         resp.setContentType(docType.getMimeType());
         if (printAction)
         {
            resp.setHeader( "Content-Disposition:", "inline");
         }
         else
         {
            resp.setHeader( "Content-Disposition:", "attachment;filename=" + fileName);
         }
         resp.setContentLength((int) mergedDoc.length());
         resp.setBufferSize(BUFFER_SIZE);
         
         ServletOutputStream outputStream = resp.getOutputStream();
         
         try (FileInputStream fin = new FileInputStream(mergedDoc))
         {
            // Write merged content
            int data = 0;
            while ((data = fin.read()) != -1)
            {
               outputStream.write(data);
            }
            outputStream.flush();
         }
         System.err.println(fileName + " was delivered within " + (System.currentTimeMillis() - startDeliver) + "ms");
      }
      catch (COSLoadException e)
      {
         e.printStackTrace();
      }
      finally
      {
         desPDDoc.close();
         doc.close();
         mergedDoc.delete();
      }
   }
}
