/*
** Module   : PdfReportBuilder.java
** Abstract : PdfReportBuilder defines a PDF report builder functionality.
**
** Copyright (c) 2017, Golden Code Development Corporation.
**
** -#- -I- --Date-- ----------------Description----------------------
** 001 SBI 20170601 Created initial version.
*/
/*
** This program is free software: you can redistribute it and/or modify
** it under the terms of the GNU Affero General Public License as
** published by the Free Software Foundation, either version 3 of the
** License, or (at your option) any later version.
**
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
** GNU Affero General Public License for more details.
**
** You may find a copy of the GNU Affero GPL version 3 at the following
** location: https://www.gnu.org/licenses/agpl-3.0.en.html
** 
** Additional terms under GNU Affero GPL version 3 section 7:
** 
**   Under Section 7 of the GNU Affero GPL version 3, the following additional
**   terms apply to the works covered under the License.  These additional terms
**   are non-permissive additional terms allowed under Section 7 of the GNU
**   Affero GPL version 3 and may not be removed by you.
** 
**   0. Attribution Requirement.
** 
**     You must preserve all legal notices or author attributions in the covered
**     work or Appropriate Legal Notices displayed by works containing the covered
**     work.  You may not remove from the covered work any author or developer
**     credit already included within the covered work.
** 
**   1. No License To Use Trademarks.
** 
**     This license does not grant any license or rights to use the trademarks
**     Golden Code, FWD, any Golden Code or FWD logo, or any other trademarks
**     of Golden Code Development Corporation. You are not authorized to use the
**     name Golden Code, FWD, or the names of any author or contributor, for
**     publicity purposes without written authorization.
** 
**   2. No Misrepresentation of Affiliation.
** 
**     You may not represent yourself as Golden Code Development Corporation or FWD.
** 
**     You may not represent yourself for publicity purposes as associated with
**     Golden Code Development Corporation, FWD, or any author or contributor to
**     the covered work, without written authorization.
** 
**   3. No Misrepresentation of Source or Origin.
** 
**     You may not represent the covered work as solely your work.  All modified
**     versions of the covered work must be marked in a reasonable way to make it
**     clear that the modified work is not originating from Golden Code Development
**     Corporation or FWD.  All modified versions must contain the notices of
**     attribution required in this license.
*/


package com.goldencode.p2j.admin.server.reports;

import org.apache.pdfbox.pdmodel.PDDocument;
import org.apache.pdfbox.pdmodel.common.PDRectangle;
import org.apache.pdfbox.pdmodel.font.PDFont;

import com.goldencode.p2j.admin.shared.PaperOrientation;


/**
 * PdfReportBuilder defines a PDF report builder functionality.
 */
public interface PdfReportBuilder
extends ReportBuilder<PdfReportBuilder.PdfReportSettings, PdfReportBuilder.PdfDocument>
{
   /** Defines the array index to locate the top margin within margins array holder */
   int TOP    = 1;
   
   /** Defines the array index to locate the left margin within margins array holder */
   int LEFT   = 0;
   
   /** Defines the array index to locate the right margin within margins array holder */
   int RIGHT  = 2;
   
   /** Defines the array index to locate the bottom margin within margins array holder */
   int BOTTOM = 3;
   
   /**
    * Defines the PDF report settings.
    */
   interface PdfReportSettings
   extends ReportBuilder.ReportSettings
   {
      /**
       * Gets the report rectangle.
       *  
       * @return   The report rectangle
       */
      PDRectangle getReportSize();
      
      /**
       * Gets the report paper orientation.
       * 
       * @return   The paper orientation
       */
      PaperOrientation getPaperOrientation();
      
      /**
       * Gets the normal report font.
       * 
       * @return   The normal report font
       */
      PDFont getFont();
      
      /**
       * Gets the bold report font.
       * 
       * @return   The bold report font
       */
      PDFont getBoldFont();
      
      /**
       * Gets the base font size.
       * 
       * @return   The base font size
       */
      int getFontSize();
      
      /**
       * Gets the column header font size.
       * 
       * @return   The column header font size
       */
      int getColumnFontSize();
      
      /**
       * Gets the cell font size.
       * 
       * @return   The cell font size
       */
      int getCellFontSize();

      /**
       * Gets the page header font size.
       * 
       * @return   The page header font size
       */
      int getHeaderFontSize();
      
      /**
       * Gets the page footer font size.
       * 
       * @return   The page footer font size
       */
      int getFooterFontSize();
      
      /**
       * Gets the page margin in this order: left, top, right and bottom. The values are
       * measured by points relative to the user space.
       * 
       * @return   The page margin
       */
      float[] getPageMargin();
      
      /**
       * Gets the header margin measured by points relative to the user space.
       * 
       * @return   The header margin
       */
      float getHeaderMargin();
      
      /**
       * Gets the footer margin measured by points relative to the user space.
       * 
       * @return   The footer margin
       */
      float getFooterMargin();
      
      /**
       * Gets the cell margin measured by points relative to the user space.
       * 
       * @return   The cell margin
       */
      float getCellMargin();
      
      /**
       * Gets the cell padding measured by points relative to the user space.
       * 
       * @return   The cell padding
       */
      float getCellPadding();
      
      /**
       * Gets the maximal pages number.
       * 
       * @return   The maximal pages number
       */
      int getMaxPagesNumber();
   }

   /**
    * Defines the PDF report document interface.
    */
   interface PdfDocument
   extends ReportBuilder.Document
   {
      /**
       * Gets the PDF document.
       * 
       * @return   The PDF document
       */
      PDDocument getDocument();
      
      /**
       * Retrieves the document pages number.
       * 
       * @return   The document pages number
       */
      int getPagesNumber();
   }
}
