/*
** Module   : ComOleHelper.java
** Abstract : Helper methods to handle COM Automation processing.
**
** Copyright (c) 2017, Golden Code Development Corporation.
**
** -#- -I- --Date-- ---------------------------------Description----------------------------------
** 001 EVL 20171023 Created initial version.
** 002 CA  20171116 Added support for COM property setter with indices.
*/
/*
** This program is free software: you can redistribute it and/or modify
** it under the terms of the GNU Affero General Public License as
** published by the Free Software Foundation, either version 3 of the
** License, or (at your option) any later version.
**
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
** GNU Affero General Public License for more details.
**
** You may find a copy of the GNU Affero GPL version 3 at the following
** location: https://www.gnu.org/licenses/agpl-3.0.en.html
** 
** Additional terms under GNU Affero GPL version 3 section 7:
** 
**   Under Section 7 of the GNU Affero GPL version 3, the following additional
**   terms apply to the works covered under the License.  These additional terms
**   are non-permissive additional terms allowed under Section 7 of the GNU
**   Affero GPL version 3 and may not be removed by you.
** 
**   0. Attribution Requirement.
** 
**     You must preserve all legal notices or author attributions in the covered
**     work or Appropriate Legal Notices displayed by works containing the covered
**     work.  You may not remove from the covered work any author or developer
**     credit already included within the covered work.
** 
**   1. No License To Use Trademarks.
** 
**     This license does not grant any license or rights to use the trademarks
**     Golden Code, FWD, any Golden Code or FWD logo, or any other trademarks
**     of Golden Code Development Corporation. You are not authorized to use the
**     name Golden Code, FWD, or the names of any author or contributor, for
**     publicity purposes without written authorization.
** 
**   2. No Misrepresentation of Affiliation.
** 
**     You may not represent yourself as Golden Code Development Corporation or FWD.
** 
**     You may not represent yourself for publicity purposes as associated with
**     Golden Code Development Corporation, FWD, or any author or contributor to
**     the covered work, without written authorization.
** 
**   3. No Misrepresentation of Source or Origin.
** 
**     You may not represent the covered work as solely your work.  All modified
**     versions of the covered work must be marked in a reasonable way to make it
**     clear that the modified work is not originating from Golden Code Development
**     Corporation or FWD.  All modified versions must contain the notices of
**     attribution required in this license.
*/

package com.goldencode.p2j.comauto;

import com.goldencode.p2j.ui.chui.ThinClient;
import com.goldencode.p2j.util.*;

/**
 * Helper methods to handle COM Automation processing.  This is a client-side specific class, do
 * not use these APIs from server-side, use {@link ComOleOps} instead.
 */
public class ComOleHelper
{
   /** Flag indicating if the current system has proper support for COM Automation. */
   private static final boolean isSupported;
   
   static
   {
      isSupported = initComBridge();
   }
   
   /**
    * Performs native initialization for COM Automation subsystem if underlying OS supports COM.
    * 
    * @return  <code>TRUE</code> in success,
    *          <code>FALSE</code> for Non-Windows or in case of setup failure.
    */
   public static native boolean initComBridge();
   
   /**
    * Release and destroy the COM object with the specified ID.
    * 
    * @param    comId
    *           The COM object ID.
    */
   public static native void release(long comId);

   /**
    * Obtain a reference to an automation. It can be a new instance or a reference to an
    * already existing top-level automation. 
    * 
    * @param   comObjectType
    *          The name of the automation.
    * @param   topLevel
    *          Only return top-level. If the requested {@code comObjectType} is not a top-level, 
    *          return {@code null}.
    * @param   noCreate
    *          Must exist. Do not allow creation of new instances. Only used with {@code topLevel}
    *          requests.
    *          
    * @return  The ID of the COM object, or {@code 0} if such automation does not exist
    *          or it does not satisfy the parameter constraints. 
    */
   public static native long create(String comObjectType, boolean topLevel, boolean noCreate);

   /**
    * Get the specified COM object property.
    * 
    * @param    comId
    *           The COM object ID.
    * @param    prop
    *           The COM property name to retrieve.
    *           
    * @return   A {@link BaseDataType} instance with the COM property's details.
    */
   public static native BaseDataType getProperty(long comId, String prop);

   /**
    * Set the specified COM object property.
    * 
    * @param    comId
    *           The COM object ID.
    * @param    prop
    *           The COM property name to set.
    * @param    val
    *           The details about property's new value.
    * @param    indices
    *           A variable number of indices used to access this property's element.
    *           
    * @return   <code>true</code> if the property was set successfully.
    */
   public static native boolean setProperty(long           comId,
                                            String         prop,
                                            ComParameter   val, 
                                            ComParameter[] indices);

   /**
    * Execute a COM object method call.
    * 
    * @param    comId
    *           The COM object ID.
    * @param    methodName
    *           The COM method name to call.
    * @param    aParams
    *           An array with this call's arguments supplied to the COM object method call.
    *           
    * @return   An array where on index 0 are details of the returned value, and further indices
    *           correspond with argument's 1-based indices, where there will be a value for each 
    *           OUTPUT or INPUT-OUTPUT argument.
    */
   public static native BaseDataType[][] call(long comId, String methodName, ComParameter[] aParams);

   /**
    * Get the COM object name associated with the given COM object ID.
    * 
    * @param    comId
    *           The COM object ID.
    *           
    * @return   See above.
    */
   public static native String getName(long comId);

   /**
    * Reports if this object is valid for use.  
    *
    * @param    comId
    *           The COM object ID.
    *           
    * @return   <code>true</code> if we are valid (can be used).
    */
   public static native boolean valid(long comId);
   
   /**
    * Fire a new COM event.  This will immediately call the server-side to invoke the specified
    * procedure.
    * 
    * @param    callback
    *           The callback procedure which needs to be executed.
    * @param    params
    *           The parameters to pass to the callback procedure.
    */
   public static void fire(String callback, BaseDataType[] params)
   {
      ThinClient tc = ThinClient.getInstance();
      Object[] opars = (Object[]) params;
      
      // TODO: what happens if the procedure doesn't exist?
      
      tc.invokeOnServer(() -> tc.getServer().invokeProcedure(callback, opars));
   }
   
   /**
    * Gets COM Automation support level for particular OS.  Only Windows system has this
    * functionality.
    * 
    * @return  <code>TRUE</code> if COM objects are supported, <code>FALSE</code> otherwise.
    */
   public static boolean isSupported()
   {
      return isSupported;
   }
}
