/*
** Module   : NodeHelper.java
** Abstract : Defines helper methods to add, delete nodes in the directory.
**
** Copyright (c) 2017, Golden Code Development Corporation.
**
** -#- -I- --Date-- ---------------------------------Description----------------------------------
** 001 SBI 20171116 Added helper methods to manipulate nodes in the directory by reusing these
**                  private methods: addNode/deleteNode/openBatch/closeBatch of SSLCertGenUtil.
*/
/*
** This program is free software: you can redistribute it and/or modify
** it under the terms of the GNU Affero General Public License as
** published by the Free Software Foundation, either version 3 of the
** License, or (at your option) any later version.
**
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
** GNU Affero General Public License for more details.
**
** You may find a copy of the GNU Affero GPL version 3 at the following
** location: https://www.gnu.org/licenses/agpl-3.0.en.html
** 
** Additional terms under GNU Affero GPL version 3 section 7:
** 
**   Under Section 7 of the GNU Affero GPL version 3, the following additional
**   terms apply to the works covered under the License.  These additional terms
**   are non-permissive additional terms allowed under Section 7 of the GNU
**   Affero GPL version 3 and may not be removed by you.
** 
**   0. Attribution Requirement.
** 
**     You must preserve all legal notices or author attributions in the covered
**     work or Appropriate Legal Notices displayed by works containing the covered
**     work.  You may not remove from the covered work any author or developer
**     credit already included within the covered work.
** 
**   1. No License To Use Trademarks.
** 
**     This license does not grant any license or rights to use the trademarks
**     Golden Code, FWD, any Golden Code or FWD logo, or any other trademarks
**     of Golden Code Development Corporation. You are not authorized to use the
**     name Golden Code, FWD, or the names of any author or contributor, for
**     publicity purposes without written authorization.
** 
**   2. No Misrepresentation of Affiliation.
** 
**     You may not represent yourself as Golden Code Development Corporation or FWD.
** 
**     You may not represent yourself for publicity purposes as associated with
**     Golden Code Development Corporation, FWD, or any author or contributor to
**     the covered work, without written authorization.
** 
**   3. No Misrepresentation of Source or Origin.
** 
**     You may not represent the covered work as solely your work.  All modified
**     versions of the covered work must be marked in a reasonable way to make it
**     clear that the modified work is not originating from Golden Code Development
**     Corporation or FWD.  All modified versions must contain the notices of
**     attribution required in this license.
*/


package com.goldencode.p2j.directory;


/**
 * Defines helper methods to add, delete nodes in the directory.
 */
public class NodeHelper
{

   /** Default constructor for directory node utilities */
   private NodeHelper()
   {
   }

   /**
    * Delete the specified node and all of its children.
    * 
    * @param    ds
    *           The directory service
    * @param    nodeId
    *           The node ID.
    * 
    * @throws   RuntimeException
    *           If the node or one of its children could not be deleted.
    */
   public static void deleteNode(DirectoryService ds, String nodeId)
   {
      if (ds.getNodeClass(nodeId) != null)
      {
         String[] children = ds.enumerateNodes(nodeId);
         for (int i = 0; i < children.length; i++)
         {
            String fullId = nodeId + "/" + children[i];
            
            // remove previous entry
            if (!ds.deleteNode(fullId))
            {
               throw new RuntimeException("deleteNode() failed for the " + fullId + " node");
            }
         }

         if (!ds.deleteNode(nodeId))
         {
            throw new RuntimeException("deleteNode() failed for the " + nodeId + " node");
         }
      }
   }

   /**
    * Add a new node to the directory, having the specified class and a single attribute
    * with the same class and specified value.
    * 
    * @param    ds
    *           The directory service
    * @param    nodeId
    *           The node ID.
    * @param    cls
    *           The node's and attribute's class.
    * @param    value
    *           The attribute's value.
    * 
    * @throws   RuntimeException
    *           If the node could not be added.
    */
   public static void addNode(DirectoryService ds, String nodeId, String cls, Object value)
   {
      NodeAttribute nodeAttr = ds.getClassNodeAttribute(cls, "value");
      Attribute attr = new Attribute(nodeAttr, new Object[] { value });
      addNode(ds, nodeId, cls, new Attribute[] { attr });
   }

   /**
    * Add a new node to the directory, having the specified class and attributes.
    * 
    * @param    ds
    *           The directory service
    * @param    nodeId
    *           The node ID.
    * @param    cls
    *           The node's class.
    * @param    attrs
    *           The node's attributes.
    * 
    * @throws   RuntimeException
    *           If the node could not be added.
    */
   public static void addNode(DirectoryService ds, String nodeId, String cls, Attribute[] attrs)
   {
      if (!ds.addNode(nodeId, cls, attrs))
      {
         throw new RuntimeException("addNode() failed for the " + nodeId + " node");
      }
   }

   /**
    * Open a batch editing session for the specified node.
    * 
    * @param    ds
    *           The directory service
    * @param    node
    *           The node ID.
    * 
    * @throws   RuntimeException
    *           If the batch editing session could not be opened.
    */
   public static void openBatch(DirectoryService ds, String node)
   {
      if (!ds.openBatch(node))
      {
         throw new RuntimeException("openBatch() failed for the " + node + " node");
      }
   }

   /**
    * Close a batch editing session.
    * 
    * @param    ds
    *           The directory service
    * @param    node
    *           The node ID.
    * 
    * @throws   RuntimeException
    *           If the batch editing session could not be closed.
    */
   public static void closeBatch(DirectoryService ds, String node)
   {
      if (!ds.closeBatch(true))
         throw new RuntimeException("closeBatch() failed for " + node);
   }
}