/*
** Module   : e4gl.g
** Abstract : embedded 4GL (WebSpeed) lexer and parser grammar
**
** Copyright (c) 2007-2009, Golden Code Development Corporation.
**
** -#- -I- --Date-- --JPRM-- ----------------Description-----------------
** 001 GES 20070820   @34961 First version implementing a complete lexer
**                           and parser for embedded 4GL. This takes HTML
**                           files with E4GL and preprocesses these into
**                           4GL programs.
** 002 GES 20090428   @42020 Shifted to use common code for lexer dumping.
**/

header
{
/*
** Module   : E4GLParser.java
**            E4GLTokenTypes.java
**            E4GLLexer.java
** Abstract : lex and parse E4GL source code (html files)
**
** Copyright (c) 2007-2017, Golden Code Development Corporation.
**
** -#- -I- --Date-- -T- --JPRM-- ----------------Description-----------------
** 001 GES 20070820 NEW   @348xx WARNING, THIS IS A GENERATED FILE!!!
**                               DO NOT EDIT THIS FILE. The original source
**                               file is e4gl.g!
*/ 
/*
** This program is free software: you can redistribute it and/or modify
** it under the terms of the GNU Affero General Public License as
** published by the Free Software Foundation, either version 3 of the
** License, or (at your option) any later version.
**
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
** GNU Affero General Public License for more details.
**
** You may find a copy of the GNU Affero GPL version 3 at the following
** location: https://www.gnu.org/licenses/agpl-3.0.en.html
** 
** Additional terms under GNU Affero GPL version 3 section 7:
** 
**   Under Section 7 of the GNU Affero GPL version 3, the following additional
**   terms apply to the works covered under the License.  These additional terms
**   are non-permissive additional terms allowed under Section 7 of the GNU
**   Affero GPL version 3 and may not be removed by you.
** 
**   0. Attribution Requirement.
** 
**     You must preserve all legal notices or author attributions in the covered
**     work or Appropriate Legal Notices displayed by works containing the covered
**     work.  You may not remove from the covered work any author or developer
**     credit already included within the covered work.
** 
**   1. No License To Use Trademarks.
** 
**     This license does not grant any license or rights to use the trademarks
**     Golden Code, FWD, any Golden Code or FWD logo, or any other trademarks
**     of Golden Code Development Corporation. You are not authorized to use the
**     name Golden Code, FWD, or the names of any author or contributor, for
**     publicity purposes without written authorization.
** 
**   2. No Misrepresentation of Affiliation.
** 
**     You may not represent yourself as Golden Code Development Corporation or FWD.
** 
**     You may not represent yourself for publicity purposes as associated with
**     Golden Code Development Corporation, FWD, or any author or contributor to
**     the covered work, without written authorization.
** 
**   3. No Misrepresentation of Source or Origin.
** 
**     You may not represent the covered work as solely your work.  All modified
**     versions of the covered work must be marked in a reasonable way to make it
**     clear that the modified work is not originating from Golden Code Development
**     Corporation or FWD.  All modified versions must contain the notices of
**     attribution required in this license.
*/

package com.goldencode.p2j.e4gl;

import java.io.*;
import com.goldencode.ast.*;
import antlr.*;
}

/**
 * Parses an HTML file that has embedded 4GL (E4GL) source code and acts like
 * a filter in writing the result out as a Progress 4GL program. Any pure
 * HTML is output as 4GL source code that duplicates this HTML via writes
 * to a stream.
 * <p>
 * This is NOT a general purpose HTML parser. In fact, it is "dumbed down"
 * from the pure HTML specification to a subset that is supported by the
 * very limited parsing that WebSpeed supports. That parsing ignores certain
 * features of HTML encoding that are valid, thus this parser can similarly
 * ignore those features.
 * <p>
 * The WebSpeed product supports 4 types of program:
 * <p>
 * <ol>
 *    <li> Static HTML - pure, valid HTML that has no embedded 4GL source
 *         code
 *    <li> E4GL - HTML with embedded 4GL source code in sections delineated
 *         by statement (multi-line arbitrary 4GL code) start/end escapes or
 *         by expression (a single 4GL expression) start/end escapes
 *    <li> CGI Wrapper - hand-coded Progress source code that emits HTML
 *    <li> HTML Mapping - static HTML combined with a separate hard coded 4GL
 *         procedure that uses a configuration file containing field offsets
 *         to map data to/from the static HTML (this merging and enhancement
 *         of the static HTML is done at runtime by the WebSpeed code in
 *         cooperation with the coded 4GL procedure)
 * </ol>
 * <p>
 * This preprocessor handles cases 1 and 2 above. Cases 3 and 4 do not have
 * any embedded 4GL to preprocess so they do not need to run through this
 * parser. On first glance, case 1 (static HTML with no E4GL) seems like it
 * can be ignored since turning it into 4GL is costly from a runtime 
 * efficiency perspective. When run through this preprocessor, the output of
 * such files is still a 4GL procedure (or include file). Since this procedure
 * or include file can be itself included in other E4GL programs, it is
 * necessary to convert all static HTML and E4GL code into 4GL source code.
 * <p>
 * Interestingly, the 4GL source code result of this preprocessor is the same
 * as if one had coded a CGI Wrapper (case 3 above) instead of using static
 * HTML or E4GL. There is no difference between the 3 cases at runtime, it is
 * only a difference in whether this WebSpeed preprocessor is used first. 
 * <p>
 * HTML Mapping (case 4) is not addressed by this preprocessor since the HTML
 * involved in that solution is processed at runtime.
 * <p>
 * The filtering that happens can be customized to match the output of a
 * specific E4GL implementation. At this time both WebSpeed and Blue Diamond
 * implementations are supported.
 * <p>
 * This parser is heavily dependent upon the following code:
 * <p>
 * <ul>
 *    <li> {@link Options} - stores configuration that can be shared from
 *         multiple modules
 *    <li> {@link CompatibilityHelper} - this hides the difference in modes
 *         between WebSpeed and Blue Diamond
 *    <li> {@link E4GLPreprocessor} - this provides the external interface
 *         for the parser
 * </ul>
 * <p>
 * There are 2 external entry points for the parser: {@link #preprocess} and
 * {@link #scan_options}.
 * <p>
 * Parser features:
 * <p>
 * <ul>
 *    <li> statement escapes (the "...4GL..." indicates where the embedded
 *         4GL code will be found)
 *    <ul>
 *       <li> &lt;SCRIPT LANGUAGE="SpeedScript"&gt; ...4GL... &lt;/SCRIPT&gt;
 *       <li> &lt;SCRIPT LANGUAGE="WebSpeed4GL"&gt; ...4GL... &lt;/SCRIPT&gt;
 *       <li> &lt;SCRIPT LANGUAGE="Progress"&gt; ...4GL... &lt;/SCRIPT&gt;
 *       <li> &lt;!--WSS ...4GL... --&gt;
 *       <li> &lt;!--WS4GL ...4GL... --&gt;
 *       <li> &lt;?WS&gt; ...4GL... &lt;/?WS&gt;
 *       <li> &lt;SERVER&gt; ...4GL... &lt;/SERVER&gt;
 *       <li> &lt;% ...4GL... %&gt;
 *    </ul>
 *    <li> expression escapes (the "...4GL..." indicates where the embedded
 *         4GL expression will be found)
 *    <ul>
 *       <li> &lt;!--WSE ...4GL... --&gt;
 *       <li> ` ...4GL... `
 *       <li> {= ...4GL... =}
 *       <li> &lt;%= ...4GL... %&gt;
 *    </ul>
 *    <li> parsing of a <code>META</code> or <code>WSMETA</code> elements
 *         with attributes of <code>NAME="wsoptions"</code> or 
 *         <code>HTTP-EQUIV="content-type"</code>; in both cases the
 *         <code>CONTENT</code> attribute is read with the following options
 *         being honored
 *    <ul>
 *       <li> include
 *       <li> web-object
 *       <li> keep-meta-content-type
 *       <li> content-type
 *    </ul>
 *    <li> everything other than the above contructs is treated as raw HTML
 *         and is output with a line prefix (effectively the same as)
 *         <code>PUT Stream WebStream UNFORMATTED '</code>) and a line suffix
 *         <code>~n'.</code> (there are many variations on these depending on
 *         what is being output and in which order; in additions the actual
 *         text used is different depending on the compatibility helper that
 *         is in use, WebSpeed or BlueDiamond)
 *    <li> raw HTML is converted into 4GL strings and any characters that
 *         cannot be safely represented in 4GL source code is escaped
 *    <li> 4GL expressions and statements which have HTML entities encoded
 *         are decoded into the matching character
 *    <li> 4GL expressions used in HTML URLs that have the %XX character
 *         encoding for spaces and other special characters are decoded into
 *         the matching character (using the 2 hexidecimal digits that follow
 *         the percent sign)
 * </ul>
 * <p>
 * <b>TODO:</b> before this output can be optimally converted, it is likely
 * that the expansions, replacements and other state of the preprocessing
 * will need to be recorded by the parser into a hints file. Much of the
 * data to be stored will be in the form of a "section definition" which
 * stores the start line/column and end line/column of a section of either
 * the original source file or of the target file. The following will be
 * needed:
 * <p>
 * <ul>
 *    <li> original HTML source file name
 *    <li> any specified wsoptions
 *    <li> any specified content-type
 *    <li> include flag
 *    <li> keep content type flag
 *    <li> compatibility mode
 *    <li> for each raw HTML section that was converted to "PUT" statements
 *         there will be a section definition for both the source and target
 *         files (this will allow the 4GL source code to be removed and
 *         replaced by the original raw HTML from the source file should
 *         the conversion need it)
 *    <li> a section definition in the target file for the generated header
 *    <li> a section definition in the target file for the generated footer
 * </ul>
 */
class E4GLParser extends Parser;

options
{   
   k = 2;
   buildAST = false;
   codeGenBitsetTestThreshold = 999;
}

// helper members and methods
{
   /** Newline character to search for for line operations. */
   private static char NL = '\n';

   /** Configuration values to honor during preprocessing. */
   private Options cfg = null;
   
   /** Source for raw input. */
   private InputStream in = null;

   /** Destination for preprocessed output. */
   private PrintStream out = null;
   
   /** Helper to emit headers, footers, comments and other text. */
   private CompatibilityHelper hlp = null;
   
   /** Flag to denote that the output is starting a new line. */
   private boolean lineBegin = true;
   
   /** Flag to denote that the last output was HTML. */
   private boolean wasHtml = true;
   
   /** Flag to denote that the last output was an expression escape close. */
   private boolean wasExprClose = false;
   
   /** Flag to denote that the last output was an statement escape close. */
   private boolean wasStmtClose = false;
   
   static
   {
      // by default we search for '\n' which handles both "\n" and "\r\n"
      // cases, but if the platform newline is "\r" then we must search for
      // '\r' instead
      if ("\r".equals(CompatibilityHelper.NEWLINE))
      {
         NL = '\r';
      }
   }
   
   /**
    * Get the input stream. This is not used internally but is stored here
    * to allow the stream to be easily obtained by the user of this class.
    *
    * @return   Source stream.
    */
   public InputStream getInput()
   {
      return in;
   }

   /**
    * Set the input stream. This is not used internally but is stored here
    * to allow the stream to be easily obtained by the user of this class.
    *
    * @param    in
    *           Source stream.
    */
   public void setInput(InputStream in)
   {
      this.in = in;
   }

   /**
    * Close the input stream.
    */
   public void closeInput()
   {
      try
      {
         if (in != null)
            in.close();
      }
      
      catch (IOException ioe)
      {
         // ignore
      }
   }

   /**
    * Writes error data to <code>System.err</code>, including a full stack
    * trace.
    *
    * @param    re
    *           The error on which to report.
    */
   public void reportError(RecognitionException re)
   {
      re.printStackTrace(System.err);
   }
   
   /**
    * Output the given text to the stream while maintaining the new line
    * state. If requested, this method will also convert HTML entities into
    * the valid characters as they would appear in 4GL source code. This
    * occurs based on the <code>cvt</code> parameter. If the type of output
    * is <code>html</code>, then any unsafe characters are escaped before
    * being output as 4GL code.
    * <p>
    * This method is highly sensitive to the member state variables that
    * are maintained in cooperation with <code>outputOpenComment()</code> and
    * <code>outputCloseComment()</code>. Together these 3 methods handle all
    * output for the preprocessor.
    * <p>
    * Based on the current state variables, the parameters and the
    * {@link CompatibilityHelper} instance, this method handles all output
    * that isn't a start or end element of a statement of expression escape.
    * This means that all raw HTML and all (embedded 4GL source) content of
    * expression/statement escapes is handled here.
    * <p>
    * Of great importance is that each line of output is started and ended
    * with the proper text (in HTML mode). Likewise, when shifting into and
    * out of expression and statement escapes, any raw HTML must be properly
    * terminated and resumed as needed.  Some deferred suffix processing of
    * end elements for statement and expression escapes is handled here where
    * the following content is known such that the variance in results
    * can be determined.
    *
    * @param    txt
    *           Contents to write to the stream. Must be non-empty and
    *           not <code>null</code>.
    * @param    cvt
    *           <code>true</code> to force HTML entity conversion before the
    *           text is output.
    * @param    html
    *           <code>true</code> to add end of line and beginning of line
    *           output around new line characters in HTML. Since this
    *           preprocessor is converting this HTML into strings that are
    *           output via streams in the 4GL source, these strings need to
    *           be augmented at the beginning and end of each line.
    */
   private void output(String txt, boolean cvt, boolean html)
   {
      int last = txt.length() - 1;
      
      if (last == -1)
         return;
   
      txt = cvt  ? hlp.entityConvert(txt) : txt;
      // txt = html ? hlp.escapeHTML(txt)    : txt;
      
      if (lineBegin)
      {
         // if we are at the beginning of a line and in HTML mode, then we
         // need to output the line prefix
         if (html)
         {
            out.print(hlp.openHTML(true));
         }
         
         // clear the counter, since we are doing output
         lineBegin = false;
      }
      else
      {
         // check if we just exited an expression escape
         if (wasExprClose)
         {
            if (html && txt.indexOf(NL) != 0)
            {
               // re-open our string for raw HTML output
               out.print(hlp.closeExprEscape());
            }
            else
            {
               // no more HTML so we need to close the last HTML output
               out.print(hlp.closeHTML(false));
            }
         }
         else if (wasStmtClose)
         {
            // HTML text that follows a statement escape needs to be opened
            // if there is no newline char as the first character of our text
            if (html && txt.indexOf(NL) != 0)
            {
               out.print(hlp.openHTML(true));
            }
         }
      }
      
      if (html)
      {
         StringBuilder sb = new StringBuilder();
         
         boolean found = false;
         int prev = 0;
         int next = txt.indexOf(NL);
         
         while (next >= 0)
         {
            if (found)
            {
               // subsequent time through, act like the lineBegin flag is
               // true and output the line prefix
               sb.append(hlp.openHTML(true));
            }
         
            // if the text to be output ends in a newline, then we need to
            // reset the lineBegin flag to ensure that subsequent output will
            // honor the fact (if needed by context) that we haven't yet output
            // any line prefix; if this newline is in the middle of the text
            // there is no need to update the state of this flag since we are
            // about to reset its state again before this method returns...
            if (next == last)
            {
               lineBegin = true;
            }
            
            // it is possible that there are blank lines and lines with only
            // whitespace that are output here, but these can't be dropped
            // in case we are inside a PRE block and the whitespace is
            // important; put the text and the line suffix and the new line
            sb.append(hlp.escapeHTML(txt.substring(prev, next)));
            
            // on the first pass, even though we are outputting HTML now, if
            // the newline is the first character and the previous mode was
            // not HTML, then we don't output the close HTML text
            boolean avoid = (!found && !wasHtml && next == 0);
            
            if (!avoid || found)
              sb.append(hlp.closeHTML(true));
              
            sb.append(hlp.NEWLINE);
         
            prev = next + 1;
            next = txt.indexOf(NL, prev);
            
            // not the first time through
            found = true;
         }
         
         if (found)
         {
            // output the rest of the string if there is any
            if (prev < txt.length())
            {
               sb.append(hlp.openHTML(true));
               sb.append(hlp.escapeHTML(txt.substring(prev)));
            }
         
            // reset our output text to the modified version
            txt = sb.toString();
         }
         else
         {
            txt = hlp.escapeHTML(txt);
         }
      }
      else
      {
         // see the above comment regarding why we set this flag and why it
         // is only if the last character is a newline
         if (txt.lastIndexOf(NL) == last)
         {
            lineBegin = true;
         }
      }
      
      out.print(txt);
      
      // clear our indicators, any deferred output is complete
      wasExprClose = false;
      wasStmtClose = false;
      
      // save state
      wasHtml = html;
   }
   
   /**
    * Comment out the given text of an expression or statement open element
    * and output it to the stream.
    * <p>
    * This method is highly sensitive to the member state variables that
    * are maintained in cooperation with <code>output()</code> and
    * <code>outputCloseComment()</code>. Together these 3 methods handle all
    * output for the preprocessor.
    * <p>
    * Based on the current state variables, the parameters and the
    * {@link CompatibilityHelper} instance, this method handles all output
    * that is a start element of a statement of expression escape.
    * <p>
    * Of great importance is that, when shifting into expression and
    * statement escapes, any raw HTML must be properly terminated or
    * started as needed. Some deferred suffix processing of end elements
    * for statement and expression escapes is handled here where the
    * following content is known such that the variance in results
    * can be determined.
    *
    * @param    escape
    *           The token with the text to comment.
    * @param    expr
    *           <code>true</code> if this is an expression escape,
    *           <code>false</code> for a statement escape.
    */
   private void outputOpenComment(Token escape, boolean expr)
   {
      outputOpenComment(escape.getText(), expr);
   }
   
   /**
    * Comment out the given text of an expression or statement open element
    * and output it to the stream.
    * <p>
    * This method is highly sensitive to the member state variables that
    * are maintained in cooperation with <code>output()</code> and
    * <code>outputCloseComment()</code>. Together these 3 methods handle all
    * output for the preprocessor.
    * <p>
    * Based on the current state variables, the parameters and the
    * {@link CompatibilityHelper} instance, this method handles all output
    * that is a start element of a statement of expression escape.
    * <p>
    * Of great importance is that, when shifting into expression and
    * statement escapes, any raw HTML must be properly terminated or
    * started as needed. Some deferred suffix processing of end elements
    * for statement and expression escapes is handled here where the
    * following content is known such that the variance in results
    * can be determined.
    *
    * @param    escape
    *           The text to comment.
    * @param    expr
    *           <code>true</code> if this is an expression escape,
    *           <code>false</code> for a statement escape.
    */
   private void outputOpenComment(String escape, boolean expr)
   {
      String prefix = "";
   
      if (expr)
      {
         if (lineBegin)
         {
            // no content to terminate, but we do need to start the line
            prefix = hlp.openHTML(false);
         }
         else
         {
            if (wasHtml)
            {
               // content to terminate before the comment emits
               prefix = hlp.openExprEscape();
            }
         }
      }
      else
      {
         // this is a statement escape (nothing to do unless this is in
         // the middle of a line)
         if (!lineBegin)
         {
            if (wasHtml)
            {
               // there is unterminated content on the line, terminate it
               // properly
               prefix = hlp.closeHTML(true);
            }
            else
            {
               if (wasExprClose)
               {
                  // close off the HTML mode BUT there is no content to
                  // terminate
                  prefix = hlp.closeHTML(false);
               }
            }
         }
      }
      
      if (prefix.length() > 0)
         out.print(prefix);
      
      // reset our state vars
      lineBegin    = false;
      wasHtml      = false;
      wasExprClose = false;
      wasStmtClose = false;
      
      generateComment(escape, expr);
   }
   
   /**
    * Comment out the given text of an expression or statement close element
    * and output it to the stream.
    * <p>
    * This method is highly sensitive to the member state variables that
    * are maintained in cooperation with <code>output()</code> and
    * <code>outputOpenComment()</code>. Together these 3 methods handle all
    * output for the preprocessor.
    * <p>
    * Based on the current state variables, the parameters and the
    * {@link CompatibilityHelper} instance, this method handles all output
    * that is an end element of a statement of expression escape.
    * <p>
    * Of great importance is that, when shifting into and out of expression
    * and statement escapes, any raw HTML must be properly
    * resumed and terminated as needed. Some suffix processing of end elements
    * for statement and expression escapes is also deferred to the
    * <code>output()</code> and <code>outputOpenComment()</code> methods
    * where the following content is known such that the variance in results
    * can be determined.
    *
    * @param    escape
    *           The token with the text to comment.
    * @param    expr
    *           <code>true</code> if this is an expression escape,
    *           <code>false</code> for a statement escape.
    */
   private void outputCloseComment(Token escape, boolean expr)
   {
      outputCloseComment(escape.getText(), expr);
   }
   
   /**
    * Comment out the given text of an expression or statement close element
    * and output it to the stream.
    * <p>
    * This method is highly sensitive to the member state variables that
    * are maintained in cooperation with <code>output()</code> and
    * <code>outputOpenComment()</code>. Together these 3 methods handle all
    * output for the preprocessor.
    * <p>
    * Based on the current state variables, the parameters and the
    * {@link CompatibilityHelper} instance, this method handles all output
    * that is an end element of a statement of expression escape.
    * <p>
    * Of great importance is that, when shifting into and out of expression
    * and statement escapes, any raw HTML must be properly
    * resumed and terminated as needed. Some suffix processing of end elements
    * for statement and expression escapes is also deferred to the
    * <code>output()</code> and <code>outputOpenComment()</code> methods
    * where the following content is known such that the variance in results
    * can be determined.
    *
    * @param    escape
    *           The text to comment.
    * @param    expr
    *           <code>true</code> if this is an expression escape,
    *           <code>false</code> for a statement escape.
    */
   private void outputCloseComment(String escape, boolean expr)
   {
      // reset our state vars
      lineBegin    = false;
      wasHtml      = false;
      
      // remember that we just exited an expression or statement escape so
      // that following output can terminate the stream output or start
      // another string of raw HTML, depending on the context
      wasExprClose = expr;
      wasStmtClose = !expr;
      
      generateComment(escape, expr);
   }
   
   /**
    * Generate a comment out the given text of an escape and output it to the
    * stream. The compatibility helper will be used to create the open and
    * close comment text. The compatibility helper may also specify that the
    * escape is to be silently dropped, in which case no output will be
    * generated.
    *
    * @param    escape
    *           The text to comment.
    * @param    expr
    *           <code>true</code> if this is an expression escape,
    *           <code>false</code> for a statement escape.
    */
   private void generateComment(String escape, boolean expr)
   {
      String open = hlp.openComment(expr);
      
      // only output if the comment text is non-null, otherwise silently
      // eat the escape
      if (open != null)
      {
         out.print(String.format("%s%s%s",
                                 open,
                                 escape,
                                 hlp.closeComment(expr)));
      }
   }
   
   /**
    * Get the text of the token unless it is of type <code>ENCODED_CHAR</code>
    * in which case the contents will be converted from the '%XX' form into
    * the hexidecimal character that is being encoded.
    *
    * @param    next
    *           The token to potentially decode.
    *
    * @return   The text of the token or its encoded character as text.
    */
   private String getTokenText(Token next)
   {
      String txt = next.getText();
      
      if (next.getType() == ENCODED_CHAR)
      {
         char encoded = (char) Integer.parseInt(txt.substring(1, 3), 16);
         txt = Character.toString(encoded);
      }
      
      return txt;
   }
}

/**
 * Top-level entry point that parses the input stream and preprocesses the
 * HTML and embedded 4GL into a valid 4GL program. All statement and
 * expression escapes are supported by calling a sub-rule for each type.
 * This sub-rule will not return until the entire escape has been parsed
 * and output. State is maintained such that transitions into and out of
 * escapes are prefixed or suffixed with the proper text. In some cases,
 * the suffix cannot be determined until following output is parsed, in which
 * case that suffix processing is deferred via member state variables. 
 * <p>
 * Uses the following expression escape rules:
 * <p>
 * <ul>
 *    <li> {@link #backtick_escape}
 *    <li> {@link #percent_eq_escape}
 *    <li> {@link #curly_escape}
 *    <li> {@link #wse_escape}
 * </ul>
 * <p>
 * Uses the following statement escape rules:
 * <p>
 * <ul>
 *    <li> {@link #script_escape}
 *    <li> {@link #question_escape}
 *    <li> {@link #percent_escape}
 *    <li> {@link #server_escape}
 *    <li> {@link #wss_escape}
 * </ul>
 * <p>
 * This rule also processes the {@link #meta_tag} rule in the case where
 * <code>keep-meta-content-type</code> is not set as an option. In this
 * case, a <code>META</code> or <code>WSMETA</code> element with an attribute
 * of <code>HTTP-EQUIV</code> set to <code>CONTENT-TYPE</code> will be
 * removed from the output stream using comments.  Anything else will be
 * passed through a raw HTML.
 * <p>
 * Anything that is not handled by one of the above listed rules is considered
 * raw HTML and is output with little regard to the content. The exceptions to
 * this include newline processing (since the beginning and end of each line
 * must have text added) and the proper escaping of certain valid HTML
 * characters that cannot be directly included in 4GL programs.
 * <p>
 * Internally, all of the state management and output processing is
 * centralized in the following methods: <code>output()</code>,
 * <code>outputOpenComment()</code> and <code>outputCloseComment()</code>.
 * <p>
 * All rules are peers at this level to ensure that they do not cause
 * parser ambiguity.
 *
 * @param    cfg
 *           Storage container for the options that are found.
 */
preprocess [Options cfg, PrintStream out]
   {
      this.cfg = cfg;
      this.out = out;
      hlp = cfg.getCompatibilityHelper();
      
      boolean httpEquiv = false;
      boolean keep      = cfg.isKeepContentType();
      StringBuilder sb = new StringBuilder();
      
      // print the file header 
      output(hlp.header(), false, false);
      wasHtml = true;
      lineBegin = true;
   }
   :
      (
         options { generateAmbigWarnings = false; }
         :
           // expression escapes
           backtick_escape
         | percent_eq_escape
         | curly_escape
         | wse_escape
         
           // statement escapes
         | script_escape
         | question_escape
         | percent_escape
         | server_escape
         | wss_escape
         
           // process META and WSMETA flags to remove content-type if
           // specified by options (this is the default setting so this code
           // will be executed unless the keep-meta-content-type is specified
           // as a wsoption
         | { !keep }?
           httpEquiv = meta_tag[null, sb]
           {
              String open  = httpEquiv ? hlp.openContentTypeComment()  : "";
              String close = httpEquiv ? hlp.closeContentTypeComment() : "";
              String core  = sb.toString();
              
              // are we supposed to remove the entry
              if (open.length() > 0)
              {
                 // remove the start and end character from the core buffer,
                 // which should correspond to the '<' and '>' that must
                 // always be present
                 core = core.substring(1, core.length() - 1);
              }
              
              output(String.format("%s%s%s", open, core, close), false, true);
           }
         
           // anything else is raw HTML
         | { LA(1) != EOF }?
           txt:.
           {
              output(txt.getText(), false, true);
           }
      )*
      EOF
      
      {
         // print the file footer 
         output(hlp.footer(), false, false);
      }
   ;
   
/**
 * Match a starting <code>OPEN_START_TAG SCRIPT</code>, arbitrary
 * 4GL statement text and then <code>CLOSE_QUESTION WSPEED CLOSE_TAG</code>.
 */   
script_escape
   {
      StringBuilder sb = new StringBuilder();
      boolean e4gl = false;
   }
   :
      // open an stmt escape
      e4gl = script_tag[sb]
      {
         if (e4gl)
         {
            outputOpenComment(sb.toString(), false);
            sb = new StringBuilder();
         }
      }
      
      // match and emit the 4GL stmts
      (
         options { generateAmbigWarnings = false; }
         :
         { !(LA(1) == OPEN_END_TAG && LA(2) == SCRIPT) }?
         txt:.
         {
            sb.append(getTokenText(txt));
         }
      )*
      {
         output(sb.toString(), e4gl, !e4gl);
      }
      
      // close the stmt escape
      oe:OPEN_END_TAG s:SCRIPT ct:CLOSE_TAG
      {
         String combined = String.format("%s%s%s",
                                         oe.getText(),
                                         s.getText(),
                                         ct.getText());
      
         if (e4gl)
         {
            // comment out the close text
            outputCloseComment(combined, false);
         }
         else
         {
            // emit as raw HTML
            output(combined, false, true);
         }
      }
   ;
   
/**
 * Match a starting <code>OPEN_START_TAG SCRIPT</code>, arbitrary
 * attributes and then <code>CLOSE_TAG</code>. The attributes will be
 * scanned and if a <code>LANGUAGE</code> attribute with a value of
 * <code>SpeedScript</code>, <code>WebSpeed4GL</code> or <code>Progress</code>
 * is found then this rule will return <code>true</code>. Either way the text
 * of all matched tokens will be returned in the given buffer so that the
 * calling rule may handle the output however is needed. 
 *
 * @param    sb
 *           Buffer which will contain all text associated with the parsed
 *           tokens.
 *
 * @return   <code>true</code> if this script element includes a language
 *           attribute that identifies the content as embedded 4GL.
 */   
script_tag [StringBuilder sb] returns [boolean e4gl]
   {
      String prop = null;
      e4gl = false;
   }
   :
      ot:OPEN_START_TAG s:SCRIPT
      {
         sb.append(ot.getText()).append(s.getText());
      }
      (
           l:LANGUAGE
           {
              sb.append(l.getText());
           }
           prop = attribute_value[sb]
           {
              if ("speedscript".equalsIgnoreCase(prop) ||
                  "webspeed4gl".equalsIgnoreCase(prop) ||
                  "progress".equalsIgnoreCase(prop))
              {
                 e4gl = true;
              }
           }
           
         | other:~CLOSE_TAG
           {
              sb.append(other.getText());
           }
      )*
      ct:CLOSE_TAG
      {
         sb.append(ct.getText());
      }
   ;

   
/**
 * Match a starting <code>OPEN_QUESTION WSPEED CLOSE_TAG</code>, arbitrary
 * 4GL statement text and then <code>CLOSE_QUESTION WSPEED CLOSE_TAG</code>.
 */   
question_escape
   {
      StringBuilder sb = new StringBuilder();
   }
   :
      // open an stmt escape
      oq:OPEN_QUESTION ws1:WSPEED ct1:CLOSE_TAG
      {
         oq.setText(String.format("%s%s%s",
                                  oq.getText(),
                                  ws1.getText(),
                                  ct1.getText()));
         
         // comment out the open text
         outputOpenComment(oq, false);
      }
      
      // match and emit the 4GL stmts
      (
         txt:~CLOSE_QUESTION
         {
            sb.append(getTokenText(txt));
         }
      )*
      {
         output(sb.toString(), true, false);
      }
      
      // close the stmt escape
      cq:CLOSE_QUESTION ws2:WSPEED ct2:CLOSE_TAG
      {
         cq.setText(String.format("%s%s%s",
                                  cq.getText(),
                                  ws2.getText(),
                                  ct2.getText()));
                                  
         // comment out the close text
         outputCloseComment(cq, false);
      }
   ;
   
/**
 * Match a starting <code>OPEN_PCT</code>, arbitrary 4GL statement text
 * and then <code>CLOSE_PCT</code>.
 */   
percent_escape
   {
      StringBuilder sb = new StringBuilder();
   }
   :
      // open an stmt escape
      op:OPEN_PCT
      {
         // comment out the open text
         outputOpenComment(op, false);
      }
      
      // match and emit the 4GL stmts
      (
         txt:~CLOSE_PCT
         {
            sb.append(getTokenText(txt));
         }
      )*
      {
         output(sb.toString(), true, false);
      }
      
      // close the stmt escape
      cp:CLOSE_PCT
      {
         // comment out the close text
         outputCloseComment(cp, false);
      }
   ;
   
/**
 * Match a starting <code>OPEN_START_TAG SERVER CLOSE_TAG</code>, arbitrary
 * 4GL statement text and then <code>CLOSE_QUESTION WSPEED CLOSE_TAG</code>.
 */   
server_escape
   {
      StringBuilder sb = new StringBuilder();
   }
   :
      // open an stmt escape
      ot:OPEN_START_TAG s1:SERVER ct1:CLOSE_TAG
      {
         ot.setText(String.format("%s%s%s",
                                  ot.getText(),
                                  s1.getText(),
                                  ct1.getText()));
         
         // comment out the open text
         outputOpenComment(ot, false);
      }
      
      // match and emit the 4GL stmts
      (
         txt:~OPEN_END_TAG
         {
            sb.append(getTokenText(txt));
         }
      )*
      {
         output(sb.toString(), true, false);
      }
      
      // close the stmt escape
      oe:OPEN_END_TAG s2:SERVER ct2:CLOSE_TAG
      {
         oe.setText(String.format("%s%s%s",
                                  oe.getText(),
                                  s2.getText(),
                                  ct2.getText()));
                                  
         // comment out the close text
         outputCloseComment(oe, false);
      }
   ;
   
/**
 * Match a starting <code>OPEN_COMMENT</code> followed by a <code>WSS</code>
 * or <code>WS4GL</code>, arbitrary 4GL statement text and then
 * <code>CLOSE_COMMENT</code>.
 */   
wss_escape
   {
      StringBuilder sb = new StringBuilder();
   }
   :
      // open an stmt escape
      oc:OPEN_COMMENT (wss:WSS | ws4gl:WS4GL)
      {
         Token word = (wss == null) ? ws4gl : wss;
         oc.setText(String.format("%s%s",
                                  oc.getText(),
                                  word.getText()));
         
         // comment out the open text
         outputOpenComment(oc, false);
      }
      
      // match and emit the 4GL stmts
      (
         txt:~CLOSE_COMMENT
         {
            sb.append(getTokenText(txt));
         }
      )*
      {
         output(sb.toString(), true, false);
      }
      
      // close the stmt escape
      cc:CLOSE_COMMENT
      {
         // comment out the close text
         outputCloseComment(cc, false);
      }
   ;
   
/**
 * Match a starting <code>BACK_TICK</code>, arbitrary 4GL expression text and
 * a closing <code>BACK_TICK</code>.
 */   
backtick_escape
   {
      StringBuilder sb = new StringBuilder();
   }
   :
      // open an expression escape
      bt1:BACK_TICK
      {
         // comment out the open text
         outputOpenComment(bt1, true);
      }
      
      // match and emit the 4GL expression
      (
         txt:~BACK_TICK
         {
            sb.append(getTokenText(txt));
         }
      )*
      {
         output(sb.toString(), true, false);
      }
      
      // close the expression escape
      bt2:BACK_TICK
      {
         // comment out the close text
         outputCloseComment(bt2, true);
      }
   ;
   
/**
 * Match a starting <code>OPEN_PCT_EQ</code>, arbitrary 4GL expression text
 * and a closing <code>CLOSE_PCT</code>.
 */   
percent_eq_escape
   {
      StringBuilder sb = new StringBuilder();
   }
   :
      // open an expression escape
      op:OPEN_PCT_EQ
      {
         // comment out the open text
         outputOpenComment(op, true);
      }
      
      // match and emit the 4GL expression
      (
         txt:~CLOSE_PCT
         {
            sb.append(getTokenText(txt));
         }
      )*
      {
         output(sb.toString(), true, false);
      }
      
      // close the expression escape
      cp:CLOSE_PCT
      {
         // comment out the close text
         outputCloseComment(cp, true);
      }
   ;
   
/**
 * Match a starting <code>OPEN_CURLY_EQ</code>, arbitrary 4GL expression text
 * and a closing <code>CLOSE_CURLY_EQ</code>.
 */   
curly_escape
   {
      StringBuilder sb = new StringBuilder();
   }
   :
      // open an expression escape
      oc:OPEN_CURLY_EQ
      {
         // comment out the open text
         outputOpenComment(oc, true);
      }
      
      // match and emit the 4GL expression
      (
         txt:~CLOSE_CURLY_EQ
         {
            sb.append(getTokenText(txt));
         }
      )*
      {
         output(sb.toString(), true, false);
      }
      
      // close the expression escape
      cc:CLOSE_CURLY_EQ
      {
         // comment out the close text
         outputCloseComment(cc, true);
      }
   ;

/**
 * Match a starting <code>OPEN_COMMENT WSE</code>, arbitrary 4GL expression
 * text and a closing <code>CLOSE_COMMENT</code>.
 */   
wse_escape
   {
      StringBuilder sb = new StringBuilder();
   }
   :
      // open an expression escape
      (ow:OPEN_COMMENT w:WSE)
      {
         // put the text back together as if it was a single token
         ow.setText(String.format("%s%s", ow.getText(), w.getText()));
         
         // comment out the open text
         outputOpenComment(ow, true);
      }
      
      // match and emit the 4GL expression
      (
         txt:~CLOSE_COMMENT
         {
            sb.append(getTokenText(txt));
         }
      )*
      {
         output(sb.toString(), true, false);
      }
      
      // close the expression escape
      cc:CLOSE_COMMENT
      {
         // comment out the close text
         outputCloseComment(cc, true);
      }
   ;

/**
 * Top-level entry point that parses the input stream to read the E4GL
 * options embedded in any <code>META</code> or <code>WSMETA</code> elements.
 * <p>
 * Uses {@link #meta_tag} to parse the options themselves. Anything else
 * is simply ignored until <code>EOF</code>.
 *
 * @param    opts
 *           Storage container for the options that are found.
 */
scan_options [Options opts]
   :
      {
         boolean bogus;
      }
      (
         options { generateAmbigWarnings = false; }
         :
            bogus = meta_tag[opts, null]
          | . // anything else
      )*
      EOF
   ;

/**
 * Match <code>OPEN_START_TAG META</code> or <code>OPEN_COMMENT WSMETA</code>,
 * the following <code>NAME</code> or <code>HTTP_EQUIV</code> and then the
 * <code>CONTENT</code> and a closing <code>CLOSE_TAG</code> or
 * <code>CLOSE_COMMENT</code>. This parses the meta tag processing and
 * may set option values that are embedded in the attributes. This is
 * called from {@link #scan_options} and {@link #preprocess}.
 *
 * @param    opts
 *           The options to update with the found values. If <code>null</code>
 *           then this rule matches the given text but does not update any
 *           options even if those attributes are present.
 * @param    sb
 *           The buffer in which to accumulate the parsed text. May be
 *           <code>null</code> to bypass text accumulation.
 *
 * @return   <code>true</code> if the parsed meta tag has an 
 *           <code>HTTP-EQUIV</code> attribute with a value of
 *           "content-type".
 */
meta_tag [Options opts, StringBuilder sb] returns [boolean contype]
   {
      boolean wsmeta    = false;
      boolean wsoptions = false;
      
      String  prop      = null;
      String  content   = null;
      
      contype = false;
   }
   :
      (    
           os:OPEN_START_TAG m:META
         | oc:OPEN_COMMENT wm:WSMETA { wsmeta = true; }
      )
      {
         if (sb != null)
         {
            if (os != null)
            {
               sb.append(os.getText()).append(m.getText());
            }
            else
            {
               sb.append(oc.getText()).append(wm.getText());
            }
         }
      }
      (
           // tolerate and ignore any whitespace
           ws:WS
           {
              if (sb != null)
              {
                 sb.append(ws.getText());
              }
           }
           
           // process name attribute
         | n:NAME
           {
              if (sb != null)
              {
                 sb.append(n.getText());
              }
           }
           prop = attribute_value[sb]
           {   
              if ("wsoptions".equalsIgnoreCase(prop))
              {
                 wsoptions = true;
              }
           }
           
           // process http-equiv attribute
         | he:HTTP_EQUIV
           {
              if (sb != null)
              {
                 sb.append(he.getText());
              }
           }
           prop = attribute_value[sb]
           {
              if ("content-type".equalsIgnoreCase(prop))
              {
                 contype = true;
              }
           }
           
           // process content attribute
         | c:CONTENT
           {
              if (sb != null)
              {
                 sb.append(c.getText());
              }
           }
           content = attribute_value[sb]
           
           // ignore any other attribute
         | s:SYMBOL
           {
              if (sb != null)
              {
                 sb.append(s.getText());
              }
           }
           prop = attribute_value[sb]
      )*
      (
           { !wsmeta }?
           ct:CLOSE_TAG
         | cc:CLOSE_COMMENT
      )
      {
         if (sb != null)
         {
            sb.append((ct != null) ? ct.getText() : cc.getText());
         }
      
         // only process if there is some content and we are in options
         // processing mode
         if (content != null && opts != null)
         {
            // only process specific matches to NAME = WSOPTIONS and
            // to HTTP-EQUIV = CONTENT-TYPE, otherwise ignore
            if (wsoptions)
            {
               opts.processWSOptions(content, true);
            }
            else if (contype)
            {
               opts.setContentType(content);
            }
         }
      }
   ;
   
/**
 * Parse the <code>=</code> followed by a <code>SYMBOL</code> or 
 * <code>STRING</code>. Whitespace on either side of the <code>EQUALS</code>
 * is tolerated.
 *
 * @param    sb
 *           Buffer in which to store all matched text if not
 *           <code>null</code>.
 *
 * @return   The text of the <code>SYMBOL</code> or <code>STRING</code>.
 */
attribute_value [StringBuilder sb] returns [String txt]
   {
      txt = null;
   }
   :
      (ws1:WS!)? eq:EQUALS! (ws2:WS!)? (s1:SYMBOL | txt = string)
      {
         if (s1 != null)
            txt = s1.getText();
         
         // capture all the text if needed
         if (sb != null)
         {
            if (ws1 != null)
               sb.append(ws1.getText());
               
            sb.append(eq.getText());
               
            if (ws2 != null)
               sb.append(ws2.getText());
               
            sb.append(txt);
         }
         
         if (s1 == null)
         {
            // remove the enclosing quotes
            txt = txt.substring(1, txt.length() - 1);
         }
         
         // no need for leading or trailing whitespace
         txt = txt.trim();
      }
   ;
   
/**
 * Matches various string types (data delimited by two matching single or
 * double quote characters). This is not implemented in the lexer since
 * there is no lookahead that can inform the lexer when it needs to match
 * paired quotes versus when it is acceptable for the quote characters to
 * stand alone. Since both are possible depending on parser context AND
 * since there is no reliable way to share state between the parser and
 * lexer, it was necessary to move the matching here.
 * <p>
 * Uses {@link #sstring} and {@link #dstring}.
 *
 * @return   The parsed string text. 
 */
string returns [String txt]
   {
      txt = null;
   }
   :
      txt = dstring | txt = sstring
   ;
   
/**
 * Matches any data delimited by two matching double quote characters.
 * <p>
 * See {@link #string}.
 *
 * @return   The parsed string text. 
 */
dstring returns [String retstr]
   {
      StringBuilder sb = new StringBuilder();
      retstr = null;
   }
   :
      start:DQUOTE      { sb.append(start.getText()); } 
      ( 
         txt:~DQUOTE    { sb.append(txt.getText()); }
      )*
      end:DQUOTE
      {
         sb.append(start.getText());
         retstr = sb.toString();
      }
   ;
   
/**
 * Matches any data delimited by two matching single quote characters.
 * <p>
 * See {@link #string}.
 *
 * @return   The parsed string text. 
 */
sstring returns [String retstr]
   {
      StringBuilder sb = new StringBuilder();
      retstr = null;
   }
   :
      start:SQUOTE      { sb.append(start.getText()); } 
      ( 
         txt:~SQUOTE    { sb.append(txt.getText()); }
      )*
      end:SQUOTE
      {
         sb.append(start.getText());
         retstr = sb.toString();
      }
   ;
   
/**
 * Reads an HTML file and provides a stream of tokens that can be readily
 * parsed. The central design point of this lexer is for all possible input
 * to be matched and passed through on output. For example, all whitespace
 * (with one exception, see below) will be matched and passed through. The
 * minimum number of tokens needed by the parser is generated.
 * <p> 
 * Most HTML constructs are NOT directly matched. Instead, this only
 * matches those constructs needed to unambiguously parse embedded 4GL
 * in an HTML page and translate that input into a valid 4GL program. No
 * 4GL constructs are matched here as the parser is not aware of any 4GL
 * syntax. Since the parser is only a filter (preprocessor), it only needs
 * to match on a small subset of HTML syntax such that it can make its
 * state transitions and so forth. Generally, there is a top-level rule
 * or artificial token type for every critical parser decision. Many of
 * these are single character matches but some more complicated 
 * multi-character constructs are matched as needed. Every possible ASCII
 * and extended ASCII character can be matched by itself. For any character
 * that doea not match a unique top-level rule or artificial token type, the
 * character can be returned as part of {@link #mWS} or {@link #mUNKNOWN}.
 * There are some places in the parser that matches on <code>WS</code> but
 * the content (token text) is not really important. Anything that is matched
 * as <code>UNKNOWN</code> is simply passed through by the parser grammar
 * rules, though on output the parser my do some tranformations as needed
 * to represent 4GL source code (e.g. unsafe character escaping).
 * <p>
 * Whitespace processing takes into account the platform-specific newline
 * string and maintains the newline counter as appropriate. In the case of
 * a platform's newline of "\r\n", the '\r' character immediately preceding
 * a '\n' character will be dropped. This allows the parser to match on a
 * simple '\n' in that case and on output the full platform-specific newline
 * will be inserted as needed. This also allows the '\r' to be dropped from
 * output if run on a system where the newline string is "\n".
 * <p>
 * Symbol processing is very simplistic. It leverages the built-in literals
 * testing (case-insensitively) of ANTLR. The key factor is that any text
 * that must be converted into an artificial token type must first match via
 * some other top-level lexer rule (see {@link #mSYMBOL}). The end of the
 * {@link #nextToken} rule then does a case-insensitive literal text lookup
 * with the list of hard coded string literals that are associated with
 * artificial token types. If a match is found, the token type of the
 * <code>SYMBOL</code> is changed to the corresponding artificial type.
 */
class E4GLLexer extends Lexer;

options
{
   k = 4;                                 // lookahead depth is 4 chars
   charVocabulary = '\000'..'\377';       // specifies range 0x00-0xFF
   caseSensitive = false;
   caseSensitiveLiterals = false;
}

tokens
{
   CONTENT      = "content";
   HTTP_EQUIV   = "http-equiv";
   LANGUAGE     = "language";
   META         = "meta";
   NAME         = "name";
   SCRIPT       = "script";
   SERVER       = "server";
   WSMETA       = "wsmeta";
   WSPEED       = "ws";
   WSE          = "wse";
   WSS          = "wss";
   WS4GL        = "ws4gl";
}
                                        
{
   /** Flag to force dropping CR characters from newlines. */
   private boolean dropCR = false;
   
   /** Flag to force maintaining newline based solely on CR characters. */
   private boolean newlineCR = false;
   
   // initializer
   {
      String nl = System.getProperty("line.separator");
      
      // newline of "\n" or "\r\n" will process as "\n", only a newline that
      // is a lone "\r" will trigger newline processing on the CR character
      if ("\r".equals(nl))
      {
         newlineCR = true;
      }
      
      // newline of "\n" will drop CR characters immediately preceding the
      // "\n", but in all other cases these "\r" will be left behind
      if ("\n".equals(nl))
      {
         dropCR = true;
      }
   }
   
   /**
    * Command line driver to allow lexer testing. The only argument should
    * be the source filename that should be lexed.
    *
    * @param    args                                           
    *           List of command line arguments.
    */
   public static void main(String[] args)
   {
      if (args.length != 1)
      {
         System.err.println("Syntax: java E4GLLexer <source_filename>");
         System.exit(-1);
      }
         
      try
      {
         // setup
         FileInputStream file   = new FileInputStream(args[0]);
         DataInputStream data   = new DataInputStream(file);
         E4GLLexer       lexer  = new E4GLLexer(data);
         LexerDumpHelper test   = new LexerDumpHelper(lexer,
                                                      E4GLParser._tokenNames,
                                                      System.out);
         
         // start the lexing and print a record for each token                  
         test.dump();
      }
       
      catch(Exception excpt)
      {
         excpt.printStackTrace();
      }
   }
}

/**                 
 * Matches any amount of whitespace in a program. Each newline character
 * causes the lexer's line counter to be incremented in order to properly 
 * maintain each token's line information.
 * <p>
 * Spaces, tabs and carriage returns and line feeds (newlines) are all
 * matched.
 * <p>
 * Non-visible ASCII codes decimal 127 and below are matched here as well. 
 * <p>
 * This is a top level lexer rule which means that there is an associated
 * <code>WS</code> token.
 */
WS
   :
      (
           ' '
         | '\t'
         | { dropCR && LA(2) == '\n' }?
           '\r'!
         | '\r' { if (newlineCR)  newline(); }
         | '\n' { if (!newlineCR) newline(); }
         | JUNK
      )+
   ;

/**
 * Matches the double quote character.
 */
DQUOTE
   :
      '\"'
   ;
   
/**
 * Matches a single quote character.
 */
SQUOTE
   :
      '\''
   ;
   
/**
 * Match any valid HTML name or id. This includes all alphanumeric characters
 * hyphen, underscore, colon and period. Once this matches, the resulting
 * text will be compared to the list of hard-coded literals on a 
 * case-insensitive basis. The token type will be converted to that of the
 * matching literal if a match is found via this literal testing.
 */
SYMBOL
   :
      (LETTER | DIGIT) ( LETTER | DIGIT | HYPHEN | UNDERSCORE | COLON | DOT )*
   ;

/**
 * Match '%' followed by 2 hexidecimal digits.
 */
ENCODED_CHAR
   {
      $setType(UNKNOWN);
   }
   :
      // part of this rule is optional to allow lone front parts to be seen
      // as UNKNOWN, ANTLR is forced to generate a full lookahead to
      // these cases AND it eliminates runtime ambiguity
      PERCENT HEX_DIGIT ( HEX_DIGIT { $setType(ENCODED_CHAR); } )?
   ;
   
/**
 * Match '&lt;!--WSE' used to open one form of expression escape.
 */
OPEN_PCT_EQ
   :
      LT PERCENT EQUALS
   ;
   
/**
 * Match '&lt;!--WSE' used to open one form of expression escape.
 */
OPEN_PCT
   :
      LT PERCENT
   ;
   
/**
 * Match '&lt;!--WSE' used to open one form of expression escape.
 */
CLOSE_PCT
   :
      PERCENT GT
   ;
   
/**
 * Match '&lt;?' used to open one form of statement escape.
 */
OPEN_QUESTION
   :
      LT QUESTION
   ;
   
/**
 * Match '&lt;?' used to open one form of statement escape.
 */
CLOSE_QUESTION
   :
      OPEN_END_TAG QUESTION
   ;
   
/**
 * Match '{=' used to open one form of expression escape.
 */
OPEN_CURLY_EQ
   :
      LEFT_CURLY EQUALS
   ;
   
/**
 * Match '=}' used to close one form of expression escape.
 */
CLOSE_CURLY_EQ
   :
      EQUALS RIGHT_CURLY
   ;
   
/**
 * Match everything else (this is the inverse of all other top-level lexer
 * rules) as unknown text.
 */
UNKNOWN
   :
      (
           PERCENT
         | SLASH
         | EXCLAIM
         | HYPHEN
         | UNDERSCORE
         | COLON
         | QUESTION
         | LEFT_CURLY
         | RIGHT_CURLY
         | DOT
         | OTHER
      )
   ;
   
/**
 * Matches a single '=' character.
 */
EQUALS
   :
      '='
   ;
  
/**
 * Matches a single '`' (back-tick) character.
 */
BACK_TICK
   :
      '`'
   ;
  
/**
 * Matches the HTML open comment "&lt;!--".
 */
OPEN_COMMENT
   {
      $setType(UNKNOWN);
   }
   :
      // we make some of this rule optional to match lone front parts of
      // that can occur, this forces ANTLR to generate a full lookahead to
      // differentiate and it eliminates ambiguity at runtime
      LT EXCLAIM ( HYPHEN ( HYPHEN { $setType(OPEN_COMMENT); })? )?
   ;
  
/**
 * Matches a single '&lt;' character, normally used to start an opening HTML
 * element.
 */
OPEN_START_TAG
   :
      LT
   ;
  
/**
 * Matches a single '&lt;' character followed by a '/', normally
 * used to start a closing HTML element.
 */
OPEN_END_TAG
   :
      LT SLASH
   ;
  
/**
 * Matches the HTML close comment "--&gt;".
 */
CLOSE_COMMENT
   {
      $setType(UNKNOWN);
   }
   :
      // we make some of this rule optional to match the lone front part
      // that can occur, this forces ANTLR to generate a full lookahead to
      // differentiate and it eliminates ambiguity at runtime
      HYPHEN HYPHEN (GT { $setType(CLOSE_COMMENT); })?
   ;
  
/**
 * Matches a '/' followed by a '&gt;' normally used to close an HTML element
 * that has no content.
 */
CLOSE_TAG_NO_CONTENT
   :
      SLASH GT
   ;
  
/**
 * Matches a single '&gt;' character normally used to close an HTML element.
 */
CLOSE_TAG
   :
      GT
   ;
  
/**
 * String delimited by 2 double-quote characters. May include any text inside
 * and all will be matched. NOT USED but left for reference.
 */
protected
HTML_DSTRING
   :
      (
         '\"'                             // opening "
              (
                 options { generateAmbigWarnings = false; }
                 :
                   { dropCR && LA(2) == '\n' }?
                   '\r'!
                 | '\r' { if (newlineCR)  newline(); } // preserve counter
                 | '\n' { if (!newlineCR) newline(); } // preserve counter
                 | ~'\"'                   // anything else except closing "
              )*
         '\"'                              // closing "
      )
   ;
  
/**
 * String delimited by 2 single-quote characters. May include any text inside
 * and all will be matched. NOT USED but left for reference.
 */
protected
HTML_SSTRING
   :
      (
         '\''                              // opening '
              (
                 options { generateAmbigWarnings = false; }
                 :
                   { dropCR && LA(2) == '\n' }?
                   '\r'!
                 | '\r' { if (newlineCR)  newline(); } // preserve counter
                 | '\n' { if (!newlineCR) newline(); } // preserve counter
                 | ~'\''                   // anything else except closing "
              )*
         '\''                             // closing '
      )
   ;
  
/**
 * Matches a single '&gt;' character.
 */
protected
GT
   :
      '>'
   ;
  
/**
 * Matches a single '&lt;' character.
 */
protected
LT
   :
      '<'
   ;
  
/**
 * Matches a single '/' character.
 */
protected
SLASH
   :
      '/'
   ;
  
/**
 * Matches a single '%' character.
 */
protected
PERCENT
   :
      '%'
   ;
  
/**
 * Matches a single '!' character.
 */
protected
EXCLAIM
   :
      '!'
   ;
  
/**
 * Matches a single '-' character.
 */
protected
HYPHEN
   :
      '-'
   ;
  
/**
 * Matches a single '_' character.
 */
protected
UNDERSCORE
   :
      '_'
   ;
  
/**
 * Matches a single '.' character.
 */
protected
DOT
   :
      '.'
   ;
  
/**
 * Matches a single ':' character.
 */
protected
COLON
   :
      ':'
   ;
  
/**
 * Matches a single '?' character.
 */
protected
QUESTION
   :
      '?'
   ;
  
/**
 * Matches a single '{' character.
 */
protected
LEFT_CURLY
   :
      '{'
   ;
  
/**
 * Matches a single '}' character.
 */
protected
RIGHT_CURLY
   :
      '}'
   ;
  
/**
 * Matches any alphabetic character: <code>a - z</code>.
 */   
protected   
LETTER
   :
      'a'..'z'
   ;
   
/**
 * Matches any numeric digit: <code>0 - 9</code>.
 */
protected
DIGIT
   :
      '0'..'9'
   ;

/**
 * Matches any hexidecimal digit: <code>0 - 9</code> and <code>a - f</code>.
 */
protected
HEX_DIGIT
   :
      '0'..'9' | 'a'..'f' 
   ;

/**
 * Matches all other unknown, non-whitespace characters, including the
 * extended ASCII range.
 */
protected
OTHER
   :
      '#' | '$' | '&'  | '(' | ')' | '*' | '+' | ',' | ';' |
      '@' | '[' | '\\' | ']' | '^' | '|' | '~' | '\200'..'\377'
   ;

/**
 * Matches non-visible ASCII codes that should be silently ignored.
 */
protected
JUNK
   :
      '\000'..'\010' | '\013'..'\014' | '\016'..'\037' | '\177'
   ;