/*
** Module   : MultiReader.java
** Abstract : a reader which allows its input to be switched
**
** Copyright (c) 2008-2017, Golden Code Development Corporation.
**
** -#- -I- --Date-- -T- --JPRM-- ----------------Description-----------------
** 001 GES 20080308 NEW   @37541 First version which multiplexes readers
**                               into a single reader.
*/
/*
** This program is free software: you can redistribute it and/or modify
** it under the terms of the GNU Affero General Public License as
** published by the Free Software Foundation, either version 3 of the
** License, or (at your option) any later version.
**
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
** GNU Affero General Public License for more details.
**
** You may find a copy of the GNU Affero GPL version 3 at the following
** location: https://www.gnu.org/licenses/agpl-3.0.en.html
** 
** Additional terms under GNU Affero GPL version 3 section 7:
** 
**   Under Section 7 of the GNU Affero GPL version 3, the following additional
**   terms apply to the works covered under the License.  These additional terms
**   are non-permissive additional terms allowed under Section 7 of the GNU
**   Affero GPL version 3 and may not be removed by you.
** 
**   0. Attribution Requirement.
** 
**     You must preserve all legal notices or author attributions in the covered
**     work or Appropriate Legal Notices displayed by works containing the covered
**     work.  You may not remove from the covered work any author or developer
**     credit already included within the covered work.
** 
**   1. No License To Use Trademarks.
** 
**     This license does not grant any license or rights to use the trademarks
**     Golden Code, FWD, any Golden Code or FWD logo, or any other trademarks
**     of Golden Code Development Corporation. You are not authorized to use the
**     name Golden Code, FWD, or the names of any author or contributor, for
**     publicity purposes without written authorization.
** 
**   2. No Misrepresentation of Affiliation.
** 
**     You may not represent yourself as Golden Code Development Corporation or FWD.
** 
**     You may not represent yourself for publicity purposes as associated with
**     Golden Code Development Corporation, FWD, or any author or contributor to
**     the covered work, without written authorization.
** 
**   3. No Misrepresentation of Source or Origin.
** 
**     You may not represent the covered work as solely your work.  All modified
**     versions of the covered work must be marked in a reasonable way to make it
**     clear that the modified work is not originating from Golden Code Development
**     Corporation or FWD.  All modified versions must contain the notices of
**     attribution required in this license.
*/

package com.goldencode.p2j.io;

import java.io.*;

/**
 * Implements a simple switchable reader based on any number of underlying
 * readers. Readers can be switched to at any time. There is no history or
 * buffering kept inside this class. This means that there is no reading
 * ahead and no stacking or automatic "switch back" capability.  Likewise
 * only one reader is used at a time and once replaced, the previous reader
 * is not referenced or remembered in any way. 
 * <p>
 * When instantiated, there may be no connection to any reader. If the
 * {@link #read} method is called while in this state, a <code>-1</code> is
 * returned signalling the <code>EOF</code> condition.
 * <p>
 * Once the {@link #switchTo} method is called, <code>read()</code>
 * delegates the work to the <code>read()</code> method of the current
 * underlying reader.
 * <p>
 * This class may be useful in many situations. For instance,
 * <code>PushbackReader</code> can provide pushback functionality for multiple
 * underlying readers once in an efficient way.
 */
public class MultiReader
extends FilterReader
{
   /**
    * Default constructor. Creates an instance with no associated reader.
    */
   public MultiReader()
   {
      super(new StringReader(""));
   }

   /**
    * Creates an instance with the specified associated reader as the
    * current reader.
    *
    * @param    reader
    *           The reader that should be used first.
    */
   public MultiReader(Reader reader)
   {
      super(reader);
   }

   /**
    * Switches input to the specified reader.
    *
    * @param   reader
    *          The reader to be switched to.
    */
   public void switchTo(Reader reader)
   {
      super.in = reader;
   }

   /**
    * Reads the next byte from the current stream. If no stream was
    * switched to, returns <code>-1</code> as the indication of the
    * <code>EOF</code> condition.
    *
    * @return   The character read, as an integer in the range 0 to 65535
    *           (0x0000-0xFFFF), or -1 if the end of the stream has been
    *           reached or if no reader was switched to.
    *
    * @throws   IOException
    *           If any error occurs in reading from the underlying reader.
    */
   public int read()
   throws IOException
   {
      if (super.in == null)
      {
         return -1;
      }
      
      return super.read();
   }
}