/*
** Module   : NativeTypeArray.java
** Abstract : an array of descriptors passed as a single argument to a native function 
**
** Copyright (c) 2013-2017, Golden Code Development Corporation.
**
** -#- -I- --Date-- ---------------------------------Description---------------------------------
** 001 GES 20131211 First version.
** 002 GES 20140130 Even arrays need to be passed as a pointer to a pointer (one more level of
**                  indirection than we were providing).
*/
/*
** This program is free software: you can redistribute it and/or modify
** it under the terms of the GNU Affero General Public License as
** published by the Free Software Foundation, either version 3 of the
** License, or (at your option) any later version.
**
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
** GNU Affero General Public License for more details.
**
** You may find a copy of the GNU Affero GPL version 3 at the following
** location: https://www.gnu.org/licenses/agpl-3.0.en.html
** 
** Additional terms under GNU Affero GPL version 3 section 7:
** 
**   Under Section 7 of the GNU Affero GPL version 3, the following additional
**   terms apply to the works covered under the License.  These additional terms
**   are non-permissive additional terms allowed under Section 7 of the GNU
**   Affero GPL version 3 and may not be removed by you.
** 
**   0. Attribution Requirement.
** 
**     You must preserve all legal notices or author attributions in the covered
**     work or Appropriate Legal Notices displayed by works containing the covered
**     work.  You may not remove from the covered work any author or developer
**     credit already included within the covered work.
** 
**   1. No License To Use Trademarks.
** 
**     This license does not grant any license or rights to use the trademarks
**     Golden Code, FWD, any Golden Code or FWD logo, or any other trademarks
**     of Golden Code Development Corporation. You are not authorized to use the
**     name Golden Code, FWD, or the names of any author or contributor, for
**     publicity purposes without written authorization.
** 
**   2. No Misrepresentation of Affiliation.
** 
**     You may not represent yourself as Golden Code Development Corporation or FWD.
** 
**     You may not represent yourself for publicity purposes as associated with
**     Golden Code Development Corporation, FWD, or any author or contributor to
**     the covered work, without written authorization.
** 
**   3. No Misrepresentation of Source or Origin.
** 
**     You may not represent the covered work as solely your work.  All modified
**     versions of the covered work must be marked in a reasonable way to make it
**     clear that the modified work is not originating from Golden Code Development
**     Corporation or FWD.  All modified versions must contain the notices of
**     attribution required in this license.
*/

package com.goldencode.p2j.library;

import java.io.*;
import com.goldencode.p2j.util.*;

/**
 * An array of descriptors passed as a single argument to a native function.  4GL RETURN
 * parameters cannot have an extent, so this can only be used for INPUT, INPUT-OUTPUT and
 * OUTPUT parameters.  This class hides the complexity of rendering/restoring arrays to
 * native memory.
 */
public class NativeTypeArray
extends BaseNativeType
{
   /** Stores the contained elements for the array. */
   private BaseNativeType[] elements = new BaseNativeType[0];
   
   /**
    * Default constructor.
    */
   public NativeTypeArray()
   {
      super();
      setPassByPointer(true);
      type = LibraryManager.TYPE_POINTER;
   }
   
   /**
    * Constructor that takes a list of elements.
    *
    * @param    elements
    *           The elements for this array.
    */
   public NativeTypeArray(BaseNativeType[] elements)
   {
      this();
      this.elements = elements;
   }
   
   /**
    * Access the array of data elements.
    *
    * @return   The array of elements.
    */
   public BaseNativeType[] getElements()
   {
      return elements;
   }
   
   /**
    * Assign the array of data elements.
    *
    * @param    elements
    *           The array of elements.
    */
   public void setElements(BaseNativeType[] elements)
   {
      this.elements = elements;
   }
   
   /**
    * Report the size (in bytes) of the native memory needed to store this instance.
    * <p>
    * This should never be called on the server, only on the client side.
    *
    * @return   The number of bytes needed.
    */
   public int size()
   {
      return (elements.length > 0) ? elements.length * elements[0].size() : 0;
   }
   
   /**
    * Render the data into native memory, which will be allocated if the address is not
    * given.
    * <p>
    * This should never be called on the server, only on the client side.
    * <p>
    * Arrays are rendered as a pointer to a block of memory in which each element is
    * written contiguously from lowest address to highest address and the size of the
    * block is the number of elements multiplied by the size of each element (which 
    * must be the same for all elements). This is consistent with how native code ('C'
    * language) handles arrays.
    *
    * @param    addr
    *           The memory address to which the data should be rendered, if the caller
    *           is supplying a buffer or 0 if the buffer should be allocated.
    *
    * @return   The memory address of the rendered data or 0 if there is no data to
    *           output.
    */
   public long render(long addr)
   {
      if (elements.length > 0)
      {
         // all elements of the array must have the same size
         int sz = elements[0].size();
         
         memptr local = new memptr();
         
         if (addr == 0)
         {
            // save off for restore use later
            ref = local;
            
            // allocate the block which must be big enough to store the array
            local.setLength(elements.length * sz);
            
            addr = local.getPointerValue().longValue();
         }
         else
         {
            local.setPointerValue(addr);
         }
         
         if (copyIn || elements[0].isPassByPointer())
         {
            for (int i = 0; i < elements.length; i++)
            {
               // calculate the address at which to write our next element
               long idx = addr + (i * sz);
               
               // give each element a chance to render itself
               elements[i].render(idx);
            }
         }
      }
      else
      {
         addr = 0;
      }
      
      return addr;
   }
   
   /**
    * Copy back (if necessary) data from the native memory and deallocate the memory
    * if this instance allocated it in the first place.
    * <p>
    * This should never be called on the server, only on the client side.
    *
    * @param    addr
    *           The memory address from which the data should be restored, if the caller
    *           is supplying a buffer or 0 if the buffer should already exist.
    */
   public void restore(long addr)
   {
      memptr local = ref;
      
      if (local == null && addr != 0)
      {
         local = new memptr();
         local.setPointerValue(addr);
      }
      
      if (copyOut && local != null && elements.length > 0)
      {
         addr = local.getPointerValue().longValue();
         
         // all elements of the array must have the same size
         int sz = elements[0].size();
         
         // each element must have a chance to restore and deallocate the native resources
         for (int i = 0; i < elements.length; i++)
         {
            // calculate the address at which to read our next element
            long idx = addr + (i * sz);
            
            elements[i].restore(idx);
         }
      }
      
      // deallocate memory if we allocated it
      if (ref != null)
         ref.setLength(0);
   }
   
   /**
    * Replacement for the default object reading method. The latest
    * state is read from the input source.  The parent's state is
    * read first.
    * 
    * @param    in
    *           The input source from which fields will be restored.
    *
    * @throws   IOException
    *           In case of I/O errors.
    * @throws   ClassNotFoundException
    *           If payload can't be instantiated.
    */
   public void readExternal(ObjectInput in)
   throws IOException,
          ClassNotFoundException
   {
      super.readExternal(in);
      
      int len = in.readInt();
      
      elements = new BaseNativeType[len];
      
      for (int i = 0; i < len; i++)
      {
         elements[i] = (BaseNativeType) in.readObject();
      }
   }

   /**
    * Replacement for the default object writing method. The latest
    * state is written to the output destination.  The parent's state is
    * written first.
    * 
    * @param    out
    *           The output destination to which fields will be saved.
    *
    * @throws   IOException
    *           In case of I/O errors.
    */
   public void writeExternal(ObjectOutput out)
   throws IOException
   {
      super.writeExternal(out);
      
      out.writeInt(elements.length);
      
      for (int i = 0; i < elements.length; i++)
      {
         out.writeObject(elements[i]);
      }
   }
}