/*
** Module   : ClientBuilderOptions.java
** Abstract : store the parameters read from directory.
**
** Copyright (c) 2013-2017, Golden Code Development Corporation.
**
** -#- -I- --Date-- -------------------------------Description------------------------------------
** 001 MAG 20131128 First version.
** 002 MAG 20140128 Add WebSocket timeout.
** 003 MAG 20140204 Add Watchdog timeout.
** 004 CA  20140206 Refactored to make it unaware of the type of the P2J client being spawned.
** 005 GES 20150211 Removed application arguments support. Forced headless to be always present
**                  in JVM args. No other JVM arg is defaulted.
** 006 GES 20160203 Moved core initialization logic into a private method.  Added a constructor
**                  for cloning, which doesn't call the core init but instead copies state from
**                  the given instance. This is useful for subclasses which need to avoid
**                  executing the initialization logic at runtime (e.g. because of security
**                  context constraints).
*/
/*
** This program is free software: you can redistribute it and/or modify
** it under the terms of the GNU Affero General Public License as
** published by the Free Software Foundation, either version 3 of the
** License, or (at your option) any later version.
**
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
** GNU Affero General Public License for more details.
**
** You may find a copy of the GNU Affero GPL version 3 at the following
** location: https://www.gnu.org/licenses/agpl-3.0.en.html
** 
** Additional terms under GNU Affero GPL version 3 section 7:
** 
**   Under Section 7 of the GNU Affero GPL version 3, the following additional
**   terms apply to the works covered under the License.  These additional terms
**   are non-permissive additional terms allowed under Section 7 of the GNU
**   Affero GPL version 3 and may not be removed by you.
** 
**   0. Attribution Requirement.
** 
**     You must preserve all legal notices or author attributions in the covered
**     work or Appropriate Legal Notices displayed by works containing the covered
**     work.  You may not remove from the covered work any author or developer
**     credit already included within the covered work.
** 
**   1. No License To Use Trademarks.
** 
**     This license does not grant any license or rights to use the trademarks
**     Golden Code, FWD, any Golden Code or FWD logo, or any other trademarks
**     of Golden Code Development Corporation. You are not authorized to use the
**     name Golden Code, FWD, or the names of any author or contributor, for
**     publicity purposes without written authorization.
** 
**   2. No Misrepresentation of Affiliation.
** 
**     You may not represent yourself as Golden Code Development Corporation or FWD.
** 
**     You may not represent yourself for publicity purposes as associated with
**     Golden Code Development Corporation, FWD, or any author or contributor to
**     the covered work, without written authorization.
** 
**   3. No Misrepresentation of Source or Origin.
** 
**     You may not represent the covered work as solely your work.  All modified
**     versions of the covered work must be marked in a reasonable way to make it
**     clear that the modified work is not originating from Golden Code Development
**     Corporation or FWD.  All modified versions must contain the notices of
**     attribution required in this license.
*/

package com.goldencode.p2j.main;

import java.io.*;
import java.net.*;
import java.util.*;
import java.util.logging.*;

import com.goldencode.p2j.util.*;

/**
 * The parameters stored inside this structure are read from directory and are used as default 
 * parameters to create and spawn a P2J client process.
 */
public abstract class ClientBuilderOptions
{
   /** Logger. */
   private static final Logger LOG = LogHelper.getLogger(ClientBuilderOptions.class.getName());
   
   /** Directory node name. */
   private static final String DIRECTORY_NODE_ID = "clientConfig";
   
   /** Main class, must have a main method */
   private final static Class<?> mainClass = ClientDriver.class;
   
   /** Runtime calculated path to our main jar file. */
   private final static String defaultPath = getDefaultPath();

   /** Secure socket flag. */
   private boolean secure = true;
   
   /** Java launcher executable name. */
   private String command = "java";

   /** User's working directory. */
   private String workingDirectory = "./";

   /** Class path. */
   private String classPath;
   
   /** Native library path. */
   private String libPath;

   /** Spawner executable name. */
   private String spawner;
   
   /** Client configuration file. */
   private String configFile;

   /** JVM arguments. */
   private List<String> jvmArguments = new LinkedList<String>();

   /** P2J client bootstrap config */
   private List<String> bootstrapConfig = new LinkedList<String>();
   
   /**
    * All the subject IDs associated with this builder or <code>null</code> to use the 
    * context's subject IDs.
    */
   protected final String[] accountIds;

   /**
    * Initialize. Read all configuration parameters from directory.
    * 
    * @param   accountIds
    *          All the subject IDs associated with this builder or <code>null</code> to use the 
    *          context's subject IDs.
    */
   protected ClientBuilderOptions(String[] accountIds)
   {
      this.accountIds = accountIds;
      
      initialize();
   }
   
   /**
    * Initialize by reading all configuration parameters from a previously created instance.
    * This avoids reading from the directory which can be useful in the case where security
    * context would otherwise change the result.
    *
    * @param    copy
    *           The instance from which to initialize.
    * @param    accountIds
    *           All the subject IDs associated with this builder or <code>null</code> to use the 
    *           context's subject IDs.
    */
   protected ClientBuilderOptions(ClientBuilderOptions copy, String[] accountIds)
   {
      this.accountIds = accountIds;
      
      this.secure           = copy.secure;
      this.command          = copy.command;
      this.classPath        = copy.classPath;
      this.libPath          = copy.libPath;
      this.spawner          = copy.spawner;
      this.workingDirectory = copy.workingDirectory;
      this.configFile       = copy.configFile;
      this.jvmArguments     = new LinkedList<String>(copy.jvmArguments);
      this.bootstrapConfig  = new LinkedList<String>(copy.bootstrapConfig);
   }
   
   /**
    * Get the authentication mode for this spawner.
    * 
    * @return   {@code true} indicating that the user's password is required for authentication.
    */
   protected abstract boolean isPasswordAuthentication();
   
   /**
    * Get the security mode.
    * 
    * @return  <code>true</code> use a secured connection <code>false</code> otherwise.
    */
   public boolean isSecure()
   {
      return secure;
   }

   /**
    * Get command
    * 
    * @return  Command name string.
    */
   public String getCommand()
   {
      return command;
   }
   
   /**
    * Get classpath
    * 
    * @return  Class path string.
    */
   public String getClassPath()
   {
      return classPath;
   }

   /**
    * Get native library path.
    * 
    * @return  Native library path string.
    */
   public String getLibPath()
   {
      return libPath;
   }

   /**
    * Get user working directory.
    * 
    * @return  User working directory.
    */
   public String getWorkingDirectory()
   {
      return workingDirectory;
   }

   /**
    * Get configuration file name.
    * 
    * @return  Configuration file name.
    */
   public String getConfigFile()
   {
      return configFile;
   }
   
   /**
    * Get JVM arguments
    * 
    * @return   A List with JVM arguments.
    */
   public List<String> getJvmArguments()
   {
      return jvmArguments;
   }
   
   /**
    * Get P2J client's bootstrap config.
    * 
    * @return   A List with bootstrap config settings.
    */
   public List<String> getBootstrapConfig()
   {
      return bootstrapConfig;
   }

   /**
    * Get spawner command.
    * 
    * @return  The spawner command;
    */
   public String getSpawner()
   {
      return spawner;
   }

   /**
    * Get the Java process main class.
    * 
    * @return  The main class.
    */
   public Class<?> getMainClass()
   {
      return mainClass;
   }

   /**
    * When the string contains space characters a quoted string is build.
    * On Windows OS spaces could be used in file path, file name, font name and other
    * strings. In order to use such kind of strings in API calls the string should be
    * transformed in quoted form. 
    * 
    * @param   text  Text to search for spaces.
    * 
    * @return  If no spaces found return the initial string otherwise return a quoted string. 
    */
   protected static String quotedString(String text)
   {
      if (text == null)
      {
         return null;
      }
      
      if (text.startsWith("\"") && text.endsWith("\""))
      {
         return text;
      }
      return (text.indexOf(" ") > -1) ? "\\\"" + text +"\\\"" : text;
   }

   /**
    * Get a string value from a directory node.
    *  
    * @param   parent
    *          The name of the parent node.
    * @param   node
    *          Node name.
    * @param   defa
    *          Default value if node not found.
    *          
    * @return  Node value if exists or the default value otherwise.
    */
   protected String getNode(String parent, String node, String defa)
   {
      return Utils.getDirectoryNodeString(null, parent + "/" + node, defa, false);
   }

   /**
    * Get an integer value from a directory node.
    * 
    * @param   parent
    *          The name of the parent node.
    * @param   node
    *          Node name.
    * @param   defa
    *          Default value if node not found.
    *          
    * @return  Node value if exists or the default value otherwise.
    */
   protected int getNode(String parent, String node, int defa)
   {
      return Utils.getDirectoryNodeInt(null, parent + "/" + node, defa, false);
   }

   /**
    * Get a boolean value from a directory node.
    *  
    * @param   parent
    *          The name of the parent node.
    * @param   node
    *          Node name.
    * @param   defa
    *          Default value if node not found.
    *          
    * @return  Node value if exists or the default value otherwise.
    */
   protected boolean getNode(String parent, String node, boolean defa)
   {
      return Utils.getDirectoryNodeBoolean(null, parent + "/" + node, defa, false);
   }

   /**
    * Get default path as the p2j.jar file location.
    * 
    * @return  Path to p2j.jar file.
    */
   private static String getDefaultPath()
   {
      String message = "Could not determine the location of the p2j.jar file.";
      String defaultPath = "./";
      
      try
      {
         URL url = mainClass.getProtectionDomain().getCodeSource().getLocation();
         String classPath = url.toURI().getPath();
         File jarFile = new File(classPath);
         
         // if the source is in a folder and not jar, then this can cause problems
         if (!jarFile.exists())
         {
            LOG.logp(Level.WARNING,
                     "ClientBuilderOptions.getDefaultPath()",
                     "",
                     LogHelper.generate(message));
         }
         else
         {
            // default path
            defaultPath = jarFile.getParentFile().getPath();
            
            if (!defaultPath.endsWith(File.separator))
            {
               defaultPath += File.separator;
            }
         }
      }
      
      catch (URISyntaxException e)
      {
         LOG.logp(Level.WARNING,
                  "ClientBuilderOptions.getDefaultPath()",
                  "",
                  LogHelper.generate(message), 
                  e);
      }
      
      return defaultPath;
   }
   
   /**
    * Read the configuration from the directory and calculate certain runtime values that will
    * be used by all client launching.
    */
   private void initialize()
   {
      // set default values using default path
      libPath = defaultPath;
      classPath = defaultPath  + "p2j.jar";
      spawner = defaultPath + "spawn";
      
      // secure (TODO: we should not support insecure mode)
      secure = getNode(DIRECTORY_NODE_ID, "secure", true);
      
      command = quotedString(getNode(DIRECTORY_NODE_ID, "command", command));
      classPath = quotedString(getNode(DIRECTORY_NODE_ID, "classpath", classPath));
      libPath = quotedString(getNode(DIRECTORY_NODE_ID, "libPath", libPath));
      spawner = quotedString(getNode(DIRECTORY_NODE_ID, "spawner", spawner)); 
      workingDirectory = quotedString(getNode(DIRECTORY_NODE_ID, "workingDir", workingDirectory)); 
      configFile = quotedString(getNode(DIRECTORY_NODE_ID, "configFile", configFile));
      
      // JVM arguments
      String jvm = getNode(DIRECTORY_NODE_ID, "jvmArgs", "");
      
      // we always must pass headless mode
      if (jvm.indexOf("-Djava.awt.headless=true") == -1)
      {
         String pad = (jvm.length() == 0) ? "" : " ";
         
         jvm = jvm + pad + "-Djava.awt.headless=true";
      }
      
      jvmArguments.addAll(Arrays.asList(jvm.split(" ")));
      
      // bootstrap config overrides
      String bootstrapCfg = getNode(DIRECTORY_NODE_ID, "cfgOverrides", "");
      bootstrapConfig.addAll(Arrays.asList(bootstrapCfg.split(" ")));
   }
}