/*
** Module   : ClientSpawner.java
** Abstract : define a generic client Spawner.
**
** Copyright (c) 2013-2017, Golden Code Development Corporation.
**
** -#- -I- --Date-- -------------------------------Description------------------------------------
** 001 MAG 20131119 First version.
** 002 CA  20140206 Refactored to make it unaware of the type of the P2J client being spawned.
** 003 MAG 20140710 Implements remote launcher.
** 004 OM  20150108 Added optional appserver specific environment settings to spawned process.
** 005 GES 20150211 Added P2J_HOME as a standard part of the environment of a spawned child.
** 006 GES 20160219 Added logging of non-zero exit codes.
** 007 SBI 20160628 Added new methods to allocate and to release the client's system resources.
** 008 SBI 20171019 Changed to spawn remote clients on dedicated hosts.
*/
/*
** This program is free software: you can redistribute it and/or modify
** it under the terms of the GNU Affero General Public License as
** published by the Free Software Foundation, either version 3 of the
** License, or (at your option) any later version.
**
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
** GNU Affero General Public License for more details.
**
** You may find a copy of the GNU Affero GPL version 3 at the following
** location: https://www.gnu.org/licenses/agpl-3.0.en.html
** 
** Additional terms under GNU Affero GPL version 3 section 7:
** 
**   Under Section 7 of the GNU Affero GPL version 3, the following additional
**   terms apply to the works covered under the License.  These additional terms
**   are non-permissive additional terms allowed under Section 7 of the GNU
**   Affero GPL version 3 and may not be removed by you.
** 
**   0. Attribution Requirement.
** 
**     You must preserve all legal notices or author attributions in the covered
**     work or Appropriate Legal Notices displayed by works containing the covered
**     work.  You may not remove from the covered work any author or developer
**     credit already included within the covered work.
** 
**   1. No License To Use Trademarks.
** 
**     This license does not grant any license or rights to use the trademarks
**     Golden Code, FWD, any Golden Code or FWD logo, or any other trademarks
**     of Golden Code Development Corporation. You are not authorized to use the
**     name Golden Code, FWD, or the names of any author or contributor, for
**     publicity purposes without written authorization.
** 
**   2. No Misrepresentation of Affiliation.
** 
**     You may not represent yourself as Golden Code Development Corporation or FWD.
** 
**     You may not represent yourself for publicity purposes as associated with
**     Golden Code Development Corporation, FWD, or any author or contributor to
**     the covered work, without written authorization.
** 
**   3. No Misrepresentation of Source or Origin.
** 
**     You may not represent the covered work as solely your work.  All modified
**     versions of the covered work must be marked in a reasonable way to make it
**     clear that the modified work is not originating from Golden Code Development
**     Corporation or FWD.  All modified versions must contain the notices of
**     attribution required in this license.
*/

package com.goldencode.p2j.main;

import java.io.Serializable;
import java.util.*;
import java.util.concurrent.*;
import java.util.logging.*;

import com.goldencode.p2j.util.*;

/**
 * This defines a way of spawning new P2J client processes.  Temporary credentials are sent to
 * the P2J client side, which, if used, will establish a P2J session with the P2J server and do
 * some custom work (depending on the implementation).
 * <p>
 * Actual implementations of this class will need to provide the following info:
 * <ul>
 *    <li>A {@link ClientBuilder} instance, to spawn the P2J client process.</li>
 *    <li>A {@link ClientBuilderParameters} instance, to configure the P2J client.</li>
 *    <li>Implementation for the {@link SpawnerListener#getServerData()}, to send custom data
 *        back to the P2J client.</li>
 *    <li>Implementation for the {@link SpawnerListener#getTemporaryClient()}, with the work
 *        needed to be done after authenticating with the temporary credentials.</li>
 * </ul>
 * 
 * @param    R
 *           The type that holds the system resources allocated for the spawned client. 
*/
public abstract class ClientSpawner<R extends Serializable>
implements SpawnerListener<R>
{
   /** Logger. */
   private static final Logger LOG = LogHelper.getLogger(ClientSpawner.class.getName());
   
   /** Timeout */
   private static final int TIME_OUT = 30;
   
   /** The {@link ClientBuilder} used to spawn the P2J client process. */
   protected ClientBuilder cb;
   
   /** The configuration for this spawner. */
   protected final ClientBuilderParameters cfg;

   /** Spawner exit code */
   private int exitCode = 0;

   /** Synchronize */
   private CountDownLatch latch;

   /** Flag indicating the P2J client has been started. */
   private volatile boolean ready = false;

   /**
    * Configure a new client spawner.
    * 
    * @param    cfg
    *           The spawner configuration.
    */
   public ClientSpawner(ClientBuilderParameters cfg)
   {
      this.cfg = cfg;
   }
   
   /**
    * Get a {@link ClientBuilder} used to spawn P2J clients.
    * 
    * @param    cfg
    *           The P2J client configuration.
    *           
    * @return   See above.
    */
   protected abstract ClientBuilder getBuilder(ClientBuilderParameters cfg);

   /**
    * Launch a P2J client local or remote. Get a list of brokers available for this user.
    * If no brokers are registered for this user the client is spawn local.
    * If at least one broker is registered the client is spawn remote.
    * 
    * @param   userName
    *          User name.
    * @param   password
    *          User password.
    * @param   envVars
    *          Environment variables to set. May be <code>null</code>.
    */
   public void spawn(String userName, String password, Map<String, String> envVars)
   {
      if (userName == null)
      {
         throw new NullPointerException("The username must be provided!");
      }
      
      // Get a temporary account from pool.
      TemporaryAccount tmpAcc = TemporaryAccount.open();
      
      if (tmpAcc == null)
      {
         String message =
            "Temporary Account Pool is empty! Please call your system administrator.";
         
         LOG.logp(Level.WARNING, "ClientSpawner.spawn()", "", LogHelper.generate(message));
         
         throw new RuntimeException(message);
      }
      
      // Get a list of brokers for this user
      Brokers brokers = BrokerManager.getBrokers(userName);
      
      cfg.setUserName(userName);
      cfg.setPassword(password);
      
      cb = getBuilder(cfg);
      
      // register listener
      SpawnerImpl spawner = (SpawnerImpl) SpawnerImpl.getInstance();
      spawner.addListener(cfg.getUuid(), this);
      
      boolean errLogged = false;
      
      // spawn
      Process shell = null;
      ready = false;
      try
      {
         latch = new CountDownLatch(1);
         
         // initial environment map contains user and password 
         Map<String, String> environmentMap = tmpAcc.getEnvironmentMap();
         
         // add P2J_HOME so that all P2J client sessions spawned below this one can
         // find their way home
         environmentMap.put("P2J_HOME", cfg.getLibPath());
         
         // merge appserver specific environment settings 
         if (envVars != null && !envVars.isEmpty())
         {
            environmentMap.putAll(envVars);
         }
         
         // if no brokers spawn local 
         if (brokers.getBrokers().isEmpty())
         {
            // spawn local
            shell = cb.localStart(environmentMap);
            
            // wait for notification
            latch.await(TIME_OUT, TimeUnit.SECONDS);
            
            shell.waitFor();
            exitCode = shell.exitValue();
         }
         else
         {
            // spawn remote
            BrokerSpawnResult result = cb.remoteStart(userName, brokers, environmentMap);
            
            // wait for notification
            latch.await(TIME_OUT, TimeUnit.SECONDS);
            
            exitCode = result.getExitCode();
            
            if (result.getException() != null)
            {
               errLogged = true;
               
               LOG.logp(Level.SEVERE, 
                        "ClientSpawner.spawn()", 
                        "",
                        LogHelper.generate("Cannot spawn remote process. Exit code: %d",
                                           exitCode),
                        result.getException());
            }
         }
      }
      
      catch (Exception e)
      {
         errLogged = true;
         
         LOG.logp(Level.SEVERE, 
                  "ClientSpawner.spawn()", 
                  "",
                  LogHelper.generate("Cannot spawn the process."),
                  e);
      }
      
      finally
      {
         spawner.removeListener(cfg.getUuid());
         tmpAcc.close();
         
         // cleanup the command
         ClientBuilder.removeCommand(cfg.getUuid());
         
         if (exitCode != 0 && !errLogged)
         {
            LOG.logp(Level.SEVERE, 
                     "ClientSpawner.spawn()", 
                     "",
                     LogHelper.generate("Failed spawn with exit code: %d", exitCode));
         }
      }

      if (!ready && shell != null)
      {
         shell.destroy();
      }
   }

   /**
    * Return spawner exit code
    * 
    * @return  Spawn exit value.
    */
   public int getExitCode()
   {
      return exitCode;
   }
   
   /**
    * Notify when the client has started.
    * 
    * @param    data
    *           Custom data sent by the P2J client back to the server.
    */
   @Override
   public void clientIsReady(Object data)
   {
      this.ready = true;
      latch.countDown();
   }
}