/*
** Module   : TemporaryAccountWorker.java
** Abstract : a Runnable object used to manage user accounts.
**
** Copyright (c) 2013-2017, Golden Code Development Corporation.
**
** -#- -I- --Date-- -------------------------------Description------------------------------------
** 001 MAG 20131216 Created initial version.
** 002 CA  20140214 Fixed NPE at thread shutdown when the pool is never used. 
** 003 MAG 20140217 Fixed thread termination.
** 004 CA  20150909 Fixed a race condition resulting in a NPE for callerLatch.
*/
/*
** This program is free software: you can redistribute it and/or modify
** it under the terms of the GNU Affero General Public License as
** published by the Free Software Foundation, either version 3 of the
** License, or (at your option) any later version.
**
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
** GNU Affero General Public License for more details.
**
** You may find a copy of the GNU Affero GPL version 3 at the following
** location: https://www.gnu.org/licenses/agpl-3.0.en.html
** 
** Additional terms under GNU Affero GPL version 3 section 7:
** 
**   Under Section 7 of the GNU Affero GPL version 3, the following additional
**   terms apply to the works covered under the License.  These additional terms
**   are non-permissive additional terms allowed under Section 7 of the GNU
**   Affero GPL version 3 and may not be removed by you.
** 
**   0. Attribution Requirement.
** 
**     You must preserve all legal notices or author attributions in the covered
**     work or Appropriate Legal Notices displayed by works containing the covered
**     work.  You may not remove from the covered work any author or developer
**     credit already included within the covered work.
** 
**   1. No License To Use Trademarks.
** 
**     This license does not grant any license or rights to use the trademarks
**     Golden Code, FWD, any Golden Code or FWD logo, or any other trademarks
**     of Golden Code Development Corporation. You are not authorized to use the
**     name Golden Code, FWD, or the names of any author or contributor, for
**     publicity purposes without written authorization.
** 
**   2. No Misrepresentation of Affiliation.
** 
**     You may not represent yourself as Golden Code Development Corporation or FWD.
** 
**     You may not represent yourself for publicity purposes as associated with
**     Golden Code Development Corporation, FWD, or any author or contributor to
**     the covered work, without written authorization.
** 
**   3. No Misrepresentation of Source or Origin.
** 
**     You may not represent the covered work as solely your work.  All modified
**     versions of the covered work must be marked in a reasonable way to make it
**     clear that the modified work is not originating from Golden Code Development
**     Corporation or FWD.  All modified versions must contain the notices of
**     attribution required in this license.
*/

package com.goldencode.p2j.main;

import java.util.concurrent.*;
import java.util.logging.*;

import com.goldencode.p2j.util.*;

/**
 * A Runnable object used to manage user accounts.
 * In order to manage user account admin rights are mandatory.
 * This object is used as Runnable target in an AssociatedThread instance
 * which run in standard server security context and inherit server admin rights.
 * This object execute the run method of another Runnable object the task to be executed.
 * The execute method is synchronized between this thread and caller.
 * The caller waits for end of execution of task that he provide as parameter.
 *  
 * @author mag
 *
 */
public class TemporaryAccountWorker 
implements Runnable
{
   /** Logger. */
   private static final Logger LOG = LogHelper.getLogger(TemporaryAccountWorker.class.getName());
   
   /**  Termination flag */
   private volatile boolean terminated = false;
   
   /** Thread synchronization */
   private volatile CountDownLatch workerLatch;

   /** Caller synchronization */
   private volatile CountDownLatch callerLatch;
   
   /** Task to be executed */
   private volatile Runnable task;
   
   /**
    * Constructor. 
    * Initialize thread synchronization latch.
    */
   public TemporaryAccountWorker()
   {
      workerLatch = new CountDownLatch(1);      
   }

   /**
    * Execute specified task.
    * 
    * @param   task
    *          Task to be executed.
    */
   public void execute(Runnable task)
   {
      try
      {
         if (!terminated)
         {
            // Set task instance.
            this.task = task;
            // Arm caller latch.
            callerLatch = new CountDownLatch(1);
            // Execute task.
            workerLatch.countDown();
            // Wait for task execution to end.
            callerLatch.await();
         }
      }
      catch (InterruptedException e)
      {
         LOG.logp(Level.WARNING,
                  "TemporaryAccountWorker.execute()",
                  "",
                  LogHelper.generate("InterruptedException!"),
                  e);         
      }
   }
   
   /**
    * Task executor.
    */
   public void run()
   {
      while (!terminated)
      {
         try
         {
            // Waiting for a call
            workerLatch.await();
            // Execute task
            if (!terminated && task != null)
            {
               task.run();
               task = null;
            }
         }
         catch (InterruptedException e)
         {
            LOG.logp(Level.WARNING,
                     "TemporaryAccountWorker.run()",
                     "",
                     LogHelper.generate("InterruptedException!"),
                     e);
         }
         finally
         {
            if (!terminated && workerLatch.getCount() == 0)
            {
               workerLatch = new CountDownLatch(1);
            }
            if (callerLatch != null && callerLatch.getCount() == 1)
            {
               // Notify caller
               callerLatch.countDown();
            }
         }
      }
      
      LOG.logp(Level.INFO,
               "TemporaryAccountWorker.run()",
               "",
               LogHelper.generate("Thread exit."));
   }

   /**
    * Terminate thread.
    */
   public void terminate()
   {
      this.terminated = true;
      if (workerLatch.getCount() == 1)
      {
         workerLatch.countDown();
      }
   }
}