/*
** Module   : WebClientBuilderOptions.java
** Abstract : store web client parameters read from the directory
**
** Copyright (c) 2014-2017, Golden Code Development Corporation.
**
** -#- -I- --Date-- ---------------------------------Description----------------------------------
** 001 CA  20140206 First version, extracted from ClientBuilderOptions.
** 002 MAG 20140711 Enable/Disable clipboard.
** 003 GES 20150310 Made this class generic for web clients, not just for ChUI web clients.
** 004 GES 20150717 Added some GUI configuration values.
** 005 GES 20160202 Added trusted mode support and removed some dead code.
**     SBI 20160321 Added maxBinaryMessage, maxTextMessage, maxIdleTime websocket's parameters and
**                  utilized "webClient/embedded" to set the web client embedded mode.
** 006 SBI 20160427 Added graphicsCached parameter to turn off the graphics cache for testing.
** 007 SBI 20170628 Fixed misspelled variable to set watchdogTimeout correctly.
*/
/*
** This program is free software: you can redistribute it and/or modify
** it under the terms of the GNU Affero General Public License as
** published by the Free Software Foundation, either version 3 of the
** License, or (at your option) any later version.
**
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
** GNU Affero General Public License for more details.
**
** You may find a copy of the GNU Affero GPL version 3 at the following
** location: https://www.gnu.org/licenses/agpl-3.0.en.html
** 
** Additional terms under GNU Affero GPL version 3 section 7:
** 
**   Under Section 7 of the GNU Affero GPL version 3, the following additional
**   terms apply to the works covered under the License.  These additional terms
**   are non-permissive additional terms allowed under Section 7 of the GNU
**   Affero GPL version 3 and may not be removed by you.
** 
**   0. Attribution Requirement.
** 
**     You must preserve all legal notices or author attributions in the covered
**     work or Appropriate Legal Notices displayed by works containing the covered
**     work.  You may not remove from the covered work any author or developer
**     credit already included within the covered work.
** 
**   1. No License To Use Trademarks.
** 
**     This license does not grant any license or rights to use the trademarks
**     Golden Code, FWD, any Golden Code or FWD logo, or any other trademarks
**     of Golden Code Development Corporation. You are not authorized to use the
**     name Golden Code, FWD, or the names of any author or contributor, for
**     publicity purposes without written authorization.
** 
**   2. No Misrepresentation of Affiliation.
** 
**     You may not represent yourself as Golden Code Development Corporation or FWD.
** 
**     You may not represent yourself for publicity purposes as associated with
**     Golden Code Development Corporation, FWD, or any author or contributor to
**     the covered work, without written authorization.
** 
**   3. No Misrepresentation of Source or Origin.
** 
**     You may not represent the covered work as solely your work.  All modified
**     versions of the covered work must be marked in a reasonable way to make it
**     clear that the modified work is not originating from Golden Code Development
**     Corporation or FWD.  All modified versions must contain the notices of
**     attribution required in this license.
*/

package com.goldencode.p2j.main;

import java.util.*;

import com.goldencode.p2j.cfg.*;
import com.goldencode.p2j.net.*;
import com.goldencode.p2j.ui.client.chui.driver.*;

// must be explicit to avoid abmiguity with Java
import com.goldencode.p2j.security.SecurityManager;

/**
 * The parameters stored inside this structure are read from directory on server start-up 
 * and are used as default parameters to create and spawn a web client process.
 * <p>
 * This class MUST NOT BE USED from client side, as it uses a direct access to the server.  Also,
 * this class can be instantiated only from threads with a P2J context for the running P2J server.
 */
public class WebClientBuilderOptions
extends ClientBuilderOptions
{   
   /** Directory node name in which our configuration can be found. */
   public static final String DIRECTORY_NODE_ID = "webClient";
   
   /** The default maximal binary message size excepted by the configured websocket */
   private static final int MAX_BINARY_MESSAGE_SIZE = 32768;
   
   /** The default maximal text message size excepted by the configured websocket */
   private static final int MAX_TEXT_MESSAGE_SIZE = 4096;
   
   /** The default maximal idle time of the configured websocket */
   private static final int MAX_IDLE_TIMEOUT = 90000;

   /** Options storage. */
   private static final Map<String, String> options = new HashMap<>();
   
   /** Copy of state used for initializing the superclass without reading the directory. */
   private static ClientBuilderOptions cbo = null;
   
   /** Enables trusted launch mode (no password is used and the OS account is forced). */
   private boolean trusted = false;

   /**
    * Create a new instance.
    *
    * @param    trusted
    *           If this instance will be used to launch clients in trusted mode.
    * 
    * @throws   IllegalStateException
    *           If this is called from P2J clients.
    */
   public WebClientBuilderOptions(boolean trusted)
   {
      super(cbo, null);
      this.trusted = trusted;
   }
   
   /**
    * Read the configuration from the directory and calculate certain runtime values that will
    * be used by web client launching. This should be executed from a thread that has the
    * server account context.
    *
    * @param    sm
    *           The initialized instance of <code>SecurityManager</code>.
    * @param    cfg
    *           The bootstrap configuration information for this server.
    */
   public static synchronized void initialize(SecurityManager sm, BootstrapConfig cfg)
   {
      // TODO: this will not allow changing the options via the AdminConsole - add a way to 
      // invalidate and recreate this instance
      
      if (SessionManager.get() == null || SessionManager.get().isLeaf())
      {
         throw new IllegalStateException("This class can be used only on the server-side!");
      }
      
      // save off our parent class state now, instead of deferring until runtime
      cbo = new ClientBuilderOptions(null)
      {
         /**
          * This is not used, but must be present to create an instance of the abstract class.
          */
         protected boolean isPasswordAuthentication()
         {
            return true;
         }
      };

      // read the configuration from the directory
      readOptions(sm, cfg);
   }

   /**
    * Get web client driver options. This contains configuration for both GUI and ChUI web
    * clients.  It is up to the caller to use only those portions that make sense for the
    * client type in use.
    * 
    * @return   The web client options.
    */
   public Map<String, String> getClientOptions()
   {
      return new HashMap<>(options);
   }

   /**
    * Get the authentication mode for this spawner.
    * 
    * @return   <code>true</code> if not in trusted mode.
    */
   protected boolean isPasswordAuthentication()
   {
      return !trusted;
   }

   /**
    * Read all web client driver options from the directory and store the defaults and any
    * overridden configuration in a map for configuration of both GUI and ChUI web drivers.
    * It is up to the downstream user of the map to use only those portions that make sense
    * for the client type in use.
    *
    * @param    sm
    *           The initialized instance of <code>SecurityManager</code>.
    * @param    cfg
    *           The bootstrap configuration information for this server.
    */
   private static void readOptions(SecurityManager sm, BootstrapConfig cfg)
   {
      // default ChUI terminal size
      int     rows       = TerminalOptions.DEF_ROWS;
      int     columns    = TerminalOptions.DEF_COLUMNS;
      
      // default GUI desktop size in pixels
      int     height     = 768;
      int     width      = 1024;
      
      // for colors, don't include any alpha on bits 24-31, if emitted as FF the result is not
      // interpreted properly by the browser
      String  background = TerminalOptions.DEF_BACKGROUND.asHexString();
      String  foreground = TerminalOptions.DEF_FOREGROUND.asHexString();
      String  selection  = TerminalOptions.DEF_SELECTION.asHexString();
      
      String  fontname   = TerminalOptions.DEF_FONTNAME;
      int     fontsize   = TerminalOptions.DEF_FONTSIZE;
      long    stimeout   = -1;
      long    wtimeout   = -1;
      int     port       = 0;
      String  host       = null;
      boolean clipboard  = true;
      boolean taskbar    = true;

      // get registered terminal options
      rows       = cbo.getNode(DIRECTORY_NODE_ID, "rows", rows);
      columns    = cbo.getNode(DIRECTORY_NODE_ID, "columns", columns);
      height     = cbo.getNode(DIRECTORY_NODE_ID, "desktopHeight", height);
      width      = cbo.getNode(DIRECTORY_NODE_ID, "desktopWidth", width);
      background = cbo.getNode(DIRECTORY_NODE_ID, "background", background);
      foreground = cbo.getNode(DIRECTORY_NODE_ID, "foreground", foreground);
      selection  = cbo.getNode(DIRECTORY_NODE_ID, "selection", selection);
      fontname   = quotedString(cbo.getNode(DIRECTORY_NODE_ID, "fontname", fontname));
      fontsize   = cbo.getNode(DIRECTORY_NODE_ID, "fontsize", fontsize);
      stimeout   = cbo.getNode(DIRECTORY_NODE_ID, "webSocketTimeout", (int) stimeout);
      wtimeout   = cbo.getNode(DIRECTORY_NODE_ID, "watchdogTimeout",  (int) wtimeout);
      port       = cbo.getNode(DIRECTORY_NODE_ID, "port", port);
      host       = cbo.getNode(DIRECTORY_NODE_ID, "host", host);
      clipboard  = cbo.getNode(DIRECTORY_NODE_ID, "clipboard", true);
      taskbar    = cbo.getNode(DIRECTORY_NODE_ID, "taskbar", true);
   
      // common websocket parameters
      int maxBinaryMessage = cbo.getNode(DIRECTORY_NODE_ID,
                                         "maxBinaryMessage",
                                         MAX_BINARY_MESSAGE_SIZE);
      int maxTextMessage   = cbo.getNode(DIRECTORY_NODE_ID,
                                         "maxTextMessage",
                                         MAX_TEXT_MESSAGE_SIZE);
      int maxIdleTime      = cbo.getNode(DIRECTORY_NODE_ID,
                                         "maxIdleTime",
                                         MAX_IDLE_TIMEOUT);
      // read the web client embedded mode
      boolean embedded     = cbo.getNode(DIRECTORY_NODE_ID,
                                         "embedded",
                                         false);
      // Read if the drawing batches are cached 
      boolean graphicsCached = cbo.getNode(DIRECTORY_NODE_ID, "graphicsCached", true);
      
      // build client options map
      options.put("rows", String.valueOf(rows));
      options.put("columns", String.valueOf(columns));
      options.put("desktopHeight", String.valueOf(height));
      options.put("desktopWidth", String.valueOf(width));
      options.put("background", background);
      options.put("foreground", foreground);
      options.put("selection", selection);
      options.put("fontname", fontname);
      options.put("fontsize", String.valueOf(fontsize));
      options.put("webSocketTimeout", String.valueOf(stimeout));
      options.put("watchdogTimeout", String.valueOf(wtimeout));
      options.put("port", String.valueOf(port));
      options.put("host", host);
      options.put("clipboard", clipboard ? "true" : "false");
      options.put("taskbar", taskbar ? "true" : "false");
      options.put("maxBinaryMessage", String.valueOf(maxBinaryMessage));
      options.put("maxTextMessage", String.valueOf(maxTextMessage));
      options.put("maxIdleTime", String.valueOf(maxIdleTime));

      // sets true or false if the web client embedded mode is on or off respectively
      options.put("embedded", embedded ? "true" : "false");
      // sets true if the web client uses a buffer to cache drawings, otherwise false.
      options.put("graphicsCached", graphicsCached ? "true" : "false");
      
      // in trusted mode, we must authenticate by using the NativeSecureConnection
      // to access the P2J server and get our temporary account; this requires some
      // additional secure connection configuration to be available
      options.put("securePort", String.valueOf(cfg.getInt("net", "server", "secure_port", -1)));
      options.put("serverAlias", sm.getServerAlias());
   }
}