/*
** Module   : WebClientSpawner.java
** Abstract : customizations for spawning a web client
**
** Copyright (c) 2013-2017, Golden Code Development Corporation.
**
** -#- -I- --Date-- --------------------------------Description-----------------------------------
** 001 MAG 20131119 First version.
** 002 MAG 20140206 Send URL as referrer address via configuration file.
** 003 CA  20140206 Changes related to refactoring the spawner infrastructure (allow P2J process 
**                  clients to be started, beside web chui clients).
** 004 GES 20150312 Added support for the GUI web client.
** 005 CA  20150820 Wait for the remote web side to notify that it has initialized all P2J modules
**                  before starting sending requests.
** 006 SBI 20151119 Added the trust server certificate checking.
** 007 GES 20160201 Moved auth token processing here to be better encapsulated/hidden. Modified
**                  the constructor to handle trusted mode (and to pass the referrer directly
**                  instead of requiring a setter to be called later).
**     SBI 20160301 Fixed the web server to redirect the authorized users to the dedicated web
**                  application handler.
** 008 CA  20160428 Added command-line configuration options for web-clients.
** 009 SBI 20160628 Added new methods to allocate and to release web clients.
** 010 SBI 20171023 Changed to allocate new system resources in order to run new web client on
**                  the target host.
*/
/*
** This program is free software: you can redistribute it and/or modify
** it under the terms of the GNU Affero General Public License as
** published by the Free Software Foundation, either version 3 of the
** License, or (at your option) any later version.
**
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
** GNU Affero General Public License for more details.
**
** You may find a copy of the GNU Affero GPL version 3 at the following
** location: https://www.gnu.org/licenses/agpl-3.0.en.html
** 
** Additional terms under GNU Affero GPL version 3 section 7:
** 
**   Under Section 7 of the GNU Affero GPL version 3, the following additional
**   terms apply to the works covered under the License.  These additional terms
**   are non-permissive additional terms allowed under Section 7 of the GNU
**   Affero GPL version 3 and may not be removed by you.
** 
**   0. Attribution Requirement.
** 
**     You must preserve all legal notices or author attributions in the covered
**     work or Appropriate Legal Notices displayed by works containing the covered
**     work.  You may not remove from the covered work any author or developer
**     credit already included within the covered work.
** 
**   1. No License To Use Trademarks.
** 
**     This license does not grant any license or rights to use the trademarks
**     Golden Code, FWD, any Golden Code or FWD logo, or any other trademarks
**     of Golden Code Development Corporation. You are not authorized to use the
**     name Golden Code, FWD, or the names of any author or contributor, for
**     publicity purposes without written authorization.
** 
**   2. No Misrepresentation of Affiliation.
** 
**     You may not represent yourself as Golden Code Development Corporation or FWD.
** 
**     You may not represent yourself for publicity purposes as associated with
**     Golden Code Development Corporation, FWD, or any author or contributor to
**     the covered work, without written authorization.
** 
**   3. No Misrepresentation of Source or Origin.
** 
**     You may not represent the covered work as solely your work.  All modified
**     versions of the covered work must be marked in a reasonable way to make it
**     clear that the modified work is not originating from Golden Code Development
**     Corporation or FWD.  All modified versions must contain the notices of
**     attribution required in this license.
*/

package com.goldencode.p2j.main;

import java.net.URI;
import java.net.URISyntaxException;

import com.goldencode.p2j.cfg.*;
import com.goldencode.p2j.directory.*;
import com.goldencode.p2j.main.WebClientConfig;
import com.goldencode.p2j.main.Spawner.OutOfResources;
import com.goldencode.p2j.ui.client.driver.web.*;
import com.goldencode.p2j.ui.client.chui.driver.web.*;
import com.goldencode.p2j.ui.client.gui.driver.web.*;
import com.goldencode.p2j.ui.client.driver.*;

/**
 * Class used to spawn a new web client process.
 * <p>
 * This class is tricky to understand because the main part of it runs on the server while the
 * <code>TemporaryClientTask</code> inner class is serialized and sent down to execute on the
 * client side.  That client side code then calls methods which are implemented as a remote
 * object with the {@link Spawner} interface, which the server portions of this class implement. 
 * <p>
 * In the interactive login mode, the spawned process first tries to authenticate the userid and
 * password at the OS level. On success a new process (client) is spawned with those credentials.
 * <p>
 * In trusted mode, the spawner process will connect to the server using a secure socket and
 * temporary credentials to establish a secure session.  If that works, it will lookup the
 * credentials to use for spawning the client based on a unique UUID for this spawning session.
 * Using those credentials it will start the client and notify the server about success or
 * failure.
 * <p>
 * In both modes, the spawned process will call a remote object method to get the server's
 * KeyStore which is used to start the client's embedded web server.  Finally, the client will
 * call a remote object method to notify this (server side) instance about the URI on which the
 * client's embedded web server is listening. If a notification does not occur within a specific
 * amount of time, a failure in spawning is asserted.  
 */
public class WebClientSpawner
extends ClientSpawner<WebClientConfig>
{
   /** Authorization token query parameter name */
   public static final String PARAM_TOKEN = "token";
   
   /** Remote server URI */
   private volatile String remoteUri = null;
   
   /** Flag that denotes client type (<code>true</code> for GUI, <code>false</code> for ChUI). */
   private boolean gui;

   /** The web clients manager */
   private final WebClientsManager webClientsManager = WebClientsManager.getInstance();
   
   /** The request parameters */
   private final String[] requestParameters;
   
   /**
    * Create a new web client spawner.
    *
    * @param    requestParameters
    *           Holds additional parameters if they are provided. It is expected that the first
    *           requestParameters[0] holds the forwarded host of "X-Forwarded-Host" HTTP header,
    *           requestParameters[1] holds the corresponding forwarded protocol that is provided
    *           by "X-Forwarded-Proto" value and requestParameters[2] holds the client IP address.
    * @param    gui
    *           Flag that denotes client type (<code>true</code> for GUI, <code>false</code> for
    *           ChUI).
    * @param    trusted
    *           If this instance will be used to launch clients in trusted mode.
    * @param    referrer
    *           The referrer string to provide the web client. If the client gets disconnected
    *           it will be redirected back to this URL. 
    * @param    options
    *           The command-line configuration options.  May be <code>null</code>.
    */
   public WebClientSpawner(String[] requestParameters,
                           boolean gui,
                           boolean trusted,
                           String referrer,
                           String[] options)
   {
      super(new WebClientBuilderParameters(gui, trusted, referrer, options));
      this.gui = gui;
      this.requestParameters = requestParameters;
   }

   /**
    * Notify when the client has started.
    * 
    * @param    data
    *           Custom data sent by the P2J client back to the server.  For web clients, this is
    *           the remote URI to which it will redirect the browser.
    */
   @Override
   public void clientIsReady(Object data)
   {
      remoteUri = (String) data;
      super.clientIsReady(data);
   }

   /**
    * Get the server key store.
    * 
    * @return   See above.
    */
   @Override
   public Object getServerData()
   {
      return ((WebClientBuilder) cb).getServerKeyStore();
   }
   
   /**
    * Get a {@link WebClientBuilder} used to spawn P2J clients.
    * 
    * @param    cfg
    *           The P2J web client configuration.
    *           
    * @return   See above.
    */
   protected ClientBuilder getBuilder(ClientBuilderParameters cfg)
   {
      return new WebClientBuilder(cfg);
   }

   /**
    * Get the remote server URI to which the browser will be redirected.
    * 
    * @return   The remote server URI or <code>null</code>.
    * @throws URISyntaxException 
    */
   public String getRemoteUri() throws URISyntaxException
   {
      String uri = remoteUri;
      
      if (uri != null)
      {
         // set authorization token for the redirect
         Object store = getServerData();
         
         if (store instanceof ServerKeyStore)
         {
            String token = ((ServerKeyStore) store).getAuthorizationToken();
            
            if (token != null)
            {
               uri = uri + "index.html?" + PARAM_TOKEN + "=" + token;
            }
         }
      }
      
      return uri;
   }
   
   /**
    * Get a {@link TemporaryClient} worker which will do the work after authenticating on the P2J
    * server using the temporary credentials.
    * 
    * @return   See above.
    */
   @Override
   public TemporaryClient getTemporaryClient()
   {
      return new TemporaryClientTask(gui, requestParameters);
   }

   /**
    * On P2J client side, it will do custom work after authenticating using the temporary 
    * credentials.
    * <p>
    * This will create a {@link ChuiWebDriver} or {@link GuiWebDriver} and start the 
    * embedded web server which will serve this client's requests.
    */
   private static class TemporaryClientTask
   implements TemporaryClient
   {
      /** Flag that denotes client type. */
      private boolean gui;
      
      private String[] proxyServerParameters;
      
      /**
       * Create a new web client spawner.
       *
       * @param    gui
       *           Flag that denotes client type (<code>true</code> for GUI, <code>false</code>
       *           for ChUI).
       * @param    requestParameters
       *           Holds additional parameters if they are provided. It is expected that the first
       *           requestParameters[0] holds the forwarded host of "X-Forwarded-Host" HTTP header
       *           and requestParameters[1] holds the corresponding forwarded protocol that is 
       *           provided by "X-Forwarded-Proto" value, and requestParameters[2] holds
       *           the client IP address.
       */
      public TemporaryClientTask(boolean gui, String[] requestParameters)
      {
         this.gui = gui;
         this.proxyServerParameters = requestParameters;
      }
      
      /**
       * On P2J client side, it will do custom work after authenticating using the temporary 
       * credentials.
       * <p>
       * This will create a {@link ChuiWebDriver} or {@link GuiWebDriver} and start the 
       * embedded web server which will serve this client's requests.
       * 
       * @param    spawner
       *           The spawner remote access interface. 
       * @param    uuid
       *           Remote client identifier uuid.
       * @param    config
       *           The configuration to use for setup of the client and for the server connection.
       *
       * @return   An explicit web screen driver to be used by the P2J client.
       * 
       * @throws   RuntimeException
       *           If the embedded web server could not be started.
       */
      @Override
      public ScreenDriver doWork(Spawner spawner, String uuid, BootstrapConfig config)
      {
         // import the server key store
         ServerKeyStore keyStore = (ServerKeyStore) spawner.exportData(uuid);
         
         if (keyStore == null)
         {
            throw new RuntimeException("Cannot import server key store.");
         }
         
         String bytes = Base64.byteArrayToBase64(keyStore.getKeyStore());
         String pw    = keyStore.getKeyStorePassword();
         
         config.setConfigItem("security", "truststore", "alias", keyStore.getTrustServerAlias());
         config.setConfigItem("security", "truststore", "bytes", bytes);
         config.setConfigItem("access", "password", "truststore", pw);
         config.setConfigItem("security", "certificate", "validate", "true");
         config.setConfigItem("security", "trust_mgr", "disable", "false");
         config.setConfigItem("security", "transport", "refresh", "true");

         // get an explicit port and host set by the client
         int    port = config.getInt("client", "web", "port", 0);
         String host = config.getString("client", "web", "host", null);
         
         boolean dedicatedMode = config.getBoolean("client", "web", "dedicatedMode", false);
         
         WebClientConfig webClientConfig = null;
         
         Object clientResources;
         try
         {
            clientResources = spawner.allocateClient(uuid, host);
         }
         catch (OutOfResources e)
         {
            throw new RuntimeException("All client sessions are allocated.", e); 
         }
         
         if (clientResources instanceof WebClientConfig)
         {
            webClientConfig = (WebClientConfig) clientResources;
            
            //override host and port if new values are provided
            if (webClientConfig.getPort() > 0)
            {
               port = webClientConfig.getPort();
               config.setConfigItem("client", "web", "port", String.valueOf(port));
            }
            if (webClientConfig.getHost() != null    &&
                !webClientConfig.getHost().isEmpty() && 
                !"localhost".equals(webClientConfig.getHost()))
            {
               host = webClientConfig.getHost();
               config.setConfigItem("client", "web", "host", host);
            }
            // set web root
            // getWebRoot() returns "/" or non empty path like /server/client1
            if (!"/".equals(webClientConfig.getWebRoot()))
            {
               config.setConfigItem("client", "web", "webRoot", webClientConfig.getWebRoot());
            }
         }
         
         // create web screen driver
         ScreenDriver<?> driver = gui ? new GuiWebDriver(config) : new ChuiWebDriver(config);
         driver.init();
         
         if (dedicatedMode)
         {
            host = "localhost";
            config.setConfigItem("client", "web", "host", host);
         }
         
         String uri = ((EmbeddedWebServer) driver).startupServer(keyStore, config, host, port);
         
         if (proxyServerParameters != null       &&
             proxyServerParameters.length > 0    &&
             proxyServerParameters[0] != null    &&
             !proxyServerParameters[0].isEmpty() &&
             webClientConfig != null             &&
             uri != null)
         {
            uri = rewriteUri(webClientConfig);
         }

         // client initialization is done, report embedded server URI back to the server
         spawner.clientIsReady(uuid, uri);
         
         ((EmbeddedWebServer) driver).waitInitialization();
         
         if (uri == null)
         {
            spawner.releaseClient(uuid);
            throw new RuntimeException("Cannot start embedded server."); 
         }
         
         return driver;
      }
      
      /**
       * Rewrite uri in the case if a web client is accessed via the reverse proxy server.
       * 
       * @param    webClient
       *           The web client system resources
       * 
       * @return   The external uri of this web client
       */
      private String rewriteUri(WebClientConfig webClient)
      {
         String forwardedProto = webClient.getForwardedProto();
         String forwardedHost = webClient.getForwardedHost();
         String forwardedPath = webClient.getWebRoot();
         URI uri = URI.create(forwardedProto + "://" + forwardedHost +  forwardedPath + "/");
         //rewrite uri
         return uri.toString();
      }
   }

   /**
    * Allocates the system resources for the spawned client if it is necessarily.
    * 
    * @param    host
    *           The remote host for the spawned client
    * 
    * @return   The system resources for the spawned client.
    * 
    * @throws   OutOfResources
    *           If the required resources can't be allocated.
    */
   @Override
   public WebClientConfig allocateClient(String host) throws OutOfResources
   {
      if (webClientsManager != null)
      {
         WebClientConfig clientConfig = webClientsManager.allocateClient(requestParameters, host);
         
         if (clientConfig == null)
         {
            throw new OutOfResources();
         }
         webClientsManager.registerWebClientUuid(cfg.getUuid(), clientConfig);
         
         return clientConfig;
      }
      
      return null;
   }

   /**
    * Releases the system resources for this spawned client.
    */
   @Override
   public void releaseClient()
   {
      if (webClientsManager != null)
      {
         webClientsManager.freeWebClientResources(cfg.getUuid());
      }
   }
}