/*
** Module   : NetRights.java
** Abstract : Implements the Rights interface for the "net" abstract resource.
**
** Copyright (c) 2005-2017, Golden Code Development Corporation.
**
** -#- -I- --Date-- --JPRM-- ----------------Description----------------------
** 001 NVS 20090630   @43004 Made it a separate class due to a new requirement
**                           for all rights classes to be separate and
**                           serializable.
** 002 NVS 20090703   @43042 Implemented toDirectoryNode() method, which was
**                           added to the Rights interface.
** 003 NVS 20090713   @43164 Added getPermissions() method.
** 004 ECF 20150715          Replace StringBuffer with StringBuilder.
** 005 SBI 20170224          Changed to be a simple DTO, moved business methods out.
*/
/*
** This program is free software: you can redistribute it and/or modify
** it under the terms of the GNU Affero General Public License as
** published by the Free Software Foundation, either version 3 of the
** License, or (at your option) any later version.
**
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
** GNU Affero General Public License for more details.
**
** You may find a copy of the GNU Affero GPL version 3 at the following
** location: https://www.gnu.org/licenses/agpl-3.0.en.html
** 
** Additional terms under GNU Affero GPL version 3 section 7:
** 
**   Under Section 7 of the GNU Affero GPL version 3, the following additional
**   terms apply to the works covered under the License.  These additional terms
**   are non-permissive additional terms allowed under Section 7 of the GNU
**   Affero GPL version 3 and may not be removed by you.
** 
**   0. Attribution Requirement.
** 
**     You must preserve all legal notices or author attributions in the covered
**     work or Appropriate Legal Notices displayed by works containing the covered
**     work.  You may not remove from the covered work any author or developer
**     credit already included within the covered work.
** 
**   1. No License To Use Trademarks.
** 
**     This license does not grant any license or rights to use the trademarks
**     Golden Code, FWD, any Golden Code or FWD logo, or any other trademarks
**     of Golden Code Development Corporation. You are not authorized to use the
**     name Golden Code, FWD, or the names of any author or contributor, for
**     publicity purposes without written authorization.
** 
**   2. No Misrepresentation of Affiliation.
** 
**     You may not represent yourself as Golden Code Development Corporation or FWD.
** 
**     You may not represent yourself for publicity purposes as associated with
**     Golden Code Development Corporation, FWD, or any author or contributor to
**     the covered work, without written authorization.
** 
**   3. No Misrepresentation of Source or Origin.
** 
**     You may not represent the covered work as solely your work.  All modified
**     versions of the covered work must be marked in a reasonable way to make it
**     clear that the modified work is not originating from Golden Code Development
**     Corporation or FWD.  All modified versions must contain the notices of
**     attribution required in this license.
*/

package com.goldencode.p2j.net;

import com.goldencode.p2j.security.*;

import java.io.Serializable;
import java.util.LinkedHashMap;
import java.util.Map;


/**
 * Implements the "net" rights objects.  
 * <p>
 * Net resources come with one field rights objects made of a bitfield
 * of 4 permissions RWXN.
 */
public class NetRights
implements Rights,
           Serializable
{
   /** READ permissions bit for net resource */
   public static final int BIT_READ_ACCESS = 0;

   /** WRITE permissions bit for net resource */
   public static final int BIT_WRITE_ACCESS = 1;

   /** EXECUTE permissions bit for net resource */
   public static final int BIT_EXECUTE_ACCESS = 2;

   /** DENIED permissions bit for net resource */
   public static final int BIT_DENIED_ACCESS = 3;

   /* the only field of this class - permissions */
   BitSet permissions = null;

   /**
    * Default constructor. 
    */
   public NetRights()
   {
      permissions = new BitSet(BIT_DENIED_ACCESS + 1);
      permissions.unset(BIT_READ_ACCESS);
      permissions.unset(BIT_DENIED_ACCESS);
   }

   /**
    * Constructor. 
    * Just copies the argument into its member.
    *
    * @param permissions
    *        bitfield of permissions RWXN
    */
   public NetRights(BitSet permissions)
   {
      this.permissions = permissions;
   }

   /**
    * Gets the indexed bit of permissions. 
    *
    * @param  index
    *         index of the permissions bit to query
    * @return the bit's value
    */
   boolean get(int index)
   {
      return permissions.isSet(index);
   }

   /**
    * Gets the entire set of permissions. 
    *
    * @return all permissions as a <code>BitField</code>
    */
   public BitSet getPermissions()
   {
      return permissions;
   }
   
   /**
    * Converts this object to string. 
    *
    * @return string representation of this object
    */
   public String toString()
   {
      StringBuilder sb = new StringBuilder();

      sb.append("{");
      if (permissions.isSet(BIT_READ_ACCESS))
         sb.append("R");
      else
         sb.append(".");

      if (permissions.isSet(BIT_WRITE_ACCESS))
         sb.append("W");
      else
         sb.append(".");

      if (permissions.isSet(BIT_EXECUTE_ACCESS))
         sb.append("X");
      else
         sb.append(".");

      if (permissions.isSet(BIT_DENIED_ACCESS))
         sb.append("N");
      else
         sb.append(".");
      sb.append("}");

      return new String(sb);
   }
   
   /**
    * Returns rights name, a directory known class name.
    * 
    * @return   a directory known class name
    */
   @Override
   public String getRightsName()
   {
      return "netRights";
   }

   /**
    * Returns its attributes map.
    * 
    * @return   A map of an attribute name to its value
    */
   @Override
   public Map<String, Object[]> getAttributes()
   {
      Map<String, Object[]> map = new LinkedHashMap<>(1);
      
      map.put("permissions", new Object[] {permissions});
      
      return map;
   }
}
