/*
** Module   : NetSocket.java
** Abstract : combines a socket and related input/output streams into a
**            manageable unit
**
** Copyright (c) 2005-2017, Golden Code Development Corporation.
**
** -#- -I- --Date-- --JPRM-- ---------------------------- Description ----------------------------
** 001 NVS 20050308   @20245 Created. This class extends and replaces
**                           SSLSocket in net and security packages.
** 002 NVS 20050311   @20287 Simplified by removing DataInput and Output
**                           streams. Authentication now uses ObjectInput
**                           and Output streams as well.
** 003 GES 20081103   @40327 Rework to remove unnecessary features and
**                           to allow non-secure sockets to be used.
** 004 GES 20090609   @42628 Provided access to local socket info.
** 005 OM  20131022          Added a list of SessionListeners and minimum management methods.
*/ 
/*
** This program is free software: you can redistribute it and/or modify
** it under the terms of the GNU Affero General Public License as
** published by the Free Software Foundation, either version 3 of the
** License, or (at your option) any later version.
**
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
** GNU Affero General Public License for more details.
**
** You may find a copy of the GNU Affero GPL version 3 at the following
** location: https://www.gnu.org/licenses/agpl-3.0.en.html
** 
** Additional terms under GNU Affero GPL version 3 section 7:
** 
**   Under Section 7 of the GNU Affero GPL version 3, the following additional
**   terms apply to the works covered under the License.  These additional terms
**   are non-permissive additional terms allowed under Section 7 of the GNU
**   Affero GPL version 3 and may not be removed by you.
** 
**   0. Attribution Requirement.
** 
**     You must preserve all legal notices or author attributions in the covered
**     work or Appropriate Legal Notices displayed by works containing the covered
**     work.  You may not remove from the covered work any author or developer
**     credit already included within the covered work.
** 
**   1. No License To Use Trademarks.
** 
**     This license does not grant any license or rights to use the trademarks
**     Golden Code, FWD, any Golden Code or FWD logo, or any other trademarks
**     of Golden Code Development Corporation. You are not authorized to use the
**     name Golden Code, FWD, or the names of any author or contributor, for
**     publicity purposes without written authorization.
** 
**   2. No Misrepresentation of Affiliation.
** 
**     You may not represent yourself as Golden Code Development Corporation or FWD.
** 
**     You may not represent yourself for publicity purposes as associated with
**     Golden Code Development Corporation, FWD, or any author or contributor to
**     the covered work, without written authorization.
** 
**   3. No Misrepresentation of Source or Origin.
** 
**     You may not represent the covered work as solely your work.  All modified
**     versions of the covered work must be marked in a reasonable way to make it
**     clear that the modified work is not originating from Golden Code Development
**     Corporation or FWD.  All modified versions must contain the notices of
**     attribution required in this license.
*/

package com.goldencode.p2j.net;

import java.io.*;
import java.net.*;
import java.util.*;
import javax.net.ssl.*;

/**
 * This class represents a secure socket and all related input/output streams
 * as a one unit. This solves the problem of temporary streams closing other
 * streams and the underlying socket when finalized.
 */
public class NetSocket
{
   /** Embedded socket which may or may not be secure. */
   private Socket socket = null;

   /** SSL session. */
   private SSLSession session = null;

   /** Object input stream. */
   private ObjectInputStream objectInp = null;
   
   /** Object output stream. */
   private ObjectOutputStream objectOut = null;

   /** Stores a list of <code>SessionListener</code>s */
   private ArrayList<SessionListener> sessionListeners = null;
   
   /**
    * Constructor
    *
    * @param    socket
    *           A connected socket to embed into this class. This may be a
    *           secure socket but it is not required.
    *
    * @throws   IOException
    *           If one of the streams being built on top of the socket throws
    *           the same exception.
    */
   public NetSocket(Socket socket)
   throws IOException
   {
      this.socket = socket;

      if (socket instanceof SSLSocket)
      {
         SSLSocket ssl = (SSLSocket) socket;
         
         // get the SSL session, if this is a secure socket (NOTE: this will
         // cause the SSL handshake to occur if it has not already!)
         session = ssl.getSession();
      }

      // create object streams
      objectOut = new ObjectOutputStream(socket.getOutputStream());
      objectInp = new ObjectInputStream(socket.getInputStream());
   }
   
   /**
    * Gets the SSL session.
    *
    * @return   The SSL session or <code>null</code> if the socket is not
    *           secure.
    */
   public SSLSession getSession()
   {
      return session;
   }

   /**
    * Gets the object input stream.
    *
    * @return   The embedded object input stream.
    */
   public ObjectInputStream getObjectInput()
   {
      return objectInp;
   }

   /**
    * Gets the object output stream.
    *
    * @return   The embedded object output stream.
    */
   public ObjectOutputStream getObjectOutput()
   {
      return objectOut;
   }
   
   /**
    * Closes the socket. 
    */
   public void close()
   {
      try
      {
         socket.close();
      }
      catch (IOException ex)
      {
         // best efforts
      }
   }

   /**
    * Reports if the socket is closed (and thus can no longer be used).
    *
    * @return   <code>true</code> if the socket is closed. 
    */
   public boolean isClosed()
   {
      return socket.isClosed();
   }
   
   /**
    * Access the address of the local system, through which we are connecting.
    *
    * @return   The local address.
    */
   public InetAddress getLocalAddr()
   {
      return socket.getLocalAddress();
   }
   
   /**
    * Access the address of the local system, through which we are connecting.
    *
    * @return   The local socket address.
    */
   public InetSocketAddress getLocalSockAddr()
   {
      return new InetSocketAddress(socket.getLocalAddress(), socket.getLocalPort());
   }
   
   /**
    * Access the address of the remote system, to which we are connecting.
    *
    * @return   The remote address.
    */
   public InetAddress getRemoteAddr()
   {
      return socket.getInetAddress();
   }
   
   /**
    * Access the address of the remote system, to which we are connecting.
    *
    * @return   The remote socket address.
    */
   public InetSocketAddress getRemoteSockAddr()
   {
      return new InetSocketAddress(socket.getInetAddress(), socket.getPort());
   }

   /**
    * Gets the list with added session listeners. May return <code>null</code> if none has
    * been added.
    *
    * @return  The current session listener.
    */
   public List<SessionListener> getSessionListeners()
   {
      return sessionListeners;
   }

   /**
    * Adds a session listener.
    *
    * @param   sessionListener
    *          A session listener to be added.
    */
   public void addSessionListener(SessionListener sessionListener)
   {
      if (sessionListeners == null)
      {
         sessionListeners = new ArrayList<>();
      }

      sessionListeners.add(sessionListener);
   }
}