/*
** Module   : VirtualSession.java
** Abstract : Session implementation for multiplexed connections between nodes
**
** Copyright (c) 2004-2017, Golden Code Development Corporation.
**
** -#- -I- --Date-- --JPRM-- ------------------Description-------------------
** 001 ECF 20071101  @35863  Created initial version. Session
**                           implementation for multiplexed connections
**                           between nodes.
** 002 CA  20090901  @43817  Removed final modifier for preprocess method, to
**                           be able to override it in RequesterSession. Now
**                           the virtual sessions keep the local context ID 
**                           too, to be able to deregister from the session
**                           manager's local maps.
*/
/*
** This program is free software: you can redistribute it and/or modify
** it under the terms of the GNU Affero General Public License as
** published by the Free Software Foundation, either version 3 of the
** License, or (at your option) any later version.
**
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
** GNU Affero General Public License for more details.
**
** You may find a copy of the GNU Affero GPL version 3 at the following
** location: https://www.gnu.org/licenses/agpl-3.0.en.html
** 
** Additional terms under GNU Affero GPL version 3 section 7:
** 
**   Under Section 7 of the GNU Affero GPL version 3, the following additional
**   terms apply to the works covered under the License.  These additional terms
**   are non-permissive additional terms allowed under Section 7 of the GNU
**   Affero GPL version 3 and may not be removed by you.
** 
**   0. Attribution Requirement.
** 
**     You must preserve all legal notices or author attributions in the covered
**     work or Appropriate Legal Notices displayed by works containing the covered
**     work.  You may not remove from the covered work any author or developer
**     credit already included within the covered work.
** 
**   1. No License To Use Trademarks.
** 
**     This license does not grant any license or rights to use the trademarks
**     Golden Code, FWD, any Golden Code or FWD logo, or any other trademarks
**     of Golden Code Development Corporation. You are not authorized to use the
**     name Golden Code, FWD, or the names of any author or contributor, for
**     publicity purposes without written authorization.
** 
**   2. No Misrepresentation of Affiliation.
** 
**     You may not represent yourself as Golden Code Development Corporation or FWD.
** 
**     You may not represent yourself for publicity purposes as associated with
**     Golden Code Development Corporation, FWD, or any author or contributor to
**     the covered work, without written authorization.
** 
**   3. No Misrepresentation of Source or Origin.
** 
**     You may not represent the covered work as solely your work.  All modified
**     versions of the covered work must be marked in a reasonable way to make it
**     clear that the modified work is not originating from Golden Code Development
**     Corporation or FWD.  All modified versions must contain the notices of
**     attribution required in this license.
*/

package com.goldencode.p2j.net;

/**
 * A concrete implementation of the {@link Session} interface, built upon
 * <code>BaseSession</code>.  This implementation is to be used for virtual,
 * multiplexed connections between network nodes.
 * <p>
 * This class implements the superclass' abstract {@link #preprocess(Message)}
 * to set the session's context ID as the message's routing key's context ID.
 * This is necessary to support the multiplexing of virtual sessions across a
 * single transport queue.
 */
class VirtualSession
extends BaseSession
{
   /** 
    * The context ID to which this virtual session is registered locally
    * on this side of the connection. 
    */
   private int localContextID = 0;

   /**
    * Constructor.
    * 
    * @param   queue
    *          Message transport.
    * @param   context
    *          Security context.  Must not be <code>null</code>.
    * @param   peerID
    *          Unique identifier for this session's peer security context on
    *          the other side of the network connection.
    * @param   contextID
    *          Unique identifier for this session's local security context on
    *          the this side of the network connection.
    */
   VirtualSession(Queue queue, Object context, int peerID, int contextID)
   {
      super(queue, context, peerID);
      
      this.localContextID = contextID;
   }
   
   /**
    * Overrides the parent class' implementation to disallow the registraction
    * of a state synchronizer.  This is necessary because a virtual session is
    * multiplexed and a state synchronizer currently supports only a one to
    * one connection.
    *
    * @param   sync
    *          The synchronizer to register or <code>null</code> if the
    *          currently registered synchronizer should be deregistered.
    * 
    * @throws  UnsupportedOperationException
    *          always.
    */
   public void registerSynchronizer(StateSynchronizer sync)
   {
      throw new UnsupportedOperationException();
   }
   
   /**
    * Get the context ID to which the virtual session is registered on this 
    * side.
    * 
    * @return    the context ID to which the virtual session is registered on 
    *            this side.
    */
   int getLocalContextID()
   {
      return localContextID;
   }

   /**
    * Update the security context ID of the given message to this session's
    * (peer) context ID, before it is sent.  This will the other side of the
    * connection to demultiplex the inbound message to provide the appropriate
    * security context.
    * 
    * @param   message
    *          Message to process.
    */
   protected void preprocess(Message message)
   {
      RoutingKey key = message.getKey();
      int id = getContextID();
      key.setContextID(id);
   }
}