/*
** Module   : DataTarget.java
** Abstract : Describes an external target to which data is written from a buffer.
**
** Copyright (c) 2017, Golden Code Development Corporation.
**
** -#- -I- --Date-- ---------------------------------Description---------------------------------
** 001 ECF 20170820 Created initial version.
** 002 CA  20170830 Added missing c'tor.
** 003 ECF 20171008 Rolled back #002. Implemented file as an output resource.
*/
/*
** This program is free software: you can redistribute it and/or modify
** it under the terms of the GNU Affero General Public License as
** published by the Free Software Foundation, either version 3 of the
** License, or (at your option) any later version.
**
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
** GNU Affero General Public License for more details.
**
** You may find a copy of the GNU Affero GPL version 3 at the following
** location: https://www.gnu.org/licenses/agpl-3.0.en.html
** 
** Additional terms under GNU Affero GPL version 3 section 7:
** 
**   Under Section 7 of the GNU Affero GPL version 3, the following additional
**   terms apply to the works covered under the License.  These additional terms
**   are non-permissive additional terms allowed under Section 7 of the GNU
**   Affero GPL version 3 and may not be removed by you.
** 
**   0. Attribution Requirement.
** 
**     You must preserve all legal notices or author attributions in the covered
**     work or Appropriate Legal Notices displayed by works containing the covered
**     work.  You may not remove from the covered work any author or developer
**     credit already included within the covered work.
** 
**   1. No License To Use Trademarks.
** 
**     This license does not grant any license or rights to use the trademarks
**     Golden Code, FWD, any Golden Code or FWD logo, or any other trademarks
**     of Golden Code Development Corporation. You are not authorized to use the
**     name Golden Code, FWD, or the names of any author or contributor, for
**     publicity purposes without written authorization.
** 
**   2. No Misrepresentation of Affiliation.
** 
**     You may not represent yourself as Golden Code Development Corporation or FWD.
** 
**     You may not represent yourself for publicity purposes as associated with
**     Golden Code Development Corporation, FWD, or any author or contributor to
**     the covered work, without written authorization.
** 
**   3. No Misrepresentation of Source or Origin.
** 
**     You may not represent the covered work as solely your work.  All modified
**     versions of the covered work must be marked in a reasonable way to make it
**     clear that the modified work is not originating from Golden Code Development
**     Corporation or FWD.  All modified versions must contain the notices of
**     attribution required in this license.
*/

package com.goldencode.p2j.persist;

import java.io.*;
import com.goldencode.p2j.util.*;

/**
 * Object which describes an external target into which data is written from a temp-table buffer
 * or data set member buffer.
 */
public class DataTarget
{
   /** Normalized output stream */
   private OutputStream stream;
   
   /**
    * Constructor.
    * 
    * @param   type
    *          {@code FILE} or {@code STREAM}.
    * @param   target
    *          Name of a file or stream to which data is written.
    */
   public DataTarget(character type, character target)
   {
      initializeRemote(type.toStringMessage(), target.toStringMessage());
   }
   
   /**
    * Constructor.
    * 
    * @param   type
    *          {@code FILE} or {@code STREAM}.
    * @param   target
    *          Name of a file or stream to which data is written.
    */
   public DataTarget(String type, character target)
   {
      initializeRemote(type, target.toStringMessage());
   }
   
   /**
    * Constructor.
    * 
    * @param   type
    *          {@code FILE} or {@code STREAM}.
    * @param   target
    *          Name of a file or stream to which data is written.
    */
   public DataTarget(character type, String target)
   {
      initializeRemote(type.toStringMessage(), target);
   }
   
   /**
    * Constructor.
    * 
    * @param   type
    *          {@code FILE} or {@code STREAM}.
    * @param   target
    *          Name of a file or stream to which data is written.
    */
   public DataTarget(String type, String target)
   {
      initializeRemote(type, target);
   }
   
   /**
    * Constructor.
    * 
    * @param   target
    *          Pointer to an area of memory into which data is written.
    */
   public DataTarget(memptr target)
   {
      UnimplementedFeature.missing("DataTarget c'tor");
   }
   
   /**
    * Constructor.
    * 
    * @param   type
    *          {@code STREAM-HANDLE} or {@code HANDLE}.
    * @param   target
    *          Handle of a stream object or X-document/X-noderef object into which to write data.
    */
   public DataTarget(character type, handle target)
   {
      initializeRemote(type.toStringMessage(), target);
   }
   
   /**
    * Constructor.
    * 
    * @param   type
    *          {@code STREAM-HANDLE} or {@code HANDLE}.
    * @param   target
    *          Handle of a stream object or X-document/X-noderef object into which to write data.
    */
   public DataTarget(String type, handle target)
   {
      initializeRemote(type, target);
   }
   
   /**
    * Constructor.
    * 
    * @param   target
    *          Variable into which to write data.
    */
   public DataTarget(longchar target)
   {
      UnimplementedFeature.missing("DataTarget c'tor");
   }
   
   /**
    * Get the output stream representing the target resource to which data will be written.
    * This may be a remote or a local resource.
    * 
    * @return  Output stream.
    */
   public OutputStream getOutputStream()
   {
      return stream;
   }
   
   /**
    * Close the output stream underlying this data source.
    * 
    * @throws  IOException
    *          if there is an error closing the stream.
    */
   void close()
   throws IOException
   {
      stream.close();
   }
   
   /**
    * Initialize a remote file or stream resource as an output stream.
    * <p>
    * The stream resource is unimplemented currently.
    * 
    * @param   type
    *          Type of output resource: FILE or STREAM.
    * @param   name
    *          The name of the resource (path for a file; name for a stream).
    */
   private void initializeRemote(String type, String name)
   {
      switch(type.toUpperCase())
      {
         case "FILE":
            stream = new OutputStreamWrapper(StreamFactory.openFileStream(name, true, false));
            break;
         case "STREAM":
            // TODO: get the stream and wrap it in OutputStreamWrapper
            UnimplementedFeature.missing("DataTarget STREAM support");
            break;
         default:
            throw new IllegalArgumentException("Data target type must be FILE or STREAM");
      }
   }
   
   /**
    * Initialize a remote stream handle or handle resource as an output stream.
    * <p>
    * Both types are unimplemented features currently.
    * 
    * @param   type
    *          Type of output resource: STREAM-HANDLE or HANDLE.
    * @param   target
    *          The handle to the resource.
    */
   private void initializeRemote(String type, handle target)
   {
      switch(type.toUpperCase())
      {
         case "STREAM-HANDLE":
            UnimplementedFeature.missing("DataTarget STREAM-HANDLE support");
            break;
         case "HANDLE":
            UnimplementedFeature.missing("DataTarget HANDLE support");
            break;
         default:
            throw new IllegalArgumentException(
               "Data target type must be STREAM-HANDLE or HANDLE");
      }
   }
}
