/*
** Module   : DynamicConversionHelper.java
** Abstract : Base class with common functions for DynamicQueryHelper and DynamicValidationHelper.
**
** Copyright (c) 2017, Golden Code Development Corporation.
**
** -#- -I- --Date-- ---------------------------------Description----------------------------------
** 001 SVL 20171024 Created initial version.
*/
/*
** This program is free software: you can redistribute it and/or modify
** it under the terms of the GNU Affero General Public License as
** published by the Free Software Foundation, either version 3 of the
** License, or (at your option) any later version.
**
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
** GNU Affero General Public License for more details.
**
** You may find a copy of the GNU Affero GPL version 3 at the following
** location: https://www.gnu.org/licenses/agpl-3.0.en.html
**
** Additional terms under GNU Affero GPL version 3 section 7:
**
**   Under Section 7 of the GNU Affero GPL version 3, the following additional
**   terms apply to the works covered under the License.  These additional terms
**   are non-permissive additional terms allowed under Section 7 of the GNU
**   Affero GPL version 3 and may not be removed by you.
**
**   0. Attribution Requirement.
**
**     You must preserve all legal notices or author attributions in the covered
**     work or Appropriate Legal Notices displayed by works containing the covered
**     work.  You may not remove from the covered work any author or developer
**     credit already included within the covered work.
**
**   1. No License To Use Trademarks.
**
**     This license does not grant any license or rights to use the trademarks
**     Golden Code, FWD, any Golden Code or FWD logo, or any other trademarks
**     of Golden Code Development Corporation. You are not authorized to use the
**     name Golden Code, FWD, or the names of any author or contributor, for
**     publicity purposes without written authorization.
**
**   2. No Misrepresentation of Affiliation.
**
**     You may not represent yourself as Golden Code Development Corporation or FWD.
**
**     You may not represent yourself for publicity purposes as associated with
**     Golden Code Development Corporation, FWD, or any author or contributor to
**     the covered work, without written authorization.
**
**   3. No Misrepresentation of Source or Origin.
**
**     You may not represent the covered work as solely your work.  All modified
**     versions of the covered work must be marked in a reasonable way to make it
**     clear that the modified work is not originating from Golden Code Development
**     Corporation or FWD.  All modified versions must contain the notices of
**     attribution required in this license.
*/

package com.goldencode.p2j.persist;

import antlr.*;
import com.goldencode.ast.*;
import com.goldencode.p2j.cfg.*;
import com.goldencode.p2j.convert.*;
import com.goldencode.p2j.schema.*;
import com.goldencode.p2j.uast.*;
import com.goldencode.p2j.util.*;
import org.hibernate.mapping.*;
import java.lang.reflect.*;
import java.util.*;
import java.util.List;
import java.util.Map;

import static com.goldencode.p2j.persist.DynamicTablesHelper.normalizeName;

/**
 * Base class with common functions for {@link DynamicQueryHelper} and
 * {@link DynamicValidationHelper}.
 */
abstract class DynamicConversionHelper
{
   /** Map of P2J wrappers to their associated token type. */
   private static final Map<String, Integer> wrapperToTokenType = new HashMap<>();

   static
   {
      wrapperToTokenType.put(ParmType.CHAR.toString(),   ProgressParserTokenTypes.FIELD_CHAR);
      wrapperToTokenType.put(ParmType.INT.toString(),    ProgressParserTokenTypes.FIELD_INT);
      wrapperToTokenType.put(ParmType.INT64.toString(),  ProgressParserTokenTypes.FIELD_INT64);
      wrapperToTokenType.put(ParmType.DEC.toString(),    ProgressParserTokenTypes.FIELD_DEC);
      wrapperToTokenType.put(ParmType.LOG.toString(),    ProgressParserTokenTypes.FIELD_LOGICAL);
      wrapperToTokenType.put(ParmType.DATE.toString(),   ProgressParserTokenTypes.FIELD_DATE);
      wrapperToTokenType.put(ParmType.DT.toString(),
                             ProgressParserTokenTypes.FIELD_DATETIME);
      wrapperToTokenType.put(ParmType.DTTZ.toString(),ProgressParserTokenTypes.FIELD_DATETIME_TZ);
      wrapperToTokenType.put(ParmType.HANDLE.toString(), ProgressParserTokenTypes.FIELD_HANDLE);
      wrapperToTokenType.put(ParmType.RECID.toString(),  ProgressParserTokenTypes.FIELD_RECID);
      wrapperToTokenType.put(ParmType.ROWID.toString(),  ProgressParserTokenTypes.FIELD_ROWID);
   }

   /**
    * Create a {@link DataModelAst} with the given token type and text.
    *
    * @param    type
    *           The node's token type.
    * @param    text
    *           The node's text.
    * @param    p2oFile
    *           The name of the {@link com.goldencode.p2j.schema.DataModelAst} tree with the
    *           temp-table objects.
    *
    * @return   A {@link DataModelAst} instance setup as noted above.
    */
   private static DataModelAst createDataModelAst(int type, String text, String p2oFile)
   {
      long treeId = AstManager.get().getTreeId(p2oFile);
      DataModelAst ast = new DataModelAst(new CommonToken(type, text.intern()));
      ast.setId(AstManager.get().getNextNodeId(treeId));

      return ast;
   }

   /**
    * Build an {@link DataModelAst index component} associated with the given property AST.
    *
    * @param    comp
    *           The index component
    * @param    propAst
    *           The associated {@link DataModelAst property AST}.
    * @param    p2oFile
    *           The name of the {@link com.goldencode.p2j.schema.DataModelAst} tree with the
    *           temp-table objects.
    *
    * @return   See above.
    */
   private static DataModelAst buildIndexComponent(P2JIndexComponent comp,
                                                   DataModelAst propAst,
                                                   String p2oFile)
   {
      String name        = comp.getName();
      String historical  = (String) propAst.getAnnotation(P2OLookup.HISTORICAL);
      String datatype    = (String) propAst.getAnnotation(P2OLookup.DATATYPE);

      DataModelAst root = createDataModelAst(DataModelTokenTypes.INDEX_COL,
                                             name.intern(),
                                             p2oFile);
      root.putAnnotation(P2OLookup.HISTORICAL, historical.intern());
      root.putAnnotation(P2OLookup.DATATYPE, datatype.intern());
      if (comp.isIgnoreCase())
      {
         root.putAnnotation(P2OLookup.IGNORE_CASE, comp.isIgnoreCase());
      }
      if (comp.isDescending())
      {
         root.putAnnotation(P2OLookup.DESCEND, comp.isDescending());
      }
      if (comp.isAbbreviated())
      {
         root.putAnnotation(P2OLookup.ABBREVIATED, comp.isAbbreviated());
      }
      root.putAnnotation(P2OLookup.REFID, propAst.getId());

      return root;
   }

   /**
    * Obtain the other operand of a binary operator node. This method works for binary and unary
    * operators since the next sibling is {@code null}.
    *
    * @param   progAst
    *          The node to search for sibling.
    *
    * @return  The sibling of the passed in node. If such node does not exist, {@code null} is
    *          returned. In case of ternary operator, the result is irrelevant.
    */
   private static ProgressAst getSibling(ProgressAst progAst)
   {
      ProgressAst first = (ProgressAst) progAst.getParent().getFirstChild();
      if (first == progAst)
      {
         // I am the first child, the second child if exists, is my next sibling
         return (ProgressAst) progAst.getNextSibling();
      }
      else
      {
         // there is at least one sibling before me, return that
         return first;
      }
   }

   /**
    * Build the root {@link ProgressAst} node for the schema dictionary AST.
    *
    * @param    schema
    *           The schema name.
    *
    * @return   See above.
    */
   private static ProgressAst buildRootSchema(String schema)
   {
      ProgressAst root = createProgressAst(SchemaParserTokenTypes.DATABASE, schema);
      root.putAnnotation("srcfile", "in-memory");

      return root;
   }

   /**
    * Build the {@link DataModelAst root} AST and associated with the given schema.
    *
    * @param    schema
    *           The schema name.
    * @param    p2oFile
    *           The name of the {@link com.goldencode.p2j.schema.DataModelAst} tree with the
    *           temp-table objects.
    *
    * @return   See above.
    */
   private static DataModelAst buildRootDataModel(String schema, String p2oFile)
   {
      String pkg = Configuration.getParameter("pkgroot") + ".dmo";

      // this is a reverse generation of what schema/p2o.xml does
      DataModelAst root = createDataModelAst(DataModelTokenTypes.DATA_MODEL, schema, p2oFile);
      root.putAnnotation(P2OLookup.PACKAGE, pkg);

      return root;
   }

   /**
    * Create a {@link DataModelAst data model object} AST for the specified buffer.
    *
    * @param    buf
    *           The buffer instance.
    * @param    rbuf
    *           The {@link RecordBuffer} instance for the buffer.
    * @param    p2oFile
    *           The name of the {@link com.goldencode.p2j.schema.DataModelAst} tree with the
    *           temp-table objects.
    *
    * @return   A {@link DataModelAst} buffer definition.
    *
    * @throws   PersistenceException
    *           In case of problems during preparation of the temp-table schema and p2o ASTs.
    */
   private static DataModelAst buildDataModelClass(Buffer buf, RecordBuffer rbuf, String p2oFile)
   throws PersistenceException
   {
      String historical = normalizeName(TableMapper.getLegacyName(buf)).intern();
      String dmoName = rbuf.getDMOName().intern();

      DataModelAst root = createDataModelAst(DataModelTokenTypes.CLASS, dmoName, p2oFile);
      root.putAnnotation(P2OLookup.HISTORICAL, historical);

      // build the ID property
      DataModelAst idProp = createDataModelAst(DataModelTokenTypes.PRIMARY, "id", p2oFile);
      idProp.putAnnotation(P2OLookup.DATATYPE, "Long");
      root.addChild(idProp);

      Map<String, DataModelAst> propMap = new HashMap<>();

      // go through all non-extent properties
      Class<?> dmoClass = rbuf.getDMOImplementationClass();
      Field[] fields = dmoClass.getDeclaredFields();
      for (Field field : fields)
      {
         if (!BaseDataType.class.isAssignableFrom(field.getType()))
         {
            continue;
         }

         DataModelAst propAst = buildProperty(buf, field, 0, p2oFile);

         root.addChild(propAst);

         String histName = (String) propAst.getAnnotation(P2OLookup.HISTORICAL);
         propMap.put(normalizeName(histName), propAst);
      }

      // add the extent properties
      Map<String, Integer> extentMap = rbuf.getExtentMap();
      Class<?>[] innerClasses = dmoClass.getDeclaredClasses();
      for (Class<?> inner : innerClasses)
      {
         int extent = 0;

         fields = inner.getDeclaredFields();
         for (Field field : fields)
         {
            if (!BaseDataType.class.isAssignableFrom(field.getType()))
            {
               continue;
            }

            if (extent == 0)
            {
               String fname = field.getName();
               extent = extentMap.get(fname);
            }

            root.addChild(buildProperty(buf, field, extent, p2oFile));
         }
      }

      // composite info is not needed by P2OLookup

      Iterator<P2JIndex> idxIter = DMOIndex.databaseIndexes(rbuf.getSchema(),
            rbuf.getDatabase(),
            rbuf.getDMOInterface());
      while (idxIter.hasNext())
      {
         P2JIndex index = idxIter.next();
         String idx = DBUtils.extractIndexName(index.getName());

         String legacyName = TableMapper.getLegacyIndexName(buf, idx);
         if (legacyName != null)
         {
            // include only legacy indexes
            root.addChild(buildIndex(buf, rbuf, index, idx, propMap, p2oFile));
         }
      }

      // DataModelTokenTypes.UNIQUE, ONE_TO_ONE and ONE_TO_MANY nodes are not needed

      return root;
   }

   /**
    * Create a {@link DataModelAst property} AST for the specified DMO java field.
    *
    * @param    buf
    *           The {@link Buffer} instance to which the property belongs.
    * @param    field
    *           The java field defining this property at the DMO.
    * @param    extent
    *           The legacy extent of this property.
    * @param    p2oFile
    *           The name of the {@link com.goldencode.p2j.schema.DataModelAst} tree with the
    *           temp-table objects.
    *
    * @return   A {@link DataModelAst} property definition.
    */
   private static DataModelAst buildProperty(Buffer buf, Field field, int extent, String p2oFile)
   {
      String property = field.getName().intern();
      RecordBuffer recordBuffer = ((BufferImpl) buf).buffer();
      String historical;
      if (DynamicTablesHelper.isDynamicTableBuffer(recordBuffer))
      {
         historical = DynamicTablesHelper.get4GLFieldName(recordBuffer.getDMOInterface(),
                                                          property);
      }
      else
      {
         historical = TableMapper.getLegacyFieldName(buf, property);
      }
      String datatype = field.getType().getName();
      datatype = datatype.substring(datatype.lastIndexOf('.') + 1).intern();

      DataModelAst root = createDataModelAst(DataModelTokenTypes.PROPERTY, property, p2oFile);
      root.putAnnotation(P2OLookup.HISTORICAL, historical.intern());
      root.putAnnotation(P2OLookup.DATATYPE, datatype);
      if (extent != 0)
      {
         root.putAnnotation(P2OLookup.EXTENT, new Long(extent));
      }
      if ("character".equals(datatype))
      {
         LegacyField legacy = field.getAnnotation(LegacyField.class);
         if (legacy.caseSensitive())
         {
            root.putAnnotation(P2OLookup.CASE_SENSITIVE, true);
         }
      }

      // INITIAL is not needed for P2OLookup

      return root;
   }

   /**
    * Build a {@link DataModelAst index} for the specified buffer.
    *
    * @param    buf
    *           The {@link Buffer} instance to which the index belongs.
    * @param    rbuf
    *           The {@link RecordBuffer} instance.
    * @param    index
    *           The index definition.
    * @param    idxName
    *           The index name.
    * @param    propMap
    *           A map with the {@link DataModelAst property} ASTs by their legacy name.
    * @param    p2oFile
    *           The name of the {@link com.goldencode.p2j.schema.DataModelAst} tree with the
    *           temp-table objects.
    *
    * @return   A {@link DataModelAst index} definition.
    */
   private static DataModelAst buildIndex(Buffer                    buf,
                                          RecordBuffer              rbuf,
                                          P2JIndex                  index,
                                          String                    idxName,
                                          Map<String, DataModelAst> propMap,
                                          String                    p2oFile)
   {
      String historical = TableMapper.getLegacyIndexName(buf, idxName).intern();
      boolean unique    = index.isUnique();
      boolean primary   = index.isPrimary();
      boolean word      = index.isWord();

      DataModelAst root = createDataModelAst(DataModelTokenTypes.INDEX, idxName, p2oFile);
      root.putAnnotation(P2OLookup.HISTORICAL, historical);
      if (unique)
      {
         root.putAnnotation(P2OLookup.UNIQUE, unique);
      }
      if (primary)
      {
         root.putAnnotation(P2OLookup.PRIMARY, primary);
      }
      if (word)
      {
         root.putAnnotation(P2OLookup.WORD, word);
      }

      PersistentClass pclass = DatabaseManager.getClassMapping(rbuf.getDatabase(),
                                                               rbuf.getDMOImplementationClass());

      Map<String, String> propsByField = new HashMap<>();
      Iterator<P2JIndexComponent> compIter = index.components();
      while (compIter.hasNext())
      {
         P2JIndexComponent comp = compIter.next();
         String fieldName = comp.getName();
         String propertyName = propsByField.get(fieldName);
         if (propertyName == null)
         {
            Iterator<Property> itrProps = pclass.getDeclaredPropertyIterator();
            while (itrProps.hasNext())
            {
               Property hprop = itrProps.next();
               if (hprop.getColumnSpan() == 1)
               {
                  String field = ((Column) hprop.getColumnIterator().next()).getName();
                  if (field.equals(fieldName))
                  {
                     propertyName = hprop.getName();
                     break;
                  }
               }
               else
               {
                  // TODO: multi-col span properties
               }
            }

            propsByField.put(fieldName, propertyName);
         }

         String histName = TableMapper.getLegacyFieldName(buf, propertyName);
         if (histName != null)
         {
            // include only legacy fields
            root.addChild(buildIndexComponent(comp,
                                              propMap.get(normalizeName(histName)),
                                              p2oFile));
         }
      }

      return root;
   }

   /**
    * Build a temp-table definition to be loaded in the {@link SchemaDictionary}. Use the given
    * {@link DataModelAst data model object} AST as the source.
    *
    * @param    p2o
    *           The data model object AST.
    * @param    rbuf
    *           The record buffer instance.
    * @param    order
    *           The order of this temp-table AST.
    *
    * @return   See above.
    */
   private static ProgressAst buildTempTable(DataModelAst p2o, RecordBuffer rbuf, integer order)
   {
      String schema      = rbuf.getSchema();
      Class<?> iface     = rbuf.getDMOInterface();
      String legacyTable = (String) p2o.getAnnotation("historical");

      ProgressAst root = createProgressAst(SchemaParserTokenTypes.TEMP_TABLE, legacyTable);
      root.putAnnotation("lastOrder", order.toJavaLongType());

      int fieldIdx = 0;
      for (int i = 0; i < p2o.getNumImmediateChildren(); i++)
      {
         DataModelAst child = (DataModelAst) p2o.getChildAt(i);

         if (child.getType() == DataModelTokenTypes.PROPERTY)
         {
            String datatype = (String) child.getAnnotation("datatype");
            int ftype = wrapperToTokenType.get(datatype);
            String legacyField = (String) child.getAnnotation("historical");

            ProgressAst field = createProgressAst(ftype, legacyField);

            // the following properties are not needed by SchemaDictionary:
            // 1. sqlWidth
            // 2. extent
            // 3. decimals
            // 4. length
            // 5. maxwidth

            if (DMOIndex.isCaseSensitive(schema, iface, child.getText()))
            {
               ProgressAst pcase = createProgressAst(SchemaParserTokenTypes.KW_CASE_SEN,
                                                     "case-sensitive");
               pcase.addChild(createProgressAst(SchemaParserTokenTypes.BOOL_TRUE, "true"));
               field.addChild(pcase);
            }

            // add ORDER node
            ProgressAst pord = createProgressAst(SchemaParserTokenTypes.ORDER, "ORDER");
            pord.addChild(createProgressAst(SchemaParserTokenTypes.NUM_LITERAL,
                                            Long.toString(fieldIdx)));
            fieldIdx = fieldIdx + 10;

            field.addChild(pord);

            Long extent = (Long) child.getAnnotation("extent");
            if (extent != null)
            {
               ProgressAst pext = createProgressAst(SchemaParserTokenTypes.KW_EXTENT, "extent");
               pext.addChild(createProgressAst(SchemaParserTokenTypes.NUM_LITERAL,
                     extent.toString()));

               field.addChild(pext);
            }

            root.addChild(field);
         }
         else if (child.getType() == DataModelTokenTypes.INDEX)
         {
            // add a new index
            String legacyIndex = (String) child.getAnnotation("historical");
            ProgressAst rootIdx = createProgressAst(SchemaParserTokenTypes.INDEX, legacyIndex);

            // build index properties
            ProgressAst props = createProgressAst(SchemaParserTokenTypes.PROPERTIES, "");
            if (child.isAnnotation(P2OLookup.UNIQUE))
            {
               props.addChild(createProgressAst(SchemaParserTokenTypes.KW_UNIQUE, "UNIQUE"));
            }
            if (child.isAnnotation(P2OLookup.PRIMARY))
            {
               props.addChild(createProgressAst(SchemaParserTokenTypes.KW_PRIMARY, "PRIMARY"));
            }
            if (child.isAnnotation(P2OLookup.WORD))
            {
               props.addChild(createProgressAst(SchemaParserTokenTypes.KW_WORD_IDX, "WORD"));
            }

            rootIdx.addChild(props);

            // build index components
            DataModelAst idxCol =
                  (DataModelAst) child.getImmediateChild(DataModelTokenTypes.INDEX_COL, null);
            while (idxCol != null)
            {
               String legacyField = (String) child.getAnnotation("historical");

               ProgressAst idxField = createProgressAst(SchemaParserTokenTypes.INDEX_FIELD,
                                                        legacyField);
               if (idxCol.isAnnotation(P2OLookup.ABBREVIATED))
               {
                  idxField.addChild(createProgressAst(SchemaParserTokenTypes.KW_ABBV,
                                                      "ABBREVIATED"));
               }
               if (idxCol.isAnnotation(P2OLookup.DESCEND))
               {
                  idxField.addChild(createProgressAst(SchemaParserTokenTypes.KW_DESCEND,
                                                      "DESCENDING"));
               }
               else
               {
                  idxField.addChild(createProgressAst(SchemaParserTokenTypes.KW_ASCEND,
                                                      "ASCENDING"));
               }

               rootIdx.addChild(idxField);

               idxCol = (DataModelAst) child.getImmediateChild(DataModelTokenTypes.INDEX_COL,
                                                               idxCol);
            }

            root.addChild(rootIdx);
         }
      }

      order.assign(order.intValue() + 10);

      return root;
   }

   /**
    * Create a {@link ProgressAst} with the given token type and text.
    *
    * @param    type
    *           The node's token type.
    * @param    text
    *           The node's text.
    *
    * @return   A {@link ProgressAst} instance setup as noted above.
    */
   private static ProgressAst createProgressAst(int type, String text)
   {
      return new ProgressAst(new CommonToken(type, text));
   }

   /**
    * Injects a new variable in place of a literal or field reference. This method makes the
    * following changes in the structure of the {@code progAst} tree:
    * <ul>
    *    <li>created the DEFINE statement structure/subtree of a new {@code varName} variable of
    *          the type required</li>
    *    <li>adds all necessary annotations to the newly created nodes as the
    *          {@code post-parse-fixup.xml} was run against it.</li>
    *    <li>replaces the literal or field reference with variable reference (morph it by changing
    *          its text/type and) the required annotations (including refid to newly created
    *          DEFINE VAR stmt)</li>
    *    <li>infers the type of the variable and converts the old text of the literal to a
    *          {@link BaseDataType} that will be returned. This will be used to initialize the
    *          variable just created.</li>
    * </ul>
    *
    * @param   litAst
    *          The AST of a literal or field references to be replaced by a variable.
    * @param   varName
    *          The name of the variable to be created.
    * @param   progAst
    *          The root of the procedure. It is only needed to inject the definition of the new
    *          variable.
    *
    * @return  The value as {@link BaseDataType} of the literal replaced by the new variable. If
    *          a field reference is replaced, then unknown value of proper {@link BaseDataType} is
    *          returned.
    */
   static Object injectVariable(ProgressAst litAst, String varName, ProgressAst progAst)
   {
      return injectVariable(new ProgressAst[]{litAst}, varName, progAst);
   }

   /**
    * Injects a new variable in place of a literal or field reference. This method makes the
    * following changes in the structure of the {@code progAst} tree:
    * <ul>
    *    <li>created the DEFINE statement structure/subtree of a new {@code varName} variable of
    *          the type required</li>
    *    <li>adds all necessary annotations to the newly created nodes as the
    *          {@code post-parse-fixup.xml} was run against it.</li>
    *    <li>replaces the literal or field reference with variable reference (morph it by changing
    *          its text/type and) the required annotations (including refid to newly created
    *          DEFINE VAR stmt)</li>
    *    <li>infers the type of the variable and converts the old text of the literal to a
    *          {@link BaseDataType} that will be returned. This will be used to initialize the
    *          variable just created.</li>
    * </ul>
    *
    * @param   litAsts
    *          The list of ASTs of literals or field references to be replaced by a variable. All
    *          ASTs are replaced with the same variable.
    * @param   varName
    *          The name of the variable to be created.
    * @param   progAst
    *          The root of the procedure. It is only needed to inject the definition of the new
    *          variable.
    *
    * @return  The value as {@link BaseDataType} of the literal replaced by the new variable. If
    *          a field reference is replaced, then unknown value of proper {@link BaseDataType} is
    *          returned.
    */
   static Object injectVariable(ProgressAst[] litAsts,
                                String varName,
                                ProgressAst progAst)
   {
      BaseDataType ret = null;
      String className = null;

      ProgressAst stmt = new ProgressAst();
      stmt.setType(ProgressParserTokenTypes.STATEMENT);
      stmt.setText("statement");
      progAst.graftAt(stmt, 0);  // insert in front, before buffer definitions

      ProgressAst define = new ProgressAst();
      define.setType(ProgressParserTokenTypes.DEFINE_VARIABLE);
      define.setText("DEFINE");
      define.putAnnotation("name", varName);
      define.putAnnotation("vardef", true);
      stmt.graft(define);

      ProgressAst symbol = new ProgressAst();
      symbol.setType(ProgressParserTokenTypes.SYMBOL);
      symbol.setText(varName);
      define.graft(symbol);

      ProgressAst as = new ProgressAst();
      as.setType(ProgressParserTokenTypes.KW_AS);
      as.setText("AS");
      define.graft(as);

      // step 1: evaluate the current literal as a BDT with type identical to original literal
      ProgressAst litAst = litAsts[0];
      int literalType = litAst.getType();
      switch (literalType)
      {
         case ProgressParserTokenTypes.STRING:
            ret = new character(character.progressToJavaString(litAst.getText()));
            break;
         case ProgressParserTokenTypes.NUM_LITERAL:
            ret = new int64(litAst.getText());
            break;
         case ProgressParserTokenTypes.DATE_LITERAL:
            ret = new date(litAst.getText());
            break;
         case ProgressParserTokenTypes.DATETIME_LITERAL:
            ret = datetime.fromLiteral(litAst.getText());
            break;
         case ProgressParserTokenTypes.DATETIME_TZ_LITERAL:
            ret = datetimetz.fromLiteral(litAst.getText());
            break;
         case ProgressParserTokenTypes.DEC_LITERAL:
            ret = decimal.fromLiteral(litAst.getText());
            break;
         case ProgressParserTokenTypes.BOOL_TRUE:
         case ProgressParserTokenTypes.BOOL_FALSE:
            ret = new logical(literalType == ProgressParserTokenTypes.BOOL_TRUE);
            break;
         case ProgressParserTokenTypes.UNKNOWN_VAL: /* disabled */
            ret = new unknown();
            break;
         default:
            // is field reference
      }

      // step 2. create the TYPE node of a type required by the sibling. If needed, an implicit
      //         conversion is applied to value obtained in step 1. If sibling does not exist or
      //         both operands are of the same type, the conversion is skipped and the type of
      //         the original literal is used.
      int var4glType = literalType;
      switch (litAst.getParent().getType())
      {
         case ProgressParserTokenTypes.KW_EQ:
         case ProgressParserTokenTypes.EQUALS:
         case ProgressParserTokenTypes.KW_NE:
         case ProgressParserTokenTypes.NOT_EQ:
         case ProgressParserTokenTypes.KW_LT:
         case ProgressParserTokenTypes.LT:
         case ProgressParserTokenTypes.KW_LTE:
         case ProgressParserTokenTypes.LTE:
         case ProgressParserTokenTypes.KW_GT:
         case ProgressParserTokenTypes.GT:
         case ProgressParserTokenTypes.KW_GTE:
         case ProgressParserTokenTypes.GTE:
            ProgressAst sibling = getSibling(litAst);
            if (sibling != null)
            {
               if (sibling.getType() > ProgressParserTokenTypes.BEGIN_FIELDTYPES &&
                   sibling.getType() < ProgressParserTokenTypes.END_FIELDTYPES)
               {
                  // only fields are taken into consideration to avoid database CASTing issues
                  var4glType = sibling.getType();
               }
               else if (sibling.getType() == ProgressParserTokenTypes.FUNC_RECID)
               {
                  // we want our operand to be bigint to match the record.id field in db
                  var4glType = ProgressParserTokenTypes.NUM_LITERAL;
               }
            }
            break;
      }
      ProgressAst type = new ProgressAst();
      int varType = 0;
      switch (var4glType)
      {
         case ProgressParserTokenTypes.STRING:
         case ProgressParserTokenTypes.FIELD_CHAR:
            type.setType(ProgressParserTokenTypes.KW_CHAR);
            type.setText("CHARACTER");
            className = "character";
            varType = ProgressParserTokenTypes.VAR_CHAR;
            if (literalType != ProgressParserTokenTypes.STRING)
            {
               ret = new character(ret);
            }
            break;
         case ProgressParserTokenTypes.NUM_LITERAL:
         case ProgressParserTokenTypes.FIELD_INT:
         case ProgressParserTokenTypes.FIELD_INT64:
            type.setType(ProgressParserTokenTypes.KW_INT64);
            type.setText("INT64");
            className = "int64";
            varType = ProgressParserTokenTypes.VAR_INT64;
            if (literalType != ProgressParserTokenTypes.NUM_LITERAL)
            {
               ret = new int64(ret);
            }
            break;
         case ProgressParserTokenTypes.DATE_LITERAL:
         case ProgressParserTokenTypes.FIELD_DATE:
            type.setType(ProgressParserTokenTypes.KW_DATE);
            type.setText("DATE");
            className = "date";
            varType = ProgressParserTokenTypes.VAR_DATE;
            if (literalType != ProgressParserTokenTypes.DATE_LITERAL)
            {
               ret = new date(ret);
            }
            break;
         case ProgressParserTokenTypes.DATETIME_LITERAL:
         case ProgressParserTokenTypes.FIELD_DATETIME:
            type.setType(ProgressParserTokenTypes.KW_DATETIME);
            type.setText("DATETIME");
            className = "datetime";
            varType = ProgressParserTokenTypes.VAR_DATETIME;
            if (literalType != ProgressParserTokenTypes.DATETIME_LITERAL)
            {
               ret = new datetime(ret);
            }
            break;
         case ProgressParserTokenTypes.DATETIME_TZ_LITERAL:
         case ProgressParserTokenTypes.FIELD_DATETIME_TZ:
            type.setType(ProgressParserTokenTypes.KW_DATE_TZ);
            type.setText("DATETIME-TZ");
            className = "datetimetz";
            varType = ProgressParserTokenTypes.VAR_DATETIME_TZ;
            if (literalType != ProgressParserTokenTypes.DATETIME_TZ_LITERAL)
            {
               ret = new datetimetz(ret);
            }
            break;
         case ProgressParserTokenTypes.DEC_LITERAL:
         case ProgressParserTokenTypes.FIELD_DEC:
            type.setType(ProgressParserTokenTypes.KW_DEC);
            type.setText("DECIMAL");
            className = "decimal";
            varType = ProgressParserTokenTypes.VAR_DEC;
            if (literalType != ProgressParserTokenTypes.DEC_LITERAL)
            {
               ret = new decimal(ret);
            }
            break;
         case ProgressParserTokenTypes.BOOL_TRUE:
         case ProgressParserTokenTypes.BOOL_FALSE:
         case ProgressParserTokenTypes.FIELD_LOGICAL:
            type.setType(ProgressParserTokenTypes.KW_LOGICAL);
            type.setText("LOGICAL");
            className = "logical";
            varType = ProgressParserTokenTypes.VAR_LOGICAL;
            if (literalType != ProgressParserTokenTypes.BOOL_TRUE &&
                literalType != ProgressParserTokenTypes.BOOL_FALSE)
            {
               ret = new logical(ret);
            }
            break;
         default:
            // throw exception or something?
      }

      as.graft(type);

      // morphing the literal node to variable
      for (ProgressAst ast : litAsts)
      {
         ast.setText(varName);
         ast.setType(varType);
         ast.removeAnnotation("is-literal");    // remove old annotation
         ast.putAnnotation("oldtype", Long.valueOf(ProgressParserTokenTypes.SYMBOL));
         ast.putAnnotation("refid", define.getId());
      }

      // annotate the variable definition with correct type
      define.putAnnotation("type", Long.valueOf(varType));
      define.putAnnotation("javaname", varName);
      define.putAnnotation("classname", className);

      return ret;
   }

   /**
    * Process the definitions of the temp-tables (associated with the temporary buffers in the
    * given list) to create the associated dictionary and p2o ASTs.
    * <p>
    * Also, it appends to the progress code string any needed 4GL code (like DEFINE BUFFER
    * statements for explicit buffers, to refer back the originating table.
    *
    * @param    dict
    *           The schema dictionary where the temp-tables will be loaded.
    * @param    pcode
    *           Append any new progress code here.
    * @param    p2oFile
    *           The name of the {@link DataModelAst} tree with the temp-table objects for the
    *           query or validation statement.
    * @param    buffers
    *           The buffers used by the query or validation statement.
    *
    * @throws   PersistenceException
    *           In case of problems during preparation of the temp-table schema and p2o ASTs.
    */
   static void prepareTempTable(SchemaDictionary dict,
                                StringBuilder pcode,
                                String p2oFile,
                                List<Buffer> buffers)
   throws PersistenceException
   {
      DataModelAst p2oRoot = buildRootDataModel(DatabaseManager.TEMP_TABLE_SCHEMA, p2oFile);
      p2oRoot.brainwash(p2oFile);
      ProgressAst schemaRoot = buildRootSchema(DatabaseManager.TEMP_TABLE_SCHEMA);

      integer tableOrder = new integer(0);
      for (Buffer buf : buffers)
      {
         BufferImpl impl = (BufferImpl) buf;
         RecordBuffer rbuf = impl.buffer();
         rbuf.initialize();

         String legacyBufName = impl.name().toStringMessage();
         String legacyTableName = TableMapper.getLegacyName(buf);

         if (rbuf.isTemporary())
         {
            // create the schema and p2o definition for this temp-table
            DataModelAst model = buildDataModelClass(buf, rbuf, p2oFile);
            ProgressAst table = buildTempTable(model, rbuf, tableOrder);

            schemaRoot.graft(table);

            p2oRoot.graft(model);
         }

         // must create explicit buffers, if the legacy buffer name is not the same as the legacy
         // default buffer name (which is the same as the legacy table name).
         if (!legacyBufName.equalsIgnoreCase(legacyTableName))
         {
            pcode.append("DEFINE BUFFER ").append(legacyBufName).append(" FOR ");
            if (rbuf.isTemporary())
            {
               pcode.append("TEMP-TABLE ");
            }
            pcode.append(legacyTableName).append(".\n");
         }
      }

      // save the P2O tree, to expose it to the conversion rules
      AstManager.get().saveTree(p2oRoot, p2oFile, true);

      // load the temp schema
      dict.loadFromAst(schemaRoot, false);
   }

   /**
    * Prepare the query AST by placing the appropriate annotations with the converted java names
    * for the external program and also by created bogus
    * {@link ProgressParserTokenTypes#DEFINE_TEMP_TABLE} nodes (which would be created from a
    * {@code DEFINE TEMP-TABLE tt.} statement, with no fields); the later is needed to allow
    * proper conversion of the default buffer name.
    *
    * @param    root
    *           The root AST of the converted progress code.
    * @param    buffers
    *           The buffers used by the query or validation statement.
    * @param    fileName
    *           The file name of the class containing the converted code.
    * @param    javaClassName
    *           The name of the generated java class.
    */
   static void prepareTree(AnnotatedAst root,
                           List<Buffer> buffers,
                           String fileName,
                           String javaClassName)
   {
      // Create a bogus DEFINE TEMP-TABLE for each temp buffer (just to let their converted
      // buffer names resolve properly).
      // Dynamic buffers have unique DMO aliases, so "some-buffer" name may be converted to the
      // alias like "someTable__1", while aliases in predicate will be converted using standard
      // P2J rules, which will result to the converted alias "someTable". So below we explicitly
      // specify converted DMO alias like "someTable__1" for the conversion rules using
      // "force_dmo_alias" annotation.
      for (Buffer buf : buffers)
      {
         BufferImpl impl = (BufferImpl) buf;
         String legacyTableName = normalizeName(TableMapper.getLegacyName(buf));

         if (!impl.buffer().isTemporary())
         {
            continue;
         }

         ProgressAst stmt = createProgressAst(ProgressParserTokenTypes.STATEMENT, "statement");
         stmt.putAnnotation("reachable", true);

         ProgressAst defChild = createProgressAst(ProgressParserTokenTypes.DEFINE_TEMP_TABLE,
               "DEFINE");
         defChild.putAnnotation("reachable", true);
         defChild.putAnnotation("schemaname", legacyTableName);
         defChild.putAnnotation("bufname", legacyTableName);
         defChild.putAnnotation("dbname", "");
         if (impl.buffer().isDynamic())
         {
            defChild.putAnnotation("force_dmo_alias", impl.buffer().getDMOAlias());
            if (DynamicTablesHelper.isDynamicTableBuffer(impl.buffer()))
            {
               defChild.putAnnotation("is_dynamic_table", true);
            }
         }
         stmt.graft(defChild);

         ProgressAst ttChild = createProgressAst(ProgressParserTokenTypes.KW_TEMP_TAB,
               "TEMP-TABLE");
         ttChild.putAnnotation("reachable", true);
         defChild.graft(ttChild);

         ProgressAst symChild = createProgressAst(ProgressParserTokenTypes.SYMBOL,
               legacyTableName);
         symChild.putAnnotation("reachable", true);
         ttChild.graft(symChild);

         root.graftAt(stmt, 0);
      }

      // put annotation related to name, package, etc
      root.putAnnotation("file", fileName);
      root.putAnnotation("classname", javaClassName);
      root.putAnnotation("javaname", javaClassName);
      root.putAnnotation("relative-name", fileName);
      root.putAnnotation("basename", fileName);
      root.putAnnotation("package-base", "");
      root.putAnnotation("pkgname", "com.goldencode.p2j.persist.dynquery");

      // make "force_dmo_alias" annotations, errors are ignored at this point, they will
      // be raised later anyway
      Iterator<Aast> iter = root.iterator();
      while (iter.hasNext())
      {
         Aast chAst = iter.next();

         if (chAst.getType() == ProgressParserTokenTypes.RECORD_PHRASE ||
             chAst.getType() == ProgressParserTokenTypes.DEFINE_BUFFER)
         {
            Aast bufAst = (Aast) chAst.getFirstChild();
            String bufName = (String) bufAst.getAnnotation("bufname");
            if (bufName != null)
            {
               int dotPos = bufName.lastIndexOf(".");
               if (dotPos >= 0)
               {
                  bufName = bufName.substring(dotPos + 1);
               }

               int idx = locateBuffer(buffers, bufName);
               if (idx >= 0)
               {
                  BufferImpl impl = (BufferImpl) buffers.get(idx);
                  bufAst.putAnnotation("force_dmo_alias", impl.buffer().getDMOAlias());
                  if (DynamicTablesHelper.isDynamicTableBuffer(impl.buffer()))
                  {
                     bufAst.putAnnotation("is_dynamic_table", true);
                  }
               }
            }
         }
      }
   }

   /**
    * Locate a {@link Buffer} with the given {@code name}, from the specified list of buffers,
    * and return its position.
    *
    * @param    buffers
    *           List of buffers to be searched.
    * @param    name
    *           The buffer name.
    *
    * @return   The 0-based position of this buffer or -1 if not found.
    */
   static int locateBuffer(List<Buffer> buffers, String name)
   {
      int idx = 0;
      for (Buffer buf : buffers)
      {
         if (name.equalsIgnoreCase(buf.name().toStringMessage()))
         {
            return idx;
         }

         idx++;
      }

      return -1;
   }
}
