/*
** Module   : ForeignResolver.java
** Abstract : Helper object for foreign relation joins
**
** Copyright (c) 2006-2017, Golden Code Development Corporation.
**
** -#- -I- --Date-- -JPRM-- ----------------------------Description------------------------------
** 001 ECF 20060302  @25012 Created initial version. Resolves foreign
**                          records using legacy keys.
** 002 ECF 20060713  @28045 Use new DBUtils class.
** 003 ECF 20061114  @31137 Modified to support expanded foreign
**                          association synchronization. Also embedded
**                          rtrim() function into character queries to
**                          match database indices.
** 004 ECF 20070629  @35298 Minor optimization. Replaced StringBuffer
**                          with StringBuilder.
** 005 CA  20080819  @39460 Support API change in FieldReference.
** 006 SVL 20140210         Use HQLExpression instead of HQL string.
** 007 ECF 20160202         Replaced Apache commons logging with J2SE logging.
** 008 OM  20160603         Fixed rtrimming to space character only.
*/
/*
** This program is free software: you can redistribute it and/or modify
** it under the terms of the GNU Affero General Public License as
** published by the Free Software Foundation, either version 3 of the
** License, or (at your option) any later version.
**
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
** GNU Affero General Public License for more details.
**
** You may find a copy of the GNU Affero GPL version 3 at the following
** location: https://www.gnu.org/licenses/agpl-3.0.en.html
** 
** Additional terms under GNU Affero GPL version 3 section 7:
** 
**   Under Section 7 of the GNU Affero GPL version 3, the following additional
**   terms apply to the works covered under the License.  These additional terms
**   are non-permissive additional terms allowed under Section 7 of the GNU
**   Affero GPL version 3 and may not be removed by you.
** 
**   0. Attribution Requirement.
** 
**     You must preserve all legal notices or author attributions in the covered
**     work or Appropriate Legal Notices displayed by works containing the covered
**     work.  You may not remove from the covered work any author or developer
**     credit already included within the covered work.
** 
**   1. No License To Use Trademarks.
** 
**     This license does not grant any license or rights to use the trademarks
**     Golden Code, FWD, any Golden Code or FWD logo, or any other trademarks
**     of Golden Code Development Corporation. You are not authorized to use the
**     name Golden Code, FWD, or the names of any author or contributor, for
**     publicity purposes without written authorization.
** 
**   2. No Misrepresentation of Affiliation.
** 
**     You may not represent yourself as Golden Code Development Corporation or FWD.
** 
**     You may not represent yourself for publicity purposes as associated with
**     Golden Code Development Corporation, FWD, or any author or contributor to
**     the covered work, without written authorization.
** 
**   3. No Misrepresentation of Source or Origin.
** 
**     You may not represent the covered work as solely your work.  All modified
**     versions of the covered work must be marked in a reasonable way to make it
**     clear that the modified work is not originating from Golden Code Development
**     Corporation or FWD.  All modified versions must contain the notices of
**     attribution required in this license.
*/

package com.goldencode.p2j.persist;

import java.io.*;
import java.util.*;
import java.util.logging.*;
import org.hibernate.type.*;
import com.goldencode.p2j.util.*;

/**
 * Resolves records on either end of a foreign key association, based upon
 * the values of one or more properties in a DMO, where those properties
 * represent the legacy "foreign key" in the pre-conversion application.
 * This allows a record's foreign key to be kept in synch with changes to
 * these legacy key properties.  This is necessary because a true foreign key
 * is only added to a table as part of the conversion process;  the legacy
 * application knows nothing of these new keys, so the runtime must keep them
 * in synch as the legacy key values change.
 */
final class ForeignResolver
{
   /** Logger */
   private static final Logger LOG = LogHelper.getLogger(ForeignResolver.class.getName());
   
   /** Information about the primary/foreign key relation */
   private final RelationInfo info;
   
   /** HQL query statement */
   private final String hql;
   
   /** Field references used to retrieve query substitution values */
   private final List fieldRefs = new ArrayList();
   
   /** Does resolver walk the foreign key association from the foreign end? */
   private final boolean inverse;
   
   /** The associated database */
   private final Database database;

   /**
    * Constructor.
    *
    * @param   info
    *          Descriptor of the relation between the local and foreign DMOs.
    * @param   buffer
    *          A record buffer which represents the "local" (i.e., refering)
    *          DMO if <code>inverse</code> is <code>false</code>, or which
    *          represents the foreign (i.e., referent) DMO if
    *          <code>inverse</code> is <code>true</code>.
    * @param   inverse
    *          <code>true</code> to construct an object which resolves the
    *          "local" (i.e., refering) DMO(s) given a foreign (i.e.,
    *          referent) DMO;  <code>false</code> to construct an object
    *          which resolves the foreign DMO given a local DMO.
    */
   ForeignResolver(RelationInfo info, RecordBuffer buffer, boolean inverse)
   {
      this.info = info;
      this.database = buffer.getDatabase();
      this.inverse = inverse;
      this.hql = generateHQL(info);
      Iterator propIter = (inverse
                           ? info.foreignProperties()
                           : info.localProperties());
      createFieldRefs(buffer.getDMOInterface(), propIter);
   }
   
   /**
    * Resolve a foreign record using the values of one or more legacy key
    * properties in <code>buffer</code>'s current record.
    *
    * @param   buffer
    *          The record buffer containing the current, local record for
    *          which a related, foreign record must be found.
    *
    * @return  The primary key ID of the foreign record associated with the
    *          current value(s) of the legacy key(s) in <code>buffer</code>'s
    *          current record, or <code>null</code> if none was found.
    *
    * @throws  PersistenceException
    *          if any error occurs querying the foreign record from the
    *          database.
    * @throws  IllegalStateException
    *          if the current record in the buffer is <code>null</code>, or
    *          if this object has been configured to retrieve the record for
    *          the inverse, foreign key association.
    */
   Serializable resolveForeign(RecordBuffer buffer)
   throws PersistenceException
   {
      if (inverse)
      {
         throw new IllegalStateException(
            "Resolver configured to resolve inverse foreign association");
      }
      
      Persistable local = buffer.getCurrentRecord();
      if (local == null)
      {
         throw new IllegalStateException(
            "Cannot resolve foreign DMO for a null record");
      }
      
      Persistence persistence = buffer.getPersistence();
      Object[] args = getArguments(local);
      Type[] types = DBUtils.makeTypeArray(args);
      
      if (LOG.isLoggable(Level.FINE))
      {
         LOG.fine("HQL:  " + hql);
         
         StringBuilder buf = new StringBuilder("PARMS:  ");
         int len = args.length;
         if (len == 0)
         {
            buf.append("N/A");
         }
         
         for (int k = 0; k < len; k++)
         {
            if (k > 0)
            {
               buf.append(", ");
            }
            buf.append(String.valueOf(args[k]));
         }
         
         LOG.fine(buf.toString());
         
         buf.setLength(0);
         buf.append("LOCAL DMO:  ");
         buf.append(local);
         
         if (local != null)
         {
            buf.append(" [");
            buf.append(local.getId());
            buf.append("]");
         }
      
         LOG.fine(buf.toString());
      }
      
      List list = persistence.list(hql, args, types, 0, 0, true);
      Serializable id = (Serializable) (list == null ? null : list.get(0));
      
      if (LOG.isLoggable(Level.FINE))
      {
         StringBuilder buf = new StringBuilder("FOREIGN DMO:  ");
         buf.append(id != null ? id : "N/A");
         
         LOG.fine(buf.toString());
      }
      
      return id;
   }
   
   /**
    * Resolve zero or more records which refer via foreign key association to
    * the record currently stored in <code>buffer</code>, using the values
    * of one or more legacy key properties in that record.
    *
    * @param   buffer
    *          The record buffer containing the current, foreign record for
    *          which one or more related, "local" records must be found.
    *
    * @return  A list of zero or more primary key IDs for the record(s), if
    *          any, associated with the current value(s) of the legacy key(s)
    *          in <code>buffer</code>'s current record.
    *
    * @throws  PersistenceException
    *          if any error occurs querying the database.
    * @throws  IllegalStateException
    *          if the current record in the buffer is <code>null</code>, or
    *          if this object has not been configured to retrieve the record
    *          for the inverse, foreign key association..
    */
   List resolveLocal(RecordBuffer buffer)
   throws PersistenceException
   {
      if (!inverse)
      {
         throw new IllegalStateException(
            "Resolver not configured to resolve inverse foreign association");
      }
      
      Persistable dmo = buffer.getCurrentRecord();
      if (dmo == null)
      {
         throw new IllegalStateException("Cannot resolve local DMO for a null record");
      }
      
      Persistence persistence = buffer.getPersistence();
      Object[] args = getArguments(dmo);
      Type[] types = DBUtils.makeTypeArray(args);
      
      if (LOG.isLoggable(Level.FINE))
      {
         LOG.fine("HQL:  " + hql);
         
         StringBuilder buf = new StringBuilder("PARMS:  ");
         int len = args.length;
         if (len == 0)
         {
            buf.append("N/A");
         }
         
         for (int k = 0; k < len; k++)
         {
            if (k > 0)
            {
               buf.append(", ");
            }
            buf.append(String.valueOf(args[k]));
         }
         
         LOG.fine(buf.toString());
         
         buf.setLength(0);
         buf.append("FOREIGN DMO:  ");
         buf.append(dmo);
         buf.append(" [");
         buf.append(dmo.getId());
         buf.append("]");
      
         LOG.fine(buf.toString());
      }
      
      List list = persistence.list(hql, args, types, 0, 0, true);
      
      if (LOG.isLoggable(Level.FINE))
      {
         StringBuilder buf = new StringBuilder();
         buf.append("LOCAL DMO(s):  ");
         buf.append(list);
         
         LOG.fine(buf.toString());
      }
      
      return list;
   }
   
   /**
    * Get the descriptor object which contains information about the relation
    * associated with this resolver.
    *
    * @return  Relation descriptor.
    */
   RelationInfo getInfo()
   {
      return info;
   }
   
   /**
    * Get the current values of all legacy key properties for this relation,
    * from the given record.  These will become query substitution values
    * when searching for the foreign record.
    *
    * @param   dmo
    *          Record currently stored in the record buffer.
    *
    * @return  Array of current, legacy key property values.
    */
   private Object[] getArguments(Persistable dmo)
   {
      Object[] args = new Object[fieldRefs.size()];
      
      Iterator iter = fieldRefs.iterator();
      for (int i = 0; iter.hasNext(); i++)
      {
         FieldReference next = (FieldReference) iter.next();
         args[i] = next.get(dmo);
      }
      
      return args;
   }
   
   /**
    * Generate the HQL statement which will be used for record resolution,
    * using the information stored in the relation descriptor.
    *
    * @param   info
    *          Relation descriptor.
    *
    * @return  HQL query statement.
    */
   private String generateHQL(RelationInfo info)
   {
      String alias = (inverse
                      ? info.getLocalAlias()
                      : info.getForeignAlias());
      HQLExpression buf = new HQLExpression("select ");
      buf.append(alias);
      buf.append(".id from ");
      buf.append(inverse
                 ? info.getLocalClassName()
                 : info.getForeignClassName());
      buf.append(" as ");
      buf.append(alias);
      buf.append(" where ");
      
      Iterator iter = (inverse
                       ? info.localProperties()
                       : info.foreignProperties());
      for (int i = 0; iter.hasNext(); i++)
      {
         if (i > 0)
         {
            buf.append(" and ");
         }
         
         String prop = (String) iter.next();
         
         // TODO:  support computed columns
         if (info.isIgnoreCase(prop))
         {
            buf.append("upper(rtrim(");
            buf.append(alias);
            buf.append(".");
            buf.append(prop);
            buf.append(")) = upper(rtrim(", true);
            buf.append("))");
         }
         else
         {
            buf.append(alias);
            buf.append(".");
            buf.append(prop);
            buf.append(" = ", true);
         }
      }
      
      return buf.toFinalExpression();
   }
   
   /**
    * Construct one {@link FieldReference} object for each legacy key
    * property which will be used as a query substitution parameter during
    * record resolution.  Store these objects in a list for later access.
    *
    * @param   dmoIface
    *          Business interface which the field references will invoke to
    *          retrieve property values.
    * @param   propIter
    *          Iterator on the names of those DMO properties which form the
    *          legacy foreign key.
    */
   private void createFieldRefs(Class dmoIface, Iterator propIter)
   {
      while (propIter.hasNext())
      {
         String prop = (String) propIter.next();
         fieldRefs.add(new FieldReference(database, dmoIface, prop));
      }
      
      ((ArrayList) fieldRefs).trimToSize();
   }
}