/*
** Module   : OutputTableHandleCopier.java
** Abstract : Finalizable for copying data in OUTPUT direction in table-handle parameter case.
**
** Copyright (c) 2013-2017, Golden Code Development Corporation.
**
** -#- -I- --Date-- -----------------------Description-------------------------------------------
** 001 SVL 20140221 Created initial version.
** 002 SVL 20140616 Fixed output copying for remote parameters.
** 003 SVL 20140709 Reflect changed signature of copyAllRows, Set TableWrapper.tableHandle flag
*                   for output TABLE-HANDLE parameters.
*/
/*
** This program is free software: you can redistribute it and/or modify
** it under the terms of the GNU Affero General Public License as
** published by the Free Software Foundation, either version 3 of the
** License, or (at your option) any later version.
**
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
** GNU Affero General Public License for more details.
**
** You may find a copy of the GNU Affero GPL version 3 at the following
** location: https://www.gnu.org/licenses/agpl-3.0.en.html
** 
** Additional terms under GNU Affero GPL version 3 section 7:
** 
**   Under Section 7 of the GNU Affero GPL version 3, the following additional
**   terms apply to the works covered under the License.  These additional terms
**   are non-permissive additional terms allowed under Section 7 of the GNU
**   Affero GPL version 3 and may not be removed by you.
** 
**   0. Attribution Requirement.
** 
**     You must preserve all legal notices or author attributions in the covered
**     work or Appropriate Legal Notices displayed by works containing the covered
**     work.  You may not remove from the covered work any author or developer
**     credit already included within the covered work.
** 
**   1. No License To Use Trademarks.
** 
**     This license does not grant any license or rights to use the trademarks
**     Golden Code, FWD, any Golden Code or FWD logo, or any other trademarks
**     of Golden Code Development Corporation. You are not authorized to use the
**     name Golden Code, FWD, or the names of any author or contributor, for
**     publicity purposes without written authorization.
** 
**   2. No Misrepresentation of Affiliation.
** 
**     You may not represent yourself as Golden Code Development Corporation or FWD.
** 
**     You may not represent yourself for publicity purposes as associated with
**     Golden Code Development Corporation, FWD, or any author or contributor to
**     the covered work, without written authorization.
** 
**   3. No Misrepresentation of Source or Origin.
** 
**     You may not represent the covered work as solely your work.  All modified
**     versions of the covered work must be marked in a reasonable way to make it
**     clear that the modified work is not originating from Golden Code Development
**     Corporation or FWD.  All modified versions must contain the notices of
**     attribution required in this license.
*/
package com.goldencode.p2j.persist;

import com.goldencode.p2j.util.*;
import java.util.*;

/**
 * Class which is registered as a top-level finalizable and copies data in OUTPUT direction
 * in table-handle parameter case.
 */
public class OutputTableHandleCopier
implements Finalizable
{
   /** Calling-side table parameter. */
   private final TableParameter calling;

   /** Called-side table handle. */
   private final handle called;

   /** Table postponed for deletion. */
   private TempTable tableToDelete = null;

   /**
    * Map of all output copiers for the procedure/function for which this copier is registered, to
    * their 0-based indexes in the set of all TABLE-HANDLE parameters for the procedure/function
    * (excluding other types of parameters).
    */
   private Map<OutputTableHandleCopier, Integer> siblings;

   /**
    * Constructor.
    *
    * @param calling
    *        Calling-side table parameter.
    * @param called
    *        Called-side table handle.
    */
   public OutputTableHandleCopier(TableParameter calling, handle called)
   {
      this.calling = calling;
      this.called = called;
   }

   /** Copies data in OUTPUT direction in table-handle parameter case.*/
   @Override
   public void finished()
   {
      boolean res = true;
      if (calling.getTableHandle()._isValid())
      {
         if (called._isValid())
         {
            // append data to existing table
            Temporary calledDMO = TemporaryBuffer.getDefaultBuffer(called);
            res = TemporaryBuffer.copyAllRows(calledDMO,
                                              calling.getTable(),
                                              calling.isAppend(),
                                              TemporaryBuffer.CopyTableMode.OUTPUT_TABLE_HANDLE_PARAM_MODE);
         }
         else if (!calling.isAppend())
         {
            Temporary callingBuffer = TemporaryBuffer.getDefaultBuffer(calling.getTableHandle());
            ((BufferImpl) callingBuffer).deleteAll();
         }
      }
      else if (called._isValid())
      {
         if (!calling.isRemoteParameter())
         {
            // create new table
            TempTableBuilder builder = new TempTableBuilder();
            res = builder.copyTempTable(called,
                                        calling.isAppend(),
                                        false,
                                        false,
                                        "").booleanValue();
            if (res)
               calling.getTableHandle().assign(builder);
         }
         else
         {
            // fill result set
            Temporary calledDMO = TemporaryBuffer.getDefaultBuffer(called);
            TableWrapper oldWrapper = calling.getResultSet();
            try
            {
               TempTableResultSet rs = new TempTableResultSet(calledDMO,
                                                              oldWrapper.isInput(),
                                                              oldWrapper.isOutput(),
                                                              oldWrapper.isAppend());
               TableWrapper wrapper = new TableWrapper(rs);
               RecordBuffer calledBuffer = RecordBuffer.get(calledDMO);
               wrapper.setTableName(calledBuffer.getParentTable().name().getValue());
               wrapper.setTableHandle(true);
               calling.setResultSet(wrapper);
            }
            catch (ClassNotFoundException e)
            {
               throw new IllegalStateException(
                     "Could not create result set for DMO " + calledDMO.getClass().getName(), e);
            }
         }
      }

      if (!res)
         TemporaryBuffer.deferCopyError();

      if (tableToDelete != null &&
          (tableToDelete.equals(called.getResource()) ||
           (!called._isValid() && (siblings.size() == 1 || siblings.get(this) != 0) )))
      {
         // If we have:
         // 1. table postponed for delete AND
         // 2. corresponding TABLE-HANDLE value hasn't changed OR ...
         // 3. TABLE-HANDLE value is unknown and it is the single TABLE-HANDLE parameter OR ...
         // 4. TABLE-HANDLE value is unknown, there are multiple TABLE-HANDLE parameter and it
         //    is not the first one.
         BufferManager.get().unregisterOutputTableHandleCopier(this);
         ((TempTableBuilder) tableToDelete).forceDelete();
      }
   }

   /** No-op. */
   @Override
   public void deleted()
   {
   }

   /** No-op. */
   @Override
   public void iterate()
   {
   }

   /** No-op. */
   @Override
   public void retry()
   {
   }

   /**
    * Set table for postponed deletion.
    *
    * @param tableToDelete
    *        Table for postponed deletion.
    */
   public void setTableToDelete(TempTable tableToDelete)
   {
      this.tableToDelete = tableToDelete;
   }

   /**
    * Get called-side table handle.
    *
    * @return called-side table handle.
    */
   public handle getCalled()
   {
      return called;
   }

   /**
    * Set the map of all output copiers for the procedure/function for which this copier is
    * registered, to their 0-based indexes in the set of all TABLE-HANDLE parameters for the
    * procedure/function (excluding other types of parameters).
    *
    * @param siblings
    *        See above.
    */
   public void setSiblingsMap(Map<OutputTableHandleCopier, Integer> siblings)
   {
      this.siblings = siblings;
   }
}

