/*
** Module   : QueryOffEndException.java
** Abstract : exception thrown when certain query instances run off end
**
** Copyright (c) 2006-2017, Golden Code Development Corporation.
**
** -#- -I- --Date-- --JPRM-- --------------------------Description-------------------------------
** 001 ECF 20060531  @26847  Created initial version. Thrown to indicate a
**                           query has run off the end of its result set.
** 002 CA  20090727  @43394  All listeners interested in this off-end 
**                           exception must be passed to the constructor.
** 003 ECF 20150421          Removed unused constructors and implemented fillInStackTrace to do
**                           nothing; we do not use the stack trace information and profiling has
**                           indicated that we pay a noticeable performance penalty during
**                           construction.
*/
/*
** This program is free software: you can redistribute it and/or modify
** it under the terms of the GNU Affero General Public License as
** published by the Free Software Foundation, either version 3 of the
** License, or (at your option) any later version.
**
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
** GNU Affero General Public License for more details.
**
** You may find a copy of the GNU Affero GPL version 3 at the following
** location: https://www.gnu.org/licenses/agpl-3.0.en.html
** 
** Additional terms under GNU Affero GPL version 3 section 7:
** 
**   Under Section 7 of the GNU Affero GPL version 3, the following additional
**   terms apply to the works covered under the License.  These additional terms
**   are non-permissive additional terms allowed under Section 7 of the GNU
**   Affero GPL version 3 and may not be removed by you.
** 
**   0. Attribution Requirement.
** 
**     You must preserve all legal notices or author attributions in the covered
**     work or Appropriate Legal Notices displayed by works containing the covered
**     work.  You may not remove from the covered work any author or developer
**     credit already included within the covered work.
** 
**   1. No License To Use Trademarks.
** 
**     This license does not grant any license or rights to use the trademarks
**     Golden Code, FWD, any Golden Code or FWD logo, or any other trademarks
**     of Golden Code Development Corporation. You are not authorized to use the
**     name Golden Code, FWD, or the names of any author or contributor, for
**     publicity purposes without written authorization.
** 
**   2. No Misrepresentation of Affiliation.
** 
**     You may not represent yourself as Golden Code Development Corporation or FWD.
** 
**     You may not represent yourself for publicity purposes as associated with
**     Golden Code Development Corporation, FWD, or any author or contributor to
**     the covered work, without written authorization.
** 
**   3. No Misrepresentation of Source or Origin.
** 
**     You may not represent the covered work as solely your work.  All modified
**     versions of the covered work must be marked in a reasonable way to make it
**     clear that the modified work is not originating from Golden Code Development
**     Corporation or FWD.  All modified versions must contain the notices of
**     attribution required in this license.
*/

package com.goldencode.p2j.persist;

import com.goldencode.p2j.util.*;

/**
 * Exception thrown when certain query instances run off end.
 */
public class QueryOffEndException
extends EndConditionException
{
   /** A list of listeners interested in this off-end exception. */
   private final QueryOffEndListener[] listeners;
   
   /**
    * Constructor which accepts an error message.
    *
    * @param   message
    *          Error message.
    * @param   listeners
    *          The listeners interested in this off-end exception
    */
   public QueryOffEndException(String message, QueryOffEndListener[] listeners)
   {
      super(message);
      
      this.listeners = listeners;
   }
   
   /**
    * Override the ancestor class' implementation to do nothing. We do not actually use the stack
    * trace information when we catch this exception (it simply terminates an iterating query
    * loop), so there is no need to pay the performance penalty of collecting the data, which is
    * expensive.
    */
   @Override
   public final Throwable fillInStackTrace()
   {
      return this;
   }
   
   /**
    * Get the listeners interested in this off-end exception
    * 
    * @return   the listeners interested in this off-end exception
    */
   public QueryOffEndListener[] getListeners()
   {
      return listeners;
   }
}