/*
** Module   : H2Helper.java
** Abstract : Helper class which initializes embedded H2 database
**
** Copyright (c) 2004-2018, Golden Code Development Corporation.
**
** -#- -I- --Date-- --JPRM-- -----------------------------Description-----------------------------
** 001 ECF 20080311   @37516 Created initial version. Helper class which
**                           initializes embedded H2 database.
** 002 ECF 20080320   @37649 Added new steps to database preparation. Set
**                           sorting of nulls to high (NULL values last).
**                           Set collation.
** 003 SVL 20080429   @38135 Collation is loaded from the directory.
** 004 ECF 20080508   @38487 Changed createFunctionAliases(). Use Database
**                           object when registering functions with the
**                           HQLPreprocessor rather than database name
**                           String. This allows differentiation of the
**                           same function across multiple databases with
**                           the same name.
** 005 ECF 20080724   @39169 Fixed bugs in executeDDL(). Connection and
**                           SessionFactory must not be closed after use.
** 006 ECF 20090729   @43451 Fixed logging in executeDDL(). It was possible to
**                           trigger NPE in BatchUpdateException catch block.
** 007 CA  20101029          Added method to retrieve a list of database 
**                           prepare statements.
** 009 CA  20101202          Added method to prepare a permanent database, as
**                           the custom functions need to  be registered with 
**                           the HQL preprocessor. Removed redundant code 
**                           (collation was set more than 1 time).
**                           The "h2.sortNullsHigh" setting was removed, as
**                           it wasn't taken into consideration anyway; also,
**                           more testing needs to be done to exactly 
**                           determine how 4GL sorts the unknown value in the
**                           temporary and permanent database.
** 010 SVL 20130331          Upgraded to Hibernate 4.
** 011 ECF 20131001          Added metadata support and getH2Version method.
** 012 OM  20140728          Added support for HQLFunction annotation. Moved to dialect package.
** 013 EVL 20160225          Javadoc fixes to make compatible with Oracle Java 8 for Solaris 10.
** 014 OM  20160608          Fixed support for HQLFunction annotation.
** 015 OM  20160905          Small optimizations that got lost in previous commit.
** 016 ECF 20160912          Added setCommonInMemoryProperties. Established en_US_P2J as the
**                           default H2 collation.
** 017 ECF 20180816          Upgrade to H2 1.4.197; needed to add mvcc=false to JDBC URL.
*/
/*
** This program is free software: you can redistribute it and/or modify
** it under the terms of the GNU Affero General Public License as
** published by the Free Software Foundation, either version 3 of the
** License, or (at your option) any later version.
**
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
** GNU Affero General Public License for more details.
**
** You may find a copy of the GNU Affero GPL version 3 at the following
** location: https://www.gnu.org/licenses/agpl-3.0.en.html
** 
** Additional terms under GNU Affero GPL version 3 section 7:
** 
**   Under Section 7 of the GNU Affero GPL version 3, the following additional
**   terms apply to the works covered under the License.  These additional terms
**   are non-permissive additional terms allowed under Section 7 of the GNU
**   Affero GPL version 3 and may not be removed by you.
** 
**   0. Attribution Requirement.
** 
**     You must preserve all legal notices or author attributions in the covered
**     work or Appropriate Legal Notices displayed by works containing the covered
**     work.  You may not remove from the covered work any author or developer
**     credit already included within the covered work.
** 
**   1. No License To Use Trademarks.
** 
**     This license does not grant any license or rights to use the trademarks
**     Golden Code, FWD, any Golden Code or FWD logo, or any other trademarks
**     of Golden Code Development Corporation. You are not authorized to use the
**     name Golden Code, FWD, or the names of any author or contributor, for
**     publicity purposes without written authorization.
** 
**   2. No Misrepresentation of Affiliation.
** 
**     You may not represent yourself as Golden Code Development Corporation or FWD.
** 
**     You may not represent yourself for publicity purposes as associated with
**     Golden Code Development Corporation, FWD, or any author or contributor to
**     the covered work, without written authorization.
** 
**   3. No Misrepresentation of Source or Origin.
** 
**     You may not represent the covered work as solely your work.  All modified
**     versions of the covered work must be marked in a reasonable way to make it
**     clear that the modified work is not originating from Golden Code Development
**     Corporation or FWD.  All modified versions must contain the notices of
**     attribution required in this license.
*/

package com.goldencode.p2j.persist.dialect;

import java.io.*;
import java.lang.reflect.*;
import java.sql.*;
import java.util.*;
import java.util.logging.*;
import org.hibernate.cfg.*;
import org.hibernate.internal.*;
import org.h2.engine.Constants;
import org.h2.tools.*;
import org.h2.tools.Server;
import org.hibernate.*;
import com.goldencode.p2j.persist.*;
import com.goldencode.p2j.persist.pl.*;
import com.goldencode.p2j.util.*;
import com.goldencode.p2j.util.LogHelper;
import com.goldencode.util.*;

/**
 * Helper class which provides H2-specific services, such as initializing an embedded H2 database,
 * running scripts against it, and running a server to provide mixed mode access to it.
 */
public final class H2Helper
{
   /** Logger */
   private static final Logger log = LogHelper.getLogger(H2Helper.class.getName());
   
   /** Default embedded collation */
   private static final String DEFAULT_COLLATION = "en_US_P2J";
   
   /**
    * Private constructor.  All access to this class is via static methods.
    */
   private H2Helper()
   {
   }
   
   /**
    * Report the full H2 database version.
    */
   public static String getH2Version()
   {
      return Constants.getFullVersion();
   }
   
   /**
    * Set the properties commonly needed by an in-memory, embedded H2 database into the given
    * properties object. This includes the JDBC connection URL, JDBC driver, randomly generated
    * user name and password, and several Hibernate options.
    * 
    * @param   database
    *          Database name to be embedded in the connection URL.
    * @param   props
    *          Object into which to store properties.
    * 
    * @return  The populated properties object (same object instance as {@code props}).
    */
   public static Properties setCommonInMemoryProperties(String database, Properties props)
   {
      StringBuilder url = new StringBuilder();
      url.append("jdbc:h2:mem:");
      url.append(database);
      url.append(";db_close_delay=-1;multi_threaded=1;mv_store=false;mvcc=false;");
      
      props.put(AvailableSettings.URL, url.toString());
      
      props.put(AvailableSettings.USER, RandomWordGenerator.create(128, 16, 16));
      props.put(AvailableSettings.PASS, RandomWordGenerator.create(128, 16, 16));
      
      props.put(AvailableSettings.USE_SECOND_LEVEL_CACHE, "false");
      props.put(AvailableSettings.USE_QUERY_CACHE, "false");
      props.put(AvailableSettings.RELEASE_CONNECTIONS, "on_close");
      
      props.put(AvailableSettings.DRIVER, "org.h2.Driver");
      
      props.put(AvailableSettings.DIALECT, "com.goldencode.p2j.persist.dialect.P2JH2Dialect");
      
      return props;
   }
   
   /**
    * Prepare the specified database by creating all necessary function
    * aliases and setting global parameters.
    * <p>
    * The "h2.sortNullsHigh" setting was removed, as the H2 driver was already
    * initiated at the time when this method is called, so the setting has no
    * effect.
    * 
    * @param   database
    *          Database to prepare.
    *          
    * @throws  PersistenceException
    *          if there is any error executing DDL.
    */
   public static void prepareDatabase(Database database)
   throws PersistenceException
   {
      // create built-in function aliases
      createFunctionAliases(database, BuiltIns.getPublicFunctions(), true);
      createFunctionAliases(database, BuiltIns.getSupportFunctions(), false);
      
      // set collation
      String path = String.format("database/%s/p2j/embedded-collation", database.getName());
      
      String collation = Utils.getDirectoryNodeString(null, path, DEFAULT_COLLATION, false);
      if (collation != null)
      {
         try
         {
            List<String> ddl = new ArrayList<>(1);
            ddl.add(String.format("set collation \"%s\"", collation));
            executeDDL(ddl, database);
         }
         catch (PersistenceException exc)
         {
            log.log(
               Level.SEVERE,
               "Error setting P2J-specific collation;  please ensure " +
               "p2jspi.jar is installed in Java extension directory");
            
            throw exc;
         }
      }
      else if (log.isLoggable(Level.WARNING))
      {
         log.log(Level.WARNING,
                 "Embedded " + (database.isDirty() ? "dirty" : "primary") +
                 " database '" + database.getName() +
                 "' collation not set;  using default Java collation");
      }
   }
   
   /**
    * Register the function aliases with the HQL preprocessor, for the 
    * permanent database. 
    * 
    * @param   database
    *          Database to prepare.
    */
   public static void preparePermanentDatabase(Database database)
   {
      getFunctionAliases(database, BuiltIns.getPublicFunctions(), true, true);
      getFunctionAliases(database, BuiltIns.getSupportFunctions(), false, true);
   }
   
   /**
    * Get a list of database prepare DDL statements.
    * 
    * @param   database
    *          Database to prepare.
    *          
    * @return  the list of DDL statements.
    */
   public static List<String> getPrepareStatements(Database database)
   {
      List<String> ddl = new ArrayList<>();
      
      // Create built-in function aliases.
      ddl.addAll(getFunctionAliases(database, BuiltIns.getPublicFunctions(), true, false));
      ddl.addAll(getFunctionAliases(database, BuiltIns.getSupportFunctions(), false, false));
      
      return ddl;
   }
   
   /**
    * Run a script of SQL statements using the specified database connection.
    *
    * @param   connection
    *          Connection to target database; caller should clean up the connection.
    * @param   reader
    *          Reader from which the script will be read.
    *
    * @throws  PersistenceException
    *          if there is any error running the script.
    */
   public static void runScript(Connection connection, Reader reader)
   throws PersistenceException
   {
      try
      {
         RunScript.execute(connection, reader);
         connection.commit();
      }
      catch (SQLException exc)
      {
         try
         {
            connection.rollback();
         }
         catch (SQLException sqle)
         {
         }
         
         throw new PersistenceException("Error running script against embedded database", exc);
      }
   }
   
   /**
    * Start a mixed mode TCP server using the specified arguments, which are passed directly to
    * the H2 server.
    *
    * @param   args
    *          Arguments to be passed to the H2 server.
    *
    * @throws  PersistenceException
    *          if there is an error launching the server.
    */
   public static void startMixedModeServer(String... args)
   throws PersistenceException
   {
      try
      {
         Server tcpSrv = Server.createTcpServer(args);
         tcpSrv.start();
      }
      catch (SQLException exc)
      {
         throw new PersistenceException("Error starting up mixed mode database server", exc);
      }
   }
   
   /**
    * Get function aliases for the given list of static methods, optionally
    * making the alias names unique by appending a unique number for each
    * overloaded method.
    * 
    * @param   database
    *          Database in which to create aliases.
    * @param   funcs
    *          List of static methods for which function aliases should be
    *          created.
    * @param   overload
    *          <code>true</code> to create unique function names for each
    *          overloaded method
    * @param   register
    *          <code>true</code> if the function should be registered with
    *          the {@link HQLPreprocessor}.
    */
   private static List<String> getFunctionAliases(Database database,
                                                  List<Method> funcs,
                                                  boolean overload,
                                                  boolean register)
   {
      List<String> ddl = new ArrayList<>();
      Map<String, Integer> overloads = new HashMap<>();
      for (Method method : funcs)
      {
         HQLFunction hqlFn = method.getAnnotation(HQLFunction.class);
         if (hqlFn == null)
         {
            // not a support function
            continue;
         }
         
         String name = method.getName();
         String hqlFnName = hqlFn.name();
         String hqlName = hqlFnName.isEmpty() ? name : hqlFnName;
         
         Integer count = null;
         if (overload)
         {
            count = overloads.get(name);
            if (count == null)
            {
               count = 1;
            }
            else
            {
               count++;
            }
            overloads.put(name, count);
         }
         
         StringBuilder buf = new StringBuilder();
         
         buf.append(hqlName);
         
         if (overload)
         {
            buf.append('_');
            buf.append(count);
         }
         String uniqueName = buf.toString();
         
         buf.insert(0, "create alias if not exists ");
         buf.append(" for ");
         buf.append(formatMethod(method));
         buf.append(";");
         
         ddl.add(buf.toString());
         
         if (overload && register)
         {
            HQLPreprocessor.registerFunction(database, hqlName, uniqueName, method);
         }
      }
      
      return ddl;
   }
   
   /**
    * Create function aliases for the given list of static methods, optionally
    * making the alias names unique by appending a unique number for each
    * overloaded method.
    * 
    * @param   database
    *          Database in which to create aliases.
    * @param   funcs
    *          List of static methods for which function aliases should be
    *          created.
    * @param   overload
    *          <code>true</code> to create unique function names for each
    *          overloaded method
    * 
    * @throws  PersistenceException
    *          if there is any error executing DDL.
    */
   private static void createFunctionAliases(Database database,
                                             List<Method> funcs,
                                             boolean overload)
   throws PersistenceException
   {
      List<String> ddl = getFunctionAliases(database, funcs, overload, true);
      
      executeDDL(ddl, database);
   }
   
   /**
    * Generate a quoted method name and signature, suitable for use within a
    * CREATE ALIAS DDL statement.
    * 
    * @param   method
    *          Method which is to be mapped to a function alias.
    * 
    * @return  Quoted method name and signature.
    */
   private static String formatMethod(Method method)
   {
      StringBuilder buf = new StringBuilder();
      buf.append("\"");
      buf.append(method.getDeclaringClass().getName());
      buf.append('.');
      buf.append(method.getName());
      buf.append('(');
      Class<?>[] sig = method.getParameterTypes();
      for (int i = 0; i < sig.length; i++)
      {
         if (i > 0)
         {
            buf.append(", ");
         }
         
         buf.append(sig[i].getName());
      }
      buf.append(')');
      buf.append("\"");
      
      return buf.toString();
   }
   
   /**
    * Execute a batch of DDL statements for the given database.  The batch is
    * executed within a single transaction.
    * 
    * @param ddl
    * @param database
    * @throws PersistenceException
    */
   private static void executeDDL(List<String> ddl, Database database)
   throws PersistenceException
   {
      SessionFactory factory = DatabaseManager.getSessionFactory(database);
      Session session = factory.openSession();
      Connection conn = ((SessionImpl) session).connection();
      Statement statement = null;
      try
      {
         try
         {
            boolean debug = log.isLoggable(Level.FINEST);
            conn.setAutoCommit(false);
            statement = conn.createStatement();
            for (String next : ddl)
            {
               statement.addBatch(next);
               if (debug)
               {
                  log.log(Level.FINEST, "DDL:  " + next);
               }
            }
            
            statement.executeBatch();
            conn.commit();
         }
         catch (BatchUpdateException exc)
         {
            conn.rollback();
            
            SQLException next = exc.getNextException();
            
            throw new PersistenceException(
               "Error initializing embedded database",
               (next != null ? next : exc));
         }
         catch (SQLException exc)
         {
            conn.rollback();
            
            throw new PersistenceException(
               "Error initializing embedded database", exc);
         }
         finally
         {
            if (statement != null)
            {
               statement.close();
            }
            
            if (session != null && session.isOpen())
            {
               session.close();
            }
         }
      }
      catch (SQLException exc)
      {
      }
   }
}