/*
** Module   : StoredProcKeyRetriever.java
** Abstract : KeyRetriever implementation which retrieves keys using stored
**            procedure.
**
** Copyright (c) 2005-2017, Golden Code Development Corporation.
**
** -#- -I- --Date-- --JPRM-- ----------------Description-----------------
** 001 SVL 20080827   @39782 Created initial version. Retrieves free keys
**                           using a stored procedure.
** 002 SVL 20120331          Upgraded to Hibernate 4.
*/
/*
** This program is free software: you can redistribute it and/or modify
** it under the terms of the GNU Affero General Public License as
** published by the Free Software Foundation, either version 3 of the
** License, or (at your option) any later version.
**
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
** GNU Affero General Public License for more details.
**
** You may find a copy of the GNU Affero GPL version 3 at the following
** location: https://www.gnu.org/licenses/agpl-3.0.en.html
** 
** Additional terms under GNU Affero GPL version 3 section 7:
** 
**   Under Section 7 of the GNU Affero GPL version 3, the following additional
**   terms apply to the works covered under the License.  These additional terms
**   are non-permissive additional terms allowed under Section 7 of the GNU
**   Affero GPL version 3 and may not be removed by you.
** 
**   0. Attribution Requirement.
** 
**     You must preserve all legal notices or author attributions in the covered
**     work or Appropriate Legal Notices displayed by works containing the covered
**     work.  You may not remove from the covered work any author or developer
**     credit already included within the covered work.
** 
**   1. No License To Use Trademarks.
** 
**     This license does not grant any license or rights to use the trademarks
**     Golden Code, FWD, any Golden Code or FWD logo, or any other trademarks
**     of Golden Code Development Corporation. You are not authorized to use the
**     name Golden Code, FWD, or the names of any author or contributor, for
**     publicity purposes without written authorization.
** 
**   2. No Misrepresentation of Affiliation.
** 
**     You may not represent yourself as Golden Code Development Corporation or FWD.
** 
**     You may not represent yourself for publicity purposes as associated with
**     Golden Code Development Corporation, FWD, or any author or contributor to
**     the covered work, without written authorization.
** 
**   3. No Misrepresentation of Source or Origin.
** 
**     You may not represent the covered work as solely your work.  All modified
**     versions of the covered work must be marked in a reasonable way to make it
**     clear that the modified work is not originating from Golden Code Development
**     Corporation or FWD.  All modified versions must contain the notices of
**     attribution required in this license.
*/

package com.goldencode.p2j.persist.id;

import com.goldencode.p2j.persist.Database;
import java.util.*;
import org.hibernate.*;
import org.hibernate.type.*;

/**
 * Provides a service for retrieving available primary keys from database
 * by calling stored procedure.
 */
public class StoredProcKeyRetriever
implements KeyRetriever
{
   /** Query which is used for stored procedure call. */
   private String retrieveQuery;

   /**
    * Name of the stored procedure used for id retrieval.
    * Procedure should be declared as 'FUNCTION procedure_name(min_id bigint, 
    * max_id bigint) RETURNS SETOF range_type'
    */
   private static String STORED_PRUCEDURE_NAME = "get_free_ids";

   /**
    * Initialize key retriever for the given database.
    *
    * @param database
    *        Database for which will be scanned for available keys.
    */
   public void initialize(Database database)
   {
      StringBuilder buf = new StringBuilder();
      buf.append("select lower_bound, upper_bound from ");
      buf.append(STORED_PRUCEDURE_NAME);
      buf.append("(:minId, :maxId)");

      retrieveQuery = buf.toString();
   }

   /**
    * Get available keys.
    *
    * @param  minKey
    *         Lower bound of a range of identifiers on which search will be
    *         made.
    * @param  maxKey
    *         Upper bound of a range of identifiers on which search will be
    *         made.
    * @param  session
    *         Hibernate session which will be used to perform queries.
    * @return A list of ranges of available keys
    */
   public List<Range> retrieveFreeKeys(Long minKey,
                                       Long maxKey,
                                       Session session)
   {
      // Execute the query which will return all ranges of free ids in the
      // given global range. 
      SQLQuery query = session.createSQLQuery(retrieveQuery);
      query.addScalar("lower_bound", StandardBasicTypes.LONG);
      query.addScalar("upper_bound", StandardBasicTypes.LONG);
      query.setLong("minId", minKey);
      query.setLong("maxId", maxKey);

      List<Range> res = new ArrayList<Range>();
      List<Object[]> bounds = (List<Object[]>) query.list();
      for (Object[] bound : bounds)
      {
         res.add(new Range((Long)bound[0], (Long)bound[1]));
      }

      return res;
   }
}