/*
** Module   : TransactionTableUpdater.java
** Abstract : Manages metadata transaction records for a particular, primary database
**
** Copyright (c) 2017, Golden Code Development Corporation.
**
** -#- -I- --Date-- --------------------------------Description----------------------------------
** 001 ECF 20171030 Created initial version.
*/
/*
** This program is free software: you can redistribute it and/or modify
** it under the terms of the GNU Affero General Public License as
** published by the Free Software Foundation, either version 3 of the
** License, or (at your option) any later version.
**
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
** GNU Affero General Public License for more details.
**
** You may find a copy of the GNU Affero GPL version 3 at the following
** location: https://www.gnu.org/licenses/agpl-3.0.en.html
** 
** Additional terms under GNU Affero GPL version 3 section 7:
** 
**   Under Section 7 of the GNU Affero GPL version 3, the following additional
**   terms apply to the works covered under the License.  These additional terms
**   are non-permissive additional terms allowed under Section 7 of the GNU
**   Affero GPL version 3 and may not be removed by you.
** 
**   0. Attribution Requirement.
** 
**     You must preserve all legal notices or author attributions in the covered
**     work or Appropriate Legal Notices displayed by works containing the covered
**     work.  You may not remove from the covered work any author or developer
**     credit already included within the covered work.
** 
**   1. No License To Use Trademarks.
** 
**     This license does not grant any license or rights to use the trademarks
**     Golden Code, FWD, any Golden Code or FWD logo, or any other trademarks
**     of Golden Code Development Corporation. You are not authorized to use the
**     name Golden Code, FWD, or the names of any author or contributor, for
**     publicity purposes without written authorization.
** 
**   2. No Misrepresentation of Affiliation.
** 
**     You may not represent yourself as Golden Code Development Corporation or FWD.
** 
**     You may not represent yourself for publicity purposes as associated with
**     Golden Code Development Corporation, FWD, or any author or contributor to
**     the covered work, without written authorization.
** 
**   3. No Misrepresentation of Source or Origin.
** 
**     You may not represent the covered work as solely your work.  All modified
**     versions of the covered work must be marked in a reasonable way to make it
**     clear that the modified work is not originating from Golden Code Development
**     Corporation or FWD.  All modified versions must contain the notices of
**     attribution required in this license.
*/

package com.goldencode.p2j.persist.meta;

import java.lang.reflect.*;
import java.util.*;
import java.util.logging.*;
import org.hibernate.type.Type;
import com.goldencode.p2j.persist.*;
import com.goldencode.p2j.security.SecurityManager;
import com.goldencode.p2j.util.*;
import com.goldencode.proxy.*;

/**
 * A {@link com.goldencode.p2j.persist.meta.TransactionListener transaction listener
 * implementation} which receives notifications of transaction events and updates the metadata
 * transaction table accordingly. This involves the creation, lookup, and deletion of metadata
 * transaction records. Currently, this work is done inline. If this becomes too expensive, it
 * will need to be offloaded to a worker thread).
 */
public final class TransactionTableUpdater
implements TransactionListener
{
   /** Logger */
   private static final Logger log = LogHelper.getLogger(TransactionTableUpdater.class);
   
   /** Metadata transaction table name */
   private static final String TRANS_TABLE = "meta_trans";
   
   /** HQL query to find a trans record by transaction ID */
   private static final String HQL_TRANS = "from MetaTransImpl m where m.transId = ?0";
   
   /** Parameter types for transaction record query */
   private static final Type[] TYPES_TRANS = DBUtils.makeTypeArray(new Object[] { new int64() });
   
   /** MetaTransImpl DMO implementation class */
   private static final Class<?> dmoClass;
   
   /** Map of <code>MinimalTrans</code> methods to <code>MetaTransImpl</code> methods */
   private static final Map<Method, Method> methodMap = new HashMap<>();
   
   static
   {
      String clsName = String.format("%s._meta.impl.MetaTransImpl", DMOIndex.getBasePackage());
      try
      {
         dmoClass = Class.forName(clsName);
         
         // map MinimalTrans methods to corresponding dmoClass methods; if we try to use
         // MinimalTrans methods with a MetaTransImpl object, we will raise
         // IllegalArgumentException during Method.invoke(), because our object will not be an
         // instance of MinimalTrans
         for (Method ifcMeth : MinimalTrans.class.getMethods())
         {
            Method dmoMeth = dmoClass.getMethod(ifcMeth.getName(),
                                                ifcMeth.getParameterTypes());
            methodMap.put(ifcMeth, dmoMeth);
         }
      }
      catch (ClassNotFoundException | NoSuchMethodException exc)
      {
         throw new RuntimeException("Error preparing MetaTrans implementation class", exc);
      }
   }
   
   /** Next available transaction number */
   private static int nextTransNum = 0;
   
   /** Persistence helper object for metadata database */
   private final Persistence persistence;
   
   /** Security manager */
   private final SecurityManager securityManager;
   
   /** Proxy object for the current metadata transaction DMO */
   private final MinimalTrans proxy;
   
   /** Invocation handler which maps proxy method calls to underlying DMO methods */
   private final Handler handler = new Handler();
   
   /** Object on which updates are synchronized */
   private final Object lock = new Object();
   
   /**
    * Constructor which stores a persistence object for the associated metadata database, and
    * initializes a proxy for the application-specific DMO.
    * 
    * @param   metaDB
    *          Metadata database helper object.
    */
   public TransactionTableUpdater(Database metaDB)
   {
      this.persistence = PersistenceFactory.getInstance(metaDB);
      this.securityManager = SecurityManager.getInstance();
      this.proxy = ProxyFactory.getProxy(dmoClass,
                                         new Class<?>[] { MinimalTrans.class },
                                         handler);
   }
   
   /**
    * Create a new trans record in the metadata database, in response to a transaction event.
    * 
    * @param   transID
    *          Unique identifier for the transaction.
    */
   public void transactionStarted(int64 transID)
   {
      // get current session id as the user number
      Integer sid = securityManager.getSessionId();
      
      // character now = new character(datetime.now().toStringMessage());
      
      // fairly coarse synchronization should be OK here; if this becomes a bottleneck, we can
      // replace this with a lock manager
      synchronized (lock)
      {
         Persistable dmo = null;
         boolean commit = false;
         
         try
         {
            commit = persistence.beginTransaction();
            
            // create new DMO instance and back the proxy with it
            dmo = (Persistable) dmoClass.newInstance();
            handler.setDelegate(dmo);
            
            // assign primary key
            Long key = (Long) persistence.nextPrimaryKey(TRANS_TABLE);
            dmo.setId(key);
            
            // assign transaction ID (required by unique index)
            proxy.setTransId(transID);
            
            // assign fields with known values
            proxy.setTransNum(new integer(++nextTransNum));
            proxy.setTransUsrnum(new integer(sid));
            proxy.setTransFlags(new character("FWD"));   // TODO: research possibilities
            proxy.setTransState(new character("BEGIN")); // TODO: research possibilities
            
            // assign other fields which are mandatory and default to unknown
            proxy.setTransCoord(new character("?"));
            proxy.setTransCoordtx(new integer(0));
            proxy.setTransDuration(new integer(0));
            proxy.setTransTxtime(new character("?"));
            proxy.setTransTenantid(new integer(0));
            proxy.setTransCounter(new integer(0));
            proxy.setTransMisc(new integer(0));
            
            // save the new record
            persistence.save(dmo, key);
            
            // flush explicitly
            persistence.flush();
            
            // commit
            if (commit)
            {
               commit = false;   // to avoid double-rollback in catch block if commit fails
               persistence.commit();
            }
         }
         catch (Exception exc)
         {
            try
            {
               if (commit)
               {
                  persistence.rollback();
               }
            }
            catch (PersistenceException pe)
            {
               // error already logged in finally block
            }
            finally
            {
               if (log.isLoggable(Level.SEVERE))
               {
                  String msg = String.format("Error inserting into _trans table for %s (%s)",
                                             persistence.getDatabase().toString(),
                                             transID.toStringMessage());
                  log.log(Level.SEVERE, msg, exc);
               }
            }
         }
      }
   }
   
   /**
    * Remove the trans record with the given ID from the metadata database, in response to a
    * transaction event.
    * 
    * @param   transID
    *          Unique identifier for the transaction.
    */
   public void transactionEnded(int64 transID)
   {
      // fairly coarse synchronization may be OK here; if this becomes a bottleneck, we can
      // replace this with a lock manager
      synchronized (lock)
      {
         Persistable dmo = null;
         boolean commit = false;
         
         try
         {
            commit = persistence.beginTransaction();
            
            // retrieve DMO with the given transaction ID
            Object[] args = new Object[] { transID };
            List<Persistable> list = persistence.list(HQL_TRANS,
                                                      args,
                                                      TYPES_TRANS,
                                                      1,
                                                      0,
                                                      true);
            if (list == null)
            {
               throw new PersistenceException("Unable to retrieve transaction metadata record " +
                                              transID.toStringMessage());
            }
            
            // delete the record
            dmo = list.get(0);
            persistence.delete(dmo);
            
            // flush explicitly
            persistence.flush();
            
            // commit
            if (commit)
            {
               commit = false;   // to avoid double-rollback in catch block if commit fails
               persistence.commit();
            }
         }
         catch (Exception exc)
         {
            try
            {
               if (commit)
               {
                  persistence.rollback();
               }
            }
            catch (PersistenceException pe)
            {
               // error already logged in finally block
            }
            finally
            {
               if (log.isLoggable(Level.SEVERE))
               {
                  String msg = String.format("Error deleting from _trans table for %s (%s)",
                                             persistence.getDatabase().toString(),
                                             transID.toStringMessage());
                  log.log(Level.SEVERE, msg, exc);
               }
            }
         }
      }
   }
   
   /**
    * Invocation handler which delegates calls to the {@link MinimalTrans} proxy methods to the
    * backing DMO instance.
    */
   private static class Handler
   implements InvocationHandler
   {
      /** DMO delegate whose methods ultimately are invoked */
      private Object delegate = null;
      
      /**
       * Process method invocations on the {@link MinimalTrans} proxy by delegating them to the
       * backing DMO instance.
       * 
       * @param   proxy
       *          Proxy on which the enclosing class invokes methods.
       * @param   method
       *          Method which was invoked.
       * @param   args
       *          Method arguments.
       * 
       * @throws  Throwable
       *          if the delegated method call throws an exception.
       */
      @Override
      public Object invoke(Object proxy, Method method, Object[] args)
      throws Throwable
      {
         Method dmoMethod = methodMap.get(method);
         
         return dmoMethod.invoke(delegate, args);
      }
      
      /**
       * Set the backing DMO delegate instance.
       * 
       * @param   delegate
       *          DMO whose methods will be invoked.
       */
      void setDelegate(Object delegate)
      {
         this.delegate = delegate;
      }
   }
   
   /**
    * Interface which defines the minimally required methods of the <code>MetaTrans</code>
    * interface. This interface defines the proxy used to create and manipulate metadata
    * transaction DMO instances. The DMO's methods cannot be invoked directly from code here,
    * since the <code>MetaTrans</code> interface resides in an application-specific package,
    * which is not known at compile time.
    */
   public interface MinimalTrans
   extends Persistable
   {
      /**
       * Setter: transId
       * 
       * @param   transId
       *          transId
       */
      public void setTransId(NumberType transId);
      
      /**
       * Setter: Trans State
       * 
       * @param   transState
       *          Trans State
       */
      public void setTransState(Text transState);
      
      /**
       * Setter: Trans Flags
       * 
       * @param   transFlags
       *          Trans Flags
       */
      public void setTransFlags(Text transFlags);
      
      /**
       * Setter: Trans Usernum
       * 
       * @param   transUsrnum
       *          Trans Usernum
       */
      public void setTransUsrnum(NumberType transUsrnum);
      
      /**
       * Setter: Trans Num
       * 
       * @param   transNum
       *          Trans Num
       */
      public void setTransNum(NumberType transNum);
      
      /**
       * Setter: Trans rl counter
       * 
       * @param   transCounter
       *          Trans rl counter
       */
      public void setTransCounter(NumberType transCounter);
      
      /**
       * Setter: Trans Start Time
       * 
       * @param   transTxtime
       *          Trans Start Time
       */
      public void setTransTxtime(Text transTxtime);
      
      /**
       * Setter: Coord Name
       * 
       * @param   transCoord
       *          Coord Name
       */
      public void setTransCoord(Text transCoord);
      
      /**
       * Setter: Coordinator Transaction
       * 
       * @param   transCoordtx
       *          Coordinator Transaction
       */
      public void setTransCoordtx(NumberType transCoordtx);
      
      /**
       * Setter: Tx Duration (in secs)
       * 
       * @param   transDuration
       *          Tx Duration (in secs)
       */
      public void setTransDuration(NumberType transDuration);
      
      /**
       * Setter: Trans Misc
       * 
       * @param   transMisc
       *          Trans Misc
       */
      public void setTransMisc(NumberType transMisc);
      
      /**
       * Setter: Tenant Id
       * 
       * @param   transTenantid
       *          Tenant Id
       */
      public void setTransTenantid(NumberType transTenantid);
   }
}