/*
** Module   : XmlExport.java
** Abstract : Write temp-table data into an XML stream.
**
** Copyright (c) 2017-2018, Golden Code Development Corporation.
**
** -#- -I- --Date-- ---------------------------------Description---------------------------------
** 001 ECF 20171016 Created initial version.
** 002 CA  20180511 Added XML-NODE-NAME table option support.
*/
/*
** This program is free software: you can redistribute it and/or modify
** it under the terms of the GNU Affero General Public License as
** published by the Free Software Foundation, either version 3 of the
** License, or (at your option) any later version.
**
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
** GNU Affero General Public License for more details.
**
** You may find a copy of the GNU Affero GPL version 3 at the following
** location: https://www.gnu.org/licenses/agpl-3.0.en.html
** 
** Additional terms under GNU Affero GPL version 3 section 7:
** 
**   Under Section 7 of the GNU Affero GPL version 3, the following additional
**   terms apply to the works covered under the License.  These additional terms
**   are non-permissive additional terms allowed under Section 7 of the GNU
**   Affero GPL version 3 and may not be removed by you.
** 
**   0. Attribution Requirement.
** 
**     You must preserve all legal notices or author attributions in the covered
**     work or Appropriate Legal Notices displayed by works containing the covered
**     work.  You may not remove from the covered work any author or developer
**     credit already included within the covered work.
** 
**   1. No License To Use Trademarks.
** 
**     This license does not grant any license or rights to use the trademarks
**     Golden Code, FWD, any Golden Code or FWD logo, or any other trademarks
**     of Golden Code Development Corporation. You are not authorized to use the
**     name Golden Code, FWD, or the names of any author or contributor, for
**     publicity purposes without written authorization.
** 
**   2. No Misrepresentation of Affiliation.
** 
**     You may not represent yourself as Golden Code Development Corporation or FWD.
** 
**     You may not represent yourself for publicity purposes as associated with
**     Golden Code Development Corporation, FWD, or any author or contributor to
**     the covered work, without written authorization.
** 
**   3. No Misrepresentation of Source or Origin.
** 
**     You may not represent the covered work as solely your work.  All modified
**     versions of the covered work must be marked in a reasonable way to make it
**     clear that the modified work is not originating from Golden Code Development
**     Corporation or FWD.  All modified versions must contain the notices of
**     attribution required in this license.
*/

package com.goldencode.p2j.persist.serial;

import java.io.*;
import java.lang.reflect.*;
import java.util.function.*;
import javax.xml.stream.*;
import com.goldencode.p2j.persist.*;
import com.goldencode.p2j.util.*;

/**
 * An implementation of the 4GL WRITE-XML method, which writes the records of a temp-table into
 * an XML stream.
 * <p>
 * Runtime support is partial at this time; not all modes and features are supported. PRODATASET
 * is not supported at this time.
 */
public final class XmlExport
{
   /** 3-space indent */
   private static final String INDENT = "   ";
   
   /** XML namespace attribute */
   private static final String ATTR_NS_XSI = "xmlns:xsi";
   
   /** Namespace URI for XML schema */
   private static final String NS_URI = "http://www.w3.org/2001/XMLSchema-instance";
   
   /** Nil attribute */
   private static final String ATTR_XSI_NIL = "xsi:nil";
   
   /** Platform-specific line separator */
   // TODO: we need client's line separate, not server's
   private static final String lineSep = System.lineSeparator();
   
   /** Temp-table buffer used for export */
   private final TemporaryBuffer buffer;
   
   /** Temp-table schema information */
   private final TempTableSchema schema;
   
   /** Flag indicating whether XML output should be pretty-printed */
   private final boolean format;
   
   /** XML stream writer */
   private XMLStreamWriter writer = null;
   
   /**
    * Constructor which immediately performs the export.
    * 
    * @param   buffer
    *          Temp-table buffer in which data is stored.
    * @param   target
    *          XML data target which normalizes access of various media to an output stream.
    * @param   formatted
    *          {@code true} if serialized output should be pretty-printed for human-friendly
    *          reading; {@code false} to omit formatting.
    * @param   encoding
    *          Character encoding (not currently used).
    * @param   schemaLocation
    *          Path to XML schema (not currently used).
    * @param   writeXmlSchema
    *          {@code true} to write XML schema information; {@code false} to write only data.
    *          If {@code true}, schema will be written to {@code schemaLocation}, if provided;
    *          if not provided, schema will be embedded in XML output. Not currently honored.
    * @param   minXmlSchema
    *          {@code true} to write minimal schema information; {@code false} to write verbose
    *          schema information. Ignored if {@code writeXmlSchema} is {@code false}. Not
    *          currently honored.
    * @param   writeBeforeImage
    *          Not currently honored.
    * @param   omitInitialValues
    *          If {@code true}, record values which match default initial values will be omitted
    *          from serialized output, else they will be included. Not currently honored.
    * 
    * @throws  PersistenceException
    *          if there is an error writing XML data or accessing the database.
    */
   public XmlExport(TemporaryBuffer buffer,
                    DataTarget target,
                    logical formatted,
                    character encoding,
                    character schemaLocation,
                    logical writeXmlSchema,
                    logical minXmlSchema,
                    logical writeBeforeImage,
                    logical omitInitialValues)
   throws PersistenceException
   {
      this.buffer = buffer;
      this.schema = new TempTableSchema(buffer);
      this.format = formatted.booleanValue();
      
      XMLOutputFactory factory = XMLOutputFactory.newInstance();
      
      OutputStream stream = null;
      
      try
      {
         stream = target.getOutputStream();
         writer = factory.createXMLStreamWriter(stream);
         
         writer.writeStartDocument("1.0");
         lineBreak();
         
         String tableName = schema.getTableXmlNodeName();
         
         // TODO: set namespace-prefix and namespace-uri
         
         writer.writeStartElement(tableName);
         writer.writeAttribute(ATTR_NS_XSI, NS_URI);
         lineBreak();
         
         String rowName = tableName + "Row";
         
         // called once per record
         Consumer<Persistable> worker = (dmo) ->
         {
            writeRecord(dmo, rowName);
         };
         
         // reads records in order of primary index, if available, else in ascending primary key
         // order
         buffer.readAllRows(worker);
         
         writer.writeEndElement();
         lineBreak();
         
         writer.writeEndDocument();
      }
      catch (XMLStreamException exc)
      {
         throw new PersistenceException(exc);
      }
      finally
      {
         if (writer != null)
         {
            try
            {
               writer.close();
            }
            catch (XMLStreamException exc)
            {
               // TODO: log
            }
         }
         
         if (stream != null)
         {
            try
            {
               stream.close();
            }
            catch (IOException exc)
            {
               // TODO: log
            }
         }
      }
   }
   
   /**
    * Write a temp-table record's data to the XML output stream.
    * 
    * @param   dmo
    *          Object containing temp-table record's data.
    * @param   rowName
    *          Name applied to the element or attribute associated with the row data in the XML
    *          output.
    */
   private void writeRecord(Persistable dmo, String rowName)
   {
      try
      {
         indent(1);
         writer.writeStartElement(rowName);
         
         // write all columns which must be serialized to attributes first, while the start
         // element for the row is open
         writeColumns(dmo, schema.attributeColumns());
         
         lineBreak();
         
         // write all columns which must be serialized to elements as separate XML elements
         writeColumns(dmo, schema.elementColumns());
         
         indent(1);
         writer.writeEndElement();
         lineBreak();
      }
      catch (XMLStreamException exc)
      {
         ErrorManager.recordOrThrowError(0, exc.getMessage());
      }
      catch (IllegalAccessException | IllegalArgumentException | InvocationTargetException exc)
      {
         throw new RuntimeException("Unable to access DMO data", exc);
      }
   }
   
   /**
    * Write data for the given columns from a single record into the XML output stream.
    * 
    * @param   dmo
    * @param   columns
    * 
    * @throws  IllegalAccessException
    *          if a DMO getter method cannot be accessed using reflection.
    * @throws  IllegalArgumentException
    *          if an illegal argument is passed to the DMO getter method for a column.
    * @throws  InvocationTargetException
    *          if there is a problem invoking a getter method on the given DMO.
    * @throws  XMLStreamException
    *          if there is an error writing to the XML stream.
    */
   private void writeColumns(Persistable dmo, Iterable<TempTableSchema.Column> columns)
   throws IllegalAccessException,
          IllegalArgumentException,
          InvocationTargetException,
          XMLStreamException
   {
      for (TempTableSchema.Column column : columns)
      {
         Integer extent = column.getExtent();
         Method getter = column.getGetter();
         
         if (extent != null)
         {
            for (int i = 0; i < extent; i++)
            {
               BaseDataType datum = (BaseDataType) getter.invoke(dmo, i);
               writeDatum(column, datum);
            }
         }
         else
         {
            BaseDataType datum = (BaseDataType) getter.invoke(dmo);
            writeDatum(column, datum);
         }
      }
   }
   
   /**
    * Write a single data value to the XML output stream, according to the instructions stored
    * in the temp-table schema.
    * 
    * @param   column
    *          Temp-table column schema information.
    * @param   datum
    *          Datum to be written to XML output.
    * 
    * @throws  XMLStreamException
    *          if there is an error writing to the XML stream.
    */
   private void writeDatum(TempTableSchema.Column column, BaseDataType datum)
   throws XMLStreamException
   {
      String name = column.getName();
      String value = datum.toStringMessage();
      
      switch (column.getNodeType())
      {
         case ATTRIBUTE:
            // TODO: how to handle nil?
            writer.writeAttribute(name, value);
            break;
            
         case ELEMENT:
            indent(2);
            writer.writeStartElement(name);
            if (datum.isUnknown())
            {
               writer.writeAttribute(ATTR_XSI_NIL, "true");
            }
            else
            {
               writer.writeCharacters(value);
            }
            writer.writeEndElement();
            lineBreak();
            break;
      }
   }
   
   /**
    * Write a line break to the XML output stream, if formatting is enabled.
    * 
    * @throws  XMLStreamException
    *          if there is an error writing to the XML stream.
    */
   private void lineBreak()
   throws XMLStreamException
   {
      if (format)
      {
         writer.writeCharacters(lineSep);
      }
   }
   
   /**
    * Write a soft indent to the XML output stream, if formatting is enabled.
    * 
    * @param   levels
    *          Levels of indent (3 spaces per indent level) to apply. Must be non-negative.
    *          
    * @throws  XMLStreamException
    *          if there is an error writing to the XML stream.
    */
   private void indent(int levels)
   throws XMLStreamException
   {
      if (format)
      {
         for (int i = 0; i < levels; i++)
         {
            writer.writeCharacters(INDENT);
         }
      }
   }
}
