/*
** Module   : ast.js
** Abstract : AST web client java script functionality.
**
** Copyright (c) 2017, Golden Code Development Corporation.
**
** -#- -I- --Date-- ------------------------------Description----------------------------------
** 001 ECF 20170523 First version.
*/

/*
** This program is free software: you can redistribute it and/or modify
** it under the terms of the GNU Affero General Public License as
** published by the Free Software Foundation, either version 3 of the
** License, or (at your option) any later version.
**
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
** GNU Affero General Public License for more details.
**
** You may find a copy of the GNU Affero GPL version 3 at the following
** location: https://www.gnu.org/licenses/agpl-3.0.en.html
** 
** Additional terms under GNU Affero GPL version 3 section 7:
** 
**   Under Section 7 of the GNU Affero GPL version 3, the following additional
**   terms apply to the works covered under the License.  These additional terms
**   are non-permissive additional terms allowed under Section 7 of the GNU
**   Affero GPL version 3 and may not be removed by you.
** 
**   0. Attribution Requirement.
** 
**     You must preserve all legal notices or author attributions in the covered
**     work or Appropriate Legal Notices displayed by works containing the covered
**     work.  You may not remove from the covered work any author or developer
**     credit already included within the covered work.
** 
**   1. No License To Use Trademarks.
** 
**     This license does not grant any license or rights to use the trademarks
**     Golden Code, FWD, any Golden Code or FWD logo, or any other trademarks
**     of Golden Code Development Corporation. You are not authorized to use the
**     name Golden Code, FWD, or the names of any author or contributor, for
**     publicity purposes without written authorization.
** 
**   2. No Misrepresentation of Affiliation.
** 
**     You may not represent yourself as Golden Code Development Corporation or FWD.
** 
**     You may not represent yourself for publicity purposes as associated with
**     Golden Code Development Corporation, FWD, or any author or contributor to
**     the covered work, without written authorization.
** 
**   3. No Misrepresentation of Source or Origin.
** 
**     You may not represent the covered work as solely your work.  All modified
**     versions of the covered work must be marked in a reasonable way to make it
**     clear that the modified work is not originating from Golden Code Development
**     Corporation or FWD.  All modified versions must contain the notices of
**     attribution required in this license.
*/

/**
 * AST-related functionality.
 */
define(["wsbase"], function(ws)
{
   "use strict";
   /** The top-level object which will be returned from this function. */
   var me = {};
   
   /** Starting value for AST-related message codes. */
   var MSG_BASE = ws.MSG_AST_BASE;
   
   /** AST token type to name array for token name lookups. */
   var tokenMap;
   
   /** AST symbolic name to token type map. */
   var nameMap = {};
   
   /** Message types. */
   var msgTypes =
   {
      TOKEN_MAP            :  MSG_BASE + 1,
   }
   
   /**
    * Initialize this module with token type/name information from the server.
    */
   me.init = function()
   {
      ws.sendMessage(msgTypes.TOKEN_MAP, null, function(err, data)
      {
         tokenMap = data;
         
         // initialize our reverse lookup map (names to types)
         for (var n = 0; n < data.length; n++)
         {
            if (typeof data[n] !== "undefined")
            {
               nameMap[data[n]] = n;
            }
         }
      });
   }
   
   me.tokenName = function(type)
   {
      // TODO: range check and error handling
      return tokenMap[type];
   }
   
   /**
    * Obtain the Progress Parser token type that matches the given name.
    *
    * @param    name
    *           The symbolic name of the token.
    *
    * @return   The matching token type or -1 if there is no match;
    */
   me.tokenType = function(name)
   {
      var type = -1;
      
      if (typeof name === "string" || name instanceof String)
      {
         var key = name.toUpperCase();
         
         if (typeof nameMap[key] !== "undefined")
         {
            type = nameMap[key];
         }
      }
      
      return type;
   }
   
   /**
    * Given an array of AST nodes, compose a hierarchical tree structure, using the parent ID of
    * each node. Each node will reference its parent, and each parent will contain an array of
    * children nodes.
    * 
    * @param   {Object[]} nodes
    *          Array of AST nodes.
    * 
    * @return  {Object} Root node of the tree.
    */
   me.composeAst = function(nodes)
   {
      var nodeMap = new Map();
      
      nodes.forEach(function(node)
      {
         node.token = fwd.ast.tokenName(node.type);
         var parent = nodeMap.get(node.pid);
         if (parent)
         {
            if (!parent.children)
            {
               parent.children = [];
            }
            parent.children.push(node);
            node.parent = parent;
            node.depth = parent.depth + 1;
         }
         else
         {
            node.depth = 0;
         }
         
         nodeMap.set(node.id, node);
      });
      
      return nodes[0];
   }
   
   /**
    * Given an array of AST nodes, create a map of IDs to nodes.
    * 
    * @param   {Object[]} nodes
    *          Array of AST nodes.
    * 
    * @return  {Map} ID to node map.
    */
   me.mapNodes = function(nodes)
   {
      var nodeMap = new Map();
      nodes.forEach(function(node) { nodeMap.set(node.id, node); });
      
      return nodeMap;
   }
   
   /**
    * Add an information panel describing the properties and annotations of an AST node to the
    * the specified, D3 selection. It is assumed that the selection represents an SVG or SVG
    * group element, and the panel is added as a group of SVG elements. The panel will be
    * positioned in the lower, left corner of the element represented by the given selection.
    * 
    * @param   {D3 Selection} selection
    *          Selection representing the SVG element to which to attach the info panel elements.
    * @param   {Object} ast
    *          An AST node.
    */
   me.showInfoPanel = function(selection, ast)
   {
      // append a group node to contain the elements of the panel
      var g = selection.append("g")
                       .attr("class", "ast-info-panel")
                       .style("opacity", 1e-6);
      
      // append a rectangle to contain the property information visually; dimensions and size
      // are computed later, based on the data to be displayed
      g.append("rect")
       .style("fill", "rgba(255, 255, 255, 0.8")
       .style("stroke", "steelblue")
       .style("stroke-width", 2)
       .attr("rx", 5)
       .attr("ry", 5)
      
      // store standard AST property data in a two-dimensional array
      var data =
      [
         [ "AST ID"     , ast.id                  ],
         [ "type"       , ast.token               ],
         [ "text"       , ast.text                ],
         [ "line : col" , ast.l0 + " : " + ast.c0 ],
      ];
      
      if (ast.depth != undefined)
      {
         data.push(["depth", ast.depth]);
      }
      
      // add any AST annotation information into the array
      // TODO: handle list annotations
      if (ast.notes)
      {
         data.push(["", ""]);
         for (var key in ast.notes)
         {
            data.push([key, ast.notes[key]]);
         }
      }
      
      // add a group element for each row of label/value pairs
      var row = g.selectAll("g")
                 .data(data)
                 .enter()
                 .append("g")
                 .attr("transform", function(d, i) { return "translate(7," + i * 15 + ")"; });
      
      // add a text element for each label and value
      var col = row.selectAll("text")
                   .data(function(d) { return d; })
                   .enter()
                   .append("text")
                   .attr("text-anchor", function(d, i) { return i == 0 ? "end" : "start"; })
                   .attr("x", function(d, i) { return i * 4 - 2; })
                   .text(function(d) { return d; })
                   .style("fill", function(d, i) { return i == 0 ? "#444" : "steelblue"; })
                   .style("font-weight", function(d, i) { return i == 0 ? "bold" : "normal"; });
      
      // determine the widest row
      var w0 = 0;
      var w1 = 0;
      row.selectAll("text")
         .each(function(d, i)
              {
                 var bbw = this.getBBox().width;
                 if (i == 0)
                 {
                    w0 = Math.max(w0, bbw);
                 }
                 else
                 {
                    w1 = Math.max(w1, bbw);
                 }
              });
      
      // calculate the location and size of the panel
      var svgh = selection.attr("height");
      var w = w0 + w1 + 14;
      var h = 15 * data.length + 10;
      var x = w0;
      var y = svgh - h + 15;
      
      // size and move the background rectangle relative to its parent group
      g.select("rect")
       .attr("width", w)
       .attr("height", h)
       .attr("transform", "translate(" + -w0 + ",-16)");
      
      // translate the group to place it at the bottom left of the parent
      var t = "translate(" + x + "," + y + ")";
      g.attr("transform", t)
       .transition()
       .duration(200)
       .style("opacity", 1);
   }
   
   /**
    * Remove an AST information panel previously added to an SVG.
    * 
    * @param   {D3 Selection} selection
    *          Selection representing the parent SVG element of the panel to be removed.
    */
   me.removeInfoPanel = function(selection)
   {
      selection.select(".ast-info-panel")
               .remove();
   }
   
   /* exports */
   return me;
});