/*
** Module   : BrowseJasperDataSource.java
** Abstract : Data source for a FWD Jasper report backed by a browse.
**
** Copyright (c) 2018, Golden Code Development Corporation.
**
** -#- -I- --Date-- ---------------------------------Description----------------------------------
** 001 SVL 20180207 Created initial version.
*/
/*
** This program is free software: you can redistribute it and/or modify
** it under the terms of the GNU Affero General Public License as
** published by the Free Software Foundation, either version 3 of the
** License, or (at your option) any later version.
**
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
** GNU Affero General Public License for more details.
**
** You may find a copy of the GNU Affero GPL version 3 at the following
** location: https://www.gnu.org/licenses/agpl-3.0.en.html
**
** Additional terms under GNU Affero GPL version 3 section 7:
**
**   Under Section 7 of the GNU Affero GPL version 3, the following additional
**   terms apply to the works covered under the License.  These additional terms
**   are non-permissive additional terms allowed under Section 7 of the GNU
**   Affero GPL version 3 and may not be removed by you.
**
**   0. Attribution Requirement.
**
**     You must preserve all legal notices or author attributions in the covered
**     work or Appropriate Legal Notices displayed by works containing the covered
**     work.  You may not remove from the covered work any author or developer
**     credit already included within the covered work.
**
**   1. No License To Use Trademarks.
**
**     This license does not grant any license or rights to use the trademarks
**     Golden Code, FWD, any Golden Code or FWD logo, or any other trademarks
**     of Golden Code Development Corporation. You are not authorized to use the
**     name Golden Code, FWD, or the names of any author or contributor, for
**     publicity purposes without written authorization.
**
**   2. No Misrepresentation of Affiliation.
**
**     You may not represent yourself as Golden Code Development Corporation or FWD.
**
**     You may not represent yourself for publicity purposes as associated with
**     Golden Code Development Corporation, FWD, or any author or contributor to
**     the covered work, without written authorization.
**
**   3. No Misrepresentation of Source or Origin.
**
**     You may not represent the covered work as solely your work.  All modified
**     versions of the covered work must be marked in a reasonable way to make it
**     clear that the modified work is not originating from Golden Code Development
**     Corporation or FWD.  All modified versions must contain the notices of
**     attribution required in this license.
*/

package com.goldencode.p2j.reporting;

import com.goldencode.p2j.persist.*;
import com.goldencode.p2j.ui.*;
import com.goldencode.p2j.util.*;
import net.sf.jasperreports.engine.*;
import java.util.HashMap;

import static com.goldencode.p2j.reporting.ReportFactory.showError;

/**
 * Data source for a FWD Jasper report backed by a browse. Rows are passed to the report in the
 * same way they look in browse (with the same formatting, after ROW-DISPLAY trigger has been
 * applied (if any)).
 */
class BrowseJasperDataSource
implements JasperDataSource
{
   /** Browse which provides data for the report. */
   private final BrowseWidget browse;

   /** Jasper report field name to column number (1-based). */
   private HashMap<String, Integer> fieldMapping;

   /** Browse rows which represent the report data. */
   private BrowseRow[] browseRows;

   /** Index of the current browse row in {@link #browseRows} while iterating the data set. */
   private int browseRowIndex;

   /**
    * Constructs a data source object for the given browse.
    *
    * @param browse
    *        Underlying browse.
    */
   BrowseJasperDataSource(BrowseWidget browse)
   {
      this.browse = browse;
   }

   /**
    * Get compiled Jasper report for the given report template design file.
    *
    * @param  designTemplateFileName
    *         File name of the Jasper report design template file.
    *
    * @return compiled Jasper report for the given report design template file.
    */
   @Override
   public JasperReport getReport(String designTemplateFileName)
   {
      return ReportFactory.getBrowseReport(browse, designTemplateFileName);
   }

   /**
    * Executed before report is filled from the data source.
    *
    * @param  report
    *         Jasper report to be filled.
    *
    * @return <code>true</code> on success, <code>false</code> if the data source cannot provide
    *         data.
    */
   @Override
   public boolean reportStarted(JasperReport report)
   {
      P2JQuery query = browse.getQuery();

      if (query == null)
      {
         showError("A query is not assigned to the browse.");
         return false;
      }

      if (!query.isOpen().booleanValue())
      {
         showError("Browse query should be open before the report is run.");
         return false;
      }

      if (!buildFieldMapping(report))
      {
         return false;
      }

      browseRows = browse.getRows(0, -1, true, -1, null, null);
      browseRowIndex = -1;
      return true;
   }

   /**
    * Executed after report was filled from the data source.
    *
    * @param  error
    *         <code>true</code> if there were problems while getting data from the data source.
    *         <code>false</code> on success,
    */
   @Override
   public void reportFinished(boolean error)
   {
      browseRows = null;
   }

   /**
    * Get underlying 4GL browse object which provides data for the report.
    *
    * @return handle to the underlying 4GL browse object which provides data for the report.
    */
   @Override
   public handle asProgressObject()
   {
      return new handle(browse);
   }

   /**
    * Tries to position the cursor on the next element in the data source.
    *
    * @return <code>true</code> if there is a next record, <code>false</code> otherwise.
    *
    * @throws JRException
    *         if any error occurs while trying to move to the next element.
    */
   @Override
   public boolean next()
   throws JRException
   {
      return browseRows != null && ++browseRowIndex < browseRows.length;
   }

   /**
    * Gets the field value for the current position.
    *
    * @param  field
    *         Field to get.
    *
    * @return an object containing the field value. The object type must be the field object type.
    *
    * @throws JRException
    *         if any error occurs while trying to get the field value.
    */
   @Override
   public Object getFieldValue(JRField field)
   throws JRException
   {
      if (browseRows == null || browseRowIndex < 0 || browseRowIndex >= browseRows.length)
      {
         throw new JRException("Cannot access browse row at index " + browseRowIndex);
      }

      int columnIndex = fieldMapping.get(field.getName()) - 1;
      return browseRows[browseRowIndex].text[columnIndex];
   }

   /**
    * Build and store field mapping (Jasper report field names to 1-based browse column index) for
    * the given report.
    *
    * @param  report
    *         Jasper report for which we need to build a field mapping.
    *
    * @return <code>true</code> on success.
    */
   private boolean buildFieldMapping(JasperReport report)
   {
      fieldMapping = new HashMap<>();
      for (JRField jrField : report.getFields())
      {
         String name = jrField.getName();
         try
         {
            Integer columnNumber = Integer.valueOf(name.replaceAll("column", ""));
            fieldMapping.put(name, columnNumber);
         }
         catch (NumberFormatException e)
         {
            showError("Invalid column name inside a browse report: " + name, e);
            return false;
         }
      }

      return true;
   }
}
