/*
** Module   : AmbiguousSchemaNameException.java
** Abstract : Exception caused by ambiguous namespace search
**
** Copyright (c) 2004-2017, Golden Code Development Corporation.
**
** -#- -I- --Date-- -T- --JPRM-- ----------------Description-----------------
** 001 ECF 20050101 ADD   @19160 Created initial version.
** 002 ECF 20050209 ADD   @19740 Added constructor which accepts message and
**                               root cause.
** 003 ECF 20050210 CHG   @19746 Added rootCause argument to constructor added
**                               for #002 (@19740). This was necessary to
**                               properly construct a useful string in the
**                               toString method when this exception is
**                               constructed by a subclass.
*/
/*
** This program is free software: you can redistribute it and/or modify
** it under the terms of the GNU Affero General Public License as
** published by the Free Software Foundation, either version 3 of the
** License, or (at your option) any later version.
**
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
** GNU Affero General Public License for more details.
**
** You may find a copy of the GNU Affero GPL version 3 at the following
** location: https://www.gnu.org/licenses/agpl-3.0.en.html
** 
** Additional terms under GNU Affero GPL version 3 section 7:
** 
**   Under Section 7 of the GNU Affero GPL version 3, the following additional
**   terms apply to the works covered under the License.  These additional terms
**   are non-permissive additional terms allowed under Section 7 of the GNU
**   Affero GPL version 3 and may not be removed by you.
** 
**   0. Attribution Requirement.
** 
**     You must preserve all legal notices or author attributions in the covered
**     work or Appropriate Legal Notices displayed by works containing the covered
**     work.  You may not remove from the covered work any author or developer
**     credit already included within the covered work.
** 
**   1. No License To Use Trademarks.
** 
**     This license does not grant any license or rights to use the trademarks
**     Golden Code, FWD, any Golden Code or FWD logo, or any other trademarks
**     of Golden Code Development Corporation. You are not authorized to use the
**     name Golden Code, FWD, or the names of any author or contributor, for
**     publicity purposes without written authorization.
** 
**   2. No Misrepresentation of Affiliation.
** 
**     You may not represent yourself as Golden Code Development Corporation or FWD.
** 
**     You may not represent yourself for publicity purposes as associated with
**     Golden Code Development Corporation, FWD, or any author or contributor to
**     the covered work, without written authorization.
** 
**   3. No Misrepresentation of Source or Origin.
** 
**     You may not represent the covered work as solely your work.  All modified
**     versions of the covered work must be marked in a reasonable way to make it
**     clear that the modified work is not originating from Golden Code Development
**     Corporation or FWD.  All modified versions must contain the notices of
**     attribution required in this license.
*/

package com.goldencode.p2j.schema;

import java.util.*;

/**
 * Exception thrown when a lookup of a schema entity name produces more than
 * one match within a namespace. This can occur if the namespace contains
 * duplicate entries, or if an abbreviation matches more than one
 * unabbreviated name nodes.
 */
public class AmbiguousSchemaNameException
extends SchemaException
{
   /** List of name nodes which matched criteria in ambiguous search */
   private List matches = null;
   
   /**
    * Constructor which accepts an error message and a list of matches.
    *
    * @param   message
    *          Error message.
    * @param   matches
    *          List of ambiguous name matches which were the cause of this
    *          exception.
    */
   public AmbiguousSchemaNameException(String message, List matches)
   {
      super(message);
      
      this.matches = matches;
   }
   
   /**
    * Constructor which accepts another exception as its root cause, as well
    * as an error message and list of ambiguous name matches. This variant is
    * used when the error condition is detected as the result of catching
    * another exception.
    *
    * @param   message
    *          Error message.
    * @param   matches
    *          List of ambiguous name matches which were the cause of this
    *          exception.
    * @param   rootCause
    *          Root cause exception.
    */
   public AmbiguousSchemaNameException(String message,
                                       List matches,
                                       Throwable rootCause)
   {
      super(message, rootCause);
      
      this.matches = matches;
   }
    
   /**
    * Get the list of {@link NameNode} objects which matched the name used in
    * the ambiguous search which caused this exception.
    *
    * @return  List of name nodes matching ambiguous search name.
    */
   public List getMatches()
   {
      return matches;
   }
   
   /**
    * Get the string representation of this exception, which consists of the
    * error message and a listing of all matching name nodes.
    *
    * @return  String representation of this exception.
    */
   public String toString()
   {
      StringBuffer buf = new StringBuffer(getMessage());
      buf.append(' ')
         .append(getMatches());
      
      return buf.toString();
   }
}