/*
** Module   : AcmeClient.java
** Abstract : Implements ACME client to get certificates signed by well-known trusted authority.
**
** Copyright (c) 2017, Golden Code Development Corporation.
**
** -#- -I- --Date-- ---------------------------------Description----------------------------------
** 001 SBI 20171030 First version.
** 002 SBI 20171110 Fixed a NPE in askCertificates.
** 003 SBI 20171116 Added CertificateSuite usage.
*/
/*
** This program is free software: you can redistribute it and/or modify
** it under the terms of the GNU Affero General Public License as
** published by the Free Software Foundation, either version 3 of the
** License, or (at your option) any later version.
**
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
** GNU Affero General Public License for more details.
**
** You may find a copy of the GNU Affero GPL version 3 at the following
** location: https://www.gnu.org/licenses/agpl-3.0.en.html
** 
** Additional terms under GNU Affero GPL version 3 section 7:
** 
**   Under Section 7 of the GNU Affero GPL version 3, the following additional
**   terms apply to the works covered under the License.  These additional terms
**   are non-permissive additional terms allowed under Section 7 of the GNU
**   Affero GPL version 3 and may not be removed by you.
** 
**   0. Attribution Requirement.
** 
**     You must preserve all legal notices or author attributions in the covered
**     work or Appropriate Legal Notices displayed by works containing the covered
**     work.  You may not remove from the covered work any author or developer
**     credit already included within the covered work.
** 
**   1. No License To Use Trademarks.
** 
**     This license does not grant any license or rights to use the trademarks
**     Golden Code, FWD, any Golden Code or FWD logo, or any other trademarks
**     of Golden Code Development Corporation. You are not authorized to use the
**     name Golden Code, FWD, or the names of any author or contributor, for
**     publicity purposes without written authorization.
** 
**   2. No Misrepresentation of Affiliation.
** 
**     You may not represent yourself as Golden Code Development Corporation or FWD.
** 
**     You may not represent yourself for publicity purposes as associated with
**     Golden Code Development Corporation, FWD, or any author or contributor to
**     the covered work, without written authorization.
** 
**   3. No Misrepresentation of Source or Origin.
** 
**     You may not represent the covered work as solely your work.  All modified
**     versions of the covered work must be marked in a reasonable way to make it
**     clear that the modified work is not originating from Golden Code Development
**     Corporation or FWD.  All modified versions must contain the notices of
**     attribution required in this license.
*/


package com.goldencode.p2j.security;


import java.io.*;
import java.net.*;
import java.security.*;
import java.security.cert.*;
import java.util.*;
import java.util.concurrent.*;
import java.util.stream.Collectors;

import org.bouncycastle.jce.provider.*;
import org.kohsuke.args4j.*;
import org.shredzone.acme4j.*;
import org.shredzone.acme4j.Session;
import org.shredzone.acme4j.Certificate;
import org.shredzone.acme4j.challenge.*;
import org.shredzone.acme4j.exception.*;
import org.shredzone.acme4j.util.*;
import org.slf4j.*;

import com.goldencode.p2j.security.SSLCertGenUtil.*;
import com.goldencode.p2j.security.SSLCertFactory.*;


/**
*  Defines the ACME client to get trusted certificates for the target domains. 
*/
public class AcmeClient
{
   /** The user private certificate */
   static final File USER_KEY_FILE = new File("user.key");
   
   /** The user registration URI */
   static final File USER_REG_FILE = new File("user.reg");
   
   /** The target domain private certificate */
   static final File DOMAIN_KEY_FILE = new File("domain.key");
   
   /**  The certificate request for the signed target domain certificate */
   static final File DOMAIN_CSR_FILE = new File("domain.csr");
   
   /**
    * The signed certificate full chain that comprises from the requested leaf certificate and
    * the trusted certificates chain.
    */
   static final File DOMAIN_FULLCHAIN_FILE = new File("domain-fullchain.crt");
   
   /** The signed certificate chain without a requested leaf certificate */
   static final File DOMAIN_CHAIN_FILE = new File("domain-chain.crt");
   
   /** The signed domain certificate */
   static final File DOMAIN_CRT_FILE = new File("domain.crt");
   
   /** The certificate domain registration file */
   static final File DOMAIN_REG_FILE = new File("domain.reg");
   
   /** If the challenge is not accepted for this period, then the client process is stopped. */
   private static final long ACCEPT_CHALLENGE_TIMEOUT = TimeUnit.MINUTES.toMillis(1);
   
   /** The number of tries to get validated challenge */
   private static final int CHALLENGE_TRIES = 100;
   
   /** The polling interval between challenge updates */
   private static final long UPDATE_CHALLENGE_TIMEOUT = 3000;
   
   /** The default RSA key size of generated key pairs */
   private static final int KEY_SIZE = 2048;

   /** The class logger */
   private static final Logger LOG = LoggerFactory.getLogger(AcmeClient.class);

   /** ACME server URI "acme://letsencrypt.org/staging" */
   private final String acmeServerUri;
   
   /** The managed server host address, the domain IP address */
   private final String host;
   
   /** The managed server port */ 
   private final int port;
   
   /** The instance of the managed web server */
   private ManagedWebServer mws;
   
   /** The current session with ACME server */
   private Session session;
   
   /** The ACME client account */
   private Registration registration;
   
   /**
    * Setups the ACME client to use the provided ACME server for its requests and the given
    * ACME accessible host and port.
    * 
    * @param    acmeServerUri
    *           The ACME server URI
    * @param    host
    *           The ACME client host address
    * @param    port
    *           The ACME client port
    */
   public AcmeClient(String acmeServerUri, String host, int port)
   {
      this.acmeServerUri = acmeServerUri;
      this.host          = host;
      this.port          = port;
   }
   
   /**
    * Returns the list of requested domains for the given string presentation in which domains
    * separated by spaces.
    * 
    * @param    domains
    *           The domains listed in a string and separated by spaces.
    * 
    * @return   The list of domains.
    */
   public static List<String> parseDomains(String domains)
   {
      List<String> list = Arrays.asList(domains.trim().split(" "))
                                .stream().filter(v -> {return !(v.trim().isEmpty());})
                                .collect(Collectors.toList());
      
      return list;
   }
   
   /**
    * Requests certificates for the given domains.
    * 
    * @param    certRequestInfo
    *           The certificate request information that contains these keys and their values:
    *           OU="organization unit", O="organization", L="locality(city)", S="state" and
    *           C="Country Code".
    * @param    domains
    *           Domain to get a common certificate for
    * 
    * @return   The certificate suite.
    */
   public CertificateSuite askCertificates(Map<RequestInfo, String> certRequestInfo,
                                           List<String> domains)
   throws IOException, AcmeException
   {
      if (session == null)
      {
         throw new IllegalStateException("The session has not been created yet.");
      }

      // Separately authorize every requested domain.
      for (String domain : domains)
      {
         authorize(domain);
         try
         {
            // the web server is not started because of request is valid
            if (mws != null)
            {
               mws.shutdown();
            }
         }
         catch (Exception e)
         {
            throw new AcmeException("Can't shutdown the managed server", e);
         }
      }

      // Load or create a key pair for the domains.
      KeyPair domainKeyPair = loadOrCreateKeyPair(DOMAIN_KEY_FILE);

      // Generate a CSR for all of the domains, and sign it with the domain key pair.
      RequestBuilder requestBuilder = new RequestBuilder(domainKeyPair, certRequestInfo, domains);

      requestBuilder.saveAsCrtFile(DOMAIN_CSR_FILE);

      Certificate certificate = registration.requestCertificate(requestBuilder.toBytes());

      LOG.info("Success! The certificate for domains " + domains + " has been generated!");
      LOG.info("Certificate URI: " + certificate.getLocation());
      serializeURI(certificate.getLocation(), DOMAIN_REG_FILE);

      // Download the leaf certificate and certificate chain.
      X509Certificate cert = certificate.download();
      
      // Write the leaf certificate
      try (FileWriter fw = new FileWriter(DOMAIN_CRT_FILE))
      {
         CertificateUtils.writeX509Certificate(cert, fw);
      }
      
      X509Certificate[] chain = certificate.downloadChain();

      // Write a combined file containing the certificate and chain.
      try (FileWriter fw = new FileWriter(DOMAIN_FULLCHAIN_FILE))
      {
         CertificateUtils.writeX509CertificateChain(fw, cert, chain);
      }
      
      // Write certificate chain in order to support older servers
      try (FileWriter fw = new FileWriter(DOMAIN_CHAIN_FILE))
      {
         CertificateUtils.writeX509CertificateChain(fw, null, chain);
      }
      
      CertificateSuite suite = new CertificateSuite();
      
      suite.privateKey = domainKeyPair.getPrivate();
      
      suite.publicKey = cert;
      
      suite.chain = chain;
      
      return suite;
   }
   
   /**
    * Connects to the ACME server on behalf of the provided user private key and gets signed
    * certificates for the requested domains.
    * 
    * @param    contact
    *           The provided contact information
    * @param    subjectFields
    *           The provided 
    * @param    domains
    *           The requested domains
    * @param    userKeyFile
    *           The user private certificate
    * @param    userKeyRegFile
    *           The user registration data file
    * 
    * @return   The certificate suite.
    * 
    * @throws   AcmeException
    *           If the request for getting target certificates is failed.
    */
   public CertificateSuite getCertificates(String                   contact,
                                           Map<RequestInfo, String> subjectFields,
                                           List<String>             domains,
                                           File                     userKeyFile,
                                           File                     userKeyRegFile)
   throws AcmeException
   {
      CertificateSuite suite = null;
      try
      {
         if (domains == null || domains.isEmpty())
         {
            throw new AcmeException("Domains must be provided");
         }
         
         KeyPair userKeyPair = loadOrCreateKeyPair(userKeyFile);
         
         URI regAccount = null;
         
         try
         {
            regAccount = readURI(userKeyRegFile);
            loginAccount(userKeyPair, regAccount);
         }
         catch (IOException | ClassNotFoundException ex)
         {
            List<String> regContacts = new ArrayList<String>();
            
            if (contact != null && !contact.isEmpty())
            {
               regContacts.add(contact);
            }
            
            regAccount = registerAccount(userKeyPair, regContacts);
            serializeURI(regAccount, userKeyRegFile);
         }
         
         LOG.info("The client account is " + regAccount);
         
         suite = askCertificates(subjectFields, domains);
      }
      catch (Exception ex)
      {
         LOG.error("Failed to get a certificate for domains " + domains, ex);
      }

      shutdown();
      
      return suite;
   }
   
   /**
    * Shutdown the managed web server.
    */
   public void shutdown()
   {
      if (mws != null)
      {
         try
         {
            mws.shutdown();
         }
         catch (Exception e)
         {
         }
      }
   }
   
   /**
    * Loads a key pair from the specified file.
    * 
    * @param    file
    *           The file containing the target PEM private certificate.
    * 
    * @return   {@link KeyPair}.
    */
   private static KeyPair loadKeyPair(File file)
   throws IOException
   {
      try (FileReader fr = new FileReader(file))
      {
         return KeyPairUtils.readKeyPair(fr);
      }
   }

   /**
    * Creates a new private certificate and saves it in the specified file.
    * 
    * @param    file
    *           The file to store the created PEM private certificate.
    * 
    * @return   {@link KeyPair}.
    */
   private static KeyPair createKeyPair(File file)
   throws IOException
   {
      KeyPair domainKeyPair = KeyPairUtils.createKeyPair(KEY_SIZE);
      
      try (FileWriter fw = new FileWriter(file))
      {
         KeyPairUtils.writeKeyPair(domainKeyPair, fw);
      }
      
      return domainKeyPair;
   }
   
   /**
    * Loads a key pair from specified file. If the file does not exist, a new key pair is
    * generated and saved.
    * 
    * @param    file
    *           The file containing the target PEM private certificate.
    * 
    * @return   {@link KeyPair}.
    */
   private static KeyPair loadOrCreateKeyPair(File file)
   throws IOException
   {
      if (file.exists())
      {
         return loadKeyPair(file);
      }
      else
      {
         return createKeyPair(file);
      }
   }
   
   /**
    * Serializes the target registration URI into the given file.
    * 
    * @param    regUri
    *           The registration URI
    * @param    file
    *           The given file to store the client registration.
    */
   private static void serializeURI(URI regUri, File file)
   {
      try (ObjectOutputStream oos = new ObjectOutputStream(new FileOutputStream(file)))
      {
         oos.writeObject(regUri);
         oos.flush();
      }
      catch(IOException ex)
      {
         LOG.info("Can't save the client registration URI '" + regUri + "'");
      }
   }
   
   /**
    * Reads the target registration URI from the given file.
    * 
    * @param    file
    *           The given file with the client registration.
    * 
    * @throws   IOException
    *           Iff IO read operation is failed
    * @throws   FileNotFoundException 
    *           Iff the target file doesn't exist
    * @throws   ClassNotFoundException 
    *           Iff the serialized class is not found
    */
   private static URI readURI(File file)
   throws FileNotFoundException, IOException, ClassNotFoundException
   {
      try (ObjectInputStream ois = new ObjectInputStream(new FileInputStream(file)))
      {
         Object uri = ois.readObject();
         if (!(uri instanceof URI))
         {
            throw new IOException("The file is corrupted");
         }
         
         return (URI) uri;
      }
   }
   
   
   /**
    * Delete file and log results
    * 
    * @param    file
    *           The target file to delete
    */
   private static void deleteFile(File file)
   {
      if (file.exists())
      {
         if (!file.delete())
         {
            LOG.info( file.getPath() + " wasn't deleted. Please remove it manually.");
         }
      }
   }
   
   /**
    * Authorize a domain. It will be associated with your account, so you will be able to
    * retrieve a signed certificate for the domain later. It is required to do separate
    * authorizations for subdomains (e.g. "www" subdomain). Wildcard certificates are
    * not currently supported.
    *
    * @param    domain
    *           Name of the domain to authorize
    * 
    * @throws   AcmeException
    *           iff the tls-sni-01 challenge validation is failed.
    */
   private void authorize(String domain)
   throws AcmeException
   {
      // Authorize the domain.
      Authorization auth = registration.authorizeDomain(domain);
      LOG.info("Authorization for domain " + domain);
      
      // Find the desired challenge and prepare it.
      Challenge challenge = tlsSniChallenge(auth, domain);
      
      if (challenge == null)
      {
         throw new AcmeException("No challenge found");
      }

      // If the challenge is already verified, there's no need to execute it again.
      if (challenge.getStatus() == Status.VALID)
      {
         return;
      }
      
      if (!acceptTlsSniChallenge(((TlsSni01Challenge) challenge).getSubject(),
                                 ACCEPT_CHALLENGE_TIMEOUT))
      {
         throw new AcmeException("Challenge is not accepted");
      }
      
      // Now trigger the challenge.
      challenge.trigger();
      
      // Poll for the challenge to complete.
      try
      {
         int attempts = CHALLENGE_TRIES;
         while (challenge.getStatus() != Status.VALID && attempts-- > 0)
         {
            // Did the authorization fail?
            if (challenge.getStatus() == Status.INVALID)
            {
               throw new AcmeException("Challenge failed... Giving up.");
            }

            // Wait for a few seconds
            Thread.sleep(UPDATE_CHALLENGE_TIMEOUT);

            // Then update the status
            challenge.update();
         }
      }
      catch (InterruptedException ex)
      {
         LOG.error("interrupted", ex);
         Thread.currentThread().interrupt();
      }

      // All reattempts are used up and there is still no valid authorization?
      if (challenge.getStatus() != Status.VALID)
      {
         throw new AcmeException("Failed to pass the challenge for domain " + domain
                  + ", ... Giving up.");
      }
   }

   /**
    * Prepares the TLS-SNI challenge. The verification of this challenge expects that
    * the web server returns a special validation certificate.
    *
    * @param    auth
    *           {@link Authorization} to find the challenge in
    * @param    domain
    *           Domain name to be authorized
    * 
    * @return   {@link Challenge} to verify
    */
   @SuppressWarnings("deprecation")
   // until tls-sni-02 is supported
   private Challenge tlsSniChallenge(Authorization auth, String domain)
   throws AcmeException
   {
      // Find a single tls-sni-01 challenge
      TlsSni01Challenge challenge = auth.findChallenge(TlsSni01Challenge.TYPE);
      if (challenge == null)
      {
         throw new AcmeException("Found no " + TlsSni01Challenge.TYPE +
                                 " challenge, don't know what to do...");
      }

      // Get the Subject
      String subject = challenge.getSubject();

      // Create a validation key pair
      KeyPair domainKeyPair;
      try (FileWriter fw = new FileWriter("tlssni.key"))
      {
         domainKeyPair = KeyPairUtils.createKeyPair(2048);
         KeyPairUtils.writeKeyPair(domainKeyPair, fw);
      }
      catch (IOException ex)
      {
         throw new AcmeException("Could not write keypair", ex);
      }

      // Create a validation certificate
      try (FileWriter fw = new FileWriter("tlssni.crt"))
      {
         X509Certificate cert = CertificateUtils.createTlsSniCertificate(domainKeyPair, subject);
         CertificateUtils.writeX509Certificate(cert, fw);
      }
      catch (IOException ex)
      {
         throw new AcmeException("Could not write certificate", ex);
      }

      LOG.info("In order to pass tls-sni-01 challenge the web server must return " +
               "the certificate 'tlssni.crt' on a SNI request to:");
      LOG.info(subject);
      LOG.info("The matching keypair must be available at 'tlssni.key'.");

      StringBuilder message = new StringBuilder();
      message.append("Please use 'tlssni.key' and 'tlssni.crt' cert for SNI requests to:\n\n");
      message.append("https://").append(subject).append("\n\n");
      LOG.info(message.toString());
      
      return challenge;
   }

   /**
    * Prepares the challenge validation that starts the managed web server for the provided
    * subject.
    *
    * @param    subject
    *           The provided subject
    * @param    timeout
    *           The waiting time until the server is ready or failed.
    */
   private boolean acceptTlsSniChallenge(String subject, long timeout)
   throws AcmeException
   {
      try
      {
         mws = new ManagedWebServer(host, port, subject);
         
         return mws.waitUntilReady(timeout);
      }
      catch (Exception e)
      {
         throw new AcmeException("Managed web server " + host + ":" + port  + " can't be started",
                                 e);
      }
   }

   /**
    * Revoke the target certificate.
    * 
    * @param    domainCertUri
    *           The domain certificate URI to revoke
    * @param    userKeyPair
    *           The user private certificate
    * 
    * @throws   AcmeException
    *           Iff the certificate revocation is not succeeded. 
    */
   private void revoke(URI domainCertUri, KeyPair userKeyPair)
   throws AcmeException
   {
      Session curr = new Session(acmeServerUri, userKeyPair);
      Certificate domainCert = Certificate.bind(curr, domainCertUri);
      domainCert.revoke(RevocationReason.KEY_COMPROMISE);
   }
   
   /**
    * Remove the target registration account.
    * 
    * @param    regAccount
    *           The registration account URI to remove
    * @param    userKeyPair
    *           The user private certificate
    * 
    * @throws   AcmeException
    *           Iff the removing registration account is not succeeded. 
    */
   private void remove(URI regAccount, KeyPair userKeyPair)
   throws AcmeException
   {
      Session curr = new Session(acmeServerUri, userKeyPair);
      Registration reg = Registration.bind(curr, regAccount);
      reg.deactivate();
      
      deleteFile(USER_KEY_FILE);
      
      deleteFile(USER_REG_FILE);
   }

   /**
    * Registers new account or login into the existing one.
    * 
    * @param    userKeyPair
    *           The user private certificate
    * @param    contacts
    *           Contact information
    * 
    * @return   The registered account URI
    * 
    * @throws   AcmeException
    *           iff registration is failed
    */
   private URI registerAccount(KeyPair userKeyPair, List<String> contacts)
   throws AcmeException
   {
      // Create a session for Let's Encrypt.
      // Use "acme://letsencrypt.org" for production server
      // https://acme-staging.api.letsencrypt.org/directory
      session = new Session(acmeServerUri, userKeyPair);
      
      RegistrationBuilder builder = new RegistrationBuilder();
      
      for(String contact : contacts)
      {
         builder.addContact(contact);
      }
      
      try
      {
         registration = builder.create(session);
         
         URI agreement = registration.getAgreement();
         setAgreement(agreement);
      }
      catch (AcmeConflictException ex)
      {
         registration = Registration.bind(session, ex.getLocation());
      }
      
      return registration.getLocation();
   }
   
   /**
    * Login the registered account.
    * 
    * @param    userKeyPair
    *           The user private certificate associated with its account.
    * @param    regAccount
    *           The registered account URI 
    * 
    * @throws   AcmeException
    *           iff registration is failed
    */
   private void loginAccount(KeyPair userKeyPair, URI regAccount)
   throws AcmeException
   {
      session = new Session(acmeServerUri, userKeyPair);
      
      registration = Registration.bind(session, regAccount);
      
      setAgreement(registration.getAgreement());
   }
   
   /**
    * Updates the registered account with new contact information.
    * 
    * @param    contact
    *           New contact information
    * 
    * @throws   AcmeException
    *           iff this transaction is failed.
    */
   private void addContact(String contact)
   throws AcmeException
   {
      registration.modify().addContact(contact).commit();
   }
   
   /**
    * Confirms the Terms of Service given by its URI.
    *
    * @param    agreement
    *           The Terms of Service given by its URI
    * 
    * @throws   AcmeException
    *           iff this transaction is failed.
    */
   private void setAgreement(URI agreement)
   throws AcmeException
   {
      registration.modify().setAgreement(agreement).commit();
   }
   
   /**
    * Change the registered account key.
    * 
    * @param    userKeyPair
    *           New user private certificate
    * 
    * @throws   AcmeException
    *           iff this transaction is failed.
    */
   private void changeUserKey(KeyPair userKeyPair)
   throws AcmeException
   {
      registration.changeKey(userKeyPair);
   }
   
   /**
    * Runs ACME client to get certificates for the target domains. The required parameters are
    * -server "acme://letsencrypt.org/staging"
    * -domains "test8.acme.com test7.acme.com"
    * -host "127.0.0.1"
    * -port 8888
    *
    * @param    args
    *           The provided parameters
    */
   public static void main(String[] args)
   {
      InputParameters inputParameters = new InputParameters();
      
      CmdLineParser parser = new CmdLineParser(inputParameters);
      
      try
      {
         parser.parseArgument(args);
      }
      catch (CmdLineException e)
      {
         System.err.println(e.getMessage());
         parser.printUsage(System.err);
         return;
      }

      LOG.info("AcmeClient is starting up...");

      Security.addProvider(new BouncyCastleProvider());
      
      AcmeClient client = new AcmeClient(inputParameters.serverUri,
                                         inputParameters.host,
                                         inputParameters.port);
      switch(inputParameters.acmeOp)
      {
         case REQUEST:
            try
            {
               List<String> domains = parseDomains(inputParameters.domains);
               
               client.getCertificates(inputParameters.contact,
                                      inputParameters.subjectFields,
                                      domains,
                                      USER_KEY_FILE,
                                      USER_REG_FILE);
            }
            catch (AcmeException ex)
            {
               LOG.info("ACME client failed in getCertificates due to this reason", ex);
            }
            break;
         case REVOKE:
            try
            {
               KeyPair userKeyPair = loadKeyPair(USER_KEY_FILE);
               URI domainCertUri = readURI(DOMAIN_REG_FILE);
               client.revoke(domainCertUri, userKeyPair);
            }
            catch (Exception ex)
            {
               LOG.info("ACME client failed to revoke due to this reason", ex);
            }
            break;
         case UPDATE_ACME_ACCOUNT:
            try
            {
               KeyPair userKeyPair = loadKeyPair(USER_KEY_FILE);
               
               URI regAccount = readURI(USER_REG_FILE);
               
               client.loginAccount(userKeyPair, regAccount);
               
               if (inputParameters.contact != null && !inputParameters.contact.isEmpty())
               {
                  client.addContact(inputParameters.contact);
               }
               
               if (inputParameters.key != null && !inputParameters.key.isEmpty())
               {
                  KeyPair newKeyPair = loadKeyPair(new File(inputParameters.key));
                  
                  client.changeUserKey(newKeyPair);
                  
                  client.loginAccount(newKeyPair, regAccount);
                  //substitute the old user key file if login account is successful
                  try(FileWriter writer = new FileWriter(USER_KEY_FILE))
                  {
                     KeyPairUtils.writeKeyPair(newKeyPair, writer);
                  }
               }
            }
            catch (Exception ex)
            {
               LOG.info("ACME client failed to update registartion account due to this reason",
                        ex);
            }
            
            break;
         case REMOVE_ACME_ACCOUNT:
            try
            {
               KeyPair userKeyPair = loadKeyPair(USER_KEY_FILE);
               
               URI regAccount = readURI(USER_REG_FILE);
               
               client.remove(regAccount, userKeyPair);
            }
            catch(Exception ex)
            {
               LOG.info("ACME client failed to deactivate registartion account for this reason",
                        ex);
            }
            
            break;
         default:
      }
      
      LOG.info("AcmeClient is done.");
   }

   /**
    * Defines certificate requests data fields. 
    */
   enum RequestInfo
   {
      /** Organization unit */
      OU,
      
      /** Organization */
      O,
      
      /** Locality (city) */
      L,
      
      /** State (province name) */
      ST,
      
      /** Country code */
      C
   }
   
   /**
    * The ACME client command.
    */
   enum AcmeOps
   {
      /** Request target certificates */
      REQUEST,
      
      /** Revoke target certificates */
      REVOKE,
      
      /** Update ACME registration account info */
      UPDATE_ACME_ACCOUNT,
      
      /** Remove ACME registration account */
      REMOVE_ACME_ACCOUNT
   }
   
   /**
    * Wraps functionality of acme4j certificate request builder required to build a certificate
    * request for the target domains.
    */
   class RequestBuilder
   {
      /** Certificate request builder */
      private final CSRBuilder csrb = new CSRBuilder();
      
      /**
       * Creates a wrapper object around acme4j certificate request builder to build a certificate
       * request for the target domains.
       * 
       * @param    domainKeyPair
       *           The domain private certificate
       * @param    certRequestInfo
       *           The certificate request information
       * @param    domains
       *           The requested domains
       * 
       * @throws   IOException
       *           iff the underlined certificate request can't be signed by the provided domain
       *           private key.  
       */
      public RequestBuilder(KeyPair                  domainKeyPair,
                            Map<RequestInfo, String> certRequestInfo,
                            List<String>             domains)
      throws IOException
      {
         if (certRequestInfo != null)
         {
            String country = certRequestInfo.get(RequestInfo.C);
            
            if (country != null)
            {
               csrb.setCountry(country);
            }
            
            String loc = certRequestInfo.get(RequestInfo.L);
            
            if (loc != null)
            {
               csrb.setLocality(loc);
            }
            
            String org = certRequestInfo.get(RequestInfo.O);
            
            if (org != null)
            {
               csrb.setOrganization(org);
            }
            
            String orgUnit = certRequestInfo.get(RequestInfo.OU);
            
            if (orgUnit != null)
            {
               csrb.setOrganizationalUnit(orgUnit);
            }
            
            String state = certRequestInfo.get(RequestInfo.ST);
            
            if (state != null)
            {
               csrb.setState(state);
            }
         }
         
         csrb.addDomains(domains);
         
         csrb.sign(domainKeyPair);
      }
      
      /**
       * Save this request as a PEM certificate request file.
       * 
       * @param    crtFile
       *           The given certificate request file to save this certificate request.
       * 
       * @throws   IOException
       *           Iff the requested IO write is failed.
       */
      public void saveAsCrtFile(File crtFile)
      throws IOException
      {
         try (Writer out = new FileWriter(crtFile))
         {
            csrb.write(out);
         }
      }
      
      /**
       * Gets this certificate request as PKCS#10 data bytes.
       * 
       * @return   The PKCS#10 data bytes
       * 
       * @throws   IOException
       *           iff this IO operation is failed.
       */
      public byte[] toBytes() throws IOException
      {
         return csrb.getEncoded();
      }
   }
   
   /**
    * Defines the ACME client input parameters.
    */
   static class InputParameters
   {
      /**
       * The domain host address. Can be set via "-host" option.
       */
      @Option(name="-host",
              usage="The domain host address.")
      public String host;

      /**
       * The internal redirected port or 443 standard HTTPS port. Can be set via "-port" option.
       * It must be 443 only or the redirected port number. ACME client must be available
       * for external ACME servers via the requested domain name and the port number 443.
       */
      @Option(name="-port",
              usage="The domain port.")
      public Integer port;

      /**
       * The ACME server URI. Can be set via "-server" option. For the testing purpose it must be
       * "acme://letsencrypt.org/staging".
       */
      @Option(name="-server",
              depends={"-host","-port", "-domains"},
              required=true,
              usage="The ACME server URI.")
      public String serverUri;
      
      /**
       * The list of requested domains separated by spaces that is similar to this example,
       * "test1.acme.com test2.acme.com test3.acme.com". The wild cards are not supported. 
       */
      @Option(name="-domains",
              usage="The requested domains enclosed in one string within double quoters and " +
                    "separated by spaces.")
      public String domains;
      
      /**
       * Sets mandatory subject fields for generated certificates:
       * -S OU="organization unit"
       * -S O="organization"
       * -S L="locality(city)"
       * -S ST="state(province name)"
       * -S C="Country Code"
       */
      @Option(name="-S", handler=RequestInfoMapper.class)
      public Map<RequestInfo, String> subjectFields;
      
      /**
       * The registration contact information. Can be set via "-contact" option,
       * "-contact admin@acme.com"
       */
      @Option(name="-contact",
              usage="The registration contact information.")
      public String contact;
      
      /**
       * The ACME client command to request or to revoke target certificates and to update or
       * to remove ACME client account. 
       */
      @Option(name="-acme",
              aliases="--acme-command",
              depends={"-host", "-port", "-server"},
              usage="The ACME client command.")
      public AcmeOps acmeOp = AcmeOps.REQUEST;
      
      /**
       * The user private key to update. Can be set via "-key" option.
       */
      @Option(name="-key",
              usage="The user private key.")
      public String key;
   }
}