/*
** Module   : Authenticator.java
** Abstract : hooks to implement custom authentication logic
**
** Copyright (c) 2005-2017, Golden Code Development Corporation.
**
** -#- -I- --Date-- --JPRM-- ---------------------------- Description ----------------------------
** 001 NVS 20050208   @19696 Created the initial version.
** 002 NVS 20050914   @22709 Added clientFinalize() to the interface.
** 003 GES 20061003   @30107 Added support for return codes and cleaned
**                           up code formatting/javadoc.
** 004 GES 20061220   @31797 Removed use of AuthUIHelper since it is
**                           generally redundant with the idea of a
**                           client-side hook anyway.
** 005 SVL 20100923          Added input parameter "parameters" for the
**                           clientAuthHook function. 
** 006 OM  20131018          Added support for optionally returning a SessionListener.
** 007 EVL 20160224          Javadoc fixes to make compatible with Oracle Java 8 for Solaris 10.
*/
/*
** This program is free software: you can redistribute it and/or modify
** it under the terms of the GNU Affero General Public License as
** published by the Free Software Foundation, either version 3 of the
** License, or (at your option) any later version.
**
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
** GNU Affero General Public License for more details.
**
** You may find a copy of the GNU Affero GPL version 3 at the following
** location: https://www.gnu.org/licenses/agpl-3.0.en.html
** 
** Additional terms under GNU Affero GPL version 3 section 7:
** 
**   Under Section 7 of the GNU Affero GPL version 3, the following additional
**   terms apply to the works covered under the License.  These additional terms
**   are non-permissive additional terms allowed under Section 7 of the GNU
**   Affero GPL version 3 and may not be removed by you.
** 
**   0. Attribution Requirement.
** 
**     You must preserve all legal notices or author attributions in the covered
**     work or Appropriate Legal Notices displayed by works containing the covered
**     work.  You may not remove from the covered work any author or developer
**     credit already included within the covered work.
** 
**   1. No License To Use Trademarks.
** 
**     This license does not grant any license or rights to use the trademarks
**     Golden Code, FWD, any Golden Code or FWD logo, or any other trademarks
**     of Golden Code Development Corporation. You are not authorized to use the
**     name Golden Code, FWD, or the names of any author or contributor, for
**     publicity purposes without written authorization.
** 
**   2. No Misrepresentation of Affiliation.
** 
**     You may not represent yourself as Golden Code Development Corporation or FWD.
** 
**     You may not represent yourself for publicity purposes as associated with
**     Golden Code Development Corporation, FWD, or any author or contributor to
**     the covered work, without written authorization.
** 
**   3. No Misrepresentation of Source or Origin.
** 
**     You may not represent the covered work as solely your work.  All modified
**     versions of the covered work must be marked in a reasonable way to make it
**     clear that the modified work is not originating from Golden Code Development
**     Corporation or FWD.  All modified versions must contain the notices of
**     attribution required in this license.
*/

package com.goldencode.p2j.security;

import com.goldencode.p2j.net.SessionListener;
import java.util.*;

/**
 * This interface defines methods that provide a way to implement custom authentication logic.
 * The <code>Securitymanager</code> will send the implementing class to client for executing and
 * will also instantiate an object on server-side. If not configured with, the 
 * <code>Securitymanager</code> will use its own <code>Authenticator</code> interface to handle
 * server-side authentication processing of response from client-side.
 * <p>
 * The <code>Securitymanager</code> will create iterate all the entities that the server side of 
 * the hook reports with <code>getAuthenticationEntities</code> and call <code>clientAuthHook</code>
 * for obtaining credentials for each entity. Then the credentials packaged as a byte array 
 * are decoded and checked by <code>serverAuthHook</code> on server-side.
 */
public interface Authenticator
extends SecurityConstants
{
   /**
    * Implements client side custom authentication logic.
    *
    * @param    parameters
    *           Additional configuration parameters.
    * @param    code
    *           The result of the most recent attempt to authenticate or
    *           <code>AUTH_RESULT_NONE</code> if this is the first attempt.
    *
    * @return   Array of bytes to be transmitted to the server as the
    *           authorization input.
    */
   public byte[] clientAuthHook(Map<String, Object> parameters, int code);

   /**
    * Finalizes any resources allocated during authentication by the client.
    */
   public void clientFinalize();

   /**
    * Implements server side custom authentication logic.
    * <p>
    * Accepts the byte array produced by the client side authentication hook
    * as authentication input and custom parameters.
    *
    * @param    auth
    *           The authorization input from the client.
    * @param    entity
    *           The entity to be authenticated.
    *
    * @return   The result of the authentication processing.
    */
   public AuthenticationResponse serverAuthHook(byte[] auth, String entity);

   /**
    * Optionally the server-side may return a session listener. If some additional settings should
    * be done based on <code>Session</code> events (initialization / termination).
    * The initialization callback will be executed on the client thread, after the correct
    * context will be created.
    * The termination callback will be executed when the session ends (the client disconnects).
    *
    * @return  The current session listener (if there is any).
    */
   public SessionListener getSessionListener();

   /**
    * Configures the Authenticator by setting the "option" parameter from directory.xml.
    * The meaning of the parameter depends on the implementation class.
    *
    * @param   option
    *          The value of "option" entry for the auth plugin.
    */
   public void configure(String option);

   /**
    * Returns a set of entities that this class handles.
    * They will later be processed one at a time and allowed a number of retries for each entry.
    * This runs on the server-side of application.
    *
    * @return  The set of entities which need authentication that the Authenticator has detected
    *          by inspecting the directory or other means.
    *          <p>
    *          <code>null</code> if enumeration of entities could not be executed because of
    *          insufficient information or other non-fatal errors.
    */
   public Set<String> getAuthenticationEntities();
}