/*
** Module   : BitSet.java
** Abstract : Bit set implementation.
**
** Copyright (c) 2017, Golden Code Development Corporation.
**
** -#- -I- --Date-- ------------------------------Description----------------------------------
** 001 HC  20170612 Initial version.
*/
/*
** This program is free software: you can redistribute it and/or modify
** it under the terms of the GNU Affero General Public License as
** published by the Free Software Foundation, either version 3 of the
** License, or (at your option) any later version.
**
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
** GNU Affero General Public License for more details.
**
** You may find a copy of the GNU Affero GPL version 3 at the following
** location: https://www.gnu.org/licenses/agpl-3.0.en.html
**
** Additional terms under GNU Affero GPL version 3 section 7:
**
**   Under Section 7 of the GNU Affero GPL version 3, the following additional
**   terms apply to the works covered under the License.  These additional terms
**   are non-permissive additional terms allowed under Section 7 of the GNU
**   Affero GPL version 3 and may not be removed by you.
**
**   0. Attribution Requirement.
**
**     You must preserve all legal notices or author attributions in the covered
**     work or Appropriate Legal Notices displayed by works containing the covered
**     work.  You may not remove from the covered work any author or developer
**     credit already included within the covered work.
**
**   1. No License To Use Trademarks.
**
**     This license does not grant any license or rights to use the trademarks
**     Golden Code, FWD, any Golden Code or FWD logo, or any other trademarks
**     of Golden Code Development Corporation. You are not authorized to use the
**     name Golden Code, FWD, or the names of any author or contributor, for
**     publicity purposes without written authorization.
**
**   2. No Misrepresentation of Affiliation.
**
**     You may not represent yourself as Golden Code Development Corporation or FWD.
**
**     You may not represent yourself for publicity purposes as associated with
**     Golden Code Development Corporation, FWD, or any author or contributor to
**     the covered work, without written authorization.
**
**   3. No Misrepresentation of Source or Origin.
**
**     You may not represent the covered work as solely your work.  All modified
**     versions of the covered work must be marked in a reasonable way to make it
**     clear that the modified work is not originating from Golden Code Development
**     Corporation or FWD.  All modified versions must contain the notices of
**     attribution required in this license.
*/
package com.goldencode.p2j.security;

import java.io.*;

/**
 * This class provides GWT-compatible bit set implementation, the usage of java.util.BitSet
 * breaks GWT serialization.
 */
public class BitSet
implements Serializable
{
   /** data array */
   private int[] data;

   /** number of bits in the set */
   private int bits;

   /**
    * Default ctor creates a bit set of size 64.
    */
   public BitSet()
   {
      this(64);
   }

   /**
    * Ctor.
    *
    * @param   bits
    *          The desired set size.
    */
   public BitSet(int bits)
   {
      data = new int[(bits >> 5) + 1];
      this.bits = bits;
   }

   /**
    * Sets a bit to the provided value.
    *
    * @param   bit
    *          Bit index.
    * @param   state
    *          New bit state.
    */
   public void set(int bit, boolean state)
   {
      if (state)
      {
         set(bit);
      }
      else
      {
         unset(bit);
      }
   }

   /**
    * Turns a bit on.
    *
    * @param   bit
    *          Bit index.
    */
   public void set(int bit)
   {
      int remainder = bit & 0x1F;
      int div = bit >> 5;
      data[div] = data[div] | (1 << remainder);
   }

   /**
    * Turns a bit off.
    *
    * @param   bit
    *          Bit index.
    */
   public void unset(int bit)
   {
      int remainder = bit & 0x1F;
      int div = bit >> 5;
      data[div] = data[div] & (~(1 << remainder));
   }

   /**
    * Checks bit state.
    *
    * @param   bit
    *          Bit index.
    *
    * @return  <code>true</code> if the bit is set, <code>false</code> otherwise
    */
   public boolean isSet(int bit)
   {
      int remainder = bit & 0x1F;
      int div = bit >> 5;
      return (data[div] & (1 << remainder)) != 0;
   }

   /**
    * Returns the bit set size.
    *
    * @return  see above
    */
   public int size()
   {
      return bits;
   }
}