/*
** Module   : CustomHookSample.java
** Abstract : authentication hook sample code
**
** Copyright (c) 2005-2017, Golden Code Development Corporation.
**
** -#- -I- --Date-- --JPRM-- ---------------------------- Description ----------------------------
** 001 NVS 20050314   @20317 Created the file. This sample hook uses ID
**                           and password authentication as illustration.
** 002 NVS 20050901   @22445 Changed the decision making logic. Now it
**                           calls a method in SecurityManager.
** 003 NVS 20050914   @22707 Made it to comply with the changes to the
**                           Authenticator interface.
** 004 GES 20061003   @30108 More interface changes and cleanups.
** 005 GES 20061220   @31798 Matched interface changes.
** 006 SVL 20100923          Changed to conform the new Authenticator API.
** 007 OM  20131018          Updated to conform the new Authenticator API.
*/
/*
** This program is free software: you can redistribute it and/or modify
** it under the terms of the GNU Affero General Public License as
** published by the Free Software Foundation, either version 3 of the
** License, or (at your option) any later version.
**
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
** GNU Affero General Public License for more details.
**
** You may find a copy of the GNU Affero GPL version 3 at the following
** location: https://www.gnu.org/licenses/agpl-3.0.en.html
** 
** Additional terms under GNU Affero GPL version 3 section 7:
** 
**   Under Section 7 of the GNU Affero GPL version 3, the following additional
**   terms apply to the works covered under the License.  These additional terms
**   are non-permissive additional terms allowed under Section 7 of the GNU
**   Affero GPL version 3 and may not be removed by you.
** 
**   0. Attribution Requirement.
** 
**     You must preserve all legal notices or author attributions in the covered
**     work or Appropriate Legal Notices displayed by works containing the covered
**     work.  You may not remove from the covered work any author or developer
**     credit already included within the covered work.
** 
**   1. No License To Use Trademarks.
** 
**     This license does not grant any license or rights to use the trademarks
**     Golden Code, FWD, any Golden Code or FWD logo, or any other trademarks
**     of Golden Code Development Corporation. You are not authorized to use the
**     name Golden Code, FWD, or the names of any author or contributor, for
**     publicity purposes without written authorization.
** 
**   2. No Misrepresentation of Affiliation.
** 
**     You may not represent yourself as Golden Code Development Corporation or FWD.
** 
**     You may not represent yourself for publicity purposes as associated with
**     Golden Code Development Corporation, FWD, or any author or contributor to
**     the covered work, without written authorization.
** 
**   3. No Misrepresentation of Source or Origin.
** 
**     You may not represent the covered work as solely your work.  All modified
**     versions of the covered work must be marked in a reasonable way to make it
**     clear that the modified work is not originating from Golden Code Development
**     Corporation or FWD.  All modified versions must contain the notices of
**     attribution required in this license.
*/

package com.goldencode.p2j.security;

import java.io.*;
import java.util.*;
import com.goldencode.p2j.net.SessionListener;
import com.goldencode.p2j.util.*;

/**
 * This class is a sample of a custom authentication hook implementation. It
 * represents the minimum necessary to implement the interface.  Note that
 * no client-side user interface has been provided at all, that is an 
 * exercise left to the reader.
 */
class CustomHookSample
implements Authenticator
{
   /** The value of "option" node from AuthPlugin for this Authenticator */
   private String authOption = null;

   /**
    * Package private constructor. 
    */
   CustomHookSample()
   {
   }
   
   /**
    * Implements client side custom authentication logic. This simply uses
    * <code>stdin</code> and <code>stdout</code> to interact with the user.
    *
    * @param    parameters
    *           Additional configuration parameters.
    * @param    code
    *           The result of the most recent attempt to authenticate or
    *           <code>AUTH_RESULT_NONE</code> if this is the first attempt.
    *
    * @return   Array of bytes to be transmitted to the server as the
    *           authorization input. 
    */
   @Override
   public byte[] clientAuthHook(Map<String, Object> parameters, int code)
   {
      String id = "";
      String pw = ""; 
      
      try
      {
         id = new String(Utils.prompt("User ID  :"));
         pw = new String(Utils.prompt("Password :"));
      }
      
      catch (IOException ioe)
      {
         // allow invalid input to pass through
      }
      
      return SecurityManager.packageIdPassword(id, pw);
   }
   
   /**
    * Implements server side custom authentication logic.
    * <p>
    * Accepts the byte array produced by the client side authentication hook
    * as authentication input as well as any custom parameters.
    *
    * @param    auth
    *           The authorization input from the client.
    * @param    entity
    *           The entity to be authenticated (additionall configuration parameters taken
    *           from the directory).
    *
    * @return   The result of the authentication processing.
    */
   @Override
   public AuthenticationResponse serverAuthHook(byte[] auth, String entity)
   {
      SecurityManager sm = SecurityManager.getInstance();
      return sm.serverAuthHook(auth, entity);
   }

   /**
    * Finalizes any resources allocated during authentication by the client.
    */
   @Override
   public void clientFinalize()
   {
   }

   /**
    * Always returns <code>null</code>.
    *
    * @return  Always <code>null</code>.
    */
   @Override
   public SessionListener getSessionListener()
   {
      return null;
   }

   /**
    * Returns a set of entities that this class handles. The list of entities / parameters will
    * be configured using other ways.
    *
    * @return  A set with the value of option that was configured or an empty set if option was
    *          not configured.
    */
   @Override
   public Set<String> getAuthenticationEntities()
   {
      Set<String> ret = new HashSet<>();
      if (authOption != null)
      {
         ret.add(authOption);
      }
      return ret;
   }

   /**
    * Configures the Authenticator by setting the "option" parameter from directory.xml.
    *
    * @param   option
    *          The value of "option" entry for the auth plugin.
    */
   @Override
   public void configure(String option)
   {
      this.authOption = option;
   }
}