/*
** Module   : Decision.java
** Abstract : Represents cached access decisions.
**
** Copyright (c) 2004-2017, Golden Code Development Corporation.
**
** -#- -I- --Date-- -T- --JPRM-- ----------------Description-----------------
** 001 NVS 20050322 NEW   @20454 Created the initial version. This class 
**                               represents cached access decisions.
** 002 ECF 20080704 CHG   @39142 Minor memory optimization. Use Boolean
**                               constants instead of instantiating new
**                               Boolean objects.
*/ 
/*
** This program is free software: you can redistribute it and/or modify
** it under the terms of the GNU Affero General Public License as
** published by the Free Software Foundation, either version 3 of the
** License, or (at your option) any later version.
**
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
** GNU Affero General Public License for more details.
**
** You may find a copy of the GNU Affero GPL version 3 at the following
** location: https://www.gnu.org/licenses/agpl-3.0.en.html
** 
** Additional terms under GNU Affero GPL version 3 section 7:
** 
**   Under Section 7 of the GNU Affero GPL version 3, the following additional
**   terms apply to the works covered under the License.  These additional terms
**   are non-permissive additional terms allowed under Section 7 of the GNU
**   Affero GPL version 3 and may not be removed by you.
** 
**   0. Attribution Requirement.
** 
**     You must preserve all legal notices or author attributions in the covered
**     work or Appropriate Legal Notices displayed by works containing the covered
**     work.  You may not remove from the covered work any author or developer
**     credit already included within the covered work.
** 
**   1. No License To Use Trademarks.
** 
**     This license does not grant any license or rights to use the trademarks
**     Golden Code, FWD, any Golden Code or FWD logo, or any other trademarks
**     of Golden Code Development Corporation. You are not authorized to use the
**     name Golden Code, FWD, or the names of any author or contributor, for
**     publicity purposes without written authorization.
** 
**   2. No Misrepresentation of Affiliation.
** 
**     You may not represent yourself as Golden Code Development Corporation or FWD.
** 
**     You may not represent yourself for publicity purposes as associated with
**     Golden Code Development Corporation, FWD, or any author or contributor to
**     the covered work, without written authorization.
** 
**   3. No Misrepresentation of Source or Origin.
** 
**     You may not represent the covered work as solely your work.  All modified
**     versions of the covered work must be marked in a reasonable way to make it
**     clear that the modified work is not originating from Golden Code Development
**     Corporation or FWD.  All modified versions must contain the notices of
**     attribution required in this license.
*/

package com.goldencode.p2j.security;

import java.util.*;

/**
 * Associates a resource instance name, access mode and access decision so 
 * that repeatable accesses can reuse the access decision.
 * <p>
 * Access decisions are cached in the maps held by the respective security 
 * contexts. The top level map contains second level maps and has the instance
 * name as a key. Second level maps have access mode as a key and contain 
 * Boolean bearing the decisions.
 */
class Decision
{
   /** resource instance name */
   private String instanceName = null;

   /** access mode AKA requested rights */
   private int mode = 0;

   /** access decision */
   private boolean decision = false;

   /**
    * Package private constructor. 
    *
    * @param instance
    *        resource instance name 
    *
    * @param mode
    *        resource access mode 
    * @param decision
    *        access decision 
    */
   Decision(String instance, int mode, boolean decision)
   {
      this.instanceName = instance;
      this.mode = mode;
      this.decision = decision;
   }

   /**
    * Searches the specified map for {instance}, then searches the returned 
    * map for {mode} and returns the associated decision or null.
    *
    * @param cache
    *        map representing aspecific cache to search 
    *
    * @param instance
    *        resource instance name 
    *
    * @param mode
    *        resource access mode 
    *
    * @return <code>Boolean</code> bearing the decision or <code>null</code>
    */
   static Boolean searchDecisions(Map cache, String instance, int mode)
   {
      if (cache == null)
         return null;

      Map internal = (Map)cache.get(instance);
      if (internal == null)
         return null;

      return (Boolean)internal.get(new Integer(mode));
   }

   /**
    * Removes the access decision from the specified cache map. 
    *
    * @param cache
    *        map representing aspecific cache 
    *
    * @param instance
    *        resource instance name 
    *
    * @param mode
    *        resource access mode 
    */
   static void clearDecisions(Map cache, String instance, int mode)
   {
      if (cache == null)
         return;

      Map internal = (Map)cache.get(instance);
      if (internal == null)
         return;

      internal.remove(new Integer(mode));
   }

   /**
    * Puts the access decision into the specified cache map. 
    *
    * @param cache
    *        map representing aspecific cache to search 
    *
    * @param instance
    *        resource instance name 
    *
    * @param mode
    *        resource access mode 
    *
    * @param decision
    *        value to cache 
    */
   static void cacheDecision(Map cache, String instance, int mode,
                             boolean decision)
   {
      if (cache == null)
         return;

      Map internal = (Map)cache.get(instance);
      if (internal == null)
      {
         internal = new HashMap();
         cache.put(instance, internal);
      }

      internal.put(new Integer(mode),
                   (decision ? Boolean.TRUE : Boolean.FALSE));
   }
}