/*
** Module   : GuestAccess.java
** Abstract : provides open access to the specified P2J account
**
** Copyright (c) 2008-2017, Golden Code Development Corporation.
**
** -#- -I- --Date-- --JPRM-- ---------------------------- Description ----------------------------
** 001 NVS 20081418   @38031 Created the initial version. This class is
**                           a custom P2J authenticator. It provides two
**                           sides: client side login logic and the server
**                           side authentication hook that handles the
**                           input. Both provide unchecked access to the
**                           user account specified with the parameter.
** 002 SVL 20100923          Changed to conform the new Authenticator API.
** 003 LMR 20101105          Changed clientAuthHook() to look up "option" 
**                           entry instead of "subjectId" in the parameter 
**                           map ("subjectId" is used with a different 
**                           plug-in).
** 004 OM  20131018          Updated to conform the new Authenticator API.
*/
/*
** This program is free software: you can redistribute it and/or modify
** it under the terms of the GNU Affero General Public License as
** published by the Free Software Foundation, either version 3 of the
** License, or (at your option) any later version.
**
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
** GNU Affero General Public License for more details.
**
** You may find a copy of the GNU Affero GPL version 3 at the following
** location: https://www.gnu.org/licenses/agpl-3.0.en.html
** 
** Additional terms under GNU Affero GPL version 3 section 7:
** 
**   Under Section 7 of the GNU Affero GPL version 3, the following additional
**   terms apply to the works covered under the License.  These additional terms
**   are non-permissive additional terms allowed under Section 7 of the GNU
**   Affero GPL version 3 and may not be removed by you.
** 
**   0. Attribution Requirement.
** 
**     You must preserve all legal notices or author attributions in the covered
**     work or Appropriate Legal Notices displayed by works containing the covered
**     work.  You may not remove from the covered work any author or developer
**     credit already included within the covered work.
** 
**   1. No License To Use Trademarks.
** 
**     This license does not grant any license or rights to use the trademarks
**     Golden Code, FWD, any Golden Code or FWD logo, or any other trademarks
**     of Golden Code Development Corporation. You are not authorized to use the
**     name Golden Code, FWD, or the names of any author or contributor, for
**     publicity purposes without written authorization.
** 
**   2. No Misrepresentation of Affiliation.
** 
**     You may not represent yourself as Golden Code Development Corporation or FWD.
** 
**     You may not represent yourself for publicity purposes as associated with
**     Golden Code Development Corporation, FWD, or any author or contributor to
**     the covered work, without written authorization.
** 
**   3. No Misrepresentation of Source or Origin.
** 
**     You may not represent the covered work as solely your work.  All modified
**     versions of the covered work must be marked in a reasonable way to make it
**     clear that the modified work is not originating from Golden Code Development
**     Corporation or FWD.  All modified versions must contain the notices of
**     attribution required in this license.
*/

package com.goldencode.p2j.security;

import com.goldencode.p2j.net.SessionListener;
import java.io.*;
import java.util.*;

/**
 * The class to request authentication to a specified user account without
 * providing password.
 */
public class GuestAccess
implements Authenticator
{
   /** The value of "option" node from AuthPlugin for this Authenticator */
   private String authOption = null;

   /**
    * Default constructor. 
    */
   public GuestAccess()
   {
      // nothing to do here
   }

   /**
    * Implements client side custom authentication logic.
    *
    * @param    parameters
    *           Additional configuration parameters. This plugin uses
    *           "subjectId" parameter which corresponds the name of the user
    *           account to be used.
    * @param    code
    *           The result of the most recent attempt to authenticate or
    *           <code>AUTH_RESULT_NONE</code> if this is the first attempt.
    *
    * @return   Array of bytes to be transmitted to the server as the
    *           authorization input. 
    */
   @Override
   public byte[] clientAuthHook(Map<String, Object> parameters, int code)
   {
      String id = "";
      if (parameters != null)
         id = (String) parameters.get("option");
      
      if (id == null)
         id = "";
      
      String pw = "";

      // return a byte array
      return SecurityManager.packageIdPassword(id, pw);
   }
   
   /**
    * Finalizes any resources allocated during authentication by the client.
    */
   @Override
   public void clientFinalize()
   {
      // nothing to do here
   }

   /**
    * Not used for the client side.
    *
    * @param    auth
    *           The authorization input from the client.
    * @param    entity
    *           The entity to be authenticated.
    *           Additional configuration parameters taken from the directory.
    *
    * @return   Always <code>null</code>.
    */
   @Override
   public AuthenticationResponse serverAuthHook(byte[] auth, String entity)
   {
      // parse the byte array input
      ByteArrayInputStream bis = new ByteArrayInputStream(auth);
      DataInputStream      dis = new DataInputStream(bis);
      
      String userId = null;
      
      try
      {
         userId = dis.readUTF();
      }
      catch (IOException ioe)
      {
         return new AuthenticationResponse(null, AUTH_RESULT_UNSPECIFIED_FAILURE);
      }
      
      // Safely query cached data
      SecurityCache sc =  SecurityManager.getInstance().getCache(); 

      // locate the account and its password hash
      Account acc = sc.getAccountById(userId);
      
      if (acc == null)
      {
         return new AuthenticationResponse(null, AUTH_RESULT_INVALID_USERID);
      }
      
      if (acc.getAccountType() != Account.ACC_USER)
      {
         return new AuthenticationResponse(null, AUTH_RESULT_UNSPECIFIED_FAILURE);
      }
      
      UserAccount user = (UserAccount) acc;
      
      // no password verification 
      return new AuthenticationResponse(userId, AUTH_RESULT_SUCCESS);
   }

   /**
    * Always returns <code>null</code>.
    *
    * @return  Always <code>null</code>.
    */
   @Override
   public SessionListener getSessionListener()
   {
      return null;
   }

   /**
    * Returns a set of entities that this class handles. The list of entities / parameters will
    * be configured using other ways.
    *
    * @return  A set with the value of option that was configured or an empty set if option was
    *          not configured.
    */
   @Override
   public Set<String> getAuthenticationEntities()
   {
      Set<String> ret = new HashSet<>();
      if (authOption != null)
      {
         ret.add(authOption);
      }
      return ret;
   }

   /**
    * Configures the Authenticator by setting the "option" parameter from directory.xml.
    *
    * @param   option
    *          The value of "option" entry for the auth plugin.
    */
   @Override
   public void configure(String option)
   {
      this.authOption = option;
   }
}