/*
** Module   : BaseConfig.java
** Abstract : container for the most common widget state
**
** Copyright (c) 2005-2018, Golden Code Development Corporation.
**
** -#- -I- --Date-- --JPRM-- ----------------------------Description-----------------------------
** 001 EVL 20050715   @21704 Creation based on general concept. Moving
**                           the widget type constants to eliminate the
**                           unusable class WidgetTransferrable.
** 002 EVL 20050729   @21909 To perform automatic layout usage the row
**                           and column default values must be 0.
** 003 SIY 20050831   @22485 Refactoring (see @@22394-22409).
** 004 SIY 20050907   @22621 Fixed use of @inheritDoc tag in javadoc,
**                           Temporary reverting to 0-based coordinates.
** 005 EVL 20050919   @22752 Excluding the widget's name from transferring
**                           through the network.
** 006 SIY 20050920   @22818 Moved setSensitive() below setWidthChars() to
**                           make sure that it will not cause any negative 
**                           side effects, updated support for SENSITIVE
**                           and SELECTED attributes, using ColorTable for 
**                           the colors.
** 007 SIY 20051129   @23552 Changed type of PARENT attribute to 
**                           GenericWidget, changed type of PRIVATE-DATA
**                           to String.
** 008 SIY 20051130   @23586 Added support for ENTERED/NOT ENTERED
**                           property.
** 009 GES 20060116   @23945 Cleanup of some docs and formatting.
** 010 SIY 20060203   @24237 Support for TO option.
** 011 SIY 20060209   @24486 Major refactoring.
** 012 NVS 20060310   @24966 setModified() method also changes the widget
**                           state to CHANGED.
** 013 SIY 20060412   @25595 Encapsulated 'state' field.
** 014 SIY 20060424   @25679 Added getter/setter for parent ID.
** 015 SIY 20060621   @27534 Removed 'transient' specification for name 
**                           attribute.
** 016 SIY 20060628   @27658 Moved 'title' attribute from ContainerConfig. 
** 017 GES 20060803   @28388 Made externalizable for better performance.
**                           Many code standard cleanups.
** 018 SIY 20061102   @30937 Simplified and generalized for TO/COLON/AT 
**                           options.
** 019 SIY 20061103   @30967 Set coordinate when instance is available
**                           even if alignment is not ALIGN_LEFT.
** 020 NVS 20061204   @31550 "selected" flag is no longer part of the
**                           configuration.
** 021 SIY 20070214   @32129 Added isLeftAligned() method.
** 022 NVS 20070426   @33246 SENSITIVE attribute is no longer maintained
**                           in the configuration.
** 023 SIY 20070627   @34277 Do not set DCOLOR to -1.
** 024 SIY 20070627   @34810 Added accessor for true column value.
** 025 SIY 20080429   @38140 Added saveColors()/restoreColors() methods.
** 026 CA  20080905   @39744 Negative coordinates can be received by the 
**                           client, so INV_COORD was changed to a
**                           smaller value. Fixed COLUMN/ROW getters.
** 027 SIY 20090313   @41595 Added commend about INV_COORD value. 
** 028 SIY 20100619          Switching to own implementation of character UI.
** 029 SIY 20100804          Changes in imports, some refactorings.
** 030 SIY 20100822          Refactored AttributeContainer support.
** 031 SIY 20100915          locateFrame() is moved to UiUtils.
** 032 SIY 20110112          Changes related to new color handling scheme.
** 033 SIY 20111101          Moved from CHUI.
** 034 SIY 20111115          Added default font constant.
** 035 SIY 20111221          Minor cleanups.
** 035 MAG 20140410          Add SIZE-PIXELS, FGCOLOR, BGCOLOR, TOOLTIP options.
** 036 MAG 20140512          Enhanced color support.
** 037 CA  20140711          Enhanced font support, for GUI clients.
** 038 CA  20140805          Changes to isolate ChUI-specific code from common code: added 
**                           generics.
** 039 MAG 20140912          Change FGCOLOR and BGCOLOR to integer types.
** 040 HC  20140911          GUI coordinate support - data types refactoring.
** 041 CA  20140926          Added shared widget configuration support, to allow GUI/ChUI
**                           concrete implementation for the same widget ID.  Refactored the
**                           widget configuration classes: all fields were made public; these
**                           classes need to be as dumb as possible, all logic related to
**                           setting/getting a certain field should be at the widget or at the
**                           caller. Refs #2254
** 042 MAG 20141006          Set default value to -1 for BGCOLOR and FGCOLOR
** 043 HC  20141021          Implemented window move/position support.
** 044 HC  20141207          Implemented runtime support for widget coordinate attributes.
** 045 GES 20141124          Added movable attr, enabled selected attr.
** 046 HC  20150102          Fixed handling of uninitialized values for the coordinate fields.
** 047 GES 20150104          Added a TODO.
** 048 HC  20150128          Added support for frame widget position changes induced by setting
**                           ROW/COL/X/Y attributes.
** 049 EVL 20150225          X and Y coordinates need to be set by special methods in
**                           applyConfig().
** 050 CA  20150323          Added flags to track if widget dimension is set outside of the frame
**                           setup.
** 051 HC  20150323          Removed business logic from config classes.
** 052 VIG 20150407          The attribute "modified" was moved to WidgetConfig.
** 053 CA  20150725          Added FRAME-COL, FRAME-ROW, FRAME-X, FRAME-Y support: for down or top
**                           label frames, an offset from which the widget has been positioned 
**                           (relative to the frame's top label header) is maintained.  These 
**                           offsets are used so that the FRAME- attributes return the location 
**                           from the frame's top-left corner, while the X,Y,ROW,COL attributes
**                           return the location from their parent field-group.
** 054 CA  20151019          Track widget placement on frames (to fix H053 when the widget has its
**                           location set before or after is attached to a frame).
**                           Fixed auto-resize for FILL-IN widgets.
** 055 HC  20151110          Added client-side size fields.
** 056 HC  20160304          Fixed serialization of client size config fields.
** 057 EVL 20170703          Adding RESIZABLE attribute runtime support.
** 058 CA  20171111          Added WIDGET:BGCOLOR-RGB and WIDGET:FGCOLOR-RGB FWD extension
**                           attributes
** 059 CA  20180208          Added locationChanged flag to track location assigned by server-side.
*/
/*
** This program is free software: you can redistribute it and/or modify
** it under the terms of the GNU Affero General Public License as
** published by the Free Software Foundation, either version 3 of the
** License, or (at your option) any later version.
**
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
** GNU Affero General Public License for more details.
**
** You may find a copy of the GNU Affero GPL version 3 at the following
** location: https://www.gnu.org/licenses/agpl-3.0.en.html
** 
** Additional terms under GNU Affero GPL version 3 section 7:
** 
**   Under Section 7 of the GNU Affero GPL version 3, the following additional
**   terms apply to the works covered under the License.  These additional terms
**   are non-permissive additional terms allowed under Section 7 of the GNU
**   Affero GPL version 3 and may not be removed by you.
** 
**   0. Attribution Requirement.
** 
**     You must preserve all legal notices or author attributions in the covered
**     work or Appropriate Legal Notices displayed by works containing the covered
**     work.  You may not remove from the covered work any author or developer
**     credit already included within the covered work.
** 
**   1. No License To Use Trademarks.
** 
**     This license does not grant any license or rights to use the trademarks
**     Golden Code, FWD, any Golden Code or FWD logo, or any other trademarks
**     of Golden Code Development Corporation. You are not authorized to use the
**     name Golden Code, FWD, or the names of any author or contributor, for
**     publicity purposes without written authorization.
** 
**   2. No Misrepresentation of Affiliation.
** 
**     You may not represent yourself as Golden Code Development Corporation or FWD.
** 
**     You may not represent yourself for publicity purposes as associated with
**     Golden Code Development Corporation, FWD, or any author or contributor to
**     the covered work, without written authorization.
** 
**   3. No Misrepresentation of Source or Origin.
** 
**     You may not represent the covered work as solely your work.  All modified
**     versions of the covered work must be marked in a reasonable way to make it
**     clear that the modified work is not originating from Golden Code Development
**     Corporation or FWD.  All modified versions must contain the notices of
**     attribution required in this license.
*/

package com.goldencode.p2j.ui;

import java.io.*;

import com.goldencode.p2j.ui.client.*;

/**
 * The container for the most common widget state variables, except some special cases.
 * <p>
 * All the positional fields in the class hold either characters or pixels. The character fields
 * are: {@link #initColumn}, {@link #initRow}, {@link #column}, {@link #row},
 * {@link #clientColumn}, {@link #clientRow}, {@link #widthChars}, {@link #heightChars}. The pixel
 * fields are: {@link #initX}, {@link #initY}, {@link #x}, {@link #y}, {@link #clientX},
 * {@link #clientY}, {@link #widthPixels}, {@link #heightPixels}.
 * <p>
 * In addition to character and pixel units the client also deals with native (also called
 * physical) units. Native units are used by the output subsystems. The conversion from
 * pixels/characters to native units and also between character and pixel values is handled by
 * {@link CoordinatesConversion} class. Also, assignments to the positional fields in this class
 * are converted automatically by {@link com.goldencode.p2j.aspects.ui.SyncCoordinatesAspect}, for
 * example character value assigned to {@link #column} is automatically converted to pixel value
 * and assigned to {@link #x}.
 * <p>
 * Do we need to convert character to pixel values in ChUI mode and pixel to character values in
 * GUI mode? Yes we do. In 4GL character attributes are synchronized with the pixel counterpart
 * attributes and vice versa regardless whether in ChUI or GUI mode.
 * <p>
 * Why is there so many derivatives of the positional fields? Why is there {@link #initColumn},
 * {@link #column}, {@link #clientColumn} and not just one "column" field? The reason is the
 * ability to encapsulate all the layout inputs in config instances and to honor the "single
 * responsibility principle". That is, limit the number of different semantics for the positional
 * fields depending on actual context.
 * <p>
 * The initial* fields hold the values assigned during frame definition. They must be stored
 * separately from values assigned to positional 4GL attributes outside of frame definition as
 * they have special meaning during frame layout. The client* fields are used to communicate the
 * actual client widget location back to server, which may differ from the server-assigned
 * location.
 */
public class BaseConfig
extends WidgetConfig
{
   /** 
    * Invalid coordinate value.   
    */
   // Note that changes of constant value must be properly reflected in
   // ScreenBitmap.INV_COORD
   public static final int INV_COORD = Integer.MIN_VALUE;
   
   /** The integer index of the parent in the array of configs. */ 
   public int parentId = -1;
   
   /** Flag indicating if the location has been set by the server-side. */
   public boolean locationChanged = false;
   
   /**
    * The initial COLUMN position of the widget set during frame setup
    * phase.
    * <p>
    * Use {@link ConfigHelper} to access the positional fields.
    */
   public double initColumn = INV_COORD;
   
   /**
    * The initial ROW position of the widget set during frame setup
    * phase.
    * <p>
    * Use {@link ConfigHelper} to access the positional fields.
    */
   public double initRow = INV_COORD;
   
   /**
    * The initial X position of the widget set during frame setup
    * phase.
    * <p>
    * Use {@link ConfigHelper} to access the positional fields.
    */
   public int initX = INV_COORD;
   
   /**
    * The initial Y position of the widget set during frame setup
    * phase.
    * <p>
    * Use {@link ConfigHelper} to access the positional fields.
    */
   public int initY = INV_COORD;
   
   /** 
    * The server-assigned COLUMN position of the widget. 
    * <p>
    * Use {@link ConfigHelper} to access the positional fields.
    */
   public double column = INV_COORD;
   
   /** 
    * The server-assigned ROW position of the widget. 
    * <p>
    * Use {@link ConfigHelper} to access the positional fields.
    */
   public double row = INV_COORD;
   
   /** 
    * The server-assigned X position of the widget in pixel units.
    * <p>
    * Use {@link ConfigHelper} to access the positional fields.
    */
   public int x = INV_COORD;
   
   /** 
    * The server-assigned Y position of the widget in pixel units.
    * <p>
    * Use {@link ConfigHelper} to access the positional fields.
    */
   public int y = INV_COORD;
   
   /**
    * An offset from the left-side of the frame from which the widget is determined to be drawn,
    * at initial frame layout.  This column offset value is in pixel units.
    */
   public int frameXOffset = 0;

   /**
    * An offset from the left-side of the frame from which the widget is determined to be drawn,
    * at initial frame layout.  This column offset value is in character units.
    */
   public double frameColumnOffset = 0;

   /**
    * An offset from the top of the frame from which the widget is determined to be drawn,
    * at initial frame layout.  This row offset value is in pixel units.
    */
   public int frameYOffset = 0;

   /**
    * An offset from the top of the frame from which the widget is determined to be drawn,
    * at initial frame layout.  This row offset value is in character units.
    */
   public double frameRowOffset = 0;
   
   /**
    * The client COLUMN position of the widget.
    * <p>
    * The "client" positional fields reflect the actual position set
    * by client. This covers the cases when the server-assigned (and
    * reported) position differs with the position calculated by
    * client. The "client" positional fields are never assigned by
    * server.
    * <p>
    * Use {@link ConfigHelper} to access the positional fields.
    */
   public double clientColumn = INV_COORD;
   
   /** 
    * The client ROW position of the widget. 
    * <p>
    * The "client" positional fields reflect the actual position set
    * by client. This covers the cases when the server-assigned (and
    * reported) position differs with the position calculated by
    * client. The "client" positional fields are never assigned by
    * server.
    * <p>
    * Use {@link ConfigHelper} to access the positional fields.
    */
   public double clientRow = INV_COORD;
   
   /** 
    * The client X position of the widget in pixel units. 
    * <p>
    * The "client" positional fields reflect the actual position set
    * by client. This covers the cases when the server-assigned (and
    * reported) position differs with the position calculated by
    * client. The "client" positional fields are never assigned by
    * server.
    * <p>
    * Use {@link ConfigHelper} to access the positional fields.
    */
   public int clientX = INV_COORD;
   
   /** 
    * The client Y position of the widget in pixel units.
    * <p>
    * The "client" positional fields reflect the actual position set
    * by client. This covers the cases when the server-assigned (and
    * reported) position differs with the position calculated by
    * client. The "client" positional fields are never assigned by
    * server.
    * <p>
    * Use {@link ConfigHelper} to access the positional fields.
    */
   public int clientY = INV_COORD;
   
   /** 
    * This flag is set to true when WIDTH-CHARS or WIDTH-PIXELS attribute is set outside of the 
    * frame setup phase.
    */
   public boolean fixedWidth = false;

   /** 
    * This flag is set to true when HEIGHT-CHARS or HEIGHT-PIXELS attribute is set outside of the 
    * frame setup phase.
    */
   public boolean fixedHeight = false;
   
   /** The color is used to render the widget. */
   public Color dcolor = null;

   /** The color of the widget under the focus. */
   public Color pfcolor = Color.INPUT;

   /** The height of the widget. */
   public double heightChars = 0;
   
   /** The width of the widget. */
   public double widthChars = 0;

   /** 
    * The height of the widget as calculated by the client. 
    * This extra field is needed when the actual widget size 
    * (as calculated by the client) differs from the assigned size.
    */
   public double clientHeightChars = BaseConfig.INV_COORD;
   
   /** 
    * The width of the widget as calculated by the client.
    * This extra field is needed when the actual widget size 
    * (as calculated by the client) differs from the assigned size.
    */
   public double clientWidthChars = BaseConfig.INV_COORD;
   
   /** Rectangle width pixels */
   public int widthPixels;
   
   /** Rectangle height pixels */
   public int heightPixels;
   
   /** 
    * The width of the widget as calculated by the client.
    * This extra field is needed when the actual widget size 
    * (as calculated by the client) differs from the assigned size.
    */
   public int clientWidthPixels = BaseConfig.INV_COORD;
   
   /** 
    * The height of the widget as calculated by the client.
    * This extra field is needed when the actual widget size 
    * (as calculated by the client) differs from the assigned size.
    */
   public int clientHeightPixels = BaseConfig.INV_COORD;
   
   /** Set to true if component is created dynamically by the runtime. */
   public boolean dynamic = false;
   
   /** The tab-stop attribute. */
   public boolean tabStop = true;
   
   /** The name attribute. */
   public String name = "";
   
   /** The title of the container. */
   public String title = "";

   /** The title font of the container. */
   public int titleFont = FontManager.UNDEFINED_FONT;

   /** TITLE-DCOLOR attribute variable. */
   public Color titleDColor = Color.MESSAGES;

   /** Widget alignment. */
   public int align = ControlEntity.ALIGN_LEFT;
   
   /** Font number. Corresponds to FONT attribute. */
   public int font = FontManager.UNDEFINED_FONT;
   
   /** Flag indicating if font has been set via the widget's FONT attribute. */
   public boolean fontChanged = false;
   
   /** TOOLTIP option */
   public String tooltip;   
         
   /** FGCOLOR option */
   public int fgcolor = -1;

   /** BGCOLOR option */
   public int bgcolor = -1;
   
   /** FGCOLOR-RGB option */
   public int fgcolorRgb = -1;

   /** BGCOLOR-RGB option */
   public int bgcolorRgb = -1;
   
   /** default FGCOLOR option */
   public int sysfgcolor = -1;

   /** default BGCOLOR option */
   public int sysbgcolor = -1;
   
   /** SELECTED attribute. TODO: its client-side runtime is not implemented. */
   public boolean selected = false;
   
   /** MOVABLE attribute. */
   public boolean movable = false;
   
   /** RESIZABLE attribute. */
   public boolean resizable = false;
   
   /**
    * Flag indicating if the widget has been placed by the layout.  For top-label frames, 
    * this is required to distinguish between widgets which have their location set before or 
    * after they were attached to a frame.
    */
   public transient boolean widgetPlaced = false;
   
   /**
    * Default constructor (only used in deserialization).
    */
   public BaseConfig()
   {
   }

   /**
    * Create a new config and associate it with the given widget.
    * 
    * @param    id
    *           The ID of the widget to which this configuration belongs.
    */
   protected BaseConfig(int id)
   {
      super(id);
   }

   /**
    * Create a new config and associate it with the given widget.
    * 
    * @param    id
    *           The ID of the widget to which this configuration belongs.
    */
   protected BaseConfig(WidgetId id)
   {
      super(id);
   }

   /**
    * Set new values from the provided configuration instance.
    * 
    * @param    config
    *           The instance from which to take values.
    */
   @Override
   public void applyConfig(WidgetConfig config)
   {
      super.applyConfig(config);
      
      BaseConfig cfg = (BaseConfig) config;
      
      parentId = cfg.parentId;
      locationChanged = cfg.locationChanged;
      initColumn = cfg.initColumn;
      initRow = cfg.initRow;
      initX = cfg.initX;
      initY = cfg.initY;
      column = cfg.column;
      row = cfg.row;
      x = cfg.x;
      y = cfg.y;
      clientColumn = cfg.clientColumn;
      clientRow = cfg.clientRow;
      clientX = cfg.clientX;
      clientY = cfg.clientY;
      fixedWidth = cfg.fixedWidth;
      fixedHeight = cfg.fixedHeight;
      dcolor = cfg.dcolor;
      pfcolor = cfg.pfcolor;
      heightChars = cfg.heightChars;
      widthChars = cfg.widthChars;
      widthPixels = cfg.widthPixels;
      heightPixels = cfg.heightPixels;
      clientHeightChars = cfg.clientHeightChars;
      clientWidthChars = cfg.clientWidthChars;
      clientWidthPixels = cfg.clientWidthPixels;
      clientHeightPixels = cfg.clientHeightPixels;
      dynamic = cfg.dynamic;
      tabStop = cfg.tabStop;
      name = cfg.name;
      title = cfg.title;
      titleFont = cfg.titleFont;
      titleDColor = cfg.titleDColor;
      align = cfg.align;
      font = cfg.font;
      fontChanged = cfg.fontChanged;
      tooltip = cfg.tooltip;
      fgcolor = cfg.fgcolor;
      bgcolor = cfg.bgcolor;
      fgcolorRgb = cfg.fgcolorRgb;
      bgcolorRgb = cfg.bgcolorRgb;
      sysfgcolor = cfg.sysfgcolor;
      sysbgcolor = cfg.sysbgcolor;
      selected = cfg.selected;
      movable = cfg.movable;
      resizable = cfg.resizable;
      frameColumnOffset = cfg.frameColumnOffset;
      frameRowOffset = cfg.frameRowOffset;
      frameXOffset = cfg.frameXOffset;
      frameYOffset = cfg.frameYOffset;
   }
   
   /**
    * Replacement for the default object reading method. The latest
    * state is read from the input source.  The parent's state is
    * read first.
    * 
    * @param    in
    *           The input source from which fields will be restored.
    *
    * @throws   IOException
    *           In case of I/O errors.
    * @throws   ClassNotFoundException
    *           If payload can't be instantiated.
    */
   @Override
   public void readExternal(ObjectInput in)
   throws IOException,
          ClassNotFoundException
   {
      super.readExternal(in);
      parentId       = in.readInt();
      locationChanged = in.readBoolean();
      initColumn     = in.readDouble();
      initRow        = in.readDouble();
      initX          = in.readInt();
      initY          = in.readInt();
      column         = in.readDouble();
      row            = in.readDouble();
      x              = in.readInt();
      y              = in.readInt();
      clientColumn   = in.readDouble();
      clientRow      = in.readDouble();
      clientX        = in.readInt();
      clientY        = in.readInt();
      fixedWidth   = in.readBoolean();
      fixedHeight  = in.readBoolean();
      font         = in.readInt();
      fontChanged  = in.readBoolean();
      dcolor       = (Color) in.readObject();
      pfcolor      = (Color) in.readObject();
      fgcolor      = in.readInt();
      bgcolor      = in.readInt();
      fgcolorRgb   = in.readInt();
      bgcolorRgb   = in.readInt();
      heightChars  = in.readDouble();
      widthChars   = in.readDouble();
      widthPixels  = in.readInt();
      heightPixels = in.readInt();
      clientHeightChars  = in.readDouble();
      clientWidthChars   = in.readDouble();
      clientWidthPixels  = in.readInt();
      clientHeightPixels = in.readInt();
      align        = in.readInt();
      dynamic      = in.readBoolean();
      tabStop      = in.readBoolean();
      name         = (String) in.readObject();
      title        = (String) in.readObject();
      tooltip      = (String) in.readObject();
      titleFont    = in.readInt();
      titleDColor  = (Color) in.readObject();
      selected     = in.readBoolean();
      movable      = in.readBoolean();
      resizable      = in.readBoolean();
      frameColumnOffset = in.readDouble();
      frameRowOffset    = in.readDouble();
      frameXOffset      = in.readInt();
      frameYOffset      = in.readInt();
   }

   /**
    * Replacement for the default object writing method. The latest
    * state is written to the output destination.  The parent's state is
    * written first.
    * 
    * @param    out
    *           The output destination to which fields will be saved.
    *
    * @throws   IOException
    *           In case of I/O errors.
    */
   @Override
   public void writeExternal(ObjectOutput out)
   throws IOException
   {
      super.writeExternal(out);
      out.writeInt(parentId);
      out.writeBoolean(locationChanged);
      out.writeDouble(initColumn);
      out.writeDouble(initRow);
      out.writeInt(initX);
      out.writeInt(initY);
      out.writeDouble(column);
      out.writeDouble(row);
      out.writeInt(x);
      out.writeInt(y);
      out.writeDouble(clientColumn);
      out.writeDouble(clientRow);
      out.writeInt(clientX);
      out.writeInt(clientY);
      out.writeBoolean(fixedWidth);
      out.writeBoolean(fixedHeight);
      out.writeInt(font);
      out.writeBoolean(fontChanged);
      out.writeObject(dcolor);
      out.writeObject(pfcolor);
      out.writeInt(fgcolor);
      out.writeInt(bgcolor);
      out.writeInt(fgcolorRgb);
      out.writeInt(bgcolorRgb);
      out.writeDouble(heightChars);
      out.writeDouble(widthChars);
      out.writeInt(widthPixels);
      out.writeInt(heightPixels);
      out.writeDouble(clientHeightChars);
      out.writeDouble(clientWidthChars);
      out.writeInt(clientWidthPixels);
      out.writeInt(clientHeightPixels);
      out.writeInt(align);
      out.writeBoolean(dynamic);
      out.writeBoolean(tabStop);
      out.writeObject(name);
      out.writeObject(title);
      out.writeObject(tooltip);
      out.writeInt(titleFont);
      out.writeObject(titleDColor);
      out.writeBoolean(selected);
      out.writeBoolean(movable);
      out.writeBoolean(resizable);
      out.writeDouble(frameColumnOffset);
      out.writeDouble(frameRowOffset);
      out.writeInt(frameXOffset);
      out.writeInt(frameYOffset);
   }
}