/*
** Module   : BaseEntity.java
** Abstract : server side base class for most widgets
**
** Copyright (c) 2005-2018, Golden Code Development Corporation.
**
** -#- -I- --Date-- --JPRM-- ----------------------------Description-----------------------------
** 001 NVS 20051026   @23479 Created initial version.
** 002 GES 20060116   @23947 Implemented conversion of 1-based to 0-based positions.
** 003 SIY 20060209   @24432 Added missing method for width setting, other minor changes.
** 004 SIY 20060424   @25680 Changed setParent() signature.
** 005 GES 20060803   @28363 Major cleanup and refactoring.
** 006 NVS 20061204   @31549 SELECTED attribute is server side only from now on.
** 007 NVS 20070426   @33241 SENSITIVE attribute is no longer maintained separately, but is backed
**                           up with ENABLED flag of CHARVA on the client.
** 008 NVS 20080307   @37409 setRow() and setColumn() methods now support dynamic settings that is
**                           required for ROW and COL widget attributes support.
** 009 GES 20080319   @37579 Added NEXT-CHILD and PREV-CHILD support.
** 010 CA  20080925   @39960 Support for dynamic browse title - base field is now protected.
** 011 NVS 20081211   @40872 getParent() method takes into account field group the widget is in.
** 012 GES 20081230   @41001 Return a NullWidget instance instead of null for safety.
** 013 SIY 20101004          Added necessary method stubs.
** 014 SIY 20101005          Generalized setX/setY.
** 015 SVL 20101006          Added FGCOLOR, BGCOLOR attributes.
** 016 SIY 20101007          Use NumberType instead of decimal.
** 017 SIY 20110112          Changes related to new color handling scheme.
** 018 SIY 20111101          Updated imports.
** 019 CA  20130112          getNextSibling and getPrevSibling must return a handle instance.
** 020 CA  20130222          Added NPE protection for cases when a widget is used before is 
**                           attached to a frame (after a CREATE WIDGET statement is used).
** 021 CA  20130323          The parent must be a CommonWidget, not Object.
** 022 OM  20130527          Added int64 support for colors.
** 023 CA  20130927          Fixed getNext/PrevSibling - NullWidget is not needed.
** 024 CA  20131115          Changes related to properly instantiate dynamic resources.
** 025 HC  20131215          Removed PRIVATE-DATA attribute, it is already implemented 
**                           in the base class.
** 026 MAG 20140410          Implements FGCOLOR, BGCOLOR, TOOLTIP and SIZE-PIXELS attributes.           
** 027 VIG 20140314          Changed constructor signature, adding new "before" semantic in
**                           the logic of inserting new BaseEntity to HandleChain.
**                           Added possibility to use HandleChain siblings methods.
** 028 MAG 20140508          Set FGCOLOR, BGCOLOR using integer values.           
** 029 MAG 20140512          Enhanced color support.           
** 030 CA  20140711          Enhanced font support, for GUI clients.
** 031 MAG 20140912          Change FGCOLOR and BGCOLOR to integer types.
** 032 HC  20140911          GUI coordinate support - data types refactoring.
** 033 CA  20140926          Added shared widget configuration support, to allow GUI/ChUI
**                           concrete implementation for the same widget ID.  Refactored the
**                           widget configuration classes: all fields were made public; these
**                           classes need to be as dumb as possible, all logic related to
**                           setting/getting a certain field should be at the widget or at the
**                           caller. Refs #2254
** 034 CA  20141104          Removed setHeight/Width: all widgets need to use the explicit char or
**                           pixel unit APIs.
** 035 HC  20141021          Implemented window move/position support.
** 036 EVL 20141111          Adding getter methods for DCOLOR, PFCOLOR attributes.
** 037 HC  20141207          Implemented runtime support for widget coordinate attributes.
** 038 GES 20141124          Added movable attr, enabled selected attr. Moved setColon() from
**                           ControlEntity since customer code shows it can be used widely for
**                           other widgets (undocumented feature).
** 039 EVL 20141217          Adding TOOLTIP setter implementation for character variable.
** 040 HC  20150111          Fixed widget position change in a frame.
** 041 HC  20150128          Added support for frame widget position changes induced by
**                           setting ROW/COL/X/Y attributes.
** 042 CA  20150219          Allow the ROW/COL/X/Y to be set for dynamic widgets not yet attached
**                           to a frame.
** 043 CA  20150222          Fixed label's COLUMN and WIDTH-CHARS attributes (in ChUI).
** 044 EVL 20150224          Sometimes we need to know if the widget is dynamic on the client
**                           side.  So save the flag inside config container.
** 045 CA  20150323          Track changes in the widget dimension outside of the frame setup.
** 046 HC  20150323          Removed business logic from config classes.
** 047 CA  20150416          Fixed server-side support for PARENT attribute.
**                           Fixed support for NEXT/PREV-SIBLING in relation with tree widget
**                           structures.
** 048 CA  20150722          Added some TODOs related to widget positioning and dimension.
** 049 CA  20150725          Added FRAME-COL, FRAME-ROW, FRAME-X, FRAME-Y support: for down or top
**                           label frames, an offset from which the widget has been positioned 
**                           (relative to the frame's top label header) is maintained.  These 
**                           offsets are used so that the FRAME- attributes return the location 
**                           from the frame's top-left corner, while the X,Y,ROW,COL attributes
**                           return the location from their parent field-group.
**                           Changed canSetPixelPositionInChui to return true by default: only
**                           WindowWidget should return false.
** 050 HC  20150803          Improved runtime support for WINDOW size attributes.
** 051 IAS 20150804          Added TAB-STOP attribute support.
** 052 IAS 20150923          Added check for ChUI mode.
** 053 OM  20150924          setAlign() pulled from ControlEntity.
** 054 EVL 20151016          Adding TODO entry to explain size settings in attributes different
**                           considerations between frames and other widgets.
**                           Fixed auto-resize for FILL-IN widgets.
** 055 CA  20151026          Allow dimensions between 0 and 1.  Fixed setting the PARENT attribute.
**                           If the widget is not yet realized and its width/height is set, then
**                           the virtual width/height is adjusted, too; also, the widget's 
**                           dimension is marked as fixed.
**     EVL 20151030          Adding check for the new widget location to not to exceed frame size
**                           while setting row or column.  4gl does not allow this change once
**                           the owning frame become realized.
**     HC  20151110          Added client-side size fields.
**     EVL 20151110          Fix for getWidthChars() call.  Also now we consider as true the
**                           fixed(Width|Height) flag no matter when the size was explicitly set,
**                           inside the widget setup or later via the attribute.
** 056 HC  20151213          Improved DIALOG-BOX window parenting logic and related changes.
** 057 HC  20160203          Improved runtime support of FRAME:VIRTUAL* attributes.
** 058 EVL 20160224          Javadoc fixes to make compatible with Oracle Java 8 for Solaris 10.
** 059 HC  20160304          Fixed various window sizing/positioning issues. Window layout
**                           improvements.
** 060 CA  20160303          Made attribute setters be a no-op if the value doesn't change.
** 061 EVL 20160411          Fix for coordinate change issue after widget realzation when virtual
**                           size of the frame the widget belongs to is not initialized.
** 062 GES 20160311          Removed getHeight() and getWidth() which were dead code.
** 063 HC  20160922          Fixed positioning of nested frames.
** 064 CA  20161123          Centralized the logic for the COL/ROW/X/Y setters in GenericFrame, to
**                           use the APIs from FrameWidget.
** 065 HC  20170131          Added runtime support for dynamic MENU* widgets.
**     CA  20161101          Fixed X/Y attribute setters for dynamic widgets, which have their 
**                           FRAME set after these attributes are set.
** 066 CA  20170320          Removed the color save/restoration  code - it's no longer useful,
**                           and it breaks down frames after the improvements for down frame
**                           configs.
** 067 EVL 20170627          Unlock default support for MOVABLE attribute.  Add support for
**                           RESIZABLE attribute.
** 068 SVL 20170813          Refactored so common code can be used for ROW-HEIGHT attribute.
** 069 HC  20171003          Changed several errors to warnings because they should be displayed
**                           in warning alert boxes (in interactive session) and should be
**                           sensitive to SESSION:SUPPRESS-WARNINGS.
** 070 SBI 20161011          Changed setTabStop to add this widget at the end of its tab item list
**                           if this widget was a TAB-STOP disabled.
** 071 CA  20171111          Added WIDGET:BGCOLOR-RGB and WIDGET:FGCOLOR-RGB FWD extension
**                           attributes
**     CA  20171122          Avoid a trip to the client when checking redirection state.
** 072 SBI 20171204          Fixed setTabStop to propagate changes to clients if it has been
**                           realized.
** 073 CA  20180130          Replaced pushScreenDefinition() with pushWidgetAttr() for cases when
**                           business logic is assigning a widget attribute.
** 074 HC  20180205          Removed redundant rounding methods from Coordinate class.
** 075 CA  20180208          Added locationChanged flag to track location assigned by server-side.
** 076 CA  20180529          Made setXorY protected.
**     HC  20180611          Fixed several issues in runtime support of FRAME:VIRTUAL-*
**                           attributes: wrong size reported by the VIRTUAL* attributes, wrong
**                           frame outer size calculated, wrong inner size calculated. And
**                           related changes.
** 077 HC  20180627          Fixed double scale issue when checking widget to fit its parent container.
*/
/*
** This program is free software: you can redistribute it and/or modify
** it under the terms of the GNU Affero General Public License as
** published by the Free Software Foundation, either version 3 of the
** License, or (at your option) any later version.
**
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
** GNU Affero General Public License for more details.
**
** You may find a copy of the GNU Affero GPL version 3 at the following
** location: https://www.gnu.org/licenses/agpl-3.0.en.html
** 
** Additional terms under GNU Affero GPL version 3 section 7:
** 
**   Under Section 7 of the GNU Affero GPL version 3, the following additional
**   terms apply to the works covered under the License.  These additional terms
**   are non-permissive additional terms allowed under Section 7 of the GNU
**   Affero GPL version 3 and may not be removed by you.
** 
**   0. Attribution Requirement.
** 
**     You must preserve all legal notices or author attributions in the covered
**     work or Appropriate Legal Notices displayed by works containing the covered
**     work.  You may not remove from the covered work any author or developer
**     credit already included within the covered work.
** 
**   1. No License To Use Trademarks.
** 
**     This license does not grant any license or rights to use the trademarks
**     Golden Code, FWD, any Golden Code or FWD logo, or any other trademarks
**     of Golden Code Development Corporation. You are not authorized to use the
**     name Golden Code, FWD, or the names of any author or contributor, for
**     publicity purposes without written authorization.
** 
**   2. No Misrepresentation of Affiliation.
** 
**     You may not represent yourself as Golden Code Development Corporation or FWD.
** 
**     You may not represent yourself for publicity purposes as associated with
**     Golden Code Development Corporation, FWD, or any author or contributor to
**     the covered work, without written authorization.
** 
**   3. No Misrepresentation of Source or Origin.
** 
**     You may not represent the covered work as solely your work.  All modified
**     versions of the covered work must be marked in a reasonable way to make it
**     clear that the modified work is not originating from Golden Code Development
**     Corporation or FWD.  All modified versions must contain the notices of
**     attribution required in this license.
*/

package com.goldencode.p2j.ui;

import java.util.*;

import com.goldencode.p2j.ui.client.*;
import com.goldencode.p2j.util.*;

/**
 * The server side base class for most widgets. 
 */
public abstract class BaseEntity<T extends BaseConfig>
extends GenericWidget<T>
{
   /** PARENT attribute */
   private CommonWidget parent = null;
   
   /** Group container for the widget and its siblings. */
   protected Group group = null;

   /**
    * Constructor.
    *
    * @param    dynamic
    *           Flag indicating if this is a static or dynamic resource.
    * @param    base
    *           The container for all state to be transferred to the client.
    */
   public BaseEntity(boolean dynamic, T base)
   {
      super(dynamic, base);
      config.dynamic = dynamic;
   }
   
   /**
    * Constructor.
    *
    * @param    dynamic
    *           Flag indicating if this is a static or dynamic resource.
    * @param    base
    *           The container for all state to be transferred to the client.
    * @param    before
    *           The right neighbor of the BaseEntity been created or null if new BaseEntity
    *           will be the last in the chain. If <b>before</b> is the first element in the chain
    *           the new BaseEntity will become the new first element in the chain.
    * @param    firstAsLast
    *           The flag indicating that this new entity must be inserted as last element 
    *           in the chain in the case when <b>before</b> - is the first element of the chain.
    *           When <b>before</b> is the first element of the chain, default behavior (expected
    *           from the name of this parameter is to insert new element as the first. 
    *           This flag allows to change the expected behavior of <b>before</b> parameter when 
    *           it needs and makes sense only if <b>before</b> is the first element.
    */
   public BaseEntity(boolean dynamic, T base, BaseEntity<T> before, boolean firstAsLast)
   {
      super(dynamic, base, before, firstAsLast);
      config.dynamic = dynamic;
   }

   /**
    * Get validated character and pixel values of the character attribute. Before conversion
    * attribute value is validated, truncated, rounded.
    *
    * @param  value
    *         The attribute value, in pixels. <code>null</code> denotes an unknown attribute
    *         value.
    * @param  attr
    *         The attribute being converted.
    * @param  widget
    *         Widget to which the attribute belongs.
    *
    * @return character and pixel value of the attribute or <code>null</code> if attribute value
    *         is not valid.
    */
   static CharsPixels prepareSizeChars(Double value, SizeAttribute attr, GenericWidget widget)
   {
      // TODO: This is the general widget character size setter for every widgets except frame.
      //       Frame widget has it's own width and height setters, see GenericFrame.setWidthChars
      //       and GenericFrame.setHeigthChars. We need to unify both approaches to have only
      //       one setter place for all widgets. But until this is ready consider the requirement
      //       to change GenericFrame specific methods too if planning to change this method.
      boolean isValid = validateSize(value, attr, widget);

      if (!isValid)
         return null;

      boolean isChui = LogicalTerminal.isChui();

      value = isChui ? Coordinate.scale(value, 0) : Coordinate.scale(value);

      if (value > Coordinate.MAX_GUI_CH && !attr.isVirt())
      {
         String msg;
         if (value.doubleValue() == value.longValue())
         {
            msg = String.format("Invalid character unit value %d changed to %d",
                  value.longValue(), (long) Coordinate.MAX_GUI_CH);
         }
         else
         {
            msg = String.format("Invalid character unit value %s changed to %d",
                  value, (long) Coordinate.MAX_GUI_CH);
         }

         ErrorManager.recordOrShowError(13734, msg, false, true, false);

         value = Coordinate.MAX_GUI_CH;
      }

      CoordinatesConversion cc = LogicalTerminal.coordinates();
      boolean virt = attr.isVirt();
      int valuePixels;

      // setting the WIDTH or HEIGHT attribute?
      // TODO: widget can not go outside of the frame's size (when realized)!
      if (attr.isWidth())
      {
         valuePixels = cc.columnToPixels(value);
         valuePixels = widget.truncatePixelSize(valuePixels, attr);

         // QUIRK: VIRT attributes may be assigned to values 0 or less.
         //        Also the converted pixel value is first incremented by 1
         //        (with a max of 0) before converted back to characters.
         if (attr.isVirt() && value < 0)
         {
            valuePixels++;

            if (valuePixels > 0)
               valuePixels = 0;
         }

         // adjust the input value to the widget minimum/maximum width
         if (widget._isRealized() && !isChui && valuePixels > Coordinate.MAX_GUI_WIDTH_PIXELS &&
               !virt)
         {
            valuePixels = Coordinate.MAX_GUI_WIDTH_PIXELS;
         }

         // the physical location must be recalculated back to the character
         // so that the corresponding exact character value is determined
         value = cc.columnFromPixels(valuePixels,
               virt ? Coordinate.RoundingMode.TRUNCATE :
                     Coordinate.RoundingMode.ROUND);
      }
      else
      {
         valuePixels = cc.rowToPixels(value);
         valuePixels = widget.truncatePixelSize(valuePixels, attr);

         // QUIRK: VIRT attributes may be assigned to values 0 or less.
         //        Also the converted pixel value is first incremented by 1
         //        (with a ceiling of 0) before converted back to characters.
         if (virt && value < 0)
         {
            valuePixels++;

            if (valuePixels > 0)
               valuePixels = 0;
         }

         if (widget._isRealized() && !isChui && valuePixels > Coordinate.MAX_GUI_HEIGHT_PIXELS &&
               !virt)
         {
            valuePixels = Coordinate.MAX_GUI_HEIGHT_PIXELS;
         }

         // the physical location must be recalculated back to the character
         // so that the corresponding exact character value is determined
         value = cc.rowFromPixels(valuePixels,
               virt ? Coordinate.RoundingMode.TRUNCATE :
                     Coordinate.RoundingMode.ROUND);
      }

      return new CharsPixels(value, valuePixels);
   }

   /**
    * Get validated character and pixel values of the pixel attribute. Before conversion
    * attribute value is validated, truncated, rounded.
    *
    * @param  value
    *         The attribute value, in characters. <code>null</code> denotes an unknown attribute
    *         value.
    * @param  attr
    *         The attribute being converted.
    * @param  widget
    *         Widget to which the attribute belongs.
    *
    * @return character and pixel value of the attribute or <code>null</code> if attribute value
    *         is not valid.
    */
   static CharsPixels prepareSizePixels(Integer value, SizeAttribute attr, GenericWidget widget)
   {
      // TODO: This is the general widget pixel size setter for every widgets except frame.
      //       Frame widget has it's own width and height setters, see GenericFrame.setWidthChars
      //       and GenericFrame.setHeigthChars. We need to unify both approaches to have only
      //       one setter place for all widgets. But until this is ready consider the requirement
      //       to change GenericFrame specific methods too if planning to change this method.
      boolean isValid = validateSize(value, attr, widget);

      if (!isValid)
         return null;

      value = widget.truncatePixelSize(value, attr);

      CoordinatesConversion cc = LogicalTerminal.coordinates();
      double charValue = 0;

      // negative VIRT*CHARS values truncate to two decimal places instead of
      // the usual RoundingMode.HALF_UP
      Coordinate.RoundingMode rmode = attr.isVirt() && value < 0 ?
            Coordinate.RoundingMode.TRUNCATE :
            Coordinate.RoundingMode.ROUND;

      // TODO: widget can not go outside of the frame's size (when realized)!

      if (widget._isRealized() && !LogicalTerminal.isChui())
      {
         if (!attr.isVirt())
         {
            // pixel values exceeding the pixel limits for width and height are
            // silently adjusted to the max values
            if (attr.isWidth() && value > Coordinate.MAX_GUI_WIDTH_PIXELS)
            {
               value = Coordinate.MAX_GUI_WIDTH_PIXELS;
            }

            if (!attr.isWidth() && value > Coordinate.MAX_GUI_HEIGHT_PIXELS)
            {
               value = Coordinate.MAX_GUI_HEIGHT_PIXELS;
            }
         }

         charValue = attr.isWidth() ?
               cc.columnFromPixels(value, rmode) :
               cc.rowFromPixels(value, rmode);
      }
      else
      {
         charValue = attr.isWidth() ?
               cc.columnFromPixels(value, rmode) :
               cc.rowFromPixels(value, rmode);

         if (charValue > Coordinate.MAX_GUI_CH)
         {
            String msg;
            if (charValue == (long) charValue)
            {
               msg = String.format("Invalid character unit value %d changed to %d",
                     (long) charValue, (long) Coordinate.MAX_GUI_CH);
            }
            else
            {
               msg = String.format("Invalid character unit value %s changed to %d",
                     charValue, (long) Coordinate.MAX_GUI_CH);
            }
            ErrorManager.recordOrShowError(13734, msg, false, true, false);

            charValue = Coordinate.MAX_GUI_CH;

            // TODO: the following condition is suspicious, retest
            // pixel value for HEIGHT is not adjusted back (at least for window widget)
            if (attr.isWidth() && !attr.isVirt())
            {
               value = cc.columnToPixels(charValue);
            }
         }
      }

      return new CharsPixels(charValue, value);
   }

   /**
    * Performs a validation of widget size value.
    *
    * @param    value
    *           The input value to validate.
    * @param    attr
    *           Denotes the kind of size attribute.
    *
    * @return   <code>true</code> when valid, <code>false</code> otherwise
    */
   private static boolean validateSize(Number value, SizeAttribute attr, GenericWidget widget)
   {
      if (attr.isVirt())
      {
         if (value == null)
         {
            String msg = String.format("**Unable to assign UNKNOWN value to attribute %s on %s widget",
                  attr.shortName(), widget.type());
            ErrorManager.displayWarning(4083, msg, false);
            return false;
         }

         // VIRT* attributes may be assigned values 0 or less
         return true;
      }

      if (value == null || (Double.compare(value.doubleValue(), 0.0d) <= 0 && !attr.isVirt()))
      {
         int num = attr == SizeAttribute.HEIGHT_CHARS  || attr == SizeAttribute.WIDTH_CHARS ||
               attr == SizeAttribute.HEIGHT_PIXELS || attr == SizeAttribute.WIDTH_PIXELS ||
               attr == SizeAttribute.ROW_HEIGHT_CHARS ||
               attr == SizeAttribute.ROW_HEIGHT_PIXELS ?
               4084 : 4082;

         String msg = String.format("**Attribute %s must be greater than zero on %s widget",
               attr.shortName(), widget.type());
         ErrorManager.displayWarning(num, msg, false);
         return false;
      }

      return true;
   }

   /**
    * Sets the column to which widget's colon between label and widget should
    * be aligned.
    * 
    * @param   colon
    *          The 1-based column number. This is converted to a 0-based
    *          column number for the internal runtime.
    */
   public void setColon(NumberType colon)
   {
      setColon(colon.doubleValue());
   }

   /**
    * Sets the column to which widget's colon between label and widget should
    * be aligned.
    * 
    * @param   colon
    *          The 1-based column number. This is converted to a 0-based
    *          column number for the internal runtime.
    */
   public void setColon(double colon)
   {
      ConfigHelper.setColon(config, colon - 1);
      pushWidgetAttr(new String[] {"initColumn", "align"},
                     new Object[] {config.initColumn, config.align});
   }

   /**
    * Gets the COLUMN writable attribute.
    *
    * @return   The 1-based column number at which to start displaying the 
    *           widget.  This is converted from a 0-based column number for
    *           the internal runtime.
    */
   @Override
   public decimal getColumn()
   {
      double col = ConfigHelper.getEffectiveColumn(config);
      col = col < 0 ? col - 1 : col + 1;
      
      return new decimal(col - config.frameColumnOffset);
   }

   /**
    * Sets the column position for this widget.
    *
    * @param    c
    *           The 1-based column number at which to start displaying the 
    *           widget.  This is converted to a 0-based column number for the
    *           internal runtime.
    */
   @Override
   public void setColumn(NumberType c)
   {
      boolean unknown = c.isUnknown();
      setColumnOrRow(unknown, unknown ? 0 : c.doubleValue(), false);
   }

   /**
    * Sets the column position for this widget.
    *
    * @param    c
    *           The 1-based column number at which to start displaying the 
    *           widget.  This is converted to a 0-based column number for the
    *           internal runtime.
    */
   @Override
   public void setColumn(double c)
   {
      setColumnOrRow(false, c, false);
   }
   
   /**
    * Gets the display color writable attribute.
    *
    * @return   The current value of the display color attribute.
    */
   @Override
   public integer getDColor()
   {
      int color = config.dcolor.color();
      return (color == -1) ? new integer() : new integer(color);
   }
   
   /**
    * Sets the display color.
    *
    * @param    param
    *           The display color.
    */
   public void setDColor(NumberType param)
   {
      config.dcolor = new ColorSpec(param.intValue()).convert();
      pushWidgetAttr("dcolor", config.dcolor);
   }

   /**
    * Sets the display color.
    * 
    * @param    displayColor
    *           The display color.
    */
   public void setDColor(Color displayColor)
   {
      config.dcolor = displayColor;
      pushWidgetAttr("dcolor", config.dcolor);
   }

   /**
    * Set widget DCOLOR
    * 
    * @param   color
    *          Integer color value.
    */
   public void setDColor(int color)
   {
      setDColor(new ColorSpec(color).convert());
   }
   
   /**
    * Gets the FGCOLOR writable attribute.
    *
    * @return   The current value of the FGCOLOR attribute.
    */
   public integer getFgColor()
   {
      int color = config.fgcolor;
      return color == -1 ? new integer() : new integer(color);
   }

   /**
    * Sets the FGCOLOR writable attribute.
    *
    * @param    fgcolor
    *           The new value for the FGCOLOR attribute.
    */
   @Override
   public void setFgColor(integer fgcolor)
   {
      int color = (fgcolor == null || fgcolor.isUnknown()) ? -1 : fgcolor.intValue();
      setFgColor(color);
   }   

   /**
    * Sets the FGCOLOR writable attribute.
    *
    * @param    fgcolor
    *           The new value for the FGCOLOR attribute.
    */
   public void setFgColor(int64 fgcolor)
   {
      int color = (fgcolor == null || fgcolor.isUnknown()) ? -1 : fgcolor.intValue(); 
      setFgColor(color);
   }

   /**
    * Sets the FGCOLOR writable attribute.
    *
    * @param    fgcolor
    *           The new value for the FGCOLOR attribute.
    */
   public void setFgColor(int fgcolor)
   {
      if (config.fgcolor == fgcolor)
      {
         return;
      }

      config.fgcolor = fgcolor;
      // reset the explicit FGCOLOR-RGB
      config.fgcolorRgb = -1;
      pushWidgetAttr(new String[] {"fgcolor", "fgcolorRgb"},
                     new Object[] {config.fgcolor, config.fgcolorRgb});
   }

   /**
    * Gets the BGCOLOR writable attribute.
    *
    * @return   The current value of the BGCOLOR attribute.
    */
   public integer getBgColor()
   {
      int color = config.bgcolor;
      return (color == -1) ? new integer() : new integer(color);
   }
   
   /**
    * Sets the BGCOLOR writable attribute.
    *
    * @param    bgcolor
    *           The new value for the BGCOLOR attribute.
    */
   @Override
   public void setBgColor(integer bgcolor)
   {
      int color = (bgcolor == null || bgcolor.isUnknown()) ? -1 : bgcolor.intValue();
      setBgColor(color);
   }

   /**
    * Sets the BGCOLOR writable attribute.
    *
    * @param    bgcolor
    *           The new value for the BGCOLOR attribute.
    */
   public void setBgColor(int64 bgcolor)
   {
      int color = (bgcolor == null || bgcolor.isUnknown()) ? -1 : bgcolor.intValue(); 
      setBgColor(color);
   }

   /**
    * Sets the BGCOLOR writable attribute.
    *
    * @param    bgcolor
    *           The new value for the BGCOLOR attribute.
    */
   public void setBgColor(int bgcolor)
   {
      if (config.bgcolor == bgcolor)
      {
         return;
      }

      config.bgcolor = bgcolor;
      // reset the explicit BGCOLOR-RGB
      config.bgcolorRgb = -1;
      pushWidgetAttr(new String[] {"bgcolor", "bgcolorRgb"},
                     new Object[] {config.bgcolor, config.bgcolorRgb});
   }
   
   /**
    * Gets the FGCOLOR-RGB writable attribute.
    *
    * @return   The current value of the FGCOLOR-RGB attribute.
    */
   public integer getFgColorRgb()
   {
      int color = config.fgcolorRgb;
      return color == -1 ? new integer() : new integer(color);
   }

   /**
    * Sets the FGCOLOR-RGB writable attribute.
    *
    * @param    fgcolorRgb
    *           The new value for the FGCOLOR-RGB attribute.
    */
   @Override
   public void setFgColorRgb(integer fgcolorRgb)
   {
      int color = (fgcolorRgb == null || fgcolorRgb.isUnknown()) ? -1 : fgcolorRgb.intValue();
      setFgColorRgb(color);
   }   

   /**
    * Sets the FGCOLOR-RGB writable attribute.
    *
    * @param    fgcolorRgb
    *           The new value for the FGCOLOR-RGB attribute.
    */
   public void setFgColorRgb(int64 fgcolorRgb)
   {
      int color = (fgcolorRgb == null || fgcolorRgb.isUnknown()) ? -1 : fgcolorRgb.intValue(); 
      setFgColorRgb(color);
   }

   /**
    * Sets the FGCOLOR-RGB writable attribute.
    *
    * @param    fgcolorRgb
    *           The new value for the FGCOLOR-RGB attribute.
    */
   public void setFgColorRgb(int fgcolorRgb)
   {
      if (config.fgcolorRgb == fgcolorRgb)
      {
         return;
      }

      config.fgcolorRgb = fgcolorRgb;
      // reset the explicit FGCOLOR
      config.fgcolor = -1;
      pushWidgetAttr(new String[] {"fgcolor", "fgcolorRgb"},
                     new Object[] {config.fgcolor, config.fgcolorRgb});
   }

   /**
    * Gets the BGCOLOR-RGB writable attribute.
    *
    * @return   The current value of the BGCOLOR-RGB attribute.
    */
   public integer getBgColorRgb()
   {
      int color = config.bgcolorRgb;
      return (color == -1) ? new integer() : new integer(color);
   }
   
   /**
    * Sets the BGCOLOR-RGB writable attribute.
    *
    * @param    bgcolorRgb
    *           The new value for the BGCOLOR-RGB attribute.
    */
   @Override
   public void setBgColorRgb(integer bgcolorRgb)
   {
      int color = (bgcolorRgb == null || bgcolorRgb.isUnknown()) ? -1 : bgcolorRgb.intValue();
      setBgColorRgb(color);
   }

   /**
    * Sets the BGCOLOR-RGB writable attribute.
    *
    * @param    bgcolorRgb
    *           The new value for the BGCOLOR-RGB attribute.
    */
   public void setBgColorRgb(int64 bgcolorRgb)
   {
      int color = (bgcolorRgb == null || bgcolorRgb.isUnknown()) ? -1 : bgcolorRgb.intValue(); 
      setBgColorRgb(color);
   }

   /**
    * Sets the BGCOLOR-RGB writable attribute.
    *
    * @param    bgcolorRgb
    *           The new value for the BGCOLOR-RGB attribute.
    */
   public void setBgColorRgb(int bgcolorRgb)
   {
      if (config.bgcolorRgb == bgcolorRgb)
      {
         return;
      }

      config.bgcolorRgb = bgcolorRgb;
      // reset the explicit BGCOLOR
      config.bgcolor = -1;
      pushWidgetAttr(new String[] {"bgcolor", "bgcolorRgb"},
                     new Object[] {config.bgcolor, config.bgcolorRgb});
   }
   
   /**
    * Set alignment.
    *
    * @param   mode
    *          Set alignment mode.
    */
   public void setAlign(int mode)
   {
      config.align = mode;
      pushWidgetAttr("align", config.align);
   }

   /**
    * Gets the HEIGHT-CHARS writable attribute.
    *
    * @return   The current value of the HEIGHT-CHARS attribute.
    */
   @Override
   public decimal getHeightChars()
   {
      return _getHeightChars();
   }

   /**
    * Sets the HEIGHT-CHARS writable attribute.
    *
    * @param    height
    *           The new value for the HEIGHT-CHARS attribute.
    */
   public void setHeightChars(NumberType height)
   {
      boolean unknown = height.isUnknown();
      setSizeChars(unknown ? null : height.doubleValue(), SizeAttribute.HEIGHT_CHARS);
   }
   
   /**
    * Sets the HEIGHT-CHARS writable attribute.
    *
    * @param    height
    *           The new value for the HEIGHT-CHARS attribute.
    */
   public void setHeightChars(double height)
   {
      setSizeChars(height, SizeAttribute.HEIGHT_CHARS);
   }
   
   /**
    * Gets the WIDTH-CHARS writable attribute.
    *
    * @return   The current value of the WIDTH-CHARS attribute.
    */
   public decimal getWidthChars()
   {
      return _getWidthChars();
   }

   /**
    * Sets the WIDTH-CHARS writable attribute.
    *
    * @param    width
    *           The new value for the WIDTH-CHARS attribute.
    */
   public void setWidthChars(NumberType width)
   {
      boolean unknown = width.isUnknown();
      setSizeChars(unknown ? null : width.doubleValue(), SizeAttribute.WIDTH_CHARS);
   }
   
   /**
    * Sets the WIDTH-CHARS writable attribute.
    *
    * @param    width
    *           The new value for the WIDTH-CHARS attribute.
    */
   public void setWidthChars(double width)
   {
      setSizeChars(width, SizeAttribute.WIDTH_CHARS);
   }
   
   /**
    * Get rectangle height pixels.
    * SIZE-PIXELS option.
    * 
    * @return  The height of the widget in pixels.
    */
   public integer getHeightPixels()
   {
      return _getHeightPixels();
   }   

   /**
    * Get rectangle width pixels.
    * SIZE-PIXELS option.
    * 
    * @return  The width of the widget in pixels.
    */
   public integer getWidthPixels()
   {
      return _getWidthPixels();
   }   
   
   /**
    * Gets the PARENT writable attribute.
    * If the parent attribute was set explicitly, it is returned.
    * Otherwise, the field group of this widget is returned as parent.
    * For the regular widgets, it always has a value. It's the default field
    * group. Non-regular widgets override this method to return a different
    * result.  
    *
    * @return   the parent widget for this widget
    */
   @Override
   public CommonWidget getParent()
   {
      // explicitly set parent attribute
      if (parent != null)
      {
         return parent;
      }

      // normal case
      return group;
   }

   /**
    * Returns the parent widget of the specified type. If not found
    * <code>null</code> is returned.
    *
    * @return   the parent widget for this widget
    */
   public <T extends CommonWidget> T getParent(Class<T> clazz)
   {
      CommonWidget parent = getParent();
      while(parent != null && !(clazz.isAssignableFrom(parent.getClass())))
      {
         parent = parent.getParent();
      }

      return (T) parent;
   }

   /**
    * Sets the PARENT writable attribute.
    *
    * @param    widget
    *           The new value for the PARENT attribute.
    */
   public void setParent(GenericWidget<?> widget)
   {
      if (!validateParent(widget))
      {
         // parent is not valid, do not proceed
         return;
      }
      
      GenericWidget parent = LogicalTerminal.getWidgetForId(config.parentId);
      
      if (parent instanceof WidgetContainer)
      {
         WidgetContainer cont = (WidgetContainer) parent;
         if (cont != null)
         {
            cont.removeWidget(this);
         }
      }

      this.parent = widget;
      config.parentId = widget.getId();
      
      // update the parent container
      parent = LogicalTerminal.getWidgetForId(config.parentId);
      if (parent instanceof WidgetContainer)
      {
         WidgetContainer cont = (WidgetContainer) parent;
         cont.addWidget(this);

         // attach it to the frame
         GenericFrame frame = widget.getFrame();
         if (frame != null)
         {
            this.config.frameId = frame.getFrameId();
            frame.addDynamicWidget(this);
         }
         else
         {
            this.config.frameId = -1;
         }
      }

      pushScreenDefinition();
   }

   /**
    * Establishes a relationship between the container and the elements in
    * a given widget group.
    *
    * @param    group
    *           The container of the widget.
    */
   @Override
   public void setGroup(Group group)
   {
      this.group = group;
   }
   
   /**
    * Obtains the next sibling widget in the current field group.
    *
    * @return   The next sibling or an unknown handle if this is the only widget or if this is the 
    *           last widget.
    */
   @Override
   public handle getNextSibling()
   {
      if (useHandleChainSiblings())
      {
         return super.getNextSibling((HandleChain) getParent());
      }
      else
      {
         GenericWidget<?> widget = (group == null) ? null : group.getNextChild(this);
         return new handle(widget);
      }
   }
   
   /**
    * Obtains the previous sibling widget in the current field group.
    *
    * @return   The previous sibling or an unknown handle if this only widget or if this is the 
    *           first widget.
    */
   @Override
   public handle getPrevSibling()
   {
      if (useHandleChainSiblings())
      {
         return super.getPrevSibling((HandleChain) getParent());
      }
      else
      {
         GenericWidget<?> widget = (group == null) ? null : group.getPrevChild(this);
         return new handle(widget);
      }
   }

   /**
    * Queries if the user is allowed to move this widget using direct manipulation.
    * 
    * @return  If <code>true</code>, the widget is movable via direct manipulation.
    */
   @Override
   public logical isMovable()
   {
      return new logical(config.movable);
   }
   
   /**
    * Defines if the user is allowed to move this widget using direct manipulation.
    * 
    * @param   movable
    *          <code>true</code> to enable movable direct manipulation.
    */
   @Override
   public void setMovable(boolean movable)
   {
      if (config.movable == movable)
      {
         return;
      }

      config.movable = movable;
      pushWidgetAttr("movable", config.movable);
   }
   
   /**
    * Defines if the user is allowed to move this widget using direct manipulation.
    * 
    * @param   movable
    *          <code>true</code> to enable movable direct manipulation.
    */
   @Override
   public void setMovable(logical movable)
   {
      if (movable == null || movable.isUnknown())
      {
         ErrorManager.displayWarning(4083,
               String.format(
               "Unable to assign UNKNOWN value to attribute MOVABLE on %s %s", 
               type(), widgetName()),
               false);
         return;
      }
      
      setMovable(movable.booleanValue()); 
   }
   
   /**
    * Getter for RESIZABLE attribute.
    *
    * @return  {@code true} if widget is resizable.
    */
   @Override
   public logical isResizable()
   {
      return new logical(config.resizable);
   }
   
   /**
    * Setter for RESIZABLE attribute.
    *
    * @param   on
    *          {@code true} to make the widget resizable
    */
   @Override
   public void setResizable(boolean on)
   {
      if (config.resizable == on)
      {
         return;
      }

      config.resizable = on;
      pushWidgetAttr("resizable", config.resizable);
   }
   
   /**
    * Setter for RESIZABLE attribute.
    *
    * @param   on
    *          {@code true} to make the widget resizable
    */
   @Override
   public void setResizable(logical on)
   {
      if (on == null || on.isUnknown())
      {
         ErrorManager.displayWarning(4083,
               String.format(
               "Unable to assign UNKNOWN value to attribute RESIZABLE on %s %s", 
               type(), widgetName()),
               false);
         return;
      }
      
      setResizable(on.booleanValue()); 
   }
   
   /**
    * Gets the input (prompt/edit mode) color writable attribute.
    *
    * @return   The current value of the input (prompt/edit mode) color attribute.
    */
   @Override
   public integer getPfColor()
   {
      int color = config.pfcolor.color();
      return (color == -1) ? new integer() : new integer(color);
   }
   
   /**
    * Sets the input (prompt/edit mode) color.
    *
    * @param    param
    *           The color to use.
    */
   public void setPfColor(NumberType param)
   {
      config.pfcolor = new ColorSpec(param.intValue()).convert();
      pushWidgetAttr("pfcolor", config.pfcolor);
   }

   /**
    * Sets the input (prompt/edit mode) color.
    * 
    * @param    inputColor
    *           The color to use for when in "input" mode.
    */
   public void setPfColor(Color inputColor)
   {
      config.pfcolor = inputColor;
      pushWidgetAttr("pfcolor", config.pfcolor);
   }

   /**
    * Set widget PFCOLOR
    * 
    * @param   color
    *          Integer color value.
    */
   public void setPfColor(int color)
   {
      setPfColor(new ColorSpec(color).convert());
   }
   
   /**
    * Gets the ROW writable attribute.
    *
    * @return   The 1-based row number at which to start displaying the 
    *           widget.  This is converted from a 0-based row number for the
    *           internal runtime.
    */
   @Override
   public decimal getRow()
   {
      double row = ConfigHelper.getEffectiveRow(config);
      row = row < 0 ? row - 1 : row + 1;
      
      return new decimal(row - config.frameRowOffset);
   }

   /**
    * Sets the row position for this widget.
    *
    * @param    r
    *           The 1-based row number at which to start displaying the 
    *           widget.  This is converted to a 0-based row number for the
    *           internal runtime.
    */
   @Override
   public void setRow(NumberType r)
   {
      boolean unknown = r.isUnknown();
      setColumnOrRow(unknown, unknown ? 0 : r.doubleValue(), true);
   }

   /**
    * Sets the row position for this widget.
    *
    * @param    r
    *           The 1-based row number at which to start displaying the 
    *           widget.  This is converted to a 0-based row number for the
    *           internal runtime.
    */
   @Override
   public void setRow(double r)
   {
      setColumnOrRow(false, r, true);   
   }

   /**
    * Gets the SELECTED writable attribute.
    *
    * @return    The current value of the SELECTED attribute.
    */
   @Override
   public logical isSelected()
   {
      return new logical(config.selected);
   }

   /**
    * Sets the SELECTED writable attribute.
    *
    * @param    sel
    *           The new value for the SELECTED attribute.
    */
   @Override
   public void setSelected(logical sel)
   {
      if (sel == null || sel.isUnknown())
      {
         ErrorManager.displayWarning(4083,
               String.format(
               "Unable to assign UNKNOWN value to attribute SELECTED on %s %s", 
               type(), widgetName()),
               false);
         return;
      }

      setSelected(sel.booleanValue());
   }

   /**
    * Sets the SELECTED writable attribute.
    *
    * @param    sel
    *           The new value for the SELECTED attribute.
    */
   @Override
   public void setSelected(boolean sel)
   {
      if (config.selected == sel)
      {
         return;
      }
      
      config.selected = sel;
      pushWidgetAttr("selected", config.selected);
   }
   
   /**
    * Gets the TITLE writable attribute.
    *
    * @return   The current value of the TITLE attribute.
    */
   @Override
   public character getTitle()
   {
      return new character(config.title);
   }

   /**
    * Sets the TITLE writable attribute.
    *
    * @param    title
    *           The new value for the TITLE attribute.
    */
   @Override
   public void setTitle(character title)
   {
      setTitle(title.isUnknown() ? "" : title.toStringMessage());
   }

   /**
    * Sets the TITLE writable attribute.
    *
    * @param    title
    *           The new value for the TITLE attribute.
    */
   @Override
   public void setTitle(String title)
   {
      if (Utils.equals(config.title, title))
      {
         return;
      }

      config.title = title;
      pushWidgetAttr("title", config.title);
   }

   /**
    * Get TOOLTIP attribute.
    *
    * @return   The current value of the TOOLTIP attribute.
    */
   @Override
   public character getTooltip()
   {
      return new character(config.tooltip);
   }

   /**
    * Gets the X writable attribute.
    *
    * @return   The current value of the X attribute.
    */
   @Override
   public integer getX()
   {
      int x = ConfigHelper.getEffectiveX(config);
      return new integer(x - config.frameXOffset);
   }

   /**
    * Sets the X writable attribute.
    *
    * @param    x
    *           The new value for the X attribute.
    */
   @Override
   public void setX(NumberType x)
   {
      boolean unknown = x.isUnknown();
      setXorY(unknown, unknown ? 0 : x.intValue(), true);
   }

   /**
    * Sets the X writable attribute.
    *
    * @param    x
    *           The new value for the X attribute.
    */
   public void setX(int x)
   {
      setXorY(false, x, true);
   }
   
   /**
    * Gets the Y writable attribute.
    *
    * @return    The current value of the Y attribute.
    */
   @Override
   public integer getY()
   {
      int y = ConfigHelper.getEffectiveY(config);
      return new integer(y - config.frameYOffset);
   }

   /**
    * Sets the Y writable attribute.
    *
    * @param    y
    *           The new value for the Y attribute.
    */
   @Override
   public void setY(NumberType y)
   {
      boolean unknown = y.isUnknown();
      setXorY(unknown, unknown ? 0 : y.intValue(), false);
   }
   
   /**
    * Sets the Y writable attribute.
    *
    * @param    y
    *           The new value for the Y attribute.
    */
   public void setY(int y)
   {
      setXorY(false, y, false);
   }

   /**
    * Set new value of TOOLTIP attribute.
    *
    * @param tooltip
    *        new value of TOOLTIP attribute
    */
   @Override
   public void setTooltip(String tooltip)
   {
      if (Utils.equals(config.tooltip, tooltip))
      {
         return;
      }

      config.tooltip = tooltip;
      pushWidgetAttr("tooltip", config.tooltip);
   }   
   
   /**
    * Set new value of TOOLTIP attribute.
    *
    * @param tooltip
    *        new value of TOOLTIP attribute
    */
   @Override
   public void setTooltip(character tooltip)
   {
      if (tooltip == null || tooltip.isUnknown())
      {
         if (config.tooltip == null)
         {
            return;
         }
         
         setTooltip((String)null);
         // deactivate tooltip feature for this widget not depending on SESSION:TOOLTIP value
         // the client side call to deactivate tooltip for this widget
         LogicalTerminal.getClient().deactivateTooltipForWidget(config.id.asInt());
      }
      else
      {
         if (Utils.equals(config.tooltip, tooltip.toStringMessage()))
         {
            return;
         }
         
         setTooltip(tooltip.toStringMessage());
         // activate tooltip feature for this widget but if SESSION:TOOLTIP is true
         if (SessionUtils.getSessionTooltips().booleanValue())
         {
            // the client side call to activate tooltip for this widget
            LogicalTerminal.getClient().activateTooltipForWidget(config.id.asInt());
         }
      }
   }   
   
   /**
    * Sets the WIDTH-PIXELS writable attribute.
    *
    * @param    widthPixels
    *           The new value for the WIDTH-PIXELS attribute.
    */
   public void setWidthPixels(NumberType widthPixels)
   {
      boolean unknown = widthPixels.isUnknown();
      setSizePixels(unknown ? null : widthPixels.intValue(), 
                             SizeAttribute.WIDTH_PIXELS);
   }
   
   /**
    * Sets the WIDTH-PIXELS writable attribute.
    *
    * @param    widthPixels
    *           The new value for the WIDTH-PIXELS attribute.
    */
   public void setWidthPixels(int widthPixels)
   {
      setSizePixels(widthPixels, SizeAttribute.WIDTH_PIXELS);
   }
   
   /**
    * Sets new value of HEIGHT-PIXELS attribute.
    *
    * @param heightPixels
    *        new value of HEIGHT-PIXELS attribute
    */
   public void setHeightPixels(NumberType heightPixels)
   {
      boolean unknown = heightPixels.isUnknown();
      setSizePixels(unknown ? null : heightPixels.intValue(), 
                             SizeAttribute.HEIGHT_PIXELS);
   }

   /**
    * Sets new value of HEIGHT-PIXELS attribute.
    *
    * @param heightPixels
    *        new value of HEIGHT-PIXELS attribute
    */
   public void setHeightPixels(int heightPixels)
   {
      setSizePixels(heightPixels, SizeAttribute.HEIGHT_PIXELS);
   }

   /**
    * Set the FONT attribute of this widget.
    * 
    * @param    fontNum
    *           An entry in the font-table or unknown to refer the default font.
    */
   @Override
   public void setFont(int64 fontNum)
   {
      if (LogicalTerminal.isChui())
      {
         ErrorManager.recordOrShowError(4088, 
               String.format(
               "Progress does not support the FONT attribute for the %s %s " +
               "in this display environment", 
               type(), widgetName()),
               false);
         return;
      }
      
      // WARNING: this can't be a no-op, as setting the font (even to the same value), the widget 
      // gets realized/resized
      
      // if the font is unknown, it gets set to UNDEFINED.
      int font = (fontNum.isUnknown() ? FontManager.UNDEFINED_FONT : fontNum.intValue());

      changeFontWorker(font);
      
      pushWidgetAttr(new String[] {"font", "fontChanged", "fixedWidth"},
                     new Object[] {config.font, config.fontChanged, config.fixedWidth});
   }
   
   /**
    * Get the FONT attribute of this widget.
    * 
    * @return   A value representing an entry in the user's font-table.
    */
   @Override
   public integer getFont()
   {
      if (LogicalTerminal.isChui())
      {
         return new integer();
      }
      int font = config.font;
      
      return (font == FontManager.UNDEFINED_FONT ? new integer() : new integer(font));
   }
   
   /**
    * Set value of the TAB-STOP attribute.
    *
    * @param    value
    *           Value of the attribute.
    */
   @Override
   public void setTabStop(boolean value)
   {
      if (config.tabStop == value)
      {
         return;
      }
      
      boolean oldValue = config.tabStop;
      
      config.tabStop = value;
      
      if (this.group instanceof FieldGroup)
      {
         FieldGroup fg = (FieldGroup) this.group;
         // add widget to the end of NEXT-TAB-ITEM list iff the widget is not in this list
         if (value)
         {
            fg.updateTabItemsOnTabStop(this, !oldValue);
         }
         
         GenericFrame ownerFrame = fg.getFrame();
         
         if (_isRealized() && ownerFrame != null && ownerFrame.getFrameWidget()._isRealized())
         {
            LogicalTerminal.updateTabItemList(fg);
         }
      }

      pushWidgetAttr("tabStop", config.tabStop);
   }

   /**
    * Set value of the TAB-STOP attribute.
    *
    * @param    value
    *           Value of the attribute.
    */
   @Override
   public void setTabStop(logical value)
   {
      if (value == null || value.isUnknown())
      {
         ErrorManager.displayWarning(4083,
               String.format(
               "Unable to assign UNKNOWN value to attribute TAB-STOP on %s %s", 
               type(), widgetName()),
               false);
         return;
      }
      setTabStop(value.booleanValue());
   }

   /**
    * Get value of the TAB-STOP attribute.
    *
    * @return   value of the attribute.
    */
   @Override
   public logical isTabStop()
   {
      return new logical(config.tabStop);
   }
   
   /**
    * Gets the FRAME-COL writable attribute.
    *
    * @return   The 1-based frame column number.
    */
   @Override
   public decimal getFrameColumn()
   {
      return MathOps.plus(getColumn(), config.frameColumnOffset);
   }

   /**
    * Gets the FRAME-ROW writable attribute.
    *
    * @return   The 1-based frame column number.
    */
   @Override
   public decimal getFrameRow()
   {
      return MathOps.plus(getRow(), config.frameRowOffset);
   }

   /**
    * Gets the FRAME-X writable attribute.
    *
    * @return   The 0-based frame column number.
    */
   @Override
   public integer getFrameX()
   {
      // if we are in ChUI, this actually returns the FRAME-COL
      return new integer(LogicalTerminal.isChui() ? getFrameColumn() 
                                                  : MathOps.plus(getX(), config.frameXOffset));
   }

   /**
    * Gets the FRAME-Y writable attribute.
    *
    * @return   The 0-based frame Y number.
    */
   @Override
   public integer getFrameY()
   {
      // if we are in ChUI, this actually returns the FRAME-ROW
      return new integer(LogicalTerminal.isChui() ? getFrameRow() 
                                                  : MathOps.plus(getY(), config.frameYOffset));
   }

   /**
    * Sets the value for the FONT attribute.
    * Only valid width values must be passed in the font argument.
    * 
    * @param    font
    *           Valid font value.
    */
   protected void changeFontWorker(int font)
   {
      config.font = font;
   }
   
   /**
    * Indicates an implementation of siblings methods. Child classes can override
    * this method to specify the usage of HandleChain siblings methods.
    * 
    * @return   TRUE  - if the logic of parent class {@link HandleChain} is used,
    *           FALSE - if own siblings implementation logic is used. 
    */
   protected boolean useHandleChainSiblings()
   {
      return false;
   }
   
   /**
    * Set AUTO-RESIZE to false if applicable
    */
   protected void resetAutoResize()
   {
      if (this instanceof AutoResizableWidget)
      {
         ((AutoResizableWidget)this).resetAutoResize();
      }
   }

   /**
    * Sets the value for the attribute WIDTH-CHARS in the associated config instance.
    * Only valid width values must be passed in the width argument.
    * 
    * @param    width
    *           Valid width value.
    */
   protected void setWidthCharsWorker(double width)
   {
      config.widthChars = width;
      
      // override the client-calculated value
      config.clientWidthChars = BaseConfig.INV_COORD;

      resetAutoResize();
   }

   /**
    * Sets the value for the attribute HEIGHT-CHARS in the associated config instance.
    * Only valid height values must be passed in the height argument.
    * 
    * @param    height
    *           Valid height value.
    */
   protected void setHeightCharsWorker(double height)
   {
      config.heightChars = height;
      
      // override the client-calculated value
      config.clientHeightChars = BaseConfig.INV_COORD;
      
      resetAutoResize();
   }

   /**
    * Sets the value for the attribute WIDTH-PIXELS in the associated config instance.
    * Only valid width values must be passed in the width argument.
    * 
    * @param    width
    *           Valid width value.
    */
   protected void setWidthPixelsWorker(int width)
   {
      config.widthPixels = width;
      
      // override the client-calculated value
      config.clientWidthPixels = BaseConfig.INV_COORD;
      
      resetAutoResize();
   }

   /**
    * Sets the value for the attribute HEIGHT-PIXELS in the associated config instance.
    * Only valid height values must be passed in the height argument.
    * 
    * @param    height
    *           Valid height value.
    */
   protected void setHeightPixelsWorker(int height)
   {
      config.heightPixels = height;
      
      // override the client-calculated value
      config.clientHeightPixels = BaseConfig.INV_COORD;
      
      resetAutoResize();
   }   
   
   /**
    * Sets the value for the attribute MIN-WIDTH-CHARS.
    * Only valid width values must be passed in the width argument.
    * 
    * @param    width
    *           Valid width value.
    */
   protected void setMinWidthCharsWorker(double width)
   {
      throw new RuntimeException("Operation not supported.");
   }

   /**
    * Sets the value for the attribute MIN-HEIGHT-CHARS.
    * Only valid height values must be passed in the height argument.
    * 
    * @param    height
    *           Valid height value.
    */
   protected void setMinHeightCharsWorker(double height)
   {
      throw new RuntimeException("Operation not supported.");
   }

   /**
    * Sets the value for the attribute MIN-WIDTH-PIXELS.
    * Only valid width values must be passed in the width argument.
    * 
    * @param    width
    *           Valid width value.
    */
   protected void setMinWidthPixelsWorker(int width)
   {
      throw new RuntimeException("Operation not supported.");
   }

   /**
    * Sets the value for the attribute MIN-HEIGHT-PIXELS.
    * Only valid height values must be passed in the height argument.
    * 
    * @param    height
    *           Valid height value.
    */
   protected void setMinHeightPixelsWorker(int height)
   {
      throw new RuntimeException("Operation not supported.");
   }   
   
   /**
    * Sets the value for the attribute MAX-WIDTH-CHARS.
    * Only valid width values must be passed in the width argument.
    * 
    * @param    width
    *           Valid width value.
    */
   protected void setMaxWidthCharsWorker(double width)
   {
      throw new RuntimeException("Operation not supported.");
   }

   /**
    * Sets the value for the attribute MAX-HEIGHT-CHARS.
    * Only valid height values must be passed in the height argument.
    * 
    * @param    height
    *           Valid height value.
    */
   protected void setMaxHeightCharsWorker(double height)
   {
      throw new RuntimeException("Operation not supported.");
   }

   /**
    * Sets the value for the attribute MAX-WIDTH-PIXELS.
    * Only valid width values must be passed in the width argument.
    * 
    * @param    width
    *           Valid width value.
    */
   protected void setMaxWidthPixelsWorker(int width)
   {
      throw new RuntimeException("Operation not supported.");
   }

   /**
    * Sets the value for the attribute MAX-HEIGHT-PIXELS.
    * Only valid height values must be passed in the height argument.
    * 
    * @param    height
    *           Valid height value.
    */
   protected void setMaxHeightPixelsWorker(int height)
   {
      throw new RuntimeException("Operation not supported.");
   }   

   /**
    * Sets the value for the attribute VIRTUAL-WIDTH-CHARS.
    * Only valid width values must be passed in the width argument.
    * 
    * @param    width
    *           Valid width value.
    */
   protected void setVirtWidthCharsWorker(double width)
   {
      // no-op
   }

   /**
    * Sets the value for the attribute VIRTUAL-HEIGHT-CHARS.
    * Only valid height values must be passed in the height argument.
    * 
    * @param    height
    *           Valid height value.
    */
   protected void setVirtHeightCharsWorker(double height)
   {
      // no-op
   }

   /**
    * Sets the value for the attribute VIRTUAL-WIDTH-PIXELS.
    * Only valid width values must be passed in the width argument.
    * 
    * @param    width
    *           Valid width value.
    */
   protected void setVirtWidthPixelsWorker(int width)
   {
      // no-op
   }

   /**
    * Sets the value for the attribute VIRTUAL-HEIGHT-PIXELS.
    * Only valid height values must be passed in the height argument.
    * 
    * @param    height
    *           Valid height value.
    */
   protected void setVirtHeightPixelsWorker(int height)
   {
      // no-op
   }
   
   /**
    * Sets the row or column position for this widget.
    * <p>
    * Sub-classes can override this method; at this point, the parameters are valid.
    * 
    * @param    isUnknown
    *           If <code>true</code> the attribute value being set is an unknown value.
    * @param    value
    *           The 1-based row or column number.
    * @param    row
    *           If <code>true</code> <code>value</code> represents a row number, otherwise 
    *           a column number.
    */
   protected void setColumnOrRowWorker(boolean isUnknown, double value, boolean row)
   {
      boolean isChui  = LogicalTerminal.isChui();
      boolean isRedir = isRedirected();
      
      // make the value 0-based, it is treated as such by the rest of the framework
      value = value < 0 ? value + 1 : value - 1;
      
      CoordinatesConversion coordinates = LogicalTerminal.coordinates();
      GenericFrame gfr = getFrame();
      boolean isInsideSetup = (gfr != null && gfr.isInsideSetup());

      // setting the row or column attribute?
      if (row)
      {
         int y = coordinates.rowToPixels(value);
         
         if (_isRealized() && !isChui && !isRedir)
         {
            y = Coordinate.adjustMinMaxGuiPixel(y);
         }
         
         // the physical location must be recalculated back to the character
         // so that the corresponding exact character location is determined
         value = coordinates.rowFromPixels(y);
   
         // bring back to 1-based, to do more validation
         value = value < 0 ? value - 1 : value + 1;
         
         // strangely, the value -1 has a special behavior, where it is silently changed to 1 when
         // assigned to ROW/COLUMN
         if (value == -1)
         {
            value = 1;
         }
         
         if (!isRedir)
         {
            if (isChui)
            {
               value = Coordinate.adjustMinMaxChuiChar(value);
            }
            else
            {
               value = Coordinate.adjustMinMaxGuiChar(value);
            }
         }
         
         // config stores 0-based value
         value = value < 0 ? value + 1 : value - 1;
         
         if (!config.realized || isInsideSetup)
         {
            config.initRow = value;
            config.initY = y;
            
            // reset the client-related coordinates, as they need to be re-computed
            config.clientRow = BaseConfig.INV_COORD;
            config.clientY = BaseConfig.INV_COORD;
         }
         
         if (!isInsideSetup)
         {
            // adjust for frame offset
            value += config.frameRowOffset;
            y += config.frameYOffset;
            
            // last check for the new position does not place the widget outside the frame
            // virtual region
            //
            // get possible frame/window owning this widget
            PaneEntity<?> p = getParent(FrameWidget.class);
            if (p == null)
            {
               p = getParent(WindowWidget.class);
            }

            double maxHeight = Coordinate.scale(value + config.heightChars);

            // TODO: Progress doesn't prevent assignment of the new value when it doesn't fit, it only
            //       shows a warning
            // need to check if the widget with new location fits the realized frame
            if (p == null || !_isRealized() || maxHeight <= ConfigHelper.getEffectiveVirtualHeight(p.config))
            {
               config.row = value;
               config.y = y;
            }
         }
      }
      else
      {
         int x = coordinates.columnToPixels(value);
         
         if (_isRealized() && !isChui && !isRedir)
         {
            x = Coordinate.adjustMinMaxGuiPixel(x);
         }
         
         // the physical location must be recalculated back to the character
         // so that the corresponding exact character location is determined
         value = coordinates.columnFromPixels(x);
   
         // bring back to 1-based, to do more validation
         value = value < 0 ? value - 1 : value + 1;
         
         // strangely, the value -1 has a special behavior, where it is silently changed to 1 when
         // assigned to ROW/COLUMN
         if (value == -1)
         {
            value = 1;
         }
         
         if (isRedir)
         {
            if (isChui)
            {
               value = Coordinate.adjustMinMaxChuiChar(value);
            }
            else
            {
               value = Coordinate.adjustMinMaxGuiChar(value);
            }
         }
         
         // config stores 0-based value
         value = value < 0 ? value + 1 : value - 1;
         
         if (!config.realized || isInsideSetup)
         {
            config.initColumn = value;
            config.initX = x;
            
            // reset the client-related coordinates, as they need to be re-computed
            config.clientColumn = BaseConfig.INV_COORD;
            config.clientX = BaseConfig.INV_COORD;
         }

         if (!isInsideSetup)
         {
            // adjust for frame offset
            value += config.frameColumnOffset;
            x += config.frameXOffset;

            // last check for the new position does not place the widget outside the frame
            // virtual region
            //
            // get possible frame/window owning this widget
            PaneEntity<?> p = getParent(FrameWidget.class);
            if (p == null)
            {
               p = getParent(WindowWidget.class);
            }

            double maxWidth = Coordinate.scale(value + config.widthChars);

            // TODO: Progress doesn't prevent assignment of the new value when it doesn't fit, it only
            //       shows a warning
            // need to check if the widget with new location fits the realized frame
            if (p == null || !_isRealized() || maxWidth <= ConfigHelper.getEffectiveVirtualWidth(p.config))
            {
               config.column = value;
               config.x = x;
            }
         }
         
         config.align = ControlEntity.ALIGN_LEFT;
      }
   }
   
   /**
    * Push the location attributes to the client-side.
    * 
    * @param   row
    *          Flag indicating if we are pushing row location or column.
    */
   protected final void pushLocAttrs(boolean row)
   {
      config.locationChanged = true;
      
      if (row)
      {
         pushWidgetAttr(new String[] {"initRow", "initY",
                                      "clientRow", "clientY",
                                      "row", "y",
                                      "locationChanged"},
                        new Object[] {config.initRow, config.initY,
                                      config.clientRow, config.clientY,
                                      config.row, config.y,
                                      config.locationChanged});
         
      }
      else
      {
         pushWidgetAttr(new String[] {"initColumn", "initX",
                                      "clientColumn", "clientX",
                                      "column", "x",
                                      "align", "locationChanged"},
                        new Object[] {config.initColumn, config.initX,
                                      config.clientColumn, config.clientX,
                                      config.column, config.x,
                                      config.align, config.locationChanged});
      }

      config.locationChanged = false;
   }
   
   /**
    * Sets the x or y pixel position for this widget.
    * <p>
    * Sub-classes can override this method; at this point, the parameters are
    * valid.
    * 
    * @param    isUnknown
    *           If <code>true</code> the attribute value being set is an unknown value.
    * @param    value
    *           The x or y pixel value.
    * @param    x
    *           If <code>true</code> <code>value</code> represents an x
    *           position, otherwise y.
    */
   protected void setXOrYWorker(boolean isUnknown, int value, boolean x)
   {
      boolean isRedir = isRedirected();
      
      GenericFrame gfr = getFrame();
      boolean isInsideSetup = (gfr != null && gfr.isInsideSetup());
      
      boolean realized = _isRealized();
      
      if (!isRedir && realized)
      {
         value = Coordinate.adjustMinMaxGuiPixel(value);
      }
      
      CoordinatesConversion cc = LogicalTerminal.coordinates();
     
      // setting the X or Y attribute?
      if (x)
      {
         if (!isInsideSetup)
         {
            // adjust for frame offset first
            value += config.frameXOffset;
         }
         
         double chValue = cc.columnFromPixels(value);
         if (!isRedir)
         {
            chValue = Coordinate.adjustMinMaxGuiChar(chValue);
         }
         
         if (!config.realized || isInsideSetup)
         {
            config.initX = value;
            config.initColumn = chValue;
            
            // reset the client-related coordinates, as they need to be re-computed
            config.clientColumn = BaseConfig.INV_COORD;
            config.clientX = BaseConfig.INV_COORD;
         }

         // TODO: check the value fits the parent frame, see setColumnOrRowWorker
         if (!isInsideSetup)
         {
            config.x = value;
            config.column = chValue;
         }
      }
      else
      {
         if (!isInsideSetup)
         {
            // adjust for frame offset first
            value += config.frameYOffset;
         }
         
         double chValue = cc.rowFromPixels(value);
         if (!isRedir)
         {
            chValue = Coordinate.adjustMinMaxGuiChar(chValue);
         }
         
         if (!config.realized || isInsideSetup)
         {
            config.initY = value;
            config.initRow = chValue;
            
            // reset the client-related coordinates, as they need to be re-computed
            config.clientRow = BaseConfig.INV_COORD;
            config.clientY = BaseConfig.INV_COORD;
         }

         // TODO: check the value fits the parent frame, see setColumnOrRowWorker
         if (!isInsideSetup)
         {
            config.y = value;
            config.row = chValue;
         }
      }
   }
   
   /**
    * The method must return <code>true</code> if this entity accepts assignment of X or Y
    * attributes in ChUI mode. If the method returns <code>false</code>, error 4079
    * is emitted when these attributes are assigned.
    * <p>
    * The default implementation always returns <code>true</code>.  
    * 
    * @return   See above.
    */
   protected boolean canSetPixelPositionInChui()
   {
      return true;
   }
   
   /**
    * Checks if the row or column position for this widget can be set.
    *
    * @param    isUnknown
    *           If <code>true</code> the attribute value being set is an unknown value.
    * @param    value
    *           The 1-based row or column number.
    * @param    row
    *           If <code>true</code> <code>value</code> represents a row number, otherwise 
    *           a column number.
    * 
    * @return   <code>true</code> if the given value is valid for the row or col attribute.
    */
   protected boolean validColumnOrRow(boolean isUnknown, double value, boolean row)
   {
      boolean isChui = LogicalTerminal.isChui();
      
      if (isChui)
      {
         value = Coordinate.scale(value, 0);
         
         if (isUnknown || value < Coordinate.MIN_CHUI_CH)
         {
            String msg = String.format(
                  "**Attribute %s must be greater than or equal to 1 on %s widget", 
                  row ? "ROW" : "COL",
                  type());
            ErrorManager.displayWarning(4081, msg, false);
            return false;
         }
         
         if (value > Coordinate.MAX_CHUI_CH)
         {
            String msg = String.format("Invalid character unit value %d changed to %d", 
                                       Math.abs((int) value), Coordinate.MAX_CHUI_CH);
            ErrorManager.recordOrShowError(13734, msg, false, true, false);
            return false;
         }
         
         value = Coordinate.adjustMinMaxChuiChar(value);
      }
      else
      {
         value = Coordinate.scale(value);
         
         if (isUnknown || (value > -1 && value < 1))
         {
            String msg = String.format(
                  "**Attribute %s must be greater than or equal to 1 on %s widget", 
                  row ? "ROW" : "COL",
                  type());
            ErrorManager.displayWarning(4081, msg, false);
            return false;
         }
         
         if (value < Coordinate.MIN_GUI_CH || value > Coordinate.MAX_GUI_CH)
         {
            double origValue = value;
            value = Coordinate.adjustMinMaxGuiChar(value); 
            String msg;
            if (origValue == (long) origValue)
            {
               msg = String.format("Invalid character unit value %d changed to %d", 
                     Math.abs((long) origValue), Math.abs((long) value));
            }
            else
            {
               msg = String.format("Invalid character unit value %s changed to %s", 
                     Math.abs(origValue), Math.abs(value));
            }
            ErrorManager.recordOrShowError(13734, msg, false, true, false);
            return false;
         }
      }
      // TODO: widget can not go outside of the frame's size (when realized)!

      return true;
   }
   
   /**
    * Sets the row or column position for this widget.
    *
    * @param    isUnknown
    *           If <code>true</code> the attribute value being set is an unknown value.
    * @param    value
    *           The 1-based row or column number.
    * @param    row
    *           If <code>true</code> <code>value</code> represents a row number, otherwise 
    *           a column number.
    */
   protected void setColumnOrRow(boolean isUnknown, double value, boolean row)
   {
      // TO-DO:
      // value > screen-width - frame-width should produce PROGRESS error 4054
      // "**Unable to set COL.  FRAME f does not fit in WINDOW widget. (4054)"
      if (!validColumnOrRow(isUnknown, value, row) && config.realized)
      {
         // if the widget was already realized and this is not a valid coordinate, do nothing
         return;
      }
      
      boolean isChui = LogicalTerminal.isChui();

      value = isChui ? Coordinate.scale(value, 0) : Coordinate.scale(value);
      value = isChui ? Coordinate.adjustMinMaxChuiChar(value) : Coordinate.adjustMinMaxGuiChar(value);
      
      if ((row && config.row == value - 1) || (!row && config.column == value - 1))
      {
         return;
      }
      
      setColumnOrRowWorker(isUnknown, value, row);

      pushLocAttrs(row);
   }
   
   /**
    * Sets the X or Y position for this widget.
    *
    * @param    isUnknown
    *           If <code>true</code> the attribute value being set is an unknown value.
    * @param    value
    *           The X or Y position value.
    * @param    x
    *           If <code>true</code> <code>value</code> represents an X position, otherwise 
    *           a Y position.
    */
   protected void setXorY(boolean isUnknown, int value, boolean x)
   {
      if (isUnknown)
      {
         String msg = String.format(
               "**Unable to assign UNKNOWN value to attribute %s on %s widget", 
               x ? "X" : "Y", 
               type());
         
         ErrorManager.displayWarning(4083, msg, false);
         return;
      }
      
      boolean isChui = LogicalTerminal.isChui();
      
      if (isChui && !canSetPixelPositionInChui())
      {
         String msg = String.format(
               "**Unable to set attribute %s on %s widget in the current environment", 
               x ? "X" : "Y", 
               type());
         
         ErrorManager.recordOrShowError(4079, msg, false, false, false);
         return;
      }
      
      if ((x && config.x == value - 1) || (!x && config.y == value - 1))
      {
         return;
      }
      
      setXOrYWorker(isUnknown, value, x);
      
      pushLocAttrs(!x);
   }

   /**
    * Sets one of the size char attributes, that is WIDTH-CHARS, HEIGHT-CHARS,
    * MIN-WIDTH-CHARS, MIN-HEIGHT-CHARS, MAX-WIDTH-CHARS, MAX-HEIGHT-CHARS,
    * VIRTUAL-WIDTH_CHARS, VIRTUAL-HEIGHT-CHARS.
    *
    * @param    value
    *           The attribute value. <code>null</code> denotes an unknown attribute value.
    * @param    attr
    *           The attribute being set.
    */
   @NoSyncCoordinates
   protected void setSizeChars(Double value, SizeAttribute attr)
   {
      CharsPixels charsPixels = prepareSizeChars(value, attr, this);
      if (charsPixels == null)
      {
         return;
      }

      int valuePixels = charsPixels.pixels;
      value = charsPixels.chars;

      // setting the WIDTH or HEIGHT attribute?
      // TODO: widget can not go outside of the frame's size (when realized)!
      if (attr.isWidth())
      {
         if (attr.isSize())
         {
            if (config.widthPixels == valuePixels && config.fixedWidth)
            {
               // AUTO-RESIZE must be set to OFF, don't push this to client-side until next call
               resetAutoResize();
               return;
            }

            // no matter how we set the size, inside setup or outside, this means fixed size
            config.fixedWidth = true;

            setWidthCharsWorker(value);
            setWidthPixelsWorker(valuePixels);
         }
         else if (attr.isMin())
         {
            setMinWidthCharsWorker(value);
            setMinWidthPixelsWorker(valuePixels);
         }
         else if (attr.isMax())
         {
            setMaxWidthCharsWorker(value);
            setMaxWidthPixelsWorker(valuePixels);
         }
         else if (attr.isVirt())
         {
            setVirtWidthCharsWorker(value);
            setVirtWidthPixelsWorker(valuePixels);
         }
         else
         {
            throw new IllegalStateException();
         }
      }
      else
      {
         if (attr.isSize())
         {
            if (config.heightPixels == valuePixels && config.fixedHeight)
            {
               // AUTO-RESIZE must be set to OFF, don't push this to client-side until next call
               resetAutoResize();
               return;
            }
            
            setHeightCharsWorker(value);
            setHeightPixelsWorker(valuePixels);

            // no matter how we set the size, inside setup or outside, this means fixed size
            config.fixedHeight = true;
         }
         else if (attr.isMin())
         {
            setMinHeightCharsWorker(value);
            setMinHeightPixelsWorker(valuePixels);
         }
         else if (attr.isMax())
         {
            setMaxHeightCharsWorker(value);
            setMaxHeightPixelsWorker(valuePixels);
         }
         else if (attr.isVirt())
         {
            setVirtHeightCharsWorker(value);
            setVirtHeightPixelsWorker(valuePixels);
         }
         else
         {
            throw new IllegalStateException();
         }
      }

      pushSizeAttrs();
   }

   /**
    * Performs a validation of widget size value.
    * 
    * @param    value
    *           The input value to validate.
    * @param    attr
    *           Denotes the kind of size attribute.
    * 
    * @return   <code>true</code> when valid, <code>false</code> otherwise
    */
   protected boolean validateSize(Number value, SizeAttribute attr)
   {
      return validateSize(value, attr, this);
   }

   /**
    * Sets one of the size pixel attributes, that is WIDTH-PIXELS, HEIGHT-PIXELS,
    * MIN-WIDTH-PIXELS, MIN-HEIGHT-PIXELS, MAX-WIDTH-PIXELS, MAX-HEIGHT-PIXELS,
    * VIRTUAL-WIDTH_PIXELS, VIRTUAL-HEIGHT-PIXELS.
    *
    * @param    value
    *           The attribute value. <code>null</code> denotes an unknown attribute value.
    * @param    attr
    *           The attribute being set.
    */
   @NoSyncCoordinates
   protected void setSizePixels(Integer value, SizeAttribute attr)
   {
      CharsPixels charsPixels = prepareSizePixels(value, attr, this);
      if (charsPixels == null)
      {
         return;
      }

      value = charsPixels.pixels;
      double charValue = charsPixels.chars;
      
      if (attr.isWidth())
      {
         if (attr.isSize())
         {
            if (config.widthPixels == value && config.fixedWidth)
            {
               // AUTO-RESIZE must be set to OFF, don't push this to client-side until next call
               resetAutoResize();
               return;
            }

            // no matter how we set the size, inside setup or outside, this means fixed size
            config.fixedWidth = true;
            
            setWidthPixelsWorker(value);
            setWidthCharsWorker(charValue);
         }
         else if (attr.isMin())
         {
            setMinWidthPixelsWorker(value);
            setMinWidthCharsWorker(charValue);
         }
         else if (attr.isMax())
         {
            setMaxWidthPixelsWorker(value);
            setMaxWidthCharsWorker(charValue);
         }
         else if (attr.isVirt())
         {
            setVirtWidthPixelsWorker(value);
            setVirtWidthCharsWorker(charValue);
         }
         else
         {
            throw new IllegalStateException();
         }
      }
      else
      {
         if (attr.isSize())
         {
            if (config.heightPixels == value && config.fixedHeight)
            {
               // AUTO-RESIZE must be set to OFF, don't push this to client-side until next call
               resetAutoResize();
               return;
            }

            setHeightPixelsWorker(value);
            setHeightCharsWorker(charValue);
   
            // no matter how we set the size, inside setup or outside, this means fixed size
            config.fixedHeight = true;
         }
         else if (attr.isMin())
         {
            setMinHeightPixelsWorker(value);
            setMinHeightCharsWorker(charValue);
            
         }
         else if (attr.isMax())
         {
            setMaxHeightPixelsWorker(value);
            setMaxHeightCharsWorker(charValue);
         }
         else if (attr.isVirt())
         {
            setVirtHeightPixelsWorker(value);
            setVirtHeightCharsWorker(charValue);
         }
         else
         {
            throw new IllegalStateException();
         }
      }

      pushSizeAttrs();
   }
   
   /**
    * Checks whether the supplied widget can be used as a valid parent
    * of this widget.
    * 
    * @param    widget
    *           The widget reference to check.
    * 
    * @return   <code>true</code> if the supplied widget can be used
    *           as a valid parent, <code>false</code> otherwise.
    */
   protected boolean validateParent(GenericWidget<?> widget)
   {
      String parentType = widget.type();
      String[] validTypes = getValidParentTypes();
      boolean match = false;
      
      for (String validType : validTypes)
      {
         if (validType.equals(parentType))
         {
            match = true;
            break;
         }
      }
      
      if (!match)
      {
         String msg = String.format("**%s widget cannot be a parent of a %s widget", 
                                    parentType, type());
         ErrorManager.displayWarning(4075, msg, false);
      }
      
      return match;
   }

   /**
    * Returns the widget types that can act as parent for this widget.
    *
    * @return   see above.
    */
   protected String[] getValidParentTypes()
   {
      // parent must be a field-group
      return new String[] {LegacyResource.FIELD_GROUP};
   }
   
   /**
    * Collect the size-related attributes which can be assigned by this widget.
    * 
    * @param    names
    *           The collection where to add the field names.
    * @param    vals
    *           The collection where to add the field values.
    */
   protected void getSizeAttrs(List<String> names, List<Object> vals)
   {
      names.addAll(Arrays.asList("fixedWidth", "fixedHeight", 
                                 "heightChars", "widthChars", 
                                 "heightPixels", "widthPixels", 
                                 "clientHeightChars", "clientWidthChars", 
                                 "clientHeightPixels", "clientWidthPixels"));
      vals.addAll(Arrays.asList(config.fixedWidth, config.fixedHeight, 
                                config.heightChars, config.widthChars, 
                                config.heightPixels, config.widthPixels, 
                                config.clientHeightChars, config.clientWidthChars, 
                                config.clientHeightPixels, config.clientWidthPixels));
   }

   /**
    * Push all the size-related attributes to the client-size.
    */
   protected void pushSizeAttrs()
   {
      if (frame == null || frame.isInsideSetup())
      {
         return;
      }
      
      List<String> names = new ArrayList<>();
      List<Object> vals = new ArrayList<>();
      
      getSizeAttrs(names, vals);
      
      pushWidgetAttr(names.toArray(new String[names.size()]), vals.toArray(new Object[vals.size()]));
   }

   /**
    * Gets the WIDTH-CHARS writable attribute.
    *
    * @return   The current value of the WIDTH-CHARS attribute.
    */
   protected decimal _getWidthChars()
   {
      return new decimal(config.clientWidthChars != BaseConfig.INV_COORD ?
                         config.clientWidthChars : config.widthChars);
   }

   /**
    * Gets the HEIGHT-CHARS writable attribute.
    *
    * @return   The current value of the HEIGHT-CHARS attribute.
    */
   protected decimal _getHeightChars()
   {
      return new decimal(config.clientHeightChars != BaseConfig.INV_COORD ?
                         config.clientHeightChars : config.heightChars);
   }

   /**
    * Get rectangle width pixels.
    * SIZE-PIXELS option.
    *
    * @return  The width of the widget in pixels.
    */
   protected integer _getWidthPixels()
   {
      return new integer(config.clientWidthPixels != BaseConfig.INV_COORD ?
                         config.clientWidthPixels : config.widthPixels);
   }

   /**
    * Gets the HEIGHT-PIXELS writable attribute.
    *
    * @return  The height of the widget in pixels.
    */
   protected integer _getHeightPixels()
   {
      return new integer(config.clientHeightPixels != BaseConfig.INV_COORD ?
                         config.clientHeightPixels : config.heightPixels);
   }

   /**
    * Check if we are in a redirected state.
    * 
    * @return   See above.
    */
   private boolean isRedirected()
   {
      return UnnamedStreams.isOut() || 
             (config.frameId != -1 && LogicalTerminal.getFrame(config.frameId).isOutRedirected());
   }

   /**
    * Container for size in chars and pixels.
    */
   static class CharsPixels
   {
      /** Size in chars. */
      public double chars;

      /** Size in pixels. */
      public int pixels;

      /**
       * Constructor.
       *
       * @param chars
       *        Size in chars.
       * @param pixels
       *        Size in pixels.
       */
      CharsPixels(double chars, int pixels)
      {
         this.chars = chars;
         this.pixels = pixels;
      }
   }
}