/*
** Module   : ColorPalette.java
** Abstract : Declares the supported terminal colors.
**
** Copyright (c) 2014-2017, Golden Code Development Corporation.
**
** -#- -I- --Date-- ---------------------------------Description----------------------------------
** 001 HC  20140119 Created initial version.
** 002 MAG 20140211 Map entries to AWT Color.
** 003 MAG 20140821 Add new color entries for Windows OS.  Provide conversion method for color 
*                   number to palette index and from palette index to color number. 
*/
/*
** This program is free software: you can redistribute it and/or modify
** it under the terms of the GNU Affero General Public License as
** published by the Free Software Foundation, either version 3 of the
** License, or (at your option) any later version.
**
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
** GNU Affero General Public License for more details.
**
** You may find a copy of the GNU Affero GPL version 3 at the following
** location: https://www.gnu.org/licenses/agpl-3.0.en.html
** 
** Additional terms under GNU Affero GPL version 3 section 7:
** 
**   Under Section 7 of the GNU Affero GPL version 3, the following additional
**   terms apply to the works covered under the License.  These additional terms
**   are non-permissive additional terms allowed under Section 7 of the GNU
**   Affero GPL version 3 and may not be removed by you.
** 
**   0. Attribution Requirement.
** 
**     You must preserve all legal notices or author attributions in the covered
**     work or Appropriate Legal Notices displayed by works containing the covered
**     work.  You may not remove from the covered work any author or developer
**     credit already included within the covered work.
** 
**   1. No License To Use Trademarks.
** 
**     This license does not grant any license or rights to use the trademarks
**     Golden Code, FWD, any Golden Code or FWD logo, or any other trademarks
**     of Golden Code Development Corporation. You are not authorized to use the
**     name Golden Code, FWD, or the names of any author or contributor, for
**     publicity purposes without written authorization.
** 
**   2. No Misrepresentation of Affiliation.
** 
**     You may not represent yourself as Golden Code Development Corporation or FWD.
** 
**     You may not represent yourself for publicity purposes as associated with
**     Golden Code Development Corporation, FWD, or any author or contributor to
**     the covered work, without written authorization.
** 
**   3. No Misrepresentation of Source or Origin.
** 
**     You may not represent the covered work as solely your work.  All modified
**     versions of the covered work must be marked in a reasonable way to make it
**     clear that the modified work is not originating from Golden Code Development
**     Corporation or FWD.  All modified versions must contain the notices of
**     attribution required in this license.
*/

package com.goldencode.p2j.ui;

/**
 * Set of supported terminal colors.
 * <p>
 * The implicit ordinal created by Java for this enum directly maps 
 * each enum "element" to the "COLOR_element" constant in NCurses/Curses. 
 * This mapping is implicitly utilized in NCurses/Curses drivers without 
 * any further translation. Other drivers will have to provide some kind 
 * of mapping from these values to their native rendering. 
 * <p>
 * WARNING: Do not change the order of enum items.
 * <p>
 * The color palette is implemented as a Java enum type.  The first 8 entries in color palette
 * are used on Linux / Unix only.  On Windows all 16 entries are used. The color phrase should
 * works on both ChUI and GUI interfaces.  On Windows GUI the colors are indexes in COLOR-TABLE
 * structure. Because the ordinal numbers of palette enum are different than color numbers
 * conversion methods has been designed to do the conversions from color number to palette index
 * and from palette index to color number.
 * <p>
 * Conversion Table:
 * <p>
 * <pre>
 * -----------------------------------------------------------------------------------------------
 * Index Color          Abbreviation            Palette     Color          dos-hex    Color
 * -----------------------------------------------------------------------------------------------
 *  0    Black          Bla, Blk                 0          BLACK          0          Black
 *  1    Blue           Blu                     11          DARK_BLUE      1          Blue
 *  2    Green          Gre, Grn                 9          DARK_GREEN     2          Green
 *  3    Cyan           C                       13          DARK_CYAN      3          Aqua
 *  4    Red            Red                      8          DARK_RED       4          Red
 *  5    Magenta        Ma                      12          DARK_MAGENTA   5          Purple
 *  6    Brown          Bro, Brn                10          BROWN          6          Yellow
 *  7    Gray           Gra, Gry                14          GRAY           7          White
 *  8    Dark-Gray      D-Gra                   15          DARK_GRAY      8          Gray
 *  9    Light-Blue     Lt-Blu                   4          BLUE           9          Light Blue
 * 10    Light-Green    Lt-Gre                   2          GREEN          A          Light Green
 * 11    Light-Cyan     Lt-C                     6          CYAN           B          Light Green
 * 12    Light-Red      Lt-Red                   1          RED            C          Light Green
 * 13    Light-Magenta  Lt-Ma                    5          MAGENTA        D          Light Purple
 * 14    Yellow         Y Light-Brown Lt-Br      3          YELLOW         E          Light Yellow
 * 15    White          W                        7          WHITE          F          Bright White
 * -----------------------------------------------------------------------------------------------
 * </pre>
 */

/**
 * ChUI color palette implementation.
 */
public enum ColorPalette
{   
   BLACK(0, 0, 0, 0),
   RED(12, 255, 0, 0),
   GREEN(10, 0, 255, 0),
   YELLOW(14, 255, 255, 0),
   BLUE(9, 0, 0, 255),
   MAGENTA(13, 255, 0, 255),
   CYAN(11, 0, 255, 255),
   WHITE(15, 255, 255, 255),
   DARK_RED(4, 128, 0, 0),
   DARK_GREEN(2, 0, 128, 0),
   BROWN(6, 128, 128, 0),
   DARK_BLUE(1, 0, 0, 128),                          
   DARK_MAGENTA(5, 128, 0, 128),
   DRAK_CYAN(3, 0, 128, 128),
   GRAY(7, 128, 128, 128),
   DARK_GRAY(8, 192, 192, 192);
         
   /** Conversion table color number to palette index */
   private static int[] paletteIndexes;

   /** Conversion table palette index to color number */
   private static int[] colorIndexes;

   /** Color number (index in COLOR-TABLE) */
   private int index;
   
   /** Red. */
   private int r;

   /** Green. */
   private int g;
   
   /** Blue. */
   private int b;
   
   static
   {
      ColorPalette[] values = values();
      
      paletteIndexes = new int[values.length];
      colorIndexes = new int[values.length];
      
      for (int i = 0; i < values.length; i++)
      {
         int colorNum = values[i].getIndex();
         paletteIndexes[colorNum] = i;
         colorIndexes[i] = colorNum;
      }
   }
   
   /**
    * Convert color number to palette index.
    * 
    * @param    color
    *           Color number. Must be between 0 and 15.
    *           
    * @return   Palette index.
    */
   public static int colorToPalette(int color)
   {
      return paletteIndexes[color & 0x0F];
   }

   /**
    * Convert color number to palette index.
    * 
    * @param    index
    *           Palette index. Must be between 0 and 15.
    *                                                 
    * @return   Color number.
    */
   public static int paletteToColor(int index)
   {
      return colorIndexes[index & 0x0F];
   }
   
   /**
    * Constructor.
    * 
    * @param    index
    *           Color index in COLOR-TBALE.
    * @param    r
    *           Red component.
    * @param    g
    *           Green component.
    * @param    b
    *           Blue component
    */
   private ColorPalette(int index, int r, int g, int b)
   {
      this.index = index;
      this.r     = r;
      this.g     = g;
      this.b     = b;
   }
                                                             
   /**
    * Get color index number in COLOR-TABLE.
    * 
    * @return   The color-table index.
    */
   public int getIndex()
   {
      return index;
   }

   /**
    * Get the red component of this color.
    * 
    * @return   The red component.
    */
   public int getRed()
   {
      return r;
   }

   /**
    * Get the green component of this color.
    * 
    * @return   The green component.
    */
   public int getGreen()
   {
      return g;
   }

   /**
    * Get the blue component of this color.
    *                                                        
    * @return   The blue component.
    */
   public int getBlue()
   {
      return b;
   }
}