/*
** Module   : ColorSpec.java
** Abstract : Implements Progress color-phrase details.
**
** Copyright (c) 2005-2017, Golden Code Development Corporation.
**
** -#- -I- --Date-- --JPRM-- ----------------------------Description------------------------------
** 001 NVS 20051026   @23274 Created initial version.
** 002 SIY 20080331   @37736 Improved support for COLOR_VALUE mode. Ref: #37732
** 003 NVS 20080420   @38047 Color value that encodes one of the regular color specifications 
**                           MESSAGE/INPUT/NORMAL is not stored as value but converts into the
**                           corresponding specification right away.
** 004 SIY 20080429   @38143 Return invalid color if it is not present in color map.
** 005 SIY 20101122          Minor cleanups.
** 006 SIY 20101223          Complete rewrite due to change of color handling approach.
** 007 CS  20121031          Removed client specific references
** 008 HC  20140119          Implemented runtime support for the COLOR protermcap-attribute.
** 009 MAG 20140820          Implements WINDOWS colors.
** 010 CA  20171009          Reworked to all both ChUI and GUI color conversions.
** 011 CA  20171019          Javadoc fix.
*/
/*
** This program is free software: you can redistribute it and/or modify
** it under the terms of the GNU Affero General Public License as
** published by the Free Software Foundation, either version 3 of the
** License, or (at your option) any later version.
**
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
** GNU Affero General Public License for more details.
**
** You may find a copy of the GNU Affero GPL version 3 at the following
** location: https://www.gnu.org/licenses/agpl-3.0.en.html
** 
** Additional terms under GNU Affero GPL version 3 section 7:
** 
**   Under Section 7 of the GNU Affero GPL version 3, the following additional
**   terms apply to the works covered under the License.  These additional terms
**   are non-permissive additional terms allowed under Section 7 of the GNU
**   Affero GPL version 3 and may not be removed by you.
** 
**   0. Attribution Requirement.
** 
**     You must preserve all legal notices or author attributions in the covered
**     work or Appropriate Legal Notices displayed by works containing the covered
**     work.  You may not remove from the covered work any author or developer
**     credit already included within the covered work.
** 
**   1. No License To Use Trademarks.
** 
**     This license does not grant any license or rights to use the trademarks
**     Golden Code, FWD, any Golden Code or FWD logo, or any other trademarks
**     of Golden Code Development Corporation. You are not authorized to use the
**     name Golden Code, FWD, or the names of any author or contributor, for
**     publicity purposes without written authorization.
** 
**   2. No Misrepresentation of Affiliation.
** 
**     You may not represent yourself as Golden Code Development Corporation or FWD.
** 
**     You may not represent yourself for publicity purposes as associated with
**     Golden Code Development Corporation, FWD, or any author or contributor to
**     the covered work, without written authorization.
** 
**   3. No Misrepresentation of Source or Origin.
** 
**     You may not represent the covered work as solely your work.  All modified
**     versions of the covered work must be marked in a reasonable way to make it
**     clear that the modified work is not originating from Golden Code Development
**     Corporation or FWD.  All modified versions must contain the notices of
**     attribution required in this license.
*/

package com.goldencode.p2j.ui;

import java.util.*;

import org.apache.commons.logging.*;

import com.goldencode.p2j.net.SessionManager;
import com.goldencode.p2j.security.*;
import com.goldencode.p2j.ui.ColorTable.ColorPair;
import com.goldencode.p2j.ui.chui.ThinClient;
import com.goldencode.p2j.util.*;

/**
 * Class ColorSpec supports the color-phrase specification.
 * <ul>
 *   <li>- defines color-phrase instance
 *   <li>- NORMAL | INPUT | MESSAGES
 *   <li>- protermcap-attribute
 *   <li>- BLINK- BRIGHT- fgnd-color bgnd-color (WINDOWS only)
 *   <li>- BLINK- RVV- UNDERLINE- BRIGHT- fgnd-color (WINDOWS only).
 *   <li>- VALUE(expression)
 * </ul>
 * The following options are NOT SUPPORTED:
 * <ul>
 *   <li>- dos-hex-attribute
 *         This option is fully documented but not yet implemented.
 *         Testing on OpenEdge for Windows environment found that isn't supported.
 *         Seems to be supported only in P4GL for MS-DOS version.  
 * </ul>
 * <p>
 * This class does not interpret the meaning of the color value.
 *
 * @author   NVS
 * @version  2.0
 */
public class ColorSpec
{
   /** Logger */
   private static final Log LOG = LogFactory.getLog(ColorSpec.class);
   
   /** Context-local global data area for color state */
   private static ContextLocal<WorkArea> workArea = new ContextLocal<WorkArea>()
   {
      /**
       * Initializes the work area, the first time it is requested within a new context.
       *
       * @return   The newly instantiated work area.
       */
      protected synchronized WorkArea initialValue()
      {
         boolean onClientSide = SessionManager.get().isLeaf();
         boolean isChUI = onClientSide ? ThinClient.getInstance().isChui()
                                       : LogicalTerminal.isChui();
         
         return isChUI ? new ChUIWorkArea() : new GUIWorkArea();
      }   
   };

   /** COLOR_VALUE definition */
   public static final int COLOR_VALUE = 3;

   /** Color name constant for NORMAL color. */
   public static final String COLOR_NORMAL   = "NORMAL";

   /** Color name constant for INPUT color. */
   public static final String COLOR_INPUT    = "INPUT";

   /** Color name constant for MESSAGES color. */
   public static final String COLOR_MESSAGES = "MESSAGES";

   /** Color name constant for BOLD color. */
   public static final String COLOR_BOLD     = "BOLD";

   /** Color name constant for BLINK color. */
   public static final String COLOR_BLINK    = "BLINK";
   
   /** Color name (String) or index (Integer) definition value */
   private Object colorValue = COLOR_NORMAL;

   /**
    * Copy constructor.
    *
    * @param   color
    *          existing color specification to duplicate
    */
   public ColorSpec(ColorSpec color)
   {
      this.colorValue = color.colorValue;
   }

   /**
    * Main constructor.
    *
    * @param   value
    *          color name.
    */
   public ColorSpec(String value)
   {
      this.colorValue = (value == null) ? null : value.trim().toUpperCase();
   }

   /**
    * Build color specification using integer value.
    *
    * @param   intValue
    *          Color number.
    */
   public ColorSpec(int intValue)
   {
      this.colorValue = intValue;
   }

   /**
    * Get the {@link WorkArea} instance for this context.
    * 
    * @return   See above.
    */
   private static <T extends WorkArea> T locate()
   {
      return (T) workArea.get();
   }
   
   /**
    * Process color attribute.
    * 
    * @param   attribute
    *          Current attributes.
    * @param   attr
    *          Attribute string.
    * 
    * @return  Update attributes.
    */
   private static int processAttribute(int attribute, String attr)
   {
      if (attr != null && !attr.isEmpty())
      {
         if ("BLINK-".equals(attr))
         {
            attribute |= ColorAttribute.BLINK.getValue();
         }
         else if ("RVV-".equals(attr))
         {
            attribute |= ColorAttribute.REVERSE.getValue();
         }
         else if ("UNDERLINE-".equals(attr))
         {
            attribute |= ColorAttribute.UNDERLINE.getValue();
         }
         else if ("BRIGHT-".equals(attr))
         {
            attribute |= ColorAttribute.BOLD.getValue();
         }
      }
      return attribute;
   }
   
   /**
    * Convert color specification into real color.
    *
    * @return  color value
    */
   public Color convert()
   {
      WorkArea wa = locate();
      
      return wa.colorConvertor.convert(this);
   }

   /**
    * Convert the color based on the current client type.
    */
   private static interface ColorConvertor
   {
      /**
       * Convert color specification into real color.
       *
       * @return  color value
       */
      public Color convert(ColorSpec cs);
   }
   
   /**
    * Color convertor for ChUI clients.
    */
   private static class ChUIColorConvertor
   implements ColorConvertor
   {
      /**
       * Convert color specification into real color.
       *
       * @return  color value
       */
      @Override
      public Color convert(ColorSpec cs)
      {
         ChUIWorkArea wa = locate();
         
         // try the global static map first
         Color result = wa.globalColorMap.get(cs.colorValue);
         if (result != null)
         {
            return result;
         }
         // search on legacy client runtime platform color table
         if (EnvironmentOps.OS_WIN.equals(EnvironmentOps.getLegacyPlatform()))
         {
            // search on Windows color map first.
            result = wa.windowsColorMap.get(cs.colorValue);
            if (result == null)
            {
               // if not in the map process it.
               result = processWindowsColor(cs);
            }
         }
         else
         {
            result = processLinuxColor(cs);
         }
         // fall back to the NORMAL color
         return (result == null) ? wa.globalColorMap.get(COLOR_NORMAL) : result;
      }
      
      /**
       * Process Unix/Linux colors.
       * 
       * @return  Instance of Color.
       */
      private Color processLinuxColor(ColorSpec cs)
      {
         ChUIWorkArea wa = locate();

         Color result = null;
         // now the colors from the terminal specific color map, if the map was initialized
         if (wa.terminalColorMap != null)
         {
            character terminalType = LogicalTerminal.getTerminalType();
            if (terminalType != null && !terminalType.isUnknown())
            {
               String terminalName = terminalType.getValue().toUpperCase();
               Map<Object, Color> terminalColors = wa.terminalColorMap.get(terminalName);
               if (terminalColors != null)
               {
                  result = terminalColors.get(cs.colorValue);
               }
            }
         }      
         return result;
      }
      
      /**
       * Process Windows colors.
       * BLINK- BRIGHT- fgnd-color bgnd-color
       * BLINK- RVV- UNDERLINE- BRIGHT- fgnd-color
       * 
       * @return  Instance of Color. 
       */
      private Color processWindowsColor(ColorSpec cs)
      {
         ChUIWorkArea wa = locate();

         Color color =  wa.globalColorMap.get(COLOR_NORMAL);      

         if (cs.colorValue instanceof String)
         {
            int attribute = ColorAttribute.NORMAL.getValue();
            
            String[] parts = ((String) cs.colorValue).split("/");
            switch (parts.length)
            {
               case 1:
                  // BLINK- RVV- UNDERLINE- BRIGHT- fgnd-color
                  if (parts[0].matches(wa.pattern1))
                  {
                     // foreground
                     for (int i = 2; i < 6; i++)
                     {
                        String attr = parts[0].replaceFirst(wa.pattern1, "$" + i);
                        attribute = processAttribute(attribute, attr);
                     }
                     // color name
                     String fgColorName = parts[0].replaceFirst(wa.pattern1, "$6");
                     int fg = getStandardColors(fgColorName);
                     if (fg == -1)
                     {
                        Integer fgColorIndex = wa.windowsColors.get(fgColorName);
                        fg = (fgColorIndex == null) ? 0 : 
                              ColorPalette.colorToPalette(fgColorIndex.intValue()) + 1;
                     }
                     else
                     {
                        fg = (fg >> 16) & 0xFFFF; 
                     }
                     // build and save color
                     color = makeColor(attribute, fg, 0);
                     // add a new entry into Windows color table
                     wa.windowsColorMap.put(cs.colorValue, color);
                     wa.windowsColorMap.put(wa.index++, color);
                  }
                  break;
               case 2:
                  // BLINK- BRIGHT- fgnd-color bgnd-color
                  if (parts[0].matches(wa.pattern0) && parts[1].matches(wa.pattern0))
                  {
                     // foreground
                     for (int i = 2; i < 4; i++)
                     {
                        String attr = parts[0].replaceFirst(wa.pattern0, "$" + i);
                        attribute = processAttribute(attribute, attr);
                     }
                     // color name
                     String fgColorName = parts[0].replaceFirst(wa.pattern0, "$4");
                     int fg = getStandardColors(fgColorName);
                     if (fg == -1)
                     {
                        Integer fgColorIndex = wa.windowsColors.get(fgColorName);
                        fg = (fgColorIndex == null) ? 0 :
                              ColorPalette.colorToPalette(fgColorIndex.intValue()) + 1;
                     }
                     else
                     {
                        fg = (fg >> 16) & 0xFFFF; 
                     }
                     
                     // background
                     for (int i = 2; i < 4; i++)
                     {
                        String attr = parts[1].replaceFirst(wa.pattern0, "$" + i);
                        processAttribute(attribute, attr);
                     }
                     // color name
                     String bgColorName = parts[1].replaceFirst(wa.pattern0, "$4");
                     int bg = getStandardColors(bgColorName);
                     if (bg == -1)
                     {
                        Integer bgColorIndex = wa.windowsColors.get(bgColorName);
                        bg = (bgColorIndex == null) ? 0 :
                              ColorPalette.colorToPalette(bgColorIndex.intValue()) + 1;
                     }
                     else
                     {
                        bg = bg & 0xFFFF; 
                     }
   
                     // build and save color
                     color = makeColor(attribute, fg, bg);
                     // add a new entry into Windows color table
                     wa.windowsColorMap.put(cs.colorValue, color);
                     wa.windowsColorMap.put(wa.index++, color);
                  }
                  break;
            }
         }
         
         return color;
      }

      /**
       * Build a Color object.
       * 
       * @param   attribute
       *          Color attributes.
       * @param   fg
       *          Foreground color index.
       * @param   bg
       *          Background color index.
       * 
       * @return  A new Color object instance.
       */
      private Color makeColor(int attribute, int fg, int bg)
      {
         int color = (bg & 0xFFFF) | ((fg << 16) & 0xFFFF0000);
         return new Color(attribute, color);
      }

      /**
       * Check if color is missing or is a standard color NORMAL, INPUT or MESSAGES
       * If yes return the standard color otherwise -1.
       * 
       * @param   colorName
       *          color name.
       *          
       * @return  see above
       */
      private int getStandardColors(String colorName)
      {
         ChUIWorkArea wa = locate();

         // color is missing
         if (colorName == null || colorName.trim().isEmpty())
         {
            return wa.globalColorMap.get(COLOR_NORMAL).color();
         }
         // NORMAL color
         else if (COLOR_NORMAL.equals(colorName))
         {
            return wa.globalColorMap.get(COLOR_NORMAL).color();         
         }
         // INPUT color
         else if (COLOR_INPUT.equals(colorName))
         {
            return wa.globalColorMap.get(COLOR_INPUT).color();         
         }
         // MESSAGES color
         else if (COLOR_MESSAGES.equals(colorName))
         {
            return wa.globalColorMap.get(COLOR_MESSAGES).color();         
         }
         // other color
         return -1;
      }
   }
   
   /**
    * Color convertor for ChGUI clients.
    */
   private static class GUIColorConvertor
   implements ColorConvertor
   {
      /**
       * Check if color is missing or is a standard color NORMAL, INPUT or MESSAGES.
       * If yes return the standard color otherwise -1.
       * 
       * @param   colorName
       *          color name.
       * @param   bgColor
       *          Flag indicating to return either the background or foreground color.
       *          
       * @return  see above
       */
      private static int getStandardColors(String colorName, boolean bgColor)
      {
         character colorVal = null;
         
         // NORMAL color
         if (colorName == null || colorName.trim().isEmpty() || COLOR_NORMAL.equals(colorName))
         {
            colorVal = EnvironmentOps.getKeyValue(ColorTable.COLOR_SECTION, COLOR_NORMAL);
         }
         // INPUT color
         else if (COLOR_INPUT.equals(colorName))
         {
            colorVal = EnvironmentOps.getKeyValue(ColorTable.COLOR_SECTION, COLOR_INPUT);
         }
         // MESSAGES color
         else if (COLOR_MESSAGES.equals(colorName))
         {
            colorVal = EnvironmentOps.getKeyValue(ColorTable.COLOR_SECTION, COLOR_MESSAGES);
         }
         
         if (colorVal != null)
         {
            ColorPair pair = ColorTable.createColorPair(colorVal.toJavaType());
            
            if (pair == null)
            {
               // TODO: there should be a default, in-memory environment, with all the default 
               // values (font-table, color-table, etc); for now, just hardcode this...
               
               if (colorName == null || colorName.trim().isEmpty() || COLOR_NORMAL.equals(colorName))
               {
                  pair = GUIWorkArea.NORMAL;
               }
               // INPUT color
               else if (COLOR_INPUT.equals(colorName))
               {
                  pair = GUIWorkArea.INPUT;
               }
               // MESSAGES color
               else if (COLOR_MESSAGES.equals(colorName))
               {
                  pair = GUIWorkArea.MESSAGES;
               }
            }
            
            return bgColor ? pair.bgColor : pair.fgColor;
         }

         // other color
         return -1;
      }
      
      /**
       * Build a Color object.
       * 
       * @param   attribute
       *          Color attributes.
       * @param   index
       *          The color index.
       *          
       * @return  The real color index for the given color.
       */
      private static int makeColor(int attribute, int index)
      {
         final int[] BLINK     = { 0, 129, 130, 131, 132, 133, 134, 135, 135, 137, 138, 139, 140, 141, 142, 142, 143 };
         final int[] BOLD      = { 0,   9,  10,  11,  12,  13,  14,  15,  15,   9,  10,  11,  12,  13,  14,  14,  15 };
         final int[] UNDERLINE = { 0,   1,   2,   3,   4,   5,   6,   7,   7,   9,  10,  11,  12,  13,  14,  14,  15 };
         final int[] REVERSE   = { 0,   1,   2,   3,   4,   5,   6,   7,   7,   9,  10,  11,  12,  13,  14,  14,  15 };
         final int[] NORMAL    = { 0,   1,   2,   3,   4,   5,   6,   7,   7,   9,  10,  11,  12,  13,  14,  14,  15 };
         
         ColorAttribute ca = ColorAttribute.valueOf(attribute);
         int[] colIdx = null;
         switch (ca)
         {
            case BLINK:
               colIdx = BLINK;
               break;
            case BOLD:
               colIdx = BOLD;
               break;
            case UNDERLINE:
               colIdx = UNDERLINE;
               break;
            case REVERSE:
               colIdx = REVERSE;
               break;
            case NORMAL:
               colIdx = NORMAL;
               break;
         }
         
         return (index < 0 || index >= colIdx.length ? -1 : colIdx[index]);
      }

      /**
       * Convert color specification into real color.
       *
       * @return  color value
       */
      @Override
      public Color convert(ColorSpec cs)
      {
         GUIWorkArea wa = locate();

         Color color = null;

         if (cs.colorValue instanceof String)
         {
            int attribute = ColorAttribute.NORMAL.getValue();
            int bg = -1;
            int fg = -1;
            
            String[] parts = ((String) cs.colorValue).split("/", -1);
            switch (parts.length)
            {
               case 2:
                  // BLINK- RVV- UNDERLINE- BRIGHT- bgnd-color
                  if (parts[1].matches(wa.pattern))
                  {
                     // background
                     for (int i = 2; i < 6; i++)
                     {
                        String attr = parts[1].replaceFirst(wa.pattern, "$" + i);
                        attribute = processAttribute(attribute, attr);
                     }
                     // color name
                     String bgColorName = parts[1].replaceFirst(wa.pattern, "$6");
                     if (bgColorName.isEmpty())
                     {
                        bg = 255;
                     }
                     else if ("?".equals(bgColorName))
                     {
                        // unknown, do not change
                        bg = -1;
                     }
                     else
                     {
                        bg = getStandardColors(bgColorName, true);
                        if (bg == -1)
                        {
                           Integer bgColorIndex = wa.windowsColors.get(bgColorName);
                           bg = (bgColorIndex == null) ? -1 : makeColor(attribute, bgColorIndex);
                        }
                     }
                  }
                  
               // fallback to foreground color
               case 1:
                  // BLINK- RVV- UNDERLINE- BRIGHT- fgnd-color
                  if (parts[0].matches(wa.pattern))
                  {
                     // foreground
                     for (int i = 2; i < 6; i++)
                     {
                        String attr = parts[0].replaceFirst(wa.pattern, "$" + i);
                        attribute = processAttribute(attribute, attr);
                     }
                     // color name
                     String fgColorName = parts[0].replaceFirst(wa.pattern, "$6");
                     
                     if (fgColorName.isEmpty() || "?".equals(fgColorName))
                     {
                        fg = -1;
                     }
                     else
                     {
                        fg = getStandardColors(fgColorName, false);
                        if (fg == -1)
                        {
                           Integer fgColorIndex = wa.windowsColors.get(fgColorName);
                           fg = (fgColorIndex == null) ? -1 : makeColor(attribute, fgColorIndex);
                        }
                        else if (parts.length == 1)
                        {
                           // if only one standard color mentioned, then background is computed, too
                           bg = getStandardColors(fgColorName, true);
                        }
                     }
                  }
                  break;
            }

            // build and save color
            color = new ColorGui(fg, bg);

            // add a new entry into Windows color table
            wa.windowsColorMap.put(cs.colorValue, color);
            wa.windowsColorMap.put(wa.index++, color);
         }
         
         return color;      
      }
   }
   
   /**
    * GUI-specific work area with context-local data.
    */
   private static class GUIWorkArea
   extends WorkArea
   {
      /** The standard NORMAL GUI color pair. */
      private static final ColorPair NORMAL = new ColorPair(0, 15);
      
      /** The standard INPUT GUI color pair. */
      private static final ColorPair INPUT = new ColorPair(15, 0);
       
      /** The standard MESSAGES GUI color pair. */
      private static final ColorPair MESSAGES = new ColorPair(15, 1);

      /**
       * A map of color with all their abbreviations, to their color index in the color table
       * provided by the legacy 4GL documentation.
       * <p>
       * If the second value is <code>null</code>, then it means that color is not a valid 
       * abbreviation.
       */
      private static final String[][] COLOR_MAP =
      {
         {"BLACK",          "0"},
         {"BLAC",           "0"},
         {"BLA",            "0"},
         {"BLK",            "0"},
         {"BLC",            null},
         {"BL",             null},
         {"B",              null},
         {"BLUE",           "1"},
         {"BLU",            "1"},
         {"BL",             null},
         {"GREEN",          "2"},
         {"GREE",           "2"},
         {"GRE",            "2"},
         {"GR",             null},
         {"G",              null},
         {"GRN",            "2"},
         {"CYAN",           "3"},
         {"CYA",            "3"},
         {"CY"  ,           "3"},
         {"C",              "3"},
         {"RED",            "4"},
         {"RE",             null},
         {"R",              null},
         {"RD",             null},
         {"MAGENTA",        "5"},
         {"MAGENT",         "5"},
         {"MAGEN",          "5"},
         {"MAGE",           "5"},
         {"MAG",            "5"},
         {"MA",             "5"},
         {"M",              null},
         {"BROWN",          "6"},
         {"BROW",           "6"},
         {"BRO",            "6"},
         {"BR",             null},
         {"B",              null},
         {"BRN",            "6"},
         {"GREY",           "2"}, // correct 2/green
         {"GRE",            "2"}, // correct 2/green
         {"GRY",            "7"},
         {"GRAY",           "7"},
         {"GRA",            "7"},
         {"GR",             null},
         {"G",              null},
         {"DARK-GREY",      "2"}, // correct 2/green
         {"DARK-GRE",       "2"}, // correct 2/green
         {"DARK-GRY",       "8"},
         {"DARK-GRAY",      "8"},
         {"DARK-GRA",       "8"},
         {"DARK-GR",        null},
         {"DARK-G",         null},
         {"D-GREY",         "2"}, // correct 2/green
         {"D-GRE",          "2"}, // correct 2/green
         {"D-GRY",          "8"},
         {"D-GRAY",         "8"},
         {"D-GRA",          "8"},
         {"D-GR",           null},
         {"D-G",            null},
         {"LIGHT-BLUE",     "9"},
         {"LIGHT-BLU",      "9"},
         {"LIGHT-BL",       null},
         {"LIGHT-B",        null},
         {"LT-BLUE",        "9"},
         {"LT-BLU",         "9"},
         {"LT-BL",          null},
         {"LT-B",           null},
         {"LIGHT-GRN",     "10"},
         {"LIGHT-GREEN",   "10"},
         {"LIGHT-GREE",    "10"},
         {"LIGHT-GRE",     "10"},
         {"LIGHT-GR",      null},
         {"LIGHT-G",       null},
         {"LIGHT-GRN",     "10"},
         {"LT-GREEN",      "10"},
         {"LT-GREE",       "10"},
         {"LT-GRE",        "10"},
         {"LT-GR",         null},
         {"LT-G",          null},
         {"LT-GRN",        "10"},
         {"LIGHT-CYAN",    "11"},
         {"LIGHT-CYA",     "11"},
         {"LIGHT-CY",      "11"},
         {"LIGHT-C",       "11"},
         {"LT-CYAN",       "11"},
         {"LT-CYA",        "11"},
         {"LT-CY",         "11"},
         {"LT-C",          "11"},
         {"LIGHT-RED",     "12"},
         {"LIGHT-RE",      null},
         {"LIGHT-R",       null},
         {"LT-RED",        "12"},
         {"LT-RE",         null},
         {"LT-R",          null},
         {"LIGHT-MAGENTA", "13"},
         {"LIGHT-MAGENT",  "13"},
         {"LIGHT-MAGEN",   "13"},
         {"LIGHT-MAGE",    "13"},
         {"LIGHT-MAG",     "13"},
         {"LIGHT-MA",      "13"},
         {"LIGHT-M",       null},
         {"LT-MAGENTA",    "13"},
         {"LT-MAGENT",     "13"},
         {"LT-MAGEN",      "13"},
         {"LT-MAGE",       "13"},
         {"LT-MAG",        "13"},
         {"LT-MA",         "13"},
         {"LT-M",          null},
         {"LIGHT-BROWN",   "14"},
         {"LIGHT-BROW",    "14"},
         {"LIGHT-BRO",     "14"},
         {"LIGHT-BR",      null},
         {"LIGHT-B",       null},
         {"LIGHT-BRN",     "14"},
         {"LT-BROWN",      "14"},
         {"LT-BROW",       "14"},
         {"LT-BRO",        "14"},
         {"LT-BR",         null},
         {"LT-B",          null},
         {"LT-BRN",        "14"},
         {"YELLOW",        "15"},
         {"YELLO",         "15"},
         {"YELL",          "15"},
         {"YEL",           "15"},
         {"YE",            "15"},
         {"Y",             "15"},
         {"WHITE",         "16"},
         {"WHIT",          "16"},
         {"WHI",           "16"},
         {"WH",            "16"},
         {"W",             "16"},
      };

      /** Color mapping table for Windows colors */
      private Map<Object, Color> windowsColorMap = new LinkedHashMap<>();

      /** Map of WINDOWS color names and indexes. */
      private Map<String, Integer> windowsColors;

      /** WINDOWS colors abbreviations */
      private String colorsPattern;

      /** BLINK- RVV- UNDERLINE- BRIGHT- fgnd-color */
      private String pattern;

      /** Color table index used on WINDOWS only*/
      private int index;

      /**
       * Default c'tor which initializes a GUI-specific workarea.
       */
      public GUIWorkArea()
      {
         super(new GUIColorConvertor());
         
         initWindowsColors();
      }

      /**
       * Initialize Windows color names and indexes.  Build patterns used as regular expressions. 
       * BLINK- BRIGHT- fgnd-color bgnd-color
       * BLINK- RVV- UNDERLINE- BRIGHT- fgnd-color
       */
      private void initWindowsColors()
      {
         windowsColors = new LinkedHashMap<>();
         for (String[] col : COLOR_MAP)
         {
            if (col[1] == null)
            {
               continue;
            }
            
            String colName = col[0];
            int colIdx = Integer.parseInt(col[1]);
            
            windowsColors.put(colName, colIdx);
         }
         
         // build regular expressions
         StringBuilder sb = new StringBuilder("(");
         for (String name : windowsColors.keySet())
         {
            sb.append(name).append("|");
         }
         // add standard colors
         sb.append("NORMAL|INPUT|MESSAGES)?");
         
         colorsPattern = sb.toString();
         pattern = "^((.*BLINK-)?(.*RVV-)?(.*UNDERLINE-)?(.*BRIGHT-)?)" + colorsPattern;            
      }
   }
   
   /**
    * ChUI-specific work area with context-local data.
    */
   private static class ChUIWorkArea
   extends WorkArea
   {
      /** Color mapping table common to all terminals */
      private Map<String, Color> globalColorMap = new LinkedHashMap<String, Color>();
      
      /** Map of terminal color mappings. */
      private Map<String, Map<Object, Color>> terminalColorMap;
      
      /** Color mapping table for Windows colors */
      private Map<Object, Color> windowsColorMap = new LinkedHashMap<>();

      /** Map of WINDOWS color names and indexes. */
      private Map<String, Integer> windowsColors;

      /** WINDOWS colors abbreviations */
      private String colorsPattern;

      /** BLINK- BRIGHT- fgnd-color bgnd-color */
      private String pattern0;

      /** BLINK- RVV- UNDERLINE- BRIGHT- fgnd-color */
      private String pattern1;

      /** Color table index used on WINDOWS only*/
      private int index;

      /**
       * Default c'tor which initializes a ChUI-specific workarea.
       */
      public ChUIWorkArea()
      {
         super(new ChUIColorConvertor());

         // the default mappings, the keys must be upper case!
         globalColorMap.put(COLOR_NORMAL  , Color.DEFAULT);
         globalColorMap.put(COLOR_INPUT   , Color.DEFAULT.withUnderline());
         globalColorMap.put(COLOR_MESSAGES, Color.DEFAULT.withReverse());
         globalColorMap.put(COLOR_BOLD    , Color.DEFAULT.withBold());
         globalColorMap.put(COLOR_BLINK   , Color.DEFAULT.withBlink());

         // initialize Linux/Unix OS colors.
         initLinuxColors();
         
         // initialize Windows OS colors.
         index = globalColorMap.size();
         initWindowsColors();
      }

      /**
       * Init colors for Unix/Linux.  The values are read from protermcap file.
       */
      private void initLinuxColors()
      {
         // load the color map
         ColorMapLoader loader = new ColorMapLoader();
            
         Map<String, Map<Object, Color>> loadedMap = loader.loadAllMappings();
         if (loadedMap != null)
         {
            // fill the terminalColorMap field, don't use the loaded map directly, we need to
            // normalize the keys
            Map<String, Map<Object, Color>> localTerminalMap = new LinkedHashMap<String, Map<Object, Color>>();
            for (Map.Entry<String, Map<Object, Color>> termEntry : loadedMap.entrySet())
            {
               Map<Object, Color> localColorMap = new LinkedHashMap<Object, Color>();
               for (Map.Entry<Object, Color> colorEntry : termEntry.getValue().entrySet())
               {
                  Object key = colorEntry.getKey();
                  if (key instanceof String)
                  {
                     // normalize the string key
                     localColorMap.put(((String) key).trim().toUpperCase(), colorEntry.getValue());
                  }
                  else
                  {
                     localColorMap.put(key, colorEntry.getValue());
                  }
               }
               localTerminalMap.put(termEntry.getKey().trim().toUpperCase(), termEntry.getValue());
            }
            terminalColorMap = localTerminalMap;
         } 
         else
         {
            LOG.info("Terminal color map was not loaded.");
         }      
      }
      
      /**
       * Initialize Windows color names and indexes.  Build patterns used as regular expressions. 
       * BLINK- BRIGHT- fgnd-color bgnd-color
       * BLINK- RVV- UNDERLINE- BRIGHT- fgnd-color
       */
      private void initWindowsColors()
      {
         windowsColors = new LinkedHashMap<>();
         windowsColors.put("BLACK", 0);
         windowsColors.put("BLA", 0);
         windowsColors.put("BLK", 0);
         windowsColors.put("BLUE", 1);
         windowsColors.put("BLU", 1);
         windowsColors.put("GREEN", 2);
         windowsColors.put("GRE", 2);
         windowsColors.put("GRN", 2);
         windowsColors.put("CYAN", 3);
         windowsColors.put("C", 3);
         windowsColors.put("RED", 4);
         windowsColors.put("MAGENTA", 5);
         windowsColors.put("MA", 5);
         windowsColors.put("BROWN", 6);
         windowsColors.put("BRO", 6);
         windowsColors.put("BRN", 6);
         windowsColors.put("GRAY", 7);
         windowsColors.put("GRA", 7);
         windowsColors.put("GRY", 7);
         windowsColors.put("DARK-GRAY", 8);
         windowsColors.put("D-GRA", 8);
         windowsColors.put("LIGHT-BLUE", 9);
         windowsColors.put("LT-BLU", 9);
         windowsColors.put("LIGHT-GREEN", 10);
         windowsColors.put("LT-GRE", 10);
         windowsColors.put("LIGHT-CYAN", 11);
         windowsColors.put("LT-C", 11);
         windowsColors.put("LIGHT-RED", 12);
         windowsColors.put("LT-RED", 12);
         windowsColors.put("LIGHT-MAGENTA", 13);
         windowsColors.put("LT-MA", 13);
         windowsColors.put("LIGHT-BROWN", 14);
         windowsColors.put("LT-BRO", 14);
         windowsColors.put("YELLOW", 14);
         windowsColors.put("Y", 14);
         windowsColors.put("WHITE", 15);
         windowsColors.put("W", 15);
         
         // build regular expressions
         StringBuilder sb = new StringBuilder("(");
         for (String name : windowsColors.keySet())
         {
            sb.append(name).append("|");
         }
         // add standard colors
         sb.append("NORMAL|INPUT|MESSAGES)?");
         
         colorsPattern = sb.toString();
         pattern0 = "^((.*BLINK-)?(.*BRIGHT-)?)" + colorsPattern;
         pattern1 = "^((.*BLINK-)?(.*RVV-)?(.*UNDERLINE-)?(.*BRIGHT-)?)" + colorsPattern;            
      }
   }
   
   /**
    * Context-local data for color specification.
    */
   private static class WorkArea
   {
      /** The color convertor. */
      protected final ColorConvertor colorConvertor;
      
      /**
       * Create a new instance which uses the specified color convertor.
       * 
       * @param    colorConvertor
       *           The client-specific color convertor.
       */
      private WorkArea(ColorConvertor colorConvertor)
      {
         this.colorConvertor = colorConvertor;
      }
   }
}