/*
** Module   : ColorTable.java
** Abstract : Implementation of the COLOR-TABLE resource plus other misc color-related APIs.
**
** Copyright (c) 2014-2018, Golden Code Development Corporation.
**
** -#- -I- --Date-- --------------------------------Description-----------------------------------
** 001 MAG 20140812 Created initial version.
** 002 MAG 20140911 Integrate environment API.
** 003 CA  20140926 Added shared widget configuration support, to allow GUI/ChUI concrete
**                  implementation for the same widget ID.  Refactored the widget configuration
**                  classes: all fields were made public; these classes need to be as dumb as
**                  possible, all logic related to setting/getting a certain field should be at
**                  the widget or at the caller. Refs #2254
** 004 EVL 20141206 The default color table that always exists added.  If there are color table
**                  entries in directory file the color values with same index will be overridden
**                  by the colors from directory.  If no colors in directory the default values
**                  will be used.
** 005 CA  20150502 The resource ID needs to be kept at the resource, so it can be accessed after
**                  the resource is deleted.
** 006 GES 20150518 Javadoc cleanup.
** 007 EVL 20160224 Javadoc fixes to make compatible with Oracle Java 8 for Solaris 10.
** 008 OM  20170320 Added support for Themes.
** 009 OM  20170720 Added ColorChooserDialog builder.
** 010 CA  20170910 Added createColorPair.
** 011 OM  20171027 If a new color is selected by user, the old entry is replaced in ColorTable.
** 012 EVL 20180207 Fix for invalid char in comment issue.
** 013 OM  20180321 Added ColorChooserDialog.getColor() static worker for obtaining a RGB color
**                  on server-side. Avoid creating intermediary objects in rgbValue(). Updated
**                  javadoc.
** 014 CA  20180712 Fixed the system color dialog owner.
*/

/*
** This program is free software: you can redistribute it and/or modify
** it under the terms of the GNU Affero General Public License as
** published by the Free Software Foundation, either version 3 of the
** License, or (at your option) any later version.
**
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
** GNU Affero General Public License for more details.
**
** You may find a copy of the GNU Affero GPL version 3 at the following
** location: https://www.gnu.org/licenses/agpl-3.0.en.html
** 
** Additional terms under GNU Affero GPL version 3 section 7:
** 
**   Under Section 7 of the GNU Affero GPL version 3, the following additional
**   terms apply to the works covered under the License.  These additional terms
**   are non-permissive additional terms allowed under Section 7 of the GNU
**   Affero GPL version 3 and may not be removed by you.
** 
**   0. Attribution Requirement.
** 
**     You must preserve all legal notices or author attributions in the covered
**     work or Appropriate Legal Notices displayed by works containing the covered
**     work.  You may not remove from the covered work any author or developer
**     credit already included within the covered work.
** 
**   1. No License To Use Trademarks.
** 
**     This license does not grant any license or rights to use the trademarks
**     Golden Code, FWD, any Golden Code or FWD logo, or any other trademarks
**     of Golden Code Development Corporation. You are not authorized to use the
**     name Golden Code, FWD, or the names of any author or contributor, for
**     publicity purposes without written authorization.
** 
**   2. No Misrepresentation of Affiliation.
** 
**     You may not represent yourself as Golden Code Development Corporation or FWD.
** 
**     You may not represent yourself for publicity purposes as associated with
**     Golden Code Development Corporation, FWD, or any author or contributor to
**     the covered work, without written authorization.
** 
**   3. No Misrepresentation of Source or Origin.
** 
**     You may not represent the covered work as solely your work.  All modified
**     versions of the covered work must be marked in a reasonable way to make it
**     clear that the modified work is not originating from Golden Code Development
**     Corporation or FWD.  All modified versions must contain the notices of
**     attribution required in this license.
*/

package com.goldencode.p2j.ui;

import java.io.*;
import java.util.*;
import java.util.logging.*;

import com.goldencode.p2j.security.*;
import com.goldencode.p2j.ui.client.*;
import com.goldencode.p2j.util.*;
import com.goldencode.p2j.util.ErrorManager;
import com.goldencode.proxy.*;

/**
 * Implementation of the COLOR-TABLE system handle (the APIs from the {@link ColorTableResource}
 * interface).  In addition, this implements:
 * <ul>
 *    <li>the {@code RGB-VALUE} statement, via the {@link #rgbValue} API</li>
 *    <li>the {@code SYSTEM-DIALOG COLOR} statement, via the
 *          {@link #createColorChooserDialog} API</li>
 *    <li>the {@code PUT-KEY-VALUE COLOR ALL} statement, via the {@link #saveColorTable} API </li>
 *    <li>the {@code PUT-KEY-VALUE COLOR color-num} statement, via the {@link #saveColor} API
 *    </li>
 * </ul>
 */
public class ColorTable
extends BaseSysTable
{
   /** logger */
   private static final Logger LOG = LogHelper.getLogger(ColorTable.class.getName());
   
   /** Color INI file section name */
   public static final String COLOR_SECTION = "colors";
   
   /** Color INI file key name */
   public static final String COLOR_KEY = "color";

   /** The name of the node containing the color-table. */
   public static final String COLOR_TABLE = "color-table";

   /** The name of the node containing the color-table. */
   public static final String STANDARD_COLORS = "standard-colors";
   
   /** The name of the node containing the color-table. */
   public static final String SYSTEM_COLORS = "system-colors";
   
   /** Default colors. */
   private static final List<ColorTableEntry> defaultColorTable;

   // initialize default color table
   static
   {
      defaultColorTable = new ArrayList<>();
      // this is the legacy 16 colors palette
      defaultColorTable.add(new ColorTableEntry(0,  0,   0,   0));   // black
      defaultColorTable.add(new ColorTableEntry(1,  0,   0,   128)); // dark blue
      defaultColorTable.add(new ColorTableEntry(2,  0,   128, 0));   // dark green
      defaultColorTable.add(new ColorTableEntry(3,  0,   128, 128)); // dark cyan
      defaultColorTable.add(new ColorTableEntry(4,  128, 0,   0));   // red
      defaultColorTable.add(new ColorTableEntry(5,  128, 0,   128)); // purple
      defaultColorTable.add(new ColorTableEntry(6,  128, 128, 0));   // brown
      defaultColorTable.add(new ColorTableEntry(7,  128, 128, 128)); // dark gray
      defaultColorTable.add(new ColorTableEntry(8,  192, 192, 192)); // light gray
      defaultColorTable.add(new ColorTableEntry(9,  0,   0,   255)); // blue
      defaultColorTable.add(new ColorTableEntry(10, 0,   255, 0));   // green
      defaultColorTable.add(new ColorTableEntry(11, 0,   255, 255)); // cyan
      defaultColorTable.add(new ColorTableEntry(12, 255, 0,   0));   // light red
      defaultColorTable.add(new ColorTableEntry(13, 255, 0,   255)); // magenta
      defaultColorTable.add(new ColorTableEntry(14, 255, 255, 0));   // yellow
      defaultColorTable.add(new ColorTableEntry(15, 255, 255, 255)); // white
   }

   /** Context local instance of this class. */
   private static final ContextLocal<WorkArea> local = new ContextLocal<WorkArea>()
   {
      protected WorkArea initialValue()
      { 
         return new WorkArea();
      };
   };
   
   /**
    * Get system colors table.
    * 
    * @return  An array containing system colors.
    */
   public static ColorTableEntry[] getSystemColors()
   {
      return locate().systemColors;
   }
   
   /**
    * Get color table of the specified environment.
    * 
    * @param   envName
    *          Environment name. Could be <code>null</code> if is default environment. 
    *          
    * @return  see above.
    */
   public static EnvironmentColorTable getColorTable(String envName)
   {
      return locate().envColorTables.get(envName);
   }

   /**
    * Getter for the INSTANTIATING-PROCEDURE attribute.
    * <p>
    * The returned handle will refer a valid external procedure (persistent or not) or will be
    * unknown, in case its instantiating procedure was persistent and has been deleted.
    * 
    * @return   See above.
    */
   public static handle instantiatingProcedure()
   {
      // TODO: implement this
      throw new RuntimeException("INSTANTIATING-PROCEDURE attribute is not implemented.");
   }

   /**
    * API needed to implement read-only attribute assignment (a 4GL 
    * "feature").
    * 
    * @param    attribute
    *           The attribute's name.
    *           
    * @see      handle#readOnlyError(handle, String)
    */
   public static void readOnlyError(String attribute)
   {
      handle.readOnlyError(asHandle(), attribute);
   }
   
   /**
    * Get the type of its associated handle.
    * 
    * @return   Always return the PSEUDO-WIDGET value.
    */
   public static character getResourceType()
   {
      return new character(LegacyResource.PSEUDO_WIDGET);
   }
   
   /**
    * Get an instance for the COLOR-TABLE system handle.  Is obtained using
    * a call to the {@link StaticProxy#obtain(Class, Class[])}, using the
    * {@link ColorTableResource} interface and its methods implemented by these
    * classes: {@link ColorTable}.
    * 
    * @return   See above.
    */
   public static handle asHandle()
   {
      WorkArea wa = local.get();
      
      Class<?>[] classes = { ColorTable.class };
      ColorTableResource proxy = StaticProxy.obtain(ColorTableResource.class, classes);
      
      if (wa.id == null)
      {
         wa.id = handle.resourceId(proxy);
      }
      
      return new handle(proxy);
   }
   
   /**
    * Get this resource's ID.
    * 
    * @return   The resource's ID.
    */
   public static Long id()
   {
      WorkArea wa = local.get();
      
      return wa.id;
   }
   
   /**
    * Set this resource's ID.
    * <p>
    * This is a no-op for system handles.
    *  
    * @param    id
    *           The resource's ID.
    */
   public static void id(long id)
   {
      // no-op
   }
   
   /**
    * Implementation for the {@link WrappedResource#valid()} API.
    * 
    * @return   Always true.
    */
   public static boolean valid()
   {
      return true;
   }
   
   /**
    * Implementation for the {@link WrappedResource#unknown()} API.
    * 
    * @return   Always false.
    */
   public static boolean unknown()
   {
      return false;
   }
   
   /**
    * Get the number of entries in the color-table. For ChUI interfaces always return 0.
    * 
    * @return   See above.
    */
   public static integer getNumEntries()
   {
      WorkArea wa = locate();
      EnvironmentColorTable ect = getCurrentColorTable(wa);
      return new integer(wa.isChui ? 0 : ect == null ? 0 : ect.colors.size());
   }
   
   /**
    * Set the number of entries in the color-table.
    * 
    * @param    numEntries
    *           The new size of the color-table. 
    */
   public static void setNumEntries(long numEntries)
   {
      setNumEntries(new int64(numEntries));
   }
   
   /**
    * Set the number of entries in the color-table.
    * 
    * @param    numEntries
    *           The new size of the color-table. 
    */
   public static void setNumEntries(int64 numEntries)
   {
      WorkArea wa = locate();
      
      if (!numEntries.isUnknown() && !wa.isChui)
      {
         // TODO: what happens if the color-table is shrinked and there are widgets referring those
         // colors?
         int size =  numEntries.intValue();
     
         EnvironmentColorTable ect = getCurrentColorTable(wa);
         if (ect == null || ect.colors.size() == size)
         {
            return;
         }
   
         while (ect.colors.size() != size)
         {
            if (size < ect.colors.size())
            {
               // shrink it
               ect.colors.remove(ect.colors.size() - 1);
            }
            else
            {
               // enlarge it with the default color
               ect.colors.add(new ColorTableEntry());
            }
         }
         
         synch(ect);
      }
   }
   
   /**
    * Returns the red component of an entry in the color table.
    * 
    * @param    index
    *           An integer expression that specifies an entry in the color table.
    *           
    * @return   See above.
    */
   public static integer getRedValue(int64 index)
   {
      if (locate().isChui)
      {
         return new integer(0);
      }      
      ColorTableEntry color = null;
      if (!index.isUnknown())
      {
         EnvironmentColorTable ect = getCurrentColorTable(locate());
         color = getColorTableEntry(ect, index.intValue());
      }
      return (color == null) ? new integer() : new integer(color.getRed());
   }
   
   /**
    * Returns the red component of an entry in the color table.
    * 
    * @param    index
    *           An integer expression that specifies an entry in the color table.
    *           
    * @return   See above.
    */
   public static integer getRedValue(long index)
   {
      return getRedValue(new int64(index));
   }
   
   /**
    * Specifies the red component of an entry in the color table.
    * 
    * @param    index
    *           An integer expression that specifies an entry in the color table.
    * @param    redValue
    *           An integer expression that specifies the red RGB component of an entry in the
    *           color table. The value must be in the range 0 to 255.
    *           
    * @return   If the operation is successful, the method returns TRUE.
    */
   public static logical setRedValue(int64 index, int64 redValue)
   {
      boolean result = false;
      WorkArea wa = locate();
      if (!index.isUnknown() && !redValue.isUnknown() && !wa.isChui)
      {
         EnvironmentColorTable ect = getCurrentColorTable(wa);
         ColorTableEntry color = getColorTableEntry(ect, index.intValue());
         if (color != null && color.isDynamic())
         {
            color.setRed(redValue.intValue() & 0xFF);
            synch(ect);
            result = true;
         }
         else
         {
            warning("SET-RED-VALUE");
         }
      }
      return new logical(result);
   }
   
   /**
    * Specifies the red component of an entry in the color table.
    * 
    * @param    index
    *           An integer expression that specifies an entry in the color table.
    * @param    redValue
    *           An integer expression that specifies the red RGB component of an entry in the
    *           color table. The value must be in the range 0 to 255.
    *           
    * @return   If the operation is successful, the method returns TRUE.
    */
   public static logical setRedValue(int64 index, long redValue)
   {
      return setRedValue(index, new int64(redValue));
   }

   /**
    * Specifies the red component of an entry in the color table.
    * 
    * @param    index
    *           An integer expression that specifies an entry in the color table.
    * @param    redValue
    *           An integer expression that specifies the red RGB component of an entry in the
    *           color table. The value must be in the range 0 to 255.
    *           
    * @return   If the operation is successful, the method returns TRUE.
    */
   public static logical setRedValue(long index, int64 redValue)
   {
      return setRedValue(new int64(index), redValue);
   }

   /**
    * Specifies the red component of an entry in the color table.
    * 
    * @param    index
    *           An integer expression that specifies an entry in the color table.
    * @param    redValue
    *           An integer expression that specifies the red RGB component of an entry in the
    *           color table. The value must be in the range 0 to 255.
    *           
    * @return   If the operation is successful, the method returns TRUE.
    */
   public static logical setRedValue(long index, long redValue)
   {
      return setRedValue(new int64(index), new int64(redValue));
   }

   /**
    * Returns the green component of an entry in the color table.
    * 
    * @param    index
    *           An integer expression that specifies an entry in the color table.
    *           
    * @return   See above.
    */
   public static integer getGreenValue(int64 index)
   {
      if (locate().isChui)
      {
         return new integer(0);
      }      
      ColorTableEntry color = null;
      if (!index.isUnknown())
      {
         EnvironmentColorTable ect = getCurrentColorTable(locate());
         color = getColorTableEntry(ect, index.intValue());
      }
      return (color == null) ? new integer() : new integer(color.getGreen());
   }
   
   /**
    * Returns the green component of an entry in the color table.
    * 
    * @param    index
    *           An integer expression that specifies an entry in the color table.
    *           
    * @return   See above.
    */
   public static integer getGreenValue(long index)
   {
      return getGreenValue(new int64(index));
   }

   /**
    * Specifies the green component of an entry in the color table.
    * 
    * @param    index
    *           An integer expression that specifies an entry in the color table.
    * @param    greenValue
    *           An integer expression that specifies the green RGB component of an entry in the
    *           color table. The value must be in the range 0 to 255.
    *           
    * @return   If the operation is successful, the method returns TRUE.
    */
   public static logical setGreenValue(int64 index, int64 greenValue)
   {
      boolean result = false;
      WorkArea wa = locate();
      if (!index.isUnknown() && !greenValue.isUnknown() && !wa.isChui)
      {
         EnvironmentColorTable ect = getCurrentColorTable(wa);
         ColorTableEntry color = getColorTableEntry(ect, index.intValue());
         if (color != null && color.isDynamic())
         {
            color.setGreen(greenValue.intValue() & 0xFF);
            synch(ect);
            result = true;
         }
         else
         {
            warning("SET-GREEN-VALUE");
         }
      }
      return new logical(result);
   }

   /**
    * Specifies the green component of an entry in the color table.
    * 
    * @param    index
    *           An integer expression that specifies an entry in the color table.
    * @param    greenValue
    *           An integer expression that specifies the green RGB component of an entry in the
    *           color table. The value must be in the range 0 to 255.
    *           
    * @return   If the operation is successful, the method returns TRUE.
    */
   public static logical setGreenValue(int64 index, long greenValue)
   {
      return setGreenValue(index, new int64(greenValue));
   }

   /**
    * Specifies the green component of an entry in the color table.
    * 
    * @param    index
    *           An integer expression that specifies an entry in the color table.
    * @param    greenValue
    *           An integer expression that specifies the green RGB component of an entry in the
    *           color table. The value must be in the range 0 to 255.
    *           
    * @return   If the operation is successful, the method returns TRUE.
    */
   public static logical setGreenValue(long index, int64 greenValue)
   {
      return setGreenValue(new int64(index), greenValue);
   }

   /**
    * Specifies the green component of an entry in the color table.
    * 
    * @param    index
    *           An integer expression that specifies an entry in the color table.
    * @param    greenValue
    *           An integer expression that specifies the green RGB component of an entry in the
    *           color table. The value must be in the range 0 to 255.
    *           
    * @return   If the operation is successful, the method returns TRUE.
    */
   public static logical setGreenValue(long index, long greenValue)
   {
      return setGreenValue(new int64(index), new int64(greenValue));
   }

   /**
    * Returns the blue component of an entry in the color table.
    * 
    * @param    index
    *           An integer expression that specifies an entry in the color table.
    *           
    * @return   See above.
    */
   public static integer getBlueValue(int64 index)
   {
      if (locate().isChui)
      {
         return new integer(0);         
      }      
      ColorTableEntry color = null;
      if (!index.isUnknown())
      {
         EnvironmentColorTable ect = getCurrentColorTable(locate());
         color = getColorTableEntry(ect, index.intValue());
      }
      return (color == null) ? new integer() : new integer(color.getBlue());
   }
   
   /**
    * Returns the blue component of an entry in the color table.
    * 
    * @param    index
    *           An integer expression that specifies an entry in the color table.
    *           
    * @return   See above.
    */
   public static integer getBlueValue(long index)
   {
      return getBlueValue(new int64(index));
   }

   /**
    * Specifies the blue component of an entry in the color table.
    * 
    * @param    index
    *           An integer expression that specifies an entry in the color table.
    * @param    blueValue
    *           An integer expression that specifies the blue RGB component of an entry in the
    *           color table. The value must be in the range 0 to 255.
    *           
    * @return   If the operation is successful, the method returns TRUE.
    */
   public static logical setBlueValue(int64 index, int64 blueValue)
   {
      boolean result = false;
      WorkArea wa = locate();
      if (!index.isUnknown() && !blueValue.isUnknown() && !wa.isChui)
      {
         EnvironmentColorTable ect = getCurrentColorTable(wa);
         ColorTableEntry color = getColorTableEntry(ect, index.intValue());
         if (color != null && color.isDynamic())
         {
            color.setBlue(blueValue.intValue() & 0xFF);
            synch(ect);
            result = true;
         }
         else
         {
            warning("SET-BLUE-VALUE");
         }
      }
      return new logical(result);
   }

   /**
    * Specifies the blue component of an entry in the color table.
    * 
    * @param    index
    *           An integer expression that specifies an entry in the color table.
    * @param    blueValue
    *           An integer expression that specifies the blue RGB component of an entry in the
    *           color table. The value must be in the range 0 to 255.
    *           
    * @return   If the operation is successful, the method returns TRUE.
    */
   public static logical setBlueValue(int64 index, long blueValue)
   {
      return setBlueValue(index, new int64(blueValue));
   }

   /**
    * Specifies the blue component of an entry in the color table.
    * 
    * @param    index
    *           An integer expression that specifies an entry in the color table.
    * @param    blueValue
    *           An integer expression that specifies the blue RGB component of an entry in the
    *           color table. The value must be in the range 0 to 255.
    *           
    * @return   If the operation is successful, the method returns TRUE.
    */
   public static logical setBlueValue(long index, int64 blueValue)
   {
      return setBlueValue(new int64(index), blueValue);
   }

   /**
    * Specifies the blue component of an entry in the color table.
    * 
    * @param    index
    *           An integer expression that specifies an entry in the color table.
    * @param    blueValue
    *           An integer expression that specifies the blue RGB component of an entry in the
    *           color table. The value must be in the range 0 to 255.
    *           
    * @return   If the operation is successful, the method returns TRUE.
    */
   public static logical setBlueValue(long index, long blueValue)
   {
      return setBlueValue(new int64(index), new int64(blueValue));
   }

   /**
    * Returns the rgb component of an entry in the color table.
    * 
    * @param    index
    *           An integer expression that specifies an entry in the color table.
    *           
    * @return   See above.
    */
   public static integer getRgbValue(int64 index)
   {
      if (locate().isChui)
      {
         return new integer(0);
      }      
      ColorTableEntry color = null;
      if (!index.isUnknown())
      {
         EnvironmentColorTable ect = getCurrentColorTable(locate());
         color = getColorTableEntry(ect, index.intValue());
      }
      return (color == null) ? new integer() : new integer(color.getRgb());
   }
   
   /**
    * Returns the rgb component of an entry in the color table.
    * 
    * @param    index
    *           An integer expression that specifies an entry in the color table.
    *           
    * @return   See above.
    */
   public static integer getRgbValue(long index)
   {
      return getRgbValue(new int64(index));
   }

   /**
    * Specifies a combination of the red, green, and blue values of an entry in the color table.
    * 
    * @param    index
    *           An integer expression that specifies an entry in the color table.
    * @param    rgbValue
    *           An integer expression that specifies the RGB component of an entry in the color
    *           table. You can obtain this value from the color property of an ActiveX control,
    *           by using the RGB-VALUE function, or by using the GET-RGB-VALUE() method.
    *           
    * @return   If the operation is successful, the method returns TRUE.
    */
   public static logical setRgbValue(int64 index, int64 rgbValue)
   {
      boolean result = false;
      WorkArea wa = locate();
      if (!index.isUnknown() && !rgbValue.isUnknown() && !wa.isChui)
      {
         EnvironmentColorTable ect = getCurrentColorTable(wa);
         ColorTableEntry color = getColorTableEntry(ect, index.intValue());
         if (color != null && color.isDynamic())
         {
            color.setRgb(rgbValue.intValue());
            synch(ect);
            result = true;
         }
         else
         {
            warning("SET-RGB-VALUE");
         }
      }
      return new logical(result);
   }

   /**
    * Specifies a combination of the red, green, and blue values of an entry in the color table.
    * 
    * @param    index
    *           An integer expression that specifies an entry in the color table.
    * @param    rgbValue
    *           An integer expression that specifies the RGB component of an entry in the color
    *           table. You can obtain this value from the color property of an ActiveX control,
    *           by using the RGB-VALUE function, or by using the GET-RGB-VALUE() method.
    *           
    * @return   If the operation is successful, the method returns TRUE.
    */
   public static logical setRgbValue(int64 index, long rgbValue)
   {
      return setRgbValue(index, new int64(rgbValue));
   }

   /**
    * Specifies a combination of the red, green, and blue values of an entry in the color table.
    * 
    * @param    index
    *           An integer expression that specifies an entry in the color table.
    * @param    rgbValue
    *           An integer expression that specifies the RGB component of an entry in the color
    *           table. You can obtain this value from the color property of an ActiveX control,
    *           by using the RGB-VALUE function, or by using the GET-RGB-VALUE() method.
    *           
    * @return   If the operation is successful, the method returns TRUE.
    */
   public static logical setRgbValue(long index, int64 rgbValue)
   {
      return setRgbValue(new int64(index), rgbValue);
   }

   /**
    * Specifies a combination of the red, green, and blue values of an entry in the color table.
    * 
    * @param    index
    *           An integer expression that specifies an entry in the color table.
    * @param    rgbValue
    *           An integer expression that specifies the RGB component of an entry in the color
    *           table. You can obtain this value from the color property of an ActiveX control,
    *           by using the RGB-VALUE function, or by using the GET-RGB-VALUE() method.
    *           
    * @return   If the operation is successful, the method returns TRUE.
    */
   public static logical setRgbValue(long index, long rgbValue)
   {
      return setRgbValue(new int64(index), new int64(rgbValue));
   }

   /**
    * Returns TRUE if the entry in the color table is a dynamic color.
    *  
    * @param    index
    *           An integer expression that specifies an entry in the color table.
    *           
    * @return   See above.
    */
   public static logical getDynamic(int64 index)
   {
      if (locate().isChui)
      {
         return new logical(false);
      }      
      ColorTableEntry color = null;
      if (!index.isUnknown())
      {
         EnvironmentColorTable ect = getCurrentColorTable(locate());
         color = getColorTableEntry(ect, index.intValue());
      }
      return (color == null) ? new logical() : new logical(color.isDynamic());
   }
   
   /**
    * Returns TRUE if the entry in the color table is a dynamic color.
    * 
    * @param    index
    *           An integer expression that specifies an entry in the color table.
    *           
    * @return   See above.
    */
   public static logical getDynamic(long index)
   {
      return getDynamic(new int64(index));
   }
   
   /**
    * Sets a color entry to a dynamic or static color.
    * 
    * @param    index
    *           An integer expression that specifies an entry in the color table.
    * @param    dynamic
    *           Logical expression that specifies the dynamic status of an entry in the color
    *           table. If dynamic is TRUE sets the entry to a dynamic color and sets the entry
    *           to a static color if dynamic is FALSE.
    *
    * @return   If the operation is successful, the method returns TRUE.
    */
   public static logical setDynamic(int64 index, logical dynamic)
   {
      boolean result = false;
      if (!index.isUnknown() && !dynamic.isUnknown() && !locate().isChui)
      {
         EnvironmentColorTable ect = getCurrentColorTable(locate());
         ColorTableEntry color = getColorTableEntry(ect, index.intValue());
         if (color != null)
         {
            color.setDynamic(dynamic.booleanValue());
            result = true;
         }
         else
         {
            warning("SET-DYNAMIC");
         }
      }
      return new logical(result);
   }

   /**
    * Sets a color entry to a dynamic or static color.
    * 
    * @param    index
    *           An integer expression that specifies an entry in the color table.
    * @param    dynamic
    *           Logical expression that specifies the dynamic status of an entry in the color
    *           table. If dynamic is TRUE sets the entry to a dynamic color and sets the entry
    *           to a static color if dynamic is FALSE.    
    *
    * @return   If the operation is successful, the method returns TRUE.
    */
   public static logical setDynamic(int64 index, boolean dynamic)
   {
      return setDynamic(index, new logical(dynamic));
   }

   /**
    * Sets a color entry to a dynamic or static color.
    * 
    * @param    index
    *           An integer expression that specifies an entry in the color table.
    * @param    dynamic
    *           Logical expression that specifies the dynamic status of an entry in the color
    *           table. If dynamic is TRUE sets the entry to a dynamic color and sets the entry
    *           to a static color if dynamic is FALSE.
    *
    * @return   If the operation is successful, the method returns TRUE.
    */
   public static logical setDynamic(long index, logical dynamic)
   {
      return setDynamic(new int64(index), dynamic);
   }

   /**
    * Sets a color entry to a dynamic or static color.
    * 
    * @param    index
    *           An integer expression that specifies an entry in the color table.
    * @param    dynamic
    *           Logical expression that specifies the dynamic status of an entry in the color
    *           table. If dynamic is TRUE sets the entry to a dynamic color and sets the entry
    *           to a static color if dynamic is FALSE.    
    *                      
    * @return   If the operation is successful, the method returns TRUE.
    */
   public static logical setDynamic(long index, boolean dynamic)
   {
      return setDynamic(new int64(index), new logical(dynamic));
   }

   /**
    * Get RGB value for specified index from color table associate to a Window.
    * 
    * @param   envName
    *          Environment name.
    * @param   index
    *          Color table index.
    *          
    * @return  RGB value in Java format.
    */
   public static int getColor(String envName, int index)
   {
      int rgb = -1;
      
      EnvironmentColorTable ect = locate().envColorTables.get(envName);
      if (ect != null)
      {
         if (index >= 0 && index < ect.colors.size())
         {
            ColorTableEntry color = ect.colors.get(index);
            if (color != null)
            {
               rgb = color.getGuiRgb();
            }
         }
      }
      
      return rgb;
   }
   
   /**
    * Save the user's color-table to the environment. This implements the PUT-KEY-VALUE COLOR ALL
    * statement.
    */
   public static void saveColorTable()
   {
      EnvironmentColorTable ect = getCurrentColorTable(locate());
      if (ect == null)
      {
         return;
      }
      
      if (ect.envName == null)
      {
         // TODO: default environment should be saved on server directory
      }
      else
      {
         // save color table
         int n = 0;
         for (ColorTableEntry color : ect.colors)
         {
            EnvironmentOps.setKeyValue(COLOR_SECTION, COLOR_KEY + n, color.asString());
            n++;
         }
         // save standard colors
         for (Map.Entry<String, ColorPair> color : ect.standardColors.entrySet())
         {
            EnvironmentOps.setKeyValue(COLOR_SECTION, color.getKey(), color.getValue().asString());
         }
      }
   }
   
   /**
    * Save the the specified color, from the user's color-table, to the environment. This
    * implements the <code>PUT-KEY-VALUE COLOR color-num</code> statement.
    * 
    * @param    colorNum
    *           The color-table entry to be saved.
    */
   public static void saveColor(long colorNum)
   {
      saveColor(new int64(colorNum));
   }
   
   /**
    * Save the the specified color, from the user's color-table, to the environment. This
    * implements the <code>PUT-KEY-VALUE COLOR color-num</code> statement.
    * 
    * @param    colorNum
    *           The color-table entry to be saved.
    */
   public static void saveColor(int64 colorNum)
   {
      if (colorNum.isUnknown())
      {
         return;
      }
         
      EnvironmentColorTable ect = getCurrentColorTable(locate());
      if (ect == null)
      {
         return;
      }
      
      int n = colorNum.intValue();
      
      ColorTableEntry color = getColorTableEntry(ect, n);
      if (color == null)
      {
         return;
      }

      if (ect.envName == null)
      {
         // TODO: default environment should be saved on server directory
      }
      else
      {
         EnvironmentOps.setKeyValue(COLOR_SECTION, COLOR_KEY + n, color.asString());
      }
   }
   
   /**
    * Factory method for creating a color chooser dialog.
    * 
    * @param   colorRef
    *          The color-table entry to be edited in the dialog
    * 
    * @return  A new color dialog. Calling {@link ColorChooserDialog#execute()} on this will
    *          result the client to display a theme specific color chooser dialog.
    */
   public static ColorChooserDialog createColorChooserDialog(int64 colorRef)
   {
      return new ColorChooserDialog(colorRef);
   }
   
   /**
    * Factory method for creating a color chooser dialog.
    * 
    * @param   colorRef
    *          The color-table entry to be edited in the dialog
    * 
    * @return  A new color dialog. Calling {@link ColorChooserDialog#execute()} on this will
    *          result the client to display a theme specific color chooser dialog.
    */
   public static ColorChooserDialog createColorChooserDialog(long colorRef)
   {
      return new ColorChooserDialog(colorRef);
   }
   
   /**
    * Returns an integer that represents a combination of a red, green, and blue color value.
    * This function allows you to define an arbitrary color, expanding beyond those colors
    * defined in the color table.
    * <p>
    * Return value is calculated as follow: red + green &lt;&lt; 8 + blue &lt;&lt; 16
    * <p>
    * This is the implementation of <code>RGB-VALUE</code> function.
    *  
    * @param   r
    *          Red value.
    * @param   g
    *          Green value.
    * @param   b
    *          Blue value.
    * 
    * @return  An integer that represents a combination of a red, green, and blue color value.
    */
   public static integer rgbValue(int r, int g, int b)
   {
      return new integer(r | (g << 8) | (b << 16));
   }
   
   /**
    * Returns an integer that represents a combination of a red, green, and blue color value.
    * This function allows you to define an arbitrary color, expanding beyond those colors 
    * defined in the color table.
    * <p>
    * Return value is calculated as follow: red + green &lt;&lt; 8 + blue &lt;&lt; 16
    * <p>
    * This is the implementation of <code>RGB-VALUE</code> function.
    * 
    * @param   r
    *          Red value.
    * @param   g
    *          Green value.
    * @param   b
    *          Blue value.
    * 
    * @return  An integer that represents a combination of a red, green, and blue color value.
    */
   public static integer rgbValue(int64 r, int64 g, int64 b)
   {
      if (r.isUnknown() || g.isUnknown() || b.isUnknown())
      {
         return new integer();
      }
      return rgbValue(r.intValue(), g.intValue(), b.intValue());
   }
   
   /**
    * Environments are synchronized between server and client using server state synchronization
    * mechanism. Each environment has a dirty flag which is used to indicate when environments
    * must be synchronized. Before the synchronization is effective made dirty flags are cleared.
    * {@link LogicalTerminal#getChanges()}  
    *  
    * @param   envs
    *          Environments list.
    */
   public static void clearDirtyFlag(List<EnvironmentColorTable> envs)
   {
      if (envs != null)
      {
         // clear dirty flags
         for (EnvironmentColorTable env : envs)
         {
            env.dirty = false;
         }
      }
   }
   
   /**
    * Call when a new window is created.
    * 
    * @param   window
    *          Window configuration.
    */
   public static void registerWindow(WindowConfig window)
   {
      WorkArea wa = local.get(false);
      if (wa != null)
      {
         // load color table if not already loaded.
         getCurrentColorTable(wa);
         window.envName = wa.envName;
      }
   }
   
   /**
    * Set current environment name. A notification from environment API whenever an
    * <code>USE environment</code> statement is executed.
    * 
    * @param   envName
    *          Environment name.
    */
   public static void use(String envName)
   {
      WorkArea wa = local.get(false);
      if (wa != null)
      {
         wa.envName = envName;
      }
   }
      
   /**
    * Unload environment. A notification from environment API whenever an
    * <code>UNLOAD environment</code> statement is executed.
    * 
    * @param   envName
    *          Environment name.
    */
   public static void unload(String envName)
   {
      WorkArea wa = local.get(false);
      if (wa != null && envName != null)
      {
         wa.envColorTables.remove(envName);
         if (envName.equals(wa.envName))
         {
            wa.envName = null;
         }
      }
   }
   
   /**
    * Create a color pair from the specified value.
    * 
    * @param    value
    *           The pair of color indices.
    *           
    * @return   The {@link ColorPair}.
    */
   static ColorPair createColorPair(String value)
   {
      if (value == null)
      {
         return null;
      }

      int fgColor = 0;
      int bgColor = 0;

      String[] colors = value.split(",");
      if (colors.length == 2)
      {
         try
         {
            fgColor = Integer.valueOf(colors[0]);
            bgColor = Integer.valueOf(colors[1]);
         }
         catch (NumberFormatException e)
         {
            LOG.log(Level.SEVERE, "Could not parse color pair", e);
            return null;
         }
      }
      else
      {
         return null;
      }

      return new ColorPair(fgColor, bgColor);
   }

   /**
    * Display a warning message if the attribute was not set.
    * 
    * @param   attribute
    *          Attribute name.
    */
   private static void warning(String attribute)
   {
      String msg = String.format("**Unable to set attribute %s for %s.",
                                  attribute,
                                  LegacyResource.PSEUDO_WIDGET);
      // display as warning
      ErrorManager.warningModeEnable();
      ErrorManager.displayError(4078, msg);
      ErrorManager.warningModeDisable();   
   }

   /**
    * Load color-table from server directory or from environment.
    * 
    * @param   userSystemColors
    *          User system colors.
    * @param   custom
    *          If <code>false</code> get default form server directory.
    * 
    * @return  A string array of colors RGB values.
    */
   private static ColorTableEntry[] loadColorTable(Map<Integer, ColorRgb> userSystemColors,
                                                   boolean custom)
   {
      String[] colors = (custom) ? readTable(COLOR_TABLE, COLOR_SECTION, COLOR_KEY) :
                                   read(COLOR_TABLE, COLOR_SECTION, COLOR_KEY, custom);
      return makeColorTable(colors, userSystemColors);
   }
   
   /**
    * Load system-colors from server directory.
    * 
    * @return  A string array of system colors RGB values.
    */
   private static ColorTableEntry[] loadSystemColors()
   {
      String[] colors = read(SYSTEM_COLORS, "sys-colors", COLOR_KEY, false);
      return makeColorTable(colors, null);
   }

   /**
    * Load a standard color NORMAL, INPUT or MESSAGES.
    *  
    * @param   name
    *          Standard color name.
    * @param   fgColor
    *          Default foreground color.
    * @param   bgColor
    *          Default background color.
    * 
    * @return  Standard color pair.
    */
   private static ColorPair loadStandardColor(String name, int fgColor, int bgColor)
   {
      String value = readValue(STANDARD_COLORS, COLOR_SECTION, name);
      
      ColorPair pair = createColorPair(value);
      
      if (pair == null)
      {
         pair = new ColorPair(fgColor, bgColor);
      }
      
      return pair;
   }

   /**
    * Make a collection of ColorTableEntry items from color strings.
    * <p>
    * Color table strings are in the following format: colorn = { R , G , B , | colorname }
    * where:
    *    - n color number (max. 255)
    *    - R, G and B are red, green and blue integer values example 127,0,127
    *    - colorname is the name of a system color example COLOR-SCROLLBAR
    *       
    * @param   colors
    *          Color strings.
    * @param   userSystemColors
    *          User system colors.
    *          
    * @return  A collection of ColorTableEntry items.
    */
   private static ColorTableEntry[] makeColorTable(String[] colors, 
                                                   Map<Integer, ColorRgb> userSystemColors)
   {
      // read table-color form server directory
      ColorTableEntry[] colorTable = new ColorTableEntry[colors.length];
      
      for (int i = 0; i < colors.length; i++)
      {
         String[] rgb = colors[i].split(",");
         if (rgb.length == 3)
         {
            // R,G and B values
            int r = Integer.valueOf(rgb[0]);
            int g = Integer.valueOf(rgb[1]);
            int b = Integer.valueOf(rgb[2]);
            
            colorTable[i] = new ColorTableEntry(i, r, g, b);
         }
         else if (rgb.length == 1 && userSystemColors != null)
         {
            // System color name
            String colorName = colors[i].toUpperCase();
            Integer index = SysColor.systemColorNames.get(colorName);
            if (index == null)
            {
               colorTable[i] = new ColorTableEntry(i, 0, 0, 0);
               log(i, "invalid system color name provided on input");
            }
            else
            {
               ColorRgb sysc = userSystemColors.get(index);
               if (sysc == null)
               {
                  colorTable[i] = new ColorTableEntry(i, 0, 0, 0);
                  log(i, "no system color is defined for the given (valid) system color name");
               }
               else
               {
                  colorTable[i] = new ColorTableEntry(i, sysc);
               }
            }
            colorTable[i].setName(colorName);
         }
         else
         {
            colorTable[i] = new ColorTableEntry(i, 0, 0, 0);
            log(i, "improperly formatted color string was provided on input");
         }
      }
      
      return colorTable;
   }
   
   /**
    * Log warning messages related to color mapping errors building color tables.
    * 
    * @param   n
    *          Color number.
    * @param   reason
    *          A string specifying the reason of the warning message.
    */
   private static void log(int n, String reason)
   {
      String fmt = "Color [%d] has been mapped to a 0,0,0 RGB value!\n Reason: %s.";
      LOG.warning(String.format(fmt, n, reason));
   }
   
   /**
    * Locate the context-local data of this class.
    * 
    * @return   A {@link WorkArea} instance.
    */
   private static WorkArea locate()
   {
      return local.get();
   }

   /**
    * Get color entry for the given index.
    * 
    * @param   ect
    *          Environment color table.
    * @param   index
    *          Color table index.
    *          
    * @return  see above.
    */
   private static ColorTableEntry getColorTableEntry(EnvironmentColorTable ect, int index)
   {
      if (ect != null)
      {
         List<ColorTableEntry> colors = ect.colors;
         
         if (colors != null && !colors.isEmpty())
         {
            if (index >= 0 && index < colors.size())
            {
               return colors.get(index);
            }
         }
      }
      
      return null;
   }
   
   /**
    * Load color table from current environment.  Colors of default environment should be read
    * from server directory. For custom environments INI files should be used.
    * 
    * @param   userSystemColors   
    *          User system colors.
    * @param   custom
    *          If <code>false</code> get default form server directory.
    *          
    * @return  environment color table
    */
   private static EnvironmentColorTable readColorTable(Map<Integer, ColorRgb> userSystemColors,
                                                       boolean custom)
   {
      // first get the default table
      ColorTableEntry[] colorTable =
         defaultColorTable.toArray(new ColorTableEntry[defaultColorTable.size()]);
      // load color table
      ColorTableEntry[] colorTableOverride = loadColorTable(userSystemColors, custom);

      // if found overrides in directory file, replace in final color table
      for (ColorTableEntry aColor : colorTableOverride)
      {
         // check color index in override
         int ndx = aColor.getIndex();
         // scan default table for same index
         for (int j = 0; j < colorTable.length; j++)
         {
            // found first entry with same index
            if (colorTable[j].getIndex() == ndx)
            {
               // replace default color with override
               colorTable[j] = aColor;
               // and stop scanning default table
               break;
            }
         }
      }

      // load standard colors NORMAL, INPUT and MESSAGES
      Map<String, ColorPair> standardColors = new HashMap<>();
      ColorPair normal = loadStandardColor(ColorSpec.COLOR_NORMAL, 0, 15);
      standardColors.put(ColorSpec.COLOR_NORMAL, normal);
      ColorPair input = loadStandardColor(ColorSpec.COLOR_INPUT, 15, 0);
      standardColors.put(ColorSpec.COLOR_INPUT, input); 
      ColorPair messages = loadStandardColor(ColorSpec.COLOR_MESSAGES, 15, 1);
      standardColors.put(ColorSpec.COLOR_MESSAGES, messages);
     
      // build environment colors
      return new EnvironmentColorTable(null, colorTable, standardColors);
   }      

   /**
    * Synchronize environments. Whenever changes are made on color table of environments a
    * synchronization should be made between server state and client state. Environments
    * that should be synchronized are marked as dirty ad add to the list on environments
    * into the current server state structure. If environment is already marked as dirty
    * server state already contains the environment.
    * 
    * @param   ect
    *          Environment to synchronize.
    */
   private static void synch(EnvironmentColorTable ect)
   {
      if (!ect.dirty)
      {
         ect.dirty = true;
         LogicalTerminal.synch(ect);
      }
   }
   
   /**
    * Get current environment.
    * 
    * @param   wa
    *          Work area instance.
    *          
    * @return  see above.
    */
   private static EnvironmentColorTable getCurrentColorTable(WorkArea wa)
   {
      String envName = wa.envName;
      
      // check if already loaded.
      EnvironmentColorTable ect = wa.envColorTables.get(envName);
      
      if (ect == null)
      {
         // load current environment color table if not loaded
         ect = readColorTable(wa.userSystemColors, true);
         if (ect == null)
         {
            // normally should not happen
            LOG.warning(String.format("Cannot load [%s] environment!", envName));
         }
         else
         {
            // set environment name
            ect.envName = envName;
            // add to environments cache
            wa.envColorTables.put(envName, ect);
         }
      }
      
      return ect;
   }
  
   /**
    * An immutable structure to hold color pairs. 
    */
   @SuppressWarnings("serial")
   public static class ColorPair
   implements Serializable
   {
      /** Foreground color index */
      public final int fgColor;
      
      /** Background color index */
      public final int bgColor;
      
      /**
       * Constructor.
       * 
       * @param   fgColor
       *          Foreground color;
       * @param   bgColor
       *          Background color; 
       */
      public ColorPair(int fgColor, int bgColor)
      {
         this.fgColor = fgColor;
         this.bgColor = bgColor;
      }
      
      /**
       * Return a color pair string comma separated.
       * 
       * @return  see above.
       */
      public final String asString()
      {
         return fgColor + "," + bgColor;
      }
   }

   /**
    * The color-table associated with a certain environment.
    */
   @SuppressWarnings("serial")
   public static class EnvironmentColorTable
   implements Serializable
   {
      /** Dirty flag */
      private transient volatile boolean dirty = false;
      
      /** The environment's name. */
      public String envName;
      
      /** The environment's color-table. */
      public final List<ColorTableEntry> colors;
      
      /** The environment's standard colors NORMAL, INPUT and MESSAGES */
      public final Map<String, ColorPair> standardColors;

      /**
       * Create a new color-table, for the specified environment.
       * 
       * @param    name
       *           The environment's name. Default environment has <code>null</code> name.
       * @param    colorTable
       *           The colors part of this color-table.
       */
      public EnvironmentColorTable(String name, 
                                   ColorTableEntry[] colorTable, 
                                   Map<String, ColorPair> standardColors)
      {
         this.envName        = name;
         this.colors         = new ArrayList<>(Arrays.asList(colorTable));
         this.standardColors = new HashMap<>(standardColors);
      }
   }
      
   /**
    * Context-local data for this color manager.
    */
   private static class WorkArea
   {  
      /** The resource's ID. */
      private Long id = null;

      /** System colors for this session. */
      private Map<Integer, ColorRgb> userSystemColors;

      /** System colors */
      private ColorTableEntry[] systemColors;
      
      /** A map of environment names to their associated color-table. */
      private Map<String, EnvironmentColorTable> envColorTables = new HashMap<>();
      
      /** Client driver mode */
      private boolean isChui = false;
      
      /** Current environment name */
      private String envName = null;
      
      /**
       * Create a new work-area.
       */
      public WorkArea()
      {
         // client graphical mode
         isChui = LogicalTerminal.isChui();
         
         // get system colors always from server directory
         initSystemColorTable();
         
         // get default color table from server directory
         EnvironmentColorTable ect = readColorTable(userSystemColors, false);
         if (ect != null)
         {
            // add to environments collection
            envColorTables.put(null, ect);
         }
         
         // get current environment name
         envName = EnvironmentOps.getCurrentEnvironmentName();
         if (envName != null)
         {
            // current environment color table
            getCurrentColorTable(this);
         }
      }
      
      /**
       * Initialize system colors table.
       */
      private void initSystemColorTable()
      {
         systemColors = loadSystemColors();
         
         // if system colors are missing use defaults
         if (systemColors.length == 0)
         {
            // call ThemeManager.getCurrentTheme().getDefaultSystemPalette() from the client
            userSystemColors = LogicalTerminal.getThemeSystemPalette();
         }
         else
         {
            userSystemColors = new HashMap<>();
            
            for (ColorTableEntry color : systemColors)
            {
               userSystemColors.put(color.getIndex(), color);
            }
         }
      }
   }
   
   /**
    * A class that collects option information about a change color statement using chained
    * methods.
    * <p>
    * Each public method emulates an option of the statement. They do a quick validation. If
    * successful, they return a reference to this object to allow other method to be chained.
    * <p>
    * The {@code execute()} method does the final checks, send the request with gathered options
    * to client, awaits the user to interact and finally sets up the response and confirmation
    * variable {@code update}.
    */
   public static class ColorChooserDialog
   {
      /**
       * The (index of the) selected color that is to be modified by user. The color index must be
       * a number between 0 and 255 inclusive.
       */
      private int colorRef = 0;
      
      /**
       * A logical variable used as output parameter that will contain the status of the user's
       * dialog interaction: if the user validates it by selecting the {@code OK} button, the
       * value of the variable is set to {@code TRUE} on return. If the user rejects the dialog by
       * pressing the {@code Cancel} button, the variable's value on return will be {@code FALSE}.
       */
      private logical toUpdate = null;
      
      /**
       * Specifies the parent window for the dialog box to be displayed. The value must be a
       * handle to a window.
       */
      private handle inWindow = null;
      
      /**
       * The constructor. Configures the dialog in the simplest form when only the colorRef
       * reference is passed. To activate it, just call {@code execute}.
       *
       * @param   colRef
       *          The (index of the) selected color that is to be selected by user.
       */
      private ColorChooserDialog(int64 colRef)
      {
         if (colRef == null || colRef.isUnknown())
         {
            ErrorManager.recordOrThrowError(3189,
                                            "Unable to extract expression for system-dialog",
                                            false);
            return;
         }
         
         int color = colRef.intValue();
         if (color <= 0 || color > getNumEntries().intValue())
         {
            String msg = "Invalid color number, " + color + ", given for common dialog"; 
            ErrorManager.recordOrThrowError(3187, msg, false);
            return;
         }
         
         logical dynamic = getDynamic(colRef);
         if (dynamic.isUnknown() || !dynamic.booleanValue())
         {
            String msg = "Invalid color number, " + color + ", given for common dialog";
            ErrorManager.recordOrThrowError(3187, msg, false);
            return;
         }
         
         colorRef = color;
      }
      
      /**
       * The constructor. Configures the dialog in the simplest form when only the colorRef
       * reference is passed. To activate it, just call {@code execute}.
       *
       * @param   colRef
       *          The (index of the) selected color that is to be selected by user.
       */
      private ColorChooserDialog(long colRef)
      {
         if (colRef <= 0 || colRef > getNumEntries().intValue())
         {
            String msg = "Invalid color number, " + colRef + ", given for common dialog";
            ErrorManager.recordOrThrowError(3187, msg, false);
            return;
         }
         
         colorRef = (int) colRef;
      }
      
      /**
       * Configures the file dialog to be created as child of a certain windows.
       *
       * @param   window
       *          A handle to a window that will be the parent of the dialog. The handle must
       *          refer a window or an error condition will be issued.
       *
       * @return  The reference to current {@code FileDialog}. Used for chaining.
       *
       * @see     #inWindow
       */
      public ColorChooserDialog setWindow(handle window)
      {
         if (window.isUnknown())
         {
            String msg = "Could not find widget handle for window from expression";
            ErrorManager.recordOrThrowError(3138, msg, false);
         }
         else
         {
            WrappedResource r = window.getResource();
            if (r instanceof WindowWidget)
            {
               this.inWindow = window;
            }
            else
            {
               String msg = "Wrong type of object handle (%s)  -- must be a window handle";
               ErrorManager.recordOrThrowError(3139,
                                               String.format(msg, r.getClass().toString()),
                                               false);
            }
         }
         return this;
      }
      
      /**
       * Configures the variable to be set when the file selection dialog is closed.
       *
       * @param   update
       *          A variable that will hold the return status of the dialog. That is {@code true}
       *          if the dialog was accepted and {@code false} otherwise.
       *
       * @return  The reference to current {@code FileDialog}. Used for chaining.
       *
       * @see     #toUpdate
       */
      public ColorChooserDialog update(logical update)
      {
         this.toUpdate = update;
         return this;
      }
      
      /**
       * Instructs the client to show the {@code Color Chooser} dialog using the configured
       * options using previously chained methods. When the client returns, updates the
       * {@link #toUpdate} variable, if set, to let the programmer know whether the client
       * accepted the dialog or rejected it. If a new color is selected, the entry in the palette
       * is updated to the newly selected RGB value.
       */
      public void execute()
      {
         if (inWindow == null)
         {
            inWindow = LogicalTerminal.activeWindow();
         }
         WindowWidget window = (WindowWidget) inWindow.getResource();
         integer oldRGB = getRgbValue(colorRef);
         
         ColorRgb rgb = getColor(window, new ColorRgb(oldRGB.intValue()));
         
         if (rgb != null)
         {
            // update the palette entry if dialog was not cancelled
            setRgbValue(colorRef, rgb.getRgb());
         }
         
         if (toUpdate != null)
         {
            // update the output variable if specified
            toUpdate.assign(rgb != null);
         }
      }
      
      /**
       * Worker method for getting the color from user using the specific dialog.
       * 
       * @param   window
       *          The parent window. If it is {@code null} the default window is used.
       * @param   init
       *          The initial color.
       *
       * @return  The color the user has chosen or {@code null} if the color chooser dialog was 
       *          rejected.
       */
      public static ColorRgb getColor(WindowWidget window, ColorRgb init)
      {
         if (window == null)
         {
            window = (WindowWidget) LogicalTerminal.defaultWindow().getResource();
         }
         
         int newRGB = LogicalTerminal.getClient().chooseColor(init.getGuiRgb(), window.getId());
         
         // check the result. If the dialog was cancelled, -1 is returned
         return (newRGB == -1) ? null : new ColorRgb(newRGB);
      }
   }
}
