/*
** Module   : ConfigSyncManager.java
** Abstract : Implements synchronization of config changes. 
**
** Copyright (c) 2015-2018, Golden Code Development Corporation.
**
** -#- -I- --Date-- --------------------------------Description-----------------------------------
** 001 HC  20150323 Created initial version
** 002 EVL 20160224 Javadoc fixes to make compatible with Oracle Java 8 for Solaris 10.
** 003 IAS 20160331 Fixed typo in the markScopeEnd() method
** 004 ECF 20171122 Minor optimization in registerConfig.
** 005 ECF 20180101 Minimize thread-local access to improve performance.
*/
/*
** This program is free software: you can redistribute it and/or modify
** it under the terms of the GNU Affero General Public License as
** published by the Free Software Foundation, either version 3 of the
** License, or (at your option) any later version.
**
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
** GNU Affero General Public License for more details.
**
** You may find a copy of the GNU Affero GPL version 3 at the following
** location: https://www.gnu.org/licenses/agpl-3.0.en.html
** 
** Additional terms under GNU Affero GPL version 3 section 7:
** 
**   Under Section 7 of the GNU Affero GPL version 3, the following additional
**   terms apply to the works covered under the License.  These additional terms
**   are non-permissive additional terms allowed under Section 7 of the GNU
**   Affero GPL version 3 and may not be removed by you.
** 
**   0. Attribution Requirement.
** 
**     You must preserve all legal notices or author attributions in the covered
**     work or Appropriate Legal Notices displayed by works containing the covered
**     work.  You may not remove from the covered work any author or developer
**     credit already included within the covered work.
** 
**   1. No License To Use Trademarks.
** 
**     This license does not grant any license or rights to use the trademarks
**     Golden Code, FWD, any Golden Code or FWD logo, or any other trademarks
**     of Golden Code Development Corporation. You are not authorized to use the
**     name Golden Code, FWD, or the names of any author or contributor, for
**     publicity purposes without written authorization.
** 
**   2. No Misrepresentation of Affiliation.
** 
**     You may not represent yourself as Golden Code Development Corporation or FWD.
** 
**     You may not represent yourself for publicity purposes as associated with
**     Golden Code Development Corporation, FWD, or any author or contributor to
**     the covered work, without written authorization.
** 
**   3. No Misrepresentation of Source or Origin.
** 
**     You may not represent the covered work as solely your work.  All modified
**     versions of the covered work must be marked in a reasonable way to make it
**     clear that the modified work is not originating from Golden Code Development
**     Corporation or FWD.  All modified versions must contain the notices of
**     attribution required in this license.
*/

package com.goldencode.p2j.ui;

import java.util.*;

/**
 * The class implements synchronization of changes performed in
 * configs (types inheriting from {@link WidgetConfig}) with their
 * respective config owners ({@link ConfigOwner}) - usually widgets.
 * Do not reference this class directly, instead use {@link SyncConfigChanges}
 * or {@link ConfigManager#syncConfigChanges(Runnable)} to take advantage
 * of the synchronization feature this class provides. 
 * <p>
 * Widget state is distributed between a widget instance itself and 
 * its associated config. Any changes performed in the config instance must
 * be synchronized with its owning widget. The core synchronization logic lies 
 * in the concrete implementations of {@link ConfigOwner#afterConfigUpdate(WidgetConfig)}.
 * This class frees the developer from the burden of keeping track of when
 * this method should be called, providing proper parameters, handling edge cases 
 * like recursion or synchronization of complex widget structures, batching the 
 * changes so that the method {@link ConfigOwner#afterConfigUpdate(WidgetConfig)}
 * is not called after every config change and determining the correct config owner.
 * <p>
 * In principal the logic implemented by this class is very simple. When the developer
 * starts synchronization the class will monitor all the config field assignments.
 * When a config field is assigned its target instance is considered "dirty" and
 * saved for later. When the developer stops synchronization all the
 * dirty instances are synchronized - for each instance its config
 * owner is determined and {@link ConfigOwner#afterConfigUpdate(WidgetConfig)} called.
 * The code between the points when synchronization is started and stopped is called
 * synchronization scope. When scopes are nested, the synchronization is performed only after
 * the root scope is exited.
 * <p>
 * Here's an example.
 * 
 * <pre><code>
 * {@literal @}SyncConfigChanges
 * void methodWhereConfigsAreChanged()
 * {
 *    configRef.widthChars = 10;
 *    innerMethodWhereConfigsAreChanged();
 * }
 * 
 * void innerMethodWhereConfigsAreChanged()
 * {
 *    ...
 *    yetAnotherInnerMethodWhereConfigsAreChanged();
 *    ...
 * }
 * 
 * {@literal @}SyncConfigChanges
 * void yetAnotherInnerMethodWhereConfigsAreChanged()
 * {
 *    configRef2.heightChars = 20;
 * }
 * 
 * void case1()
 * {
 *    // assuming no other sync scope exists on the call stack
 *    // methodWhereConfigsAreChanged will start the root 
 *    // sync scope
 *    
 *    methodWhereConfigsAreChanged();
 *    
 *    // after the method exits, the owners of configRef and configRef2
 *    // will be synchronized (by invocation of 
 *    // {@link ConfigOwner#afterConfigUpdate(WidgetConfig)})
 * }
 * 
 * void case2()
 * {
 *    // assuming no other sync scope exists on the call stack
 *    // yetAnotherInnerMethodWhereConfigsAreChanged will start 
 *    // the root sync scope
 *    
 *    innerMethodWhereConfigsAreChanged();
 *    
 *    // after yetAnotherInnerMethodWhereConfigsAreChanged exits, 
 *    // the owner of configRef2 will be synchronized (by invocation of 
 *    // {@link ConfigOwner#afterConfigUpdate(WidgetConfig)})
 * }
 * </code></pre>
 * <p>
 * Synchronization scopes don't cross thread boundaries.
 *
 * <pre><code>
 * {@literal @}SyncConfigChanges
 * void methodWhereConfigsAreChanged()
 * {
 *    configRef.widthChars = 10;
 *    
 *    new Thread().run(innerMethodWhereConfigsAreChanged());
 * }
 * 
 * void innerMethodWhereConfigsAreChanged()
 * {
 *    configRef2.heightChars = 20;
 * }
 * 
 * void case3()
 * {
 *    methodWhereConfigsAreChanged();
 *    
 *    // after methodWhereConfigsAreChanged exits, 
 *    // the owner of configRef2 will NOT be synchronized
 * }
 * </code></pre>
 * <p>
 * The semantics of {@link ConfigManager#syncConfigChanges(Runnable)} 
 * is the same as {@link SyncConfigChanges}. The only difference
 * is its imperative form.
 */
public class ConfigSyncManager
{
   /** Thread-local state */
   private static ThreadLocal<State> localState = new ThreadLocal<State>()
   {
      @Override
      protected State initialValue()
      {
         return new State();
      }
   };
   
   /**
    * Marks the start of config synchronization scope on the current thread. 
    */
   public static void markScopeStart()
   {
      localState.get().scopeDepth++;
   }
   
   /**
    * Marks the end of config synchronization scope on the current thread.
    * When the ended is a root scope then the method also handles notification
    * of config owners to let them synchronize with the config changes. 
    */
   @SuppressWarnings("unchecked")
   public static void markScopeEnd()
   {
      State state = localState.get();
      
      int depth = state.scopeDepth;
      state.scopeDepth--;
      
      // exiting root scope, sync the config updates
      if (depth == 1)
      {
         ConfigManager cm = null;
         
         Map<WidgetConfig, ConfigOwner<?>> ownerMap = state.configOwnerMap;
         state.configOwnerMap = new HashMap<WidgetConfig, ConfigOwner<?>>();
         
         Map<WidgetConfig, WidgetConfig> configs = state.pendingConfigs;
         state.pendingConfigs = new HashMap<WidgetConfig, WidgetConfig>();
         
         for (Map.Entry<WidgetConfig, WidgetConfig> c : configs.entrySet())
         {
            WidgetConfig baseConfig = c.getValue();
            
            if (baseConfig == null)
            {
               throw new IllegalStateException("Unable to resolve base configuration!");
            }
            
            ConfigOwner<WidgetConfig> owner = 
               (ConfigOwner<WidgetConfig>) ownerMap.get(c.getKey());
            
            if (owner == null && c.getKey().id != null)
            {
               if (cm == null)
               {
                  cm = baseConfig.getConfigManager();
               }
               owner = (ConfigOwner<WidgetConfig>) cm.resolveConfigOwner(c.getKey().id);
            }
            
            if (owner == null)
            {
               // this may happen, the registered config may not have
               // any owner, just ignore
               continue;
            }
            
            if (state.ownerCallstack.contains(owner))
            {
               // prevent recursion - do not run afterConfigUpdate
               // on the same target recursively
               continue;
            }
            
            try
            {
               state.ownerCallstack.push(owner);
               owner.afterConfigUpdate(baseConfig);
            } 
            finally
            {
               state.ownerCallstack.pop();
            }
         }
      }
   }
   
   /**
    * Registers a modified config for state synchronization
    * with its owner. When the passed in owner is <code>null</code>
    * it is later resolved with the help of the config id.
    * 
    * @param    owner
    *           {@link ConfigOwner} reference, may be null.
    * @param    config
    *           {@link WidgetConfig} reference, must not be null.
    */
   public static void registerConfig(ConfigOwner<?> owner, WidgetConfig config)
   {
      if (config == null)
      {
         throw new IllegalArgumentException("Config must not be null!");
      }
      
      State state = localState.get();
      
      if (state.scopeDepth == 0)
      {
         return;
      }
      
      if (owner != null)
      {
         state.configOwnerMap.put(config, owner);
      }
      
      if (!state.duplicateInProgress)
      {
         Map<WidgetConfig, WidgetConfig> configs = state.pendingConfigs;
         
         if (!configs.containsKey(config))
         {
            state.duplicateInProgress = true;
            
            try
            {
               WidgetConfig baseConfig = config.getConfigManager().duplicate(config);
               configs.put(config, baseConfig);
            }
            finally
            {
               state.duplicateInProgress = false;
            }
         }
      }
   }
   
   /**
    * Thread-local state.
    */
   private static class State
   {
      /** The depth of sync scope on the current thread. **/
      int scopeDepth = 0;
      
      /** Map of config references and their respective owners. **/
      Map<WidgetConfig, ConfigOwner<?>> configOwnerMap = new HashMap<>();
      
      /** Map of configs modified in the current sync scope. **/
      Map<WidgetConfig, WidgetConfig> pendingConfigs = new HashMap<>();
      
      /** Flag to prevent recursion during config duplication. **/
      boolean duplicateInProgress = false;
      
      /**
       * This helps to prevent recursion when config update notification
       * causes recursive notification on the same owner.
       **/
      Deque<ConfigOwner<?>> ownerCallstack = new ArrayDeque<>(128);
   }
}