/*
** Module   : ControlTextWidget.java
** Abstract : server side text widget implementation
**
** Copyright (c) 2005-2018, Golden Code Development Corporation.
**
** -#- -I- --Date-- --JPRM-- -----------------------------Description-----------------------------
** 001 NVS 20051026   @23488 Created initial version.
** 002 SIY 20060202   @24234 Added support for NO-LABELS widget flag.
** 003 GES 20060208   @24381 Removed setFormat() since this was hiding
**                           the version implemented in ComponentConfig
**                           and it was also redundant.
** 004 SIY 20060209   @24431 Moved methods to base class.
** 005 SIY 20060316   @25128 Added support for header flag.
** 006 SIY 20060426   @25912 Changed signatures of getNextTabItem.
** 007 GES 20060627   @27649 Removed setPfColor/getPfColor methods which
**                           were only placed here to block usage by
**                           throwing an exception in our implementation.
**                           This attribute is valid and must not be
**                           blocked (they are implemented in one of the
**                           parent classes in our inheritance hierarchy). 
** 008 SIY 20060710   @27983 Added static widget content flag.
** 009 SIY 20060717   @28082 Added @ widget flag.
** 010 GES 20060803   @28372 Cleanup and refactoring.
** 011 NVS 20070418   @33104 Removed getNextTabItem() method which was
**                           only placed here to block usage by
**                           throwing an exception in our implementation.
**                           This attribute is valid and must not be
**                           blocked (they are implemented in one of the
**                           parent classes in our inheritance hierarchy).
** 012 GES 20080320   @37584 Added TYPE attribute.
** 013 SIY 20111101          Updated imports.
** 014 CA  20130221          Refactored usage of TYPE attribute.
** 015 CA  20130222          Added NPE protection for cases when a widget is used before is 
**                           attached to a frame (after a CREATE WIDGET statement is used).
** 016 CA  20130927          Made it abstract, as instances of TextWidget or LiteralWidget need
**                           to be emitted when string literals are used in frames.
** 017 CA  20131115          Changes related to properly instantiate dynamic resources.
** 018 CA  20140926          Added shared widget configuration support, to allow GUI/ChUI
**                           concrete implementation for the same widget ID.  Refactored the
**                           widget configuration classes: all fields were made public; these
**                           classes need to be as dumb as possible, all logic related to
**                           setting/getting a certain field should be at the widget or at the
**                           caller. Refs #2254
** 019 CA  20150106          Changes for SIDE-LABEL-HANDLE runtime support.
** 020 CA  20150323          Added API to set the format length from the widget's @ phrase(s).
** 021 HC  20150323          Removed business logic from config classes.
** 022 IAS 20150801          Added support fote BLANK attribute.
** 023 CA  20160303          Made attribute setters be a no-op if the value doesn't change.
** 024 OM  20171122          Used WriteProtectable interface for READ-ONLY attribute.
** 025 CA  20180130          Replaced pushScreenDefinition() with pushWidgetAttr() for cases when
**                           business logic is assigning a widget attribute.
*/ 

/*
** This program is free software: you can redistribute it and/or modify
** it under the terms of the GNU Affero General Public License as
** published by the Free Software Foundation, either version 3 of the
** License, or (at your option) any later version.
**
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
** GNU Affero General Public License for more details.
**
** You may find a copy of the GNU Affero GPL version 3 at the following
** location: https://www.gnu.org/licenses/agpl-3.0.en.html
** 
** Additional terms under GNU Affero GPL version 3 section 7:
** 
**   Under Section 7 of the GNU Affero GPL version 3, the following additional
**   terms apply to the works covered under the License.  These additional terms
**   are non-permissive additional terms allowed under Section 7 of the GNU
**   Affero GPL version 3 and may not be removed by you.
** 
**   0. Attribution Requirement.
** 
**     You must preserve all legal notices or author attributions in the covered
**     work or Appropriate Legal Notices displayed by works containing the covered
**     work.  You may not remove from the covered work any author or developer
**     credit already included within the covered work.
** 
**   1. No License To Use Trademarks.
** 
**     This license does not grant any license or rights to use the trademarks
**     Golden Code, FWD, any Golden Code or FWD logo, or any other trademarks
**     of Golden Code Development Corporation. You are not authorized to use the
**     name Golden Code, FWD, or the names of any author or contributor, for
**     publicity purposes without written authorization.
** 
**   2. No Misrepresentation of Affiliation.
** 
**     You may not represent yourself as Golden Code Development Corporation or FWD.
** 
**     You may not represent yourself for publicity purposes as associated with
**     Golden Code Development Corporation, FWD, or any author or contributor to
**     the covered work, without written authorization.
** 
**   3. No Misrepresentation of Source or Origin.
** 
**     You may not represent the covered work as solely your work.  All modified
**     versions of the covered work must be marked in a reasonable way to make it
**     clear that the modified work is not originating from Golden Code Development
**     Corporation or FWD.  All modified versions must contain the notices of
**     attribution required in this license.
*/

package com.goldencode.p2j.ui;

import com.goldencode.p2j.util.*;

/**
 * Represents common function for text oriented widgets.
 */
public abstract class ControlTextWidget<T extends ControlTextConfig>
extends ControlEntity<T>
implements WriteProtectable
{
   /**
    * Constructor for use by sub-classes.
    *
    * @param    dynamic
    *           Flag indicating if this is a static or dynamic resource.
    * @param    ctc
    *           Configuration container.
    */
   public ControlTextWidget(boolean dynamic, T ctc)
   {
      super(dynamic, ctc);
   }

   /**
    * Checks whether the widget is write-protected.
    * 
    * @return   <code>true</code> if the widget is write-protected.
    */
   @Override
   public logical isReadOnly() 
   {
      return new logical(config.readOnly);
   }

   /**
    * Sets the READ-ONLY attribute of the widget.
    * 
    * @param    r
    *           <code>true</code> if the widget should be write-protected.
    */
   @Override
   public void setReadOnly(logical r) 
   {
      if (config.readOnly == r.booleanValue())
      {
         return;
      }
      
      config.readOnly = r.booleanValue();
      pushWidgetAttr("readOnly", config.readOnly);
   }

   /**
    * Force displaying of label.
    *
    * @param    force
    *           New value for the flag.
    */
   @Override
   public void setForceLabel(final boolean force)
   {
      super.setForceLabel(force);

      checkLabelsField();

      // this is a frame definition option, can't be used via widget handle
   }

   /**
    * Set header flag which marks widget used as frame header.
    * 
    * @param   b
    *          New state of the flag.
    */
   public void setHeader(final boolean b)
   {
      if (config.header == b)
      {
         return;
      }
      
      config.header = b;
      checkLabelsField();

      // this is a frame definition option, can't be used via widget handle
   }

   /**
    * Set @ widget flag.
    * 
    * @param   flag
    *          New value for the flag.
    */
   public void setAt(boolean flag)
   {
      config.flagAt = flag;
   }
   
   /**
    * Set the length of the format for the widget's @ phrases.
    * <p>
    * This API is called only during frame setup, thus the frame doesn't need to push the screen
    * definition explicitly.
    * 
    * @param   length
    *          The maximum format length from the widget's @ phrases.
    */
   public void setAtFormatLength(int length)
   {
      config.atFormatLength = length;
   }

   /**
    * Set static widget content flag.
    * <p>
    * When this widget was created by the use of static data (a constant
    * or literal in the source code) and was created in a FORM or DEFINE
    * FRAME statement, then this flag should be set.
    * 
    * @param   flagStatic
    *          New value for the flag.
    */
   protected void setStatic(boolean flagStatic)
   {
      config.flagStatic = flagStatic;
      config.possibleSideLabel = !flagStatic;
   }
   
   /**
    * Check the value of the {@link ControlConfig#labels} field, as it depends on the
    * {@link WidgetConfig#header} and {@link ControlConfig#forceLabel} values.
    */
   private void checkLabelsField()
   {
      // set no-labels, but only if labels are not already forced explicitly
      if (config.header && !config.forceLabel)
      {
         config.labels = false;
      }
   }
}
