/*
** Module   : EnhancedBrowseConfigKey.java
** Abstract : Enhanced browse configuration key.
**
** Copyright (c) 2018, Golden Code Development Corporation.
**
** -#- -I- --Date-- ---------------------------------Description----------------------------------
** 001 SVL 20180529 Created initial version.
*/
/*
** This program is free software: you can redistribute it and/or modify
** it under the terms of the GNU Affero General Public License as
** published by the Free Software Foundation, either version 3 of the
** License, or (at your option) any later version.
**
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
** GNU Affero General Public License for more details.
**
** You may find a copy of the GNU Affero GPL version 3 at the following
** location: https://www.gnu.org/licenses/agpl-3.0.en.html
**
** Additional terms under GNU Affero GPL version 3 section 7:
**
**   Under Section 7 of the GNU Affero GPL version 3, the following additional
**   terms apply to the works covered under the License.  These additional terms
**   are non-permissive additional terms allowed under Section 7 of the GNU
**   Affero GPL version 3 and may not be removed by you.
**
**   0. Attribution Requirement.
**
**     You must preserve all legal notices or author attributions in the covered
**     work or Appropriate Legal Notices displayed by works containing the covered
**     work.  You may not remove from the covered work any author or developer
**     credit already included within the covered work.
**
**   1. No License To Use Trademarks.
**
**     This license does not grant any license or rights to use the trademarks
**     Golden Code, FWD, any Golden Code or FWD logo, or any other trademarks
**     of Golden Code Development Corporation. You are not authorized to use the
**     name Golden Code, FWD, or the names of any author or contributor, for
**     publicity purposes without written authorization.
**
**   2. No Misrepresentation of Affiliation.
**
**     You may not represent yourself as Golden Code Development Corporation or FWD.
**
**     You may not represent yourself for publicity purposes as associated with
**     Golden Code Development Corporation, FWD, or any author or contributor to
**     the covered work, without written authorization.
**
**   3. No Misrepresentation of Source or Origin.
**
**     You may not represent the covered work as solely your work.  All modified
**     versions of the covered work must be marked in a reasonable way to make it
**     clear that the modified work is not originating from Golden Code Development
**     Corporation or FWD.  All modified versions must contain the notices of
**     attribution required in this license.
*/
package com.goldencode.p2j.ui;

import java.io.*;
import java.util.*;

/**
 * Enhanced browse configuration key. Uniquely identifies browse in an application using
 * combination of the parent external procedure name, browse name and set of column keys.
 * If the parent external procedure name is <code>null</code> then this key references to all
 * browses.
 */
public class EnhancedBrowseConfigKey
implements Serializable
{
   /*
    * The name of the parent external procedure which contains the browse.
    * <code>null</code> if the key is applied to all browses.
    */
   public String ehKeyProcName = null;

   /**
    * Browse name.
    */
   public String ehKeyBrowseName = null;

   /**
    * Set of unique (browse-wide) column keys. For regular columns keys are
    * <code>buffer_alias.field_name</code>. For calculated and computed columns it is an
    * expression containing column parameters like column data type, format, initial value and
    * label.
    */
   public Set<String> ehColumnKeys;

   /**
    * Default constructor.
    */
   public EnhancedBrowseConfigKey()
   {
   }

   /**
    * Constructor.
    *
    * @param ehKeyProcName
    *        The name of the parent external procedure which contains the browse.
    *        <code>null</code> if the key is applied to all browses.
    * @param ehKeyBrowseName
    *        Browse name.
    * @param ehColumnKeys
    *        Set of column keys.
    */
   public EnhancedBrowseConfigKey(String ehKeyProcName,
                                  String ehKeyBrowseName,
                                  Set<String> ehColumnKeys)
   {
      this.ehKeyProcName = ehKeyProcName;
      this.ehKeyBrowseName = ehKeyBrowseName;
      this.ehColumnKeys = ehColumnKeys;
   }

   /**
    * Copy constructor.
    *
    * @param key
    *        Key to copy.
    */
   public EnhancedBrowseConfigKey(EnhancedBrowseConfigKey key)
   {
      ehKeyProcName = key.ehKeyProcName;
      ehKeyBrowseName = key.ehKeyBrowseName;

      if (key.ehColumnKeys != null)
      {
         ehColumnKeys = new HashSet<>(key.ehColumnKeys);
      }
   }

   /**
    * Checks that the key is valid. Note that this function returns <code>false</code> if this key
    *        is a key that references ALL browses.
    *
    * @return <code>true</code> if the key is valid. <code>false</code> if the key is invalid
    *         OR if it is a key that references ALL browses.
    */
   public boolean isValid()
   {
      if (ehKeyProcName == null || ehKeyProcName.isEmpty() || ehColumnKeys == null)
      {
         return false;
      }

      for (String key : ehColumnKeys)
      {
         if (key == null || key.isEmpty())
         {
            return false;
         }
      }

      return true;
   }

   /**
    * Compare two instances.
    *
    * @param    obj
    *           An object to compare with.
    *
    * @return   <code>true</code> if the two keys are identical.
    */
   @Override
   public boolean equals(Object obj)
   {
      if (this == obj)
      {
         return true;
      }
      if (obj == null || getClass() != obj.getClass())
      {
         return false;
      }

      EnhancedBrowseConfigKey key = (EnhancedBrowseConfigKey) obj;
      return Objects.equals(ehKeyProcName, key.ehKeyProcName) &&
             Objects.equals(ehKeyBrowseName, key.ehKeyBrowseName) &&
             Objects.equals(ehColumnKeys, key.ehColumnKeys);
   }

   /**
    * Return a hash code for this object.
    *
    * @return  hash code for this object.
    */
   @Override
   public int hashCode()
   {
      int result = 17;
      result = 37 * result + (ehKeyProcName == null ? 0 : ehKeyProcName.hashCode());
      result = 37 * result + (ehKeyBrowseName == null ? 0 : ehKeyBrowseName.hashCode());
      result = 37 * result + (ehColumnKeys == null ? 0 : ehColumnKeys.hashCode());

      return result;
   }
}
