/* 
** Module   : FillInConfig.java
** Abstract : stores FillIn specific state
**
** Copyright (c) 2005-2018, Golden Code Development Corporation.
**
** -#- -I- --Date-- --JPRM-- ----------------Description-----------------
** 001 EVL 20050725   @21810 Creation based on general concept.
** 002 SIY 20050831   @22502 Refactoring (see @@22394-22409). 
** 003 SIY 20050907   @22635 Fixed use of @inheritDoc tag in javadoc
** 004 SIY 20050913   @22838 Removed redundant disableAutoZap attribute,
**                           fixed handling of the cursorOffset, 
**                           moved getDataType() to superclass, cleanups.
** 005 SIY 20060209   @24467 Major refactoring.
** 006 SIY 20060324   @25281 Added isAutoReturn().
** 007 SIY 20060404   @25258 Support for the text group.
** 008 SIY 20060428   @25086 Moved some members to ControlTextConfig.
** 009 GES 20060803   @28406 Made externalizable for better performance.
**                           Many code standard cleanups.
** 010 SIY 20060926   @30023 Added getters for the next/previous ID's.
** 011 SIY 20100619          Switching to own implementation of character UI.
** 012 SIY 20100804          Changes in imports, some refactorings.
** 013 SVL 20101006          Added NATIVE attribute.
** 014 SIY 20111101          Moved from CHUI package. Minor 
**                           cleanups/refactorings in preparation to GUI client.
** 015 SIY 20111202          Updated imports.
** 016 CS  20121031          Removed client specific references
** 017 CA  20140926          Added shared widget configuration support, to allow GUI/ChUI
**                           concrete implementation for the same widget ID.  Refactored the
**                           widget configuration classes: all fields were made public; these
**                           classes need to be as dumb as possible, all logic related to
**                           setting/getting a certain field should be at the widget or at the
**                           caller. Refs #2254
** 018 CA  20141206          Added support for SELECTION-START/END, needed for GUI FILL-IN.
** 019 CA  20141220          Moved some client-side state common for both ChUI and GUI from the 
**                           FillIn class.   This exists only on client-side, and is not 
**                           transferred during widget config sync.
** 020 CA  20150105          Changes for SIDE-LABEL-HANDLE runtime support.
** 021 HC  20150323          Removed business logic from config classes.
** 022 SVL 20150605          Added zapState.
** 023 IAS 20150606          Added support for the DISABLE-AUTO-ZAP, moved text selection 
**                           related fields to the ControlConfig, removed obsolete zatState
** 024 CA  20150722          In LABEL case, the server and client-side configs are different 
**                           types (TextConfig vs LabelConfig) - so, applyConfig is protected to
**                           ignore instances which are not of this type.
** 025 IAS 20150801          Added PASSWORD-FIELD attribute  
** 026 IAS 20151214          LOAD-MOUSE-POINTER support.
** 027 SVL 20180529          Added enhanced font for working with enhanced editable browses.
*/
/*
** This program is free software: you can redistribute it and/or modify
** it under the terms of the GNU Affero General Public License as
** published by the Free Software Foundation, either version 3 of the
** License, or (at your option) any later version.
**
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
** GNU Affero General Public License for more details.
**
** You may find a copy of the GNU Affero GPL version 3 at the following
** location: https://www.gnu.org/licenses/agpl-3.0.en.html
** 
** Additional terms under GNU Affero GPL version 3 section 7:
** 
**   Under Section 7 of the GNU Affero GPL version 3, the following additional
**   terms apply to the works covered under the License.  These additional terms
**   are non-permissive additional terms allowed under Section 7 of the GNU
**   Affero GPL version 3 and may not be removed by you.
** 
**   0. Attribution Requirement.
** 
**     You must preserve all legal notices or author attributions in the covered
**     work or Appropriate Legal Notices displayed by works containing the covered
**     work.  You may not remove from the covered work any author or developer
**     credit already included within the covered work.
** 
**   1. No License To Use Trademarks.
** 
**     This license does not grant any license or rights to use the trademarks
**     Golden Code, FWD, any Golden Code or FWD logo, or any other trademarks
**     of Golden Code Development Corporation. You are not authorized to use the
**     name Golden Code, FWD, or the names of any author or contributor, for
**     publicity purposes without written authorization.
** 
**   2. No Misrepresentation of Affiliation.
** 
**     You may not represent yourself as Golden Code Development Corporation or FWD.
** 
**     You may not represent yourself for publicity purposes as associated with
**     Golden Code Development Corporation, FWD, or any author or contributor to
**     the covered work, without written authorization.
** 
**   3. No Misrepresentation of Source or Origin.
** 
**     You may not represent the covered work as solely your work.  All modified
**     versions of the covered work must be marked in a reasonable way to make it
**     clear that the modified work is not originating from Golden Code Development
**     Corporation or FWD.  All modified versions must contain the notices of
**     attribution required in this license.
*/

package com.goldencode.p2j.ui;

import java.io.*;

import com.goldencode.p2j.ui.client.*;
import com.goldencode.p2j.ui.client.format.*;
import com.goldencode.p2j.util.*;

/**
 * Contains FillIn specific widget state.
 */
public class FillInConfig
extends ControlTextConfig
{
   /** The Progress CURSOR-OFFSET attribute. */
   public int cursorOffset = 0;
   
   /** Determines if AUTO-ZAP is allowed. */
   public boolean autoZap = true;

   /** DISABLE-AUTO-ZAP attribute. */
   public boolean disableAutoZap = false;

   /** ID of previous widget in the group. */
   public int prevId = -1;

   /** ID of next widget in the group. */
   public int nextId = -1;

   /** The Progress DEBLANK attribute */
   public boolean deblank = false;

   /** The Progress AUTO-RETURN attribute. */
   public boolean autoReturn = false;
   
   /** Group mode flag. */
   public boolean grouping = false;

   /** The Progress NATIVE attribute. */
   public boolean nativ = false;
   
   /** The Progress PASSWORD-FIELD attribute. */
   public boolean passwordField = false;
   
   /** The custom mouse pointer */
   public String mousePointer = null;

   /** Enhanced font. */
   public FontDetails ehFont = null;

   // START_CLIENT_SIDE_MARKER
   // on client-side, following are fields containing shared state between different widget 
   // implementations (i.e. GUI and ChUI)
   /**
    * Internal presentation of displayed data. May be <code>null</code>
    * which means that value can't be represented with the current format.
    * Such value is rendered as several question marks and can only be reset
    * to a valid value by ZAP or CLEAR actions. At resetting corresponding
    * Presentation instance is generated and is assigned to this member.
    */
   public transient DisplayFormat.Presentation appDataPres;
   
   /** Instance of display format. */ 
   public transient DisplayFormat appFormat;

   /** Alternative variable to display. */
   public transient BaseDataType displayVar = null;
   
   /** Alternative format to display. */ 
   public transient DisplayFormat displayFormat = null;
   
   /** Alternative data presentation to display. */
   public transient DisplayFormat.Presentation displayDataPres = null;
   
   /** Widget in focus flag */
   public transient boolean widgetFocused = false;
   // END_CLIENT_SIDE_MARKER

   /**
    * Default constructor (only used in de-serialization).
    */
   public FillInConfig()
   {
   }

   /**
    * Create a new config and associate it with the given widget.
    * 
    * @param    id
    *           The ID of the widget to which this configuration belongs.
    */
   protected FillInConfig(int id)
   {
      super(id);
   }

   /**
    * Create a new config and associate it with the given widget.
    * 
    * @param    id
    *           The ID of the widget to which this configuration belongs.
    */
   public FillInConfig(WidgetId id)
   {
      super(id);
   }

   /**
    * Set new values from the provided configuration instance.
    * 
    * @param    config
    *           The instance from which to take values.
    */
   @Override
   public void applyConfig(WidgetConfig config)
   {
      super.applyConfig(config);
      
      if (!(config instanceof FillInConfig))
      {
         return;
      }

      FillInConfig cfg = (FillInConfig) config;
      
      cursorOffset = cfg.cursorOffset;
      autoZap = cfg.autoZap;
      disableAutoZap = cfg.disableAutoZap;
      prevId = cfg.prevId;
      nextId = cfg.nextId;
      deblank = cfg.deblank;
      autoReturn = cfg.autoReturn;
      grouping = cfg.grouping;
      nativ = cfg.nativ;
      passwordField = cfg.passwordField;
      mousePointer = cfg.mousePointer;
      ehFont = cfg.ehFont;
   }

   /**
    * Replacement for the default object reading method. The latest
    * state is read from the input source.  The parent's state is
    * read first.
    * 
    * @param    in
    *           The input source from which fields will be restored.
    *
    * @throws   IOException
    *           In case of I/O errors.
    * @throws   ClassNotFoundException
    *           If payload can't be instantiated.
    */
   @Override
   public void readExternal(ObjectInput in)
   throws IOException,
          ClassNotFoundException
   {
      super.readExternal(in);
      cursorOffset   = in.readInt();
      prevId         = in.readInt();
      nextId         = in.readInt();
      autoZap        = in.readBoolean();
      disableAutoZap = in.readBoolean();
      deblank        = in.readBoolean();
      autoReturn     = in.readBoolean();
      grouping       = in.readBoolean();
      nativ          = in.readBoolean();
      passwordField  = in.readBoolean();
      mousePointer   = (String) in.readObject();
      ehFont         = (FontDetails) in.readObject();
   }

   /**
    * Replacement for the default object writing method. The latest
    * state is written to the output destination.  The parent's state is
    * written first.
    * 
    * @param    out
    *           The output destination to which fields will be saved.
    *
    * @throws   IOException
    *           In case of I/O errors.
    */
   @Override
   public void writeExternal(ObjectOutput out)
   throws IOException
   {
      super.writeExternal(out);
      out.writeInt(cursorOffset);
      out.writeInt(prevId);
      out.writeInt(nextId);
      out.writeBoolean(autoZap);
      out.writeBoolean(disableAutoZap);
      out.writeBoolean(deblank);
      out.writeBoolean(autoReturn);
      out.writeBoolean(grouping);
      out.writeBoolean(nativ);
      out.writeBoolean(passwordField);
      out.writeObject(mousePointer);
      out.writeObject(ehFont);
  }
}