/*
** Module   : FormattedValue.java
** Abstract : provides a simple container for a value and an associated
**            format string
**
** Copyright (c) 2006-2017, Golden Code Development Corporation.
**
** -#- -I- --Date-- -T- --JPRM-- ----------------Description-----------------
** 001 GES 20060915 NEW   @29618 First version as a simple container.
** 002 NVS 20071018 CHG   @35481 Added tostring() method useful for debugging.
*/
/*
** This program is free software: you can redistribute it and/or modify
** it under the terms of the GNU Affero General Public License as
** published by the Free Software Foundation, either version 3 of the
** License, or (at your option) any later version.
**
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
** GNU Affero General Public License for more details.
**
** You may find a copy of the GNU Affero GPL version 3 at the following
** location: https://www.gnu.org/licenses/agpl-3.0.en.html
** 
** Additional terms under GNU Affero GPL version 3 section 7:
** 
**   Under Section 7 of the GNU Affero GPL version 3, the following additional
**   terms apply to the works covered under the License.  These additional terms
**   are non-permissive additional terms allowed under Section 7 of the GNU
**   Affero GPL version 3 and may not be removed by you.
** 
**   0. Attribution Requirement.
** 
**     You must preserve all legal notices or author attributions in the covered
**     work or Appropriate Legal Notices displayed by works containing the covered
**     work.  You may not remove from the covered work any author or developer
**     credit already included within the covered work.
** 
**   1. No License To Use Trademarks.
** 
**     This license does not grant any license or rights to use the trademarks
**     Golden Code, FWD, any Golden Code or FWD logo, or any other trademarks
**     of Golden Code Development Corporation. You are not authorized to use the
**     name Golden Code, FWD, or the names of any author or contributor, for
**     publicity purposes without written authorization.
** 
**   2. No Misrepresentation of Affiliation.
** 
**     You may not represent yourself as Golden Code Development Corporation or FWD.
** 
**     You may not represent yourself for publicity purposes as associated with
**     Golden Code Development Corporation, FWD, or any author or contributor to
**     the covered work, without written authorization.
** 
**   3. No Misrepresentation of Source or Origin.
** 
**     You may not represent the covered work as solely your work.  All modified
**     versions of the covered work must be marked in a reasonable way to make it
**     clear that the modified work is not originating from Golden Code Development
**     Corporation or FWD.  All modified versions must contain the notices of
**     attribution required in this license.
*/

package com.goldencode.p2j.ui;

import java.io.*;
import com.goldencode.p2j.util.*;

/**
 * A simple container for a value and an associated format string.
 */
public class FormattedValue
implements Externalizable
{
   /** The data value. */
   private BaseDataType value = null;
   
   /** The value's associated format string. */
   private String fmt = null;
   
   /**
    * Default constructor, only used for <code>Externalizable</code>.
    */
   public FormattedValue()
   {
   }
   
   /**
    * Construct a new instance.
    *
    * @param    value
    *           The data value for this instance.
    * @param    fmt
    *           An optional format string to use with this value.
    */
   public FormattedValue(BaseDataType value, String fmt)
   {
      this.value = value;
      this.fmt   = fmt;
   }
   
   /**
    * Get the data value.
    * 
    * @return   The data value for this instance.
    */
   public BaseDataType getValue()
   {
      return value;
   }
   
   /**
    * Get the format string.
    * 
    * @return   The format string for this instance.
    */
   public String getFormat()
   {
      return fmt;
   }
   
   /**
    * 
    * @see java.lang.Object#toString()
    */
   public String toString()
   {
      return "Value "  + value.toString() + " " + 
             "Format " + fmt;
   }

   /**
    * Replacement for the default object reading method.
    * 
    * @param    in
    *           Input source from which fields will be restored.
    *
    * @throws   IOException
    *           In case of I/O errors.
    * @throws   ClassNotFoundException
    *           If payload can't be instantiated.
    */
   public void readExternal(ObjectInput in)
   throws IOException,
          ClassNotFoundException
   {
      value = (BaseDataType) in.readObject();
      fmt   = (String) in.readObject();
       
      if (fmt.length() == 0)
      {
         fmt = null;
      }
   }

   /**
    * Replacement for the default object writing method.
    * 
    * @param    out
    *           The output destination to which fields will be saved.
    *
    * @throws   IOException
    *           In case of I/O errors.
    */
   public void writeExternal(ObjectOutput out)
   throws IOException
   {
      out.writeObject(value);
      out.writeObject(fmt == null ? "" : fmt);
   }
}