/*
** Module   : HelpService.java
** Abstract : Help management class.
**
** Copyright (c) 2017, Golden Code Development Corporation.
**
** -#- -I- --Date-- --------------------------------Description--------------------------------
** 001 OM  20170721 Initial commit with stubbed methods.
*/

/*
** This program is free software: you can redistribute it and/or modify
** it under the terms of the GNU Affero General Public License as
** published by the Free Software Foundation, either version 3 of the
** License, or (at your option) any later version.
**
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
** GNU Affero General Public License for more details.
**
** You may find a copy of the GNU Affero GPL version 3 at the following
** location: https://www.gnu.org/licenses/agpl-3.0.en.html
** 
** Additional terms under GNU Affero GPL version 3 section 7:
** 
**   Under Section 7 of the GNU Affero GPL version 3, the following additional
**   terms apply to the works covered under the License.  These additional terms
**   are non-permissive additional terms allowed under Section 7 of the GNU
**   Affero GPL version 3 and may not be removed by you.
** 
**   0. Attribution Requirement.
** 
**     You must preserve all legal notices or author attributions in the covered
**     work or Appropriate Legal Notices displayed by works containing the covered
**     work.  You may not remove from the covered work any author or developer
**     credit already included within the covered work.
** 
**   1. No License To Use Trademarks.
** 
**     This license does not grant any license or rights to use the trademarks
**     Golden Code, FWD, any Golden Code or FWD logo, or any other trademarks
**     of Golden Code Development Corporation. You are not authorized to use the
**     name Golden Code, FWD, or the names of any author or contributor, for
**     publicity purposes without written authorization.
** 
**   2. No Misrepresentation of Affiliation.
** 
**     You may not represent yourself as Golden Code Development Corporation or FWD.
** 
**     You may not represent yourself for publicity purposes as associated with
**     Golden Code Development Corporation, FWD, or any author or contributor to
**     the covered work, without written authorization.
** 
**   3. No Misrepresentation of Source or Origin.
** 
**     You may not represent the covered work as solely your work.  All modified
**     versions of the covered work must be marked in a reasonable way to make it
**     clear that the modified work is not originating from Golden Code Development
**     Corporation or FWD.  All modified versions must contain the notices of
**     attribution required in this license.
*/

package com.goldencode.p2j.ui;

import com.goldencode.p2j.util.*;

/**
 * A class that collects option information about a SYSTEM-HELP statement using chained
 * methods.
 * <p>
 * Each public method emulates an option of the statement. They do a quick validation. If
 * successful they return a reference to this object to allow other method to be chained.
 * <p>
 * The {@code execute()} method does the final checks, send the request with gathered options
 * to client.
 */
public class HelpService
{
   /** The help system primitive to be executed. */
   private SystemHelpAction action = null;
   
   /** The list of specific parameters to current action (optional). */
   private Object[] specificParams = null;
   
   /**
    * Specifies the pathname of a help file. On ABL:
    * <ul>
    *    <li>if the file has a .chm extension (the extension for compiled Microsoft HTML Help
    *       files), the Microsoft HTML Help viewer is launched; 
    *    <li>if the file has a .hlp file extension, the Microsoft Windows Help viewer is launched.
    * </ul>
    * TODO: We will decide in #1827 how this will be implemented in FWD.  
    */
   private String helpFile = null;
   
   /**
    * The window-name parameter represents the primary or secondary window name as defined in the
    * [WINDOWS] section of the help project file. If the window name is omitted, or if "main" is
    * specified, the primary help window is used.
    */
   private String windowName = null;
   
   /** The flag for {@code CONTENTS} option. */ 
   private boolean contents = false;
   
   /**
    * Flag that {@code CONTEXT} mode was configured. In this mode, the statements displays the
    * help topic that the context number identifies. The context numbers are defined in the [MAP]
    * section of the help project file. The int-expr parameter is the context number for the help
    * topic.
    */
   private int context = -1;
   
   /**
    * The {@code HELP-TOPIC} option active if this is not null. The value that indicates the
    * topic (.htm/.html file) within the compiled Microsoft HTML Help (.chm) file to display.
    */
   private String topic = null;
   
   /** The keyword for the desired help topic. */
   private String key = null;
   
   /** The alternative keyword for the desired help topic. */
   private String altKey = null;
   
   /** The help macro to execute. */
   private String command = null;
   
   /** The partial key for the desired help topic in a {@code PARTIAL-KEY} statement. */
   private String partialKey = null;
   
   /**
    * The single character keyword table identifier for the required table in a 
    * {@code MULTIPLE-KEY} help statement.
    */
   private String tableId = null;
   
   /**
    * The keyword that is located in the keyword table in a {@code MULTIPLE-KEY} help statement.
    */
   private String multipleKey = null;
   
   /** Flags the help application that help is no longer required. */
   private boolean quit = false;
   
   /** Flags the {@code FINDER} option is activated */
   private boolean finder = false;
   
   /** Flags the {@code FORCE-FILE} option is activated */
   private boolean forceFile = false;
   
   /** Flags the {@code HELP} option is activated */
   private boolean help = false;
   
   /** The context number for the help topic in a {@code CONTEXT-POPUP} help statement. */
   private int contextPopup = -1;
   
   /** The context number for the new contents help topic in a {@code SET-CONTENTS} help stmt. */
   private int contentTopic = -1;
   
   /** Flag to maximize an existing (already opened) help window. */
   private boolean maximized = false;
   
   /** The X coordinate for the help window. */
   private int x = -1;
   
   /** The Y coordinate for the help window. */
   private int y = -1;
   
   /** The WIDTH of the help window. */
   private int width = -1;
   
   /** The HEIGHT of the help window. */
   private int height = -1;
   
   /**
    * Private constructor. Use the {@link #prepare} factory method to create an object.
    */
   private HelpService()
   {
   }
   
   /**
    * Prepares a new configuration based on {@code helpFile}.
    *
    * @param   helpFile
    *          The pathname of a help file
    *
    * @return  The reference to a new {@code HelpService}. Used for chaining.
    */
   public static HelpService prepare(String helpFile)
   {
      HelpService dialog = new HelpService();
      dialog.helpFile = helpFile;
      return dialog;
   }
   
   /**
    * Prepares a new configuration based on {@code helpFile}.
    * 
    * @param   helpFile
    *          The pathname of a help file
    *
    * @return  The reference to a new {@code HelpService}. Used for chaining.
    */
   public static HelpService prepare(Text helpFile)
   {
      return prepare(helpFile.getValue());
   }
   
   /**
    * The chain-terminator method. Evaluates all the options and issues commands to client to
    * show on help window.
    */
   public void execute()
   {
      if (action == null)
      {
         System.err.println("HELP primitive not specified!");
         return;
      }
      
      // TODO: sent a message to client to show the help file using configured options
      LogicalTerminal.getClient().systemHelp(helpFile, windowName, action, specificParams);
   }
   
   /**
    * Configures the name of the  to be created as child of a certain windows.
    *
    * @param   windowName
    *          The help window name.
    *
    * @return  The reference to current {@code HelpService}. Used for chaining.
    *
    * @see     #windowName
    */
   public HelpService setWindowName(Text windowName)
   {
      if (windowName.isUnknown())
      {
         String msg = "Could not extract argument for SYSTEM-HELP statement";
         ErrorManager.recordOrThrowError(3218, msg, false);
      }
      else
      {
         this.windowName = windowName.toJavaType();
      }
      return this;
   }
   
   /**
    * Configures the name of the  to be created as child of a certain windows.
    *
    * @param   windowName
    *          The help window name.
    *
    * @return  The reference to current {@code HelpService}. Used for chaining.
    *
    * @see     #windowName
    */
   public HelpService setWindowName(String windowName)
   {
      this.windowName = windowName;
      return this;
   }
   
   /**
    * Activates the {@code CONTENTS} mode. 
    *
    * @return  The reference to current {@code HelpService}. Used for chaining.
    *
    * @see     #contents
    */
   public HelpService contents()
   {
      setAction(SystemHelpAction.CONTENTS);
      this.contents = true;
      return this;
   }
   
   /**
    * Activates the {@code HELP-TOPIC} mode. 
    *
    * @param   topic
    *          The help window name.
    *          
    * @return  The reference to current {@code HelpService}. Used for chaining.
    *
    * @see     #topic
    */
   public HelpService setTopic(String topic)
   {
      setAction(SystemHelpAction.TOPIC);
      this.topic = topic;
      specificParams = new String[] { this.topic };
      return this;
   }
   
   /**
    * Activates the {@code HELP-TOPIC} mode. 
    *
    * @param   topic
    *          The help window name.
    *
    * @return  The reference to current {@code HelpService}. Used for chaining.
    *
    * @see     #topic
    */
   public HelpService setTopic(Text topic)
   {
      setAction(SystemHelpAction.TOPIC);
      if (topic.isUnknown())
      {
         String msg = "Could not extract argument for SYSTEM-HELP statement";
         ErrorManager.recordOrThrowError(3218, msg, false);
      }
      else
      {
         this.topic = topic.getValue();
      }
      specificParams = new String[] { this.topic };
      return this;
   }
   
   /**
    * Activates the {@code CONTEXT} mode. 
    *
    * @param   context
    *          The context number for the help topic. Displays the help topic that the context
    *          number identifies. You define context numbers in the [MAP] section of the help
    *          project file.
    *
    * @return  The reference to current {@code HelpService}. Used for chaining.
    *
    * @see     #context
    */
   public HelpService setContext(long context)
   {
      setAction(SystemHelpAction.CONTEXT);
      this.context = (int) context;
      specificParams = new Integer[] { this.context };
      return this;
   }
   
   /**
    * Activates the {@code CONTEXT} mode. 
    *
    * @param   context
    *          The context number for the help topic. Displays the help topic that the context
    *          number identifies. You define context numbers in the [MAP] section of the help
    *          project file.
    *
    * @return  The reference to current {@code HelpService}. Used for chaining.
    *
    * @see     #context
    */
   public HelpService setContext(NumberType context)
   {
      setAction(SystemHelpAction.CONTEXT);
      if (context.isUnknown())
      {
         String msg = "Could not extract argument for SYSTEM-HELP statement";
         ErrorManager.recordOrThrowError(3218, msg, false);
      }
      else
      {
         this.context = context.intValue();
      }
      specificParams = new Integer[] { this.context };
      return this;
   }
   
   /**
    * Activates the {@code KEY} mode.
    * <p>
    * For HTML Help, this option displays the topic matching the string found in the keyword
    * index. Use semicolons in the string parameter to delimit multiple keywords. If no match is
    * found, Microsoft HTML Help displays the help viewer with the Index tab on top.
    * <p>
    * For Windows Help, this option displays the help topic matching the string found in the index
    * keyword list. If there is more than one match, it displays the first topic containing the
    * keyword. If there is no match or the string is omitted, a message is displayed indicating
    * that the keyword is invalid.
    *
    * @param   key
    *          The keyword for the desired help topic.
    *
    * @return  The reference to current {@code HelpService}. Used for chaining.
    *
    * @see     #key
    */
   public HelpService setKey(String key)
   {
      setAction(SystemHelpAction.KEY);
      this.key = key;
      specificParams = new String[] { this.key };
      return this;
   }
   
   /**
    * Activates the {@code KEY} mode.
    * <p>
    * For HTML Help, this option displays the topic matching the string found in the keyword
    * index. Use semicolons in the string parameter to delimit multiple keywords. If no match is
    * found, Microsoft HTML Help displays the help viewer with the Index tab on top.
    * <p>
    * For Windows Help, this option displays the help topic matching the string found in the index
    * keyword list. If there is more than one match, it displays the first topic containing the
    * keyword. If there is no match or the string is omitted, a message is displayed indicating
    * that the keyword is invalid.
    *
    * @param   key
    *          The keyword for the desired help topic.
    *
    * @return  The reference to current {@code HelpService}. Used for chaining.
    *
    * @see     #key
    */
   public HelpService setKey(Text key)
   {
      setAction(SystemHelpAction.KEY);
      this.key = key.getValue();
      specificParams = new String[] { this.key };
      return this;
   }
   
   /**
    * Activates the {@code ALTERNATE-KEY} mode. Displays a help topic matching the string found in
    * the alternate keyword (Alink) index.
    *
    * @param   altKey
    *          The keyword in the alternate keyword index.
    *
    * @return  The reference to current {@code HelpService}. Used for chaining.
    *
    * @see     #altKey
    */
   public HelpService setAlternateKey(String altKey)
   {
      setAction(SystemHelpAction.ALTERNATE_KEY);
      this.altKey = altKey;
      specificParams = new String[] { this.altKey };
      return this;
   }
   
   /**
    * Activates the {@code ALTERNATE-KEY} mode. Displays a help topic matching the string found in
    * the alternate keyword (Alink) index.
    *
    * @param   altKey
    *          The keyword in the alternate keyword index.
    *
    * @return  The reference to current {@code HelpService}. Used for chaining.
    *
    * @see     #altKey
    */
   public HelpService setAlternateKey(Text altKey)
   {
      setAction(SystemHelpAction.ALTERNATE_KEY);
      this.altKey = altKey.getValue();
      specificParams = new String[] { this.altKey };
      return this;
   }
   
   /**
    * Activates the {@code PARTIAL-KEY} mode. Displays the help topic matching the string found
    * in the keyword list. In Windows, if there is more than one match, no match, or if the string
    * is omitted, it displays the Help Topics: Window Help Topics dialog box with the Index tab on
    * top.
    *
    * @param   partialKey
    *          The partial key for the desired help topic.
    *
    * @return  The reference to current {@code HelpService}. Used for chaining.
    *
    * @see     #partialKey
    */
   public HelpService setPartialKey(String partialKey)
   {
      setAction(SystemHelpAction.PARTIAL_KEY);
      this.partialKey = partialKey;
      specificParams = new String[] { this.partialKey };
      return this;
   }
   
   /**
    * Activates the {@code PARTIAL-KEY} mode. Displays the help topic matching the string found
    * in the keyword list. In Windows, if there is more than one match, no match, or if the string
    * is omitted, it displays the Help Topics: Window Help Topics dialog box with the Index tab on
    * top.
    *
    * @param   partialKey
    *          The partial key for the desired help topic.
    *
    * @return  The reference to current {@code HelpService}. Used for chaining.
    *
    * @see     #partialKey
    */
   public HelpService setPartialKey(Text partialKey)
   {
      setAction(SystemHelpAction.PARTIAL_KEY);
      this.partialKey = partialKey.getValue();
      specificParams = new String[] { this.partialKey };
      return this;
   }
   
   /**
    * Activates the {@code MULTIPLE-KEY} mode. Displays the help topic matching a keyword from an
    * alternate keyword table.
    *
    * @param   tableId
    *          The single character keyword table identifier for the required table.
    * @param   multipleKey
    *          The keyword that is located in the keyword table.
    *
    * @return  The reference to current {@code HelpService}. Used for chaining.
    *
    * @see     #tableId
    * @see     #multipleKey
    */
   public HelpService setMultipleKey(String tableId, String multipleKey)
   {
      setAction(SystemHelpAction.MULTIPLE_KEY);
      this.tableId = tableId;
      this.multipleKey = multipleKey;
      specificParams = new String[] { this.tableId, this.multipleKey };
      return this;
   }
   
   /**
    * Activates the {@code MULTIPLE-KEY} mode. Displays the help topic matching a keyword from an
    * alternate keyword table.
    *
    * @param   tableId
    *          The single character keyword table identifier for the required table.
    * @param   multipleKey
    *          The keyword that is located in the keyword table.
    *
    * @return  The reference to current {@code HelpService}. Used for chaining.
    *
    * @see     #tableId
    * @see     #multipleKey
    */
   public HelpService setMultipleKey(Text tableId, String multipleKey)
   {
      setAction(SystemHelpAction.MULTIPLE_KEY);
      this.tableId = tableId.getValue();
      this.multipleKey = multipleKey;
      specificParams = new String[] { this.tableId, this.multipleKey };
      return this;
   }
   
   /**
    * Activates the {@code MULTIPLE-KEY} mode. Displays the help topic matching a keyword from an
    * alternate keyword table.
    *
    * @param   tableId
    *          The single character keyword table identifier for the required table.
    * @param   multipleKey
    *          The keyword that is located in the keyword table.
    *
    * @return  The reference to current {@code HelpService}. Used for chaining.
    *
    * @see     #tableId
    * @see     #multipleKey
    */
   public HelpService setMultipleKey(String tableId, Text multipleKey)
   {
      setAction(SystemHelpAction.MULTIPLE_KEY);
      if (multipleKey.isUnknown())
      {
         String msg = "Could not extract argument for SYSTEM-HELP statement";
         ErrorManager.recordOrThrowError(3218, msg, false);
      }
      else
      {
         this.multipleKey = multipleKey.getValue();
      }
      this.tableId = tableId;
      specificParams = new String[] { this.tableId, this.multipleKey };
      return this;
   }
   
   /**
    * Activates the {@code MULTIPLE-KEY} mode. Displays the help topic matching a keyword from an
    * alternate keyword table.
    *
    * @param   tableId
    *          The single character keyword table identifier for the required table.
    * @param   multipleKey
    *          The keyword that is located in the keyword table.
    *
    * @return  The reference to current {@code HelpService}. Used for chaining.
    *
    * @see     #tableId
    * @see     #multipleKey
    */
   public HelpService setMultipleKey(Text tableId, Text multipleKey)
   {
      setAction(SystemHelpAction.MULTIPLE_KEY);
      if (tableId.isUnknown())
      {
         String msg = "Could not extract argument for SYSTEM-HELP statement";
         ErrorManager.recordOrThrowError(3218, msg, false);
      }
      else
      {
         this.tableId = tableId.getValue();
      }
   
      if (multipleKey.isUnknown())
      {
         String msg = "Could not extract argument for SYSTEM-HELP statement";
         ErrorManager.recordOrThrowError(3218, msg, false);
      }
      else
      {
         this.multipleKey = multipleKey.getValue();
      }
      specificParams = new String[] { this.tableId, this.multipleKey };
      return this;
   }
   
   /**
    * Activates the {@code COMMAND} mode. Executes a help macro 
    *
    * @param   command
    *          The command to be executed.
    *
    * @return  The reference to current {@code HelpService}. Used for chaining.
    *
    * @see     #command
    */
   public HelpService setCommand(String command)
   {
      setAction(SystemHelpAction.COMMAND);
      this.command = command;
      specificParams = new String[] { this.command };
      return this;
   }
   
   /**
    * Activates the {@code COMMAND} mode. Executes a help macro 
    *
    * @param   command
    *          The command to be executed.
    *
    * @return  The reference to current {@code HelpService}. Used for chaining.
    *
    * @see     #command
    */
   public HelpService setCommand(Text command)
   {
      setAction(SystemHelpAction.COMMAND);
      if (command.isUnknown())
      {
         String msg = "Could not extract argument for SYSTEM-HELP statement";
         ErrorManager.recordOrThrowError(3218, msg, false);
      }
      else
      {
         this.command = command.getValue();
      }
      specificParams = new String[] { this.command };
      return this;
   }
   
   /**
    * Activates the {@code QUIT} mode. Informs the help application that help is no longer
    * required. The system closes the help window. 
    *
    * @return  The reference to current {@code HelpService}. Used for chaining.
    *
    * @see     #quit
    */
   public HelpService quit()
   {
      setAction(SystemHelpAction.QUIT);
      this.quit = true;
      return this;
   }
   
   /**
    * Activates the {@code FINDER} mode. Displays the Help Topics: Windows Help Topics dialog box,
    * which contains an Index tab, a Find tab, and optionally a Contents tab, with the most
    * recently used tab displayed on top. If a Contents tab file (.CNT file) is present when the 
    * user initially call the Help Topics: Windows Help dialog box, then the Content tab displays
    * on top. However, if a .CNT file is not present, then the dialog box displays with the Index
    * tab on top; the Contents tab is not available.
    *
    * @return  The reference to current {@code HelpService}. Used for chaining.
    *
    * @see     #finder
    */
   public HelpService finder()
   {
      setAction(SystemHelpAction.FINDER);
      this.finder = true;
      return this;
   }
   
   /**
    * Activates the {@code FORCE-FILE} mode: Ensures that the correct help file is open and
    * displayed.
    *
    * @return  The reference to current {@code HelpService}. Used for chaining.
    *
    * @see     #forceFile
    */
   public HelpService forceFile()
   {
      setAction(SystemHelpAction.FORCE_FILE);
      this.forceFile = true; 
      return this;
   }
   
   /**
    * Activates the {@code HELP} mode. In this mode, teh statement displays the contents of the
    * Help-on-Help file. In Windows, HELP displays the Help Topics: Windows Help Topics dialog
    * box.
    *
    * @return  The reference to current {@code HelpService}. Used for chaining.
    *
    * @see     #help
    */
   public HelpService help()
   {
      setAction(SystemHelpAction.HELP);
      this.help = true;
      return this;
   }
   
   /**
    * Activates the {@code CONTEXT-POPUP} mode. In this mode, when executed, it displays the help
    * topic in a pop-up window that the context number identifies. The context numbers are 
    * defined in the [MAP] section of the help project file. If a non-scrolling region exists in
    * a help topic, only that region displays when you use the CONTEXT-POPUP option to display
    * the topic. 
    *
    * @param   contextPopup
    *          The context number for the help topic.
    *
    * @return  The reference to current {@code HelpService}. Used for chaining.
    *
    * @see     #contextPopup
    */
   public HelpService setContextPopup(NumberType contextPopup)
   {
      setAction(SystemHelpAction.POPUP);
      if (contextPopup.isUnknown())
      {
         String msg = "Could not extract argument for SYSTEM-HELP statement";
         ErrorManager.recordOrThrowError(3218, msg, false);
      }
      else
      {
         this.contextPopup = contextPopup.intValue();
      }
      specificParams = new Integer[] { this.contextPopup };
      return this;
   }
   
   /**
    * Activates the {@code CONTEXT-POPUP} mode. In this mode, when executed, it displays the help
    * topic in a pop-up window that the context number identifies. The context numbers are 
    * defined in the [MAP] section of the help project file. If a non-scrolling region exists in
    * a help topic, only that region displays when you use the CONTEXT-POPUP option to display
    * the topic. 
    *
    * @param   contextPopup
    *          The context number for the help topic.
    *
    * @return  The reference to current {@code HelpService}. Used for chaining.
    *
    * @see     #contextPopup
    */
   public HelpService setContextPopup(long contextPopup)
   {
      setAction(SystemHelpAction.POPUP);
      this.contextPopup = (int) contextPopup;
      specificParams = new Integer[] { this.contextPopup };
      return this;
   }
   
   /**
    * Activates the {@code POSITION MAXIMIZE} mode, when executed, it maximizes an existing 
    * (already opened) help window.
    * 
    * @return  The reference to current {@code HelpService}. Used for chaining.
    *
    * @see     #maximized
    */
   public HelpService maximized()
   {
      setAction(SystemHelpAction.MAXIMIZE);
      this.maximized = true;
      return this;
   }
   
   /**
    * Activates the {@code SET-CONTENTS} mode.
    * <p>
    * Dynamically re-maps the contents help topic from what is defined in the [OPTIONS] section
    * of the help project file. When a CONTENTS call is made, the new contents help topic is
    * displayed.
    *
    * @param   contents
    *          The help window name.
    *
    * @return  The reference to current {@code HelpService}. Used for chaining.
    *
    * @see     #contentTopic
    */
   public HelpService setContents(long contents)
   {
      setAction(SystemHelpAction.SET_CONTENTS);
      this.contentTopic = (int) contents;
      specificParams = new Integer[] { this.contentTopic };
      return this;
   }
   
   /**
    * Activates the {@code SET-CONTENTS} mode.
    * <p>
    * Dynamically re-maps the contents help topic from what is defined in the [OPTIONS] section
    * of the help project file. When a CONTENTS call is made, the new contents help topic is
    * displayed.
    *
    * @param   contents
    *          The help window name.
    *
    * @return  The reference to current {@code HelpService}. Used for chaining.
    *
    * @see     #contentTopic
    */
   public HelpService setContents(NumberType contents)
   {
      setAction(SystemHelpAction.SET_CONTENTS);
      if (contents.isUnknown())
      {
         String msg = "Could not extract argument for SYSTEM-HELP statement";
         ErrorManager.recordOrThrowError(3218, msg, false);
      }
      else
      {
         this.contentTopic = contents.intValue();
      }
      specificParams = new Integer[] { this.contentTopic };
      return this;
   }
   
   /**
    * Sets the new {@code POSITION} and geometry to help window. 
    *
    * @param   x
    *          The new X coordinate.
    * @param   y
    *          The new Y coordinate.
    * @param   width
    *          The new width of the window.
    * @param   height
    *          The new height of the window.
    *
    * @return  The reference to current {@code HelpService}. Used for chaining.
    *
    * @see     #x
    * @see     #y
    * @see     #height
    * @see     #width
    */
   public HelpService setPosition(NumberType x, NumberType y, NumberType width, NumberType height)
   {
      setAction(SystemHelpAction.POSITION);
      this.x = x.intValue();
      this.y = y.intValue();
      this.width = width.intValue();
      this.height = height.intValue();
      specificParams = new Integer[] { this.x, this.y, this.width, this.height  };
      return this;
   }
   
   /**
    * Sets the new {@code POSITION} and geometry to help window. 
    *
    * @param   x
    *          The new X coordinate.
    * @param   y
    *          The new Y coordinate.
    * @param   width
    *          The new width of the window.
    * @param   height
    *          The new height of the window.
    *
    * @return  The reference to current {@code HelpService}. Used for chaining.
    *
    * @see     #x
    * @see     #y
    * @see     #height
    * @see     #width
    */
   public HelpService setPosition(long x, long y, long width, long height)
   {
      setAction(SystemHelpAction.POSITION);
      this.x = (int) x;
      this.y = (int) y;
      this.width = (int) width;
      this.height = (int) height;
      specificParams = new Integer[] { this.x, this.y, this.width, this.height  };
      return this;
   }
   
   /**
    * Validates and sets the current action to be executed.
    * 
    * @param   a
    *          The action to be executed.
    */
   private void setAction(SystemHelpAction a)
   {
      if (action != null)
      {
         System.err.println("ABL does not allow multiple help actions to be executed in a " +
                            "single batch.");
      }
      
      action = a;
   }
}
