/*
** Module   : JastValidationExpr.java
** Abstract : Validation expression backed by a Java AST which is interpreted to run validation.
**
** Copyright (c) 2017, Golden Code Development Corporation.
**
** -#- -I- --Date-- ---------------------------------Description----------------------------------
** 001 SVL 20171024 Created initial version.
*/
/*
** This program is free software: you can redistribute it and/or modify
** it under the terms of the GNU Affero General Public License as
** published by the Free Software Foundation, either version 3 of the
** License, or (at your option) any later version.
**
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
** GNU Affero General Public License for more details.
**
** You may find a copy of the GNU Affero GPL version 3 at the following
** location: https://www.gnu.org/licenses/agpl-3.0.en.html
**
** Additional terms under GNU Affero GPL version 3 section 7:
**
**   Under Section 7 of the GNU Affero GPL version 3, the following additional
**   terms apply to the works covered under the License.  These additional terms
**   are non-permissive additional terms allowed under Section 7 of the GNU
**   Affero GPL version 3 and may not be removed by you.
**
**   0. Attribution Requirement.
**
**     You must preserve all legal notices or author attributions in the covered
**     work or Appropriate Legal Notices displayed by works containing the covered
**     work.  You may not remove from the covered work any author or developer
**     credit already included within the covered work.
**
**   1. No License To Use Trademarks.
**
**     This license does not grant any license or rights to use the trademarks
**     Golden Code, FWD, any Golden Code or FWD logo, or any other trademarks
**     of Golden Code Development Corporation. You are not authorized to use the
**     name Golden Code, FWD, or the names of any author or contributor, for
**     publicity purposes without written authorization.
**
**   2. No Misrepresentation of Affiliation.
**
**     You may not represent yourself as Golden Code Development Corporation or FWD.
**
**     You may not represent yourself for publicity purposes as associated with
**     Golden Code Development Corporation, FWD, or any author or contributor to
**     the covered work, without written authorization.
**
**   3. No Misrepresentation of Source or Origin.
**
**     You may not represent the covered work as solely your work.  All modified
**     versions of the covered work must be marked in a reasonable way to make it
**     clear that the modified work is not originating from Golden Code Development
**     Corporation or FWD.  All modified versions must contain the notices of
**     attribution required in this license.
*/

package com.goldencode.p2j.ui;

import com.goldencode.p2j.persist.*;
import com.goldencode.p2j.uast.*;
import com.goldencode.p2j.util.*;
import com.goldencode.p2j.util.ErrorManager;

import java.util.*;
import java.util.logging.*;

/**
 * Validation expression backed by a Java AST which is interpreted to run validation.
 */
public class JastValidationExpr
implements ValidationExpr<BaseDataType>
{
   /** Logger. */
   private static final Logger LOG = LogHelper.getLogger(JastValidationExpr.class.getName());

   /** Interpreter for running Java AST. */
   private final RuntimeJastInterpreter interpreter;

   /** Java AST representing validation expression. */
   private final JavaAst javaAst;

   /**
    * Default constructor.
    *
    * @param javaAst
    *        Java AST representing validation expression.
    * @param buffers
    *        Buffers of the query which iterates the validated buffer.
    */
   public JastValidationExpr(JavaAst javaAst, List<Buffer> buffers)
   {
      this.javaAst = javaAst;
      interpreter = new RuntimeJastInterpreter(buffers, null);
      interpreter.prepare(javaAst);
   }

   /**
    * Evaluate the validation expression and return the calculated result.
    *
    * @param    value
    *           The new widget value to validate.
    *
    * @return   The result of the validation expression evaluation. Must NOT be
    *           <code>null</code>.
    */
   @Override
   public logical validate(BaseDataType value)
   {
      try
      {
         Map<String, Object> params = new HashMap<>(1);
         params.put(DynamicValidationHelper.VAL_EXPR_VAR_NAME, value);
         interpreter.setVariables(params);
         return (logical) interpreter.interpret("execute");
      }
      catch (Throwable t)
      {
         // Look for ErrorConditionException
         Throwable cause;
         Throwable error = t;
         while ((cause = error.getCause()) != null  && (error != cause))
         {
            error = cause;

            if (error instanceof ErrorConditionException)
            {
               ErrorConditionException e = (ErrorConditionException) error;
               ErrorManager.recordOrShowError(e.getProgressErrorCode(), e.getMessage(), false);
               return new logical(false);
            }
         }

         if (LOG.isLoggable(Level.SEVERE))
         {
            LOG.log(Level.SEVERE,
                  "Failed to interpret JAST " + javaAst.dumpTree(true), t);
         }

         ErrorManager.recordOrShowError(-1,
                                        "Failed to interpret validation expression's JAST",
                                        false);
         return new logical(false);
      }
   }
}
