/*
** Module   : LogicalTerminalConversionParams.java
** Abstract : Server-side resolution of coordinate conversion parameters.
**
** Copyright (c) 2014-2017, Golden Code Development Corporation.
**
** -#- -I- --Date-- ---------------------------------Description---------------------------------
** 001 HC  20141021 Created initial version.
** 002 HC  20141207 Implemented runtime support for several SESSION attributes.
*/
/*
** This program is free software: you can redistribute it and/or modify
** it under the terms of the GNU Affero General Public License as
** published by the Free Software Foundation, either version 3 of the
** License, or (at your option) any later version.
**
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
** GNU Affero General Public License for more details.
**
** You may find a copy of the GNU Affero GPL version 3 at the following
** location: https://www.gnu.org/licenses/agpl-3.0.en.html
** 
** Additional terms under GNU Affero GPL version 3 section 7:
** 
**   Under Section 7 of the GNU Affero GPL version 3, the following additional
**   terms apply to the works covered under the License.  These additional terms
**   are non-permissive additional terms allowed under Section 7 of the GNU
**   Affero GPL version 3 and may not be removed by you.
** 
**   0. Attribution Requirement.
** 
**     You must preserve all legal notices or author attributions in the covered
**     work or Appropriate Legal Notices displayed by works containing the covered
**     work.  You may not remove from the covered work any author or developer
**     credit already included within the covered work.
** 
**   1. No License To Use Trademarks.
** 
**     This license does not grant any license or rights to use the trademarks
**     Golden Code, FWD, any Golden Code or FWD logo, or any other trademarks
**     of Golden Code Development Corporation. You are not authorized to use the
**     name Golden Code, FWD, or the names of any author or contributor, for
**     publicity purposes without written authorization.
** 
**   2. No Misrepresentation of Affiliation.
** 
**     You may not represent yourself as Golden Code Development Corporation or FWD.
** 
**     You may not represent yourself for publicity purposes as associated with
**     Golden Code Development Corporation, FWD, or any author or contributor to
**     the covered work, without written authorization.
** 
**   3. No Misrepresentation of Source or Origin.
** 
**     You may not represent the covered work as solely your work.  All modified
**     versions of the covered work must be marked in a reasonable way to make it
**     clear that the modified work is not originating from Golden Code Development
**     Corporation or FWD.  All modified versions must contain the notices of
**     attribution required in this license.
*/
package com.goldencode.p2j.ui;

import com.goldencode.p2j.security.*;

/**
 * The class resolves coordinate conversion parameters based on the static methods
 * of {@link LogicalTerminal}.
 */
class LogicalTerminalConversionParams
implements CoordinatesConversionParams
{
   /** Context-local instance of the cached parameters */
   private static final ContextLocal<Params> params =
   new ContextLocal<Params>()
   {
      @Override
      protected Params initialValue()
      {
         Params result = new Params();
         ClientExports client = LogicalTerminal.getClient();
         result.isChui = client.isChui();
         result.pixelsPerColumn = client.getPixelsPerColumn();
         result.pixelsPerRow = client.getPixelsPerRow();
         return result;
      }
   };

   /**
    * Flag indicating the ChUI or GUI mode.
    * 
    * @return  Boolean value indicating the ChUI or GUI mode.
    */
   @Override
   public boolean isChui()
   {
      return params.get().isChui;
   }

   /**
    * Returns the number of pixels in each column of the display. This usually corresponds
    * to the 4GL attribute SESSION:PIXELS-PER-COLUMN.
    * 
    * @return  Number of pixels in each column.
    */
   @Override
   public int getPixelsPerColumn()
   {
      return params.get().pixelsPerColumn;
   }

   /**
    * Returns the number of pixels in each row of the display. This usually corresponds
    * to the 4GL attribute SESSION:PIXELS-PER-ROW.
    * 
    * @return  Number of pixels in each row.
    */
   @Override
   public int getPixelsPerRow()
   {
      return params.get().pixelsPerRow;
   }

   /**
    * Get the width of the given font.
    * 
    * @param    windowId
    *           The ID of the window from which the font-table is used.
    * @param    font
    *           Font index.
    *          
    * @return   Width of the font.
    */
   @Override
   public int getFontWidth(WidgetId windowId, int font)
   {
      throw new RuntimeException("Not implemented!");
   }

   /**
    * Get the height of the given font.
    * 
    * @param    windowId
    *           The ID of the window from which the font-table is used.
    * @param    font
    *           Font index.
    *          
    * @return   Height of the font.
    */
   @Override
   public int getFontHeight(WidgetId windowId, int font)
   {
      throw new RuntimeException("Not implemented!");
   }
   
   /**
    * Container for storing the conversion parameters. 
    */
   private static class Params
   {
      /**
       * Flag indicating the ChUI or GUI mode. 
       */
      public boolean isChui;
      
      /**
       * Pixels per column.
       */
      public int pixelsPerColumn;
      
      /**
       * Pixels per row.
       */
      public int pixelsPerRow;
   }
}