/*
** Module   : MarkEntry.java
** Abstract : stores data to make deferred hiding/pausing state changes
**
** Copyright (c) 2006-2017, Golden Code Development Corporation.
**
** -#- -I- --Date-- --JPRM-- ----------------Description-----------------
** 001 GES 20060731   @28240 Created initial version to contain a simple
**                           set of server-side state.
** 002 GES 20060801   @28298 Converted to Externalizable to improve
**                           efficiency because Serializable has unneeded
**                           processing (for versioning...) that is
**                           executed on "inflation".
** 003 SIY 20060804   @28456 Replaced two booleans with mode of operation
**                           and added constructors.
** 004 GES 20061127   @31411 Added DESTROY_SHARED operation.
** 005 NVS 20061129   @31475 Added OPEN_SCOPE operation.
** 006 SIY 20070119   @31902 Removed DESTROY_SHARED operation.
** 007 SIY 20070521   @33754 Added CLOSE_SCOPE operation.
** 008 SIY 20071126   @36040 Added frame DOWN data. Ref: #36034
** 009 GES 20080613   @38880 Removed frame DOWN data. The return sync
**                           uses ClientState now, instead of making this
**                           class have multiple roles.
** 010 SIY 20081003   @40019 Added constant for resetting 'can hide' flag.
** 011 SIY 20081127   @40712 Added constant for setting 'can hide others' 
**                           flag.
** 011 SIY 20090422   @41842 Added constant for setting 'can hide others' 
**                           and 'can wipe screen' flags.
** 012 CA  20131013          Added constant for cleaning up a frame when its instantiating 
**                           procedure gets deleted. 
** 013 CA  20160304          Added DESTROY_WIDGETS.
**/

package com.goldencode.p2j.ui;

import java.io.*;

/**
 * This class is a container for server-side state for hiding/pausing.
 */
/*
** This program is free software: you can redistribute it and/or modify
** it under the terms of the GNU Affero General Public License as
** published by the Free Software Foundation, either version 3 of the
** License, or (at your option) any later version.
**
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
** GNU Affero General Public License for more details.
**
** You may find a copy of the GNU Affero GPL version 3 at the following
** location: https://www.gnu.org/licenses/agpl-3.0.en.html
** 
** Additional terms under GNU Affero GPL version 3 section 7:
** 
**   Under Section 7 of the GNU Affero GPL version 3, the following additional
**   terms apply to the works covered under the License.  These additional terms
**   are non-permissive additional terms allowed under Section 7 of the GNU
**   Affero GPL version 3 and may not be removed by you.
** 
**   0. Attribution Requirement.
** 
**     You must preserve all legal notices or author attributions in the covered
**     work or Appropriate Legal Notices displayed by works containing the covered
**     work.  You may not remove from the covered work any author or developer
**     credit already included within the covered work.
** 
**   1. No License To Use Trademarks.
** 
**     This license does not grant any license or rights to use the trademarks
**     Golden Code, FWD, any Golden Code or FWD logo, or any other trademarks
**     of Golden Code Development Corporation. You are not authorized to use the
**     name Golden Code, FWD, or the names of any author or contributor, for
**     publicity purposes without written authorization.
** 
**   2. No Misrepresentation of Affiliation.
** 
**     You may not represent yourself as Golden Code Development Corporation or FWD.
** 
**     You may not represent yourself for publicity purposes as associated with
**     Golden Code Development Corporation, FWD, or any author or contributor to
**     the covered work, without written authorization.
** 
**   3. No Misrepresentation of Source or Origin.
** 
**     You may not represent the covered work as solely your work.  All modified
**     versions of the covered work must be marked in a reasonable way to make it
**     clear that the modified work is not originating from Golden Code Development
**     Corporation or FWD.  All modified versions must contain the notices of
**     attribution required in this license.
*/
public class MarkEntry
implements Externalizable
{
   /** Do nothing. */
   public static final int NOP = 0;
   
   /** Mark frame for hiding. */
   public static final int HIDE = 1;
   
   /** Mark frame for hiding and reset pause flag.*/
   public static final int HIDE_NO_PAUSE = 2;
   
   /** Destroy the frame. */
   public static final int DESTROY = 3;
   
   /** Enable hiding of the frames. */
   public static final int SET_CAN_HIDE = 4;
   
   /** Disable hiding of the frames. */
   public static final int RESET_CAN_HIDE = 5;

   /** Disable hiding of the frames. */
   public static final int RESET_FRAME = 6;
   
   /** Signal entering a new scope. */
   public static final int OPEN_SCOPE = 7;
   
   /** Signal end of frame scope. */
   public static final int CLOSE_SCOPE = 8;

   /** Reset frames' 'can hide' flag. */
   public static final int UNHIDE = 9;

   /** Set frames' 'can hide others' flag. */
   public static final int CAN_HIDE_OTHERS = 10;

   /** Set frames' 'can hide others' flag and 'can wipe screen' flag. */
   public static final int CAN_HIDE_OTHERS_EXT = 11;
   
   /** Signal deletion of frame's instantiating procedure. */
   public static final int DELETED_SCOPE = 12;
   
   /** Destroy the widgets. */
   public static final int DESTROY_WIDGETS = 13;

   /** List of frame or widget IDs to process. */
   public int[] frames = null;
   
   /** Mode of operation. */
   public int mode = NOP;
   
   /**
    * Construct an instance for specified array of IDs and operating mode.
    * 
    * @param   frames
    * @param   mode
    */
   public MarkEntry(int[] frames, int mode)
   {
      this.frames = frames;
      this.mode   = mode;
   }

   /**
    * Construct an empty instance.
    */
   public MarkEntry()
   {
   }
   
   /**
    * Replacement for the default object reading method.
    * 
    * @param    in
    *           Input source from which fields will be restored.
    *
    * @throws   IOException
    *           In case of I/O errors.
    * @throws   ClassNotFoundException
    *           If payload can't be instantiated.
    */
   public void readExternal(ObjectInput in)
   throws IOException,
          ClassNotFoundException
   {
      frames  = (int[]) in.readObject();
      mode    = in.readInt();
   }

   /**
    * Replacement for the default object writing method.
    * 
    * @param    out
    *           The output destination to which fields will be saved.
    *
    * @throws   IOException
    *           In case of I/O errors.
    */
   public void writeExternal(ObjectOutput out)
   throws IOException
   {
      out.writeObject(frames);
      out.writeInt(mode);
   }
}