/*
** Module   : PaneEntity.java
** Abstract : server side base class for DIALOG, FRAME, WINDOW widgets
**
** Copyright (c) 2014-2018, Golden Code Development Corporation.
**
** -#- -I- --Date-- ---------------------------------Description---------------------------------
** 001 MAG 20140514 Created initial version.      
** 002 CA  20140926 Added shared widget configuration support, to allow GUI/ChUI concrete
**                  implementation for the same widget ID.  Refactored the widget configuration
**                  classes: all fields were made public; these classes need to be as dumb as
**                  possible, all logic related to setting/getting a certain field should be at
**                  the widget or at the caller. Refs #2254
** 003 HC  20141021 Implemented window move/position support.
** 004 HC  20141207 Implemented runtime support for widget coordinate attributes.
** 005 HC  20150803 Improved runtime support for WINDOW size attributes.
** 006 IAS 20151015 Implemented setters  for the THREE-D attribute
** 007 HC  20160203 Improved runtime support of FRAME:VIRTUAL* attributes.
** 008 CA  20160303 Made attribute setters be a no-op if the value doesn't change.
** 009 HC  20160320 Improved frame virtual size handling.
** 010 EVL 20170529 Adding Droppable and DirectManipulation interfaces to handle file list
**                  dropping and widget move resize capability.
** 011 CA  20180130 Replaced pushScreenDefinition() with pushWidgetAttr() for cases when business 
**                  logic is assigning a widget attribute.
** 012 HC  20180611 Fixed several issues in runtime support of FRAME:VIRTUAL-* attributes: wrong
**                  size reported by the VIRTUAL* attributes, wrong frame outer size calculated,
**                  wrong inner size calculated. And related changes.
*/
/*
** This program is free software: you can redistribute it and/or modify
** it under the terms of the GNU Affero General Public License as
** published by the Free Software Foundation, either version 3 of the
** License, or (at your option) any later version.
**
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
** GNU Affero General Public License for more details.
**
** You may find a copy of the GNU Affero GPL version 3 at the following
** location: https://www.gnu.org/licenses/agpl-3.0.en.html
** 
** Additional terms under GNU Affero GPL version 3 section 7:
** 
**   Under Section 7 of the GNU Affero GPL version 3, the following additional
**   terms apply to the works covered under the License.  These additional terms
**   are non-permissive additional terms allowed under Section 7 of the GNU
**   Affero GPL version 3 and may not be removed by you.
** 
**   0. Attribution Requirement.
** 
**     You must preserve all legal notices or author attributions in the covered
**     work or Appropriate Legal Notices displayed by works containing the covered
**     work.  You may not remove from the covered work any author or developer
**     credit already included within the covered work.
** 
**   1. No License To Use Trademarks.
** 
**     This license does not grant any license or rights to use the trademarks
**     Golden Code, FWD, any Golden Code or FWD logo, or any other trademarks
**     of Golden Code Development Corporation. You are not authorized to use the
**     name Golden Code, FWD, or the names of any author or contributor, for
**     publicity purposes without written authorization.
** 
**   2. No Misrepresentation of Affiliation.
** 
**     You may not represent yourself as Golden Code Development Corporation or FWD.
** 
**     You may not represent yourself for publicity purposes as associated with
**     Golden Code Development Corporation, FWD, or any author or contributor to
**     the covered work, without written authorization.
** 
**   3. No Misrepresentation of Source or Origin.
** 
**     You may not represent the covered work as solely your work.  All modified
**     versions of the covered work must be marked in a reasonable way to make it
**     clear that the modified work is not originating from Golden Code Development
**     Corporation or FWD.  All modified versions must contain the notices of
**     attribution required in this license.
*/

package com.goldencode.p2j.ui;

import java.util.*;

import com.goldencode.p2j.util.*;

/**
 * The server side base class for DIALOG, FRAME, WINDOW widgets. 
 */
public class PaneEntity<T extends PaneConfig>
extends BaseEntity<T>
implements Droppable,
           DirectManipulation
{
   /**
    * Constructor.
    *
    * @param    dynamic
    *           Flag indicating if this is a static or dynamic resource.
    * @param    paneConfig
    *           The container for all state to be transferred to the client.
    */
   public PaneEntity(boolean dynamic, T paneConfig)
   {
      super(dynamic, paneConfig);
   }
   
   /**
    * Constructor.
    *
    * @param    dynamic
    *           Flag indicating if this is a static or dynamic resource.
    * @param    paneConfig
    *           The container for all state to be transferred to the client.
    * @param    before
    *           The right neighbor of the BaseEntity been created or null if new BaseEntity
    *           will be the last in the chain. If <b>before</b> is the first element in the chain
    *           the new BaseEntity will become the new first element in the chain.
    * @param    firstAsLast
    *           The flag indicating that this new entity must be inserted as last element 
    *           in the chain in the case when <b>before</b> - is the first element of the chain.
    *           When <b>before</b> is the first element of the chain, default behavior (expected
    *           from the name of this parameter is to insert new element as the first. 
    *           This flag allows to change the expected behavior of <b>before</b> parameter when 
    *           it needs and makes sense only if <b>before</b> is the first element.
    */
   public PaneEntity(boolean dynamic, T paneConfig, PaneEntity<T> before, boolean firstAsLast)
   {
      super(dynamic, paneConfig, before, firstAsLast);
   }
   
   /**
    * Check if window has THREE-D option set.
    * 
    * @return <code>true</code> if window has THREE-D option set.
    */
   @Override
   public logical isThreeD()
   {
      return new logical(config.threeD);
   }

   /**
    * Enable/disable 3D look.
    * 
    * @param   threeD
    *          New flag value.
    */
   @Override
   public void setThreeD(logical threeD)
   {
      if (threeD == null || threeD.isUnknown())
      {
         ErrorManager.recordOrShowError(4083, 
               String.format(
               "Unable to assign UNKNOWN value to attribute THREE-D on %s %s", 
               type(), widgetName()
               ),
               false);
         return;
      }
      setThreeD(threeD.booleanValue());
   }

   /**
    * Enable/disable 3D look.
    * 
    * @param   threeD
    *          New flag value.
    */
   @Override
   public void setThreeD(boolean threeD)
   {
      if (config.realized)
      {
         ErrorManager.recordOrShowError(4053, 
               String.format(
                  "Unable to set THREE-D because the %s %s has been realised", 
                  type(), widgetName()
               ), 
               false);
         return;
      }
      
      if (config.threeD == threeD)
      {
         return;
      }
      
      config.threeD = threeD;
      pushWidgetAttr("threeD", config.threeD);
   }

   /**
    * Get VIRTUAL-HEIGHT-CHARS attribute. 
    */
   @Override
   public decimal getVirtHeight()
   {
      decimal v = getHeightChars();
      if (config.virtualHeightChars <= v.doubleValue())
      {
         return v;
      }
      else
      {
         return new decimal(config.virtualHeightChars);
      }
   }

   /**
    * Set VIRTUAL-HEIGHT-CHARS attribute. 
    *
    * @param   height
    *          VIRTUAL-HEIGHT-CHARS value
    */
   @Override
   public void setVirtHeight(NumberType height)
   {
      setSizeChars(height.isUnknown() ? null : height.doubleValue(), 
                   SizeAttribute.VIRT_HEIGHT_CHARS);
   }

   /**
    * Get VIRTUAL-WIDTH-CHARS attribute. 
    */
   @Override
   public decimal getVirtWidth()
   {
      decimal v = getWidthChars();
      if (config.virtualWidthChars <= v.doubleValue())
      {
         return v;
      }
      else
      {
         return new decimal(config.virtualWidthChars);
      }
   }

   /**
    * Set VIRTUAL-WIDTH-CHARS attribute. 
    *
    * @param   width
    *          VIRTUAL-WIDTH-CHAR value
    */
   @Override
   public void setVirtWidth(NumberType width)
   {
      setSizeChars(width.isUnknown() ? null : width.doubleValue(), 
                   SizeAttribute.VIRT_WIDTH_CHARS);
   }
   
   /**
    * Get VIRTUAL-HEIGHT-PIXELS attribute. 
    */
   @Override
   public integer getVirtHeightPixels()
   {
      integer v = getHeightPixels();
      if (config.virtualHeightPixels <= v.intValue())
      {
         return v;
      }
      else
      {
         return new integer(config.virtualHeightPixels);
      }
   }

   /**
    * Set VIRTUAL-HEIGHT-PIXELS attribute.
    *
    * @param   height
    *          VIRTUAL-HEIGHT-PIXELS value
    */
   @Override
   public void setVirtHeightPixels(NumberType height)
   {
      setSizePixels(height.isUnknown() ? null : height.intValue(), 
                    SizeAttribute.VIRT_HEIGHT_PIXELS);
   }

   /**
    * Get VIRTUAL-WIDTH-PIXELS attribute. 
    */
   @Override
   public integer getVirtWidthPixels()
   {
      integer v = getWidthPixels();
      if (config.virtualWidthPixels <= v.intValue())
      {
         return v;
      }
      else
      {
         return new integer(config.virtualWidthPixels);
      }
   }

   /**
    * Set VIRTUAL-WIDTH-PIXELS attribute. 
    *
    * @param   width
    *          VIRTUAL-WIDTH-PIXELS value
    */
   @Override
   public void setVirtWidthPixels(NumberType width)
   {
      setSizePixels(width.isUnknown() ? null : width.intValue(), 
                    SizeAttribute.VIRT_WIDTH_PIXELS);
   }
   
   // start implementation of the DirectManupulationInterface
   /**
    * Gets BOX-SELECTABLE attribute.
    * 
    * @return  BOX-SELECTABLE current value.
    */
   @Override
   public logical isBoxSelectable()
   {
      return new logical(config.boxSelectable);
   }
   
   /**
    * Sets BOX-SELECTABLE attribute.
    * 
    * @param   flag
    *          New value for the attribute.
    */
   @Override
   public void setBoxSelectable(boolean flag)
   {
      if (flag != config.boxSelectable)
      {
         config.boxSelectable = flag;
         pushWidgetAttr("boxSelectable", config.boxSelectable);
      }
   }
   
   /**
    * Gets NUM-SELECTED-WIDGETS attribute.
    * 
    * @return  NUM-SELECTED-WIDGETS current value.
    */
   @Override
   public integer getNumSelectedWidgets()
   {
      return new integer(config.numSelectedWidgets);
   }

   /**
    * Gets GRID-FACTOR-HORIZONTAL attribute.
    * 
    * @return  GRID-FACTOR-HORIZONTAL current value.
    */
   @Override
   public integer getGridFactorHorizontal()
   {
      return new integer(config.gridFactorHorizontal);
   }
   
   /**
    * Sets GRID-FACTOR-HORIZONTAL attribute.
    * 
    * @param   value
    *          New value for the attribute.
    */
   @Override
   public void setGridFactorHorizontal(int value)
   {
      if (value > 0 && value != config.gridFactorHorizontal)
      {
         config.gridFactorHorizontal = value;
         pushWidgetAttr("gridFactorHorizontal", config.gridFactorHorizontal);
      }
   }
   
   /**
    * Gets GRID-FACTOR-VERTICAL attribute.
    * 
    * @return  GRID-FACTOR-VERTICAL current value.
    */
   @Override
   public integer getGridFactorVertical()
   {
      return new integer(config.gridFactorVertical);
   }
   
   /**
    * Sets GRID-FACTOR-VERTICAL attribute.
    * 
    * @param   value
    *          New value for the attribute.
    */
   @Override
   public void setGridFactorVertical(int value)
   {
      if (value > 0 && value != config.gridFactorVertical)
      {
         config.gridFactorVertical = value;
         pushWidgetAttr("gridFactorVertical", config.gridFactorVertical);
      }
   }

   /**
    * Gets GRID-SNAP attribute.
    * 
    * @return  GRID-SNAP current value.
    */
   @Override
   public logical isGridSnap()
   {
      return new logical(config.gridSnap);
   }
   
   /**
    * Sets GRID-SNAP attribute.
    * 
    * @param   flag
    *          New value for the attribute.
    */
   @Override
   public void setGridSnap(boolean flag)
   {
      if (flag != config.gridSnap)
      {
         config.gridSnap = flag;
         pushWidgetAttr("gridSnap", config.gridSnap);
      }
   }
   
   /**
    * Gets GRID-UNIT-HEIGHT-CHARS attribute.
    * 
    * @return  GRID-UNIT-HEIGHT-CHARS current value.
    */
   @Override
   public decimal getGridUnitHeightChars()
   {
      return new decimal(LogicalTerminal.coordinates().rowFromPixels(config.gridUnitHeightPixels));
   }
   
   /**
    * Sets GRID-UNIT-HEIGHT-CHARS attribute.
    * 
    * @param   value
    *          New value for the attribute.
    */
   @Override
   public void setGridUnitHeightChars(double value)
   {
      setGridUnitHeightPixels(LogicalTerminal.coordinates().rowToPixels(value));
   }
   
   /**
    * Gets GRID-UNIT-HEIGHT-PIXELS attribute.
    * 
    * @return  GRID-UNIT-HEIGHT-PIXELS current value.
    */
   @Override
   public integer getGridUnitHeightPixels()
   {
      return new integer(config.gridUnitHeightPixels);
   }
   
   /**
    * Sets GRID-UNIT-HEIGHT-PIXELS attribute.
    * 
    * @param   value
    *          New value for the attribute.
    */
   @Override
   public void setGridUnitHeightPixels(int value)
   {
      if (value > 0 && value != config.gridUnitHeightPixels)
      {
         config.gridUnitHeightPixels = value;
         pushWidgetAttr("gridUnitHeightPixels", config.gridUnitHeightPixels);
      }
   }
   
   /**
    * Gets GRID-UNIT-WIDTH-CHARS attribute.
    * 
    * @return  GRID-UNIT-WIDTH-CHARS current value.
    */
   @Override
   public decimal getGridUnitWidthChars()
   {
      return new decimal(LogicalTerminal.coordinates().columnFromPixels(
                         config.gridUnitWidthPixels));
   }
   
   /**
    * Sets GRID-UNIT-WIDTH-CHARS attribute.
    * 
    * @param   value
    *          New value for the attribute.
    */
   @Override
   public void setGridUnitWidthChars(double value)
   {
      setGridUnitWidthPixels(LogicalTerminal.coordinates().columnToPixels(value));
   }
   
   /**
    * Gets GRID-UNIT-WIDTH-PIXELS attribute.
    * 
    * @return  GRID-UNIT-WIDTH-PIXELS current value.
    */
   @Override
   public integer getGridUnitWidthPixels()
   {
      return new integer(config.gridUnitWidthPixels);
   }
   
   /**
    * Sets GRID-UNIT-WIDTH-PIXELS attribute.
    * 
    * @param   value
    *          New value for the attribute.
    */
   @Override
   public void setGridUnitWidthPixels(int value)
   {
      if (value > 0 && value != config.gridUnitWidthPixels)
      {
         config.gridUnitWidthPixels = value;
         pushWidgetAttr("gridUnitWidthPixels", config.gridUnitWidthPixels);
      }
   }
   
   /**
    * Gets GRID-VISIBLE attribute.
    * 
    * @return  GRID-VISIBLE current value.
    */
   @Override
   public logical isGridVisible()
   {
      return new logical(config.gridVisible);
   }
   
   /**
    * Sets GRID-VISIBLE attribute.
    * 
    * @param   flag
    *          New value for the attribute.
    */
   @Override
   public void setGridVisible(boolean flag)
   {
      if (flag != config.gridVisible)
      {
         config.gridVisible = flag;
         pushWidgetAttr("gridVisible", config.gridVisible);
      }
   }
   // end implementation of the DirectManupulationInterface
   
   /**
    * Gets DROP-TARGET attribute.
    * 
    * @return  DROP-TARGET current value.
    */
   @Override
   public logical isDropTarget()
   {
      return new logical(config.dropTarget);
   }
   
   /**
    * Sets DROP-TARGET attribute.
    * 
    * @param   flag
    *          New value for the attribute.
    */
   @Override
   public void setDropTarget(boolean flag)
   {
      if (config.dropTarget != flag)
      {
         config.dropTarget = flag;
         pushWidgetAttr("dropTarget", config.dropTarget);
      }
   }
   
   /**
    * Sets the value for the attribute VIRTUAL-WIDTH-CHARS.
    * Only valid width values must be passed in the width argument.
    * 
    * @param    width
    *           Valid width value.
    */
   protected void setVirtWidthCharsWorker(double width)
   {
      decimal w = _getWidthChars();
      double dw = w.doubleValue();

      if (width < dw)
      {
         // decreasing below regular size also affects regular size
         setWidthChars(width);
      }

      if (width <= dw)
      {
         // make the effective virtual size the regular size
         config.virtualWidthChars = BaseConfig.INV_COORD;
      }
      else
      {
         config.virtualWidthChars = width;
      }
   }

   /**
    * Sets the value for the attribute VIRTUAL-HEIGHT-CHARS.
    * Only valid height values must be passed in the height argument.
    * 
    * @param    height
    *           Valid height value.
    */
   protected void setVirtHeightCharsWorker(double height)
   {
      decimal h = _getHeightChars();
      double dh = h.doubleValue();

      if (height < dh)
      {
         // decreasing below regular size also affects regular size
         setHeightChars(height);
      }

      if (height <= dh)
      {
         // make the effective virtual size the regular size
         config.virtualHeightChars = BaseConfig.INV_COORD;
      }
      else
      {
         config.virtualHeightChars = height;
      }
   }

   /**
    * Sets the value for the attribute VIRTUAL-WIDTH-PIXELS.
    * Only valid width values must be passed in the width argument.
    * 
    * @param    width
    *           Valid width value.
    */
   protected void setVirtWidthPixelsWorker(int width)
   {
      integer w = _getWidthPixels();
      int dw = w.intValue();

      if (width < dw)
      {
         // decreasing below regular size also affects regular size
         setWidthPixels(width);
      }

      if (width <= dw)
      {
         // make the effective virtual size the regular size
         config.virtualWidthPixels = BaseConfig.INV_COORD;
      }
      else
      {
         config.virtualWidthPixels = width;
      }
   }

   /**
    * Sets the value for the attribute VIRTUAL-HEIGHT-PIXELS.
    * Only valid height values must be passed in the height argument.
    * 
    * @param    height
    *           Valid height value.
    */
   protected void setVirtHeightPixelsWorker(int height)
   {
      integer h = _getHeightPixels();
      int dh = h.intValue();

      if (height < dh)
      {
         // decreasing below regular size also affects regular size
         setHeightPixels(height);
      }

      if (height <= dh)
      {
         // make the effective virtual size the regular size
         config.virtualHeightPixels = BaseConfig.INV_COORD;
      }
      else
      {
         config.virtualHeightPixels = height;
      }
   }
   
   /**
    * Collect the size-related attributes which can be assigned by this widget.
    * 
    * @param    names
    *           The collection where to add the field names.
    * @param    vals
    *           The collection where to add the field values.
    */
   @Override
   protected void getSizeAttrs(List<String> names, List<Object> vals)
   {
      super.getSizeAttrs(names, vals);

      names.add("virtualHeightChars");
      vals.add(config.virtualHeightChars);
      
      names.add("virtualWidthChars");
      vals.add(config.virtualWidthChars);
      
      names.add("virtualHeightPixels");
      vals.add(config.virtualHeightPixels);
      
      names.add("virtualWidthPixels");
      vals.add(config.virtualWidthPixels);
   }
}