/*
** Module   : ServerConfigManager.java
** Abstract : Management of widget configurations, on server-side.
**
** Copyright (c) 2014-2018, Golden Code Development Corporation.
**
** -#- -I- --Date-- ---------------------------------Description----------------------------------
** 001 CA  20140929 Created initial version.
** 002 CA  20141014 Separated data which needs to survive the appserver agent's context reset into 
**                  a different container.
** 003 HC  20150323 Removed business logic from config classes.
** 004 CA  20160304 Added isSharedConfig().
** 005 OM  20161207 Removed dead code.
** 006 CA  20180130 Replaced MethodHandle with ReflectASM.  Other misc performance improvements.
*/ 
/*
** This program is free software: you can redistribute it and/or modify
** it under the terms of the GNU Affero General Public License as
** published by the Free Software Foundation, either version 3 of the
** License, or (at your option) any later version.
**
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
** GNU Affero General Public License for more details.
**
** You may find a copy of the GNU Affero GPL version 3 at the following
** location: https://www.gnu.org/licenses/agpl-3.0.en.html
** 
** Additional terms under GNU Affero GPL version 3 section 7:
** 
**   Under Section 7 of the GNU Affero GPL version 3, the following additional
**   terms apply to the works covered under the License.  These additional terms
**   are non-permissive additional terms allowed under Section 7 of the GNU
**   Affero GPL version 3 and may not be removed by you.
** 
**   0. Attribution Requirement.
** 
**     You must preserve all legal notices or author attributions in the covered
**     work or Appropriate Legal Notices displayed by works containing the covered
**     work.  You may not remove from the covered work any author or developer
**     credit already included within the covered work.
** 
**   1. No License To Use Trademarks.
** 
**     This license does not grant any license or rights to use the trademarks
**     Golden Code, FWD, any Golden Code or FWD logo, or any other trademarks
**     of Golden Code Development Corporation. You are not authorized to use the
**     name Golden Code, FWD, or the names of any author or contributor, for
**     publicity purposes without written authorization.
** 
**   2. No Misrepresentation of Affiliation.
** 
**     You may not represent yourself as Golden Code Development Corporation or FWD.
** 
**     You may not represent yourself for publicity purposes as associated with
**     Golden Code Development Corporation, FWD, or any author or contributor to
**     the covered work, without written authorization.
** 
**   3. No Misrepresentation of Source or Origin.
** 
**     You may not represent the covered work as solely your work.  All modified
**     versions of the covered work must be marked in a reasonable way to make it
**     clear that the modified work is not originating from Golden Code Development
**     Corporation or FWD.  All modified versions must contain the notices of
**     attribution required in this license.
*/

package com.goldencode.p2j.ui;

import java.util.*;

import com.goldencode.p2j.security.*;

/**
 * Management of widget configurations, on server-side.
 */
public class ServerConfigManager
extends ConfigManager
{
   /** Context-local state which needs to survive the context reset. */
   protected static ContextLocal<WorkArea> survivals = new ContextLocal<WorkArea>()
   {
      protected WorkArea initialValue()
      {
         return new WorkArea();
      };
      
      @Override
      protected boolean isResetAllowed()
      {
         return false;
      }
   };

   /**
    * On server-side, synchronize the field IDs for the server-side widget configurations with 
    * the IDs provided by the associated list.
    * 
    * @param    configs
    *           A map of configurations, having as hey a class name and as value a list of field
    *           names (the order in this list gives the field ID).
    */
   public void initWidgetConfigurations(Map<String, List<String>> configs)
   {
      WorkArea wa = survivals.get();
      
      wa.cfgDefs.clear();
      
      // initialize the widget config fields on the remote side
      for (String clsName : configs.keySet())
      {
         List<String> fieldNames = configs.get(clsName);
         
         try
         {
            @SuppressWarnings("unchecked")
            Class<? extends WidgetConfig> cfgCls = 
               (Class<? extends WidgetConfig>) Class.forName(clsName);
            
            WidgetConfigDef cfg = new WidgetConfigDef(cfgCls, fieldNames);
            wa.cfgDefs.put(cfgCls, cfg);
         } 
         catch (ClassNotFoundException e)
         {
            final String msg = "Could not initialize remote widget configs for class %s.";
            throw new RuntimeException(String.format(msg, clsName), e);
         }
      }
   }
   
   /**
    * Determine if the specified config instance can be tracked.
    * <p>
    * Sever-side does not track the config changes.
    * 
    * @param    cfg
    *           The {@link WidgetConfig} instace which needs to be tracked.
    *           
    * @return   See above.
    */
   @Override
   public boolean trackConfig(WidgetConfig cfg)
   {
      return false;
   }

   /**
    * Register the specified field in the configuration for the specified widget ID as being 
    * dirty.
    * <p>
    * Sever-side does not track the config changes, so this is a no-op.
    * 
    * 
    * @param    cfg
    *           The widget config.
    * @param    field
    *           The changed field.
    */
   @Override
   public void addDirtyConfig(WidgetConfig cfg, String field)
   {
      // no-op
   }

   /**
    * The method attempts to resolve a {@link ConfigOwner} reference from
    * the widget <code>widgetId</code>.
    * 
    * @param    widgetId
    *           Widget Id.
    *           
    * @return   A {@link ConfigOwner} reference or <code>null</code> if
    *           such cannot be find. 
    */
   @Override
   ConfigOwner<?> resolveConfigOwner(WidgetId widgetId)
   {
      GenericWidget<?> widget = LogicalTerminal.getWidgetForId(widgetId.asInt());
      if (widget instanceof ConfigOwner)
      {
         return (ConfigOwner<?>) widget;
      }
      
      return null;
   }
   
   /**
    * Check if the given ID is associated with a registered widget.
    * 
    * @param    id
    *           The widget ID.
    *           
    * @return   <code>true</code> if a widget exists in the registry, with the given ID.
    */
   @Override
   protected boolean isRegisteredWidget(WidgetId id)
   {
      return id != null && LogicalTerminal.isWidgetIdInUse(id.asInt());
   }

   /**
    * Get the {@link WidgetConfigDef} instance associated with the given {@link WidgetConfig} 
    * implementation.
    * 
    * @param    cls
    *           The widget config class.
    *           
    * @return   See above.
    */
   @Override
   protected WidgetConfigDef getConfigDef(Class<? extends WidgetConfig> cls)
   {
      WorkArea wa = survivals.get();
      
      return wa.cfgDefs.get(cls);
   }

   /**
    * Update the local active and backup configurations with the received changes.
    * 
    * @param    cfgUpdate
    *           The container with the config updates for a given widget ID.
    */
   @SuppressWarnings("unchecked")
   protected void applyConfigUpdates(GenericWidget widget, WidgetConfigUpdates cfgUpdate)
   {
      // there is a "chicken-and-the-egg" problem here.  if the widget has just been created, the
      // state sync is applied BEFORE the pushScreenDefinition API is executed.  this means that
      // we don't know the widget's config class (and other info), as we are too early.  for this
      // reason, all calls for which there is no widget config available are dropped.  it is 
      // expected for the server-side to setup the widget properly (including the active/backup 
      // config instances) via pushScreenDefinition calls.
      
      // the server-side holds all client-side state in a single config...
      WidgetId wid = new WidgetId(cfgUpdate.widgetId);
      
      WidgetConfigDef wdef = widgetCfg.get(wid);
      if (wdef == null)
      {
         return;
      }
      
      final WidgetConfig active = activeConfigs.get(wid);
      final WidgetConfig backup = backupConfigs.get(wid);

      // apply the changed fields to active
      for (int fidx = 0; fidx < cfgUpdate.changedAttributes.length; fidx++)
      {
         Object val = cfgUpdate.updates[fidx];
         
         int fid = cfgUpdate.changedAttributes[fidx];
         
         setWidgetConfig(active, val, wdef, wdef.fieldNames.get(fid));
      }
      
      // do post-processing - backup always reflects the previous state
      if (widget != null)
      {
         widget.afterConfigUpdate(backup);
      }

      // apply the changed fields to backup
      for (int fidx = 0; fidx < cfgUpdate.changedAttributes.length; fidx++)
      {
         Object val = cfgUpdate.updates[fidx];
         
         int fid = cfgUpdate.changedAttributes[fidx];
         
         setWidgetConfig(backup, val, wdef, wdef.fieldNames.get(fid));
      }
   }

   /**
    * Widget-specific data which needs to survive the context-reset.
    */
   static class WorkArea
   {
      /** Map of widget configuration definitions by their config class. */
      protected final Map<Class<? extends WidgetConfig>, WidgetConfigDef> cfgDefs = new HashMap<>();
   }
}