/*
** Module   : SliderWidget.java
** Abstract : Server side slider widget
**
** Copyright (c) 2015-2018, Golden Code Development Corporation.
**
** -#- -I- --Date-- ---------------------------------Description---------------------------------
** 001 EVL 20150911 Created initial version.
** 002 EVL 20160127 Adding dynamic widget constructor.  Default size calculations.
**                  The NO-CURRENT-VALUE should be inverted while setting to reflect 4GL logic.
**                  Changed direct check for congig.realized to generic call _isRealized().
** 003 IAS 20160217 LOAD-MOUSE-POINTER support.
** 004 CA  20160303 Made attribute setters be a no-op if the value doesn't change.
** 005 SVL 20171025 Fixed widget name in error messages.
** 006 CA  20180130 Replaced pushScreenDefinition() with pushWidgetAttr() for cases when business 
**                  logic is assigning a widget attribute.
** 007 OM  20180713 Fixed TIC-MARKS attribute.
*/ 

/*
** This program is free software: you can redistribute it and/or modify
** it under the terms of the GNU Affero General Public License as
** published by the Free Software Foundation, either version 3 of the
** License, or (at your option) any later version.
**
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
** GNU Affero General Public License for more details.
**
** You may find a copy of the GNU Affero GPL version 3 at the following
** location: https://www.gnu.org/licenses/agpl-3.0.en.html
** 
** Additional terms under GNU Affero GPL version 3 section 7:
** 
**   Under Section 7 of the GNU Affero GPL version 3, the following additional
**   terms apply to the works covered under the License.  These additional terms
**   are non-permissive additional terms allowed under Section 7 of the GNU
**   Affero GPL version 3 and may not be removed by you.
** 
**   0. Attribution Requirement.
** 
**     You must preserve all legal notices or author attributions in the covered
**     work or Appropriate Legal Notices displayed by works containing the covered
**     work.  You may not remove from the covered work any author or developer
**     credit already included within the covered work.
** 
**   1. No License To Use Trademarks.
** 
**     This license does not grant any license or rights to use the trademarks
**     Golden Code, FWD, any Golden Code or FWD logo, or any other trademarks
**     of Golden Code Development Corporation. You are not authorized to use the
**     name Golden Code, FWD, or the names of any author or contributor, for
**     publicity purposes without written authorization.
** 
**   2. No Misrepresentation of Affiliation.
** 
**     You may not represent yourself as Golden Code Development Corporation or FWD.
** 
**     You may not represent yourself for publicity purposes as associated with
**     Golden Code Development Corporation, FWD, or any author or contributor to
**     the covered work, without written authorization.
** 
**   3. No Misrepresentation of Source or Origin.
** 
**     You may not represent the covered work as solely your work.  All modified
**     versions of the covered work must be marked in a reasonable way to make it
**     clear that the modified work is not originating from Golden Code Development
**     Corporation or FWD.  All modified versions must contain the notices of
**     attribution required in this license.
*/

package com.goldencode.p2j.ui;

import com.goldencode.p2j.util.*;

/**
 * Implements the server side slider widget.
 */
@LegacyResource(resource = LegacyResource.SLIDER)
public class SliderWidget
extends ControlEntity<SliderConfig>
implements SliderInterface,
           Hoverable
{
   /**
    * Default constructor (creates static vertical slider).
    */
   public SliderWidget()
   {
      this(false);
   }

   /**
    * Default constructor.
    * 
    * @param   dynamic
    *          Flag indicating if this is a static or dynamic resource.
    */
   public SliderWidget(boolean dynamic)
   {
      super(dynamic, new SliderConfig());
   }
   
   /**
    * Gets the MIN-VALUE writable attribute.
    *
    * @return   The current value of the MIN-VALUE attribute.
    */
   @Override
   public integer getMinValue()
   {
      return new integer(config.minValue);
   }

   /**
    * Sets the MIN-VALUE writable attribute.
    *
    * @param    minValue
    *           The new value for the MIN-VALUE attribute.
    */
   @Override
   public void setMinValue(NumberType minValue)
   {
      if (minValue != null && !minValue.isUnknown())
      {
         setMinValue(minValue.intValue());
      }
   }
   
   /**
    * Sets the MIN-VALUE writable attribute.
    *
    * @param    minValue
    *           The new value for the MIN-VALUE attribute.
    */
   @Override
   public void setMinValue(int minValue)
   {
      // change is possible only if widget is not realized 
      if (!_isRealized())
      {
         if (config.minValue == minValue)
         {
            return;
         }
         
         config.minValue = minValue;
         config.currentValue = config.minValue; 
         updateSize();
         pushWidgetAttr(new String[] {"minValue", "currentValue", "widthPixels", "heightPixels"},
                        new Object[] {config.minValue, config.currentValue, 
                                      config.widthPixels, config.heightPixels});
      }
      else
      {
         ErrorManager.recordOrShowError(4053, String.format(
               "Unable to set MIN-VALUE because the %s %s has been realised", type(),
                  widgetName()), false);
      }
   }

   /**
    * Gets the MAX-VALUE writable attribute.
    *
    * @return   The current value of the MAX-VALUE attribute.
    */
   @Override
   public integer getMaxValue()
   {
      return new integer(config.maxValue);
   }

   /**
    * Sets the MAX-VALUE writable attribute.
    *
    * @param    maxValue
    *           The new value for the MAX-VALUE attribute.
    */
   @Override
   public void setMaxValue(NumberType maxValue)
   {
      if (maxValue != null && !maxValue.isUnknown())
      {
         setMaxValue(maxValue.intValue());
      }
   }
   
   /**
    * Sets the MAX-VALUE writable attribute.
    *
    * @param    maxValue
    *           The new value for the MAX-VALUE attribute.
    */
   @Override
   public void setMaxValue(int maxValue)
   {
      // change is possible only if widget is not realized 
      if (!_isRealized())
      {
         if (config.maxValue == maxValue)
         {
            return;
         }
         
         config.maxValue = maxValue;
         updateSize();
         pushWidgetAttr(new String[] {"maxValue", "widthPixels", "heightPixels"},
                        new Object[] {config.maxValue, config.widthPixels, config.heightPixels});
      }
      else
      {
         ErrorManager.recordOrShowError(4053, String.format(
               "Unable to set MAX-VALUE because the %s %s has been realised", type(),
                  widgetName()), false);
      }
   }

   /**
    * Resets the value of the HORIZONTAL attribute for slider.
    * 
    * @param   value
    *          The new value of the HORIZONTAL attribute.
    */
   @Override
   public void setHorizontal(logical value)
   {
      if (value.isUnknown())
      {
         ErrorManager.recordOrShowError(4083, String.format(
               "Unable to assign UNKNOWN value to attribute HORIZONTAL on %s %s", type(),
                  widgetName()), false);
      }
      setHorizontal(value.booleanValue());
   }

   /**
    * Resets the value of the HORIZONTAL attribute for slider.
    * 
    * @param   value
    *          The new value of the HORIZONTAL attribute.
    */
   @Override
   public void setHorizontal(boolean value)
   {
      if (!_isRealized())
      {
         if (config.horizontal == value)
         {
            return;
         }

         config.horizontal = value;
         updateSize();
         pushWidgetAttr(new String[] {"horizontal", "widthPixels", "heightPixels"},
                        new Object[] {config.horizontal, config.widthPixels, config.heightPixels});
      }
      else
      {
         ErrorManager.recordOrShowError(4053, String.format(
               "Unable to set HORIZONTAL because the %s %s has been realised", type(),
                  widgetName()), false);
      }
   }

   /**
    * Get the value of the HORIZONTAL attribute for slider.
    * 
    * @return
    *          The new value of the HORIZONTAL attribute.
    */
   @Override
   public logical isHorizontal()
   {
      return new logical(config.horizontal);
   }

   /**
    * Resets the value of the HORIZONTAL attribute for slider.
    * 
    * @param   value
    *          The new value of the HORIZONTAL attribute.
    */
   @Override
   public void setVertical(logical value)
   {
      if (value.isUnknown())
      {
         ErrorManager.recordOrShowError(4083, String.format(
               "Unable to assign UNKNOWN value to attribute HORIZONTAL on %s %s", type(),
                  widgetName()), false);
      }
      setHorizontal(!value.booleanValue());
   }

   /**
    * Resets the value of the HORIZONTAL attribute for radio-set.
    * 
    * @param   value
    *          The new value of the VERTICAL option.
    */
   @Override
   public void setVertical(boolean value)
   {
      setHorizontal(!value);
   }

   /**
    * Get the value of the LARGE-TO-SMALL attribute.
    *
    * @return   the value of the attribute.
    */
   @Override
   public logical isLargeToSmall()
   {
      return new logical(config.largeToSmall);
   }
   
   /**
    * Set the value of the LARGE-TO-SMALL attribute.
    *
    * @param   value
    *          the new value of the attribute.
    */
   @Override
   public void setLargeToSmall(logical value)
   {
      if (value != null && !value.isUnknown())
      {
         setLargeToSmall(value.booleanValue());
      }
   }

   /**
    * Set the value of the LARGE-TO-SMALL attribute.
    *
    * @param   value
    *          the new value of the attribute.
    */
   @Override
   public void setLargeToSmall(boolean value)
   {
      if (config.largeToSmall == value)
      {
         return;
      }
      
      config.largeToSmall = value;
      pushWidgetAttr("largeToSmall", config.largeToSmall);
   }

   /**
    * Get the value of the NO-CURRENT-VALUE attribute.
    *
    * @return   the value of the attribute.
    */
   @Override
   public logical isNoCurrentValue()
   {
      return new logical(config.noCurrentValue);
   }
   
   /**
    * Set the value of the NO-CURRENT-VALUE attribute.  If NO-CURRENT-VALUE is <code>TRUE</code>
    * the slider displays current value.  If NO-CURRENT-VALUE is <code>FALSE</code> slider does
    * not display the current value.  The default value is <code>TRUE</code>.
    *
    * @param   value
    *          the new value of the attribute.
    */
   @Override
   public void setNoCurrentValue(logical value)
   {
      if (value != null && !value.isUnknown())
      {
         setNoCurrentValue(value.booleanValue());
      }
   }

   /**
    * Set the value of the NO-CURRENT-VALUE attribute.  If NO-CURRENT-VALUE is <code>TRUE</code>
    * the slider displays current value.  If NO-CURRENT-VALUE is <code>FALSE</code> slider does
    * not display the current value.  The default value is <code>TRUE</code>.
    *
    * @param   value
    *          the new value of the attribute.
    */
   @Override
   public void setNoCurrentValue(boolean value)
   {
      // change is possible only if widget is not realized 
      if (!_isRealized())
      {
         config.noCurrentValue = !value;
         pushWidgetAttr("noCurrentValue", config.noCurrentValue);
      }
      else
      {
         ErrorManager.recordOrShowError(4053, String.format(
               "Unable to set NO-CURRENT-VALUE because the %s %s has been realised", type(),
                  widgetName()), false);
      }
   }

   /**
    * Get the value of the TIC-MARKS attribute.
    *
    * @return   the value of the attribute.
    */
   @Override
   public character getTicMarks()
   {
      return new character(config.ticMarks);
   }

   /**
    * Set the value of the TICK-MARKS attribute.
    *
    * @param   tMarks
    *          the new value of the attribute.
    */
   @Override
   public void setTicMarks(character tMarks)
   {
      if (tMarks != null && !tMarks.isUnknown())
      {
         setTicMarks(tMarks.toStringMessage());
      }
   }
   
   /**
    * Set the value of the TIC-MARKS attribute.
    *
    * @param   tMarks
    *          the new value of the attribute.
    */
   @Override
   public void setTicMarks(String tMarks)
   {
      // if widget already realized - not possible to change tic 
      if (_isRealized())
      {
         ErrorManager.displayError(4053, String.format(
               "Unable to set TIC-MARKS because the %s %s has been realised.", type(),
                  widgetName()));
         return;
      }
      
      // WARNING: this can't be a no-op, as setting this re-sizes/realizes the widget
      tMarks = tMarks.toUpperCase();
      if (tMarks.equals(SliderConfig.TM_NONE)   || tMarks.equals(SliderConfig.TM_TOP)  ||
          tMarks.equals(SliderConfig.TM_BOTTOM) || tMarks.equals(SliderConfig.TM_LEFT) ||
          tMarks.equals(SliderConfig.TM_RIGHT)  || tMarks.equals(SliderConfig.TM_BOTH))
      {
         config.ticMarks = tMarks;
         updateSize();
         pushWidgetAttr(new String[] {"ticMarks", "widthPixels", "heightPixels"},
                        new Object[] {config.ticMarks, config.widthPixels, config.heightPixels});
      }
   }
   
   /**
    * Get the value of the FREQUENCY attribute.
    *
    * @return   the value of the attribute.
    */
   @Override
   public integer getFrequency()
   {
      return new integer(config.frequency);
   }
   
   /**
    * Set the value of the FREQUENCY attribute.
    *
    * @param   freq
    *          the new value of the attribute.
    */
   @Override
   public void setFrequency(NumberType freq)
   {
      if (freq != null && !freq.isUnknown())
      {
         setFrequency(freq.intValue());
      }
   }
   
   /**
    * Set the value of the FREQUENCY attribute.
    *
    * @param   freq
    *          the new value of the attribute.
    */
   @Override
   public void setFrequency(int freq)
   {
      if (config.frequency == freq)
      {
         return;
      }
      
      config.frequency = freq;
      pushWidgetAttr("frequency", config.frequency);
   }
   
   /**
    * Obtain the current value in the screen buffer of the backing data for
    * this widget and return it as a <code>character</code> type. If
    * there is no value in the screen buffer (if it is <code>null</code>
    * which is the representation for an uninitialized value), then return
    * the empty string.
    *
    * @return   The value as a <code>character</code> type or the empty
    *           string if the screen buffer value is uninitialized.
    */
   @Override
   public character getScreenValue()
   {
      // TODO: this sets the MODIFIED attribute to TRUE, too

      return new character(String.valueOf(config.currentValue));
   }
   
   /**
    * Set the current value in the screen buffer of the backing data for
    * this widget.  If the given value is <code>null</code> then this
    * widget will be set to the uninitialized value.
    *
    * @param    value
    *           The new value for the widget, use <code>null</code> to set
    *           the value as uninitialized.
    */
   @Override
   public void setScreenValue(character value)
   {
      if (value != null && !value.isUnknown())
      {
         String cvStr = value.toStringMessage();
         if (cvStr != null && !cvStr.isEmpty())
         {
            config.currentValue = Integer.valueOf(cvStr); 
         }
      }
   }

   /**
    * Set the new size values depending on min, max, orientation.
    */
   private void updateSize()
   {
      int baseSize = (int)(config.maxValue - config.minValue + 1.0) * 4;
      // different orientations have different approaches
      if (config.horizontal)
      {
         config.widthPixels = baseSize;
         config.heightPixels = config.widthPixels / 2;
         if (config.ticMarks.equalsIgnoreCase(SliderConfig.TM_BOTH))
         {
            config.heightPixels += 5;
         }
      }
      else
      {
         config.heightPixels = baseSize;
         config.widthPixels = (config.heightPixels * 4) / 5; 
         if (config.ticMarks.equalsIgnoreCase(SliderConfig.TM_BOTH))
         {
            config.widthPixels += 5;
         }
      }
   }

   /**
    * Set the MOUSE-POINTER attribute
    * 
    * @param name
    *        the attribute value
    */
   @Override
   public void setMousePointerName(String name)
   {
      config.mousePointer = name;
   }

   /**
    * Get the MOUSE-POINTER attribute
    * 
    * @return the attribute value 
    */
   @Override
   public String getMousePointerImpl()
   {
      return config.mousePointer;
   }
}
