/*
** Module   : ValidationResult.java
** Abstract : encapsulates the results of remote validation
**
** Copyright (c) 2006-2017, Golden Code Development Corporation.
**
** -#- -I- --Date-- --JPRM-- -------------------Description--------------------
** 001 NVS 20060320   @25157 Created the initial version, which simplifies
**                           the validation protocol by encapsulating all
**                           pieces of the validation related data into a
**                           class.
** 002 GES 20060807   @28467 Moved to externalizable interface to optimize
**                           network performance.                     
** 003 SIY 20111101          Updated imports.
** 004 EVL 20160722          Change to use screen buffer array instead of single buffer.
*/
/*
** This program is free software: you can redistribute it and/or modify
** it under the terms of the GNU Affero General Public License as
** published by the Free Software Foundation, either version 3 of the
** License, or (at your option) any later version.
**
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
** GNU Affero General Public License for more details.
**
** You may find a copy of the GNU Affero GPL version 3 at the following
** location: https://www.gnu.org/licenses/agpl-3.0.en.html
** 
** Additional terms under GNU Affero GPL version 3 section 7:
** 
**   Under Section 7 of the GNU Affero GPL version 3, the following additional
**   terms apply to the works covered under the License.  These additional terms
**   are non-permissive additional terms allowed under Section 7 of the GNU
**   Affero GPL version 3 and may not be removed by you.
** 
**   0. Attribution Requirement.
** 
**     You must preserve all legal notices or author attributions in the covered
**     work or Appropriate Legal Notices displayed by works containing the covered
**     work.  You may not remove from the covered work any author or developer
**     credit already included within the covered work.
** 
**   1. No License To Use Trademarks.
** 
**     This license does not grant any license or rights to use the trademarks
**     Golden Code, FWD, any Golden Code or FWD logo, or any other trademarks
**     of Golden Code Development Corporation. You are not authorized to use the
**     name Golden Code, FWD, or the names of any author or contributor, for
**     publicity purposes without written authorization.
** 
**   2. No Misrepresentation of Affiliation.
** 
**     You may not represent yourself as Golden Code Development Corporation or FWD.
** 
**     You may not represent yourself for publicity purposes as associated with
**     Golden Code Development Corporation, FWD, or any author or contributor to
**     the covered work, without written authorization.
** 
**   3. No Misrepresentation of Source or Origin.
** 
**     You may not represent the covered work as solely your work.  All modified
**     versions of the covered work must be marked in a reasonable way to make it
**     clear that the modified work is not originating from Golden Code Development
**     Corporation or FWD.  All modified versions must contain the notices of
**     attribution required in this license.
*/

package com.goldencode.p2j.ui;

import java.io.*;

/**
 * Represents the result of a validation request as the screen buffer,
 * <code>boolean</code> validation result and any associated error message
 * on failure.
 */
public class ValidationResult
implements Externalizable
{
   /** Validation request result. */
   private boolean result = false;
   
   /** Validation error message which may be null. */
   private String error = null;
   
   /** Validation return differential screen buffer. */
   private ScreenBuffer[] diffBuf = null;
   
   /**
    * Default constructor (only used for de-serialization).
    */
   public ValidationResult()
   {
   }
   
   /**
    * Package private constructor.
    *
    * @param    result
    *           validation request result
    *
    * @param    error
    *           validation error message or <code>null</code>
    *
    * @param    diffBuf
    *           return <code>ScreenBuffer</code> array of all editable buffers.
    */
   ValidationResult(boolean result, String error, ScreenBuffer[] diffBuf)
   {
      this.result  = result;
      this.error   = error;
      this.diffBuf = diffBuf;
   }
   
   /**
    * Gets the validation result.
    *
    * @return   <code>true</code> if validation was ssuccessful
    */
   public boolean isValidated()
   {
      return result;
   }
   
   /**
    * Gets the validation error message.
    *
    * @return   <code>null</code> if validation was ssuccessful,
    *           otherwise the error message
    */
   public String getError()
   {
      if (result)
         return null;
      
      return error;
   }

   /**
    * Gets the differential screen buffer that may have been changed
    * during validation.
    *
    * @return   <code>ScreenBuffer</code> array which may contain zero or more
    *           modified screen values.
    */
   public ScreenBuffer[] getScreenBuffers()
   {
      return diffBuf;
   }
   
   /**
    * Replacement for the default object reading method. The latest
    * state is read from the input source.
    * 
    * @param    in
    *           The input source from which fields will be restored.
    *
    * @throws   IOException
    *           In case of I/O errors.
    * @throws   ClassNotFoundException
    *           If payload can't be instantiated.
    */
   public void readExternal(ObjectInput in)
   throws IOException,
          ClassNotFoundException
   {
      result = in.readBoolean();
      
      // the rest of the data are objects which can be null so we
      // read an indicator var first to determine if there is following
      // data to be read
      
      if (in.readBoolean())
         error = (String) in.readObject();
      
      if (in.readBoolean())
         diffBuf = (ScreenBuffer[]) in.readObject();      
   }

   /**
    * Replacement for the default object writing method. The latest
    * state is written to the output destination.
    * 
    * @param    out
    *           The output destination to which fields will be saved.
    *
    * @throws   IOException
    *           In case of I/O errors.
    */
   public void writeExternal(ObjectOutput out)
   throws IOException
   {
      out.writeBoolean(result);
      
      // the rest of the data are objects which can be null so we
      // write an indicator var first to specify if there is following
      // data to be read
      
      boolean there = (error != null); 
      out.writeBoolean(there);
      if (there)
         out.writeObject(error);
      
      there = (diffBuf != null); 
      out.writeBoolean(there);
      if (there)
         out.writeObject(diffBuf);      
   }
}