/*
** Module   : WidgetConfigUpdates.java
** Abstract : helper class to transport widget changes from one side to another.
**
** Copyright (c) 2014-2018, Golden Code Development Corporation.
**
** -#- -I- --Date-- ---------------------------------Description----------------------------------
** 001 CA  20140924 Created initial version.
** 002 CA  20180130 Removed BitSet usage, send the changed field IDs directly.
*/ 
/*
** This program is free software: you can redistribute it and/or modify
** it under the terms of the GNU Affero General Public License as
** published by the Free Software Foundation, either version 3 of the
** License, or (at your option) any later version.
**
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
** GNU Affero General Public License for more details.
**
** You may find a copy of the GNU Affero GPL version 3 at the following
** location: https://www.gnu.org/licenses/agpl-3.0.en.html
** 
** Additional terms under GNU Affero GPL version 3 section 7:
** 
**   Under Section 7 of the GNU Affero GPL version 3, the following additional
**   terms apply to the works covered under the License.  These additional terms
**   are non-permissive additional terms allowed under Section 7 of the GNU
**   Affero GPL version 3 and may not be removed by you.
** 
**   0. Attribution Requirement.
** 
**     You must preserve all legal notices or author attributions in the covered
**     work or Appropriate Legal Notices displayed by works containing the covered
**     work.  You may not remove from the covered work any author or developer
**     credit already included within the covered work.
** 
**   1. No License To Use Trademarks.
** 
**     This license does not grant any license or rights to use the trademarks
**     Golden Code, FWD, any Golden Code or FWD logo, or any other trademarks
**     of Golden Code Development Corporation. You are not authorized to use the
**     name Golden Code, FWD, or the names of any author or contributor, for
**     publicity purposes without written authorization.
** 
**   2. No Misrepresentation of Affiliation.
** 
**     You may not represent yourself as Golden Code Development Corporation or FWD.
** 
**     You may not represent yourself for publicity purposes as associated with
**     Golden Code Development Corporation, FWD, or any author or contributor to
**     the covered work, without written authorization.
** 
**   3. No Misrepresentation of Source or Origin.
** 
**     You may not represent the covered work as solely your work.  All modified
**     versions of the covered work must be marked in a reasonable way to make it
**     clear that the modified work is not originating from Golden Code Development
**     Corporation or FWD.  All modified versions must contain the notices of
**     attribution required in this license.
*/

package com.goldencode.p2j.ui;

import java.io.*;

import com.goldencode.util.*;

/**
 * This is a helper class to transport widget changes (associated with public fields from the 
 * widget config class) from one side to another.
 */
public class WidgetConfigUpdates
implements Externalizable
{
   /** The widget  ID to which these changes belong. */
   public int widgetId;
   
   /** The list of changed field IDs. */
   public int[] changedAttributes;
   
   /** The new field values. */
   public Object[] updates;
   
   private NativeTypeSerializer[] workers;
   
   /**
    * Default c'tor.  Not for public use, allowed only in de-serialization.
    */
   public WidgetConfigUpdates()
   {
      // no-op
   }
   
   /**
    * Collect the information related to widget config changes, to transfer it to a remote side.
    * 
    * @param   widgetId
    *          The widget  ID to which these changes belong.
    * @param   changedAttributes
    *          The list of the changed field IDs.
    * @param   updates
    *          An array with the new field values.
    */
   public WidgetConfigUpdates(int                    widgetId,
                              int[]                  changedAttributes,
                              Object[]               updates,
                              NativeTypeSerializer[] workers)
   {
      this.widgetId = widgetId;
      this.changedAttributes = changedAttributes;
      this.updates = updates;
      this.workers = workers;
   }
   
   /**
    * Replacement for the default object writing method. The latest state is written to the output
    * destination.
    * 
    * @param    out
    *           The output destination to which fields will be saved.
    *
    * @throws   IOException
    *           In case of I/O errors.
    */
   @Override
   public void writeExternal(ObjectOutput out)
   throws IOException
   {
      out.writeInt(widgetId);

      out.writeInt(changedAttributes.length);
      for (int i = 0; i < changedAttributes.length; i++)
      {
         out.writeInt(changedAttributes[i]);
      }

      out.writeInt(updates.length);
      for (int i = 0; i < updates.length; i++)
      {
         NativeTypeSerializer.writeValue(updates[i], workers[i], out);
      }
   }
   
   /**
    * Replacement for the default object reading method. The latest state is read from the input 
    * source.
    * 
    * @param    in
    *           The input source from which fields will be restored.
    *
    * @throws   IOException
    *           In case of I/O errors.
    * @throws   ClassNotFoundException
    *           If payload can't be instantiated.
    */
   @Override
   public void readExternal(ObjectInput in)
   throws IOException,
          ClassNotFoundException
   {
      this.widgetId = in.readInt();

      int blen = in.readInt();
      this.changedAttributes = new int[blen];
      for (int i = 0; i < blen; i++)
      {
         changedAttributes[i] = in.readInt();
      }

      int ulen = in.readInt();
      this.updates = new Object[ulen];
      for (int i = 0; i < updates.length; i++)
      {
         this.updates[i] = NativeTypeSerializer.readValue(in);
      }
   }
}