/*
** Module   : WidgetDownId.java
** Abstract : represents widget identifiers for a widget part of a frame's down body
**
** Copyright (c) 2014-2017, Golden Code Development Corporation.
**
** -#- -I- --Date-- ---------------------------------Description----------------------------------
** 001 CA  20140924 Created initial version.
** 002 CA  20150114 Added the createId API, so that a new ID can be copied the location in the 
**                  down frame of an existing ID.
** 003 CA  20150115 Added a comment that equals() can't be safely used to identify widgets in a 
**                  down frame.
*/ 
/*
** This program is free software: you can redistribute it and/or modify
** it under the terms of the GNU Affero General Public License as
** published by the Free Software Foundation, either version 3 of the
** License, or (at your option) any later version.
**
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
** GNU Affero General Public License for more details.
**
** You may find a copy of the GNU Affero GPL version 3 at the following
** location: https://www.gnu.org/licenses/agpl-3.0.en.html
** 
** Additional terms under GNU Affero GPL version 3 section 7:
** 
**   Under Section 7 of the GNU Affero GPL version 3, the following additional
**   terms apply to the works covered under the License.  These additional terms
**   are non-permissive additional terms allowed under Section 7 of the GNU
**   Affero GPL version 3 and may not be removed by you.
** 
**   0. Attribution Requirement.
** 
**     You must preserve all legal notices or author attributions in the covered
**     work or Appropriate Legal Notices displayed by works containing the covered
**     work.  You may not remove from the covered work any author or developer
**     credit already included within the covered work.
** 
**   1. No License To Use Trademarks.
** 
**     This license does not grant any license or rights to use the trademarks
**     Golden Code, FWD, any Golden Code or FWD logo, or any other trademarks
**     of Golden Code Development Corporation. You are not authorized to use the
**     name Golden Code, FWD, or the names of any author or contributor, for
**     publicity purposes without written authorization.
** 
**   2. No Misrepresentation of Affiliation.
** 
**     You may not represent yourself as Golden Code Development Corporation or FWD.
** 
**     You may not represent yourself for publicity purposes as associated with
**     Golden Code Development Corporation, FWD, or any author or contributor to
**     the covered work, without written authorization.
** 
**   3. No Misrepresentation of Source or Origin.
** 
**     You may not represent the covered work as solely your work.  All modified
**     versions of the covered work must be marked in a reasonable way to make it
**     clear that the modified work is not originating from Golden Code Development
**     Corporation or FWD.  All modified versions must contain the notices of
**     attribution required in this license.
*/

package com.goldencode.p2j.ui;

import java.io.*;

/**
 * This class represents widget identifiers for a widget part of a frame's down body. 
 */
public class WidgetDownId
extends WidgetId
{
   /** The location of the widget associated with this ID in the frame's down body. */
   private int downIdx = 1;

   /**
    * Default c'tor.  Not for public use, for de-serialization only.
    */
   public WidgetDownId()
   {
      // no-op
   }
   
   /**
    * Create a new widget ID which is linked to a widget located on the given row, in the 
    * frame's down body.
    * 
    * @param    id
    *           The widget ID.
    * @param    downIdx
    *           The index in the frame's down body of the widget associated with this ID.
    */
   public WidgetDownId(int id, int downIdx)
   {
      super(id);
      this.downIdx = downIdx;
   }

   /**
    * Create a new {@link WidgetDownId} instance which is assigned the given <code>widgetId</code>
    * and inherits this id's {@link #downIdx}.
    * 
    * @param    widgetId
    *           The new widgetId.
    *           
    * @return   See above.
    */
   @Override
   public WidgetId createId(int widgetId)
   {
      return new WidgetDownId(widgetId, downIdx); 
   }

   /**
    * Get the position in the frame's down body of this widget.
    * 
    * @return   the {@link #downIdx} value.
    */
   public int getDownIdx()
   {
      return downIdx;
   }

   
   /**
    * Replacement for the default object writing method. The latest state is written to the output
    * destination.
    * 
    * @param    out
    *           The output destination to which fields will be saved.
    *
    * @throws   IOException
    *           In case of I/O errors.
    */
   @Override
   public void writeExternal(ObjectOutput out) throws IOException
   {
      super.writeExternal(out);

      out.writeInt(downIdx);
   }

   /**
    * Replacement for the default object reading method. The latest state is read from the input 
    * source.
    * 
    * @param    in
    *           The input source from which fields will be restored.
    *
    * @throws   IOException
    *           In case of I/O errors.
    * @throws   ClassNotFoundException
    *           If payload can't be instantiated.
    */
   @Override
   public void readExternal(ObjectInput in)
   throws IOException,
          ClassNotFoundException
   {
      super.readExternal(in);
      
      downIdx = in.readInt();
   }
   
   /**
    * Check if the given object is the same as this one.
    * 
    * @param    obj
    *           The candidate.
    * 
    * @return   <code>true</code> if this instance represents the same widget ID as the given one.
    */
   @Override
   public boolean equals(Object obj)
   {
      if (obj instanceof WidgetDownId)
      {
         WidgetDownId id = (WidgetDownId) obj;
         
         return super.equals(obj) && id.downIdx == this.downIdx;
      }
      else
      {
         // this will refer all widgets in a down frame to their "first-row" equivalent.
         // this is required as on server-side there is only one widget associated, even if the 
         // widget is part of a down frame (so the client-side has more than one instance)
         return super.equals(obj);
      }
   }
   
   /**
    * Compute a hash code for this widget ID.
    * 
    * @return   See above.
    */
   @Override
   public int hashCode()
   {
      return super.hashCode() * 37 + downIdx;
   }
   
   /**
    * Get a string representation of this ID.
    * 
    * @return   See above.
    */
   @Override
   public String toString()
   {
      return super.toString() + ":" + Integer.toString(downIdx);
   }
   
   /**
    * Compare this instance with the given object.  Return value matches the 
    * {@link Comparable#compareTo}.
    *
    * @param    wid
    *           The widget ID to compare against.
    * 
    * @return   See {@link Comparable#compareTo}.
    */
   @Override
   public int compareTo(WidgetId wid)
   {
      return this.equals(wid) ? this.downIdx - ((WidgetDownId) wid).downIdx 
                              : super.compareTo(wid);
   }
}