/*
** Module   : WidgetListBase.java
** Abstract : Registers the widgets created as part of a container.
**
** Copyright (c) 2015-2017, Golden Code Development Corporation.
**
** -#- -I- --Date------------Description-----------------------------
** 001 VIG 20141129          Created initial version.
** 002 IAS 20151020          Added support for the background widgets.
*/
/*
** This program is free software: you can redistribute it and/or modify
** it under the terms of the GNU Affero General Public License as
** published by the Free Software Foundation, either version 3 of the
** License, or (at your option) any later version.
**
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
** GNU Affero General Public License for more details.
**
** You may find a copy of the GNU Affero GPL version 3 at the following
** location: https://www.gnu.org/licenses/agpl-3.0.en.html
** 
** Additional terms under GNU Affero GPL version 3 section 7:
** 
**   Under Section 7 of the GNU Affero GPL version 3, the following additional
**   terms apply to the works covered under the License.  These additional terms
**   are non-permissive additional terms allowed under Section 7 of the GNU
**   Affero GPL version 3 and may not be removed by you.
** 
**   0. Attribution Requirement.
** 
**     You must preserve all legal notices or author attributions in the covered
**     work or Appropriate Legal Notices displayed by works containing the covered
**     work.  You may not remove from the covered work any author or developer
**     credit already included within the covered work.
** 
**   1. No License To Use Trademarks.
** 
**     This license does not grant any license or rights to use the trademarks
**     Golden Code, FWD, any Golden Code or FWD logo, or any other trademarks
**     of Golden Code Development Corporation. You are not authorized to use the
**     name Golden Code, FWD, or the names of any author or contributor, for
**     publicity purposes without written authorization.
** 
**   2. No Misrepresentation of Affiliation.
** 
**     You may not represent yourself as Golden Code Development Corporation or FWD.
** 
**     You may not represent yourself for publicity purposes as associated with
**     Golden Code Development Corporation, FWD, or any author or contributor to
**     the covered work, without written authorization.
** 
**   3. No Misrepresentation of Source or Origin.
** 
**     You may not represent the covered work as solely your work.  All modified
**     versions of the covered work must be marked in a reasonable way to make it
**     clear that the modified work is not originating from Golden Code Development
**     Corporation or FWD.  All modified versions must contain the notices of
**     attribution required in this license.
*/

package com.goldencode.p2j.ui;

import java.util.*;
import java.util.stream.*;

/**
 * This class provides a simple registration means for the widgets inside a container.
 * 
 *  @param   F
 *           Container class.
 */
abstract class WidgetListBase<F>
implements Settable<F>
{
   /** List of regular widgets. */
   private Map<String, GenericWidget<?>> widgetMap = new LinkedHashMap<>();

   /** Converted name to original name mapping */
   private Map<String, String> nameMap = new LinkedHashMap<String, String>();

   /** List of header widgets. */
   private Map<String, GenericWidget<?>> widgetHeaderMap = new LinkedHashMap<>();
   
   /** List of background widgets. */
   private Map<String, GenericWidget<?>> backgroundWidgets =  new LinkedHashMap<>();
   
   /**
    * Registers a widget with the frame.
    *
    * @param    name
    *           widget name
    * @param    widget
    *           instance of a widget to register
    */
   public void addWidget(String name, GenericWidget<?> widget)
   {
      widgetMap.put(name, widget);
      nameMap.put(name, name);
   }

   /**
    * Registers a widget with the frame.
    *
    * @param    convName
    *           converted widget name
    * @param    origName
    *           original widget name
    * @param    widget
    *           instance of a widget to register
    */
   public void addWidget(String convName, String origName, GenericWidget<?> widget)
   {
      widgetMap.put(convName, widget);
      nameMap.put(convName, origName);
   }

   /**
    * Registers a background widget with the frame.
    *
    * @param    convName
    *           converted widget name
    * @param    origName
    *           original widget name
    * @param    widget
    *           instance of a widget to register
    */
   public void addBackgroundWidget(String convName, String origName, GenericWidget<?> widget)
   {
      widgetMap.put(convName, widget);
      backgroundWidgets.put(convName, widget);
      nameMap.put(convName, origName);
   }

   /**
    * Registers a header widget with the frame.
    *
    * @param    name
    *           widget name
    * @param    widget
    *           instance of a widget to register
    */
   public void addHeaderWidget(String name, GenericWidget<?> widget)
   {
      widgetHeaderMap.put(name, widget);
   }
   
   /**
    * Gets the map of widgets registered with the frame.
    * 
    * @return   widget map.
    */
   public Map<String, GenericWidget<?>> getWidgets()
   {
      return widgetMap;
   }

   /**
    * Gets the map of names of widgets registered with the frame.
    * 
    * @return   name map.
    */
   public Map<String, String> getNames()
   {
      return nameMap;
   }

   /**
    * Gets the map of header widgets registered with the frame.
    * 
    * @return   widget map.
    */
   public Map<String, GenericWidget<?>> getHeaderWidgets()
   {
      return widgetHeaderMap;
   }

   /**
    * Gets the map of background widgets registered with the frame.
    * 
    * @return   widget map.
    */
   public Map<String, GenericWidget<?>> getBackgroundWidgets()
   {
      return backgroundWidgets;
   }
}