/*
** Module   : WordWrapElement.java
** Abstract : represents the relationship between frame data and a widget
**
** Copyright (c) 2006-2017, Golden Code Development Corporation.
**
** -#- -I- --Date-- -T- --JPRM-- ----------------Description-----------------
** 001 GES 20060404 NEW   @25339 Created initial version to mark a specific
**                               widget as word wrap for the current editing
**                               operation.
** 002 GES 20060915 CHG   @29628 Added support for format strings.
*/ 
/*
** This program is free software: you can redistribute it and/or modify
** it under the terms of the GNU Affero General Public License as
** published by the Free Software Foundation, either version 3 of the
** License, or (at your option) any later version.
**
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
** GNU Affero General Public License for more details.
**
** You may find a copy of the GNU Affero GPL version 3 at the following
** location: https://www.gnu.org/licenses/agpl-3.0.en.html
** 
** Additional terms under GNU Affero GPL version 3 section 7:
** 
**   Under Section 7 of the GNU Affero GPL version 3, the following additional
**   terms apply to the works covered under the License.  These additional terms
**   are non-permissive additional terms allowed under Section 7 of the GNU
**   Affero GPL version 3 and may not be removed by you.
** 
**   0. Attribution Requirement.
** 
**     You must preserve all legal notices or author attributions in the covered
**     work or Appropriate Legal Notices displayed by works containing the covered
**     work.  You may not remove from the covered work any author or developer
**     credit already included within the covered work.
** 
**   1. No License To Use Trademarks.
** 
**     This license does not grant any license or rights to use the trademarks
**     Golden Code, FWD, any Golden Code or FWD logo, or any other trademarks
**     of Golden Code Development Corporation. You are not authorized to use the
**     name Golden Code, FWD, or the names of any author or contributor, for
**     publicity purposes without written authorization.
** 
**   2. No Misrepresentation of Affiliation.
** 
**     You may not represent yourself as Golden Code Development Corporation or FWD.
** 
**     You may not represent yourself for publicity purposes as associated with
**     Golden Code Development Corporation, FWD, or any author or contributor to
**     the covered work, without written authorization.
** 
**   3. No Misrepresentation of Source or Origin.
** 
**     You may not represent the covered work as solely your work.  All modified
**     versions of the covered work must be marked in a reasonable way to make it
**     clear that the modified work is not originating from Golden Code Development
**     Corporation or FWD.  All modified versions must contain the notices of
**     attribution required in this license.
*/

package com.goldencode.p2j.ui;

import com.goldencode.p2j.util.*;

/**
 * This class marks a specific widget as needing word wrap for the current
 * editing operation.
 * <p>
 * Each such element has a group index (by default this is 0). All elements
 * with the same group index are part of the same word wrapping group.
 */
public class WordWrapElement
extends Element
{
   /** Group index. */
   private int idx = 0;
   
   /** Is this a widget-only element? */
   private boolean widgetOnly = false;
   
   /**
    * Create a widget-only instance with a group index of 0.
    *
    * @param    widget
    *           The widget instance representing the exact frame field and
    *           state of that field.  Must NOT be <code>null</code>.
    */
   public WordWrapElement(GenericWidget widget)
   {
      this((Accessor) null, widget, 0);
   }
   
   /**
    * Create a widget-only instance with a given group index.
    *
    * @param    widget
    *           The widget instance representing the exact frame field and
    *           state of that field.  Must NOT be <code>null</code>.
    * @param    idx
    *           Group index.
    */
   public WordWrapElement(GenericWidget widget, int idx)
   {
      this((Accessor) null, widget, idx);
   }
   
   /**
    * Create an instance with a group index of 0.
    *
    * @param    data
    *           The data to display/edit in the given widget.  Must NOT be
    *           <code>null</code>.
    * @param    widget
    *           The widget instance representing the exact frame field and
    *           state of that field.  Must NOT be <code>null</code>.
    */
   public WordWrapElement(Accessor data, GenericWidget widget)
   {
      this(data, widget, 0);
   }
   
   /**
    * Create an instance with a group index of 0.
    *
    * @param    data
    *           The data to display/edit in the given widget.  Must NOT be
    *           <code>null</code>.
    * @param    widget
    *           The widget instance representing the exact frame field and
    *           state of that field.  Must NOT be <code>null</code>.
    */
   public WordWrapElement(BaseDataType data, GenericWidget widget)
   {
      this(data, widget, 0);
   }
   
   /**
    * Create an instance with the specified group index.
    *
    * @param    data
    *           The data to display/edit in the given widget.  Must NOT be
    *           <code>null</code>.
    * @param    widget
    *           The widget instance representing the exact frame field and
    *           state of that field.  Must NOT be <code>null</code>.
    * @param    idx
    *           Group index.
    */
   public WordWrapElement(Accessor data, GenericWidget widget, int idx)
   {
      super(data, widget);
      this.idx = idx; 
      widgetOnly = (data == null);
   }
   
   /**
    * Create an instance with the specified group index.
    *
    * @param    data
    *           The data to display/edit in the given widget.  Must NOT be
    *           <code>null</code>.
    * @param    widget
    *           The widget instance representing the exact frame field and
    *           state of that field.  Must NOT be <code>null</code>.
    * @param    idx
    *           Group index.
    */
   public WordWrapElement(BaseDataType data, GenericWidget widget, int idx)
   {
      super(data, widget);
      this.idx = idx;
      widgetOnly = (data == null);
   }
   
   /**
    * Create an instance with a group index of 0.
    *
    * @param    data
    *           The data to display/edit in the given widget.  Must NOT be
    *           <code>null</code>.
    * @param    fmt
    *           The format string to associate with the data value. May be
    *           <code>null</code>.
    * @param    widget
    *           The widget instance representing the exact frame field and
    *           state of that field.  Must NOT be <code>null</code>.
    */
   public WordWrapElement(Accessor data, String fmt, GenericWidget widget)
   {
      this(data, fmt, widget, 0);
   }
   
   /**
    * Create an instance with a group index of 0.
    *
    * @param    data
    *           The data to display/edit in the given widget.  Must NOT be
    *           <code>null</code>.
    * @param    fmt
    *           The format string to associate with the data value. May be
    *           <code>null</code>.
    * @param    widget
    *           The widget instance representing the exact frame field and
    *           state of that field.  Must NOT be <code>null</code>.
    */
   public WordWrapElement(BaseDataType data, String fmt, GenericWidget widget)
   {
      this(data, fmt, widget, 0);
   }
   
   /**
    * Create an instance with the specified group index.
    *
    * @param    data
    *           The data to display/edit in the given widget.  Must NOT be
    *           <code>null</code>.
    * @param    fmt
    *           The format string to associate with the data value. May be
    *           <code>null</code>.
    * @param    widget
    *           The widget instance representing the exact frame field and
    *           state of that field.  Must NOT be <code>null</code>.
    * @param    idx
    *           Group index.
    */
   public WordWrapElement(Accessor      data,
                          String        fmt,
                          GenericWidget widget,
                          int           idx)
   {
      super(data, fmt, widget);
      this.idx = idx; 
      widgetOnly = (data == null);
   }
   
   /**
    * Create an instance with the specified group index.
    *
    * @param    data
    *           The data to display/edit in the given widget.  Must NOT be
    *           <code>null</code>.
    * @param    fmt
    *           The format string to associate with the data value. May be
    *           <code>null</code>.
    * @param    widget
    *           The widget instance representing the exact frame field and
    *           state of that field.  Must NOT be <code>null</code>.
    * @param    idx
    *           Group index.
    */
   public WordWrapElement(BaseDataType  data,
                          String        fmt,
                          GenericWidget widget,
                          int           idx)
   {
      super(data, fmt, widget);
      this.idx = idx;
      widgetOnly = (data == null);
   }
   
   /**
    * Determines if the element should cause word wrap to be activated during
    * editing operations.
    *
    * @return   Always <code>true</code>;
    */
   public boolean wordWrap()
   {
      return true;
   }
   
   /**
    * Accesses the group index.
    *
    * @return   The group index.
    */
   public int getIndex() 
   {
      return idx;
   }      
   
   /**
    * Determines if the element represents a widget with no data (like a
    * button).
    *
    * @return   <code>true</code> if this widget has no data or is being used
    *           without its data properties.
    */
   public boolean widgetOnly()
   {
      return widgetOnly;
   }   
}