/*
 ** Module   : AbstractTerminalBuffer.java
 ** Abstract :
 **
 ** Copyright (c) 2011-2017, Golden Code Development Corporation.
 **
** -#- -I- --Date-- ---------------------------------Description----------------------------------
** 001 SIY 20110112 Created initial version
** 002 CA  20140805 Changes related to refactoring the OutputManager into UI-specific 
**                  implementations.
 */
/*
** This program is free software: you can redistribute it and/or modify
** it under the terms of the GNU Affero General Public License as
** published by the Free Software Foundation, either version 3 of the
** License, or (at your option) any later version.
**
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
** GNU Affero General Public License for more details.
**
** You may find a copy of the GNU Affero GPL version 3 at the following
** location: https://www.gnu.org/licenses/agpl-3.0.en.html
** 
** Additional terms under GNU Affero GPL version 3 section 7:
** 
**   Under Section 7 of the GNU Affero GPL version 3, the following additional
**   terms apply to the works covered under the License.  These additional terms
**   are non-permissive additional terms allowed under Section 7 of the GNU
**   Affero GPL version 3 and may not be removed by you.
** 
**   0. Attribution Requirement.
** 
**     You must preserve all legal notices or author attributions in the covered
**     work or Appropriate Legal Notices displayed by works containing the covered
**     work.  You may not remove from the covered work any author or developer
**     credit already included within the covered work.
** 
**   1. No License To Use Trademarks.
** 
**     This license does not grant any license or rights to use the trademarks
**     Golden Code, FWD, any Golden Code or FWD logo, or any other trademarks
**     of Golden Code Development Corporation. You are not authorized to use the
**     name Golden Code, FWD, or the names of any author or contributor, for
**     publicity purposes without written authorization.
** 
**   2. No Misrepresentation of Affiliation.
** 
**     You may not represent yourself as Golden Code Development Corporation or FWD.
** 
**     You may not represent yourself for publicity purposes as associated with
**     Golden Code Development Corporation, FWD, or any author or contributor to
**     the covered work, without written authorization.
** 
**   3. No Misrepresentation of Source or Origin.
** 
**     You may not represent the covered work as solely your work.  All modified
**     versions of the covered work must be marked in a reasonable way to make it
**     clear that the modified work is not originating from Golden Code Development
**     Corporation or FWD.  All modified versions must contain the notices of
**     attribution required in this license.
*/

package com.goldencode.p2j.ui.chui;

import com.goldencode.p2j.ui.*;
import com.goldencode.p2j.ui.client.*;
import com.goldencode.p2j.ui.client.chui.driver.*;

/**
 * This class is a base with most possible common code for implementations
 * which hold internal screen buffer for drawing operations.
 *
 * @author 	SIY
 * @version 1.0
 */
public abstract class AbstractTerminalBuffer
extends BasePrimitives
{
   /** Number of columns on the screen. */
   protected final int cols;

   /** Number of lines on the screen. */
   protected final int rows;

   /** Current rendering buffer. */
   private final Cell[][] render;

   /** The current drawing X coordinate of the cursor. */
   protected int cx = 0;

   /** The current drawing Y coordinate of the cursor. */
   protected int cy = 0;

   /** The current stay X coordinate of the cursor. */
   protected int csx = 0;

   /** The current stay Y coordinate of the cursor. */
   protected int csy = 0;

   /** The default clipping bitmap. */
   private ScreenBitmap bitmap;

   /**
    * Constructor.
    *
    * @param   rows
    *          Number of rows in buffer.
    * @param   cols
    *          Number of columns in buffer.
    */
   public AbstractTerminalBuffer(int rows, int cols)
   {
      this.rows = rows;
      this.cols = cols;
      bitmap = new ScreenBitmap(rows, cols);
      render = new Cell[rows][cols];

      fillScreen(EMPTY_CELL);
   }

   /**
    * Add a string to the virtual terminal at the current cursor position.
    *
    * @param    str
    *           The text to output.
    * @param    color
    *           The color to use
    */
   @Override
   public void append(String str, Color color)
   {
      for (int i = 0; i < str.length(); i++)
      {
         append(str.charAt(i), color);
      }
   }

   /**
    * Draw a horizontal line of the specified length starting at the
    * current cursor position.
    *
    * @param    length
    *           The length of the line to draw in columns.
    * @param    color
    *           The color to use
    */
   @Override
   public void hLine(int length, Color color)
   {
      for (int i = 0; i < length; i++)
      {
         append(ChuiOutputManager.UACS_HLINE, color);
      }
   }

   /**
    * Draw a vertical line of the specified length starting at the current
    * cursor position.
    *
    * @param    length
    *           The length of the line to draw in rows.
    * @param    color
    *           The color to use
    */
   @Override
   public void vLine(int length, Color color)
   {
      int cxNow = cx;
      int cyNow = cy;

      for (int i = 0; i < length; i++)
      {
         append(ChuiOutputManager.UACS_VLINE, color);
         cursorAt(cxNow, cyNow + i + 1);
      }
   }

   /**
    * Blank out a box using the specified attributes and color pair.
    *
    * @param    left
    *           The left column of the rectangle.
    * @param    top
    *           The top row of the rectangle.
    * @param    right
    *           The right column of the rectangle.
    * @param    bottom
    *           The bottom row of the rectangle.
    * @param    color
    *           The color to use.
    */
   @Override
   public void clearArea(int left, int top, int right,
                         int bottom, Color color)
   {
      for (int i = top; i <= bottom; i++)
      {
         cursorAt(left, i);
         for (int j = left; j <= right; j++)
         {
            append(' ', color);
         }
      }
   }

   /**
    * Draw a box using the specified dimensions and color pair.
    *
    * @param    left
    *           The left column of the rectangle.
    * @param    top
    *           The top row of the rectangle.
    * @param    right
    *           The right column of the rectangle.
    * @param    bottom
    *           The bottom row of the rectangle.
    * @param    color
    *           The color to use
    */
   @Override
   public void box(int left, int top, int right, int bottom,
                      Color color)
   {
      drawULCorner(left,  top, color);
      hLine(right - left - 1, color);
      drawURCorner(right, top, color);

      drawLLCorner(left,  bottom, color);
      hLine(right - left - 1, color);
      drawLRCorner(right, bottom, color);

      drawVBorder(left,  top + 1, bottom - top - 1, color);
      drawVBorder(right, top + 1, bottom - top - 1, color);
   }

   /**
    * Get absolute X position of the cursor.
    *
    * @return   The X position of the cursor (0-based column).
    */
   @Override
   public int getX()
   {
      return cx;
   }

   /**
    * Get absolute Y position of the cursor.
    *
    * @return   The Y position of the cursor (0-based row).
    */
   @Override
   public int getY()
   {
      return cy;
   }

   /**
    * Position the virtual cursor during the drawing.
    *
    * @param    x
    *           column
    *
    * @param    y
    *           row
    *
    * @return   <code>true</code> if cursor positioning succeeded.
    */
   @Override
   public boolean cursorAt(int x, int y)
   {
      // check the new position for validity
      if (x < 0 || x >= cols || y < 0 || y >= rows)
      {
         // reject the request
         return false;
      }

      // accept the request and relocate the cursor
      cx = x;
      cy = y;

      return true;
   }

   /**
    * Get the screen bitmap for this output primitives implementation.
    *
    * @return   The screen bitmap.
    */
   @Override
   public ScreenBitmap getScreenBitmap()
   {
      return bitmap;
   }

   /**
    * Replace the current bitmap instance with this one.
    *
    * @param    bitmap
    *           The new bitmap instance.
    */
   @Override
   public void setScreenBitmap(ScreenBitmap bitmap)
   {
      this.bitmap = bitmap;
   }

   /**
    * Draw the upper-left border corner.
    *
    * @param    col
    *           The column number where to draw.
    * @param    row
    *           The row number where to draw.
    * @param    color
    *           The number of the color-pair (foreground+background) to use
    *           for blanking the rectangle.
    */
   @Override
   public void drawULCorner(int col, int row, Color color)
   {
      if (!cursorAt(col, row))
      {
         return;
      }

      append(ChuiOutputManager.UACS_ULCORNER, color);
   }

   /**
    * Draw the upper-right border corner.
    *
    * @param    col
    *           The column number where to draw.
    * @param    row
    *           The row number where to draw.
    * @param    color
    *           The number of the color-pair (foreground+background) to use
    *           for blanking the rectangle.
    */
   @Override
   public void drawURCorner(int col, int row, Color color)
   {
      if (!cursorAt(col, row))
      {
         return;
      }

      append(ChuiOutputManager.UACS_URCORNER, color);
   }

   /**
    * Draw the lower-left border corner.
    *
    * @param    col
    *           The column number where to draw.
    * @param    row
    *           The row number where to draw.
    * @param    color
    *           The number of the color-pair (foreground+background) to use
    *           for blanking the rectangle.
    */
   @Override
   public void drawLLCorner(int col, int row, Color color)
   {
      if (!cursorAt(col, row))
      {
         return;
      }

      append(ChuiOutputManager.UACS_LLCORNER, color);
   }

   /**
    * Draw the lower-right border corner.
    *
    * @param    col
    *           The column number where to draw.
    * @param    row
    *           The row number where to draw.
    * @param    color
    *           The number of the color-pair (foreground+background) to use
    *           for blanking the rectangle.
    */
   @Override
   public void drawLRCorner(int col, int row, Color color)
   {
      if (!cursorAt(col, row))
      {
         return;
      }

      append(ChuiOutputManager.UACS_LRCORNER, color);
   }

   /**
    * Draw the upper or lower border.
    *
    * @param    col
    *           The column number where to draw.
    * @param    row
    *           The row number where to draw.
    * @param    width
    *           The border width.
    * @param    color
    *           The number of the color-pair (foreground+background)
    */
   @Override
   public void drawHBorder(int col, int row, int width,
                              Color color)
   {
      if (!cursorAt(col, row))
      {
         return;
      }

      hLine(width, color);
   }

   /**
    * Draw the left or right border.
    *
    * @param    col
    *           The column number where to draw.
    * @param    row
    *           The row number where to draw.
    * @param    height
    *           The border height.
    * @param    color
    *           The number of the color-pair (foreground+background)
    */
   @Override
   public void drawVBorder(int col, int row, int height,
                              Color color)
   {
      if (!cursorAt(col, row))
      {
         return;
      }

      vLine(height, color);
   }

   /**
    * Fill entire terminal screen with specified cell.
    *
    * @param   cell
    *          Source cell.
    */
   private void fillScreen(Cell cell)
   {
      int i = 0;
      int j = 0;

      // get the video color and attribute to clear screen

      // initialize the render buffer, so master and render are different now

      for (i = 0; i < rows; i ++)
      {
         for (j = 0; j < cols; j ++)
            render[i][j] = cell;
      }
   }

   /**
    * Get access to rendering buffer.
    *
    * @return rendering buffer.
    */
   protected Cell[][] render()
   {
      return render;
   }

   /**
    * Advance the cursor to the next screen position with wrapping around
    * the line and and the screen end.
    */
   protected void advanceCursor()
   {
      // advance the cursor with the single wrapping (no wrapping home)
      if (cx < cols - 1)
      {
         cx++;
      }
      else
      {
         cx = 0;
         if (cy < rows - 1)
         {
            cy++;
         }
      }
   }

   /**
    * Position the real cursor.
    *
    * @param    x
    *           Cursor X.
    * @param    y
    *           Cursor Y.
    *
    * @return   <code>true</code> if cursor positioning succeeded.
    */
   @Override
   public boolean cursorStay(int x, int y)
   {
      // check the new position for validity
      if (x < 0 || x >= cols || y < 0 || y >= rows)
      {
         // reject the request
         return false;
      }

      // accept the request and relocate the cursor
      if (csx != x || csy != y)
      {
         csx = x;
         csy = y;
         cursorRelocated(csx, csy);
      }

      cx = csx;
      cy = csy;
      return true;
   }

   /**
    * Callback method when cursor is actually relocated at given position.
    *
    * @param   csx
    *          Cursor X.
    * @param   csy
    *          Cursor Y.
    */
   protected void cursorRelocated(int csx, int csy)
   {
   }

   /**
    * Gets the number of columns on the screen.
    *
    * @return   The number of columns (in characters).
    */
   @Override
   public int screenWidth()
   {
      return cols;
   }

   /**
    * Gets the number of rows on the screen.
    *
    * @return   The number of rows (lines of characters).
    */
   @Override
   public int screenHeight()
   {
      return rows;
   }

}