/*
** Module   : ComboBoxImpl.java
** Abstract : chui implementation of the combo-box
**
** Copyright (c) 2011-2017, Golden Code Development Corporation.
**
** -#- -I- --Date-- ---------------------------------Description---------------------------------
** 001 SIY 20111122 Extracted from ComboBox.
** 002 GES 20140627 Changed color attribute method names to match naming standards.
** 003 CA  20140805 Do not allow ChUI impl classes access to APIs outside of ChuiOutputManager.
** 004 HC  20140911 GUI coordinate support - data types refactoring.
** 005 CA  20140926 Added shared widget configuration support, to allow GUI/ChUI concrete
**                  implementation for the same widget ID.  Refactored the widget configuration
**                  classes: all fields were made public; these classes need to be as dumb as
**                  possible, all logic related to setting/getting a certain field should be at
**                  the widget or at the caller. Refs #2254
** 006 GES 20150225 Fixed the drawing of the drop-down indicator (and the space char) to use the
**                  correct colors (normal for the space and reverse for the indicator). The
**                  4GL always seems to draw it this way, whether enabled or not. I don't
**                  understand how we missed this in the past.
** 007 HC  20150323 Removed business logic from config classes.
** 008 CA  20150603 Removed requestSync from drawCaret.
** 009 IAS 20150924 Added AUTO-RESIZE attribute support
** 010 EVL 20151217 Drop-down approach rework.  Separation ChUI specific drop-down registration
**                  and focus gaining.
** 011 OM  20170717 ChUI comboboxes take into account the display format set.
*/
/*
** This program is free software: you can redistribute it and/or modify
** it under the terms of the GNU Affero General Public License as
** published by the Free Software Foundation, either version 3 of the
** License, or (at your option) any later version.
**
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
** GNU Affero General Public License for more details.
**
** You may find a copy of the GNU Affero GPL version 3 at the following
** location: https://www.gnu.org/licenses/agpl-3.0.en.html
** 
** Additional terms under GNU Affero GPL version 3 section 7:
** 
**   Under Section 7 of the GNU Affero GPL version 3, the following additional
**   terms apply to the works covered under the License.  These additional terms
**   are non-permissive additional terms allowed under Section 7 of the GNU
**   Affero GPL version 3 and may not be removed by you.
** 
**   0. Attribution Requirement.
** 
**     You must preserve all legal notices or author attributions in the covered
**     work or Appropriate Legal Notices displayed by works containing the covered
**     work.  You may not remove from the covered work any author or developer
**     credit already included within the covered work.
** 
**   1. No License To Use Trademarks.
** 
**     This license does not grant any license or rights to use the trademarks
**     Golden Code, FWD, any Golden Code or FWD logo, or any other trademarks
**     of Golden Code Development Corporation. You are not authorized to use the
**     name Golden Code, FWD, or the names of any author or contributor, for
**     publicity purposes without written authorization.
** 
**   2. No Misrepresentation of Affiliation.
** 
**     You may not represent yourself as Golden Code Development Corporation or FWD.
** 
**     You may not represent yourself for publicity purposes as associated with
**     Golden Code Development Corporation, FWD, or any author or contributor to
**     the covered work, without written authorization.
** 
**   3. No Misrepresentation of Source or Origin.
** 
**     You may not represent the covered work as solely your work.  All modified
**     versions of the covered work must be marked in a reasonable way to make it
**     clear that the modified work is not originating from Golden Code Development
**     Corporation or FWD.  All modified versions must contain the notices of
**     attribution required in this license.
*/

package com.goldencode.p2j.ui.chui;

import com.goldencode.p2j.ui.*;
import com.goldencode.p2j.ui.client.*;
import com.goldencode.p2j.ui.client.chui.driver.*;
import com.goldencode.p2j.ui.client.format.*;
import com.goldencode.p2j.ui.client.model.*;
import com.goldencode.p2j.ui.client.widget.*;
import com.goldencode.p2j.util.*;
import com.goldencode.util.*;

/**
 * Specific ComboBox implementation for CHUI.
 */
public class ComboBoxImpl
extends ComboBox<ChuiOutputManager>
implements AutoResizable<ComboBoxConfig>
{
   /**
    * Special constructor used to restore the component using a given
    * configuration.
    * 
    * @param    id
    *           The widget's ID.
    * @param    cfg
    *           Configuration data.
    */
   public ComboBoxImpl(WidgetId id, ComboBoxConfig cfg)
   {
      super(id, cfg);
   }
   
   /**
    * Draws the combo box as a list item.
    */
   @Override
   public void draw()
   {
      int selIdx = model().getSelectedIndex();
      String display = null;
      if (selIdx >= 0 && selIdx < model().size())
      {
         display = model().get(selIdx); // model().selected();
      }
      if (display == null)
      {
         display = "";
      }
      
      // limit the data portion of the field to the maximum allowed size
      // in chui char coordinates are always whole numbers
      if (display.length() > (int) limit)
      {
         display = display.substring(0, (int) limit);
      }
      else if (display.length() < (int) limit)
      {
         display = StringHelper.leftAlignText(display, (int) limit);
      }
      
      Color dc = UiUtils.getDColor(this);
      if (focusTraversable())
      {
         dc = dc.withUnderline();
      }
      
      screen().at(screenLocation())
              .append(display, dc)
              .at(screenLocation().offset(display.length(), 0))
              .append(' ', Color.NORMAL)
              .append("[V]", Color.MESSAGES);
      
      if (hasFocus() && isEnabled())
      {
         drawCaret();
      }
   }
   
   /**
    * Get widget screen value.
    *
    * @return  Current widget screen value.
    */
   @Override
   public BaseDataType getScreenValue()
   {
      ControlSetItem[] items = config.items;
      ComboBoxModel<String> model = model();
      int ndx = model.getSelectedIndex();
      
      if (ndx >= 0 && ndx < items.length)
      {
         return model.getSelectedValue();
      }
      
      // invalid selection? return new an unknown value of the appropriate type
      BaseDataType var = DisplayFormat.instanceOfType(config.dataType);
      var.setUnknown();
      return var;
   }
   
   /**
    * Sets position of terminal's cursor to the first char of the combo-box.
    */
   @Override
   public void drawCaret()
   {
      // this check is required since the method can be called from multiple drivers
      if (!hasFocus() || !isEnabled())
      {
         return;
      }

      Point origin = screenLocation();
      screen().at(origin);
   }

   /**
    * Obtain the input focus.
    */
   @Override
   public void requestFocus()
   {
      if (isEnabled())
      {
         // generate the FOCUS_GAINED event
         super.requestFocus();

         screen().at(screenLocation());
      }
   }
   /**
    * Update widget's size after config update
    * 
    * @param beforeUpdate
    *        previous value of the widget's config
    */
   @Override
   public void updateSize(ComboBoxConfig beforeUpdate)
   {
      if (!StringHelper.equal(config.format, beforeUpdate.format))
      {
         config.widthChars = 0;
         limit = calcLimit(false);
         config.widthChars =  width();
         config.heightChars = height();
         if (config.widthChars < beforeUpdate.widthChars)
         {
            parent().repaint();
         }
      }
   }

   /**
    * Get the AUTO-RESIZE status
    * 
    * @return <code>true</code> is AUTO-RESIZE is enabled
    */
   @Override
   public boolean isAutoResize()
   {
      return config.autoResize;
   }

   /**
    * Registers drop-down frame in current window system.  ChUI specific version.
    */
   @Override
   protected void registerDropDownFrame()
   {
       TopLevelWindow<?> window = ComboBoxImpl.this.topLevelWindow();
       window.addFrame((Widget) dropDown, 0, 0);
       window.moveToTop((Widget) dropDown);
   }

   /**
    * Requesting focus to opened drop-down.
    */
   @Override
   protected void dropDownRequestFocus()
   {
      Widget<ChuiOutputManager> focus = dropDown.defaultFocus();
      focus.requestFocus();
   }
}
