/*
** Module   : SelectionListImpl.java
** Abstract :
**
** Copyright (c) 2011-2017, Golden Code Development Corporation.
**
** -#- -I- --Date-- ---------------------------------Description----------------------------------
** 001 SIY 20111202 Extracted from SelectionList (former List).
** 002 CA  20140805 Do not allow ChUI impl classes access to APIs outside of ChuiOutputManager.
** 003 HC  20140911 GUI coordinate support - data types refactoring.
** 004 CA  20140926 Added shared widget configuration support, to allow GUI/ChUI concrete
**                  implementation for the same widget ID.  Refactored the widget configuration
**                  classes: all fields were made public; these classes need to be as dumb as
**                  possible, all logic related to setting/getting a certain field should be at
**                  the widget or at the caller. Refs #2254
** 005 HC  20150323 Removed business logic from config classes.
** 006 IAS 20150801 Fixed bug with the trigger activation on focus lost.
** 007 EVL 20150820 Separation of ChUI specific code.
** 008 HC  20150831 Scrolling support overhaul.
** 009 SBI 20180503 Changed adjustSizeImpl, added validateLayout(SelectionListConfig).
*/
/*
** This program is free software: you can redistribute it and/or modify
** it under the terms of the GNU Affero General Public License as
** published by the Free Software Foundation, either version 3 of the
** License, or (at your option) any later version.
**
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
** GNU Affero General Public License for more details.
**
** You may find a copy of the GNU Affero GPL version 3 at the following
** location: https://www.gnu.org/licenses/agpl-3.0.en.html
** 
** Additional terms under GNU Affero GPL version 3 section 7:
** 
**   Under Section 7 of the GNU Affero GPL version 3, the following additional
**   terms apply to the works covered under the License.  These additional terms
**   are non-permissive additional terms allowed under Section 7 of the GNU
**   Affero GPL version 3 and may not be removed by you.
** 
**   0. Attribution Requirement.
** 
**     You must preserve all legal notices or author attributions in the covered
**     work or Appropriate Legal Notices displayed by works containing the covered
**     work.  You may not remove from the covered work any author or developer
**     credit already included within the covered work.
** 
**   1. No License To Use Trademarks.
** 
**     This license does not grant any license or rights to use the trademarks
**     Golden Code, FWD, any Golden Code or FWD logo, or any other trademarks
**     of Golden Code Development Corporation. You are not authorized to use the
**     name Golden Code, FWD, or the names of any author or contributor, for
**     publicity purposes without written authorization.
** 
**   2. No Misrepresentation of Affiliation.
** 
**     You may not represent yourself as Golden Code Development Corporation or FWD.
** 
**     You may not represent yourself for publicity purposes as associated with
**     Golden Code Development Corporation, FWD, or any author or contributor to
**     the covered work, without written authorization.
** 
**   3. No Misrepresentation of Source or Origin.
** 
**     You may not represent the covered work as solely your work.  All modified
**     versions of the covered work must be marked in a reasonable way to make it
**     clear that the modified work is not originating from Golden Code Development
**     Corporation or FWD.  All modified versions must contain the notices of
**     attribution required in this license.
*/

package com.goldencode.p2j.ui.chui;

import com.goldencode.p2j.ui.*;
import com.goldencode.p2j.ui.client.*;
import com.goldencode.p2j.ui.client.border.*;
import com.goldencode.p2j.ui.client.chui.driver.*;
import com.goldencode.p2j.util.*;

/**
 * CHUI-specific implementation of SelectionList
 */
public class SelectionListImpl
extends SelectionList<ChuiOutputManager>
{
   /**
    * Special constructor used to restore component using given configuration.
    * 
    * @param   id
    *          The widget's ID.
    * @param   cfg
    *          Configuration data.
    */
   public SelectionListImpl(WidgetId id, SelectionListConfig cfg)
   {
      super(id, cfg);
   }
   
   /**
    * Draw all the components in this container.
    */
   @Override
   public void draw() 
   {
      if (!isVisible())
         return;

      ThinClient tc = ThinClient.getInstance();

      if (tc.viewRefresh)
      {
         // if whole screen needs refresh, draw the decorations as well
         scrollpane.draw();
      }
      else if (tc.viewDisplay)
      {
         // only data need refresh, redraw the content
         scrollpane.viewport().draw();
      }
   }

   /**
    * Select next focusable widget.
    */
   @Override
   public void nextFocus()
   {
      ThinClient.nextFocus(this);
   }

   /**
    * Select previous focusable widget.
    */
   @Override
   public void prevFocus()
   {
      ThinClient.previousFocus(this);
   }

   /**
    * Adjust SelectionList size ChUI specific part.
    * 
    * @param    beforeUpdate
    *           The previous configuration values
    */
   @Override
   protected void adjustSizeImpl(SelectionListConfig beforeUpdate)
   {
      Border<ChuiOutputManager> borderImpl = scrollpane.getBorder();
      Insets border = borderImpl.insets();
      
      if (config.widthChars != 0 && config.heightChars != 0)
      {
         // in chui char coordinates are always whole numbers      
         config.innerChars = (int) (config.widthChars - border.widthDelta() - 2);
         config.innerLines = (int) (config.heightChars - border.heightDelta());
      }
      
      // size including borders
      Dimension outerSize = new Dimension(config.innerChars + border.widthDelta() + 2, 
                                          config.innerLines + border.heightDelta());
      setSize(outerSize);
      scrollpane.setSize(outerSize);
      list.setColumns(config.innerChars);
      list.setVisibleRows(config.innerLines);
   }

   /**
    * Validate its own layout.
    * 
    * @param    beforeUpdate
    *           The previous configuration values
    */
   @Override
   protected void validateLayout(SelectionListConfig beforeUpdate)
   {
      if (beforeUpdate.innerLines != config.innerLines)
      {
         Frame<ChuiOutputManager> frame = getFrame();
         
         if (frame == null)
         {
            return;
         }
         
         FrameConfig frameConfig = frame.config();
         
         int limit = (int)(frameConfig.heightChars - Math.max(config.row, 0));
         
         if (frameConfig.box)
         {
            limit -= 2;
         }
         
         if (config.initialInnerLines > 0 ? config.innerLines > config.initialInnerLines :
                                            config.innerLines > limit)
         {
            if (config.dynamic)
            {
               ErrorManager.recordOrShowError(4474, String
                        .format("New height for %s %s is too large to fit in frame", getType(),
                                 getName()), false);
            }
            
            String str = frameConfig.name;
            ErrorManager.recordOrShowError(4041, String.format(
                     "%s %s does not fit in parent FRAME %s", getType(), getName(),
                     str == null ? "" : str.substring(0, str.indexOf('#'))), false);
         }
      }
   }
}