/*
** Module   : TraceableTerminal.java
** Abstract : an API for rendering terminal output with call tracing option
**
** Copyright (c) 2006-2017, Golden Code Development Corporation.
**
** -#- -I- --Date-- ---------------------------------Description----------------------------------
** 001 NVS 20060705   @27878 A plugin which provides a compatible back-
**                           end for terminal output using the default
**                           ncurses backend, but also provides tracing of
**                           primitives into a file.
** 002 CA  20081114   @40448 All output primitives implementations will
**                           always draw data. Clipping responsibility
**                           belongs to the Toolkit class.
** 003 NVS 20090209   @41278 Due to changes in OutputDirectives interface
**                           added setCursorStay() and refresh() methods.
** 004 SIY 20090309   @41541 Added two stubs in order to satisfy interface
**                           definition.
** 005 SIY 20090319   @41641 Changed addArray() signature.
** 006 SIY 20090911   @43891 Now class is inherited from BasePrimitives.
** 007 SIY 20100619          Switching to own implementation of character UI.
** 008 SIY 20100804          Changes in imports, some refactorings.
** 009 SIY 20110112          Changes related to new color handling scheme.
** 010 CA  20140805          Changes related to OutputManager refactoring.
** 011 HC  20140911          GUI coordinate support - data types refactoring.
*/
/*
** This program is free software: you can redistribute it and/or modify
** it under the terms of the GNU Affero General Public License as
** published by the Free Software Foundation, either version 3 of the
** License, or (at your option) any later version.
**
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
** GNU Affero General Public License for more details.
**
** You may find a copy of the GNU Affero GPL version 3 at the following
** location: https://www.gnu.org/licenses/agpl-3.0.en.html
** 
** Additional terms under GNU Affero GPL version 3 section 7:
** 
**   Under Section 7 of the GNU Affero GPL version 3, the following additional
**   terms apply to the works covered under the License.  These additional terms
**   are non-permissive additional terms allowed under Section 7 of the GNU
**   Affero GPL version 3 and may not be removed by you.
** 
**   0. Attribution Requirement.
** 
**     You must preserve all legal notices or author attributions in the covered
**     work or Appropriate Legal Notices displayed by works containing the covered
**     work.  You may not remove from the covered work any author or developer
**     credit already included within the covered work.
** 
**   1. No License To Use Trademarks.
** 
**     This license does not grant any license or rights to use the trademarks
**     Golden Code, FWD, any Golden Code or FWD logo, or any other trademarks
**     of Golden Code Development Corporation. You are not authorized to use the
**     name Golden Code, FWD, or the names of any author or contributor, for
**     publicity purposes without written authorization.
** 
**   2. No Misrepresentation of Affiliation.
** 
**     You may not represent yourself as Golden Code Development Corporation or FWD.
** 
**     You may not represent yourself for publicity purposes as associated with
**     Golden Code Development Corporation, FWD, or any author or contributor to
**     the covered work, without written authorization.
** 
**   3. No Misrepresentation of Source or Origin.
** 
**     You may not represent the covered work as solely your work.  All modified
**     versions of the covered work must be marked in a reasonable way to make it
**     clear that the modified work is not originating from Golden Code Development
**     Corporation or FWD.  All modified versions must contain the notices of
**     attribution required in this license.
*/

package com.goldencode.p2j.ui.chui;

import com.goldencode.p2j.ui.*;
import com.goldencode.p2j.ui.client.*;
import com.goldencode.p2j.ui.client.chui.driver.*;

/**
 * This implementation can be used to trace all output calls for testing and/or
 * debugging purposes.
 */
public class TraceableTerminal
extends BasePrimitives
{
   /** default plugin */
   BasePrimitives std = null;
   
   /**
    * Creates an instance.
    */
   public TraceableTerminal()
   {
   }
   
   /**
    * Controls the tracing of the primitives.
    *
    * @param    tracing
    *           <code>true</code> turns tracing on
    */
   public void trace(boolean tracing)
   {
      if (tracing)
         std = ((ChuiOutputManager) OutputManager.instance()).switchOutput(this);
      else
         std = ((ChuiOutputManager) OutputManager.instance()).switchOutput(std);
   }

   /**
    * Add a character to the virtual terminal at the current cursor position.
    *
    * @param    chr
    *           The character to output.
    * @param    color
    *           Color.
    */
   @Override
   public void append(int chr, Color color)
   {
      cp();
      tn("addChar <" + chr + "> " + color);

      std.append(chr, color);
   }
   
   /**
    * Add a string to the virtual terminal at the current cursor position.
    *
    * @param    str
    *           The text to output.
    * @param    color
    *           Color.
    */
   @Override
   public void append(String str, Color color)
   {
      cp();
      tn("addString <" + str + "> " + color);

      std.append(str, color);
   }
   
   /**
    * Draw a horizontal line of the specified length starting at the current
    * cursor position.  Always uses a '-' to draw.
    *
    * @param    length
    *           The length of the line to draw in columns.
    * @param    color
    *           Color.
    */
   @Override
   public void hLine(int length, Color color)
   {
      cp();
      tn("addHLine " + length + " " + color);

      std.hLine(length, color);
   }

   /**
    * Draw a vertical line of the specified length starting at the current
    * cursor position.  Always uses a ' ' (space character) to draw.
    *
    * @param    length
    *           The length of the line to draw in rows.
    * @param    color
    *           Color.
    */
   @Override
   public void vLine(int length, Color color)
   {
      cp();
      tn("addVLine " + length + " " + color);

      std.vLine(length, color);
   }
   
   /**
    * Blank out a box using the specified attributes and color pair.
    *
    * @param    left
    *           The left column of the rectangle.
    * @param    top
    *           The top row of the rectangle.
    * @param    right
    *           The right column of the rectangle.
    * @param    bottom
    *           The bottom row of the rectangle.
    * @param    color
    *           Color.
    */
   @Override
   public void clearArea(int left, int top, int right, int bottom, Color color)
   {
      cp();
      tn("blankBox " + left + "," + top + "," + right + "," + bottom + " "
         + color);

      std.clearArea(left, top, right, bottom, color);
   }
   
   /**
    * Draw a box using the specified dimensions and color pair.
    *
    * @param    left
    *           The left column of the rectangle.
    * @param    top
    *           The top row of the rectangle.
    * @param    right
    *           The right column of the rectangle.
    * @param    bottom
    *           The bottom row of the rectangle.
    * @param    color
    *           Color.
    */
   @Override
   public void box(int left, int top, int right, int bottom, Color color)
   {
      cp();
      tn("drawBox " + left + "," + top + "," + right + "," + bottom + " "
         + color);

      std.box(left, top, right, bottom, color);
   }
   
   /**
    * Get absolute X position of the cursor.
    *
    * @return   The X position of the cursor.
    */
   @Override
   public int getX()
   {
      cp();
      tn("getX");

      return std.getX();
   }
   
   /**
    * Get absolute Y position of the cursor.
    *
    * @return   The Y position of the cursor.
    */
   @Override
   public int getY()
   {
      cp();
      tn("getY");

      return std.getY();
   }
   
   /**
    * Position the virtual cursor during the drawing.
    * 
    * @param     x
    *            Cursor X coordinate. 
    * @param     y 
    *            Cursor Y coordinate.
    *             
    * @return    <code>true</code> if cursor positioning succeeded.
    */
   @Override
   public boolean cursorAt(int x, int y)
   {
      cp();
      tn("setCursor " + x + "," + y);

      return std.cursorAt(x, y);
   }
   
   /**
    * Position the real cursor.
    * 
    * @param     x
    *            Cursor X coordinate. 
    * @param     y 
    *            Cursor Y coordinate.
    *             
    * @return    <code>true</code> if cursor positioning succeeded.
    */
   @Override
   public boolean cursorStay(int x, int y)
   {
      cp();
      tn("setCursorStay " + x + "," + y);

      return std.cursorAt(x, y);
   }

   /**
    * Mark a rectangle as needing repainting.
    * Makes sense with the double buffered output only; ignored here.
    * 
    * @param    rect
    *           rectangle to mark for repainting
    */
   @Override
   public void refresh(NativeRectangle rect)
   {
   }

   /**
    * Gets the number of columns on the screen.
    *
    * @return   The number of columns (in characters).
    */
   @Override
   public int screenWidth()
   {
      cp();
      tn("getScreenColumns");

      return std.screenWidth();
   }

   /**
    * Gets the number of rows on the screen.
    *
    * @return   The number of rows (lines of characters).
    */
   @Override
   public int screenHeight()
   {
      cp();
      tn("getScreenRows");

      return std.screenHeight();
   }
   
   /**
    * Synchronize the state of the physical terminal with the state of
    * the virtual terminal.
    */
   @Override
   public void sync()   
   {
      cp();
      tn("sync");
      
      std.sync();   
   }
   
   /**
    * Prints a string with new line.
    *
    * @param    s
    *           string to print
    */
   private void tn(String s)
   {
      System.err.println(s);
   }
   
   /**
    * Prints cursor position.
    */
   private void cp()
   {
      System.err.print(std.getX() + "," + std.getY() + " ");
   }

   /**
    * Get the screen bitmap for this output primitives implementation. 
    * 
    * @return   The screen bitmap.
    */
   @Override
   public ScreenBitmap getScreenBitmap()
   {
      return std.getScreenBitmap();
   }
   
   /**
    * Replace the current bitmap instance with this one.
    * 
    * @param    bitmap
    *           The new bitmap instance.
    */
   @Override
   public void setScreenBitmap(ScreenBitmap bitmap)
   {
      std.setScreenBitmap(bitmap);
   }

   /**
    * Draw the upper-left border corner.
    *  
    * @param    col 
    *           The column number where to draw.
    * @param    row 
    *           The row number where to draw.
    * @param    color
    *           The number of the color-pair (foreground+background) to use
    *           for blanking the rectangle.
    */
   @Override
   public void drawULCorner(int col, int row, Color color)
   {
      std.drawULCorner(col, row, color);
   }
   
   /**
    * Draw the upper-right border corner. 
    *  
    * @param    col 
    *           The column number where to draw.
    * @param    row 
    *           The row number where to draw.
    * @param    color
    *           The number of the color-pair (foreground+background) to use
    *           for blanking the rectangle.
    */
   @Override
   public void drawURCorner(int col, int row, Color color)
   {
      std.drawURCorner(col, row, color);
   }
   
   /**
    * Draw the lower-left border corner. 
    *  
    * @param    col 
    *           The column number where to draw.
    * @param    row 
    *           The row number where to draw.
    * @param    color
    *           The number of the color-pair (foreground+background) to use
    *           for blanking the rectangle.
    */
   @Override
   public void drawLLCorner(int col, int row, Color color)
   {
      std.drawLLCorner(col, row, color);
   }
   
   /**
    * Draw the lower-right border corner. 
    *  
    * @param    col 
    *           The column number where to draw.
    * @param    row 
    *           The row number where to draw.
    * @param    color
    *           The number of the color-pair (foreground+background) to use
    *           for blanking the rectangle.
    */
   @Override
   public void drawLRCorner(int col, int row, Color color)
   {
      std.drawLRCorner(col, row, color);
   }
   
   /**
    * Draw the upper or lower border. 
    *  
    * @param    col
    *           The column number where to draw.
    * @param    row
    *           The row number where to draw.
    * @param    width
    *           The border width.
    * @param    color
    *           The number of the color-pair (foreground+background)
    */
   @Override
   public void drawHBorder(int col, int row, int width, Color color)
   {
      std.drawHBorder(col, row, width, color);
   }
   
   /**
    * Draw the left or right border. 
    *  
    * @param    col
    *           The column number where to draw.
    * @param    row
    *           The row number where to draw.
    * @param    height
    *           The border height.
    * @param    color
    *           The number of the color-pair (foreground+background)
    */
   @Override
   public void drawVBorder(int col, int row, int height, Color color)
   {
      std.drawVBorder(col, row, height, color);
   }
   
   /**
    * Check if this output primitives implementation requires border corners
    * when drawing a box.
    * 
    * @return   <code>true</code> if the box should be drawn with border 
    *           corners.
    */
   @Override
   public boolean hasBoxBorderCorners()
   {
      return std.hasBoxBorderCorners();
   }

   /**
    * Get character at specified position.
    * 
    * @param    row
    *           Character row.
    * @param    col
    *           Character column.
    *
    * @return   character from specified position or space if operation is not
    *           supported.
    */
   @Override
   public Cell cellAt(int row, int col)
   {
      return std.cellAt(row, col);
   }

   /**
    * Perform a bulk output operation for the whole screen at once with
    * output masking when desired. Set the cursor afterwards.
    *
    * @param   screen
    *          Container for whole screen content and cursor position.
    */
   @Override
   public void updateScreen(ScreenData screen)
   {
      std.updateScreen(screen);
   }
}