/*
** Module   : AbstractFileChooserDialog.java
** Abstract : Gathers common code from file/directory selection dialogs.
**
** Copyright (c) 2017-2018, Golden Code Development Corporation.
**
** -#- -I- --Date-- ---------------------------------Description----------------------------------
** 001 OM  20170815 Initial commit.
** 002 OM  20171026 The object now inherits FrameGuiImpl and implements ModalBox.
** 003 OM  20180727 Added handle for system close button.
*/

/*
** This program is free software: you can redistribute it and/or modify
** it under the terms of the GNU Affero General Public License as
** published by the Free Software Foundation, either version 3 of the
** License, or (at your option) any later version.
**
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
** GNU Affero General Public License for more details.
**
** You may find a copy of the GNU Affero GPL version 3 at the following
** location: https://www.gnu.org/licenses/agpl-3.0.en.html
** 
** Additional terms under GNU Affero GPL version 3 section 7:
** 
**   Under Section 7 of the GNU Affero GPL version 3, the following additional
**   terms apply to the works covered under the License.  These additional terms
**   are non-permissive additional terms allowed under Section 7 of the GNU
**   Affero GPL version 3 and may not be removed by you.
** 
**   0. Attribution Requirement.
** 
**     You must preserve all legal notices or author attributions in the covered
**     work or Appropriate Legal Notices displayed by works containing the covered
**     work.  You may not remove from the covered work any author or developer
**     credit already included within the covered work.
** 
**   1. No License To Use Trademarks.
** 
**     This license does not grant any license or rights to use the trademarks
**     Golden Code, FWD, any Golden Code or FWD logo, or any other trademarks
**     of Golden Code Development Corporation. You are not authorized to use the
**     name Golden Code, FWD, or the names of any author or contributor, for
**     publicity purposes without written authorization.
** 
**   2. No Misrepresentation of Affiliation.
** 
**     You may not represent yourself as Golden Code Development Corporation or FWD.
** 
**     You may not represent yourself for publicity purposes as associated with
**     Golden Code Development Corporation, FWD, or any author or contributor to
**     the covered work, without written authorization.
** 
**   3. No Misrepresentation of Source or Origin.
** 
**     You may not represent the covered work as solely your work.  All modified
**     versions of the covered work must be marked in a reasonable way to make it
**     clear that the modified work is not originating from Golden Code Development
**     Corporation or FWD.  All modified versions must contain the notices of
**     attribution required in this license.
*/

package com.goldencode.p2j.ui.client;

import com.goldencode.p2j.ui.*;
import com.goldencode.p2j.ui.client.gui.*;
import com.goldencode.p2j.ui.client.gui.driver.*;
import com.goldencode.p2j.ui.client.widget.*;

/**
 * This abstract class groups common code from all dialogs related to file/directory selection.
 */
public abstract class AbstractFileChooserDialog
extends FrameGuiImpl
implements Container<GuiOutputManager>,
           IsModal
{
   /** The initial location the dialog shows to the user, if set. */
   protected String initialDir;
   
   /** Flag that turns to true only if the user closed the dialog by accepting it. */
   protected boolean accepted = false;
   
   /** Flag which tracks user input. */
   protected boolean inputDone = false;
   
   /** The selected path. It will be sent back to server if the dialog is accepted. */
   protected String selectedPath = null;
   
   /**
    * The constructor build an object by forwarding the required parameters to super class. 
    *
    * @param   id
    *          The {@link WidgetId} of the dialog. 
    * @param   cfg
    *          The {@link FrameConfig} of the dialog.
    */
   public AbstractFileChooserDialog(WidgetId id, FrameConfig cfg)
   {
      super(id, cfg);
   }
   
   /**
    * Checks whether the user accepted the selection from the dialog.
    * 
    * @return  {@code true} if the dialog was accepted, and {@code false} if the dialog was not
    *          yet displayed or it was rejected by user.
    */
   public boolean isAccepted()
   {
      return accepted;
   }
   
   /**
    * Configures the initial directory (the location where the dialog opens). 
    * 
    * @param   initialDir
    *          The initial directory for this dialog.
    */
   public void setInitialDir(String initialDir)
   {
      this.initialDir = initialDir;
   }
   
   /**
    * Obtain the path that the user has selected.
    *
    * @return  path that the user has selected or {@code null} if the dialog was rejected or not
    *          yet accepted.
    */
   public String getSelectedPath()
   {
      return selectedPath;
   }
   
   /**
    * Make the dialog visible and allowing the user to select the entity as the dialog was
    * configured. This kind of dialog can be opened a single time, it will be destroyed after the
    * user accepts it or rejects it the first time.
    * 
    * @param   visible
    *          {@code true} shows the dialog, {@code false} hides it.
    */
   @Override
   public void setVisible(boolean visible)
   {
      init();
      // this method should only be called with [visible == true]
      if (visible && parent() == null)
      {
         setParent(WindowManager.getDefaultWindow());
      }
      
      super.setVisible(visible);
   }
   
   /**
    * Show widget.
    */
   @Override
   public void show()
   {
      init();
      
      super.show();
   }
   
   /**
    * Hide widget.
    */
   @Override
   public void hide()
   {
      super.hide();
      
      destroy();
      
      inputDone = true;
   }
   
   /**
    * Get status of the input processing.
    *
    * @return  current status of the input processing.
    */
   @Override
   public boolean isDismissed()
   {
      return inputDone;
   }
   
   /**
    * Let the modal UI element now that it should finish its modal input processing.
    */
   @Override
   public void dismiss()
   {
      inputDone = true;
      destroy();
      hide();
   }
   
   /**
    * Sets the value to be returned by the dialog.
    *
    * @param   path
    *          The full path of the selected file/dir. 
    */
   protected void setSelectedPath(String path)
   {
      this.selectedPath = path;
   }
   
   /**
    * Initialize the dialog by performing tasks that was not possible in the constructor because
    * it was not fully created.
    */
   protected void init()
   {
      // the titlebar can be null when the web client is in embedded mode
      WindowTitleBar tbar = topLevelWindow().getTitleBar();
      if (tbar == null)
      {
         return;
      }
      tbar.setClickHandler(CaptionButtonType.CLOSE, this::dismiss);
   }
}
