/*
** Module   : AlertBoxCommons.java
** Abstract : AlertBox logic
**
** Copyright (c) 2015-2018, Golden Code Development Corporation.
**
** -#- -I- --Date-- ----------------------------Description-----------------------------
** 001 HC  20150612 Initial version.
** 002 HC  20151020 Fixed an abend in GUI ALERT-BOX caused by wrong button initialization.
** 003 HC  20151023 Fixed GUI ALERT-BOX keyboard input processing. 
** 004 IAS 20160315 Support for the DEBUG-ALERT attribute
** 005 EVL 20160414 Fix for layout issues for stack trace window in GUI mode.
** 006 VIG 20160427 Added flag setting: mnemonic support for buttons.
** 007 EVL 20160510 Making correct stack trace dialog window height and width.
** 008 EVL 20160726 Changing to use screen buffer array in ThinClient.enable() method.
** 009 SBI 20161101 Changed showStackTrace().
** 010 HC  20171024 Runtime support for SYSTEM-DIALOG PRINT-SETUP and OUTPUT TO PRINTER.
** 011 GES 20171207 Removed forced version of assign().
** 012 CA  20180620 Ensure both on/off calls for an invalidation bracket are always called.
*/

/*
** This program is free software: you can redistribute it and/or modify
** it under the terms of the GNU Affero General Public License as
** published by the Free Software Foundation, either version 3 of the
** License, or (at your option) any later version.
**
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
** GNU Affero General Public License for more details.
**
** You may find a copy of the GNU Affero GPL version 3 at the following
** location: https://www.gnu.org/licenses/agpl-3.0.en.html
** 
** Additional terms under GNU Affero GPL version 3 section 7:
** 
**   Under Section 7 of the GNU Affero GPL version 3, the following additional
**   terms apply to the works covered under the License.  These additional terms
**   are non-permissive additional terms allowed under Section 7 of the GNU
**   Affero GPL version 3 and may not be removed by you.
** 
**   0. Attribution Requirement.
** 
**     You must preserve all legal notices or author attributions in the covered
**     work or Appropriate Legal Notices displayed by works containing the covered
**     work.  You may not remove from the covered work any author or developer
**     credit already included within the covered work.
** 
**   1. No License To Use Trademarks.
** 
**     This license does not grant any license or rights to use the trademarks
**     Golden Code, FWD, any Golden Code or FWD logo, or any other trademarks
**     of Golden Code Development Corporation. You are not authorized to use the
**     name Golden Code, FWD, or the names of any author or contributor, for
**     publicity purposes without written authorization.
** 
**   2. No Misrepresentation of Affiliation.
** 
**     You may not represent yourself as Golden Code Development Corporation or FWD.
** 
**     You may not represent yourself for publicity purposes as associated with
**     Golden Code Development Corporation, FWD, or any author or contributor to
**     the covered work, without written authorization.
** 
**   3. No Misrepresentation of Source or Origin.
** 
**     You may not represent the covered work as solely your work.  All modified
**     versions of the covered work must be marked in a reasonable way to make it
**     clear that the modified work is not originating from Golden Code Development
**     Corporation or FWD.  All modified versions must contain the notices of
**     attribution required in this license.
*/

package com.goldencode.p2j.ui.client;


import java.util.*;

import com.goldencode.p2j.ui.*;
import com.goldencode.p2j.ui.chui.*;
import com.goldencode.p2j.ui.client.driver.WidgetFactory;
import com.goldencode.p2j.ui.client.event.*;
import com.goldencode.p2j.ui.client.event.listener.ActionListener;
import com.goldencode.p2j.ui.client.widget.*;
import com.goldencode.p2j.util.*;

/**
 * The class provides logic common to all implementations of {@link AlertBox}.
 */
public class AlertBoxCommons<O extends OutputManager<?>>
implements ActionListener
{
   /** Alert box items. */ 
   public Object[] content;
   
   /** Default color scheme. */
   public Color alertColor = Color.NORMAL;

   /** The variable which is being edited. */
   public logical result = null;
   
   /** The button configuration. */
   public int buttonSet = 0;
   
   /** Flag indicating the title must be forced to left-align. */
   public boolean forceTitleLeftAlign = false;
   
   /** Cached widget factory reference. */
   public WidgetFactory<O> widgetFactory;

   /** Flag which tracks user input. */
   private boolean inputDone = false;
   
   /** The related alert box reference. */
   private AlertBox<O> alertBox;
   
   /** The buttons container. */
   private Container<O> buttonsContainer;
   
   /** The Progress stack trace */ 
   private final String stack;
   
   /**
    * The default constructor.
    *
    * @param    alertBox
    *           The owner instance.
    * @param    buttonsContainer
    *           The buttons container.
    * @param    widgetFactory
    *           Reference to widget factory.
    * @param    content
    *           Array of elements which form alert box content.
    * @param    buttonSet
    *           Button set type to be constructed for this AlertBox.
    * @param    var
    *           The variable to be edited.  The initial value sets the 
    *           default button and the resulting value is changed by the
    *           user's press of a button.
    * @param    color
    *           Color scheme to use.
    * @param    stack
    *           Progress stack trace
    */
   public AlertBoxCommons(AlertBox<O>      alertBox,
                          Container<O>     buttonsContainer,
                          WidgetFactory<O> widgetFactory,
                          Object[]         content,
                          int              buttonSet,
                          logical          var,
                          Color            color,
                          String[]         stack)
   {
      this.alertBox         = alertBox;
      this.buttonsContainer = buttonsContainer;
      this.widgetFactory    = widgetFactory;
      this.content          = content;
      this.buttonSet        = buttonSet;
      this.result           = (logical) var.duplicate();
      this.alertColor       = color;
      if (stack == null)
      {
         this.stack = null;
      }
      else 
      {
         List<String> strace = new ArrayList<>(stack.length);
         String prefix = "--> ";
         for (String s: stack)
         {
            strace.add(prefix + s + "\n");
            prefix = "    ";
         }
         
         // In ChUI mode Progress displays stack trace in a reverse order 
         if (ThinClient.getInstance().isChui())
         {
            Collections.reverse(strace);
         }
         
         StringBuilder sb = new StringBuilder();
         strace.stream().forEach(sb::append);
         this.stack = sb.toString();
      }
   }
   
   /**
    * Action handler to process the AlertBox buttons event. Each event
    * changes the return code appropriately.
    * 
    * @param    action
    *           The action event to handle is coming from defined buttons.
    */
   public void onAction(ActionEvent action)
   {
      logical rc = null;
      
      if (action.command().equals("Help"))
      {
         showStackTrace();
         return;
      }

      if (action.command().equals("OK")  ||
          action.command().equals("Yes") ||
          action.command().equals("Retry"))
      {
         rc = new logical(true);
      }
      else if (action.command().equals("No"))
      {
         rc = new logical(false);
      }
      else if (action.command().equals("Cancel"))
      {
         if (buttonSet == LogicalTerminal.BTN_YES_NO_CANCEL)
         {
            rc = new logical();
         }
         else
         {
            rc = new logical(false);
         }
      }

      setResult(rc);
      dismiss();
   }
   
   /**
    * Set the result code, hide the dialog and flag input as complete.
    *
    */
   public void dismiss()
   {
      alertBox.hide();
      inputDone = true;
   }

   /**
    * Helper method to get the edited result when finished.
    * 
    * @return   The result of the editing operation which was set depending
    *           on the pressed button.
    */
   public logical getResult()
   {
      return result;
   }

   /**
    * Sets the alert-box result value.
    *
    * @param   value
    *          The result logical value.
    */
   public void setResult(logical value)
   {
      this.result.assign(value);
   }

   /**
    * Returns <code>true</code> when the modal UI element has dismissed (finished its modal input).
    *
    * @return  see above.
    */
   public boolean isDismissed()
   {
      return inputDone;
   }
   
   /**
    * Constructs the alert box button set. Only the required
    * number of buttons are created here.  The value of the 
    * <code>logical</code> variable being edited determines the default
    * button (<code>true</code> == the first button, <code>false</code>
    * == the second button and for a 3-button dialog, <code>unknown</code> == 
    * the third button).
    */
   public List<Button<O>> createButtonSet()
   {
      List<Button<O>> result = new ArrayList<>(3);
      
      if (buttonSet == LogicalTerminal.BTN_OK)
      {
         result.add(createButton("OK"));
      }
      else if (buttonSet == LogicalTerminal.BTN_OK_CANCEL)
      {
         result.add(createButton("OK"));
         result.add(createButton("Cancel"));
      }
      else if (buttonSet == LogicalTerminal.BTN_YES_NO)
      {
         result.add(createButton("Yes"));
         result.add(createButton("No"));
      }
      else if (buttonSet == LogicalTerminal.BTN_YES_NO_CANCEL)
      {
         result.add(createButton("Yes"));
         result.add(createButton("No"));
         result.add(createButton("Cancel"));
      }
      else if (buttonSet == LogicalTerminal.BTN_RETRY_CANCEL)
      {
         result.add(createButton("Retry"));
         result.add(createButton("Cancel"));
      }
      
      if (stack != null)
      {
         result.add(createButton("Help"));
      }
      return result;
   }
   
   /** 
    * Returns the index of the default button. The index is compatible
    * with the return value of {@link #createButtonSet()}.
    * 
    * @return   Index of the default button.
    */
   public int getDefaultButton()
   {
      // calculate the default button using the value of the edited variable
      // (true == 0, false == 1 and for a 3-button dialog, unknown == 2)
      int index = 2;
      
      if (!result.isUnknown() || buttonSet != LogicalTerminal.BTN_YES_NO_CANCEL)
      {
         index = (result.booleanValue() ? 0 : 1);
      }
      
      return index;
   }

   /**
    * Move the input focus to the first widget which can receive a focus in
    * the peer container. This is required to handle the situation when the
    * current widget with the focus is the last one within the container and
    * we have to simulate the Progress behavior in which the next focused
    * widget will be the first widget which can receive a focus in a
    * container.
    */
   public void nextFocus()
   {
      Widget<O>[] c = buttonsContainer.widgets();

      int start = -1;
      
      for (int i = 0; i < c.length; i++)
      {
         if (c[i].hasFocus())
            {
            start = i;
            break;
         }
      }
      
      if (start == -1)
         return;
      
      start++;
      
      if (start >= c.length)
         start = 0;

      for (int i = start; i < c.length; i++)
      {
         if (c[i].focusTraversable())
         {
            c[i].requestFocus();
            return;
         }
      }
      
      for (int i = 0; i < c.length; i++)
      {
         if (c[i].focusTraversable())
         {
            c[i].requestFocus();
            return;
         }
      }
   }
   
   /**
    * Move the input focus to the last widget which can receive a focus in
    * the peer container. This is required to handle the situation when the
    * current widget with the focus is the first one within the container and
    * we have to simulate the Progress behavior in which the previous focused
    * widget will be the last widget which can receive a focus in a container.
    */
   public void prevFocus()
   {
      Widget<O>[] c = buttonsContainer.widgets();

      int start = -1;
      
      for (int i = 0; i < c.length; i++)
      {
         if (c[i].hasFocus())
         {
            start = i;
            break;
         }
      }
      
      if (start == -1)
         return;
      
      start--;
      
      if (start < 0)
         start = c.length - 1;

      for (int i = start; i >= 0; i--)
      {
         if (c[i].focusTraversable())
         {
            c[i].requestFocus();
            return;
         }
      }

      for (int i = c.length - 1; i >= 0; i--)
      {
         if (c[i].focusTraversable())
         {
            c[i].requestFocus();
            return;
         }
      }
   }

   /**
    * Find the button that matches a passed key event.
    * 
    * @param    keyEvent
    *           Event to check.
    * 
    * @return   <code>true</code> if matching button is found and event is
    *           processed.
    */
   public boolean findButton(KeyInput keyEvent)
   {
      Widget<O>[] c = buttonsContainer.widgets();
      
      for (int i = 0; i < c.length; i++)
      {
         if (c[i] instanceof Button && ((Button<O>) c[i]).checkKey(keyEvent))
         {
            c[i].requestFocus();
            EventManager.postEvent(new KeyInput(c[i], keyEvent.id(), keyEvent.keyCode()));
            return true;
         }
      }
      
      return false;
   }
   
   /**
    * Builds a multiline string from the alert box content array.
    * 
    * @return   See above.
    */
   public String buildMessageText()
   {
      StringBuilder sb = new StringBuilder();
      
      for (Object item : this.content)
      {
         if (item instanceof String)
         {
            sb.append(item);
         }
         else if (item instanceof Integer)
         {
            char[] newLines = new char[(Integer) item];
            Arrays.fill(newLines, '\n');
            sb.append(newLines);
         }
      }
      
      return sb.toString();
   }
   
   /**
    * Display dialog with the Progress stack trace
    */
   private void showStackTrace()
   {
      final String title = "Stack Trace";
      final ThinClient tc = ThinClient.getInstance();
      final boolean chui = tc.isChui();
      final TitledWindow win = alertBox.createStackTraceWindow(title);
      final O screen = alertBox.window().screen();
      final WidgetRegistry registry = screen.getRegistry();

      final Widget saveFocus  = UiUtils.getCurrentFocus();
      final int    focusState = FocusManager.focusState(saveFocus);

      // there are only 3 tab items in this dialog
      final int[]  tabItemList = new int[3]; 
      
      Window<O> window = alertBox.window();
      int wndId = chui ? UiUtils.getWidgetIdAsInt(window) : win.getId().asInt();
      
      // TODO: replace hard-coded offsets with createGap()
      FrameConfig fcfg = new FrameConfig(WidgetId.nextID());
      fcfg.box = false;
      fcfg.overlay = true;
      fcfg.labels = false;
      fcfg.sideLabels = false;
      if (chui)
      {
         fcfg.widthChars = 60;
         fcfg.heightChars = 13;
      }
      else
      {
         fcfg.windowID = win.getId().asInt();
         fcfg.dialog = true;
         fcfg.threeD = true;
         // l&f GUI constants
         fcfg.font = FontManager.STACK_TRACE_WINDOW_FONT;
         fcfg.widthPixels = 341;
         fcfg.heightPixels = 328;
      }
      registry.reconstructWidget(fcfg);
      Frame<O> frame =  (Frame)registry.getComponent(fcfg.id);
//      frame.addWidget(alertBox.createGap(registry)); // top space
      
      frame.addWidget(createSkip(registry, 1));
//      frame.addWidget(alertBox.createGap(registry)); // left space

      EditorConfig ecfg = new EditorConfig(WidgetId.nextID());
      ecfg.innerLines = chui ? 13 : 19;
      ecfg.innerChars = chui ? 55 : 62;
      ecfg.box = true;
      ecfg.readOnly = true;
      ecfg.enabled = true;
      ecfg.scrollbarHorizontal = !chui;
      ecfg.scrollbarVertical = !chui;
      if (chui)
      {
         ecfg.column = 2;
      }
      else
      {
         ecfg.x = 12;
         ecfg.y = 11;
         ecfg.font = FontManager.STACK_TRACE_WINDOW_FONT;
      }
      registry.reconstructWidget(ecfg);
      Editor<O> editor = (Editor)registry.getComponent(ecfg.id);
      editor.setValue(new character(stack));
      frame.addWidget(editor);
      
      // add the first tab item
      tabItemList[0] = editor.getId().asInt();
      
//    frame.addWidget(alertBox.createGap(registry)); // right space

      frame.addWidget(createSkip(registry, 1));

      ButtonConfig btnOkCfg = new ButtonConfig(WidgetId.nextID());
      btnOkCfg.label = "OK";
      btnOkCfg.fixedWidth = true;
      btnOkCfg.isDefault = true;
      if (chui)
      {
         btnOkCfg.row = 12;
         btnOkCfg.column = 22;
      }
      else
      {
         btnOkCfg.x = 116;
         btnOkCfg.y = 271;
         btnOkCfg.heightPixels = 23;
         // l&f GUI constants
         btnOkCfg.autoResize = false;
         btnOkCfg.widthPixels = 45;
         btnOkCfg.font = FontManager.STACK_TRACE_WINDOW_FONT;
      }
      btnOkCfg.dynamic = true;
      registry.reconstructWidget(btnOkCfg);
      Button<O> btnOk = (Button)registry.getComponent(btnOkCfg.id);
      frame.addWidget(btnOk);
      fcfg.defaultButtonId = btnOkCfg.id.asInt();
      btnOk.setCommand("OK");
      btnOk.addListener(event -> {
         frame.hide();
         frame.destroy();
      });

      // add the second tab item
      tabItemList[1] = btnOk.getId().asInt();

      ButtonConfig btnDebugCfg = new ButtonConfig(WidgetId.nextID());
      btnDebugCfg.label = "Debug";
      btnDebugCfg.fixedWidth = true;
      btnDebugCfg.noFocus = true;
      if (chui)
      {
         btnDebugCfg.row = 12;
         btnDebugCfg.column = 27;
      }
      else
      {
         btnDebugCfg.x = 168;
         btnDebugCfg.y = 271;
         btnDebugCfg.heightPixels = 23;
         // l&f GUI constants
         btnDebugCfg.autoResize = false;
         btnDebugCfg.widthPixels = 45;
         btnDebugCfg.font = FontManager.STACK_TRACE_WINDOW_FONT;
      }
      btnDebugCfg.dynamic = true;
      registry.reconstructWidget(btnDebugCfg);
      Button<O> btnDebug = (Button)registry.getComponent(btnDebugCfg.id);
      frame.addWidget(btnDebug);
      
      // add the third tab item
      tabItemList[2] = btnDebug.getId().asInt();
      
      frame.doLayout();
      frame.setProtected(true);

      ScreenBuffer sb = new ScreenBuffer(fcfg.id.asInt(), 5);
      screen.setInvalidate(true);
      try
      {
         tc.enable(fcfg.id.asInt(), null, new ScreenBuffer[] {sb}, true, wndId, tabItemList);
         btnDebug.setEnabled(false);
         
         tc.clear(fcfg.id.asInt(), true);
         sb.resetChanged();
      }
      finally
      {
         screen.setInvalidate(chui ? window : win, false);
      }
   }

   /**
    * Create SKIP(n) widget
    *  
    * @param registry
    *        widget registry
    * @param lines
    *        lines to skip
    * 
    * @return SKIP(n) widget
    */
   private Widget<O> createSkip(WidgetRegistry<O> registry, int lines)
   {
      SkipConfig scfg = new SkipConfig(WidgetId.nextID());
      scfg.vertical = true;
      scfg.length = lines;
      registry.reconstructWidget(scfg);
      return registry.getComponent(scfg.id); 
   }

   /**
    * Creates and initializes a button widget to be used in the alert box.
    * 
    * @return   A reference to {@link Button} instance.
    */
   private Button<O> createButton(String text)
   {
      Button<O> b = widgetFactory.createButton(WidgetId.nextID(), buttonsContainer, text);
      b.initialize(b.getId(), null);
      b.setSupportsMnemonic();
      b.setTextLabel(text);
      b.setCommand(text);
      b.setEnabled(true);
      
      b.addListener(this);
      return b;
   }
}