/*
** Module   : ChooseHandler.java
** Abstract :
**
** Copyright (c) 2010-2017, Golden Code Development Corporation.
**
** -#- -I- --Date-- ---------------------------------Description----------------------------------
** 001 SIY 20100831 Created initial version
** 002 SIY 20100915 tryFocusChange() is moved to FocusManager.
** 003 SIY 20101122 Switching from KeyEvent to KeyInput.
** 004 SIY 20110112 Changes related to new color handling scheme.
** 005 SIY 20111101 Moved from CHUI package. Minor cleanups/refactorings in preparation to GUI 
**                  client.
** 006 CA  20140805 Added generics to force the abstract widget classes to not use implementation-
**                  specific code: the abstract widgets will have access only to the base APIs
**                  provided by the OutputManager.
**                  Explicitly import classes from chui package(s).
** 007 HC  20140911 GUI coordinate support - data types refactoring.
** 008 CA  20140926 Added shared widget configuration support, to allow GUI/ChUI concrete
**                  implementation for the same widget ID.  Refactored the widget configuration
**                  classes: all fields were made public; these classes need to be as dumb as
**                  possible, all logic related to setting/getting a certain field should be at
**                  the widget or at the caller. Refs #2254
** 009 CA  20141119 Refactored drawing code to enable multiple OS windows, for GUI support.  Misc
**                  changes/fixes related to drawing on multiple windows.
** 010 HC  20150323 Removed business logic from config classes.
*/
/*
** This program is free software: you can redistribute it and/or modify
** it under the terms of the GNU Affero General Public License as
** published by the Free Software Foundation, either version 3 of the
** License, or (at your option) any later version.
**
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
** GNU Affero General Public License for more details.
**
** You may find a copy of the GNU Affero GPL version 3 at the following
** location: https://www.gnu.org/licenses/agpl-3.0.en.html
** 
** Additional terms under GNU Affero GPL version 3 section 7:
** 
**   Under Section 7 of the GNU Affero GPL version 3, the following additional
**   terms apply to the works covered under the License.  These additional terms
**   are non-permissive additional terms allowed under Section 7 of the GNU
**   Affero GPL version 3 and may not be removed by you.
** 
**   0. Attribution Requirement.
** 
**     You must preserve all legal notices or author attributions in the covered
**     work or Appropriate Legal Notices displayed by works containing the covered
**     work.  You may not remove from the covered work any author or developer
**     credit already included within the covered work.
** 
**   1. No License To Use Trademarks.
** 
**     This license does not grant any license or rights to use the trademarks
**     Golden Code, FWD, any Golden Code or FWD logo, or any other trademarks
**     of Golden Code Development Corporation. You are not authorized to use the
**     name Golden Code, FWD, or the names of any author or contributor, for
**     publicity purposes without written authorization.
** 
**   2. No Misrepresentation of Affiliation.
** 
**     You may not represent yourself as Golden Code Development Corporation or FWD.
** 
**     You may not represent yourself for publicity purposes as associated with
**     Golden Code Development Corporation, FWD, or any author or contributor to
**     the covered work, without written authorization.
** 
**   3. No Misrepresentation of Source or Origin.
** 
**     You may not represent the covered work as solely your work.  All modified
**     versions of the covered work must be marked in a reasonable way to make it
**     clear that the modified work is not originating from Golden Code Development
**     Corporation or FWD.  All modified versions must contain the notices of
**     attribution required in this license.
*/

package com.goldencode.p2j.ui.client;

import com.goldencode.p2j.ui.*;
import com.goldencode.p2j.ui.chui.ThinClient;
import com.goldencode.p2j.ui.client.chui.driver.ChuiOutputManager;
import com.goldencode.p2j.ui.client.event.*;
import com.goldencode.p2j.ui.client.widget.*;
import com.goldencode.p2j.util.*;

/**
 * Helper class which encapsulates all logic related to processing CHOOSE 
 * in frame.
 */
public class ChooseHandler<O extends OutputManager<?>>
{
   /** Set to <code>true</code> when CHOOSE processing is active. */
   private boolean chooseActive = false;

   /** 
    * Set to <code>true</code> if CHOOSE processing was finished
    * by generating internal GO event.
    */
   private boolean internalGo = false;
   
   /** Selected item color. */
   private Color chooseColor = null;

   /** CHOOSE data. */
   private ChooseData chooseData = null;
   
   /** List of components involved in choose. */
   private Widget<O>[] compList = null;

   /** Component focused during processing of choose in the ROW mode. */
   private Widget<O> focusedCol = null;
   
   /** Current selected field. */
   private int index = -1;

   /** This flag is reset when CHOOSE processes first typed key. */
   private boolean initial = true;

   /** Owner frame. */
   private Frame<O> owner;

   /**
    * Constructor.
    * 
    * @param   owner
    *          Frame which owns this handler.
    */
   public ChooseHandler(Frame<O> owner) 
   {
      this.owner = owner;
   }
   
   /**
    * Draw choice if CHOOSE processing is active. 
    */
   public void drawIfActive()
   {
      if (chooseActive && index != -1)
      {
         drawChoice(index, true);
      }
   }
   
   /**
    * Obtain current CHOOSE mode data.
    *
    * @return  Current instance.
    */
   public ChooseData getChooseData()
   {
      return chooseData;
   }
   
   /**
    * Return the help text for the selected component using the CHOOSE index.
    *
    * @return   help text or <code>null</code> if not in CHOOSE
    */
   public String getChooseHelp()
   {
      if (!chooseActive)
      {
         return null;
      }

      String helpText = null;

      if (chooseData.isFieldMode())
         helpText = chooseData.getFieldHelp(index);
      else
         helpText = chooseData.getRowHelp();

      return helpText;
   }
   
   /**
    * Process keys in CHOOSE mode.
    * 
    * @param   event
    *          Key event to process.
    */
   public void processKeyEvent(KeyInput event)
   {
      int next = index;
      int key = event.actionCode();

      if (key == Keyboard.KA_RETURN)
      {
         internalGo = true;
         EventManager.postKey(Keyboard.KA_GO);
         return;
      }

      if (key == ' '                      ||
          key == Keyboard.KA_TAB          || 
          key == Keyboard.KA_CURSOR_RIGHT || 
          key == Keyboard.KA_CURSOR_DOWN)
      {
         next = (chooseData.isFieldMode()) ? nextField(key) : nextRow(key);
         chooseData.setKeys("");
      }
      else if (key == Keyboard.KA_BACK_TAB  ||
               key == Keyboard.KA_CURSOR_UP ||
               key == Keyboard.KA_CURSOR_LEFT)
      {
         next = (chooseData.isFieldMode()) ? prevField(key) : prevRow(key);
         chooseData.setKeys("");
      }
      else
      {
         // filter out high level events
         if (key >= 0 || 
             (chooseData.isNoError() && 
             (key == Keyboard.KA_PAGE_DOWN || 
              key == Keyboard.KA_PAGE_UP)))  
         {
            char ch = Utils.toChar(key);

            if (Character.isLetterOrDigit(ch))
            {
               next = locateField(ch);
            }
            else
            {
               if (chooseData.isNoError() && key >= 0)
               {
                  internalGo = true;
               }

               invalidInput();
            }
         }
      }

      // check if the current choice is now different
      if (next != index)
      {
         drawChoice(index, false);

         index = next;

         // row mode causes the frame-line to change (the cursor is moved
         // in the down frame)
         if (!chooseData.isFieldMode())
         {
            // moveCursor() should not be used here, it has behavior (like
            // wrapping from top back to the bottom) which is wrong for a
            // choose
            owner.setCursor(next);
         }
         else
         {
            FocusManager.tryFocusChange(compList[next]);
         }

         if (!ThinClient.getInstance().isGoPending())
            drawChoice(next, true);

         owner.sync();
      }
   }

   /**
    * Initialize CHOOSE processing mode.
    * 
    * @param   list
    *          List of components.
    * @param   data
    *          Choose data.
    * @param   focusOn
    *          ID of the widget to start from.
    *           
    * @return  ID of the selected widget or -1.
    */
   public int enterChoose(Widget<O>[] list, ChooseData data, int focusOn)
   {
      if (list == null)
         throw new RuntimeException("Component list in enterChoose() is null");

      chooseData = data;
      compList = list;
      chooseColor = data.getColor();
      internalGo  = false;
      
      if (index < 0 || index > list.length)
         index = 0;

      // honor the outstanding next-prompt request if any
      if (focusOn != -1)
      {
         if (chooseData.isFieldMode())
         {
            for (int i = 0; i < list.length; i++)
            {
               if (list[i].getId().asInt() == focusOn)
               {
                  index = i;
                  break;
               }
            }
         }
      }

      if (owner.isDown() && 
          !chooseData.isFieldMode() && 
          owner.getCursor() >= 0)
      {
         index = owner.getCursor();
      }
      
      initial = true;
      
      if (!chooseData.isFieldMode())
         focusedCol = list[0];
      else
         focusedCol = null;

      int next = locateFieldByPrefix(data.getKeys());
      
      if (next >= 0)
         index = next;
    
      chooseActive = true;

      // repaint all involved components
      for (int i = 0; i < compList.length; i++)
      {
         list[i].repaint();
      }

      // repaint all rows in row choose mode
      if (!chooseData.isFieldMode())
      {
         int downs = owner.config().down;
         boolean mode = false;

         for (int i = 0; i < downs; i++)
         {
            mode = i == index ? true : false;
            drawChoice(i, mode);
         }
         owner.sync();
      }

      if (next >= 0)
      {
         // simulate Progress behavior 
         if (chooseData.isAutoReturn())
            EventManager.postKey(Keyboard.KA_GO);
      }
      
      if (chooseData.isFieldMode())
      {
         return list[index].getId().asInt();
      }

      return -1;
   }

   /**
    * Get current state of the choose activation flag.
    * 
    * @return  Current state of the flag.
    */
   public boolean isChooseActive()
   {
      return chooseActive;
   }
   
   /**
    * Check to see if a widget is part of an ongoing CHOOSE.
    * 
    * @param   widget 
    *          Widget to check.
    * @return  <code>true</code> if the widget is part of a CHOOSE
    */
   public boolean isInChoose(Widget<O> widget)
   {
      if (!chooseActive)
      {
         return false;
      }

      for (int i = 0; i < compList.length; i ++)
      {
         if (compList[i].equals(widget))
         {
            return true;
         }
      }

      return false;
   }
   
   /**
    * Stop CHOOSE mode.
    * 
    * @return  ChooseData instance as it is left after processing.
    */
   public ChooseData leaveChoose()
   {
      if (!chooseActive)
         return chooseData;
      
      chooseActive = false;
      compList = null;
      
      return chooseData;
   }
   
   /**
    * Positions the CHOOSE at the specified widget if possible and returns
    * the indication of success or failure.
    * If the specified widget is not part of the active choose, the choose
    * repositions to nowhere and the highlight gets removed from the frame.
    * 
    * @param   widget 
    *          Widget to set current CHOOSE position at.
    * @return  <code>true</code> if the widget is part of an active CHOOSE
    */
   public boolean setChooseAt(Widget<O> widget)
   {
      // check if in choose
      if (!chooseActive)
      {
         return false;
      }

      // figure out the new position in choose
      int newpos = -1;

      for (int i = 0; i < compList.length; i ++)
      {
         if (compList[i].equals(widget))
         {
            newpos = i;
            break;
         }
      }

      // quick test for no change
      if (newpos == index)
      {
         // no change
         if (newpos == -1)
         {
            return false;
         }
            
         return true;
      }

      // remove previous highlight
      if (index != -1)
      {
         drawChoice(index, false);
      }

      // put new highlight
      index = newpos;

      if (index != -1)
      {
         drawChoice(index, true);
         owner.sync();

         return true;
      }

      owner.sync();
      return false;
   }

   /**
    * Draws one of the choice fields either highlighted or not.
    *
    * @param    num
    *           field index
    * @param    highlight
    *           <code>true</code> if highlighting is to be on
    */
   @SuppressWarnings("unchecked")
   private void drawChoice(int num, boolean highlight)
   {
      FillIn<O,?> field = null;

      if (chooseData.isFieldMode() && compList[num] instanceof FillIn)
      {
         field  = (FillIn<O,?>) compList[num];
      }

      if (!chooseData.isFieldMode())
      {
         Widget<O> tmp = owner.getField(num, focusedCol);

         if (tmp instanceof FillIn)
            field  = (FillIn<O,?>) tmp;
      }

      if (field != null)
      {
         ScreenBitmap bitmap = owner.getAndResetBitmap();
         if (highlight)
         {
            field.draw(chooseColor);
            // TODO: remove this dependency
            ((ChuiOutputManager) owner.screen()).at(0, 0);
         }
         else
            field.draw();
         owner.setBitmap(bitmap);
      }
   }

   /**
    * Standard processing for invalid input (a key that should not be pressed
    * during this particular CHOOSE execution).
    */
   private void invalidInput()
   {
      if (chooseData.isNoError())
      {         
         EventManager.postKey(Keyboard.KA_GO);
      }
      else
      {
         owner.screen().beep();
      }
   }

   /**
    * Locate field using characters typed by user.
    * 
    * @param   c
    *          character typed by user.
    * @return  index of found field.
    */
   private int locateField(char c)
   {
      int next = 0;
      int found = 0;
      
      if (initial)
      {
         chooseData.setKeys("");
         initial = false;
      }
      
      String typed = chooseData.getKeys(); 
      typed += c;
      
      boolean fullMatch = false;
      
      if (!chooseData.isFieldMode())
      {
         String[] values = owner.getValues(focusedCol);
         
         for (int i = 0; i < values.length; i++)
         {
            String str = values[i].trim().toUpperCase(); 

            if (str.startsWith(typed.toUpperCase()))
            {
               found++;
               
               if (found == 1)  // first match
                  next = i;
               
               if (values[i].equalsIgnoreCase(typed))
                  fullMatch = true;
            }
         }
      }
      else
      {
         for (int i = 0; i < compList.length; i++)
         {
            if (compList[i] instanceof FillIn)
            {
               @SuppressWarnings("unchecked")
               FillIn<O,?> fin = (FillIn<O,?>) compList[i];
               String str = fin.getText().trim().toUpperCase(); 
               
               if (str.startsWith(typed.toUpperCase()))
               {
                  found++;
                  
                  if (found == 1)  // first match
                     next = i;
                  
                  if (str.equalsIgnoreCase(typed))
                     fullMatch = true;
               }
               
               // exactly one match
               if (!fullMatch && found == 1)
                  fullMatch = true;
            }
         }
      }
      
      if (found == 0)   // no such field
      {
         typed = "";
         
         if (!chooseData.isNoError())
            owner.screen().beep();
      }

      chooseData.setKeys(typed);
      
      if ((chooseData.isAutoReturn() && fullMatch) || 
          (found == 0 && chooseData.isNoError()))
      {
         EventManager.postKey(Keyboard.KA_GO);
      }
         
      return (found > 0) ? next : index;
   }

   /**
    * Locate field for specified keys.
    * 
    * @param   prefix
    *          Initial characters of the item value.
    *          
    * @return  index of first found item which matches the prefix.
    */
   private int locateFieldByPrefix(String prefix)
   {
      if (prefix == null || prefix.length() == 0)
         return -1;

      prefix = prefix.trim();
      
      if (prefix.length() == 0)
         return -1;
      
      if (!chooseData.isFieldMode())
      {
         String[] values = owner.getValues(focusedCol);
         for (int i = 0; i < values.length; i++)
         {
            if (values[i] != null && values[i].trim().startsWith(prefix))
               return i;
         }
      }
      else
      {
         for (int i = 0; i < compList.length; i++)
         {
            if (compList[i] instanceof DataContainer)
            {
               String text = ((DataContainer) compList[i]).getText();
               
               if (text != null && text.trim().startsWith(prefix))
                  return i;
            }
         }
      }
      
      return -1;
   }
   
   /**
    * Select next field.
    * 
    * @param   key
    *          Key which triggered this selection.
    *          
    * @return  index of the field to select.
    */
   private int nextField(int key)
   {
      // safety code
      if (index < 0 || index >= compList.length)
      {
         throw new IllegalStateException("Invalid choose index = " + index);
      }
      
      // check if we are already at the end
      if (index == (compList.length - 1))
      {
         if (key == ' ' || key == Keyboard.KA_TAB)
         {
            // we must wrap back to the first field
            return 0;
         }
         
         invalidInput();
         
         // cursor right or cursor down don't wrap 
         return index;
      }
      
      // the default (always works for tab or space that are not at the
      // end of the field list, sometimes works for down or right) is to
      // pick the next field in the list
      int next = index + 1;
      
      if (key == Keyboard.KA_CURSOR_RIGHT)
      {
         // right key only works where there is a field in the list that is
         // on the same line and is to the right of the current field
         Point current  = compList[index].location();      
         Point possible = compList[next].location();
         
         if (current.y != possible.y || current.x >= possible.x)
         {
            invalidInput();
            
            // reset so there is no change
            next = index;
         }
      }
      else if (key == Keyboard.KA_CURSOR_DOWN)
      {
         Point current = compList[index].location();
         double curend = current.x + compList[index].width();
         
         int first  = -1;
         double firsty = -1;
         int best   = -1;
         
         // down key only works where there is a field on the line below
         for (int i = next; i < compList.length; i++)
         {
            Point possible = compList[i].location();
            double posend  = possible.x + compList[i].width();
            
            if (current.y < possible.y)
            {
               if (first == -1)
               {
                  // store the first match
                  first  = i;
                  firsty = possible.y;
               }
               
               // only keep looking while there are other fields on the same
               // line as the first match
               if (possible.y == firsty)
               {
                  // store the first overlapping field that is below
                  if ((current.x >= possible.x && current.x <= posend) ||
                      (curend    >= possible.x && curend    <= posend) ||
                      (current.x <= possible.x && curend    >= posend))
                  {
                     best = i;
                     break;
                  }
               }
               else
               {
                  // done searching
                  break;
               }
            }
         }
         
         if (first == -1)
         {
            invalidInput();
            
            // no match was found so don't change
            next = index;
         }
         else
         {
            // choose the best match if there was one, otherwise fall back to
            // the first match
            next = (best == -1) ? first : best;
         }
      }
      
      return next;
   }
   
   /**
    * Select next row in down frame.
    * 
    * @param   key
    *          Key which triggered this selection.
    *          
    * @return  index of the row to select.
    */
   private int nextRow(int key)
   {
      int next = index;
      
      if (key == Keyboard.KA_CURSOR_RIGHT)
      {
         invalidInput();
      }
      else
      {
         next++;
         
         if (next >= owner.config().down)
         {
            if (chooseData.isNoError())
            {
               EventManager.postKey(Keyboard.KA_GO);
               return index;
            }
            
            // stop at the end of the rows and don't go past
            next = owner.config().down - 1;
         }
      }
      
      return next;
   }

   
   /**
    * Select previous field.
    * 
    * @param   key
    *          Key which triggered this selection.
    *          
    * @return  index of the field to select.
    */
   private int prevField(int key)
   {
      // safety code
      if (index < 0 || index >= compList.length)
      {
         throw new IllegalStateException("Invalid choose index = " + index);
      }
      
      // check if we are already at the beginning
      if (index == 0)
      {
         if (key == Keyboard.KA_BACK_TAB)
         {
            // we must wrap around to the last field
            return compList.length - 1;
         }
         
         invalidInput();
         
         // cursor left or cursor up don't wrap 
         return index;
      }
      
      // the default (always works for back-tab that are not at the beginning
      // of the field list, sometimes works for left or up) is to pick the
      // previous field in the list
      int prev = index - 1;
      
      if (key == Keyboard.KA_CURSOR_LEFT)
      {
         // left key only works where there is a field in the list that is
         // on the same line and is to the left of the current field
         Point current  = compList[index].location();      
         Point possible = compList[prev].location();
         
         if (current.y != possible.y || current.x <= possible.x)
         {
            invalidInput();
            
            // reset so there is no change
            prev = index;
         }
      }
      else if (key == Keyboard.KA_CURSOR_UP)
      {
         Point current = compList[index].location();
         double curend = current.x + compList[index].width();
         
         int first  = -1;
         double firsty = -1;
         int best   = -1;
         
         // up key only works where there is a field on the line above
         for (int i = prev; i >= 0; i--)
         {
            Point possible = compList[i].location();
            double posend  = possible.x + compList[i].width();
            
            if (current.y > possible.y)
            {
               if (first == -1)
               {
                  // store the first match
                  first  = i;
                  firsty = possible.y;
               }
               
               // only keep looking while there are other fields on the same
               // line as the first match
               if (possible.y == firsty)
               {
                  // store each overlapping field that is above, in turn
                  // (this will leave the leftmost match stored when the
                  // loop exits) 
                  if ((current.x >= possible.x && current.x <= posend) ||
                      (curend    >= possible.x && curend    <= posend) ||
                      (current.x <= possible.x && curend    >= posend))
                  {
                     best = i;
                  }
                  
                  // check if we can stop searching (because we already found
                  // at least 1 overlapping field and the current field is
                  // not overlapping
                  if (best != -1 && current.x > posend)
                  {
                     break;
                  }
               }
               else
               {
                  // done searching
                  break;
               }
            }
         }
         
         if (first == -1)
         {
            invalidInput();
            
            // no match was found so don't change
            prev = index;
         }
         else
         {
            // choose the best match if there was one, otherwise fall back to
            // the first match
            prev = (best == -1) ? first : best;
         }
      }
      
      return prev;
   }


   /**
    * Select previous row in down frame.
    * 
    * @param   key
    *          Key which triggered this selection.
    *          
    * @return  index of the row to select.
    */
   private int prevRow(int key)
   {
      int next = index;
      
      if (key == Keyboard.KA_CURSOR_LEFT)
      {
         invalidInput();
      }
      else
      {
         next--;
         
         if (next < 0)
         {
            if (chooseData.isNoError())
            {
               EventManager.postKey(Keyboard.KA_GO);
               return index;
            }
            
            // stop at the beginning of the rows and don't go past
            next = 0;
         }
      }

      return next;
   }

   /**
    * Get focused column widget.
    * 
    * @return  focused column widget.
    */
   public Widget<O> focusedCol()
   {
      return focusedCol;
   }

   /**
    * Get current CHOOSE index.
    * 
    * @return  current index.
    */
   public int index()
   {
      return index;
   }

   /**
    * Return <code>true</code> if last CHOOSe operation was finished by
    * internal GO event.
    * 
    * @return  Returns current state of <code>internalGo</code> flag.
    */
   public boolean isInternalGo()
   {
      return  internalGo;
   }
   
   /**
    * Check if CHOOSE is in field mode.
    * 
    * @return  <code>true</code> if field mode is active.
    */
   public boolean isFieldMode()
   {
      return chooseData.isFieldMode();
   }
}