/*
** Module   : ColorManager.java
** Abstract : A client-side color management class.
**
** Copyright (c) 2014-2017, Golden Code Development Corporation.
**
** -#- -I- --Date-- ---------------------------------Description----------------------------------
** 001 MAG 20140812 Created initial version
** 002 MAG 20140910 Added unload notification.
** 003 CA  20140926 Added shared widget configuration support, to allow GUI/ChUI concrete
**                  implementation for the same widget ID.  Refactored the widget configuration
**                  classes: all fields were made public; these classes need to be as dumb as
**                  possible, all logic related to setting/getting a certain field should be at
**                  the widget or at the caller. Refs #2254.
** 004 GES 20150425 Removed AWT dependencies and switch to a ColorRgb return value.
** 005 OM  20170320 Added support for Themes.
*/
/*
** This program is free software: you can redistribute it and/or modify
** it under the terms of the GNU Affero General Public License as
** published by the Free Software Foundation, either version 3 of the
** License, or (at your option) any later version.
**
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
** GNU Affero General Public License for more details.
**
** You may find a copy of the GNU Affero GPL version 3 at the following
** location: https://www.gnu.org/licenses/agpl-3.0.en.html
** 
** Additional terms under GNU Affero GPL version 3 section 7:
** 
**   Under Section 7 of the GNU Affero GPL version 3, the following additional
**   terms apply to the works covered under the License.  These additional terms
**   are non-permissive additional terms allowed under Section 7 of the GNU
**   Affero GPL version 3 and may not be removed by you.
** 
**   0. Attribution Requirement.
** 
**     You must preserve all legal notices or author attributions in the covered
**     work or Appropriate Legal Notices displayed by works containing the covered
**     work.  You may not remove from the covered work any author or developer
**     credit already included within the covered work.
** 
**   1. No License To Use Trademarks.
** 
**     This license does not grant any license or rights to use the trademarks
**     Golden Code, FWD, any Golden Code or FWD logo, or any other trademarks
**     of Golden Code Development Corporation. You are not authorized to use the
**     name Golden Code, FWD, or the names of any author or contributor, for
**     publicity purposes without written authorization.
** 
**   2. No Misrepresentation of Affiliation.
** 
**     You may not represent yourself as Golden Code Development Corporation or FWD.
** 
**     You may not represent yourself for publicity purposes as associated with
**     Golden Code Development Corporation, FWD, or any author or contributor to
**     the covered work, without written authorization.
** 
**   3. No Misrepresentation of Source or Origin.
** 
**     You may not represent the covered work as solely your work.  All modified
**     versions of the covered work must be marked in a reasonable way to make it
**     clear that the modified work is not originating from Golden Code Development
**     Corporation or FWD.  All modified versions must contain the notices of
**     attribution required in this license.
*/

package com.goldencode.p2j.ui.client;

import java.util.*;
import java.util.logging.*;

import com.goldencode.p2j.security.*;
import com.goldencode.p2j.ui.*;
import com.goldencode.p2j.ui.ColorTable.ColorPair;
import com.goldencode.p2j.ui.ColorTable.EnvironmentColorTable;
import com.goldencode.p2j.ui.chui.*;
import com.goldencode.p2j.ui.client.gui.theme.*;
import com.goldencode.p2j.util.*;

/**
 * Color configuration manager, for a specific client session.
 */
public class ColorManager
{   
   /** Return value when any error occurs. */
   public static final ColorRgb DEFAULT_RGB = new ColorRgb(255, 255, 255);
   
   /** Logger. */
   private static final Logger LOG = LogHelper.getLogger(ColorManager.class.getName());
   
   /** Context local instance of this class. */
   private static final ContextLocal<WorkArea> local = new ContextLocal<WorkArea>()
   {
      protected WorkArea initialValue()
      { 
         return new WorkArea();
      };
   };
   
   /**
    * Return system color for a given index. { @link SysColor }
    * 
    * @param    index
    *           System color index.
    *           
    * @return   System color RGB value.
    */
   public static ColorRgb getSysColor(int index)
   {      
      WorkArea wa = locate();
      
      if (wa.isChui)
      {
         return DEFAULT_RGB;
      }
      
      // use user system color table 
      if (checkRange(index, wa.systemColors.size()))
      {
         return wa.systemColors.get(index);
      }
      
      // use default color system table
      Map<Integer, ColorRgb> palette = ThemeManager.getCurrentTheme().getDefaultSystemPalette();
      if (checkRange(index, palette.size()))
      {
         return palette.get(index);
      }
      
      return DEFAULT_RGB;
   }

   /**
    * Get FGCOLOR RGB value for specified index from color table associate to a Window.
    * 
    * @param    window
    *           Window instance.
    * @param    index
    *           Color table index.
    *           
    * @return   FGCOLOR RGB value.
    */
   public static ColorRgb getFgColor(Window<?> window, int index)
   {
      return getColor(window, index, true);
   }
   
   /**
    * Get BGCOLOR RGB value for specified index from color table associate to a Window.
    * 
    * @param    window
    *           Window instance.
    * @param    index
    *           Color table index.
    *          
    * @return   BGCOLOR RGB value.
    */
   public static ColorRgb getBgColor(Window<?> window, int index)
   {
      return getColor(window, index, false);
   }
   
   /**
    * Get foreground or background RGB color of a standard color.
    * 
    * @param    window
    *           Window instance.
    * @param    stdColor
    *           Standard color.
    * @param    kind
    *           <code>true</code> get foreground color.
    *           <code>false</code> get background color.
    *           
    * @return   Standard color RGB value.
    */
   public static ColorRgb getStdColor(Window<?> window, String stdColor, boolean kind)
   {
      WorkArea wa = locate();
      
      // get environment color table for Window
      EnvironmentColorTable ect = wa.envColorTables.get(window.config().envName);

      return (ect == null) ? DEFAULT_RGB : getStdColor(ect, stdColor.toUpperCase(), kind);
   }

   /**
    * Update environments collection.  The environments are synchronized between server and
    * client using the state synchronization mechanism.
    * 
    * @param    envs
    *           An array of environments objects.
    */
   public static void update(List<EnvironmentColorTable> envs)
   {
      WorkArea wa = locate();
      
      for (EnvironmentColorTable env : envs)
      {
         wa.envColorTables.put(env.envName, env);
      }
   }
      
   /**
    * Get a RGB color value from color table for specified index. If index is out of range use
    * NORMAL standard color index.  The return value is in Java format.
    *   
    * @param    window
    *           Window instance.
    * @param    index
    *           Color table index.
    * @param    kind
    *           <code>true</code> get foreground color. <code>false</code> get background color.
    * 
    * @return   The RBG color value.
    */
   private static ColorRgb getColor(Window<?> window, int index, boolean kind)
   {
      WorkArea wa = locate();
      
      if (wa.isChui)
      {
         return DEFAULT_RGB;
      }
      
      // get environment color table for Window
      String envName = window.config().envName;
      
      EnvironmentColorTable ect = wa.envColorTables.get(envName);
      
      if (ect == null)
      {                         
         // if not in cache try to download from server
         ect = wa.server.readColorTable(envName);
         
         if (ect == null)
         {
            // This should never happen!
            final String msg = "Could not load color table of [%s] environment!";
            LOG.warning(String.format(msg, envName == null ? "default" : envName));
            
            return DEFAULT_RGB;
         }
         // cache environment color
         wa.envColorTables.put(envName, ect);
      }
      
      // check index range
      boolean inRange = checkRange(index, ect.colors.size());
      
      // if index is out of range get NORMAL color.
      return inRange ? ect.colors.get(index) : getStdColor(ect, ColorSpec.COLOR_NORMAL, kind);
   }

   /**
    * Check color table index range.
    *  
    * @param   index
    *          Color table index.
    * @param   size
    *          Color table size.
    * 
    * @return  <code>true</code> if index is in range 0..size <code>false</code> otherwise.
    */
   private static boolean checkRange(int index, int size)
   {
      return index >= 0 && index < size;
   }

   /**
    * Get foreground or background RGB color of a standard color.
    * 
    * @param    ect
    *           Environment color table.
    * @param    std
    *           Standard color.
    * @param    kind
    *           <code>true</code> to get the foreground color, <code>false</code> for the
    *           background color.
    *           
    * @return   Standard color RGB value.
    */
   private static ColorRgb getStdColor(EnvironmentColorTable ect, String std, boolean kind)
   {
      // get color index
      ColorPair pair = ect.standardColors.get(std);

      if (pair == null)
      {
         return DEFAULT_RGB;
      }
      
      int index = kind ? pair.fgColor : pair.bgColor;
      
      return checkRange(index, ect.colors.size()) ? ect.colors.get(index) : DEFAULT_RGB;         
   }

   /**
    * Unload environment. A notification from environment API whenever an 
    * <code>UNLOAD environment</code> statement is executed.
    * 
    * @param    envName
    *           Environment name.
    */
   public static void unload(String envName)
   {
      WorkArea wa = local.get(false);
      if (wa != null && envName != null)
      {
         wa.envColorTables.remove(envName);         
      }
   }
      
   /**
    * Locate the context-local data of this class.
    * 
    * @return   A {@link WorkArea} instance.
    */
   private static WorkArea locate()
   {
      return local.get();
   }
   
   /**
    * Context-local data for this color manager.
    */
   private static class WorkArea
   {
      /** A map of environment names to their associated color-table. */
      private Map<String, EnvironmentColorTable> envColorTables = new HashMap<>();

      /** System colors for this session. */
      private Map<Integer, ColorRgb> systemColors;

      /** The server exports. */
      private final ServerExports server;
                 
      /** Client driver mode. */
      private boolean isChui = false;

      /**
       * Create a new work-area.
       */
      public WorkArea()
      {
         isChui = OutputManager.getDriver().isChui();
         
         if (isChui)
         {
            server            = null;
            envColorTables    = null;
            systemColors      = null;
         }
         else
         {
            // initialize only in GUI mode
            server = ThinClient.getInstance().getServer();

            // get default color table
            EnvironmentColorTable ect = server.readColorTable(null);
            envColorTables.put(ect.envName, ect);
            
            // initialize system colors
            initSystemColorTable();
         }
      }
      
      /**
       * Initialize system colors table.
       */
      private void initSystemColorTable()
      {
         // get system colors from server
         ColorTableEntry[] colors = server.readSystemColorTable();
         
         // if system colors are missing use defaults
         if (colors.length == 0)
         {
            systemColors = ThemeManager.getCurrentTheme().getDefaultSystemPalette();
         }
         else
         {
            systemColors = new HashMap<>();
            
            for (ColorTableEntry color : colors)
            {
               systemColors.put(color.getIndex(), color);
            }
         }
      }
   }
}