/* 
** Module   : EditScroller.java
** Abstract : Defines APIs to allow a widget to scroll its content sideways, if the content does
**            not fit its native width (do not confuse with format length).
**
** Copyright (c) 2014-2017, Golden Code Development Corporation.
**
** -#- -I- --Date-- ---------------------------------Description----------------------------------
** 001 CA  20141208 Created initial version.
** 002 CA  20150805 Fixed FILL-IN cursor problems when editable width is smaller than content
**                  length.
*/
/*
** This program is free software: you can redistribute it and/or modify
** it under the terms of the GNU Affero General Public License as
** published by the Free Software Foundation, either version 3 of the
** License, or (at your option) any later version.
**
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
** GNU Affero General Public License for more details.
**
** You may find a copy of the GNU Affero GPL version 3 at the following
** location: https://www.gnu.org/licenses/agpl-3.0.en.html
** 
** Additional terms under GNU Affero GPL version 3 section 7:
** 
**   Under Section 7 of the GNU Affero GPL version 3, the following additional
**   terms apply to the works covered under the License.  These additional terms
**   are non-permissive additional terms allowed under Section 7 of the GNU
**   Affero GPL version 3 and may not be removed by you.
** 
**   0. Attribution Requirement.
** 
**     You must preserve all legal notices or author attributions in the covered
**     work or Appropriate Legal Notices displayed by works containing the covered
**     work.  You may not remove from the covered work any author or developer
**     credit already included within the covered work.
** 
**   1. No License To Use Trademarks.
** 
**     This license does not grant any license or rights to use the trademarks
**     Golden Code, FWD, any Golden Code or FWD logo, or any other trademarks
**     of Golden Code Development Corporation. You are not authorized to use the
**     name Golden Code, FWD, or the names of any author or contributor, for
**     publicity purposes without written authorization.
** 
**   2. No Misrepresentation of Affiliation.
** 
**     You may not represent yourself as Golden Code Development Corporation or FWD.
** 
**     You may not represent yourself for publicity purposes as associated with
**     Golden Code Development Corporation, FWD, or any author or contributor to
**     the covered work, without written authorization.
** 
**   3. No Misrepresentation of Source or Origin.
** 
**     You may not represent the covered work as solely your work.  All modified
**     versions of the covered work must be marked in a reasonable way to make it
**     clear that the modified work is not originating from Golden Code Development
**     Corporation or FWD.  All modified versions must contain the notices of
**     attribution required in this license.
*/

package com.goldencode.p2j.ui.client;

/**
 * In ChUI or GUI, widget content can be larger than the widget's editable width.  This can happen
 * in cases when the widget's WIDTH-CHARS/WIDTH-PIXELS attribute is smaller than the maximum length
 * as allowed by the widget's format, or in cases when the content width computed in native units
 * is larger than the editable width.
 * <p>
 * In such cases, the scrolling can be done via the cursor left/right, home or end keys.  This
 * class provides APIs which allow the widget to adjust its content, depending on the index on 
 * which the scrolling anchor is currently positioned and on the anchor's location.
 */
public abstract class EditScroller
{
   /** The anchor's location in the widget's content*/
   protected int anchorIndex;

   /**
    * Get the amount of characters which are over the left-most edge, and are not being displayed.
    * 
    * @return   See above.
    */
   public abstract int getDelta();

   /**
    * Adjust the location so that only the visible content will be displayed.
    *  
    * @param    txt
    *           The full content.
    * @param    np
    *           The current location of the content.
    *           
    * @return   The adjusted location.
    */
   public abstract NativePoint adjustLocation(String txt, NativePoint np);
   
   /**
    * Get the index of the leftmost visible character in the widget's displayable area.
    * 
    * @return   See above.
    */
   public abstract int getOffset();

   /**
    * Set the {@link #anchorIndex} to the new index.
    * 
    * @param    anchorIndex
    *           The new anchor index.
    */
   public void setAnchorIndex(int anchorIndex)
   {
      this.anchorIndex = anchorIndex;
   }

   /**
    * Get the anchor.
    * 
    * @return   The {@link #anchorIndex}.
    */
   public int getAnchorIndex()
   {
      return anchorIndex;
   }
}