/*
** Module   : FocusManager.java
** Abstract :
**
** Copyright (c) 2010-2018, Golden Code Development Corporation.
**
** -#- -I- --Date-- ---------------------------------Description----------------------------------
** 001 SIY 20100915 Created initial version
** 002 SIY 20101015 Changes related to refactoring of Screen.
** 003 SIY 20101122 Switching from KeyEvent to KeyInput.
** 004 SIY 20111101 Moved from CHUI package. Minor cleanups/refactorings in preparation to GUI 
**                  client.
** 005 CA  20140805 Explicitly import any chui package classes.
** 006 VIG 20140919 Changed status method call.
** 007 CA  20141119 Refactored drawing code to enable multiple OS windows, for GUI support.  Misc
**                  changes/fixes related to drawing on multiple windows.
** 008 HC  20150221 Improvements to frame tab order.
** 009 VIG 20150224 Added menu support to getFirstEnabledWidget
** 010 SVL 20150608 Reflect API change in fill-in.
** 011 IAS 20150801 TAB-STOP, WIDGET-LEAVE, WIDGET-ENTER attributes support
** 012 CA  20150910 Fixed a NPE related to focus change between frame and non-frame widgets.
** 013 CA  20151006 Changes to allow multi-window focus management in GUI.
** 014 HC  20151013 Changes to extend GUI multi-window focus management and ACTIVE-WINDOW system
**                  handle processing to modal windows.
** 015 CA  20151123 Fixed focus traversing when there are nested frames.
**     VIG 20151125 Fixed IllegalStateException exception in widgets processing in focusChange().
** 016 HC  20151213 Improved DIALOG-BOX window parenting logic and related changes.
** 017 VIG 20151226 oldFocus.isEnabled() is changed to oldFocus.navigable() in setFocusOn().
**                  Fixed NullPointerException when using WindowManager.topLevelWindow().
** 016 IAS 20160222 SET-WAIT-STATE support.
** 017 SVL 20160314 Fixed the focus processing case when the parent browse couldn't be found for
**                  an in-browse editor.
** 018 HC  20160511 Fixed COMBO-BOX focusing issues and related changes.
** 019 HC  20160607 GUI menus reworked to be displayed in dedicated top-level windows.
** 020 HC  20161018 Fixed an abend when focus was switched between BROWSE FILL-INs.
** 021 EVL 20161027 Fix for sending LEAVE event in case of the frame has only one focusable widget
**                  and become a target of the LEAVE in addition to actual focusable widget
**                  inside.
** 022 SBI 20161101 Changed focus traversal to use the provided tab item list.
** 023 CA  20171218 Changes to better handle the LAST-EVENT:WIDGET-ENTER/LEAVE on focus change.
** 024 CA  20180202 Minor performance improvement for parent() (avoid multiple calls).
** 025 OM  20180321 When processing system dialogs do not collaborate with server.
** 026 HC  20180509 Improved MENU focus management.
*/

/*
** This program is free software: you can redistribute it and/or modify
** it under the terms of the GNU Affero General Public License as
** published by the Free Software Foundation, either version 3 of the
** License, or (at your option) any later version.
**
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
** GNU Affero General Public License for more details.
**
** You may find a copy of the GNU Affero GPL version 3 at the following
** location: https://www.gnu.org/licenses/agpl-3.0.en.html
** 
** Additional terms under GNU Affero GPL version 3 section 7:
** 
**   Under Section 7 of the GNU Affero GPL version 3, the following additional
**   terms apply to the works covered under the License.  These additional terms
**   are non-permissive additional terms allowed under Section 7 of the GNU
**   Affero GPL version 3 and may not be removed by you.
** 
**   0. Attribution Requirement.
** 
**     You must preserve all legal notices or author attributions in the covered
**     work or Appropriate Legal Notices displayed by works containing the covered
**     work.  You may not remove from the covered work any author or developer
**     credit already included within the covered work.
** 
**   1. No License To Use Trademarks.
** 
**     This license does not grant any license or rights to use the trademarks
**     Golden Code, FWD, any Golden Code or FWD logo, or any other trademarks
**     of Golden Code Development Corporation. You are not authorized to use the
**     name Golden Code, FWD, or the names of any author or contributor, for
**     publicity purposes without written authorization.
** 
**   2. No Misrepresentation of Affiliation.
** 
**     You may not represent yourself as Golden Code Development Corporation or FWD.
** 
**     You may not represent yourself for publicity purposes as associated with
**     Golden Code Development Corporation, FWD, or any author or contributor to
**     the covered work, without written authorization.
** 
**   3. No Misrepresentation of Source or Origin.
** 
**     You may not represent the covered work as solely your work.  All modified
**     versions of the covered work must be marked in a reasonable way to make it
**     clear that the modified work is not originating from Golden Code Development
**     Corporation or FWD.  All modified versions must contain the notices of
**     attribution required in this license.
*/

package com.goldencode.p2j.ui.client;

import java.util.*;

import com.goldencode.p2j.ui.*;
import com.goldencode.p2j.ui.chui.ThinClient;
import com.goldencode.p2j.ui.client.event.*;
import com.goldencode.p2j.ui.client.widget.*;

/**
 * Focus manager.
 * 
 * @author 	SIY
 * @version 1.0
 */
public class FocusManager
{
   /** State of the focus on entry to waitFor: undefined */
   public static final int FOCUS_UNDEF = 0;

   /** State of the focus on entry to waitFor: invalid */
   public static final int FOCUS_INVALID = 1;
   
   /** State of the focus on entry to waitFor: valid */
   public static final int FOCUS_VALID = 2;
   
   /** 
    * State of the focus on entry to waitFor: hidden, as in not visible.
    * This is not the HIDDEN flag state.
    */
   public static final int FOCUS_HIDDEN = 3;

   /** Owner {@link ThinClient} instance. */
   private final ThinClient tc;

   /** Do not adjust focus in WAIT-FOR if set */
   private boolean noFocusAdjustment = false;

   /** Lookup mode for focusable widgets in frames: wrap */
   public static final boolean FOCUS_WRAP   = true;

   /** Lookup mode for focusable widgets in frames: no wrap */
   public static final boolean FOCUS_NOWRAP = false;

   /** Flag indicating that we are within a temporary focus change */
   private boolean tempFocusChange = false;

   /**
    * Constructor.
    * 
    * @param   tc
    *          Reference to {@link ThinClient}.
    */
   public FocusManager(ThinClient tc) 
   {
      this.tc = tc;
   }
   
   /**
    * Adjusts focus if it is either undefined or invalid and a trigger has
    * enabled at least one widget.
    */
   public void adjustFocus()
   {
      // get the current focus
      Widget focus = UiUtils.getCurrentFocus();
      Widget next  = null;
   
      // get the focus state: either undefined or invalid
      int focusState = focusState(focus);
   
      // if focus is invalid, the next enabled widget in the frame gets it
      if (focusState == FOCUS_INVALID || focusState == FOCUS_HIDDEN)
      {
         Frame frame = (Frame)UiUtils.locateFrame(focus);
         
         if (frame == null)
         {
            return;
         }

         // if invalid, search for next enabled widget, with wrapping
         next = frame.getNextEnabledWidget(focus, FocusManager.FOCUS_WRAP);
   
         if (next != null)
         {
            // set focus on the next enabled widget silently
            setFocusOn(next);
   
            return;
         }
         
         // if focus is disabled and there is no other enabled widget, 
         // then focus is not adjusted if other widget type beside combo or 
         // fill-in is in focus and the last trigger returned normal
         boolean b = tc.isTriggerWasAppliedWithoutWaitFor();
         if (b && !(focus instanceof FillIn || focus instanceof ComboBox))
         {
            // this is only if there was no nested wait-for loop executed
            // by the trigger
            
            if (frame.hasEnabled())
               return;
         }
      }
   
      // if focus is undefined, the very first enabled widget gets it
      next = getFirstEnabledWidget();
   
      if (next == null)
      {
         // still nothing
         // set focus to undefined, if no enabled widget found
         WindowManager.topLevelWindowOptional().ifPresent(w -> w.setFocus(null));
         // set the status to default if there is no enabled widget
         WindowManager.resolveWindow().status(null);
         return;
      }
   
      // set focus on the first enabled widget silently
      setFocusOn(next);
   
      // adjust zap and insert state as if ENTRY was sent
      if (focus instanceof FillIn && 
          next instanceof FillIn  && 
          focusState == FOCUS_HIDDEN)
      {
         if (((FillIn) focus).isZapActive())
         {
            FillIn fNext  = (FillIn) next;
            
            fNext.setZapMode(true);
            fNext.setInsertMode(false);
         }
      }
   }

   /**
    * Check if we are within a temporary focus change
    *  
    * @return   <code>true</code> if we are within a temporary focus change
    */
   public boolean isTempFocusChange()
   {
      return tempFocusChange;
   }

   /**
    * Sets focus on the specified widget silently and unconditionally.
    *
    * @param  widget
    *         widget to set focus on; it should be a valid widget, but
    *         can be disabled  
    */
   public void setFocusOn(final Widget widget)
   {
      Widget oldFocus = UiUtils.getCurrentFocus();
      boolean ins = Window.getInsertMode() == Window.INSERT_MODE_INS;
      tc.eventDrawingBracket(widget, false, false, new Runnable()
      {
         @Override
         public void run()
         {
            FocusManager.setInitialFocus(widget);
         }
      });
   
      Widget newFocus = UiUtils.getCurrentFocus();
      if (newFocus != null &&
          newFocus instanceof FillIn)
      {
         FillIn newFillIn = (FillIn) newFocus;
   
         if (oldFocus == null)
         {
            newFillIn.setInsertMode(ins);
         }
         else if (newFocus != oldFocus &&
                  (!oldFocus.navigable() || !oldFocus.isVisible()))
         {
            if (oldFocus instanceof Editor)
            {
               newFillIn.setInsertMode(false);
            }
            else
            {
               newFillIn.setInsertMode(ins);
   
               if (oldFocus instanceof FillIn &&
                   !((FillIn) oldFocus).isZapActive() &&
                   !newFillIn.isLinkedToBrowse())
               {
                  newFillIn.setPendingNoZap();
               }
            }
         }
      }
   }

   /**
    * Looks up the very first enabled widget among all displayed frames and menubar.
    *
    * @return the first enabled widget or <code>null</code>
    */
   public Widget getFirstEnabledWidget()
   {
      // loop through the displayed frames
      TopLevelWindow<OutputManager<?>> topLevelWnd = WindowManager.topLevelWindow();
      Frame frame = topLevelWnd == null ? null : topLevelWnd.getFirstVisibleFrame();
      
      while (frame != null)
      {
         if (frame.isAlive())
         {
            Widget comp = frame.getFirstEnabledWidget();
            
            if (comp != null) {
               return comp;
            }
         }
   
         // try next frame
         // TODO: dialog-window
         frame = frame.window().getNextVisibleFrame(frame);
      }
   
      // no enabled widgets in all frames
      
      if (frame == null)
      {
         Window<?> wnd = WindowManager.resolveWindow();
         if (wnd.config.menubarId != 0)
         {
            WidgetRegistry<?> registry = tc.getOutputManager().getRegistry();
            Menu<?> menu = (Menu<?>) registry.getComponent(wnd.config.menubarId);
            return menu.getFirstEnabledWidget();
         }
      }
      
      return null; 
   }

   /**
    * Sets initial focus for the WAIT-FOR with invalid focus on entry
    * and no explicit FOCUS specification.
    *
    * @param  inFocus
    *         widget in focus
    * @return <code>true</code> if WAIT-FOR can proceed
    */
   public boolean focusInvalidImplicit(Widget inFocus)
   {
      this.noFocusAdjustment = false;
   
      // figure out the frame
      Frame frame = (Frame)UiUtils.locateFrame(inFocus);
   
      // find the first enabled widget that follows
      Widget first = frame.getNextEnabledWidget(inFocus, FocusManager.FOCUS_WRAP);
   
      if (first == null)
      {
         // no enabled widgets within the same frame; check others
         first = getFirstEnabledWidget();
   
         if (first == null)
         {
            // no enabled widgets at all; focus becomes undefined
            WindowManager.getDefaultWindow().setFocus(null);
   
            return true;
         }
      }
   
      // set focus on the found widget
      setFocusOn(first);
   
      return true;
   }

   /**
    * Sets initial focus for the WAIT-FOR with valid focus on entry
    * and an explicit FOCUS specification.
    *
    * @param    inFocus
    *           widget in focus
    * @param    toFocus
    *           candidate widget for receiving focus
    * @param    list
    *           list of event to wait for 
    * @param    cop
    *           currently executed operation
    * @param    hidden
    *           <code>true</code> if this is the case of hidden focus
    *         
    * @return   <code>true</code> if WAIT-FOR can proceed
    */
   public boolean focusValidExplicit(Widget inFocus, 
                                      Widget toFocus,
                                      EventList list, 
                                      int       cop,
                                      boolean   hidden)
   {
      this.noFocusAdjustment = false;
   
      // check the target's enabled status
      if (!toFocus.focusTraversable())
      {
         // leave the focus where it is
         setFocusOn(inFocus);
   
         return true;
      }
      
      Widget inFrame = UiUtils.locateFrame(inFocus);
      Frame toFrame = (Frame) UiUtils.locateFrame(toFocus);
      boolean sameFrame = inFrame == toFrame;
      boolean fastExit = false;
   
      boolean isUpdate = cop == EventList.UPDATE || 
                         cop == EventList.SET    || 
                         cop == EventList.PROMPT_FOR;
      boolean isWaitFor = cop == EventList.WAIT_FOR;
   
      if (isWaitFor && hidden)
      {
         Widget first = getFirstEnabledWidget();
         Frame firstFrame = (Frame) UiUtils.locateFrame(first);
   
         // if previous focus is hidden and explicit focus is the first 
         // focusable widget on screen, adjust focus to it
         if (toFocus == first)
         {
            setFocusOn(toFocus);
         }
         else
         {
            if (inFrame.isVisible() == false && inFrame == firstFrame)
            {
               // if the previous focus frame is hidden, focus is set to
               // undefined
                WindowManager.getDefaultWindow().setFocus(null);
            }
            // else, search for focus. if explicit focus was reached, adjust
            // focus to it, but do not send ENTRY - it will be sent later
            fastExit = searchForFocusWithEntry(first, toFocus);
         }
      }
      
      // remove highlight from the previous focus
      tc.highlight(inFocus, false);
   
      Widget next = toFocus;
   
      // scan enabled widgets of the frame and chose the first accepting ENTRY
      while (next != null)
      {
         this.noFocusAdjustment = true;
         boolean state = next.focusTraversable();
   
         // send ENTRY event and see the response
         boolean check = tc.sendEntry(next);
   
         if (check || (sameFrame && next == inFocus) || fastExit)
         {
            // no-apply must be ignored if X and Y are in the same frame and
            // X is reached
            // no-apply must be ignored if it was focused due to hidden 
            // initial focus
   
            if (widgetLostFocusability(inFocus, next, state, FOCUS_VALID))
            {
               return true;
            }
   
            // ENTRY accepted
            this.noFocusAdjustment = false;
            break;
         }
         
         // ENTRY not accepted, remove highlight
         next.setHighlighted(false);
   
         // look up the next candidate
         next = toFrame.getNextEnabledWidget(next, FocusManager.FOCUS_NOWRAP);
      }
   
      if (next == null && sameFrame == false) 
      {
         // if focus was not found and X and Y are not in the same frame, then:
         // 1. if event list waits for Y frame events, end, sending leave to X
         // 2. else, focus another widget without ENTRY, starting from X
         if (isUpdate                                     ||
             list.anyEventForWidget(UiUtils.lookupWidgetId(toFocus)) ||
             list.anyEventForWidget(UiUtils.lookupWidgetId(toFrame)))
         {
            tc.sendLeave(inFocus, null);
   
            return false;
         }
   
         if (!searchForFocusWithoutEntry(inFocus))
         {
            // focus remains null if no widget found
            WindowManager.getDefaultWindow().setFocus(null);
         }
   
         return true;
      }
   
      // check the success
      if (next == null)
      {
         // no widget wants focus; check event list for global events
         if (!UiUtils.checkEventList(list))
         {
            // no global events; should exit after sending LEAVE to inFocus
            tc.sendLeave(inFocus, null);
   
            return false;
         }
         
         // leave the focus where it is
         setFocusOn(inFocus);
   
         return true;
      }
   
      // a good candidate is found; set focus on it
      setFocusOn(next);
   
      return true;
   }

   /**
    * Sets initial focus for the WAIT-FOR with undefined or hidden focus on 
    * entry and no explicit FOCUS specification.
    *
    * @param    list
    *           list of event to wait for 
    * @param    hidden
    *           <code>true</code> if this is the case of hidden initial focus
    * 
    * @return   <code>true</code> if WAIT-FOR can proceed
    */
   public boolean focusUndefImplicit(EventList list, boolean hidden)
   {
      this.noFocusAdjustment = false;
   
      // find the very first enabled widget
      Widget first = getFirstEnabledWidget();
      if (first == null)
      {
         // no enabled widgets; check event list for global events
         if (!UiUtils.checkEventList(list))
         {
            // no global events; should exit immediately
            return false;
         }
      }
   
      // found an enabled widget; figure out the owning frame
      Frame frame = (Frame)UiUtils.locateFrame(first);
   
      // find the first enabled widget of the frame which accepts ENTRY
      while (first != null)
      {
         this.noFocusAdjustment = true;
         boolean state = first.focusTraversable();
   
         // send ENTRY event and see the response
         boolean check = tc.sendEntry(first);
         
         if (check)
         {
            if (state && first.focusTraversable() != state)
            {
               if (hidden == false || 
                  searchForFocusWithoutEntry(first) == false)
               {
                  // focus was disabled during ENTRY, focus remains undefined
                  WindowManager.getDefaultWindow().setFocus(null);
               }
   
               return true;
            }
   
            // ENTRY accepted
            this.noFocusAdjustment = false;            
            break;
         }
         
         // ENTRY not accepted, remove highlight
         first.setHighlighted(false);
   
         first = frame.getNextEnabledWidget(first, FocusManager.FOCUS_NOWRAP);
      }
   
      // end of search; either found or hit end of frame
      if (first == null)
      {
         // no enabled widgets; check event list for global events
         if (!UiUtils.checkEventList(list))
         {
            // no global events; should exit immediately
            return false;
         }
         
         // focus remains undefined
         return true;
      }
   
      // set focus on the found widget
      setFocusOn(first);
   
      return true;
   }

   /**
    * Sets initial focus for the WAIT-FOR with undefined focus on entry
    * and an explicit FOCUS specification.
    *
    * @param  toFocus
    *         candidate widget for receiving focus
    * @param  list
    *         list of event to wait for
    * @param  cop
    *         currently executed operation
    * @param  explicit
    *         when <code>true</code>, it marks a real explicit focus; else, it
    *         marks explicit focus generated due to previous pending focus, in
    *         non-nested trigger mode.
    * 
    * @return <code>true</code> if WAIT-FOR can proceed
    */
   @SuppressWarnings("null")
   public
   boolean focusUndefExplicit(Widget toFocus, 
                                      EventList list, 
                                      int cop,
                                      boolean explicit)
   {
      this.noFocusAdjustment = false;
      // if focus candidate is not the very first enabled widget on the screen
      // ENTRY may be sent to the latter first
      Widget first  = getFirstEnabledWidget();
      Frame firstFrame = (Frame) UiUtils.locateFrame(first);
      Frame toFrame = (Frame) UiUtils.locateFrame(toFocus);
   
      boolean isUpdate = cop == EventList.UPDATE || 
                         cop == EventList.SET    || 
                         cop == EventList.PROMPT_FOR;
      boolean isWaitFor = cop == EventList.WAIT_FOR;
   
      // check the target's enabled status
      if (!toFocus.focusTraversable())
      {
         // find the very first enabled widget
         if (first == null)
         {
            // no enabled widgets; focus remains undefined
            return true;
         }
   
         if (isWaitFor)
         {
            
            // search for focus starting from the first widget sending ENTRY
            searchForFocusWithEntry(first, null);
            Widget focused = UiUtils.getCurrentFocus();
            
            if (focused == first && !first.focusTraversable())
            {
               // if first enabled widget on screen accepted ENTRY and got  
               // disabled, focus is undefined
               WindowManager.getDefaultWindow().setFocus(null);
            }
            // if no focus was found, then current operation must end
            if (focused == null)  
            {
               return false;
            }
         }
         else
         {
            // set focus on the found widget without sending ENTRY, if not
            // WAIT-FOR
            setFocusOn(first);
         }
   
         return true;
      }
      
   
      // send entry to the first enabled widget on screen W if this is the
      // wait-for operation and the target focus is not W
      boolean entryFirst = cop == EventList.WAIT_FOR && 
                           first != null             && 
                           first != toFocus;
      boolean fastExit = false;
   
      if (isUpdate && tc.getTriggerNesting() == 0 && tc.getPendingFocus() != null)
      {
         Frame pendingFrame = (Frame) UiUtils.locateFrame(tc.getPendingFocus());
   
         // before sending ENTRY to any widget, adjust focus to last widget
         // which had ENTRY only if the frame is not out of scope
         if (pendingFrame != null && pendingFrame.isAlive())
         {
            if (tc.getPendingFocus().focusTraversable())
            {
               setFocusOn(tc.getPendingFocus());
            }
            // in this case, if explicit focus is disabled during ENTRY, focus
            // search starts from this 'implicit' focus
            first = tc.getPendingFocus();
         }
         
         // the pending focus is consumed on first usage
         tc.setPendingFocus(null);
      }
      
      if (entryFirst)
      {
         fastExit = searchForFocusWithEntry(first, toFocus);
      }
   
      // scan enabled widgets of the frame and chose the first accepting ENTRY
      Widget next = toFocus;
   
      if (next != null && entryFirst)
      {
         // at least one more ENTRY event will be sent, remove highlight from
         // the widget
         first.setHighlighted(false);
      }
      
      while (next != null)
      {
         this.noFocusAdjustment = true;
         boolean state = next.focusTraversable();
         
         // send ENTRY event and see the response
         boolean check = tc.sendEntry(next);
   
         if (check || fastExit)
         {
            // ignore NO-APPLY if there is a fast exit
   
            if (state && next.focusTraversable() != state)
            {
               boolean undefined = false;
               Widget nextFrame = UiUtils.locateFrame(next);
   
               if (isUpdate || (isWaitFor && firstFrame == nextFrame))
               {
                  undefined = searchForFocusWithoutEntry(first) == false;
               }
               
               if (isWaitFor && firstFrame != nextFrame)
               {
                  undefined = (first != null && 
                               first.focusTraversable()) || 
                               searchForFocusWithoutEntry(first) == false;
               }
               
               if (undefined)
               {
                  WindowManager.getDefaultWindow().setFocus(null);
               }
               
               return true;
            }
   
            // ENTRY accepted
            this.noFocusAdjustment = false;
            break;
         }
   
         // ENTRY not accepted, remove highlight
         next.setHighlighted(false);
   
         // look up the next candidate
         next = toFrame.getNextEnabledWidget(next, FocusManager.FOCUS_NOWRAP);
      }
   
      // check the success
      if (next == null)
      {
         // if event list has events for toFocus or toFocus frame, 
         // wait-for ends, only if explicit is not caused by pending focus
         if (explicit &&
             (isUpdate                                     ||
              list.anyEventForWidget(UiUtils.lookupWidgetId(toFocus)) ||
              list.anyEventForWidget(UiUtils.lookupWidgetId(toFrame))))
         {
            if (tc.getPendingFocus() != null)
            {
               // send LEAVE, ignoring NO-APPLY
               tc.sendLeave(tc.getPendingFocus(), null);
   
               // set focus to undefined
               WindowManager.getDefaultWindow().setFocus(null);
               // 'consume' the focus
               tc.setPendingFocus(null);
            }
            return false;
         }
         
         // no widget wants focus; set focus on first enabled widget on screen
         // or undefined, if none found
         next = getFirstEnabledWidget();
         if (next == null)
         {
            WindowManager.getDefaultWindow().setFocus(null);
         }
         else
         {
            setFocusOn(next);
            // although ENTRY was not accepted, focus was changed
            this.noFocusAdjustment = false;
         }
   
         return true;
      }
   
      // a good candidate is found; set focus on it
      setFocusOn(next);
   
      return true;
   }

   /**
    * Sets initial focus for the WAIT-FOR with invalid focus on entry
    * and an explicit FOCUS specification.
    *
    * @param  inFocus
    *         widget in focus
    * @param  toFocus
    *         candidate widget for receiving focus
    * @param  list
    *         list of event to wait for
    *
    * @return <code>true</code> if WAIT-FOR can proceed
    */
   public boolean focusInvalidExplicit(Widget inFocus, Widget toFocus,
                                        EventList list)
   {
      this.noFocusAdjustment = false;
   
      Frame frame = null;
      Widget next = null;
   
      // check the target's enabled status
      if (!toFocus.focusTraversable())
      {
         if (searchForFocusWithoutEntry(inFocus) == false)
         {
            // no enabled widgets at all; focus becomes undefined
            WindowManager.getDefaultWindow().setFocus(null);
         }
   
         return true;
      }
   
      // scan enabled widgets of the frame and chose the first accepting ENTRY
      frame = (Frame)UiUtils.locateFrame(toFocus);
      next = toFocus;
   
      while (next != null)
      {
         this.noFocusAdjustment = true;
         boolean state = next.focusTraversable();
   
         // send ENTRY event and see the response
         boolean check = tc.sendEntry(next);
   
         if (check)
         {
            // check if the widget was disabled during trigger
            if (widgetLostFocusability(inFocus, next, state, FOCUS_INVALID))
            {
               return true;
            }
   
            // ENTRY accepted
            this.noFocusAdjustment = false;
            break;
         }
         
         // ENTRY not accepted, remove highlight
         next.setHighlighted(false);
   
         // look up the next candidate
         next = frame.getNextEnabledWidget(next, FocusManager.FOCUS_NOWRAP);
      }
   
      // check the success
      if (next == null)
      {
         // no widget wants focus; check event list for global events
         if (!UiUtils.checkEventList(list))
         {
            // no global events; should exit immediately
            return false;
         }
   
         // set focus on toFocus ignoring NO-APPLY 
         setFocusOn(toFocus);
   
         return true;
      }
   
      // a good candidate is found; set focus on it
      setFocusOn(next);
   
      return true;
   }

   /**
    * If the toFocus widget lost its focusability, then:
    * <ul>
    *    <li>if the inFocus and toFocus widgets are both in the same frame,
    *        search for focus starting from:
    *          <ul>
    *             <li>toFocus, if inFocus is invalid</li>
    *             <li>inFocus, if inFocus is valid</li>
    *          </ul>
    *    </li>
    *    <li>else, focus is undefined if:
    *       <ul>
    *          <li>inFocus is invalid</li>
    *          <li>inFocus is valid and still focusable</li>
    *          <li>inFocus is valid, not focusable: focus first enabled and
    *              visible widget on screen, without ENTRY. if none found,
    *              focus is undefined.
    *          </li>
    *       </ul>
    *    </li>
    * </ul>
    * 
    * @param    inFocus
    *           the focus before the ENTRY event was triggered
    * @param    toFocus
    *           the widget which had ENTRY event
    * @param    state
    *           previous state of this widget (focusable or not)
    * @param    focusState
    *           the state of the focused widget, as before the trigger was
    *           invoked (valid or invalid).
    * 
    * @return   <code>true</code> if this widget lost focusability
    */
   boolean widgetLostFocusability(Widget inFocus,
                                          Widget toFocus,
                                          boolean   state, 
                                          int       focusState)
   {
      if (state && toFocus.focusTraversable() != state)
      {
         boolean undefined = false;
         
         if (UiUtils.locateFrame(inFocus) == UiUtils.locateFrame(toFocus))
         {
            // Y was disabled, search for focus without ENTRY
            Widget start = focusState == FOCUS_VALID ? inFocus : toFocus;
            undefined = searchForFocusWithoutEntry(start) == false;
         }
         else
         {
            // if X is still enabled and visible or no other widget to
            // set focus to, focus is undefined
            undefined = focusState == FOCUS_VALID && 
                         (inFocus.focusTraversable() ||
                         searchForFocusWithoutEntry(inFocus) == false);
         }
         
         if (undefined)
         {
            // focus remains undefined
            WindowManager.getDefaultWindow().setFocus(null);
         }
   
         return true;
      }
      
      return false;
   }

   /**
    * Search for focus with ENTRY, starting from the specified widget,
    * without wrapping. If no enabled and visible widget which accepts entry
    * found in current frame, focus is undefined.
    * 
    * @param    start
    *           the widget from which focus search start
    * @param    toFocus 
    *           If search reaches this widget, sending ENTRY will be skipped.
    *           
    * @return   <code>true</code> if a widget was found
    */
   boolean searchForFocusWithEntry(Widget start, Widget toFocus)
   {
      Frame startFrame = (Frame) UiUtils.locateFrame(start);
      
      // send ENTRY to the specified widget
      boolean check = tc.sendEntry(start);
   
      if (check == false)
      {
         // the specified widget did not accept entry
         // send entry to next widgets, in tab-order
         Widget next = startFrame.getNextEnabledWidget(start, FocusManager.FOCUS_NOWRAP);
         while (next != null)
         {
            if (next == toFocus)
            {
               // reached explicit focus, it must receive focus before sending
               // ENTRY to it
               setFocusOn(next);
   
               return true;
            }
   
            // send ENTRY event
            check = tc.sendEntry(next);
            if (check)
            {
               // focus found
               break;
            }
            
            next = startFrame.getNextEnabledWidget(next, FocusManager.FOCUS_NOWRAP);
         }
         
         if (next == null)
         {
            // no widget accepts ENTRY in firstFrame, focus is undefined
            WindowManager.getDefaultWindow().setFocus(null);
         }
      }
      
      return false;
   }

   /**
    * Sets initial focus for the WAIT-FOR with valid focus on entry
    * and no explicit FOCUS specification.
    *
    * @param  inFocus
    *         widget in focus
    * @return <code>true</code> always to indicate success
    */
   public boolean focusValidImplicit(Widget inFocus)
   {
      this.noFocusAdjustment = false;
      boolean state = inFocus.focusTraversable();
   
      // currently focused widget receives ENTRY event
      // NO-APPLY, if returned by a trigger, is ignored
      tc.sendEntry(inFocus);
   
      if (state && inFocus.focusTraversable() != state)
      {
         // if it was disabled or hidden by the trigger, readjust focus,
         // starting from current frame
         if (searchForFocusWithoutEntry(inFocus) == false)
         {
            // focus remains undefined
            WindowManager.getDefaultWindow().setFocus(null);
         }
      }
   
      // currently focused widget remains in focus or focus is undefined
      return true;
   }

   /**
    * Sets initial focus for the WAIT-FOR with valid focus on entry
    * and no explicit FOCUS specification.
    *
    * @param  inFocus
    *         widget in focus
    * @return <code>true</code> always to indicate success
    */
   public static int focusState(Widget inFocus)
   {
      if (inFocus == null)
      {
         return FOCUS_UNDEF;
      }
      
      if (inFocus.isVisible() == false)
      {
         return FOCUS_HIDDEN;
      }
   
      if (inFocus.isEnabled())
      {
         return FOCUS_VALID;
      }
   
      return FOCUS_INVALID;
   }

   /**
    * Search for focus without ENTRY, starting from the specified widget,
    * with wrapping in its frame. If no enabled and visible widget found in
    * current frame, search for focus starting with first frame on screen.
    * 
    * @param    start
    *           the widget from which focus search start
    * 
    * @return   <code>true</code> if a widget was found
    */
   public boolean searchForFocusWithoutEntry(Widget start)
   {
      if (start == null)
      {
         return false;
      }
      
      Frame frame = (Frame) UiUtils.locateFrame(start);
      
      if (frame.isVisible())
      {
         Widget next = start;
         
         do
         {
            if (next.focusTraversable())
            {
               // widget was found, request focus
               setFocusOn(next);
               return true;
            }
            
            next = frame.getNextEnabledWidget(next, FocusManager.FOCUS_WRAP);
         }
         while (next != null);
      }
      
      Widget first = getFirstEnabledWidget();
      if (first != null)
      {
         setFocusOn(first);
         return true;
      }
   
      // no widget was found
      return false;
   }

   /**
    * Initiates the focus change procedure by creating and handling a LEAVE
    * event and, if that event is not consumed by a trigger, proceeds with
    * the focus change.
    * 
    * @param   widget
    *          <code>Component</code> that is the candidate to receive focus.
    *          
    * @return  <code>true</code> if focus change is successful.
    */
   public static boolean tryFocusChange(Widget widget)
   {
      return tryFocusChange(widget, true, true);
   }

   /**
    * Initiates the focus change procedure by creating and handling a LEAVE
    * event and, if that event is not consumed by a trigger, proceeds with
    * the focus change.
    * <p>
    * Conditionally issues ENTRY event that follows the focus change.
    * 
    * @param   widget
    *          <code>Component</code> that is the candidate to receive focus.
    *          
    * @param   doLeave
    *          <code>true</code> means LEAVE event will be issued
    *          
    * @param   doEntry
    *          <code>true</code> means ENTRY event will be issued
    *          
    * @return  <code>true</code> if focus change is successful.
    */
   public static boolean tryFocusChange(Widget widget, boolean doLeave,
                                         boolean doEntry)
   {
      if (ThinClient.isWaitState())
      {
         return false;
      }
      // safety check
      if (widget == null)
         return false;
   
      Widget inFocus = UiUtils.getCurrentFocus();
      
      if (inFocus == widget)
         return false;
      
      if (inFocus != null && inFocus.isEnabled() && doLeave)
      {
         // create and process the LEAVE event first so that any trigger
         // defined for that event gets its chance to run
         KeyInput leave = new KeyInput(inFocus,
                                       EventType.KEY_PRESSED,
                                       Keyboard.SE_LEAVE);
   
         ThinClient.getInstance().processProgressEvent(leave);
   
         // check to see if a trigger returned no-apply
         if (leave.isConsumed())
            return false;
      }
   
      // if there is no ENTRY sent, focus must not be adjusted
      if (!doEntry)
         return true;
   
      // change the focus temporarily now
      widget.requestFocus();
      
      // create and process the ENTRY event now
      KeyInput entry = new KeyInput(widget,
                                    EventType.KEY_PRESSED,
                                    Keyboard.SE_ENTRY);
      ThinClient.getInstance().processProgressEvent(entry);
   
      // check to see if a trigger returned no-apply
      if (!entry.isConsumed())
         return true;
   
      // reverse the focus change
      if (inFocus != null)
         inFocus.requestFocus();
      else
         widget.parent().setFocus(null);
   
      return false;
   }

   /**
    * Performs conditional focus change procedure to the next or previous
    * widget in the tab order. The focus change can be cancelled by a
    * trigger.
    * 
    * @param   widget
    *          <code>Component</code> that is the reference widget, which is
    *          currently retaining focus.
    * @param   afterTab
    *          flag indication if focus change was initiate by TAB/BACK-TAB 
    * @param   direction
    *          The direction of focus change. If this parameter is set to
    *          <code>true</code> then next in Tab order widget is selected.
    *          
    * @return  <code>true</code> if the focus change is successful.
    */
   public boolean focusChange(Widget widget, boolean afterTab, boolean direction)
   {
      if (widget == null)
         return false;
      
      // if there is no other focusable widget in the parent container, then
      // there won't be an actual focus change, just return before any events
      // are generated
      boolean possibleFocus = false;
      Widget test = widget;
      // initial LEAVE event target
      Widget leaveTarget = widget;
      Browse parentBrowse = (Browse) test.parent(Browse.class);
      
      Container<?> widgetParent = widget.parent();
      
      while (true)
      {
         Frame frame;
         if (test.config().frameId == -1)
         {
            if (parentBrowse != null)
            {
               possibleFocus = parentBrowse.hasFocusable(test);
               break;
            }
            else if (test.parent() instanceof Menu || test.parent(MenuStructure.class) != null)
            {
               // widget attached to inner browse
               possibleFocus = test.parent().hasFocusable(test);
               break;
            }
            else if (test instanceof Frame)
            {
               frame = (Frame) test;
            }
            else
            {
               throw new IllegalStateException("Could not resolve frame from widget " + widget);
            }
         }
         else
         {
            frame = tc.getFrame(test.config().frameId);
         }
         
         if (frame.getContentPane().hasFocusable(test) ||
                  ((widgetParent instanceof AbstractContainer) &&
                           widgetParent.focusTraversable() && widgetParent.hasFocusable(test)))
         {
            possibleFocus = true;
            break;
         }
         
         if (!frame.isRootFrame())
         {
            test = frame;
            widget = frame;
         }
         else
         {
            break;
         }
      }
      
      if (!possibleFocus)
      {
         return false;
      }
   
      Container<?> parent = parentBrowse != null ? parentBrowse : widgetParent;

      // the list of widgets that will not receive focus;
      // this covers the case when widgets are enabled/made sensitive
      // in the LEAVE trigger; such widgets are really enabled or made
      // sensitive, they are just ignored in the tab order.
      List<Widget<?>> ignoreFocus = null;
      
      // create and process the LEAVE event first so that any trigger
      // defined for that event gets its chance to run
      Widget oldFocus = null;
      if (widget.isEnabled())
      {
         // build the list of focusable widgets currently not traversable (not enabled or insensitive)
         List<? extends Widget<?>> parentTabWidgets = parent.getTabItemList();
         
         List<Widget<?>> noTraverseWidgets = new ArrayList<>(parentTabWidgets.size());
         
         for (Widget<?> w : parentTabWidgets)
         {
            if (w.isFocusable() && (!w.focusTraversable() || (afterTab && !w.isTabStop())))
            {
               noTraverseWidgets.add(w);
            }
         }
         
         oldFocus = widget;
         ignoreFocus = new ArrayList<>(noTraverseWidgets.size());
         
         // if focusable non-traversable widget was made traversable 
         // (with SENSITIVE attribute or ENABLE statement) ignore it in the tab order
         for (Widget<?> w : noTraverseWidgets)
         {
            if (w.focusTraversable())
            {
               ignoreFocus.add(w);
            }
         }

         // set the widgets that must not receive focus, see above
         parent.setIgnoreFocus(ignoreFocus);
      }
      
      // get the top-level window now, the focused widget may be destroyed after
      // the call to parent.nextFocus() or parent.prevFocus()
      TopLevelWindow widgetWindow = widget.topLevelWindow();
      
      tempFocusChange = !tc.isModalBoxActive();
      try
      {
         // change the focus temporarily now
         if (direction)
         {
            parent.nextFocus();
         }
         else
         {
            parent.prevFocus();
         }
      }
      finally
      {
         tempFocusChange = false;
         // clear the widgets that must not receive focus
         parent.setIgnoreFocus(null);
      }
      
      if (tc.isModalBoxActive())
      {
         // when a modal clint-side only box is active, no event is sent to server, the focus
         // events are handled locally
         return true;
      }
      
      // give priority to the top-level window of the reference widget, this
      // is to support widgets that consist of multiple top-level windows
      // like a COMBO-BOX
      Widget inFocus = widgetWindow != null ? widgetWindow.currentFocus() :
                                              UiUtils.getCurrentFocus();
      
      if (inFocus == widget)
      {
         if (!widget.focusTraversable())
         {
            // if the current widget is left on focus and it is not focusable,
            // focus remains undefined. this is the case when focus LEAVE has
            // disabled current focus
            WindowManager.getDefaultWindow().setFocus(null);
            if (oldFocus != null)
            {
               KeyInput leave = new KeyInput(oldFocus, EventType.KEY_PRESSED, Keyboard.SE_LEAVE);
               tc.processProgressEvent(leave);
               
               // check to see if the trigger returned no-apply
               if (leave.isConsumed())
                  return false;
            }
         }
         return true;
      }
   
      // Text widgets grouping mode support - TAB/BACKTAB in this mode
      // switches groups, not individual FillIn widgets.
      if (inFocus instanceof FillIn)
      {
         FillIn fin = (FillIn) inFocus;
         
         fin = fin.getFirstInGroup();
         
         if (!fin.hasFocus())
            fin.requestFocus();
      }
      
      // restore the focus to the original widget
      setFocusOn(widget);
   
      // for browse editing fill-in, the oldFocus fill-in gets destroyed during temporary focus 
      // change by Browse.goDown.  As the browse uses the same fill-in ID for each row during 
      // editing, we need to check if the oldFocus is still attached to a parent; TC.getWidget
      // can not be used because it will return inFocus (which has the same config ID as oldFocus).
      if (oldFocus != null && oldFocus.parent() != null)
      {
         KeyInput leaveTargetEvt = null;
         // if oldFocus is a frame, need additional LEAVE event for widget inside the frame 
         if (oldFocus != leaveTarget)
         {
            leaveTargetEvt = new KeyInput(leaveTarget, inFocus, EventType.KEY_PRESSED,
                                          Keyboard.SE_LEAVE, false);
            tc.processProgressEvent(leaveTargetEvt);
         }
         KeyInput leave = new KeyInput(oldFocus, inFocus, 
               EventType.KEY_PRESSED, Keyboard.SE_LEAVE, false);
         tc.processProgressEvent(leave);
         
         // check to see if the trigger returned no-apply
         if (leave.isConsumed() && (leaveTargetEvt == null || leaveTargetEvt.isConsumed()))
         {
            return false;
         }
      }

      // create and process the ENTRY event now. the ENTRY event will be 
      // responsible for setting focus to the required widget, if it accepts
      // it
      KeyInput entry = new KeyInput(inFocus,
                                    oldFocus,
                                    EventType.KEY_PRESSED,
                                    Keyboard.SE_ENTRY,
                                    true);
      tc.processProgressEvent(entry);
   
      // check to see if the event was consumed (due to no-apply or widget 
      // lost focusability - was disabled or hidden)
      if (!entry.isConsumed())
      {
         // pass event to widget
         inFocus.processKeyEvent(entry);
         EventManager.postEvent(new FocusEvent(widget, inFocus, EventType.FOCUS_LOST));
         EventManager.postEvent(new FocusEvent(inFocus, widget, EventType.FOCUS_GAINED));
         return true;
      }
      
      // if a trigger returned no-apply, search and focus the next widget
      // in tab-order after it, without wrapping; this widget will NOT 
      // ignore no-apply
      if (tc.isTriggerResultNoApply())
      {
         Frame frame = (Frame) UiUtils.locateFrame(inFocus);
         do 
         {
            inFocus = frame.getNextEnabledWidget(inFocus, FocusManager.FOCUS_NOWRAP);
         }
         while (inFocus != null && !tc.sendEntry(inFocus));
   
         if (inFocus == null)
         {
            // in no widget found, initial widget has focus
            widget.requestFocus();
         }
      }
      // if the widget was disabled by the triggers, focus remains on 
      // initial focus
      else
      {
         // reverse the focus change
         widget.requestFocus();
      }
   
      return false;
   }
   
   /**
    * Get state of the {@link #noFocusAdjustment} flag.
    * 
    * @return  current value.
    */
   public boolean isNoFocusAdjustment()
   {
      return noFocusAdjustment;
   }

   /**
    * Force the initial focus to a specific widget when entering an
    * interactive/editing mode.  This will generate <code>FOCUS_GAINED</code>
    * but it will not generate a <code>LEAVE</code> or <code>FOCUS_LOST</code>
    * since it is the initial focus set.
    * <p>
    * This processing will cause a repaint so it must be bracketed by
    * invalidation processing in the caller.
    *
    * @param    widget
    *           The widget to receive focus (must not be <code>null</code>).
    */
   public static void setInitialFocus(Widget widget)
   {
      // unconditional focus change since it's the initial set and should
      // not generate a LEAVE event for the previous widget
   
      EventManager.postEvent(new FocusEvent(widget, EventType.FOCUS_GAINED));
      
      ThinClient tc = ThinClient.getInstance();

      Container parent = widget.parent();

      if (parent != null)
         parent.setFocus(widget);
      
      if (tc.getInputActive() > 0)
      {
         widget.setHighlighted(true);
      }
   
      widget.repaint();
      
      Frame frame = (Frame) UiUtils.locateFrame(widget);
      
      if (frame != null && tc.getActiveFrameInt() != frame)
         tc.setLastViewed(frame);
   }
}
