/*
** Module   : ImageHelper.java
** Abstract : image helper methods common for different widgets
**
** Copyright (c) 2015-2017, Golden Code Development Corporation.
**
** -#- -I- --Date-- ---------------------------------Description----------------------------------
** 001 EVL 20150122 Created initial version.
** 002 EVL 20150415 Adding methods to work with prepackaged button images btn-*-arrow.
** 003 IAS 20160224 LOAD-MOUSE-POINTER support
** 004 OM  20160405 Renamed getLocalImageName() to getLocalFileName() from ThinClient.
*/ 
/*
** This program is free software: you can redistribute it and/or modify
** it under the terms of the GNU Affero General Public License as
** published by the Free Software Foundation, either version 3 of the
** License, or (at your option) any later version.
**
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
** GNU Affero General Public License for more details.
**
** You may find a copy of the GNU Affero GPL version 3 at the following
** location: https://www.gnu.org/licenses/agpl-3.0.en.html
** 
** Additional terms under GNU Affero GPL version 3 section 7:
** 
**   Under Section 7 of the GNU Affero GPL version 3, the following additional
**   terms apply to the works covered under the License.  These additional terms
**   are non-permissive additional terms allowed under Section 7 of the GNU
**   Affero GPL version 3 and may not be removed by you.
** 
**   0. Attribution Requirement.
** 
**     You must preserve all legal notices or author attributions in the covered
**     work or Appropriate Legal Notices displayed by works containing the covered
**     work.  You may not remove from the covered work any author or developer
**     credit already included within the covered work.
** 
**   1. No License To Use Trademarks.
** 
**     This license does not grant any license or rights to use the trademarks
**     Golden Code, FWD, any Golden Code or FWD logo, or any other trademarks
**     of Golden Code Development Corporation. You are not authorized to use the
**     name Golden Code, FWD, or the names of any author or contributor, for
**     publicity purposes without written authorization.
** 
**   2. No Misrepresentation of Affiliation.
** 
**     You may not represent yourself as Golden Code Development Corporation or FWD.
** 
**     You may not represent yourself for publicity purposes as associated with
**     Golden Code Development Corporation, FWD, or any author or contributor to
**     the covered work, without written authorization.
** 
**   3. No Misrepresentation of Source or Origin.
** 
**     You may not represent the covered work as solely your work.  All modified
**     versions of the covered work must be marked in a reasonable way to make it
**     clear that the modified work is not originating from Golden Code Development
**     Corporation or FWD.  All modified versions must contain the notices of
**     attribution required in this license.
*/

package com.goldencode.p2j.ui.client;

import java.io.*;
import java.util.*;

import com.goldencode.p2j.ui.*;
import com.goldencode.p2j.ui.chui.ThinClient;
import com.goldencode.p2j.ui.client.gui.driver.*;

/**
 * Image helper methods to handle prending image loading.
 */
public class ImageHelper
{
   /** Map to store pending images for image widget. */
   private static final Map<Integer, Object> imgMap     = new HashMap<Integer, Object>();
   
   /** Map to store pending images for button widget up image. */
   private static final Map<Integer, Object> imgUpMap   = new HashMap<Integer, Object>();
   
   /** Map to store pending images for button widget down image. */
   private static final Map<Integer, Object> imgDownMap = new HashMap<Integer, Object>();
   
   /** Map to store pending images for button widget disabled image. */
   private static final Map<Integer, Object> imgInsMap  = new HashMap<Integer, Object>();
   
   /** Map pre-defined MousePointers to AWT/JS ones */
   private static final Map<MousePointer, CursorType> mousePointerTypes = 
         Collections.unmodifiableMap(
               new EnumMap(MousePointer.class)
               {{
                  put(MousePointer.ARROW,   CursorType.DEFAULT);
                  put(MousePointer.CROSS,   CursorType.CROSS);
                  put(MousePointer.IBEAM,   CursorType.TEXT);
                  put(MousePointer.SIZE,    CursorType.MOVE);
                  put(MousePointer.SIZE_E,  CursorType.E_RESIZE);
                  put(MousePointer.SIZE_N,  CursorType.N_RESIZE);
                  put(MousePointer.SIZE_NE, CursorType.NE_RESIZE);
                  put(MousePointer.SIZE_NW, CursorType.NW_RESIZE);
                  put(MousePointer.SIZE_S,  CursorType.S_RESIZE);
                  put(MousePointer.SIZE_SE, CursorType.SE_RESIZE);
                  put(MousePointer.SIZE_SW, CursorType.SW_RESIZE);
                  put(MousePointer.SIZE_W,  CursorType.W_RESIZE);
                  put(MousePointer.WAIT,    CursorType.WAIT);
                  put(MousePointer.GLOVE,   CursorType.HAND);
               }}
         );

   /** Map pre-defined MousePointers w/o AWT/JS counterparts to data files*/
   private static final Map<MousePointer, String> mousePointerFiles = 
         Collections.unmodifiableMap(
               new EnumMap(MousePointer.class)
               {{
                  // TODO: fill
//                  put(MousePointer.APPSTARTING, "");
//                  put(MousePointer.HELP, "");
//                  put(MousePointer.NO, "");
//                  put(MousePointer.RECTANGLE, "");
//                  put(MousePointer.UPARROW, "");
//                  put(MousePointer.COMPILER_WAIT, "");
               }}
         );
   
   /**
    * Saves the new image name for further loading at widget realization stage.
    * 
    * @param    wId
    *           ID of the widget will own the image.
    * @param    type
    *           Image type to load.  Up, Down or Insensitive for buttons or image.
    * @param    fileName
    *           The file name for the image to be loaded.
    *
    * @return   <code>TRUE</code> if operation successful <code>FALSE</code> otherwise.
    */
   public static boolean storePendingImageName(int wId, int type, String fileName)
   {
      if (fileName == null || fileName.length() == 0)
      {
         return false;
      }
      
      switch (type)
      {
         case ButtonConfig.IMAGE_IMG:
         {
            imgMap.put(wId, fileName);
            break;
         }
         case ButtonConfig.IMAGE_UP:
         {
            imgUpMap.put(wId, fileName);
            break;
         }
         case ButtonConfig.IMAGE_DOWN:
         {
            imgDownMap.put(wId, fileName);
            break;
         }
         case ButtonConfig.IMAGE_INS:
         {
            imgInsMap.put(wId, fileName);
            break;
         }
      }
      
      return true;
   }

   /**
    * Saves the new image data byte array for further loading at widget realization stage.
    * 
    * @param    wId
    *           ID of the widget will own the image.
    * @param    type
    *           Image type to load.  Up, Down or Insensitive for buttons or image.
    * @param    imageBytes
    *           The byte array data for the image to be loaded.
    *
    * @return   <code>TRUE</code> if operation successful <code>FALSE</code> otherwise.
    */
   public static boolean storePendingImageData(int wId, int type, byte[] imageBytes)
   {
      if (imageBytes == null)
      {
         return false;
      }
      
      switch (type)
      {
         case ButtonConfig.IMAGE_IMG:
         {
            imgMap.put(wId, imageBytes);
            break;
         }
         case ButtonConfig.IMAGE_UP:
         {
            imgUpMap.put(wId, imageBytes);
            break;
         }
         case ButtonConfig.IMAGE_DOWN:
         {
            imgDownMap.put(wId, imageBytes);
            break;
         }
         case ButtonConfig.IMAGE_INS:
         {
            imgInsMap.put(wId, imageBytes);
            break;
         }
      }
      
      return true;
   }
   
   /**
    * Gets the new image data byte array for loading at widget realization stage.
    * 
    * @param    wId
    *           ID of the widget will own the image.
    * @param    type
    *           Image type to load.  Up, Down or Insensitive for buttons or image.
    *
    * @return   The byte array data for the image to be loaded or null if there is no data.
    */
   public static byte[] getPendingImageData(int wId, int type)
   {
      byte[] imgData = null;
      Object imgObj = null;
      
      switch (type)
      {
         case ButtonConfig.IMAGE_IMG:
         {
            imgObj = imgMap.get(wId);
            // check if we have data array, not string
            if (imgObj != null && !(imgObj instanceof String))
            {
               imgData = (byte[])imgMap.remove(wId);
            }
            break;
         }
         case ButtonConfig.IMAGE_UP:
         {
            imgObj = imgUpMap.get(wId);
            // check if we have data array, not string
            if (imgObj != null && !(imgObj instanceof String))
            {
               imgData = (byte[])imgUpMap.remove(wId);
            }
            break;
         }
         case ButtonConfig.IMAGE_DOWN:
         {
            imgObj = imgDownMap.get(wId);
            // check if we have data array, not string
            if (imgObj != null && !(imgObj instanceof String))
            {
               imgData = (byte[])imgDownMap.remove(wId);
            }
            break;
         }
         case ButtonConfig.IMAGE_INS:
         {
            imgObj = imgInsMap.get(wId);
            // check if we have data array, not string
            if (imgObj != null && !(imgObj instanceof String))
            {
               imgData = (byte[])imgInsMap.remove(wId);
            }
            break;
         }
      }
      
      return imgData;
   }
   
   /**
    * Gets the new image file name for loading at widget realization stage.
    * 
    * @param    wId
    *           ID of the widget will own the image.
    * @param    type
    *           Image type to load.  Up, Down or Insensitive for buttons or image.
    *
    * @return   The file name for the image to be loaded and found in PROPATH directory.
    */
   public static String getPendingImageName(int wId, int type)
   {
      String imgName = null;
      Object imgObj = null;
      
      switch (type)
      {
         case ButtonConfig.IMAGE_IMG:
         {
            imgObj = imgMap.get(wId);
            // check if we have file name string
            if (imgObj != null && imgObj instanceof String)
            {
               imgName = (String)imgMap.remove(wId);
            }
            break;
         }
         case ButtonConfig.IMAGE_UP:
         {
            imgObj = imgUpMap.get(wId);
            // check if we have file name string
            if (imgObj != null && imgObj instanceof String)
            {
               imgName = (String)imgUpMap.remove(wId);
            }
            break;
         }
         case ButtonConfig.IMAGE_DOWN:
         {
            imgObj = imgDownMap.get(wId);
            // check if we have file name string
            if (imgObj != null && imgObj instanceof String)
            {
               imgName = (String)imgDownMap.remove(wId);
            }
            break;
         }
         case ButtonConfig.IMAGE_INS:
         {
            imgObj = imgInsMap.get(wId);
            // check if we have file name string
            if (imgObj != null && imgObj instanceof String)
            {
               imgName = (String)imgInsMap.remove(wId);
            }
            break;
         }
      }
      
      return ThinClient.getLocalFileName(imgName);
   }

   /**
    * Checks if the given image file name belongs to the set of the prepackaged images embedded
    * into project distribution.  Currently the embedded images includes: btn-up-arrow,
    * btn-down-arrow, btn-left-arrow and btn-right-arrow.
    * 
    * @param    fileName
    *           The file name for the image to be checked.
    *
    * @return   <code>TRUE</code> if given image is prepackaged <code>FALSE</code> otherwise.
    */
   public static boolean isPrepackaged(String fileName)
   {
      // simple check for input parameters
      if (fileName == null || fileName.isEmpty())
      {
         return false;
      }
      
      // make case insensitive
      String nameLCase = fileName.toLowerCase();
      
      return nameLCase.indexOf("btn-up-arrow") != -1 ||
             nameLCase.indexOf("btn-down-arrow") != -1 ||
             nameLCase.indexOf("btn-left-arrow") != -1 ||
             nameLCase.indexOf("btn-right-arrow") != -1;
   }
   
   /**
    * Gets the new image data byte array for prepackaged images.
    * 
    * @param    fileName
    *           The file name for the image to be loaded from p2j.jar.
    *
    * @return   The byte array data for the image to be loaded or null if there is no data.
    */
   public static byte[] getPrepackagedImageData(String fileName)
   {
      byte[] imgData = null;
      
      // loading image embedded into p2j.jar
      InputStream is = ImageHelper.class.getResourceAsStream(fileName.toLowerCase());
         
      if (is != null)
      {
         imgData = UiUtils.getData(is);
      }
         
      return imgData;
   }
   
   /**
    * Get AWT/JS cursor type for the mouse pointer
    * 
    * @param ptr
    *        mouse pointer
    * @return AWT/JS cursor type
    */
   public static CursorType cursorType(MousePointer ptr)
   {
      return mousePointerTypes.get(ptr);
   }

   /**
    * Get data file name for the mouse pointer
    * 
    * @param ptr
    *        mouse pointer
    *        
    * @return data file name
    */
   public static String cursorFile(MousePointer ptr)
   {
      return mousePointerFiles.get(ptr);
   }
}
