/*
** Module   : LabeledDataContainer.java
** Abstract : P2JComponent with optional side or top label.
**
** Copyright (c) 2010-2017, Golden Code Development Corporation.
**
** -#- -I- --Date-- ---------------------------------Description----------------------------------
** 001 SIY 20100820 Refactored common code from Editor and FillIn.
** 002 SIY 20100915 Renamed class and base class from LabeledP2JComponent.
** 003 SIY 20111101 Moved from CHUI package. Minor cleanups/refactorings in preparation to GUI 
**                  client.
** 004 SIY 20111202 Updated imports.
** 005 EVL 20140528 Fix for displaying the labels for hidden widgets in a frame.  We need to keep 
**                  track when the widget become visible to differentiate the hidden widgets 
**                  rendering.
** 006 CA  20140805 Added generics to force the abstract widget classes to not use implementation-
**                  specific code: the abstract widgets will have access only to the base APIs
**                  provided by the OutputManager.
** 007 HC  20140911 GUI coordinate support - data types refactoring.
** 008 CA  20140926 Added shared widget configuration support, to allow GUI/ChUI concrete
**                  implementation for the same widget ID.  Refactored the widget configuration
**                  classes: all fields were made public; these classes need to be as dumb as
**                  possible, all logic related to setting/getting a certain field should be at
**                  the widget or at the caller. Refs #2254
** 009 CA  20150105 Changes for SIDE-LABEL-HANDLE runtime support.
** 010 CA  20150114 Fixed support for changing the ROW/COLUMN attrs from server-side, for a 
**                  side-label.
** 011 HC  20150128 Added support for frame widget position changes induced by setting
**                  ROW/COL/X/Y attributes.
** 012 CA  20150202 Side-labels needs to track the HIDDEN state, too.
** 013 HC  20150214 Improved positioning of side labels.
** 014 CA  20150223 Allow the label to be removed.
** 015 HC  20150323 Removed business logic from config classes.
** 025 CA  20150722 Added postprocessConfig - required additional processing, as layout is 
**                  performed only on the first frame definition push.
** 026 HC  20150913 Fixed a memory leak caused by label config not being unregistered.
** 027 IAS 20151214 LOAD-MOUSE-POINTER support.
** 028 HC  20161018 Added an initial support for rendering of mnemonic labels to all labeled
**                  widgets.
** 029 CA  20171020 Added support for FRAMEs sent to named or unnamed streams, in GUI.
** 030 CA  20171023 In GUI, streamed frames don't affect any state of its UI counterpart.
*/
/*
** This program is free software: you can redistribute it and/or modify
** it under the terms of the GNU Affero General Public License as
** published by the Free Software Foundation, either version 3 of the
** License, or (at your option) any later version.
**
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
** GNU Affero General Public License for more details.
**
** You may find a copy of the GNU Affero GPL version 3 at the following
** location: https://www.gnu.org/licenses/agpl-3.0.en.html
** 
** Additional terms under GNU Affero GPL version 3 section 7:
** 
**   Under Section 7 of the GNU Affero GPL version 3, the following additional
**   terms apply to the works covered under the License.  These additional terms
**   are non-permissive additional terms allowed under Section 7 of the GNU
**   Affero GPL version 3 and may not be removed by you.
** 
**   0. Attribution Requirement.
** 
**     You must preserve all legal notices or author attributions in the covered
**     work or Appropriate Legal Notices displayed by works containing the covered
**     work.  You may not remove from the covered work any author or developer
**     credit already included within the covered work.
** 
**   1. No License To Use Trademarks.
** 
**     This license does not grant any license or rights to use the trademarks
**     Golden Code, FWD, any Golden Code or FWD logo, or any other trademarks
**     of Golden Code Development Corporation. You are not authorized to use the
**     name Golden Code, FWD, or the names of any author or contributor, for
**     publicity purposes without written authorization.
** 
**   2. No Misrepresentation of Affiliation.
** 
**     You may not represent yourself as Golden Code Development Corporation or FWD.
** 
**     You may not represent yourself for publicity purposes as associated with
**     Golden Code Development Corporation, FWD, or any author or contributor to
**     the covered work, without written authorization.
** 
**   3. No Misrepresentation of Source or Origin.
** 
**     You may not represent the covered work as solely your work.  All modified
**     versions of the covered work must be marked in a reasonable way to make it
**     clear that the modified work is not originating from Golden Code Development
**     Corporation or FWD.  All modified versions must contain the notices of
**     attribution required in this license.
*/

package com.goldencode.p2j.ui.client;

import java.util.*;

import com.goldencode.p2j.ui.*;
import com.goldencode.p2j.ui.chui.ThinClient;
import com.goldencode.p2j.ui.client.widget.*;

/**
 * Common implementation of {@link LabeledWidget} interface for 
 * {@link FillIn} and {@link Editor}.
 * 
 * @author  SIY
 * @version 1.0
 */
public abstract class LabeledDataContainer<O extends OutputManager<?>, 
                                           C extends ControlTextConfig>
extends FixedSizeContainer<O>
implements WidgetWithConfig<C>,
           LabeledWidget<O>,
           DataContainer,
           MnemonicTrait<O>
{
   /** Component label. */
   private Label<O> sideLabel = null;
   
   /** Flag indicating the widget was made visible at least once. */
   private boolean wasVisible = false;
   
   /** Component attribute container */
   protected C config;

   /** Mnemonic information */
   private MnemonicInfo mnemonic = null;

   /** Column mnemonic information */
   private MnemonicInfo columnMnemonic = null;

   /**
    * Provide access to the attributes stored in the widget.
    * 
    * @return  Reference to attribute container.
    */
   @SuppressWarnings("unchecked")
   @Override
   public C config()
   {
      return config;
   }
   
   /**
    * The method is called after the configuration associated with 
    * the implementor changes. The parameter points to a reference
    * representing the original configuration state. This reference
    * can be used to detect what configuration fields changed and so
    * optimize any processing related to the configuration change.
    * 
    * @param    beforeUpdate
    */
   @Override
   public void afterConfigUpdate(C beforeUpdate)
   {
      super.afterConfigUpdateBase(beforeUpdate);

      updateMnemonic(beforeUpdate.clabel, config.clabel, true);
      updateMnemonic(beforeUpdate.label, config.label, false);
   }

   /**
    * Return current value of COLON attribute.
    * 
    * @return  Current value of the attribute.
    */
   @Override
   public double getColon()
   {
      return ConfigHelper.getColon(config);
   }

   /**
    * Return text of the column label.
    * 
    * @return  text of the column label.
    */
   @Override
   public String getColumnLabelText()
   {
      return getMnemonicLabel(true).orElse(config().clabel);
   }

   /**
    * Return reference to the instance of the <code>Label</code> attached to
    * the widget.
    * 
    * @return  Reference to <code>Label</code> instance.
    */
   @Override
   public Label<O> getLabelInstance()
   {
      return sideLabel;
   }

   /**
    * Return text of the label.
    * 
    * @return  text of the label.
    */
   @Override
   public String getLabelText()
   {
      return getMnemonicLabel(false).orElse(config.label);
   }

   /**
    * Return current value of TO attribute.
    * 
    * @return  Current value of the attribute.
    */
   @Override
   public double getTo()
   {
      return ConfigHelper.getTo(config());
   }

   /**
    * Get current state of flag which is forces presence of the label
    * regardless from the frame NO-LABELS option.
    * 
    * @return  Current state of the flag.
    */
   @Override
   public boolean isForceLabel()
   {
      return config().forceLabel;
   }
   
   /**
    * Get current state of NO-LABELS flag.
    * 
    * @return  Current state of the flag.
    */
   @Override
   public boolean isNoLabels()
   {
      return !config().labels;
   }

   /**
    * Get widget alignment policy.
    * 
    * @return  <code>true</code> if widget is formatted for right alignment.
    */
   @Override
   public boolean isRightAligned()
   {
      return false;
   }

   /**
    * Assign new instance of <code>Label</code> to the fill-in widget.
    * 
    * @param   label
    *          Reference to new <code>Label</code> instance.
    */
   @Override
   public void setLabelInstance(Label<O> label)
   {
      if (sideLabel != label && sideLabel != null)
      {
         sideLabel.destroy();
      }
      
      if (label == null)
      {
         this.sideLabel = null;
         return;
      }

      int frameId = ThinClient.frameIdFromWidgetId(this.config().id.asInt());
      
      // link the label to the frame of this widget
      label.linkTo(frameId, config().sideLabelId);

      sideLabel = label;
      if (sideLabel.isVisible() != isVisible())
         sideLabel.setVisible(isVisible());
      if (sideLabel.hidden() != hidden())
         sideLabel.setHidden(hidden());
   }
   
   /**
    * Force NO-LABELS flag.
    *
    * @param   labels
    *          New value of the flag.
    */
   @Override
   public void setNoLabels(boolean labels)
   {
      config().labels = !labels;
   }

   /**
    * Shows or hides the widget and attached label.
    *
    * @param    visible
    *           <code>true</code> to make the widget visible, 
    *           <code>false</code> to hide the widget.
    */
   @Override
   public void setVisible(boolean visible)
   {
      super.setVisible(visible);

      if (sideLabel != null)
         sideLabel.setVisible(visible);
   }

   /**
    * Tracks the hidden state of the widget and attached label.
    * 
    * @param    hidden
    *           New hidden state.
    */
   @Override
   public void setHidden(boolean hidden)
   {
      super.setHidden(hidden);
      
      if (sideLabel != null)
         sideLabel.setHidden(hidden);
   }

   /**
    * Getting the modification state of the widget.
    * 
    * @return  <code>true</code> if the widget value was modified,
    *          <code>false</code> otherwise.
    */
   @Override
   public boolean isModified()
   {
      return config().modified;
   }
   
   /**
    * Checking if the widget was made visible at least once.
    * 
    * @return  <code>true</code> if the widget was visible,
    *          <code>false</code> otherwise.
    */
   @Override
   public boolean wasVisible()
   {
      return wasVisible;
   }
   
   /**
    * Remember the widget was made visible at least once.
    */
   @Override
   public void setWasVisible()
   {
      wasVisible = true;
   }
   
   /**
    * Additional processing required when widget configuration is updated AFTER the frame layout
    * has been performed.
    */
   @Override
   public void postprocessConfig()
   {
      super.postprocessConfig();
      
      if (sideLabel != null)
      {
         sideLabel.postprocessConfig(this);
      }
   }

   /**
    * Destroy this widget.
    */
   @Override
   public void destroy()
   {
      if (sideLabel != null)
      {
         sideLabel.destroy();
         sideLabel = null;
      }
      
      super.destroy();
   }
   /**
    * Get default mouse pointer for this widget
    * 
    * @return
    *        MousePtr or <code>Optional.empty()</code>
    *        if one is inherited from the enclosing frame
    *    
    */
   public Optional<MousePtrWrapper> defaultMousePtr()
   {
      return MousePtrWrapper.TEXT_PTR;
   }

   /**
    * MnemonicInfo instance getter.
    * <p>
    * The implementors should simply return the value of the mnemonic class field. No
    * additional processing logic is advised.
    *
    * @return   See above.
    */
   public MnemonicInfo getMnemonic()
   {
      return mnemonic;
   }

   /**
    * MnemonicInfo instance setter.
    * <p>
    * The implementors should simply assign the parameter value to the mnemonic class
    * field. No additional processing logic is advised.
    */
   public void setMnemonic(MnemonicInfo mnemonic)
   {
      this.mnemonic = mnemonic;
   }

   /**
    * Column {@link MnemonicInfo} instance getter.
    * <p>
    * The implementors should simply return the value of the mnemonic class field. No
    * additional processing logic in the implemented method is advised.
    * <p>
    * The method will return a valid value when the implementing widget supports column label.
    *
    * @return   See above.
    */

   public MnemonicInfo getColumnMnemonic()
   {
      return columnMnemonic;
   }

   /**
    * Column {@link MnemonicInfo} instance setter.
    * <p>
    * The implementors should simply assign the parameter value to the mnemonic class
    * field. No additional processing logic in the implemented method is advised.
    */
   public void setColumnMnemonic(MnemonicInfo mnemonic)
   {
      this.columnMnemonic = mnemonic;
   }
   
   /**
    * Copy the widget state from the specified source widget to this instance.
    * 
    * @param    w
    *           The source widget which state needs to be copied.
    */
   @Override
   protected void copyState(Widget<?> w)
   {
      LabeledDataContainer<?, ?> widget = (LabeledDataContainer<?, ?>) w;

      try
      {
         config = (C) widget.config.getClass().newInstance();
         config.applyConfig(widget.config);
         config.id = widget.config.id;
      }
      catch (InstantiationException | IllegalAccessException e)
      {
         throw new RuntimeException("Error while copying widget state", e);
      }
      
      super.copyState(w);

      wasVisible = widget.wasVisible;
      mnemonic = widget.mnemonic;
      columnMnemonic = widget.columnMnemonic;
      
      // sideLabel will be attached by the layout
   }
}