/*
** Module   : LabeledWidget.java
** Abstract : Common interface for widgets which can have label (top or side).
**
** Copyright (c) 2006-2017, Golden Code Development Corporation.
**
** -#- -I- --Date-- --JPRM-- ----------------------------Description------------------------------
** 001 SIY 20060209  @24456  Created initial version
** 002 SIY 20060322  @25224  Added setNoLabels() method.
** 003 SIY 20060515  @26154  Added getColumnLabelText().
** 004 SIY 20070807  @34816  Added isColumnFixed() method.
** 005 SIY 20111101          Moved from CHUI package. Minor cleanups/refactorings in preparation 
**                           to GUI client.
** 006 SIY 20111202          Updated imports.
** 007 EVL 20140528          Fix for displaying the labels for hidden widgets in a frame.  We
**                           need to keep track when the widget become visible to differentiate
**                           the hidden widgets rendering.
** 008 CA  20140805          Added generics to force the abstract widget classes to not use
**                           implementation-specific code: the abstract widgets will have access
**                           only to the base APIs provided by the OutputManager.
** 009 HC  20140911          GUI coordinate support - data types refactoring.
** 010 HC  20150128          Added support for frame widget position changes induced by
**                           setting ROW/COL/X/Y attributes.
** 011 HC  20150323          Removed business logic from config classes.
** 012 HC  20161018          Added an initial support for rendering of mnemonic labels to all
**                           labeled widgets.
*/
/*
** This program is free software: you can redistribute it and/or modify
** it under the terms of the GNU Affero General Public License as
** published by the Free Software Foundation, either version 3 of the
** License, or (at your option) any later version.
**
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
** GNU Affero General Public License for more details.
**
** You may find a copy of the GNU Affero GPL version 3 at the following
** location: https://www.gnu.org/licenses/agpl-3.0.en.html
** 
** Additional terms under GNU Affero GPL version 3 section 7:
** 
**   Under Section 7 of the GNU Affero GPL version 3, the following additional
**   terms apply to the works covered under the License.  These additional terms
**   are non-permissive additional terms allowed under Section 7 of the GNU
**   Affero GPL version 3 and may not be removed by you.
** 
**   0. Attribution Requirement.
** 
**     You must preserve all legal notices or author attributions in the covered
**     work or Appropriate Legal Notices displayed by works containing the covered
**     work.  You may not remove from the covered work any author or developer
**     credit already included within the covered work.
** 
**   1. No License To Use Trademarks.
** 
**     This license does not grant any license or rights to use the trademarks
**     Golden Code, FWD, any Golden Code or FWD logo, or any other trademarks
**     of Golden Code Development Corporation. You are not authorized to use the
**     name Golden Code, FWD, or the names of any author or contributor, for
**     publicity purposes without written authorization.
** 
**   2. No Misrepresentation of Affiliation.
** 
**     You may not represent yourself as Golden Code Development Corporation or FWD.
** 
**     You may not represent yourself for publicity purposes as associated with
**     Golden Code Development Corporation, FWD, or any author or contributor to
**     the covered work, without written authorization.
** 
**   3. No Misrepresentation of Source or Origin.
** 
**     You may not represent the covered work as solely your work.  All modified
**     versions of the covered work must be marked in a reasonable way to make it
**     clear that the modified work is not originating from Golden Code Development
**     Corporation or FWD.  All modified versions must contain the notices of
**     attribution required in this license.
*/

package com.goldencode.p2j.ui.client;

import com.goldencode.p2j.ui.client.widget.*;

/**
 * Interface which describes widgets which can have top or side labels. 
 * Actual labels are assigned by the active layout manager according to
 * common frame and individual widget settings. 
 * 
 * @author  SIY
 * @version 1.0
 */
public interface  LabeledWidget<O extends OutputManager<?>>
extends Widget<O>
{
   /**
    * Return reference to the instance of the <code>Label</code> attached to
    * the widget.
    * 
    * @return  Reference to <code>Label</code> instance.
    */
   public Label<O> getLabelInstance();
   
   /**
    * Assign new instance of <code>Label</code> to the widget.
    * 
    * @param   label
    *          Reference to new <code>Label</code> instance.
    */
   public void setLabelInstance(Label<O> label);
   
   /**
    * Return text of the label.
    * 
    * @return  text of the label.
    */
   public String getLabelText();

   /**
    * Return text of the column label.
    * 
    * @return  text of the column label.
    */
   public String getColumnLabelText();
   
   /**
    * Return TO column.
    * 
    * @return  Column assigned via TO option.
    */
   public double getTo();

   /**
    * Get current state of flag which is forces presence of the label
    * regardless from the frame NO-LABELS option.
    * 
    * @return  Current state of the flag.
    */
   public boolean isForceLabel();

   /**
    * Get current state of NO-LABELS flag.
    * 
    * @return  Current state of the flag.
    */
   public boolean isNoLabels();

   /**
    * Return current value of COLON option.
    * 
    * @return  Current value of the option column.
    */
   public double getColon();

   /**
    * Get widget alignment policy.
    * 
    * @return  <code>true</code> if widget is formatted for right alignment.
    */
   public boolean isRightAligned();

   /**
    * Force NO-LABELS flag.
    * 
    * @param   labels
    *          New value of the flag.
    */
   public void setNoLabels(boolean labels);
   
   /**
    * Checking if the widget was made visible at least once.
    * 
    * @return  <code>true</code> if the widget was visible,
    *          <code>false</code> otherwise.
    */
   public boolean wasVisible();
   
   /**
    * Remember the widget was made visible at least once.
    */
   public void setWasVisible();
}