/*
** Module   : NativeRectangle.java
** Abstract : Rectangular area descriptor.
**
** Copyright (c) 2014-2017, Golden Code Development Corporation.
**
** -#- -I- --Date-- ---------------------------------Description----------------------------------
** 001 HC  20140911 Created initial version.
** 002 HC  20150508 Improvements to GUI window scrolling support.
** 003 HC  20150913 Added toString().  
** 004 SBI 20171208 Changed intersection(.), intersects(.) and union(.) methods to consider the
**                  case if this object or its parameter are represented by empty rectangles.
*/
/*
** This program is free software: you can redistribute it and/or modify
** it under the terms of the GNU Affero General Public License as
** published by the Free Software Foundation, either version 3 of the
** License, or (at your option) any later version.
**
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
** GNU Affero General Public License for more details.
**
** You may find a copy of the GNU Affero GPL version 3 at the following
** location: https://www.gnu.org/licenses/agpl-3.0.en.html
** 
** Additional terms under GNU Affero GPL version 3 section 7:
** 
**   Under Section 7 of the GNU Affero GPL version 3, the following additional
**   terms apply to the works covered under the License.  These additional terms
**   are non-permissive additional terms allowed under Section 7 of the GNU
**   Affero GPL version 3 and may not be removed by you.
** 
**   0. Attribution Requirement.
** 
**     You must preserve all legal notices or author attributions in the covered
**     work or Appropriate Legal Notices displayed by works containing the covered
**     work.  You may not remove from the covered work any author or developer
**     credit already included within the covered work.
** 
**   1. No License To Use Trademarks.
** 
**     This license does not grant any license or rights to use the trademarks
**     Golden Code, FWD, any Golden Code or FWD logo, or any other trademarks
**     of Golden Code Development Corporation. You are not authorized to use the
**     name Golden Code, FWD, or the names of any author or contributor, for
**     publicity purposes without written authorization.
** 
**   2. No Misrepresentation of Affiliation.
** 
**     You may not represent yourself as Golden Code Development Corporation or FWD.
** 
**     You may not represent yourself for publicity purposes as associated with
**     Golden Code Development Corporation, FWD, or any author or contributor to
**     the covered work, without written authorization.
** 
**   3. No Misrepresentation of Source or Origin.
** 
**     You may not represent the covered work as solely your work.  All modified
**     versions of the covered work must be marked in a reasonable way to make it
**     clear that the modified work is not originating from Golden Code Development
**     Corporation or FWD.  All modified versions must contain the notices of
**     attribution required in this license.
*/

package com.goldencode.p2j.ui.client;

import java.util.*;

/**
 * Rectangular area descriptor.
 * All coordinate values are in units native to UI type. 
 */
public class NativeRectangle
{
   /** Rectangle bottom row. */
   private final int bottom;

   /** Rectangle left column. */
   private final int left;

   /** Rectangle right column. */
   private final int right;

   /** Rectangle top row. */
   private final int top;

   /**
    * Constructor.
    * 
    * @param   top
    *          Top row.
    * @param   left
    *          Left column.
    * @param   bottom
    *          Bottom row.
    * @param   right
    *          Right column.
    */
   public NativeRectangle(int top, int left, int bottom, int right)
   {
      this.top = top;
      this.left = left;
      this.bottom = bottom;
      this.right = right;
   }

   /**
    * Constructor.
    * 
    * @param   topleft
    *          Top-left corner.
    * @param   size
    *          Rectangle dimension.
    */
   public NativeRectangle(NativePoint topleft, NativeDimension size)
   {
      top = topleft.y;
      left = topleft.x;
      bottom = top + size.height - 1;
      right = left + size.width - 1;
   }

   /**
    * Copy constructor.
    * 
    * @param   rect
    *          Source rectangle.
    */
   public NativeRectangle(NativeRectangle rect)
   {
      top = rect.top();
      left = rect.left();
      bottom = rect.bottom();
      right = rect.right();
   }

   /**
    * Get bottom coordinate of this rectangle.
    * 
    * @return  bottom.
    */
   public int bottom()
   {
      return bottom;
   }

   /**
    * Returns <code>true</code> if the specified point is inside this
    * rectangle.
    * 
    * @param   x
    *          X coordinate.
    * @param   y
    *          Y coordinate.
    * 
    * @return <code>true</code> if coordinates are inside rectangle.
    */
   public boolean contains(int x, int y)
   {
      return ((x >= left) && (x <= right) && (y >= top) && (y <= bottom));
   }

   /**
    * Returns <code>true</code> if the specified point is inside this
    * rectangle.
    * 
    * @param   p
    *          Point to check.
    *          
    * @return  <code>true</code> if specified point is inside rectangle.
    */
   public boolean contains(NativePoint p)
   {
      return contains(p.x, p.y);
   }

   /**
    * Returns <code>true</code> is specified rectangle is completely inside
    * this rectangle.
    * 
    * @param   r
    *          Rectangle to check.
    *          
    * @return  <code>true</code> if specified rectangle is inside this
    *          rectangle.
    */
   public boolean contains(NativeRectangle r)
   {
      return ((r.left >= left) && (r.right <= right) && (r.top >= top) && (r.bottom <= bottom));
   }

   /**
    * Check if rectangle is empty.
    * 
    * @return  <code>true</code> if rectangle is empty.
    */
   public boolean empty()
   {
      return top > bottom || right < left;
   }

   /**
    * Check if this rectangle is equal to specified one.
    * 
    * @param   rect
    *          Rectangle to check.
    * 
    * @return  <code>true</code> if rectangles are equal.
    */
   public boolean equals(NativeRectangle rect)
   {
      if (rect == null)
         return false;
      if (top != rect.top)
         return false;
      if (bottom != rect.bottom)
         return false;
      if (left != rect.left)
         return false;
      if (right != rect.right)
         return false;
      return true;
   }
   
   /**
    * Compute the hashcode of this rectangle.
    * 
    * @return   A hashcode uniquely describing this rectangle.
    */
   @Override
   public int hashCode()
   {
      int r = 17;
      r = 37 * r + top;
      r = 37 * r + bottom;
      r = 37 * r + left;
      r = 37 * r + right;
      
      return r;
   }

   /**
    * Calculate rectangle which is intersection between this and specified
    * rectangle.
    * 
    * @param   rect
    *          Source rectangle.
    *          
    * @return  rectangle which is an intersection between this and specified
    *          rectangle or <code>null</code> if there is no intersection.
    */
   public NativeRectangle intersection(NativeRectangle rect)
   {
      if (empty() || rect.empty())
      {
         return null;
      }
      if (top > rect.bottom)
         return null;
      if (bottom < rect.top)
         return null;
      if (left > rect.right)
         return null;
      if (right < rect.left)
         return null;

      return new NativeRectangle((top > rect.top) ? top
                                           : rect.top,
                           (left > rect.left) ? left
                                             : rect.left,
                           (bottom < rect.bottom) ? bottom
                                                 : rect.bottom,
                           (right < rect.right) ? right
                                               : rect.right);
   }

   /**
    * Check if this and specified rectangle have intersection.
    * 
    * @param   rect
    *          Source rectangle.
    * 
    * @return  <code>true</code> if rectangles have intersection.
    */
   public boolean intersects(NativeRectangle rect)
   {
      if (empty() || rect.empty())
      {
         return false;
      }
      
      if (rect.left > right)
         return false;
      if (rect.right < left)
         return false;
      if (rect.top > bottom)
         return false;
      if (rect.bottom < top)
         return false;
      
      return true;
   }

   /**
    * Get left coordinate of this rectangle.
    * 
    * @return  left.
    */
   public int left()
   {
      return left;
   }

   /**
    * Get right coordinate of this rectangle.
    * 
    * @return  right.
    */
   public int right()
   {
      return right;
   }

   /**
    * Get top coordinate of this rectangle.
    * 
    * @return top.
    */
   public int top()
   {
      return top;
   }

   /**
    * Create new rectangle which is translated by specified offset.
    * 
    * @param   dx
    *          Delta X.
    * @param   dy
    *          Delta Y.
    * 
    * @return  translated rectangle.
    */
   public NativeRectangle translate(int dx, int dy)
   {
      return new NativeRectangle(top + dy, left + dx, bottom + dy, right + dx);
   }

   /**
    * Create new rectangle which is translated by specified offset.
    * 
    * @param   p
    *          Offset represented as a point.
    *           
    * @return  translated rectangle.
    */
   public NativeRectangle translate(NativePoint p)
   {
      return translate(p.x, p.y);
   }

   /**
    * Create new rectangle which is a union of this and specified rectangle.
    * 
    * @param   rect
    *          Source rectangle.
    * 
    * @return  union rectangle.
    */
   public NativeRectangle union(NativeRectangle rect)
   {
      if (rect.empty())
      {
         return new NativeRectangle(this);
      }
      
      return new NativeRectangle((top < rect.top) ? top
                                           : rect.top,
                           (left < rect.left) ? left
                                             : rect.left,
                           (bottom > rect.bottom) ? bottom
                                                 : rect.bottom,
                           (right > rect.right) ? right
                                               : rect.right);
   }

   /**
    * Create new rectangles which is the difference between this and the specified rectangle.
    * 
    * @param   rect
    *          Removed rectangle.
    * @param   baseUnit
    *          The base measurement unit.
    *
    * @return  All the rectangles which are part of this rectangle but not part of the specified 
    *          rectangle.
    */
   public List<NativeRectangle> difference(NativeRectangle rect, int baseUnit)
   {
      List<NativeRectangle> res = new ArrayList<>(4);

      NativeRectangle intersection = this.intersection(rect);
      if (intersection == null)
      {
         // nothing to remove, return this
         res.add(this);

         return res;
      }

      // at this point, is just a matter of removing the intersection
      rect = intersection;

      if (rect.top - baseUnit >= top)
      {
         // rectangle found above the intersection
         res.add(new NativeRectangle(top, left, rect.top - baseUnit, right));
      }
      
      if (rect.bottom + baseUnit <= bottom)
      {
         // rectangle found bellow the intersection
         res.add(new NativeRectangle(rect.bottom + baseUnit, left, bottom, right));
      }
      
      if (rect.left - baseUnit >= left)
      {
         // rectangle found on the left-side of the intersection
         res.add(new NativeRectangle(rect.top, left, rect.bottom, rect.left - baseUnit));
      }
      
      if (rect.right + baseUnit <= right)
      {
         // rectangle found on the right-side of the intersection
         res.add(new NativeRectangle(rect.top, rect.right + baseUnit, rect.bottom, right));
      }

      return res;
   }

   /**
    * Compute the area of this rectangle.
    *      
    * @return   The area of this rectangle.
    */
   public int area()
   {
      return (int) (bottom - top + 1) * (int) (right - left + 1);
   }
   
   /**
    * Returns the rectangle's width.
    * 
    * @return   See above.
    */
   public int width()
   {
      return right - left + 1;
   }
   
   /**
    * Returns the rectangle's height.
    * 
    * @return   See above.
    */
   public int height()
   {
      return bottom - top + 1;
   }

   /**
    * Returns a string representation of the object.
    * 
    * @return  See above.
    */
   @Override
   public String toString()
   {
      return "NativeRectangle[top=" + top + ", left=" + left + ", bottom=" + bottom + ", right="
               + right + "]";
   }
}