/*
** Module   : PageElement.java
** Abstract : implements page-top/page-bottom frame output for reports
**
** Copyright (c) 2006-2017, Golden Code Development Corporation.
**
** -#- -I- --Date-- --JPRM-- ----------------------------Description------------------------------
** 001 GES 20060226   @24747 Implements page-top/page-bottom frame output for reports by providing  
**                           an output callback mechanism which allows the output from a given 
**                           frame to be generated at the top or bottom of a page in a paged 
**                           output stream.
** 002 NVS 20061102   @30904 render() method asserts the redirected mode.
** 003 SIY 20070125   @31969 Moved part of code to ThinClient to better reflect responsibilities 
**                           and concentrate checks in one class.
** 004 NVS 20070607   @33981 Added hashCode() and equals() methods.
**                           Equality is based on the equality of frameId fields.
** 005 CA  20080107   @36698 If the frame is boxed and with page-bottom, ignore the bottom border.
** 006 SIY 20100619          Switching to own implementation of character UI.
** 007 SIY 20100831          Changes related to refactoring of FrameHelper.
** 008 SIY 20111101          Moved from CHUI package. Minor cleanups/refactorings in preparation 
**                           to GUI client.
** 009 CA  20140805          Isolated ThinClient import from other ChUI imports.  Fixed unchecked
**                           warning.
** 010 HC  20140911          GUI coordinate support - data types refactoring.
** 011 CA  20140926          Added shared widget configuration support, to allow GUI/ChUI
**                           concrete implementation for the same widget ID.  Refactored the
**                           widget configuration classes: all fields were made public; these
**                           classes need to be as dumb as possible, all logic related to
**                           setting/getting a certain field should be at the widget or at the
**                           caller. Refs #2254
*/
/*
** This program is free software: you can redistribute it and/or modify
** it under the terms of the GNU Affero General Public License as
** published by the Free Software Foundation, either version 3 of the
** License, or (at your option) any later version.
**
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
** GNU Affero General Public License for more details.
**
** You may find a copy of the GNU Affero GPL version 3 at the following
** location: https://www.gnu.org/licenses/agpl-3.0.en.html
** 
** Additional terms under GNU Affero GPL version 3 section 7:
** 
**   Under Section 7 of the GNU Affero GPL version 3, the following additional
**   terms apply to the works covered under the License.  These additional terms
**   are non-permissive additional terms allowed under Section 7 of the GNU
**   Affero GPL version 3 and may not be removed by you.
** 
**   0. Attribution Requirement.
** 
**     You must preserve all legal notices or author attributions in the covered
**     work or Appropriate Legal Notices displayed by works containing the covered
**     work.  You may not remove from the covered work any author or developer
**     credit already included within the covered work.
** 
**   1. No License To Use Trademarks.
** 
**     This license does not grant any license or rights to use the trademarks
**     Golden Code, FWD, any Golden Code or FWD logo, or any other trademarks
**     of Golden Code Development Corporation. You are not authorized to use the
**     name Golden Code, FWD, or the names of any author or contributor, for
**     publicity purposes without written authorization.
** 
**   2. No Misrepresentation of Affiliation.
** 
**     You may not represent yourself as Golden Code Development Corporation or FWD.
** 
**     You may not represent yourself for publicity purposes as associated with
**     Golden Code Development Corporation, FWD, or any author or contributor to
**     the covered work, without written authorization.
** 
**   3. No Misrepresentation of Source or Origin.
** 
**     You may not represent the covered work as solely your work.  All modified
**     versions of the covered work must be marked in a reasonable way to make it
**     clear that the modified work is not originating from Golden Code Development
**     Corporation or FWD.  All modified versions must contain the notices of
**     attribution required in this license.
*/

package com.goldencode.p2j.ui.client;

import com.goldencode.p2j.ui.*;
import com.goldencode.p2j.ui.chui.ThinClient;
import com.goldencode.p2j.util.*;

/**
 * Implements page-top/page-bottom frame output for reports by providing an
 * output callback mechanism which allows the output from a given frame to be
 * generated at the top or bottom of a page in a paged output stream.  This 
 * enables page headers and footers.
 * <p>
 * Please note that the only requirement for this output is that the number
 * of lines which it consumes MUST be constant across all invocations of the
 * {@link #render} method.  This is necessary to properly determine when a
 * header/footer needs to be emitted.  The output within the given number of
 * lines may vary (e.g. page numbers may increment).
 */
public class PageElement
implements RepeatedOutput
{
   /** The client terminal which is redirected to a paged output stream. */
   private ThinClient term = null;
   
   /** The frame id of the frame to be output. */
   private int frameId = 0;
   
   /** The total number of rows output during rendering. */
   private int height = 0;
   
   /**
    * Construct a page element and initialize its data structures.
    *
    * @param    term
    *           The client terminal which is redirected to a paged output 
    *           stream.
    * @param    frameId
    *           The frame id of the frame to be output.
    */
   public PageElement(ThinClient term, int frameId)
   {
      this.term    = term;
      //this.server  = server;
      this.frameId = frameId;
      
      Frame<?> f = term.getFrame(frameId);
      FrameConfig config = f.config();

      // calculate and save the height
      height = (int) f.height();
      
      if (config.pageBottom && config.box) 
      {
         // if the frame is boxed and page-bottom, do not include the bottom 
         // border in the height.
         // if we don't decrement the height, an additional NL would be 
         // appended to the stream when page-footers are used
         height--;
      }
   }
   
   /**
    * Compares two objects for equality.
    * 
    * @param   x
    *          Object to compare with.
    * @return  <code>true</code> if instances are equal and
    *          <code>false</code> otherwise.
    */
   @Override
   public boolean equals(Object x)
   {
      if (x == null)
         return false;
      
      if (!(x instanceof PageElement))
         return false;
      
      return frameId == ((PageElement)x).frameId;
   }
   
   /**
    * Return a hash code for the instance.
    * 
    * @return  Hash code for the instance.
    */
   @Override
   public int hashCode()
   {
      return frameId;
   }
   
   /**
    * Returns the height of the output as the number of lines that will be
    * generated when {@link #render} is called. 
    *
    * @return   The height of the generated output in lines.
    */
   @Override
   public int height()
   {
      return height;
   }
   
   /**
    * Generate output and write it to the given stream.
    *
    * @param    out
    *           A stream that is opened for output.
    */
   @Override
   public void render(Stream out)
   {
      if (!term.isRedirected())
         throw new RuntimeException(
                   "Redirection required for PageElement to be rendered");
      
      term.view(frameId);
   }
}