/*
** Module   : Region.java
** Abstract : Operations with sets of rectangles 
**
** Copyright (c) 2008-2017 Golden Code Development Corporation.
**
** -#- -I- --Date-- --JPRM-- ----------------Description-----------------
** 001 SIY 20100629          Created initial version
** 002 SIY 20100804          Changes in imports, some refactorings.
** 003 HC  20140911          GUI coordinate support - data types refactoring.
*/
/*
** This program is free software: you can redistribute it and/or modify
** it under the terms of the GNU Affero General Public License as
** published by the Free Software Foundation, either version 3 of the
** License, or (at your option) any later version.
**
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
** GNU Affero General Public License for more details.
**
** You may find a copy of the GNU Affero GPL version 3 at the following
** location: https://www.gnu.org/licenses/agpl-3.0.en.html
** 
** Additional terms under GNU Affero GPL version 3 section 7:
** 
**   Under Section 7 of the GNU Affero GPL version 3, the following additional
**   terms apply to the works covered under the License.  These additional terms
**   are non-permissive additional terms allowed under Section 7 of the GNU
**   Affero GPL version 3 and may not be removed by you.
** 
**   0. Attribution Requirement.
** 
**     You must preserve all legal notices or author attributions in the covered
**     work or Appropriate Legal Notices displayed by works containing the covered
**     work.  You may not remove from the covered work any author or developer
**     credit already included within the covered work.
** 
**   1. No License To Use Trademarks.
** 
**     This license does not grant any license or rights to use the trademarks
**     Golden Code, FWD, any Golden Code or FWD logo, or any other trademarks
**     of Golden Code Development Corporation. You are not authorized to use the
**     name Golden Code, FWD, or the names of any author or contributor, for
**     publicity purposes without written authorization.
** 
**   2. No Misrepresentation of Affiliation.
** 
**     You may not represent yourself as Golden Code Development Corporation or FWD.
** 
**     You may not represent yourself for publicity purposes as associated with
**     Golden Code Development Corporation, FWD, or any author or contributor to
**     the covered work, without written authorization.
** 
**   3. No Misrepresentation of Source or Origin.
** 
**     You may not represent the covered work as solely your work.  All modified
**     versions of the covered work must be marked in a reasonable way to make it
**     clear that the modified work is not originating from Golden Code Development
**     Corporation or FWD.  All modified versions must contain the notices of
**     attribution required in this license.
*/

package com.goldencode.p2j.ui.client;

import java.util.*;

/**
 * The Region class provides means to define, modify and iterate regions
 * which are collections of rectangles.
 * An important feature of regions is their ability to represent the set
 * of rectangles in various ways having the same effect - they all should
 * answer the question - whether a particular point belongs to the set or not
 * - identically.
 * <p>
 * The current implementation is optimized to a level where there is no 
 * intersecting rectangles in the set and adjacent rectangles sharing a side
 * are merged into a single one.
 */
public class Region
{
   /** array of individual non-nesting rectangles */
   private ArrayList<Rectangle> rct = new ArrayList<Rectangle>();

   /** tracks the null rectangle case */
   private boolean allInvalidated = false;

   /**
    * Constructor.
    */
   public Region()
   {
   }

   /**
    * Constructor taking an initial rectangle.
    *
    * @param   rect
    *          rectangle that is added initially to the region
    */
   public Region(Rectangle rect)
   {
      addRectangle(rect);
   }

   /**
    * Adds a rectangle to this region.
    * The region is newly optimized as soon as the rectangle is added 
    *
    * @param   rect
    *          rectangle that is to be added to the region
    */
   public void addRectangle(Rectangle rect)
   {
      // special case of null rectangle - invalidate all
      // by keeping it as the only rectangle
      if (rect == null)
      {
         rct.clear();
         rct.add(rect);
         allInvalidated = true;
         return;
      }

      // trivial case of empty rectangle - no change
      if (rect.empty())
      {
         return;
      }

      // trivial case of empty set
      if (rct.isEmpty())
      {
         rct.add(rect);
         return;
      }

      // another special case - adding a rectangle to null has no effect
      if (allInvalidated)
      {
         return;
      }

      // add the new rectangle to the set, unoptimized
      rct.add(rect);
      
      // now optimize the whole set by merging nested or adjacent rectangles
      // and breaking down intersecting ones
      // this loop terminates after a full pass that detects no changes
      boolean modified = true;

      mod:
      while (modified)
      {
         modified = false;

         // outer loop
         for (int i = 0; i < rct.size() -1; i++) 
         {
            // get the current rectangle
            Rectangle ra = rct.get(i);

            // check the current rectangle against the rest of the set
            // with this inner loop which runs backward
            for (int j = rct.size() - 1; j > i; j--)
            {
               // get the current inner rectangle
               Rectangle rb = rct.get(j);

               // check 1: A and B don't intersect
               //          check whether A and B are adjacent on equal side
               //          move on if not
               if (!ra.intersects(rb))
               {
                  Rectangle newA = null;

                  // check 1-1: B is East of A
                  if (ra.right() + 1 == rb.left())
                  {
                     // sides must be equal
                     if (ra.bottom() == rb.bottom() && ra.top() == rb.top())
                     {
                        // merge A and B into A
                        newA = new Rectangle(ra.top(),
                                             ra.left(), 
                                             ra.bottom(),
                                             rb.right());
                     }
                     else
                     {
                        // no match; move on
                        continue;
                     }
                  }
                  // check 1-2: B is South of A
                  else if (ra.bottom() + 1== rb.top())
                  {
                     // sides must be equal
                     if (ra.left() == rb.left() && ra.right() == rb.right())
                     {
                        // merge A and B into A
                        newA = new Rectangle(ra.top(), 
                                             ra.left(), 
                                             rb.bottom(),
                                             ra.right());
                     }
                     else
                     {
                        // no match; move on
                        continue;
                     }
                  }
                  // check 1-3: B is West of A
                  else if (rb.right() + 1 == ra.left())
                  {
                     // sides must be equal
                     if (ra.bottom() == rb.bottom() && ra.top() == rb.top())
                     {
                        // merge A and B into A
                        newA = new Rectangle(ra.top(), 
                                             rb.left(), 
                                             ra.bottom(),
                                             ra.right());
                     }
                     else
                     {
                        // no match; move on
                        continue;
                     }
                  }
                  // check 1-4: B is North of A
                  else if (rb.bottom() + 1 == ra.top())
                  {
                     // sides must be equal
                     if (ra.left() == rb.left() && ra.right() == rb.right())
                     {
                        // merge A and B into A
                        newA = new Rectangle(rb.top(), 
                                             ra.left(), 
                                             ra.bottom(),
                                             ra.right());
                     }
                     else
                     {
                        // no match; move on
                        continue;
                     }
                  }
                  // not adjacent: move on
                  else
                  {
                     continue;
                  }

                  // common code for cases 1-*
                  // replace A with the merged rectangle
                  rct.set(i, newA);

                  // remove B
                  rct.remove(j);

                  // restart
                  modified = true;
                  continue mod;
              }

               // check 2: A includes B
               //          throw away B and move on
               if (ra.contains(rb))
               {
                  rct.remove(j);
                  continue;
               }

               // check 3: A is included in B
               //          throw away A and restart
               if (rb.contains(ra))
               {
                  rct.remove(i);
                  modified = true;
                  continue mod;
               }

               // the rest of the cases is various forms of intersections
               // check which corners of B are inside of A
               boolean p1 = ra.contains(rb.left(),  rb.top());
               boolean p2 = ra.contains(rb.right(), rb.top());
               boolean p3 = ra.contains(rb.right(), rb.bottom());
               boolean p4 = ra.contains(rb.left(),  rb.bottom());

               // corners of B inside A
               int corners = 0;
               if (p1)
               {
                  corners++;
               }
               if (p2)
               {
                  corners++;
               }
               if (p3)
               {
                  corners++;
               }
               if (p4)
               {
                  corners++;
               }

               // cases 4 and 5: no corners
               //      +--------+                +--------+ 
               //      | A      |                | A      |
               //   +---------------+    or   +---------------+
               //   |B |        |   |         |B |        |   |
               //   |  +--------+   |         +---------------+
               //   +---------------+            +--------+
               //
               //  B covers a side of A or A and B intersect by mid sections
               if (corners == 0)
               {
                  // telling the two cases by counting corners of A inside B
                  p1 = rb.contains(ra.left(),  ra.top());
                  p2 = rb.contains(ra.right(), ra.top());
                  p3 = rb.contains(ra.right(), ra.bottom());
                  p4 = rb.contains(ra.left(),  ra.bottom());
                  
                  if (p1 || p2 || p3 || p4)
                  {
                     // case 4: B covers a side of A; A and B may share coordinate.
                     // A is reduced in size
                     Rectangle newA = null;
   
                     // check 4-1: B is East of A
                     if (rb.right() >= ra.right() && rb.left() >= ra.left())
                     {
                        // shrink A from the right
                        newA = new Rectangle(ra.top(), 
                                             ra.left(), 
                                             ra.bottom(),
                                             rb.left() - 1);
                     }
   
                     // check 4-2: B is South of A
                     else if (rb.bottom() >= ra.bottom() && rb.top() >= ra.top())
                     {
                        // shrink A from the bottom
                        newA = new Rectangle(ra.top(), 
                                             ra.left(), 
                                             rb.top() - 1,
                                             ra.right());
                     }
   
                     // check 4-3: B is West of A
                     else if (rb.left() <= ra.left() && rb.right() <= ra.right())
                     {
                        // shrink A from the left
                        newA = new Rectangle(ra.top(), 
                                             rb.right() + 1, 
                                             ra.bottom(),
                                             ra.right());
                     }
   
                     // case 4-4:  B is North of A
                     else
                     {
                        // shrink A from the top
                        newA = new Rectangle(rb.bottom() + 1, 
                                             ra.left(), 
                                             ra.bottom(),
                                             ra.right());
                     }
   
                     // common code for cases 4-*
                     // replace A with the shrunken rectangle
                     rct.set(i, newA);
   
                     // move on
                     modified = true;
                     continue mod;
                  }
                  
                  // case 5: A and B intersect by mid sections
                  // B is split into C and D
                  Rectangle rc = null;
                  Rectangle rd = null;

                  if (ra.top() < rb.top())
                  {
                     rc = new Rectangle(rb.top(),
                                        rb.left(),
                                        rb.bottom(),
                                        ra.left() - 1);
                     rd = new Rectangle(rb.top(),
                                        ra.right() + 1,
                                        rb.bottom(),
                                        rb.right());
                  }
                  else
                  {
                     rc = new Rectangle(rb.top(),
                                        rb.left(),
                                        ra.top() - 1,
                                        rb.right());
                     rd = new Rectangle(ra.bottom() + 1,
                                        rb.left(),
                                        rb.bottom(),
                                        rb.right());
                  }

                  // remove B
                  rct.remove(j); 
  
                  // add C and D rectangles
                  rct.add(rc);
                  rct.add(rd);
  
                  // restart
                  modified = true;
                  continue mod;
              }

               // case 6: one corner of B is inside A
               //      +--------+
               //      | A      |
               //      |  +---------+            +---+
               //      |  | B   |   |            |   | D
               //      +--|-----+   |      +-----+---+
               //         +---------+      +---------+ C
               //
               // B is split into C and D
               else if (corners == 1)
               {
                  Rectangle rc = null;
                  Rectangle rd = null;

                  // check 6-1: B is South-East of A
                  if (p1)
                  {
                     // create partial rectangles C and D
                     rc = new Rectangle(ra.bottom() + 1,
                                        rb.left(),
                                        rb.bottom(),
                                        rb.right());
                     rd = new Rectangle(rb.top(),
                                        ra.right() + 1,
                                        ra.bottom(),
                                        rb.right());
                  }

                  // check 6-2: B is South-West of A
                  else if (p2)
                  {
                     // create partial rectangles C and D
                     rc = new Rectangle(ra.bottom() + 1,
                                        rb.left(),
                                        rb.bottom(),
                                        rb.right());
                     rd = new Rectangle(rb.top(),
                                        rb.left(),
                                        ra.bottom(),
                                        ra.left() - 1);
                  }

                  // check 6-3: B is North-West of A
                  else if (p3)
                  {
                     // create partial rectangles C and D
                     rc = new Rectangle(rb.top(),
                                        rb.left(),
                                        ra.top() - 1,
                                        rb.right());
                     rd = new Rectangle(ra.top(),
                                        rb.left(),
                                        rb.bottom(),
                                        ra.left() - 1);
                  }

                  // case  6-4: B is North-East of A
                  else
                  {
                     // create partial rectangles C and D
                     rc = new Rectangle(rb.top(),
                                        rb.left(),
                                        ra.top() - 1,
                                        rb.right());
                     rd = new Rectangle(ra.top(),
                                        ra.right() + 1,
                                        rb.bottom(),
                                        rb.right());
                  }

                  // common code for cases 6-*
                  // remove B
                  rct.remove(j); 

                  // add C and D rectangles
                  rct.add(rc);
                  rct.add(rd);

                  // restart
                  modified = true;
                  continue mod;
               }

               // case 7: two corners (A covers a side of B)
               //      +--------+
               //      | A  +-------+
               //      |    | B |   |
               //      |    +-------+
               //      +--------+
               //
               // B is reduced in size
               else
               {
                  Rectangle newB = null;

                  // check 7-1: B is East of A
                  if (rb.right() > ra.right() && rb.left() >= ra.left())
                  {
                     // shrink B from the left
                     newB = new Rectangle(rb.top(), 
                                          ra.right() + 1, 
                                          rb.bottom(),
                                          rb.right());
                  }

                  // check 7-2: B is South of A
                  else if (rb.bottom() > ra.bottom() && rb.top() >= ra.top())
                  {
                     // shrink B from the top
                     newB = new Rectangle(ra.bottom() + 1, 
                                          rb.left(), 
                                          rb.bottom(),
                                          rb.right());
                  }

                  // check 7-3: B is West of A
                  else if (rb.left() < ra.left() && rb.right() <= ra.right())
                  {
                     // shrink B from the right
                     newB = new Rectangle(rb.top(), 
                                          rb.left(),
                                          rb.bottom(),
                                          ra.left() - 1);
                  }

                  // case 7-4:  B is North of A
                  else
                  {
                     // shrink B from the bottom
                     newB = new Rectangle(rb.top(), 
                                          rb.left(), 
                                          ra.top() - 1,
                                          rb.right());
                  }

                  // common code for cases 7-*
                  // replace B with the shrunken rectangle
                  rct.set(j, newB);

                  // move on
                  modified = true;
                  continue mod;
               }
            }
         }
      }
   }

   /**
    * Clears this region entirely.
    */
   public void resetRegion()
   {
      rct.clear();
      allInvalidated = false;
   }

   /**
    * Checks whether the region is empty.
    *
    * @return  <code>true</code> when region has no rectangles
    */
   public boolean isEmpty()
   {
      return rct.isEmpty();
   }

   /**
    * Checks whether the region contains the given point.
    *
    * @param   p
    *          point to check
    * @return  index of the rectangle that contains the point 
    *          or <code>-1</code> if none
    */
   public int contains(Point p)
   {
      return contains(p.x, p.y);
   }

   /**
    * Checks whether the region contains the given point.
    *
    * @param   x
    *          x-coordinate of the point to check
    * @param   y
    *          y-coordinate of the point to check
    * @return  index of the rectangle that contains the point 
    *          or <code>-1</code> if none
    */
   public int contains(double x, double y)
   {
      for (int i = 0; i < rct.size(); i++)
      {
         Rectangle r = rct.get(i);

         if (r == null || r.contains(x, y))
         {
            return i;
         }
      }

      return -1;
   }

   /**
    * Gets an iterator for the region.
    *
    * @return  an iterator that can be used to traverse all rectangles
    *          of the region
    */
   public Iterator<Rectangle> iterator()
   {
      return rct.iterator();
   }

   /**
    * Gets a string representation of this region.
    *
    * @return  a string describing the region
    */
   @Override
   public String toString()
   {
      StringBuffer sb = new StringBuffer("{");

      for (int i = 0; i < rct.size(); i++)
      {
         sb.append(rct.get(i).toString());
         sb.append(",");
      }

      if (rct.size() > 0)
      {
         sb.setLength(sb.length() - 1);
      }
      sb.append("}");
      
      return sb.toString();
   }
}