/*
** Module   : Text.java
** Abstract : Java standard header
**
** Copyright (c) 2005-2017, Golden Code Development Corporation.
**
** -#- -I- --Date-- --JPRM-- -----------------------------Description-----------------------------
** 001 SVG 20050609   @21452 Initial version.
** 002 SIY 20050831   @22518 Refactoring (see @@22394-22409) 
** 003 SIY 20050907   @22653 Fixed use of @inheritDoc tag in javadoc.
** 004 SIY 20050918   @22850 Added support for remaining attributes.  
** 005 SIY 20060209   @24447 Updated to accommodate changes in xxxConfig classes.
** 006 SIY 20060316   @25141 Fixed a typo in comments.
** 007 SIY 20060714   @28070 Intercepted all ways to clear content or prevent displaying it 
**                           properly.
** 008 SIY 20060824   @28861 Removed getP2JType() method.
** 009 NVS 20080401   @37784 Added isFocusable() method implementation.
** 010 SIY 20100619          Switching to own implementation of character UI.
** 011 SIY 20100804          Changes in imports, some refactorings.
** 012 SIY 20100822          Refactored AttributeContainer support.
** 013 SIY 20101122          Switching from KeyEvent to KeyInput.
** 014 SIY 20110112          Changes related to new color handling scheme.
** 015 SIY 20111202          Refactored and moved from CHUI package.
** 016 SIY 20111221          Minor cleanups.
** 017 CA  20140805          Added generics to force the abstract widget classes to not use
**                           implementation-specific code: the abstract widgets will have access
**                           only to the base APIs provided by the OutputManager.
** 018 CA  20140926          Added shared widget configuration support, to allow GUI/ChUI
**                           concrete implementation for the same widget ID.  Refactored the
**                           widget configuration classes: all fields were made public; these
**                           classes need to be as dumb as possible, all logic related to
**                           setting/getting a certain field should be at the widget or at the
**                           caller. Refs #2254
** 019 CA  20141220          Refactoring to integrate the GUI TEXT implementation with the ChUI 
**                           logic.
** 020 HC  20150323          Removed business logic from config classes.
** 021 CA  20161101          Added support for mnemonic processing.
** 022 CA  20171020          Added support for FRAMEs sent to named or unnamed streams, in GUI.
*/
/*
** This program is free software: you can redistribute it and/or modify
** it under the terms of the GNU Affero General Public License as
** published by the Free Software Foundation, either version 3 of the
** License, or (at your option) any later version.
**
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
** GNU Affero General Public License for more details.
**
** You may find a copy of the GNU Affero GPL version 3 at the following
** location: https://www.gnu.org/licenses/agpl-3.0.en.html
** 
** Additional terms under GNU Affero GPL version 3 section 7:
** 
**   Under Section 7 of the GNU Affero GPL version 3, the following additional
**   terms apply to the works covered under the License.  These additional terms
**   are non-permissive additional terms allowed under Section 7 of the GNU
**   Affero GPL version 3 and may not be removed by you.
** 
**   0. Attribution Requirement.
** 
**     You must preserve all legal notices or author attributions in the covered
**     work or Appropriate Legal Notices displayed by works containing the covered
**     work.  You may not remove from the covered work any author or developer
**     credit already included within the covered work.
** 
**   1. No License To Use Trademarks.
** 
**     This license does not grant any license or rights to use the trademarks
**     Golden Code, FWD, any Golden Code or FWD logo, or any other trademarks
**     of Golden Code Development Corporation. You are not authorized to use the
**     name Golden Code, FWD, or the names of any author or contributor, for
**     publicity purposes without written authorization.
** 
**   2. No Misrepresentation of Affiliation.
** 
**     You may not represent yourself as Golden Code Development Corporation or FWD.
** 
**     You may not represent yourself for publicity purposes as associated with
**     Golden Code Development Corporation, FWD, or any author or contributor to
**     the covered work, without written authorization.
** 
**   3. No Misrepresentation of Source or Origin.
** 
**     You may not represent the covered work as solely your work.  All modified
**     versions of the covered work must be marked in a reasonable way to make it
**     clear that the modified work is not originating from Golden Code Development
**     Corporation or FWD.  All modified versions must contain the notices of
**     attribution required in this license.
*/

package com.goldencode.p2j.ui.client;

import com.goldencode.p2j.ui.*;
import com.goldencode.p2j.ui.client.event.*;
import com.goldencode.p2j.ui.client.widget.*;
import com.goldencode.p2j.util.*;

/**
 * Implementation of Progress TEXT component.
 */
public abstract class Text<O extends OutputManager<?>>
extends FillIn<O,TextConfig>
{
   /** Track value assignment.  */
   protected boolean assigned = false;

   /** Mnemonic information for the TEXT's content. */
   protected final MnemonicTrait<O> contentMnemonic;
   
   /**
    * Special constructor used to restore component using given configuration.
    * 
    * @param   id
    *          The widget's ID.
    * @param   cfg
    *          Configuration data.
    * @param   contentMnemonic
    *          The mnemonic information for the TEXT's content.
    */
   public Text(WidgetId id, TextConfig cfg, MnemonicTrait<O> contentMnemonic)
   {
      super(id, cfg);
      setUnderline(false);
      
      this.contentMnemonic = contentMnemonic;
   }
   
   /**
    * Checks whether this attribute container can receive focus generally.
    * 
    * @return  <code>true</code> if this container can be focused under
    *          certain conditions. <code>false</code> if, and only if, this
    *          attribute container can never receive focus.    
    */
   @Override
   public boolean isFocusable()
   {
      return false;
   }

   /**
    * Sets position of terminal's cursor. Really does nothing because this
    * widget will never get an input focus.
    */
   @Override
   public void drawCaret()
   {
      // dummy method. This control has no caret
   }

   /**
    * This component will not receive an input focus when Tab or Shift-Tab is
    * pressed.
    * 
    * @return  Always <code>false</code>.
    */
   @Override
   public boolean focusTraversable()
   {
      return false;
   }

   /**
    * Sets the new variable value in the widget, set appropriate flag and update the mnemonic.
    * 
    * @param   value
    *          New value for the screen buffer.
    */
   @Override
   public void setValue(BaseDataType value)
   {
      assigned = true;
      super.setValue(value);
      
      updateSize();

      contentMnemonic.updateMnemonic(null, getText(false), false);
   }
   
   /**
    * Fills data structures from variable's value and updates the mnemonic.
    */
   @Override
   public void display()
   {
      super.display();
      
      contentMnemonic.updateMnemonic(null, getText(false), false);
   }

   /**
    * Overridden to prevent clearing of the content.
    * 
    * @param    selectionStart
    *           The start selection index (inclusive).
    * @param    selectionEnd
    *           The end selection index (exclusive).
    */
   @Override
   public void clear(int selectionStart, int selectionEnd)
   {
   }
   
   /**
    * Intercept and ignore key events.
    * 
    * @param   ke
    *          Key event structure.
    */
   @Override
   public void processKeyEvent(KeyInput ke)
   {
      ke.consume();
      return;
   }

   /**
    * Compute the initial cursor position, depending on the widget's implementation.
    * <p>
    * For TEXT widgets, this always returns 0.
    *  
    * @return   See above.
    */
   @Override
   public int getInitialCursorPos()
   {
      return 0;
   }
   
   /**
    * Copy the widget state from the specified source widget to this instance.
    * 
    * @param    w
    *           The source widget which state needs to be copied.
    */
   @Override
   protected void copyState(Widget<?> w)
   {
      Text<?> widget = (Text<?>) w;
      
      super.copyState(w);

      this.assigned = widget.assigned;
   }
   
   /**
    * Get the cursor position, in native units.
    * 
    * @return   Always zero.
    */
   @Override
   protected int getNativeCursorPos()
   {
      return 0;
   }
   
   /**
    * Determine the portion of the widget's editable part.  This may be different than the widget's
    * width, as determined by the WIDTH-CHARS or WIDTH-PIXELS attribute.
    * <p>
    * For ChUI widgets, this is the same as the widget's physical dimension.
    * 
    * @return   The editable dimension of this widget, in native units.
    */
   @Override
   protected NativeDimension editableNativeDimension()
   {
      return physicalDimension();
   }
   
   /**
    * Container class for the TEXT's content mnemonic.
    */
   protected static class ContentMnemonic<G extends OutputManager<?>>
   implements MnemonicTrait<G>
   {
      /** The mnemonic info */
      private MnemonicInfo mnemonic = null;
      
      public ContentMnemonic()
      {
         // no-op
      }

      /**
       * MnemonicInfo instance getter.
       * <p>
       * The implementors should simply return the value of the mnemonic class field. No
       * additional processing logic is advised.
       *
       * @return   See above.
       */
      @Override
      public MnemonicInfo getMnemonic()
      {
         return mnemonic;
      }

      /**
       * MnemonicInfo instance setter.
       * <p>
       * The implementors should simply assign the parameter value to the mnemonic class
       * field. No additional processing logic is advised.
       */
      @Override
      public void setMnemonic(MnemonicInfo mnemonic)
      {
         this.mnemonic = mnemonic;
      }
   }
}