/*
** Module   : TypeAhead.java
** Abstract : Buffered keyboard with CTRL-C processing.
**
** Copyright (c) 2006-2018, Golden Code Development Corporation.
**
** -#- -I- --Date-- --JPRM-- ----------------Description-----------------
** 001 NVS 20060206   @24295 Initial version.
** 002 SIY 20060209   @24437 Minor cleanups.
** 003 NVS 20060213   @24584 Added suspend() and resume() methods. They
**                           are used to disable keyboard reading thread
**                           temporarily while a child process is running
**                           and, possibly using the keyboard input 
**                           directly.
** 004 SIY 20060213   @24579 Fixed trap when timeout occurs and still no 
**                           events in the queue.
** 005 SIY 20060222   @24646 Minor Javadoc cleanups.
** 006 NVS 20060301   @24873 Installs and uses P2J keyboard input factory.
** 007 GES 20060629   @27720 Split up the reading of the terminal key 
**                           code from the conversion of that into an
**                           event object.  This is needed because by its
**                           nature, type-ahead processing means that the
**                           key may be processed an arbitrary amount of
**                           time after it was generated by the user. In
**                           that time during which it resides in the
**                           buffer, the currently focused widget may
**                           change many times.  Proper subsequent 
**                           processing of the key requires that the
**                           currently focused widget is set as the source
**                           component.  The old monolithic approach 
**                           sometimes yielded a source component that
**                           was incorrect and unexpected/undefined 
**                           results would occur.
** 008 SIY 20080901   @39639 Inferred generics and some minor cleanups.
** 009 GES 20090702   @43032 Fixed init() to properly re-initialize on CTRL-C.
** 010 SIY 20100619          Switching to own implementation of character UI.
** 011 SIY 20100804          Changes in imports, some refactorings.
** 012 SIY 20101202          Removed unnecessary mouse event handling.
** 013 SIY 20110112          Changes related to refactoring of driver code.
** 014 GES 20111012          Moved from a static to instance approach. There is
**                           a single instance now per context. This allows
**                           multiple client sessions to be hosted in the
**                           server.
** 015 SIY 20111101          Moved from CHUI package. Minor 
**                           cleanups/refactorings in preparation to GUI client.
** 016 GES 20111103          Modified the code to setup and morph security
**                           contexts when run in the server process. Also in
**                           that mode, this class allows the key reading
**                           thread to be killed and restarted, which enables
**                           proper termination and CTRL-C restart.
** 017 GES 20111110          Moved security-specific code into a dedicated
**                           thread class and use that.
** 018 SIY 20111202          Updated imports.
** 019 CA  20130822          Added support of appserver async requests: posting and executing of
**                           server-side events on the client side.
** 020 EVL 20140114          Batch mode does not need the key reader thread, not starting. 
** 021 MAG 20140507          Make sure reader thread is not waiting when going to wait in suspend(). 
** 022 CA  20140805          Added mouse support.
** 023 CA  20141119          Refactored drawing code to enable multiple OS windows, for GUI
**                           support.  Misc changes/fixes related to drawing on multiple windows.
** 024 CA  20141123          The key which raises STOP condition and restarts the client is 
**                           determined from the in-use OutputManager.
** 025 CA  20141206          Avoid resolving OutputManager on each readkey loop iteration.
** 026 ECF 20150127          Added safety checks to avoid NPE on reader, which is null when in
**                           batch mode. Made suspended variable volatile.
** 027 CA  20150826          Before listening for events, register with the GUI driver all
**                           widgets which can accept mouse events. All windows listen to any
**                           mouse actions, any time.  Mouse actions can not be disabled during
**                           drawing, as on slower machines it might mean ignoring posted mouse
**                           released/clicked events, as P2J is busy drawing/processing the mouse
**                           pressed event.
** 028 SBI 20160205          Fixed deadlock that occurs only in web client.
** 029 EVL 20180308          Externalize the ability to get/set currently used keyboard source.
**                           The GuiDriver.(suspend|resume) uses these calls to save and restore
**                           key reader.
*/
/*
** This program is free software: you can redistribute it and/or modify
** it under the terms of the GNU Affero General Public License as
** published by the Free Software Foundation, either version 3 of the
** License, or (at your option) any later version.
**
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
** GNU Affero General Public License for more details.
**
** You may find a copy of the GNU Affero GPL version 3 at the following
** location: https://www.gnu.org/licenses/agpl-3.0.en.html
** 
** Additional terms under GNU Affero GPL version 3 section 7:
** 
**   Under Section 7 of the GNU Affero GPL version 3, the following additional
**   terms apply to the works covered under the License.  These additional terms
**   are non-permissive additional terms allowed under Section 7 of the GNU
**   Affero GPL version 3 and may not be removed by you.
** 
**   0. Attribution Requirement.
** 
**     You must preserve all legal notices or author attributions in the covered
**     work or Appropriate Legal Notices displayed by works containing the covered
**     work.  You may not remove from the covered work any author or developer
**     credit already included within the covered work.
** 
**   1. No License To Use Trademarks.
** 
**     This license does not grant any license or rights to use the trademarks
**     Golden Code, FWD, any Golden Code or FWD logo, or any other trademarks
**     of Golden Code Development Corporation. You are not authorized to use the
**     name Golden Code, FWD, or the names of any author or contributor, for
**     publicity purposes without written authorization.
** 
**   2. No Misrepresentation of Affiliation.
** 
**     You may not represent yourself as Golden Code Development Corporation or FWD.
** 
**     You may not represent yourself for publicity purposes as associated with
**     Golden Code Development Corporation, FWD, or any author or contributor to
**     the covered work, without written authorization.
** 
**   3. No Misrepresentation of Source or Origin.
** 
**     You may not represent the covered work as solely your work.  All modified
**     versions of the covered work must be marked in a reasonable way to make it
**     clear that the modified work is not originating from Golden Code Development
**     Corporation or FWD.  All modified versions must contain the notices of
**     attribution required in this license.
*/

package com.goldencode.p2j.ui.client;

import java.util.*;
import com.goldencode.p2j.security.*;
import com.goldencode.p2j.ui.client.driver.ScreenDriver;
import com.goldencode.p2j.ui.client.event.*;
import com.goldencode.p2j.ui.client.gui.driver.GuiDriver;
import com.goldencode.p2j.util.*;

/**
 * This class is responsible for buffering keyboard input with type ahead
 * and handling of CTRL-C keystrokes.
 */
public class TypeAhead
{
   /** Context local instance of this class. */
   private static final ContextLocal<TypeAhead> local = new ContextLocal<>();
   
   /** The type-ahead buffer which keeps a strict FIFO ordering. */
   private List<Integer> taBuf = new ArrayList<>();
   
   /** Suspend/resume flag. */
   private volatile boolean suspended = false;
   
   /** Watcher instance. */
   private Watcher watch = null;
   
   /** Key reading thread. */
   private AssociatedThread reader = null;
   
   /** Reader thread kill flag. */
   private boolean exit = false;
   
   /** Lock access to the exit flag. */
   private Object exitLock = new Object();
   
   /** Output manager used to obtain incoming keyboard events. */
   private volatile OutputManager<?> mgr;
      
   /**
    * Constructor which ensures that this class can only be used on a
    * <code>static</code> basis.
    *
    * @param    watch
    *           The CTRL-C signal handler to use.
    */
   private TypeAhead(Watcher watch) 
   {
      this.watch = watch;
      startReader();
   }
   
   /**
    * Obtain the context-local instance of this class.
    *
    * @return   The instance or <code>null</code> if the class was never
    *           initialized.
    */
   public static synchronized TypeAhead instance()
   {
      return local.get();
   }
   
   /**
    * Initializes (or re-initializes) the type ahead buffer.
    * 
    * @param    watch
    *           The CTRL-C signal handler to use.
    *
    * @return   The initialized instance.
    */
   public static synchronized TypeAhead init(Watcher watch)
   {
      TypeAhead instance = instance();
      
      if (instance != null)
      {
         instance.watch = watch;
         instance.resume();
         instance.clear();
         
         synchronized (instance.exitLock)
         {
            if (instance.exit)
            {
               instance.exit = false;
               instance.startReader();
            }
         }
      }
      else
      {
         instance = new TypeAhead(watch);
         local.set(instance);
      }
      
      return instance;
   }
   
   /**
    * Gets the key reading source currently in use.
    * 
    * @return   The output manager used to get the key events.
    */
   public static synchronized OutputManager<?> getInputSource()
   {
      TypeAhead instance = instance();
      return instance != null ? instance.mgr : null;
   }
   
   /**
    * Sets new source to read the key events.
    * 
    * @param    newMgr
    *           The output manager to use for key events reading.
    */
   public static synchronized void setInputSource(OutputManager<?> newMgr)
   {
      TypeAhead instance = instance();
      if (instance != null)
      {
         instance.mgr = newMgr;
      }
   }
   
   /**
    * Suspends the type ahead buffer activity and keyboard reading thread.
    */
   public synchronized void suspend()
   {
      // if in batch mode, reader thread is null; nothing to do
      if (reader == null)
      {
         return;
      }
      
      if (taBuf == null || suspended)
      {
         return;
      }
      
      // notify the keyboard reading thread about suspension
      suspended = true;
      
      // wait for the release
      try
      {
         // Make sure reader is not waiting.
         reader.interrupt();
         wait();
      }
      catch (InterruptedException ie)
      {
         // nothing to do
      }
   }
   
   /**
    * Resume the type ahead buffer activity and keyboard reading thread.
    */
   public synchronized void resume()
   {
      // if in batch mode, reader thread is null; nothing to do
      if (reader == null)
      {
         return;
      }
      
      if (taBuf == null || !suspended)
      {
         return;
      }
      
      // notify the keyboard reading thread to resume reading
      suspended = false;
      notify();
   }
   
   /**
    * Purges the type ahead buffer.
    */
   public synchronized void clear()
   {
      taBuf.clear();
   }
   
   /**
    * Conditionally blocks until an input appears in the buffer.
    */
   public synchronized void waitForInput()
   {
      // if in batch mode, reader thread is null; nothing to do
      if (reader == null)
      {
         return;
      }
      
      if (!taBuf.isEmpty())
      {
         return;
      }
      
      try
      {
         wait();
      }
      
      catch (InterruptedException ie)
      {
         throw new StopConditionException(ie);
      }
   }
   
   /**
    * Notify the type-ahead that a server event was posted.
    */
   public synchronized void notifyServerEvent()
   {
      notify();
   }
   
   /**
    * Notify the type-ahead that a OS event was posted.
    */
   public synchronized void notifyOSEvent()
   {
      notify();
   }
   
   /**
    * Returns the next event from the type ahead buffer or blocks up to the specified number of 
    * milliseconds waiting for input.
    * <p>
    * If no key is received and the <code>honorServerEvents</code> parameter is set to 
    * <code>true</code>, this will return <code>null</code> if the 
    * {@link EventManager#hasServerEvents()} reports that there are pending server events; this 
    * will allow the server events to be processed.
    * <p>
    * The downside of this implementation (as 4GL shows no fixed order for mixed UI and
    * server-side events, see {@link EventManager} javadoc for more details) is that keys have 
    * precedence over server-side events and a constant key press would delay server event 
    * processing.
    * <p> 
    * {@link TypeAhead#getKeystroke} was coded this way because the server events are pushed in a
    * global queue, separated from the UI/key events, as server-events can be posted even when a 
    * WAIT-FOR-like loop is not in effect.
    *
    * @param    msec
    *           The maximum time in milliseconds to wait if there is no input.
    *           Less than 0 means an indefinite wait, 0 means no wait (return
    *           an event only if there is already one in the queue) and any
    *           positive value is the maximum number of milliseconds to wait.
    * @param    honorServerEvents
    *           flag indicating if server events should be honored or not.
    *
    * @return   The next event in the queue or <code>null</code> if the read
    *           timed out.
    */
   public KeyInput getKeystroke(int msec, boolean honorServerEvents) 
   {
      KeyInput event = null;
      
      int savedMsec = msec;
      
      ScreenDriver<?> driver = OutputManager.getDriver();
      
      // mouse events are captured constantly - otherwise, if the drawing is done on a slower 
      // system, a mouse press/released/clicked chain will be interrupted, as the released/clicked
      // may be raised while the main thread is busy drawing, but they still need to be processed
      
      // before starting listening for an OS event, register with the driver all widgets which
      // are ENABLED, VISIBLE and can react to mouse events.  Do this only if the output
      // manager supports mouse events
      if (driver instanceof GuiDriver)
      {
         ((GuiDriver) driver).registerMouseWidgets();
      }
      
      synchronized (this)
      {
         try
         {
            do
            {
               msec = savedMsec;
               event = dequeueEvent();
               
               if (event == null)
               {
                  if (EventManager.hasServerEvents() || EventManager.hasOSEvents())
                  {
                     // if events were already received, let them process
                     break;
                  }
                  
                  if (msec == 0)
                  {
                     return null;
                  }
                  
                  if (msec < 0)
                  {
                     msec = 0;
                  }
                  
                  try
                  {
                     wait(msec);
                  }
                  
                  catch (InterruptedException ie)
                  {
                     throw new StopConditionException(ie);
                  }
                  
                  event = dequeueEvent();
                  
                  if (event == null)
                  {
                     if (honorServerEvents && EventManager.hasServerEvents())
                     {
                        // if no key received but a server or mouse event was received, let it
                        // process
                        break;
                     }
                     
                     if (EventManager.hasOSEvents())
                     {
                        // if there are OS events, exit
                        break;
                     }
                  }
               }
            }
            while (event == null && honorServerEvents && msec == 0);
         }
         finally
         {
            // do nothing for now
         }
         
         return event;
      }
   }
   
   /**
    * Modify state to cause the key reading thread to exit. Wait until that
    * thread has exited and then return.
    */
   public void closeReader()
   {
      synchronized (this)
      {
         // if in batch mode, reader thread is null; nothing to do
         if (reader == null)
         {
            return;
         }
      }
      
      synchronized (exitLock)
      {
         exit = true;
         
         synchronized (this)
         {
            if (suspended)
            {
               resume();
            }
            else
            {
               // if not suspended, the thread is probably blocked waiting for
               // input so we try to unblock it here
               reader.interrupt();
            }
         }
      }
      
      // wait for the reader thread to terminate
      try
      {
         reader.join();
      }
      catch (InterruptedException ie)
      {
         // nothing to do
      }
   }
   
   /**
    * Save the given context as a pending context switch which will be
    * processed later by the key reading thread.
    *
    * @param    context
    *           The security context to switch to (if non-null).
    */
   public void setPendingContextSwitch(Object context)
   {
      synchronized (this)
      {
         // if in batch mode, reader thread is null; nothing to do
         if (reader == null)
         {
            return;
         }
      }
      
      reader.setPendingContextSwitch(context);
   }
   
   /**
    * Create and start the key reading thread.
    */
   private synchronized void startReader()
   {
      // In batch mode we do not start reader thread to prevent -2(EOF) events to overflow
      // the type ahed buffer
      if (OutputManager.inBatchMode())
      {
         return;
      }
      reader = new AssociatedThread(new KeyReader());
      reader.setName("KeyReader");
      reader.setDaemon(true);
      reader.start();
   }
   
   /**
    * Dequeue a key code from the head of the queue (if the queue is not 
    * empty), convert it to an event (which resolves the source component) 
    * and return that event.
    *
    * @return   The complete event represented by the key code in the type
    *           ahead queue or <code>null</code> if the queue is empty.
    */
   private synchronized KeyInput dequeueEvent()
   {
      KeyInput event = null;
      
      if (!taBuf.isEmpty())
      {
         event = EventManager.eventFromKey((taBuf.get(0)).intValue());
         taBuf.remove(0);
      }
      
      return event;
   }
   
   /**
    * Type ahead buffer feeding class.  It runs a daemon thread which polls
    * the keyboard and feeds the keystrokes into the type-ahead buffer
    * continuously.
    */
   private class KeyReader
   implements Runnable 
   {
      /**
       * The keyboard polling thread.  It calls the driver's keyboard reading
       * method. If a keystroke is returned, this method adds it to the end of
       * the buffer and notifies a blocked thread, if one is blocked.
       */
      @Override
      public void run() 
      {
         // set up initial key reading source
         mgr = OutputManager.instance();
         
         while (true)
         {
            synchronized (exitLock)
            {
               if (exit)
               {
                  break;
               }
            }
            
            // check for suspension request
            if (suspended)
            {
               synchronized (TypeAhead.this)
               {
                  // confirm receipt
                  taBuf.clear();
                  TypeAhead.this.notify();
                  
                  // block indefinitely till resumed
                  try
                  {
                     TypeAhead.this.wait();
                  }
                  
                  catch (InterruptedException ie)
                  {
                     // nothing to do
                  }
                  
                  // recheck for suspension for safety
                  continue;
               }
            }
            
            // make sure our context is switched if a switch is pending; this
            // allows the current OutputManager to be accessed dynamically (the
            // instance changes when the context changes) 
            reader.morphContextIfNeeded();
            
            // no suspension, usual cycle
            int key = mgr.readKey();
            
            synchronized (exitLock)
            {
               if (exit)
               {
                  break;
               }
            }
            
            // make sure our context is switched if a switch is pending; we
            // may have just been paused for a long time and if a switch
            // occurred while we were waiting, we need to switch now before
            // we potentially process any CTRL-C below (it runs code in other
            // classes that accesses context-specific instances)
            reader.morphContextIfNeeded();
            
            // check for timeout
            if (key == -1)
            {
               continue;
            }
            
            // set the default disposition to enqueue
            int disp = Watcher.QUEUE;
            
            // delegate CTRL-C/CTRL-BREAK processing
            if (mgr.raiseStopCondition(key))
            {
               disp = watch.watchCtrlC();
            }
            
            if (disp == Watcher.IGNORE)
            {
               // bypass queueing because it has been handled
               continue;
            }
            
            synchronized (TypeAhead.this)
            {
               if (disp == Watcher.CLEAR_AND_QUEUE)
               {
                  taBuf.clear();
               }
               
               // enqueue the raw key code
               taBuf.add(new Integer(key));
               
               // awaken one waiting thread if there are any that are waiting
               TypeAhead.this.notify();
            }
         }
      }
   }
}