/*
** Module   : WindowChuiImpl.java
** Abstract :
**
** Copyright (c) 2015-2017, Golden Code Development Corporation.
**
** -#- -I- --Date-- ---------------------------------Description---------------------------------
** 001 MAG 20140521 First version. Extracted from WindowImpl class.
** 002 CA  20140805 Added generics to force the abstract widget classes to not use implementation-
**                  specific code: the abstract widgets will have access only to the base APIs
**                  provided by the OutputManager.
**                  Removed the WindowImpl class and moved code here - all base behavior is in the
**                  Window class.
**                  Refactoring to allow multi-window support.
** 003 HC  20140911 GUI coordinate support - data types refactoring.
** 004 CA  20140926 Added shared widget configuration support, to allow GUI/ChUI concrete
**                  implementation for the same widget ID.  Refactored the widget configuration
**                  classes: all fields were made public; these classes need to be as dumb as
**                  possible, all logic related to setting/getting a certain field should be at
**                  the widget or at the caller. Refs #2254
** 005 MAG 20141009 Added INSERT indicator widget which is ChUI specific only.
** 006 HC  20141021 Implemented window move/position support.
** 007 CA  20141119 Refactored drawing code to enable multiple OS windows, for GUI support.  Misc
**                  changes/fixes related to drawing on multiple windows.
** 008 CA  20150219 Hiding the ChUI window removes the FOCUS reference (sets it to unknown).  The
**                  ChUI WINDOW's VISIBLE attribute can not be set to FALSE.
** 009 VIG 20150121 Added MENU drawing.
** 010 HC  20150323 Removed business logic from config classes.
** 011 VIG 20150226 Added menu parent setting before drawing method call.
**                  Menu drawing is moved from setVisible to draw method.
** 012 HC  20150508 Improvements to GUI window scrolling support.
** 013 HC  20150622 Fixed GUI frame to be properly centered in its parent window.
** 014 VIG 20150618 Changed menu drawing. Fixed menu layout.
** 015 HC  20150831 Scrolling support overhaul.
** 016 CA  20150921 Added drawCaret().
** 017 VIG 20151020 Added contentPane vertical shift when the window contains MENUBAR.
*/
/*
** This program is free software: you can redistribute it and/or modify
** it under the terms of the GNU Affero General Public License as
** published by the Free Software Foundation, either version 3 of the
** License, or (at your option) any later version.
**
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
** GNU Affero General Public License for more details.
**
** You may find a copy of the GNU Affero GPL version 3 at the following
** location: https://www.gnu.org/licenses/agpl-3.0.en.html
** 
** Additional terms under GNU Affero GPL version 3 section 7:
** 
**   Under Section 7 of the GNU Affero GPL version 3, the following additional
**   terms apply to the works covered under the License.  These additional terms
**   are non-permissive additional terms allowed under Section 7 of the GNU
**   Affero GPL version 3 and may not be removed by you.
** 
**   0. Attribution Requirement.
** 
**     You must preserve all legal notices or author attributions in the covered
**     work or Appropriate Legal Notices displayed by works containing the covered
**     work.  You may not remove from the covered work any author or developer
**     credit already included within the covered work.
** 
**   1. No License To Use Trademarks.
** 
**     This license does not grant any license or rights to use the trademarks
**     Golden Code, FWD, any Golden Code or FWD logo, or any other trademarks
**     of Golden Code Development Corporation. You are not authorized to use the
**     name Golden Code, FWD, or the names of any author or contributor, for
**     publicity purposes without written authorization.
** 
**   2. No Misrepresentation of Affiliation.
** 
**     You may not represent yourself as Golden Code Development Corporation or FWD.
** 
**     You may not represent yourself for publicity purposes as associated with
**     Golden Code Development Corporation, FWD, or any author or contributor to
**     the covered work, without written authorization.
** 
**   3. No Misrepresentation of Source or Origin.
** 
**     You may not represent the covered work as solely your work.  All modified
**     versions of the covered work must be marked in a reasonable way to make it
**     clear that the modified work is not originating from Golden Code Development
**     Corporation or FWD.  All modified versions must contain the notices of
**     attribution required in this license.
*/

package com.goldencode.p2j.ui.client.chui;

import com.goldencode.p2j.ui.*;
import com.goldencode.p2j.ui.chui.*;
import com.goldencode.p2j.ui.client.*;
import com.goldencode.p2j.ui.client.chui.driver.*;
import com.goldencode.p2j.ui.client.widget.*;

@SuppressWarnings("unchecked")
public class WindowChuiImpl
extends Window<ChuiOutputManager>
{
   /** Renders the status of the current insertion mode. */
   StatusLine<ChuiOutputManager> insertModeLabel = null;
   
   /**
    * Hide base class public constructor.
    * 
    * @param    id
    *           The widget's ID.
    */
   public WindowChuiImpl(WidgetId id)
   {
      super(id);

      init();
   }
   
   /**
    * Construct a window widget based on a specific configuration.
    * 
    * @param   config
    *          Window configuration.
    * @param   id
    *          The widget's ID.
    */
   public WindowChuiImpl(WidgetId id, WindowConfig config)
   {
      super(id, config);
   }

   /**
    * Draw caret.
    */
   @Override
   public void drawCaret()
   {
      // drawing the cursor
      Widget<ChuiOutputManager> inFocus = currentFocus();
      if (inFocus instanceof FixedSizeContainer && inFocus.isEnabled())
      {
         ((FixedSizeContainer<ChuiOutputManager>) inFocus).drawCaret();
      }
   }
   
   /**
    * Lay out container.
    */
   @Override
   public void doLayout()
   {
      super.doLayout();
      if (menubar != null)
      {
         menubar.doLayout();
         
         // shift window content because the MENUBAR become visible.
         contentPane.setLocation(new Point(0, 1));
      }
   }
   
   /**                                     
    * Get dimension of screen area available for the placement of frames.
    * 
    * @return  Instance of <code>Dimension</code> which describes size of the
    *          available screen area.
    */
   public Dimension getScreenDimension()
   {
      Dimension sz = dimension();
      return new Dimension(sz.width, sz.height - RESERVED_LINES);
   }
   
   /**
    * Get messages area rectangle by message line index.
    *
    * @param   line
    *          message line specification: either 0-based line number for
    *          specific line or -1 for the whole message area.
    * @return  Instance of <code>Rectangle</code> which describes the message
    *          area in screen coordinates.
    */
   public Rectangle getMessageLineRectangle(int line)
   {
      Dimension sz = dimension();
      double    y0 = sz.height - RESERVED_LINES;
      double    y1 = sz.height;
      double    x1 = sz.width - 1;
      
      if (line < 0)
      {
         y1 -= 2;
      }
      else
      {
         y0 += line;
         y1 = y1 - RESERVED_LINES + line; 
      }
         
      return new Rectangle(y0, 0, y1, x1);
   }

   /**
    * Get status area rectangle.
    *
    * @return  Instance of <code>Rectangle</code> which describes the status
    *          area in screen coordinates.
    */
   public Rectangle getStatusLineRectangle()
   {
      Dimension sz = dimension();
      double    y  = sz.height - 1;
      double    x  = sz.width - 1;
      
      return new Rectangle(y, 0, y, x);
   }
   
   /**
    * Returns the preferred size of the window workspace.   
    * 
    * @return   See above.
    */
   @Override
   public Dimension getScrollableSize()
   {
      Dimension result = screen().screenDimension(getId().asInt());
      result.height -= Window.RESERVED_LINES;
      return result;
   }

   /**
    * When the window is no longer visible, special processing is required.  This includes hiding 
    * all frames and clearing message and status.
    * 
    * @param    visible
    *           The frame's visibility state.
    */
   @Override
   public void setVisible(boolean visible)
   {
      if (visible) 
      {
         _setVisible(visible);

         return;
      }

      // the ChUI window can't have its VISIBLE attribute set to FALSE
      _setVisible(true);
      
      String[] content = getMessageArea().getMessageTexts();

      Widget<ChuiOutputManager> focus = currentFocus();

      ThinClient tc = ThinClient.getInstance();
      
      tc.hideAll(true, UiUtils.getWidgetIdAsInt(this));

      if (focus != null && tc.getTriggerNesting() > 0)
      {
         Frame<ChuiOutputManager> frame = (Frame<ChuiOutputManager>) UiUtils.locateFrame(focus);

         if (frame != null)
            frame.setFocusReset(true);
      }

      // no more focus - is lost forever!
      setFocus(null);

      if (content[0] == null && content[1] == null)
         return;

      // message lines were not empty, so simulate hiding of message lines
      // instead of wiping out their content
      if (content[0] != null)
         content[0] = "";
      if (content[1] != null)
         content[1] = "";

      setMessageText(content, null);
   }
   
   /**
    * Get the size of the window workspace area.
    * 
    * @return   See above.
    */
   @Override
   public Dimension getWorkspaceSize()
   {
      Dimension dim = super.getWorkspaceSize();
      return new Dimension(dim.width, dim.height - Window.RESERVED_LINES);
   }

   /**
    * Reset Window instance state.
    */
   protected void reset()
   {
      setSize(1, 1);

      super.reset();      

      // INSERT indicator is ChUI specific only.
      insertModeLabel = (StatusLine<ChuiOutputManager>) new MessageLine(WidgetId.nextID());      
   }

   /**
    * Initialize window
    */
   protected void init()
   {
      super.init();
      
      add(contentPane);
      setLayout(null);
      
      reset();

      WindowConfig config = config();

      BasePrimitives direct = screen().getDirectOutput();
      
      Dimension screenSize = new Dimension(direct.screenWidth(), direct.screenHeight());
      setSize(screenSize);

      getContentPane().setSize(screenSize);
      
      // Message area with variable lines of the text
      messageArea.attach(this);
      
      // status line with default text
      statusLine.attach(this);

      // "Insert" label 
      getContentPane().add(insertModeLabel);
      insertModeLabel.setLocation(statusLine.width(), screenSize.height - 1);
      insertModeLabel.setText(null);
      insertModeLabel.setLength(INSERT_TEXT.length());
      insertModeLabel.setColor(Color.NORMAL);
      
      // Initialization of the default values for attributes
      // which are dependent from the code in this method
      // Size related attributes
      config.screenLines = Coordinate.scale(screenSize.height - RESERVED_LINES);
      config.fullHeightChars = Coordinate.scale(screenSize.height - 4);
      config.fullWidthChars = Coordinate.scale(screenSize.width);
      config.heightChars = Coordinate.scale(screenSize.height); // direct access
      config.widthChars = Coordinate.scale(screenSize.width); // direct access
      
      isRunning = true;
   }

   /**
    * Set the contents of the status line. This method is private and
    * called at service level by another static public method within the
    * class.  The status line output is truncated to a maximum of 63
    * characters, only if the <code>noLengthLimit</code> flag is not set.
    * 
    * @param   status
    *          The new string to be displayed inside the status line area.
    * @param   noLengthLimit
    *          flag indicating the status message should not be trimmed
    */
   protected void setStatusLineText(String status, boolean noLengthLimit)
   {
      int trimSize = (int) (noLengthLimit ? dimension().width : 63);
      
      // change the text
      String trim = status.length() > trimSize ? status.substring(0, trimSize)
                                               : status;
   
      getContentPane().moveToTop(statusLine);
      statusLine.setText(trim);
      statusLine.repaint();
   }

   /**
    * The private function to set up the global insert mode value.
    * 
    * @param   insMode
    *          New integer value for insert mode.
    * @param   force
    *          Whether we have to repaint the indicator even if the new mode
    *          is the same as the old one.
    */
   protected void setInsMode(int insMode, boolean force)
   {
      if (insMode == insertMode && !force)
         return;
      
      if (insMode == INSERT_MODE_INS)
      {
         if (insertMode == INSERT_MODE_OFF || insertMode == INSERT_MODE_OVR)
            insertModeLabel.setText(INSERT_TEXT);
      }
      else if (insMode == INSERT_MODE_OVR)
      {
         if (insertMode == INSERT_MODE_OFF || insertMode == INSERT_MODE_INS)
            insertModeLabel.setText(null);
      }
      else if (insMode == INSERT_MODE_OFF)
      {
         insertModeLabel.setText(null);
      }
      else
         return;
      
      getContentPane().moveToTop(insertModeLabel);
      
      // save the bitmap, so it can be restored
      ScreenBitmap bitmap = getAndResetBitmap();
      insertModeLabel.draw();
      sync();
      // restore the screen
      setBitmap(bitmap);
      
      insertMode = insMode;
      insertInAmbiguousState = false;
   }
   
   /**
    * Enables / Disables the "special components" (status line, insert mode indicator,
    * message lines, temp text components) and restores visibility to the
    * list of all previously visible top-level components.
    *
    * This is part of the switching of the client to the interactive mode.
    * 
    * @param   visibility
    *          <code>true</code> enable <code>false</code> disable.
    */
   @Override
   protected void setVisibility(boolean visibility)
   {
      super.setVisibility(visibility);
      insertModeLabel.forceVisibility(visibility);
   }   
}