/*
** Module   : ChuiSimulator.java
** Abstract : Swing-based terminal simulator
**
** Copyright (c) 2011-2017, Golden Code Development Corporation.
**
** -#- -I- --Date-- ---------------------------------Description---------------------------------
** 001 GES 20110921 Created initial version.
** 002 GES 20111026 Added safety code to only dispose the frame when it is
**                  displayable.
** 003 EVL 20121119 Adding code to initially select the first available
**                  monospaced font when application starts.
** 004 MAG 20131114 Change this class to became a base for ChuiWebSimulator.
** 005 MAG 20131201 Fix javadoc errors.
** 006 MAG 20140114 Add protected methods used by ChuiWebSimulator.
** 007 MAG 20140211 Improve drawing using double buffered option. 
** 008 MAG 20140220 Eliminate unused method. 
** 009 MAG 20140326 Implements interactive client. 
** 010 MAG 20140508 Implements early support for the GUI screen driver.
** 011 CA  20140805 Moved non-ChUI specific APIs into a base class.
** 012 CA  20141029 Each simulator needs to receive the swing keyboard reader implementation in 
**                  use.  It's the callers responsibility to ensure that the same instance is used
**                  for all simulators, if the client is ran in a multi-window mode (i.e. GUI).
** 013 GES 20150423 Refactored this to remove the JComponent base class and move functionality
**                  into a new generic (not specific to either swing or web) abstract base class.
** 014 OM  20170214 Replaced base class from JFrame to JDialog which suits better our purpose.
*/
/*
** This program is free software: you can redistribute it and/or modify
** it under the terms of the GNU Affero General Public License as
** published by the Free Software Foundation, either version 3 of the
** License, or (at your option) any later version.
**
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
** GNU Affero General Public License for more details.
**
** You may find a copy of the GNU Affero GPL version 3 at the following
** location: https://www.gnu.org/licenses/agpl-3.0.en.html
** 
** Additional terms under GNU Affero GPL version 3 section 7:
** 
**   Under Section 7 of the GNU Affero GPL version 3, the following additional
**   terms apply to the works covered under the License.  These additional terms
**   are non-permissive additional terms allowed under Section 7 of the GNU
**   Affero GPL version 3 and may not be removed by you.
** 
**   0. Attribution Requirement.
** 
**     You must preserve all legal notices or author attributions in the covered
**     work or Appropriate Legal Notices displayed by works containing the covered
**     work.  You may not remove from the covered work any author or developer
**     credit already included within the covered work.
** 
**   1. No License To Use Trademarks.
** 
**     This license does not grant any license or rights to use the trademarks
**     Golden Code, FWD, any Golden Code or FWD logo, or any other trademarks
**     of Golden Code Development Corporation. You are not authorized to use the
**     name Golden Code, FWD, or the names of any author or contributor, for
**     publicity purposes without written authorization.
** 
**   2. No Misrepresentation of Affiliation.
** 
**     You may not represent yourself as Golden Code Development Corporation or FWD.
** 
**     You may not represent yourself for publicity purposes as associated with
**     Golden Code Development Corporation, FWD, or any author or contributor to
**     the covered work, without written authorization.
** 
**   3. No Misrepresentation of Source or Origin.
** 
**     You may not represent the covered work as solely your work.  All modified
**     versions of the covered work must be marked in a reasonable way to make it
**     clear that the modified work is not originating from Golden Code Development
**     Corporation or FWD.  All modified versions must contain the notices of
**     attribution required in this license.
*/

package com.goldencode.p2j.ui.client.chui.driver.swing;

import java.awt.*;
import java.awt.font.*;
import java.util.*;
import com.goldencode.p2j.ui.client.Cell;
import com.goldencode.p2j.ui.client.driver.swing.*;
import com.goldencode.p2j.ui.client.chui.driver.*;

// we import selectively because there are many overlapping class names between
// swing and p2j, since the current class is more swing than p2j we err on the
// side of swing imports

/**
 * Swing-based terminal simulator. This control is the core of a rectangular
 * component that appears to look and act like a character terminal (TTY).
 * <p>
 * All output is done in the same fixed width font family, with character
 * attributes implemented by modifying the attributes of the common font for
 * each output character.  The output stored in a 2-dimensional buffer of
 * of cells, which store the character to be drawn and the attributes/color
 * of that character. 
 * <p>
 * The current cursor position is remembered and can be manipulated by the
 * application. The cursor visibility state can be externally controlled as
 * well. When rendering output, if the cursor is visible, it is drawn by
 * inverting the foreground/background colors of the font used to draw the
 * character at which the cursor is located.   
 * <p>
 * Input is read using key events and that is kept in a FIFO buffer as if
 * they had been sent by the remote terminal hardware.
 * <p>
 * Other common end-user functionality is implemented including printing,
 * selection of text, copy of selected text to the clipboard, paste of text
 * from the clipboard into the keyboard input buffer and copying the entire
 * terminal's visible state to the clipboard as an image.
 * <p>
 * Limitations:
 * <p>
 * <ul>
 *    <li> The BLINK attribute is not supported yet. To implement it, one would
 *         have to keep track (in a carefully synchronized way) of the list of
 *         characters (and the current font being used) that must be blinked
 *         periodically. Then a timer would have to walk that list periodically
 *         and invert the colors (using TextAttribute.SWAP_COLORS).
 *    <li> This class doesn't use Swing's double buffering yet. It may not need
 *         it ever. The redrawing seems pretty clean.
 * </ul>
 */
@SuppressWarnings("serial")
public class ChuiSimulator
extends EmulatedTerminalState
{
   /** Popup menu. */
   private TerminalMenu menu = null;
   
   /** Helper to map color specifications to the font needed for drawing. */
   private ColorMapper colorMap = null;
   
   /** The height of a single character in pixels (same as the row height). */
   private int charHeight = 0;
   
   /** The width of a single character in pixels. */
   private int charWidth = 0;
   
   /** Number of pixels between each row based on font metrics. */
   private int baseLine = 0;
   
   /** Selection start X coordinate in pixels. */
   private int x0 = -1;
   
   /** Selection end X coordinate in pixels. */
   private int x1 = -1;
   
   /** Selection start Y coordinate in pixels. */
   private int y0 = -1;
   
   /** Selection end Y coordinate in pixels. */
   private int y1 = -1;
   
   /** Selection start row number. */
   private int r0 = -1;
   
   /** Selection end row number. */
   private int r1 = -1;
   
   /** Selection start column number. */
   private int c0 = -1;
   
   /** Selection end column number. */
   private int c1 = -1;
   
   /** Flag to control if it is valid to track selection events. */
   private boolean selectionTrack = false;
   
   /** Boundary of the current selected region in pixel coordinates. */ 
   private Rectangle selected = null;
   
   /** Content pane of the containing swing window. */
   private ContentPane pane = null;
   
   /** Color used for the selection rectangle. */
   private Color selectionColor = null;
   
   /**
    * Create a new instance with the container and frame given. The component's
    * size and characteristics will be determined from the options provided.
    * When this method is done, the preferred size of the component will have
    * been established, all listeners and key processing will be initialized
    * and the control will be ready to be made visible by the caller.
    * <p>
    * For applications, the frame and the container are usually the same
    * object (of type <code>JFrame</code>). For applets, these are different
    * instances since a <code>JApplet</code> will be the container and the
    * frame will be a parent container of the <code>JApplet</code>.
    *
    * @param    window
    *           The containing window. Needed for key event processing and for the processing of
    *           modal dialog boxes.
    * @param    container
    *           The immediate container component, which can be called to
    *           force a resizing to match a changed simulator size.
    * @param    options
    *           The terminal configuration (size and appearance).
    */
   public ChuiSimulator(Window window, ContainerOps container, TerminalOptions options)
   {
      super();
      
      SwingKeyboardReader kbr = options.underWindows ? new WinKeyboardReader()
                                                     : new LinuxKeyboardReader();
      
      pane = new ContentPane(window, kbr)
      {
         public void paintComponent(Graphics graphics)
         {
            super.paintComponent(graphics);
            ChuiSimulator.this.paintComponent(graphics);
         }   
      };
      
      if (options.fontname.equalsIgnoreCase("monospaced"))
      {
         options.fontname = SwingHelper.getMonoFontName(pane, options.fontname);
      }
      
      configure(options);
      
      if (container != null)
      {
         // setup popup menus
         menu = new TerminalMenu(window, this, pane, container, options);
         pane.addMouseListener(new PopupTrigger(menu));
         
         // enable selection tracking
         SelectionTracker tracker = new SelectionTracker(this);
         pane.addMouseListener(tracker);
         pane.addMouseMotionListener(tracker);
      }
   }
   
   /**
    * Initialize the instance using the given configuration. This will obtain
    * and initialize its fonts (and the color mapper). It will use the font
    * metrics to calculate the size of the component (based on the number of
    * rows and columns).  If this is the first initialization, the 2-dimensional
    * output buffer will be created and cleared.
    *
    * @param    termopt
    *           The configuration for the terminal simulator.
    */
   public void configure(TerminalOptions termopt)
   {
      super.configure(termopt);
      
      synchronized (cfgLock)
      {
         Color back = SwingHelper.fromColorRgb(options.background);
         Color fore = SwingHelper.fromColorRgb(options.foreground);
         
         selectionColor = SwingHelper.fromColorRgb(options.selection);
         
         // this part doesn't seem to really work (the intention was to have the background
         // automatically draw itself, but in practice this doesn't occur)
         pane.setOpaque(true);
         pane.setBackground(back);
         pane.setForeground(fore);
         
         Map<TextAttribute, Object> fontAttrs = new HashMap<TextAttribute, Object>();
         fontAttrs.put(TextAttribute.FAMILY, options.fontname);
         fontAttrs.put(TextAttribute.SIZE, options.fontsize);
         fontAttrs.put(TextAttribute.FOREGROUND, fore);
         fontAttrs.put(TextAttribute.BACKGROUND, back);
         
         Font font = new Font(fontAttrs);
         
         if (colorMap == null)
         {
            colorMap = new ColorMapper(font);
         }
         else
         {
            colorMap.reset(font);
         }
         
         // save off the "official" font family name
         options.fontname = font.getFamily();
         
         pane.setFont(font);
         
         FontMetrics fm = pane.getFontMetrics(font);
         
         // must reduce this value by 1 otherwise there is a 1 pixel blank line
         // between rows
         charHeight = fm.getHeight() - 1;
         
         // previous attempts to use fm.getMaxAdvance() - 1 did not work for
         // all fonts, this hack seems to work well instead
         charWidth  = fm.charWidth('m');
         
         baseLine   = charHeight - fm.getMaxAscent();
         
         if (!SwingHelper.isMonospaced(font, fm))
         {
            throw new IllegalArgumentException("Font face '" + 
                                               options.fontname +
                                               "' must be fixed width.");
         }
         
         Dimension sz = new Dimension((charWidth * options.columns),
                                      (charHeight * options.rows));
         pane.setPreferredSize(sz);
         pane.setMinimumSize(sz);
         pane.setMaximumSize(sz);
      }
   }
   
   /**
    * Obtain the content pane for the terminal simulator.
    */
   public ContentPane getContentPane()
   {
      return pane;
   }
   
   /**
    * Paint the entire component to the given graphics instance, using
    * previously set state.  Depending on the repaint request, the graphics
    * object's clipping region may have been set to a subset of the full
    * rectangle. This method makes no such distinction, it just blindly
    * draws everything and the stuff outside of the clipping region never
    * gets sent to the hardware.  This is the core where all drawing occurs
    * and it draws the screen text (including switching out the fonts per
    * character as needed to get the right colors/attributes), the current
    * cursor (if it is visible) and the any selection box (when a selection
    * is active).
    *
    * @param    graphics
    *           The object into which to draw.
    */
   public void paintComponent(Graphics graphics)
   {
      int         rows = -1;
      int         cols = -1;
      int         base = -1;
      int         h    = -1;
      int         w    = -1;
      ColorMapper cm   = null;
      
      synchronized (cfgLock)
      {
         rows = options.rows;
         cols = options.columns;
         base = baseLine;
         h    = charHeight;
         w    = charWidth;
         cm   = colorMap;
      }
      
      // save off font
      Font oldFont = graphics.getFont();
      
      synchronized (dataLock)
      {
         // draw each row
         for (int i = 0; i < rows; i++)
         {
            // calculate the Y position (in pixels) for this row
            int y = ((i + 1) * h) - base - 1;
            
            int j = 0;
            int x = 0;
            
            Font last = null;
            Font font = null;
            StringBuilder sb = new StringBuilder();
            
            while (j < cols)
            {
               Cell next = data[i][j];
               char ch   = next.character();
               
               font = cm.toFont(next.color());
               
               if (last != null && last != font)
               {
                  // flush our current text
                  drawString(graphics, sb.toString(), last, x, y);
                  
                  // clear our text buffer
                  sb = new StringBuilder();
                  
                  // recalculate x
                  x = j * w;
               }
               
               // accumulate the text since it is either the first in a
               // new buffer or it shares the same font as the previous text
               sb.append(Character.isISOControl(ch) ? ' ' : ch);
               
               // save off the most recent font
               last = font;
               
               // look at the next char
               j++;
            }
            
            if (sb.length() > 0)
            {
               // output the remaining text in the buffer
               drawString(graphics, sb.toString(), font, x, y);
            }
         }
      }
      
      // draw cursor (if needed)
      if (showCursor)
      {
         StringBuilder cursor = new StringBuilder();
         cursor.append(data[cursorY][cursorX].character());
         int y = ((cursorY + 1) * h) - base - 1;
         int x = cursorX * w;
         
         Font fnt = cm.toFontReverse(data[cursorY][cursorX].color());
         drawString(graphics, cursor.toString(), fnt, x, y);
      }
      
      // restore font
      graphics.setFont(oldFont);
      
      Rectangle area = null;
      
      synchronized (selLock)
      {
         area = selected;
      }
      
      if (area != null)
      {
         Color old = graphics.getColor();
         graphics.setColor(selectionColor);
         graphics.drawRect((int) area.getX(),
                           (int) area.getY(),
                           (int) area.getWidth(),
                           (int) area.getHeight());
         graphics.setColor(old);
      }
   }
   
   /**
    * Cause a partial or full repaint to occur. If <code>full</code> is
    * <code>true</code> or if there is no current selection operation occurring,
    * a repaint of the entire component ("screen") will occur.  Otherwise,
    * a repaint of only the selection area will occur.
    *
    * @param    full
    *           Flag to force a the entire screen to repaint.
    */
   @Override
   public void triggerRepaint(boolean full)
   {
      // calculate the selection box as a rectangle
      Rectangle area = null;
      
      if (selected == null || full)
      {
         area = new Rectangle(pane.getPreferredSize());
      }
      else
      {
         // we must redraw one more pixel in height and width than the actual
         // rectangle, since the drawn rectangle includes 1 extra pixel to the
         // right and below in each dimension
         area = new Rectangle((int) selected.getX(),
                              (int) selected.getY(),
                              (int) selected.getWidth() + 1,
                              (int) selected.getHeight() + 1);
      }
      
      // trigger the painting
      pane.repaint(area);
   }
   
   /**
    * Read the current terminal buffer at each character position inside the
    * current selection and return the text in a string. Each row of the
    * selection (except for the last one) will be terminated with an embedded
    * newline character inside the resulting string.
    *
    * @return   The lines of text on the screen inside the current selection
    *           or an empty string if there is no current selection.
    */
   public String copy()
   {
      StringBuilder sb = new StringBuilder();
      
      int r0 = -1;
      int r1 = -1;
      int c0 = -1;
      int c1 = -1;
      
      synchronized (selLock)
      {
         r0 = this.r0;
         r1 = this.r1;
         c0 = this.c0;
         c1 = this.c1;
      }
      
      synchronized (dataLock)
      {
         for (int i = r0; i <= r1; i++)
         {
            for (int j = c0; j <= c1; j++)
            {
               sb.append(data[i][j].character());
            }
            
            if (i != r1)
               sb.append('\n');
         }
      }
      
      // now clear the current selection (this will recursively call back into
      // the simulator to trigger a cancelSelection() which will clear state)  
      menu.textSelectionDone();
      
      return sb.toString();
   }
   
   /**
    * Control whether the selection tracker is tracking mouse events.
    *
    * @param    track
    *           <code>true</code> to allow tracking, <code>false</code> to
    *           ignore all mouse events (no tracking).
    */
   public void enableSelectionTrack(boolean track)
   {
      synchronized (selLock)
      {
         selectionTrack = track;
      }
   }
   
   /**
    * Report if selection tracking is active.
    *
    * @return   <code>true</code> if selection tracking is active.
    */
   public boolean canSelectionTrack()
   {
      boolean track = false;
      
      synchronized (selLock)
      {
         track = selectionTrack;
      }
      
      return track;
   }
   
   /**
    * Start selection processing based on the given starting coordinates and
    * force a repaint of the selection region which will cause the selection
    * box to be drawn around the character in which the point is found. This
    * will also cause dependent popup menu items to be enabled.
    *
    * @param    x
    *           Most recent ending X coordinate.
    * @param    y
    *           Most recent ending Y coordinate.
    */
   public void startSelection(int x, int y)
   {
      synchronized (selLock)
      {
         x0 = x1 = x;
         y0 = y1 = y;
         selected = calcCharBoundary();
         triggerRepaint(true);
      }
      
      menu.enableCopy(true);
   }
   
   /**
    * Update the selection state based on the latest ending coordinates and
    * force a repaint of the selection region which will cause the selection
    * box to be drawn in its new position and size.
    *
    * @param    x
    *           Most recent ending X coordinate.
    * @param    y
    *           Most recent ending Y coordinate.
    */
   public void updateSelection(int x, int y)
   {
      synchronized (selLock)
      {
         // must be checked before edits are made
         boolean full = isInsideSelectionArea(x, y);
         
         x1 = x;
         y1 = y;
         selected = calcCharBoundary();
         triggerRepaint(full);
      }
   }
   
   /**
    * End the current selection operation, update the final selection state and
    * force a repaint of the selection region which will cause the selection
    * box to be drawn in its final position and size.
    *
    * @param    x
    *           Ending X coordinate.
    * @param    y
    *           Ending Y coordinate.
    */
   public void endSelection(int x, int y)
   {
      synchronized (selLock)
      {
         x1 = x;
         y1 = y;
         selected = calcCharBoundary();
         triggerRepaint(false);
      }
   }
   
   /**
    * Cancel the current selection operation, reset the selection state and
    * force a repaint of the entire screen which will cause the selection
    * box to disappear. This will also cause the dependent popup menu items
    * to be disabled.
    */
   public void cancelSelection()
   {
      synchronized (selLock)
      {
         if (selected != null)
         {
            x0 = x1 = -1;
            y0 = y1 = -1;
            selected       = null;
            selectionTrack = false;
            triggerRepaint(true);
         }
      }
      
      menu.enableCopy(false);
   }
   
   /**
    * Detect if the given point is inside the current selection rectangle.
    *
    * @param    x
    *           X coordinate to check.
    * @param    y
    *           Y coordinate to check.
    *
    * @return   <code>true</code> if the point is inside the selection area.
    */
   private boolean isInsideSelectionArea(int x, int y)
   {
      return (x0 < x1 && x0 < x && x < x1) ||
             (x0 > x1 && x0 > x && x > x1) ||
             (y0 < y1 && y0 < y && y < y1) ||
             (y0 < y1 && y0 < y && y < y1);
   }
   
   /**
    * Helper method to draw the given string into the given graphics object
    * at the given location with the given font.
    *
    * @param    graphics
    *           Instance to draw into.
    * @param    text
    *           Text to draw.
    * @param    font
    *           The common font in which to draw all characters from the string.
    * @param    x
    *           X coordinate.
    * @param    y
    *           Y coordinate.
    */
   private void drawString(Graphics graphics, String text, Font font, int x, int y)
   {
      graphics.setFont(font);
      graphics.drawString(text, x, y);
   }
   
   /**
    * Based on the X and Y coordinates in pixels from a mouse button event
    * (during selection processing), calculate the starting and ending
    * coordinates of the selection rectangle in both character positions
    * (rows/columns) and in pixels. The pixel positions will "snap" to
    * locations that are on the upper left corner of the top leftmost selected
    * character and to the lower right corner of the bottom rightmost selected
    * character. The calculated character boundaries are cached for future
    * use.
    *
    * @return   The rectangle that defines the pixel boundaries of this
    *           area.
    */
   private Rectangle calcCharBoundary()
   {
      Rectangle area = null;
      
      int tx0 = this.x0;
      int ty0 = this.y0;
      int tx1 = this.x1;
      int ty1 = this.y1;
      
      if (tx0 != -1)
      {
         // normalize the values so that (x0, y0) is the upper left and
         // (x1, y1) is the lower right
         int x0 = Math.min(tx0, tx1);
         int x1 = Math.max(tx0, tx1);
         int y0 = Math.min(ty0, ty1);
         int y1 = Math.max(ty0, ty1);
         
         // calculate in which character (row and column) each point exists
         // (note that the use of integer division will naturally leave us at
         // the character boundries)
         r0 = y0 / charHeight;
         c0 = x0 / charWidth;
         r1 = y1 / charHeight;
         c1 = x1 / charWidth;
         
         int x = c0 * charWidth;
         int y = r0 * charHeight;
         int w = ((c1 + 1) - c0) * charWidth;
         int h = ((r1 + 1) - r0) * charHeight;
         
         area = new Rectangle(x, y, w, h);
      }
      
      return area;
   }
}