/*
** Module   : SwingChuiDriver.java
** Abstract : Swing-based ChUI driver
**
** Copyright (c) 2011-2017, Golden Code Development Corporation.
**
** -#- -I- --Date-- -------------------------------Description--------------------------------
** 001 GES 20110914 Created initial version.
** 002 SIY 20111202 Updated imports.
** 003 SIY 20111221 Added widget factory.
** 004 CA  20131004 Implement the WINDOW-SYSTEM attribute for this driver.
** 005 MAG 20131114 Extends ChuiScreenDriver base class.
** 006 CA  20140805 Added generics to force the abstract widget classes to avoid implementation-
**                  specific code: the abstract widgets will have access only to the base APIs
**                  provided by the OutputManager.
** 007 GES 20150331 Moved initialization logic here from OutputManager to separate concerns.
**                  Refactored the code to move swing-specific logic here and derive from a
**                  different base class.  This eliminates swing dependencies from the parent
**                  hierarchy and results in code that is easier to understand and support.
*/
/*
** This program is free software: you can redistribute it and/or modify
** it under the terms of the GNU Affero General Public License as
** published by the Free Software Foundation, either version 3 of the
** License, or (at your option) any later version.
**
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
** GNU Affero General Public License for more details.
**
** You may find a copy of the GNU Affero GPL version 3 at the following
** location: https://www.gnu.org/licenses/agpl-3.0.en.html
** 
** Additional terms under GNU Affero GPL version 3 section 7:
** 
**   Under Section 7 of the GNU Affero GPL version 3, the following additional
**   terms apply to the works covered under the License.  These additional terms
**   are non-permissive additional terms allowed under Section 7 of the GNU
**   Affero GPL version 3 and may not be removed by you.
** 
**   0. Attribution Requirement.
** 
**     You must preserve all legal notices or author attributions in the covered
**     work or Appropriate Legal Notices displayed by works containing the covered
**     work.  You may not remove from the covered work any author or developer
**     credit already included within the covered work.
** 
**   1. No License To Use Trademarks.
** 
**     This license does not grant any license or rights to use the trademarks
**     Golden Code, FWD, any Golden Code or FWD logo, or any other trademarks
**     of Golden Code Development Corporation. You are not authorized to use the
**     name Golden Code, FWD, or the names of any author or contributor, for
**     publicity purposes without written authorization.
** 
**   2. No Misrepresentation of Affiliation.
** 
**     You may not represent yourself as Golden Code Development Corporation or FWD.
** 
**     You may not represent yourself for publicity purposes as associated with
**     Golden Code Development Corporation, FWD, or any author or contributor to
**     the covered work, without written authorization.
** 
**   3. No Misrepresentation of Source or Origin.
** 
**     You may not represent the covered work as solely your work.  All modified
**     versions of the covered work must be marked in a reasonable way to make it
**     clear that the modified work is not originating from Golden Code Development
**     Corporation or FWD.  All modified versions must contain the notices of
**     attribution required in this license.
*/

package com.goldencode.p2j.ui.client.chui.driver.swing;

import com.goldencode.p2j.ui.client.*;
import com.goldencode.p2j.ui.client.chui.driver.*;
import com.goldencode.p2j.ui.client.driver.swing.*;
import com.goldencode.p2j.cfg.*;
import com.goldencode.p2j.util.*;
import com.goldencode.util.*;
import com.goldencode.terminal.*;

/**
 * Swing-based ChUI driver.
 */
public class SwingChuiDriver
extends AbstractChuiDriver
{
   /** Core Swing component that simulates the terminal. */
   private ChuiSimulator sim = null;
   
   /** Content pane of the containing swing window. */
   private ContentPane pane = null;
   
   /** OS platform flag */
   private boolean isWin = false;
   
   /**
    * Constructor.
    *
    * @param   config
    *          The <code>BootstrapConfig</code> configuration.
    */
   public SwingChuiDriver(BootstrapConfig config)
   {
      super(new ChuiPrimitives(ChuiClient.load(convertToOptions(config))), "vt100");
      
      sim   = (ChuiSimulator) ((ChuiPrimitives) direct).getEmulatedTerminalState();
      pane  = sim.getContentPane();
      isWin = PlatformHelper.isUnderWindowsFamily();
   }
   
   /**
    * Provide an audible feedback to user by generating a beep signal.
    * <p>
    * There is no specific requirements to this method and actual implementation
    * may do nothing if there is no way to perform audible notification to the 
    * user.
    */   
   @Override
   public void beep()
   {
      java.awt.Toolkit.getDefaultToolkit().beep();
   }
   
   /**
    * Clear screen.
    * <p>
    * This method should clear the screen and perform necessary flushing of the 
    * system-specific buffers if necessary. Users of this method assume that 
    * actual screen appearance as it is visible to the user is properly changed
    * upon exit from this method.
    */
   @Override
   public void clear()
   {
      sim.clear();
   }

   /**
    * Get driver-specific child process factory.
    * 
    * @return   The reference to the factory.
    */
   @Override
   public ChildProcessFactory getChildProcessFactory()
   {
      return new ChildProcessFactory()
      {
         /**
          * Obtains a new instance of an interactive child process.
          *
          * @return   The new child process instance.
          */
         public InteractiveChildProcess getChildProcess()
         {
            return new VT100SwingEmulator();
         }
      };
   }
   
   /**
    * Read single keystroke from keyboard. Returned code should be either proper
    * Progress 4GL/P2J key code or -1 if environment-specific input cannot be 
    * properly translated.
    * 
    * @return   The pressed key code or -1 if no keys were pressed since last call.
    */
   @Override
   public int readKey()
   {
      return pane.readKey();
   }

   /**
    * Reset current screen mode.
    * <p>
    * Upon exit from this method P2J assumes that screen is in correct state
    * and can be used for interaction with the user.
    */
   @Override
   public void resetMode()
   {
      // NOP for now, what should this do?
   }

    /**
     * Resume normal driver operation. If implementation is terminal-based,
     * they terminal should be acquired again and driver should be ready to 
     * perform regular screen I/O operations.
     */
   @Override
   public void resume()
   {
      // switch keyboard back from VT100, this may be a duplicate call if waitForExit()
      // was already called, otherwise this is critical to occur
      pane.swapKeyboards(false);
   }

   /**
    * Suspend all screen I/O operations. If implementation is terminal-based, 
    * then terminal should be released to allow other application executed 
    * from P2J access terminal.
    */
   @Override
   public void suspend()
   {
      // nothing to do here
   }
   
   /**
    * Set cursor status (ON/OFF).
    * <p>
    * The implementation of this method should show/hide actual screen 
    * cursor visible to user. Remaining code assumes that actual cursor status
    * visible to user is properly changed upon exit from this method.
    * 
    * @param   on
    *          <code>true</code> - show cursor, <code>false</code> - hide cursor.
    * @return  <code>true</code> if operation was successful.
    * 
    * @throws  IllegalStateException
    */
   @Override
   public boolean setCursorStatus(boolean on)
   throws IllegalStateException
   {
      return sim.setCursorStatus(on);
   }
   
   /**
    * Swing-specific implementation of the VT100 emulator child process.
    */
   private class VT100SwingEmulator
   extends VT100EmulatorChildProcess
   {
      /**
       * Create a shell emulator and execute the given command.
       * 
       * @param    cmdline
       *           The command.
       *           
       * @return   A VT100 emulator executing the given command.
       */
      @Override
      public VT100Emulator createShellEmulator(String[] cmdline)
      {
         VT100SwingRenderer term = new VT100SwingRenderer(sim.getScreenData(),
                                                          sim.getRows(),
                                                          sim.getColumns());
         
         VT100SwingKeyboard key = new VT100SwingKeyboard();
         
         pane.setVT100Keyboard(key);
         
         // switch keyboard to VT100 (this can't happen in the suspend() because that has
         // already been called by now and the VT100SwingKeyboard instance was created after
         // suspend() was already called
         pane.swapKeyboards(true);            
   
         // swing clients on Windows runs on a separate window
         int overlay = isWin ? 0 : 1; 
         
         return new VT100Emulator(term, key, cmdline, overlay);
      }
      
      /**
       * Block execution until the interactive child process has exited. This will ensure that
       * any operating system resources are released before returning.
       *
       * @param    pid
       *           The operating system process ID of the child process on which to wait.
       * @param    silent
       *           <code>false</code> for interactive mode (there the terminal is expected to be
       *           shared by the child process and the user is allowed to directly interact with
       *           the child process.
       */
      @Override
      public void waitForExit(int pid, boolean silent)
      throws InterruptedException
      {
         super.waitForExit(pid, silent);
         
         // this path may need to use the 4GL STATUS processing to write the pause message, we
         // need an early switch back of the keyboard from VT100 mode
         pane.swapKeyboards(false);
      }
   }
   
   /**
    * Swing-specific implementation of the VT100 emulator renderer.
    */
   private class VT100SwingRenderer
   extends VT100RendererImpl
   {
      /**
       * Constructor.
       * 
       * @param    data
       *           Video buffer.
       * @param    rows
       *           Number of rows on the screen.
       * @param    cols
       *           Number of columns on the screen.
       */
      public VT100SwingRenderer(Cell[][]data, int rows, int cols)   
      {
         super(data, rows, cols);
      }
      
      /**
       * Ring the terminal bell.
       */
      @Override
      public void beep()
      {
         SwingChuiDriver.this.beep();
      }
      
      /**
       * Set cursor visibility.
       * 
       * @param    showCursor
       *           <code>true</code> to make the cursor visible, <code>false</code> to hide it.
       */
      @Override
      public void setCursorVisible(boolean showCursor)
      {
         sim.setCursorStatus(showCursor);
         updateScreen();
      }
      
      /**
       * Update screen.
       */
      @Override
      protected void updateScreen()
      {
         sim.triggerRepaint(true);
         sim.setCursor(col, row);
      }
   }
}