/*
** Module   : p2j.screen.js
** Abstract : ChUI-specific screen management
**
** Copyright (c) 2014-2017, Golden Code Development Corporation.
**
** -#- -I- --Date-- ------------------------------Description----------------------------------
** 001 MAG 20140110 First version.
** 002 MAG 20140204 Implements text attributes.
** 003 MAG 20140211 Implements COLOR.
** 004 MAG 20140217 Minor changes.
** 005 MAG 20140219 Performances improvements.
** 006 MAG 20140310 Fix font metrics for Windows.
** 006 MAG 20140702 Implements clipboard copy / paste.
** 007 MAG 20140821 Extended color palette.
** 008 GES 20150325 Renamed objects to match the more generic web client approach (not just
**                  ChUI). Moved some logic to common code locations.
** 009 SBI 20150731 Applied "use strict" directive, fixed undeclared variables,
**                  added end-lines on copy.
** 010 GES 20150716 Dynamically create the canvas that is the "default window".
** 011 SBI 20160607 Fixed the font height detection.
*/
/*
** This program is free software: you can redistribute it and/or modify
** it under the terms of the GNU Affero General Public License as
** published by the Free Software Foundation, either version 3 of the
** License, or (at your option) any later version.
**
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
** GNU Affero General Public License for more details.
**
** You may find a copy of the GNU Affero GPL version 3 at the following
** location: https://www.gnu.org/licenses/agpl-3.0.en.html
** 
** Additional terms under GNU Affero GPL version 3 section 7:
** 
**   Under Section 7 of the GNU Affero GPL version 3, the following additional
**   terms apply to the works covered under the License.  These additional terms
**   are non-permissive additional terms allowed under Section 7 of the GNU
**   Affero GPL version 3 and may not be removed by you.
** 
**   0. Attribution Requirement.
** 
**     You must preserve all legal notices or author attributions in the covered
**     work or Appropriate Legal Notices displayed by works containing the covered
**     work.  You may not remove from the covered work any author or developer
**     credit already included within the covered work.
** 
**   1. No License To Use Trademarks.
** 
**     This license does not grant any license or rights to use the trademarks
**     Golden Code, FWD, any Golden Code or FWD logo, or any other trademarks
**     of Golden Code Development Corporation. You are not authorized to use the
**     name Golden Code, FWD, or the names of any author or contributor, for
**     publicity purposes without written authorization.
** 
**   2. No Misrepresentation of Affiliation.
** 
**     You may not represent yourself as Golden Code Development Corporation or FWD.
** 
**     You may not represent yourself for publicity purposes as associated with
**     Golden Code Development Corporation, FWD, or any author or contributor to
**     the covered work, without written authorization.
** 
**   3. No Misrepresentation of Source or Origin.
** 
**     You may not represent the covered work as solely your work.  All modified
**     versions of the covered work must be marked in a reasonable way to make it
**     clear that the modified work is not originating from Golden Code Development
**     Corporation or FWD.  All modified versions must contain the notices of
**     attribution required in this license.
*/

"use strict";

p2j.screen = (function()
{
   /** Screen object which will be returned from this initialization function. */
   var me = {};

   /** Main drawing canvas. */
   var canvas;

   /** Canvas 2D graphics context. */
   var ctx;

   /** Normal canvas font. */
   var font, bold;

   /** Bold canvas font. */
   var bold;
   
   /** Rows on the screen. */
   var rows;

   /** Columns on the screen. */
   var cols;
   
   /** Font width. */
   var dx;

   /** Font height. */
   var dy;

   /** Color palette. */
   me.palette;

   /**
    * Module   : video
    * Abstract : screen video memory
    */
   var video = (function()
   {
      /** Video memory object which will be returned from this initialization function. */
      var me = {};

      /** Internal video storage. */
      var memo;

      /**
       * Initialize video memory.
       * 
       * @param {number} rows    total number of rows
       * @param {number} cols    total number of columns
       * @param {number} color   color pair FG/BG
       */
      me.init = function(rows, cols, color)
      {
         me.color = color;
         memo = new Array(rows);
         for (var i = 0; i < rows; i++)
         {
            memo[i] = new Array(cols);
            for (var j = 0; j < cols; j++)
            {
               var cell = {'c': ' ', 'f': 'white', 'b': 'black', 'a': 0, 'k': 0};
               cell.f = color.f;
               cell.b = color.b;
               memo[i][j] = cell;
            }
         }
      };

      /**
       * Clear video memory.
       * 
       * @param {number} rows total number of rows
       * @param {number} cols total number of columns
       */
      me.clear = function(rows, cols)
      {
         for (var i = 0; i < rows; i++)
         {
            for (var j = 0; j < cols; j++)
            {
               var v = memo[i][j];
               v.c = ' ';
               v.a = 0;
               v.k = 0;
               v.f = me.color.f;
               v.b = me.color.b;
            }
         }
      };

      /**
       * Get cell value.
       * 
       * @param {number} row  row number
       * @param {number} col  column number
       * 
       * @returns {object} cell value.
       */
      me.get = function(row, col)
      {
         return memo[row][col];
      };

      /**
       * Set cell value.
       * 
       * @param {char}   c   character
       * @param {number} row row 
       * @param {number} col column
       * @param {number} a   attributes
       * @param {number} k   color
       * @param {object} p   color pair FG/BG
       */
      me.put = function(c, row, col, a, k, p)
      {
         var v = memo[row][col];
         v.c = c;
         v.a = a;
         v.k = k;
         v.f = p.f;
         v.b = p.b;
      };

      /**
       * Copy selected area into clipboard.
       * 
       * @param {number} begRow start row
       * @param {number} endRow end row
       * @param {number} begCol start column
       * @param {number} endCol end column
       * 
       * @returns {object} selected text.
       */
      me.copy = function(begRow, endRow, begCol, endCol)
      {
         var text = "";
         for (var i = begRow; i <= endRow; i++)
         {
            for (var j = begCol; j <= endCol; j++)
            {
               text += memo[i][j].c;
            }
            if (i < endRow) {
               text += "\n";//add end of line
            }
         }
         return text;
      };

      return me;
   })();

   /**
    * Module   : cursor 
    * Abstract : screen cursor management
    */
   var cursor = (function(screen)
   {
      /** Cursor object which will be returned from this initialization function. */
      var me = {};

      /**
       * Hide cursor.
       */
      me.hide = function()
      {
         screen.drawCursor(false);
      };

      /**
       * Show cursor.
       */
      me.show = function()
      {
         screen.drawCursor(me.visible);
      };

      /**
       * Set cursor position.
       * 
       * @param {number} row row number
       * @param {number} col column number
       */
      me.setPosition = function(row, col)
      {
         me.hide();
         me.row = row;
         me.col = col;
         me.show();
      };

      /**
       * Module   : blinking 
       * Abstract : screen cursor blinking management
       */
      me.blinking = (function(cursor, screen)
      {
         /** Cursor blink object which will be returned from this initialization function. */
         var me = {};

         /** timer */
         var timer;

         /**
          * Blinking off.
          */
         me.off = function()
         {
            window.clearInterval(timer);
         };

         /**
          * Blinking on.
          */
         me.on = function()
         {
            var blink = true;
            timer = window.setInterval(function()
            {
               screen.drawCursor(blink & cursor.visible);
               blink = !blink;
            }, 500);
         };

         return me;
      }(me, screen));

      /**
       * Initialize cursor.
       * 
       * @param {object} cursor attributes.
       */
      me.init = function(cursor)
      {
         me.row = 0;
         me.col = 0;
         me.type = cursor.type;
         me.visible = cursor.visible;
         if (cursor.blinking)
         {
            me.blinking.on();
         }
      };

      return me;
   })(me);

   /**
    * Draw cursor.
    * 
    * @param {boolean} on  true  = show cursor
    *                      false = hide cursor
    */
   me.drawCursor = function(on)
   {
      var v = video.get(cursor.row, cursor.col);
      var p = getColors(v.k);
      var x = cursor.col * dx;
      var y = cursor.row * dy;

      if (on)
      {
         if (cursor.type === 'solid')
         {
            p.f = video.color.b;
            p.b = video.color.f;
            drawText(x, y, v.c, 0, p);
         }
         else
         {
            drawText(x, y, v.c, v.a, p);
            // draw line cursor           
            ctx.fillStyle = video.color.f;
            ctx.fillRect(x, y + dy - 4, dx, 4);
         }
      }
      else
      {
         drawText(x, y, v.c, v.a, p);         
      }      
   };

   /**
    * Draw text on canvas.
    * 
    * @param {number} x      x coordinate
    * @param {number} y      y coordinate
    * @param {string} text   text to draw
    * @param {number} a      attribute
    * @param {object} v      color pairs
    */
   var drawText = function(x, y, text, a, v)
   {
      var max = dx * text.length;

      ctx.save();
      ctx.beginPath();
      ctx.rect(x, y, max, dy);
      ctx.clip();

      ctx.font = (a & 0x02) ? bold : font;
      ctx.textBaseline = 'top';

      /** reverse */
      if (a & 0x08)
      {
         ctx.fillStyle = v.f;
         ctx.fillRect(x, y, max, dy);
         ctx.fillStyle = v.b;
      }
      else
      {
         ctx.fillStyle = v.b;
         ctx.fillRect(x, y, max, dy);
         ctx.fillStyle = v.f;
      }
      
      ctx.fillText(text, x, y, max);

      // underline
      if (a & 0x04)
      {
         ctx.fillRect(x, y + dy - 1, max, 1);
      }

      ctx.restore();                  
   };

   /**
    * Print an error message on the last line.
    * 
    * @param {string} text   text to print.
    */
   me.error = function(text)
   {
      var y = (rows - 1) * dy;

      var v = {};
      v.f = video.color.f;
      v.b = video.color.b;

      drawText(0, y, text, 0, v);

      p2j.sound.beep();
   };

   /**
    * Initialize module.
    *  
    * @param {object} cfg configuration.
    */
   me.init = function(cfg)
   {
      if (cfg.font.name === undefined || cfg.font.name.toLowerCase() === 'monospaced')
      {
         cfg.font.name = 'monospace';
      }
      
      rows = cfg.canvas.size.rows;
      cols = cfg.canvas.size.cols;
      
      video.init(rows, cols, cfg.font.color);
      
      var settings = 
      {
         id : "pwindow_1",
         tabindex : 1,
         style : { zIndex : 0, position : "absolute" }
      };

      canvas = p2j.createCanvas(null, settings);
      ctx    = canvas.getContext('2d');
      
      font = cfg.font.size + 'px ' + cfg.font.name;
      bold = 'bold ' + font;
      
      ctx.font = font;
      
      dx = ctx.measureText('\u2500').width;
      // closest integer
      dx = (dx + 0.5) | 0;
      
      dy = fontMetrics.cellHeight(font);
      
      var vertical = fontMetrics.calculatePureTextHeight(font, '\u2502');
      var offset = 0;
      if (dy > vertical)
      {
         offset = dy - vertical;
      }
      // to cover dy with vertical lines it is required to draw two '\u2502' chars 
      me.verticalOffset = offset;
      
      canvas.width  = cols * dx;
      canvas.height = rows * dy;
      
      ctx.fillStyle = cfg.font.color.b;
      ctx.fillRect(0, 0, canvas.width, canvas.height);
      
      cursor.init(cfg.cursor);
      
      if (cfg.canvas.mouse)
      {
         // mouse click
         canvas.addEventListener('click', function(evt)
         {
            var rect = canvas.getBoundingClientRect();
            var x = evt.clientX - rect.left;
            var y = evt.clientY - rect.top;
            var c = (x / dx) | 0;
            var r = (y / dy) | 0;
            if ((r >= 0 && r < rows) && (c >= 0 && c < cols))
            {
               cursor.setPosition(r, c);
            }
         }, true);
      }
      
      // export some values
      me.canvasWidth  = canvas.width;
      me.canvasHeight = canvas.height;
      me.charWidth  = dx;
      me.charHeight = dy;
      
      // container center aligned
      document.getElementById(cfg.container).style.width  = me.canvasWidth  + 'px';
      
      canvas.focus();
   };

   /**
    * Set cursor position.
    * 
    * @param {number} row   row number
    * @param {number} col   column number
    */
   me.setCursorPosition = function(row, col)
   {
      cursor.setPosition(row, col);
   };

   /**
    * Set cursor visibility
    * 
    * @param {boolean} visible   true  = visible
    *                            false = hidden
    */
   me.setCursorStatus = function(visible)
   {
      cursor.visible = visible;
      cursor.show();
   };

   /**
    * Clear the screen.
    */
   me.clear = function()
   {
      video.clear(rows, cols);
      ctx.fillStyle = font.color.b;
      ctx.fillRect(0, 0, canvas.width, canvas.height);
      cursor.setPosition(0, 0);
   };

   /**
    * Get color pair from cell color.
    *  
    * @param {number} k    cell color
    * 
    * @returns {object} color pair.
    */
   var getColors = function(k)
   {
      var p = {};

      if (k === 0)
      {
         p.f = video.color.f;
         p.b = video.color.b;
      }
      else
      {
         // check for colors
         var fg = ((k >>> 4) & 0x0f) - 1;
         var bg = (k & 0x0f) - 1;

         p.f = me.palette[fg];
         p.b = me.palette[bg];
      }

      return p;
   };

   /**
    * Draw rectangles.
    *  
    * @param {array} rects  an array of rectangles
    */
   me.drawRectangles = function(rects)
   {
      var off = 1;

      while (off < rects.length)
      {
         // package length
         var len = (rects[off] << 8) | rects[off+1];
         me.drawRectangle(rects.subarray(off, off + len));
         off += len;
      }
   };

   /**
    * Draw a rectangle.
    * 
    * @param {array} rect   rectangle to draw
    */
   me.drawRectangle = function(rect)   
   {
      var off = 8; // header size
      // 0 and 1 are used for packege length. 
      var w = rect[2]; // width
      var h = rect[3]; // height
      var row = rect[4]; // row
      var col = rect[5]; // col
      var y = row * dy;
      var x = col * dx;
      var a = rect[6]; // attributes
      var k = rect[7]; // color
      var rx = w * dx; // rectangle width in pixels
      var ry = h * dy; // rectangle height in pixels
      var ls = w * 2;  // text line size

      // colors     
      var p = getColors(k);

      ctx.save();
      ctx.beginPath();
      ctx.rect(x, y, rx, ry);
      ctx.clip();

      ctx.fillStyle = (a & 0x08) ? p.f : p.b;
      ctx.fillRect(x, y, rx, ry);
      // prepare for text draw
      ctx.font = (a & 0x02) ? bold : font;
      ctx.fillStyle = (a & 0x08) ? p.b : p.f;
      ctx.textBaseline = 'top';
      
      // draw text line by line
      for (var i = 0;i < h; i++)
      {
         me.drawRectangleLine(row + i, col, rect, off, ls, a, k , p, y);
         off += ls;
      }            
      
      ctx.restore();                  
   };

   /**
    * Draw a line within rectangle.
    * 
    * @param {number} row    row
    * @param {number} col    column
    * @param {array}  line   text
    * @param {number} off    offset
    * @param {number} len    line size
    * @param {number} a      attributes
    * @param {number} k      color
    * @param {object} p      color pair FG/BG
    */
   me.drawRectangleLine = function(row, col, line, off, len, a, k, p)
   {
      var j = col;
      
      for (var i = off; i < off + len; i += 2)
      {
         var c = String.fromCharCode(line[i] << 8 | line[i + 1]);
         me.drawCell(c, row, j);
         video.put(c, row, j++, a, k, p);
      }      

      var y = row * dy;
      var x = col * dx;
      var size = dx * len / 2;

      // underline
      if (a & 0x04)
      {
         ctx.fillRect(x, y + dy - 1, size, 1);
      }      
   };

   /**
    * Draw character in a cell.
    * 
    * @param {tring}  text charcater to draw
    * @param {number} row  row number
    * @param {number} col  columns number
    */
   me.drawCell = function(text, row, col)
   {
      var y = row * dy;
      var x = col * dx;
      
      ctx.save();
      ctx.beginPath();
      ctx.rect(x, y - me.verticalOffset, dx, dy + me.verticalOffset);
      ctx.clip();
      // TODO to separate border drawings
      // to fix vertical border drawings 
      if (me.verticalOffset > 0 && text.indexOf('\u2502') >= 0)
      {
         ctx.fillText('\u2502', x, y - me.verticalOffset);
         ctx.fillText('\u2502', x, y + me.verticalOffset);
      }
      ctx.fillText(text, x, y, dx);
      
      ctx.restore();
   };
   
   /**
    * Module   : fontMetrics
    * Abstract : measure font height. The module to detect a char height uses mean and variance
    * approach because antialiasing transforms pixels so that they become colored in gray colors.
    * But detecting a char height requires to take into account only valuable pixels.
    * As an example, the vertical char '\u2502' has only 10 valuable pixels for '12px monospace'
    * in Safari but the other pixels are faded. Otherwise, in Firefox '\u2502' has 13 valuable
    * pixels for the same font.
    */
   var fontMetrics = (function()
   {
      /** Font metrics object which will be returned from this initialization function. */
      var me = {};

      /** Canvas height. */
      var height = 128;
      
      /** Canvas width. */
      var width  = 128;

      /** Canvas use to measure metrics. */
      var fontCanvas = document.createElement('canvas');

      fontCanvas.height = height;
      fontCanvas.width  = width;
      
      /** 2D graphical context. */
      var fontCtx = fontCanvas.getContext('2d');
      
      /**
       * Get font height.
       * 
       * @param {string} font
       * 
       * @returns {number} font heigth.
       */
      me.cellHeight = function(font)
      {
         // character height
         me.charHeight = calculatePureTextHeight(font, '\u2503Eghy');
         
         return me.charHeight;
      };
      
      /**
       * Calculates the pure text height for the given font without antialiasing.
       * 
       * @param    {String}
       *           The string font representation
       * @param    {String}
       *           The text to measure its height.
       * 
       * @return   The height in pixels that is occupied by the given text for the provided font
       *           without antialiasing.
       */
      function calculatePureTextHeight(font, text)
      {
         fontCtx.clearRect(0, 0, width, height);
         fontCtx.textBaseline = 'top';
         fontCtx.fillStyle = 'white';
         fontCtx.font = font;
         fontCtx.fillText(text, 0, 0);

         var image  = fontCtx.getImageData(0, 0, width, height);
         var d = calculatePixelsDistribution(image, energy);
         var s = calculateMeanAndVariance(d, 0, height, 0);
         
         return findValuableDomain(d, 0, height, s[0], s[1], 0);
      };
      
      /**
       * Calculates the energy function over the RGBA pixels domain. It is greater than
       * I = (R + G + B) / 3.
       * 
       * @param    {Uint8ClampedArray} data
       *           The image data represented by the array of sequential RGBA values for
       *           each data pixel.
       * @param    {Integer} p
       *           The pixel's position within the image data
       * 
       * @return   The energy of a given pixel.
       */
      function energy(data, p)
      {
         return Math.sqrt(data[p] * data[p] + data[p + 1] * data[p + 1] + data[p + 2] * data[p + 2]);
      };

      /**
       * Calculate pixels horizontal distribution for the given image
       * 
       * @param    {ImageData} img
       *           The provided image to test.
       * @param    {Function} energy
       *           The pixel's energy function to calculate the pixel's intense.
       * 
       * @return   The array built from the average of the provided energy function along horizontals.
       */
      function calculatePixelsDistribution(img, energy)
      {
         var awidth  = img.width;
         var aheight = img.height;
         var data = img.data;
         var x = 0;
         var y = 0;
         var w = 0;
         var distribution = [];
         var count = 0;
         var pixelsInBytes = awidth * aheight * 4;
         for (var p = 0; p < pixelsInBytes;)
         {
            if (distribution[y] === undefined)
            {
               distribution[y] = 0;
            }
            w = energy(data, p);
            if (w > 0)
            {
               count++;
               distribution[y] += w;
            }
            x++;
            if (x === width)
            {
               distribution[y] = distribution[y] / (count > 0 ? count : 1);
               y++;
               x = 0;
               count = 0;
            }
            p += 4; 
         };
         return distribution;
      }
      
      /**
       * Calculate the mean and the variance for the target distribution along the provided
       * segment and the given threshold.
       * 
       * @param    {Array} d
       *           The target distribution.
       * @param    {Integer} i1
       *           The lower bound of the distribution domain.
       * @param    {Integer} i2
       *           The upper bound of the distribution domain.
       * @param    {Number} threshold
       *           Defines the strict lower bound for the distribution values.
       * 
       * @return   The 2-elements array filled with the mean and the variance.
       */
      function calculateMeanAndVariance(d, i1, i2, threshold)
      {
         var n = 0;
         var mean = 0;
         var M2 = 0;
         var delta = 0;
         for (var i = i1; i < i2; i++)
         {
            if (d[i] > threshold)
            {
               n++;
               delta = d[i] - mean;
               mean += delta / n;
               M2 += delta * (d[i] - mean);
            }
         }
         var v;
         if (n < 2)
         {
            v = undefined;
         }
         else
         {
            v = M2 / (n - 1);
         }
         
         return [mean, v];
      }
      
      /**
       * Defines the valuable domain for this distribution taken into account its mean and 
       * its variance throwing out values that are less or equal than the provided threshold.
       * 
       * @param    {Array} d
       *           The target distribution.
       * @param    {Integer} i1
       *           The lower bound of the distribution domain.
       * @param    {Integer} i2
       *           The upper bound of the distribution domain.
       * @param    {Number} mean
       *           The mean of the given distribution d.
       * @param    {Number} variance
       *           The variance of the given distribution d.
       * @param    {Number} threshold
       *           Defines the strict lower bound for the distribution values.
       * 
       * @return   The valuable domain width for the target distribution.
       */
      function findValuableDomain(d, i1, i2, mean, variance, threshold)
      {
         var beg = -1;
         for (var i = i1; i < i2; i++)
         {
            if (d[i] > threshold && ((d[i] - mean) * (d[i] - mean)) <= 9 * variance)
            {
               beg = i;
               break;
            }
         }
         var end = -1;
         for (var i = i2 - 1; i >= i1; i--)
         {
            if (d[i] > threshold && ((d[i] - mean) * (d[i] - mean)) <= 9 * variance)
            {
               end = i;
               break;
            }
         }
         return end - beg + 1;
      }
      
      me.calculatePureTextHeight = calculatePureTextHeight;
      
      return me;
   })();

   /**
    * Copy selected area into clipboard.
    * 
    * @param {number} begRow start row
    * @param {number} endRow end row
    * @param {number} begCol start column
    * @param {number} endCol end column
    * 
    * @returns {object} selected text.
   */
   me.copy = function(begRow, endRow, begCol, endCol)
   {
      return video.copy(begRow, endRow, begCol, endCol);
   };

   return me;
}());