/*
** Module   : WidgetFactory.java
** Abstract :
**
** Copyright (c) 2011-2017, Golden Code Development Corporation.
**
** -#- -I- --Date-- ---------------------------------Description---------------------------------
** 001 SIY 20111221 Created initial version
** 002 MAG 20140410 Add RECTANGLE widget creation.
** 003 EVL 20140610 Added TOGGLE-BOX widget support.
** 004 GES 20140625 Minor cleanup.
** 005 CA  20140711 Added MESSAGE-AREA widget support.
** 006 CA  20140805 Added generics to force the abstract widget classes to not use implementation-
**                  specific code: the abstract widgets will have access only to the base APIs
**                  provided by the OutputManager.
**                  Added support for IMAGE widget creation.
** 007 CA  20140926 Added shared widget configuration support, to allow GUI/ChUI concrete
**                  implementation for the same widget ID.  Refactored the widget configuration
**                  classes: all fields were made public; these classes need to be as dumb as
**                  possible, all logic related to setting/getting a certain field should be at
**                  the widget or at the caller. Refs #2254
** 008 MAG 20141009 Deleted createMessageLine.
** 009 EVL 20141118 Adding scroll container creation method.
** 010 CA  20150323 Pass the frame ID when creating a label (to properly resolve the font in GUI).
** 011 EVL 20150428 Adding drop-down container creation.
** 012 CA  20150509 Allow scroll container for any widgets, not just frames.
** 013 HC  20150612 Implemented GUI ALERT-BOX.
** 014 HC  20150831 Scrolling support overhaul.
** 015 IAS 20160217 Added new flavor for the createScrollPane method.
** 016 IAS 20160315 Support for the DEBUG-ALERT attribute
** 017 CA  20161101 Added createMnemonic().
** 017 OM  20170815 Added createFileSystemChooserDialog() method.
*/

/*
** This program is free software: you can redistribute it and/or modify
** it under the terms of the GNU Affero General Public License as
** published by the Free Software Foundation, either version 3 of the
** License, or (at your option) any later version.
**
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
** GNU Affero General Public License for more details.
**
** You may find a copy of the GNU Affero GPL version 3 at the following
** location: https://www.gnu.org/licenses/agpl-3.0.en.html
** 
** Additional terms under GNU Affero GPL version 3 section 7:
** 
**   Under Section 7 of the GNU Affero GPL version 3, the following additional
**   terms apply to the works covered under the License.  These additional terms
**   are non-permissive additional terms allowed under Section 7 of the GNU
**   Affero GPL version 3 and may not be removed by you.
** 
**   0. Attribution Requirement.
** 
**     You must preserve all legal notices or author attributions in the covered
**     work or Appropriate Legal Notices displayed by works containing the covered
**     work.  You may not remove from the covered work any author or developer
**     credit already included within the covered work.
** 
**   1. No License To Use Trademarks.
** 
**     This license does not grant any license or rights to use the trademarks
**     Golden Code, FWD, any Golden Code or FWD logo, or any other trademarks
**     of Golden Code Development Corporation. You are not authorized to use the
**     name Golden Code, FWD, or the names of any author or contributor, for
**     publicity purposes without written authorization.
** 
**   2. No Misrepresentation of Affiliation.
** 
**     You may not represent yourself as Golden Code Development Corporation or FWD.
** 
**     You may not represent yourself for publicity purposes as associated with
**     Golden Code Development Corporation, FWD, or any author or contributor to
**     the covered work, without written authorization.
** 
**   3. No Misrepresentation of Source or Origin.
** 
**     You may not represent the covered work as solely your work.  All modified
**     versions of the covered work must be marked in a reasonable way to make it
**     clear that the modified work is not originating from Golden Code Development
**     Corporation or FWD.  All modified versions must contain the notices of
**     attribution required in this license.
*/

package com.goldencode.p2j.ui.client.driver;

import java.util.function.*;
import com.goldencode.p2j.ui.*;
import com.goldencode.p2j.ui.client.*;
import com.goldencode.p2j.ui.client.border.*;
import com.goldencode.p2j.ui.client.layout.*;
import com.goldencode.p2j.ui.client.model.*;
import com.goldencode.p2j.ui.client.widget.*;
import com.goldencode.p2j.util.*;

/**
 * This interface provides access to driver-specific widget factory necessary to 
 * hide from client-level code implementations of driver-specific widgets.
 */
public interface WidgetFactory<O extends OutputManager<?>>
{
   /**
    * Create an object accumulating prepared label and mnemonic info.
    * 
    * @return   new {@link MnemonicInfo} instance.
    */
   MnemonicInfo createMnemonic();
   
   /**
    * Create new {@link Label} instance for given text.
    * 
    * @param   id
    *          The widget's ID.
    * @param   text
    *          Label text. 
    * @param   frameId
    *          The ID of the frame to which the label belongs, or -1 if unknown at the time of the 
    *          creation.
    * 
    * @return  new {@link Label} instance.
    */
   Label<O> createLabel(WidgetId id, String text, int frameId);

   /**
    * Create new button for given container and text.
    * 
    * @param   id
    *          The widget's ID.
    * @param   container
    *          Container to which new instance will belong.
    * @param   text
    *          Button text.
    *          
    * @return  new {@link Button} instance.
    */
   Button<O> createButton(WidgetId id, Container<O> container, String text);

   /**
    * Create new toggle-box for given container and text.
    * 
    * @param   id
    *          The widget's ID.
    * @param   container
    *          Container to which new instance will belong.
    * @param   text
    *          ToggleBox label text.
    *          
    * @return  new {@link ToggleBox} instance.
    */
   ToggleBox<O> createToggleBox(WidgetId id, Container<O> container, String text);

   /**
    * Create new rectangle for given container.
    * 
    * @param   id
    *          The widget's ID.
    * @param   container
    *          Container to which new instance will belong.
    *          
    * @return  new {@link Rect} instance.
    */
   Rect<O> createRectangle(WidgetId id, Container<O> container);

   /**
    * Instantiate UI-specific version of the specified widget class.
    * 
    * @param   <T>
    *          Widget abstract class type.
    * @param   clazz
    *          Widget abstract class for which UI-specific implementation should
    *          be instantiated.
    * @param   id
    *          The widget's ID.
    * @param   cfg
    *          Widget configuration container.
    *          
    * @return  new UI-specific instance of the requested class. 
    */
   <T extends Widget<O>> T create(Class<T> clazz, WidgetId id, WidgetConfig cfg);

   /**
    * Create UI-specific {@link ScrollableList} implementation.
    * 
    * @param   parent
    *          Parent {@link ComboBox} widget.
    * @param   model
    *          Scrollable list model.
    *          
    * @return  instance of UI-specific implementation of {@link ScrollableList}.
    */
   ScrollableList<O> createScrollableList(ComboBox<O> parent, ListModel<String> model);

   /**
    * Create UI-specific {@link DropDown} implementation.
    * 
    * @param   parent
    *          Parent {@link ComboBox} widget.
    *          
    * @return  instance of UI-specific implementation of {@link DropDown}.
    */
   DropDown<O> createDropDown(ComboBox<O> parent);

   /**
    * Create UI-specific {@link SensitiveScrollContainer} implementation.
    *          
    * @return  instance of UI-specific implementation of {@link SensitiveScrollContainer}.
    */
   SensitiveScrollContainer<O> createScrollContainer();

   /**
    * Create UI-specific {@link ScrollPane} implementation.
    * 
    * @param   widget 
    *          Widget which will be scrolled.
    *          
    * @return  instance of UI-specific implementation of {@link ScrollPane}.
    */
   ScrollPane<O> createScrollPane(ScrollableWidget<O> widget); 

   /**
    * Create UI-specific {@link ScrollPane} implementation.
    * 
    * @param   widget 
    *          Widget which will be scrolled.
    *          
    * @return  instance of UI-specific implementation of {@link ScrollPane}.
    */
   ScrollPane<O> createScrollPane(ScrollableWidget<O> widget, 
         Supplier<TopLevelWindow<O>> wnd);
   
   /**
    * Create UI-specific {@link Label} instance as a copy of given {@link Label}.
    * 
    * @param   id
    *          The widget's ID.
    * @param   labelInstance
    *          Source {@link Label} instance.
    * @param   frameId
    *          The ID of the frame to which the label belongs, or -1 if unknown at the time of the 
    *          creation.
    *          
    * @return  instance of UI-specific implementation of {@link Label}.
    */
   Label<O> createLabel(WidgetId id, Label<O> labelInstance, int frameId);
   
   /**
    * Create UI-specific {@link RadioButton} instance.
    * 
    * @param   config
    *          Parent config. 
    * @param   label
    *          Radio button label. 
    * @param   value
    *          Radio button data type. 
    * 
    * @return  instance of UI-specific implementation of {@link RadioButton}.
    */
   RadioButton<O> createRadioButton(RadioSetConfig config, String label, BaseDataType value);

   /**
    * Create UI-specific {@link Window} instance.
    * 
    * @param   id
    *          The widget's ID.
    * 
    * @return  instance of UI-specific implementation of {@link Window}.
    */
   Window<O> createWindow(WidgetId id);

   /**
    * Create UI-specific {@link MessageArea} instance.
    * 
    * @param   id
    *          The widget's ID.
    * @param   lines
    *          Number of visible lines in message area.         
    * 
    * @return  instance of UI-specific implementation of {@link MessageArea}.
    */
   MessageArea<O> createMessageArea(WidgetId id, int lines);
   
   /**
    * Create UI-specific {@link StatusLine} instance.
    * 
    * @param   id
    *          The widget's ID.
    * 
    * @return  instance of UI-specific implementation of {@link StatusLine}.
    */
   StatusLine<O> createStatusLine(WidgetId id);

   /**
    * Create UI-specific {@link Image} instance.
    * 
    * @param   id
    *          The widget's ID.
    * 
    * @return  instance of UI-specific implementation of {@link Image}.
    */
   Image<O> createImage(WidgetId id);
   
   /**
    * Create UI-specific {@link AlertBox} instance.
    * 
    * @param    alertType
    *           One of the ALERT_* constants specifying message type.
    * @param    content
    *           The list of message text components.
    * @param    buttonSet
    *           One of the BTN_* constants specifying used button set.
    * @param    var
    *           The variable to update if not <code>null</code>.
    * @param    title
    *           The dialog title.
    * @param    color
    *           Color scheme to use.
    * @param    stack
    *           Progress stack trace
    *
    * @return   instance of UI-specific implementation of {@link MessageArea}.
    */
   AlertBox<O> createAlertBox(int      alertType,
                              Object[] content,
                              int      buttonSet,
                              logical  var,
                              String   title,
                              Color    color,
                              String[] stack);
   
   /**
    * Creates new instance of {@link BorderedPanel} with {@link BorderLayout} as
    * the layout manager and no border.
    * 
    * @return  instance of UI-specific implementation of {@link BorderedPanel}.
    */
   BorderedPanel<O> createBorderedPanel();
      
   /**
    * Creates new instance of {@link BorderedPanel} with {@link BorderLayout}
    * layout manager and simple border of width 1 measured in native units. 
    * 
    * @return  instance of UI-specific implementation of {@link BorderedPanel}.
    */
   BorderedPanel<O> createLineBorderedPanel();
   
   /**
    * Creates new instance of {@link BorderedPanel} with a specific layout
    * manager and border.
    * 
    * @param   mgr
    *          Layout manager.
    * @param   border
    *          Border for this widget.
    * 
    * @return  instance of UI-specific implementation of {@link BorderedPanel}.
    */
   BorderedPanel<O> createBorderedPanel(LayoutManager<O> mgr, Border<O> border);
   
   /**
    * Creates new instance representing a default line border.
    * 
    * @return  instance of UI-specific implementation of {@link Border}.
    */
   Border<O> createLineBorder();
   
   /**
    * Creates a new instance of a dialog that allows the user to navigate and choose a 
    * file-system entity.
    * 
    * @param   title
    *          The title of the dialog. Optional. If {@code null} the dialog will use its default
    *          window title. 
    * @param   parentWindow
    *          The id of the window to be parent of the newly created dialog. 
    * @param   dir
    *          If {@code true} the dialog returned will be able to select only directories, 
    *          otherwise the dialog will allow selection of files only.
    *
    * @return  An instance of the dialog for selection of file or directories.
    */
   AbstractFileChooserDialog createFileSystemChooserDialog(String title,
                                                           int parentWindow,
                                                           boolean dir);
}
