/*
** Module   : ColorMapper.java
** Abstract : map P2J color specifications to fonts that can be used for CHUI 
**
** Copyright (c) 2011-2017, Golden Code Development Corporation.
**
** -#- -I- --Date-- -----------------------Description------------------------
** 001 GES 20110921 Created initial version.
** 002 MAG 20140211 Implements COLOR. 
** 003 MAG 20140219 Fixed NPE when color is null.
** 004 GES 20150424 Added toColor(). Moved to a driver package.
*/
/*
** This program is free software: you can redistribute it and/or modify
** it under the terms of the GNU Affero General Public License as
** published by the Free Software Foundation, either version 3 of the
** License, or (at your option) any later version.
**
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
** GNU Affero General Public License for more details.
**
** You may find a copy of the GNU Affero GPL version 3 at the following
** location: https://www.gnu.org/licenses/agpl-3.0.en.html
** 
** Additional terms under GNU Affero GPL version 3 section 7:
** 
**   Under Section 7 of the GNU Affero GPL version 3, the following additional
**   terms apply to the works covered under the License.  These additional terms
**   are non-permissive additional terms allowed under Section 7 of the GNU
**   Affero GPL version 3 and may not be removed by you.
** 
**   0. Attribution Requirement.
** 
**     You must preserve all legal notices or author attributions in the covered
**     work or Appropriate Legal Notices displayed by works containing the covered
**     work.  You may not remove from the covered work any author or developer
**     credit already included within the covered work.
** 
**   1. No License To Use Trademarks.
** 
**     This license does not grant any license or rights to use the trademarks
**     Golden Code, FWD, any Golden Code or FWD logo, or any other trademarks
**     of Golden Code Development Corporation. You are not authorized to use the
**     name Golden Code, FWD, or the names of any author or contributor, for
**     publicity purposes without written authorization.
** 
**   2. No Misrepresentation of Affiliation.
** 
**     You may not represent yourself as Golden Code Development Corporation or FWD.
** 
**     You may not represent yourself for publicity purposes as associated with
**     Golden Code Development Corporation, FWD, or any author or contributor to
**     the covered work, without written authorization.
** 
**   3. No Misrepresentation of Source or Origin.
** 
**     You may not represent the covered work as solely your work.  All modified
**     versions of the covered work must be marked in a reasonable way to make it
**     clear that the modified work is not originating from Golden Code Development
**     Corporation or FWD.  All modified versions must contain the notices of
**     attribution required in this license.
*/

package com.goldencode.p2j.ui.client.driver.swing;

import java.awt.*;
import java.awt.font.*;
import java.util.*;

// we import selectively because there are many overlapping class names between
// swing and p2j, since the current class is more swing than p2j we err on the
// side of swing imports
import com.goldencode.p2j.ui.Color;
import com.goldencode.p2j.ui.ColorPalette;

/**
 * Converts Progress-compatible color specifications (see
 * {@link com.goldencode.p2j.ui.Color}) into a fixed width font which can be
 * used to display that text in a CHUI which uses that color specification.
 * <p>
 * At this time, only the attributes member of a color specification is
 * honored.  The color member is not supported.
 * <p>
 * The <code>BLINK</code> attribute cannot be supported directly with a font
 * alone (all other attributes can be duplicated with a single font).  Instead
 * the calling code will have to use a timer to periodically redraw the font
 * while inverting the colors each time.
 */
public class ColorMapper
{
   /** Cache of fonts mapped to the color instances they represent. */
   private Map<Color, Font> map = null;
   
   /**
    * Constructor.
    *
    * @param    font
    *           The base font which defines <code>Color.NORMAL</code> and from
    *           which all other "colors" are derived.
    */
   public ColorMapper(Font font)
   {
      reset(font);
   }
   
   /**
    * Reset the cache based on the given font. This is a no-operation if the
    * given font is the same as the current base font.
    *
    * @param    font
    *           The base font which defines <code>Color.NORMAL</code> and from
    *           which all other "colors" are derived.
    */
   public void reset(Font font)
   {
      if (map == null)
      {
         // first initialization
         map = new HashMap<Color, Font>();
      }
      else
      {
         Font old = map.get(Color.NORMAL);
         
         if (!old.equals(font))
         {
            // reset
            map.clear();
         }
         else
         {
            // bypass the map insert below
            font = null;
         }
      }
      
      if (font != null)
         map.put(Color.NORMAL, font);
   }
   
   /**
    * Derive a font from <code>Color.NORMAL</code> which has the attributes
    * specified in the given color AND then return the reverse of that font.
    * parameter is not honored, only the attribute member is handled.
    * <p>
    * <code>BLINK</code> is not supported at this time.
    *
    * @param    color
    *           The color specification to convert into a font.
    *
    * @return   The font that matches the reverse of the color specification.
    */
   public Font toFontReverse(Color color)
   {
      Font normal = toFont(color);
      
      Map<TextAttribute, Object> attr = new HashMap<TextAttribute, Object>();
      attr.put(TextAttribute.SWAP_COLORS, TextAttribute.SWAP_COLORS_ON);
         
      return normal.deriveFont(attr);
   }
   
   /**
    * Derive a font from <code>Color.NORMAL</code> which has the attributes
    * specified in the given color. If that font has already been derived, it
    * will be returned from a cache. At this time, the color member of the
    * parameter is not honored, only the attribute member is handled.
    * <p>
    * <code>BLINK</code> is not supported at this time.
    *
    * @param    color
    *           The color specification to convert into a font.
    *
    * @return   The font that matches the color specification.
    */
   public Font toFont(Color color)
   {
      Font font = map.get(color);
      
      if (font == null)
      {
         Map<TextAttribute, Object> attr = new HashMap<TextAttribute, Object>();
         
         if (color != null)
         {
            // FG color.
            ColorPalette foreground = Color.toForeground(color);                  
            if (foreground != null)
            {
               attr.put(TextAttribute.FOREGROUND, toColor(foreground));
            }
   
            // BG color.
            ColorPalette background = Color.toBackground(color);
            if (background != null)
            {
               attr.put(TextAttribute.BACKGROUND, toColor(background));
            }                  
            
            if (color.hasUnderline())
            {
               attr.put(TextAttribute.UNDERLINE, TextAttribute.UNDERLINE_ON);
            }
            
            if (color.hasReverse())
            {
               attr.put(TextAttribute.SWAP_COLORS, TextAttribute.SWAP_COLORS_ON);
            }
            
            if (color.hasBold())
            {
               attr.put(TextAttribute.WEIGHT, TextAttribute.WEIGHT_BOLD);
            }
            
            if (color.hasBlink())
            {
               // TODO: needs a timer-based implementation, can't be implemented
               //       only with a font!
            }
         }
                           
         // derive the new font
         font = map.get(Color.NORMAL).deriveFont(attr);
        
         // cache it
         map.put(color, font);
      }
      
      return font;
   }
   
   /**
    * Convert the given 4GL color definition into an AWT color instance.
    *
    * @param    clr
    *           The 4GL color definition.
    *
    * @return   The AWT equivalent.
    */
   public static java.awt.Color toColor(ColorPalette clr)
   {
      return new java.awt.Color(clr.getRed(), clr.getGreen(), clr.getBlue());
   }
}