/*
** Module   : PushMessagesWorker.java
** Abstract : a thread used to push messages to the remote end point
**
** Copyright (c) 2014-2017, Golden Code Development Corporation.
**
** -#- -I- --Date-- ----------------------Description--------------------------
** 001 MAG 20140220 Created initial version.
** 002 GES 20150308 Moved to a more generic package location.
*/
/*
** This program is free software: you can redistribute it and/or modify
** it under the terms of the GNU Affero General Public License as
** published by the Free Software Foundation, either version 3 of the
** License, or (at your option) any later version.
**
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
** GNU Affero General Public License for more details.
**
** You may find a copy of the GNU Affero GPL version 3 at the following
** location: https://www.gnu.org/licenses/agpl-3.0.en.html
** 
** Additional terms under GNU Affero GPL version 3 section 7:
** 
**   Under Section 7 of the GNU Affero GPL version 3, the following additional
**   terms apply to the works covered under the License.  These additional terms
**   are non-permissive additional terms allowed under Section 7 of the GNU
**   Affero GPL version 3 and may not be removed by you.
** 
**   0. Attribution Requirement.
** 
**     You must preserve all legal notices or author attributions in the covered
**     work or Appropriate Legal Notices displayed by works containing the covered
**     work.  You may not remove from the covered work any author or developer
**     credit already included within the covered work.
** 
**   1. No License To Use Trademarks.
** 
**     This license does not grant any license or rights to use the trademarks
**     Golden Code, FWD, any Golden Code or FWD logo, or any other trademarks
**     of Golden Code Development Corporation. You are not authorized to use the
**     name Golden Code, FWD, or the names of any author or contributor, for
**     publicity purposes without written authorization.
** 
**   2. No Misrepresentation of Affiliation.
** 
**     You may not represent yourself as Golden Code Development Corporation or FWD.
** 
**     You may not represent yourself for publicity purposes as associated with
**     Golden Code Development Corporation, FWD, or any author or contributor to
**     the covered work, without written authorization.
** 
**   3. No Misrepresentation of Source or Origin.
** 
**     You may not represent the covered work as solely your work.  All modified
**     versions of the covered work must be marked in a reasonable way to make it
**     clear that the modified work is not originating from Golden Code Development
**     Corporation or FWD.  All modified versions must contain the notices of
**     attribution required in this license.
*/

package com.goldencode.p2j.ui.client.driver.web;

import java.io.*;
import java.nio.*;
import java.util.*;
import java.util.concurrent.*;

import org.eclipse.jetty.websocket.api.*;

/**
 * For performance reasons the sending of messages to the remote end point via web sockets
 * should be done asynchronously. A queue is used to store the messages. Messages are extracted
 * from the queue in a FIFO order and are then pushed down to the remote end point.
 * <p>
 * IMPORTANT: Do not use logging anywhere inside this class. When an authentication plugin
 * is running during session establishment with the P2J server using logging no messages are 
 * written into the log and the application becomes deadlocked.
 */
public class PushMessagesWorker 
extends Thread
{
   /** Time in milliseconds to wait for the thread exit. */
   private static final long WAIT_TO_DIE = 5000;

   /** Keyboard input queue. */
   private volatile Deque<Object> messages = new ConcurrentLinkedDeque<>();
   
   /** Remote session. */
   private volatile Session session;  

   /** Remote endpoint. */
   private volatile RemoteEndpoint remote;  
   
   /** Running flag. */
   private volatile boolean running = true;
   
   /**
    * Constructor.
    * 
    * @param    session
    *           Remote end point connection.
    */
   public PushMessagesWorker(Session session)
   {
      this.setName("pushworker");
      this.session = session;
      this.remote = session.getRemote();
      this.start();
   }
   
   /**
    * Add a new message at the end of the FIFO queue and notify the push thread.
    * 
    * @param    message
    *           A binary or a text message.
    */
   public void pushMessage(Object message)
   {
      if (running)
      {
         messages.offerLast(message);
         
         // notify
         synchronized (messages)
         {
            messages.notify();
         }
      }
   }
   
   /**
    * Get messages from queue and send. Wait if no messages on queue.  
    */
   @Override                                          
   public void run()
   {
      while (running)
      {
         // wait until messages are available in queue; this may also return with the
         // queue empty if the thread has been terminated
         waitForMessages();
         
         // send a message from queue (this will check if the running flag is cleared
         // so it is safe even if the queue is empty)
         sendMessage();         
      }
   }

   /**
    * Kill thread and wait to die.
    */
   public void stopPushing()
   {
      running = false;
      this.interrupt();

      // wait for thread to exit      
      try
      {
         this.join(PushMessagesWorker.WAIT_TO_DIE);
      }
      catch (InterruptedException e)
      {
         e.printStackTrace();
      }
   }
   
   /**
    * Block if the queue is empty. This will return when there are messages available in the
    * queue OR in the case where the thread has been interrupted (this should be because the
    * thread has been shutdown but it could also happen spuriously). For this reason, the
    * caller must place this in a loop and check the <code>running</code> flag.
    */
   private void waitForMessages()
   {
      synchronized (messages)
      {
         // are messages in the queue ?
         if (messages.isEmpty())
         {
            try
            {
               // no messages on the queue, wait
               messages.wait();
            }
            catch (InterruptedException e)
            {
               e.printStackTrace();
            }
         }
      }      
   }
   
   /**
    * Send a message.
    */
   private void sendMessage()
   {
      // test if interrupted 
      if (!isInterrupted() && running)
      {
         // get a message from queue, may be null if the queue is empty and there was
         // a spurious interruption in the waiting method
         Object message = messages.pollFirst();

         // send message
         if (session.isOpen())
         {
            try
            {
               // test message type
               if (message instanceof ByteBuffer)
               {
                  remote.sendBytes((ByteBuffer) message);
               }
               else if (message instanceof String)
               {
                  remote.sendString((String) message);
               }
            }
            catch (IOException e)
            {
               running = false;
               e.printStackTrace();
            }
         }    
         else
         {
            running = false;
         }
      }      
   }
}