/*
** Module   : WatchdogTimer.java
** Abstract : a watchdog timer implementation
**
** Copyright (c) 2014-2017, Golden Code Development Corporation.
**
** -#- -I- --Date-- ----------------------Description--------------------------
** 001 MAG 20140131 Created initial version.
** 002 MAG 20140213 Add log messages.
** 003 MAG 20140213 Fixed loop exit condition.
** 004 MAG 20140220 Add join statement on exit.
** 005 GES 20150308 Moved to a more generic package location.
*/
/*
** This program is free software: you can redistribute it and/or modify
** it under the terms of the GNU Affero General Public License as
** published by the Free Software Foundation, either version 3 of the
** License, or (at your option) any later version.
**
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
** GNU Affero General Public License for more details.
**
** You may find a copy of the GNU Affero GPL version 3 at the following
** location: https://www.gnu.org/licenses/agpl-3.0.en.html
** 
** Additional terms under GNU Affero GPL version 3 section 7:
** 
**   Under Section 7 of the GNU Affero GPL version 3, the following additional
**   terms apply to the works covered under the License.  These additional terms
**   are non-permissive additional terms allowed under Section 7 of the GNU
**   Affero GPL version 3 and may not be removed by you.
** 
**   0. Attribution Requirement.
** 
**     You must preserve all legal notices or author attributions in the covered
**     work or Appropriate Legal Notices displayed by works containing the covered
**     work.  You may not remove from the covered work any author or developer
**     credit already included within the covered work.
** 
**   1. No License To Use Trademarks.
** 
**     This license does not grant any license or rights to use the trademarks
**     Golden Code, FWD, any Golden Code or FWD logo, or any other trademarks
**     of Golden Code Development Corporation. You are not authorized to use the
**     name Golden Code, FWD, or the names of any author or contributor, for
**     publicity purposes without written authorization.
** 
**   2. No Misrepresentation of Affiliation.
** 
**     You may not represent yourself as Golden Code Development Corporation or FWD.
** 
**     You may not represent yourself for publicity purposes as associated with
**     Golden Code Development Corporation, FWD, or any author or contributor to
**     the covered work, without written authorization.
** 
**   3. No Misrepresentation of Source or Origin.
** 
**     You may not represent the covered work as solely your work.  All modified
**     versions of the covered work must be marked in a reasonable way to make it
**     clear that the modified work is not originating from Golden Code Development
**     Corporation or FWD.  All modified versions must contain the notices of
**     attribution required in this license.
*/

package com.goldencode.p2j.ui.client.driver.web;

/**
 * A watchdog timer thread implementation.  Once started, if the timer thread is not stopped
 * before the timer expires then the JVM process is ended using <code>System.exit()</code>. This
 * is useful to cleanup a process which is waiting for a connection that never arrives.
 * <p>
 * The thread should only be started when the process is in a state where it should be exited
 * at the end of the timeout. 
 * <p>
 * IMPORTANT: Do not use logging anywhere inside this class. When an authentication plugin
 * is running during session establishment with the P2J server using logging no messages are 
 * written into the log and the application becomes deadlocked.
 */
public class WatchdogTimer
extends Thread
{
   /** Time to wait for thread exit. */
   private static final long WAIT_TO_DIE = 5000;

   /** One minute in milliseconds. */
   private static final long ONE_MINUTE = 60000;
   
   /** System time when started. */
   private volatile long startTime;

   /** Time out value in milliseconds. */
   private volatile long timeout;
   
   /** Flag to control thread abort. */
   private volatile boolean running = true;
   
   /**
    * Create a new instance and start the timer thread.
    * 
    * @param    timeout
    *           The time out value in milliseconds.
    */
   public WatchdogTimer(long timeout)
   {
      this.setName("watchdog");
      this.timeout = timeout;
      this.start();
   }

   /**
    * Because the Thread.sleep time is not guaranteed to be accurate, the elapsed time is
    * calculated once per minute. If the elapsed time is greater than the timeout value, a 
    * JVM process exit is forced (using <code>System.exit()</code>).  Any spurious wake-ups
    * are avoided using a loop condition.
    */
   @Override
   public void run()
   {
      // start time
      startTime = System.currentTimeMillis();

      while (running)
      {
         try
         {
            Thread.sleep(WatchdogTimer.ONE_MINUTE);
         }
         catch (InterruptedException e)
         {
            e.printStackTrace();
         }
         
         // every minute check elapsed time and force an exit on timeout
         long elapsedTime = System.currentTimeMillis() - startTime;
         
         if (running && elapsedTime >= timeout)
         {
            // timer has popped, process goes down now
            System.exit(-1);
         }         
      }
   }
   
   /**
    * Terminate the watchdog timer thread.
    */
   public void kill()
   {
      running = false;
      this.interrupt();
      
      // wait for thread to exit
      try
      {
         this.join(WatchdogTimer.WAIT_TO_DIE);
      }
      catch (InterruptedException e)
      {
         e.printStackTrace();
      }
   }   
}