/*
** Module   : WebClientProtocol.java
** Abstract : provides the core WebSocket functionality for web clients
**
** Copyright (c) 2013-2017, Golden Code Development Corporation.
**
** -#- -I- --Date-- --------------------------------Description-----------------------------------
** 001 GES 20150317 Moved code from the ChuiWebSimulator and enhanced it.
** 002 GES 20150701 Added features related to web GUI implementation. Added utility methods.
** 003 CA  20150820 Implemented MSG_PAGE_LOADED.  Moved here the APIs waiting for a message to be 
**                  received (from GuiWebSocket).  Added readMessageInt32/64.
** 004 CA  20151126 When waiting for a message result to be received from the JS side, each message
**                  must be uniquely identified; otherwise, if the same message type is sent in
**                  the same batch more than once, it will collide and it will not preserve the
**                  message result.
** 005 EVL 20160224 Javadoc fixes to make compatible with Oracle Java 8 for Solaris 10.
** 006 IAS 20160229 Change write method to return new offset 
** 007 CA  20160203 Fixed deadlock for receivedMessages - it must lock on the "lock" field, as
**                  this object is used for all sock operations (including key reading).
**     SBI 20160225 Added MSG_PARTIAL and MSG_PING_PONG, to handle MSG_QUIT.
** 008 EVL 20170829 Adding method to extract byte array from message.
**     SBI 20170903 Used file channels for uploading large files.
*/
/*
** This program is free software: you can redistribute it and/or modify
** it under the terms of the GNU Affero General Public License as
** published by the Free Software Foundation, either version 3 of the
** License, or (at your option) any later version.
**
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
** GNU Affero General Public License for more details.
**
** You may find a copy of the GNU Affero GPL version 3 at the following
** location: https://www.gnu.org/licenses/agpl-3.0.en.html
** 
** Additional terms under GNU Affero GPL version 3 section 7:
** 
**   Under Section 7 of the GNU Affero GPL version 3, the following additional
**   terms apply to the works covered under the License.  These additional terms
**   are non-permissive additional terms allowed under Section 7 of the GNU
**   Affero GPL version 3 and may not be removed by you.
** 
**   0. Attribution Requirement.
** 
**     You must preserve all legal notices or author attributions in the covered
**     work or Appropriate Legal Notices displayed by works containing the covered
**     work.  You may not remove from the covered work any author or developer
**     credit already included within the covered work.
** 
**   1. No License To Use Trademarks.
** 
**     This license does not grant any license or rights to use the trademarks
**     Golden Code, FWD, any Golden Code or FWD logo, or any other trademarks
**     of Golden Code Development Corporation. You are not authorized to use the
**     name Golden Code, FWD, or the names of any author or contributor, for
**     publicity purposes without written authorization.
** 
**   2. No Misrepresentation of Affiliation.
** 
**     You may not represent yourself as Golden Code Development Corporation or FWD.
** 
**     You may not represent yourself for publicity purposes as associated with
**     Golden Code Development Corporation, FWD, or any author or contributor to
**     the covered work, without written authorization.
** 
**   3. No Misrepresentation of Source or Origin.
** 
**     You may not represent the covered work as solely your work.  All modified
**     versions of the covered work must be marked in a reasonable way to make it
**     clear that the modified work is not originating from Golden Code Development
**     Corporation or FWD.  All modified versions must contain the notices of
**     attribution required in this license.
*/

package com.goldencode.p2j.ui.client.driver.web;

import java.io.*;
import java.nio.*;
import java.nio.channels.*;
import java.nio.file.*;
import java.nio.file.attribute.*;
import java.util.*;
import java.util.concurrent.*;

import org.eclipse.jetty.websocket.api.*;
import org.eclipse.jetty.websocket.api.annotations.*;


/**
 * Implements the WebSockets protocol to drive communications with a JavaScript web client.
 * <p>
 * IMPORTANT: Do not use logging anywhere inside this class. When an authentication plugin
 * is running during session establishment with the P2J server using logging no messages are 
 * written into the log and the application is blocked.
 * <p>   
 * The payloads are in 2 formats binary and JSON text.
 * <p>
 * Input messages:
 * 
 * Keyboard input.
 * 
 * offset  value content
 * ------------------------------------
 * 0       0x01  message id
 * 1       uu    keystroke int value
 * 2       ll    keystroke = uu &lt;&lt; 8 | ll 
 * <p>
 * Keyboard pair input.
 * 
 * offset  value content
 * ------------------------------------
 * 0       0x02  message id
 * 1       uu    keyCode int value
 * 2       ll    keyCode = uu &lt;&lt; 8 | ll 
 * 3       uu    charCode int value
 * 4       ll    charCode = uu &lt;&lt; 8 | ll 
 * <p>
 * Paste.
 * 
 * offset  value content
 * ------------------------------------
 * 0       0x03  message id
 * 1       uu    charCode int value
 * 2       ll    charCode = uu &lt;&lt; 8 | ll 
 * ...
 * n-1     uu    charCode int value
 * n       ll    charCode = uu &lt;&lt; 8 | ll 
 * <p>
 * Output messages:
 * 
 * Clear screen.
 * 
 * offset  value content
 * ------------------------------------
 * 0       0x80  message id
 * <p>
 * Draw text on screen.
 * 
 * offset  value content
 * ------------------------------------
 * 0       0x81  message id
 * 1..n    data  data packets
 * 
 * structure of data packets:
 *
 * offset  value  content
 * ------------------------------------
 * 0       size   packet size high byte
 * 1       size   packet size low byte
 * 2       width  rectangle with
 * 3       height rectangle height
 * 4       row    row number
 * 5       col    column number
 * 6       attr   text attributes
 * 7       color  FG/BG color pairs
 * 8..n    text   text characters, each character on 2 bytes (UTF-16)
 * <p>
 * Set cursor position.
 * 
 * offset  value content
 * ------------------------------------
 * 0       0x82  message id
 * 1       row   row
 * 2       col   column
 * <p>
 * Set cursor status.
 * 
 * offset  value content
 * ------------------------------------
 * 0       0x83  message id
 * 1       stat  status 1 = visible 
 *                      0 = hidden 
 * <p>
 * Play a beep message.
 * 
 * offset  value content
 * ------------------------------------
 * 0       0x84  message id
 * <p>
 * Quit message.
 * 
 * offset  value content
 * ------------------------------------
 * 0       0x85  message id
 * <p>
 * Palette message is in JSON format as follow.
 * 
 *    {"c":0, "p":[color1, color2, ... color3]}
 * 
 * Example:
 *    {"c":0, "p":["#000000", "#FF0000", "#00FF00", "#0000FF", "#FFFFFF"]}
 * 
 * offset  value content
 * ------------------------------------
 * 0       0x86  message id
 * 1       mode  mode 1 = VT100 
 *                    0 = P2J 
 * <p>
 * Switch mode.
 * 
 */
@SuppressWarnings("serial")
@WebSocket
public class WebClientProtocol
implements WebClientMessageTypes
{
   /** The options used to open file channels for created temporary files */
   private static final StandardOpenOption[] TMP_FILE_OPTIONS
      = new StandardOpenOption[]
            {
               StandardOpenOption.READ,
               StandardOpenOption.WRITE,
               StandardOpenOption.DELETE_ON_CLOSE
            };
   
   /** The file permissions to create a new temporary file */
   private static final Set<PosixFilePermission> NEW_TMP_FILE_PERMISSIONS;
   
   static
   {
      PosixFilePermission[] permissions
         = new PosixFilePermission[]
               {
                  PosixFilePermission.OWNER_READ,
                  PosixFilePermission.OWNER_WRITE
               };
      NEW_TMP_FILE_PERMISSIONS = new HashSet<PosixFilePermission>(Arrays.asList(permissions));
   }
   
   /** The new created temporary file name should start from this string */
   private static final String TMP_FILE_PREFIX = "msg";
   
   /** The new created temporary file extension */
   private static final String TMP_FILE_EXT = ".tmp";
   
   /** Web socket timeout. */
   private final long timeout;
   
   /** Watchdog thread time out value. */
   private final long wdtimeout;
   
   /** Keyboard input queue. */
   private Deque<Integer> keyBuffer = new ConcurrentLinkedDeque<>();
   
   /** Watchdog thread instance. */
   private WatchdogTimer watchdog;
   
   /** Push messages worker thread. */
   private PushMessagesWorker pushWorker;
   
   /** Number of simultaneous connections. */
   private volatile int connections = 0;
   
   /** Dedicated worker to process the tasks, to release the web-socket. */
   private WebTaskWorker webWorker;
   
   /**
    * A map with the received messages from java-script side.  Requests which require a 
    * "synchronous" approach must call {@link #waitForResult} for the expected operation.
    */
   private Map<Integer, Object> receivedMessages = new HashMap<>();
   
   /** The partial messages collector */
   private MessagesCollector collector;
   
   /** Callbacks for delegated processing. */
   protected ClientProtocolHooks callbacks = null;
   
   /** Keyboard input is being processed in virtual VT100 terminal mode. */
   protected boolean isVT100 = false;
   
   /** Instance lock for synchronization of operations. */
   protected final Object lock;

   /**
    * Create a instance.
    *
    * @param    lock
    *           Object to use for locking. By passing this in, the caller can share the
    *           locking instance with this code.  Use <code>null</code> if an internal
    *           lock object should be created.
    * @param    callbacks
    *           An instance with hooks to call at specific events.
    * @param    timeout
    *           Web socket timeout.
    * @param    wdtimeout
    *           Watchdog thread timeout.
    */
   public WebClientProtocol(Object              lock,
                            ClientProtocolHooks callbacks,
                            long                timeout,
                            long                wdtimeout)
   {
      this.lock      = (lock != null) ? lock : new Object();
      this.callbacks = callbacks;
      this.timeout   = timeout;
      this.wdtimeout = wdtimeout;
      
      startWatchdog(wdtimeout);
   }
   
   /**
    * This method is called when the remote peer open a web socket connection.
    *
    * @param    session
    *           An active link of communications with a Remote WebSocket Endpoint.
    *
    */
   @OnWebSocketConnect
   public void onConnect(Session session)
   {
      synchronized (lock)
      {
         if (++connections == 1)
         {
            // incoming connection means the watchdog is no longer needed
            stopWatchdog();
            
            if (timeout > 0)
            {
               session.setIdleTimeout(timeout);
            }
            
            startWebWorker();
            startPushWorker(session);
            
            if (callbacks != null)
            {
               // synchronize the remove client with the current or initial state
               callbacks.initRemoteClient();
            }
         }
         else
         {
            // multiple connections are not allowed 
            session.close();
         }
      }
   }

   /**
    * This method is called when a text message is received.
    *
    * @param    message
    *           The text message.
    */
   @OnWebSocketMessage
   public void onMessage(String message) 
   {
      // not used
   }

   /**
    * This method is called when a binary message is received.
    *
    * @param    message
    *           Message content as an byte array.
    * @param    offset
    *           Message offset..
    * @param    length
    *           Message size.
    */
   @OnWebSocketMessage
   public void onMessage(final byte[] message, final int offset, final int length)
   {
      webWorker.post(new Runnable()
      {
         @Override
         public void run()
         {
            processBinaryMessage(message, offset, length);
         }
      });
   }
   
   /**
    * This is the worker method which is called when a binary message is received.
    *
    * @param    message
    *           Message content as an byte array.
    * @param    offset
    *           Message offset..
    * @param    length
    *           Message size.
    *
    * @return   <code>true</code> if the message has been handled, <code>false</code> otherwise.
    */
   public boolean processBinaryMessage(byte[] message, int offset, int length)
   {
      boolean handled = false;
      
      if (length == 1 && message[offset] == MSG_PAGE_LOADED)
      {
         receiveResult(MSG_PAGE_LOADED, new Object());
      }
      else if (length == 3 && message[offset] == MSG_KEY)
      {
         int keystroke = readMessageInt16(message, offset + 1);
         
         synchronized (lock)
         {
            keyBuffer.offerLast(keystroke);
            lock.notifyAll();
         }
         
         handled = true;
      }
      else if (length == 5 && message[offset] == MSG_KEY_VT100)
      {
         int keyCode  = readMessageInt16(message, offset + 1);
         int charCode = readMessageInt16(message, offset + 3);
         
         callbacks.injectVT100Key(keyCode, (char) charCode);
         
         handled = true;
      }
      else if (length > 1 && message[offset] == MSG_PASTE)
      {
         paste(message, offset, length);
         handled = true;
      }
      else if (length > 11 && message[offset] == MSG_PARTIAL)
      {
         getMessagesCollector().processPartialMesssage(message, offset);
         handled = true;
      }
      else if (message[offset] == MSG_PING_PONG && length == 1)
      {
         sendBinaryMessage(MSG_PING_PONG);
         handled = true;
      }
      else if (message[offset] == MSG_QUIT && length == 1)
      {
         // exit the client session
         System.exit(0);
      }
      
      return handled;
   }
   
   /**
    * This method is called when the connection is closed.
    *
    * @param   statusCode
    *          The status integer code.
    * @param   reason
    *          A string representing the reason.
    */
   @OnWebSocketClose
   public void onClose(int statusCode, String reason)
   {
      synchronized (lock)
      {
         if (--connections == 0)
         {
            stopPushWorker();
            stopWebWorker();
            startWatchdog(wdtimeout);
         }
      }
   }
   
   /**
    * This method is called on communication errors.
    *
    * @param   session
    *          An active link of communications with a Remote WebSocket Endpoint.
    * @param   error
    *          A <code>Throwable</code> instance representing the error condition.
    */
   @OnWebSocketError
   public void onError(Session session, Throwable error)
   {
      synchronized (lock)
      {
         if (error != null)
         {
            error.printStackTrace();
         }
         session.close(StatusCode.ABNORMAL, "");
      }
   }

   /**
    * Send a text message.
    * 
    * @param    message
    *           The fully encoded (JSON) text message.
    */   
   public void sendTextMessage(String message) 
   {
      synchronized (lock)
      {
         if (pushWorker != null)
         {
            pushWorker.pushMessage(message);
         }
      }
   }

   /**
    * Send a binary message of arbitrary size.
    * 
    * @param    message
    *           The fully encoded binary message contents.
    */   
   public void sendBinaryMessage(byte[] message) 
   {
      synchronized (lock)
      {
         if (pushWorker != null)
         {
            pushWorker.pushMessage(ByteBuffer.wrap(message));
         }
      }
   }

   /**
    * Send a binary message with text of arbitrary size.
    * 
    * @param    type
    *           The message type.
    * @param    text
    *           The binary message contents as text.
    */   
   public void sendBinaryMessage(byte type, String text) 
   {
      int len = text.length();
      
      // TODO: all unicode chars won't necessarily fit in 16-bits
      byte[] message = new byte[(len * 2) + 1];
      
      message[0] = type;
      writeMessageText(message, 1, text);
      
      sendBinaryMessage(message);
   }
   
   /**
    * Send a binary message with a 32-bit integer of arbitrary size.
    * 
    * @param    type
    *           The message type.
    * @param    num
    *           The binary message contents as a num.
    */   
   public void sendBinaryMessage(byte type, int num) 
   {
      byte[] message = new byte[5];
      
      message[0] = type;
      writeMessageInt32(message, 1, num);
      
      sendBinaryMessage(message);
   }
   
   /**
    * Convenience method to send a single-byte binary message.  The core functionality is
    * implemented in {@link #sendBinaryMessage(byte[])}.
    * 
    * @param    message
    *           The byte which comprises the entire binary message.
    */   
   public void sendBinaryMessage(byte message) 
   {
      sendBinaryMessage(new byte[] { message });
   }

   /**
    * Convenience method to send a two-byte binary message.  The core functionality is
    * implemented in {@link #sendBinaryMessage(byte[])}.
    * 
    * @param    first
    *           The first byte of the binary message.
    * @param    second
    *           The second byte of the binary message.
    */   
   public void sendBinaryMessage(byte first, byte second) 
   {
      sendBinaryMessage(new byte[] { first, second });
   }

   /**
    * Convenience method to send a three-byte binary message.  The core functionality is
    * implemented in {@link #sendBinaryMessage(byte[])}.
    * 
    * @param    first
    *           The first byte of the binary message.
    * @param    second
    *           The second byte of the binary message.
    * @param    third
    *           The third byte of the binary message.
    */   
   public void sendBinaryMessage(byte first, byte second, byte third) 
   {
      sendBinaryMessage(new byte[] { first, second, third });
   }

   /**
    * Interpret 2 bytes from a binary message as a 16-bit integer and return the accumulated
    * result.
    *
    * @param    message
    *           Message content as an byte array.
    * @param    offset
    *           Offset into the message at which the integer begins.
    *
    * @return   The number.
    */
   public int readMessageInt16(byte[] message, int offset)
   {
      return (((message[offset] << 8) & 0xFF00) | (message[offset + 1] & 0x00FF));
   }

   /**
    * Interpret 4 bytes from a binary message as a 32-bit integer and return the accumulated
    * result.
    *
    * @param    message
    *           Message content as an byte array.
    * @param    offset
    *           Offset into the message at which the integer begins.
    *
    * @return   The number.
    */
   public int readMessageInt32(byte[] message, int offset)
   {
      return ((message[offset]     << 24) & 0xFF000000) | 
             ((message[offset + 1] << 16) & 0x00FF0000) |
             ((message[offset + 2] <<  8) & 0x0000FF00) | 
              (message[offset + 3]        & 0x000000FF);
   }

   /**
    * Interpret 8 bytes from a binary message as a 64-bit integer and return the accumulated
    * result.
    *
    * @param    message
    *           Message content as an byte array.
    * @param    offset
    *           Offset into the message at which the integer begins.
    *
    * @return   The number.
    */
   public long readMessageInt64(byte[] message, int offset)
   {
      return (((long) message[offset]     << 56) & 0xFF00000000000000L) |
             (((long) message[offset + 1] << 48) & 0x00FF000000000000L) | 
             (((long) message[offset + 2] << 40) & 0x0000FF0000000000L) |
             (((long) message[offset + 3] << 32) & 0x000000FF00000000L) | 
             (((long) message[offset + 4] << 24) & 0x00000000FF000000L) | 
             (((long) message[offset + 5] << 16) & 0x0000000000FF0000L) |
             (((long) message[offset + 6] <<  8) & 0x000000000000FF00L) | 
             (((long) message[offset + 7])       & 0x00000000000000FFL);
   }
         
   /**
    * Interpret bytes from a binary message as text and return the accumulated result.
    *
    * @param    message
    *           Message content as an byte array.
    * @param    offset
    *           Offset into the message at which the text begins.
    * @param    length
    *           Location in the message where the text ends.
    *
    * @return   The message text.
    */
   public String readMessageText(byte[] message, int offset, int length)
   {
      StringBuilder sb = new StringBuilder();
      
      for (int i = offset; i < length; i += 2)
      {
         sb.append((char) readMessageInt16(message, i));
      }
      
      return sb.toString();
   }

   /**
    * Interpret bytes from a binary message as byte array and return the accumulated result.
    *
    * @param    message
    *           Message content as an byte array.
    * @param    offset
    *           Offset into the message at which the data begins.
    * @param    size
    *           The length of the array of byte.
    *
    * @return   The message text.
    */
   public byte [] readByteArray(byte[] message, int offset, int size)
   {
      byte b[] = new byte[size];
      
      // making array copy
      System.arraycopy(message, offset, b, 0, size);
      
      return b;
   }

   /**
    * Write the lower 2 bytes from a given value into a binary message at a given offset.
    *
    * @param    message
    *           Message content as an byte array.
    * @param    offset
    *           Offset into the message at which the integer should be written.
    * @param    value
    *           The 16-bit value to write.
    *           
    * @return   the new offset (after the end of the written data)       
    */
   public int writeMessageInt16(byte[] message, int offset, int value)
   {
      message[offset]     = (byte)((value >> 8) & 0x00FF);
      message[offset + 1] = (byte)(value & 0x00FF);
      return offset + 2;
   }
         
   /**
    * Write a 32-bit integer value into a binary message at a given offset.
    *
    * @param    message
    *           Message content as an byte array.
    * @param    offset
    *           Offset into the message at which the integer should be written.
    * @param    value
    *           The 32-bit value to write.
    *           
    * @return   the new offset (after the end of the written data)       
    */
   public int writeMessageInt32(byte[] message, int offset, int value)
   {
      message[offset]     = (byte)((value >> 24) & 0x00FF);
      message[offset + 1] = (byte)((value >> 16) & 0x00FF);
      message[offset + 2] = (byte)((value >> 8)  & 0x00FF);
      message[offset + 3] = (byte)( value        & 0x00FF);
      return offset + 4;
   }
         
   /**
    * Write the given text into the binary message at the specified offset.
    *
    * @param    message
    *           Message content as an byte array.  This must be large enough to hold the text.
    * @param    offset
    *           Offset into the message at which the text should begin.
    * @param    text
    *           The message text.
    *           
    * @return   the new offset (after the end of the written data)       
    */
   public int writeMessageText(byte[] message, int offset, String text)
   {
      if(text == null)
      {
         return offset;
      }
      int len = text.length();
      
      for (int i = 0; i < len; i++)
      {
         // TODO: all unicode chars won't necessarily fit in 16-bits
         offset = writeMessageInt16(message, offset, (int) text.charAt(i));
      }
      return offset;
   }

   /**
    * Read the next key from the FIFO key input buffer.  
    *
    * @return   The next key in the buffer or -1 if buffer is empty.
    */
   public int readKey()
   {
      synchronized (lock)
      {
         int result = -1;
         
         while (keyBuffer.isEmpty())
         {
            try
            {
               lock.wait();
            }
            catch (InterruptedException e)
            {
               break;
            }
         }

         Integer key = keyBuffer.pollFirst();
         
         if (key != null)
         {
            result = key.intValue();
         }

         return result;
      }
   }
   
   /**
    * Process the incoming PASTE message from the client.
    *
    * @param    message
    *           Message content as an byte array.
    * @param    offset
    *           Offset into the message at which the text begins.
    * @param    length
    *           Location in the message where the text ends.
    */
   public void paste(byte[] message, int offset, int length)
   {
      for (int i = offset + 1; i < length; i += 2)
      {
         injectKey(readMessageInt16(message, i));
      }
   }
   
   /**
    * Inject the given key into the current keystroke buffer.
    *
    * @param    key
    *           The keystroke to inject.
    */
   public void injectKey(int key)
   {
      if (isVT100)
      {
         callbacks.injectVT100Key(key, (char) key);
      }
      else
      {
         synchronized (lock)
         {
            keyBuffer.offerLast(key);
            
            // TODO: is it better to notify only after all the message is queued?
            lock.notifyAll();
         }
      }
   }
   
   /**
    * The beep() should be implemented by sending a command on the remote end point.
    * The command is part of the communication protocol.
    *
    */
   public void beep()
   {
      sendBinaryMessage(MSG_BEEP);         
   }
   
   /**
    * Remote application is about to end. Notify the JS client about that.
    * The client should be redirected to referrer address.
    */
   public void quit()
   {
      sendBinaryMessage(MSG_QUIT);         
   }

   /**
    * Switch between P2J client and interactive client.
    * 
    * @param    isVT100
    *           If <code>true</code>, process keys in virtual VT100 terminal mode otherwise
    *           process for internal client screen.
    */
   public void swapKeyboardMode(boolean isVT100)
   {
      synchronized (lock)
      {
         this.isVT100 = isVT100;
      }
      sendSwitchMode();
   }   

   /**
    * Switch mode message.
    */
   public void sendSwitchMode()
   {
      sendBinaryMessage(MSG_VT100, (byte)(this.isVT100 ? 1 : 0));               
   }
   
   /**
    * Wait for a result to be {@link #lock received} for the given message type.
    * 
    * @param    msgId
    *           The response message ID.
    * 
    * @return   The result.
    */
   public Object waitForResult(int msgId)
   {
      synchronized (lock)
      {
         while (true)
         {
            // this needs to be checked first in case the message was already received before
            // starting the waiting loop
            if (receivedMessages.containsKey(msgId))
            {
               return receivedMessages.remove(msgId);
            }
            
            try
            {
               lock.wait();
            }
            catch (InterruptedException e)
            {
               break;
            }
         }
         
         return receivedMessages.remove(msgId);
      }
   }
  
   /**
    * Returns the unique instance of MessageCollector
    *  
    * @return The messages collector.
    */
   public MessagesCollector getMessagesCollector()
   {
      if (collector == null)
      {
         collector = new MessagesCollector();
      }
      
      return collector;
   }
   
   /**
    * Notify the {@link #lock} that a message was received with the given type.
    * 
    * @param    msgId
    *           The response message ID.
    * @param    result
    *           The received result.
    */
   protected void receiveResult(int msgId, Object result)
   {
      synchronized (lock)
      {
         receivedMessages.put(msgId, result);
         lock.notifyAll();
      }
   }

   /**
    * Looks ahead the given message channel and takes a decision to read and parse this channel or
    * to process it by the default. This method must be overridden in its super class
    * if this channel contains new message type that must have a special treatment.
    * 
    * @param    channel
    *           The given file channel.
    * 
    * @return   The true value indicates that this channel has been parsed, otherwise 
    *           the channel content is not handled by this method.
    */
   protected boolean processChannel(FileChannel channel)
   {
      return false;
   }
   
   /**
    * Handles the given message channel by the default if it has not been processed yet and then
    * closes this message channel and removes its reference from the map of the currently tracked
    * messages.
    * 
    * @param    channel
    *           The given message channel
    * @param    appendTask
    *           The append task associated with the given message channel.
    * @param    isNotProcessed
    *           The true value means that this channel hasn't been parsed yet.
    * 
    */
   private void processChannel(FileChannel channel, Closeable appendTask, boolean isNotProcessed)
   {
      try
      {
         if (isNotProcessed)
         {
            int size = (int) channel.size();
            ByteBuffer dst = ByteBuffer.allocate(size);
            final int readedBytes = channel.read(dst, 0);
            dst.flip();
            
            if (readedBytes == size)
            {
               processBinaryMessage(dst.array(), 0, readedBytes);
            }
         }
      }
      catch(IOException | OutOfMemoryError ex)
      {
         // TODO report exception
         // ex.printStackTrace();
      }
      finally
      {
         try
         {
            appendTask.close();
         }
         catch(IOException closeChannelEx)
         {
         }
      }
   }

   /**
    * Start watchdog timer.
    * <p>
    * This must only be called by a caller that owns the instance lock.
    * 
    * @param    wdtimeout
    *           Watchdog thread timeout value.
    */
   private void startWatchdog(long wdtimeout)
   {
      if (watchdog == null)
      {
         watchdog = new WatchdogTimer(wdtimeout);
      }
   }

   /**
    * Stop watchdog timer.
    * <p>
    * This must only be called by a caller that owns the instance lock.
    */
   private void stopWatchdog()
   {
      if (watchdog != null)
      {
         watchdog.kill();
         watchdog = null;
      }
   }
   
   /**
    * Start the web worker which processes tasks sent from javascript.
    */
   private void startWebWorker()
   {
      webWorker = new WebTaskWorker();
   }
   
   /**
    * Stop the web worker which processes tasks sent from javascript.
    */
   private void stopWebWorker()
   {
      if (webWorker != null)
      {
         webWorker.kill();
         webWorker = null;
      }
   }

   /**
    * Start pushWorker thread.
    * <p>
    * This must only be called by a caller that owns the instance lock.
    * 
    * @param   session
    *          An active link of communications with a Remote WebSocket end point.
    */
   private void startPushWorker(Session session)
   {
      if (pushWorker == null)
      {
         pushWorker = new PushMessagesWorker(session);
      }
   }

   /**
    * Stop pushWorker thread.
    * <p>
    * This must only be called by a caller that owns the instance lock.
    */
   private void stopPushWorker()
   {
      if (pushWorker != null)
      {
         pushWorker.stopPushing();
         pushWorker = null;
      }
   }
   
   /**
    * MessagesColector is responsible to collect the original message from the messages of
    * this type MSG_PARTIAL. 
    */
   class MessagesCollector
   {
      /**
       * Executes tasks to collect partial messages having the same id number to a one large
       * binary message.
       */
      private final ExecutorService asynchIOExecutor = Executors.newSingleThreadExecutor();

      /** Partial messages are mapped to temporary files channels. */
      private final Map<Integer, FileChannel> partialMessages = new HashMap<>();

      /** Partial messages are mapped to their payload data types. */
      private final Map<Integer, Byte> payloadMessagesTypes = new HashMap<>();
      
      /** The directory to save partial messages to the corresponding temporary file */
      private final Path tmpDir;

      /**
       * Creates the instance.
       */
      private MessagesCollector()
      {
         Path tmp;
         try
         {
            tmp = Files.createTempDirectory(null);
         }
         catch (IOException e)
         {
            // try to use the current working directory
            tmp = new File("./").toPath();
         }
         this.tmpDir = tmp;
      }
      
      /**
       * Processes messages of the MSG_PARTIAL type.
       * 
       * @param    message
       *           The received message.
       * @param    offset
       *           The data offset.
       */
      public void processPartialMesssage(byte[] message, int offset)
      {
         // [MSG_PARTIAL_TYPE][MSG_ID][PARTIAL_STATUS][PAYLOAD_LENGTH][PAYLOAD_DATA]
         // 1 + 4 + 1 + 4 + length(payload data) > 10 
         final int msgId = readMessageInt32(message, offset + 1);
         // PARTIAL_STATUS is 1 byte that equals 0 if it is the last message with MSG_ID or
         // 1 if a next message with the same MSG_ID should follow this one.
         final boolean isLast = message[offset + 5] == 0x0;
         final int payloadLength = readMessageInt32(message, offset + 6);
         final ByteBuffer payloadData = ByteBuffer.wrap(message, offset + 10, payloadLength);
         
         // retrieve payload data type
         Byte payloadType = payloadMessagesTypes.get(msgId);
         
         if (payloadType == null)
         {
            payloadType = Byte.valueOf(message[offset + 10]);
            payloadMessagesTypes.put(msgId, payloadType);
         }
         
         FileChannel channel = partialMessages.get(msgId);
         
         if (channel == null)
         {
            try
            {
               
               Path channelPath = Files.createTempFile(tmpDir, TMP_FILE_PREFIX, TMP_FILE_EXT, 
                        PosixFilePermissions.asFileAttribute(NEW_TMP_FILE_PERMISSIONS));
               channel = FileChannel.open(channelPath,TMP_FILE_OPTIONS);
               partialMessages.put(msgId, channel);
            }
            catch (IOException e)
            {
               // TODO:report exception and notify the JS client that the message is not accepted
               // e.printStackTrace();
            }
         }

         if (channel != null && channel.isOpen())
         {
            asynchIOExecutor.submit(new AppendMessageTask(msgId, payloadData, isLast));
         }
         
      }
      
      /**
       * The task that appends a new partial message to the already received messages having the same
       * id number (or belonging to the same packet of messages).
       */
      class AppendMessageTask
      implements Runnable, Closeable
      {

         /** The target message id to collect */
         private final int msgId;
      
         /** Payload data that should be added to build the original message*/
         private final ByteBuffer payloadData;
         
         /**
          * Indicates that it is a final packet of the original message 
          */
         private final boolean isLast;


         /**
          * Appends the provided payload data to the original message.
          * 
          * @param    msgId
          *           The original message id.
          * @param    payloadData
          *           The payload data nested in the .
          * @param    isLast
          */
         public AppendMessageTask(int msgId, ByteBuffer payloadData, boolean isLast)
         {
            this.msgId = msgId;
            this.payloadData = payloadData;
            this.isLast = isLast;
         }
         
         /**
          * Executes this task.
          */
         @Override
         public void run()
         {
            // Find a temporary file's channel by id
            FileChannel channel = partialMessages.get(msgId);
            // Write a payload data
            try
            {
               channel.write(payloadData);
            }
            catch(IOException ex)
            {
               // TODO report exception
               // ex.printStackTrace();
               try
               {
                  channel.close();
               }
               catch(IOException closeChannelEx)
               {
               }
            }
            if (isLast)
            {
               // if message is the last message in this messages packet
               // then read all gathered payload data and deliver it
               // to the messages processor
               webWorker.post(new Runnable()
               {
                  @Override
                  public void run()
                  {
                     boolean handled = processChannel(channel);
                     processChannel(channel, AppendMessageTask.this, !handled);
                  }
               });
            }
         }

         /**
          * Closes this append task and releases the system resources associated with the opened 
          * file channel and cleans the map of the currently tracked  messages.
          */
         @Override
         public void close()
         throws IOException
         {
            payloadMessagesTypes.remove(msgId);
            FileChannel channel = partialMessages.remove(msgId);
            try
            {
               if (channel != null && channel.isOpen())
               {
                  channel.close();
               }
            }
            catch (IOException e)
            {
               // TODO report exception
               // e.printStackTrace();
            }
         }
      }
   }
}