/*
** Module   : DirectManipulationEvent.java
** Abstract : Direct manipulation specific event encoding.
**
** Copyright (c) 2017, Golden Code Development Corporation.
**
** -#- -I- --Date-- ---------------------------------Description---------------------------------
** 001 EVL 20170714 Created initial version
*/
/*
** This program is free software: you can redistribute it and/or modify
** it under the terms of the GNU Affero General Public License as
** published by the Free Software Foundation, either version 3 of the
** License, or (at your option) any later version.
**
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
** GNU Affero General Public License for more details.
**
** You may find a copy of the GNU Affero GPL version 3 at the following
** location: https://www.gnu.org/licenses/agpl-3.0.en.html
** 
** Additional terms under GNU Affero GPL version 3 section 7:
** 
**   Under Section 7 of the GNU Affero GPL version 3, the following additional
**   terms apply to the works covered under the License.  These additional terms
**   are non-permissive additional terms allowed under Section 7 of the GNU
**   Affero GPL version 3 and may not be removed by you.
** 
**   0. Attribution Requirement.
** 
**     You must preserve all legal notices or author attributions in the covered
**     work or Appropriate Legal Notices displayed by works containing the covered
**     work.  You may not remove from the covered work any author or developer
**     credit already included within the covered work.
** 
**   1. No License To Use Trademarks.
** 
**     This license does not grant any license or rights to use the trademarks
**     Golden Code, FWD, any Golden Code or FWD logo, or any other trademarks
**     of Golden Code Development Corporation. You are not authorized to use the
**     name Golden Code, FWD, or the names of any author or contributor, for
**     publicity purposes without written authorization.
** 
**   2. No Misrepresentation of Affiliation.
** 
**     You may not represent yourself as Golden Code Development Corporation or FWD.
** 
**     You may not represent yourself for publicity purposes as associated with
**     Golden Code Development Corporation, FWD, or any author or contributor to
**     the covered work, without written authorization.
** 
**   3. No Misrepresentation of Source or Origin.
** 
**     You may not represent the covered work as solely your work.  All modified
**     versions of the covered work must be marked in a reasonable way to make it
**     clear that the modified work is not originating from Golden Code Development
**     Corporation or FWD.  All modified versions must contain the notices of
**     attribution required in this license.
*/

package com.goldencode.p2j.ui.client.event;

import com.goldencode.p2j.ui.client.*;
import com.goldencode.p2j.ui.client.gui.driver.*;
import com.goldencode.p2j.ui.client.widget.*;

/**
 * Direct manipulation event. This event is generated as a reactions on direct manipulation
 * related mouse events.
 */
public class DirectManipulationEvent
extends Event
{
   /** Optional parent container. */
   private Container<?> parentContainer = null;
   
   /** Flag indicating the parent container to be repainted after message handling. */
   private boolean forceParentRepaint = false;
   
   /** Variable to store X location or width changes value. */
   private int xChange = 0;
   
   /** Variable to store Y location or height changes value. */
   private int yChange = 0;
   
   /** Resize direction for widget resize event. */
   private int resizeDirection = -1;
   
   /** Pointer to optional hanler to allow callback. */
   private MouseDirectManipulation handler = null;
   
   /**
    * Constructor.  Passing the parent container as additional option.  This means the message
    * handler should repaint it for the changes made visible.
    * 
    * @param   source
    *          Event source.
    * @param   code
    *          Event code defining the action performed.
    */
   public DirectManipulationEvent(Widget source, int code)
   {
      this(source, null, false, code);
   }
   
   /**
    * Constructor.  Passing the parent container as additional option.  This means the message
    * handler should repaint it for the changes made visible.
    * 
    * @param   handler
    *          Direct manupulation handler to be called inside event target context.
    * @param   source
    *          Event source.
    * @param   code
    *          Event code defining the action performed.
    */
   public DirectManipulationEvent(MouseDirectManipulation handler, Widget source, int code)
   {
      this(source, null, false, code);
      this.handler = handler;
   }
   
   /**
    * Constructor.  Passing the parent container as additional option.  This means the message
    * handler should repaint it for the changes made visible.
    * 
    * @param   source
    *          Event source.
    * @param   parent
    *          The parent container for event source.
    * @param   code
    *          Event code defining the action performed.
    */
   public DirectManipulationEvent(Widget source, Container parent, int code)
   {
      this(source, parent, false, code);
   }
   
   /**
    * Constructor.  Passing the parent container as additional option.  This means the message
    * handler should repaint it for the changes made visible.
    * 
    * @param   source
    *          Event source.
    * @param   parent
    *          The parent container for event source.
    * @param   repaint
    *          The parent container needs repainting after message.
    * @param   code
    *          Event code defining the action performed.
    */
   public DirectManipulationEvent(Widget source, Container parent, boolean repaint, int code)
   {
      super(source, code);
      this.parentContainer = parent;
      this.forceParentRepaint = repaint;
   }
   
   /**
    * Encodes X and Y shift pixels value.  Used while widget move or resize event processing.
    * 
    * @param   xChange
    *          X distance to move or resize, can be negative.
    * @param   yChange
    *          Y distance to move or resize, can be negative.
    */
   public void setDistance(int xChange, int yChange)
   {
      this.xChange = xChange;
      this.yChange = yChange;
   }
   
   /**
    * Extracts horizontal direction shift pixels value.  Used while widget move or resize event
    * processing.
    * 
    * @return   X distance to move or resize.
    */
   public int getXChange()
   {
      return xChange;
   }
   
   /**
    * Extracts vertical direction shift pixels value.  Used while widget move or resize event
    * processing.
    * 
    * @return   Y distance to move or resize.
    */
   public int getYChange()
   {
      return yChange;
   }
   
   /**
    * Encodes widget resize direction value.  Used with widget resize event processing.
    * 
    * @param   resizeDirection
    *          Widget resize direction.
    */
   public void setResizeDirection(int resizeDirection)
   {
      this.resizeDirection = resizeDirection;
   }
   
   /**
    * Extracts widget resize direction value.  Used with widget resize event processing.
    * 
    * @return   Widget resize direction.
    */
   public int getResizeDirection()
   {
      return resizeDirection;
   }
   
   /**
    * Checks if the repaint is necessary for changes to make visible.
    * 
    * @return   <code>TRUE</code> if the repaint is possible,
    *           <code>FALSE</code> otherwise.
    */
   public boolean isRepaintRequired()
   {
      return parentContainer != null && forceParentRepaint;
   }
   
   /**
    * Gets the parent container for widget target of the message.
    * 
    * @return   Container instance.
    */
   public Container<?> getParentContainer()
   {
      return parentContainer;
   }
   
   /**
    * Checks if the event has handler integrated.  This can be used for some event that have
    * dependency on whether previous events were consumed or not.
    * 
    * @return   <code>TRUE</code> if handler exists, <code>FALSE</code> otherwise.
    */
   public boolean hasHandler()
   {
      return handler != null;
   }
   
   /**
    * Gets the direct manipulation handler the event was generated within.
    * 
    * @return   Handler instance.
    */
   public MouseDirectManipulation getHandler()
   {
      return handler;
   }
}