/*
** Module   : PaintEvent.java
** Abstract : Repaint request event.
**
** Copyright (c) 2010-2017, Golden Code Development Corporation.
**
** -#- -I- --Date-- ---------------------------------Description----------------------------------
** 001 SIY 20100629 Created initial version
** 002 SIY 20100804 Changes in imports, some refactorings.
** 003 VIG 20160620 Fixed imports.
** 004 CA  20171127 Cache PaintEvent's and process them before the application gets into 
**                  a user-interactive state.
*/
/*
** This program is free software: you can redistribute it and/or modify
** it under the terms of the GNU Affero General Public License as
** published by the Free Software Foundation, either version 3 of the
** License, or (at your option) any later version.
**
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
** GNU Affero General Public License for more details.
**
** You may find a copy of the GNU Affero GPL version 3 at the following
** location: https://www.gnu.org/licenses/agpl-3.0.en.html
** 
** Additional terms under GNU Affero GPL version 3 section 7:
** 
**   Under Section 7 of the GNU Affero GPL version 3, the following additional
**   terms apply to the works covered under the License.  These additional terms
**   are non-permissive additional terms allowed under Section 7 of the GNU
**   Affero GPL version 3 and may not be removed by you.
** 
**   0. Attribution Requirement.
** 
**     You must preserve all legal notices or author attributions in the covered
**     work or Appropriate Legal Notices displayed by works containing the covered
**     work.  You may not remove from the covered work any author or developer
**     credit already included within the covered work.
** 
**   1. No License To Use Trademarks.
** 
**     This license does not grant any license or rights to use the trademarks
**     Golden Code, FWD, any Golden Code or FWD logo, or any other trademarks
**     of Golden Code Development Corporation. You are not authorized to use the
**     name Golden Code, FWD, or the names of any author or contributor, for
**     publicity purposes without written authorization.
** 
**   2. No Misrepresentation of Affiliation.
** 
**     You may not represent yourself as Golden Code Development Corporation or FWD.
** 
**     You may not represent yourself for publicity purposes as associated with
**     Golden Code Development Corporation, FWD, or any author or contributor to
**     the covered work, without written authorization.
** 
**   3. No Misrepresentation of Source or Origin.
** 
**     You may not represent the covered work as solely your work.  All modified
**     versions of the covered work must be marked in a reasonable way to make it
**     clear that the modified work is not originating from Golden Code Development
**     Corporation or FWD.  All modified versions must contain the notices of
**     attribution required in this license.
*/
 
package com.goldencode.p2j.ui.client.event;

import com.goldencode.p2j.ui.client.*;
import com.goldencode.p2j.ui.client.widget.*;

/**
 * Area repaint request event.
 * <p>
 * Note that this event does not trigger immediate repainting. Instead it is 
 * used to mark specified area as such that requires repainting and actual 
 * repainting happens at the end of event processing cycle.
 */
public class PaintEvent
extends Event
{
   /** Rectangle which describes the are which need to be repainted. */
   private Rectangle updateRect;

   /** Flag indicating this repaint is being postponed. */
   private boolean delayed = true;
   
   /**
    * Construct an instance of Paint event from given widget and rectangle.
    * 
    * @param   source
    *          The event source.
    * @param   rect
    *          Rectangle to repaint.
    */
   public PaintEvent(Widget source, Rectangle rect)
   {
      super(source, EventType.PAINT_EVENT);
      updateRect = rect;
      
      delayed = !FocusEvent.isDispatchingFocus();
   }

   /**
    * Set the {@link #delayed} state.
    * 
    * @param    delayed
    *           <code>true</code> if the repaint is delayed.
    */
   public void setDelayed(boolean delayed)
   {
      this.delayed = delayed;
   }
   
   /**
    * Check the {@link #delayed} state.
    * 
    * @return   <code>true</code> if the repaint is delayed.
    */
   public boolean isDelayed()
   {
      return delayed;
   }
   
   /**
    * Returns the rectangle representing the area that needs to be repainted
    * in response to this event.
    * 
    * @return  The rectangle which need to be repainted. 
    */
   public Rectangle getUpdateRect()
   {
      return new Rectangle(updateRect);
   }

   /**
    * Override default comparison method.
    * 
    * @see java.lang.Object#equals(java.lang.Object)
    */
   @Override
   public boolean equals(Object obj)
   {
      if (obj instanceof PaintEvent)
      {
         PaintEvent pe = (PaintEvent) obj;

         if (pe.source().hashCode() == source().hashCode()
             && pe.updateRect.equals(updateRect))
            return true;
      }
      return false;
   }
   
   /**
    * Hash code implementation which is consistent with {@link #equals}.
    * 
    * @return  Hash code value for this object instance.
    */
   @Override
   public int hashCode()
   {
      int result = 17;

      result = 37 * result + source().hashCode();
      result = 37 * result + updateRect.hashCode();

      return result;
   }
}