/*
** Module   : BoolFormat.java
** Abstract : formatting/editing logic for logical variables 
**
** Copyright (c) 2005-2017, Golden Code Development Corporation.
**
** -#- -I- --Date-- --JPRM-- ----------------Description-----------------
** 001 SVG 20050609   @21446 Initial version.
** 002 SVG 20050620   @21523 DisplayFormat related classes are redesigned.
**                           Now cursor's position and insert mode are
**                           maintained inside Presentation. This allows
**                           to not pass them by reference into
**                           Presentation edit methods. All
**                           com.goldencode.p2j.ui.chui.dataref package
**                           is obsoleted and removed.
** 003 SIY 20050831   @22525 Formatting and spelling fixes.
** 004 SIY 20050907   @22622 Fixed use of @inheritDoc tag in javadoc
** 005 GES 20060504   @25966 Character comparisons are now case 
**                           insensitive. Massive formatting cleanup and
**                           removal of all of the ridiculous inheritDoc
**                           usage which should never, ever be present in
**                           GCD source code.
** 006 NVS 20060519   @26348 Logical formats "x", "x/", "/x" and even
**                           "x/x" are now accepted and work properly.
** 007 GES 20060920   @26728 Added support for isUnknown().
** 008 SVL 20080922   @39895 Insert mode reset removed from the
**                           activate().
** 009 SVL 20090402   @41718 Insert mode is saved on clear().
** 010 SIY 20090721   @43291 Moved activate() into base class.
** 011 GES 20090722   @43326 Fixed auto-return. The last char entered flag
**                           should only be set when the last character in the
**                           field is successfully filled or overwritten. The
**                           previous implementation was triggering this
**                           condition as soon as any auto-completion was
**                           caused, which is not how it works in the 4GL.
** 012 SIY 20100619          Switching to own implementation of character UI.
** 013 SIY 20100804          Changes in imports, some refactorings.
** 014 SIY 20111101          Updated imports.
** 015 CA  20141220          Changes related to GUI FILL-IN implementation.
** 016 SBI 20180807          Changed fromVar in order to fix class cast to logical type.
*/
/*
** This program is free software: you can redistribute it and/or modify
** it under the terms of the GNU Affero General Public License as
** published by the Free Software Foundation, either version 3 of the
** License, or (at your option) any later version.
**
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
** GNU Affero General Public License for more details.
**
** You may find a copy of the GNU Affero GPL version 3 at the following
** location: https://www.gnu.org/licenses/agpl-3.0.en.html
** 
** Additional terms under GNU Affero GPL version 3 section 7:
** 
**   Under Section 7 of the GNU Affero GPL version 3, the following additional
**   terms apply to the works covered under the License.  These additional terms
**   are non-permissive additional terms allowed under Section 7 of the GNU
**   Affero GPL version 3 and may not be removed by you.
** 
**   0. Attribution Requirement.
** 
**     You must preserve all legal notices or author attributions in the covered
**     work or Appropriate Legal Notices displayed by works containing the covered
**     work.  You may not remove from the covered work any author or developer
**     credit already included within the covered work.
** 
**   1. No License To Use Trademarks.
** 
**     This license does not grant any license or rights to use the trademarks
**     Golden Code, FWD, any Golden Code or FWD logo, or any other trademarks
**     of Golden Code Development Corporation. You are not authorized to use the
**     name Golden Code, FWD, or the names of any author or contributor, for
**     publicity purposes without written authorization.
** 
**   2. No Misrepresentation of Affiliation.
** 
**     You may not represent yourself as Golden Code Development Corporation or FWD.
** 
**     You may not represent yourself for publicity purposes as associated with
**     Golden Code Development Corporation, FWD, or any author or contributor to
**     the covered work, without written authorization.
** 
**   3. No Misrepresentation of Source or Origin.
** 
**     You may not represent the covered work as solely your work.  All modified
**     versions of the covered work must be marked in a reasonable way to make it
**     clear that the modified work is not originating from Golden Code Development
**     Corporation or FWD.  All modified versions must contain the notices of
**     attribution required in this license.
*/

package com.goldencode.p2j.ui.client.format;


import com.goldencode.p2j.ui.client.*;
import com.goldencode.p2j.util.*;

/**
 * Format and editing for <code>logical</code> variables.
 */
public class BoolFormat
extends DisplayFormat
{
   /** FALSE presentation. */
   protected String fmtFalse;

   /** TRUE presentation. */
   protected String fmtTrue;

   /** Length of common part in FALSE and TRUE strings. */
   protected int fmtCommonLength;

   /** Lowercased common part of both strings. */
   protected String fmtCommonPart = null;

   /**
    * Constructs logical format from the Progress format string.
    * 
    * @param   pformat
    *          Progress format string.
    * 
    * @throws  DisplayFormatParsingException
    *          Format is invalid.
    */
   public BoolFormat(String pformat)
   throws DisplayFormatParsingException
   {
      super(pformat);

      String[] parts = logical.splitFormat(pformat);
      if (parts == null)
         return;
      
      fmtTrue = parts[0];
      fmtFalse = parts[1];

      int minLen = Math.min(fmtFalse.length(), fmtTrue.length());
      fmtCommonLength = 0;
      
      if (fmtTrue.compareToIgnoreCase(fmtFalse) != 0)
      {   
         while (fmtCommonLength < minLen &&
                (fmtFalse.charAt(fmtCommonLength) == 
                 fmtTrue.charAt(fmtCommonLength)))
         {
            fmtCommonLength++;
         }
      }

      if (fmtCommonLength > 0)
      {
         fmtCommonPart = fmtFalse.substring(0, fmtCommonLength).toLowerCase(); 
      }
      
      fmtScreenWidth = Math.max(fmtTrue.length(), fmtFalse.length());
   }

   /**
    * Indicates that ZAP operation can be performed on field with this format.
    * 
    * @return  Is ZAP should be performed for this format when it is possible
    *          (AUTO-ZAP set etc.).
    */
   @Override
   public boolean isZapSupported()
   {
      return false;
   }

   /**
    * Returns a new presentation instance with a default initialized insert
    * mode and cursor position.
    * 
    * @param   pres
    *          A Presentation instance which was constructed by this format.
    *          May be <code>null</code>. This presentation instance is not 
    *          changed in this method.
    * 
    * @return  New presentation instance.
    */
   @Override
   public Presentation clear(Presentation pres)
   {
      BoolBuf buf = new BoolBuf();

      if (pres != null)
      {
         buf.presInsertMode = pres.presInsertMode;
         pres = null;
      }

      return buf;
   }

   /**
    * Initializes a presentation for variable's value.
    * 
    * @param   var
    *          Variable which should be compatible with format.
    * 
    * @return  Data presentation
    */
   @Override
   public Presentation fromVar(BaseDataType var)
   {
      return new BoolBuf(new logical(var));
   }

   /**
    * Internal form of logical data.
    */
   private class BoolBuf
   extends Presentation
   {
      /**
       * Presentation chars. The same as screen value. This member may be
       * <code>null</code> which means UNKNOWN value. Unfortunately logical
       * editor may contain incomplete content so it is not possible to keep
       * only logical value.
       */
      StringBuffer bufChars;

      /**
       * Constructs clear presentation. This is FALSE value.
       */
      public BoolBuf()
      {
         presCursorPos = 0;
         presInsertMode = false;
         bufChars = new StringBuffer(fmtFalse);
      }

      /**
       * Constructs presentation initialized with given logical variable.
       * 
       * @param varLogical
       *        Initial value.
       */
      public BoolBuf(logical varLogical)
      {
         if (!varLogical.isUnknown())
         {
            bufChars = new StringBuffer(getScreenWidth());
            if (varLogical.getValue())
            {
               bufChars.append(fmtTrue);
            }
            else
            {
               bufChars.append(fmtFalse);
            }
         }
      }

      /**
       * Determine if this format must use the DEFAULT-FIXED-FONT, if no font is specified.
       * 
       * @return   Always <code>false</code>.
       */
      @Override
      public boolean isFixedFont()
      {
         return false;
      }
      
      /**
       * No action. Presentation is maintained to always be correct.
       */
      @Override
      public void checkFormat()
      {
         // dummy method
      }

      /**
       * Returns if the presentation currently represents the 
       * <code>unknown value</code>.  If any editing has occurred, then this
       * will return <code>false</code> even if the editing is not complete.
       *
       * @return   <code>true</code> if this is <code>unknown</code>.
       */
      @Override
      public boolean isUnknown()
      {
         return (bufChars == null);
      }

      /**
       * Erases one symbol before the cursor position.
       * 
       * @return   <code>true</code> if input is accepted (nevertheless it
       *           may be silently ignored in reality).
       */
      @Override
      public boolean deleteBack()
      {
         if (presCursorPos > 0)
         {
            presCursorPos--;
            return true;
         }
         
         return false;
      }

      /**
       * Erases one symbol in the cursor position.
       * 
       * @return   <code>true</code> if input is accepted (nevertheless it
       *           may be silently ignored in reality).
       */
      @Override
      public boolean deleteCurrent()
      {
         return false;
      }

      /**
       * Inputs character. The result depends from the current insert mode.
       * 
       * @param   ch
       *          Input character
       * 
       * @return   <code>true</code> if input is accepted (nevertheless it
       *           may be silently ignored in reality).
       */
      @Override
      public boolean input(char ch)
      {
         if (presCursorPos >= getScreenWidth())
         {
            return false;
         }

         // NULL value key
         if (ch == '?')
         {
            bufChars = null;
            presCursorPos = 0;
            return true;
         }

         if (bufChars == null)
         {
            if (presCursorPos != 0)
            {
               return false;
            }

            // restarting method from empty content
            bufChars = new StringBuffer(getScreenWidth());
            boolean res = input(ch);
            if (!res)
            {
               bufChars = null;
            }
            return res;
         }
         
         // comparisons must be case insensitive, prepare for this
         char   lowch  = Character.toLowerCase(ch);
         String lowbuf = bufChars.toString().toLowerCase();

         // enabling typing the same chars which are already in buffer or its
         // spaces complement
         if ((presCursorPos < lowbuf.length() &&
              lowch == lowbuf.charAt(presCursorPos)) ||
             (presCursorPos >= lowbuf.length() && lowch == ' '))
         {
            presCursorPos++;
            
            if (presCursorPos < getScreenWidth())
            {
               // still typing, suppress bells
               setLastCharEntered(false);
            }
            else
            {
               // done, enable auto-return
               setLastCharEntered(true);
            }
            
            return true;
         }

         // chars may be entered only into buffer or appended to it
         if (presCursorPos > lowbuf.length())
         {
            OutputManager.instance().beep();
            return false;
         }

         if (presCursorPos < fmtCommonLength)
         { 
            // we are in common part, no completion here
            if (lowch != fmtCommonPart.charAt(presCursorPos))
            {
               OutputManager.instance().beep();
               return false;
            }

            if (presCursorPos == 0)
            { 
               // buffer may contain "?"
               bufChars.setLength(0);
            }

            // we can be here only when cursorPos==bufChars.length() because
            // the case when cursorPos<=bufChars.length() is already handled
            bufChars.append(ch);
         }
         else if (presCursorPos == fmtCommonLength)
         { 
            // first char after common part, perform automatic completion
            String boolStr = guessBool(ch);
            
            if (boolStr == null)
            { 
               // char is absent in both of boolean strings
               OutputManager.instance().beep();
               return false;
            }

            bufChars.setLength(0);
            bufChars.append(boolStr);
            
            // if we are the last character in the field, remember that fact
            // so auto-return will work; note that the cursor pos is about to
            // be incremented so we work with the "updated" position and
            // compare it to the last char position
            if ((presCursorPos + 1) == getScreenWidth())
            {
               setLastCharEntered(true);
            }
         }
         else
         {
            // completion is already done but char is not from buffer. This
            // means that it is invalid
            OutputManager.instance().beep();
            return false;
         }

         presCursorPos++;
         return true;
      }

      /**
       * Moves cursor one character to the left.
       * 
       * @return   <code>true</code> if input is accepted (nevertheless it
       *           may be silently ignored in reality).
       */
      @Override
      public boolean moveCursorLeft()
      {
         if (presCursorPos > 0)
         {
            presCursorPos--;
         }
         return true;
      }

      /**
       * Moves cursor one character to the right. Also updates insert mode.
       * 
       * @return   <code>true</code> if input is accepted (nevertheless it
       *           may be silently ignored in reality).
       */
      @Override
      public boolean moveCursorRight()
      {
         if (presCursorPos < getScreenWidth() - 1)
         {
            presCursorPos++;
         }
         return true;
      }

      /**
       * Moves cursor to the beginning of the field.
       * 
       * @return   <code>true</code> if input is accepted (nevertheless it
       *           may be silently ignored in reality).
       */
      @Override
      public boolean moveCursorBegin()
      {
         presCursorPos = 0;
         return true;
      }

      /**
       * Moves cursor to the end of the field.
       * 
       * @return   <code>true</code> if input is accepted (nevertheless it
       *           may be silently ignored in reality).
       */
      @Override
      public boolean moveCursorEnd()
      {
         presCursorPos = getScreenWidth() - 1;
         return true;
      }

      /**
       * Indicates completion of an edit. The presentation instance may fill 
       * some unfilled places.
       * 
       * @return   Always <code>false</code>.
       */
      @Override
      public boolean finishEdit()
      {
         return false;
      }

      /**
       * Formats the data as a string that is ready for display.
       * 
       * @return   The formatted string.
       */
      @Override
      public String toScreenValue()
      {
         return (bufChars != null) ? bufChars.toString() : "?";
      }

      /**
       * Instantiates a variable that is equivalent to the presentation's
       * internal representation.
       * 
       * @param    var
       *           A variable which must be compatible with the format.
       */
      @Override
      public void toVar(BaseDataType var)
      {
         logical varLogical = (logical)var;
         
         if (bufChars == null)
         {
            varLogical.setUnknown();
         }
         else
         {
            String screen = toScreenValue();
            
            if (screen.equals(fmtTrue))
            {
               varLogical.setValue(true);
               return;
            }
            
            if (screen.equals(fmtFalse))
            {
               varLogical.setValue(false);
            }
            else
            {
               // This covers two cases: 1) equality; 2) incomplete string
               // e.g. just "o" for "on/off" format etc.
               varLogical.setValue(true);
            }
         }
      }

      /**
       * Strings for boolean values may have the same beginning. This method
       * resolves a boolean value based on the next char after this common
       * providing a form of automatic completion. 
       * 
       * @param    distinct
       *           The first char after the common part.
       * 
       * @return   The matched text or <code>null</code>.
       */
      private String guessBool(char distinct)
      {
         String boolStr = null;
         char   lowch   = Character.toLowerCase(distinct);
         
         if (fmtTrue.length() > fmtCommonLength)
         {
            char ch = Character.toLowerCase(fmtTrue.charAt(fmtCommonLength));
            
            if (lowch == ch)
            {
               boolStr = fmtTrue;
               return boolStr;
            }
         }
         
         if (fmtFalse.length() > fmtCommonLength)
         {
            char ch = Character.toLowerCase(fmtFalse.charAt(fmtCommonLength));
            
            if (lowch == ch)
            {
               boolStr = fmtFalse;
            }
         }
         
         return boolStr;
      }
   }
}