/*
** Module   : BrowseGuiImpl.java
** Abstract : GUI browse implementation
**
** Copyright (c) 2015-2018, Golden Code Development Corporation.
**
** -#- -I- --Date-- ---------------------------------Description---------------------------------
** 001 SVL 20151013 Initial version.
** 002 HC  20151013 Changes to extend GUI multi-window focus management and ACTIVE-WINDOW system
**                  handle processing to modal windows.
** 003 SVL 20151019 Layout procedure refactoring. Proper horizontal and vertical scrolling.
** 004 SVL 20151027 Fixed layout and re-layout (on column addition). Proper handling of scrolling
**                  keys. Added support for FIT-LAST-COLUMN.
** 005 SVL 20151202 Update column.WIDTH attribute after browse layout. Fixed right border of a
**                  fitted column.
** 006 SVL 20151230 Implemented editable browse.
** 007 EVL 20160224 Javadoc fixes to make compatible with Oracle Java 8 for Solaris 10.
** 008 IAS 20160229 LOAD-MOUSE-POINTER support.
** 009 CA  20160216 Cache the intermediate widget states.
** 010 SVL 20160314 Fixed clicking on a locked column in web version, Page Up/Down for editing
**                  browse. Fixed issues related to the case when a new edited row is out of the
**                  view. Added moveColumn.
** 011 VIG 20160107 Added parameters to horizontalScrollbar.onButtonClick() calls and to 
**                  positionUpdated() (scroll step type), fixed javadoc, fixed VerticalRowScroller.
** 012 SVL 20160412 Updated getViewportDimension, clipping rectangles. Added ColumnsScrollPane.
**                  Implemented column resizing.
** 013 SVL 20160608 VALUE-CHANGE event is fired on mouse click.
** 014 SVL 20160719 Updated translate() signature.
** 015 SVL 20160722 Implement browse multiple selection with Shift and Ctrl keys. Implemented
**                  scrolling for LONG step.
** 016 HC  20160712 Improved next/prev tab focus in nested frames.
** 017 EVL 20160705 Fix for thumb size precision loss in calculation.
** 018 EVL 20160915 Fix for vertical scrollbar enable state and thumb presenting when no data
**                  inside browse.
** 019 SBI 20160921 Fixed NPE caused by markersColumn.
** 020 SVL 20160922 Code refactoring because of switching from BrowseConfig.queryId to
**                  BrowseConfig.queryAssigned. Safety call of initQueryAssigned() in draw().
** 021 EVL 20161014 Fixing the left coordinate of the selection rectangle.  Row marker presence
**                  should include checking the browse widget config instance.
** 022 SVL 20161029 Query model initialization was removed from draw(). Fixed NPE in
**                  resolveColors().
** 023 SVL 20161031 Reverted H018.
** 024 SVL 20170126 Added support for pressed columns and START-SEARCH/END-SEARCH events.
**                  Fix for separator default color.
** 025 OM  20170320 Added support for Themes.
** 026 SVL 20170418 Sort numbers are drawn as images. Implemented search by key.
** 027 EVL 20170723 Adding browse specific installer and other changes for direct manipulation
**                  features.  Installing hover handler for browse widget must install the similar
**                  processing for browse column widgets to properly change mouse cursor.
** 028 SVL 20170814 Implemented server-side ROW-DISPLAY triggering, NO-SCROLLBAR-VERTICAL.
** 029 OM  20170713 LineBorderGuiImpl c'tor changed signature.
** 030 SVL 20170831 Support for toggle-box columns.
**     SVL 20170918 Fixed NPE in moveColumn().
** 031 EVL 20170824 Adding file list drag-and-drop support.
** 032 SVL 20171026 Fixes and changes related to in-browse editing.
** 033 SVL 20171116 Fixes for row and column resizing.
** 034 OM  20171124 TODO: validate realization changes.
**     CA  20171126 Changed afterConfigUpdate to work with the Browser.afterConfigUpdate.
** 035 OM  20171127 Called invokeRowEntryTrigger() when row changes.
**     SVL 20171204 Moved isCurrentRowInView to parent Browse class.
**     CA  20171218 Fixes for BROWSE/BROWSE-COLUMN:ENTRY/LEAVE, BROWSE:ROW-LEAVE/ENTER,
**                  BROWSE:VALUE-CHANGED events.
**     SVL 20171219 Perform layout for static browses earlier. Select the current row if the query
**                  was repositioned before the initial fetching of rows.
** 036 SVL 20180115 Locked columns set is properly created if NUM-LOCKED-COLUMNS was adjusted
**                  after the browse was created. Refactoring related to minimum column width.
**     SVL 20180126 Added beforeColumnHeaderPressed().
**     EVL 20180130 File header time signature changed to 2018.
** 037 HC  20180223 Prevented unexpected exception during an attempt to start edit mode on empty browse.
**
** 038 SVL 20180209 Added support for mouse right-click. Support for invisible columns.
** 039 SBI 20180319 Changed rowClicked to throw the default action on double click.
** 040 SVL 20180321 Data for invisible columns should be preserved on the client side too. Fix for
**                  disappearing markers column in enhanced mode.
** 050 EVL 20180405 Fix for widget resize for enclosing window restore event on the server side
**                  by attribute change.
** 051 HC  20180420 Implemented popup menu resolution in the tree of widgets.
** 052 HC  20180503 Fixed FILL-IN stops responding to mouse clicks after its popup menu
**                  displayed. Improved MENU focus management.
** 053 SBI 20180606 Changed to override ColumnSetContainer.getActualBounds() to update its hover
**                  widget region correctly.
** 054 OM  20180613 Select whole row when the marker is clicked.
**     EVL 20180614 Fix for stop editing mode when editor's FillIn widget is loosing focus.
**     EVL 20180627 Fix for toggle-box based editor issue with proper value commit on losing
**                  focus.  Now committing the change when leaving the browse (SE_LEAVE).
** 055 SVL 20180703 Fixed toggle-box missing the click when the browse gains the focus.
** 056 SVL 20180529 Implemented enhanced browse. Fixed non-clickable locked column set. Removed
**                  getEffectiveMousePointer() with default cursor.
** 057 OM  20180808 Fixed focus behaviour (TAB key handling) of non populated browses.
**         20180811 Fixed behaviour on mouse click events.
** 058 HC  20180813 Implemented initial version of dynamic layout.
** 059 OM  20180818 Added minimal support for flat Theme.
*/

/*
** This program is free software: you can redistribute it and/or modify
** it under the terms of the GNU Affero General Public License as
** published by the Free Software Foundation, either version 3 of the
** License, or (at your option) any later version.
**
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
** GNU Affero General Public License for more details.
**
** You may find a copy of the GNU Affero GPL version 3 at the following
** location: https://www.gnu.org/licenses/agpl-3.0.en.html
** 
** Additional terms under GNU Affero GPL version 3 section 7:
** 
**   Under Section 7 of the GNU Affero GPL version 3, the following additional
**   terms apply to the works covered under the License.  These additional terms
**   are non-permissive additional terms allowed under Section 7 of the GNU
**   Affero GPL version 3 and may not be removed by you.
** 
**   0. Attribution Requirement.
** 
**     You must preserve all legal notices or author attributions in the covered
**     work or Appropriate Legal Notices displayed by works containing the covered
**     work.  You may not remove from the covered work any author or developer
**     credit already included within the covered work.
** 
**   1. No License To Use Trademarks.
** 
**     This license does not grant any license or rights to use the trademarks
**     Golden Code, FWD, any Golden Code or FWD logo, or any other trademarks
**     of Golden Code Development Corporation. You are not authorized to use the
**     name Golden Code, FWD, or the names of any author or contributor, for
**     publicity purposes without written authorization.
** 
**   2. No Misrepresentation of Affiliation.
** 
**     You may not represent yourself as Golden Code Development Corporation or FWD.
** 
**     You may not represent yourself for publicity purposes as associated with
**     Golden Code Development Corporation, FWD, or any author or contributor to
**     the covered work, without written authorization.
** 
**   3. No Misrepresentation of Source or Origin.
** 
**     You may not represent the covered work as solely your work.  All modified
**     versions of the covered work must be marked in a reasonable way to make it
**     clear that the modified work is not originating from Golden Code Development
**     Corporation or FWD.  All modified versions must contain the notices of
**     attribution required in this license.
*/

package com.goldencode.p2j.ui.client.gui;

import java.awt.event.*;
import java.util.*;
import java.util.logging.*;

import com.goldencode.p2j.ui.*;
import com.goldencode.p2j.ui.chui.*;
import com.goldencode.p2j.ui.client.*;
import com.goldencode.p2j.ui.client.ScrollBar.ScrollStep;
import com.goldencode.p2j.ui.client.border.*;
import com.goldencode.p2j.ui.client.driver.*;
import com.goldencode.p2j.ui.client.event.*;
import com.goldencode.p2j.ui.client.event.FocusEvent;
import com.goldencode.p2j.ui.client.event.listener.FocusListener;
import com.goldencode.p2j.ui.client.gui.driver.*;
import com.goldencode.p2j.ui.client.gui.theme.*;
import com.goldencode.p2j.ui.client.layout.*;
import com.goldencode.p2j.ui.client.widget.*;
import com.goldencode.p2j.util.*;
import com.goldencode.p2j.util.ErrorManager;

import static com.goldencode.p2j.ui.BrowseColumnWidgetType.*;
import static com.goldencode.p2j.ui.BrowseWidget.*;
import static com.goldencode.p2j.ui.EnhancedBrowseParameter.CELLS_FONT;

/**
 * GUI implementation of BROWSE widget.
 */
public class BrowseGuiImpl
extends Browse<GuiOutputManager>
implements Hoverable,
           HasOptimalSize
{
   /** Top offset for text area in a cell. */
   static final int CELL_OFFSET_TOP = 1;

   /** Bottom offset for text area in a cell. */
   static final int CELL_OFFSET_BOTTOM = 2;

   /** Width of browse/column border. */
   public static final int BORDER_THICKNESS = 1;

   /** Width of the thumb of the horizontal scroll bar. */
   static final int H_BAR_THUMB_WIDTH = 16;

   /** Right offset for text area in a cell. */
   static final int DATA_OFFSET_RIGHT = 2;

   /** Width of a sort arrow displayed on column labels, in pixels. */
   public static final int SORT_ARROW_WIDTH = 9;

   /** Height of a sort arrow displayed on column labels, in pixels. */
   public static final int SORT_ARROW_HEIGHT = 5;

   /** Width of a sort number displayed on column labels, in pixels. */
   public static final int SORT_NUMBER_WIDTH = 4;

   /** Height of a sort number displayed on column labels, in pixels. */
   public static final int SORT_NUMBER_HEIGHT = 7;

   /** Minimum number of rows in a browse. */
   public static final int MIN_BROWSE_ROWS = 2;

   /** Minimum allowed height of a row during resize. */
   public static final int MIN_RESIZE_ROW_HEIGHT = 9;

   /** Offset from the bottom of the browse which limits the row height during resizing. */
   public static final int ROW_RESIZE_BOTTOM_OFFSET = 2;

   /**
    * Width of a resize area. See
    * {@link com.goldencode.p2j.ui.client.gui.BrowseColumnGuiImpl} enum ResizeArea for more
    * information.
    */
   static final int COLUMN_RESIZE_AREA_WIDTH = 5;

   /** Top or bottom offset for a column resize contour. */
   static final int RESIZE_LINE_VERTICAL_OFFSET = 3;

   /** Part of the cell the can be hidden at bottom. */
   private static final int ALLOWED_HIDDEN_CELL_BOTTOM_PART = 1;

   /** Default separator color. */
   private static final ColorRgb SEPARATOR_DEFAULT_COLOR = new ColorRgb(192, 192, 192);

   /**
    * Default separator color if background browse color matches {@link #SEPARATOR_DEFAULT_COLOR}.
    */
   private static final ColorRgb SEPARATOR_DEFAULT_COLOR2 = new ColorRgb(128, 128, 128);

   /** Default DOWN value for a dynamic browse. */
   private static final int DEFAULT_DYNAMIC_DOWN_SIZE = 4;

   /** Default width for a dynamic browse. */
   private static final int DEFAULT_DYNAMIC_WIDTH_CHARS = 30;

   /** Name of the sorting arrow up image. */
   private static final String SORT_ARROW_UP_IMAGE_NAME = "sort-arr-en-up";

   /** Name of the sorting arrow down image. */
   private static final String SORT_ARROW_DOWN_IMAGE_NAME = "sort-arr-en-down";

   /** Template of a sorting number image name. */
   private static final String SORT_NUMBER_IMAGE_NAME_TEMPLATE = "sort-num-%d";

   /** Logger. */
   private static final Logger LOG = LogHelper.getLogger(BrowseGuiImpl.class);

   /**
    * When the current row is out of the view below, then if we press UP key, the new current row
    * becomes <code>max(1, DOWN - UP_KEY_BOTTOM_OFFSET_ROWS)</code>th row in the viewport
    * (1-based).
    */
   private static final int UP_KEY_BOTTOM_OFFSET_ROWS = 2;

   /**
    * How many rows is scrolled per one mouse wheel rotation (at maximum).
    */
   private static final int MAX_SCROLL_ROWS_PER_WHEEL_ROTATION = 3;

   /** Height of a row resize area (of the part with overlaps with a cell). */
   private static final int ROW_RESIZE_AREA_HEIGHT = 4;

   /** Minimal width of a column, in pixels. */
   private static final int MIN_COL_WIDTH_PIXELS = 4;

   /** Cache of save targets for enhanced browse. */
   private static EnhancedBrowseSaveTarget[] enhancedSaveTargets = null;

   /** Screen driver */
   private GuiDriver gd;

   /** Font resolver for the browse title font. */
   private GuiFontResolver titleFontResolver;

   /** Font resolver. */
   private GuiFontResolver browseFontResolver;

   /**
    * Content pane for browse columns and data (scrollable part). Browse has the option called
    * NUM-LOCKED-COLUMNS, which defines the number of "pinned down" columns (starting from the
    * leftmost). These columns are not scrollable. The rest of the columns are scrollable, they go
    * "under" the locked columns when they are scrolled. If NUM-LOCKED-COLUMNS is adjusted,
    * columns are moved from {@link #lockedColumnSet} to <code>scrollableColumnSet</code> or vice
    * versa.
    */
   private ColumnSetContainer scrollableColumnSet;

   /**
    * Content pane for browse columns and data (locked part). Browse has the option called
    * NUM-LOCKED-COLUMNS, which defines the number of "pinned down" columns (starting from the
    * leftmost). These columns are not scrollable. The rest of the columns are scrollable, they go
    * "under" the locked columns when they are scrolled. If NUM-LOCKED-COLUMNS is adjusted,
    * columns are moved from <code>lockedColumnSet</code> to {@link #scrollableColumnSet} or vice
    * versa. This field can be <code>null</code>.
    */
   private ColumnSetContainer lockedColumnSet;

   /**
    * Scrollable wrapper for {@link #scrollableColumnSet}, used for handling empty (it may be
    * colored) area between the rightmost column and the vertical scroller.
    * See {@link ColumnsScrollPane} for more information.
    */
   private ScrollPaneGuiImpl scrollPane;

   /** Browse color resolver. */
   private GuiColorResolver colorResolver;

   /** Browse title height (without title borders). */
   private int titleHeight;

   /** Coordinates conversion. */
   private CoordinatesConversion cc;

   /** Separator color. */
   private ColorRgb separatorColor;

   /** Background color. */
   private ColorRgb backgroundColor;

   /**
    * Row height (in pixels) excluding borders and separators. <code>paddedRowHeight =
    * config.rowHeightPixels + CELL_OFFSET_TOP + CELL_OFFSET_BOTTOM</code>.
    */
   private int paddedRowHeight;

   /** Title panel. */
   private BorderedPanelGuiImpl titlePanel;
   
   /** Total columns width. */
   private int totalColumnsWidth;

   /** Total width of the locked columns. */
   private int lockedColumnsWidth;

   /** Horizontal scrolling step for non-COLUMN-SCROLLING mode. */
   private int hScrollingStep;

   /** Markers column (leftmost). */
   private BrowseColumnGuiImpl markersColumn;

   /** Horizontal scrollbar. */
   private ScrollBarGuiImpl horizontalScrollbar;

   /** Vertical scrollbar. */
   private ScrollBarGuiImpl verticalScrollbar;

   /**
    * Maximum for horizontal scrolling in column scrolling mode. Depends on the number of last
    * columns that fit into the view.
    */
   private int horizontalColumnScrollMax = 0;

   /** If browse gained focus by mouse click rather than other ways. */
   private boolean mouseFocusGained = false;

   /** Vertical scroller. */
   private ScrollBarControllerGuiImpl verticalScroller;

   /** Horizontal scroller. */
   private ScrollBarControllerGuiImpl horizontalScroller;

   /** Reference to the column which is being dragged. */
   private BrowseColumnGuiImpl draggedColumn = null;

   /** Reference to the column which is being resized. */
   private BrowseColumnGuiImpl resizedColumn = null;

   /**
    * Offset from the left edge of the column which is being dragged to the position of the
    * pointer inside the column.
    */
   private int leftDragColumnMouseOffset = 0;

   /**
    * Offset from the left edge of the browse (or the edge of the markers column) to the left
    * edge of the column which is being dragged.
    */
   private int leftDragColumnOffset = 0;

   /**
    * Offset from the left edge of the browse to the left edge of the contour of the column which
    * is being resized.
    */
   private int leftResizeColumnOffset = 0;

   /**
    * Offset from the left edge of the browse to the right edge of the contour of the column which
    * is being resized.
    */
   private int rightResizeColumnOffset = 0;

   /** Cached sort arrow up image. */
   private ImageWrapper sortArrowUpImage = null;

   /** Cached sort arrow down image. */
   private ImageWrapper sortArrowDownImage = null;

   /** Cached images of sort numbers. */
   private ImageWrapper[] sortNumberImages = null;

   /**
    * <code>true</code> if initial reposition was done (i.e. reposition on opening of the first
    * assigned query). Used in handling of the selection state of the first row on initial display
    * of a single-selection browse.
    */
   private boolean initialRepositionDone = false;

   /** <code>true</code> if the browse is in layout mode. */
   private boolean browseLayoutMode = false;

   /** Undo configuration for enhanced browse. */
   private EnhancedBrowseConfig undoConfig = null;

   /**
    * All hyperlink widgets for the current view. Keyed by column and viewport row index.
    */
   private Map<BrowseColumnGuiImpl, Map<Integer, Hyperlink>> hyperlinkWidgets = null;
   
   /**
    * The column that generates the focus events. It is possible to get multiple focus gain 
    * notifications. The source column is reset when events generated by mouse events on it are
    * not dispatched any more.
    */
   private BrowseColumnGuiImpl meColumn = null;
   
   /**
    * Constructs the Browse widget.
    * <p>
    * Call {@link #initialize} to finish the widget initialization.
    *
    * @param id
    *        The widget's ID.
    * @param cfg
    *        Browse config,
    */
   public BrowseGuiImpl(WidgetId id, BrowseConfig cfg)
   {
      super(id, cfg);

      gd = (GuiDriver) OutputManager.getDriver();
      cc = screen().coordinates();
   }

   /**
    * Replacement API for the logic which should exist in the widget's c'tor, but the logic
    * can't be executed as it depends on the widget being registered into the official registry.
    * <p>
    * This API is called after the widget is created and added to the registry.
    *
    * @param    id
    *           The widget's ID.
    * @param    wcfg
    *           The config used to initialize this widget.
    */
   @Override
   public void initialize(WidgetId id, BrowseConfig wcfg)
   {
      super.initialize(id, wcfg);
      if (wcfg != null && wcfg.visible)
      {
         _setVisible(true); // sets config.realized
      }

      if (config.performLayout &&
          !config.dynamic)          // limit to static browse for now
      {
         doLayout();
      }
   }
   
   /**
    * Set container size.
    *
    * @param   width
    *          New width.
    * @param   height
    *          New height.
    */
   @Override
   public void setSize(double width, double height)
   {
      originalWidth = width;
      originalHeight = height;

      if (width == 0 || height == 0)
         return;

      // TODO set ilFirstScrollCol ?

      super.setSize(width, height);

      config.widthPixels = physicalDimension().width;
      config.heightPixels = physicalDimension().height;

      config.widthChars = dimension().width;
      config.heightChars = dimension().height;

      if (ThinClient.isDynamicLayoutEnabled())
      {
         doLayout();
      }
   }

   /**
    * The method is called after the configuration associated with the
    * owner changes. The parameter points to a reference representing
    * the original configuration state. This reference can be used to
    * detect what configuration fields changed and so optimize any
    * processing related to the change.
    * <p>
    * The implemented logic must be idempotent - the method must produce
    * the same results if executed once or multiple times not relying
    * on any particular call sequence or related state. This is
    * necessary so that the method can be repeated or retried as often as
    * necessary without causing unintended effects
    *
    * @param   beforeUpdate
    *          Config reference capturing the config state before
    *          the owner's config was modified.
    */
   @Override
   public void afterConfigUpdate(BrowseConfig beforeUpdate)
   {
      super.afterConfigUpdate(beforeUpdate);

      if (config.queryAssigned)
      {
         initQueryAssigned();
      }

      Window window = window();
      boolean graphicsAvail = gd != null && window != null;
      if (graphicsAvail)
      {
         if (isVisible() &&
             config.widthChars > 0 && config.heightChars > 0  &&
             beforeUpdate.widthChars > 0 && beforeUpdate.heightChars > 0 &&
             (config.widthChars != beforeUpdate.widthChars ||
              config.heightChars != beforeUpdate.heightChars))
         {
            // browse has been resized
            doLayout();
         }

         updateBrowseDimensions();
      }

      setSelectionMultiple(config.multiple);
      setReadOnly(config.readOnly);

      if (!config.performLayout)
      {
         resolveFonts();
         if (colorResolver != null  // init colorResolver for the first time in doLayout
             && graphicsAvail)
         {
            gd.withSelectedWindow(window.getId().asInt(), this::resolveColors);
         }

         if (beforeUpdate.numLockedColumns != config.numLockedColumns)
         {
            notifyColumnsUpdated();
         }

         if (config.realized)
         {
            if (beforeUpdate.columnScrolling != config.columnScrolling)
            {
               setColumnScrolling(config.columnScrolling);
            }

            if (beforeUpdate.fitLastColumn != config.fitLastColumn ||
                beforeUpdate.noEmptySpace  != config.noEmptySpace)
            {
               fitLastColumn();
            }

            if (beforeUpdate.rowHeightPixels > 0 &&   // make sure it is row height update case
                                                      // rather than an update due to initial row
                                                      // height calculation
                beforeUpdate.rowHeightPixels != config.rowHeightPixels)
            {
               updateRowHeight();
            }
         }
      }

      if (beforeUpdate.rowHeightPixels != config.rowHeightPixels &&
          beforeUpdate.rowHeightPixels > 0 &&
          config.rowHeightPixels <= 0)
      {
         // deny default row height sent from server, if it was already calculated in doLayout on
         // the client side
         config.rowHeightPixels = beforeUpdate.rowHeightPixels;
         config.rowHeightChars = beforeUpdate.rowHeightChars;
      }

      // GUI allows to switch labels visibility after the browse has been realized
      if (!config.performLayout && config.labels != beforeUpdate.labels)
      {
         // TODO fix minor issues
         updateLabelsVisible();
         config.down = calcDown();
         notifyColumnsUpdated();
      }
   }

   /**
    * Get attribute container for the given cell.
    *
    * @param  viewportIndex
    *         0-based row index in the viewport.
    * @param  columnOrdinal
    *         0-based column index.
    *
    * @return attribute container for the given cell.
    */
   public GuiCellAttributes getCellAttr(int viewportIndex, int columnOrdinal)
   {
      return getCellAttr(dcData, viewportIndex, columnOrdinal);
   }

   /**
    * Get attribute container for the given cell from the cache containing intermediate rows
    * ({@link #intermediateDcData}), if it is available, or normat cache ({@link #dcData})
    * otherwise.
    *
    * @param  viewportIndex
    *         0-based row index in the viewport.
    * @param  columnOrdinal
    *         0-based column index.
    *
    * @return attribute container for the given cell.
    */
   public GuiCellAttributes getCellAttrWithIntermediate(int viewportIndex, int columnOrdinal)
   {
      return intermediateDcData != null ?
               getCellAttr(intermediateDcData, viewportIndex, columnOrdinal) :
               getCellAttr(viewportIndex, columnOrdinal);
   }

   /**
    * Get attribute container for the given cell from the specified row cache.
    *
    * @param  rowCache
    *         Cache of rows: {@link #dcData} or {@link #intermediateDcData}.
    * @param  viewportIndex
    *         0-based row index in the viewport.
    * @param  columnOrdinal
    *         0-based column index.
    *
    * @return attribute container for the given cell.
    */
   public GuiCellAttributes getCellAttr(BrowseRow[] rowCache, int viewportIndex, int columnOrdinal)
   {
      if (rowCache == null || viewportIndex >= rowCache.length)
      {
         return null;
      }

      return (GuiCellAttributes) rowCache[viewportIndex].attr[columnOrdinal];
   }

   /**
    * Notify the browse that some columns have been added to it or visibility of columns has been
    * changed.
    */
   @Override
   public void notifyColumnsUpdated()
   {
      int scrollingPosition = 0;
      if (!config.performLayout && horizontalScrollbar.isVisible())
      {
         scrollingPosition = horizontalScroller.getPosition();
      }

      setSize(originalWidth, originalHeight);

      if (!config.performLayout)
      {
         boolean wasHorizontalScrollbar = totalColumnsWidth > getViewportDimension().width;
         rebuildColumns();
         boolean needHorizontalScrollbar = totalColumnsWidth > getViewportDimension().width;

         if (wasHorizontalScrollbar != needHorizontalScrollbar)
         {
            showHorizontalScrollbar(needHorizontalScrollbar);

            if (config.heightChars <= 0)
            {
               // implicit height - increase browse height
               //TODO fix drawing issue
               int barHeight = horizontalScrollbar.physicalDimension().height;
               setHeight(cc.heightFromNative(physicalDimension().height + barHeight));
               scrollPane.setHeight(
                     cc.heightFromNative(scrollPane.physicalDimension().height + barHeight));

            }
            else
            {
               // explicit height - recalculate DOWN
               config.down = calcDown();
            }

            updateColumnSetsHeight();
         }

         setColumnScrolling(config.columnScrolling);

         fitLastColumn();
      }
      else
      {
         resetColumnsCache();
         browseModel.resetEditorsConfigsCache();
      }

      refresh(false);

      if (!config.performLayout &&
          horizontalScrollbar.isVisible() &&
          scrollingPosition <= horizontalScrollbar.getMax())
      {
         horizontalScrollbar.positionUpdated(scrollingPosition, ScrollStep.CUSTOM, false);
      }

      repaint();
   }

   /**
    * Set "intermediate" rows (i.e. the rows that were already fetched in the process of getting
    * new rows by
    * {@link BrowseWidget#getRows(int, int, boolean, int, ScreenBuffer[], RowDisplayStrategy)}),
    * so these rows can be drawn if a pause occurs in ROW-DISPLAY trigger.
    *
    * @param rows
    *        Rows to set.
    */
   @Override
   public void setIntermediateRows(BrowseRow[] rows)
   {
      if (rows == null)
      {
         return;
      }

      if (currentRowDisplayStrategy == RowDisplayStrategy.INCREMENTAL_SCROLL_FWD ||
          currentRowDisplayStrategy == RowDisplayStrategy.INCREMENTAL_SCROLL_BACK)
      {
         int down = getDown();
         int dcSize = dcData == null ? 0 : dcData.length;
         intermediateDcData = new BrowseRow[Math.min(down, dcSize + rows.length)];
         int dcCopyLen = Math.min(dcSize, intermediateDcData.length - rows.length);

         if (currentRowDisplayStrategy == RowDisplayStrategy.INCREMENTAL_SCROLL_FWD)
         {
            if (dcData != null)
            {
               System.arraycopy(dcData, dcSize - dcCopyLen, intermediateDcData, 0, dcCopyLen);
            }
            System.arraycopy(rows, 0, intermediateDcData, dcCopyLen, rows.length);
         }
         else
         {
            // copy in reverse order
            for (int i = 0; i < rows.length; i++)
            {
               intermediateDcData[rows.length - i - 1] = rows[i];
            }

            if (dcData != null)
            {
               System.arraycopy(dcData, 0, intermediateDcData, rows.length, dcCopyLen);
            }
         }
      }

      //TODO other ROW-DISPLAY strategies
   }

   /**
    * Notification of a mouse clicked event occurred for this widget.
    * <p>
    * If the widget is not already focused, it forces the focus to it.  After the focus was
    * received, it computes the caret location, based on the mouse position over the widget's
    * content.
    *
    * @param    e
    *           The mouse event.
    */
   @Override
   public void mouseClicked(MouseEvent e)
   {
      super.mouseClicked(e);

      if (e.getButton() == MouseEvent.BUTTON1)
      {
         if (!hasFocus())
         {
            requestFocus();
         }
      }
   }

   /**
    * Notification of a mouse dragged event occurred for this widget. 
    *
    * @param    e
    *           The mouse event.
    */
   @Override
   public void mouseDragged(MouseEvent e)
   {
      if (!isEnabled() || (!hasFocus() && !isEditing()))
      {
         return;
      }

      if (isColumnDrag())
      {
         NativePoint point = translate(e);
         leftDragColumnOffset = point.x - leftDragColumnMouseOffset;
         
         TopLevelWindow<GuiOutputManager> wnd = topLevelWindow();

         int leftLimit = getLeftColumnContourLimit();
         if (leftDragColumnOffset < leftLimit)
         {
            leftDragColumnOffset = leftLimit;
            horizontalScrollbar.onButtonClick(ScrollBar.Position.LEFT, 
                                              ScrollStep.SHORT, 
                                              new MouseEvt(e, wnd));
         }
         else
         {
            int rightLimit = getRightColumnContourLimit();
            int colWidth = draggedColumn.physicalDimension().width;
            if (leftDragColumnOffset + colWidth > rightLimit)
            {
               leftDragColumnOffset = rightLimit - colWidth;
               horizontalScrollbar.onButtonClick(ScrollBar.Position.RIGHT, 
                                                 ScrollStep.SHORT, 
                                                 new MouseEvt(e, wnd));
            }
         }

         repaint();
      }
      else if (isColumnResize())
      {
         int pointerX = translate(e).x;
         int leftLimit = leftResizeColumnOffset + getEffectiveMinColumnWidth();

         if (pointerX < leftLimit)
         {
            pointerX = leftLimit;
         }
         else
         {
            int rightLimit = getRightColumnContourLimit();
            if (pointerX > rightLimit)
            {
               pointerX = rightLimit;
            }
         }

         rightResizeColumnOffset = pointerX;

         repaint();
      }
   }
   
   /**
    * Notification of a mouse wheel moved event occurred for this widget.  No-op by default.
    *
    * @param    e
    *           The mouse event.
    */
   @Override
   public void mouseWheelMoved(MouseWheelEvent e)
   {
      if (getCacheSize() == 0)
      {
         return;
      }

      int down = getDown();
      int scrollRowsPerRotation = down <= MAX_SCROLL_ROWS_PER_WHEEL_ROTATION ?
                                  down - 1 :
                                  MAX_SCROLL_ROWS_PER_WHEEL_ROTATION;

      int rows = e.getWheelRotation() * scrollRowsPerRotation;
      if (rows == 0)
      {
         return;
      }

      if (rows > 0 && getCacheSize() < getDown())
      {
         return;
      }

      incrementalScroll(rows, false, true);
      repaint();
   }
   
   /**
    * Get the height of the title, if any.
    * 
    * @return  the title's height.
    */
   public int getTitleHeight()
   {
      return titleHeight;
   }
   
   /**
    * Lay out the browse.
    */
   @Override
   public void doLayout()
   {
      // TODO update firstEditable column

      config.performLayout = false;

      resolveFonts();
      resolveColors();
      updateLabelsVisible();

      paddedRowHeight = calcRowHeight();
      config.rowHeightPixels = paddedRowHeight - CELL_OFFSET_TOP - CELL_OFFSET_BOTTOM;
      config.rowHeightChars = cc.heightFromNative(config.rowHeightPixels);
      titleHeight = calcTitleHeight();

      buildPanels();

      config.down = calcDown();

      if (config.down < MIN_BROWSE_ROWS)
      {
         ErrorManager.displayError(3520,
               String.format("BROWSE widget must have at least %d DOWN", MIN_BROWSE_ROWS), false);
         ErrorManager.displayError(4025,
               String.format("Unable to realise BROWSE %s", config.name), true);

         // TODO 4GL doesn't realise widget at this point, we leave standard row size

         config.rowHeightChars = -1;
         config.rowHeightPixels = -1;
         paddedRowHeight = calcRowHeight();;
         config.rowHeightPixels = paddedRowHeight - CELL_OFFSET_TOP - CELL_OFFSET_BOTTOM;
         config.rowHeightChars = cc.heightFromNative(config.rowHeightPixels);
         config.down = calcDown();
      }

      // MIN-HEIGHT-PIXELS equals the size of the browse with the same fonts and characteristics,
      // but with two rows.
      int minHeightPixels = (config.box ? 2 * BORDER_THICKNESS : 0) +
                            (hasTitle() ? titleHeight + BORDER_THICKNESS : 0) +
                            (drawLabels ? getLabelsAreaHeight() : 0) +
                            getRowHeight() * MIN_BROWSE_ROWS +
                            BORDER_THICKNESS * (MIN_BROWSE_ROWS - 1) +
                            (horizontalScrollbar.isVisible() ?
                                 horizontalScrollbar.physicalDimension().height : 0);
      config.minHeightChars = cc.heightFromNative(minHeightPixels);

      createTitlePanel();

      setColumnScrolling(config.columnScrolling);

      // Update column.WIDTH attribute.
      int colNo = getColumnCount();
      for (int i = 0; i < colNo; i++)
      {
         BrowseColumnGuiImpl column = getColumn(i);

         int width = column.physicalDimension().width;
         width -= column.getColumnPadding();
         column.config().widthChars = cc.widthFromNative(width);
         column.config().widthPixels = width;
      }

      stretchColumns();
   }

   /**
    * Get row height (in pixels), excluding borders and separators.
    *
    * @return see above.
    */
   public int getRowHeight()
   {
      return paddedRowHeight;
   }

   /**
    * Get height of the labels area (in pixels), excluding borders.
    *
    * @return see above.
    */
   public int getLabelsHeight()
   {
      return config.labelsHeightPixels;
   }

   /**
    * Get height of the labels area (in pixels), including borders.
    *
    * @return see above.
    */
   public int getLabelsAreaHeight()
   {
      return getLabelsHeight() + 2 * BORDER_THICKNESS;
   }

   /**
    * Number of lines in the labels area (matches the number of lines in the tallest multi-line
    * label).
    *
    * @return number of lines in the labels area.
    */
   public int getLabelLinesNum()
   {
      return ilLabelsHeight;
   }

   /**
    * Sets position of terminal's cursor. Redirects this positioning to the in-place editor if it
    * is present. Otherwise does nothing because BROWSE itself has no purpose for the hardware
    * cursor.
    */
   @Override
   public void drawCaret()
   {

   }

   /**
    * Draws the widget: border, header, selection marks and line, internal FillIn etc.
    */
   @Override
   public void draw()
   {
      Window window = window();
      if (window == null)
      {
         return;
      }

      if (config.performLayout)
      {
         doLayout();
      }

      if (!config.realized || !isVisible())
      {
         return;
      }

      // colors must be resolved here, to reflect latest state of the widget (enabled/disabled,etc)
      resolveColors();
      
      NativePoint p = physicalLocation();
      final NativeDimension d = physicalDimension();

      final NativeRectangle clip = clipRectangle(p, d);

      gd.draw(getId(), p, clip, () ->
      {
         if (titlePanel != null)
         {
            titlePanel.draw();
         }

         // draw browse contents
         scrollPane.draw();
         
         if (lockedColumnSet != null)
         {
            lockedColumnSet.draw();
         }
         
         Theme theme = ThemeManager.getCurrentTheme();
         
         if (hasFocus() && isEnabled() && !isEditing())
         {
            theme.drawBrowseRowHighlight(this, gd, 0, 0, d.width, d.height);
         }
         
         if (isColumnDrag())
         {
            NativeDimension viewport = getViewportDimension();
            theme.drawBrowseColumnDrag(this,
                                      gd,
                                      leftDragColumnOffset,
                                      hasTitle() ? titleHeight : 0,
                                      draggedColumn.physicalDimension().width,
                                      viewport.height);
         }
         else if (isColumnResize())
         {
            NativeDimension viewport = getViewportDimension();
            
            int topY = (hasTitle() ? titleHeight : 0) +
                        (config.box ? BORDER_THICKNESS : 0) + RESIZE_LINE_VERTICAL_OFFSET;
            int bottomY = viewport.height + topY - 2 * RESIZE_LINE_VERTICAL_OFFSET;
            
            theme.drawBrowseResizingColumn(this,
                                           gd,
                                           leftResizeColumnOffset,
                                           topY,
                                           rightResizeColumnOffset,
                                           bottomY);
         }

         if (isLayoutMode())
         {
            // draw red frame indicating layout mode for enhanced browse
            gd.setColor(new ColorRgb(255, 0, 0));
            gd.setLineStroke(LineStroke.DOTTED_LARGE);
            for (int i = 0; i < 3; i++)
            {
               int y = i;
               gd.drawLine(i, y, physicalDimension().width, y);

               int x = physicalDimension().width - 1 - i;
               gd.drawLine(x, i, x, physicalDimension().height);

               y = physicalDimension().height - 1 - i;
               gd.drawLine(physicalDimension().width - i, y, 0, y);

               x = i;
               gd.drawLine(x, physicalDimension().height - i, x, 0);
            }
            gd.setLineStroke(LineStroke.DEFAULT);
         }
      });
   }

   /**
    * Destroy this widget.  Notify all of its components that the widget has been destroyed.
    */
   @Override
   public void destroy()
   {
      if (scrollPane != null)
      {
         OutputManager.instance().getRegistry().removeWidget(scrollPane.getId());
      }

      if (titlePanel != null)
      {
         OutputManager.instance().getRegistry().removeWidget(titlePanel.getId());
      }

      super.destroy();
   }

   /**
    * Handle situation when <code>Browse</code> widget receives the focus.
    * <p>
    * When a help string is defined for the browse it will be displayed on
    * the status line.  This method also sets the selected attribute on.
    *
    * @param    event
    *           Reference to the event.
    */
   @Override
   public void onFocusGained(FocusEvent event)
   {
      if (isEditPossible() && !mouseFocusGained && getCacheSize() > 0)
      {
         setCurrentSingleRowSelected();
         startEdit(event.other() == null ? null : event.other().getId());
      }

      super.onFocusGained(event);
   }

   /**
    * Handle situation when <code>Browse</code> widget loses the focus.
    * <p>
    * This method clears the focused flag.
    *
    * @param    event
    *           Reference to the event.
    */
   @Override
   public void onFocusLost(FocusEvent event)
   {
      leaveRow();
      super.onFocusLost(event);
   }

   /**
    * Process direct manipulation event.
    * 
    * @param   event
    *          Event to process.
    */
   @Override
   public void processDirectManipulationEvent(DirectManipulationEvent event)
   {
      int id = event.id();
      
      // row resize messages first
      if (id == Keyboard.SE_START_ROW_RESIZE)
      {
      }
      else if (id == Keyboard.SE_END_ROW_RESIZE)
      {
         int change = event.getYChange();
         if (change != 0)
         {
            int newRowHeight = config.rowHeightPixels + event.getYChange();
            if (calcDown(newRowHeight) < MIN_BROWSE_ROWS)
            {
               ErrorManager.displayError(6422, String.format(ERROR_TEXT_6422,
                                                             "ROW-HEIGHT-PIXELS",
                                                             config.name,
                                                             MIN_BROWSE_ROWS));
            }
            else
            {
               // commit new height
               BrowseConfig beforeUpdate = new BrowseConfig();
               beforeUpdate.applyConfig(config);
               config.rowHeightPixels = newRowHeight;
               config.rowHeightChars = cc.heightFromNative(newRowHeight);

               // refresh browse
               afterConfigUpdate(beforeUpdate);
            }
         }
      }
      else
      {
         super.processDirectManipulationEvent(event);
      }
   }
   
   /**
    * Enable/disable widget.
    * 
    * @param   enabled
    */
   @Override
   public void setEnabled(boolean enabled)
   {
      super.setEnabled(enabled);
      // the code below propagates the hoverable handlers install/remove for browse columns
      if (!config.performLayout)
      {
         int colNum = getColumnCount();
         for (int i = 0; i < colNum; i++)
         {
            getColumn(i).setEnabled(enabled);
         }
      }
   }

   /**
    * Resizes visible columns to utilize all the available width of the columns container.
    * The columns are resized proportionally to the individual column data widths.
    */
   void stretchColumns()
   {
      if (!ThinClient.isDynamicLayoutEnabled())
      {
         return;
      }

      int totalWidth = 0;
      int totalDataWidth = 0;

      HashMap<BrowseColumnGuiImpl, Integer> dataWidths = new HashMap<>();
      HashMap<BrowseColumnGuiImpl, Integer> widths = new HashMap<>();

      int colNum = getVisibleColumnCount();
      for (int i = 0; i < colNum; i++)
      {
         BrowseColumnGuiImpl column = getVisibleColumn(i);

         int width = column.physicalDimension().width;
         totalWidth += width;

         // only stretch the unlocked set of columns
         if (i < config.numLockedColumns)
         {
            continue;
         }

         widths.put(column, width);

         int colDataWidth = column.calcDataWidth(true);
         dataWidths.put(column, colDataWidth);
         totalDataWidth += colDataWidth;
      }

      int columnSetLocation = scrollableColumnSet.physicalLocation().x;
      int viewportWidth = getViewportDimension().width;
      int availableWidth = viewportWidth - columnSetLocation;
      totalWidth = 0;

      double minRatio = DynamicLayoutConfig.getInstance().getMinColumnWidthRatio();

      for (int i = 0; i < colNum; i++)
      {
         BrowseColumnGuiImpl column = getVisibleColumn(i);

         // only stretch the unlocked set of columns
         if (i < config.numLockedColumns)
         {
            totalWidth += column.physicalDimension().width;
            continue;
         }

         int dataWidth = dataWidths.get(column);
         double minWidth = cc.widthFromNative((int) (minRatio * dataWidth));

         if (i == colNum -1)
         {
            // rounding error can make the column spill over the remaining available width
            column.setWidth(Math.max(cc.widthFromNative(availableWidth - totalWidth), minWidth));
            totalWidth += column.physicalDimension().width;
            break;
         }

         BrowseColumnConfig c = column.config();
         int newWidth = (int) (availableWidth * ((double) dataWidth / totalDataWidth));
         column.setWidth(Math.max(cc.widthFromNative(newWidth), minWidth));
         totalWidth += column.physicalDimension().width;
      }

      totalColumnsWidth = totalWidth + lockedColumnsWidth;
   }
   
   /**
    * Activates direct manipulation engine for this widget.
    */
   @Override
   protected void activateDirectManipulation()
   {
      if (!config.performLayout)
      {
         BrowseConfig brc = config();
         if (isRowResizable() || brc.selectable)
         {
            super.activateDirectManipulation();

            int colNum = getColumnCount();
            for (int i = 0; i < colNum; i++)
            {
               BrowseColumnGuiImpl column = getColumn(i);
               column.config().selectable = brc.selectable;
               ((GuiDriver) screen().getDriver()).registerDirectManipulation(column);
            }
         }
      }
   }
   
   /**
    * Deactivates direct manipulation engine for this widget.
    */
   @Override
   protected void deactivateDirectManipulation()
   {
      BrowseConfig brc = config();
      if (!isRowResizable() && !brc.selectable)
      {
         super.deactivateDirectManipulation();

         if (lockedColumnSet != null)
         {
            for (Widget w : lockedColumnSet.widgets())
            {
               MouseHandler.deregisterWidget(w.getId().asInt());
            }
         }

         for (Widget w : scrollableColumnSet.widgets())
         {
            MouseHandler.deregisterWidget(w.getId().asInt());
         }
      }
   }
   
   /**
    * Checks if direct manipulation feature must be installed or removed in GUI mode.  If config
    * is upadting in a way that stops using direct manipulation it must be removed from system.
    * Different widgets can use different criteria to make this decision.
    * 
    * @return   <code>TRUE</code> if direct manipulation is need to be installed,
    *           <code>FALSE</code> if direct manipulation is need to be removed.
    */
   @Override
   protected boolean isDirectManipulationRequired()
   {
      BrowseConfig brc = config();
      return brc.selectable || isRowResizable();
   }
   
   /**
    * Get viewport index of the row which corresponds given vertical position.
    *
    * @param  y
    *         Vertical position (in native units in browse coordinates) that is supposed to be
    *         within one of the viewport rows.
    *
    * @return 0-based viewport index of the row which corresponds the given vertical position.
    *         It can be <code>-1</code> if position is higher than the first row or can be more
    *         than the index of the last row if position is lower than the last row.
    */
   public int getRowByCoordinates(int y)
   {
      if (isLabelsVisible())
      {
         y -= getLabelsAreaHeight();
      }

      if (y < 0)
      {
         return -1;
      }
      else
      {
         return y / (getRowHeight() + BORDER_THICKNESS);
      }
   }
   
   /**
    * Get Y coordinate minus possible labels height.
    *
    * @param  y
    *         Vertical position (in native units in browse coordinates) that is supposed to be
    *         within one of the viewport rows.
    *
    * @return Vertical position without optional label height in case of labels exist or
    *         unchanged position if no labels.
    */
   public int fixupNoLabelsY(int y)
   {
      return isLabelsVisible() ? y - getLabelsAreaHeight() : y;
   }

   /**
    * If the pointer is currently positioned over a row resize area, return that area.
    *
    * @param  browseColumnPoint
    *         Coordinates of the mouse pointer relatively the column being hovered.
    *
    * @return {@link BrowseColumnGuiImpl.ResizeArea#UPPER} if the pointer is positioned over the
    *         upper row resize area of a cell. {@link BrowseColumnGuiImpl.ResizeArea#LOWER} if the
    *         pointer is positioned over the lower row resize area of a cell. <code>null</code>
    *         otherwise.
    */
   public BrowseColumnGuiImpl.ResizeArea getRowResizeArea(NativePoint browseColumnPoint)
   {
      int y = browseColumnPoint.y;
      int rowNum = getRowByCoordinates(y);
      if (rowNum > -1)
      {
         y = fixupNoLabelsY(y);
         if (y < ROW_RESIZE_AREA_HEIGHT +
             rowNum * (getRowHeight() + BrowseGuiImpl.BORDER_THICKNESS))
         {
            return rowNum > 0 ? BrowseColumnGuiImpl.ResizeArea.UPPER : null;
         }
         else if (y >= (rowNum + 1) * (getRowHeight() + BrowseGuiImpl.BORDER_THICKNESS) -
               ROW_RESIZE_AREA_HEIGHT - BrowseGuiImpl.BORDER_THICKNESS)
         {
            return BrowseColumnGuiImpl.ResizeArea.LOWER;
         }
      }

      return null;
   }

   /**
    * Gets scrollpane container to draw something in coordinates related to it.
    *
    * @return Scrollpane containing browse columns.
    */
   public ScrollPaneGuiImpl getScrollPane()
   {
      return scrollPane;
   }
   
   /**
    * Gets vertical position of the given row related to the browse column vertical start.
    *
    * @param  rowNdx
    *         0-base row index to compute the Y coordinate.
    *
    * @return Y coordinate of the top border for the row specified.
    */
   public int getTopYforRow(int rowNdx)
   {
      int res = rowNdx * (getRowHeight() + BORDER_THICKNESS);
      
      if (isLabelsVisible())
      {
         res += getLabelsAreaHeight();
      }
      
      return res;
   }

   /**
    * Get cached sorting arrow image.
    *
    * @param  ascending
    *         <code>true</code> to return soring arrow for ascending sorting (arrow up).
    *         <code>false</code> to return soring arrow for descending sorting (arrow down).
    *
    * @return cached sorting arrow image (up or down arrow).
    */
   public ImageWrapper getSortArrowImage(boolean ascending)
   {
      if (sortArrowUpImage == null)
      {
         sortArrowUpImage =
               VectorGraphicsHelper.getScalableImage(SORT_ARROW_UP_IMAGE_NAME,
                     SORT_ARROW_WIDTH, SORT_ARROW_HEIGHT);

         sortArrowDownImage =
               VectorGraphicsHelper.getScalableImage(SORT_ARROW_DOWN_IMAGE_NAME,
                     SORT_ARROW_WIDTH, SORT_ARROW_HEIGHT);
      }

      return ascending ? sortArrowUpImage : sortArrowDownImage;
   }

   /**
    * Get image of the given sort number.
    *
    * @param  num
    *         Sort number, from <code>1</code> to <code>9</code>.
    *
    * @return image of the given sort number.
    */
   public ImageWrapper getSortNumImage(int num)
   {
      if (num < 1 || num > 9)
      {
         throw new IllegalArgumentException("Specify a valid number (1-9).");
      }

      if (sortNumberImages == null)
      {
         sortNumberImages = new ImageWrapper[9];

         for (int i = 0; i < sortNumberImages.length; i++)
         {
            sortNumberImages[i] = VectorGraphicsHelper.getScalableImage(
                                     String.format(SORT_NUMBER_IMAGE_NAME_TEMPLATE, i + 1),
                                     SORT_NUMBER_WIDTH,
                                     SORT_NUMBER_HEIGHT);
         }
      }

      return sortNumberImages[num - 1];
   }

   /**
    * Check if the widget supports LOAD-MOUSE-POINTER() method 
    * 
    * @return <code>true</code> if the widget supports 
    *         LOAD-MOUSE-POINTER() method
    */
   @Override
   public boolean supportsCustomMousePointer()
   {
      return true;
   }

   /**
    * Set the new enhanced font for this browse, which includes recalculation of row and label
    * heights and recalculation of column widths for the columns with implicit width.
    *
    * @param param
    *        Type of the font to set. Can be {@link EnhancedBrowseParameter#CELLS_FONT} or
    *        {@link EnhancedBrowseParameter#LABEL_FONT}.
    * @param font
    *        Font to set.
    */
   void updateEnhancedFont(EnhancedBrowseParameter param, FontDetails font)
   {
      switch (param)
      {
         case CELLS_FONT:
            config().ehFont = font;
            break;
         case LABEL_FONT:
            config().ehLabelFont = font;
            break;
      }

      resolveFonts();

      if (param.equals(CELLS_FONT))
      {
         int heightPixels = calcImplicitRowHeight();
         config().rowHeightPixels = heightPixels;
         config().rowHeightChars = cc.heightFromNative(heightPixels);
         updateRowHeight();

         int colNo = getColumnCount();
         for (int i = 0; i < colNo; i++)
         {
            BrowseColumnGuiImpl column = getColumn(i);
            column.updateImplicitColumnWidth();
         }
         notifyColumnsUpdated();
      }

      updateLabelsHeight();
      repaint();
   }

   /**
    * Returns the left limit for the resize column contour (relative to browse).
    *
    * @return see above.
    */
   protected int getLeftColumnContourLimit()
   {
      return (hasRowMarkers() ? markersColumn.physicalDimension().width : 0) -
               (config.box ? 0 : 1);
   }

   /**
    * Returns the right limit for the resize column contour (relative to browse).
    *
    * @return see above.
    */
   protected int getRightColumnContourLimit()
   {
      return getViewportDimension().width + (config.box ? 1 : 0);
   }

   /**
    * Sub-classes will override this to return a list of mouse actions on which the widget
    * responds.  This is compatible with the AWT {@link MouseEvent} codes and P2J specific
    * {@link com.goldencode.p2j.ui.client.event.MouseEvt} actions.
    *
    * @return   See above.
    */
   @Override
   protected int[] mouseActions()
   {
      return new int[]
      {
         MouseEvent.MOUSE_CLICKED,
         MouseEvent.MOUSE_PRESSED,
         MouseEvent.MOUSE_RELEASED,
         MouseEvent.MOUSE_DRAGGED,
         MouseEvent.MOUSE_WHEEL
      };
   }

   /**
    * Handle TAB key press.
    */
   protected void handleTabKey()
   {
      if (isEditPossible() && getCacheSize() != 0)
      {
         switchToNextCell();
      }
      else
      {
         ThinClient.nextTabStop(this);
      }
   }

   /**
    * Scroll to the editable row if it is out of the view.
    */
   protected void scrollToEditableRow()
   {
      int currRow = getCurrentRow();
      if (!isRowInView(currRow))
      {
         if (isNewEditingRowOutOfTheView())
         {
            scrollToRow(editingExistingRowBefore + 1);
         }
         else
         {
            int topRow = getFirstExistingRow();
            if (currRow < topRow || topRow == -1)
            {
               scrollToRow(currRow);
            }
            else
            {
               scrollToRow(currRow, true, false, false, true, true, getPageSize(), false);
            }
         }
      }
   }

   /**
    * Get browse color resolver.
    *
    * @return see above.
    */
   public GuiColorResolver getBrowseColorResolver()
   {
      return colorResolver;
   }

   /**
    * Get separator color.
    *
    * @return separator color.
    */
   public ColorRgb getSeparatorColor()
   {
      return separatorColor;
   }

   /**
    * Get top offset of the row (relatively to the place where label drawing starts).
    *
    * @param  viewportIndex
    *         0-based index of the row in the viewport.
    *
    * @return top offset of the row.
    */
   public int getTopRowOffset(int viewportIndex)
   {
      int topRowOffset = 0;
      if (isLabelsVisible())
      {
         topRowOffset += getLabelsAreaHeight();
      }
      topRowOffset += (getRowHeight() + BORDER_THICKNESS) * viewportIndex;

      return topRowOffset;
   }

   /**
    * Check if the browse should be drawn as a focused widget.
    *
    * @return <code>true</code> is if the browse should be drawn as a focused widget.
    */
   public boolean isDrawnAsFocused()
   {
      return hasFocus() || isEditing();
   }

   /**
    * Determine if the labels area is visible.
    *
    * @return <code>true</code> if the labels area is visible.
    */
   boolean isLabelsVisible()
   {
      return drawLabels;
   }

   /**
    * Return <code>true</code> if browse layout needs to be performed.
    *
    * @return see above.
    */
   boolean isPerformLayout()
   {
      return config.performLayout;
   }

   /**
    * Get viewport dimension.
    *
    * @return viewport dimension.
    */
   public NativeDimension getViewportDimension()
   {
      NativeDimension dim = scrollableColumnSet.getVisibleDimension();

      ScrollBarGuiImpl horizontal = scrollPane.getScrollBar(ScrollBar.Orientation.HORIZONTAL);
      if (horizontal.isVisible())
      {
         dim.height -= horizontal.physicalDimension().height;
      }

      ScrollBarGuiImpl vertical = scrollPane.getScrollBar(ScrollBar.Orientation.VERTICAL);
      if (vertical.isVisible())
      {
         dim.width -= vertical.physicalDimension().width;
      }

      return dim;
   }

   /**
    * Get viewport bounded rectangle.
    *
    * @return   The viewport bounded rectangle
    */
   public NativeRectangle getViewportBoundedRectangle()
   {
      NativeDimension viewPortDim = getViewportDimension();
      
      NativeRectangle viewPortRect = new NativeRectangle(screenPhysicalLocation(), viewPortDim);
      
      return viewPortRect;
   }

   /**
    * Resolve color basing on input parameters.
    *
    * @param  columnValue
    *         Column-specific color number.
    * @param  columnEhValue
    *         Column-specific enhanced color.
    * @param  browseValue
    *         Browse-specific color number.
    * @param  browseEhValue
    *         Browse-specific enhanced color.
    * @param  fg
    *         <code>true</code> if it is a foreground color. <code>false</code> if it is a
    *         background color.
    * @param  label
    *         <code>true</code> if it is a label color. <code>false</code> if it is a cell color.
    *
    * @return resolved color.
    */
   public ColorRgb resolveColor(int columnValue,
                                ColorRgb columnEhValue,
                                int browseValue,
                                ColorRgb browseEhValue,
                                boolean fg,
                                boolean label)
   {
      if (columnEhValue != null)
      {
         return columnEhValue;
      }

      Window window = window();
      int color = columnValue;
      if (color < 0)
      {
         if (browseEhValue != null)
         {
            return browseEhValue;
         }

         color = browseValue;
      }

      if (color < 0)
      {
         if (label && !fg)
         {
            return gd.getSysColor(SysColor.COLOR_3DFACE);
         }
         else
         {
            GuiColorResolver browseResolver = getBrowseColorResolver();
            return fg ? browseResolver.fgColor : browseResolver.bgColor;
         }
      }
      else
      {
         return fg ? ColorManager.getFgColor(window, color) :
               ColorManager.getBgColor(window, color);
      }
   }

   /**
    * Start column dragging.
    *
    * @param column
    *        Column which is being dragged.
    * @param leftDragColumnMouseOffset
    *        Offset from the left edge of the column which is being dragged to the position of the
    *        pointer inside the column.
    */
   void startColumnDrag(BrowseColumnGuiImpl column, int leftDragColumnMouseOffset)
   {
      draggedColumn = column;
      this.leftDragColumnMouseOffset = leftDragColumnMouseOffset;
   }

   /**
    * Stop column dragging.
    */
   void stopColumnDrag(MouseEvent e)
   {
      if (draggedColumn == null)
      {
         return;
      }

      int pointerX = translate(e).x;
      int leftColumnsOffset = BrowseGuiImpl.this.config.box ? BORDER_THICKNESS : 0;
      int targetColumnIndex = -1;
      if (lockedColumnSet != null)
      {
         for (int i = 0; i < config.numLockedColumns; i++)
         {
            BrowseColumnGuiImpl column = getVisibleColumn(i);
            if (pointerX < leftColumnsOffset + column.physicalLocation().x +
                              column.physicalDimension().width)
            {
               targetColumnIndex = i;
               break;
            }
         }
      }

      int columnCount = getVisibleColumnCount();
      if (targetColumnIndex == -1)
      {
         for (int i = config.numLockedColumns; i < columnCount; i++)
         {
            BrowseColumnGuiImpl column = getVisibleColumn(i);
            if (pointerX < leftColumnsOffset + scrollableColumnSet.physicalLocation().x +
                              column.physicalLocation().x + column.physicalDimension().width)
            {
               targetColumnIndex = i;
               break;
            }
         }
      }

      if (targetColumnIndex == -1)
      {
         targetColumnIndex = columnCount - 1;
      }

      int draggedColumnIndex = draggedColumn.getVisibleColumnIndex();
      if (targetColumnIndex != draggedColumnIndex && targetColumnIndex != draggedColumnIndex - 1)
      {
         int srcOrdinal = visibleToOrdinal(draggedColumnIndex);
         int destOrdinal = visibleToOrdinal(targetColumnIndex);
         browseModel.moveColumn(srcOrdinal,
                                targetColumnIndex < draggedColumnIndex ? destOrdinal + 1 :
                                                                         destOrdinal);
      }

      leftDragColumnMouseOffset = 0;
      draggedColumn = null;
   }

   /**
    * Start column resizing.
    *
    * @param visibleColumnIndex
    *        Index of the column which is being resized (among the visible columns).
    */
   void startColumnResize(int visibleColumnIndex)
   {
      resizedColumn = getVisibleColumn(visibleColumnIndex);

      int browseLeftEdge = hasRowMarkers() ? markersColumn.physicalDimension().width : 0;
      int columnLeftEdge = resizedColumn.physicalLocation().x +
                              (visibleColumnIndex < config.numLockedColumns ? 0 :
                                 resizedColumn.parent().physicalLocation().x);
      leftResizeColumnOffset = Math.max(columnLeftEdge, browseLeftEdge);
   }

   /**
    * Stop column resizing.
    */
   void stopColumnResize()
   {
      if (resizedColumn == null)
      {
         return;
      }

      // True even if the part of the column is out of the view. "+1" is 4GL quirk - resize
      // contour is narrower that the column size.
      int newWidth = rightResizeColumnOffset - leftResizeColumnOffset + 1;
      resizedColumn.setWidth(cc.widthFromNative(newWidth));
      rebuildColumns();

      int columnCount = getVisibleColumnCount();
      if ((config.fitLastColumn || config.noEmptySpace) && columnCount > 0)
      {
         // if last column can be fit (it will not be smaller than min column width) then force
         // resize the last column
         if (needsColumnFit())
         {
            fitLastColumn();

            // update total column sizes
            rebuildColumns();
         }
      }

      boolean horizontalScrollbarWasVisible = horizontalScrollbar.isVisible();
      showHorizontalScrollbar(totalColumnsWidth > getViewportDimension().width);

      // recalculate DOWN
      int oldDown = config.down;
      config.down = calcDown();

      if (oldDown < config.down)
      {
         // get more rows
         int lastRow = getLastExistingRow();
         if (lastRow != -1 && dcData.length == oldDown)
         {
            int diff = config.down - oldDown;
            BrowseRow[] serverData = getRows(lastRow + 1,
                                             diff,
                                             true,
                                             RowDisplayStrategy.NO_DISPLAY);
            if (serverData.length > 0)
            {
               BrowseRow[] oldData = dcData;
               dcData = new BrowseRow[oldData.length + serverData.length];
               System.arraycopy(oldData, 0, dcData, 0, oldData.length);
               System.arraycopy(serverData, 0, dcData, oldData.length, serverData.length);
            }
         }
      }
      else if (oldDown > config.down)
      {
         // removed extra rows
         if (dcData.length > config.down)
         {
            BrowseRow[] oldData = dcData;
            dcData = new BrowseRow[config.down];
            System.arraycopy(oldData, 0, dcData, 0, dcData.length);
         }
      }

      if (horizontalScrollbarWasVisible != horizontalScrollbar.isVisible())
      {
         if (config.columnScrolling)
         {
            updateHorizontalColumnScrollMax();
         }
         scrollableColumnSet.setPhysicalLocation(0, 0);

         updateColumnSetsHeight();
      }

      updateEditorAfterColumnModification();
      leftResizeColumnOffset = 0;
      rightResizeColumnOffset = 0;
      resizedColumn = null;
      prepareHyperlinks();
      layoutHyperlinks();
      repaint();
   }

   /**
    * Returns <code>true</code> if a column is being dragged.
    *
    * @return <code>true</code> if a column is being dragged.
    */
   boolean isColumnDrag()
   {
      return draggedColumn != null;
   }

   /**
    * Returns <code>true</code> if a column is being resized.
    *
    * @return <code>true</code> if a column is being resized.
    */
   boolean isColumnResize()
   {
      return resizedColumn != null;
   }

   /**
    * Returns <code>true</code> if the mouse pointer is positioned over the resize area after the
    * right edge of the last locked column.
    *
    * @param  e
    *         Mouse event.
    *
    * @return <code>true</code> if the mouse pointer is positioned over the resize area after the
    *         right edge of the last locked column.
    */
   boolean isLastLockedColumnResizeArea(MouseEvent e)
   {
      if (config.numLockedColumns == 0)
      {
         return false;
      }

      NativePoint point = translate(e);
      int leftOffset = lockedColumnSet.physicalDimension().width +
                        (config.box ? BORDER_THICKNESS : 0);
      return point.x >= leftOffset && point.x < leftOffset + COLUMN_RESIZE_AREA_WIDTH;
   }

   /**
    * Get total width of the locked columns,
    *
    * @return see above.
    */
   int getLockedColumnsWidth()
   {
      return lockedColumnsWidth;
   }

   /**
    * Perform actions which should happen when a header of a column is pressed. These actions can
    * cancel header pressing.
    *
    * @return <code>true</code> if the header can be pressed, <code>false</code> otherwise.
    */
   boolean beforeColumnHeaderPressed()
   {
      return !isEditing() || leaveRow();
   }

   /**
    * Change visibility of the column.
    *
    * @param columnOrdinal
    *        0-based column ordinal (among all columns).
    * @param visible
    *        <code>true</code> if the column should be visible. <code>false</code> if the column
    *        should be hidden.
    */
   void changeColumnVisibility(int columnOrdinal, boolean visible)
   {
      BrowseColumnGuiImpl column = getColumn(columnOrdinal);
      column.setVisible(visible);
      column.config().hidden = !visible;
      notifyColumnsUpdated();
   }

   /**
    * Get browse font.
    *
    * @return browse font.
    */
   FontDetails getBrowseFont()
   {
      return browseFontResolver.font().font;
   }

   /**
    * Get browse labels font.
    *
    * @return browse labels font.
    */

   FontDetails getBrowseLabelFont()
   {
      if (config().ehLabelFont != null)
      {
         return config().ehLabelFont;
      }

      int labelFont = config().labelFont;

      if (labelFont != FontManager.UNDEFINED_FONT)
      {
         return FontManager.resolveFont(window(), labelFont, gd);
      }
      else
      {
         return getBrowseFont();
      }
   }

   /**
    * Setups scrolling to fully (as possible) show the given column.
    *
    * @param visibleColumnIndex
    *        0-based index of the column among the visible columns.
    */
   @Override
   protected void columnToView(int visibleColumnIndex)
   {
      //TODO implement (is it really needed?)
   }

   /**
    * Setups scrolling to fully (as possible) show the given column.
    * TODO does the same thing as columnToView in ChUI, however in ChUI columnToView is possibly
    * called in wrong places which may cause unnecessary scrolling. This should be investigated.
    *
    * @param visibleColumnIndex
    *        0-based index of the column among the visible columns.
    */
   protected void scrollToColumn(int visibleColumnIndex)
   {
      boolean visible = true;
      boolean toTheLeft = true;

      if (visibleColumnIndex >= config.numLockedColumns)
      {
         BrowseColumnGuiImpl column = getVisibleColumn(visibleColumnIndex);
         int scrollPosition = - scrollableColumnSet.physicalLocation().x;
         int columnPosition = column.physicalLocation().x;
         int columnWidth = column.physicalDimension().width;
         int lockedWidth = lockedColumnSet != null ?
                              lockedColumnSet.physicalDimension().width : 0;
         int columnPosInViewport = columnPosition - scrollPosition - lockedWidth;
         int viewportWidth = getViewportDimension().width;
         int effectiveViewportWidth = viewportWidth - lockedWidth;

         if (columnPosInViewport < 0)
         {
            visible = false;
         }
         else
         {
            int rightColumnEdge = columnPosition + columnWidth;
            int rightViewportEdge = scrollPosition + viewportWidth;
            int allowedRightOffset = (int) (1.9 * hScrollingStep); // approximately

            if (rightColumnEdge > rightViewportEdge + allowedRightOffset)
            {
               visible = false;
               toTheLeft = false;
            }
         }

         if (!visible)
         {
            // scroll to column
            if (toTheLeft)
            {
               scrollableColumnSet.setPhysicalLocation(-getColumnOffset(visibleColumnIndex), 0);
            }
            else
            {
               if (config.columnScrolling)
               {
                  int totalWidth = 0;
                  int index;
                  for (index = visibleColumnIndex; index >= config.numLockedColumns; index--)
                  {
                     totalWidth += getVisibleColumn(index).physicalDimension().width;
                     if (totalWidth > effectiveViewportWidth)
                     {
                        break;
                     }
                  }

                  index = Math.min(index + 1, visibleColumnIndex);
                  scrollableColumnSet.setPhysicalLocation(-getColumnOffset(index), 0);
               }
               else
               {
                  int x = column.physicalLocation().x - getViewportDimension().width +
                           column.physicalDimension().width;
                  scrollableColumnSet.setPhysicalLocation(-x, 0);
               }
            }
         }
      }
      else
      {
         // reset scroll pane if a locked column is edited
         scrollableColumnSet.setPhysicalLocation(0, 0);
      }
   }

   /**
    * Creates an editor for the current cell and adds it to the browse. Adjusts selection data if
    * necessary.
    *
    * @param    prevFocus
    *           The previous focused widget, which will lose focus.
    * @param    updateSelection
    *           For multi-selection browse: if <code>true</code>, current row becomes the only
    *           selected row.
    * @param    useValue
    *           Use the specified value for the editor instead of getting it from the browse cache.
    * @param    value
    *           Value to use in the editor. Can be <code>null</code> (valid value). Set
    *           <code>useValue</code> to <code>true</code> if you want to use this parameter.
    * 
    * @return   <code>true</code> if editing was on and ENTRY triggered allowed it to stop.
    */
   @Override
   protected boolean startEdit(WidgetId     prevFocus,
                               boolean      updateSelection,
                               boolean      useValue,
                               BaseDataType value)
   {
      if (!hasFocus() || !isEnabled() || isLayoutMode())
      {
         return false;
      }
      
      // a mouse event is in progress, check the right actions
      if (meColumn != null)
      {
         BrowseColumnConfig[] cols = getVisibleColumns();
         
         if (meColumn.config().readOnly)
         {
            // cannot start editing on this column
            return false;
         }
      }
      
      if (!super.startEdit(prevFocus, updateSelection, useValue, value))
      {
         return false;
      }
      
      int currentRow = getCurrentRow();
      if (editedRowIndex != currentRow)
      {
         editedRow = dcData[dcGetRowIndex(currentRow)];
         editedRowIndex = currentRow;
      }

      editingExistingRowBefore = null;
      if (dcData != null && isNewRow(currentRow))
      {
         int viewportRow = getViewportRow();

         // search for closest row at the top
         for (int i = viewportRow - 1; i >= 0; i--)
         {
            int modelIndex = getCachedRowNumber(i);
            if (!isNewRow(modelIndex))
            {
               editingExistingRowBefore = modelIndex;
               break;
            }
         }

         if (editingExistingRowBefore == null)
         {
            for (int i = viewportRow + 1; i < getCacheSize(); i++)
            {
               int modelIndex = getCachedRowNumber(i);
               if (!isNewRow(modelIndex))
               {
                  editingExistingRowBefore = modelIndex - 1;
                  break;
               }
            }
         }
      }
      
      return true;
   }

   /**
    * Create editor widget for the current column. Type of editor corresponds column's widget
    * type (fill-in, toggle-box or combo-box).
    *
    * @param  useValue
    *         Use the specified value for the editor instead of getting it from the browse cache.
    * @param  value
    *         Value to use in the editor. Can be <code>null</code> (valid value). Set
    *         <code>useValue</code> to <code>true</code> if you want to use this parameter.
    *
    * @return editor widget for the current column.
    */
   @Override
   protected Widget createEditor(boolean useValue, BaseDataType value)
   {
      BrowseColumnConfig[] cols = getVisibleColumns();
      if (TOGGLE_BOX.equals(cols[currColumn].widgetType))
      {
         return createEditingToggleBox(useValue, value);
      }
      else
      {
         return createEditingFillIn(useValue, value);
      }
      // TODO combo-box
   }

   /**
    * Create toggle-box editor widget for the current column.
    *
    * @param  useValue
    *         Use the specified value for the editor instead of getting it from the browse cache.
    * @param  value
    *         Value to use in the editor. Can be <code>null</code> (valid value). Set
    *         <code>useValue</code> to <code>true</code> if you want to use this parameter.
    *
    * @return toggle-box editor widget for the current column.
    */
   protected ToggleBox createEditingToggleBox(boolean useValue, BaseDataType value)
   {
      ConfigManager mgr =         ConfigManager.getInstance();
      WidgetConfig widgetConfig =   browseModel.getEditorsConfigs()[currColumn];
      BrowseColumnConfig[] cols = getVisibleColumns();
      ToggleBoxConfig cfg =       (ToggleBoxConfig) widgetConfig;

      if (mgr.getActiveConfig(cfg.id) == null)
      {
         // Widget is unregistered when it is destroyed in stopEdit. Register it again with
         // empty backup config so all updated values are pushed to server.
         ToggleBoxConfig backupConfig = new ToggleBoxConfig();
         mgr.addWidgetConfig(cfg, backupConfig);
      }

      cfg.pfcolor = config.pfcolor;
      cfg.dcolor = config.dcolor;
      cfg.help = cols[currColumn].help;
      cfg.modified = false;
      cfg.visible = false;

      initEditorConfig(cfg);

      WidgetFactory<?> factory = screen().getFactory();
      WidgetId editorId = cfg.id;
      ToggleBox toggleBox = factory.create(ToggleBox.class, editorId, cfg);

      screen().getRegistry().addWidget(editorId, toggleBox);
      toggleBox.initialize(editorId, cfg);
      ((ToggleBoxGuiImpl) toggleBox).setBrowse(this);

      // reset underline attribute and force Browse colors
      //toggleBox.setBrowse(this);

      BaseDataType toggleValue = useValue ? value : getValue(visibleToOrdinal(currColumn));
      if (toggleValue != null)
      {
         toggleBox.setValue(toggleValue);
         toggleBox.setState(ScreenBuffer.CHANGED);
      }
      else
      {
         toggleBox.setState(ScreenBuffer.UNINITIALIZED);
      }

      addEditor(toggleBox);

      toggleBox.setEnabled(true);
      toggleBox.setVisible(true);
      return toggleBox;
   }

   /**
    * Initialize editor widget config.
    *
    * @param cfg
    *        Editor widget config.
    */
   @Override
   protected void initEditorConfig(BaseConfig cfg)
   {
      BrowseColumnGuiImpl column = getVisibleColumn(currColumn);
      column.initEditorConfig(cfg);
   }

   /**
    * Add editor widget to the browse.
    *
    * @param editor
    *        Editor widget.
    */
   @Override
   protected void addEditor(Widget editor)
   {
      if (currColumn < config.numLockedColumns)
      {
         lockedColumnSet.add(editor);
      }
      else
      {
         scrollableColumnSet.add(editor);
      }

      if (editor instanceof FillIn)
      {
         editor.addFocusListener(new FocusListener()
         {
            @Override
            public void onFocusGained(FocusEvent event)
            {
               if (mouseFocusGained)
               {
                  FillIn fillIn = (FillIn) editor;
                  fillIn.config().selectionStart = FillInConfig.INVALID_SELECTION;
                  fillIn.config().selectionEnd = FillInConfig.INVALID_SELECTION;
                  fillIn.setCursorOffset(0);
                  fillIn.setZapMode(false);
                  fillIn.setInsertMode(true);
               }

               mouseFocusGained = false;
            }

            @Override
            public void onFocusLost(FocusEvent event)
            {
               // why? focus may move to other editor on same row/browse. BrowseGuiImpl.this.onFocusLost(event);
               // the answer is the browse not always gets focus loosing event when editor gets  
               if (isEditing())
               {
                  stopEdit(event.other() == null ? null : event.other().getId());
               }
            }
         });
      }

      scrollToColumn(currColumn);
   }

   /**
    * Clean up editor widget.
    *
    * @param editor
    *        Editor widget.
    */
   @Override
   protected void cleanupEditor(Widget editor)
   {
      gd.withSelectedWindow(window().getId().asInt(),
                            () -> gd.deregisterWidget(editor.getId().asInt()));
   }

   /**
    * Returns <code>true</code> if the current row of the single-selection browse is selected on
    * FOCUS-GAINED and some other events.
    *
    * @return see above.
    */
   protected boolean isAggressiveCurrentRowSelection()
   {
      return false;
   }

   /**
    * Get page size used for Page Up / Page Down.
    *
    * @return increment size (in rows) used for Page Up / Page Down.
    */
   @Override
   protected int getPageSize()
   {
      return getDown() - 1;
   }

   /**
    * Update height of the labels area.
    */
   @Override
   protected void updateLabelsHeight()
   {
      ilLabelsHeight = 1;
      int colNum = getVisibleColumnCount();
      for (int i = 0; i < colNum; i++)
      {
         BrowseColumnGuiImpl column = getVisibleColumn(i);
         column.updateLabel();
         String[] label = column.getLabelLines();
         ilLabelsHeight = Math.max(ilLabelsHeight, label.length);
      }

      int rowHeight;
      if (config().ehLabelFont != null)
      {
         rowHeight = FontManager.getFontHeight(window(), config().ehLabelFont, null, gd);
      }
      else
      {
         rowHeight = browseFontResolver.font().fontHeight;
      }
      config.labelsHeightPixels = rowHeight * ilLabelsHeight + CELL_OFFSET_TOP + CELL_OFFSET_BOTTOM;
   }

   /**
    * Notify that the browse has been scrolled (i.e. the top row has changed).
    */
   protected void notifyBrowseScrolled()
   {
      super.notifyBrowseScrolled();

      Widget editor = getActiveEditor();
      if (editor != null)
      {
         BaseConfig cfg = (BaseConfig) editor.config();
         if (isCurrentRowInView())
         {
            cfg.y = getCurrentEditorY();
         }
         else
         {
            // put the editor outside of the view so it is not visible but can receive key presses
            cfg.y = (int) (- 1.5 * editor.physicalDimension().height);
         }
      }

      // update hyperlinks and web text metrics for right-aligned columns
      Map<FontDetails, Set<String>> texts = prepareHyperlinks();
      if (gd.needsExplicitTextMetricsCaching())
      {
         for (BrowseColumnConfig col : getVisibleColumns())
         {
            if (!col.hyperlinked && Browse.isRightAligned(col))
            {
               texts = collectColumnTexts(getColumn(col.ordinal), texts);
            }
         }
         gd.cacheTextMetrics(texts, null);
      }
      layoutHyperlinks();
   }

   /**
    * Create hyperlink widgets without performing layout: destroy all old hyperlink widgets from
    * the old browse view and create new widgets for the current view. In order to perform layout
    * of these hyperlinks, call {@link #layoutHyperlinks()}. Layout should be performed after we
    * got text metrics for the hyperlinks.
    *
    * @return map containing texts of all created hyperlinks, keyed by fonts which are used for
    *         drawing specific hyperlink texts.
    */
   protected Map<FontDetails, Set<String>> prepareHyperlinks()
   {
      if (hyperlinkWidgets != null && !hyperlinkWidgets.isEmpty())
      {
         hyperlinkWidgets.forEach((col, map) -> map.forEach((idx, link) -> link.destroy()));
         hyperlinkWidgets.clear();
      }

      if (dcData == null || dcData.length == 0 || !hasHyperlinkedColumns())
      {
         return null;
      }

      BrowseColumnConfig[] visibleColumns = getVisibleColumns();
      if (visibleColumns == null)
      {
         return null;
      }

      if (hyperlinkWidgets == null)
      {
         hyperlinkWidgets = new HashMap<>();
      }

      Map<FontDetails, Set<String>> texts = null;
      boolean needsFontsMetricsCaching = gd.needsExplicitTextMetricsCaching();
      if (needsFontsMetricsCaching)
      {
         texts = new HashMap<>();
      }

      for (BrowseColumnConfig config : visibleColumns)
      {
         if (config.hyperlinked)
         {
            BrowseColumnGuiImpl column = getColumn(config.ordinal);

            int size = getCacheSizeWithIntermediate();
            for (int viewportIndex = 0; viewportIndex < size; viewportIndex++)
            {
               String text = getCellTextWithIntermediate(column, viewportIndex);

               if (text != null && !text.trim().isEmpty())
               {
                  text = text.trim();
                  Hyperlink hyperlink = new Hyperlink(column, text, viewportIndex);

                  Map<Integer, Hyperlink> columnMap =
                        hyperlinkWidgets.computeIfAbsent(column, k -> new HashMap<>());
                  columnMap.put(viewportIndex, hyperlink);
                  FontDetails font = hyperlink.getFont();

                  if (needsFontsMetricsCaching)
                  {
                     Set<String> fontTexts = texts.computeIfAbsent(font, k -> new HashSet<>());
                     fontTexts.add(text);
                  }
               }
            }
         }
      }

      return texts;
   }

   /**
    * Collect texts of all cells in the given column in the current view and add them to the
    * specified map.
    *
    * @param  column
    *         Column containing texts to be collected.
    * @param  texts
    *         Map into which collected texts are added, keyed by fonts which are used for drawing
    *         specific texts. If <code>null</code>, a new map is created for the collected texts
    *         (if there are any texts to collect).
    *
    * @return map containing collected texts, keyed by fonts which are used for drawing specific
    *         texts.
    */
   protected Map<FontDetails, Set<String>> collectColumnTexts(BrowseColumnGuiImpl column,
                                                              Map<FontDetails, Set<String>> texts)
   {
      int cacheSize = getCacheSizeWithIntermediate();
      if (cacheSize == 0)
      {
         return texts;
      }

      if (texts == null)
      {
         texts = new HashMap<>();
      }

      FontDetails font = column.getCellsFont();
      Set<String> fontTexts = texts.computeIfAbsent(font, k -> new HashSet<>());

      for (int i = 0; i < cacheSize; i++)
      {
         fontTexts.add(getCellTextWithIntermediate(column, i));
      }

      return texts;
   }

   /**
    * Perform layout for all hyperlinks (determine their size, position etc.). Hyperlinks should
    * be already created by {@link #prepareHyperlinks()}. Layout should be performed after we
    * got text metrics for the hyperlinks.
    */
   protected void layoutHyperlinks()
   {
      if (hyperlinkWidgets != null)
      {
         hyperlinkWidgets.forEach((col, map) -> map.forEach((idx, link) -> link.doLayout()));
      }
   }

   /**
    * Get Y coordinate for the current editor (relatively to the containing column set).
    *
    * @return see above.
    */
   protected int getCurrentEditorY()
   {
      BrowseColumnGuiImpl column = getVisibleColumn(currColumn);
      switch (column.config().widgetType)
      {
         case TOGGLE_BOX:
            int offset = ThemeManager.getCurrentTheme().getRowToggleOffset(column, true).y;
            return getTopRowOffset(getViewportRow()) + offset;
         case COMBO_BOX:
            return 0; // TODO combo-box
         case FILL_IN:
         default:
            return getTopRowOffset(getViewportRow());
      }
   }

   /**
    * Get X coordinate for the current editor (relatively to the containing column set).
    *
    * @return see above.
    */
   protected int getCurrentEditorX()
   {
      BrowseColumnGuiImpl column = getVisibleColumn(currColumn);
      switch (column.config().widgetType)
      {
         case TOGGLE_BOX:
            int offset = ThemeManager.getCurrentTheme().getRowToggleOffset(column, true).x;
            return column.physicalLocation().x + offset;
         case COMBO_BOX:
            return 0; // TODO combo-box
         case FILL_IN:
         default:
            return column.physicalLocation().x;
      }
   }

   /**
    * Determines if browse refreshed on browse widget enabling.
    *
    * @return <code>false</code>
    */
   @Override
   protected boolean isRefreshedOnEnable()
   {
      return false;
   }

   /**
    * The method shows the popup menu for this widget. The menu is resolved with the help of
    * {@link #resolvePopupMenuId()}.
    *
    * @param   e
    *          The related mouse event that caused a popup to be shown.
    * @param   dismissHandler
    *          When not null, the runnable instance is executed when the popup menu is dismissed, regardless
    *          whether an item is chosen or not.
    *
    * @return  Returns the popup menu shown or {@code null} if no popup is resolved for this widget.
    */
   @Override
   protected Menu showPopupMenu(MouseEvent e, Runnable dismissHandler)
   {
      return showPopupMenu(null, e, dismissHandler);
   }

   /**
    * Show browse popop menu on mouse right-click.
    *
    * @param column
    *        Column which was right-clicked or <code>null</code> if the browse was clicked outside
    *        of the columns area.
    * @param e
    *        Mouse event.
    * @param dismissHandler
    *        When not null, the runnable instance is executed when the popup menu is dismissed,
    *        regardless whether an item is chosen or not.
    */
   protected Menu showPopupMenu(BrowseColumnGuiImpl column, MouseEvent e, Runnable dismissHandler)
   {
      // enhanced browse's popup takes precedence over the legacy popup
      if (config.enhancedBrowse)
      {
         BrowsePopupMenu menu = BrowsePopupMenu.create(this, column);
         menu.showPopup(this, new NativePoint(e.getX(), e.getY()), dismissHandler);
         return menu;
      }

      return super.showPopupMenu(e, dismissHandler);
   }

   /**
    * Update specified cell in current row.
    *
    * @param   columnOrdinal
    *          0-based index of the cell to update.
    * @param   var
    *          New value of the cell.
    *
    * @return  <code>true</code> if column value was changed.
    */
   protected boolean updateColumn(int columnOrdinal, BaseDataType var)
   {
      boolean res = super.updateColumn(columnOrdinal, var);
      if (res && isCurrentRowInView())
      {
         boolean needsFontsMetricsCaching = gd.needsExplicitTextMetricsCaching();
         BrowseColumnGuiImpl column = getColumn(columnOrdinal);
         if (column.config().hyperlinked)
         {
            // update hyperlink for the edited cell
            String text = getCellTextWithIntermediate(column, getViewportRow());
            if (text != null && !text.trim().isEmpty())
            {
               text = text.trim();

               // destroy the old hyperlink
               if (hyperlinkWidgets != null)
               {
                  Map<Integer, Hyperlink> columnMap = hyperlinkWidgets.get(column);
                  if (columnMap != null)
                  {
                     Hyperlink hyperlink = columnMap.remove(getViewportRow());
                     if (hyperlink != null)
                     {
                        hyperlink.destroy();
                     }
                  }
               }
               Hyperlink hyperlink = new Hyperlink(column, text, getViewportRow());

               if (hyperlinkWidgets == null)
               {
                  hyperlinkWidgets = new HashMap<>();
               }
               Map<Integer, Hyperlink> columnMap =
                     hyperlinkWidgets.computeIfAbsent(column, k -> new HashMap<>());
               columnMap.put(getViewportRow(), hyperlink);

               if (needsFontsMetricsCaching)
               {
                  cacheTextMetrics(hyperlink.getFont(), text, window().getId().asInt());
               }
               hyperlink.doLayout();
            }
         }
         else if (needsFontsMetricsCaching && Browse.isRightAligned(column))
         {
            // update text metric for the edited right-aligned cell
            String text = getCellTextWithIntermediate(column, getViewportRow());
            cacheTextMetrics(column.getCellsFont(), text, window().getId().asInt());
         }
      }
      return res;
   }

   /**
    * Cache metric of the given text.
    *
    * @param font
    *        The font used to draw the given text.
    * @param text
    *        Text which metrics should be cached.
    * @param windowId
    *        ID of the parent window. Can be <code>null</code> if there is no need to select a
    *        window.
    */
   private void cacheTextMetrics(FontDetails font, String text, Integer windowId)
   {
      Map<FontDetails, Set<String>> texts = new HashMap<>();
      Set<String> set = new HashSet<>();
      set.add(text);
      texts.put(font, set);
      gd.cacheTextMetrics(texts, windowId);
   }

   /**
    * Update row height.
    */
   private void updateRowHeight()
   {
      int topRow = getFirstExistingRow();
      if (topRow < 0)
      {
         topRow = 0;
      }

      paddedRowHeight = calcRowHeight();
      config.down = calcDown();
      scrollToRow(topRow, false, false);

      if (isEditing())
      {
         if (isEditorFillIn())
         {
            FillIn editor = getActiveFillIn();
            editor.config().heightPixels = getRowHeight();
            editor.config().heightChars = cc.heightFromNative(editor.config().heightPixels);
            editor.setHeight(editor.config().heightChars);
            editor.repaint();
         }
         else if (isEditorToggleBox())
         {
            ToggleBox editor = getActiveToggleBox();
            editor.config().y = getCurrentEditorY();
            editor.repaint();
         }
         else
         {
            // TODO combo-box
         }
      }
   }

   /**
    * Update value of the {@link #drawLabels} attribute which defines if the labels area is
    * visible.
    */
   private void updateLabelsVisible()
   {
      drawLabels = config.labels;

      if (!drawLabels)
      {
         for (BrowseColumnConfig column : getVisibleColumns())
         {
            if (column.forceLabel)
            {
               drawLabels = true;
               return;
            }
         }
      }
   }

   /**
    * Calculate title height.
    *
    * @return title height.
    */
   private int calcTitleHeight()
   {
      if (!hasTitle())
      {
         return 0;
      }
   
      return 20 - 2 * BORDER_THICKNESS; // used the same formula as for frames but without border
   }

   /**
    * Create title panel.
    */
   private void createTitlePanel()
   {
      if (hasTitle())
      {
         Theme currentTheme = ThemeManager.getCurrentTheme();
         NativeDimension browseDimension = physicalDimension();
         
         // create title panel
         boolean flat = currentTheme.isFlat();
         titlePanel = new BorderedPanelGuiImpl(flat
               ? new EmptyBorder<>()
               : new LineBorderGuiImpl(null, null, this));
         titlePanel.setSize(cc.dimensionFromNative(
               new NativeDimension(browseDimension.width, titleHeight + 2 * BORDER_THICKNESS)));
         titlePanel.setCustomDraw(() ->
         {
            NativeDimension d = titlePanel.physicalDimension();
            // for non-flat theme, the title panel is smaller to accomodate the non-flat drawings
            int width = d.width - (flat ? 0 : 2 * BORDER_THICKNESS);
            int height = d.height - (flat ? 0 : 2 * BORDER_THICKNESS);
            currentTheme.drawBrowseTitlePanel(
                  this, gd, colorResolver, titleFontResolver, 0, 0, width, height);
         });
         
         add(titlePanel);
         OutputManager.instance().getRegistry().addWidget((Widget) titlePanel);
      }
   }

   /**
    * Create column panels and widgets.
    */
   private void buildPanels()
   {
      boolean implicitHeight = config.heightChars <= 0;
      boolean implicitWidth = config.widthChars <= 0;

      if (implicitWidth && config.dynamic)
      {
         setWidth(DEFAULT_DYNAMIC_WIDTH_CHARS);
         implicitWidth = false;
      }

      // create scrollable columns pane it even if it is empty
      scrollableColumnSet = new ColumnSetContainer(false);
      scrollableColumnSet.initialize();

      // calc horizontal scroll step before creating scroll pane
      hScrollingStep = FontManager.getFontWidth(window(), browseFontResolver.resolveFontNum(), gd);

      // create scroll pane with borders and scrollbars
      scrollPane = new ColumnsScrollPane(scrollableColumnSet);
      scrollPane.setScrollBars(ScrollBar.Bars.BOTH);
      scrollPane.setLocation(getScrollPaneLocation());
      
      add(scrollPane);

      createLockedColumnsSet();

      OutputManager.instance().getRegistry().addWidget((Widget) scrollPane);

      // init vertical scrolling
      verticalScroller = new VerticalRowScroller();
      verticalScrollbar = scrollPane.getScrollBar(ScrollBar.Orientation.VERTICAL);
      scrollPane.setScrollBarController(ScrollBar.Orientation.VERTICAL, verticalScroller);

      // init horizontal scrolling
      horizontalScroller = new HorizontalColumnScroller();
      horizontalScrollbar = scrollPane.getScrollBar(ScrollBar.Orientation.HORIZONTAL);
      scrollPane.setScrollBarController(ScrollBar.Orientation.HORIZONTAL, horizontalScroller);
      
      if (config.box && !ThemeManager.getCurrentTheme().isFlat())
      {
         double x = cc.baseUnits().x * BORDER_THICKNESS;
         double y = cc.baseUnits().y * BORDER_THICKNESS;
         Border<GuiOutputManager> border =
                      new LineBorderGuiImpl(new Insets(hasTitle() ? 0 : y, x, y, x), null, this);
         scrollPane.setBorder(border);
      }
      
      updateVerticalScrollbarVisibility();
      rebuildColumns();

      int hBarHeight = getScrollbarThickness(ScrollBar.Orientation.HORIZONTAL);
      int vBarWidth = config.scrollbarVertical ?
                        getScrollbarThickness(ScrollBar.Orientation.VERTICAL) : 0;

      scrollableColumnSet.setWidth(cc.widthFromNative(totalColumnsWidth));

      // resize browse according to columns width
      if (implicitWidth)
      {
         int widthNative = totalColumnsWidth + vBarWidth +
                              (config.box ? 2 * BORDER_THICKNESS : 0);
         setWidth(cc.widthFromNative(widthNative));
      }

      int viewportWidth = physicalDimension().width - vBarWidth -
                              (config.box ? 2 * BORDER_THICKNESS : 0);
      boolean useHorizontalScrollbar = totalColumnsWidth > viewportWidth;
      showHorizontalScrollbar(useHorizontalScrollbar);

      if (implicitHeight)
      {
         int nativeHeight = calcImplicitBrowseHeightNoScrollbar();
         if (useHorizontalScrollbar)
         {
            nativeHeight += hBarHeight;
         }

         setHeight(cc.heightFromNative(nativeHeight));
      }

      // set scroll pane size
      NativeDimension scrollPaneDimension = getScrollPaneDimension();
      scrollPane.setSize(cc.dimensionFromNative(scrollPaneDimension));

      // adjust sizes of panels
      adjustLockedColumnsPanel();

      NativeDimension viewportDimension = getViewportDimension();
      scrollableColumnSet.setHeight(cc.heightFromNative(viewportDimension.height));

      if (fitLastColumn())
      {
         rebuildColumns();

         showHorizontalScrollbar(totalColumnsWidth > getViewportDimension().width);
         config.down = calcDown();

         if (config.columnScrolling)
         {
            updateHorizontalColumnScrollMax();
         }
         updateColumnSetsHeight();
      }
   }

   /**
    * Create and initialize locked columns set (if it wasn't created before).
    */
   private void createLockedColumnsSet()
   {
      if (lockedColumnSet != null)
      {
         return;
      }

      if (hasRowMarkers() || config.numLockedColumns > 0)
      {
         // create locked columns pane (markers column also belongs to this column)
         lockedColumnSet = new ColumnSetContainer(true);
         lockedColumnSet.initialize();
         lockedColumnSet.setLocation(getLockedPaneLocation());
         add(lockedColumnSet);
      }
   }

   /**
    * Re-initiate column widgets.
    */
   private void rebuildColumns()
   {
      // detach old columns first
      if (markersColumn != null && markersColumn.parent() != null)
      {
         markersColumn.parent().detach(markersColumn);
      }

      int colNum = getVisibleColumnCount();
      for (int i = 0; i < colNum; i++)
      {
         BrowseColumnGuiImpl column = getVisibleColumn(i);

         if (column.parent() != null)
         {
            column.parent().detach(column);
         }
      }

      resetColumnsCache();

      colNum = getVisibleColumnCount();
      createLockedColumnsSet();

      // iterate columns and get total width
      totalColumnsWidth = 0;
      lockedColumnsWidth = 0;
      if (hasRowMarkers() || (config.realized && markersColumn != null))
      {
         createMarkersColumn();

         lockedColumnSet.add(markersColumn);
         int width = markersColumn.physicalDimension().width;
         totalColumnsWidth += width;
         lockedColumnsWidth += width;;
      }

      for (int i = 0; i < colNum; i++)
      {
         BrowseColumnGuiImpl column = getVisibleColumn(i);

         if (i < config.numLockedColumns)
         {
            lockedColumnSet.add(column);
            column.init(this, totalColumnsWidth, i);
            lockedColumnsWidth += column.physicalDimension().width;
         }
         else
         {
            scrollableColumnSet.add(column);
            column.init(this, totalColumnsWidth, i);
         }
         totalColumnsWidth += column.physicalDimension().width;
      }

      if (colNum > 0)
      {
         // right border is drawn if possible but not included in width calculations
         totalColumnsWidth -= BORDER_THICKNESS;
      }

      adjustLockedColumnsPanel();
      updateLabelsHeight();
   }
   
   /**
    * Get the total width of the columns of this browse widget.
    * 
    * @return  the total width of the columns. 
    */
   public int getTotalColumnsWidth()
   {
      return totalColumnsWidth;
   }
   
   /**
    * Get the scrollable column widget.
    * 
    * @return the scrollable column widget.
    */
   public Widget getScrollableColumnSet()
   {
      return scrollableColumnSet;
   }

   /**
    * Returns the same as <code>getDefaultFocus()</code> because focus placement
    * inside BROWSE can't be switched from initial component until subcomponent
    * creation or deletion.
    * 
    * @see com.goldencode.p2j.ui.client.widget.Container#currentFocus()
    */
   @Override
   public Widget<GuiOutputManager> currentFocus()
   {
      if (isEditPossible() && scrollableColumnSet != null)
      {
         Widget<GuiOutputManager> focus = scrollableColumnSet.currentFocus();
         if (focus != null)
         {
            return focus;
         }
      }
      
      return super.currentFocus();
   }
   
   /**
    * Check if widget has focus.  If any of its sub-components has the focus, than this will 
    * report  <code>true</code>.
    * 
    * @return  <code>true</code> if widget has focus.
    */
   @Override
   public boolean hasFocus()
   {
      if (super.hasFocus())
      {
         return true;
      }
      
      Widget<?> focus = currentFocus();
      return (focus != null && focus.parent(Browse.class) == this);
   }

   /**
    * Set the width of the {@link #lockedColumnSet} panel according to
    * {@link #lockedColumnsWidth}.
    */
   private void adjustLockedColumnsPanel()
   {
      // adjust panel sizes according to columns width
      if (lockedColumnSet != null)
      {
         NativeDimension viewportDimension = getViewportDimension();
         lockedColumnSet.setSize(cc.dimensionFromNative(
                                 new NativeDimension(lockedColumnsWidth, viewportDimension.height)));
      }
   }

   /**
    * Fit the last column if FIT-LAST-COLUMN attribute is set.
    *
    * @return <code>true</code> if width of the last column was updated in order to fit it in the
    *         view.
    */
   private boolean fitLastColumn()
   {
      if (!config.fitLastColumn && !config.noEmptySpace)
      {
         return false;
      }

      int columnCount = getVisibleColumnCount();
      if (columnCount == 0)
      {
         return false;
      }

      if (needsColumnFit())
      {
         BrowseColumnGuiImpl lastColumn = getVisibleColumn(columnCount - 1);
         totalColumnsWidth = getViewportDimension().width;
         int width = totalColumnsWidth - lastColumn.physicalLocation().x + BORDER_THICKNESS;
         lastColumn.setWidth(cc.widthFromNative(width));
         return true;
      }

      return false;
   }

   /**
    * Returns <code>true</code> if last column should be fit according to FIT-LAST-COLUMN or
    * NO-EMPTY-SPACE option.
    *
    * @return see above.
    */
   private boolean needsColumnFit()
   {
      if (!config.fitLastColumn && !config.noEmptySpace)
      {
         return false;
      }

      int columnCount = getVisibleColumnCount();
      BrowseColumnGuiImpl lastColumn = getVisibleColumn(columnCount - 1);
      int lastColumnWidth = lastColumn.physicalDimension().width;
      int viewportWidth = getViewportDimension().width;

      return config.fitLastColumn ?
            viewportWidth >
                     totalColumnsWidth - lastColumnWidth + getEffectiveMinColumnWidth() :
            viewportWidth > totalColumnsWidth;
   }

   /**
    * Update browse width and height without actual layout.
    */
   private void updateBrowseDimensions()
   {
      boolean implicitWidth = config.widthChars <= 0;
      boolean implicitHeight = config.heightChars <= 0;

      double width = config.widthChars;
      double height = config.heightChars;

      if (!implicitHeight && !implicitWidth)
      {
         setSize(width, height);
         return;
      }

      resolveFonts();
      updateLabelsVisible();

      // iterate columns and get total width
      int totalColumnsWidth = 0;
      if (hasRowMarkers())
      {
         createMarkersColumn();
         totalColumnsWidth += markersColumn.calcColumnWidth();
      }

      int colNum = getVisibleColumnCount();
      for (int i = 0; i < colNum; i++)
      {
         BrowseColumnGuiImpl column = getVisibleColumn(i);
         column.init(this, totalColumnsWidth, i);
         totalColumnsWidth += column.calcColumnWidth();
      }
      
      int scrollBarWidth = ThemeManager.getCurrentTheme().getScrollBarGaugeWidth();
      // rightmost vertical separator is hidden if columns width is more that viewport size
      totalColumnsWidth -= BORDER_THICKNESS;

      if (implicitWidth)
      {
         if (config.dynamic)
         {
            width = DEFAULT_DYNAMIC_WIDTH_CHARS;
         }
         else
         {
            int totalWidth = totalColumnsWidth + scrollBarWidth +
                  (config.box ? 2 * BORDER_THICKNESS : 0);

            width = cc.widthFromNative(totalWidth);
         }
      }

      if (implicitHeight)
      {
         int nativeHeight = calcImplicitBrowseHeightNoScrollbar();

         int viewportWidth = cc.widthToNative(width) - scrollBarWidth -
                                (config.box ? 2 * BORDER_THICKNESS : 0);

         if (totalColumnsWidth > viewportWidth)
         {
            nativeHeight += scrollBarWidth;
         }

         height = cc.heightFromNative(nativeHeight);
      }

      setSize(width, height);
   }

   /**
    * Get browse native height if it is not specified explicitly (based on explicit or implicit
    * DOWN value). Height is calculated as if the horizontal scrollbar is not used.
    *
    * @return see above.
    */
   private int calcImplicitBrowseHeightNoScrollbar()
   {
      int rowHeight = calcRowHeight();

      int nativeHeight = 0;
      if (hasTitle())
      {
         nativeHeight += calcTitleHeight() + 2 * BORDER_THICKNESS;
      }

      if (config.box)
      {
         nativeHeight += hasTitle() ? BORDER_THICKNESS : 2 * BORDER_THICKNESS;
      }

      if (isLabelsVisible())
      {
         nativeHeight += rowHeight + 2 * BORDER_THICKNESS;
      }

      int down = config.down > 0 ? config.down : DEFAULT_DYNAMIC_DOWN_SIZE;
      nativeHeight += down * rowHeight + (down - 1) * BORDER_THICKNESS;

      return nativeHeight;
   }
   
   /**
    * Obtain the current marker column. If if does not exist, return {@code null}. 
    * 
    * @return  the current marker column or {@code null} if not yet created. 
    */
   public BrowseColumnGuiImpl getCurrentMarkersColumn()
   {
      return markersColumn;
   }
   
   /**
    * Create markers (leftmost) column if it doesn't exist yet.
    */
   private void createMarkersColumn()
   {
      if (markersColumn == null)
      {
         markersColumn = new BrowseColumnGuiImpl(true);
         markersColumn.initMarkersColumn(this);
      }
   }
   
   /**
    * Calculate DOWN value.
    */
   private int calcDown()
   {
      return calcDown(getRowHeight() - CELL_OFFSET_TOP - CELL_OFFSET_BOTTOM);
   }

   /**
    * Calculate DOWN value for the browse basing on the given row height.
    *
    * @param  rowHeightPixels
    *         Row height in pixels.
    *
    * @return calculated DOWN value.
    */
   @Override
   public int calcDown(int rowHeightPixels)
   {
      // Calculate DOWN. It is allowed to hide bottom separator + 1 px of the last cell.
      int availableSpace = getViewportDimension().height;
      if (isLabelsVisible())
      {
         availableSpace -= getLabelsAreaHeight();
      }

      availableSpace += ALLOWED_HIDDEN_CELL_BOTTOM_PART;

      int paddedHeight = rowHeightPixels + CELL_OFFSET_TOP + CELL_OFFSET_BOTTOM;

      // exclude bottom border from the calculations
      return ((availableSpace - paddedHeight) / (paddedHeight + BORDER_THICKNESS)) + 1;
   }

   /**
    * Get scroll panel location.
    *
    * @return scroll panel location.
    */
   private Point getScrollPaneLocation()
   {
      int y = 0;
      if (hasTitle())
      {
         y += titleHeight + 2 * BORDER_THICKNESS;
      }

      return cc.pointFromNative(new NativePoint(0, y));
   }

   /**
    * Get locked panel location.
    *
    * @return locked panel location.
    */
   private Point getLockedPaneLocation()
   {
      int y = 0;
      int x = 0;
      if (config.box)
      {
         y += BORDER_THICKNESS;
         x += BORDER_THICKNESS;
      }

      if (hasTitle())
      {
         y += titleHeight + BORDER_THICKNESS;
      }

      return cc.pointFromNative(new NativePoint(x, y));
   }

   /**
    * Get scroll pane dimension.
    *
    * @return scroll pane dimension.
    */
   private NativeDimension getScrollPaneDimension()
   {
      NativeDimension dimension = physicalDimension();

      if (hasTitle())
      {
         dimension.height -= titleHeight + 2 * BORDER_THICKNESS;
      }

      return dimension;
   }

   /**
    * Calculate row height basing on the ROW-HEIGHT-CHARS or font height.
    *
    * @return row height.
    */
   private int calcRowHeight()
   {
      if (!config.realized || config.rowHeightChars <= 0)
      {
         return calcImplicitRowHeight();
      }
      else
      {
         return cc.heightToNative(config.rowHeightChars) + CELL_OFFSET_TOP + CELL_OFFSET_BOTTOM;
      }
   }

   /**
    * Calculate implicit row height basing on the font height.
    *
    * @return implicit row height.
    */
   private int calcImplicitRowHeight()
   {
      return browseFontResolver.font().fontHeight + CELL_OFFSET_TOP + CELL_OFFSET_BOTTOM;
   }

   /**
    * Determines if this browse has title.
    *
    * @return <code>true</code> if this browse has title.
    */
   public boolean hasTitle()
   {
      return !config.title.isEmpty();
   }

   /**
    * Resolve colors.
    */
   private void resolveColors()
   {
      Window window = window();

      if (colorResolver == null)
      {
         colorResolver = new GuiColorResolver(gd, config, SysColor.COLOR_WINDOW,
                                                          SysColor.COLOR_WINDOWTEXT);
      }

      colorResolver.colors();
      colorResolver.titleFgColor = colorResolver.sysBgColor;
      colorResolver.titleBgColor = gd.getSysColor(SysColor.COLOR_ACTIVECAPTION);
      colorResolver.boxColor = gd.getSysColor(SysColor.COLOR_WINDOWTEXT);

      backgroundColor = colorResolver.bgColor;

      if (config().separators)
      {
         int color = config().separatorFgColor;
         if (color >= 0)
         {
            separatorColor = ColorManager.getFgColor(window, color);
         }
         else if (!SEPARATOR_DEFAULT_COLOR.equals(backgroundColor))
         {
            separatorColor = SEPARATOR_DEFAULT_COLOR;
         }
         else
         {
            separatorColor = SEPARATOR_DEFAULT_COLOR2;
         }
      }
      else
      {
         separatorColor = backgroundColor;
      }
   }

   /**
    * Resolve fonts for browse and its columns.
    */
   public void resolveFonts()
   {
      if (config.enhancedBrowse)
      {
         int windowId = window().getId().asInt();
         config.ehFont = FontManager.registerFontDetails(windowId, config.ehFont);
         config.ehLabelFont = FontManager.registerFontDetails(windowId, config.ehLabelFont);
      }

      Window window = window();
      browseFontResolver = new GuiFontResolver(window, gd, config(), false, config().ehFont);

      if (hasTitle())
      {
         // browse title uses the window title font
         LabelConfig labelConfig = new LabelConfig();
         labelConfig.font = FontManager.WINDOW_TITLE_FONT;
         titleFontResolver = new GuiFontResolver(window, gd, labelConfig, false);
      }

      int colNo = getColumnCount();
      for (int i = 0; i < colNo; i++)
      {
         getColumn(i).resolveFonts();
      }
   }

   /**
    * Get the visible browse column with the specific index.
    *
    * @param   visibleColumnIndex
    *          0-based index of the column among the visible columns.
    *
    * @return widget representing the target column.
    */
   private BrowseColumnGuiImpl getVisibleColumn(int visibleColumnIndex)
   {
      return getColumn(getVisibleColumns(), visibleColumnIndex);
   }

   /**
    * Get the browse column with the specific index.
    *
    * @param   columnOrdinal
    *          0-based index of the column among all columns (i.e. column ordinal).
    *
    * @return widget representing the target column.
    */
   public BrowseColumnGuiImpl getColumn(int columnOrdinal)
   {
      return getColumn(getColumns(), columnOrdinal);
   }

   /**
    * Get the browse column with the specific index in the given array.
    *
    * @param   columns
    *          Array of column configurations.
    * @param   index
    *          0-based index of the column in the given array.
    *
    * @return widget representing the target column.
    */
   private BrowseColumnGuiImpl getColumn(BrowseColumnConfig[] columns, int index)
   {
      if (columns == null)
         return null;

      return (BrowseColumnGuiImpl) ThinClient.getInstance().getWidget(columns[index].id);
   }

   /**
    * Get thickness (width or height) of the corresponding scroll bar.
    *
    * @param  orientation
    *         Target scroll bar.
    *
    * @return thickness (width or height) of the corresponding scroll bar.
    */
   private int getScrollbarThickness(ScrollBar.Orientation orientation)
   {
      ScrollBar scrollBar = scrollPane.getScrollBar(orientation);
      NativeDimension d = scrollBar.physicalDimension();

      return orientation == ScrollBar.Orientation.VERTICAL ? d.width : d.height;
   }

   /**
    * Get text of the cell from the cache containing intermediate rows
    * ({@link #intermediateDcData}), if it is available, or normal cache ({@link #dcData})
    *
    * @param  column
    *         The column the target cell belongs to.
    * @param  viewportIndex
    *         0-based index of the target row in the viewport the target cell belongs to.
    *
    * @return text of the target cell.
    */
   public String getCellTextWithIntermediate(BrowseColumnGuiImpl column, int viewportIndex)
   {
      if (intermediateDcData != null)
      {
         return intermediateDcData[viewportIndex].text[column.config().ordinal];
      }

      return getCellText(viewportIndex, column.config().ordinal);
   }

   /**
    * Get value of the cell from the cache containing intermediate rows
    * ({@link #intermediateDcData}), if it is available, or normal cache ({@link #dcData})
    *
    * @param  column
    *         The column the target cell belongs to.
    * @param  viewportIndex
    *         0-based index of the target row in the viewport the target cell belongs to.
    *
    * @return value of the target cell.
    */
   public BaseDataType getCellValueWithIntermediate(BrowseColumnGuiImpl column, int viewportIndex)
   {
      if (intermediateDcData != null)
      {
         return intermediateDcData[viewportIndex].data[column.config().ordinal];
      }

      return getCellValue(viewportIndex, column.config().ordinal);
   }

   /**
    * Returns <code>true</code> if the specified row is drawn as selected.
    *
    * @param  viewportRowIndex
    *         0-based index of the target row in the viewport.
    *
    * @return <code>true</code> if the specified row is drawn as selected.
    */
   public boolean drawRowAsSelected(int viewportRowIndex)
   {
      return !isEditing() && !isSavingRow() &&
             isRowSelected(getCachedRowNumberWithIntermediate(viewportRowIndex));
   }

   /**
    * Get value of the cell.
    *
    * @param  viewportIndex
    *         0-based index of the target row in the viewport the target cell belongs to.
    * @param  columnOrdinal
    *         0-based index of the column the target cell belongs to.
    *
    * @return value of the target cell.
    */
   public BaseDataType getCellValue(int viewportIndex, int columnOrdinal)
   {
      BaseDataType var = null;

      int rowIndex = getCachedRowNumber(viewportIndex);
      if (getCurrentRow() == rowIndex &&
            currRowUpdates != null &&
            currRowUpdates[columnOrdinal] != null)
      {
         var = currRowUpdates[columnOrdinal];
      }

      if (var != null)
      {
         return var;
      }
      else if (isEditing() &&
            rowIndex == editedRowIndex &&
            editedRow != null)
      {
         return editedRow.data[columnOrdinal];
      }
      else
      {
         return dcData[viewportIndex].data[columnOrdinal];
      }
   }

   /**
    * Returns <code>true</code> if the given column is the current column and editing is in
    * progress.
    *
    * @param  column
    *         Column to check.
    *
    * @return see above.
    */
   public boolean isCurrentEditingColumn(BrowseColumnGuiImpl column)
   {
      return isEditing() && column.getVisibleColumnIndex() == currColumn;
   }

   /**
    * Get text of the cell.
    *
    * @param  viewportIndex
    *         0-based index of the target row in the viewport the target cell belongs to.
    * @param  columnOrdinal
    *         0-based index of the column the target cell belongs to.
    *
    * @return text of the target cell.
    */
   public String getCellText(int viewportIndex, int columnOrdinal)
   {
      BaseDataType var = null;

      int rowIndex = getCachedRowNumber(viewportIndex);
      if (getCurrentRow() == rowIndex &&
          currRowUpdates != null &&
          currRowUpdates[columnOrdinal] != null)
      {
         var = currRowUpdates[columnOrdinal];
      }

      if (var != null)
      {
         return getCellTextNoError(getColumns()[columnOrdinal], var);
      }
      else if (isEditing() &&
               rowIndex == editedRowIndex &&
               editedRow != null)
      {
         return editedRow.text[columnOrdinal];
      }
      else
      {
         return dcData[viewportIndex].text[columnOrdinal];
      }
   }

   /**
    * Returns <code>true</code> if CTRL key was held down when a mouse button was pressed.
    *
    * @param  e
    *         Mouse event.
    *
    * @return <code>true</code> if CTRL key was held down when a mouse button was pressed.
    */
   boolean isCtrlModifier(MouseEvent e)
   {
      return (e.getModifiers()   & MouseEvent.CTRL_MASK)      != 0 ||
             (e.getModifiersEx() & MouseEvent.CTRL_DOWN_MASK) != 0;
   }
   
   /**
    * Register column that will generate next focus events. When the events are all dispatched,
    * the {@link BrowseColumnGuiImpl} is reponsible for resetting {@code meColumn} back to
    * {@code null}.
    * 
    * @param   col
    *          The new column to be registered or {@code null} for unregister process.
    */
   void setMouseActionColumn(BrowseColumnGuiImpl col)
   {
      meColumn = col;
   }
   
   /**
    * Handle row clicks.
    *
    * @param viewportRowIndex
    *        0-based viewport index of the clicked row.
    * @param visibleColumnIndex
    *        0-based index of the clicked column among the visible columns. If -1 then the
    *        whole row is selected as in read-only mode.
    * @param e
    *        Original mouse event.
    */
   void rowClicked(int viewportRowIndex, int visibleColumnIndex, MouseEvent e)
   {
      // Special case when the area UNDER a toggle-box widget is clicked. That can happen when
      // we click a toggle-box in a browse and browse gains the focus. Toggle-box is created on
      // focus gaining, but the click is received by the browse and toggle-box misses the click.
      if (isEditorToggleBox() &&
          visibleColumnIndex == currColumn &&
          viewportRowIndex == getViewportRow() &&
          getVisibleColumn(visibleColumnIndex).isInToggleBoxArea(e))
      {
         getActiveToggleBox().mouseClicked(e);
         return;
      }

      if (activeEditor != null &&
          visibleColumnIndex == currColumn &&
          viewportRowIndex == getViewportRow())
      {
         return;
      }
      
      //TODO is viewportRowIndex < 0 column labels are pushed down
      mouseFocusGained = true;
      requestFocus();   // TODO: is this really necessary? Focus will be requested by system call
      
      if (viewportRowIndex < getCacheSize() && viewportRowIndex >= 0)
      {
         if (isEditing())
         {
            if (viewportRowIndex == getViewportRow() && isCurrentRowInView())
            {
               if (!stopEdit(null, false, true))
               {
                  return;
               }
            }
            else if (!leaveRow())
            {
               mouseFocusGained = false; // in this case zap mode and selection is on
               return;
            }
         }

         int oldRow = getCurrentRow();

         int row = getCachedRowNumber(viewportRowIndex);
         setCurrentRow(row);

         if (isSelectionMultiple())
         {
            if (isShiftModifier(e))
            {
               int min = Math.min(oldRow, getCurrentRow());
               int max = Math.max(oldRow, getCurrentRow());

               selectedRowIndices.clear();
               for (int i = min; i <= max; i++)
               {
                  selectedRowIndices.add(new SelectedRow(i));
               }
            }
            else if (isCtrlModifier(e))
            {
               switchCurrentRowSelection();
            }
            else
            {
               deselectAll();
               selectFocusedRow();
            }
         }

         if (visibleColumnIndex >= 0 && isEditPossible())
         {
            BrowseColumnGuiImpl column = getVisibleColumn(visibleColumnIndex);
            if (!column.config().readOnly)
            {
               currColumn = visibleColumnIndex;
               setCurrentSingleRowSelected();

               boolean toggleClicked = false;
               if (TOGGLE_BOX.equals(column.config().widgetType))
               {
                  toggleClicked = column.isInToggleBoxArea(e);
               }

               startEdit();

               if (toggleClicked)
               {
                  // inverse-check toggle-box immediately
                  getActiveToggleBox().mouseClicked(e);
               }
            }
         }
         else 
         {
            finishStopEdit();
            ThinClient.setHelp(config.help, this);
         }
         
         if (oldRow != row)
         {
            invokeRowEntryTrigger(true);
            fireValueChanged();
         }

         // check double click
         if (e.getClickCount() >= 2 && !isEditing())
         {
            fireDefaultAction();
         }

         resetPressedColumns(false);

         repaint();
      }
   }

   /**
    * Set all column in non-pressed state.
    *
    * @param trigger
    *        If <code>true</code> and a column was in pressed state, raise END-SEARCH event.
    */
   void resetPressedColumns(boolean trigger)
   {
      BrowseColumnConfig[] cols = getVisibleColumns();
      for (BrowseColumnConfig col : cols)
      {
         if (col.pressed && trigger)
         {
            ThinClient.getInstance().invokeTriggers(this, Keyboard.SE_END_SEARCH, false);
         }

         col.pressed = false;
      }
   }

   /**
    * Returns <code>true</code> if SHIFT key was held down when a mouse button was pressed.
    *
    * @param  e
    *         Mouse event.
    *
    * @return <code>true</code> if SHIFT key was held down when a mouse button was pressed.
    */
   private boolean isShiftModifier(MouseEvent e)
   {
      return (e.getModifiers()   & MouseEvent.SHIFT_MASK)      != 0 ||
             (e.getModifiersEx() & MouseEvent.SHIFT_DOWN_MASK) != 0;
   }

   /**
    * Set COLUMN-SCROLLING or non-COLUMN-SCROLLING modes.
    *
    * @param columnScrolling
    *        <code>true</code> for COLUMN-SCROLLING mode, <code>false</code> for
    *        non-COLUMN-SCROLLING mode,
    */
   private void setColumnScrolling(boolean columnScrolling)
   {
      if (columnScrolling)
      {
         updateHorizontalColumnScrollMax();
      }

      scrollableColumnSet.setPhysicalLocation(0, 0);
   }

   /**
    * Show or hide horizontal scrollbar.
    *
    * @param show
    *        <code>true</code> to show, <code>false</code> to hide.
    */
   private void showHorizontalScrollbar(boolean show)
   {
      scrollPane.setShowBars(getShowBars(show, config.scrollbarVertical));
      // updates viewport dimension
      horizontalScrollbar.setVisible(show);
   }

   /**
    * Show or hide vertical scrollbar depending on (NO-)SCROLLBAR-VERTICAL flag state.
    */
   private void updateVerticalScrollbarVisibility()
   {
      scrollPane.setShowBars(getShowBars(horizontalScrollbar.isVisible(),
                                         config.scrollbarVertical));
      // updates viewport dimension
      verticalScrollbar.setVisible(config.scrollbarVertical);
   }

   /**
    * Get {@link ScrollPaneGuiImpl.ShowBars} option which corresponds given visibility of the
    * scroll bars.
    *
    * @param  horizontal
    *         <code>true</code> if horizontal scroll bar is visible.
    * @param  vertical
    *         <code>true</code> if vertical scroll bar is visible.
    *
    * @return {@link ScrollPaneGuiImpl.ShowBars} option which corresponds given visibility of
    *         the scroll bars.
    */
   private ScrollPaneGuiImpl.ShowBars getShowBars(boolean horizontal, boolean vertical)
   {
      ScrollPaneGuiImpl.ShowBars showBars;
      if (vertical)
      {
         showBars = horizontal ? ScrollPaneGuiImpl.ShowBars.ALWAYS :
                                 ScrollPaneGuiImpl.ShowBars.VERTICAL_ALWAYS;
      }
      else
      {
         showBars = horizontal ? ScrollPaneGuiImpl.ShowBars.HORIZONTAL_ALWAYS :
                                 ScrollPaneGuiImpl.ShowBars.NEVER;
      }
      return showBars;
   }

   /**
    * Select current row for single-selection case.
    */
   private void setCurrentSingleRowSelected()
   {
      if (!isSelectionMultiple() && getCacheSize() > 0)
      {
         currRowSelected = true;
         repaint();
      }
   }

   /**
    * Perform search by key: find the row which starts with the given character. Complex rules are
    * applied, see {@link BrowseWidget#getRows(char, int, int, int, int, ScreenBuffer[])} for more
    * information.
    *
    * @param searchChar
    *        Character to be searched for.
    */
   private void searchByKey(char searchChar)
   {
      int searchColumn = getSearchColumn();
      if (searchColumn < 0)
      {
         return;
      }

      // search within the current view first
      if (isCurrentRowInView())
      {
         for (int i = getViewportRow() + 1; i < getCacheSize(); i++)
         {
            if (cellStartsWith(getCellText(i, searchColumn), searchChar))
            {
               setCurrentRow(getCachedRowNumber(i));
               repaint();
               return;
            }
         }
      }

      dcData = getRows(searchChar, searchColumn, getFirstExistingRow(), getLastExistingRow());
      setCurrentRow(browseModel.getCurrentRow());
      repaint();
   }

   /**
    * Get the search column, i.e. column which is searched when a key is pressed. If
    * ALLOW-COLUMN-SEARCHING is <code>false</code>, then it is the first column. If
    * ALLOW-COLUMN-SEARCHING is <code>true</code> then it is the pressed column.
    *
    * @return 0-based index of the search column or <code>-1</code> if there is no column to
    *         search.
    */
   private int getSearchColumn()
   {
      BrowseColumnConfig[] columns = getVisibleColumns();

      if (columns.length == 0)
      {
         return -1;
      }
      else if (!config.allowColumnSearching)
      {
         // ALLOW-COLUMN-SEARCHING is disabled
         return 0;
      }
      else
      {
         // ALLOW-COLUMN-SEARCHING is enabled

         int i = 0;
         for (BrowseColumnConfig column : columns)
         {
            if (column.pressed)
            {
               return i;
            }
            i++;
         }

         // no row is selected
         return -1;
      }
   }

   /**
    * Get ancestor Window for this widget.
    *
    * @return  ancestor Window or <code>null</code> if the widget is not yet attached to a Window.
    */
   @Override
   public Window window()
   {
      Window window = parent(Window.class);

      if (window == null && (config.frameId != -1))
      {
         window = ThinClient.getInstance().getFrame(config.frameId).window();
      }

      return window;
   }

   /**
    * Get value of the given column in current row.
    *
    * @param   visibleIndex
    *          0-based index of the column among visible columns.
    *
    * @return  The value of the given column in current row.
    */
   public BaseDataType getColumnValue(int visibleIndex)
   {
      if (!isCurrentRowInView())
      {
         return new character();
      }

      return super.getColumnValue(visibleIndex);
   }

   /**
    * Process key event.
    *
    * @param   event
    *          Reference to keyboard event.
    */
   @Override
   public void onKeyPressed(KeyInput event)
   {
      int key = event.actionCode();

      if (key == Keyboard.KA_CURSOR_RIGHT)
      {
         event.consume();
         if (horizontalScrollbar.isVisible())
         {
            horizontalScrollbar.onButtonClick(ScrollBar.Position.RIGHT, ScrollStep.SHORT, null);
         }
      }
      else if (key == Keyboard.KA_CURSOR_LEFT)
      {
         event.consume();
         if (horizontalScrollbar.isVisible())
         {
            horizontalScrollbar.onButtonClick(ScrollBar.Position.LEFT, ScrollStep.SHORT, null);
         }
      }
      else if (key == Keyboard.KA_CURSOR_DOWN &&
               !isSelectionMultiple() &&
               !currRowSelected &&
               getCacheSize() > 0 &&
               getCurrentRow() == 0)
      {
         // TODO handle the case when the first row is deleted

         // Do not change current row, just select the first row.
         setCurrentSingleRowSelected();

         if (!isCurrentRowInView())
         {
            scrollToRow(0);
         }

         event.consume();
         return;
      }
      else if ((key == Keyboard.KA_CURSOR_DOWN || key == Keyboard.KA_CURSOR_UP) &&
               !isCurrentRowInView() &&
               getCacheSize() > 0)
      {
         event.consume();
         boolean upper = getCurrentRow() < getFirstExistingRow();

         // TODO check if current row is a new row

         if (key == Keyboard.KA_CURSOR_DOWN)
         {
            int oldRow = getCurrentRow();
            int newRow = oldRow + 1;
            if (isRowInView(newRow))
            {
               setCurrentRow(newRow);
               repaint();
            }
            else if (upper)
            {
               scrollToRow(newRow, true);
            }
            else
            {
               scrollToRow(newRow, true, false, false, true, true, getPageSize(), false);
            }

            if (oldRow != getCurrentRow())
            {
               setCurrentSingleRowSelected();
            }
         }
         else // Keyboard.KA_CURSOR_UP
         {
            int newRow = Math.max(0, getCurrentRow() - 1);
            if (upper)
            {
               scrollToRow(newRow, true);
               if (!isCurrentRowInView() && getCacheSize() > 0)
               {
                  setCurrentRow(getCachedRowNumber(0));
                  repaint();
               }
            }
            else
            {
               int repoRow = Math.max(0, getPageSize() - UP_KEY_BOTTOM_OFFSET_ROWS);
               scrollToRow(newRow, true, false, false, true, true, repoRow, false);
            }

            // UP always selects row
            setCurrentSingleRowSelected();
         }

         return;
      }
      else if (key == Keyboard.KA_PAGE_DOWN)
      {
         // TODO check for new rows
         event.consume();

         boolean editing = isEditing();
         if (editing && !leaveRow())
         {
            return;
         }

         int oldRow = getCurrentRow();
         if (isCurrentRowInView())
         {
            int newViewportRow = getViewportRow() + getPageSize();
            if (getCacheSize() < getDown())
            {
               newViewportRow = Math.min(newViewportRow, getCacheSize() - 1);
            }

            if (newViewportRow <= getCacheSize() - 1)
            {
               setCurrentRow(getCachedRowNumber(newViewportRow));
               if (oldRow != getCurrentRow())
               {
                  setCurrentSingleRowSelected();
               }

               if (editing)
               {
                  startEdit();
               }

               repaint();
               return;
            }
         }

         int newRow = getCurrentRow() + getPageSize();
         // TODO if there is overlap, it should be iterative scrolling
         // or may be even it should be iterative scrolling only
         scrollToRow(newRow, true, false, false, true, true, getPageSize(), false);
         if (oldRow != getCurrentRow())
         {
            setCurrentSingleRowSelected();
         }

         if (editing)
         {
            startEdit();
         }
         repaint();
         return;
      }
      else if (key == Keyboard.KA_PAGE_UP)
      {
         // TODO check for new rows
         event.consume();

         boolean editing = isEditing();
         if (editing && !leaveRow())
         {
            return;
         }

         int oldRow = getCurrentRow();
         if (isCurrentRowInView())
         {
            int newViewportRow = getViewportRow() - getPageSize();
            if (newViewportRow >= 0)
            {
               setCurrentRow(getCachedRowNumber(newViewportRow));
               if (oldRow != getCurrentRow())
               {
                  setCurrentSingleRowSelected();
               }

               if (editing)
               {
                  startEdit();
               }
               repaint();
               return;
            }
         }

         int newRow = Math.max(0, getCurrentRow() - getPageSize());
         scrollToRow(newRow, true);
         if (oldRow != getCurrentRow())
         {
            setCurrentSingleRowSelected();
         }

         if (editing)
         {
            startEdit();
         }
         repaint();
         return;
      }
      else if (!isEditing() &&
               !event.isTriggerFired() &&
               Keyboard.isPrintableKey(key) &&
               key != ' ')
      {
         searchByKey((char) key);
         return;
      }

      int oldRow = getCurrentRow();
      super.onKeyPressed(event);

      if (key == Keyboard.KA_CURSOR_UP)
      {
         setCurrentSingleRowSelected();
      }
      else if ((key == Keyboard.KA_HOME || key == Keyboard.KA_END ||
                key == Keyboard.KA_LEFT_END || key == Keyboard.KA_RIGHT_END) &&
               (oldRow != getCurrentRow()))
      {
         setCurrentSingleRowSelected();
      }
   }

   /**
    * Intercept special events. Using regular keyTyped/keyPressed may not work
    * if component is not in focus.
    *
    * @see  Widget#processEvent(Event)
    */
   @Override
   public void processEvent(Event evt)
   {
      if (evt instanceof KeyInput)
      {
         KeyInput ki = (KeyInput) evt;
         int action = ki.actionCode();
         if (action == Keyboard.SE_END_SEARCH)
         {
            if (!ki.isConsumed())
            {
               resetPressedColumns(false);
               repaint();
            }
            return;
         }
         else if (action == Keyboard.SE_LEAVE && isEditing())
         {
            // commit changes when leaving the browse
            releaseCurrentRow(true);
         }
      }

      super.processEvent(evt);
   }

   /**
    * Notify the browse that the underlying query has been repositioned..
    *
    * @param   repositionError
    *          For reposition mode - if an error has happened during reposition.
    * @param   targetRepositionRow
    *          For reposition mode - row used as the target for reposition. If it was deleted,
    *          reposition with fetching may end up on a different row.
    */
   @Override
   public void queryRepositioned(boolean repositionError, int targetRepositionRow)
   {
      super.queryRepositioned(repositionError, targetRepositionRow);

      if (!initialRepositionDone)
      {
         initialRepositionDone = true;
         if (isVisible())
         {
            setCurrentSingleRowSelected();
         }
      }
      else
      {
         setCurrentSingleRowSelected();
      }
   }

   /**
    * Refresh the browse cache and redisplay the browse.
    *
    * @param deleteCurrentRowMode
    *        <code>true</code> if this function is used for update on DELETE-CURRENT-ROW.
    *        <code>false</code> if this function is used for update on REFRESH.
    */
   @Override
   public void refresh(boolean deleteCurrentRowMode)
   {
      boolean initialRowSet = initialRow != null;

      super.refresh(deleteCurrentRowMode);

      if (initialRowSet)
      {
         setCurrentSingleRowSelected();
      }
   }

   /**
    * Set the MOUSE-POINTER attribute
    * 
    * @param name
    *        the attribute value
    */
   @Override
   public void setMousePointerName(String name)
   {
      config.mousePointer = name;
   }

   /**
    * Get the MOUSE-POINTER attribute
    * 
    * @return the attribute value 
    *        
    */
   @Override
   public String getMousePointerImpl()
   {
      return config.mousePointer;
   }

   /**
    * Implements the client side of the MOVE-COLUMN() widget method.<p>
    * Repositions a column in a browse widget.
    *
    * @param  srcIndex
    *         0-based position of the column to be moved (column ordinal).
    * @param  destIndex
    *         0-based position to which the column is moved (column ordinal).
    */
   @Override
   public void moveColumn(int srcIndex, int destIndex)
   {
      if (!config.performLayout)
      {
         // update currColumn using the old cache
         int curOrdinal = visibleToOrdinal(currColumn);
         if (srcIndex == curOrdinal)
         {
            currColumn = ordinalToVisible(destIndex);
         }
         else if (srcIndex < curOrdinal && destIndex >= curOrdinal)
         {
            currColumn--;
         }
         else if (srcIndex > curOrdinal && destIndex <= curOrdinal)
         {
            currColumn++;
         }
      }

      // columns have been reordered in BrowseWidget.columns and in config.columnIds, update
      // ordinals accordingly
      populateColumnsCache();
      BrowseColumnConfig[] configs = getColumns();
      for (int i = 0; i < configs.length; i++)
      {
         configs[i].ordinal = i;
      }

      if (config.performLayout)
      {
         return;
      }

      rebuildColumns();

      if (dcData != null)
      {
         for (BrowseRow row : dcData)
         {
            moveColumnInArray(row.data, srcIndex, destIndex);
            moveColumnInArray(row.text, srcIndex, destIndex);
            moveColumnInArray(row.attr, srcIndex, destIndex);
         }
      }
      if (editedRow != null && !isRowInView(editedRowIndex))
      {
         moveColumnInArray(editedRow.data, srcIndex, destIndex);
         moveColumnInArray(editedRow.text, srcIndex, destIndex);
         moveColumnInArray(editedRow.attr, srcIndex, destIndex);
      }
      moveColumnInArray(currRowUpdates, srcIndex, destIndex);
      browseModel.resetEditorsConfigsCache();

      if (config.columnScrolling)
      {
         // 4GL quirk - scroll to the first column only if COLUMN-SCROLLING is on
         scrollToColumn(0);
      }

      updateEditorAfterColumnModification();
      repaint();
   }

   /**
    * Client-side part of the code which applies the given enhanced configuration to the browse.
    *
    * @param  config
    *         Enhanced browse config.
    */
   @Override
   public void applyEnhancedConfigClient(EnhancedBrowseConfig config)
   {
      currColumn = 0;

      // columns have been reordered in BrowseWidget.columns and in config.columnIds, update
      // ordinals accordingly
      populateColumnsCache();
      BrowseColumnConfig[] configs = getColumns();
      for (int i = 0; i < configs.length; i++)
      {
         configs[i].ordinal = i;
      }

      if (this.config.performLayout)
      {
         return;
      }

      rebuildColumns();
      browseModel.resetEditorsConfigsCache();

      refresh(false);
      scrollToColumn(0);
      repaint();
   }

   /**
    * Checks if the given widget is able to be the target of the drop operation.
    * 
    * @return   <code>TRUE</code> if this widget can be droppable at this time,
    *           <code>FALSE</code> otherwise.
    */
   @Override
   public boolean isDroppable()
   {
      return config.dropTarget; 
   }

   /**
    * Get position of the horizontal scroll bar.
    *
    * @return position of the horizontal scroll bar.
    */
   public int getHorizontalScrollingPosition()
   {
      return horizontalScroller.getPosition();
   }

   /**
    * Start browse layout mode.
    */
   public void startLayoutMode()
   {
      if (browseLayoutMode || (isEditing() && !leaveRow()))
      {
         return;
      }

      EnhancedBrowseConfig backupState = new EnhancedBrowseConfig(this);
      if (!backupState.getKey().isValid())
      {
         ThinClient tc = ThinClient.getInstance();
         tc.invokeLaterOS(() ->
               tc.messageBox(new Object[]{"Cannot create unique key for this browse, " +
                                          "enhanced mode is not available!"},
                             null,
                             LogicalTerminal.ALERT_ERROR,
                             LogicalTerminal.BTN_OK,
                             "Enhanced browse mode",
                             null,
                             WindowConfig.RESOLVE_WINDOW,
                             null));
         return;
      }

      this.undoConfig = backupState;
      browseLayoutMode = true;
      afterConfigUpdate(config);
      repaint();
   }

   /**
    * Stop browse layout mode.
    */
   public void stopLayoutMode(boolean undo, EnhancedBrowseSaveTarget set)
   {
      if (!browseLayoutMode)
      {
         return;
      }

      if (undo)
      {
         browseModel.applyEnhancedConfig(undoConfig);
      }
      else
      {
         if (!browseModel.saveEnhancedConfig(set))
         {
            return;
         }
      }

      afterConfigUpdate(config);
      undoConfig = null;
      browseLayoutMode = false;
      repaint();
   }

   /**
    * Apply default 4GL configuration (non-enhanced) for this browse.
    *
    * @return <code>true</code> on success.
    */
   public boolean applyDefault4glConfig()
   {
      return browseModel.applyDefault4glConfig();
   }

   /**
    * Get save targets for an enhanced browse configuration, which are available for the current
    * user.
    *
    * @return see above.
    */
   public EnhancedBrowseSaveTarget[] getEnhancedBrowseSaveTargets()
   {
      if (enhancedSaveTargets == null)
      {
         enhancedSaveTargets = browseModel.getEnhancedBrowseSaveTargets();
      }

      return enhancedSaveTargets;
   }

   /**
    * Returns <code>true</code> if browse layout mode is active.
    *
    * @return <code>true</code> if browse layout mode is active.
    */
   public boolean isLayoutMode()
   {
      return browseLayoutMode;
   }

   /**
    * Returns <code>true</code> if columns are movable.
    *
    * @return <code>true</code> if columns are movable.
    */
   public boolean isColumnMovable()
   {
      return config().columnMovable || isLayoutMode();
   }

   /**
    * Returns <code>true</code> if rows are resizable.
    *
    * @return <code>true</code> if rows are resizable.
    */
   public boolean isRowResizable()
   {
      return config().rowResizable || isLayoutMode();
   }

   /**
    * Returns <code>true</code> if columns are resizable.
    *
    * @return <code>true</code> if columns are resizable.
    */
   public boolean isColumnResizable()
   {
      return config().columnResizable || isLayoutMode();
   }

   /**
    * Returns optimal size.
    *
    * @return  see above
    */
   @Override
   public double getOptimalWidth()
   {
      return getTotalColumnsWidth() + 2 * BORDER_THICKNESS;
   }
   
   /**
    * Repositions the element in the array.
    *
    * @param array
    *        Target array.
    * @param srcIndex
    *        0-based position of the element to be moved.
    * @param destIndex
    *        0-based position to which the element is moved.
    */
   private void moveColumnInArray(Object[] array, int srcIndex, int destIndex)
   {
      if (array == null)
      {
         return;
      }

      Object edgeElement = array[srcIndex];
      if (destIndex > srcIndex)
      {
         System.arraycopy(array, srcIndex + 1, array, srcIndex, destIndex - srcIndex);
      }
      else
      {
         System.arraycopy(array, destIndex, array, destIndex + 1, srcIndex - destIndex);
      }
      array[destIndex] = edgeElement;
   }
   
   /**
    * Set of arranged columns. In browse there are two sets: the set of locked columns (including
    * row markers column) and the set of scrollable columns.
    */
   private class ColumnSetContainer
   extends NativeScrollContainer<GuiOutputManager>
   {
      /**
       * <code>true</code> for locked column set. <code>false</code> for scrollable column set.
       */
      private final boolean locked;

      /**
       * Constructor.
       *
       * @param locked
       *        <code>true</code> for locked column set. <code>false</code> for scrollable column
       *        set.
       */
      ColumnSetContainer(boolean locked)
      {
         this.locked = locked;
      }

      /**
       * Get the scroll pane actual bounds.
       * 
       * @return   The scroll pane actual bounds
       */
      @Override
      public NativeRectangle getActualBounds()
      {
         NativeRectangle viewPortRect = getViewportBoundedRectangle();
         NativeRectangle actualBounds = viewPortRect.intersection(super.getActualBounds());
         
         return actualBounds;
      }
      
      /**
       * Initialize the widget.
       */
      public void initialize()
      {
         config = new BaseConfig();
         config.id = WidgetId.nextID();
         OutputManager.instance().getRegistry().addWidget((Widget) this);
      }

      /**
       * Draw the widget.
       */
      @Override
      public void draw()
      {
         final NativePoint p = this.physicalLocation();
         final NativeDimension d = this.physicalDimension();
         final NativeDimension clipDimension = new NativeDimension(d.width, d.height);
         if (locked &&
             BrowseGuiImpl.this.config.separators &&
             BrowseGuiImpl.this.config.numLockedColumns < getVisibleColumnCount())
         {
            clipDimension.width += BORDER_THICKNESS; // for separator
         }
         final NativeRectangle clip = clipRectangle(p, clipDimension);

         gd.draw(p, clip, () ->
         {
            // draw the background. It will only be visible the extreme right and bottom where
            // cells might not overlap it
            gd.setColor(backgroundColor);

            if (locked && isLabelsVisible())
            {
               // Do not fill the top of the rightmost separator of the locked columns set
               // because of the quirk described in ClassicTheme#isBrowseLabelBorderDrawQuirk.
               int labelsHeight = getLabelsAreaHeight();
               gd.fillRect(0, 0, d.width - BORDER_THICKNESS, labelsHeight);
               gd.fillRect(0, labelsHeight, d.width, d.height - labelsHeight);
            }
            else
            {
               gd.fillRect(0, 0, d.width, d.height);
            }
            
            ColumnSetContainer.super.draw();
         });
      }

      /**
       * Notification of a mouse clicked event occurred for this widget.
       * <p>
       * In GUI mode, if the widget has a popup-menu attached, it will display it.
       *
       * @param    e
       *           The mouse event.
       */
      @Override
      public void mouseClicked(MouseEvent e)
      {
         super.mouseClicked(e);

         if (e.getButton() == MouseEvent.BUTTON1)
         {
            if (!BrowseGuiImpl.this.hasFocus())
            {
               BrowseGuiImpl.this.requestFocus();
            }

            if (BrowseGuiImpl.this.isEnabled() && e.getID() == MouseEvent.MOUSE_CLICKED)
            {
               NativePoint point = translate(e);
               int rowIndex = getRowByCoordinates(point.y);
               int columnIndex = point.x > (markersColumn != null ?
                                            markersColumn.physicalDimension().width : 0)
                  ? getVisibleColumnCount() - 1 // forces the cell to enter edit mode if possible
                  : -1;                         // will only select the clicked row
               rowClicked(rowIndex, columnIndex, e);
            }
         }
      }

      /**
       * Notification of a mouse wheel moved event occurred for this widget.
       *
       * @param    e
       *           The mouse event.
       */
      @Override
      public void mouseWheelMoved(MouseWheelEvent e)
      {
         BrowseGuiImpl.this.mouseWheelMoved(e);
      }

      /**
       * Destroy this widget.  Notify all of its components that the widget has been destroyed.
       */
      @Override
      public void destroy()
      {
         OutputManager.instance().getRegistry().removeWidget(config.id);
         super.destroy();
      }

      /**
       * Directs the widget to scroll to the specified X and Y positions.
       * It is the widget's responsibility to update its visuals accordingly.
       * Note that the position values are in the same units as the returned
       * dimensions from {@link #getScrollDimension()}, {@link #getVisibleDimension()}
       * and {@link #getScrollStep()}.
       *
       * @param    newXPosition
       *           New X scroll position.
       * @param    newYPosition
       *           New Y scroll position.
       */
      @Override
      public void scroll(Integer newXPosition, Integer newYPosition)
      {
         if (newXPosition != null)
         {
            horizontalScroller.positionUpdated(newXPosition, ScrollStep.CUSTOM, false);
         }

         if (newYPosition != null)
         {
            verticalScroller.positionUpdated(newYPosition, ScrollStep.CUSTOM, false);
         }
      }

      /**
       * The method shows the popup menu for this widget. The menu is resolved with the help of
       * {@link #resolvePopupMenuId()}.
       *
       * @param   e
       *          The related mouse event that caused a popup to be shown.
       * @param   dismissHandler
       *          When not null, the runnable instance is executed when the popup menu is dismissed, regardless
       *          whether an item is chosen or not.
       *
       * @return  Returns the popup menu shown or {@code null} if no popup is resolved for this widget.
       */
      @Override
      protected Menu showPopupMenu(MouseEvent e, Runnable dismissHandler)
      {
         return BrowseGuiImpl.this.showPopupMenu(e, dismissHandler);
      }

      /**
       * Sub-classes will override this to return a list of mouse actions on which the widget
       * responds.  This is compatible with the AWT {@link MouseEvent} codes and P2J specific
       * {@link com.goldencode.p2j.ui.client.event.MouseEvt} actions.
       *
       * @return   See above.
       */
      @Override
      protected int[] mouseActions()
      {
         return new int[]
         {
            MouseEvent.MOUSE_CLICKED,
            MouseEvent.MOUSE_PRESSED
         };
      }
   }

   /**
    * Get offset of the column from the left edge of the viewport or from the right edge of
    * the last locked column if there are locked columns (if the is scroll pane is on its initial
    * position). I.e. first scrollable column always has offset 0 (regardless of presence of
    * locked columns).
    *
    * @param  visibleColumnIndex
    *         0-based index of the column among the visible columns.
    *
    * @return see above.
    */
   private int getColumnOffset(int visibleColumnIndex)
   {
      BrowseColumnGuiImpl column = getVisibleColumn(visibleColumnIndex);
      int offset = column.physicalLocation().x;
      offset -= lockedColumnSet != null ? lockedColumnSet.physicalDimension().width : 0;
      return offset;
   }

   /**
    * Update height of scrollable and locked column sets in order to to match the viewport height.
    */
   private void updateColumnSetsHeight()
   {
      double height = cc.heightFromNative(getViewportDimension().height);
      scrollableColumnSet.setHeight(height);
      if (lockedColumnSet != null)
      {
         lockedColumnSet.setHeight(height);
      }
   }

   /**
    * Update editor after column moving or resizing.
    */
   private void updateEditorAfterColumnModification()
   {
      if (!isEditing())
      {
         return;
      }

      Widget editor = getActiveEditor();
      Container parent = editor.parent();
      parent.moveToTop(editor);

      // update editor position
      ((BaseConfig) editor.config()).x = getCurrentEditorX();

      // update editor width
      if (isEditorFillIn())
      {
         FillIn fillIn = getActiveFillIn();
         BrowseColumnGuiImpl column = getVisibleColumn(currColumn);
         fillIn.config().widthPixels = column.physicalDimension().width;;
         fillIn.config().widthChars = cc.widthFromNative(fillIn.config().widthPixels);
         fillIn.setWidth(fillIn.config().widthChars);
         fillIn.repaint();
      }
      else if (isEditorToggleBox())
      {
      }
      else
      {
         // TODO combo-box
      }

      // update editor vertical position
      notifyBrowseScrolled();
   }

   /**
    * Update maximum for horizontal scrolling in column scrolling mode. See
    * {@link #horizontalColumnScrollMax}.
    */
   private void updateHorizontalColumnScrollMax()
   {
      int colCount = getVisibleColumnCount();

      // how many full columns can fit on the last page?
      int availWidth = getViewportDimension().width;
      if (lockedColumnSet != null)
      {
         availWidth -= lockedColumnSet.physicalDimension().width;
      }

      int lastPageColumns = 0;
      int lastColumnsWidth = 0;
      for (int i = colCount - 1; i >= config.numLockedColumns; i--)
      {
         BrowseColumnGuiImpl column = getVisibleColumn(i);
         lastColumnsWidth += column.physicalDimension().width;

         if (lastColumnsWidth > availWidth)
         {
            break;
         }

         lastPageColumns++;
      }

      lastPageColumns = Math.max(0, --lastPageColumns);

      horizontalColumnScrollMax = Math.max(0, colCount - config.numLockedColumns - lastPageColumns - 1);

      if (horizontalScroller.getPosition() > horizontalColumnScrollMax)
      {
         horizontalScroller.positionUpdated(horizontalColumnScrollMax, ScrollStep.CUSTOM, false);
      }
   }

   /**
    * Get minimum visual width of a column (data width + offsets).
    *
    * @return minimum visual width of a column.
    */
   private int getEffectiveMinColumnWidth()
   {
      return MIN_COL_WIDTH_PIXELS;
   }

   /**
    * Returns <code>true</code> if there is at least one visible column with hyperlinks.
    *
    * @return see above.
    */
   private boolean hasHyperlinkedColumns()
   {
      BrowseColumnConfig[] visibleColumns = getVisibleColumns();
      if (visibleColumns != null)
      {
         for (BrowseColumnConfig config : visibleColumns)
         {
            if (config.hyperlinked)
            {
               return true;
            }
         }
      }

      return false;
   }

   /**
    * Class for handling horizontal scrolling.
    */
   private class HorizontalColumnScroller
   extends ScrollBarControllerGuiImpl
   implements ScrollBarController
   {
      /**
       * Get maximum value for the scroll bar.
       *
       * @return maximum value for the scroll bar.
       */
      @Override
      public int getMax()
      {
         return isColumnScrolling() ? horizontalColumnScrollMax :
                                      totalColumnsWidth - getViewportDimension().width;
      }

      /**
       * Get step for the scroll bar.
       *
       * @return step for the scroll bar.
       */
      @Override
      public int getStep()
      {
         return isColumnScrolling() ? 1 : hScrollingStep;
      }

      /**
       * Get visibility state for the scroll bar.
       *
       * @return <code>true</code> if scroll bar is visible.
       */
      @Override
      public boolean isVisible()
      {
         return isEnabled();
      }

      /**
       * Get enable state for the scroll bar.
       *
       * @return <code>true</code> if scroll bar is enabled.
       */
      @Override
      public boolean isEnabled()
      {
         return isColumnScrolling() ? totalColumnsWidth > getViewportDimension().width:
                                      super.isEnabled();
      }

      /**
       * Get position for the scroll bar.
       *
       * @return position for the scroll bar.
       */
      @Override
      public int getPosition()
      {
         int offset = -scrollableColumnSet.physicalLocation().x;

         if (isColumnScrolling())
         {
            int colNum = getVisibleColumnCount();
            for (int i = config.numLockedColumns; i < colNum; i++)
            {
               //TODO drag calculations for the last column are more tolerant, so normally
               //     you shouldn't drag thumb to the very end to get to the last column
               if (offset <= getColumnOffset(i))
               {
                  return i - config.numLockedColumns;
               }
            }

            return 0;
         }
         else
         {
            return offset;
         }
      }

      /**
       * Get thumb size for the scroll bar.
       *
       * @return thumb size for the scroll bar (in pixels).
       */
      @Override
      public int getThumbSize()
      {
         return H_BAR_THUMB_WIDTH;
      }

      /**
       * Get thumb position for the scroll bar.
       *
       * @return thumb position for the scroll bar (distance between left/top button and thumb's
       * left/top edge).
       */
      @Override
      public int getThumbPosition()
      {
         if (drag)
         {
            return dragPos;
         }

         if (scrollBar == null || totalColumnsWidth == 0 || getMax() == 0)
         {
            return 0;
         }

         int availSize = scrollBar.getTravelSize() - getThumbSize();
         int tPos = Coordinate.toInt(Coordinate.scale(availSize * getPosition() / getMax()));
         tPos = Math.max(0, tPos);
         tPos = Math.min(tPos, availSize);
         return tPos;
      }

      /**
       * Notify controller that position has been updated.
       *
       * @param   position 
       *          New position of the scroll bar.
       * @param   scrollStep 
       *          Scroll step type.
       * @param   drag     
       *          <code>true</code> if position was changed because of thumb dragging.
       */
      @Override
      public void positionUpdated(int position, ScrollStep scrollStep, boolean drag)
      {
         int colCount = getVisibleColumnCount();
         int oldPosition = getPosition();
         if (position == oldPosition || colCount == 0 || config.numLockedColumns >= colCount)
         {
            return;
         }

         // Mouse click between the thumb and a scroll button.
         if (scrollStep.equals(ScrollStep.LONG))
         {
            int viewportWidth = getViewportDimension().width;

            if (isColumnScrolling())
            {
               // Absolute value of the new position can be incorrect, but it is comparable with the
               // old position. Here we must calculate actual position value.
               if (position > oldPosition)
               {
                  // Scroll right - to the last column in view, or at least to the next column.
                  position = Math.max(oldPosition + 1, getIntersectedColumn(viewportWidth));
               }
               else if (position < oldPosition)
               {
                  // Scroll left by (viewport width - locked column width -
                  // width of the first scrollable column in the view), or at least to the previous
                  // column.
                  int oldColumn = oldPosition + config.numLockedColumns;

                  if (oldColumn + 1 == getVisibleColumnCount())
                  {
                     // The last scrollable column is the single scrollable row in the view- scroll
                     // by a single row.
                     position = oldPosition - 1;
                  }
                  else
                  {
                     int secondColumnOffset = getColumnOffset(oldColumn + 1);
                     int leftOffset = secondColumnOffset + lockedColumnsWidth +
                           scrollableColumnSet.physicalLocation().x;
                     if (leftOffset > viewportWidth)
                     {
                        // There is the single scrollable column in the view.
                        position = oldPosition - 1;
                     }
                     else
                     {
                        int rightOffset = viewportWidth - leftOffset;
                        position = Math.min(oldPosition - 1,
                              getIntersectedColumn(-rightOffset + lockedColumnsWidth));
                     }
                  }
               }
            }
            else
            {
               int effectiveViewportWidth = Math.max(0, viewportWidth - lockedColumnsWidth);
               position = position < oldPosition ? position - effectiveViewportWidth :
                                                   position + effectiveViewportWidth;
            }
         }

         position = Math.max(0, position);
         position = Math.min(position, getMax());

         if (isColumnScrolling())
         {
            scrollableColumnSet.setPhysicalLocation(
                  -getColumnOffset(config.numLockedColumns + position), 0);
         }
         else
         {
            scrollableColumnSet.setPhysicalLocation(-position, 0);
         }
         repaint();
      }

      /**
       * Notify controller that thumb drag has ended.
       */
      @Override
      public void stopThumbDrag()
      {
         if (drag)
         {
            super.stopThumbDrag();

            int availSize = scrollBar.getTravelSize() - getThumbSize();

            int position = availSize > 0 ?
                              Coordinate.toInt(Coordinate.scale(dragPos * getMax() / availSize)) :
                              0;

            positionUpdated(position, ScrollStep.CUSTOM, true);
         }
      }

      /**
       * Return value of the COLUMN-SCROLLING attribute.
       *
       * @return see above
       */
      private boolean isColumnScrolling()
      {
         return config.columnScrolling;
      }

      /**
       * Get the column which body intersects given x coordinate.
       *
       * @param  x
       *         X coordinate, offset relatively to left viewport edge (regardless
       *         presence/absence of locked columns).
       *
       * @return 0-based index of the target column or <code>-1</code> if there is no such column.
       */
      private int getIntersectedColumn(int x)
      {
         int lockedWidth = getLockedColumnsWidth();
         int colNo = getVisibleColumnCount();
         for (int i = config.numLockedColumns; i < colNo; i++)
         {
            int offset = getColumnOffset(i) + lockedWidth +
                  scrollableColumnSet.physicalLocation().x;
            if (offset > x)
            {
               return i - config.numLockedColumns - 1;
            }
         }

         return -1;
      }
   }

   /**
    * Class for handling vertical scrolling.
    */
   private class VerticalRowScroller
   extends ScrollBarControllerGuiImpl
   {
      /**
       * Constructor.
       */
      public VerticalRowScroller()
      {
         super();
      }

      /**
       * Notify controller that position has been updated.
       *
       * @param   position
       *          New position of the scroll bar.
       * @param   scrollStep
       *          Scroll step type.       
       * @param   drag
       *          <code>true</code> if position was changed because of thumb dragging.
       *          <code>false</code> if position was changed because a scroll button was pressed.
       */
      @Override
      public void positionUpdated(int position, ScrollStep scrollStep, boolean drag)
      {
         if (getCacheSize() > 0)
         {
            if (!drag &&
                isNewEditingRowOutOfTheView() &&
                position < 0 &&
                editingExistingRowBefore < 0)
            {
               // special case - first row is a new row and is out of the view
               incrementalScroll(-1);
               repaint();
               return;
            }

            int oldRow = getPosition();
            if (position < 0)
            {
               position = 0;
            }

            if (oldRow == position)
            {
               repaint();
               return;
            }

            if (drag || ScrollStep.LONG.equals(scrollStep))
            {
               int oldPosition = getPosition();

               if (position == oldPosition)
               {
                  return;
               }
               else
               {
                  if (ScrollStep.LONG.equals(scrollStep))
                  {
                     if (position < oldPosition)
                     {
                        position = oldPosition - (getDown() - 1);
                     }
                     else
                     {
                        position = oldPosition + (getDown() - 1);
                     }
                  }

                  position += getDown() - 1;
                  scrollToRow(position, false, false, false, true, true, getDown() - 1, false);
               }
            }
            else
            {
               if (oldRow < position)
               {
                  if (getCacheSize() == getDown())
                  {
                     if (ScrollStep.MAX.equals(scrollStep))
                     {
                        EventManager.postEvent(new KeyInput(BrowseGuiImpl.this,
                                                            EventType.KEY_PRESSED,
                                                            Keyboard.KA_END));
                     }
                     else if (ScrollStep.LONG.equals(scrollStep))
                     {
                        if (isEditing() && !leaveRow())
                        {
                           return;
                        }
                        
                        int newRow = getCurrentRow() + getPageSize();
                        scrollToRow(newRow, true, false, false, true, true, getPageSize(), false);
                     }
                     else
                     {
                        incrementalScroll(1);
                     }
                  }
               }
               else
               {
                  if (ScrollStep.MIN.equals(scrollStep))
                  {
                     EventManager.postEvent(new KeyInput(BrowseGuiImpl.this,
                                                         EventType.KEY_PRESSED,
                                                         Keyboard.KA_HOME));
                  }
                  else if (ScrollStep.LONG.equals(scrollStep))
                  {
                     if (isEditing() && !leaveRow())
                     {
                        return;
                     }
                     
                     int newRow = getCurrentRow() - getPageSize();
                     scrollToRow(newRow, true, false, false, true, true, getPageSize(), false);
                  }
                  else
                  {
                     incrementalScroll(-1);
                  }
               }
            }
            repaint();
         }
      }

      /**
       * Get step for the scroll bar.
       *
       * @return step for the scroll bar.
       */
      @Override
      public int getStep()
      {
         return 1;
      }

      /**
       * Get maximum value for the scroll bar.
       *
       * @return maximum value for the scroll bar.
       */
      @Override
      public int getMax()
      {
         return Math.max(0, config.maxDataGuess - getDown());
      }

      /**
       * Get visibility state for the scroll bar.
       *
       * @return <code>true</code> if scroll bar is visible.
       */
      @Override
      public boolean isVisible()
      {
         return config.scrollbarVertical;
      }

      /**
       * Get position for the scroll bar.
       *
       * @return position for the scroll bar.
       */
      @Override
      public int getPosition()
      {
         int row = getFirstExistingRow();
         if (row < 0)
         {
            if (editingExistingRowBefore != null)
            {
               row = editingExistingRowBefore;
            }
            else
            {
               row = 0; // safe option
            }

         }
         return row;
      }

      /**
       * Notify controller that thumb drag has ended.
       */
      @Override
      public void stopThumbDrag()
      {
         if (drag)
         {
            super.stopThumbDrag();

            int scrollSize = getAvailableTravelSize() - getThumbSize();
            int maxVal = getMax();
            int position = Coordinate.toInt(Coordinate.scale(dragPos * maxVal / scrollSize));
            position = Math.max(0, position);
            position = Math.min(position, maxVal);

            positionUpdated(position, ScrollStep.CUSTOM, true);
         }
      }

      /**
       * Get thumb size for the scroll bar.
       *
       * @return thumb size for the scroll bar (in pixels).
       */
      @Override
      public int getThumbSize()
      {
         int thumbMinSize = ThemeManager.getCurrentTheme().getScrollBarThumbMinSize();
         if (config.maxDataGuess == 0)
         {
            return thumbMinSize;
         }

         int availSize = getAvailableTravelSize();

         int tHeight = Coordinate.toInt(Coordinate.scale(
                           (double)availSize * (double)getDown() / (double)config.maxDataGuess));

         tHeight = Math.max(tHeight, thumbMinSize);
         tHeight = Math.min(tHeight, availSize);
         return tHeight;
      }

      /**
       * Get thumb position for the scroll bar.
       *
       * @return thumb position for the scroll bar (distance between left/top button and thumb's
       *         left/top edge).
       */
      @Override
      public int getThumbPosition()
      {
         if (drag)
         {
            return dragPos;
         }

         if (config.maxDataGuess - getDown() <= 0)
         {
            return 0;
         }

         int thumbSize = getThumbSize();
         int availSize = getAvailableTravelSize() - thumbSize;

         int pos = Coordinate.toInt(Coordinate.scale(
               availSize * getPosition() / (config.maxDataGuess - getDown())));
         pos = Math.min(pos, availSize);

         return pos;
      }

      /**
       * Get size of the travel area available for thumb scrolling, excluding the bottom gap which
       * is used to indicate scrolling towards the end of the result set.
       *
       * @return size of travel area, in pixels.
       */
      private int getAvailableTravelSize()
      {
         if (scrollBar == null)
         {
            return 0;
         }

         int travelSize = scrollBar.getTravelSize();

         // real data, no estimation
         if (config.maxDataReached)
         {
            return travelSize;
         }

         // default bottom gap
         int gap = Coordinate.toInt(Coordinate.scale(travelSize * 0.01));

         // have we moved past row 100?
         if (config.maxDataGuess > BrowseConfig.DATA_GUESS_DEFAULT_VALUE)
         {
            int topRow = getPosition();
            int shrinkBy = 0;
            int calcRow = topRow + getDown() - BrowseConfig.DATA_GUESS_DEFAULT_VALUE + 1;
            if (calcRow > 0)
            {
               shrinkBy = Coordinate.toInt(Coordinate.scale((gap / 3) * Math.log10(calcRow)));
            }
            gap = Math.max(0, gap - shrinkBy);
         }

         return travelSize - gap;
      }
   }

   /**
    * Scroll pane for the {@link #scrollableColumnSet}. The purpose of this class is to increase
    * the width of the {@link #scrollableColumnSet} if there should be an background-colored empty
    * area after the last column. This can happen only in column-scrolling mode.
    */
   private class ColumnsScrollPane extends ScrollPaneGuiImpl
   {
      /**
       * Constructor.
       *
       * @param widget Child widget.
       */
      ColumnsScrollPane(ScrollableWidget<GuiOutputManager> widget)
      {
         super(widget);
      }

      /**
       * Adjusts viewport size, scroll bar visibility and positions. Additionally increases
       * the width of the {@link #scrollableColumnSet} if there should be an background-colored
       * empty area with after the last column. This can happen only in column-scrolling mode.
       */
      @Override
      protected void adjustScrollLayoutImpl()
      {
         super.adjustScrollLayoutImpl();

         int columnSetWidth = scrollableColumnSet.physicalDimension().width;
         int columnSetLocation = scrollableColumnSet.physicalLocation().x;
         int viewportWidth = getViewportDimension().width;
         if (columnSetLocation + columnSetWidth < viewportWidth)
         {
            scrollableColumnSet.setWidth(viewportWidth - columnSetLocation);
         }
      }
   }
}
