/*
** Module   : BrowsePopupMenu.java
** Abstract : Popup menu for browse.
**
** Copyright (c) 2018, Golden Code Development Corporation.
**
** -#- -I- --Date-- ---------------------------------Description----------------------------------
** 001 SVL 20180209 Created initial version.
**     CA  20180210 Allow the report to download via the driver.
** 002 SBI 20180410 Changed due to deliverPrintOutput was renamed to deliverDocumentOutput.
** 003 HC  20180511 Added exception logging.
** 004 SVL 20180529 Implemented complete menu.
*/
/*
** This program is free software: you can redistribute it and/or modify
** it under the terms of the GNU Affero General Public License as
** published by the Free Software Foundation, either version 3 of the
** License, or (at your option) any later version.
**
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
** GNU Affero General Public License for more details.
**
** You may find a copy of the GNU Affero GPL version 3 at the following
** location: https://www.gnu.org/licenses/agpl-3.0.en.html
**
** Additional terms under GNU Affero GPL version 3 section 7:
**
**   Under Section 7 of the GNU Affero GPL version 3, the following additional
**   terms apply to the works covered under the License.  These additional terms
**   are non-permissive additional terms allowed under Section 7 of the GNU
**   Affero GPL version 3 and may not be removed by you.
**
**   0. Attribution Requirement.
**
**     You must preserve all legal notices or author attributions in the covered
**     work or Appropriate Legal Notices displayed by works containing the covered
**     work.  You may not remove from the covered work any author or developer
**     credit already included within the covered work.
**
**   1. No License To Use Trademarks.
**
**     This license does not grant any license or rights to use the trademarks
**     Golden Code, FWD, any Golden Code or FWD logo, or any other trademarks
**     of Golden Code Development Corporation. You are not authorized to use the
**     name Golden Code, FWD, or the names of any author or contributor, for
**     publicity purposes without written authorization.
**
**   2. No Misrepresentation of Affiliation.
**
**     You may not represent yourself as Golden Code Development Corporation or FWD.
**
**     You may not represent yourself for publicity purposes as associated with
**     Golden Code Development Corporation, FWD, or any author or contributor to
**     the covered work, without written authorization.
**
**   3. No Misrepresentation of Source or Origin.
**
**     You may not represent the covered work as solely your work.  All modified
**     versions of the covered work must be marked in a reasonable way to make it
**     clear that the modified work is not originating from Golden Code Development
**     Corporation or FWD.  All modified versions must contain the notices of
**     attribution required in this license.
*/

package com.goldencode.p2j.ui.client.gui;

import com.goldencode.p2j.ui.*;
import com.goldencode.p2j.ui.chui.*;
import com.goldencode.p2j.ui.client.*;
import com.goldencode.p2j.ui.client.gui.driver.*;
import com.goldencode.p2j.util.*;
import com.google.common.io.*;
import com.goldencode.p2j.ui.client.gui.theme.*;

import java.awt.event.*;
import java.io.*;
import java.util.logging.*;

import static com.goldencode.p2j.ui.EnhancedBrowseParameter.*;

/**
 * Popup menu for browse
 */
class BrowsePopupMenu
extends MenuGuiImpl
{
   /** Logger. */
   private static Logger log = Logger.getLogger(BrowsePopupMenu.class.getSimpleName());

   /** Minimum size of the font allowed for selection. */
   private static final int MIN_FONT_SIZE = 6;

   /** Maximum size of the font allowed for selection. */
   private static final int MAX_FONT_SIZE = 30;

   /**
    * Constructs a popup menu.
    *
    * @param browse
    *        Parent browse.
    */
   private BrowsePopupMenu(Browse browse)
   {
      super(null, new MenuConfig());
      config().popupOnly = true;
      config().ownerId = browse.getId().asInt();
   }

   /**
    * Create a popup menu for the given browse.
    *
    * @param  browse
    *         Parent browse.
    * @param  column
    *         Column which was right-clicked, or <code>null</code> if the right click was
    *         performed outside of the columns area.
    *
    * @return popup menu for the given browse.
    */
   public static BrowsePopupMenu create(BrowseGuiImpl browse, BrowseColumnGuiImpl column)
   {
      BrowsePopupMenu menu = new BrowsePopupMenu(browse);
      menu.initialize(WidgetId.nextID(), menu.config());

      boolean layoutMode = browse.isLayoutMode();

      // Export submenu
      if (!layoutMode)
      {
         SubMenuGuiImpl submenu = SubMenuGuiImpl.createDynamicSubmenu("Export");
         menu.add(submenu);

         submenu.add(createExportMenuItem(browse, MediaType.PDF));
         submenu.add(createExportMenuItem(browse, MediaType.CSV));
         submenu.add(createExportMenuItem(browse, MediaType.XLS));
         submenu.add(createExportMenuItem(browse, MediaType.XLSX));
      }

      if (layoutMode)
      {
         // Hide column
         if (column != null)
         {
            StringBuilder itemLabel = new StringBuilder("Hide Column");
            String[] labelLines = column.getLabelLines();
            if (labelLines.length > 0)
            {
               itemLabel.append(" ");
               for (int i = 0; i < labelLines.length; i++)
               {
                  itemLabel.append(labelLines[i]);
                  if (i != 0)
                  {
                     itemLabel.append(" ");
                  }
               }
            }

            menu.add(MenuItemGuiImpl.createDynamicItem(itemLabel.toString(), new MouseAdapter()
            {
               @Override
               public void mouseClicked(MouseEvent mouseEvent)
               {
                  browse.changeColumnVisibility(column.config().ordinal, false);
               }
            }));
         }

         // Show column
         BrowseColumnConfig[] invisibleColumns = browse.getInvisibleColumns();
         if (invisibleColumns.length > 0)
         {
            SubMenuGuiImpl submenu = SubMenuGuiImpl.createDynamicSubmenu("Show Column");
            menu.add(submenu);

            for (BrowseColumnConfig columnConfig : invisibleColumns)
            {
               String label = columnConfig.label;
               if (label == null || label.trim().isEmpty())
               {
                  label = "(no column label)";
               }

               submenu.add(MenuItemGuiImpl.createDynamicItem(label,
                     new MouseAdapter()
                     {
                        @Override
                        public void mouseClicked(MouseEvent mouseEvent)
                        {
                           browse.changeColumnVisibility(columnConfig.ordinal, true);
                        }
                     }));
            }
         }

         // Browse colors
         SubMenuGuiImpl submenu = SubMenuGuiImpl.createDynamicSubmenu("Change Browse Colors");
         menu.add(submenu);

         submenu.add(MenuItemGuiImpl.createDynamicItem("Text Color",
                                       getBrowseColorChangeAdapter(browse, FG_COLOR)));

         submenu.add(MenuItemGuiImpl.createDynamicItem("Cells Background Color",
                                       getBrowseColorChangeAdapter(browse, BG_COLOR)));

         submenu.add(MenuItemGuiImpl.createDynamicItem("Label Text Color",
                                       getBrowseColorChangeAdapter(browse, LABEL_FG_COLOR)));

         submenu.add(MenuItemGuiImpl.createDynamicItem("Label Background Color",
                                       getBrowseColorChangeAdapter(browse, LABEL_BG_COLOR)));

         submenu.add(MenuItemGuiImpl.createDynamicItem("Separators Color",
                                       getBrowseColorChangeAdapter(browse, SEPARATORS_COLOR)));

         // Browse fonts
         submenu = SubMenuGuiImpl.createDynamicSubmenu("Change Browse Fonts");
         menu.add(submenu);

         submenu.add(MenuItemGuiImpl.createDynamicItem("Cells Font",
                                       getBrowseFontChangeAdapter(browse, CELLS_FONT)));
         submenu.add(MenuItemGuiImpl.createDynamicItem("Label Font",
                                       getBrowseFontChangeAdapter(browse, LABEL_FONT)));

         // Column colors
         if (column != null)
         {
            submenu = SubMenuGuiImpl.createDynamicSubmenu("Change Column Colors");
            menu.add(submenu);

            submenu.add(MenuItemGuiImpl.createDynamicItem("Text Color",
                                       getColumnColorChangeAdapter(column, FG_COLOR)));

            submenu.add(MenuItemGuiImpl.createDynamicItem("Cells Background Color",
                                       getColumnColorChangeAdapter(column, BG_COLOR)));

            submenu.add(MenuItemGuiImpl.createDynamicItem("Label Text Color",
                                       getColumnColorChangeAdapter(column, LABEL_FG_COLOR)));

            submenu.add(MenuItemGuiImpl.createDynamicItem("Label Background Color",
                                       getColumnColorChangeAdapter(column, LABEL_BG_COLOR)));

            // Column fonts
            submenu = SubMenuGuiImpl.createDynamicSubmenu("Change Column Fonts");
            menu.add(submenu);

            submenu.add(MenuItemGuiImpl.createDynamicItem("Cells Font",
                                       getColumnFontChangeAdapter(column, CELLS_FONT)));
            submenu.add(MenuItemGuiImpl.createDynamicItem("Label Font",
                                       getColumnFontChangeAdapter(column, LABEL_FONT)));
         }

         // Separators on/off
         menu.add(MenuItemGuiImpl.createDynamicItem(
               "Separators " + (browse.config().separators ? "Off" : "On"),
               new MouseAdapter()
               {
                  @Override
                  public void mouseClicked(MouseEvent mouseEvent)
                  {
                     browse.config().separators = !browse.config().separators;
                     browse.repaint();
                  }
               }));
      }

      // Change layout
      if (!layoutMode)
      {
         menu.add(MenuItemGuiImpl.createDynamicItem("Change Layout",
            new MouseAdapter()
            {
               @Override
               public void mouseClicked(MouseEvent mouseEvent)
               {
                  browse.startLayoutMode();
               }
            }));
      }

      // Commit layout changes
      if (layoutMode)
      {
         menu.add(MenuItemGuiImpl.createDynamicItem("Revert to default 4GL state",
               new MouseAdapter()
               {
                  @Override
                  public void mouseClicked(MouseEvent mouseEvent)
                  {
                     browse.applyDefault4glConfig();
                  }
               }));

         menu.add(MenuItemGuiImpl.createDynamicItem("Undo Changes",
               new MouseAdapter()
               {
                  @Override
                  public void mouseClicked(MouseEvent mouseEvent)
                  {
                     browse.stopLayoutMode(true, null);
                  }
               }));

         EnhancedBrowseSaveTarget[] options = browse.getEnhancedBrowseSaveTargets();
         if (options.length > 0)
         {
            SubMenuGuiImpl submenu = SubMenuGuiImpl.createDynamicSubmenu("Save Changes");
            menu.add(submenu);

            for (EnhancedBrowseSaveTarget option : options)
            {
               submenu.add(MenuItemGuiImpl.createDynamicItem(getSaveTargetDescription(option),
                                                        getSaveTargetAdapter(browse, option)));
            }
         }
      }

      return menu;
   }

   /**
    * Create menu item that calls export to the specified format.
    *
    * @param  browse
    *         Parent browse.
    * @param  outputFormat
    *         Report output format.
    *
    * @return menu item that calls export to the specified format.
    */
   private static MenuItemGuiImpl createExportMenuItem(Browse browse,
                                                       MediaType outputFormat)
   {
      String format = outputFormat.name();
      return MenuItemGuiImpl.createDynamicItem("Export to " + format, new MouseAdapter()
         {
            @Override
            public void mouseClicked(MouseEvent e)
            {
               ThinClient tc = ThinClient.getInstance();
               String fileName = "output." + format.toLowerCase();
               boolean res = tc.runReport(browse.getId().asInt(), fileName, outputFormat);
               if (res)
               {
                  GuiDriver gd = (GuiDriver) tc.getOutputManager().getInstanceDriver();
                  gd.deliverDocumentOutput(stream ->
                  {
                     try
                     {
                        BufferedOutputStream bos =
                              new BufferedOutputStream((OutputStream) stream);
                        BufferedInputStream fis =
                              new BufferedInputStream(new FileInputStream(fileName));
                           
                        ByteStreams.copy(fis, bos);

                        fis.close();
                        bos.flush();
                     }
                     catch (IOException ex)
                     {
                        log.log(Level.WARNING, format + " export failed", ex);
                     }
                  }, null, outputFormat);
               }
                  
               tc.messageBox(new Object[] { "Report output to \"" + fileName + "\" done!" },
                             null,
                             res ? LogicalTerminal.ALERT_INFO : LogicalTerminal.ALERT_ERROR,
                             LogicalTerminal.BTN_OK,
                             "Report output",
                             null,
                             WindowConfig.RESOLVE_WINDOW,
                             null);
            }
         });
   }

   /**
    * Create a mouse adapter that calls color change dialog for the browse column.
    *
    * @param   column
    *          Target browse column.
    * @param   param
    *          Parameter which specifies which type of color should be changed: can be
    *          {@link EnhancedBrowseParameter#FG_COLOR}, {@link EnhancedBrowseParameter#BG_COLOR},
    *          {@link EnhancedBrowseParameter#LABEL_FG_COLOR} or
    *          {@link EnhancedBrowseParameter#LABEL_BG_COLOR}.
    *
    * @return  mouse adapter that calls color change dialog for the browse column.
    */
   private static MouseAdapter getColumnColorChangeAdapter(BrowseColumnGuiImpl column,
                                                           EnhancedBrowseParameter param)
   {
      return new MouseAdapter()
      {
         @Override
         public void mouseClicked(MouseEvent mouseEvent)
         {
            Theme theme = ThemeManager.getCurrentTheme();
            ColorRgb color;

            switch (param)
            {
               case FG_COLOR:
                  color = theme.getBrowseCellForegroundColor(column, false, false, -1);
                  break;
               case BG_COLOR:
                  color = theme.getColumnBackgroundColor(column);
                  break;
               case LABEL_FG_COLOR:
                  color = theme.getBrowseCellForegroundColor(column, true, false, -1);
                  break;
               case LABEL_BG_COLOR:
                  color = theme.getColumnLabelBackgroundColor(column);
                  break;
               default:
                  throw new IllegalArgumentException("Not a color parameter: " + param);
            }

            Browse browse = column.getBrowse();
            ThinClient tc = ThinClient.getInstance();

            int res = tc.chooseColor(color.getGuiRgb(), browse.window().getId().asInt());
            if (res >= 0 && res != color.getGuiRgb())
            {
               switch (param)
               {
                  case FG_COLOR:
                     column.config().ehFgColor = new ColorRgb(res);
                     break;
                  case BG_COLOR:
                     column.config().ehBgColor = new ColorRgb(res);
                     break;
                  case LABEL_FG_COLOR:
                     column.config().ehLabelFgColor = new ColorRgb(res);
                     break;
                  case LABEL_BG_COLOR:
                     column.config().ehLabelBgColor = new ColorRgb(res);
                     break;
               }
               browse.repaint();
            }
         }
      };
   }

   /**
    * Create a mouse adapter that calls color change dialog for the browse.
    *
    * @param   browse
    *          Target browse.
    * @param   param
    *          Parameter which specifies which type of color should be changed: can be
    *          {@link EnhancedBrowseParameter#FG_COLOR}, {@link EnhancedBrowseParameter#BG_COLOR},
    *          {@link EnhancedBrowseParameter#LABEL_FG_COLOR},
    *          {@link EnhancedBrowseParameter#LABEL_BG_COLOR} or
    *          {@link EnhancedBrowseParameter#SEPARATORS_COLOR}.
    *
    * @return  mouse adapter that calls color change dialog for the browse.
    */
   private static MouseAdapter getBrowseColorChangeAdapter(BrowseGuiImpl browse,
                                                           EnhancedBrowseParameter param)
   {
      return new MouseAdapter()
      {
         @Override
         public void mouseClicked(MouseEvent mouseEvent)
         {
            ThinClient tc = ThinClient.getInstance();
            Theme theme = ThemeManager.getCurrentTheme();
            ColorRgb color;

            switch (param)
            {
               case FG_COLOR:
                  color = browse.resolveColor(-1,
                                              null,
                                              browse.config().columnFgColor,
                                              browse.config().ehFgColor,
                                              true,
                                              false);
                  break;
               case BG_COLOR:
                  color = browse.resolveColor(-1,
                                              null,
                                              browse.config().columnBgColor,
                                              browse.config().ehBgColor,
                                              false,
                                              false);
                  break;
               case LABEL_FG_COLOR:
                  color = browse.resolveColor(-1,
                                              null,
                                              browse.config().labelFgColor,
                                              browse.config().ehLabelFgColor,
                                              true,
                                              true);
                  break;
               case LABEL_BG_COLOR:
                  color = browse.resolveColor(-1,
                                              null,
                                              browse.config().labelBgColor,
                                              browse.config().ehLabelBgColor,
                                              false,
                                              true);
                  break;
               case SEPARATORS_COLOR:
                  color = theme.getBrowseCellSeparatorColor(browse);
                  break;
               default:
                  throw new IllegalArgumentException("Not a color parameter: " + param);
            }

            int res = tc.chooseColor(color.getGuiRgb(), browse.window().getId().asInt());
            if (res >= 0 && res != color.getGuiRgb())
            {
               switch (param)
               {
                  case FG_COLOR:
                     browse.config().ehFgColor = new ColorRgb(res);
                     break;
                  case BG_COLOR:
                     browse.config().ehBgColor = new ColorRgb(res);
                     break;
                  case LABEL_FG_COLOR:
                     browse.config().ehLabelFgColor = new ColorRgb(res);
                     break;
                  case LABEL_BG_COLOR:
                     browse.config().ehLabelBgColor = new ColorRgb(res);
                     break;
                  case SEPARATORS_COLOR:
                     browse.config().ehSeparatorsColor = new ColorRgb(res);
                     break;
               }
               browse.repaint();
            }
         }
      };
   }

   /**
    * Create a mouse adapter that calls font change dialog for the browse column.
    *
    * @param   column
    *          Target browse column.
    * @param   param
    *          Parameter which specifies which type of font should be changed: can be
    *          {@link EnhancedBrowseParameter#CELLS_FONT} or
    *          {@link EnhancedBrowseParameter#LABEL_FONT}.
    *
    * @return  mouse adapter that calls font color change dialog for the browse column.
    */
   private static MouseAdapter getColumnFontChangeAdapter(BrowseColumnGuiImpl column,
                                                          EnhancedBrowseParameter param)
   {
      return new MouseAdapter()
      {
         @Override
         public void mouseClicked(MouseEvent mouseEvent)
         {
            ThinClient tc = ThinClient.getInstance();
            FontDetails<?> currentDetails;

            switch (param)
            {
               case CELLS_FONT:
                  currentDetails = column.getCellsFont();
                  break;
               case LABEL_FONT:
                  currentDetails = column.getLabelFont();
                  break;
               default:
                  throw new IllegalArgumentException("Not a font parameter: " + param);
            }

            FontDetails<?> newDetails = tc.chooseFont(currentDetails,
                                                      null,
                                                      column.window().getId().asInt(),
                                                      false,
                                                      false,
                                                      MIN_FONT_SIZE,
                                                      MAX_FONT_SIZE);

            if (newDetails != null && !newDetails.equals(currentDetails))
            {
               switch (param)
               {
                  case CELLS_FONT:
                     column.config().ehFont = newDetails;
                     break;
                  case LABEL_FONT:
                     column.config().ehLabelFont = newDetails;
                     break;
               }

               column.resolveFonts();
               column.getBrowse().repaint();
            }
         }
      };
   }

   /**
    * Create a mouse adapter that calls font change dialog for the browse.
    *
    * @param   browse
    *          Target browse.
    * @param   param
    *          Parameter which specifies which type of font should be changed: can be
    *          {@link EnhancedBrowseParameter#CELLS_FONT} or
    *          {@link EnhancedBrowseParameter#LABEL_FONT}.
    *
    * @return  mouse adapter that calls font color change dialog for the browse .
    */
   private static MouseAdapter getBrowseFontChangeAdapter(BrowseGuiImpl browse,
                                                          EnhancedBrowseParameter param)
   {
      return new MouseAdapter()
      {
         @Override
         public void mouseClicked(MouseEvent mouseEvent)
         {
            ThinClient tc = ThinClient.getInstance();
            FontDetails<?> currentDetails;

            switch (param)
            {
               case CELLS_FONT:
                  currentDetails = browse.getBrowseFont();
                  break;
               case LABEL_FONT:
                  currentDetails = browse.getBrowseLabelFont();
                  break;
               default:
                  throw new IllegalArgumentException("Not a font parameter: " + param);
            }

            FontDetails<?> newDetails = tc.chooseFont(currentDetails,
                                                      null,
                                                      browse.window().getId().asInt(),
                                                      false,
                                                      false,
                                                      MIN_FONT_SIZE,
                                                      MAX_FONT_SIZE);

            if (newDetails != null && !newDetails.equals(currentDetails))
            {
               browse.updateEnhancedFont(param, newDetails);
            }
         }
      };
   }

   /**
    * Get mouse adapter which saves the enhanced browse configuration into the specified save
    * target.
    *
    * @param  browse
    *         Target browse.
    * @param  saveTarget
    *         Save target (target server, target account, applied to all browses or not).
    *
    * @return mouse adapter which saves the enhanced browse configuration into the specified save
    *         target.
    */
   private static MouseAdapter getSaveTargetAdapter(BrowseGuiImpl browse,
                                                    EnhancedBrowseSaveTarget saveTarget)
   {
      return new MouseAdapter()
      {
         @Override
         public void mouseClicked(MouseEvent mouseEvent)
         {
            browse.stopLayoutMode(false, saveTarget);
         }
      };
   }

   /**
    * Get description of the specified save target.
    *
    * @param  saveTarget
    *         Save target (target server, target account, applied to all browses or not).
    * @return description of the specified save target.
    */
   private static String getSaveTargetDescription(EnhancedBrowseSaveTarget saveTarget)
   {
      return String.format("SERVER: %s,  USER/GROUP: %s,  FOR ALL BROWSES: %s",
                           saveTarget.serverId,
                           saveTarget.accountId,
                           saveTarget.allBrowses ? "yes" : "no");
   }
}
