/*
** Module   : DragDropHelper.java
** Abstract : drag and drop helper methods common for all widgets in session
**
** Copyright (c) 2017, Golden Code Development Corporation.
**
** -#- -I- --Date-- ---------------------------------Description----------------------------------
** 001 EVL 20170609 Created initial version.
** 002 EVL 20170823 Adding implementation logic.
**     SBI 20170903 Used file channels for uploading large files.
**     EVL 20170905 Moving web client specific code to web gui driver.
*/ 
/*
** This program is free software: you can redistribute it and/or modify
** it under the terms of the GNU Affero General Public License as
** published by the Free Software Foundation, either version 3 of the
** License, or (at your option) any later version.
**
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
** GNU Affero General Public License for more details.
**
** You may find a copy of the GNU Affero GPL version 3 at the following
** location: https://www.gnu.org/licenses/agpl-3.0.en.html
** 
** Additional terms under GNU Affero GPL version 3 section 7:
** 
**   Under Section 7 of the GNU Affero GPL version 3, the following additional
**   terms apply to the works covered under the License.  These additional terms
**   are non-permissive additional terms allowed under Section 7 of the GNU
**   Affero GPL version 3 and may not be removed by you.
** 
**   0. Attribution Requirement.
** 
**     You must preserve all legal notices or author attributions in the covered
**     work or Appropriate Legal Notices displayed by works containing the covered
**     work.  You may not remove from the covered work any author or developer
**     credit already included within the covered work.
** 
**   1. No License To Use Trademarks.
** 
**     This license does not grant any license or rights to use the trademarks
**     Golden Code, FWD, any Golden Code or FWD logo, or any other trademarks
**     of Golden Code Development Corporation. You are not authorized to use the
**     name Golden Code, FWD, or the names of any author or contributor, for
**     publicity purposes without written authorization.
** 
**   2. No Misrepresentation of Affiliation.
** 
**     You may not represent yourself as Golden Code Development Corporation or FWD.
** 
**     You may not represent yourself for publicity purposes as associated with
**     Golden Code Development Corporation, FWD, or any author or contributor to
**     the covered work, without written authorization.
** 
**   3. No Misrepresentation of Source or Origin.
** 
**     You may not represent the covered work as solely your work.  All modified
**     versions of the covered work must be marked in a reasonable way to make it
**     clear that the modified work is not originating from Golden Code Development
**     Corporation or FWD.  All modified versions must contain the notices of
**     attribution required in this license.
*/

package com.goldencode.p2j.ui.client.gui;

import java.io.*;
import java.nio.channels.*;
import java.nio.file.*;
import java.util.*;

import com.goldencode.p2j.ui.*;
import com.goldencode.p2j.ui.chui.ThinClient;
import com.goldencode.p2j.ui.client.event.*;
import com.goldencode.p2j.ui.client.widget.*;

/**
 * Helper methods to handle drag and drop client side processing.
 */
public class DragDropHelper
{
   /** The default file options set */
   private static final Set<? extends OpenOption> OPEN_OPTIONS = EnumSet.of(
                                                            StandardOpenOption.CREATE,
                                                            StandardOpenOption.WRITE,
                                                            StandardOpenOption.TRUNCATE_EXISTING);
   
   /** Storage to keep dropped file names with current operation.  */
   private static List<String> fileNames = new ArrayList<String>();
   
   /** Flag indicating we are now currently in a middle of drop operation. */
   private static boolean inProgress = false;
   
   /** File ready counter. */
   private static int readyCounter = 0;

   /** The widget is pending for drop notify event. */
   private static Widget<?> widDropTargetPending = null;
  
   /**
    * Terminates file dropping operation and clean up resources.
    * 
    * @return  <code>TRUE</code> in success, <code>FALSE</code> otherwise.
    */
   public static boolean endFileDrop()
   {
      // clean up is optional when it was not already done
      if (inProgress)
      {
         // and file list storage
         fileNames.clear();
         // clean up pending widget cache
         widDropTargetPending = null;
         // current operation completed
         inProgress = false;
      }
      return true;
   }

   /**
    * Gets the number of the dropped files.
    * 
    * @return  The integer number for dropped file list size.
    */
   public static int getNumDroppedFiles()
   {
      return fileNames.size();
   }

   /**
    * Gets the name of the dropped file from internal 1-based list list.
    * 
    * @param   index
    *          1-based number of the file name to get.
    *          
    * @return  The name of the file with given index.
    */
   public static String getDroppedFile(int index)
   {
      return index > 0 && index <= fileNames.size() ? fileNames.get(index-1) : "?";
   }

   /**
    * Initiates drop operation for given widget.  The 4GL only provides file list support as data
    * transfer.
    * 
    * @param    widDropTarget
    *           The widget to post notification about drop event.
    * @param    fileList
    *           The file list the widget must handle on drop operation.
    * @param    useTmpDir
    *           The flag indicating we need to use files uploaded into temporary directory.
    */
   public static void startDropTarget(Widget<?> widDropTarget,
                                      List<File> fileList,
                                      boolean useTmpDir)
   {
      // mark itself we are now busy
      inProgress = true;
      // clean up pending drop target cache
      if (useTmpDir)
      {
         widDropTargetPending =  null;
      }
      // checking the incoming list
      for (File f : fileList)
      {
         addFileName(f.getPath());
         if (useTmpDir)
         {
            readyCounter++;
         }
      }

      // in case of web client uploading we post the message later when all files are ready
      if (useTmpDir && readyCounter != 0)
      {
         widDropTargetPending = widDropTarget;
      }
      else
      {
         // post the message to inform target widget the file list is ready
         ThinClient.getInstance().postOSEvent(new DirectManipulationEvent(widDropTarget,
                                                                   Keyboard.SE_DROP_FILE_NOTIFY));
      }
   }
   
   /**
    * Stores the given byte array date as file in Java temporary directory with the given name.
    * 
    * @param    name
    *           The file name to create.
    * @param    data
    *           The data to store in the file.
    * @param    size
    *           The size of the to create, used to check the input array.
    */
   public static void saveFile(String name, byte [] data, int size)
   {
      try
      {
         // getting the new file
         Path fileToWrite = Paths.get(name);
         // request delete on JVM exits
         fileToWrite.toFile().deleteOnExit();
         // write binary data
         Files.write(fileToWrite, data);
      }
      catch (IOException ioe)
      {
         ioe.printStackTrace();
      }
      finally
      {
         // yet another file is ready
         readyCounter--;
         // all files uploaded, it's time to post end drop target message
         if (readyCounter == 0 && widDropTargetPending != null)
         {
            // post the message to inform target widget the file list is ready
            ThinClient.getInstance().postOSEvent(
                  new DirectManipulationEvent(widDropTargetPending, Keyboard.SE_DROP_FILE_NOTIFY));
         }
      }
   }
   
   /**
    * Reads the data from the given file channel starting from its current position and stores
    * the retrieved data into the file in Java temporary directory with the given name.
    * 
    * @param    name
    *           The file name to create
    * @param    channel
    *           The file channel
    * @param    length
    *           The length of the content
    */
   public static void readDataAndSaveAsFile(String name, FileChannel channel, int length)
   {
      try
      {
         // getting the new file
         Path fileToWrite = Paths.get(name);
         // request delete on JVM exits
         fileToWrite.toFile().deleteOnExit();
         
         // write binary data from the opened file channel into new named file
         try(WritableByteChannel target = Files.newByteChannel(fileToWrite, OPEN_OPTIONS))
         {
            channel.transferTo(channel.position(), length, target);
         }
      }
      catch (IOException ioe)
      {
         ioe.printStackTrace();
      }
      finally
      {
         // yet another file is ready
         readyCounter--;
         // all files uploaded, it's time to post end drop target message
         if (readyCounter == 0 && widDropTargetPending != null)
         {
            // post the message to inform target widget the file list is ready
            ThinClient.getInstance().postOSEvent(
                 new DirectManipulationEvent(widDropTargetPending, Keyboard.SE_DROP_FILE_NOTIFY));
         }
      }
   }
   
   /**
    * Adds next file name to the list of the dropped files.
    * 
    * @param   fileName
    *          The filename to add.
    */
   private static void addFileName(String fileName)
   {
      if (fileName != null && !fileName.isEmpty())
      {
         fileNames.add(fileName);
      }
   }
}