/*
** Module   : FontChooserGuiImpl.java
** Abstract : The dialog that allows to select a font and its attributes in GUI.
**
** Copyright (c) 2017-2018, Golden Code Development Corporation.
**
** -#- -I- --Date-- ---------------------------------Description----------------------------------
** 001 OM  20171016 Created initial version.
** 002 SVL 20180529 Use -1 instead of 0 as default font metrics to force font metrics calculation
**                  and avoid bug when no font is rendered.
** 003 OM  20180727 Added handle for system close button.
*/

/*
** This program is free software: you can redistribute it and/or modify
** it under the terms of the GNU Affero General Public License as
** published by the Free Software Foundation, either version 3 of the
** License, or (at your option) any later version.
**
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
** GNU Affero General Public License for more details.
**
** You may find a copy of the GNU Affero GPL version 3 at the following
** location: https://www.gnu.org/licenses/agpl-3.0.en.html
** 
** Additional terms under GNU Affero GPL version 3 section 7:
** 
**   Under Section 7 of the GNU Affero GPL version 3, the following additional
**   terms apply to the works covered under the License.  These additional terms
**   are non-permissive additional terms allowed under Section 7 of the GNU
**   Affero GPL version 3 and may not be removed by you.
** 
**   0. Attribution Requirement.
** 
**     You must preserve all legal notices or author attributions in the covered
**     work or Appropriate Legal Notices displayed by works containing the covered
**     work.  You may not remove from the covered work any author or developer
**     credit already included within the covered work.
** 
**   1. No License To Use Trademarks.
** 
**     This license does not grant any license or rights to use the trademarks
**     Golden Code, FWD, any Golden Code or FWD logo, or any other trademarks
**     of Golden Code Development Corporation. You are not authorized to use the
**     name Golden Code, FWD, or the names of any author or contributor, for
**     publicity purposes without written authorization.
** 
**   2. No Misrepresentation of Affiliation.
** 
**     You may not represent yourself as Golden Code Development Corporation or FWD.
** 
**     You may not represent yourself for publicity purposes as associated with
**     Golden Code Development Corporation, FWD, or any author or contributor to
**     the covered work, without written authorization.
** 
**   3. No Misrepresentation of Source or Origin.
** 
**     You may not represent the covered work as solely your work.  All modified
**     versions of the covered work must be marked in a reasonable way to make it
**     clear that the modified work is not originating from Golden Code Development
**     Corporation or FWD.  All modified versions must contain the notices of
**     attribution required in this license.
*/

package com.goldencode.p2j.ui.client.gui;

import com.goldencode.p2j.ui.*;
import com.goldencode.p2j.ui.chui.*;
import com.goldencode.p2j.ui.client.*;
import com.goldencode.p2j.ui.client.Dimension;
import com.goldencode.p2j.ui.client.event.ActionEvent;
import com.goldencode.p2j.ui.client.event.listener.ActionListener;
import com.goldencode.p2j.ui.client.gui.driver.*;
import com.goldencode.p2j.ui.client.gui.filechooser.*;
import com.goldencode.p2j.ui.client.layout.LayoutManager;
import com.goldencode.p2j.ui.client.model.*;
import com.goldencode.p2j.ui.client.widget.*;
import com.goldencode.p2j.ui.client.widget.Container;
import com.goldencode.p2j.util.*;

/* Note: This class does not use the AWT/Graphics2D to render directly on the screen. Instead it
 *       uses the AWT/Graphics2D framework to render offscreen a preview of the selected font
 *       which will be wrapped in a FWD specific Image container and sent to FWD driver to be
 *       displayed on client endpoint.*/
import java.awt.*;
import java.awt.Color;
import java.awt.event.*;
import java.awt.image.*;
import java.util.*;

/**
 * This is a frame viewed as dialog that allows to select a font and its properties. It mimics the
 * similar dialog from Windows.
 * <p>
 * NOTE: Known issues:
 *    1. at this moment, the list of fonts is obtained from the AWT graphic environment. This is
 *       fine for Swing FWD driver, but it does not make sense for web driver. Alternatively, we
 *       can use a predefined list of fonts like the DEFAULT_FONTS (or maybe request them from
 *       server);
 *    2. I could not find a way to get properties of a specified font. Properties like ANSI
 *       compatibility, FIXED-SIZE and SCRIPT are not fully functional. The exclusions for ANSI
 *       and FIXED fonts are implemented by hardcoded (constant) lists;
 *    3. Strikeout property has no support in {@link FontStyle}; FWD cannot handle it.
 */
public class FontChooserGuiImpl
extends FrameGuiImpl
implements Container<GuiOutputManager>,
           IsModal,
           ActionListener
{
   /** The native width of the dialog. */
   private final int NATIVE_WIDTH = 431;
   
   /** The native height of the dialog. */
   private final int NATIVE_HEIGHT = 319;
   
   /** Flag that turns to true only if the user closed the dialog by accepting it. */
   private boolean accepted = false;
   
   /** Flag which tracks user input. */
   private boolean inputDone = false;
   
   /** Cached driver. */
   private GuiDriver gd = null;
   
   /** Cached coordinates. */
   private CoordinatesConversion cc = null;
   
   /** A label for the combobox allowing to chose the font face. */
   private SimpleLabel fontLabel = null;
   
   /** A label for the combobox allowing to chose the font style. */
   private SimpleLabel styleLabel = null;
   
   /** A label for the combobox allowing to chose the font size. */
   private SimpleLabel sizeLabel = null;
   
   /** A label for the combobox allowing to chose the font scripting. */
   private SimpleLabel scriptLabel = null;
   
   /** The label of the EFFECTS groupbox. */
   private SimpleLabel effectsLabel = null;
   
   /** The label of the SAMPLE groupbox. */
   private SimpleLabel sampleLabel = null;
   
   /** The effects groupbox. */
   private RectangleGuiImpl effectsGroup = null;
   
   /** The sample groupbox. */
   private RectangleGuiImpl sampleGroup = null;
   
   /** The combobox allowing to chose the font face. */
   private ComboBoxGuiImpl fontFaceCbx = null;
   
   /** The combobox allowing to chose the font style. */
   private ComboBoxGuiImpl fontStyleCbx = null;
   
   /** The combobox allowing to chose the font size. */
   private ComboBoxGuiImpl fontSizeCbx = null;
   
   /** The combobox allowing to chose the font's scripting. */
   private ComboBoxGuiImpl scriptCbx = null;
   
   /** Configures whether the font instance is stroke at the middle (strikeout). */
   private ToggleBoxGuiImpl strikeChk = null;
   
   /** Configures whether the font instance is underlined. */
   private ToggleBoxGuiImpl underlineChk = null;
   
   /** The button for accepting the current selection. */
   private ButtonGuiImpl accept = null;
   
   /** The button for rejecting hte dialog. */
   private ButtonGuiImpl reject = null;
   
   /** A quick preview of the selected font. */
   private Preview preview = null;
   
   /** Locally saves the minimum size of the font for validation. */
   private int minSize = 0;
   
   /** Locally saves the maximum size of the font for validation. */
   private int maxSize = 100;
   
   /** The list of default font names. */
   private static final String[] DEFAULT_FONTS = 
         { "Courier New", "Fixedsys", "Microsoft Sans Serif", "MS Sans Serif", "Segue UI", 
           "Tahoma" };
   
   /** 
    * A list of known fixed-size fonts. This is used as a workaround for missing of FIXED-SIZE
    * property of the font in Java.
    */
   private static final String[] FIXED_FONTS = { "Courier New", "Fixedsys" };
   
   /**
    * A list of known ANSI fonts. This is used as a workaround for missing of ANSI property of the
    * font in Java.
    */
   private static final String[] ANSI_FONTS = 
         { "Courier New", "Fixedsys", "Microsoft Sans Serif", "MS Sans Serif", "Segue UI",
           "Tahoma" };
   
   /**
    * A list of known fonts to have both FIXED-SIZE and ANSI compatible property. This is used as
    * a workaround for missing of these properties of the font in Java.
    */
   private static final String[] ANSI_FIXED_FONTS = { "Courier New", "Fixedsys" };
   
   /** The list of standard sizes for the fonts. */
   private static final int[] DEFAULT_SIZES = 
         {8, 9, 10, 11, 12, 14, 16, 18, 20, 24, 26, 28, 36, 48, 72};
   
   /**
    * The names of script support. Not all fonts support all these scripts. For the moment we are
    * not able to properly support this in Java.
    */
   private static final String[] DEFAULT_SCRIPTS = 
         { "Western", "Hebrew", "Arabic", "Greek", "Turkish", "Baltic", "Central European", 
           "Cyrillic", "Vietnamese" };
   
   /** The styles of the fonts: all combinations of bold and italic attributes. */
   private static final String[] STYLES = { "Regular", "Italic", "Bold", "Bold Italic" };
   
   /**
    * Factory method. Creates and configures a {@code FontChooserGuiImpl} dialog. 
    *
    * @param   windowId
    *          The parent window of the newly created dialog.
    * @param   ansiOnly
    *          The dialog will present to user to select only fonts that are ANSI compatible (that
    *          do not include graphic symbols).
    * @param   fixedOnly
    *          The dialog will present to user to select only fixed-size fonts (also known as
    *          monospace in which all characters take the same width when displayed).
    * @param   minSize
    *          Constraint on font-size. User is not allowed to select (or key-in) a font-size that
    *          is less than this value. In case he does, a warning dialog will remind him of the
    *          constraint.
    * @param   maxSize
    *          Constraint on font-size. User is not allowed to select (or key-in) a font-size that
    *          is greater than this value. In case he does, a warning dialog will remind him of
    *          the constraint.
    *
    * @return  A new instance of {@code FontChooserGuiImpl} configured as requested. 
    */
   public static FontChooserGuiImpl createInstance(int windowId,
                                                   boolean ansiOnly,
                                                   boolean fixedOnly,
                                                   int minSize,
                                                   int maxSize)
   {
      WidgetId fid = WidgetId.nextID();
      FrameConfig fCfg = new FrameConfig(fid);
      fCfg.dialog = true;
      fCfg.font = FontManager.ALERT_BOX_FONT;
      fCfg.widthPixels = 431 - 4;
      fCfg.heightPixels = 319 - 4;
      fCfg.windowID = -1; // parent windowId;
      fCfg.scrollable = false;
      fCfg.scrollbarVertical = false;
      fCfg.scroll = 0; 
      fCfg.overlay = false;
      fCfg.title = "Font";
      fCfg.sideLabels = false;
      fCfg.labels = false;
      fCfg.threeD = true;
      
      FontChooserGuiImpl ret = new FontChooserGuiImpl(fid, fCfg);
      ret.initialize(fid, fCfg);
      ret.init(ansiOnly, fixedOnly, minSize, maxSize);
      return ret;
   }
   
   /**
    * Private constructor. To create this kind of object use the {@link #createInstance} factory
    * method.
    * 
    * @param   id
    *          The {@link WidgetId} of the dialog. 
    * @param   cfg
    *          The {@link FrameConfig} of the dialog.
    */
   private FontChooserGuiImpl(WidgetId id, FrameConfig cfg)
   {
      super(id, cfg);
      setLayout(new FontChooserLayout());
   }
   
   /**
    * Initialize the dialog by performing tasks that was not possible in the constructor because
    * it was not fully created.
    * 
    * @param   ansiOnly
    *          If {@code true} filter only the ANSI fonts.
    * @param   fixedOnly
    *          If {@code true} filter only fixed-sized the ANSI fonts.
    * @param   minSize
    *          The minimum size of the fonts allowed to be selected.
    * @param   maxSize
    *          The maximum size of the fonts allowed to be selected.
    */
   protected void init(boolean ansiOnly, boolean fixedOnly, int minSize, int maxSize)
   {
      if (gd != null)
      {
         return; // already initialized
      }
      
      // the titlebar can be null when the web client is in embedded mode
      WindowTitleBar tbar = topLevelWindow().getTitleBar();
      if (tbar == null)
      {
         return;
      }
      tbar.setClickHandler(CaptionButtonType.CLOSE, this::dismiss);
      
      gd = (GuiDriver<?, ?>) OutputManager.getDriver();
      cc = screen().coordinates();
      
      fontLabel = createSimpleLabel("&Font:", 147, 13, false);
      addWidget(fontLabel);
      
      styleLabel = createSimpleLabel("Font st&yle:", 111, 13, false);
      addWidget(styleLabel);
      
      sizeLabel = createSimpleLabel("&Size:", 54, 13, false);
      addWidget(sizeLabel);
      
      effectsGroup = createGoupbox(11, 164, 148, 112);
      addWidget(effectsGroup);
      
      effectsLabel = createSimpleLabel("Effects", 37, 13, false);
      effectsLabel.setOpaque();
      addWidget(effectsLabel);
      
      sampleGroup = createGoupbox(165, 164, 175, 65);
      addWidget(sampleGroup);
      
      sampleLabel = createSimpleLabel("Sample", 39, 13, false);
      sampleLabel.setOpaque();
      addWidget(sampleLabel);
      
      scriptLabel = createSimpleLabel("Sc&ript:", 174, 13, false);
      addWidget(scriptLabel);
      
      String[] fonts = ansiOnly 
            ? (fixedOnly ? ANSI_FIXED_FONTS : ANSI_FONTS)
            : (fixedOnly ? FIXED_FONTS : getAvailableFonts());
      ControlSetItem[] fontsCSI = new ControlSetItem[fonts.length];
      for (int i = 0; i < fontsCSI.length; i++)
      {
         fontsCSI[i] = new ControlSetItem(i,
                                          new character(fonts[i]),
                                          new character(fonts[i]));
      }
      fontFaceCbx = createCombobox(ComboBoxConfig.Mode.SIMPLE, 11, 26, 147, 119,
                                   fontsCSI, fonts[0], "x(50)");
      addWidget(fontFaceCbx);
      //fontFaceCbx.valueChangedCommon();
      fontFaceCbx.setVisible(true); // must be called AFTER addWidget() 
      
      ControlSetItem[] stylesCSI = new ControlSetItem[STYLES.length];
      for (int i = 0; i < STYLES.length; i++)
      {
         stylesCSI[i] = new ControlSetItem(i, new character(STYLES[i]), new character(STYLES[i]));
      }
      fontStyleCbx = createCombobox(ComboBoxConfig.Mode.SIMPLE, 165, 26, 111, 119,
                                    stylesCSI, STYLES[0], "x(50)");
      addWidget(fontStyleCbx);
      //fontStyleCbx.valueChangedCommon();
      fontStyleCbx.setVisible(true);  // must be called AFTER addWidget()
      
      // filter only the sizes that match our constraints
      this.minSize = minSize;
      this.maxSize = maxSize;
      int[] sizes = Arrays.stream(DEFAULT_SIZES)
                          .filter(v -> v >= minSize)
                          .filter(v -> v <= maxSize)
                          .toArray();
      ControlSetItem[] sizesCSI = new ControlSetItem[sizes.length];
      for (int i = 0; i < sizesCSI.length; i++)
      {
         sizesCSI[i] = new ControlSetItem(i,
                                          new character(String.valueOf(sizes[i])),
                                          new integer(sizes[i]));
      }
      fontSizeCbx = createCombobox(ComboBoxConfig.Mode.SIMPLE, 285, 26, 54, 119,
                                   sizesCSI, String.valueOf(sizes[0]), ">>9");
      addWidget(fontSizeCbx);
      //fontSizeCbx.valueChangedCommon();
      fontSizeCbx.setVisible(true);  // must be called AFTER addWidget()
      
      ControlSetItem[] scriptCSI = new ControlSetItem[DEFAULT_SCRIPTS.length];
      for (int i = 0; i < scriptCSI.length; i++)
      {
         scriptCSI[i] = new ControlSetItem(i,
                                           new character(DEFAULT_SCRIPTS[i]),
                                           new character(DEFAULT_SCRIPTS[i]));
      }
      scriptCbx = createCombobox(ComboBoxConfig.Mode.DROP_DOWN_LIST, 165, 255, 174, 20,
                                 scriptCSI, DEFAULT_SCRIPTS[0], "x(50)");
      addWidget(scriptCbx);
      //scriptCbx.valueChangedCommon();
      scriptCbx.setVisible(true);
      
      strikeChk = createTogglebox("Stri&keout", 20, 178, 130, 17);
      addWidget(strikeChk);
      
      underlineChk = createTogglebox("&Underline", 20, 199, 130, 17);
      addWidget(underlineChk);
      
      preview = new Preview();
      addWidget(preview);
      
      accept = createButton("OK", "OK", 68, 23, true);
      config().defaultButtonId = accept.getId().asInt();
      addWidget(accept);
      
      reject = createButton("Cancel", "CANCEL", 68, 23, false);
      config().cancelButtonId = reject.getId().asInt();
      addWidget(reject);
      
      screen().getRegistry().addFrame(this);
      setSize(cc.dimensionFromNative(new NativeDimension(NATIVE_WIDTH, NATIVE_HEIGHT)));
   }
   
   /**
    * Checks whether the user accepted the selection from the dialog.
    *
    * @return  {@code true} if the dialog was accepted, and {@code false} if the dialog was not
    *          yet displayed or it was rejected by user.
    */
   public boolean isAccepted()
   {
      return accepted;
   }
   
   /**
    * Get status of the input processing.
    *
    * @return  current status of the input processing.
    */
   @Override
   public boolean isDismissed()
   {
      return inputDone;
   }
   
   /**
    * Set the result code, hide the dialog and flag input as complete.
    */
   @Override
   public void dismiss()
   {
      inputDone = true;
      destroy();
   }
   
   /**
    * This method is invoked when action event is generated. Only processes events generated by 
    * member widgets.
    *
    * @param   event
    *          Event which triggered this call.
    */
   @Override
   public void onAction(ActionEvent event)
   {
      String cmd = event.command();
      if (cmd == null)
      {
         return; // unknown command
      }
      
      switch (cmd)
      {
         case "OK":
            // validate first
            String strSize = fontSizeCbx.getScreenValue().toStringMessage().toLowerCase();
            int size = getFontSize();
            
            if (size < minSize || size > maxSize)
            {
               String msg = "Size must be between " + minSize + " and " + maxSize + " points.";
               ThinClient.getInstance().messageBox(new Object[] { msg },
                                                   null,
                                                   LogicalTerminal.ALERT_INFO,
                                                   LogicalTerminal.BTN_OK,
                                                   "Font",
                                                   null,
                                                   WindowConfig.RESOLVE_WINDOW,
                                                   null);
               return;
            }
            
            inputDone = true;
            accepted = true; // first check whether a file was selected
            dismiss();
            return;
         
         case "CANCEL":
            inputDone = true;
            accepted = false;
            dismiss();
            return;
      }
   }
   
   /**
    * Configures the dialog by setting the widgets to reflect the properties of the current font.  
    * 
    * @param   fontName
    *          The font name.
    * @param   fontSize
    *          The font size.
    * @param   bold
    *          {@code true} if the font is bold.
    * @param   italic
    *          {@code true} if the font is italic.
    * @param   underline
    *          {@code true} if the font is underlined.
    * @param   strikeout
    *          {@code true} if the font is stroke at the middle.
    */
   public void configure(String fontName,
                         int fontSize,
                         boolean bold,
                         boolean italic,
                         boolean underline,
                         boolean strikeout)
   {
      strikeChk.config().checked = strikeout;
      underlineChk.config().checked = underline;
      
      fontFaceCbx.setValue(new character(fontName));
      fontFaceCbx.valueChangedCommon();
      
      String style = ((bold ? "Bold" : "") + (italic ? " Italic" : "")).trim();
      fontStyleCbx.setValue(new character(style.isEmpty() ? "Regular" : style));
      fontStyleCbx.valueChangedCommon();
      
      fontSizeCbx.setValue(new integer(fontSize));
      fontSizeCbx.valueChangedCommon();
      
      //scriptCbx.valueChangedCommon();
      
      repaint(); // update the screen with configured values
   }
   
   /**
    * Obtain the currently selected size of the font.
    * 
    * @return  the selected size of the font.
    */
   public int getFontSize()
   {
      String strSize = fontSizeCbx.getScreenValue().toStringMessage().toLowerCase();
      int size = 8;
      try
      {
         size = Integer.parseInt(strSize);
      }
      catch (NumberFormatException e)
      {
         // defaulting to size 8
      }
      return size;
   }
   
   /**
    * Returns the font that the user has selected.
    * 
    * @return  if the dialog was successfully accepted, a {@link FontDetails} object with all
    *          properties from the dialog set and {@code null} if the dialog was not accepted
    *          (still running or rejected).
    */
   public FontDetails<?> getSelectedFont()
   {
      if (!accepted)
      {
         return null;
      }
      String strFace = fontFaceCbx.getScreenValue().toStringMessage();
      String strStyle = fontStyleCbx.getScreenValue().toStringMessage().toLowerCase();
      int size = getFontSize();
      boolean isBold = strStyle.contains("bold");
      boolean isItalic = strStyle.contains("italic");
      boolean isUnderline = underlineChk.config().checked;
      boolean isStroke = strikeChk.config().checked;
      
      FontStyle style = FontStyle.createStyle(isBold, isItalic, isUnderline);
      String key = FontManager.getFontKey(strFace, size, isBold, isItalic, isUnderline);
      
      return new FontDetails(key, strFace, style, size, -1, -1, -1, null, false);
   }
   
   /**
    * Adding new widget to the container. This method does not require the coordinates to be
    * defined. The location will be selected automatically. Makes sure this frame is also set on
    * target {@code widget}.
    *
    * @param   widget
    *          The new widget to be added to the container.
    */
   @Override
   public void addWidget(Widget<GuiOutputManager> widget)
   {
      super.addWidget(widget);
      widget.config().frameId = getId().asInt();
   }
   
   /**
    * Obtain available fonts from the graphic environment. This is useful for Swing driver but
    * not applicable on Web driver.
    * 
    * @return  A list containing all installed fonts. 
    */
   private static String[] getAvailableFonts()
   {
      GraphicsEnvironment env = java.awt.GraphicsEnvironment.getLocalGraphicsEnvironment();
      String fonts[] = env.getAvailableFontFamilyNames();
      // Font[] allFonts = env.getAllFonts();
      return fonts;
   }
   
   /**
    * Creates and initializes a customized button widget.
    * 
    * @param   text
    *          The text (label) displayed on the button.  
    * @param   action
    *          The action to trigger when button is pressed.
    * @param   width
    *          The width of the button (native units).
    * @param   height
    *          The height of the button (native units). 
    * @param   def
    *          {@code true} if this is the default button for the frame.
    * 
    * @return  A reference to {@link ButtonGuiImpl} instance.
    */
   private ButtonGuiImpl createButton(String text,
                                      String action,
                                      int width,
                                      int height,
                                      boolean def)
   {
      WidgetId bid = WidgetId.nextID();
      
      ButtonConfig bcfg = new ButtonConfig(bid);
      bcfg.widthPixels = width;
      bcfg.heightPixels = height;
      bcfg.isDefault = def;
      bcfg.flatButton = false;
      bcfg.autoResize = false;
      bcfg.fixedHeight = true;
      bcfg.fixedWidth = true;
      
      ButtonGuiImpl b = new ButtonGuiImpl(bid, bcfg);
      b.initialize(bid, bcfg);
      b.setSupportsMnemonic();
      b.setTextLabel(text);
      b.setCommand(action);
      b.setEnabled(true);
      b.setVisible(true);
      b.addListener(this);
      
      return b;
   }
   
   /**
    * Creates a customized {@code SimpleLabel} to be used on this dialog. 
    * 
    * @param   text
    *          The text to be displayed.
    * @param   w
    *          The physical width of the label.
    * @param   h
    *          The physical height of the label.
    * @param   rightAlign
    *          Whether the text is right aligned.
    *          
    * @return  A customized {@code SimpleLabel}.
    */
   private SimpleLabel createSimpleLabel(String text, int w, int h, boolean rightAlign)
   {
      SimpleLabel label = new SimpleLabel(text);
      label.setSize(cc.dimensionFromNative(new NativeDimension(w, h)));
      LabelConfig lcfg = (LabelConfig) label.config();
      lcfg.align = rightAlign ? ControlEntity.ALIGN_RIGHT : ControlEntity.ALIGN_LEFT;
      return label;
   }
   
   /**
    * Creates a customized {@code RectangleGuiImpl} to be used on this dialog. 
    *
    * @param   w
    *          The physical width of the widget.
    * @param   h
    *          The physical height of the widget.
    * @param   x
    *          The physical X coordinate of the widget. 
    * @param   y
    *          The physical Y coordinate of the widget.
    *
    * @return  A customized {@code RectangleGuiImpl}.
    */
   private RectangleGuiImpl createGoupbox(int x, int y, int w, int h)
   {
      WidgetId rid = WidgetId.nextID();
      RectangleConfig rcfg = new RectangleConfig(rid);
      rcfg.clientX = x;
      rcfg.clientY = y;
      rcfg.heightPixels = w;
      rcfg.widthPixels = h;
      rcfg.heightChars = cc.heightFromNative(h);
      rcfg.widthChars = cc.widthFromNative(w);
      rcfg.autoResize = false;
      rcfg.groupBox = true;
      
      RectangleGuiImpl group = new RectangleGuiImpl(rid, rcfg);
      group.initialize(rid, rcfg);
      group.setVisible(true);
      group.setEnabled(true);
      return group;
   }
   
   /**
    * Creates a customized {@code ComboBoxGuiImpl} to be used on this dialog. 
    *
    * @param   mode
    *          The type of combobox to be returned.
    * @param   w
    *          The physical width of the widget.
    * @param   h
    *          The physical height of the widget.
    * @param   x
    *          The physical X coordinate of the widget. 
    * @param   y
    *          The physical Y coordinate of the widget.
    * @param   items
    *          An array of items to be used in the combobox.
    * @param   selVal
    *          The value to be selected bu default.
    * @param   format
    *          The format of the displayed data. 
    *
    * @return  A customized {@code ComboBoxGuiImpl}.
    */
   private ComboBoxGuiImpl createCombobox(ComboBoxConfig.Mode mode,
                                          int x,
                                          int y,
                                          int w,
                                          int h,
                                          ControlSetItem[] items,
                                          String selVal,
                                          String format)
   {
      WidgetId rid = WidgetId.nextID();
      ComboBoxConfig ccfg = new ComboBoxConfig();
      ccfg.id = rid;
      ccfg.clientX = x;
      ccfg.clientY = y;
      ccfg.widthPixels = h;
      ccfg.heightPixels = w;
      ccfg.widthChars = cc.widthFromNative(w);
      ccfg.heightChars = cc.heightFromNative(h);
      ccfg.clientWidthChars = cc.widthFromNative(w);
      ccfg.clientHeightChars = cc.heightFromNative(h);
      ccfg.autoResize = false;
      ccfg.fixedWidth = true;
      ccfg.fixedHeight = true;
      ccfg.mode = mode;
      ccfg.items = items;
      ccfg.format = format;
      
      ComboBoxGuiImpl combo = new ComboBoxGuiImpl(rid, ccfg);
      combo.initialize(rid, ccfg);
      combo.setEnabled(true);
      combo.setPhysicalLocation(x, y);
      ComboBoxModel<String> comboModel = combo.model();
      for (ControlSetItem it : items)
      {
         comboModel.add(it.getLabel().toStringMessage());
      }
      combo.setValue(new character(selVal));
      combo.valueChangedCommon();
      comboModel.select(selVal);
      return combo;
   }
   
   /**
    * Creates a customized {@code ToggleBoxGuiImpl} to be used on this dialog. 
    *
    * @param   text
    *          The text (label) of the toggle button.
    * @param   w
    *          The physical width of the widget.
    * @param   h
    *          The physical height of the widget.
    * @param   x
    *          The physical X coordinate of the widget. 
    * @param   y
    *          The physical Y coordinate of the widget.
    *
    * @return  A customized {@code ToggleBoxGuiImpl}.
    */
   private ToggleBoxGuiImpl createTogglebox(String text, int x, int y, int w, int h)
   {
      WidgetId tid = WidgetId.nextID();
      ToggleBoxConfig tcfg = new ToggleBoxConfig();
      tcfg.id = tid;
      tcfg.clientX = x;
      tcfg.clientY = y;
      tcfg.widthPixels = w;
      tcfg.heightPixels = h;
      tcfg.autoResize = false;
      tcfg.fixedWidth = true;
      tcfg.fixedHeight = true;
      tcfg.label = text;
      
      ToggleBoxGuiImpl toggle = new ToggleBoxGuiImpl(tid, this, text)
      {
         /**
          * Calculate the width of this widget.
          * <p>
          * NOTE: initialize() calls width() which needs [gf] which is assigned in initialize(),
          *       after returning from super.initialize()  
          *
          * @return    The width in characters.
          */
         @Override
         public double width()
         {
            return cc.widthFromNative(w);
         }
         
         /**
          * Intercept the inverse checked state processing and force refresh the preview.
          *
          * @param  e
          *         Mouse event or null, that means, that the source event is a key press.
          */
         @Override
         protected void check(MouseEvent e)
         {
            super.check(e);
            preview.repaint();
         }
      };
      toggle.initialize(tid, tcfg);
      toggle.setVisible(true);
      toggle.setEnabled(true);
      toggle.setText(text);
      toggle.setTextLabel(text);
      toggle.setPhysicalLocation(x, y);
      return toggle;
   }
   
   /**
    * The manager that does the layout of the widgets of this frame. 
    */
   private class FontChooserLayout
   implements LayoutManager<GuiOutputManager>
   {
      /**
       * Calculate minimum size required for widgets stored in passed container.
       *
       * @param   container
       *          Container to process.
       *
       * @return  A constant dimension for this unresizable dialog.
       */
      @Override
      public Dimension minimumSize(Container<GuiOutputManager> container)
      {
         // constant minim size:
         return cc.dimensionFromNative(new NativeDimension(NATIVE_WIDTH - 5, NATIVE_HEIGHT - 5));
      }
      
      /**
       * Place components in specified container.
       *
       * @param   container
       *          Container to process.
       */
      @Override
      public void doLayout(Container<GuiOutputManager> container)
      {
         fontLabel.setPhysicalLocation(11, 11);
         styleLabel.setPhysicalLocation(165, 11);
         sizeLabel.setPhysicalLocation(284, 11);
         effectsLabel.setPhysicalLocation(20, 158);
         effectsGroup.setPhysicalLocation(11, 164);
         sampleLabel.setPhysicalLocation(174, 158);
         sampleGroup.setPhysicalLocation(165, 164);
         scriptLabel.setPhysicalLocation(165, 239);
         
         fontFaceCbx.setPhysicalLocation(11, 26);
         fontStyleCbx.setPhysicalLocation(165, 26);
         fontSizeCbx.setPhysicalLocation(285, 26);
         
         preview.setPhysicalLocation(172, 172);
         
         accept.setPhysicalLocation(347, 26);
         reject.setPhysicalLocation(347, 52);
      }
   }
   
   /**
    * Shows a quick preview of the font properties selected in dialog. It is automatically updated
    * each time a property is changed.
    */
   private class Preview
   extends AbstractWidget
   implements ConfigOwner
   {
      /** The widget configuration for {@code Preview}s. */
      BaseConfig myConfig;
      
      /** Constant representing the total width of this widget. */
      private final int NATIVE_WIDTH = 160;
      
      /** Constant representing the height of this widget. */
      private final int NATIVE_HEIGHT = 49;
      
      /** The short text used to show the preview.  */
      private static final String PREVIEW_STRING = "AaBbYyZz";
      
      /** The constructor creates the preview object. */
      private Preview()
      {
         myConfig = new BaseConfig();
         myConfig.id = WidgetId.nextID();
         myConfig.fixedHeight = true;
         myConfig.fixedWidth = true;
         myConfig.heightChars = cc.heightFromNative(NATIVE_HEIGHT);
         myConfig.widthChars = cc.widthFromNative(NATIVE_WIDTH);
         myConfig.name = "preview"; // direct access
         setEnabled(false);
      }
   
      /**
       * Draw the picker widget. Called by this Button's parent container.
       */
      @Override
      public void draw()
      {
         int bgColor = isThreeD() ? SysColor.COLOR_3DFACE : SysColor.COLOR_FILLIN_BG;
         GuiColorResolver gc = new GuiColorResolver(gd, null, bgColor, SysColor.COLOR_FILLIN_FG);
         gc.colors();
         
         // get location and size
         NativePoint p = physicalLocation();
         NativeDimension dim = physicalDimension();
         NativeRectangle clip = clipRectangle(p, dim);
         
         // clippable drawing start 
         gd.draw(this.getId(), p, clip, () -> {
            String strFace = fontFaceCbx.getScreenValue().toStringMessage();
            String strStyle = fontStyleCbx.getScreenValue().toStringMessage().toLowerCase();
            int size = getFontSize();
            boolean bold = strStyle.contains("bold");
            boolean italic = strStyle.contains("italic");
            int style = (bold ? Font.BOLD : Font.PLAIN) + (italic ? Font.ITALIC : Font.PLAIN);
            boolean underline = underlineChk.config().checked;
            boolean stroke = strikeChk.config().checked;
            
            BufferedImage offImage = new BufferedImage(NATIVE_WIDTH,
                                                       NATIVE_HEIGHT,
                                                       BufferedImage.TYPE_INT_ARGB);
            Graphics2D g = offImage.createGraphics();
            g.setFont(new Font(strFace, style, size * 4 / 3)); // convert point to pixel size
            g.setColor(Color.black);
            FontMetrics metrics = g.getFontMetrics();
            int textWidth = metrics.stringWidth(PREVIEW_STRING);
            int x = (NATIVE_WIDTH - textWidth) / 2;
            int y = ((NATIVE_HEIGHT - metrics.getHeight()) / 2) + metrics.getAscent();
            
            // finally draw it to of graphic:
            g.drawString(PREVIEW_STRING, x, y);
            if (underline)
            {
               g.drawLine(x, y + 1, x + textWidth, y + 1);
            }
            if (stroke)
            {
               g.drawLine(x, NATIVE_HEIGHT / 2, x + textWidth, NATIVE_HEIGHT / 2);
            }
               
            ImageWrapper<BufferedImage> offImg = new ImageWrapper<>(offImage,
                                                                    NATIVE_HEIGHT,
                                                                    NATIVE_WIDTH);
            // actual drawing using FWD driver
            gd.setColor(gc.bgColor);
            gd.fillRect(0, 0, dim.width, dim.height);
            gd.drawImage(offImg, 0, 5, NATIVE_WIDTH, NATIVE_HEIGHT, 0, 0, false, false, true, false);
         });
      }
   
      /**
       * Get widget width.
       *
       * @return  widget width.
       */
      @Override
      public double width()
      {
         return myConfig.widthChars;
      }
   
      /**
       * Get widget height.
       *
       * @return  widget height.
       */
      @Override
      public double height()
      {
         return myConfig.heightChars;
      }
   
      /**
       * Get widget dimension.
       *
       * @return  widget dimension.
       */
      @Override
      public Dimension dimension()
      {
         return cc.dimensionFromNative(new NativeDimension(NATIVE_WIDTH, NATIVE_HEIGHT));
      }
      
      /**
       * Get widget minimum size.
       *
       * @return  widget minimum size.
       */
      @Override
      public Dimension minimumSize()
      {
         return dimension();
      }
      
      /**
       * Returns config instance associated with this instance.
       *
       * @return  Config reference.
       */
      @Override
      public WidgetConfig config()
      {
         return myConfig;
      }
      
      /**
       * The method is called after the configuration associated with the owner changes. It will 
       * never be called since this object does not have a server-side correspondent.
       *
       * @param   beforeUpdate
       *          Config reference capturing the config state before 
       *          the owner's config was modified.
       */
      @Override
      public void afterConfigUpdate(WidgetConfig beforeUpdate)
      {
      }
   }
}
