/*
** Module   : LineEditor.java
** Abstract : a line editor based on FillIn
**
** Copyright (c) 2017, Golden Code Development Corporation.
**
** -#- -I- --Date-- ---------------------------------Description---------------------------------
** 001 OM  20171113 Initial commit.
*/

/*
** This program is free software: you can redistribute it and/or modify
** it under the terms of the GNU Affero General Public License as
** published by the Free Software Foundation, either version 3 of the
** License, or (at your option) any later version.
**
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
** GNU Affero General Public License for more details.
**
** You may find a copy of the GNU Affero GPL version 3 at the following
** location: https://www.gnu.org/licenses/agpl-3.0.en.html
** 
** Additional terms under GNU Affero GPL version 3 section 7:
** 
**   Under Section 7 of the GNU Affero GPL version 3, the following additional
**   terms apply to the works covered under the License.  These additional terms
**   are non-permissive additional terms allowed under Section 7 of the GNU
**   Affero GPL version 3 and may not be removed by you.
** 
**   0. Attribution Requirement.
** 
**     You must preserve all legal notices or author attributions in the covered
**     work or Appropriate Legal Notices displayed by works containing the covered
**     work.  You may not remove from the covered work any author or developer
**     credit already included within the covered work.
** 
**   1. No License To Use Trademarks.
** 
**     This license does not grant any license or rights to use the trademarks
**     Golden Code, FWD, any Golden Code or FWD logo, or any other trademarks
**     of Golden Code Development Corporation. You are not authorized to use the
**     name Golden Code, FWD, or the names of any author or contributor, for
**     publicity purposes without written authorization.
** 
**   2. No Misrepresentation of Affiliation.
** 
**     You may not represent yourself as Golden Code Development Corporation or FWD.
** 
**     You may not represent yourself for publicity purposes as associated with
**     Golden Code Development Corporation, FWD, or any author or contributor to
**     the covered work, without written authorization.
** 
**   3. No Misrepresentation of Source or Origin.
** 
**     You may not represent the covered work as solely your work.  All modified
**     versions of the covered work must be marked in a reasonable way to make it
**     clear that the modified work is not originating from Golden Code Development
**     Corporation or FWD.  All modified versions must contain the notices of
**     attribution required in this license.
*/

package com.goldencode.p2j.ui.client.gui;

import com.goldencode.p2j.ui.*;
import com.goldencode.p2j.ui.client.event.*;
import com.goldencode.p2j.ui.client.event.ActionEvent;
import com.goldencode.p2j.ui.client.event.FocusEvent;
import com.goldencode.p2j.ui.client.event.listener.ActionListener;
import com.goldencode.p2j.ui.client.event.listener.FocusListener;
import java.awt.event.*;
import java.util.*;

/**
 * A line editor based on {@link FillInGuiImpl}. It will notify the listeners when the user
 * validates the content by pressing the RETURN key.
 * Other features that lacks in standard {@code FillInGuiImpl}:
 * - catches focus on click
 * - improved mouse selection
 * - validates changes on focus lost (Windows explorer specific, might need to make it togglable)
 * - notifies listeners for ENTER and ESCAPE key pressed
 */
public class LineEditor
extends FillInGuiImpl
{
   /** The action that is fired when ENTER/RETURN key is pressed while editing the text. */
   public static final String AC_TEXT_VALIDATED = "TEXT_VALIDATED";
   
   /** The action that is fired when ESCAPE key is pressed while editing the text. */
   public static final String AC_CANCEL_EDIT = "CANCEL_EDIT";
   
   /** 
    * The editor will emit fire {@code AC_TEXT_VALIDATED} on focus lost only if this flag is set
    * to {@code true} (default value).
    */
   private boolean validateOnFocusLost = true;
   
   /** The list of active listeners. */
   private List<ActionListener> actionListeners = new ArrayList<>();
   
   /**
    * Special constructor used to restore component using given configuration.
    * <p>
    * Call {@link #initialize} to finish the widget initialization.
    *
    * @param   fid
    *          The widget's ID.
    * @param   fcfg
    *          Configuration data.
    */
   public LineEditor(WidgetId fid, FillInConfig fcfg)
   {
      super(fid, fcfg);
   
      addFocusListener(new FocusListener()
      {
         /**
          * This method is invoked when focus is gained.
          *
          * @param   event
          *          Event which triggered this call.
          */
         @Override
         public void onFocusGained(FocusEvent event)
         {
            // NO-OP
         }
         
         /**
          * This method is invoked when focus is lost.
          *
          * @param   event
          *          Event which triggered this call.
          */
         @Override
         public void onFocusLost(FocusEvent event)
         {
            if (validateOnFocusLost)
            {
               // on windows, when focus is lost, the file chooser will validate the changes for
               //       the windows users too lazy to confirm with ENTER
               for (ActionListener listener : actionListeners)
               {
                  listener.onAction(new ActionEvent(LineEditor.this, AC_TEXT_VALIDATED));
               }
            }
         }
      });
   }
   
   /**
    * Notification of a mouse pressed event occurred for this widget.
    * <p>
    * If the widget is not already focused, it forces the focus to it.  After the focus was
    * received, it computes the caret location, based on the mouse position over the widget's
    * content.
    *
    * @param   e
    *          The mouse event.
    */
   @Override
   public void mousePressed(MouseEvent e)
   {
      // just focus request, no superclass call to avoid click event duplication
      // see FillInGuiImpl for clarification mouse events handling
      if (isEnabled())
      {
         requestFocus();
      }
   }
   
   /**
    * Notification of a mouse clicked event occurred for this widget.
    * <p>
    * If the widget is not already focused, it forces the focus to it.  After the focus was
    * received, it computes the caret location, based on the mouse position over the widget's
    * content.
    *
    * @param    e
    *           The mouse event.
    */
   @Override
   public void mouseClicked(MouseEvent e)
   {
      super.mouseClicked(e);
      
      if (e.getClickCount() == 2)
      {
         // select the word
         int cp = config.appDataPres.getCursorPos();
         String text = getText();
         if (text == null || text.isEmpty())
         {
            return;
         }
         int sw = cp; // find the start of the word
         while (sw > 0 && (sw >= text.length() || text.charAt(sw) != ' '))
         {
            --sw; 
         }
         if (text.charAt(sw) == ' ' && sw < text.length())
         {
            ++sw; 
         }
         int ew = cp; // find the end of the word
         while (ew < text.length() && text.charAt(ew) != ' ')
         {
            ++ew; 
         }
         config.selectionStart = sw;
         config.selectionEnd = ew;
         config.appDataPres.setCursorPos(ew);
         setCursorOffset(ew);
         repaint();
      }
      else if (e.getClickCount() == 3)
      {
         // select all text
         config.selectionStart = 0;
         config.selectionEnd = getLastOffset() + 1;
         moveCursorToLast();
         repaint();
      }
   }
   
   /**
    * Process specific keys.
    *
    * @param   key
    *          Key event structure.
    */
   @Override
   public void processKeyEvent(KeyInput key)
   {
      super.processKeyEvent(key);
      
      ActionEvent event = null;
      if (key.keyCode() == Key.VK_ENTER)
      {
         event = new ActionEvent(this, AC_TEXT_VALIDATED);
      }
      else if (key.keyCode() == Key.VK_ESCAPE)
      {
         event = new ActionEvent(this, AC_CANCEL_EDIT);
      }
      
      if (event != null)
      {
         for (ActionListener listener : actionListeners)
         {
            listener.onAction(event);
         }
      }
   }
   
   /**
    * Adds an {@code ActionListener} to this editor. It will be notified when user confirms the
    * current text by typing {@code RETURN}.
    *
    * @param  listener
    *         The {@code ActionListener} to be added.
    */
   public void addListener(ActionListener listener)
   {
      actionListeners.add(listener);
   }
   
   /**
    * Removes an {@code ActionListener} from the list of listeners. The listener will not be
    * notified any more when user confirms the current text by typing {@code RETURN}.
    *
    * @param  listener
    *         The {@code ActionListener} to be removed.
    */
   public void removeListener(ActionListener listener)
   {
      actionListeners.remove(listener);
   }
   
   /**
    * Checks if this editor will emit {@code AC_TEXT_VALIDATED} action event when it loses focus.
    * 
    * @return  current value of {@code validateOnFocusLost} flag.
    */
   public boolean isValidateOnFocusLost()
   {
      return validateOnFocusLost;
   }
   
   /**
    * Sets whether  this editor will emit {@code AC_TEXT_VALIDATED} action event when it loses
    * focus.
    *
    * @param   valid
    *          new value for {@code validateOnFocusLost} flag.
    */
   public void setValidateOnFocusLost(boolean valid)
   {
      this.validateOnFocusLost = valid;
   }
}
