/*
** Module   : MenuGuiImpl.java
** Abstract : GUI client implementation of Menu.
**
** Copyright (c) 2015-2018, Golden Code Development Corporation.
**
** -#- -I- --Date-- -----------------------------Description-------------------------------------
** 001 VIG 20150525 Created initial version.
** 002 VIG 20150818 Added keyMode flag to indicate key navigation mode.
** 003 CA  20150910 Fixed some issues related to menu acting as the system popup menu.
** 004 CA  20151007 Workaround for mismatch of clipping behavior.
** 005 HC  20151013 Changes to extend GUI multi-window focus management and ACTIVE-WINDOW system
**                  handle processing to modal windows.
** 006 CA  20151026 Enclose driver API calls in select/releaseWindow().
**     VIG 20151103 Fixed NPE.
** 007 VIG 20151226 Added key navigation mode flag and focus reset to mouseClicked().
** 008 EVL 20160224 Javadoc fixes to make compatible with Oracle Java 8 for Solaris 10.
** 009 IAS 20160331 Fixed potential NPE
** 010 CA  20160216 Cache the intermediate widget states.
** 011 VIG 20160128 If this popup is focused in some container - reset focus after popup hiding.
**                  Added initSystemPopupMenu(), added currentPopup and popupShown().
** 012 HC  20160607 GUI menus reworked to be displayed in dedicated top-level windows.
** 013 VIG 20160620 Added darkShadow color, fixed popup menu sizes, fixed height/width methods.
** 014 VIG 20160628 Cleaned draw(), added width fix to width() method.
**                  Added getMenubarParentOrSelf() method, added flag mbarPressedState,
**                  fixed width and drawing. Added menubar state static flags.
** 015 VIG 20161105 getParent() call is refactored to getParentMenuContainer().
** 016 OM  20170320 Added support for Themes.
** 017 EVL 20180108 Adding fix for NPE caused by keyboard navigation.  Adding menubar highlight
**                  reset method.
** 018 CA  20180208 Notify the driver when it switches control to server-side (and vice-versa).
** 019 HC  20180309 Fixed multiple menu-originated abends.
** 020 HC  20180419 Fixed popup menu-related key processing and focus handling. Fixed popup menu not show up
**                  after closed due to its overlay window losing focus. Fixed menu input events sent to
**                  wrong widgets in web GUI.
** 021 HC  20180503 Fixed FILL-IN stops responding to mouse clicks after its popup menu
**                  displayed. Improved MENU focus management.
** 022 HC  20180528 Fixed focus-related menu abend, MENU widget in some cases didn't have its
**                  Container.current set, which caused an NPE in focus management logic. Also
**                  fixed focus-related draw artifacts.
*/
/*
** This program is free software: you can redistribute it and/or modify
** it under the terms of the GNU Affero General Public License as
** published by the Free Software Foundation, either version 3 of the
** License, or (at your option) any later version.
**
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
** GNU Affero General Public License for more details.
**
** You may find a copy of the GNU Affero GPL version 3 at the following
** location: https://www.gnu.org/licenses/agpl-3.0.en.html
** 
** Additional terms under GNU Affero GPL version 3 section 7:
** 
**   Under Section 7 of the GNU Affero GPL version 3, the following additional
**   terms apply to the works covered under the License.  These additional terms
**   are non-permissive additional terms allowed under Section 7 of the GNU
**   Affero GPL version 3 and may not be removed by you.
** 
**   0. Attribution Requirement.
** 
**     You must preserve all legal notices or author attributions in the covered
**     work or Appropriate Legal Notices displayed by works containing the covered
**     work.  You may not remove from the covered work any author or developer
**     credit already included within the covered work.
** 
**   1. No License To Use Trademarks.
** 
**     This license does not grant any license or rights to use the trademarks
**     Golden Code, FWD, any Golden Code or FWD logo, or any other trademarks
**     of Golden Code Development Corporation. You are not authorized to use the
**     name Golden Code, FWD, or the names of any author or contributor, for
**     publicity purposes without written authorization.
** 
**   2. No Misrepresentation of Affiliation.
** 
**     You may not represent yourself as Golden Code Development Corporation or FWD.
** 
**     You may not represent yourself for publicity purposes as associated with
**     Golden Code Development Corporation, FWD, or any author or contributor to
**     the covered work, without written authorization.
** 
**   3. No Misrepresentation of Source or Origin.
** 
**     You may not represent the covered work as solely your work.  All modified
**     versions of the covered work must be marked in a reasonable way to make it
**     clear that the modified work is not originating from Golden Code Development
**     Corporation or FWD.  All modified versions must contain the notices of
**     attribution required in this license.
*/

package com.goldencode.p2j.ui.client.gui;

import java.awt.event.MouseEvent;
import java.util.*;

import com.goldencode.p2j.ui.*;
import com.goldencode.p2j.ui.chui.*;
import com.goldencode.p2j.ui.client.*;
import com.goldencode.p2j.ui.client.event.*;
import com.goldencode.p2j.ui.client.gui.driver.*;
import com.goldencode.p2j.ui.client.gui.theme.*;
import com.goldencode.p2j.ui.client.widget.*;

/**
 * <code>Menu</code> GUI client implementation class. 
 */
@SuppressWarnings("unchecked")
public class MenuGuiImpl
extends Menu<GuiOutputManager>
implements MouseWidgetsEnumerator
{
   /** Screen driver */
   @SuppressWarnings("rawtypes")
   private GuiDriver gd;
   
   /** Number of MENUBAR lines */
   private int menubarLines = 1;
   
   /** 
    * If <code>true</code> =&gt; mnemonics are drawn.
    * This mode is intended only for MENUBAR. It activated using ALT key.
    */
   private boolean mnemonicsOn = false;
   
   /**
    * Special constructor used to restore component using given configuration.
    * 
    * @param   id
    *          The widget's ID.
    * @param   cfg
    *          Configuration data.
    */
   public MenuGuiImpl(WidgetId id, MenuConfig cfg)
   {
      super(id, cfg);
   }
   
   /**
    * Find the mouse source in this menu, recursive.
    * 
    * @param    menu
    *           The (sub)menu parent.
    * @param    p
    *           The mouse location, relative to parent.
    * @param    bodyBounds
    *           The bounds of the (sub)menu's body.
    *           
    * @return   The selected sub-menu or menu-item; <code>null</code> if the mouse is not over
    *           a child in the menu tree.
    */
   protected static Widget<GuiOutputManager> findMenuMouseSource(AbstractContainer menu,
                                                                 NativePoint       p, 
                                                                 NativeRectangle   bodyBounds)
   {
      CoordinatesConversion cc = menu.screen().coordinates();
      
      // we have two cases:
      // 1. the mouse is somewhere over one of this sub-menu's direct children.  in this case, 
      // search the "title" dimension
      // 2. the mouse is somewhere over one of this sub-menu's opened hierarchy children (down
      // the tree) - in this case, delegate search to each of the sub-menu
      
      Widget<GuiOutputManager>[] widgets = menu.widgets();
      
      if (bodyBounds != null && bodyBounds.contains(p))
      {
         for (int i = widgets.length - 1; i >= 0; i--)
         {
            Widget<GuiOutputManager> w = widgets[i];
            
            if (!w.isVisible())
            {
               continue;
            }
            
            NativeRectangle wTitleBounds = null;
            
            if (w instanceof MenuItemGuiImpl)
            {
               // menu-items don't have children, so use their bounds as is
               wTitleBounds = w.physicalBounds();
            }
            else if (w instanceof SubMenuGuiImpl)
            {
               // sub-menus can have children, but search the "title" bounds first
               SubMenuGuiImpl sm = (SubMenuGuiImpl) w;
               
               NativePoint smTitleLoc = sm.physicalLocation();
               NativeDimension smTitleBounds = 
                  cc.dimensionToNative(new Dimension(sm.titleWidthWithBorders(), 
                                                     SubMenuGuiImpl.titleHeightWithBorders(sm)));
               
               wTitleBounds = new NativeRectangle(smTitleLoc, smTitleBounds);
            }
            
            if (wTitleBounds != null && wTitleBounds.contains(p))
            {
               return w;
            }
         }
      }
      else
      {
         // search the sub-menus
         for (int i = widgets.length - 1; i >= 0; i--)
         {
            Widget<GuiOutputManager> w = widgets[i];
            
            if (!w.isVisible())
            {
               continue;
            }
            
            if (w instanceof SubMenuGuiImpl)
            {
               NativePoint tp = new NativePoint(p);
               tp = tp.minus(w.physicalLocation());
               
               Widget<GuiOutputManager> found = w.findMouseSource(tp);
               
               if (found != null)
               {
                  return found;
               }
            }
         }
      }
      
      // nothing found, return null; don't delegate to super.findMouseSource()!
      return null;
   }

   /**
    * Replacement API for the logic which should exist in the widget's c'tor, but the logic 
    * can't be executed as it depends on the widget being registered into the official registry.
    * <p>
    * This API is called after the widget is created and added to the registry.
    * 
    * @param    id
    *           The widget's ID.
    * @param    cfg
    *           The config used to initialize this widget.
    */
   @SuppressWarnings("rawtypes")
   @Override
   public void initialize(WidgetId id, MenuConfig cfg)
   {
      super.initialize(id, cfg);
      
      gd = (GuiDriver) OutputManager.getDriver();
      
      Widget<?> parent = ThinClient.getInstance().getWidget(cfg.frameId);
      Window<?> window = parent != null ? parent.window() : WindowManager.resolveWindow(this);
   }
   
   /**
    * Get keyboard mnemonics mode.
    * 
    * @return  <code>true</code> if mnemonics are active in the menu.
    */
   public boolean isMnemonicsOn()
   {
      return this.mnemonicsOn;
   }
   
   /**
    * Set mnemonics mode.
    * 
    * @param   mnemonicsOn
    *          <code>true</code> if mnemonics should become active in the menu.
    */
   public void setMnemonicsOn(boolean mnemonicsOn)
   {
      this.mnemonicsOn = mnemonicsOn;
   }
   
   /**
    * Resets all highlight related variables to default state.
    */
   public void clearItemsHighlight()
   { 
      for (Widget<GuiOutputManager> w : widgets())
      {
         // allow submenu body to be closed
         if (w instanceof SubMenuGuiImpl)
         {
            // this call cleans activation flag and body pointer
            // method uses recusive calls inside SubMenuGuiImpl so can potentially be dangerous 
            ((SubMenuGuiImpl)w).resetBody();
         }
         w.setHighlighted(false);
      }
   }

   /**
    * Determine the mouse actions processed by this widget.
    * 
    * @return   See above.
    */
   @Override
   protected int[] mouseActions()
   {
      return new int [] { MouseEvent.MOUSE_CLICKED };
   }
   
   /**
    * Get default focusable widget.
    *
    * @return  default focusable widget.
    */
   @Override
   public Widget<GuiOutputManager> defaultFocus()
   {
      // We don't set default focus for sub-menus in GUI
      return null;
   }
   
   /**
    * Draw the peer container. The methods calls the draw() method for
    * ScrollPane instance for the peer container. In addition the required
    * clipping rectangle is calculated for ChildFrame containers because they
    * must be rendered inside the parent container.
    * <p>
    * Drawing for the redirected terminal differs. Labels/headers are drawn
    * only once per page, instead of the usual once per menu.
    */
   @Override
   public void draw()
   {
      NativePoint p = physicalLocation();
      NativeDimension dim = physicalDimension();
      NativeRectangle clipRect = clipRectangle(p, dim);
      
      final Window<?> wnd = WindowManager.resolveWindow(this);
      if (!config().popupOnly)
      {
         // if there is a non-popup menu, always re-draw the window workspace here... this is a
         // workaround from a "feature" which requires the redrawn area to be the rectangle which 
         // encloses all expanded sub-menus
         ((WindowGuiImpl) wnd).getWorkspace().draw();
         
         ScreenBitmap bitmap = wnd.screen().getBitmapCopy(wnd.getId().asInt());
         NativeRectangle nr = bitmap.getOuterRectangle();
         NativePoint screenLoc = screenPhysicalLocation();
         
         // the clipping rectangle is always with a (0,0) origin: thus we need to translate before
         // intersection
         if (nr != null && nr.intersects(clipRect.translate(screenLoc)))
         {
            // the resulting rectangle will have the same relative origin as your "clip" rectangle
            // so you can intersect safely
            nr = nr.translate(-screenLoc.x, -screenLoc.y); 
            clipRect = clipRect.intersection(nr);
         }
      }
      
      // prepare the background for all menu:
      gd.draw(getId(), p, clipRect, () ->
            ThemeManager.getCurrentTheme().drawMenuBackground(
                  gd, 0, 0, dim.width, dim.height, config().popupOnly));
      
      // then let the super class cal draw() on each components (menu items)
      gd.draw(getId(), p, clipRect, MenuGuiImpl.super::draw);
   }
   
   /**
    * Replace original version to take into account P2J specific behavior.
    * 
    * @see  com.goldencode.p2j.ui.client.widget.Widget#width()
    */
   @Override
   public double width()
   {
      return width(true); 
   }
   
   /**
    * Replace original version to take into account P2J specific behavior.
    * 
    * @see  com.goldencode.p2j.ui.client.widget.Widget#height()
    */
   @Override
   public double height()
   {
      return SubMenuGuiImpl.getMaxSubMenuHeight(this);
   }
   
   /**
    * Lay out container. Set children locations.
    */
   @Override
   public void doLayout()
   {
      TitledWindow<?> wnd = WindowManager.resolveWindow(this);
      CoordinatesConversion c = wnd.screen().coordinates();
      int menuItemHeight = c.heightToNative(SubMenuGuiImpl.titleHeightWithBorders(this));
      
      Widget<GuiOutputManager>[] w = widgets();
      boolean isMenuBar = !config().popupOnly;
      NativeInsets menuInsets = ThemeManager.getCurrentTheme().getMenuInsets();
      int x = isMenuBar ? 0 : menuInsets.left;
      int y = isMenuBar ? 0 : menuInsets.top;
      menubarLines = 1;
      int wndWidth = c.widthToNative(wnd.width());
      
      for (Widget<GuiOutputManager> mi : w)
      {
         // we redo the layout, invalidate the existing sizes
         if (mi instanceof MenuItemGuiImpl)
         {
            ((MenuItemGuiImpl) mi).invalidateCachedSize();
         }
         else if (mi instanceof SubMenuGuiImpl)
         {
            ((SubMenuGuiImpl) mi).invalidateCachedSize();
         }
         
         int itemHeight = c.heightToNative(mi.height());
         
         if (isMenuBar) // menu-bar
         {
            // including window borders
            if (x + c.widthToNative(mi.width()) > wndWidth - 10)
            {
               x = menuInsets.left;
               y += menuItemHeight;
               menubarLines++;
            }
            
            mi.setPhysicalLocation(x, y);
            mi.setLocation(c.widthFromNative(x), c.heightFromNative(y));
            x += c.widthToNative(mi.width());
            if (menuItemHeight < itemHeight)
            {
               // we found a taller menu item. use its height for further calculations.
               menuItemHeight = itemHeight;
            }
         }
         else
         {
            mi.setPhysicalLocation(x, y);
            mi.setLocation(c.widthFromNative(x), c.heightFromNative(y));
            y += itemHeight;
         }
         
         if (mi instanceof SubMenu)
         {
            ((SubMenu<?>) mi).doLayout();
         }
      }
      
      if (isMenuBar)
      {
         setHeight(c.heightFromNative(menubarLines * menuItemHeight));
      }
   }
   
   /**
    * Notification of a mouse clicked event occurred for this widget.
    * 
    * @param    e
    *           The mouse event.
    */
   @Override
   public void mouseClicked(MouseEvent e)
   {
      if (config().popupOnly && config().displayed)
      {
         // determine the real mouse source
         Widget<?> source = topLevelWindow().findMouseSource(new NativePoint(e.getX(), e.getY()));
         if (source != this && !(source instanceof MenuItem || source instanceof SubMenu))
         {
            // call repaint now, to re-draw all widgets bellow it
            repaint();
            setFocusInt(null);
            setVisible(false);
         }
      }
   }
   
   /**
    * Hide/show widget.
    * <p>
    * If the menu is being shown and not yet attached to a window, attached it to the default
    * window.
    * 
    * @param   visible
    *          New visible state.
    */
   @Override
   public void setVisible(boolean visible)
   {
      super.setVisible(visible);

      if (gd != null && config().popupOnly)
      {
         int mid = UiUtils.getWidgetIdAsInt(this);

         TopLevelWindow<?> window = parent(TopLevelWindow.class);
         if (window != null && !visible)
         {
            gd.withSelectedWindow(window.getId().asInt(), () ->
            {
               gd.deregisterWidget(mid);
            });
         }
      }
   }
   
   /**
    * Find the widget positioned just bellow the specified mouse position (in physical units),
    * in this menu bar, walking the entire opened menu tree, recursively.  
    * 
    * @param    p
    *           The mouse physical location.
    * 
    * @return   A widget over which the mouse is positioned or <code>null</code> if none found.
    */
   @Override
   public Widget<GuiOutputManager> findMouseSource(NativePoint p)
   {
      // search only the "title" bounds
      CoordinatesConversion cc = screen().coordinates();
      
      NativePoint menuLoc = physicalLocation();
      NativeDimension menuDim = cc.dimensionToNative(new Dimension(width(false), height()));
      NativeRectangle menuBounds = new NativeRectangle(menuLoc, menuDim);
      
      NativePoint tp = new NativePoint(p).translate(menuLoc);
      if (menuBounds.contains(tp))
      {
         return super.findMouseSource(p);
      }
      else
      {
         return findMenuMouseSource(this, p, null);
      }
   }
   
   /**
    * Looks up the very first enabled widget.
    * 
    * @return  the first enabled widget or <code>null</code>
    */
   @Override
   public Widget<GuiOutputManager> getFirstEnabledWidget()
   {
      return null;
   }
   
   /**
    * Returns the tab item list for this container. If the tab item list is not given, then
    * an empty list is returned.
    * 
    * @return   The tab item list.
    */
   @Override
   public List<Widget<GuiOutputManager>> getTabItemList()
   {
      // for menu get the menu item list
      return widgetsAsList();
   }

   /**
    * Hide widget.
    */
   @Override
   public void hide()
   {
      leaveMenu();
      super.hide();
   }

   /**
    * Close all opened bodies and remove focus from the menu.
    */
   @Override
   public void leaveMenu()
   {
      // do not proceed when menu is not "entered" (i.e. it doesn't have focus)
      if (!isVisible() || !isCurrent())
      {
         return;
      }

      boolean popup = config().popupOnly;

      if (!popup)
      {
         SubMenuGuiImpl.hideBodies(this, null);
      }
      
      setFocusInt(null);

      if (popup)
      {
         Widget<?> owner = screen().getRegistry().getComponent(config().ownerId);
         if (owner instanceof AbstractContainer)
         {
            AbstractContainer<?> container = (AbstractContainer<?>) owner;
            if (this.equals(container.getFocusInt()))
               container.setFocusInt(null);
         }
         
         TopLevelWindow<GuiOutputManager> wnd = parentOrSelf(TopLevelWindow.class);
         if (wnd != null)
         {
            // do not let the menu be destroyed
            detachFromPopupWindow(wnd);
            wnd.close();
         }
         
         config().displayed = false;
      }

      mnemonicsOn = false;

      ThinClient.getInstance().invokeLaterOS(this::triggerMenuClosed);

      for (Widget w : widgets())
      {
         w.setHighlighted(false);
      }

      repaint();
   }

   /**
    * Get widget which is in focus.
    * <p>
    * The method allows to choose whether the focused widget is searched in the nested
    * containers (<code>recurse == true</code>) or whether only this container is searched
    * (<code>recurse == false</code>).
    *
    * @param   recurse
    *          The parameter controls whether the method should recurse to the nested
    *          containers when searching for the focused widget.
    *
    * @return  focused widget.
    */
   @Override
   public Widget<GuiOutputManager> currentFocus(boolean recurse)
   {
      Widget focus = super.currentFocus(recurse);

      // this override ensures the (key) events are received by the popup menu when it is displayed
      if (focus == null)
      {
         focus = this;
      }

      return focus;
   }

   /**
    * Returns the focused direct child Component.
    *
    * This overrides makes sure {@link #focus()} and {@link #currentFocus(boolean)} are consistent.
    *
    * @return  The focused direct child Component.
    */
   @Override
   public Widget<GuiOutputManager> focus()
   {
      return currentFocus(false);
   }

   /**
    * Returns the list of widgets that will be considered when registering mouse widgets by GUI driver.
    *
    * @return  list of widgets, may be {@code null}.
    */
   @Override
   public List<Widget> getMouseWidgets()
   {
      return Arrays.asList(widgets());
   }

   /**
    * Process key event.
    *
    * @param event Key event to process.
    */
   @Override
   public void processKeyEvent(KeyInput event)
   {
      int actionCode = event.actionCode();
      Widget focus = null;

      if (actionCode == Keyboard.KA_CURSOR_DOWN)
      {
         focus = getFirstFocusableItem();
      }
      else if (actionCode == Keyboard.KA_CURSOR_UP)
      {
         focus = getLastFocusableItem();
      }
      else if (actionCode == Keyboard.KA_END_ERROR || event.keyCode() == Key.VK_ALT)
      {
         leaveMenu();
         event.consume();
      }

      if (focus != null)
      {
         focus.requestFocus(true);
      }
   }

   /**
    * Returns the first focusable child.
    *
    * @return  focusable child.
    */
   protected Widget getFirstFocusableItem()
   {
      Widget firstChild = null;
      for (Widget w : widgets())
      {
         if (w.focusTraversable())
         {
            firstChild = w;
            break;
         }
      }

      return firstChild;
   }

   /**
    * Returns the last focusable child.
    *
    * @return  focusable child.
    */
   protected Widget getLastFocusableItem()
   {
      Widget[] widgets = widgets();
      for (int i = widgets.length - 1; i >= 0; i--)
      {
         Widget w = widgets[i];
         if (w.focusTraversable())
         {
            return w;
         }
      }

      return null;
   }

   /**
    * The abstract portion of the logic making pop-up menu shown.
    * 
    * @param   src
    *          Source widget.
    * @param   location
    *          The popup location.
    * @param   dismissHandler
    *          When not null, the runnable instance is executed when the popup menu is dismissed, regardless
    *          whether an item is chosen or not.
    */
   @Override
   protected void doShowPopup(Widget<GuiOutputManager> src, NativePoint location, Runnable dismissHandler)
   {
      // force the driver to be in control, as this affects UI
      gd.setControl(true);

      OverlayWindow wnd = new OverlayWindow(this, src.topLevelWindow());
      wnd.setDestroyingListener(() ->
      {
         detachFromPopupWindow(wnd);
         if (dismissHandler != null)
         {
            dismissHandler.run();
         }
      });

      // make the menu itself the only focused widget - when popup is displayed it must not show any
      // sub items focused, but it must be itself focused in order to receive (key) events
      parent().setFocus(this);
      clearItemsHighlight();

      setPhysicalLocation(location.x, location.y);
      doLayout();
      setVisible(true);
      wnd.show();
      
      ThinClient.getInstance().eventDrawingBracket(this, this::repaint);
   }
   
   /**
    * Process accelerator event. If some MENUBAR sub-menu is displayed,
    * than event must be ignored.
    * 
    * @param   event
    *          KeyInput object, related to key event.
    * @param   m
    *          Target accelerator source menu-item.
    *          
    * @return  <code>true</code> means accelerator was processed.
    *          <code>false</code> means accelerator was ignored.
    */
   @Override
   protected boolean processAccelerator(KeyInput event, MenuItem<?> m)
   {   
      for (Widget<?> w : widgets())
      {
         if (w instanceof SubMenu && ((SubMenu<?>) w).config().displayed)
         {
            return false;
         }
      }
      
      return super.processAccelerator(event, m);
   }
   
   /**
    * Replace original version to take into account P2J specific behavior.
    * 
    * @param   includeNested
    *          <code>true</code> if we should process nested sub-menus
    *          
    * @return  See above.
    */
   private double width(boolean includeNested)
   {
      double width = !config().popupOnly ? // MENUBAR ? 
                      super.width() : 
                      SubMenuGuiImpl.getMaxSubMenuItemWidth(this, includeNested);
      
      NativeInsets insets = ThemeManager.getCurrentTheme().getMenuInsets();
      CoordinatesConversion c = screen().coordinates();
      return width + c.widthFromNative(insets.left + insets.right);
   }

   /**
    * Detahces this widget from the parent window.
    *
    * @param   wnd
    *          Parent window to detach from.
    */
   private void detachFromPopupWindow(TopLevelWindow wnd)
   {
      wnd.getContentPane().detach(this);
   }
}