/*
** Module   : SliderGuiImpl.java
** Abstract : GUI slider implementation
**
** Copyright (c) 2016-2018, Golden Code Development Corporation.
**
** -#- -I- --Date-- ---------------------------------Description----------------------------------
** 001 EVL 20160127 Initial version.
** 002 EVL 20160217 LOAD-MOUSE-POINTER support.
** 003 SVL 20160719 Updated translate() signature.
** 004 OM  20170320 VectorGraphicsHelper.getScalableImage() signature change.
** 005 EVL 20170824 Adding file list drag-and-drop support.
** 006 EVL 20180709 Adding highlighted image support.
**     OM  20180710 Delegating widget drawing code to Themes.
**     EVL 20180710 Fix for negative slider values support.  Small slider size is valid for
**                  TicMarks.NONE in classic theme only.
**     OM  20180713 Added TicMarks.NOT_SET that uses small slider size for all themes.
**                  Added mouse sensitiveness and wheel events. Fine-tunning mouse handling.
** 007 EVL 20180720 Changed Z-order class level to default value WIDGET.
*/

/*
** This program is free software: you can redistribute it and/or modify
** it under the terms of the GNU Affero General Public License as
** published by the Free Software Foundation, either version 3 of the
** License, or (at your option) any later version.
**
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
** GNU Affero General Public License for more details.
**
** You may find a copy of the GNU Affero GPL version 3 at the following
** location: https://www.gnu.org/licenses/agpl-3.0.en.html
** 
** Additional terms under GNU Affero GPL version 3 section 7:
** 
**   Under Section 7 of the GNU Affero GPL version 3, the following additional
**   terms apply to the works covered under the License.  These additional terms
**   are non-permissive additional terms allowed under Section 7 of the GNU
**   Affero GPL version 3 and may not be removed by you.
** 
**   0. Attribution Requirement.
** 
**     You must preserve all legal notices or author attributions in the covered
**     work or Appropriate Legal Notices displayed by works containing the covered
**     work.  You may not remove from the covered work any author or developer
**     credit already included within the covered work.
** 
**   1. No License To Use Trademarks.
** 
**     This license does not grant any license or rights to use the trademarks
**     Golden Code, FWD, any Golden Code or FWD logo, or any other trademarks
**     of Golden Code Development Corporation. You are not authorized to use the
**     name Golden Code, FWD, or the names of any author or contributor, for
**     publicity purposes without written authorization.
** 
**   2. No Misrepresentation of Affiliation.
** 
**     You may not represent yourself as Golden Code Development Corporation or FWD.
** 
**     You may not represent yourself for publicity purposes as associated with
**     Golden Code Development Corporation, FWD, or any author or contributor to
**     the covered work, without written authorization.
** 
**   3. No Misrepresentation of Source or Origin.
** 
**     You may not represent the covered work as solely your work.  All modified
**     versions of the covered work must be marked in a reasonable way to make it
**     clear that the modified work is not originating from Golden Code Development
**     Corporation or FWD.  All modified versions must contain the notices of
**     attribution required in this license.
*/

package com.goldencode.p2j.ui.client.gui;

// this usage of AWT events is intentional, they are being used separately from the
// AWT/Swing event model itself; no other AWT/Swing usage is occurring in this class
import java.awt.event.*;

import com.goldencode.p2j.ui.*;
import com.goldencode.p2j.ui.client.*;
import com.goldencode.p2j.ui.client.gui.driver.*;
import com.goldencode.p2j.ui.client.gui.theme.*;
import com.goldencode.p2j.ui.client.widget.*;

/**
 * Specific GUI implementation of SLIDER widget.
 */
public class SliderGuiImpl
extends Slider<GuiOutputManager>
{
   /** The size is used to compute different positions for parts. */
   private static final int FIXED_SIZE = 16;
   
   /** The size of the internal slider area for none tic marks defined. */
   private static final int TM_NONE_SIZE = 30;
   
   /** The size of the internal slider area for none tic marks defined. */
   private static final int TM_NOT_SET_SIZE = 18;
   
   /** The size of the internal slider area for both tic marks mode defined. */
   private static final int TM_BOTH_SIZE = 40;
   
   /** The size of the internal slider area for single side tic marks mode defined. */
   private static final int TM_SINGLE_SIZE = 35;
   
   /** Working tic-marks mode. */
   public static enum TicMarks
   {
      /** 
       * ABL programmer has not set the tic-marks mode. The slider is thin because it is assumed
       * that this is migrated OLD ABL code. In all other case, trackbar on the slider widget will
       * be relatively large. Also, the track is drawn 3 pixel longer when in {@code NOT-SET}
       * mode.
       */
      NOT_SET,
      
      /** No tic marks will be drawn. */
      NONE,
      
      /** 
       * The tic marks will be drawn on top or left of the slider, depending on its orientation.
       */
      TOP,
      
      /**
       *  The tic marks will be drawn on top or left of the slider, depending on its orientation.
       */
      BOTTOM,
      
      /**
       * The tic marks will be drawn on bottom or right of the slider, depending on its 
       * orientation.
       */
      LEFT,
      
      /** 
       * The tic marks will be drawn on top or left of the slider, depending on its orientation.
       */
      RIGHT,
      
      /** 
       * The tic marks will be drawn on bottom or right of the slider, depending on its 
       * orientation.
       */
      BOTH
   }
   
   /** Cached tic marks mode in format. */
   private TicMarks tmMode = TicMarks.NOT_SET;
   
   /** Screen driver. */
   private GuiDriver gd;
   
   /** Output manager. */
   private GuiOutputManager screen;
   
   /** Coordinate converter. */
   private CoordinatesConversion cc;
   
   /** Color resolver. */
   private GuiColorResolver gc;
   
   /** Font resolver */
   private GuiFontResolver gf;
      
   /** Mode dependent thumb dimension of the slider. */
   private NativeDimension thumbDim;
   
   /** Thumb X position of the slider. */
   private int thumbX = -1;
   
   /** Thumb Y position of the slider. */
   private int thumbY = -1;
   
   /** The rectangle where the mouse events must be handled. */
   private NativeRectangle mouseArea;
   
   /** Mouse drag indicator. */
   private boolean dragInProgress = false;
   
   /** Flag for mouse hovering the thumb. */
   private boolean mouseHover = false;
   
   /** Flag indicating internals has been completed. */
   private boolean initCompleted = false;
   
   /**
    * This constructor creates a slider.
    * 
    * @param    id
    *           The widget's ID.
    * @param    owner
    *           The parent container, frame which will include this slider.
    */
   SliderGuiImpl(WidgetId id, Container<GuiOutputManager> owner) 
   {
      super(id, owner);
   }

   /**
    * Special constructor used to restore component using given configuration.
    * 
    * @param   id
    *          The widget's ID.
    * @param   cfg
    *          Configuration data.
    */
   public SliderGuiImpl(WidgetId id, SliderConfig cfg)
   {
      super(id, cfg);
   }
   
   /**
    * Replacement API for the logic which should exist in the widget's c'tor, but the logic 
    * can't be executed as it depends on the widget being registered into the official registry.
    * <p>
    * This API is called after the widget is created and added to the registry.
    * 
    * @param    id
    *           The widget's ID.
    * @param    wcfg
    *           The config used to initialize this widget.
    */
   @Override
   public void initialize(WidgetId id, SliderConfig wcfg)
   {
      super.initialize(id, wcfg);
      
      // screen driver
      gd = (GuiDriver) OutputManager.getDriver();
      
      // output manager
      screen = screen();
      cc = screen.coordinates();
      
      // font resolver
      gf = new GuiFontResolver(WindowManager.resolveWindow(this), gd, config, false);
      
      // adding widget to the registry
      screen.getRegistry().addWidget(this);
   }
   
   /**
    * Test whether the mouse hovers the thumb. Useful for mouse hover sensitive themes.
    * 
    * @return  {@code true} if mouse hovers the thumb.
    */
   public boolean isMouseHover()
   {
      return mouseHover;
   }
   
   /**
    * Draw the slider. Called by this rectangle's parent slider.
    */
   @Override
   public void draw()
   {
      // rectangle options need to be refreshed before every painting
      SliderConfig config = config();
      boolean horiz = config.horizontal;
      
      gc = new GuiColorResolver(gd, config, SysColor.COLOR_WINDOW, SysColor.COLOR_WINDOWTEXT);
      
      // determine colors to use
      gc.colors();
   
      // get location and size
      NativePoint p = physicalLocation();
      final NativeDimension d = physicalDimension();
      
      // set up clipping rectangle
      NativeRectangle clip = clipRectangle(p, new NativeDimension(d.width, d.height));
      
      // initialize internal variables
      initInternals();
      
      // clippable drawing start 
      gd.draw(p, clip, () -> {
         Theme theme = ThemeManager.getCurrentTheme();
         
         // fill the background first
         theme.drawSliderBackground(this, gd, gc, 0, 0, d.width, d.height);
         
         // inside this call widget origin is always (0,0)
         int nativeLine = cc.heightToNative(1.0) * 3 / 4;
         // starting point to the area of dotted focused frame 
         NativePoint po = new NativePoint(0, horiz
            ? tmMode == TicMarks.TOP || tmMode == TicMarks.BOTTOM ||
              tmMode == TicMarks.BOTH ? nativeLine : nativeLine * 2 : 0);
         
         // width and height depend on the orientation and tic mark mode
         int w = horiz ? d.width
                       : tmMode == TicMarks.NONE ? TM_NONE_SIZE
                                                 : tmMode == TicMarks.BOTH ? TM_BOTH_SIZE
                                                                           : tmMode == TicMarks.NOT_SET ? TM_NOT_SET_SIZE
                                                                                                        : TM_SINGLE_SIZE;
         
         int h = horiz ? tmMode == TicMarks.NONE ? TM_NONE_SIZE
                                                 : tmMode == TicMarks.BOTH ? TM_BOTH_SIZE
                                                                           : tmMode == TicMarks.NOT_SET ? TM_NOT_SET_SIZE
                                                                                                        : TM_SINGLE_SIZE 
                       : d.height;
         
         NativeDimension dim =  new NativeDimension(w, h);
         
         // compute coordinates and draw the slider components
         // horizontal left edge
         int x = po.x + (horiz ? tmMode == TicMarks.NOT_SET ? FIXED_SIZE / 2 - 2 
                                                            : FIXED_SIZE / 2
                               : tmMode == TicMarks.BOTH ? dim.width / 2 - 1
                                                         : tmMode == TicMarks.LEFT ? 18
                                                                                   : tmMode == TicMarks.RIGHT ? 8
                                                                                                              : tmMode == TicMarks.NONE ? 8
                                                                                                                                        : 6);
         // vertical top edge
         int y = po.y + (horiz ? tmMode == TicMarks.BOTH ? dim.height / 2 - 1
                                                         : tmMode == TicMarks.TOP ? 18
                                                                                  : tmMode == TicMarks.BOTTOM ? 8
                                                                                                              : tmMode == TicMarks.NONE ? 8
                                                                                                                                        : 6
                               : tmMode == TicMarks.NOT_SET ? FIXED_SIZE / 2 - 2
                                                            : FIXED_SIZE / 2);
         // vertical or horizontal length
         int l = horiz ? dim.width - (tmMode == TicMarks.NOT_SET ? FIXED_SIZE - 3 : FIXED_SIZE)
                       : dim.height - (tmMode == TicMarks.NOT_SET ? FIXED_SIZE - 3 : FIXED_SIZE);
         
         // store the mouse sensitive area
         if (mouseArea == null)
         {
            setupMouseEventsArea(x, y, l, horiz);
         }
         
         final boolean DEBUG = false;
         if (DEBUG)
         {
            gd.setColor(java.awt.Color.YELLOW);
            gd.fillRect(mouseArea.left(), mouseArea.top(), mouseArea.width(), mouseArea.height());
         }
         
         // the ruler drawing depends on orientation, simulate 3D effect
         theme.drawSliderTrack(this, gd, gc, x, y, l);
         
         // the processing below needs the coord and length to be modified
         if (horiz)
         {
            x += thumbDim.width / 2;
         }
         else
         {
            y += thumbDim.width / 2;
         }
         // adjust length
         l -= thumbDim.width;
         
         // draw slider tic marks, if tic marks not visible, do nothing
         if (tmMode != TicMarks.NONE && tmMode != TicMarks.NOT_SET)
         {
            theme.drawSliderTicMarks(this, gd, gc, x, y, l);
         }
         
         // calculate pixel based thumb position within the ruler line depending on widget value
         // base is the X or Y coord for the ruler start depending on orientation
         int pos = indexToCoord(horiz ? x : y, l, horiz, getScreenValueAsInt());
         
         // new thumb coordinates
         thumbX = horiz ? pos : x;
         thumbY = horiz ? y : pos;
         
         // draw slider pointer
         theme.drawSliderThumb(this, gd, gc, thumbX, thumbY, thumbDim.width, thumbDim.height); 
         
         // check if the widget is highlighted and enabled
         if (highlighted() && isEnabled())
         {
            theme.drawSliderFrame(this, gd, gc, po.x, po.y, w, h);
         }
         // check if the current value must be visible
         // note: NO-CURRENT-VALUE == TRUE means to display the current value
         if (config.noCurrentValue)
         {
            theme.drawSliderValue(this, gd, gc, gf, po.x, po.y, w, h);
         }
      });
   }

   /**
    * Notification of a mouse pressed event occurred for this widget.  Moves the slider pointer
    * and changes the current screen value according the mouse coordinate.
    * 
    * @param    e
    *           The mouse event.
    */
   @Override
   public void mousePressed(MouseEvent e)
   {
      // we consider only for enabled widget
      if (!isEnabled())
      {
         return;
      }

      synchronized (this)
      {
         NativePoint n = translate(e);
         // check if the mouse is over the pointer
         if (isOverThumb(n.x, n.y))
         {
            dragInProgress = true;
         }
         else
         {
            // check if the mouse coords is inside the value change area and get the suggested
            // value
            int newPos = getNewSliderValue(n.x, n.y);
            // less than minValue means the mouse is not in meaning area, do nothing for this
            if (newPos >= config.minValue && newPos != config.currentValue)
            {
               // change the slider value by mouse
               changeScreenValue((newPos > config.currentValue ? INCREMENT : -INCREMENT));
            }
         }
      }
   }
   
   /**
    * Notification of a mouse released event occurred for this widget.
    * 
    * @param    e
    *           The mouse event.
    */
   @Override
   public void mouseReleased(MouseEvent e)
   {
      // check if the widget is enabled
      if (isEnabled())
      {
         // here we can stop the drag
         synchronized (this)
         {
            dragInProgress = false;
         }
      }
   }

   /**
    * Notification of a mouse dragged event occurred for this widget.
    * 
    * @param    e
    *           The mouse event.
    */
   @Override
   public void mouseDragged(MouseEvent e)
   {
      // check if the widget is enabled
      // we will use only left button to drag the thumb
      int dragEventMask = MouseEvent.BUTTON1_DOWN_MASK;
      if (dragInProgress && isEnabled() && (e.getModifiersEx() & dragEventMask) == dragEventMask)
      {
         synchronized (e)
         {
            NativePoint n = translate(e);
            // get new slider position
            int newPos = getNewSliderValue(n.x, n.y);
            // less than minValue means the mouse is not in meaning area, do nothing for this
            if (newPos >= config.minValue)
            {
               // change the slider value by mouse drag
               changeScreenValue(newPos - config.currentValue);
            }
         }
      }
   }
   
   /**
    * Interprets the mouse wheel events.
    * 
    * @param   e
    *          The mouse wheel event.
    */
   @Override
   public void mouseWheelMoved(MouseWheelEvent e)
   {
      super.mouseWheelMoved(e);
      
      if (!highlighted())
      {
         LOG.warning("Wrong widget (?)");
         return;
      }
   
      int dir = e.getWheelRotation();
      if (!config().horizontal)
      {
         dir = -dir;
      }
      changeScreenValue(config.largeToSmall ? -dir : dir);
   }
   
   /**
    * Interprets the mouse hover events.
    *
    * @param   e
    *          The mouse move event.
    */
   @Override
   public void mouseMoved(MouseEvent e)
   {
      super.mouseMoved(e);
   
      if (!ThemeManager.getCurrentTheme().isMouseHoverSensitive())
      {
         return;
      }
   
      NativePoint np = translate(e);
      boolean overThumb = isOverThumb(np.x, np.y);
      if (overThumb != mouseHover)
      {
         repaintThumb();
         mouseHover = overThumb;
      }
   }
   
   /**
    * Interprets the mouse enter events.
    *
    * @param   e
    *          The mouse enter event.
    */
   @Override
   public void mouseEntered(MouseEvent e)
   {
      super.mouseEntered(e);
      
      if (!ThemeManager.getCurrentTheme().isMouseHoverSensitive())
      {
         return;
      }
   
      NativePoint np = translate(e);
      if (isOverThumb(np.x, np.y))
      {
         mouseHover = true;
         repaintThumb();
      }
   }
   
   /**
    * Interprets the mouse exit events.
    *
    * @param   e
    *          The mouse exit event.
    */
   @Override
   public void mouseExited(MouseEvent e)
   {
      super.mouseExited(e);
      
      if (!ThemeManager.getCurrentTheme().isMouseHoverSensitive())
      {
         return;
      }
      
      if (mouseHover)
      {
         repaintThumb();
      }
      mouseHover = false;
   }
   
   /**
    * Check if the widget supports LOAD-MOUSE-POINTER() method 
    * 
    * @return {@code true} if the widget supports LOAD-MOUSE-POINTER() method
    */
   @Override
   public boolean supportsCustomMousePointer()
   {
      return true;
   }

   /**
    * Checks if the given widget is able to be the target of the drop operation.
    * 
    * @return  {@code true} if this widget can be droppable at this time, {@code false} otherwise.
    */
   @Override
   public boolean isDroppable()
   {
      return config.dropTarget; 
   }
   
   /**
    * Obtain the tic marks' mode.
    * 
    * @return  the tic marks' mode.
    */
   public TicMarks getTicmarkMode()
   {
      return tmMode;
   }
   
   /**
    * Calculate X or Y coordinate for given slider index tha must be in range from slider
    * min value to slider max value.
    *
    * @param    start
    *           The X or Y coordinate to use as the base.
    * @param    length
    *           The length in pixels of the slider ruler.
    * @param    horizontal
    *           The flag indicating if the orientation is horizontal or not.
    * @param    ndx
    *           The value for the slider internal position to transform to coord.
    *
    * @return   The new coordinate, in X or Y dimension for the given slider index value.
    */
   public int indexToCoord(int start, int length, boolean horizontal, int ndx)
   {
      // preparing to precise calculations
      float baseDbl = (float) start;
      float lengthDbl = (float) length;
      float range = (getRange() - 1);
      
      // main proportion considering orientation and LARGE-TO-SMALL option
      float res = horizontal && !config.largeToSmall || !horizontal && config.largeToSmall
            ? baseDbl + lengthDbl * (float)(ndx - config.minValue) / range
            : baseDbl + lengthDbl - lengthDbl * (float)(ndx - config.minValue) / range;
      
      // perform honest rounding to get the result
      return Math.round(res);
   }
   
   /**
    * Determine the mouse actions processed by this widget.
    * 
    * @return   See above.
    */
   @Override
   protected int[] mouseActions()
   {
      return new int[]
      {
         MouseEvent.MOUSE_CLICKED, 
         MouseEvent.MOUSE_MOVED, 
         MouseEvent.MOUSE_WHEEL, 
         MouseEvent.MOUSE_PRESSED, 
         MouseEvent.MOUSE_RELEASED,
         MouseEvent.MOUSE_ENTERED,
         MouseEvent.MOUSE_EXITED,
         MouseEvent.MOUSE_DRAGGED
      };
   }
   
   /**
    * Check if the given X,Y coordinates fall into value changed area and calculate the new
    * suggested one.
    * 
    * @param    x
    *           The X coordinate to check.
    * @param    y
    *           The Y coordinate to check.
    * 
    * @return   The new suggested value for slider or value less than min if the coordinate is
    *           not in area of changes.
    */
   private int getNewSliderValue(int x, int y)
   {
      int result = config.minValue - 1;
      // check if the mouse inside work area
      if (mouseArea != null && mouseArea.contains(x, y))
      {
         // take into account LARGE-TO-SMALL attribute too
         int newPx = config.horizontal
                        ? config.largeToSmall ? mouseArea.right() - x : x - mouseArea.left()
                        : config.largeToSmall ? y - mouseArea.top() : mouseArea.bottom() - y;
         newPx -= thumbDim.width / 2;
         // area length depending on orientation
         int size = (config.horizontal ? mouseArea.width() : mouseArea.height()) - thumbDim.width;
         // new slider screen value
         result = Math.round(config.minValue + newPx * (getRange() - 1.0F) / size);
      }
      
      return result;
   }
   
   /**
    * Defines the area the mouse events should be handled.
    * 
    * @param    xbase
    *           The X coordinate of the main ruler.
    * @param    ybase
    *           The Y coordinate of the main ruler.
    * @param    lbase
    *           The length of the main ruler.
    * @param    horiz
    *           The flag indicating if the orientation is horizontal or not.
    */
   private void setupMouseEventsArea(int xbase, int ybase, int lbase, boolean horiz)
   {
      int left   = horiz ? xbase : xbase - thumbDim.height / 2;
      int top    = horiz ? ybase - thumbDim.height / 2 : ybase;
      int right  = horiz ? xbase + lbase : xbase + thumbDim.height / 2 + 2;
      int bottom = horiz ? ybase + thumbDim.height / 2 + 2: ybase + lbase;
      
      mouseArea = new NativeRectangle(top, left, bottom, right);
   }
   
   /**
    * Fires a repaint event only for the area of the thumb. 
    */
   private void repaintThumb()
   {
      boolean horiz = config().horizontal;
      
      NativePoint p = new NativePoint(
            thumbX - (horiz ? thumbDim.width / 2 : thumbDim.height / 2),
            thumbY - (horiz ? thumbDim.height / 2 : thumbDim.width / 2));
      
      Dimension size = cc.dimensionFromNative(
            horiz ? thumbDim : new NativeDimension(thumbDim.height, thumbDim.width));
      repaint(new Rectangle(cc.pointFromNative(p).translate(location()),
                            size, cc.baseUnits()));
   }
   
   /**
    * Check if the given X,Y coordinates fall into slider pointer rectangle.
    * 
    * @param    x
    *           The X coordinate to check.
    * @param    y
    *           The Y coordinate to check.
    * 
    * @return   {@code true} if X,Y is over pointer, {@code false} otherwise.
    */
   private boolean isOverThumb(int x, int y)
   {
      boolean horiz = config().horizontal;
      
      // check the pointer area
      int left   = horiz ? thumbX - thumbDim.width / 2  : thumbX - thumbDim.height / 2;
      int top    = horiz ? thumbY - thumbDim.height / 2 : thumbY - thumbDim.width / 2;
      int right  = horiz ? thumbX + thumbDim.width / 2  : thumbX + thumbDim.height / 2;
      int bottom = horiz ? thumbY + thumbDim.height / 2 : thumbY + thumbDim.width / 2;
      
      // true if inside this
      return x >= left && x <= right && y >= top && y <= bottom;
   }
   
   /**
    * Initialize internal variables.
    */
   private void initInternals()
   {
      // already done, just leave
      if (initCompleted)
      {
         return;
      }
      
      // setting up the current tic marks mode
      switch (config.ticMarks)
      {
         case SliderConfig.TM_NONE:
            tmMode = TicMarks.NONE;
            break;
         case SliderConfig.TM_TOP:
            tmMode = TicMarks.TOP;
            break;
         case SliderConfig.TM_BOTTOM:
            tmMode = TicMarks.BOTTOM;
            break;
         case SliderConfig.TM_LEFT:
            tmMode = TicMarks.LEFT;
            break;
         case SliderConfig.TM_RIGHT:
            tmMode = TicMarks.RIGHT;
            break;
         case SliderConfig.TM_BOTH:
            tmMode = TicMarks.BOTH;
            break;
      }
      
      // setting up the thumb dimension depending on the current mode no tic marks available
      if (tmMode == TicMarks.NOT_SET)
      {
         // if tic mark mode is not set, ABL uses a small thumb.
         thumbDim = new NativeDimension(8, 12);
      }
      else
      {
         // thumb size when tic marks are set by the ABL programmer
         thumbDim = new NativeDimension(11, 22);
      }
      
      // to avoid double init when already done
      initCompleted = true;
   }
}
