/*
** Module   : VectorGraphicsHelper.java
** Abstract : Image helper methods to handle SVG vector graphics formats.
**
** Copyright (c) 2015-2018, Golden Code Development Corporation.
**
** -#- -I- --Date-- ---------------------------------Description----------------------------------
** 001 EVL 20150908 Created initial version.  Adding methods to work with SVG images embedded into
**                  p2j.jar.  Newly created images are storing in cache map for further re-usage.
** 002 HC  20160428 Introduced new method getScalableImage() to allow turning antialiasing
**                  on/off.
** 003 HC  20160502 Javadoc fix.
** 004 OM  20170320 Removed dependency on GuiDriver.
** 005 OM  20180503 Used Theme.getBinaryResourceLocation() instead of hardcoded "classic" location
**                  for image resources.
*/

/*
** This program is free software: you can redistribute it and/or modify
** it under the terms of the GNU Affero General Public License as
** published by the Free Software Foundation, either version 3 of the
** License, or (at your option) any later version.
**
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
** GNU Affero General Public License for more details.
**
** You may find a copy of the GNU Affero GPL version 3 at the following
** location: https://www.gnu.org/licenses/agpl-3.0.en.html
** 
** Additional terms under GNU Affero GPL version 3 section 7:
** 
**   Under Section 7 of the GNU Affero GPL version 3, the following additional
**   terms apply to the works covered under the License.  These additional terms
**   are non-permissive additional terms allowed under Section 7 of the GNU
**   Affero GPL version 3 and may not be removed by you.
** 
**   0. Attribution Requirement.
** 
**     You must preserve all legal notices or author attributions in the covered
**     work or Appropriate Legal Notices displayed by works containing the covered
**     work.  You may not remove from the covered work any author or developer
**     credit already included within the covered work.
** 
**   1. No License To Use Trademarks.
** 
**     This license does not grant any license or rights to use the trademarks
**     Golden Code, FWD, any Golden Code or FWD logo, or any other trademarks
**     of Golden Code Development Corporation. You are not authorized to use the
**     name Golden Code, FWD, or the names of any author or contributor, for
**     publicity purposes without written authorization.
** 
**   2. No Misrepresentation of Affiliation.
** 
**     You may not represent yourself as Golden Code Development Corporation or FWD.
** 
**     You may not represent yourself for publicity purposes as associated with
**     Golden Code Development Corporation, FWD, or any author or contributor to
**     the covered work, without written authorization.
** 
**   3. No Misrepresentation of Source or Origin.
** 
**     You may not represent the covered work as solely your work.  All modified
**     versions of the covered work must be marked in a reasonable way to make it
**     clear that the modified work is not originating from Golden Code Development
**     Corporation or FWD.  All modified versions must contain the notices of
**     attribution required in this license.
*/

package com.goldencode.p2j.ui.client.gui;

import java.io.*;
import java.util.*;

import com.goldencode.p2j.ui.client.gui.theme.*;
import org.apache.batik.transcoder.*;
import org.apache.batik.transcoder.image.PNGTranscoder;

import com.goldencode.p2j.ui.client.gui.driver.*;

/**
 * Image helper methods to handle pending image loading.
 */
public class VectorGraphicsHelper
{
   /** Map to store cached images to simplify the scaling image usage. */
   private static Map<String, ImageWrapper> svgToRasterImgCache = new HashMap<>();
   
   /**
    * Gets the new image data byte array for prepackaged scalable images.
    * No antialiasing is applied to the resulting image.
    * 
    * @param    name
    *           The name for image to be loaded from p2j.jar.
    * @param    width
    *           The required image width.
    * @param    height
    *           The required image height.
    *
    * @return   The byte array data for the image to be loaded or null if there is no data.
    */
   public static ImageWrapper getScalableImage(String name, int width, int height)
   {
      // the location is not specified: use the theme specific root
      return getScalableImage(name,
                              ThemeManager.getCurrentTheme().getBinaryResourceLocation(),
                              width,
                              height);
   }
   
   /**
    * Gets the new image data byte array for prepackaged scalable images.
    * No antialiasing is applied to the resulting image.
    * 
    * @param    name
    *           The name for image to be loaded from p2j.jar.
    * @param    location
    *           The theme specific location for image to be loaded.
    * @param    width
    *           The required image width.
    * @param    height
    *           The required image height.
    *
    * @return   The byte array data for the image to be loaded or null if there is no data.
    */
   public static ImageWrapper getScalableImage(String name, String location, int width, int height)
   {
      return getScalableImage(name, location, width, height, false);
   }
   
   /**
    * Gets the new image data byte array for prepackaged scalable images.
    * 
    * @param    name
    *           The name for image to be loaded from p2j.jar.
    * @param    location
    *           The theme specific location for image to be loaded.
    * @param    width
    *           The required image width.
    * @param    height
    *           The required image height.
    * @param    antialias
    *           Set {@code false} to skip antialias processing the output
    *           image, {@code true} otherwise.
    *
    * @return   The byte array data for the image to be loaded or null if there is no data.
    */
   public static ImageWrapper getScalableImage(String name,
                                               String location,
                                               int width,
                                               int height,
                                               boolean antialias)
   {
      ImageWrapper imgRes = null;
      
      // full image name
      StringBuilder sb = new StringBuilder("theme/");
      // append the current theme name
      sb.append(location);
      sb.append("/");
      // append the requested name
      sb.append(name);
      // and extension
      sb.append(".svg");
      // construct the full name
      String fullName = sb.toString();
      
      // continue making the cache map name
      sb.append("-");
      // add resolution as unique identity
      sb.append(width);
      sb.append("x");
      sb.append(height);
      // construct the name to put into cache
      String inCacheName = sb.toString();
      
      // first try to search the image in cache
      imgRes = svgToRasterImgCache.get(inCacheName);
      if (imgRes != null)
      {
         // found in cache no need to create it again
         return imgRes;
      }
      
      // loading image embedded into p2j.jar
      InputStream is = VectorGraphicsHelper.class.getResourceAsStream(fullName);
      // pre-allocale 4K byte buffer for conversion result
      ByteArrayOutputStream os = new ByteArrayOutputStream(4096);
      
      // got input stream from jar, start converting to raster image
      if (is != null)
      {
         // create transcoder
         PNGTranscoder t = antialias ? new PNGTranscoder() : new SharpPNGTranscoder();
         // set up conversion options
         t.addTranscodingHint(PNGTranscoder.KEY_WIDTH, new Float(width));
         t.addTranscodingHint(PNGTranscoder.KEY_HEIGHT, new Float(height));
         // setting transcoder input
         TranscoderInput ti = new TranscoderInput(is);
         // create transcoder output
         TranscoderOutput to = new TranscoderOutput(os);
         
         // convert SVG to raster image data according to transcoder mime type
         try
         {
            t.transcode(ti, to);
         }
         catch (TranscoderException te)
         {
            // something happen while converting
            os = null;
         }
         
         // check if conversion completed successfully
         if (os != null)
         {
            // get the image data from byte array
            byte [] imgData = os.toByteArray();
            // and create new image
            if (imgData != null)
            {
               imgRes = AbstractGuiDriver.createImage(imgData);
            }
         }
         
         // put the created image into cache for further re-usage
         if (imgRes != null)
         {
            svgToRasterImgCache.put(inCacheName, imgRes);
         }
         
         // close opened streams
         try
         {
            is.close();
         }
         catch (IOException e)
         {
            // ignore
         }
      }
      
      return imgRes;
   }
}
