/*
** Module   : WindowLayout.java
** Abstract : Layout the elements of a window.
**
** Copyright (c) 2014-2018, Golden Code Development Corporation.
**
** -#- -I- --Date-- ---------------------------------Description----------------------------------
** 001 CA  20140805 Created initial version.
** 002 MAG 20140912 Set a new BitmapScreen object when window is resized.
** 003 HC  20140911 GUI coordinate support - data types refactoring.
** 004 CA  20140926 Added shared widget configuration support, to allow GUI/ChUI concrete
**                  implementation for the same widget ID.  Refactored the widget configuration
**                  classes: all fields were made public; these classes need to be as dumb as
**                  possible, all logic related to setting/getting a certain field should be at
**                  the widget or at the caller. Refs #2254
** 005 MAG 20140923 GUI message area layout.
** 006 CA  20141021 Fixes related to window layout during resize.
** 007 MAG 20141021 Clear window borders whenever is resized.
** 008 HC  20141021 Implemented window move/position support.
** 009 CA  20141119 Refactored drawing code to enable multiple OS windows, for GUI support.  Misc
**                  changes/fixes related to drawing on multiple windows.
** 010 CA  20141123 Fixed initial dimension for dynamic windows.
** 011 HC  20141207 Improved window layout algorithm to honor the requested window size.
** 012 HC  20150323 Removed business logic from config classes.
** 013 CA  20150507 Fixed problems related to window sizing (explicit or implicit).
** 014 CA  20150508 Fixed H013 regression when resizing window via mouse (not attributes).
** 015 HC  20150508 Improvements to GUI window scrolling support.
** 016 GES 20150425 Match class name change for GuiOptions to GuiConstants.
** 017 HC  20150612 Implemented support for modal windows.
** 018 HC  20150620 Fixed layout not adding the status area size to the overall window size.
** 019 HC  20150803 Improved runtime support for WINDOW size attributes.
** 020 VIG 20150618 Added menubar layouting. Fixed workspace height calc-n using menubar height.
** 021 VIG 20150813 Some cleanup.
** 022 CA  20150822 Fixed drawing of window border.
** 023 CA  20150929 Fixed multi-window drawing issues caused by concurrent drawing access to the
**                  GuiDriver: all access to the driver is done exclusively, each thread taking
**                  and releasing ownership for the duration of the drawing operation or other
**                  GUI API invocation which requires access to the underlying physical window.
** 024 CA  20151015 Fixed window resize via attributes (they target the workspace, not the full 
**                  window).
** 025 HC  20151027 Implemented WINDOW MESSAGE-AREA and STATUS-AREA attributes.
**     CA  20151029 If workspace is resized, ensure the new window dimension includes the 
**                  decorations and the border.
** 026 VIG 20151122 Added isVisible check for menubar layout.
** 027 HC  20160304 Fixed various window sizing/positioning issues. Window layout improvements.
** 028 SBI 20160321 Changed to hide a window title bar and a border iff the graphics environment
**                  doesn't support window decorations.
** 029 VIG 20151229 Added tooltips initialization for caption buttons.
** 030 CA  20160419 Fixed a NPE when the client is running with non-decorated window (the titlebar
**                  is missing).
** 031 HC  20160428 GUI window border drawing changed to use the Border interface mechanism. Also
**                  removed hardcoded assumptions of the border size.
** 032 EVL 20160623 Sometimes for dialog box we have to ignore adding extra size to compute the
**                  final dimension, when dialog size refers to outer size.
** 033 SBI 20160909 Fixed erroneous substitutions ignoreExtraHeight <-> ignoreExtraWidth.
** 034 SBI 20161006 Fixed the message area width to layout it after window.setWorkspaceNativeSize.
** 035 EVL 20161013 Fix for dialog window title bar width in several conditions.
** 036 CA  20170104 Some improvements in window decoration support for embedded mode (modal 
**                  windows always have decorations).
** 037 HC  20170823 Added runtime support of WINDOW:SHOW-IN-TASKBAR. Implemented runtime support
**                  for WINDOW:ALWAYS-ON-TOP and SHOW-IN-TASKBAR.
** 038 CA  20171019 Javadoc fix.
** 039 EVL 20180411 Fix for window regular layout with menubar defined.  Must be taken into
**                  account in all modes.
** 040 EVL 20180712 Fix for dialog box border painting and size issue.  There are the cases for
**                  dialogs when we need to bypass title height adding.
** 041 HC  20180813 A quick workaround for an issue that prevents window resizing to smaller
**                  sizes.
*/
/*
** This program is free software: you can redistribute it and/or modify
** it under the terms of the GNU Affero General Public License as
** published by the Free Software Foundation, either version 3 of the
** License, or (at your option) any later version.
**
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
** GNU Affero General Public License for more details.
**
** You may find a copy of the GNU Affero GPL version 3 at the following
** location: https://www.gnu.org/licenses/agpl-3.0.en.html
** 
** Additional terms under GNU Affero GPL version 3 section 7:
** 
**   Under Section 7 of the GNU Affero GPL version 3, the following additional
**   terms apply to the works covered under the License.  These additional terms
**   are non-permissive additional terms allowed under Section 7 of the GNU
**   Affero GPL version 3 and may not be removed by you.
** 
**   0. Attribution Requirement.
** 
**     You must preserve all legal notices or author attributions in the covered
**     work or Appropriate Legal Notices displayed by works containing the covered
**     work.  You may not remove from the covered work any author or developer
**     credit already included within the covered work.
** 
**   1. No License To Use Trademarks.
** 
**     This license does not grant any license or rights to use the trademarks
**     Golden Code, FWD, any Golden Code or FWD logo, or any other trademarks
**     of Golden Code Development Corporation. You are not authorized to use the
**     name Golden Code, FWD, or the names of any author or contributor, for
**     publicity purposes without written authorization.
** 
**   2. No Misrepresentation of Affiliation.
** 
**     You may not represent yourself as Golden Code Development Corporation or FWD.
** 
**     You may not represent yourself for publicity purposes as associated with
**     Golden Code Development Corporation, FWD, or any author or contributor to
**     the covered work, without written authorization.
** 
**   3. No Misrepresentation of Source or Origin.
** 
**     You may not represent the covered work as solely your work.  All modified
**     versions of the covered work must be marked in a reasonable way to make it
**     clear that the modified work is not originating from Golden Code Development
**     Corporation or FWD.  All modified versions must contain the notices of
**     attribution required in this license.
*/

package com.goldencode.p2j.ui.client.gui;

import com.goldencode.p2j.ui.*;
import com.goldencode.p2j.ui.client.*;
import com.goldencode.p2j.ui.client.border.*;
import com.goldencode.p2j.ui.client.gui.driver.*;
import com.goldencode.p2j.ui.client.layout.*;
import com.goldencode.p2j.ui.client.widget.*;

import java.util.*;
import java.util.stream.*;

/**
 * Layout the elements of a legacy-style window. This includes the window title, the workspace,
 * message area and status area.
 */
public class WindowLayout
implements LayoutManager<GuiOutputManager>
{
   /** Minimal window width. */
   public static final int MIN_WIDTH_PIXELS = 104;

   /** Minimal window height. */
   public static final int MIN_HEIGHT_PIXELS = 100;

   public static final int MIN_SLOT_WIDTH = 160;

   public static final int MIN_SLOTS_TOTAL = 101;

   /** Last desktop minimized location */
   private static NativePoint lastDesktopMinimizedLocation;

   /** Cached GUI driver. */
   private GuiDriver<?,?> gd;
   
   /** Maximum window size. */
   private NativeDimension maxSize;
   
   /** Flag indicating if the window has decorations. */
   private boolean decorated = true;
   
   /**
    * Sets maximum window size.
    * 
    * @param    width
    *           Maximum window width.
    * @param    height
    *           Maximum window height.
    */
   public void setMaxSize(int width, int height)
   {
      maxSize = new NativeDimension(width, height);
   }
   
   /**
    * Get the minimum size of the specified container.
    * 
    * @param    container
    *           The container.
    *          
    * @return   Always {@code null}.
    */
   @Override
   public Dimension minimumSize(Container<GuiOutputManager> container)
   {
      return null;
   }

   /**
    * Set the {@link #decorated} flag.
    * 
    * @param    decorated
    *           The window's decorated state.
    */
   public void setDecorated(boolean decorated)
   {
      this.decorated = decorated;
   }
   
   /**
    * Place components in specified container.
    * 
    * @param   container
    *          Container to process - must be a window.
    */
   @Override
   public void doLayout(Container<GuiOutputManager> container)
   {
      GuiWindow window = (GuiWindow) container;

      if (window.isDesktopMinimized())
      {
         layoutMinimized((WindowGuiImpl) window);
      }
      else
      {
         layoutRegular(window);
      }
   }

   /**
    * Layouts a top-level gui window. This performs the layout when the window is in its
    * {@linkplain CommonWindow#WINDOW_STATE_NORMAL} or
    * {@linkplain CommonWindow#WINDOW_STATE_MAXIMIZED} states.
    *
    * @param     window
    *            The window to layout.
    */
   private void layoutRegular(GuiWindow window)
   {
      // TODO: only use pixel/native coordinates, syncing of character values
      //       must be handled by the framework
      
      gd = (GuiDriver<?, ?>) GuiOutputManager.getDriver();

      CoordinatesConversion c = window.coordinates();

      // select renderer
      gd.selectWindow(window.getId().asInt());

      // 0. prerequisites for computing full width/height:
      boolean resizing = window.isResizing();
      
      WindowConfig wcfg = window.config();
      
      // - the menu bar height
      int mbarHeight = 0;
         
      if (window instanceof Window)
      {
         Menu<?> menubar = ((Window<?>) window).getMenubar();
         if (menubar != null && menubar.isVisible())
         {
            menubar.doLayout();
            mbarHeight = c.heightToNative(menubar.height());
         }
      }

      // - the message area height
      int mheight = 0;
      MessageAreaGuiImpl marea = window.getGuiMessageArea();
      
      if (marea != null && marea.isVisible() && wcfg.messageArea)
      {
         mheight = c.heightToNative(marea.calculateMessageAreaHeight());
      }
      
      // - the status area height
      int sheight = 0;
      StatusLine<GuiOutputManager> sarea = window.getStatusArea();
      
      if (sarea != null && sarea.isVisible() && wcfg.statusArea)
      {
         sheight = c.heightToNative(sarea.height());
      }

      // - the border delta
      int borderDeltaWidth = 0;
      int borderDeltaHeight = 0;
      if (decorated)
      {
         Border<?> b = window.getBorder();
         if (b != null)
         {
            NativeInsets ins = window.screen().coordinates().insetsToNative(b.insets());
            borderDeltaWidth = ins.left + ins.right;
            borderDeltaHeight = ins.top + ins.bottom;
         }
      }
      
      // - window size
      int    wheight;
      int    wwidth;
      double cwheight;
      double cwwidth;
      
      // - workspace size
      NativeDimension workSize = window.getWorkspaceNativeSize();
      
      WindowConfig wc = window.config();
      WindowTitleBar tbar = window.getTitleBar();
      
      if (resizing)
      {
         NativeDimension oldSize     = window.physicalDimension();
         NativeDimension resizedSize = window.getResizedSize();
         
         // - get the new workspace height/width
         if (window.isResizeWorkspace())
         {
            workSize.width = resizedSize.width;
            workSize.height = resizedSize.height;
         }
         else
         {
            workSize.width  = workSize.width + (resizedSize.width - oldSize.width);
            workSize.height = workSize.height + (resizedSize.height - oldSize.height);
         }
         
         int minmaxWidthDelta = 0;
         int minmaxHeightDelta = 0;
         
         int maxWidthPixels = ConfigHelper.getMaxWidthPixels(wc);
         int maxHeightPixels = ConfigHelper.getMaxHeightPixels(wc);

         // TODO: the actual size was assigned here and so the window was unable to resize down
         //int minWidthPixels = ConfigHelper.getMinWidthPixels(wc);
         //int minHeightPixels = ConfigHelper.getMinHeightPixels(wc);
         int minWidthPixels = 0;
         int minHeightPixels = 0;
         
         // limit the new width by the config min/max width
         if (workSize.width < minWidthPixels)
         {
            minmaxWidthDelta = minWidthPixels - workSize.width;
            workSize.width = minWidthPixels;
         }
         else if (workSize.width > maxWidthPixels)
         {
            minmaxWidthDelta = maxWidthPixels - workSize.width;
            workSize.width = maxWidthPixels;
         }
         
         // limit the new height by the config min height
         if (workSize.height < minHeightPixels)
         {
            minmaxHeightDelta = minHeightPixels - workSize.height;
            workSize.height = minHeightPixels;
         }
         else if (workSize.height > maxHeightPixels)
         {
            minmaxHeightDelta = maxHeightPixels - workSize.height;
            workSize.height = maxHeightPixels;
         }
         
         if (window.isResizeWorkspace())
         {
            // if workspace is resized, ensure the new window dimension includes the decorations
            // and the border
            minmaxHeightDelta += mbarHeight + sheight + mheight + borderDeltaHeight;
            if (tbar != null)
            {
               minmaxHeightDelta += c.heightToNative(tbar.height());
            }
            minmaxWidthDelta += borderDeltaWidth;
         }
         
         wheight  = resizedSize.height + minmaxHeightDelta;
         wwidth   = resizedSize.width + minmaxWidthDelta;
         if (tbar != null)
         {
            // - set the titlebar's width
            double ctbwidth = c.widthFromNative(wwidth - borderDeltaWidth);
            tbar.setWidth(ctbwidth);
            tbar.doLayout();
         }
      }
      else
      {
         // - layout the titlebar now, is needed to compute the window height
         double ctbwidth = c.widthFromNative(workSize.width);
         int tbheight;
         if (tbar != null)
         {
            tbar.setWidth(ctbwidth);
            tbar.doLayout();
            tbheight = c.heightToNative(tbar.height());
         }
         else
         {
            tbheight = 0;
         }

         boolean ignoreExtraWidth = false;
         boolean ignoreExtraHeight = false;
         boolean isDialogBox = window instanceof DialogBoxWindow;
         
         if (isDialogBox)
         {
            FrameGuiImpl frame = ((DialogBoxWindow) window).getFrame();
            FrameConfig fcfg = frame.config();
            ignoreExtraWidth = fcfg.clientWidthChars != BaseConfig.INV_COORD && fcfg.fixedWidth;
            ignoreExtraHeight = fcfg.clientHeightChars != BaseConfig.INV_COORD && fcfg.fixedHeight;
         }
         
         wheight  = workSize.height;
         if (!ignoreExtraHeight)
         {
            // dialog box already has the titlebar added
            wheight += mheight + sheight + mbarHeight + borderDeltaHeight;
            if (!isDialogBox)
            {
               wheight += tbheight;
            }
         }
         wwidth   = workSize.width;
         if (!ignoreExtraWidth)
         {
            wwidth += borderDeltaWidth;
         }
         
         if (maxSize != null)
         {
            // adjust to maximum window size if needed
            
            int deltaW = Math.max(0, wwidth - maxSize.width);
            int deltaH = Math.max(0, wheight - maxSize.height);
            
            wwidth =- deltaW;
            wheight =- deltaH;
            
            workSize.width =- deltaW;
            workSize.height =- deltaH;
         }
         
         if (tbar != null)
         {
            // - set the titlebar's width
            tbar.setWidth(c.widthFromNative(workSize.width -
                                            (ignoreExtraWidth ? borderDeltaWidth : 0)));
            tbar.doLayout();
         }
      }
      
      NativeDimension minSize = window.getMinSize();
      
      // handle window size limits
      if (workSize.width < minSize.width)
      {
         wwidth += minSize.width - workSize.width;
         workSize.width += minSize.width - workSize.width;
      }
      
      if (workSize.height < minSize.height)
      {
         wheight += minSize.height - workSize.height;
         workSize.height += minSize.height - workSize.height;
      }
      
      cwheight = c.heightFromNative(wheight);
      cwwidth  = c.widthFromNative(wwidth); 
      
      window.setWorkspaceNativeSize(workSize.width, workSize.height);
      
      // TODO: in addition to MIN/MAX-HEIGHT attrs the window seems to have
      //       physical size limits - for width it is a minimum window title bar plus borders
      //       and for height it is the title bar height plus message area plus window status bar
      //       plus borders.
      
      if (!resizing)
      {
         gd.resizeWindow(wwidth, wheight);
      }

      window.setSize(cwwidth, cwheight);
      
      // update screen bitmap
      GuiOutputManager.instance().setBitmap(window.getId().asInt(), new ScreenBitmap(wheight, wwidth));
      
      // all coordinates for the following widgets are relative to the OS window (the container)
      int y = 0;
      
      int tbheight;
      if (tbar != null)
      {
         tbheight = c.heightToNative(tbar.height());
         // 1. set the title into position:
         tbar.setPhysicalLocation(0, y);
      }
      else
      {
         tbheight = 0;
      }
      
      // 2. set the window workarea (this is the legacy 4GL window, where frames are placed)
      // - bellow the title and above the message area
      // - its size does not include the borders
      // - its size needs to be computed first, as the general window size depends on it
      y = y + tbheight;
      
      if (window instanceof WindowGuiImpl)
      {
         Menu<GuiOutputManager> menubar = ((WindowGuiImpl)window).getMenubar();
         if (menubar != null && menubar.isVisible())
         {
            MenuConfig mc = menubar.config();
            menubar.setPhysicalLocation(0, y);
            menubar.setWidth(cwwidth);
            mc.widthChars = cwwidth;
            mc.widthPixels = wwidth;
            menubar.doLayout();
            y = y + c.heightToNative(menubar.height());
         }
      }
      
      window.setWorkspaceLocation(0, y);
      
      // the y position for the next UI element
      y = y + workSize.height;
      
      // 3. set the message area
      if (marea != null)
      {
         // message area layout is based on the workspace width
         marea.doLayout();
         marea.setPhysicalLocation(0, y);
      }
      
      // 4. set the status area
      if (marea != null)
      {
         y = y + mheight;
      }
      
      if (sarea != null)
      {
         window.getStatusArea().setPhysicalLocation(0, y);
      }
      
      if (tbar != null)
      {
         tbar.activateTooltips();
      }

      gd.releaseWindow();
   }

   /**
    * Layouts a desktop-minimized top-level window. A window is desktop-minimized when its
    * attribute SHOW-IN-TASKBAR is set to false and it is minimized using the minimized title bar
    * button. In this state the window is not displayed in the task bar (note there is a Swing GUI
    * limitation, see TODO) and so must stay on the desktop in a minimized form - only title bar
    * visible resized to a minimal width.
    *
    * @param    window
    *           The window to layout.
    */
   private void layoutMinimized(WindowGuiImpl window)
   {
      gd = (GuiDriver<?, ?>) GuiOutputManager.getDriver();

      CoordinatesConversion cc = window.coordinates();

      // select renderer
      gd.selectWindow(window.getId().asInt());
      WindowTitleBar tbar = window.getTitleBar();

      int decorationsWidth = 0;
      int decorationsHeight = 0;

      Border<?> b = window.getBorder();
      if (b != null)
      {
         NativeInsets ins = window.screen().coordinates().insetsToNative(b.insets());
         decorationsWidth = ins.left + ins.right;
         decorationsHeight = ins.top + ins.bottom;
      }

      // make the window width equal to the minimized-window slot width
      double ctbwidth = cc.widthFromNative(MIN_SLOT_WIDTH - decorationsWidth);

      tbar.setWidth(ctbwidth);
      tbar.doLayout();
      int tbheight = cc.heightToNative(tbar.height());

      window.setSize(cc.widthFromNative(MIN_SLOT_WIDTH),
                     cc.heightFromNative(tbheight + decorationsHeight));
      window.setWorkspaceNativeSize(MIN_SLOT_WIDTH - decorationsWidth, 0);

      gd.resizeWindow(MIN_SLOT_WIDTH, tbheight + decorationsHeight);
      setMinimizedLocation(window);
      gd.releaseWindow();

   }

   /**
    * The method calculates a default window desktop-minimized location based on the location
    * of previously desktop-minimized window. Desktop-minimized windows (in 4GL
    * WINDOW:SHOW-IN-TASKBAR == false) "dock" in the bottom left corner of the active screen work
    * area. The minimized windows are docked next to each other from left to right and from bottom
    * up.
    */
   private void setMinimizedLocation(WindowGuiImpl wnd)
   {
      // find bounds of all the desktop-minimized windows
      Set<NativeRectangle> rects = WindowManager.windowList().stream()
         .filter(w -> w != wnd && w instanceof WindowGuiImpl &&
                      ((WindowGuiImpl) w).isDesktopMinimized())
         .map(w -> new NativeRectangle(w.physicalLocation(), w.physicalDimension()))
         .collect(Collectors.toSet());

      NativeDimension wndDim = wnd.physicalDimension();
      int slotsPerRow = getSlotsPerRow(wnd);
      NativeRectangle display = gd.getDisplayWorkArea(wnd.getId().asInt());

      // iterate all the desktop-minimize slots and test each whether it is being occupied by
      // any desktop-minimized window, the first slot from left to right bottom up the screen
      // will be used as the minimized window location
      for (int i = 0; i < MIN_SLOTS_TOTAL; i++)
      {
         int row = i / slotsPerRow + 1;
         int xpos = MIN_SLOT_WIDTH * (i % slotsPerRow);

         int slotx = display.left() + xpos;
         int sloty = display.bottom() - wndDim.height * row;
         NativeRectangle testRect = new NativeRectangle(new NativePoint(slotx, sloty), wndDim);

         boolean slotFree = true;
         for (NativeRectangle r : rects)
         {
            if (r.intersects(testRect))
            {
               slotFree = false;
               break;
            }
         }

         // the last slot is always used if no free slot was found
         if (slotFree || i == MIN_SLOTS_TOTAL - 1)
         {
            wnd.setPhysicalLocation(slotx, sloty);
            break;
         }
      }
   }

   /**
    * Calculates the number of "slots" that fit the width of the screen. A slot is an area that
    * fits a window that is minimized and not shown on the taskbar. The supplied window is used
    * as a hint to get the correct display work area on multi-screen systems.
    *
    * @param   wnd
    *          A window reference.
    *
    * @return  number of slots
    */
   private int getSlotsPerRow(WindowGuiImpl wnd)
   {
      NativeRectangle display = gd.getDisplayWorkArea(wnd.getId().asInt());
      int dispWidth = display.width();
      return dispWidth / MIN_SLOT_WIDTH;
   }
}