/*
** Module   : MouseHandler.java
** Abstract : Provides generic mouse event support for the GUI driver.
**
** Copyright (c) 2014-2018, Golden Code Development Corporation.
**
** -#- -I- --Date-- ---------------------------------Description----------------------------------
** 001 CA  20140805 Created initial version.
** 002 CA  20141021 Improved support for mouse events: fixed cases when the widgets need to be 
**                  notified the mouse has exited/entered it and cases when a widget needs to be
**                  notified of any mouse event (i.e. to hide a popup menu).
** 003 CA  20141119 Refactored drawing code to enable multiple OS windows, for GUI support.  Misc
**                  changes/fixes related to drawing on multiple windows.
** 004 EVL 20141222 Adding methods to extract tooltip handlers from action map.
** 005 EVL 20150106 Fixed javadoc error.
** 006 HC  20150323 Removed business logic from config classes.
** 007 GES 20150501 Moved to a more generic package.
** 008 CA  20150828 Protected processAnywhereActions against concurrent modifications when 
**                  iterating.
** 009 CA  20150821 Refactored to move some complex OS actions at the driver level.
** 010 CA  20150925 Made canProcessMouse protected.
** 011 HC  20151022 Fixed an abend when mouse event causes widget destroyed.
** 012 IAS 20151214 SET-WAIT-STATE support.
** 013 OM  20171027 Upgraded code to Java8.
** 014 CA  20180530 Avoid using widgets not attached to a window, when processing mouse events.
*/

/*
** This program is free software: you can redistribute it and/or modify
** it under the terms of the GNU Affero General Public License as
** published by the Free Software Foundation, either version 3 of the
** License, or (at your option) any later version.
**
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
** GNU Affero General Public License for more details.
**
** You may find a copy of the GNU Affero GPL version 3 at the following
** location: https://www.gnu.org/licenses/agpl-3.0.en.html
** 
** Additional terms under GNU Affero GPL version 3 section 7:
** 
**   Under Section 7 of the GNU Affero GPL version 3, the following additional
**   terms apply to the works covered under the License.  These additional terms
**   are non-permissive additional terms allowed under Section 7 of the GNU
**   Affero GPL version 3 and may not be removed by you.
** 
**   0. Attribution Requirement.
** 
**     You must preserve all legal notices or author attributions in the covered
**     work or Appropriate Legal Notices displayed by works containing the covered
**     work.  You may not remove from the covered work any author or developer
**     credit already included within the covered work.
** 
**   1. No License To Use Trademarks.
** 
**     This license does not grant any license or rights to use the trademarks
**     Golden Code, FWD, any Golden Code or FWD logo, or any other trademarks
**     of Golden Code Development Corporation. You are not authorized to use the
**     name Golden Code, FWD, or the names of any author or contributor, for
**     publicity purposes without written authorization.
** 
**   2. No Misrepresentation of Affiliation.
** 
**     You may not represent yourself as Golden Code Development Corporation or FWD.
** 
**     You may not represent yourself for publicity purposes as associated with
**     Golden Code Development Corporation, FWD, or any author or contributor to
**     the covered work, without written authorization.
** 
**   3. No Misrepresentation of Source or Origin.
** 
**     You may not represent the covered work as solely your work.  All modified
**     versions of the covered work must be marked in a reasonable way to make it
**     clear that the modified work is not originating from Golden Code Development
**     Corporation or FWD.  All modified versions must contain the notices of
**     attribution required in this license.
*/

package com.goldencode.p2j.ui.client.gui.driver;

// this usage of AWT events/Component is intentional, they are being used separately from the
// AWT/Swing event model; no other AWT/Swing usage is occurring in this class
import java.awt.event.*;
import java.util.*;

import com.goldencode.p2j.security.*;
import com.goldencode.p2j.ui.chui.ThinClient;
import com.goldencode.p2j.ui.client.*;
import com.goldencode.p2j.ui.client.widget.*;

/**
 * Provide generic mouse event support for the GUI driver.
 */
public abstract class MouseHandler
extends MouseAdapter
{
   /** Context-local data for this mouse-handler. */
   private static final ContextContainer local = new ContextContainer();
      
   /** The GUI window emulator with which this mouse handler is registered. */
   protected EmulatedWindowState ews;
   
   /** Access to {@link ThinClient}. */
   protected final ThinClient tc;
   
   /**
    * Create a new mouse handler, for the specified window emulator.
    */
   public MouseHandler()
   {
      this.tc = ThinClient.getInstance();
   }

   /**
    * Enable or disable capture of mouse events.
    * 
    * @param    state
    *           {@code true} to capture mouse events, {@code false} to ignore them.
    *           
    * @return   The previous state.
    */
   public static boolean captureMouseEvents(boolean state)
   {
      WorkArea wa = local.get();

      synchronized (wa)
      {
         boolean prev = wa.state;
         
         wa.state = state;

         return prev;
      }
   }

   /**
    * Notify the widget of any mouse events, generated after the given timestamp (in millis).
    * 
    * @param    wid
    *           The widget to be notified.
    * @param    when
    *           Only events generated after this timestamp will be notified.
    */
   public static void listenAnyMouseEvent(int wid, long when)
   {
      WorkArea wa = local.get();
      
      wa.anyActions.put(wid, when);
   }
   
   /**
    * Register the specified mouse action with the specified widget.
    * 
    * @param    wid
    *           The widget ID.
    * @param    action
    *           The action to be triggered for this widget, via mouse.
    */
   public static void registerWidgetAction(int wid, MouseWidgetAction action)
   {
      WorkArea wa = local.get();
   
      List<MouseWidgetAction> actions = wa.actions.computeIfAbsent(wid, k -> new ArrayList<>());
   
      actions.add(action);
   }
   
   /**
    * Deregister the specified widget for mouse events.
    * 
    * @param    wid
    *           The widget ID.
    */
   public static void deregisterWidget(int wid)
   {
      WorkArea wa = local.get();
      
      wa.actions.remove(wid);
      wa.anyActions.remove(wid);
   }
   
   /**
    * Getting the list of the widgets with registered tooltip workers.
    *
    * @return   The widget list for ones that have the registered worker.
    */
   public static Integer[] getWidgetListForTooltipWorkers()
   {
      // the results will be here
      ArrayList<Integer> widLst = new ArrayList<>();
      
      WorkArea wa = local.get();
      // enumerate actions map
      for (Integer wId : wa.actions.keySet())
      {
         // get the widget ID
         // get the list of the actions for this widget
         List<MouseWidgetAction> handlers = wa.actions.get(wId);
         // look the handler array for tooltip worker
         if (handlers != null)
         {
            // walk through handlers
            for (MouseWidgetAction handler : handlers)
            {
               if (handler instanceof MouseTooltipWorker)
               {
                  // add widget ID to result list 
                  widLst.add(wId);
                  // only one tooltip worker per widget, go to next widget ID
                  break;
               }
            }
         }
      }
      
      return widLst.toArray(new Integer[widLst.size()]);
   }
   
   /**
    * Deregisters the specified widget for tooltip worker events.  The returned worker object
    * should be stored somewhere in caller to be registered further.
    * 
    * @param    wId
    *           The widget ID.
    *
    * @return   The previously registered tooltip worker for given widget or {@code null} if
    *           the tooltip worker was not registered for this widget.
    */
   public static MouseWidgetAction deregisterTooltipWorker(int wId)
   {
      WorkArea wa = local.get();
      
      // get the list of the actions for this widget
      List<MouseWidgetAction> handlers = wa.actions.get(wId);
      // look the handler for tooltip worker
      if (handlers != null)
      {
         // walk through handlers
         Iterator<MouseWidgetAction> iterAction = handlers.iterator();
         while (iterAction.hasNext())
         {
            MouseWidgetAction action = iterAction.next(); 
            // found the tooltip handler
            if (action instanceof MouseTooltipWorker)
            {
               // remove handler from the list, no more tooltip events will be worked for
               // this widget
               iterAction.remove();
               // the handler removed, return copy to the caller, only one tooltip handler
               // per widget possible
               return action;
            }
         }
      }
      
      // if we are here - tooltip worker not found for the given widget
      return null;
   }
   
   /**
    * Handle the specified event, in the context of the specified widget.
    * 
    * @param    sourceID
    *           The source widget ID.
    * @param    evt
    *           The mouse event.
    */
   public static void handleMouseEvent(int sourceID, final MouseEvent evt)
   {
      @SuppressWarnings("unchecked")
      Widget<GuiOutputManager> source = 
         (Widget<GuiOutputManager>) OutputManager.resolveWidget(sourceID);
      
      if (source == null)
      {
         System.err.println("Received mouse event for zombie source " + sourceID + ": " +
                            evt.toString());
         return;
      }
      
      applyMouseEvent(source, evt);
      
      // at this point the source may be destroyed (for example window closed on button click),
      // do not process the event further in this case
      // TODO: this condition is fragile, find a better way to detect this condition
      if (!(source instanceof TopLevelWindow) && source.ancestor() == null)
      {
         return;
      }
      
      ThinClient.getInstance().eventDrawingBracket(source, false, true,
                                                   () -> processAnywhereActions(evt));

      processWidgetActions(sourceID, evt);
   }
   
   /**
    * Get the {@link MouseWidgetAction#getEventId() event IDs }for all actions registered with the
    * specified widget.
    * 
    * @param    widgetId
    *           The widget ID.
    *           
    * @return   See above.
    */
   public static Set<Integer> getWidgetActions(int widgetId)
   {
      WorkArea wa = local.get();
      
      Set<Integer> actions = new TreeSet<>();
      
      if (wa.actions.containsKey(widgetId))
      {
         for (MouseWidgetAction mwa : wa.actions.get(widgetId))
         {
            actions.add(mwa.getEventId());
         }
      }
      
      if (wa.anyActions.containsKey(widgetId))
      {
         // TODO: handle any action
      }
      
      return actions;
   }
   
   /**
    * Determine if the specified widget is registered to listen for any mouse action.
    * 
    * @param   widgetId
    *          The widget ID.
    *
    * @return  {@code true} if the widget ID is registered in the {@link WorkArea#anyActions} map.
    */
   public static boolean hasAnyAction(int widgetId)
   {
      WorkArea wa = local.get();

      return (wa.anyActions.containsKey(widgetId));
   }

   /**
    * Apply the specified mouse event to the specified source.
    * 
    * @param    mouseSource
    *           The mouse source.
    * @param    evt
    *           The mouse event.
    */
   protected static void applyMouseEvent(Widget<GuiOutputManager> mouseSource, MouseEvent evt)
   {
      switch (evt.getID())
      {
         case MouseEvent.MOUSE_CLICKED:
            mouseSource.mouseClicked(evt);
            break;
         case MouseEvent.MOUSE_PRESSED:
            mouseSource.mousePressed(evt);
            break;
         case MouseEvent.MOUSE_RELEASED:
            mouseSource.mouseReleased(evt);
            break;
         case MouseEvent.MOUSE_ENTERED:
            mouseSource.mouseEntered(evt);
            break;
         case MouseEvent.MOUSE_EXITED:
            mouseSource.mouseExited(evt);
            break;
         case MouseEvent.MOUSE_DRAGGED:
            mouseSource.mouseDragged(evt);
            break;
         case MouseEvent.MOUSE_MOVED:
            mouseSource.mouseMoved(evt);
            break;
         case MouseEvent.MOUSE_WHEEL:
            mouseSource.mouseWheelMoved((MouseWheelEvent) evt);
            break;
      }
   }

   /**
    * Notify all widgets listening for any mouse events.
    * 
    * @param   evt
    *          The mouse event.
    */
   private static void processAnywhereActions(MouseEvent evt)
   {
      Map<Integer, Long> copy = local.obtainAnyActions();
      
      for (int wid : copy.keySet())
      {
         long regTime = copy.get(wid);
         long evtTime = evt.getWhen();
         
         if (evtTime <= regTime)
         {
            continue;
         }

         @SuppressWarnings("unchecked")
         Widget<GuiOutputManager> source = 
            (Widget<GuiOutputManager>) OutputManager.resolveWidget(wid);
         
         applyMouseEvent(source, evt);
      }
   }
   
   /**
    * Process the mouse actions registered for the specified source.
    * 
    * @param    sourceID
    *           The source widget ID.
    * @param    evt
    *           The mouse event.
    */
   protected static void processWidgetActions(int sourceID, MouseEvent evt)
   {
      WorkArea wa = local.get();

      List<MouseWidgetAction> actions = wa.actions.get(sourceID);
      
      if (actions != null)
      {
         for (MouseWidgetAction action : actions)
         {
            switch (evt.getID())
            {
               case MouseEvent.MOUSE_CLICKED:
                  action.mouseClicked(evt);
                  break;
               
               case MouseEvent.MOUSE_PRESSED:
                  action.mousePressed(evt);
                  break;
               
               case MouseEvent.MOUSE_RELEASED:
                  action.mouseReleased(evt);
                  break;
               
               case MouseEvent.MOUSE_ENTERED:
                  action.mouseEntered(evt);
                  break;
               
               case MouseEvent.MOUSE_EXITED:
                  action.mouseExited(evt);
                  break;
               
               case MouseEvent.MOUSE_DRAGGED:
                  action.mouseDragged(evt);
                  break;
               
               case MouseEvent.MOUSE_MOVED:
                  action.mouseMoved(evt);
                  break;
               
               case MouseEvent.MOUSE_WHEEL:
                  action.mouseWheelMoved((MouseWheelEvent) evt);
                  break;
            }
         }
      }
   }
   
   /**
    * Set the window on which this mouse handler acts.
    * 
    * @param    ews
    *           The window.
    */
   public void setWindow(EmulatedWindowState ews)
   {
      this.ews = ews;
   }
   
   /**
    * Notification of a mouse clicked event.  Event will be posted via 
    * {@link ThinClient#postMouseEvent} only if mouse processing is active.
    * 
    * @param    e
    *           The mouse event.
    */
   @Override
   public void mouseClicked(MouseEvent e)
   {
      if (canProcessMouse())
      {
         tc.postMouseEvent(e, ews.getWindowId());
      }
   }

   /**
    * Notification of a mouse pressed event.  Event will be posted via 
    * {@link ThinClient#postMouseEvent} only if mouse processing is active.
    * 
    * @param    e
    *           The mouse event.
    */
   @Override
   public void mousePressed(MouseEvent e)
   {
      if (canProcessMouse())
      {
         tc.postMouseEvent(e, ews.getWindowId());
      }
   }
   
   /**
    * Notification of a mouse dragged event.  Event will be posted via 
    * {@link ThinClient#postMouseEvent} only if mouse processing is active.
    * 
    * @param    e
    *           The mouse event.
    */
   @Override
   public void mouseDragged(MouseEvent e)
   {
      if (canProcessMouse())
      {
         tc.postMouseEvent(e, ews.getWindowId());
      }
   }
   
   /**
    * Notification of a mouse released event.  Event will be posted via 
    * {@link ThinClient#postMouseEvent} only if mouse processing is active.
    * 
    * @param    e
    *           The mouse event.
    */
   @Override
   public void mouseReleased(MouseEvent e)
   {
      if (canProcessMouse())
      {
         tc.postMouseEvent(e, ews.getWindowId());
      }
   }
   
   /**
    * Notification of a mouse exited event.  Event will be posted via 
    * {@link ThinClient#postMouseEvent} only if mouse processing is active.
    * 
    * @param    e
    *           The mouse event.
    */
   @Override
   public void mouseExited(MouseEvent e)
   {
      if (canProcessMouse())
      {
         tc.postMouseEvent(e, ews.getWindowId());
      }
   }
   
   /**
    * Notification of a mouse entered event.  Event will be posted via 
    * {@link ThinClient#postMouseEvent} only if mouse processing is active.
    * 
    * @param    e
    *           The mouse event.
    */
   @Override
   public void mouseEntered(MouseEvent e)
   {
      if (canProcessMouse())
      {
         tc.postMouseEvent(e, ews.getWindowId());
      }
   }
   
   /**
    * Notification of a mouse moved event.  Event will be posted via 
    * {@link ThinClient#postMouseEvent} only if mouse processing is active.
    * 
    * @param    e
    *           The mouse event.
    */
   @Override
   public void mouseMoved(MouseEvent e)
   {
      if (canProcessMouse())
      {
         tc.postMouseEvent(e, ews.getWindowId());
      }
   }
   
   /**
    * Notification of a mouse wheel moved event.  Event will be posted via 
    * {@link ThinClient#postMouseEvent} only if mouse processing is active.
    * 
    * @param    e
    *           The mouse event.
    */
   @Override
   public void mouseWheelMoved(MouseWheelEvent e)
   {
      if (canProcessMouse())
      {
         tc.postMouseEvent(e, ews.getWindowId());
      }
   }
   
   /**
    * Determine if mouse events can be processed.
    * 
    * @return   See above.
    */
   protected static boolean canProcessMouse()
   {
      WorkArea wa = local.get();
      synchronized (wa)
      {
         return wa.state && !ThinClient.isWaitState();
      }
   }
   
   /**
    * Container for mouse-related data.
    */
   private static class WorkArea
   {
      /**
       * Map of widgets listening for any mouse action.  Only events after the specified 
       * timestamp will be sent to the widget. 
       */
      private Map<Integer, Long> anyActions = new HashMap<>();
      
      /** Registered widgets which react on mouse events. */
      private Map<Integer, List<MouseWidgetAction>> actions = new HashMap<>();
      
      /** The state of mouse event capturing: when {@code false}, mouse events are ignored. */
      private boolean state = false;
   }
   
   /**
    * Simple container that stores and returns a context-local instance of the global work area.
    */
   private static class ContextContainer
   extends ContextLocal<WorkArea>
   {
      /**
       * Obtain the copy of {@link WorkArea#anyActions} synchronously.
       * 
       * @return  A new {@code Map<Integer, Long>} object.
       */
      public synchronized Map<Integer, Long> obtainAnyActions()
      {
         return new HashMap<>(get().anyActions);
      }
      
      /**
       * Initializes the work area, the first time it is requested within a new context.
       *
       * @return   The newly instantiated work area.
       */
      protected WorkArea initialValue()
      {
         return new WorkArea();
      }
   }
}
