/*
** Module   : PaintStructure.java
** Abstract : A structure used to store paint parameters.
**
** Copyright (c) 2014-2018, Golden Code Development Corporation.
**
** -#- -I- --Date-- ---------------------------------Description---------------------------------
** 001 MAG 20140508 First version.
** 002 MAG 20140912 Add fillPoly parameters.
** 003 EVL 20141116 Adding round rectangles draw and fill.
** 004 EVL 20141125 Adding dotted line draw capability.
** 005 EVL 20150203 Adding image offset options.
** 006 EVL 20150218 Adding support TRANSPARENT, STRETCH-TO-FIT, RETAIN-SHAPE and
**                  CONVERT-3D-COLORS options to image drawing.
** 007 CA  20150408 Added window title.
** 008 CA  20150507 Added flag to center text vertically, when drawing.
** 009 GES 20150417 Removed all Swing-specific data structures/imports. Moved the class to a
**                  generic location that is naturally shared across all GUI drivers.
** 010 SBI 20151012 Removed unnecessary import.
** 011 HC  20170223 Improved clipping to allow to clip with arbitrary set of overlapping or non-
**                  overlapping regions.
** 012 OM  20170330 Added debugging support by adding verbose toString() implementation.
** 013 EVL 20170706 Adding SET_XOR_MODE debug info, handling DRAW_SELECTION.
** 014 CA  20180207 flush() is handled internally by the driver's window implementation; this is
**                  required because flush can be done only after batch, translate and clip 
**                  brackets reach 0 (solves an issue in web mode draw cache, which requires to
**                  not have any open brackets).
**                  Warning: in web mode, unbalanced translate push/pop or CLIP/NO_CLIP will cause
**                  an infinite loop.  In Swing driver, it will not flush until all batch, 
**                  translate and clip brackets reach 0.
** 015 OM  20180815 Added support for drawing ellipse primitives.
*/

package com.goldencode.p2j.ui.client.gui.driver;

import com.goldencode.p2j.ui.*;
import com.goldencode.p2j.ui.client.*;

import java.util.*;

/**
 * A structure used to store paint parameters. Depending on paint operation different fields are
 * used.
 */
/*
** This program is free software: you can redistribute it and/or modify
** it under the terms of the GNU Affero General Public License as
** published by the Free Software Foundation, either version 3 of the
** License, or (at your option) any later version.
**
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
** GNU Affero General Public License for more details.
**
** You may find a copy of the GNU Affero GPL version 3 at the following
** location: https://www.gnu.org/licenses/agpl-3.0.en.html
** 
** Additional terms under GNU Affero GPL version 3 section 7:
** 
**   Under Section 7 of the GNU Affero GPL version 3, the following additional
**   terms apply to the works covered under the License.  These additional terms
**   are non-permissive additional terms allowed under Section 7 of the GNU
**   Affero GPL version 3 and may not be removed by you.
** 
**   0. Attribution Requirement.
** 
**     You must preserve all legal notices or author attributions in the covered
**     work or Appropriate Legal Notices displayed by works containing the covered
**     work.  You may not remove from the covered work any author or developer
**     credit already included within the covered work.
** 
**   1. No License To Use Trademarks.
** 
**     This license does not grant any license or rights to use the trademarks
**     Golden Code, FWD, any Golden Code or FWD logo, or any other trademarks
**     of Golden Code Development Corporation. You are not authorized to use the
**     name Golden Code, FWD, or the names of any author or contributor, for
**     publicity purposes without written authorization.
** 
**   2. No Misrepresentation of Affiliation.
** 
**     You may not represent yourself as Golden Code Development Corporation or FWD.
** 
**     You may not represent yourself for publicity purposes as associated with
**     Golden Code Development Corporation, FWD, or any author or contributor to
**     the covered work, without written authorization.
** 
**   3. No Misrepresentation of Source or Origin.
** 
**     You may not represent the covered work as solely your work.  All modified
**     versions of the covered work must be marked in a reasonable way to make it
**     clear that the modified work is not originating from Golden Code Development
**     Corporation or FWD.  All modified versions must contain the notices of
**     attribution required in this license.
*/
public class PaintStructure<F, I>
{
   /** Paint operation */
   public PaintPrimitives id;
   
   /** X parameter */
   public int x;
   
   /** Y parameter */
   public int y;
   
   /** Width parameter */
   public int width;
   
   /** Height parameter */
   public int height;
   
   /** X offset to start drawing image. */
   public int xOffset;
   
   /** Y offset to start drawing image. */
   public int yOffset;
   
   /** Transparency image flag. */
   public boolean transparent;
   
   /** Fit image to surrounding area flag. */
   public boolean stretchToFit;
   
   /** Keep image aspect ratio when fitting image to area flag. */
   public boolean retainShape;
   
   /** Replace certain colors with predefined system wide 3D colors flag. */
   public boolean convert3D;
   
   /** Diameter of the arc at the four corner. */
   public int arcDiameter;
   
   /** Polygon X coordinates */
   public int[] xPoints;
   
   /** Polygon Y coordinates */
   public int[] yPoints;
   
   /** Polygon number of coordinates */
   public int nPoints;
      
   /** Raised parameter for 3D operations */
   public boolean raised;
   
   /** Font parameter */
   public FontDetails<F> font;
   
   /** Font style constant. */
   public FontStyle style;
   
   /** Color parameter */
   public ColorRgb color;
   
   /** Text to draw */
   public String text;
   
   /** Image to draw */
   public ImageWrapper<I> img;

   /** Pattern to draw the lines. */
   public LineStroke stroke;
   
   /** Optional width for the stroke. */
   public float strokeWidth;

   /** The OS window's title. */
   public String title;
   
   /** Flag indicating if the text needs to be centered vertically. */
   public boolean centered;

   /** Flag indicating if we need to clean area instead of drawing. */
   public boolean clear;

   /** Clipping rectangles */
   public Set<NativeRectangle> clip;

   /**
    * Constructor.
    * 
    * @param   id
    *          Paint operation id.
    */
   public PaintStructure(PaintPrimitives id)
   {
      this.id = id;
   }
   
   /**
    * Only used for debugging. Each primitives is described as shortly as possible on a single
    * line. Complex info or binary data was skipped from the returned value.
    * 
    * @return  a short description of this drawing primitive.
    */
   @Override
   public String toString()
   {
      StringBuilder sb = new StringBuilder(id.toString());
      sb.append('('); 
      switch (id)
      {
         case SET_LOCATION:
            sb.append(x).append(',').append(y);
            break;
         
         case RESIZE_WINDOW:
            sb.append(width).append('x').append(height);
            break;
         
         case CLEAR:
         case CLEAR_AREA: // alias for CLEAR ? 
         case DRAW_RECT:
         case FILL_RECT:
         case DRAW_ELLIPSE:
         case FILL_ELLIPSE:
         case SET_WINDOW_BOUNDS:
         case START_RECORDING:
         case SET_ICON:
         case DRAW_IMAGE:
         case DRAW_SELECTION:
            sb.append(x).append(',').append(y).append(' ');
            sb.append(width).append('x').append(height);
            if (id == PaintPrimitives.SET_ICON || id == PaintPrimitives.DRAW_IMAGE)
            {
               sb.append(" (...)"); // additional data missing
            }
            else if (id == PaintPrimitives.DRAW_SELECTION)
            {
               sb.append(clear ? " clear" : " draw");
            }
            break;
         
         case DRAW_STRING:
            sb.append('\"').append(text).append("\" @ ").append(x).append(',').append(y);
            if (centered)
            {
               sb.append(" C");
            }
            break;
         
         case DRAW_STRING_SCALED:
            sb.append('\"').append(text).append("\" @ ").append(x).append(',').append(y);
            if (centered)
            {
               sb.append(" C");
            }
            sb.append(' ').append(width).append('x').append(height);
            break;
         
         case DRAW_HIGHLIGHT:
            sb.append('\"').append(text).append("\" @ ").append(x).append(',').append(y);
            sb.append(' ').append(width).append('x').append(height);
            if (color != null)
            {
               sb.append(" #").append(Integer.toHexString(color.getGuiRgb()));
            }
            break;
         
         case SET_TITLE:
            sb.append('\"').append(title).append('\"');
            break;
         
         case DRAW_LINE:
            sb.append(x).append(',').append(y).append(" to ");
            sb.append(width).append('x').append(height);
            break;
         
         case DRAW_ROUND_RECT:
         case FILL_ROUND_RECT:
            sb.append(x).append(',').append(y).append(' ');
            sb.append(width).append('x').append(height).append(" R:").append(arcDiameter);
            break;
         
         case DRAW_3D_RECT:
         case FILL_3D_RECT:
            sb.append(x).append(',').append(y).append(' ');
            sb.append(width).append('x').append(height);
            sb.append(raised ? " raised" : " sunken");
            break;
         
         case SET_DESKTOP_BGCOLOR:
         case SET_COLOR:
         case SET_XOR_MODE:
            if (color != null)
            {
               sb.append("#").append(Integer.toHexString(color.getGuiRgb()));
            }
            break;
         
         case SET_LINE_STROKE:
            sb.append(stroke.ordinal()).append(',').append(Math.round(strokeWidth));
            break;
         
         case START_BATCH:
         case END_BATCH:
            break;
         case FLUSH:
            sb.append("@deprecated");
            break;
         case NO_CLIP:
         case STOP_RECORDING:
         case SET_XOR_COMPOSITE:
         case RESET_COMPOSITE:
         case CLEAR_HIGHLIGHT:
         case RESET_SELECTION:
            // no params
            break;
         
         case TRANSLATE_PUSH:
         case TRANSLATE_POP:
            sb.append(x).append(',').append(y);
            break;
         
         case SET_FONT_STYLE:
            sb.append(style);
            break;
         
         case SET_FONT:
            sb.append(font);
            break;
         
         case DRAW_PARAGRAPH:
            sb.append('\"').append(text).append("\" @ ").append(x).append(',').append(y);
            sb.append(" W:").append(width).append(' ').append(font);
            break;
         
         case DRAW_CACHED:
            sb.append("..."); // binary data
            break;
         
         case CLIP:
            if (clip != null)
            {
               sb.append(clip.size()).append(' ').append(Objects.toString(clip));
            }
            break;
         
         case FILL_POLYGON:
            sb.append(nPoints); // missing actual points
            break;
      }
      
      sb.append(")");
      return sb.toString();
   }
}
