/*
** Module   : MouseMoveable.java
** Abstract : Implementation of a mouse moveable widget.
**
** Copyright (c) 2014-2018, Golden Code Development Corporation.
**
** -#- -I- --Date-- ---------------------------------Description----------------------------------
** 001 CA  20140805 Created initial version.
** 002 CA  20141119 Refactored drawing code to enable multiple OS windows, for GUI support.  Misc
**                  changes/fixes related to drawing on multiple windows.
** 003 HC  20141207 Improved window mouse drag and mouse resize behavior.
** 004 HC  20150323 WidgetId.getId() renamed to asInt().
** 005 GES 20150501 Moved to a more generic package.
** 006 HC  20150612 Implemented support for modal windows.
** 007 CA  20150826 Moved to the Swing driver.
** 008 CA  20150929 Fixed multi-window drawing issues caused by concurrent drawing access to the
**                  GuiDriver: all access to the driver is done exclusively, each thread taking
**                  and releasing ownership for the duration of the drawing operation or other
**                  GUI API invocation which requires access to the underlying physical window.
** 009 HC  20151013 Changes to extend GUI multi-window focus management and ACTIVE-WINDOW system
**                  handle processing to modal windows.
** 010 IAS 20160217 Added null check in the mouseReleased method.
** 011 HC  20160304 Fixed moving window by its title-bar, the mouse cursor did not stay put.
** 012 OM  20170222 Drop move event if both dx and dy are null.
** 013 CA  20180604 If window was moved, then set this location as an explicitly set location (not
**                  client-side).
*/
/*
** This program is free software: you can redistribute it and/or modify
** it under the terms of the GNU Affero General Public License as
** published by the Free Software Foundation, either version 3 of the
** License, or (at your option) any later version.
**
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
** GNU Affero General Public License for more details.
**
** You may find a copy of the GNU Affero GPL version 3 at the following
** location: https://www.gnu.org/licenses/agpl-3.0.en.html
** 
** Additional terms under GNU Affero GPL version 3 section 7:
** 
**   Under Section 7 of the GNU Affero GPL version 3, the following additional
**   terms apply to the works covered under the License.  These additional terms
**   are non-permissive additional terms allowed under Section 7 of the GNU
**   Affero GPL version 3 and may not be removed by you.
** 
**   0. Attribution Requirement.
** 
**     You must preserve all legal notices or author attributions in the covered
**     work or Appropriate Legal Notices displayed by works containing the covered
**     work.  You may not remove from the covered work any author or developer
**     credit already included within the covered work.
** 
**   1. No License To Use Trademarks.
** 
**     This license does not grant any license or rights to use the trademarks
**     Golden Code, FWD, any Golden Code or FWD logo, or any other trademarks
**     of Golden Code Development Corporation. You are not authorized to use the
**     name Golden Code, FWD, or the names of any author or contributor, for
**     publicity purposes without written authorization.
** 
**   2. No Misrepresentation of Affiliation.
** 
**     You may not represent yourself as Golden Code Development Corporation or FWD.
** 
**     You may not represent yourself for publicity purposes as associated with
**     Golden Code Development Corporation, FWD, or any author or contributor to
**     the covered work, without written authorization.
** 
**   3. No Misrepresentation of Source or Origin.
** 
**     You may not represent the covered work as solely your work.  All modified
**     versions of the covered work must be marked in a reasonable way to make it
**     clear that the modified work is not originating from Golden Code Development
**     Corporation or FWD.  All modified versions must contain the notices of
**     attribution required in this license.
*/

package com.goldencode.p2j.ui.client.gui.driver.swing;

// this usage of AWT events is intentional, they are being used separately from the AWT/Swing
// event model; no other AWT/Swing usage is occurring in this class
import java.awt.event.*;

import com.goldencode.p2j.ui.*;
import com.goldencode.p2j.ui.client.*;
import com.goldencode.p2j.ui.client.event.*;
import com.goldencode.p2j.ui.client.gui.driver.MouseWidgetAction;

/**
 * Listener for mouse actions which can trigger a widget to be moved.
 * TODO: currently this is hard-coded for the WINDOW widget.  If other widgets will require move
 * via mouse, it will need to be decoupled.
 */
class MouseMoveable
extends MouseWidgetAction
{
   /** The screen location of the click mouse event. */
   private NativePoint clickLoc;
   
   /** The screen location of the associated widget when left button down. */
   private NativePoint screenLoc;

   /**
    * Create a new mouse moveable action, for the specified widget.
    * 
    * @param    wid
    *           The widget ID.
    */
   public MouseMoveable(int wid)
   {
      super(wid, 
            new int[] 
            {
               MouseEvent.MOUSE_PRESSED, 
               MouseEvent.MOUSE_RELEASED, 
               MouseEvent.MOUSE_DRAGGED, 
               MouseEvent.MOUSE_EXITED 
            },
            MouseEvt.MOUSE_MOVABLE);
   }
   
   /**
    * Notification of a mouse dragged event for this {@link MouseWidgetAction#widget}.
    * 
    * @param    e
    *           The mouse event.
    */
   @Override
   public void mouseDragged(MouseEvent e)
   {
      updateWidgetLocation(e);
   }

   /**
    * Notification of a mouse pressed event for this {@link MouseWidgetAction#widget}.
    * 
    * @param    e
    *           The mouse event.
    */
   @Override
   public void mousePressed(MouseEvent e)
   {
      if (e.getButton() == MouseEvent.BUTTON1)
      {
         clickLoc = new NativePoint(e.getXOnScreen(), e.getYOnScreen());
         gd.selectWindow(widget.topLevelWindow().getId().asInt());
         screenLoc = gd.getWindowLocation();
         gd.releaseWindow();
      }
   }

   /**
    * Notification of a mouse released event for this {@link MouseWidgetAction#widget}.
    * 
    * @param    e
    *           The mouse event.
    */
   @Override
   public void mouseReleased(MouseEvent e)
   {
      updateWidgetLocation(e);

      clickLoc = null;
      screenLoc = null;
   }
   
   /**
    * Update the location of the widget being moved according to the supplied mouse event.
    * 
    * @param    e
    *           The mouse event.
    */
   private void updateWidgetLocation(MouseEvent e)
   {
      if (clickLoc == null)
      {
         return;
      }
      
      int dx = e.getXOnScreen() - clickLoc.x;
      int dy = e.getYOnScreen() - clickLoc.y;
      if (dx == 0 && dy == 0)
      {
         return; // just a click on titlebar, activate the window but there's nothing to move
      }
      
      // TODO: this is hard-coded for top-level windows currently
      TopLevelWindow<?> topLevelWindow = widget.topLevelWindow();
      if (topLevelWindow != null)
      {
         topLevelWindow.setPhysicalLocation(screenLoc.x + dx, screenLoc.y + dy);
         
         // when the windows has been moved by user, we need to adjust the explicit coordinates
         WidgetConfig wcfg = topLevelWindow.config();
         if (wcfg != null && wcfg instanceof BaseConfig)
         {
            BaseConfig bc = (BaseConfig) wcfg;
            bc.x = screenLoc.x + dx;
            bc.y = screenLoc.y + dy;
         }
      }
   }
}