/*
** Module   : SwingGuiDriver.java
** Abstract : Swing GUI screen driver.
**
** Copyright (c) 2014-2018, Golden Code Development Corporation.
**
** -#- -I- --Date-- ---------------------------------Description---------------------------------
** 001 MAG 20140508 First version.
** 002 CA  20140711 Enhanced font support.
** 003 CA  20140805 Refactored to implement new GenericScreenDriver APIs and the support provided 
**                  by it.  Added mouse support and other misc APIs related to window management.
** 004 MAG 20140902 Integrate COLOR-TABLE. Added support for BGCOLOR and FGCOLOR color conversion.
**                  Added system colors used solely for widget drawing.  
** 005 MAG 20140912 Implement GUI status bar.
** 006 HC  20140911 GUI coordinate support - data types refactoring.
** 007 MAG 20140925 Check if window has renderer.
** 008 CA  20141021 Added support to allow drawing of nested containers (the real location of the
**                  translated origin is required).  Misc changes related to image drawing and
**                  mouse processing.
** 009 VIG 20141022 TC.screenSize replaced with window.size.
** 010 HC  20141031 Implemented support for drawing highlighted areas.
** 011 MAG 20141029 Added clip rectangle parameter to draw method.
** 012 CA  20141029 Each simulator needs to receive the swing keyboard reader implementation in 
**                  use.  It's the callers responsibility to ensure that the same instance is used
**                  for all simulators, if the client is ran in a multi-window mode (i.e. GUI).
**                  The GUI driver ensures that only one instance is used, for each client.
** 013 CA  20141119 Refactored drawing code to enable multiple OS windows, for GUI support.  Misc
**                  changes/fixes related to drawing on multiple windows.
** 014 EVL 20141116 Adding round rectangles draw and fill.
** 015 CA  20141123 Added APIs to allow enforcement of window stacking at OS level.
** 016 CA  20141125 Invalidate the current simulator, if the currently selected window is being 
**                  destroyed.  Added API to conditionally enable/disable OS-level events.
** 017 EVL 20141125 Adding dotted line draw capability.
** 018 CA  20141205 Added API to draw relative to window's real origin (top-left corner), 
**                  regardless of the current translated origin.
**                  Added API to register editable widgets (thus the driver can adjust the mouse 
**                  cursor accordingly).
** 019 CA  20141220 Added APIs for clipboard access.  Added APIs to allow text drawing and font 
**                  selection matching legacy sizes.
** 020 HC  20141207 Implemented runtime support for several SESSION attributes.
** 021 EVL 20141211 Adding tooltip worker registration and deregistration.
** 022 EVL 20150203 Adding image offset drawing capability.
** 023 EVL 20150218 Adding support TRANSPARENT, STRETCH-TO-FIT, RETAIN-SHAPE and CONVERT-3D-COLORS
**                  options to image drawing.
** 024 CA  20150317 Added API to determine the font's max width.
** 025 CA  20150408 Added APIs to set the OS window's title and to de-iconify the window.
** 026 HC  20150323 Removed business logic from config classes.
** 027 GES 20150505 Massively refactored this class to minimize the Swing-specific logic and move
**                  the maximum logic to a new generic abstract base class.
** 028 CA  20150507 Added APIs to draw text centered vertically.
** 029 CA  20150519 Added support for custom fonts defined in the directory.
** 030 HC  20160216 Support for modal windows.
** 031 HC  20150702 Implemented window modality independently on the actual UI driver.
** 032 GES 20150730 Modified translate() methods to match API changes.
** 033 CA  20150808 Added support for font aliases.
** 034 GES 20150818 Added restackWindows().  On font creation, set its style.  Move some mouse 
**                  actions at the driver level.
**                  deriveFont() now is driver specific - it scales the font depending on the DPI.
** 035 CA  20150929 Fixed multi-window drawing issues caused by concurrent drawing access to the
**                  GuiDriver: all access to the driver is done exclusively, each thread taking
**                  and releasing ownership for the duration of the drawing operation or other
**                  GUI API invocation which requires access to the underlying physical window.
** 036 CA  20151013 Added support for window resize, z-order operations for the GUI web client.
**     HC  20151022 A fix for JWindow not able to receive focus when its owner window is not
**                  visible. This resolves ALERT-BOX and DIALOG-BOX not receiving key input when
**                  no main window visible.
** 037 SBI 20151102 To do the keyboard reader a final driver's field, added the driver's level
**                  keyboard focus manager.
** 038 HC  20151213 Improved DIALOG-BOX window parenting logic and related changes.
** 039 EVL 20160107 Adding overlay window positioning.
** 040 EVL 20160123 Removing isWindowActive() method.  Logic is now in WindowManager. 
** 041 IAS 20160130 LOAD-MOUSE-POINTER support
** 042 SBI 20160321 Implemented new method isWindowDecorationSupported().
** 043 HC  20160407 Overhaul of window-activation logic. Removed setOverlayWindowLocation(), the
**                  method was unnecessary and was causing the COMBO-BOX drop-down location
**                  incorrect.
** 044 SBI 20160412 Added openURL(String path) to open the target page in the browser.
** 045 CA  20170104 Added flag to track window's modal state.
** 046 OM  20170214 Replaced base class from JFrame to JDialog which suits better our purpose.
** 047 EVL 20170520 The font creation method must copy anti-aliasing flag for new font.
** 048 OM  20170412 Dropped window management. Swing handle this automatically.
** 049 HC  20170819 Added extension widget method GET-MOUSE-POSITION. Implemented runtime support
**                  for WINDOW:ALWAYS-ON-TOP and SHOW-IN-TASKBAR.
** 050 HC  20171024 Runtime support for SYSTEM-DIALOG PRINT-SETUP and OUTPUT TO PRINTER.
** 051 HC  20171123 Fixed a stack overlow condition in GUI redirected mode.
** 052 CA  20171204 Moved some code to FontHelper, to allow the Web driver to measure text on the
**                  client-side, via a AWT font instance.
** 053 CA  20180210 Renamed deliverPDFPrintOutput to deliverPrintOutput and added report format 
**                  parameter.
** 054 SVL 20180310 getPDFPrintOutputDir renamed to getPrintOutputDir.
** 055 SBI 20180410 Added openMimeResource, changed deliverPrintOutput to open the target file,
**                  renamed deliverPrintOutput to deliverDocumentOutput.
** 056 CA  20180313 Javadoc fix.
** 057 SBI 20180517 Changed openMimeResource to use openDocument(URI uri).
** 058 HC  20180525 Fixed mouse pointer loading.
** 059 SBI 20180815 Changed according to GuiDriver.isFontInstalled.
*/
/*
** This program is free software: you can redistribute it and/or modify
** it under the terms of the GNU Affero General Public License as
** published by the Free Software Foundation, either version 3 of the
** License, or (at your option) any later version.
**
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
** GNU Affero General Public License for more details.
**
** You may find a copy of the GNU Affero GPL version 3 at the following
** location: https://www.gnu.org/licenses/agpl-3.0.en.html
** 
** Additional terms under GNU Affero GPL version 3 section 7:
** 
**   Under Section 7 of the GNU Affero GPL version 3, the following additional
**   terms apply to the works covered under the License.  These additional terms
**   are non-permissive additional terms allowed under Section 7 of the GNU
**   Affero GPL version 3 and may not be removed by you.
** 
**   0. Attribution Requirement.
** 
**     You must preserve all legal notices or author attributions in the covered
**     work or Appropriate Legal Notices displayed by works containing the covered
**     work.  You may not remove from the covered work any author or developer
**     credit already included within the covered work.
** 
**   1. No License To Use Trademarks.
** 
**     This license does not grant any license or rights to use the trademarks
**     Golden Code, FWD, any Golden Code or FWD logo, or any other trademarks
**     of Golden Code Development Corporation. You are not authorized to use the
**     name Golden Code, FWD, or the names of any author or contributor, for
**     publicity purposes without written authorization.
** 
**   2. No Misrepresentation of Affiliation.
** 
**     You may not represent yourself as Golden Code Development Corporation or FWD.
** 
**     You may not represent yourself for publicity purposes as associated with
**     Golden Code Development Corporation, FWD, or any author or contributor to
**     the covered work, without written authorization.
** 
**   3. No Misrepresentation of Source or Origin.
** 
**     You may not represent the covered work as solely your work.  All modified
**     versions of the covered work must be marked in a reasonable way to make it
**     clear that the modified work is not originating from Golden Code Development
**     Corporation or FWD.  All modified versions must contain the notices of
**     attribution required in this license.
*/

package com.goldencode.p2j.ui.client.gui.driver.swing;

import java.awt.*;
import java.awt.Point;
import java.awt.datatransfer.*;
import java.awt.datatransfer.Clipboard;
import java.awt.image.*;

import javax.swing.*;

import java.lang.reflect.*;
import java.net.*;
import java.io.*;
import java.nio.file.*;
import java.util.*;
import java.util.function.*;
import java.util.logging.*;

import com.goldencode.p2j.cfg.*;
import com.goldencode.p2j.directory.*;
import com.goldencode.p2j.ui.client.*;
import com.goldencode.p2j.ui.client.gui.driver.*;
import com.goldencode.p2j.ui.client.driver.swing.*;
import com.goldencode.p2j.util.*;

/**
 * Swing-backed GUI driver.
 */
public class SwingGuiDriver
extends AbstractGuiDriver<Font>
{
   /**
    * The key reader.  This is global per the entire application, no matter how many frames are
    * alive.
    */
   private final SwingKeyboardReader keyReader = new WinKeyboardReader();
   
   /**
    * The driver's level keyboard focus manager.
    */
   private final KeyboardFocusManager keyboardFocusManager = new SwingEmulatedKeyboardFocusManager();

   
   /**
    * Create a new GUI screen driver.
    *
    * @param   config
    *          The <code>BootstrapConfig</code> configuration.
    */
   public SwingGuiDriver(BootstrapConfig config)
   {
      super(new GuiPrimitivesImpl()
      {
         @Override
         public void sync()
         {
            final SwingEmulatedWindow window = (SwingEmulatedWindow) ews;
            SwingUtilities.invokeLater(window::repaint);
         }
      }, config);
   }
   
   /**
    * Provide an audible feedback to user by generating a beep signal.
    */
   @Override
   public void beep()
   {
      Toolkit.getDefaultToolkit().beep();
   }
   
   /**
    * Register the specified widget as a moveable widget, via mouse.
    * 
    * @param    wid
    *           The widget ID.
    */
   @Override
   public void registerMoveableWidget(int wid)
   {
      ((SwingMouseHandler) ews.getMouseHandler()).registerAction(wid, new MouseMoveable(wid));
   }
   
   /**
    * Register this widget as being editable, so that the mouse cursor will adjust properly when
    * the widget is entered.
    * 
    * @param    action
    *           The widget mouse action.
    */
   @Override
   public void registerHoverableWidget(MouseHoverAction action)
   {
      ((SwingMouseHandler) ews.getMouseHandler()).registerAction(action.wid(), action);
   }

   /**
    * Register the specified widget as a resizeable widget, via mouse.
    * 
    * @param    wid
    *           The widget ID.
    */
   @Override
   public void registerResizeableWidget(int wid)
   {
      MouseHandler.registerWidgetAction(wid, new MouseResizeable(wid));
   }

   /**
    * Set the window details for resize; this includes the minimum and maximun dimensions, and 
    * also the window's resizeable state.
    * <p>
    * This is a no-op for Swing.
    * 
    * @param    resize
    *           Flag indicating if the window can be resized or not.
    * @param    minWidth
    *           The window's minimum width.
    * @param    minHeight
    *           The window's minimum height.
    * @param    maxWidth
    *           The window's maximum width.
    * @param    maxHeight
    *           The window's maximum height.
    */
   @Override
   public void setResizeableWindow(boolean resize,
                                   int     minWidth, 
                                   int     minHeight, 
                                   int     maxWidth, 
                                   int     maxHeight)
   {
      // no-op
   }

   /**
    * Deregister the specified widget for mouse events.
    * 
    * @param    wid
    *           The widget ID.
    */
   @Override
   public void deregisterWidget(int wid)
   {
      super.deregisterWidget(wid);
      ((SwingMouseHandler) ews.getMouseHandler()).deregisterAction(wid);
   }
   
   /**
    * Reports the screen resolution in dots per inch (DPI).
    *
    * @return   The DPI of the screen.
    */
   @Override
   public int getScreenResolution()
   {
      return Toolkit.getDefaultToolkit().getScreenResolution();
   }
   
   /**
    * Read single keystroke from keyboard. Returned code should be either proper
    * Progress 4GL/P2J key code or -1 if environment-specific input cannot be 
    * properly translated.
    * 
    * @return  key code or -1 if no keys were pressed since last call.
    */
   @Override
   public int readKey()
   {
      return keyReader.read();
   }

   /**
    * Create a new window emulator.
    * 
    * @param    windowId
    *           The window ID.
    * @param    showInTaskbar
    *           This flag is not supported by swing driver.
    *
    * @return   The newly instantiated window emulator.
    */
   @Override
   public EmulatedWindowState newEmulator(int windowId,
                                          boolean showInTaskbar)
   {
      final EmulatedWindowState ews;

      try
      {
         ews = SwingGuiClient.load(keyboardFocusManager, keyReader, null, false);
         
         // the real window size is unknown at this time; it will be set later, when the layout
         // is performed
         ews.setScreenBitmap(new ScreenBitmap(0, 0));
      }
      catch (InvocationTargetException | InterruptedException e)
      {
         throw new IllegalStateException("Could not create window emulator!", e);
      }
      
      return ews;
   }
   
   /**
    * Create a new child window emulator.
    * 
    * @param    windowId
    *           The window ID.
    * @param    owner
    *           The owning window.
    * @param    windowTitle
    *           Title of the window emulator.
    * @param    resizable
    *           <code>true</code> if the window is resizable.
    * @param    isShareActivationWithOwner
    *           Indicates whether the window shares activation state with its owner.
    * @param    modal
    *           <code>true</code> if this is a modal window.
    *           
    * @return   The newly instantiated window emulator.
    */
   @Override
   public EmulatedWindowState newChildWindowEmulator(
                               int                 windowId,
                               EmulatedWindowState owner, 
                               String              windowTitle,
                               boolean             resizable,
                               boolean             isShareActivationWithOwner,
                               boolean             modal)
   {
      EmulatedWindowState ews;
      
      try
      {
         SwingEmulatedWindow sew = (SwingEmulatedWindow) owner;
         ews = SwingGuiClient.load(keyboardFocusManager,
                                   keyReader,
                                   sew.getContainingWindow(),
                                   isShareActivationWithOwner);
         
         // the real window size is unknown at this time; it will be set later, when the layout
         // is performed
         ews.setScreenBitmap(new ScreenBitmap(0, 0));
      }
      catch (InvocationTargetException | InterruptedException e)
      {
         throw new IllegalStateException("Could not create window emulator!", e);
      }
      
      return ews;
   }
   
   /**
    * Create a font with the given details and register it with the driver.
    * 
    * @param    specs
    *           The font specifications.
    * 
    * @return   A font instance.
    */
   @Override
   public FontDetails<Font> createFont(FontDetails<Font> specs)
   {
      FontDetails<Font> details = (specs.fontAlias == null ? specs : specs.fontAlias);

      Font f = FontHelper.createFont(details);
      
      // here we use a copy of the original specs - as we want to replace only the drawing font,
      // not the legacy metrics
      FontDetails<Font> fd = specs.copy();
      fd.font = f;
      fd.pointSize = details.pointSize;
      fd.fontName = f.getName();
      // override the antialiasing too
      fd.swingAntiAliasing = details.swingAntiAliasing;

      return fd;
   }
   
   /**
    * Scale the given font to match the screen resolution.
    *
    * @param    fd
    *           Font to scale.  This instance's contained font will be modified. 
    */
   @Override
   public void deriveFont(FontDetails<Font> fd)
   {
      fd.font = FontHelper.deriveFont(fd.font, fd, getScreenResolution());
   }

   /**
    * Copy the given text to the system clipboard.
    * 
    * @param    txt
    *           The text to be copied.
    */
   @Override
   public void copyToClipboard(String txt)
   {
      StringSelection stringSelection = new StringSelection(txt);
      Clipboard clipboard = Toolkit.getDefaultToolkit().getSystemClipboard();
      clipboard.setContents(stringSelection, stringSelection);
   }
   
   /**
    * Get the current contents from the OS clipboard.
    * 
    * @return   The clipboard contents.
    */
   @Override
   public String clipboardContents()
   {
      Clipboard clipboard = Toolkit.getDefaultToolkit().getSystemClipboard();

      Transferable content = clipboard.getContents(null);
      boolean hasTransferableText = 
         (content != null && content.isDataFlavorSupported(DataFlavor.stringFlavor));
      
      String result = null;
      if (hasTransferableText)
      {
         try
         {
           result = (String) content.getTransferData(DataFlavor.stringFlavor);
         }
         catch (UnsupportedFlavorException | IOException ex)
         {
            // ignore
         }
       }
      
      return result;
   }
   
   /**
    * Force the list of all GUI windows to be stacked as in the specified order.
    * <p>
    * This is a no-op. Swing handle this automatically.
    * 
    * @param    windowIds
    *           The stack order to be enforced.
    */
   @Override
   public void restackWindows(int[] windowIds)
   {
      // the driver uses Swing's native window ordering, thus full control over z-ordering cannot
      // be achieved
   }
   
   /**
    * Move the specified window to the top of the z-order, at the OS level.
    * <p>
    * This is a no-op. Swing handle this automatically.
    *
    * @param    windowId
    *           The window to be moved.
    * @param    focus
    *           If <code>true</code>, the window will become the focused window 
    *           and its state changed to active. 
    */
   @Override
   public void moveToTop(int windowId, boolean focus)
   {
      SwingEmulatedWindow sew = (SwingEmulatedWindow) direct.getWindowEmulator(windowId);
      sew.moveToTop(focus);
   }
   
   /**
    * Move the specified window to the bottom of the z-order, at the OS level.
    * <p>
    * This is a no-op. Swing handle this automatically.
    *
    * @param    windowId
    *           The window to be moved.
    */
   @Override
   public void moveToBottom(int windowId)
   {
      SwingEmulatedWindow sew = (SwingEmulatedWindow) direct.getWindowEmulator(windowId);
      sew.moveToBottom();
   }
   
   /**
    * Sets the current selection.
    * 
    * @param txt
    *        The current selection.
    */
   @Override
   public void setCurrentSelection(String txt)
   {
      // not needed for clipboard support in Swing
   }

   /**
    * Load mouse pointer from the InputStream
    * 
    * @param data
    *        cursor data
    * @param ptrName
    *        pointer name
    *        
    * @return
    *        mouse pointer or <code>null</code>
    */
   @Override
   public MousePtrWrapper createMousePointer(InputStream data, String ptrName)
   {
      try
      {
         if (data == null)
         {
            return null;
         }
         BufferedImage image = makeImage(data);
         Object hotSpotProperty = image.getProperty("cursor_hotspot");
         java.awt.Point hotSpot = hotSpotProperty instanceof java.awt.Point ?
         (java.awt.Point) hotSpotProperty : new java.awt.Point(0,0);

         return MousePtrWrapper.createStaticSwingInstance(
         java.awt.Toolkit.getDefaultToolkit().createCustomCursor(image, hotSpot, ptrName)
         );
      }
      catch (Exception e)
      {
         LOG.log(Level.WARNING, "Error loading mouse pointer.", e);
      }
      return null;
   }

   /**
    * Set the mouse cursor
    * 
    * @param ptr
    *        mouse cursor
    */
   @Override
   public void setCursor(MousePtrWrapper ptr)
   {
      if (ptr.isPredefined())
      {
         ((SwingEmulatedWindow) ews).setCursor(ptr.type());
      }
      else
      {
         ((SwingEmulatedWindow) ews).setCursor(ptr.cursor());
      }
   }

   /**
    * Set the mouse cursor to the given type.
    * 
    * @param    type
    *           The cursor type.
    */
   public void setCursor(CursorType type)
   {
      ((SwingEmulatedWindow) ews).setCursor(type);
   }

   /**
    * Set the mouse cursor to the given one.
    * 
    * @param    c
    *           The custom cursor.
    */
   public void setCursor(Cursor c)
   {
      ((SwingEmulatedWindow) ews).setCursor(c);
   }
   
   /**
    * Tests if the target font is known to JRE.
    * 
    * @param    fontName
    *           The target font name.
    * @param    boldFont
    *           Indicates that the bold font is queried.
    * 
    * @return   True if the target font is installed, otherwise false.
    */
   @Override
   public boolean isFontInstalled(String fontName, boolean boldFont)
   {
      GraphicsEnvironment ge = GraphicsEnvironment.getLocalGraphicsEnvironment();
      // Enumerates all physical fonts known to JRE 
      String[] fontFamilies = ge.getAvailableFontFamilyNames();
      
      return (Arrays.asList(fontFamilies).indexOf(fontName) > -1);
   }

   /**
    * Returns true if a window can be decorated in this graphics environment, otherwise false.
    * 
    * @return   true if a window can have its title, caption buttons and border,
    *           otherwise false.
    */
   @Override
   public boolean isWindowDecorationSupported()
   {
      return true;
   }

   /**
    * Returns current mouse cursor display position relative to the display origin.
    *
    * @return mouse cursor position, never <code>null</code>
    */
   @Override
   public NativePoint getMousePosition()
   {
      Point p = MouseInfo.getPointerInfo().getLocation();
      return new NativePoint(p.x, p.y);
   }

   /**
    * The method delivers document output to the client. The supplied consumer is invoked by
    * the driver providing a suitable output stream. The consumer implementation is responsible
    * for sending the data to the provided stream. The consumer implementation must not close
    * the provided stream. When the finalizer is supplied, it will be called by the driver when
    * the document output is no longer needed. The finalizer should free up any additional
    * resources the document output may need.
    *
    * <p>
    * This implementation outputs the document data to a file. The output directory is given by
    * {@link #getPrintOutputDir()} and file name by {@link #getPrintOutputFile(MediaType format)}.
    *
    * @param   consumer
    *          Output consumer, must not be null. See above.
    * @param   finalizer
    *          Finalizer callback, may be null. See above.
    * @param   format
    *          Report output format. 
    */
   @Override
   public void deliverDocumentOutput(Consumer<OutputStream> consumer, 
                                     Runnable               finalizer, 
                                     MediaType format)
   {
      try
      {
         File outFile = new File(getPrintOutputDir(), getPrintOutputFile(format));
         try(OutputStream os = new FileOutputStream(outFile))
         {
            consumer.accept(os);
         }
         
         openMimeResource(format.getMimeType(), outFile.toURI().normalize().toASCIIString(), true);
      }
      catch (IOException e)
      {
         throw new RuntimeException(e);
      }
      finally
      {
         if (finalizer != null)
         {
            finalizer.run();
         }
      }
   }

   /**
    * Opens the specified resource in a browser's window.
    * 
    * @param    mimeType
    *           The resource mime type.
    * @param    urlString
    *           The string presentation of the target resource url.
    * @param    embedded
    *           The document is embedded in the template document.
    * 
    * @throws   MalformedURLException
    *           If the string presentation of the target resource url is malformed. 
    */
   @Override
   public void openMimeResource(String mimeType, String urlString, boolean embedded)
   throws MalformedURLException
   {
      // test that urlString is well-formed. 
      URL url = new URL(urlString);
      
      URI uri;
      
      try
      {
         uri = url.toURI();
      }
      catch (URISyntaxException ex)
      {
         throw new MalformedURLException(ex.getMessage());
      }
      
      if (embedded)
      {
         try
         {
            uri = buildOpenResourcePage(mimeType, urlString);
         }
         catch (IOException ex)
         {
            throw new RuntimeException(ex);
         }
      }
      
      try
      {
         openDocument(uri);
      }
      catch (IOException ex)
      {
         throw new RuntimeException(ex);
      }
   }

   /**
    * Makes the supplied window to be on top of other windows. This method is used to
    * implement the top-only functionality directly in the driver when
    * {@link #stackWindows(int[])} is not fully supported by the driver.
    *
    * @param   windowId
    *          The window to be changed.
    * @param   topOnly
    *          Corresponds to the WINDOW:TOP-ONLY attribute.
    * @param   alwaysOnTop
    *          Corresponds to the WINDOW:ALWAYS-ON-TOP attribute.
    */
   @Override
   public void setAlwaysOnTop(int windowId, boolean topOnly, boolean alwaysOnTop)
   {
      SwingEmulatedWindow sew = (SwingEmulatedWindow) direct.getWindowEmulator(windowId);
      sew.setAlwaysOnTop(topOnly, alwaysOnTop);
   }

   /**
    * Returns the directory for print output. This directory will store the files output from the
    * internal printer. The directory path is read from client:driver:print_output bootstrap
    * configuration item. If not specified or the configured path is no a valid directory the current
    * work directory will be used.
    *
    * @return  A {@link File} object representing the output directory.
    */
   private File getPrintOutputDir()
   {
      String val = config.getString("client", "driver", "print_output", null);

      File dir = null;
      if (val != null)
      {
         dir = new File(val);
         if (!dir.isDirectory())
         {
            LOG.warning("client:driver:print_output must point to a valid directory. " +
                        "Using the current work directory instead.");
            dir = null;
         }
      }

      if (dir == null)
      {
         dir = Paths.get(".").toFile();
      }

      return dir;
   }
   
   /**
    * Builds the html document that embeds the target resource given by its url path and its mime
    * type.
    * 
    * @param    mimeType
    *           The resource mime type.
    * @param    url
    *           The url of the target resource.
    * 
    * @return   The absolute file system path to the new created html document given by file URI.
    * 
    * @throws   IOException
    *           Iff the target IO operation is failed.
    */
   private URI buildOpenResourcePage(String mimeType, String url)
   throws IOException
   {
      String templatePagePath = Utils.calculateResourcePath(
                                                      "/",
                                                      OpenResourceTemplate.class.getPackage(),
                                                      OpenResourceTemplate.OPEN_RESOURCE_TEMPLATE);
      
      InputStream templatePage = getClass().getResourceAsStream(templatePagePath);
      
      File doc = new File(getPrintOutputDir(), getPrintOutputFile(MediaType.HTML));
      
      PrintWriter writer = new PrintWriter(doc);
      
      Map<String, Supplier<String>> keysProvider = new LinkedHashMap<>();
      
      keysProvider.put("isStreamed",
             () -> {
                      return "false";
                   });
      keysProvider.put("documentTitle",
             () -> {
                      return doc.getAbsolutePath();
                   });
      keysProvider.put("documentPath",
             () -> {
                      return url;
                   });
      keysProvider.put("documentType",
             () -> {
                      return mimeType;
                   });

      TemplateHelper.fill(templatePage, keysProvider, writer);

      writer.flush();

      writer.close();

      return doc.toURI();
   }

   /**
    * Open the target document given by its URI.
    * 
    * @param    uri
    *           The document URI
    * 
    * @throws   IOException
    *           Iff the target IO operation is failed
    */
   private void openDocument(URI uri)
   throws IOException
   {
      if (Desktop.isDesktopSupported())
      {
         if ("file".equalsIgnoreCase(uri.getScheme()))
         {
            Desktop.getDesktop().open(new File(uri));
         }
         else
         {
            Desktop.getDesktop().browse(uri);
         }
      }
      else
      {
         // uri has a correct syntax
         Directory dir = DirectoryManager.getInstance();
         
         // find the default browse settings in the directory
         String    brw = dir.getString(Directory.ID_RELATIVE, "browser-executable", "");
         
         String[] args = new BrowserHelper().getCmdLine(brw);
         
         // args[0] is the browser executable and args[1] represents its options
         ProcessBuilder pb = new ProcessBuilder(args[0], args[1] + " " + uri.toASCIIString());
         
         pb.start();
      }
   }
   
   /**
    * Parses the provided command line and returns the executable and its options.
    */
   private static class BrowserHelper
   {
      private static String OS = System.getProperty("os.name");
      
      /** the default browser executable */
      private String defExe;
      
      /** the default browser options */
      private String defOptions;
      
      /**
       * BrowserHelper initializes the default browser and its options.
       */
      public BrowserHelper()
      {
         defOptions = "";
         if (isWin())
         {
            defExe = "iexplore.exe";
         }
         else if (isMac())
         {
            defExe     = "open";
            defOptions = "-a safari";
         }
         else if (isUnix() || isSolaris())
         {
            defExe = "firefox";
         }
      }
      
      /**
       * Parses the command line to extract an executable and its options.
       * 
       * @param    commandLine
       *           
       * @return   The string array filled with the browser executable and its options or null
       *           if the provided command line is empty.
       */
      private String[] parse(String commandLine)
      {
         if (commandLine == null)
         {
            return null;
         }
         String defCmdLine = commandLine.trim();
         if (!defCmdLine.isEmpty())
         {
            int index = defCmdLine.indexOf(' ');
            String executable;
            String options;
            if (index > 0)
            {
               executable = defCmdLine.substring(0, index);
               options    = defCmdLine.substring(index + 1);
            }
            else
            {
               executable = defCmdLine;
               options    = "";
            }
            return new String[] { executable, options };
         }
         
         return null;
      }
      
      /**
       * Parses the command line and returns the browser executable and its options. If it isn't
       * provided and returns the default OS browser with its options.
       *  
       * @param    commandLine
       *           The command line to be parsed.
       * @return   The string array filled with the browser executable and its options.
       * @throws   UnsupportedOperationException if the command line is not provided and the
       *           given operation system isn't supported.
       */
      public String[] getCmdLine(String commandLine)
      {
         String[] args = parse(commandLine);
         if (args != null)
         {
            return args;
         }
         
         if (defExe == null || defOptions == null)
         {
            throw new UnsupportedOperationException("Unsupported OS = " + OS);
         }
         
         return new String[] {defExe, defOptions};
      }
      
      /**
       * Tests if the operation system belongs the Windows os family.
       * 
       * @return   true if os is Windows.
       */
      private static boolean isWin()
      {
         return OS.indexOf("win") >= 0;
      }
      
      /**
       * Tests if the operation system belongs the Mac os family.
       * 
       * @return   true if os is Mac.
       */
      private static boolean isMac()
      {
         return OS.indexOf("mac") >= 0;
      }
      
      /**
       * Tests if the operation system belongs the Unix os family.
       * 
       * @return   true if os is Unix.
       */
      private static boolean isUnix()
      {
         return OS.indexOf("nix") >= 0 || OS.indexOf("nux") >= 0 || OS.indexOf("aix") > 0;
      }
      
      /**
       * Tests if the operation system belongs the Solaris os family.
       * 
       * @return   true if os is Solaris.
       */
      private static boolean isSolaris()
      {
         return OS.indexOf("sunos") >= 0;
      }
   }
}
