/*
** Module   : OpenResourceHandler.java
** Abstract : Open resource handler is responsible to deliver target resources to clients.
**
** Copyright (c) 2018, Golden Code Development Corporation.
**
** -#- -I- --Date-- ---------------------------------Description----------------------------------
** 001 SBI 20180402 Initial version.
*/
/*
** This program is free software: you can redistribute it and/or modify
** it under the terms of the GNU Affero General Public License as
** published by the Free Software Foundation, either version 3 of the
** License, or (at your option) any later version.
**
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
** GNU Affero General Public License for more details.
**
** You may find a copy of the GNU Affero GPL version 3 at the following
** location: https://www.gnu.org/licenses/agpl-3.0.en.html
**
** Additional terms under GNU Affero GPL version 3 section 7:
**
**   Under Section 7 of the GNU Affero GPL version 3, the following additional
**   terms apply to the works covered under the License.  These additional terms
**   are non-permissive additional terms allowed under Section 7 of the GNU
**   Affero GPL version 3 and may not be removed by you.
**
**   0. Attribution Requirement.
**
**     You must preserve all legal notices or author attributions in the covered
**     work or Appropriate Legal Notices displayed by works containing the covered
**     work.  You may not remove from the covered work any author or developer
**     credit already included within the covered work.
**
**   1. No License To Use Trademarks.
**
**     This license does not grant any license or rights to use the trademarks
**     Golden Code, FWD, any Golden Code or FWD logo, or any other trademarks
**     of Golden Code Development Corporation. You are not authorized to use the
**     name Golden Code, FWD, or the names of any author or contributor, for
**     publicity purposes without written authorization.
**
**   2. No Misrepresentation of Affiliation.
**
**     You may not represent yourself as Golden Code Development Corporation or FWD.
**
**     You may not represent yourself for publicity purposes as associated with
**     Golden Code Development Corporation, FWD, or any author or contributor to
**     the covered work, without written authorization.
**
**   3. No Misrepresentation of Source or Origin.
**
**     You may not represent the covered work as solely your work.  All modified
**     versions of the covered work must be marked in a reasonable way to make it
**     clear that the modified work is not originating from Golden Code Development
**     Corporation or FWD.  All modified versions must contain the notices of
**     attribution required in this license.
*/

package com.goldencode.p2j.ui.client.gui.driver.web;


import java.net.*;
import java.util.*;
import java.util.function.*;

import org.eclipse.jetty.server.*;

import com.goldencode.p2j.cfg.*;
import com.goldencode.p2j.ui.client.gui.driver.OpenResourceTemplate;
import com.goldencode.p2j.web.*;


/**
 * Defines a resource handler to deliver target resources to clients.
 */
public class OpenResourceHandler
extends HtmlResourceHandler
implements OpenResourceTemplate
{
   /** The target request */
   private static final String REQUEST   = "/open/resource/";
   
   /** Configuration. */
   private BootstrapConfig config;
   
   /**
    * Creates this resource handler.
    * 
    * @param    config
    *           The bootstrap configuration
    * @param    skeletonPage
    *           Fully qualified resource name of the HTML file which should be used.
    */
   public OpenResourceHandler(BootstrapConfig config, String skeletonPage)
   {
      super(REQUEST, skeletonPage);
      this.config = config;
   }

   /**
    * Generates a map of template keys with their value suppliers in order to fill gaps in
    * the html skeleton template page.
    * 
    * @param    base
    *           The request
    * 
    * @return   The target map of template keys with their value suppliers
    */
   @Override
   protected Map<String, Supplier<String>> getTemplateKeys(Request base)
   {
      OpenResourceKeysProvider provider = new OpenResourceKeysProvider(base);
      
      return provider.getTemplateKeys();
   }
   
   /**
    * Builds the target map of keys and their value providers for the associated template
    * open-resource.html.
    */
   class OpenResourceKeysProvider
   extends HtmlTemplateKeysProvider
   {
      /**
       * Creates this key and value provider.
       * 
       * @param    base
       *           The request
       */
      public OpenResourceKeysProvider(Request base)
      {
         super(base);
         
         add("isStreamed",
             () -> {
                      String value = getRequest().getParameter(DOCUMENT_STREAMED);
                      
                      if (value == null || value.isEmpty())
                      {
                         return "false";
                      }
                      else
                      {
                         return "true";
                      }
                   });
         add("webRoot",
             () -> {
                      return config.getString("client", "web", "webRoot", "");
                   });
         add("documentTitle",
             () -> {
                      return decodeUrlEncodedParameter(DOCUMENT_TITLE, "");
                   });
         add("documentHandler",
             () -> {
                      return decodeUrlEncodedParameter(DOCUMENT_HANDLER, "document");
                   });
         add("documentPath",
             () -> {
                      return decodeUrlEncodedParameter(DOCUMENT_PATH, "");
                   });
         add("documentType",
             () -> {
                      return decodeUrlEncodedParameter(DOCUMENT_TYPE, "application/octet-stream");
                   });

      }
      
      /**
       * Decodes url encoded parameter given by its name if it is found in the request, otherwise
       * returns the given default value. If "UTF-8" is unsupported, then it catches exception silently and
       * returns unchanged parameter's value.
       * 
       * @param    name
       *           The parameter name
       * @param    defValue
       *           The default value
       * 
       * @return   The decoded parameter value if this decoding operation is succeeded
       */
      private String decodeUrlEncodedParameter(String name, String defValue)
      {
         String value = getRequest().getParameter(name);
         
         if (value == null)
         {
            return defValue;
         }
         
         try
         {
            value = URLDecoder.decode(value, "UTF-8");
         }
         catch (Exception ex)
         {
            // catch silently
         }
         
         return value;
      }
   }
}