/*
** Module   : RawEncoder.java
** Abstract : Implements Raw encoding.
**
** Copyright (c) 2015-2017, Golden Code Development Corporation.
**
** -#- -I- --Date-- --------------------------------Description----------------------------------
** 001 SBI 20150807 First version which describes a raw encoding of a rectangle
**                  on a provided virtual screen based on sRGB model.
*/
/*
** This program is free software: you can redistribute it and/or modify
** it under the terms of the GNU Affero General Public License as
** published by the Free Software Foundation, either version 3 of the
** License, or (at your option) any later version.
**
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
** GNU Affero General Public License for more details.
**
** You may find a copy of the GNU Affero GPL version 3 at the following
** location: https://www.gnu.org/licenses/agpl-3.0.en.html
** 
** Additional terms under GNU Affero GPL version 3 section 7:
** 
**   Under Section 7 of the GNU Affero GPL version 3, the following additional
**   terms apply to the works covered under the License.  These additional terms
**   are non-permissive additional terms allowed under Section 7 of the GNU
**   Affero GPL version 3 and may not be removed by you.
** 
**   0. Attribution Requirement.
** 
**     You must preserve all legal notices or author attributions in the covered
**     work or Appropriate Legal Notices displayed by works containing the covered
**     work.  You may not remove from the covered work any author or developer
**     credit already included within the covered work.
** 
**   1. No License To Use Trademarks.
** 
**     This license does not grant any license or rights to use the trademarks
**     Golden Code, FWD, any Golden Code or FWD logo, or any other trademarks
**     of Golden Code Development Corporation. You are not authorized to use the
**     name Golden Code, FWD, or the names of any author or contributor, for
**     publicity purposes without written authorization.
** 
**   2. No Misrepresentation of Affiliation.
** 
**     You may not represent yourself as Golden Code Development Corporation or FWD.
** 
**     You may not represent yourself for publicity purposes as associated with
**     Golden Code Development Corporation, FWD, or any author or contributor to
**     the covered work, without written authorization.
** 
**   3. No Misrepresentation of Source or Origin.
** 
**     You may not represent the covered work as solely your work.  All modified
**     versions of the covered work must be marked in a reasonable way to make it
**     clear that the modified work is not originating from Golden Code Development
**     Corporation or FWD.  All modified versions must contain the notices of
**     attribution required in this license.
*/


package com.goldencode.p2j.ui.client.gui.driver.web;

import java.awt.image.DirectColorModel;

import com.goldencode.p2j.ui.client.gui.driver.VirtualScreen;

/**
 * RawEncoder implements Raw encoding based on RFB specification documents:
 * https://www.realvnc.com/docs/rfbproto.pdf
 * https://tools.ietf.org/html/rfc6143
 * This implementation supports only sRGB pixels models with 32 bits per a pixel.
 */
class RawEncoder
implements PixelsEncoder
{
   /** Defines a 32 bits per a pixel integer constant. */
   private final static int TRUE_COLOR_BITS_PER_PIXEL = 32;

   /**
    * Encode the rectangular area on the screen to transfer.
    * 
    * @param    screen
    *           The virtual screen with the defined color model.
    * @param    x
    *           The position of the left side of the target rectangle along the screen horizon.
    * @param    y
    *           The position of the upper side of the target rectangle along the screen's
    *           vertical.
    * @param    w
    *           The width of the target rectangular area.
    * @param    h
    *           The height of the target rectangular area.
    * 
    * @return   The encoded message contents.
    */
   @Override
   public byte[] packToBinaries(VirtualScreen screen, int x, int y, int w, int h)
   {
      byte[] encodedRectangle = null;
      int b = 0;
      int i = 0;
      int s = 0;
      int pixel;
      int numberOfPixels = w * h;
      int scanline = screen.getScreenWidth();
      int jump = scanline - w;
      int p = y * scanline + x;
      int[] pixels = screen.getPixels();
      DirectColorModel colorModel = screen.getColorModel();
      if (colorModel.getPixelSize() == TRUE_COLOR_BITS_PER_PIXEL)
      {
         encodedRectangle = new byte[numberOfPixels << 2];
         for (; i < numberOfPixels; i++, s++, p++)
         {
            if (s == w)
            {
               s = 0;
               p += jump;
            }
            pixel = pixels[p];
            encodedRectangle[b++] = (byte) colorModel.getRed(pixel);
            encodedRectangle[b++] = (byte) colorModel.getGreen(pixel);
            encodedRectangle[b++] = (byte) colorModel.getBlue(pixel);
            encodedRectangle[b++] = (byte) colorModel.getAlpha(pixel);
         }
      }
      else
      {
         throw new UnsupportedOperationException("Supported only 32 bits per a pixel");
      }

      return encodedRectangle;
   }
}