/*
** Module   : p2j.chips.js
** Abstract : Defines a chips component based on Google Material Components.
**
** Copyright (c) 2018, Golden Code Development Corporation.
**
** -#- -I- --Date-- -------------------------------Description-----------------------------------
** 001 CA  20180723 Created initial version.
** 002 CA  20180801 Do not hide the chip's parent.
** 003 CA  20180816 Allow the chip interaction to be executed externally.
*/
/*
** This program is free software: you can redistribute it and/or modify
** it under the terms of the GNU Affero General Public License as
** published by the Free Software Foundation, either version 3 of the
** License, or (at your option) any later version.
**
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
** GNU Affero General Public License for more details.
**
** You may find a copy of the GNU Affero GPL version 3 at the following
** location: https://www.gnu.org/licenses/agpl-3.0.en.html
** 
** Additional terms under GNU Affero GPL version 3 section 7:
** 
**   Under Section 7 of the GNU Affero GPL version 3, the following additional
**   terms apply to the works covered under the License.  These additional terms
**   are non-permissive additional terms allowed under Section 7 of the GNU
**   Affero GPL version 3 and may not be removed by you.
** 
**   0. Attribution Requirement.
** 
**     You must preserve all legal notices or author attributions in the covered
**     work or Appropriate Legal Notices displayed by works containing the covered
**     work.  You may not remove from the covered work any author or developer
**     credit already included within the covered work.
** 
**   1. No License To Use Trademarks.
** 
**     This license does not grant any license or rights to use the trademarks
**     Golden Code, FWD, any Golden Code or FWD logo, or any other trademarks
**     of Golden Code Development Corporation. You are not authorized to use the
**     name Golden Code, FWD, or the names of any author or contributor, for
**     publicity purposes without written authorization.
** 
**   2. No Misrepresentation of Affiliation.
** 
**     You may not represent yourself as Golden Code Development Corporation or FWD.
** 
**     You may not represent yourself for publicity purposes as associated with
**     Golden Code Development Corporation, FWD, or any author or contributor to
**     the covered work, without written authorization.
** 
**   3. No Misrepresentation of Source or Origin.
** 
**     You may not represent the covered work as solely your work.  All modified
**     versions of the covered work must be marked in a reasonable way to make it
**     clear that the modified work is not originating from Golden Code Development
**     Corporation or FWD.  All modified versions must contain the notices of
**     attribution required in this license.
*/

define(function ()
{
   "use strict";
   
   var me = {};

   /** The chipset component. */
   var chipSetComponent;
   
   /** The ID of the DOM element where to attach the chips. */
   var chipSetContainer;

   /** The callback executed when a chip is removed. */
   var removeCallback;
   
   /** The callback executed when a chip is clicked. */
   var interactCallback;

   /**
    * Initialize the chips.
    *
    * @param    {string} vchipContainer
    *           The ID of the DOM element where to attach the chips.
    * @param    {function} vremoveCallback
    *           The callback executed when a chip is clicked.
    * @param    {function} vinteractCallback
    *           The callback executed when a chip is removed.
    */
   me.init = function(vchipContainer, vremoveCallback, vinteractCallback)
   {
      chipSetContainer = vchipContainer;
      removeCallback = vremoveCallback;
      interactCallback = vinteractCallback;
   };
   
   /**
    * Callback executed when a chip is removed.
    *
    * @param    {event} evt
    *           The event.
    */
   function removeChip(evt)
   {
      var root = evt.detail.root;
      var chipSetEl = document.getElementById(chipSetContainer);
      root && chipSetEl.removeChild(root);
      var chipId = $(evt.target).attr("chipId");
      
      removeCallback(chipId);
   };
   
   /**
    * Callback executed when a chip is clicked.
    *
    * @param    {event} evt
    *           The event.
    */
   function interactChip(evt)
   {
      var chipId = $(evt.target).attr("chipId");

      interactCallback(chipId);
   };
   
   /**
    * Executed the interact callback.
    *
    * @param    {number} chipId
    *           The chip ID.
    */
   me.interact = function(chipId)
   {
      interactCallback(chipId);
   };
   
   /**
    * Select the specified chip. If there are no chips, then the container is hidden.
    *
    * @param    {number} seleId
    *           The selected chip ID. 
    */
   function selectChip(seleId)
   {
      var chips = chipSetComponent.chips;
   
      for (var i = 0; i < chips.length; i++)
      {
         var chip = chips[i];
         var chipId = $(chip.root_).attr("chipId");
         var sele = seleId != undefined && seleId == chipId;
         chip.getDefaultFoundation().setSelected(sele);
      }
   };
   
   /**
    * Create the chips and select the specified one.
    * <p>
    * For each call, this is a comprehensive list - it will remove all existing chips which are
    * not found in the received list, update all the labels for existing chips and add any new
    * chips.
    *
    * @param    {object} chipList
    *           An array with the chip details, with keys the chip ID and the value its text.
    * @param    {number} seleId
    *           The selected chip ID. 
    */
   function createChips(chipList, seleId)
   {
      if (chipSetComponent == undefined)
      {
         var chipSetEl = document.getElementById(chipSetContainer);
         if (chipSetEl == undefined)
         {
            return;
         }
         
         chipSetEl.addEventListener('MDCChip:removal', removeChip);
         chipSetEl.addEventListener('MDCChip:interaction', interactChip);
         chipSetComponent = mdc.chips.MDCChipSet.attachTo(chipSetEl);
      }

      if (chipSetComponent == undefined)
      {
         return;
      }
   
      var chipSetEl = $("#" + chipSetContainer);
      // remove chips which don't exist in chipList
      var chips = chipSetComponent.chips;
      var existingChips = {};
      var changed = true;
      while (changed)
      {
         changed = false;
         for (var i = 0; i < chips.length; i++)
         {
            var chip = chips[i];
            var chipId = $(chip.root_).attr("chipId");
            if (existingChips[chipId] == undefined && 
                (chipId == undefined || chipList[chipId] == undefined))
            {
               chip.getDefaultFoundation().setSelected(false);
               chip.destroy();
               $(chip.root_).remove();
               chips.splice(i, 1);
               changed = true;
               break;
            }
            else
            {
               existingChips[chipId] = chip;
            }
         }
      }
   
      for (var chipId in chipList)
      {
         if (existingChips[chipId] != undefined)
         {
            continue;
         }
   
         // the chip remove will not work if the icon is moved to leading instead of trailing.
         var chipDiv = "<div class=\"mdc-chip\" tabindex=\"0\">";
         var chipText = "<div class=\"mdc-chip__text\">" + chipList[chipId] + "</div>";
         var chipImg = "<i class=\"material-icons mdc-chip__icon mdc-chip__icon--trailing\" tabindex=\"0\" role=\"button\">cancel</i>";
         chipDiv = chipDiv + chipText + chipImg + "</div>";
   
         var chipEl = $(chipDiv);
         chipEl.attr("chipId", chipId);
         chipEl = chipEl.get()[0];
   
         chipSetEl.append(chipEl);
         chipSetComponent.addChip(chipEl);
      }
   
      // update labels
      for (var chipId in existingChips)
      {
         var chipLabel = chipList[chipId];
         var chip = existingChips[chipId];
         $(chip.root_).children(".mdc-chip__text").text(chipLabel);
      }
   
      selectChip(seleId);
   };

   me.createChips = createChips;
   me.selectChip = selectChip;

   return me;
});
