/*
** Module   : p2j.menu.classic.js
** Abstract : Defines a simple, classic menu which can be used by the embedded application to 
**            expose a legacy window menubar via a HTML UI.
**
** Copyright (c) 2017-2018, Golden Code Development Corporation.
**
** -#- -I- --Date-- -------------------------------Description-----------------------------------
** 001 CA  20171006 Created initial version.
** 002 CA  20180313 Removed some console logging.
*/
/*
** This program is free software: you can redistribute it and/or modify
** it under the terms of the GNU Affero General Public License as
** published by the Free Software Foundation, either version 3 of the
** License, or (at your option) any later version.
**
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
** GNU Affero General Public License for more details.
**
** You may find a copy of the GNU Affero GPL version 3 at the following
** location: https://www.gnu.org/licenses/agpl-3.0.en.html
** 
** Additional terms under GNU Affero GPL version 3 section 7:
** 
**   Under Section 7 of the GNU Affero GPL version 3, the following additional
**   terms apply to the works covered under the License.  These additional terms
**   are non-permissive additional terms allowed under Section 7 of the GNU
**   Affero GPL version 3 and may not be removed by you.
** 
**   0. Attribution Requirement.
** 
**     You must preserve all legal notices or author attributions in the covered
**     work or Appropriate Legal Notices displayed by works containing the covered
**     work.  You may not remove from the covered work any author or developer
**     credit already included within the covered work.
** 
**   1. No License To Use Trademarks.
** 
**     This license does not grant any license or rights to use the trademarks
**     Golden Code, FWD, any Golden Code or FWD logo, or any other trademarks
**     of Golden Code Development Corporation. You are not authorized to use the
**     name Golden Code, FWD, or the names of any author or contributor, for
**     publicity purposes without written authorization.
** 
**   2. No Misrepresentation of Affiliation.
** 
**     You may not represent yourself as Golden Code Development Corporation or FWD.
** 
**     You may not represent yourself for publicity purposes as associated with
**     Golden Code Development Corporation, FWD, or any author or contributor to
**     the covered work, without written authorization.
** 
**   3. No Misrepresentation of Source or Origin.
** 
**     You may not represent the covered work as solely your work.  All modified
**     versions of the covered work must be marked in a reasonable way to make it
**     clear that the modified work is not originating from Golden Code Development
**     Corporation or FWD.  All modified versions must contain the notices of
**     attribution required in this license.
*/

define(function ()
{
   "use strict";
   
   var me = {};

   /** The ID of the DOM element where to attach the menu. */
   var menuContainer;
   
   /**
    * The callback executed when a menu-related event is performed (menu-drop, choose or 
    * value-changed).
    */
   var menuEventCallback;
   
   /** The menu-bar object. */
   var pmb;
   
   /**
    * Initialize the menu.
    *
    * @param    {string} vmenuContainer
    *           The ID of the DOM element where to attach the menu.
    * @param    {function} vmenuEventCallback
    *           The callback executed when a menu-related event is performed (menu-drop, choose or 
    *           value-changed).
    */
   me.init = function(vmenuContainer, vmenuEventCallback)
   {
      menuContainer = vmenuContainer;
      menuEventCallback = vmenuEventCallback;
   };

   /*
    * Destroy the menubar.
    */
   me.destroy = function()
   {
      if (pmb)
      {
         pmb.destroy();
         pmb = undefined;
      }
   };

   /*
    * Create a menubar and attach it, with the specified details.
    */
   me.create = function(jsonMenubar)
   {
      if (pmb)
      {
         pmb.destroy();
      }
   
      require(["dijit/MenuBar",
               "dijit/PopupMenuBarItem",
               "dijit/Menu",
               "dijit/MenuItem",
               "dijit/MenuBarItem",
               "dijit/DropDownMenu",
               "dijit/MenuSeparator",
               "dijit/CheckedMenuItem",
               "dojo/domReady!"],
              function(MenuBar,
                       PopupMenuBarItem,
                       Menu,
                       MenuItem,
                       MenuBarItem,
                       DropDownMenu,
                       MenuSeparator,
                       CheckedMenuItem)
              {
                 me.MenuBar          = MenuBar;
                 me.PopupMenuBarItem = PopupMenuBarItem;
                 me.Menu             = Menu;
                 me.MenuItem         = MenuItem;
                 me.MenuBarItem      = MenuBarItem;
                 me.DropDownMenu     = DropDownMenu;
                 me.MenuSeparator    = MenuSeparator;
                 me.CheckedMenuItem  = CheckedMenuItem;
                 
                 pmb = new me.MenuBar({});
                 buildMenuItems(pmb, jsonMenubar["menu-items"], true);
                 pmb.placeAt(menuContainer);
                 pmb.startup();
              });
   };
   
   /**
    * Build the menubar.
    * 
    * @param    {object} mparent
    *           The parent where to attach the menu items.
    * @param    {array} menuItems
    *           The menu-items details.
    * @param    {boolean} isMenuBarParent
    *           Flag indicating if the parent is actually the menubar, not a sub-menu.
    */
   function buildMenuItems(mparent, menuItems, isMenuBarParent)
   {
      if (menuItems == undefined)
      {
         return;
      }
      
      for (var i = 0; i < menuItems.length; i++)
      {
         var mhandle = menuItems[i]["handle"];
         var mtype   = menuItems[i]["type"];
         var mcfg    = buildMenuConfig(menuItems[i]);
         
         var mchild;
         
         if (mtype == "SUB-MENU")
         {
            mchild = new me.DropDownMenu(
            {
               "handle" : mhandle,
               "onOpen" : function() 
                          { 
                             if (menuEventCallback)
                             {
                                menuEventCallback(this.params.handle); 
                             }
                          }
            });
            
            buildMenuItems(mchild, menuItems[i]["menu-items"], false);
            
            mcfg["popup"] = mchild;
            mchild = new me.PopupMenuBarItem(mcfg);
         }
         else
         {
            var msubtype = menuItems[i]["subtype"];
            if (msubtype == "RULE")
            {
               mchild = new me.MenuSeparator();
            }
            else if (msubtype == "SKIP")
            {
               mchild = new me.MenuItem(mcfg);
            }
            else
            {
               if (menuItems[i]["toggle-box"] == true)
               {
                  if (menuItems[i]["checked"] == true)
                  {
                     mcfg["checked"] = true;
                  }
                  
                  mchild = new me.CheckedMenuItem(mcfg);
               }
               else if (isMenuBarParent)
               {
                  mchild = new me.MenuBarItem(mcfg);
               }
               else
               {
                  mchild = new me.MenuItem(mcfg);
               }
            }
         }

         mparent.addChild(mchild);
      }
   };
   
   /*
    * Build the configuration for this menubar element.
    *
    * @param    {object} menuItem
    *           The menu configuration.
    *
    * @return   An object with the dijit specific configuration for this menubar element.
    */
   function buildMenuConfig(menuItem)
   {
      var menuLabel = (menuItem["label"] ? menuItem["label"] : " ");;
      
      var mcfg = {};
      mcfg["handle"] = menuItem["handle"];

      if (menuItem["sensitive"] == false)
      {
         mcfg["disabled"] = true;
      }

      if (menuItem["accelerator"])
      {
         mcfg["accelKey"] = menuItem["accelerator"];
      }
      
      if (menuItem["mnemonic"])
      {
         var p1 = "";
         var p2 = "";
         var p3 = "";
         
         var mnemIdx = parseInt(menuItem["mnemonic"]);
         if (mnemIdx > 0)
         {
            p1 = menuLabel.substring(0, mnemIdx);
         }
         p2 = menuLabel.substring(mnemIdx, mnemIdx + 1);
         if (mnemIdx < menuLabel.length - 1)
         {
            p3 = menuLabel.substring(mnemIdx + 1);
         }
         
         menuLabel = p1 + "{" + p2 + "}" + p3;
      }
      mcfg["label"] = menuLabel;

      mcfg["onClick"] = 
      function() 
      { 
         if (menuEventCallback)
         {
            menuEventCallback(this.params.handle); 
         }
      };
      
      return mcfg;
   };
   
   return me;
});