/*
** Module   : p2j.module.classic.js
** Abstract : Defines classic module navigation widget.
**
** Copyright (c) 2018, Golden Code Development Corporation.
**
** -#- -I- --Date-- -------------------------------Description-----------------------------------
** 001 CA  20180301 Created initial version.
*/
/*
** This program is free software: you can redistribute it and/or modify
** it under the terms of the GNU Affero General Public License as
** published by the Free Software Foundation, either version 3 of the
** License, or (at your option) any later version.
**
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
** GNU Affero General Public License for more details.
**
** You may find a copy of the GNU Affero GPL version 3 at the following
** location: https://www.gnu.org/licenses/agpl-3.0.en.html
** 
** Additional terms under GNU Affero GPL version 3 section 7:
** 
**   Under Section 7 of the GNU Affero GPL version 3, the following additional
**   terms apply to the works covered under the License.  These additional terms
**   are non-permissive additional terms allowed under Section 7 of the GNU
**   Affero GPL version 3 and may not be removed by you.
** 
**   0. Attribution Requirement.
** 
**     You must preserve all legal notices or author attributions in the covered
**     work or Appropriate Legal Notices displayed by works containing the covered
**     work.  You may not remove from the covered work any author or developer
**     credit already included within the covered work.
** 
**   1. No License To Use Trademarks.
** 
**     This license does not grant any license or rights to use the trademarks
**     Golden Code, FWD, any Golden Code or FWD logo, or any other trademarks
**     of Golden Code Development Corporation. You are not authorized to use the
**     name Golden Code, FWD, or the names of any author or contributor, for
**     publicity purposes without written authorization.
** 
**   2. No Misrepresentation of Affiliation.
** 
**     You may not represent yourself as Golden Code Development Corporation or FWD.
** 
**     You may not represent yourself for publicity purposes as associated with
**     Golden Code Development Corporation, FWD, or any author or contributor to
**     the covered work, without written authorization.
** 
**   3. No Misrepresentation of Source or Origin.
** 
**     You may not represent the covered work as solely your work.  All modified
**     versions of the covered work must be marked in a reasonable way to make it
**     clear that the modified work is not originating from Golden Code Development
**     Corporation or FWD.  All modified versions must contain the notices of
**     attribution required in this license.
*/

define(function ()
{
   "use strict";
   
   var me = {};

   /** The ID of the DOM element where to attach the module. */
   var moduleContainer;
   
   /** The callback executed when a event is performed. */
   var moduleEventCallback;
   
   /** The module object. */
   var pmodule;
   
   /**
    * Initialize the module.
    *
    * @param    {string} vmoduleContainer
    *           The ID of the DOM element where to attach the module.
    * @param    {function} vmoduleEventCallback
    *           The callback executed when an event is performed.
    */
   me.init = function(vmoduleContainer, vmoduleEventCallback)
   {
      moduleContainer = vmoduleContainer;
      moduleEventCallback = vmoduleEventCallback;
   };

   /*
    * Destroy the module.
    */
   me.destroy = function()
   {
      if (pmodule)
      {
         pmodule.destroy();
         pmodule = undefined;
      }
   };

   /*
    * Create a module and attach it, with the specified details.
    *
    * @param   {array} jsonModule
    *          The array with module specifications.
    */
   me.create = function(jsonModule)
   {
      if (pmodule)
      {
         pmodule.destroy();
      }
   
      require(["dijit/Tree", 
               "dojo/data/ItemFileReadStore", 
               "dijit/tree/ForestStoreModel", 
               "dojo/domReady!"],
              function(Tree, ItemFileReadStore, ForestStoreModel)
              {
                 me.Tree = Tree;

                 buildModuleTree(jsonModule, 0);

                 var store = new ItemFileReadStore(
                 {
                    data : 
                    {
                       "identifier" : "id",
                       "label"      : "label",
                       "items"      : jsonModule 
                    }
                 });

                 var treeModel = new ForestStoreModel(
                 {
                     "store"         : store,
                     "query"         : { "parent" : 0 },
                     "childrenAttrs" : [ "items" ],
                     "rootId"        : "root",
                     "rootLabel"     : "Select a Module"
                 });
        
                 pmodule = new me.Tree(
                 {
                    model : treeModel,
                    betweenThreshold: 5,
                    onClick: function(item)
                    { 
                       if (item["items"]) return;
                       
                       moduleEventCallback(item.module[0]);
                    }
                 });
                 pmodule.placeAt(moduleContainer);
                 pmodule.startup();
              });
   };
   
   /**
    * Build the module tree, associating a parent ID to all items.
    * 
    * @param    {array} jsonModule
    *           The array tree with module specifications.
    * @param    {integer} parentId
    *           The parent ID.
    *
    * @return   {integer}
    *           The last used identifier.
    */
   function buildModuleTree(jsonModule, parentId)
   {
      if (jsonModule == undefined)
      {
         return parentId;
      }
      
      var id = parentId;
      for (var i = 0; i < jsonModule.length; i++)
      {
         id = id + 1;
         
         var cfg = jsonModule[i];
         cfg["id"] = id;
         cfg["parent"] = parentId;
         
         if (cfg["items"])
            id = buildModuleTree(cfg["items"], id);
      }
      
      return id;
   };
   
   return me;
});