/*
** Module   : p2j.toolbar.classic.js
** Abstract : Defines a replacement toolbar associated with a legacy a frame which contains 4GL
**            buttons and acts as a toolbar in the UI.
**
** Copyright (c) 2018, Golden Code Development Corporation.
**
** -#- -I- --Date-- -------------------------------Description-----------------------------------
** 001 CA  201802022 Created initial version.
*/
/*
** This program is free software: you can redistribute it and/or modify
** it under the terms of the GNU Affero General Public License as
** published by the Free Software Foundation, either version 3 of the
** License, or (at your option) any later version.
**
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
** GNU Affero General Public License for more details.
**
** You may find a copy of the GNU Affero GPL version 3 at the following
** location: https://www.gnu.org/licenses/agpl-3.0.en.html
** 
** Additional terms under GNU Affero GPL version 3 section 7:
** 
**   Under Section 7 of the GNU Affero GPL version 3, the following additional
**   terms apply to the works covered under the License.  These additional terms
**   are non-permissive additional terms allowed under Section 7 of the GNU
**   Affero GPL version 3 and may not be removed by you.
** 
**   0. Attribution Requirement.
** 
**     You must preserve all legal notices or author attributions in the covered
**     work or Appropriate Legal Notices displayed by works containing the covered
**     work.  You may not remove from the covered work any author or developer
**     credit already included within the covered work.
** 
**   1. No License To Use Trademarks.
** 
**     This license does not grant any license or rights to use the trademarks
**     Golden Code, FWD, any Golden Code or FWD logo, or any other trademarks
**     of Golden Code Development Corporation. You are not authorized to use the
**     name Golden Code, FWD, or the names of any author or contributor, for
**     publicity purposes without written authorization.
** 
**   2. No Misrepresentation of Affiliation.
** 
**     You may not represent yourself as Golden Code Development Corporation or FWD.
** 
**     You may not represent yourself for publicity purposes as associated with
**     Golden Code Development Corporation, FWD, or any author or contributor to
**     the covered work, without written authorization.
** 
**   3. No Misrepresentation of Source or Origin.
** 
**     You may not represent the covered work as solely your work.  All modified
**     versions of the covered work must be marked in a reasonable way to make it
**     clear that the modified work is not originating from Golden Code Development
**     Corporation or FWD.  All modified versions must contain the notices of
**     attribution required in this license.
*/

define(function ()
{
   "use strict";
   
   var me = {};

   /** The ID of the DOM element where to attach the toolbar. */
   var toolbarContainer;
   
   /** The callback executed when a button-related event is performed. */
   var toolbarEventCallback;
   
   /** The toolbar object. */
   var ptb;
   
   /**
    * Initialize the toolbar.
    *
    * @param    {string} vtoolbarContainer
    *           The ID of the DOM element where to attach the toolbar.
    * @param    {function} vtoolbarEventCallback
    *           The callback executed when a button-related event is performed.
    */
   me.init = function(vtoolbarContainer, vtoolbarEventCallback)
   {
      toolbarContainer = vtoolbarContainer;
      toolbarEventCallback = vtoolbarEventCallback;
   };

   /*
    * Destroy the toolbar.
    */
   me.destroy = function()
   {
      if (ptb)
      {
         ptb.destroy();
         ptb = undefined;
      }
   };

   /*
    * Create a toolbar and attach it, with the specified details.
    *
    * @param   {array} jsonToolbar
    *          The array with button specifications, which compose the toolbar.
    */
   me.create = function(jsonToolbar)
   {
      if (ptb)
      {
         ptb.destroy();
      }
   
      require(["dijit/Toolbar",
               "dijit/form/Button",
               "dijit/ToolbarSeparator",
               "dojo/domReady!"],
              function(Toolbar,
                       Button,
                       ToolbarSeparator)
              {
                 me.Toolbar          = Toolbar;
                 me.Button           = Button;
                 me.ToolbarSeparator = ToolbarSeparator;
                 
                 ptb = new me.Toolbar({});
                 buildToolbarItems(jsonToolbar);
                 ptb.placeAt(toolbarContainer);
                 ptb.startup();
              });
   };
   
   /**
    * Build the toolbar.
    * 
    * @param    {array} jsonToolbar
    *           The array with button specifications, which compose the toolbar.
    */
   function buildToolbarItems(jsonToolbar)
   {
      if (jsonToolbar == undefined)
      {
         return;
      }
      
      jsonToolbar.sort(function(a, b)
      {
         return a.x - b.x;
      });
      
      for (var i = 0; i < jsonToolbar.length; i++)
      {
         var bcfg = jsonToolbar[i];
         var bhandle = bcfg["handle"];

         var tooltip = '';
         var label = '';
         var image = undefined;
         
         if (bcfg['tooltip']) tooltip = bcfg['tooltip'];
         if (bcfg['label']) label = bcfg['label'];
         if (bcfg['image']) image = bcfg['image'];
         
         var cfg = 
         {
            "handle"    : bhandle,
            "onClick"   : function() 
            { 
               if (toolbarEventCallback)
               {
                  toolbarEventCallback(this.params.handle); 
               }
            }
         };
         
         if (!bcfg["sensitive"])
         {
            cfg["disabled"] = true;
         }
         
         if (image)
         {
            cfg["label"] = '<img style="height: 32px; width:32px"src="' + image + 
                           '" title="' + tooltip + '" alt="' + label + '"/>';
         }
         else
         {
            cfg["label"] = (label == "" ? tooltip : label);
         }
         
         var btn = new me.Button(cfg);
         
         ptb.addChild(btn);
      }
   };
   
   return me;
});