/*
** Module   : p2j.toolbar.js
** Abstract : Defines a replacement toolbar associated with a legacy frame which contains 4GL
**            buttons and acts as a toolbar in the UI.
**
** Copyright (c) 2018, Golden Code Development Corporation.
**
** -#- -I- --Date-- -------------------------------Description-----------------------------------
** 001 CA  20180419 Created initial version.
** 002 CA  20180801 Fixed disabled button state.
** 003 CA  20180816 Allow the images to be prefixed by a custom path.
*/
/*
** This program is free software: you can redistribute it and/or modify
** it under the terms of the GNU Affero General Public License as
** published by the Free Software Foundation, either version 3 of the
** License, or (at your option) any later version.
**
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
** GNU Affero General Public License for more details.
**
** You may find a copy of the GNU Affero GPL version 3 at the following
** location: https://www.gnu.org/licenses/agpl-3.0.en.html
** 
** Additional terms under GNU Affero GPL version 3 section 7:
** 
**   Under Section 7 of the GNU Affero GPL version 3, the following additional
**   terms apply to the works covered under the License.  These additional terms
**   are non-permissive additional terms allowed under Section 7 of the GNU
**   Affero GPL version 3 and may not be removed by you.
** 
**   0. Attribution Requirement.
** 
**     You must preserve all legal notices or author attributions in the covered
**     work or Appropriate Legal Notices displayed by works containing the covered
**     work.  You may not remove from the covered work any author or developer
**     credit already included within the covered work.
** 
**   1. No License To Use Trademarks.
** 
**     This license does not grant any license or rights to use the trademarks
**     Golden Code, FWD, any Golden Code or FWD logo, or any other trademarks
**     of Golden Code Development Corporation. You are not authorized to use the
**     name Golden Code, FWD, or the names of any author or contributor, for
**     publicity purposes without written authorization.
** 
**   2. No Misrepresentation of Affiliation.
** 
**     You may not represent yourself as Golden Code Development Corporation or FWD.
** 
**     You may not represent yourself for publicity purposes as associated with
**     Golden Code Development Corporation, FWD, or any author or contributor to
**     the covered work, without written authorization.
** 
**   3. No Misrepresentation of Source or Origin.
** 
**     You may not represent the covered work as solely your work.  All modified
**     versions of the covered work must be marked in a reasonable way to make it
**     clear that the modified work is not originating from Golden Code Development
**     Corporation or FWD.  All modified versions must contain the notices of
**     attribution required in this license.
*/

define(function ()
{
   "use strict";
   
   var me = {};

   /** The ID of the DOM element where to attach the toolbar. */
   var toolbarContainer;
   
   /** The callback executed when a button-related event is performed. */
   var toolbarEventCallback;

   /** A function which receives an image location and returns a HTML string. */
   var customImage;
   
   /** An optional folder to set as prefix to all image sources. */
   var imageFolder;

   /**
    * Initialize the toolbar.
    *
    * @param    {string} vtoolbarContainer
    *           The ID of the DOM element where to attach the toolbar.
    * @param    {function} vtoolbarEventCallback
    *           The callback executed when a button-related event is performed.
    * @param    {string} vimageFolder
    *           The folder to add as a prefix for all images.
    */
   me.init = function(vtoolbarContainer, vtoolbarEventCallback, vimageFolder)
   {
      toolbarContainer = $("#" + vtoolbarContainer);
      toolbarEventCallback = vtoolbarEventCallback;
      imageFolder = vimageFolder
   };

   /**
    * Destroy the toolbar.
    */
   me.destroy = function()
   {
      toolbarContainer.empty();
   };
   
   /**
    * Set the function used to compute a custom image element for a toolbar item.
    *
    * @param    {function} fn
    *           A function which receives an image location and returns a HTML string.
    */
   me.setCustomImage = function(fn)
   {
      customImage = fn;
   };

   /**
    * Create a toolbar and attach it, with the specified details.
    *
    * @param   {array} jsonToolbar
    *          The array with button specifications, which compose the toolbar.
    */
   me.create = function(jsonToolbar)
   {
      me.destroy();
   
      buildToolbarItems(jsonToolbar);
      
      if (toolbarEventCallback)
      {
         toolbarContainer.find("button").click(function(evt)
         {
            var a = $(this);
            if (a.attr("handle"))
            {
               toolbarEventCallback(a.attr("handle"));
            }
         });
      }
   };
   
   /**
    * Build the toolbar.
    * 
    * @param    {array} jsonToolbar
    *           The array with button specifications, which compose the toolbar.
    */
   function buildToolbarItems(jsonToolbar)
   {
      if (jsonToolbar == undefined)
      {
         return;
      }
      
      jsonToolbar.sort(function(a, b)
      {
         return a.x - b.x;
      });
      
      for (var i = 0; i < jsonToolbar.length; i++)
      {
         var bcfg = jsonToolbar[i];
         var bhandle = bcfg["handle"];

         var tooltip = '';
         var label = '';
         var image = undefined;
         
         if (bcfg['tooltip']) tooltip = bcfg['tooltip'];
         if (bcfg['label']) label = bcfg['label'];
         if (bcfg['image']) image = bcfg['image'];
         
         var btn = $("<button title='" + tooltip + "' class='mdc-button' id='" + bhandle + "'/>");
         btn.attr("handle", bhandle);
         if (!bcfg["sensitive"])
         {
            btn.prop("disabled", true);
         }

         if (image)
         {
            var imgSrc = image;
            if (imageFolder)
            {
               imgSrc = imageFolder + imgSrc;
            }
         
            var imgHtml = "<img style='height: 32px; width:32px' src='" + imgSrc + "' alt='" + label + "'/>";
            if (customImage)
            {
               var ihtml = customImage(image, tooltip, label);
               if (ihtml)
               {
                  imgHtml = ihtml;
               }
            }
            var img = $(imgHtml);
            btn.append(img);
         }
         else
         {
            btn.append(label == "" ? tooltip : label);
         }

         toolbarContainer.append(btn);
      }
   };
   
   return me;
});
