/*
** Module   : EmptyThumbView.java
** Abstract : The base class for all views shown in file chooser dialog.
**
** Copyright (c) 2017-2018, Golden Code Development Corporation.
**
** -#- -I- --Date-- ---------------------------------Description---------------------------------
** 001 OM  20171026 Initial commit.
** 002 OM  20180321 Added locate() and positional access methods. Added Theme and Edit support.
*/

/*
** This program is free software: you can redistribute it and/or modify
** it under the terms of the GNU Affero General Public License as
** published by the Free Software Foundation, either version 3 of the
** License, or (at your option) any later version.
**
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
** GNU Affero General Public License for more details.
**
** You may find a copy of the GNU Affero GPL version 3 at the following
** location: https://www.gnu.org/licenses/agpl-3.0.en.html
** 
** Additional terms under GNU Affero GPL version 3 section 7:
** 
**   Under Section 7 of the GNU Affero GPL version 3, the following additional
**   terms apply to the works covered under the License.  These additional terms
**   are non-permissive additional terms allowed under Section 7 of the GNU
**   Affero GPL version 3 and may not be removed by you.
** 
**   0. Attribution Requirement.
** 
**     You must preserve all legal notices or author attributions in the covered
**     work or Appropriate Legal Notices displayed by works containing the covered
**     work.  You may not remove from the covered work any author or developer
**     credit already included within the covered work.
** 
**   1. No License To Use Trademarks.
** 
**     This license does not grant any license or rights to use the trademarks
**     Golden Code, FWD, any Golden Code or FWD logo, or any other trademarks
**     of Golden Code Development Corporation. You are not authorized to use the
**     name Golden Code, FWD, or the names of any author or contributor, for
**     publicity purposes without written authorization.
** 
**   2. No Misrepresentation of Affiliation.
** 
**     You may not represent yourself as Golden Code Development Corporation or FWD.
** 
**     You may not represent yourself for publicity purposes as associated with
**     Golden Code Development Corporation, FWD, or any author or contributor to
**     the covered work, without written authorization.
** 
**   3. No Misrepresentation of Source or Origin.
** 
**     You may not represent the covered work as solely your work.  All modified
**     versions of the covered work must be marked in a reasonable way to make it
**     clear that the modified work is not originating from Golden Code Development
**     Corporation or FWD.  All modified versions must contain the notices of
**     attribution required in this license.
*/

package com.goldencode.p2j.ui.client.gui.filechooser;

import com.goldencode.p2j.ui.*;
import com.goldencode.p2j.ui.chui.*;
import com.goldencode.p2j.ui.client.*;
import com.goldencode.p2j.ui.client.event.listener.*;
import com.goldencode.p2j.ui.client.gui.*;
import com.goldencode.p2j.ui.client.gui.driver.*;
import com.goldencode.p2j.ui.client.widget.*;
import com.goldencode.p2j.util.*;
import java.io.*;
import java.nio.file.*;
import java.nio.file.attribute.*;
import java.util.*;
import java.util.logging.*;

/**
 * The base class for all views shown in file chooser dialog. It handles the drawing of an empty
 * set of file thumbnails, hence the name of the class.
 */
public abstract class EmptyThumbView
extends SensitiveScrollContainer<GuiOutputManager>
{
   /** Logger. */
   private final Logger LOG = LogHelper.getLogger(EmptyThumbView.class.getName());
   
   /** Cached driver. */
   protected GuiDriver gd = null;
   
   /** Font resolver */
   protected GuiFontResolver gf;
   
   /** The cached list of files. */
   protected List<Path> files = new ArrayList<>();
   
   /** 
    * The action listener that will handle the actions of the thumbnails when clicked. This is in
    * fact the parent {@link FileDialogGuiImpl} that will process all the actions from all
    * widgets.
    */
   protected ActionListener listener = null;
   
   /**
    * The constructor caches the graphic driver and font resolver for all views. 
    */
   EmptyThumbView()
   {
      setScrollable(true);

      gd = (GuiDriver<?,?>) OutputManager.getDriver();
      gf = new GuiFontResolver(WindowManager.resolveWindow(this), gd, config(), false);
   }
   
   /**
    * Query the preferred scroll types. Each type of view uses a different way of scrolling. In 
    * particular, the empty view does not require any scroll.
    * 
    * @return the preferred scroll type for this thumbnail view.  
    */
   public ScrollPaneGuiImpl.ShowBars getScrollType()
   {
      return ScrollPaneGuiImpl.ShowBars.NEVER;
   }
   
   /**
    * Query the preferred scroll bars. Each type of view uses a different way of scrolling. In 
    * particular, the empty view does not require any scroll.
    *
    * @return the preferred scroll type for this thumbnail view.  
    */
   public ScrollBar.Bars getScrollBars()
   {
      return ScrollBar.Bars.NONE;
   }
   
   /**
    * Check if widget is focus traversable.
    *
    * @return  Always {@code true}.
    */
   @Override
   public boolean focusTraversable()
   {
      return true;
   }
   
   /**
    * Configure the action listener that will be notified when user clicks on a file or directory 
    * thumbnail. 
    * 
    * @param   listener
    *          The new action listener.
    */
   public void setListener(ActionListener listener)
   {
      this.listener = listener;
   }
   
   /**
    * Repopulates the view with a new set of files. The old content is dropped, the old thumbnails
    * are detached and destroyed. A new layout is rebuilt using the new list of files. 
    * 
    * @param   files
    *          the new list of files.
    */
   public void populate(List<Path> files)
   {
      this.files.clear();
      
      // remove current widgets
      Widget[] widgets = widgets();
      for (Widget widget : widgets)
      {
         detach(widget);
         remove(widget);
      }
      
      // allow subclassed to rebuild their UIs:
      if (files != null && !files.isEmpty())
      {
         appendSearchResult(files);
      }
      else
      {
         doLayout();
      }
   }
   
   /**
    * Obtain the name of this view.
    * 
    * @return  the name of this view.
    */
   public abstract String getViewName();
   
   /**
    * Draw widget. Handles empty file list case. All derived classes use the super call to 
    * delegate to this method for this particular case.
    */
   @Override
   public void draw()
   {
      if (!isVisible())
      {
         return;
      }
      
      NativePoint p = physicalLocation();
      NativeDimension dim = physicalDimension();
      NativeRectangle clip = clipRectangle(p, dim);
      gd.draw(getId(), p, clip, () ->
      {
         // draw background
         gd.setColor(new ColorRgb(0xFFFFFF)); // white
         gd.fillRect(0, 0, dim.width, dim.height);
         
         if (files.isEmpty())
         {
            String label = "No items match your search.";
            gd.setColor(new ColorRgb(0x6D6D6D));
            FontDetails<?> fd = gf.font().font;
            gd.setGuiFont(fd);
            gd.setFontStyle(fd.style);
            int tY = 20;
            int tX = (dim.width - getTextWidthNative(label, gf.resolveFontNum())) / 2;
            gd.drawStringCentered(label, tX, tY);
         }
         else
         {
            // let the super class do its job:
            super.draw();
         }
      });
   }
   
   /**
    * Abstract factory method for creating file thumbnails depending on the current view.
    *
    * @param   id
    *          The widget's ID.
    * @param   cfg
    *          The widget's configuration.
    * @param   path
    *          The path this thumbnail represents.
    * @param   isDir
    *          Flags the thumbnails that represents directories.
    * @param   click2Command
    *          The action to be performed when the thumbnail is double clicked.
    */
   abstract FileThumb createThumb(WidgetId id,
                                  ButtonConfig cfg,
                                  Path path,
                                  boolean isDir,
                                  String click2Command);
   
   /**
    * Search for a {@code FileThumb} based on its name.
    * 
    * @param   name
    *          The name of file to search.
    *
    * @return  The {@code FileThumb} whose file name is {@code name} or {@code null} if
     *         such filer cannot be found among the contained files . 
    */
   FileThumb locate(String name)
   {
      // quick out if invalid name detected
      if (name == null || name.isEmpty())
      {
         return null;
      }
      
      for (Widget w : widgets())
      {
         if (w instanceof FileThumb)
         {
            FileThumb ft = (FileThumb) w;
            if (Objects.equals(ft.textLabel(), name))
            {
               return ft;
            }
         }
      }
      
      return null;
   }
   
   /**
    * Find the first to the right {@code FileThumb} relative to the {@code fileThumb}.
    * 
    * @param   fileThumb
    *          The currently selected file thumb.
    *
    * @return  The rightmost {@code FileThumb} from the {@code fileThumb}.
    */
   FileThumb findRight(FileThumb fileThumb)
   {
      FileThumb r = null;
      int rx = Integer.MAX_VALUE; 
      NativePoint p1 = fileThumb.physicalLocation();
      for (Widget<GuiOutputManager> w : widgets())
      {
         if (w != fileThumb && w instanceof FileThumb)
         {
            NativePoint p2 = w.physicalLocation();
            if (p1.y == p2.y && p1.x < p2.x && p2.x < rx)
            {
               r = (FileThumb) w;
               rx = p2.x;
            }
         }
      }
      
      return r;
   }
   
   /**
    * Find the first to the left {@code FileThumb} relative to the {@code fileThumb}.
    * 
    * @param   fileThumb
    *          The currently selected file thumb.
    *
    * @return  The leftmost {@code FileThumb} from the {@code fileThumb}.
    */
   FileThumb findLeft(FileThumb fileThumb)
   {
      FileThumb l = null;
      int lx = Integer.MIN_VALUE;
      NativePoint p1 = fileThumb.physicalLocation();
      for (Widget<GuiOutputManager> w : widgets())
      {
         if (w != fileThumb && w instanceof FileThumb)
         {
            NativePoint p2 = w.physicalLocation();
            if (p1.y == p2.y && p1.x > p2.x && p2.x > lx)
            {
               l = (FileThumb) w;
               lx = p2.x;
            }
         }
      }
      
      return l;
   }
   
   /**
    * Find the first {@code FileThumb} above to {@code fileThumb}.
    * 
    * @param   fileThumb
    *          The currently selected file thumb.
    *
    * @return  The {@code FileThumb} just above {@code fileThumb}.
    */
   FileThumb findUp(FileThumb fileThumb)
   {
      FileThumb u = null;
      int uy = Integer.MIN_VALUE;
      NativePoint p1 = fileThumb.physicalLocation();
      for (Widget<GuiOutputManager> w : widgets())
      {
         if (w != fileThumb && w instanceof FileThumb)
         {
            NativePoint p2 = w.physicalLocation();
            if (p1.x == p2.x && p1.y > p2.y && p2.y > uy)
            {
               u = (FileThumb) w;
               uy = p2.y;
            }
         }
      }
      
      return u;
   }
   
   /**
    * Find the first {@code FileThumb} below to {@code fileThumb}.
    * 
    * @param   fileThumb
    *          The currently selected file thumb.
    *
    * @return  The {@code FileThumb} just under {@code fileThumb}.
    */
   FileThumb findDown(FileThumb fileThumb)
   {
      FileThumb d = null;
      int dy = Integer.MAX_VALUE;
      NativePoint p1 = fileThumb.physicalLocation();
      for (Widget<GuiOutputManager> w : widgets())
      {
         if (w != fileThumb && w instanceof FileThumb)
         {
            NativePoint p2 = w.physicalLocation();
            if (p1.x == p2.x && p1.y < p2.y && p2.y < dy)
            {
               d = (FileThumb) w;
               dy = p2.y;
            }
         }
      }
      
      return d;
   }
   
   /**
    * Return the first {@code FileThumb} in this view.
    * 
    * @return  the first {@code FileThumb} in this view or {@code null} if there are no child
    *          thumbnails.
    */
   FileThumb findFirst()
   {
      for (Widget<GuiOutputManager> w : widgets())
      {
         if (w instanceof FileThumb)
         {
            return (FileThumb) w;
         }
      }
      
      return null;
   }
   
   /**
    * Add a subset of files to be listed in this view.
    * 
    * @param   paths
    *          The additional paths to be displayed.
    */
   void appendSearchResult(List<Path> paths)
   {
      for (Path path : paths)
      {
         String fileName = FileChooserUtils.getFileName(path);
         boolean vf = FileChooserUtils.isVirtualFolder(fileName);
         String label = FileChooserUtils.getDisplayName(path);
         boolean isDir = vf || Files.isDirectory(path);
         
         WidgetId bid = WidgetId.nextID();
         
         ButtonConfig bcfg = new ButtonConfig(bid);
         bcfg.widthPixels = -1;
         bcfg.heightPixels = -1;
         bcfg.isDefault = false;
         bcfg.autoResize = false;
         bcfg.fixedHeight = true;
         bcfg.fixedWidth = true;
         bcfg.frameId = config.frameId;
         
         FileThumb thumb = createThumb(bid, bcfg, path, isDir, "REL-NAVIGATE " + fileName);
         thumb.initialize(bid, bcfg);
         thumb.setSupportsMnemonic();
         thumb.setTextLabel(label);
         thumb.setCommand((isDir ? "SELDIR " : "SELECT ") + fileName);
         thumb.setEnabled(true);
         thumb.setVisible(true);
         thumb.addListener(listener);
         
         if (!vf && Files.exists(path))
         {
            try
            {
               BasicFileAttributes attr = Files.readAttributes(path, BasicFileAttributes.class);
               thumb.setFileInfo(attr.size(), attr.creationTime(), attr.lastModifiedTime());
            }
            catch (IOException e)
            {
               if (LOG.isLoggable(Level.WARNING))
               {
                  LOG.log(Level.WARNING, "Failed to access file attributes", e);
               }
            }
            
            String tooltip = thumb.getTooltip();
            thumb.config().tooltip = tooltip;
            if (tooltip != null && !tooltip.isEmpty())
            {
               ThinClient.setupTooltip(thumb);
            }
         }
         
         add(thumb);
      }
      
      this.files.addAll(paths);
      
      doLayout();
   }
}
