/*
** Module   : FileChooserUtils.java
** Abstract : A set of utility methods used within this package.
**
** Copyright (c) 2017-2018, Golden Code Development Corporation.
**
** -#- -I- --Date-- ---------------------------------Description---------------------------------
** 001 OM  20171026 Initial commit.
** 002 OM  20180321 Added support for Virtual Folders.
** 003 OM  20180503 Used Theme.getBinaryResourceLocation() to locate image resources.
*/

/*
** This program is free software: you can redistribute it and/or modify
** it under the terms of the GNU Affero General Public License as
** published by the Free Software Foundation, either version 3 of the
** License, or (at your option) any later version.
**
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
** GNU Affero General Public License for more details.
**
** You may find a copy of the GNU Affero GPL version 3 at the following
** location: https://www.gnu.org/licenses/agpl-3.0.en.html
** 
** Additional terms under GNU Affero GPL version 3 section 7:
** 
**   Under Section 7 of the GNU Affero GPL version 3, the following additional
**   terms apply to the works covered under the License.  These additional terms
**   are non-permissive additional terms allowed under Section 7 of the GNU
**   Affero GPL version 3 and may not be removed by you.
** 
**   0. Attribution Requirement.
** 
**     You must preserve all legal notices or author attributions in the covered
**     work or Appropriate Legal Notices displayed by works containing the covered
**     work.  You may not remove from the covered work any author or developer
**     credit already included within the covered work.
** 
**   1. No License To Use Trademarks.
** 
**     This license does not grant any license or rights to use the trademarks
**     Golden Code, FWD, any Golden Code or FWD logo, or any other trademarks
**     of Golden Code Development Corporation. You are not authorized to use the
**     name Golden Code, FWD, or the names of any author or contributor, for
**     publicity purposes without written authorization.
** 
**   2. No Misrepresentation of Affiliation.
** 
**     You may not represent yourself as Golden Code Development Corporation or FWD.
** 
**     You may not represent yourself for publicity purposes as associated with
**     Golden Code Development Corporation, FWD, or any author or contributor to
**     the covered work, without written authorization.
** 
**   3. No Misrepresentation of Source or Origin.
** 
**     You may not represent the covered work as solely your work.  All modified
**     versions of the covered work must be marked in a reasonable way to make it
**     clear that the modified work is not originating from Golden Code Development
**     Corporation or FWD.  All modified versions must contain the notices of
**     attribution required in this license.
*/

package com.goldencode.p2j.ui.client.gui.filechooser;

import com.goldencode.p2j.ui.client.gui.driver.*;
import com.goldencode.p2j.ui.client.gui.theme.*;
import com.goldencode.p2j.util.*;
import com.goldencode.p2j.util.LogHelper;
import com.goldencode.util.*;
import javax.imageio.*;
import java.awt.image.*;
import java.io.*;
import java.nio.file.*;
import java.nio.file.attribute.*;
import java.text.*;
import java.util.*;
import java.util.logging.*;

/**
 * This class is a container for utility methods used throughout this package.
 */
abstract class FileChooserUtils
{
   /** Logger. */
   private static final Logger LOG = LogHelper.getLogger(FileChooserUtils.class.getName());

   /** Favorites virtual folder. */
   static final String VF_FAVORITES = "$FAVORITES";
    
   /** Recent Places virtual folder. */
   static final String VF_RECENTPLACES = "$RECENTPLACES";
    
   /** This PC virtual folder. */
   static final String VF_THISPC = "$THISPC";
      
   /** Network virtual folder. */
   static final String VF_NETWORK = "$NETWORK";
   
   /** Libraries virtual folder. */
   static final String VF_LIBRARIES = "$LIBRARIES";
   
   /** Control Panel virtual folder. */
   static final String VF_CONTROLPANEL = "$CONTROL-PANEL";
   
   /** Recycle Bin virtual folder. */
   static final String VF_RECYCLEBIN = "$RECYCLE-BIN";
   
   /**
    * Draws an image (if valid) at specified coordinates. No scaling and no alignment is
    * performed. 
    * 
    * @param   gd
    *          The graphic driver to be used.
    * @param   img
    *          The image to be rendered.
    * @param   x
    *          X coordinate of the location where the image is drawn.
    * @param   y
    *          Y coordinate of the location where the image is drawn.
    */
   static void drawImage(GuiDriver gd, ImageWrapper<BufferedImage> img, int x, int y)
   {
      if (img == null)
      {
         return;
      }
      gd.drawImage(img, x, y, img.getWidth(), img.getHeight(), 0, 0, false, false, true, false);
   }
   
   /**
    * Utility method for loading an image resource from packaging bundle and return its binary
    * content as a byte array. The resource must be located in a special directory
    * {@code <theme-name>/filedialog}. This allows theme-dependent icons to be created.
    *
    * @param   path
    *          The path of the resource. It is computed relative to current theme.
    *
    * @return  a byte array containing the binary resource. If the named resource could not be
    *          find or in case of an exception, the {@code null} array is returned.
    */
   static byte[] getBinaryResource(String path)
   {
      String imgPath = ThemeManager.getCurrentTheme().getBinaryResourceLocation() + 
                       "/filedialog/" + path + ".png";
      InputStream res = Theme.class.getResourceAsStream(imgPath);
      if (res != null)
      {
         byte[] ret = null;
         try
         {
            ret = new byte[res.available()];
            res.read(ret);
            return ret;
         }
         catch (IOException e)
         {
            // TODO: create a dummy image ?
         }
      }
      
      LOG.log(Level.WARNING, "Failed to load resource: [" + imgPath+ "].");
      return null;
   }
   
   /**
    * Loads a image resource. The image is located relative to current theme. The resource must be
    * located in a special directory {@code <theme-name>/filedialog}. This allows theme-dependent
    * icons to be created.
    * <p>
    * The resource name must end in .png. Although the PNG file format is not required (any other
    * well known format will be automatically used by {@code ImageIO}) it is recommended because
    * it support full 32 bit images (including 8 bit alpha) and they are super compact in size.
    *
    * @param   icon
    *          The image to be loaded.
    *
    * @return  An {@code ImageWrapper} with the image resource loaded or {@code null} if resource
    *          was  not found or it could not be decoded. 
    */
   static ImageWrapper<BufferedImage> loadImage(String icon)
   {
      String imgPath = ThemeManager.getCurrentTheme().getBinaryResourceLocation() + 
                       "/filedialog/" + icon + ".png";
      InputStream res = Theme.class.getResourceAsStream(imgPath);
      if (res != null)
      {
         try
         {
            BufferedImage image = ImageIO.read(res);
            return new ImageWrapper<>(image, image.getHeight(), image.getWidth());
         }
         catch (IOException e)
         {
            // TODO: create a dummy image ?
         }
      }
      
      LOG.log(Level.WARNING, "Failed to load resource: [" + imgPath+ "].");
      return null;
   }
   
   /**
    * Convert a long number representing a file size into a string using multiplicative binary
    * units. The names are not ISO compatible, but this is how Windows prints the values.
    * 
    * @param   size
    *          The file size.
    * @param   forceUnits
    *          If {@code true} forces KB prefix even if {@code size} is less than 1 kilobyte
    * @param   decimals
    *          How many decimals the result should have. Valid values are 0 to 3. 
    *
    * @return  The string representation of {@code size} with eventual multiplication prefix.
    */
   static String getSizeAsString(long size, boolean forceUnits, int decimals)
   {
      if (size < 1024)
      {
         return forceUnits ? floatToString(size / 1024f, decimals) + " KB"
                           : Long.toString(size) + " bytes";
      }
      float s = size;
      String[] MX = {" KB", " MB", " GB", " TB"};
      
      for (String mx : MX)
      {
         s = s / 1024;
         if (s < 1024)
         {
            return floatToString(s, decimals) + mx;
         }
      }
      
      // this is really huge:
      s = s / 1024;
      return floatToString(s, decimals) + " PB";
   }
   
   /**
    * Helper worker method for {@code getSizeAsString()}. This method uses {@link DecimalFormat}
    * to format {@code f} to specified number of decimals.
    * 
    * @param   f
    *          The decimal number to convert.
    * @param   decimals
    *          The numbers of decimals.
    * 
    * @return  The string representation of {@code f} with specified number of decimals. 
    */
   private static String floatToString(float f, int decimals)
   {
      String format;
      switch (decimals)
      {
         default:
         case 0: format = "###0"; break;
         case 1: format = "###0.0"; break;
         case 2: format = "###0.00"; break;
         case 3: format = "###0.000"; break;
      }
      return new DecimalFormat(format).format(f);
   }
   
   /**
    * Utility method that converts a file timestamp to its string representation. This method uses
    * the {@code 99/99/9999 HH:MM} format of {@link datetime}.
    * 
    * @param   fileTime
    *          The timestamp to be converted.
    *          
    * @return  The string representation of the file's timestamp.
    */
   static String getDateAsString(FileTime fileTime)
   {
      if (fileTime == null)
      {
         return "N/A"; // a timestamp will be null for virtual folders
      }
      datetime dFWD = new datetime(new Date(fileTime.toMillis()));
      return dFWD.toString("99/99/9999 HH:MM");
   }
   
   /**
    * Tests whether the {@code path} is visible when navigating the file tree. This is OS
    * dependent: if we are running on Windows OS, we need to check the appropriate file attributes
    * (Hidden and System), when running on Linux and Solaris the files starting with a {@code .}
    * are not visible.
    * 
    * @param   path
    *          The path to test.
    *
    * @return  {@code true} if path should be visible in file selection dialogs. 
    */
   static boolean isPathVisible(Path path)
   {
      if (PlatformHelper.isUnderWindowsFamily())
      {
         try
         {
            DosFileAttributes dfa = Files.readAttributes(path, DosFileAttributes.class);
            return !dfa.isHidden() && !dfa.isSystem();
         }
         catch (IOException e)
         {
            return false; // hide file if access fail 
         }
      }
      else 
      {
         return !path.getFileName().toString().startsWith(".");
      }
   }
   
   /**
    * Worker method for extracting the file name from a {@code Path} as a {@code String} and avoid
    * a NPE in case of root paths.
    *
    * @param   path
    *          The path to be analyzed.
    *
    * @return  file name from {@code path}.
    */
   static String getFileName(Path path)
   {
      Path fileName = path.getFileName();
      return fileName == null ? path.toString() : fileName.toString();
   }
   
   /**
    * Obtain the name of the file to be displayed, including virtual folders.
    * 
    * @param   path
    *          An absolute path or virtual folder name.
    * 
    * @return  A string that represents the path, for showing to user.
    */
   static String getDisplayName(Path path)
   {
      Path fileName = path.getFileName();
      if (fileName == null)
      {
         return path.toString(); 
      }
      
      return getDisplayName(fileName.toString());
   }   
   /**
    * Obtain the name of the file to be displayed, including virtual folders.
    * 
    * @param   absPath
    *          An absolute path or virtual folder name.
    * 
    * @return  A string that represents the path, for showing to user.
    */
   static String getDisplayName(String absPath)
   {
      switch (absPath)
      {
         case VF_FAVORITES: return "Favorites";
         case VF_RECENTPLACES: return "Recent Places";
         case VF_THISPC: return "This PC";
         case VF_NETWORK: return "Network";
         case VF_LIBRARIES: return "Libraries";
         case VF_CONTROLPANEL: return "Control Panel";
         case VF_RECYCLEBIN: return "Recycle Bin";
         default: return getFileName(Paths.get(absPath));
      }
   }
   
   /**
    * Checks whether the absolute or relative path is a virtual folder.
    * 
    * @param   path
    *          The path to check.
    *
    * @return  {@code true} if the path provided is a virtual folder id.
    */
   static boolean isVirtualFolder(String path)
   {
      if (!path.startsWith("$"))
      {
         return false; 
      }
      
      switch (path)
      {
         case VF_FAVORITES:
         case VF_RECENTPLACES:
         case VF_THISPC:
         case VF_NETWORK:
         case VF_LIBRARIES:
         case VF_CONTROLPANEL:
         case VF_RECYCLEBIN:
            return true;
         default:
            return false;
      }
   }
}
