/*
** Module   : FileThumb.java
** Abstract : The base class for all file thumbs shown in file chooser dialog.
**
** Copyright (c) 2017-2018, Golden Code Development Corporation.
**
** -#- -I- --Date-- ---------------------------------Description---------------------------------
** 001 OM  20171026 Initial commit.
** 002 OM  20180321 Added select(), keyboard navigation. Added Theme and Edit support.
*/

/*
** This program is free software: you can redistribute it and/or modify
** it under the terms of the GNU Affero General Public License as
** published by the Free Software Foundation, either version 3 of the
** License, or (at your option) any later version.
**
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
** GNU Affero General Public License for more details.
**
** You may find a copy of the GNU Affero GPL version 3 at the following
** location: https://www.gnu.org/licenses/agpl-3.0.en.html
** 
** Additional terms under GNU Affero GPL version 3 section 7:
** 
**   Under Section 7 of the GNU Affero GPL version 3, the following additional
**   terms apply to the works covered under the License.  These additional terms
**   are non-permissive additional terms allowed under Section 7 of the GNU
**   Affero GPL version 3 and may not be removed by you.
** 
**   0. Attribution Requirement.
** 
**     You must preserve all legal notices or author attributions in the covered
**     work or Appropriate Legal Notices displayed by works containing the covered
**     work.  You may not remove from the covered work any author or developer
**     credit already included within the covered work.
** 
**   1. No License To Use Trademarks.
** 
**     This license does not grant any license or rights to use the trademarks
**     Golden Code, FWD, any Golden Code or FWD logo, or any other trademarks
**     of Golden Code Development Corporation. You are not authorized to use the
**     name Golden Code, FWD, or the names of any author or contributor, for
**     publicity purposes without written authorization.
** 
**   2. No Misrepresentation of Affiliation.
** 
**     You may not represent yourself as Golden Code Development Corporation or FWD.
** 
**     You may not represent yourself for publicity purposes as associated with
**     Golden Code Development Corporation, FWD, or any author or contributor to
**     the covered work, without written authorization.
** 
**   3. No Misrepresentation of Source or Origin.
** 
**     You may not represent the covered work as solely your work.  All modified
**     versions of the covered work must be marked in a reasonable way to make it
**     clear that the modified work is not originating from Golden Code Development
**     Corporation or FWD.  All modified versions must contain the notices of
**     attribution required in this license.
*/

package com.goldencode.p2j.ui.client.gui.filechooser;

import com.goldencode.p2j.ui.*;
import com.goldencode.p2j.ui.client.*;
import com.goldencode.p2j.ui.client.event.*;
import com.goldencode.p2j.ui.client.event.ActionEvent;
import com.goldencode.p2j.ui.client.gui.*;
import com.goldencode.p2j.ui.client.gui.driver.*;
import com.goldencode.p2j.ui.client.gui.theme.*;
import com.goldencode.p2j.ui.client.widget.*;
import com.goldencode.p2j.util.*;
import java.awt.event.*;
import java.awt.image.*;
import java.nio.file.*;
import java.nio.file.attribute.*;

/**
 * This is the base class for all file thumbs shown in file chooser dialog. It groups the common
 * properties of the thumbnail of each view mode from file chooser.
 */
public abstract class FileThumb
extends ButtonGuiImpl
{
   /**
    * The icon to be displayed.
    */
   protected final ImageWrapper<BufferedImage> img;
   
   /**
    * Cached file/folder size.
    */
   protected long fileSize;
   
   /**
    * Cached file/folder creation timestamp.
    */
   protected FileTime created;
   
   /**
    * Cached last modified timestamp.
    */
   protected FileTime lastModified;
   
   /**
    * The action to be performed when the thumbnail is double clicked.
    */
   private final String doubleClickCommand;
   
   /**
    * Flag for a double-click event.
    */
   private boolean doubleClick = false;
   
   /**
    * Flags the thumbnails that represents directories.
    */
   private final boolean isDir;
   
   /**
    * The path this thumbnail represents.
    */
   private final Path path;
   
   /**
    * The currently selected thumbnail.
    */
   private static FileThumb selected = null;
   
   /**
    * Constructor used to create component using given configuration.
    *
    * @param id            The widget's ID.
    * @param cfg           The widget's configuration.
    * @param path          The path this thumbnail represents.
    * @param image         The image to be displayed on the thumbnail.
    * @param isDir         Flags the thumbnails that represents directories.
    * @param click2Command The action to be performed when the thumbnail is double clicked.
    */
   FileThumb(WidgetId id,
             ButtonConfig cfg,
             Path path,
             String image,
             boolean isDir,
             String click2Command)
   {
      super(id, cfg);
      this.img = FileChooserUtils.loadImage(image);
      this.doubleClickCommand = click2Command;
      this.isDir = isDir;
      this.path = path;
   }
   
   /**
    * Query the preferred width of this thumbnail. It is used by some thumbnail containers in
    * order to do the internal layout. If width is not important, the thumbnail can return -1.
    * 
    * @return  the preferred width of this thumbnail.
    */
   public int getPreferredWidth()
   {
      return -1;
   }
   
   /**
    * Query the preferred height of this thumbnail. It is used by some thumbnail containers in
    * order to do the internal layout. If height is not important, the thumbnail can return -1.
    * 
    * @return  the preferred width of this thumbnail.
    */
   public int getPreferredHeight()
   {
      return -1;
   }
   
   /**
    * Obtain the path this thumbnail represents.
    * 
    * @return  the path this thumbnail represents.
    */
   public Path getPath()
   {
      return path;
   }
   
   /**
    * Notification of a mouse clicked event occurred for this widget.  Simulate ENTER key typing.
    * The main reason it is overwritten here is to set {@code doubleClick} if double click is
    * detected.
    *
    * @param    e
    *           The mouse event.
    */
   @Override
   public void mouseClicked(MouseEvent e)
   {
      if (selected != this)
      {
         FileThumb oldSel = selected;
         selected = this;
         if (oldSel != null)
         {
            oldSel.repaint();
         }
      }
      
      if (e.getClickCount() > 1)
      {
         doubleClick = true;
      }
      
      super.mouseClicked(e);
   }
   
   /**
    * Draw the thumbnail. Called by this Button's parent container. It makes sure the driver is
    * properly configured and clipped, draws the background and then calls {@code drawForeground}
    * for drawing the text/icon on the thumbnail.
    */
   @Override
   public void draw()
   {
      // get location and size
      NativePoint p = physicalLocation();
      NativeDimension d = physicalDimension();
      
      // set up clipping rectangle
      NativeRectangle clip = clipRectangle(p, d);
      
      // clippable drawing start 
      gd.draw(this.getId(), p, clip, () -> {
         Theme theme = ThemeManager.getCurrentTheme();
         theme.drawFileChooserThumbnailBackground(
               gd, 0, 0, d.width, d.height, isSelected(), isMouseOver(), hasFocus());
         drawForeground(d.width, d.height, theme);
      });
   }
   
   /**
    * Draws the text and icon on this thumbnail.
    * 
    * @param   width
    *          The width of the widget
    * @param   height
    *          The height of the widget.
    * @param   theme
    *          The current theme.
    */
   protected abstract void drawForeground(int width, int height, Theme theme);
   
   /**
    * Checks if this thumbnail is selected.
    * 
    * @return {@code true} if this thumbnail is selected.
    */
   public boolean isSelected()
   {
      return selected == this;
   }
   
   /**
    * Clears the selection flag. No thumbnail will be selected. 
    */
   public static void clearSelection()
   {
      selected = null;
   }
   
   /**
    * Query the currently selected thumbnail.
    * 
    * @return  the currently selected thumbnail.
    */
   public static FileThumb getSelected()
   {
      return selected;
   }
   
   /**
    * Checks whether this thumbnail represents a directory.
    * 
    * @return  {@code true} if this thumbnail represents a directory.
    */
   public boolean isDir()
   {
      return isDir;
   }
   
   /**
    * Obtain the file icon, if set.
    * 
    * @return  the file icon of this thumbnail.
    */
   public ImageWrapper<BufferedImage> getFileIcon()
   {
      return img;
   }
   
   /**
    * Sets the selected thumb.
    * 
    * @param   ft
    *          The new {@code FileThumb} to select. May be {@code null} to unselect.
    */
   public static void select(FileThumb ft)
   {
      if (ft == selected)
      {
         return;
      }
      
      FileThumb oldSel = selected;
      selected = ft;
      if (oldSel != null)
      {
         oldSel.repaint();
      }
      if (ft != null)
      {
         selected.repaint();
         selected.requestFocus();
      }
   }
   
   /**
    * Retrieves the action to be performed when the thumbnail is double clicked.
    * <p>
    * Note: this resets the {@code doubleClick} flag, so a subsequent call will return the simple
    *       click command.
    *
    * @return  the action to be performed when the thumbnail is double clicked and otherwise
    *          the simple-click command.
    */
   String getDoubleClickCommand()
   {
      if (doubleClick)
      {
         doubleClick = false;
         return doubleClickCommand;
      }
      else
      {
         return null;
      }
   }
   
   /**
    * Check whether the mouse was double-clicked on this thumbnail.
    *
    * @return  {@code true} if the mouse was double-clicked. 
    */
   boolean isDoubleClick()
   {
      return doubleClick;
   }
   
   /**
    * Sets information of the file that this thumbnail represents.
    * 
    * @param   fileSize
    *          The file size.
    * @param   created
    *          Creation timestamp.
    * @param   lastModified
    *          The timestamp of last modification of the file.
    */
   void setFileInfo(long fileSize, FileTime created, FileTime lastModified)
   {
      this.fileSize = fileSize;
      this.created = created;
      this.lastModified = lastModified;
   }
   
   /**
    * Process basic events - key navigation event.
    *
    * @param event Key event to process.
    */
   @Override
   protected void handleBasicEvents(KeyInput event)
   {
      boolean handled = false;
      if (this == selected)
      {
         EmptyThumbView view = (EmptyThumbView) parent();
         FileThumb newSel = null;
         switch (event.keyCode())
         {
            case Key.VK_ENTER:
               EventManager.postEvent(new ActionEvent(this, doubleClickCommand));
               handled = true;
               break;
               
            case Key.VK_ESCAPE:
               EventManager.postEvent(new ActionEvent(this, "ESCAPE"));
               handled = true;
               break;
               
            case Key.VK_F2:
               EventManager.postEvent(new ActionEvent(this, "RENAME"));
               handled = true;
               break;
               
            case Key.VK_BACKSPACE:
               EventManager.postEvent(new ActionEvent(this, "UP"));
               handled = true;
               break;
               
            case Key.VK_CURSOR_RIGHT:
               newSel = view.findRight(this);
               handled = true;
               break;
               
            case Key.VK_CURSOR_LEFT:
               newSel = view.findLeft(this);
               handled = true;
               break;
               
            case Key.VK_CURSOR_UP:
               newSel = view.findUp(this);
               handled = true;
               break;
               
            case Key.VK_CURSOR_DOWN:
               newSel = view.findDown(this);
               handled = true;
               break;
         }
         
         if (newSel != null)
         {
            selected = newSel;
            selected.requestFocus();
            EventManager.postEvent(new ActionEvent(selected, "SELECT " + selected.textLabel()));
         }
      }
      
      if (handled)
      {
         event.consume();
      }
      else
      {
         super.handleBasicEvents(event);
      }
   }
   
   /**
    * Returns a string representation of the object.
    *
    * @return   A string representation of the object.
    */
   @Override
   public String toString()
   {
      StringBuilder sb = new StringBuilder("FileThumb{");
      if (isDir)
      {
         sb.append("[");
      }
      sb.append(textLabel());
      if (isDir)
      {
         sb.append("]");
      }
      sb.append("}");
      return sb.toString();
   }
   
   /**
    * Obtain the tooltip to be displayed when hovering this thumbnail.
    *
    * @return  the tooltip to be displayed when hovering this thumbnail.
    */
   String getTooltip()
   {
      return null;
   }
   
   /**
    * Enter the rename mode for active thumbnail. 
    * 
    * @param   rename
    *          The line editor that will be used for renaming. 
    */
   public static void startRename(LineEditor rename)
   {
      if (selected == null || !selected.isEnabled() || !selected.isVisible())
      {
         // cannot rename if no thumbnail is selected, enables or visible 
         return;
      }
      
      // reparent the [rename] 
      Container<GuiOutputManager> oldParent = rename.parent();
      if (oldParent != null)
      {
         oldParent.remove(rename);
      }
      selected.parent().add(rename);
      rename.config().frameId = selected.config.frameId;
      rename.setValue(new character(selected.textLabel()));
      
      // adjust dimension and position
      NativePoint sp = selected.getEditorLocation();
      NativePoint ep = selected.physicalLocation();
      rename.setPhysicalLocation(sp.x + ep.x, sp.y + ep.y);
      NativeDimension dim = selected.getEditorDimension();
      rename.setSize(selected.cc.dimensionFromNative(dim));
      
      // ready, just show it
      rename.setEnabled(true);
      rename.setVisible(true);
      rename.requestFocus();
   }
   
   /**
    * Obtain the dimension of the editor that when renaming this thumbnail.  
    * 
    * @return  the dimension of the editor that when renaming this thumbnail.
    */
   protected abstract NativeDimension getEditorDimension();
   
   /**
    * Obtain the physical location of the editor that when renaming this thumbnail. The location
    * is relative to this widget.
    *
    * @return  the physical location of the editor that when renaming this thumbnail.
    */
   protected abstract NativePoint getEditorLocation();
}
