/*
** Module   : Windows10Theme.java
** Abstract : Windows 10 theme manager. 
**
** Copyright (c) 2017-2018, Golden Code Development Corporation.
**
** -#- -I- --Date-- ---------------------------------Description---------------------------------
** 001 OM  20170224 First version.
** 002 EVL 20170512 Adding method to get default font aliases.  Fix for legacy font names to be
**                  case sensitive.
** 003 OM  20170406 Added Rectangle and 3DBorder support. Fixed Tooltip support.
** 004 SVL 20170814 Fixed drawing of title bar for editable browse.
** 005 OM  20170705 Fixed Rectangle rendering regression. Other small patches.
**                  Added drawSelectionFocus() method.
** 006 SVL 20170901 Added getRowToggleOffset, getRowToggleSize, getRowToggleImage.
** 007 HC  20170823 Added runtime support of WINDOW:SMALL-TITLE.
** 008 OM  20171027 Small javadoc update.
** 009 EVL 20171211 Adding pop-up menu drawing issues with vertical separator line and width.
** 010 SVL 20180115 Added functions for getting offsets for a text inside browse cells and labels.
**                  Browse label quirk is properly implemented.
**     EVL 20180124 Adding browse label draw separation into theme specific calls.
** 011 SBI 20180130 Implemented getTextLookAndFeel for fillins, removed unused drawFillInSelection.
** 012 OM  20180321 Added multiline support for tooltips. Added support for file/dir choosers.
** 013 HC  20180509 Fixed drawing of menu items in the root of a MENU-BAR.
** 014 OM  20180503 Added getBinaryResourceLocation().
** 015 CA  20180502 Fixed COLOR option, COLOR statement and other color-related issues.
**     CA  20180530 Fixed unknown and empty FRAME:TITLE.
** 016 OM  20180710 Added support for Slider widgets.
** 017 SVL 20180529 Support for enhanced browse coloring.
** 018 OM  20180808 Fixed browse sub-components drawing on Windows8.
** 019 EVL 20180817 Fix for button feedback for SPACEBAR key.  The pressed state is possible even
**                  when the mouse is not over widget.
*/

/*
** This program is free software: you can redistribute it and/or modify
** it under the terms of the GNU Affero General Public License as
** published by the Free Software Foundation, either version 3 of the
** License, or (at your option) any later version.
**
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
** GNU Affero General Public License for more details.
**
** You may find a copy of the GNU Affero GPL version 3 at the following
** location: https://www.gnu.org/licenses/agpl-3.0.en.html
** 
** Additional terms under GNU Affero GPL version 3 section 7:
** 
**   Under Section 7 of the GNU Affero GPL version 3, the following additional
**   terms apply to the works covered under the License.  These additional terms
**   are non-permissive additional terms allowed under Section 7 of the GNU
**   Affero GPL version 3 and may not be removed by you.
** 
**   0. Attribution Requirement.
** 
**     You must preserve all legal notices or author attributions in the covered
**     work or Appropriate Legal Notices displayed by works containing the covered
**     work.  You may not remove from the covered work any author or developer
**     credit already included within the covered work.
** 
**   1. No License To Use Trademarks.
** 
**     This license does not grant any license or rights to use the trademarks
**     Golden Code, FWD, any Golden Code or FWD logo, or any other trademarks
**     of Golden Code Development Corporation. You are not authorized to use the
**     name Golden Code, FWD, or the names of any author or contributor, for
**     publicity purposes without written authorization.
** 
**   2. No Misrepresentation of Affiliation.
** 
**     You may not represent yourself as Golden Code Development Corporation or FWD.
** 
**     You may not represent yourself for publicity purposes as associated with
**     Golden Code Development Corporation, FWD, or any author or contributor to
**     the covered work, without written authorization.
** 
**   3. No Misrepresentation of Source or Origin.
** 
**     You may not represent the covered work as solely your work.  All modified
**     versions of the covered work must be marked in a reasonable way to make it
**     clear that the modified work is not originating from Golden Code Development
**     Corporation or FWD.  All modified versions must contain the notices of
**     attribution required in this license.
*/

package com.goldencode.p2j.ui.client.gui.theme;

import com.goldencode.p2j.ui.*;
import com.goldencode.p2j.ui.client.*;
import com.goldencode.p2j.ui.client.gui.*;
import com.goldencode.p2j.ui.client.gui.driver.*;
import com.goldencode.p2j.ui.client.gui.theme.Theme.TextLookAndFeel.*;
import com.goldencode.p2j.ui.client.widget.*;
import com.goldencode.p2j.util.*;

import javax.imageio.*;

import java.awt.Color;
import java.awt.image.*;
import java.io.*;
import java.util.*;

import static com.goldencode.p2j.ui.client.SysColor.*;

/**
 * This theme emulates the Windows 10 (1607 - Redstone 1) look.
 * <p>
 * The theme is very flat, most widgets components have minimal foreground with monochrome 
 * backgrounds. The theme is however, very sensitive to mouse hovering movements, almost all
 * components will highlight when the mouse pointer hovers over them. 
 */
public class Windows10Theme
extends ClassicTheme
{
   /** The color of the border when window is active. */
   private static final Color WINDOW_BORDER_COLOR_ACTIVE = new Color(0x3C8CCD);
   
   /** The color of the border when window is NOT active. */
   private static final Color WINDOW_BORDER_COLOR_INACTIVE = new Color(0xA5A6A6);
   
   /** The background color of the titlebar when window is active. */
   private static final Color WINDOW_TITLEBAR_COLOR_ACTIVE = new Color(0x2982CC);
   
   /** The background color of the titlebar when window is NOT active. */
   private static final Color WINDOW_TITLEBAR_COLOR_INACTIVE = new Color(0xFFFFFF);
   
   /** The foreground color of the titlebar when window is active. */
   private static final Color WINDOW_TITLE_COLOR_ACTIVE = new Color(0xFFFFFF);
   
   /** The foreground color of the titlebar when window is NOT active. */
   private static final Color WINDOW_TITLE_COLOR_INACTIVE = new Color(0xAAAAAA);
   
   /** The background color of the caption buttons when window is NOT active. */
   private static final Color CAPTION_BUTTON_COLOR_INACTIVE = new Color(0xFFFFFF);

   /** The background color of the small close caption button when window is NOT active. */
   private static final Color CAPTION_BUTTON_CLOSE_SMALL_COLOR_INACTIVE = new Color(0xE5E5E5);

   /** The background color of the caption buttons when window is active. */
   private static final Color CAPTION_BUTTON_COLOR_ACTIVE = new Color(0x2982CC);
   
   /** The background color of the caption buttons when mouse hover over it. */
   private static final Color CAPTION_BUTTON_COLOR_HIGHLIGHT = new Color(0x3F8FD1);
   
   /** The background color of the caption buttons when button is pushed down. */
   private static final Color CAPTION_BUTTON_COLOR_DOWN = new Color(0x559BD5);
   
   /** The background color of the CLOSE (X) caption buttons when mouse hover over it. */
   private static final Color CAPTION_BUTTON2_COLOR_HIGHLIGHT = new Color(0xE81123);
   
   /** The background color of the CLOSE (X) caption buttons when button is pushed down. */
   private static final Color CAPTION_BUTTON2_COLOR_DOWN = new Color(0x9B3E66);

   /** The background color of the small CLOSE (x) caption buttons when window is active. */
   private static final Color CAPTION_BUTTON_CLOSE_SMALL_ACTIVE = new Color(0xC75050);

   /** The height of the titlebar of normal window (NOT maximized). */
   private static final int TITLE_BAR_HEIGHT = 30;
   
   /** The height of the titlebar when window is maximized. */
   private static final int TITLE_BAR_HEIGHT_MAXIMIZED = 23;
   
   /** The height of the caption buttons in normal window (NOT maximized). */
   private static final int CAPTION_BUTTON_HEIGHT = 29;
   
   /** The height of the caption buttons when window is maximized. */
   private static final int CAPTION_BUTTON_HEIGHT_MAXIMIZED = 21;

   /** The height of the small close caption button. */
   private static final int CAPTION_BUTTON_CLOSE_SMALL_HEIGHT = 23;

   /** The width of the caption buttons in normal window (NOT maximized). */
   private static final int CAPTION_BUTTON_WIDTH = 45;
   
   /** The width of the caption buttons in dialogs. */
   private static final int CAPTION_BUTTON_WIDTH_DIALOG = 31;
   
   /** The color of the background of a scrollbar. */
   private static final Color SCROLL_GAUGE_COLOR = new Color(0xF0F0F0);
   
   /** The color of the corner of a scrollbar. */
   private static final Color SCROLL_CORNER_COLOR = new Color(0xF0F0F0);
   
   /** The color of the border of a scrollbar. */
   private static final Color SCROLL_BORDER_COLOR = new Color(0xFFFFFF);
   
   /** The background color of the of a scrollbar button when pressed. */
   private static final Color SCROLL_BUTTON_COLOR_PRESSED = new Color(0x606060);
   
   /** The background color of the of a scrollbar button in normal state. */
   private static final Color SCROLL_BUTTON_COLOR_NORMAL = new Color(0xF0F0F0);
   
   /** The background color of the of a scrollbar button when disabled. */
   private static final Color SCROLL_BUTTON_COLOR_DISABLED = new Color(0xF0F0F0);
   
   /** The background color of the of a scrollbar button when mouse is hovering it. */
   private static final Color SCROLL_BUTTON_COLOR_HIGHLIGHT = new Color(0xDADADA);
   
   /** The color of the of a scrollbar thumb/slider when pressed. */
   private static final Color SCROLL_SLIDER_COLOR_PRESSED = new Color(0x606060);
   
   /** The color of the of a scrollbar thumb/slider in normal state. */
   private static final Color SCROLL_SLIDER_COLOR_NORMAL = new Color(0xCDCDCD);
   
   /** The color of the of a scrollbar thumb/slider when disabled. */
   private static final Color SCROLL_SLIDER_COLOR_DISABLED = new Color(0xF0F0F0);
   
   /** The color of the of a scrollbar thumb/slider when mouse is over it. */
   private static final Color SCROLL_SLIDER_COLOR_HOVER = new Color(0xC0C0C0);
   
   /** Background color for tooltips. */
   private static final Color TOOLTIP_BACKGROUND_COLOR = new Color(0xFFFFE7);
   
   /** Background color for drop-down and popup menus. */
   private static final Color MENU_POPUP_BACKGROUND_COLOR = new Color(0xF2F2F2);
   
   /** Background color for highlighted menu items. */
   private static final Color MENU_BACKGROUND_SELECTED_COLOR = new Color(0x90C8F6);
   
   /** Background color for menu-menus. */
   private static final Color MENU_BAR_BACKGROUND_COLOR = new Color(0xFFFFFF);
   
   /** Foreground color for (all) menus. */
   private static final Color MENU_FOREGROUND_COLOR = new Color(0x000000);
   
   /** Background color of the workspace */
   private static final Color WORKSPACE_BACKGROUND_COLOR = new Color(0xA5A2A5);
   
   /** Background color for highlighted text (Eg. selected browse cell).*/
   private static final Color HIGHLIGHT_BACKGROUND_COLOR = new Color(0x319AFF);
   
   /** Main background color for 3D objects. */
   private static final Color BACKGROUND_3DFACE_COLOR = new Color(0xF7F3F7);
   
   /** The test colors on enabled buttons. */
   private static final Color BUTTON_FOREGROUND_COLOR = new Color(0x000000);
   
   /** Size of the scroll button on the direction perpendicular to scroll direction. */
   private static final int SCROLL_BUTTON_WIDTH = 15;
   
   /** Width of the thumb button on the direction perpendicular to scroll direction. */
   private static final int SCROLL_SLIDER_WIDTH = 15;
   
   /** Actual width of the scrollbar. */
   private static final int SCROLL_GAUGE_WIDTH = 17;
   
   /** Size of the scroll button on the direction of the scroll. */
   private static final int SCROLL_BUTTON_HEIGHT = 17;
   
   /** Checkbox width or length */
   private static final int CHECKBOX_SIZE = 13;
   
   /** The space between checkbox and its label. */
   private static final int CHECKBOX_SPACE = 3;
   
   /** The width of the combobox button. */
   protected static final int COMBOBOX_BUTTON_WIDTH = 17;
   
   /** The amount of space at the left of the text in a menu item. */
   protected static final int MENU_LEFT_PAD = 10;
   
   /** The amount of space at the right of the text in a menu item. */
   private static final int MENU_RIGHT_PAD = 36;

   /** Left offset of a text inside a browse label. */
   private static final int BROWSE_LABEL_OFFSET_LEFT = 1;

   /** Left offset of a text inside a browse cell. */
   private static final int BROWSE_DATA_OFFSET_LEFT = 1;
   
   /** The amount of space reserved for the checkmark used by checkable menus. */
   protected static final int MENU_CHECKMARK_WIDTH = 22;
   
   /** Cache with read /decoded images. */
   private Map<String, ImageWrapper<BufferedImage>> imgCache = new HashMap<>();
   
   /** Cached map with default system colors. */
   private Map<Integer, ColorRgb> defaultColors = null;
   
   /**
    * Get the name of this {@code Theme}.
    *
    * @return  the name of this theme.
    */
   @Override
   public String getName()
   {
      return "windows10";
   }
   
   /**
    * Release any held resources when this {@code Theme} is removed. 
    */
   public void releaseResources()
   {
      imgCache.clear();
   }
   
   /**
    * Obtain an image resource specific to this {@code Theme}. The image can be cached for
    * performance reasons. It will be automatically released when {@link #releaseResources()} is
    * called, when a new {@code Theme} is loaded and this one is removed.
    *
    * @param   name
    *          The name of the image as path within jar bundle.
    * @param   w
    *          The desired with of the returned image. This is only a hint. If available, an
    *          image with exact width is returned. One need to check the size of the returned
    *          {@link ImageWrapper} to know the exact value.  
    * @param   h
    *          The desired with of the returned image. This is only a hint. If available, an
    *          image with exact width is returned. One need to check the size of the returned
    *          {@code ImageWrapper} to know the exact value.  
    * @param   antialias
    *          if {@code true} and image is scaled to meet the size hints, an smooth scaling
    *          algorithm will be used, otherwise a faster one will be used but the image quality
    *          will be lower. If image is not scaled the parameter is ignored.
    *
    * @return  an {@code ImageWrapper} containing the requested image resource if it can be found
    *          and loaded and {@code null} otherwise.
    */
   @Override
   public ImageWrapper<BufferedImage> getImage(String name, int w, int h, boolean antialias)
   {
      return getPngImage(getBinaryResourceLocation() + "/" + name + ".png");
   }
   
   /**
    * Obtain the location of binary resources (like pixmaps and sprites) used by this theme. The 
    * location is relative to @Theme.class@ location so that the resources could be loaded easily
    * using @getResourceAsStream()@.
    *
    * @return  the location of binary resources used by this theme.
    */
   @Override
   public String getBinaryResourceLocation()
   {
      return "windows10";
   }
   
   /**
    * Check if the theme is mouse hover sensitive. Mouse hover sensitive themes need to repaint
    * the components when the mouse hovers over them (ie. without clicking them).
    *
    * @return  always {@code true}. This theme is mouse hover sensitive.
    */
   @Override
   public boolean isMouseHoverSensitive()
   {
      return true;
   }
   
   /**
    * Check whether this is a classic theme. A classic theme has a low color palette and usually
    * does not have animations which makes it better suited for low UI resources. Usually classic
    * themes are not mouse hover sensitive.
    *
    * @return  Always {@code false}.
    */
   @Override
   public boolean isClassic()
   {
      return false;
   }
   
   /**
    * Obtain the default system palette for this theme.
    *
    * @return  the default system palette for this theme.
    */
   @Override
   public Map<Integer, ColorRgb> getDefaultSystemPalette()
   {
      if (defaultColors == null)
      {
         Map<Integer, ColorRgb> pal = new HashMap<>();
         pal.put(COLOR_SCROLLBAR, new ColorRgb(new Color(212, 208, 200)));
         pal.put(COLOR_BACKGROUND, new ColorRgb(new Color(58, 110, 165)));
         pal.put(COLOR_ACTIVECAPTION, new ColorRgb(WINDOW_TITLEBAR_COLOR_ACTIVE));
         pal.put(COLOR_INACTIVECAPTION, new ColorRgb(WINDOW_TITLEBAR_COLOR_INACTIVE));
         pal.put(COLOR_MENU, new ColorRgb(MENU_POPUP_BACKGROUND_COLOR));
         pal.put(COLOR_WINDOW, new ColorRgb(new Color(0xFFFFFF)));
         pal.put(COLOR_WINDOWFRAME, new ColorRgb(new Color(100, 100, 100)));
         pal.put(COLOR_MENUTEXT, new ColorRgb(MENU_FOREGROUND_COLOR));
         pal.put(COLOR_WINDOWTEXT, new ColorRgb(new Color(0, 0, 0)));
         pal.put(COLOR_CAPTIONTEXT, new ColorRgb(new Color(255, 255, 255)));
         pal.put(COLOR_ACTIVEBORDER, new ColorRgb(WINDOW_BORDER_COLOR_ACTIVE));
         pal.put(COLOR_INACTIVEBORDER, new ColorRgb(WINDOW_BORDER_COLOR_INACTIVE));
         pal.put(COLOR_APPWORKSPACE, new ColorRgb(WORKSPACE_BACKGROUND_COLOR));
         pal.put(COLOR_HIGHLIGHT, new ColorRgb(HIGHLIGHT_BACKGROUND_COLOR));
         pal.put(COLOR_HIGHLIGHTTEXT, new ColorRgb(new Color(0xFFFFFF)));
         pal.put(COLOR_3DFACE, new ColorRgb(BACKGROUND_3DFACE_COLOR));
         pal.put(COLOR_3DSHADOW, new ColorRgb(new Color(0xA5A2A5)));
         pal.put(COLOR_GRAYTEXT, new ColorRgb(new Color(128, 128, 128)));  /* not used ? */
         pal.put(COLOR_BTNTEXT, new ColorRgb(BUTTON_FOREGROUND_COLOR));
         pal.put(COLOR_INACTIVECAPTIONTEXT, new ColorRgb(new Color(212, 208, 200)));
         pal.put(COLOR_BTNHIGHLIGHT, new ColorRgb(new Color(255, 255, 255)));
         pal.put(COLOR_3DDKSHADOW, new ColorRgb(new Color(64, 64, 64))); /* not used ? */
         pal.put(COLOR_3DLIGHT, new ColorRgb(new Color(212, 208, 200))); /* not used ? */
         pal.put(COLOR_INFOTEXT, new ColorRgb(new Color(0, 0, 0)));
         pal.put(COLOR_INFOBK, new ColorRgb(new Color(255, 255, 255)));
         pal.put(COLOR_BUTTONALTERNATEFACE, new ColorRgb(new Color(181, 181, 181)));
         pal.put(COLOR_HOTLIGHT, new ColorRgb(new Color(0, 0, 128)));
         pal.put(COLOR_GRADIENTACTIVECAPTION, new ColorRgb(new Color(166, 202, 240)));
         pal.put(COLOR_GRADIENTINACTIVECAPTION, new ColorRgb(new Color(192, 192, 192)));
         pal.put(COLOR_MENUHILIGHT, new ColorRgb(MENU_BACKGROUND_SELECTED_COLOR));
         pal.put(COLOR_MENUBAR, new ColorRgb(MENU_BAR_BACKGROUND_COLOR));
         pal.put(COLOR_TOOLTIPBK, new ColorRgb(TOOLTIP_BACKGROUND_COLOR));
         pal.put(COLOR_FILLIN_BG, new ColorRgb(new Color(255, 255, 255)));
         pal.put(COLOR_FILLIN_FG, new ColorRgb(new Color(0, 0, 0)));
         pal.put(COLOR_CARET, new ColorRgb(new Color(0, 0, 0)));
         
         defaultColors = Collections.unmodifiableMap(pal);
      }
      
      return defaultColors;
   }
   
   /**
    * Obtain the map of standard fonts specific to current {@code Theme}, to which a key defaults,
    * if not overridden in the directory.
    *
    * @return  the map of standard fonts specific to this {@code Theme}. 
    */
   @Override
   public Map<String, String> getStandardFontTable()
   {
      Map<String, String> fontTable = new HashMap<>();
      
      fontTable.put("default-font",        /* "Segoe UI, size=9"); */ "MS Sans Serif, size=8");
      fontTable.put("default-fixed-font",  "Courier New, size=8");
      
      // standard fonts
      fontTable.put("font0", "Courier New, size=8");
      fontTable.put("font1", "Segoe UI, size=9");         //"MS Sans Serif, size=8");
      fontTable.put("font2", "Courier New, size=8");
      fontTable.put("font3", "Courier New, size=8");
      fontTable.put("font4", "Segoe UI, size=9");         //"MS Sans Serif, size=8");
      fontTable.put("font5", "Segoe UI, size=10");        //"MS Sans Serif, size=10");
      fontTable.put("font6", "Segoe UI, size=9, bold");   //"MS Sans Serif, size=8, bold");
      fontTable.put("font7", "Segoe UI, size=9");         //"MS Sans Serif, size=8");
      
      // system fonts
      fontTable.put(FontTable.WINDOW_TITLE_FONT,        "Segoe UI, size=10");   //"Tahoma, size 8, bold");
      fontTable.put(FontTable.ALERT_BOX_FONT,           "Segoe UI, size=9");    //"Tahoma, size=8");
      fontTable.put(FontTable.STACK_TRACE_WINDOW_FONT,  "Segoe UI, size=8");    //"MS Sans Serif, size=8");
      fontTable.put(FontTable.MENU_FONT,                "Segoe UI, size=9");    //"Tahoma, size 8");
      fontTable.put(FontTable.SORT_NUMBER_FONT,         "Segoe UI, size=7");    //"MS Sans Serif, size=7");
      
      // freeze the map
      return Collections.unmodifiableMap(fontTable);
   }
   
   /**
    * Obtain the map of default font aliases specific to current {@code Theme}, to which a key
    * defaults, if not overridden in the directory.
    * 
    * @return  the map of standard font aliases specific to this {@code Theme}. 
    */
   @Override
   public Map<String, String> getDefaultFontAliases()
   {
      Map<String, String> fontTable = new HashMap<>();
      
      fontTable.put("default-font",        "Lato, size=8");
      fontTable.put("default-fixed-font",  "Liberation Mono, size=8");
      
      // standard fonts
      fontTable.put("font0", "Liberation Mono, size=8");
      fontTable.put("font1", "DejaVu Sans, size=10, bold");
      fontTable.put("font2", "Liberation Mono, size=8");
      fontTable.put("font3", "Liberation Mono, size=8");
      fontTable.put("font4", "Liberation Sans, size=8");
      fontTable.put("font5", "Open Sans, size=10");
      fontTable.put("font6", "Open Sans, size=8, bold");
      fontTable.put("font7", "DejaVu Sans, size=10, bold");
      
      // system fonts
      fontTable.put(FontTable.WINDOW_TITLE_FONT,        "Open Sans, size=10");
      fontTable.put(FontTable.ALERT_BOX_FONT,           "Open Sans, size=9");
      fontTable.put(FontTable.STACK_TRACE_WINDOW_FONT,  "Open Sans, size=8");
      fontTable.put(FontTable.MENU_FONT,                "Open Sans, size=9");
      fontTable.put(FontTable.SORT_NUMBER_FONT,         "Open Sans, size=7");
      
      // freeze the map
      return Collections.unmodifiableMap(fontTable);
   }
   
   /**
    * Returns the height of window's titleBar. It DOES NOT include the upper window border. 
    *
    * @param   maximized
    *          On some themes, the maximized window have a different title bar height when 
    *          maximized. 
    *
    * @return  the height of the window's title bar. 
    */
   @Override
   public int getTitleBarHeight(boolean maximized)
   {
      return maximized ? TITLE_BAR_HEIGHT_MAXIMIZED : TITLE_BAR_HEIGHT;
   }
   
   /**
    * Returns the window's border insets. In Windows 10 theme, all insets are equals to 1.
    *
    * @param   modal
    *          Usually the modal dialogs/windows have a different width if modal. 
    *
    * @return  the insets of window's border. 
    */
   @Override
   public NativeInsets getWindowBorderInsets(boolean modal)
   {
      return new NativeInsets(1, 1, 1, 1);
   }
   
   /**
    * Returns the height of a caption button on window's title bar. 
    *
    * @param   type
    *          On some specific themes, the height of buttons is not the same.
    * @param   maximized
    *          On some themes, the maximized window have a different button height when 
    *          maximized. 
    *
    * @return  the height of the caption button. 
    */
   @Override
   public int getCaptionButtonHeight(CaptionButtonType type, boolean maximized)
   {
      if (type == CaptionButtonType.CLOSE_SMALL)
      {
         // the button is squared
         return CAPTION_BUTTON_CLOSE_SMALL_HEIGHT;
      }
      else
      {
         return maximized ? CAPTION_BUTTON_HEIGHT_MAXIMIZED : CAPTION_BUTTON_HEIGHT;
      }
   }
   
   /**
    * Returns the width of a caption button on window's title bar. 
    *
    * @param   type
    *          On some specific themes, the width of buttons is not the same.
    *
    * @param   dialog
    *          {@code true} for dialogs. The dialogs may have specific width for caption buttons.  
    *
    * @return  the width of the requested caption button. 
    */
   @Override
   public int getCaptionButtonWidth(CaptionButtonType type, boolean dialog)
   {
      if (type == CaptionButtonType.CLOSE_SMALL)
      {
         return CAPTION_BUTTON_CLOSE_SMALL_HEIGHT;
      }
      else
      {
         return dialog ? CAPTION_BUTTON_WIDTH_DIALOG : CAPTION_BUTTON_WIDTH;
      }
   }
   
   /**
    * Do the layout of a titlebar. Each {@code Theme} can define how its components are arranged. 
    *
    * @param   window
    *          The window whose titlebar is laid out. 
    * @param   icon
    *          The icon widget of the window (if any).
    * @param   title
    *          The {@link Label} that displays the window's title. 
    * @param   buttons
    *          An array with the {@link CaptionButton}-s present on the titlebar of this window. 
    * @param   width
    *          The width of the window (in character units).
    * @param   c
    *          the {@link CoordinatesConversion} used for possible units conversions.
    *
    * @return  the computed titlebar height.
    */
   @Override
   public double layoutTitleBar(TopLevelWindow window,
                                Widget icon,
                                Label title,
                                CaptionButton[] buttons,
                                double width,
                                CoordinatesConversion c)
   {
      int ICON_INSETS = 2; //px
      boolean isWindowMaximized = window.isMaximized();
      int font = window.resolveTitleFont();
      // compute the window title's font height, in pixels.  this will determine the titlebar's
      // height.
      int fHeight = c.heightToNative(1, window.getId(), font);
      int barNativeHeight = Math.max(fHeight, getTitleBarHeight(false));
      double barHeight = c.heightFromNative(barNativeHeight); // to return !!
      
      int topInset = getWindowBorderInsets(false).top;
      if (icon != null)
      {
         // - icon on left-side; to ensure the character unit location is properly in sync with
         // the native location, compute the character location first and re-transform to native
         // for the physical location.
         NativeDimension id = icon.physicalDimension();
         int vAlign = getWindowIconAlignment() & Theme.VERTICAL_ALIGN_MASK;
         int dy = 0;
         switch (vAlign)
         {
            case Theme.ALIGN_BOTTOM:
               dy = barNativeHeight - id.height;
               break;
            case Theme.ALIGN_MIDDLE:
               // compute including the window upper border
               dy = (barNativeHeight + topInset - id.height) / 2;
               break;
         }
         icon.setLocation(c.widthFromNative(ICON_INSETS), c.heightFromNative(dy)); 
         NativePoint iconLoc = c.pointToNative(icon.location());
         icon.setPhysicalLocation(iconLoc.x, iconLoc.y);
      }

      if (buttons.length > 0)
      {
         // add the buttons (minimize, maximize, close buttons)
         // all coordinates are relative to this container
         double bx = width;

         // determine the vertical location so that the button is centered vertically
         double by = 0;

         CaptionButtonType fstButtonType = buttons[0].getType();
         int btnHeight = getCaptionButtonHeight(fstButtonType, isWindowMaximized);
         int alignment = getCaptionButtonsAlignment(fstButtonType);

         switch (alignment & Theme.VERTICAL_ALIGN_MASK)
         {
            case Theme.ALIGN_MIDDLE:
               by = c.heightFromNative((barNativeHeight - btnHeight) / 2);
               break;
            case Theme.ALIGN_BOTTOM:
               int loc = barNativeHeight - btnHeight;
               if (fstButtonType == CaptionButtonType.CLOSE_SMALL)
               {
                  loc -= 2;
               }
               by = c.heightFromNative(loc);
               break;
         }
         double bh = c.heightFromNative(btnHeight);

         for (int i = buttons.length - 1; i >= 0; i--)
         {
            CaptionButton btn = buttons[i];
            double bw = c.widthFromNative(getCaptionButtonWidth(btn.getType(), false));
            btn.setDimension(bw, bh);

            int buttonSpace = getCaptionButtonWidth(btn.getType(), false) +
            getSpaceBeforeButton(btn.getType());
            bx -= c.widthFromNative(buttonSpace);

            btn.setLocation(bx, by);
            NativePoint btnLoc = c.pointToNative(btn.location());
            btn.setPhysicalLocation(btnLoc.x, btnLoc.y);
         }
      }
      
      double iconColumn = 0;
      if (icon != null && !icon.hidden())
      {
         iconColumn = icon.bounds().right();
      }
      
      // title in the middle, between the icon and the minimize button
      int tAlign = getWindowTitleAlignment();
      NativePoint titleLoc = new NativePoint(c.columnToNative(iconColumn) + ICON_INSETS, 0);
      
      // compute title width
      double btnsCol = buttons.length == 0 ? width : buttons[0].bounds().left();
      double twidth = btnsCol - iconColumn; // extract 3 * ICON_INSETS
      switch (tAlign & Theme.VERTICAL_ALIGN_MASK)
      {
         case Theme.ALIGN_BOTTOM:
            titleLoc.y = barNativeHeight - fHeight;
            break;
         case Theme.ALIGN_MIDDLE:
            // compute including the window upper border
            fHeight += 4;
            titleLoc.y = (barNativeHeight + topInset - fHeight) / 2 - topInset - 2;
            break;
      }
      title.setHeight(c.heightFromNative(fHeight));
      title.setLocation(c.pointFromNative(titleLoc));
      title.setPhysicalLocation(titleLoc.x, titleLoc.y);
      
      // set title width
      title.setWidth(twidth);
      
      return barHeight;
   }
   
   /**
    * Get the alignment of caption button for this theme. The returned value combines both
    * horizontal and vertical alignment of the buttons.
    *
    * @param   type
    *          On some specific themes, the alignment of buttons is not the same.
    *
    * @return  the alignment of caption button.
    */
   @Override 
   public int getCaptionButtonsAlignment(CaptionButtonType type)
   {
      switch (type)
      {
         case CLOSE_SMALL:
            return ALIGN_RIGHT | ALIGN_BOTTOM;
         default:
            return ALIGN_RIGHT | ALIGN_MIDDLE;
      }

      // buttons are in fact aligned to the top of the titlebar (excluding frame border) and
      // excluding the small close
   }
   
   /**
    * Get the title alignment for this theme. The returned value combines both horizontal and
    * vertical alignment.
    *
    * @return  the alignment of title.
    */
   @Override
   public int getWindowTitleAlignment()
   {
      return ALIGN_LEFT | ALIGN_MIDDLE;
   }
   
   /**
    * Get the icon alignment for this theme. The returned value combines both horizontal and
    * vertical alignment.
    *
    * @return  the alignment of icon on titlebar.
    */
   @Override
   public int getWindowIconAlignment()
   {
      return ALIGN_LEFT | ALIGN_MIDDLE;
   }
   
   /**
    * Return the amount of space to the right of this button. On some themes, different spaces can
    * be found between the caption buttons. Although unusual, negative values can be returned,
    * when buttons overlap for a few pixels. 
    *
    * @param   type
    *          The caption button type.
    *
    * @return  the amount of space to the right of this button.
    */
   @Override
   public int getSpaceBeforeButton(CaptionButtonType type)
   {
      // in this theme uses one pixel between buttons, no space to the right
      return (type == CaptionButtonType.CLOSE) ? 0 : 1;
   }
   
   /**
    * Renders a {@link CaptionButton}.
    *
    * @param   button
    *          The {@code CaptionButton} to be drawn. 
    * @param   gd
    *          The {@link GuiDriver} used for drawing operations. Note that internal state of 
    *          this object can affect the resulting output.
    * @param   x
    *          The horizontal location (in pixel units) of the widget to be painted/component.
    * @param   y
    *          The vertical location (in pixel units) of the widget to be painted/component.
    * @param   width
    *          The width of the painted widget/component. 
    * @param   height
    *          The height of the painted widget/component. 
    */
   @Override
   public void drawCaptionButton(CaptionButton button, 
                                 GuiDriver gd, 
                                 int x,
                                 int y,
                                 int width,
                                 int height)
   {
      TopLevelWindow<GuiOutputManager> window = button.topLevelWindow();
      boolean active = WindowManager.isWindowActive(window);
      boolean enabled = button.isEnabled();
      boolean mouseOver = button.isMouseOver();
      boolean pressed = button.isPressed() && mouseOver; // even is pressed, mouse exited bounds
      CaptionButtonType type = button.getType();
      boolean close = (type == CaptionButtonType.CLOSE);
      
      // close button is handled specially because it uses a different palette
      java.awt.Color color = null;
      if (type == CaptionButtonType.CLOSE_SMALL)
      {
         color = !active ? CAPTION_BUTTON_CLOSE_SMALL_COLOR_INACTIVE :
                           pressed ? CAPTION_BUTTON2_COLOR_DOWN :
                                     mouseOver ? CAPTION_BUTTON2_COLOR_HIGHLIGHT :
                                                 CAPTION_BUTTON_CLOSE_SMALL_ACTIVE;
      }
      else
      {
         color = !active ? CAPTION_BUTTON_COLOR_INACTIVE :
                           pressed ? (close ? CAPTION_BUTTON2_COLOR_DOWN :
                                              CAPTION_BUTTON_COLOR_DOWN) :
                                     mouseOver ? (close ? CAPTION_BUTTON2_COLOR_HIGHLIGHT :
                                                          CAPTION_BUTTON_COLOR_HIGHLIGHT) :
                           /* normal ? */        CAPTION_BUTTON_COLOR_ACTIVE;
      }
      gd.setColor(color);
      gd.fillRect(x, y, width, height);
      
      StringBuilder sb = new StringBuilder(getBinaryResourceLocation());
      sb.append("/wnd-");
      switch (type)
      {
         case ICONIFY:
            sb.append("min");
            break;
         case MAXIMIZE:
            sb.append("max");
            break;
         case CLOSE:
            sb.append("close");
            break;
         case CLOSE_SMALL:
            sb.append("close-small");
            break;
         case HELP:
            sb.append("help");
            break;
         case RESTORE_DOWN:
         case RESTORE_UP:
            sb.append("restore");
            break;
      }
      
      if (type != CaptionButtonType.CLOSE_SMALL)
      {
         if (close)
         {
            sb.append(!active ? "-dis" : pressed ? "-down" : mouseOver ? "-hi" : "-en");
         }
         else
         {
            sb.append(active && enabled ? "-en" : "-dis");
         }
      }

      sb.append(".png");
      
      ImageWrapper img = getPngImage(sb.toString());
      if (img != null)
      {
         gd.drawImage(img,
                      x + (width - img.getWidth()) / 2, y + (height - img.getHeight()) / 2,
                      img.getWidth(), img.getHeight(),
                      0, 0, true, false, true, false);
      }
   }
   
   /**
    * Renders the background of the titlebar of a top-level window.
    *
    * @param   gd
    *          The {@code GuiDriver} used for drawing operations.
    * @param   x
    *          The horizontal location (in pixel units) of the widget to be painted/component.
    * @param   y
    *          The vertical location (in pixel units) of the widget to be painted/component.
    * @param   width
    *          The width of the painted widget/component. 
    * @param   height
    *          The height of the painted widget/component.
    * @param   active
    *          {@code true} if the drawn windows is active.
    */
   @Override
   public void drawWindowTitleBar(GuiDriver gd,
                                  int x,
                                  int y,
                                  int width,
                                  int height,
                                  boolean active)
   {
      int color = active ? SysColor.COLOR_ACTIVECAPTION : SysColor.COLOR_INACTIVECAPTION;
      
      gd.setColor(gd.getSysColor(color));
      gd.fillRect(x, y, width, height);
   }

   /**
    * Renders the text on the titlebar of a top-level window.
    *
    * @param   gd
    *          The {@code GuiDriver} used for drawing operations.
    * @param   gf
    *          The {@link GuiFontResolver} for selecting the required font.
    * @param   x
    *          The horizontal location (in pixel units) of the widget to be painted/component.
    * @param   y
    *          The vertical location (in pixel units) of the widget to be painted/component.
    * @param   width
    *          The width of the painted widget/component. 
    * @param   height
    *          The height of the painted widget/component.
    * @param   title
    *          The title of the window.
    * @param   active
    *          {@code true} if the drawn windows is active.
    */
   @Override
   public void drawWindowTitle(GuiDriver gd,
                               GuiFontResolver gf,
                               int x,
                               int y,
                               int width,
                               int height,
                               String title,
                               boolean active)
   {
      // remove old title
      drawWindowTitleBar(gd, x, y, width, height, active);
      
      if (title.length() > 0)
      {
         FontDetails<?> font = gf.font().font;
         
         // draw text
         gd.setColor(active ? WINDOW_TITLE_COLOR_ACTIVE :
                              WINDOW_TITLE_COLOR_INACTIVE);
         gd.setGuiFont(font);
         gd.drawStringCentered(title, x, height / 2);
      }
   }
   
   /**
    * Renders borders of a top-level window.
    *
    * @param   gd
    *          The {@code GuiDriver} used for drawing operations.
    * @param   x
    *          The horizontal location (in pixel units) of the widget to be painted/component.
    * @param   y
    *          The vertical location (in pixel units) of the widget to be painted/component.
    * @param   width
    *          The width of the painted widget/component. 
    * @param   height
    *          The height of the painted widget/component.
    * @param   border
    *          The width of the border on each side.
    * @param   active
    *          {@code true} if the drawn windows is active.
    */
   @Override
   public void drawWindowBorder(GuiDriver gd,
                                int x,
                                int y,
                                int width,
                                int height,
                                NativeInsets border,
                                boolean active)
   {
      gd.setColor(gd.getSysColor(active ? COLOR_ACTIVEBORDER : COLOR_INACTIVEBORDER));
      gd.drawRect(x, y, width - 1, height - 1);
   }
   
   /**
    * Renders borders of a top-level window.
    *
    * @param   button
    *          The {@link ButtonGuiImpl} to be rendered.
    * @param   gd
    *          The {@code GuiDriver} used for drawing operations.
    * @param   gc
    *          The {@link GuiColorResolver} to be used for color selection.
    * @param   gf
    *          The {@code GuiFontResolver} to be used.
    * @param   x
    *          The horizontal location (in pixel units) of the widget to be painted/component.
    * @param   y
    *          The vertical location (in pixel units) of the widget to be painted/component.
    * @param   width
    *          The width of the painted widget/component. 
    * @param   height
    *          The height of the painted widget/component.
    */
   @Override
   public void drawButton(ButtonGuiImpl button,
                          GuiDriver gd,
                          GuiColorResolver gc,
                          GuiFontResolver gf,
                          int x,
                          int y,
                          int width,
                          int height)
   {
      ButtonConfig bcfg =   button.config();
      boolean flat =        bcfg.flatButton;
      boolean highlighted = button.highlighted() && !flat;
      boolean isDefault =   button.isDefaultButton();
      boolean mouseOver =   button.isMouseOver();
      boolean pressed =     button.isPressed() || mouseOver; // even is pressed, mouse exited bounds
      
      // draw button text
      FontDetails fnt = gf.font().font;
      gd.setGuiFont(fnt);
      int fontNum = gf.resolveFontNum();
      MnemonicInfo labelInfo = button.getMnemonic();
      String label = (labelInfo == null) ? button.textLabel() : labelInfo.getPrepLabel();
      // do not simplify the line below, this makes the font height to be up nearest even number
      int tY = y + (height - 1) / 2;
      int tX = x + (width - button.getTextWidthNative(label, fontNum)) / 2;
      gd.setFontStyle(fnt.style);
      
      if (!button.isEnabled()) // disabled button
      {
         // for flat buttons only the text is displayed
         if (!flat)
         {
            gd.setColor(new Color(0xCCCCCC)); // button fill
            gd.fillRect(x, y, width, height);
            
            gd.setColor(new Color(0xBFBFBF)); // button border
            gd.drawRect(x, y, width - 1, height - 1);
         }
      }
      else if (pressed)
      {
         gd.setColor(new Color(0xCCE4F7)); // button fill
         gd.fillRect(x, y, width, height);
         
         gd.setColor(new Color(0x005499)); // button border
         gd.drawRect(x, y, width - 1, height - 1);
      }
      else if (!flat || mouseOver)
      {
         // button fill
         gd.setColor(mouseOver ? new Color(0xE5F1FB) : // mouse-on button color 
                                 new Color(0xE1E1E1)); // normal button color 
         gd.fillRect(x, y, width, height);
         
         // button border
         gd.setColor(isDefault || highlighted ? new Color(0x0078D7) : // default button border
                                                new Color(0xADADAD)); // normal button border
         gd.drawRect(x, y, width - 1, height - 1);
         
         // if default, the border line is doubled
         if (isDefault)
         {
            gd.drawRect(x + 1, y + 1, width - 3, height - 3);
         }
      }
      
      // draw button content
      if (button.hasImages())
      {
         ImageGuiImpl img = null;
         
         // draw button body images
         if (pressed)
         {
            // determine what image to draw
            ImageGuiImpl imgDown = button.getImgDown();
            if (imgDown != null)
            {
               img = imgDown;
            }
            else
            {
               ImageGuiImpl imgUp = button.getImgUp();
               if (imgUp != null)
               {
                  img = imgUp;
               }
            }
         }
         else
         {
            // determine what image to draw
            ImageGuiImpl imgUp = button.getImgUp();
            if (button.isEnabled() && imgUp != null)
            {
               img = imgUp;
            }
            else
            {
               ImageGuiImpl imgDisabled = button.getImgDisabled();
               if (!button.isEnabled() && imgDisabled != null)
               {
                  img = imgDisabled;
               }
            }
         }
         
         // if we found it, draw!
         if (img != null)
         {
            img.draw();
         }
      }
      else
      {
         // draw text:
         if (button.isEnabled())
         {
            gd.setColor(gd.getSysColor(COLOR_BTNTEXT)); // normal text
         }
         else
         {
            gd.setColor(new Color(0x838383)); // disabled text, not in palette ?
         }
         if (labelInfo != null && labelInfo.getMnemIdx() != null)
         {
            NativePoint tp = new NativePoint(tX, tY);
            gd.drawStringWithMnemonic(label, tp, labelInfo, gf, button, true);
         }
         else
         {
            // no mnemonic available
            gd.drawStringCentered(label, tX, tY);
         }
      }
      
      // focused button needs special highlight
      if (button.isEnabled() && highlighted && !button.config().noFocus)
      {
         // set dotted line style
         gd.setLineStroke(LineStroke.DOTTED_SMALL);
         
         gd.setColor(pressed ? new Color(0x3C1407)   // pressed button selection dotted line
                             : new Color(0x111111)); // normal button selection dotted line
         gd.drawRect(x + 2, y + 2, width - 5, height - 5);
         
         // restore default line style
         gd.setLineStroke(LineStroke.DEFAULT);
      }
   }
   
   /**
    * Renders a GUI radio button widget.
    * @param  rb
    *         The {@link RadioButtonGuiImpl} to be rendered.
    * @param  gd
    *         The {@code GuiDriver} used for drawing operations.
    * @param   gc
    *         The {@link GuiColorResolver} to be used for color selection.
    * @param  gf
    *         The {@code GuiFontResolver} to be used.
    * @param  x
    *         The horizontal location (in pixel units) of the widget to be painted/component.
    * @param  y
    *         The vertical location (in pixel units) of the widget to be painted/component.
    * @param  width
    *         The width of the painted widget/component. 
    * @param  height
    *         The height of the painted widget/component.
    * @param  focused
    *         {@code true} if the radio button have the focus
    * @param  hover
    *         {@code true} if mouse is hovering over the checkbox. Some styles are mouse hovering
    */
   @Override
   public void drawRadioButton(RadioButtonGuiImpl rb,
                               GuiDriver gd,
                               GuiColorResolver gc,
                               GuiFontResolver gf,
                               int x,
                               int y,
                               int width,
                               int height,
                               boolean focused,
                               boolean hover)
   {
      // first draw the check mark image
      boolean isEnabled = rb.isEnabled() && rb.parent().isEnabled();
      ImageWrapper<?> imgToDraw = getRadioButtonImage(rb.isSelected(), isEnabled, hover, false);
      if (imgToDraw != null && imgToDraw.getImage() != null)
      {
         int imgW = imgToDraw.getWidth();
         int imgH = imgToDraw.getHeight();
   
         gd.drawImage(imgToDraw, x, y + (height - imgH) / 2, imgW, imgH,
                      0, 0, false, false, false, false);
      }
      
      // draw button text
      String label = rb.getText();
      FontDetails<?>  fnt = gf.font().font;
      int fntNum = gf.resolveFontNum();
      
      gd.setGuiFont(fnt);
      int wt = rb.getTextWidthNative(label, fntNum);
      int ht = rb.getTextHeightNative(label, fntNum);
      int xt = x + getRadioButtonImageSpace();
      int yt = y + height / 2;
      gd.setFontStyle(fnt.style);
      // enabled state controls text drawing
      if (!isEnabled)
      {
         gd.setColor(new Color(0x6B6D6B));
         gd.drawStringCentered(label, xt, yt);
      }
      else
      {
         gd.setColor(new Color(0x000000));
         
         MnemonicInfo info = rb.getMnemonic();
         
         if (info != null && info.getMnemIdx() != null)
         {
            NativePoint tp = new NativePoint(xt, yt);
            gd.drawStringWithMnemonic(label, tp, info, gf, rb, true);
         }
         else
         {
            gd.drawStringCentered(label, xt, yt);
         }
      }
      
      if (focused)
      {
         // set dotted line style
         gd.setLineStroke(LineStroke.DOTTED_SMALL);
         // set dotted rectangle width
         int drWidth = (xt + wt) <= width ? wt + 1 : width - xt;
         int drHeight = Math.min(ht, height);
         // draw dotted rectangle depending on centered flag and item height
         gd.drawRect(xt - 1, y + (height - drHeight) / 2, drWidth, drHeight);
         // restore default line style
         gd.setLineStroke(LineStroke.DEFAULT);
      }
      
      ThemeManager.debug(() -> {
         gd.setColor(Color.CYAN);
         gd.drawRect(x, y, width - 1, height - 1);
      });
   }
   
   /**
    * Obtain the distance before text, including the radio button (text offset).
    *
    * @return  the distance before text, including the radio button.
    */
   @Override
   public int getRadioButtonImageSpace()
   {
      return 16;
   }
   
   /**
    * Obtain the image from a radio button with specified attributes. 
    *
    * @param   on
    *          The button state. {@code true} if button is on.
    * @param   enabled
    *          {@code true} is button enabled. 
    * @param   hover
    *          {@code true} if mouse is hovering over the button. Some styles are mouse hovering
    *          sensitive. 
    * @param   pressed
    *          {@code true} if the radio button is pressed (the mouse button is down on the 
    *          button).
    *
    * @return  The {@code ImageWrapper} containing the image with requested properties. 
    */
   @Override
   protected ImageWrapper getRadioButtonImage(boolean on,
                                              boolean enabled,
                                              boolean hover,
                                              boolean pressed)
   {
      StringBuilder sb = new StringBuilder(getBinaryResourceLocation());
      sb.append("/radio-").append(on ? "on-" : "off-");
      if (!enabled)
      {
         sb.append("dis");
      } 
      else if (pressed)
      {
         sb.append("down");
      }
      else if (hover) // highlighted / mouse over
      {
         sb.append("hi");
      }
      else
      {
         sb.append("en");
      }
      sb.append(".png");
      
      return getPngImage(sb.toString());
   }
   
   /**
    * Obtains the space needed by a checkbox widget.
    *
    * @return  the space needed by a checkbox widget.
    */
   @Override
   public int getCheckBoxSpace()
   {
      return CHECKBOX_SPACE;
   }
   
   /**
    * Obtain the size of image on a checkbox widget.
    *
    * @return size of image on a checkbox widget.
    */
   @Override
   public int getCheckBoxSize()
   {
      return CHECKBOX_SIZE;
   }
   
   /**
    * Renders a GUI toggle box widget.
    * @param   cb
    *          The {@link ToggleBoxGuiImpl} to be rendered.
    * @param   gd
    *          The {@code GuiDriver} used for drawing operations.
    * @param   gc
    *          The {@link GuiColorResolver} to be used for color selection.
    * @param   gf
    *          The {@code GuiFontResolver} to be used.
    * @param   x
    *          The horizontal location (in pixel units) of the widget to be painted/component.
    * @param   y
    *          The vertical location (in pixel units) of the widget to be painted/component.
    * @param   width
    *          The width of the painted widget/component. 
    * @param   height
    *          The height of the painted widget/component.
    * @param   pressed
    *          {@code true} if the checkbox is pressed (the mouse button is down on the button).
    * @param   hover
    *          {@code true} if mouse is hovering over the checkbox. Some styles are mouse hovering
    */
   public void drawToggleBox(ToggleBoxGuiImpl cb,
                             GuiDriver gd,
                             GuiColorResolver gc,
                             GuiFontResolver gf,
                             int x,
                             int y,
                             int width,
                             int height,
                             boolean pressed,
                             boolean hover)
   {
      NativePoint po = new NativePoint(x, y);
      String text = cb.getMnemonicLabel(false).orElse(cb.config().label);
      GuiFontResolver.FontCache fntCache = gf.font();
      boolean enabled = cb.isEnabled();
      boolean highlighted = cb.highlighted(); // toggle-box state flags
      
      // clear the area
      gd.setColor(gc.bgColor);
      gd.fillRect(po.x, po.y, width, height);
      
      // draw the checkbox as an image 
      ToggleBoxConfig config = cb.config();
      StringBuilder sb = new StringBuilder(getBinaryResourceLocation());
      sb.append("/checkbox-");
      sb.append(config.checked ? "on-" : "off-"); // TODO: add support for 3-state checkbutton
      if (!enabled)
      {
         sb.append("dis");
      }
      else if (pressed)
      {
         sb.append("down");
      }
      else if (hover) // highlighted / mouse over
      {
         sb.append("hi");
      }
      else
      {
         sb.append("en");
      }
      sb.append(".png");
      ImageWrapper<BufferedImage> img = getPngImage(sb.toString());
      int imgWidth = 0;
      if (img != null)
      {
         int imgHeight = img.getHeight();
         imgWidth = img.getWidth();
         int y2 = y + (height - imgHeight) / 2;
         // TODO: enlarge the images and use automatic transparency --v
         gd.drawImage(img, x, y2, imgWidth, imgHeight, 0, 0, false, false, true, false);
      }
      
      // draw the text to the right of the checkbox
      if (text != null && text.length() > 0)
      {
         int xt = po.x + imgWidth + getCheckBoxSpace(); // space between checkbox and label
         int yt = po.y + height / 2;
         
         gd.setGuiFont(fntCache.font);
         gd.setFontStyle(fntCache.font.style);
         
         // draw the light text first!
         gd.setColor(enabled ? new Color(0x000000) :
                               new Color(0x6B6D6B));
         
         MnemonicInfo info = cb.getMnemonic();
         if (info != null && info.getMnemIdx() != null)
         {
            NativePoint tp = new NativePoint(xt, yt);
            gd.drawStringWithMnemonic(text, tp, info, gf, cb, true);
         }
         else
         {
            gd.drawStringCentered(text, xt, yt);
         }
         
         // focused toggle-box needs special highlight
         if (highlighted)
         {
            int font = gf.resolveFontNum();
            int textWidth = cb.getTextWidthNative(text, font) + 1;
            int textHeight = cb.getTextHeightNative(text, font) + 2;
            yt = po.y + (height - textHeight) / 2;
            
            gd.setColor(new Color(0x000000));
            gd.setLineStroke(LineStroke.DOTTED_SMALL_THIN);
            gd.drawRect(xt - 1, yt, textWidth, textHeight);
            gd.setLineStroke(LineStroke.DEFAULT);
         }
      }
      
      ThemeManager.debug(() -> {
         gd.setColor(Color.BLUE);
         gd.drawRect(x, y, width - 1, height - 1);
      });
   }
   
   /**
    * Obtain the size of the scroll button on the direction perpendicular to scroll direction.
    *
    * @return  size of the scroll button on the direction perpendicular to scroll direction.
    */
   @Override
   public int getScrollBarButtonWidth()
   {
      return SCROLL_BUTTON_WIDTH;
   }
   
   /**
    * Obtain the thumb/slider width in a scrollbar widget.
    *
    * @return  the thumb/slider width in a scrollbar widget.
    */
   @Override
   public int getScrollBarThumbWidth()
   {
      return SCROLL_SLIDER_WIDTH;
   }
   
   /**
    * Obtain the width of the gauge of a scrollbar widget.
    *
    * @return  width of the gauge of a scrollbar widget.
    */
   @Override
   public int getScrollBarGaugeWidth()
   {
      return SCROLL_GAUGE_WIDTH;
   }
   
   /**
    * Obtain the size of the scroll button on the direction of the scroll.
    *
    * @return  the size of the scroll button on the direction of the scroll.
    */
   @Override
   public int getScrollBarButtonHeight()
   {
      return SCROLL_BUTTON_HEIGHT;
   }
   
   /**
    * Renders the gauge (background) of the scrollbar widget. 
    *
    * @param   gd
    *          The {@code GuiDriver} used for drawing operations.
    * @param   gc
    *          The {@code GuiColorResolver} used color selection.
    * @param   x
    *          The horizontal location (in pixel units) of the widget to be painted/component.
    * @param   y
    *          The vertical location (in pixel units) of the widget to be painted/component.
    * @param   width
    *          The width of the painted widget/component. 
    * @param   height
    *          The height of the painted widget/component.
    * @param   horizontal
    *          The direction of the scrollbar, {@code true} for horizontal.
    */
   @Override
   public void drawScrollBarGauge(GuiDriver gd,
                                  GuiColorResolver gc,
                                  int x,
                                  int y,
                                  int width,
                                  int height,
                                  boolean horizontal)
   {
      gd.setColor(SCROLL_GAUGE_COLOR);
      gd.fillRect(x, y, width, height);
      
      gd.setColor(SCROLL_BORDER_COLOR);
      if (horizontal)
      {
         gd.drawLine(x, y,              x + width - 1, y);
         gd.drawLine(x, y + height - 1, x + width,     y + height - 1);
      }
      else
      {
         gd.drawLine(x,             y, x,              y + height - 1);
         gd.drawLine(x + width - 1, y, x + width - 1 , y + height - 1);
      }
   }
   
   /**
    * Renders the thumb/slider of the scrollbar widget.
    *
    * @param   gd
    *          The {@code GuiDriver} used for drawing operations.
    * @param   gc
    *          The {@code GuiColorResolver} used color selection.
    * @param   x
    *          The horizontal location (in pixel units) of the widget to be painted/component.
    * @param   y
    *          The vertical location (in pixel units) of the widget to be painted/component.
    * @param   width
    *          The width of the painted widget/component. 
    * @param   height
    *          The height of the painted widget/component.
    * @param   pressed
    *          If the mouse is dragging the thumb of the scrollbar.
    * @param   hover
    *          Is mouse hovering this component ?
    * @param   horizontal
    *          The direction of the scrollbar, {@code true} for horizontal.
    */
   @Override
   public void drawScrollBarThumb(GuiDriver gd,
                                  GuiColorResolver gc,
                                  int x,
                                  int y,
                                  int width,
                                  int height,
                                  boolean pressed,
                                  boolean hover,
                                  boolean horizontal)
   {
      gd.setColor(pressed ? SCROLL_SLIDER_COLOR_PRESSED :
                  hover ? SCROLL_SLIDER_COLOR_HOVER :
                  /*normal*/ SCROLL_SLIDER_COLOR_NORMAL);
      gd.fillRect(x, y, width, height);
   }
   
   /**
    * Renders a button of the scrollbar widget. 
    *
    * @param   gd
    *          The {@code GuiDriver} used for drawing operations.
    * @param   gc
    *          The {@code GuiColorResolver} used color selection.
    * @param   x
    *          The horizontal location (in pixel units) of the widget to be painted/component.
    * @param   y
    *          The vertical location (in pixel units) of the widget to be painted/component.
    * @param   width
    *          The width of the painted widget/component. 
    * @param   height
    *          The height of the painted widget/component.
    * @param   pos
    *          The position of the button (UP/DOWN, LEFT/RIGHT).
    * @param   horizontal
    *          The direction of the scrollbar, {@code true} for horizontal.
    * @param   enabled
    *          If the button is enabled.
    * @param   pressed
    *          If the button is down (pressed by mouse).
    * @param   highlighted
    *          If the mouse is hovering over the button.
    */
   @Override
   public void drawScrollBarButton(GuiDriver gd,
                                   GuiColorResolver gc,
                                   int x,
                                   int y,
                                   int width,
                                   int height,
                                   ScrollBar.Position pos,
                                   boolean horizontal,
                                   boolean enabled,
                                   boolean pressed,
                                   boolean highlighted)
   {
      gd.setColor(!enabled ? SCROLL_BUTTON_COLOR_DISABLED :
                  pressed ? SCROLL_BUTTON_COLOR_PRESSED :
                  highlighted ? SCROLL_BUTTON_COLOR_HIGHLIGHT :
                  /*default ?*/ SCROLL_BUTTON_COLOR_NORMAL);
      gd.fillRect(x, y, width, height);
      
      StringBuilder sb = new StringBuilder(getBinaryResourceLocation());
      sb.append("/scroll-btn-");
      if (horizontal)
      {
         sb.append(pos == ScrollBar.Position.LEFT ? "left-" : "right-");
      }
      else
      {
         sb.append(pos == ScrollBar.Position.LEFT ? "up-" : "down-");
      }
      sb.append(!enabled ? "dis" :
                pressed ? "down" :
                highlighted ? "hi" : 
                /*normal ?*/ "en");
      sb.append(".png");
      
      ImageWrapper<BufferedImage> img = getPngImage(sb.toString());
      if (img != null && img.getImage() != null)
      {
         gd.drawImage(img, x + (width - img.getWidth()) / 2, y + (height - img.getHeight()) / 2, 
                      img.getWidth(), img.getHeight(), 0, 0, false, false, true, false);
      }
   }
   
   /**
    * Obtain the character used to replace text in password fields. 
    *
    * @return  the password character for this theme is bullet char (U+2022).
    */
   @Override
   public char getPasswordCharacter()
   {
      return '\u25CF';    // black circle (U+25CF)
      // return '\u2022'; // bullet char  (U+2022)
   }
   
   /**
    * Renders the border (and possibly the background) of the fill-in widget. 
    *
    * @param   gd
    *          The {@code GuiDriver} used for drawing operations.
    * @param   gc
    *          The {@code GuiColorResolver} used color selection.
    * @param   x
    *          The horizontal location (in pixel units) of the widget to be painted/component.
    * @param   y
    *          The vertical location (in pixel units) of the widget to be painted/component.
    * @param   width
    *          The width of the painted widget/component. 
    * @param   height
    *          The height of the painted widget/component.
    * @param   border3D
    *          If fill-in widget has a 3d border. Some flat {@code Theme}s may ignore this.
    * @param   hasFocus
    *          The fill-in widget is focused.
    * @param   hover
    *          The mouse is hovering over the fill-in widget.
    */
   @Override
   public void drawFillInBorder(GuiDriver gd,
                                GuiColorResolver gc,
                                int x,
                                int y,
                                int width,
                                int height,
                                boolean border3D,
                                boolean hasFocus,
                                boolean hover)
   {
      if (!border3D)
      {
         gd.setColor(new Color(0x636563)); // non 3D always same color
      }
      else
      {
         gd.setColor(hasFocus ? new Color(0x29618c) : // has focus 
                     hover    ? new Color(0x5cbbe7) : // mouse hover
                     /*else*/   new Color(0x737173)); // non-focused
      }
      gd.drawRect(x, y, width - 1, height - 1);
   }

   /**
    * Returns text look and feel.
    * 
    * @param    gcd
    *           The gui color resolver used for normal text.
    * @param    gce
    *           The gui color resolver used for selected text.
    * 
    * @return   The Fillin text look and feel.
    */
   @Override
   public TextLookAndFeel getTextLookAndFeel(GuiColorResolver gcd, GuiColorResolver gce)
   {
      return new TextLookAndFeelImpl(gcd.bgColor,
                                     gcd.fgColor,
                                     new ColorRgb(0x33, 0x99, 0xFF),
                                     new ColorRgb(0xFF, 0xFF, 0xFF),
                                     TextBaseline.MIDDLE);
   }

   /**
    * Renders the status bar of a top-level window.
    *
    * @param   gd
    *          The {@code GuiDriver} used for drawing operations.
    * @param   gc
    *          The {@code GuiColorResolver} used for color selection.
    * @param   x
    *          The horizontal location (in pixel units) of the widget to be painted/component.
    * @param   y
    *          The vertical location (in pixel units) of the widget to be painted/component.
    * @param   width
    *          The width of the painted widget/component. 
    * @param   height
    *          The height of the painted widget/component.
    * @param   text
    *          The text in the status bar.
    * @param   cache
    *          The font cache used for text rendering. 
    * @param   bounds
    *          The {@link NativePoint} where the text is drawn.
    * @param   txtDim
    *          The {@link NativeDimension} of the text displayed in status bar.
    * @param   resizable
    *          {@code true} window is resizable. In this case a "grip" is drawn in the lower right
    *          corner.
    */
   @Override
   public void drawStatusBar(GuiDriver gd,
                             GuiColorResolver gc,
                             int x,
                             int y,
                             int width,
                             int height,
                             String text,
                             GuiFontResolver.FontCache cache,
                             NativePoint bounds,
                             NativeDimension txtDim,
                             boolean resizable)
   {
      gd.setColor(gd.getSysColor(SysColor.COLOR_WINDOW));
      gd.fillRect(x + 1, y + 1, width - 2, height - 2);
      
      gd.setColor(gd.getSysColor(SysColor.COLOR_3DSHADOW));
      gd.drawRect(x, y, width - 1, height - 1);
      
      // draw status bar text
      gd.setColor(new Color(0x000000));
      gd.setGuiFont(cache.font);
      gd.setFontStyle(cache.font.style);
      bounds.x -= 1;
      bounds.y = -1; // fix legacy from Classic Style
      gd.draw(bounds, new NativeRectangle(0, 0, height - 1, width -1),
                  () -> gd.drawStringCentered(text, 0, height / 2));
      
      if (resizable)
      {
         // draw the resize handle:
         ImageWrapper<BufferedImage> img = getPngImage(getBinaryResourceLocation() + 
                                                       "/resize-handle.png");
         if (img != null && img.getImage() != null)
         {
            gd.drawImage(img, width - img.getWidth(), height - img.getHeight(), 
                        img.getWidth(), img.getHeight(), 0, 0, true, false, true, false);
         }
      }
   }
   
   /**
    * Obtain the field extra space for combo box layout.
    *
    * @param   horiz
    *          Direction. {@code true} for horizontal space.
    * @param   is3D
    *          Whether this is widget has a 3D border.
    *
    * @return  the field extra space for combo box layout.
    */
   @Override
   public int getComboBoxFieldExtraSpace(boolean horiz, boolean is3D)
   {
      return 2; /*???*/
   }
   
   /**
    * Obtain the shifting distance in a combo box widget.
    *
    * @return  the shifting distance in a combo box widget.
    */
   @Override
   public int getComboBoxEntryShift()
   {
      return 1;
   }
   
   /**
    * Obtain the size of the button that opens the combo box.
    *
    * @return  the size of teh combo box button.
    */
   @Override
   public int getComboBoxButtonSize()
   {
      return COMBOBOX_BUTTON_WIDTH; /* width = 17, height = 21 (?) */
   }
   
   /**
    * Renders the opening button of the combo box widget. 
    *
    * @param   combo
    *          The {@code ComboBoxGuiImpl} whose border is being painted.
    * @param   gd
    *          The {@code GuiDriver} used for drawing operations.
    * @param   x
    *          The horizontal location (in pixel units) of the widget to be painted/component.
    * @param   y
    *          The vertical location (in pixel units) of the widget to be painted/component.
    * @param   width
    *          The width of the painted widget/component. 
    * @param   height
    *          The height of the painted widget/component.
    * @param   pressed
    *          {@code true} if the widget is pressed (and eventually open).
    * @param   enabled
    *          Whether the widget has the focus.
    * @param   hover
    *          The mouse is hovering over the fill-in widget.
    */
   @Override
   public void drawComboBoxButton(ComboBoxGuiImpl combo,
                                  GuiDriver gd,
                                  int x,
                                  int y,
                                  int width,
                                  int height,
                                  boolean pressed,
                                  boolean enabled,
                                  boolean hover)
   {
      boolean isDropDown = combo.config().mode == ComboBoxConfig.Mode.DROP_DOWN;
      // otherwise it's a DROP_DOWN_LIST because SIMPLE does not have a button
      
      if (enabled && isDropDown && (hover || pressed))
      {
         gd.setColor(pressed ? new Color(0xCCE4F7) :
                     /*hover*/ new Color(0xE5F1FB));
         gd.fillRect(x + width - COMBOBOX_BUTTON_WIDTH, y, x + width - 1, y + height - 1);
      }
      
      // just draw the down arrow for NON-editable cbx
      StringBuilder sb = new StringBuilder(getBinaryResourceLocation());
      sb.append("/combo-");
      sb.append(!enabled ? "dis" :
                pressed ? "down" :
                hover ? "hi" :
                /*normal*/ "en");
      sb.append(".png");
      
      ImageWrapper<BufferedImage> img = getPngImage(sb.toString());
      if (img != null && img.getImage() != null)
      {
         // this is the location of the button: 
         int x1 = x + width - COMBOBOX_BUTTON_WIDTH;
         int y1 = y;
         // centering image:
         x1 += (COMBOBOX_BUTTON_WIDTH - img.getWidth()) / 2;
         y1 += (height - img.getHeight()) / 2;
         
         gd.drawImage(img, x1, y1, img.getWidth(), img.getHeight(), 0, 0, true, false, true, false);
      }
      
      if (enabled && isDropDown && (hover || pressed)) 
      {
         gd.setColor(pressed ? new Color(0x005499) :
                     /*hover*/ new Color(0x0078D7));
         gd.drawRect(x + width - COMBOBOX_BUTTON_WIDTH - 1, y, COMBOBOX_BUTTON_WIDTH, y + height - 1);
      }
   }
   
   /**
    * Renders the border (and possibly the background) of the combo box widget. Draw 3D border 
    * around different parts of the combo-box widget.
    *
    * @param   combo
    *          The {@code ComboBoxGuiImpl} whose border is being painted.
    * @param   gd
    *          The {@code GuiDriver} used for drawing operations.
    * @param   x
    *          The horizontal location (in pixel units) of the widget to be painted/component.
    * @param   y
    *          The vertical location (in pixel units) of the widget to be painted/component.
    * @param   width
    *          The width of the painted widget/component. 
    * @param   height
    *          The height of the painted widget/component.
    * @param   pressed
    *          {@code true} if the widget is pressed (and eventually open).
    * @param   enabled
    *          Whether the widget has the focus.
    * @param   hover
    *          The mouse is hovering over the fill-in widget.
    */
   @Override
   public void drawComboBoxFieldBorder(ComboBoxGuiImpl combo,
                                       GuiDriver gd,
                                       int x,
                                       int y,
                                       int width,
                                       int height,
                                       boolean pressed,
                                       boolean enabled,
                                       boolean hover)
   {
      pressed = pressed || (combo.hasFocus() && combo.isEnabled());
      boolean simple = combo.isSimpleMode();
      
      gd.setColor(!enabled ? new Color(0xBFBFBF) :
                   pressed ? new Color(0x005499) :
                     hover ? new Color(0x0078D7) :
                    simple ? new Color(0x29618C) :
                  /*normal*/ new Color(0xADADAD));
      gd.drawRect(x, y, width - 1, height - 1);
   }
   
   /**
    * Renders the entry field of a combo box. 
    *
    * @param   combo
    *          The {@code ComboBoxGuiImpl} whose entry field is painted.
    * @param   gd
    *          The {@code GuiDriver} used for drawing operations.
    * @param   gc
    *          The {@code GuiColorResolver} used color selection.
    * @param   x
    *          The horizontal location (in pixel units) of the widget to be painted/component.
    * @param   y
    *          The vertical location (in pixel units) of the widget to be painted/component.
    * @param   width
    *          The width of the painted widget/component. 
    * @param   height
    *          The height of the painted widget/component.
    * @param   font
    *          A {@code FontDetails} used for text rendering.
    * @param   pressed
    *          {@code true} if the widget is pressed (and eventually open).
    * @param   enabled
    *          Whether the widget has the focus.
    * @param   hover
    *          The mouse is hovering over the fill-in widget.
    */
   @Override
   public void drawComboBoxEntryField(ComboBoxGuiImpl combo,
                                      GuiDriver gd,
                                      GuiColorResolver gc,
                                      int x,
                                      int y,
                                      int width,
                                      int height,
                                      FontDetails<?> font,
                                      boolean pressed,
                                      boolean enabled,
                                      boolean hover)
   {
      boolean isDropDown = combo.config().mode == ComboBoxConfig.Mode.DROP_DOWN;
   
      pressed = pressed || (combo.hasFocus() && combo.isEnabled());
   
      if (!enabled)
      {
         gd.setColor(new Color(0xCCCCCC));
      }
      else if (isDropDown)
      {
         gd.setColor(new Color(0xFFFFFF));
      }
      else
      {
         // SIMPLE & DROP_DOWN_LIST cases
         gd.setColor(pressed ? new Color(0xCCE4F7) :
                     hover ? new Color(0xE5F1FB) :
                     /*normal*/ new Color(0xE1E1E1));
      }
      gd.fillRect(x, y, width, height);
      
      // text only in DROP-DOWN-LIST mode. In SIMPLE or DROP-DOWN modes the entry field drawing
      // is happening from the frame the entry field belongs to.
      if (combo.config().mode != ComboBoxConfig.Mode.DROP_DOWN_LIST)
      {
         return;
      }
      
      // get the text to draw with field
      String display = null;
      if (combo.isSelected())
      {
         display = combo.model().selected();
      }
      else
      {
         BaseDataType initial = combo.getValue();
         if (initial != null && !initial.isUnknown())
         {
            display = combo.getInitialScreenText(initial);
         }
      }
      if (display == null)
      {
         display = "";
      }
      
      // limit the data portion of the field to the maximum allowed size
      int limit = combo.getFormatLength();
      
      if (display.length() > limit && combo.isFormatHonored())
      {
         display = display.substring(0, limit);
      }
      
      // draw entry field text
      gd.setGuiFont(font);
      // left-center alignment
      int xt = x + ComboBoxGuiImpl.TEXT_SHIFT;
      int yt = y + height / 2;
      gd.setFontStyle(font.style);
      // enabled state controls text drawing
      gd.setColor(combo.isEnabled() ? new Color(0x000000) : 
                                      new Color(0x6D6D6D));
      int textWidth = width - COMBOBOX_BUTTON_WIDTH - ComboBoxGuiImpl.TEXT_SHIFT;
      final String finalDisplay = display;
      gd.draw(new NativePoint(0, 0), new NativeRectangle(0, 0, height, textWidth),
              () -> gd.drawStringCentered(finalDisplay, xt, yt));
      
      // draw dotted rectangle around focused item text
      if (combo.hasFocus() && combo.isEnabled())
      {
         // black background for dotted line
         gd.setLineStroke(LineStroke.DOTTED_SMALL); // set dotted line style
         gd.setColor(new Color(0x1E1E1E));
         
         gd.drawRect(x + 3, y + 3, textWidth - 3, height - 7);
         gd.setLineStroke(LineStroke.DEFAULT); // restore default line style
      }
   }
   
   /**
    * Renders the background of a selection list. 
    *
    * @param   gd
    *          The {@code GuiDriver} used for drawing operations.
    * @param   x
    *          The horizontal location (in pixel units) of the widget to be painted/component.
    * @param   y
    *          The vertical location (in pixel units) of the widget to be painted/component.
    * @param   width
    *          The width of the painted widget/component. 
    * @param   height
    *          The height of the painted widget/component.
    */
   @Override
   public void drawSelectionListBackground(GuiDriver gd,
                                           GuiColorResolver gc,
                                           int x,
                                           int y,
                                           int width,
                                           int height)
   {
      // fill the background
      gd.setColor(gc.bgColor);
      gd.fillRect(x, y, width, height);
   }
   
   /**
    * Renders a single item in a combo box selection list. 
    *
    * @param   box
    *          The {@code ComboBoxGuiImpl} whose border is being painted.
    * @param   gd
    *          The {@code GuiDriver} used for drawing operations.
    * @param   x
    *          The horizontal location (in pixel units) of the widget to be painted/component.
    * @param   y
    *          The vertical location (in pixel units) of the widget to be painted/component.
    * @param   width
    *          The width of the painted widget/component. 
    * @param   height
    *          The height of the painted widget/component.
    * @param   text
    *          The text to be displayed on this item
    * @param   fontSize
    *          The size of teh font to be used.
    * @param   highlighted
    *          The mouse is hovering over the fill-in widget.
    * @param   enabled
    *          Whether the widget has the focus.
    */
   @Override
   public void drawSelectionListItem(ComboBoxGuiImpl box,
                                     GuiDriver gd,
                                     GuiColorResolver gc,
                                     int x,
                                     int y,
                                     int width,
                                     int height,
                                     String text,
                                     int fontSize,
                                     boolean enabled,
                                     boolean highlighted,
                                     boolean selected)
   {
   
      // fill the background of the selected item
      if (selected && highlighted)
      {
         gd.setColor(new Color(0x3399FF));
         gd.fillRect(x, y, width, height);
      }
      
      // draw item text
      if (selected && highlighted)
      {
         gd.setColor(new Color(0xFFFFFF));
      }
      else if (!enabled)
      {
         gd.setColor(new Color(0x6B6D6B));
      }
      else
      {
         gd.setColor(gc.fgColor);
      }
      int x1 = box != null && box.config().mode == ComboBoxConfig.Mode.SIMPLE
               ? x + getComboBoxEntryShift() + 1 
               : x + ComboBoxGuiImpl.TEXT_SHIFT - 2; /* -2 why ? */
      gd.drawStringCentered(text, x1, y + height / 2);
      
      // draw the dotted line for selected item (not for SIMPLE mode of the combo-box)
      if (enabled && selected &&
            !(box == null || box.config().mode != ComboBoxConfig.Mode.DROP_DOWN_LIST))
      {
         gd.setLineStroke(LineStroke.DOTTED_SMALL);
         gd.setColor(new Color(0xCC6600));
         // take into account possible vertical scrollbar width 
         gd.drawRect(x, y, width - 1 - box.getDropDownScrollBarWidth(), height - 1);
         // restore default line style
         gd.setLineStroke(LineStroke.DEFAULT);
      }
   }
   
   /**
    * Obtain the vertical distance between the combo box and the scrolling selection list.  
    *
    * @return the vertical offset for selection list of the combo box.
    */
   @Override
   public int getComboBoxDropDownOffset()
   {
      return 0; // dropdown list is right under combobox
   }
   
   /**
    * Obtain the minimum size of a titlebar element for this Theme, including any border. 
    *
    * @param   horiz
    *          If {@code true} returns the minimum width, if {@code false} returns minimum height.
    *
    * @return  The minimum size of a titlebar element on specified direction.
    */
   @Override
   public int getMenuBarElementMinSize(boolean horiz)
   {
      return horiz ? 32 : 20;
   }
   
   /**
    * Obtain the left padding of the test in a menu item. 
    *
    * @return  left padding of the test in a menu item. 
    */
   @Override
   public int getMenuLeftPad()
   {
      return MENU_LEFT_PAD;
   }
   
   /**
    * Obtain the right padding of the test in a menu item. 
    *
    * @return  right padding of the test in a menu item. 
    */
   @Override
   public int getMenuRightPad()
   {
      return MENU_RIGHT_PAD;
   }
   
   /**
    * Obtain the width of the check mark to for togglable menus. 
    *
    * @return  the width of the check mark in menu items.
    */
   @Override
   public int getMenuCheckmarkWidth()
   {
      return MENU_CHECKMARK_WIDTH;
   }
   
   /**
    * Obtain the fixed height of a separator in a popup menu widget.
    *
    * @return  the height of a separator in a popup menu.
    */
   @Override
   public int getMenuSeparatorHeight()
   {
      return 7;
   }
   
   /**
    * Obtain the height of a {@link MenuElement} in a menu widget.
    *
    * @return  the height of a normal item in a menu widget.
    */
   @Override
   public int getMenuItemHeight(MenuElement menuItem)
   {
      return 22; // use font size to compute it 
   }
   
   /**
    * Renders the background of a menu widget. 
    * @param   gd
    *          The {@code GuiDriver} used for drawing operations.
    * @param   x
    *          The horizontal location (in pixel units) of the widget to be painted/component.
    * @param   y
    *          The vertical location (in pixel units) of the widget to be painted/component.
    * @param   width
    *          The width of the painted widget/component. 
    * @param   height
    *          The height of the painted widget/component.
    * @param   popup
    *          {@code true} for popup menus.
    */
   @Override
   public void drawMenuBackground(GuiDriver gd,
                                  int x,
                                  int y,
                                  int width,
                                  int height,
                                  boolean popup)
   {
      if (!popup)
      {
         // this is a menu bar
         gd.setColor(gd.getSysColor(COLOR_MENUBAR));
         gd.fillRect(x, y, width, height);
         return;
      }
      
      NativeInsets insets = getMenuInsets();
      
      gd.setColor(gd.getSysColor(COLOR_MENU));
      gd.fillRect(x, y, width, height);
      
      gd.setColor(new Color(0xCCCCCC));
      gd.drawRect(x, y, x + width - 1, y + height - 1);
      
      gd.setColor(new Color(0xF0F0F0));
      gd.fillRect(x + insets.left, y + insets.top, 
            MENU_CHECKMARK_WIDTH, height - insets.top - insets.bottom);
      // draw vertical separator
      gd.setColor(new Color(0xD7D7D7));
      gd.drawLine(x + MENU_CHECKMARK_WIDTH + insets.left + MENU_LEFT_PAD / 2,
                  y + insets.top, 
                  x + MENU_CHECKMARK_WIDTH + insets.left + MENU_LEFT_PAD / 2,
                  y + height - insets.bottom - 1);
   }
   
   /**
    * Renders the separator line in a menu widget. 
    *
    * @param   gd
    *          The {@code GuiDriver} used for drawing operations.
    * @param   x
    *          The horizontal location (in pixel units) of the widget to be painted/component.
    * @param   y
    *          The vertical location (in pixel units) of the widget to be painted/component.
    * @param   width
    *          The width of the painted widget/component. 
    * @param   height
    *          The height of the painted widget/component.
    */
   @Override
   public void drawMenuSeparator(GuiDriver gd, int x, int y, int width, int height)
   {
      gd.setColor(new Color(0xD7D7D7));
      gd.drawLine(x + MENU_CHECKMARK_WIDTH + MENU_LEFT_PAD / 2, y + height / 2, 
                  x + width, y + height / 2);
   }
   
   /**
    * Draws an element of a menu in a menubar.
    *
    * @param   me
    *          The {@code MenuElement} widget to be painted.
    * @param   gd
    *          The {@code GuiDriver} used for drawing operations.
    * @param   gf
    *          GUI font resolver.
    * @param   x
    *          The horizontal location (in pixel units) of the widget to be painted/component.
    * @param   y
    *          The vertical location (in pixel units) of the widget to be painted/component.
    * @param   width
    *          The width of the painted widget/component. 
    * @param   height
    *          The height of the painted widget/component.
    * @param   hover
    *          {@code true} if this item is highlighted/focused.
    */
   @Override
   public void drawMenuBarElement(MenuElement me,
                                  GuiDriver gd,
                                  GuiFontResolver gf,
                                  int x,
                                  int y,
                                  int width,
                                  int height,
                                  boolean hover)
   {
      if (!Menu.isMenubarElement(me))
      {
         System.err.println("Who are you?"); // what ?
         return;
      }

      // drawing background if the case
      boolean pressed = me.isCurrent();
      if (pressed || hover)
      {
// a bit differen colors taken from Windows 10 screenshots
//         gd.setColor(pressed ? new Color(0xCCE8FF) : 
//                    /* hover*/ new Color(0xE5F3FF));
         gd.setColor(pressed ? new Color(0xC4E5F6) : 
                    /* hover*/ new Color(0xBEE6FD));
         
         gd.fillRect(x + 1, y + 1, width - 2, height - 2);
         
// a bit differen colors taken from Windows 10 screenshots
//         gd.setColor(pressed ? new Color(0x99D1FF) : 
//                    /* hover*/ new Color(0xCCE8FF));
         gd.setColor(pressed ? new Color(0x2C628B) : 
                    /* hover*/ new Color(0x26A0DA));
         
         gd.drawRect(x, y, width - 1, height - 1);
      }
      
      // and the foreground
      if (me.isEnabled())
      {
         gd.setColor(gd.getSysColor(COLOR_MENUTEXT));
      }
      else
      {
         gd.setColor(new Color(0x808080)); // not in palette ?
      }
      
      String title = me.getPreparedTitle();
      MnemonicInfo label = me.getPreparedLabel();
      
      int w = ((AbstractWidget) me).getTextWidth(title, gf.resolveFontNum());
      int tY = y + height / 2 - 1;
      int tX = (width - w) / 2;
      
      drawMenuTitle(me, gd, gf, title, label, tX, tY);
   }
   
   /**
    * Renders the background of an item in a menu widget. 
    *
    * @param   item
    *          The {@link MenuItemGuiImpl} or {@link SubMenuGuiImpl} to be drawn.
    * @param   gd
    *          The {@code GuiDriver} used for drawing operations.
    * @param   x
    *          The horizontal location (in pixel units) of the widget to be painted/component.
    * @param   y
    *          The vertical location (in pixel units) of the widget to be painted/component.
    * @param   width
    *          The width of the painted widget/component. 
    * @param   height
    *          The height of the painted widget/component.
    * @param   hover
    *          {@code true} if this item is highlighted/focused.
    */
   @Override
   public void drawMenuItemBackground(AbstractWidget item,
                                      GuiDriver gd,
                                      int x,
                                      int y,
                                      int width,
                                      int height,
                                      boolean hover)
   {
      if (Menu.isMenubarElement(item))
      {
         System.err.println(
               "WARNING: Use Theme.drawMenuBarElement() instead to render a Menu Bar element");
         return;
      }
      
      if (hover && item.isEnabled())
      {
         // highlight consist of filled color and thin frame around
         gd.setColor(new Color(0xDDECFF));
         gd.fillRect(1, 1, width - 2, height - 2);
         
         gd.setColor(new Color(0x25A0DB));
         gd.drawRect(0, 0, width - 1, height - 1);
      }
      
      ThemeManager.debug(() -> {
         gd.setColor(Color.GREEN);
         gd.drawRect(0, 0, width - 1, height - 1);
      });
   }
   
   /**
    * Renders the foreground content of menu item in a popup menu. 
    *
    * @param   item
    *          The {@link MenuItemGuiImpl} or {@link SubMenuGuiImpl} to be drawn.
    * @param   gd
    *          The {@code GuiDriver} used for drawing operations.
    * @param   x
    *          The horizontal location (in pixel units) of the widget to be painted/component.
    * @param   y
    *          The vertical location (in pixel units) of the widget to be painted/component.
    * @param   width
    *          The width of the painted widget/component. 
    * @param   height
    *          The height of the painted widget/component.
    * @param   textHeight
    *          The height of the text.
    * @param   focused
    *          {@code true} if this item is highlighted/focused.
    */
   @Override
   public void drawMenuItemContent(AbstractWidget item,
                                   GuiDriver gd,
                                   GuiFontResolver gf,
                                   int x,
                                   int y,
                                   int width,
                                   int height,
                                   int textHeight,
                                   boolean focused)
   {
      boolean enabled = item.isEnabled();
      int checkmarkWidth = getMenuCheckmarkWidth();
      String title;
      MnemonicInfo label;
      
      if (item instanceof MenuItemGuiImpl)
      {
         MenuItemGuiImpl mi = (MenuItemGuiImpl) item;
         MenuItemConfig cfg = mi.config();
         // the space for toggle box. Printed for all menu-items in sub-menus or pop-up menus
         if (cfg.toggleBox && cfg.checked)
         {
            ImageWrapper<BufferedImage> check = getPngImage(
                  getBinaryResourceLocation() + 
                  (enabled ? "/menu-checkmark-en.png" : "/menu-checkmark-dis.png"));
            if (check != null && check.getImage() != null)
            {
               int cmOffX = (checkmarkWidth - check.getWidth()) / 2; 
               int cmOffY = (height - check.getHeight()) / 2; 
               gd.drawImage(check, cmOffX, cmOffY, check.getWidth(), check.getHeight(),
                            0, 0, false, false, true, false);
            }
         }
         
         title = mi.getPreparedTitle();
         label = mi.getPreparedLabel();
      }
      else // SUbMenuGuiImpl case
      {
         SubMenuGuiImpl sm = (SubMenuGuiImpl) item;
         
         title = sm.getPreparedTitle();
         label = sm.getPreparedLabel();
      }
      
      if (enabled)
      {
         gd.setColor(new Color(0x000000));
      }
      else
      {
         gd.setColor(new Color(0x808080)); // not in palette ?
      }
      
      int textYoffset = y + height / 2;
      NativePoint np = new NativePoint(x + checkmarkWidth + getMenuLeftPad(), 0);
      int tw = width - checkmarkWidth - getMenuLeftPad() - getMenuSmArrowWidth();
      NativeRectangle clip = AbstractWidget.clipRectangle(np, new NativeDimension(tw, height));
      gd.draw(np, clip, () -> 
            drawMenuTitle((MenuElement) item, gd, gf, title, label, 0, textYoffset));
      
      if (item instanceof MenuItemGuiImpl)
      {
         String accel = ((MenuItemGuiImpl) item).config().accelerator;
         if (accel != null && !accel.isEmpty())
         {
            // accelerators may be drawn with different color
            gd.setColor(enabled ? new Color(0x000000) :
                                  new Color(0x808080));
            // TODO: this is right aligned, it should be right aligned instead
            String accText = Keyboard.keyName(accel);
            int accWidth = item.getTextWidth(accText, gf.resolveFontNum());
            int accOffset = width - accWidth - getMenuSmArrowWidth(); 
            gd.drawStringCentered(accText, accOffset, textYoffset);
         }
      }
      else
      {
         ImageWrapper<BufferedImage> arrow = getPngImage(
               getBinaryResourceLocation() +
               (enabled ? "/menu-submenu-en.png" : "/menu-submenu-dis.png"));
         if (arrow != null && arrow.getImage() != null)
         {
            gd.drawImage(arrow, width - getMenuSmArrowWidth(), (height - arrow.getHeight()) / 2,
                  arrow.getWidth(), arrow.getHeight(), 0, 0, false, false, true, false);
         }
      }
   }
   
   /**
    * Draws title string at the specified point.
    *
    * @param   me
    *          {@code MenuItem} or {@code SubMenu} widget.
    * @param   gd
    *          GUI driver.
    * @param   gf
    *          GUI font resolver.
    * @param   title
    *          The title of menu widget. 
    * @param   label
    *          Prepared title string.
    * @param   tX
    *          Text starting point.
    * @param   tY
    *          Offset to center text.
    *
    * @throws  IllegalArgumentException
    *          When {@code me} is not a {@code MenuItem} or {@code SubMenu}.
    */
   @SuppressWarnings("rawtypes")
   private void drawMenuTitle(MenuElement me,
                              GuiDriver gd,
                              GuiFontResolver gf,
                              String title,
                              MnemonicInfo label,
                              int tX,
                              int tY) 
   {
      gd.setGuiFont(gf.font().font);
      gd.setFontStyle(gf.font().font.style);
      
      MenuGuiImpl menu = (MenuGuiImpl) Menu.getMenu(me);
      if (menu.isMnemonicsOn() && label != null && label.getMnemIdx() != null)
      {
         gd.drawStringWithMnemonic(title, new NativePoint(tX, tY), label, gf, menu, true);
      }
      else
      {
         gd.drawStringCentered(title, tX, tY);
      }
   }
   
   /**
    * Utility method that loads a image (recommended in PNG format because it is lossless and has
    * support for full 8 bit alpha channel, but not otherwise restricted to). If resource is
    * successfully loaded, it is returned in wrapper and cached for subsequent fast accesses.
    * 
    * @param   imagePath
    *          The path to image resource inside the bundle (jar).
    *          
    * @return  an {@code ImageWrapper} containing the requested image resource or {@code null} if
    *          some error occurred (resource could not be found or it could not be decoded).
    */
   protected ImageWrapper<BufferedImage> getPngImage(String imagePath)
   {
      ImageWrapper<BufferedImage> img = imgCache.get(imagePath);
      if (img == null)
      {
         InputStream res = Windows10Theme.class.getResourceAsStream(imagePath);
         if (res != null)
         {
            try
            {
               BufferedImage image = ImageIO.read(res);
               img = new ImageWrapper<>(image, image.getHeight(), image.getWidth());
            }
            catch (IOException e)
            {
               // TODO: create a dummy image ?
            }
         }
         
         if (img == null)
         {
            System.err.println("Failed to load resource: [" + imagePath+ "].");
         }
         
         // if image can't be found, store null to avoid repeating looking up non-existent image 
         imgCache.put(imagePath, img);
      }
      return img;
   }
   
   /**
    * Draw the title panel of a {@link BrowseGuiImpl} widget.
    * 
    * @param   browse
    *          The {@code BrowseGuiImpl} widget.
    * @param   gd
    *          The {@code GuiDriver} used for drawing operations.
    * @param   gc
    *          The {@code GuiColorResolver} used color selection.
    * @param   gf
    *          The {@code GuiFontResolver} to be used.
    * @param   x
    *          The horizontal location (in pixel units) of the widget to be painted/component.
    * @param   y
    *          The vertical location (in pixel units) of the widget to be painted/component.
    * @param   width
    *          The width of the painted widget/component. 
    * @param   height
    *          The height of the painted widget/component. 
    */
   @Override
   public void drawBrowseTitlePanel(BrowseGuiImpl browse,
                                    GuiDriver gd,
                                    GuiColorResolver gc,
                                    GuiFontResolver gf,
                                    int x,
                                    int y,
                                    int width,
                                    int height)
   {
      boolean focused = browse.isDrawnAsFocused();
      String title = browse.config().title;
      
      gd.setColor(focused ? new Color(0x9CB6D6) 
                          : new Color(0xBDCFDE));
      gd.fillRect(x + 1, y + 1, width - 1, height - 1);
      // add a 3d effect
      gd.setColor(new Color(0xFFFFFF));
      gd.drawLine(x, y, x + width - 1, y);
      gd.drawLine(x, y, x,             y + height - 1);
      
      // calculate text position
      int textWidth = browse.getTextWidth(title, gf.resolveFontNum());
      x += (width - textWidth) / 2;
      
      // draw title string
      gd.setColor(focused ? new Color(0x000000)
                          : new Color(0x000000));
      gd.setGuiFont(gf.font().font);
      gd.drawStringCentered(title,
            x,
            browse.getTitleHeight() / 2 + BrowseGuiImpl.BORDER_THICKNESS);
   }
   
   /**
    * Draw label-style rectangle.
    *
    * @param   column
    *          The {@code BrowseGuiImpl} widget.
    * @param   gd
    *          The {@code GuiDriver} used for drawing operations.
    * @param   topOffset
    *          Top offset, in pixels.
    * @param   markersBody
    *          <code>true</code> if it is a cell in the markers column.
    * @param   markersLabel
    *          <code>true</code> if it is the label area in the markers column.
    * @param   selectedRow
    *          Is the current row selected ?
    */
   @Override
   protected void drawBrowseColumnLabelRect(BrowseColumnGuiImpl column,
                                            GuiDriver gd,
                                            int topOffset,
                                            boolean markersBody,
                                            boolean markersLabel,
                                            boolean selectedRow)
   {
      NativeDimension d = column.physicalDimension();
      BrowseGuiImpl browse = column.getBrowse();
      int width = d.width;
      int height = browse.getLabelsHeight() + BrowseGuiImpl.BORDER_THICKNESS;
      if (!markersBody)
      {
         // marker-body height does not include separator lines 
         height += BrowseGuiImpl.BORDER_THICKNESS;
      }
      
      gd.setLineStroke(LineStroke.DEFAULT); // be safe ?
      BrowseConfig browseConfig = browse.config();
      if (markersBody || markersLabel)
      {
         if (browseConfig.separators)
         {
            // set the dark border color
            gd.setColor(getBrowseCellSeparatorColor(browse));
            gd.drawRect(0, topOffset, width - 1, height - 1);
         }
         
         // marker label
         drawBrowseColumnMarker(gd, selectedRow, topOffset, width, height);
      }
      else
      {
         // normal label
         
         // check override color first:
         ColorRgb bkColor = null;
         
         if (column.config().ehLabelBgColor != null)
         {
            bkColor = column.config().ehLabelBgColor;
         }
         else if (browseConfig.ehLabelBgColor != null)
         {
            bkColor = browseConfig.ehLabelBgColor;
         }
         else
         {
            int overColumnColor = column.config().labelBgColor;
            if (overColumnColor >= 0)
            {
               bkColor = ColorManager.getBgColor(column.window(), overColumnColor);
            }
            else
            {
               int overBrowseColor = browseConfig.labelBgColor;
               if (overBrowseColor >= 0)
               {
                  bkColor = ColorManager.getBgColor(column.window(), overBrowseColor);
               }
            }
         }
         // background Color
         if (bkColor != null)
         {
            gd.setColor(bkColor);
         }
         else
         {
            gd.setColor(column.isDrawnAsPressed() ? new Color(0xBDDBFF)
                                                  : new Color(0xFFFFFF));
         }
         
         // one additional pixel to the left
         int dx = column.isFirstColumn() ? 0 : -1;
         // fill entire label rect
         gd.fillRect(dx + 1, topOffset + 1, width - 2 - dx, height - 2);
         
         int w = width - 1 - dx;
         // one pixel more to the height
         int h = height - 1;

         drawBrowseColumnLabelBorder(column, gd, dx, topOffset, w, h);
      }
   }
   
   /**
    * Draw browse row marker column resctangle.
    *
    * @param   gd
    *          The {@code GuiDriver} used for drawing operations.
    * @param   selected
    *          Is the current row selected ?
    * @param   topOffset
    *          Top offset, in pixels.
    * @param   w
    *          The border width in pixels.
    * @param   h
    *          The border height in pixels.
    */
   protected void drawBrowseColumnMarker(GuiDriver gd,
                                         boolean selected,
                                         int topOffset,
                                         int w,
                                         int h)
   {
      if (selected)
      {
         gd.setColor(gd.getSysColor(SysColor.COLOR_3DSHADOW));
         gd.fillRect(0, topOffset, w - 1, h - 1); // leave S and E as background color

         gd.setColor(gd.getSysColor(SysColor.COLOR_WINDOW));
         gd.drawLine(w - 2, topOffset + 2, w - 2, topOffset + h - 2);
         gd.drawLine(1, topOffset + h - 2, w - 2, topOffset + h - 2);
      }
      else
      {
         gd.setColor(gd.getSysColor(COLOR_3DFACE));
         gd.fillRect(0, topOffset, w - 1, h - 1);

         gd.setColor(gd.getSysColor(SysColor.COLOR_WINDOW));
         gd.drawLine(1, topOffset + 1, 1, topOffset + h - 2);
         gd.drawLine(1, topOffset + 1, w - 3, topOffset + 1);

         gd.setColor(gd.getSysColor(SysColor.COLOR_3DSHADOW));
         gd.drawLine(w - 2, topOffset + 2, w - 2, topOffset + h - 2);
         gd.drawLine(1, topOffset + h - 2, w - 2, topOffset + h - 2);
      }
   }
   
   /**
    * Draw label-style rectangle.
    *
    * @param   column
    *          The {@code BrowseColumnGuiImpl} widget.
    * @param   gd
    *          The {@code GuiDriver} used for drawing operations.
    * @param   xOffset
    *          Left offset, in pixels.
    * @param   yOffset
    *          Top offset, in pixels.
    * @param   w
    *          The border width in pixels.
    * @param   h
    *          The border height in pixels.
    */
   protected void drawBrowseColumnLabelBorder(BrowseColumnGuiImpl column,
                                              GuiDriver gd,
                                              int xOffset,
                                              int yOffset,
                                              int w,
                                              int h)
   {
      // set the border color
      gd.setColor(column.isDrawnAsPressed() ? new Color(0x21A2DE)
                                            : new Color(0xE7E3E7));
      // and draw the border itself depending on pressed flag
      if (isBrowseLabelBorderDrawQuirk(column))
      {
         NativePoint clipPoint = new NativePoint(xOffset, yOffset);
         NativeRectangle clipRect =
               new NativeRectangle(clipPoint, new NativeDimension(w, h + 1));
         gd.draw(clipPoint, clipRect, () -> gd.drawRect(0, 0, w, h));
      }
      else
      {
         gd.drawRect(xOffset, yOffset, w, h);
      }
   }
   
   /**
    * Sets the composite color for XOR operations on browse widgets.
    *
    * @param   gd
    *          the active {@code GuiDriver} used to draw the browse widget.
    */
   @Override
   protected void setCompositeXorColor(GuiDriver gd)
   {
      gd.setColor(new Color(0xC6C3C6));
   }
   
   /**
    * Get the cell separator color.
    *
    * @param   browse
    *          The browse widget.
    *
    * @return  cell separator color.
    */
   @Override
   public ColorRgb getBrowseCellSeparatorColor(BrowseGuiImpl browse)
   {
      return browse.config().ehSeparatorsColor != null ?
            browse.config().ehSeparatorsColor :
            new ColorRgb(new Color(0xC6C3C6));
   }
   
   /**
    * Set the selected cell background color into the {@code gd} {@link GuiDriver}.
    *
    * @param   gd
    *          The {@code GuiDriver} to set the color for separators.
    * @param   unselColor
    *          The color for unselected cells. If {@code null} then the unselected cells' 
    *          background is not drawn. 
    */
   @Override
   protected void setBrowseCellBackgroundSelectionColor(GuiDriver gd, ColorRgb unselColor)
   {
      // TODO: see whether the selection color is computed from [unselColor] like in Windows8
      gd.setColor(gd.getSysColor(COLOR_HIGHLIGHT));
   }
   
   /**
    * Get row marker image for a row.
    *
    * @param   selected
    *          {@code true} if the row is selected.
    *
    * @return row marker image for a row.
    */
   @Override
   protected ImageWrapper getBrowseMarkerArrow(boolean selected)
   {
      return getPngImage(getBinaryResourceLocation() + 
                         (selected ? "/browse-marker-sel.png" : "/browse-marker.png"));
   }
   
   /**
    * Draw a {@link ToolTip}.
    *
    * @param   gd
    *          The {@code GuiDriver} used for drawing operations.
    * @param   x
    *          The horizontal location (in pixel units) of the widget to be painted/component.
    * @param   y
    *          The vertical location (in pixel units) of the widget to be painted/component.
    * @param   width
    *          The width of the painted widget/component. 
    * @param   height
    *          The height of the painted widget/component.
    * @param   fnt
    *          The {@code FontDetails} to be used when drawing the text.
    * @param   text
    *          The actual text to be drawn.
    */
   @Override
   public void drawToolTip(GuiDriver gd,
                           int x,
                           int y,
                           int width,
                           int height,
                           FontDetails fnt,
                           String text)
   {
      // start batch drawing
      gd.startBatch();
      try
      {
         // another version of clipping
         gd.setDrawingArea(new NativeRectangle(y, x, y + height - 1, x + width - 1));
         
         // tooltip background
         gd.setColor(gd.getSysColor(COLOR_TOOLTIPBK));
         gd.fillRect(x + 1, y + 1, width - 2, height - 2);
         
         // tooltip frame
         gd.setColor(new Color(0x636563));
         gd.drawRect(x, y, width - 1, height - 1);
         
         // tooltip text
         gd.setGuiFont(fnt);
         gd.setFontStyle(fnt.style);
         gd.setColor(new Color(0x000000));
         
         String lines[] = text.trim().split("\\r?\\n");
         int tH = (height - 4) / lines.length;
         int tY = y + tH / 2 + 2;
         for (String line : lines)
         {
            gd.drawStringCentered(line, x + 2, tY);
            tY += tH;
         }
      }
      finally
      {
         // draw immediately
         gd.setDrawingArea(null);
         gd.endBatch();
      }
   }
   
   /**
    * Draw the title panel of a {@link BrowseGuiImpl} widget.
    * <br>
    * TODO: this code is very similar to {@link #drawBrowseTitlePanel}. 
    *
    * @param   frame
    *          The {@code FrameGuiImpl} widget.
    * @param   gd
    *          The {@code GuiDriver} used for drawing operations.
    * @param   gc
    *          The {@code GuiColorResolver} used color selection.
    * @param   gf
    *          The {@code GuiFontResolver} to be used.
    * @param   x
    *          The horizontal location (in pixel units) of the widget to be painted/component.
    * @param   y
    *          The vertical location (in pixel units) of the widget to be painted/component.
    * @param   width
    *          The width of the painted widget/component. 
    * @param   height
    *          The height of the painted widget/component. 
    */
   @Override
   public void drawFrameTitle(FrameGuiImpl frame,
                              GuiDriver gd,
                              GuiColorResolver gc,
                              GuiFontResolver gf,
                              int x,
                              int y,
                              int width,
                              int height)
   {
      String title = frame.config().title;
      
      // draw the frame title background depending on focused state
      gd.setColor(frame.hasFocus() ? new Color(0x9CB4D1)
                                   : new Color(0xBFCDDB));
      gd.fillRect(x, y, width, height);
      
      if (title == null)
      {
         return;
      }

      // calculate centered text position
      int textWidth = frame.getTextWidthNative(title, gf.resolveFontNum());
      int tX = x + (width - textWidth) / 2 - 1;
      int tY = y + (height - 1) / 2 - 1;
      
      // draw title string
      gd.setColor(new Color(0x000000));
      gd.setGuiFont(gf.font().font);
      gd.drawStringCentered(title, tX, tY);
   }

   /**
    * Get the offset of the in-browse toggle-box relatively to the containing cell.
    *
    * @param  column
    *         The column that contains the toggle-box.
    * @param  editor
    *         <code>true</code> if the toggle-box is contained in the current cell, i.e. it is the
    *         editor. <code>false</code> for all other inactive toggle boxes.
    *
    * @return the offset of the in-browse toggle-box relatively to the containing cell, in native
    *         units.
    */
   public NativePoint getRowToggleOffset(BrowseColumnGuiImpl column, boolean editor)
   {
      NativePoint point = getRowToggleOffsetCentered(column, editor);
      if (editor)
      {
         point.x = Math.max(0, point.x);
      }
      return point;
   }

   /**
    * Get the size of the in-browse toggle-box.
    *
    * @param  column
    *         The column that contains the toggle-box.
    * @param  editor
    *         <code>true</code> if the toggle-box is contained in the current cell, i.e. it is the
    *         editor. <code>false</code> for all other inactive toggle boxes.
    *
    * @return the size of the in-browse toggle-box, in native units.
    */
   public NativeDimension getRowToggleSize(BrowseColumnGuiImpl column, boolean editor)
   {
      if (editor)
      {
         return new NativeDimension(ROW_TOGGLE_SIZE, ROW_TOGGLE_SIZE);
      }

      int availWidth = column.physicalDimension().width;
      int availHeight = column.getBrowse().getRowHeight();
      int min = Math.min(Math.min(availHeight, availWidth), ROW_TOGGLE_SIZE);
      return new NativeDimension(min, min);
   }
   
   /**
    * Draw the line border {@link LineBorderGuiImpl} widget.
    *
    * @param   border
    *          The {@code FrameGuiImpl} widget.
    * @param   gd
    *          The {@code GuiDriver} used for drawing operations.
    * @param   gc
    *          The {@code GuiColorResolver} used color selection.
    * @param   x
    *          The horizontal location (in pixel units) of the widget to be painted/component.
    * @param   y
    *          The vertical location (in pixel units) of the widget to be painted/component.
    * @param   width
    *          The width of the painted widget/component. 
    * @param   height
    *          The height of the painted widget/component. 
    */
   @Override
   public void drawLineBorder(LineBorderGuiImpl border,
                              GuiDriver gd,
                              GuiColorResolver gc,
                              int x,
                              int y,
                              int width,
                              int height)
   {
      NativeInsets ni = border.getNativeInsets();
      
      // setup line color and width
      AbstractWidget lw = border.getLegacyWidget();
      if (lw instanceof MessageAreaGuiImpl)
      {
         boolean mouseOver = false; // TODO: add this as parameter or obtain from [lw] (somehow)
         gd.setColor(lw.hasFocus() ? new Color(0xBDBABD) :
                     mouseOver     ? new Color(0x21A2CE) :
                                     new Color(0xDEDFDE));
      }
      else if (lw instanceof FrameGuiImpl)
      {
         gd.setColor(new Color(0x636563));
      }
      else if (lw instanceof BrowseGuiImpl)
      {
         gd.setColor(new Color(0xADAEB5));
      }
      else
      {
         if (gc.boxColor != null)
         {
            gd.setColor(gc.boxColor);
         }
         else
         {
            gd.setColor(new Color(0xDEDFDE));
         }
      }
      
      // determine if insets are different; if not, set the line-stroke only once
      boolean equalInsets = ni.top == ni.left && ni.top == ni.bottom && ni.top == ni.right;
      if (equalInsets)
      {
         gd.setLineStroke(LineStroke.SPECIFIED_WIDTH, (float) ni.top);
      }
      
      // top line
      if (ni.top > 0)
      {
         if (!equalInsets)
         {
            gd.setLineStroke(LineStroke.SPECIFIED_WIDTH, (float) ni.top);
         }
         
         gd.drawLine(x, y, width, y);
      }
      
      // left line
      if (ni.left > 0)
      {
         if (!equalInsets)
         {
            gd.setLineStroke(LineStroke.SPECIFIED_WIDTH, (float) ni.left);
         }
         
         gd.drawLine(x, y, x, y + height);
      }
      
      // bottom line
      if (ni.bottom > 0)
      {
         if (!equalInsets)
         {
            gd.setLineStroke(LineStroke.SPECIFIED_WIDTH, (float) ni.bottom);
         }
         
         gd.drawLine(x, y + height - ni.bottom, x + width, y + height - ni.bottom);
      }
      
      // right line
      if (ni.right > 0)
      {
         if (!equalInsets)
         {
            gd.setLineStroke(LineStroke.SPECIFIED_WIDTH, (float) ni.right);
         }
         
         gd.drawLine(x + width - ni.right, y + height, x + width - ni.right, y);
      }
      
      // restore the line stroke
      gd.setLineStroke(LineStroke.DEFAULT);
   }
   
   /**
    * Draws a {@link RectangleGuiImpl} widget.
    *
    * @param   rect
    *          The {@code RectangleGuiImpl} widget.
    * @param   gd
    *          The {@code GuiDriver} used for drawing operations.
    * @param   gc
    *          The {@code GuiColorResolver} used color selection.
    * @param   x
    *          The horizontal location (in pixel units) of the widget to be painted/component.
    * @param   y
    *          The vertical location (in pixel units) of the widget to be painted/component.
    * @param   width
    *          The width of the painted widget/component. 
    * @param   height
    *          The height of the painted widget/component. 
    */
   @Override
   public void drawRectangle(RectangleGuiImpl rect,
                             GuiDriver gd,
                             GuiColorResolver gc,
                             int x,
                             int y,
                             int width,
                             int height)
   {
      RectangleConfig cfg = rect.config();
      if (cfg.groupBox)
      {
         // in group box, fill, roundness and line thickness are ignored
         int ROUNDNESS = 5; // rounding diameter in pixels
         gd.setColor(gd.getSysColor(SysColor.COLOR_WINDOW)); //light 
         gd.drawRoundRect(x + 1, y + 1, width - 4, height - 3, ROUNDNESS - 2);
         gd.fillRect(x,             y + height - 4, 1, 1);
         gd.fillRect(x + width - 2, y + height - 4, 1, 1);
         gd.setColor(new Color(0xD5DFE5)); //dark
         gd.drawRoundRect(x, y, width - 2, height - 3, ROUNDNESS);
         return;
      }
      
      // for other cases use the windows classic renderer
      super.drawRectangle(rect, gd, gc, x, y, width, height);
   }
   
   /**
    * Draws a {@link ThreeDBorderGuiImpl} widget.
    *
    * @param   border
    *          The {@code RectangleGuiImpl} widget.
    * @param   gd
    *          The {@code GuiDriver} used for drawing operations.
    * @param   x
    *          The horizontal location (in pixel units) of the widget to be painted/component.
    * @param   y
    *          The vertical location (in pixel units) of the widget to be painted/component.
    * @param   width
    *          The width of the painted widget/component.
    * @param   height
    *          The height of the painted widget/component.
    */
   @Override
   public void draw3DBorder(ThreeDBorderGuiImpl border,
                            GuiDriver gd,
                            int x,
                            int y,
                            int width,
                            int height)
   {
      gd.setColor(new Color(0x737173));
      gd.drawRect(x, y, x + width - 1, y + height - 1);
   }
   
   /**
    * Utility method for drawing a dotted rectangle using XOR operation.
    *
    * @param   gd
    *          The {@code GuiDriver} used for drawing operations.
    * @param   x
    *          The horizontal location (in pixel units) of the widget to be painted/component.
    * @param   y
    *          The vertical location (in pixel units) of the widget to be painted/component.
    * @param   width
    *          The width of the painted widget/component. 
    * @param   height
    *          The height of the painted widget/component.
    */
   @Override
   public void drawSelectionFocus(GuiDriver gd, int x, int y, int width, int height)
   {
      // prepare stroke:
      gd.setXORComposite();
      gd.setColor(new Color(0xFFFFFF));
      gd.setLineStroke(LineStroke.DOTTED_SMALL);
      
      // draw the dotted rectangle 
      gd.drawRect(x, y, width, height);
      
      // restore back the default line style
      gd.setLineStroke(LineStroke.DEFAULT);
      gd.resetComposite();
   }

   /**
    * Get left offset of a text inside a browse label.
    *
    * @param  rightAligned
    *         <code>true</code> if the column is right-aligned. <code>false</code> if the column
    *         is left-aligned.
    *
    * @return see above.
    */
   public int getBrowseLabelOffsetLeft(boolean rightAligned)
   {
      return BROWSE_LABEL_OFFSET_LEFT;
   }

   /**
    * Get left offset of a text inside a browse cell.
    *
    * @param  rightAligned
    *         <code>true</code> if the column is right-aligned. <code>false</code> if the column
    *         is left-aligned.
    *
    * @return see above.
    */
   public int getBrowseDataOffsetLeft(boolean rightAligned)
   {
      return BROWSE_DATA_OFFSET_LEFT;
   }

   /**
    * Get the image of a non-edited in-browse toggle-box.
    *
    * @param  checked
    *         <code>true</code> if the toggle-box is checked, <code>false</code> it is not
    *         checked.
    *
    * @return the image of a non-edited in-browse toggle-box.
    */
   protected ImageWrapper getRowToggleImage(boolean checked)
   {
      return getPngImage(getBinaryResourceLocation() + 
                         (checked ? "/checkbox-on-en.png" : "/checkbox-off-en.png"));
   }
   
   /**
    * Draws a File or Folder Thumbnail of from a file/dir chooser dialog.
    *
    * @param   gd
    *          The {@code GuiDriver} used for drawing operations.
    * @param   x
    *          The horizontal location (in pixel units) of the widget to be painted/component.
    * @param   y
    *          The vertical location (in pixel units) of the widget to be painted/component.
    * @param   width
    *          The width of the painted widget/component.
    * @param   height
    *          The height of the painted widget/component.
    * @param   selected
    *          {@code true} if the thumbnail is currently selected.
    * @param   hover
    *          {@code true} if the mouse is currently hovering the thumbnail.
    * @param   focus
    *          {@code true} if the thumbnail has focus.
    */
   @Override
   public void drawFileChooserThumbnailBackground(GuiDriver gd,
                                                  int x,
                                                  int y,
                                                  int width,
                                                  int height,
                                                  boolean selected,
                                                  boolean hover,
                                                  boolean focus)
   {
      if (selected)
      {
         gd.setColor(new ColorRgb(new Color(0xCEE3FF)));
         gd.fillRect(x + 1, y + 1, width - 1, height - 1);
         gd.setColor(new ColorRgb(new Color(0x7BA2CE)));
         gd.drawRect(x, y, width - 1, height - 1);
      }
      else if (hover)
      {
         gd.setColor(new ColorRgb(new Color(0xF4F9FF)));
         gd.fillRect(x + 1, y + 1, width - 1, height - 1);
         gd.setColor(new ColorRgb(new Color(0xBDD7FF)));
         gd.drawRect(x, y, width - 1, height - 1);
      }
      else
      {
         // leave it transparent
         // gd.setColor(new ColorRgb(new Color(0xFFFFFF)));
         // gd.fillRect(x, y, width, height);
      }
   }
   
   /**
    * Obtain color information needed to render the foreground of a thumbnail.
    *
    * @param   gd
    *          The {@code GuiDriver} used for drawing operations.
    * @param   focused
    *          {@code true} if thumbnail is focused.
    * @param   selected
    *          {@code true} if thumbnail is selected.
    * @param   hover
    *          {@code true} if mouse hovers the thumbnail.
    * @param   filename
    *          {@code true} to obtain the color of the filename.
    * @param   infoLabel
    *          {@code true} to obtain the info label (for details view).
    * @param   infodata
    *          {@code true} to obtain the info value (for details view).
    *
    * @return  The foreground color that matches the requested state.
    */
   @Override
   public ColorRgb getFileChooserForegroundColor(GuiDriver gd,
                                                 boolean focused,
                                                 boolean selected,
                                                 boolean hover,
                                                 boolean filename,
                                                 boolean infoLabel,
                                                 boolean infodata)
   {
      return filename || infodata
            ? new ColorRgb(new Color(0x000000))
            : new ColorRgb(new Color(0x6B6B6B));
   }
   
   /**
    * Draws the slider thumb of a {@code SliderGuiImpl} widget.
    *
    * @param   slider
    *          The {@code SliderGuiImpl} widget.
    * @param   gd
    *          The {@code GuiDriver} used for drawing operations.
    * @param   gc
    *          The {@code GuiColorResolver} used color selection.
    * @param   x
    *          The horizontal location (in pixel units) of the thumb to be painted/component.
    * @param   y
    *          The vertical location (in pixel units) of the thumb to be painted/component.
    * @param   w
    *          The width of the thumb. 
    * @param   h
    *          The height of the thumb.
    */
   @Override
   public void drawSliderThumb(SliderGuiImpl slider,
                               GuiDriver gd,
                               GuiColorResolver gc,
                               int x,
                               int y,
                               int w,
                               int h)
   {
      // get slider config
      SliderConfig config = slider.config();
      SliderGuiImpl.TicMarks tmMode = slider.getTicmarkMode();
      
      // the pointer drawing depends on orientation and tic marks mode
      if (tmMode != SliderGuiImpl.TicMarks.BOTH)
      {
          super.drawSliderThumb(slider, gd, gc, x, y, w, h);
          return;
      }
      
      // drawing a filled rectangle
      boolean high = (slider.highlighted() || slider.isMouseHover()) && slider.isEnabled();
      
      if (config.horizontal)
      {
         x -= w / 2;
         y -= h / 2 - 2;
      }
      else
      {
         x -= h / 2 - 2;
         y -= w / 2;
         int t = h; // swap w for h
         h = w;
         w = t;
      }
      
      gd.setColor(high ? new Color(0xBEE6FD)
                       : new Color(0xDDDDDD));
      gd.fillRect(x + 1, y + 1, w - 2, h - 2);
      
      gd.setColor(high ? new Color(0x3C7FB1)
                       : new Color(0x707070));
      gd.drawRect(x, y, w - 1, h - 1);
   }
   
   /**
    * Draws the slide track of a {@code SliderGuiImpl} widget.
    *
    * @param   slider
    *          The {@code SliderGuiImpl} widget.
    * @param   gd
    *          The {@code GuiDriver} used for drawing operations.
    * @param   gc
    *          The {@code GuiColorResolver} used color selection.
    * @param   x
    *          The horizontal location (in pixel units) of the widget to be painted/component.
    * @param   y
    *          The vertical location (in pixel units) of the widget to be painted/component.
    * @param   w
    *          The width of the painted widget/component. 
    */
   @Override
   public void drawSliderTrack(SliderGuiImpl slider,
                               GuiDriver gd,
                               GuiColorResolver gc,
                               int x,
                               int y,
                               int w)
   {
      SliderConfig config = slider.config();
      
      // center
      gd.setColor(new Color(0xE7EAEA));
      if (config.horizontal)
      {
         gd.fillRect(x + 1, y + 1, w - 2, 2);
      }
      else
      {
         gd.fillRect(x + 1, y + 1, 2, w - 2);
      }
      
      // border
      gd.setColor(new Color(0xB0B0B0));
      if (config.horizontal)
      {
         gd.drawRect(x, y, w - 1, 3);
      }
      else
      {
         gd.drawRect(x, y, 3, w - 1);
      }
   }
   
   /**
    * Obtain the color used for drawing the tic-marks for this {@code Theme}. 
    * 
    * @param   gc
    *          The {@code GuiColorResolver} used color selection.
    *
    * @return  the color used for drawing the tic-marks. 
    */
   @Override
   protected ColorRgb getSliderTicmarkColor(GuiColorResolver gc)
   {
      return new ColorRgb(new Color(0xc4C4C4));
   }
}
