/*
** Module   : DynamicLayout.java
** Abstract : Implementation of the core dynamic layout algorithm.
**
** Copyright (c) 2018, Golden Code Development Corporation.
**
** -#- -I- --Date-- ----------------------------Description-----------------------------
** 001 HC  20180813 Initial version.
*/
/*
** This program is free software: you can redistribute it and/or modify
** it under the terms of the GNU Affero General Public License as
** published by the Free Software Foundation, either version 3 of the
** License, or (at your option) any later version.
**
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
** GNU Affero General Public License for more details.
**
** You may find a copy of the GNU Affero GPL version 3 at the following
** location: https://www.gnu.org/licenses/agpl-3.0.en.html
**
** Additional terms under GNU Affero GPL version 3 section 7:
**
**   Under Section 7 of the GNU Affero GPL version 3, the following additional
**   terms apply to the works covered under the License.  These additional terms
**   are non-permissive additional terms allowed under Section 7 of the GNU
**   Affero GPL version 3 and may not be removed by you.
**
**   0. Attribution Requirement.
**
**     You must preserve all legal notices or author attributions in the covered
**     work or Appropriate Legal Notices displayed by works containing the covered
**     work.  You may not remove from the covered work any author or developer
**     credit already included within the covered work.
**
**   1. No License To Use Trademarks.
**
**     This license does not grant any license or rights to use the trademarks
**     Golden Code, FWD, any Golden Code or FWD logo, or any other trademarks
**     of Golden Code Development Corporation. You are not authorized to use the
**     name Golden Code, FWD, or the names of any author or contributor, for
**     publicity purposes without written authorization.
**
**   2. No Misrepresentation of Affiliation.
**
**     You may not represent yourself as Golden Code Development Corporation or FWD.
**
**     You may not represent yourself for publicity purposes as associated with
**     Golden Code Development Corporation, FWD, or any author or contributor to
**     the covered work, without written authorization.
**
**   3. No Misrepresentation of Source or Origin.
**
**     You may not represent the covered work as solely your work.  All modified
**     versions of the covered work must be marked in a reasonable way to make it
**     clear that the modified work is not originating from Golden Code Development
**     Corporation or FWD.  All modified versions must contain the notices of
**     attribution required in this license.
*/

package com.goldencode.p2j.ui.client.layout;

import com.goldencode.p2j.ui.client.*;
import com.goldencode.p2j.ui.client.widget.*;

import java.util.*;
import java.util.function.*;

/**
 * The class provides layout implementation that allows to rearrange widget containers (frames and windows)
 * to better utilize the available space. This is an extension to the legacy 4GL and works on top of the
 * legacy layout behavior.
 * <p>
 * The layout algorithm detects the expandable (aka dynamic) widgets (currently only browse is considered an
 * expandable widget), finds the available space and resizes them accordingly. It moves the widgets around the
 * expandable widgets to offset the resize of the expandable widgets and to make more room if available.
 * The expandable widgets are resized proportionally based on their initial sizes. There is a configurable
 * padding that is kept between widgets or between widgets and the top-level window's edge. Inner frames
 * are supported as well as multiple browse widgets anywhere in the widget tree.
 * <p>
 * See {@link DynamicLayoutConfig} for the configurable layout options.
 * <p>
 * The design intentionally doesn't extend the existing layout infrastructure represented by
 * {@link LayoutManager}. The reason for this is that dynamic layout works on top of the existing layout
 * mechanics and assumes the components are already laid out according to the legacy requirements. This
 * also means that the invocation points for the dynamic layout are different from the legacy layout
 * logic.
 */
public class DynamicLayout<O extends OutputManager<?>>
extends AbstractDynamicLayout<O>
{
   /**
    * Constructor.
    *
    * @param   container
    *          The container this instance will layout.
    */
   public DynamicLayout(Container<O> container)
   {
      super(container);
   }

   /**
    * Performs the actual layout.
    */
   @Override
   public void doLayout()
   {
      inflateWidth();
      inflateHeight();
   }

   /**
    * Rearranges the container widgets to utilize the container's availble horizontal size.
    */
   private void inflateWidth()
   {
      double windowPaddingX = config.getWindowPaddingX();
      double framePaddingX = config.getFramePaddingX();

      // parse widgets horizontally
      List<RowInfo> rows = parseRows();

      inflateImpl(rows,
                  (w, l) -> w.setLocation(l, w.location().y),
                  w -> w.location().x,
                  (w, d) -> setWidgetSize(w, d, w.height()),
                  d -> d.width,
                  windowPaddingX,
                  framePaddingX);
   }

   /**
    * Rearranges the container widgets to utilize the container's available vertical size.
    */
   private void inflateHeight()
   {
      double windowPaddingY = config.getWindowPaddingY();
      double framePaddingY = config.getFramePaddingY();

      // parse widgets vertically
      List<RowInfo> columns = parseColumns();

      inflateImpl(columns,
                  (w, l) -> w.setLocation(w.location().x, l),
                  w -> w.location().y,
                  (w, d) -> setWidgetSize(w, w.width(), d),
                  d -> d.height,
                  windowPaddingY,
                  framePaddingY);
   }

   /**
    * Rearranges the container widgets to utilize the container's available vertical or horizontal size.
    *
    * @param   rows
    *          Widgets grouped into vertical or horizontal rows.
    * @param   locationSetter
    *          Widget location setter.
    * @param   locationGetter
    *          Widget location getter.
    * @param   dimensionSetter
    *          Widget dimension setter.
    * @param   widthF
    *          Widget horizontal dimension getter.
    * @param   windowPadding
    *          Padding between frames and window egdes.
    * @param   framePadding
    *          Padding between widgets and frame edges.
    */
   private void inflateImpl(List<RowInfo> rows,
                            BiConsumer<Widget, Double> locationSetter,
                            Function<Widget, Double> locationGetter,
                            BiConsumer<Widget, Double> dimensionSetter,
                            Function<Dimension, Double> widthF,
                            double windowPadding,
                            double framePadding)
   {
      // this is a technical minimum size, this is needed in order to keep relative locations of the widgets;
      // same relative positions at all sizes ensure we can scale correctly from minimum to maximum possible
      // sizes
      final double minCellWidth = 0.5;

      Dimension containerDim = container.dimension();

      for (RowInfo row : rows)
      {
         // only care for dynamic rows
         if (!row.isDynamic())
         {
            continue;
         }

         int numCells = row.cells.size();

         double origin = container.parentOrSelf(Frame.class) != null ? framePadding : windowPadding;
         double totalPadding = 2 * origin + (numCells - 1) * framePadding;

         // the available width for the dynamic part
         double availableWidth = widthF.apply(containerDim) - row.totalStaticWidth - totalPadding;
         double usedUpWidth = 0;
         CellInfo lastDynamicCell = null;

         for (int i = numCells - 1; i >= 0; i--)
         {
            CellInfo cell = row.cells.get(i);
            if (cell.dynamic)
            {
               lastDynamicCell = cell;
               break;
            }
         }

         for (int i = 0; i < row.cells.size(); i++)
         {
            CellInfo cell = row.cells.get(i);
            double cellWidth = cell.right - cell.left;
            double newWidth = cellWidth;

            if (cell.dynamic)
            {
               if (cell != lastDynamicCell)
               {
                  newWidth = availableWidth * (cell.initialWidth / row.totalInitialWidth);
                  if (newWidth < minCellWidth)
                  {
                     newWidth = minCellWidth;
                  }
                  usedUpWidth += newWidth;
               }
               else
               {
                  newWidth = availableWidth - usedUpWidth;
                  if (newWidth < minCellWidth)
                  {
                     newWidth = minCellWidth;
                  }
               }
            }

            double actLoc = Double.MAX_VALUE;
            double actWidth = 0;

            for (Widget w : cell.widgets)
            {
               if (w instanceof Frame)
               {
                  ((Frame) w).resetAutoplacing();
               }

               locationSetter.accept(w, locationGetter.apply(w) - cell.left + origin);

               // get the location from the widget to account for roundings
               double newLoc = locationGetter.apply(w);
               if (newLoc < actLoc)
               {
                  actLoc = newLoc;
               }

               if (cell.dynamic && isDynamicWidget(w) && newWidth >= 0)
               {
                  dimensionSetter.accept(w, newWidth);

                  // get the size from the widget to account for roundings
                  double newWidgetWidth = widthF.apply(w.dimension());
                  if (newWidgetWidth > actWidth)
                  {
                     actWidth = newWidgetWidth;
                  }
               }
            }

            if (actWidth == 0)
            {
               actWidth = newWidth;
            }

            if (actLoc != Double.MAX_VALUE)
            {
               origin = actLoc;
            }

            origin += actWidth + framePadding;
         }
      }
   }
}
