/*
** Module   : ListSelectionModel.java
** Abstract : Selection model for lists.
**
** Copyright (c) 2010-2017, Golden Code Development Corporation.
**
** -#- -I- --Date-- --JPRM-- ----------------Description-----------------
** 001 SIY 20100629          Created initial version
** 002 SIY 20100804          Changes in imports, some refactorings.
** 003 SIY 20100822          Added missing javadocs.
*/
/*
** This program is free software: you can redistribute it and/or modify
** it under the terms of the GNU Affero General Public License as
** published by the Free Software Foundation, either version 3 of the
** License, or (at your option) any later version.
**
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
** GNU Affero General Public License for more details.
**
** You may find a copy of the GNU Affero GPL version 3 at the following
** location: https://www.gnu.org/licenses/agpl-3.0.en.html
** 
** Additional terms under GNU Affero GPL version 3 section 7:
** 
**   Under Section 7 of the GNU Affero GPL version 3, the following additional
**   terms apply to the works covered under the License.  These additional terms
**   are non-permissive additional terms allowed under Section 7 of the GNU
**   Affero GPL version 3 and may not be removed by you.
** 
**   0. Attribution Requirement.
** 
**     You must preserve all legal notices or author attributions in the covered
**     work or Appropriate Legal Notices displayed by works containing the covered
**     work.  You may not remove from the covered work any author or developer
**     credit already included within the covered work.
** 
**   1. No License To Use Trademarks.
** 
**     This license does not grant any license or rights to use the trademarks
**     Golden Code, FWD, any Golden Code or FWD logo, or any other trademarks
**     of Golden Code Development Corporation. You are not authorized to use the
**     name Golden Code, FWD, or the names of any author or contributor, for
**     publicity purposes without written authorization.
** 
**   2. No Misrepresentation of Affiliation.
** 
**     You may not represent yourself as Golden Code Development Corporation or FWD.
** 
**     You may not represent yourself for publicity purposes as associated with
**     Golden Code Development Corporation, FWD, or any author or contributor to
**     the covered work, without written authorization.
** 
**   3. No Misrepresentation of Source or Origin.
** 
**     You may not represent the covered work as solely your work.  All modified
**     versions of the covered work must be marked in a reasonable way to make it
**     clear that the modified work is not originating from Golden Code Development
**     Corporation or FWD.  All modified versions must contain the notices of
**     attribution required in this license.
*/

package com.goldencode.p2j.ui.client.model;

import java.util.*;

import com.goldencode.p2j.ui.client.event.listener.*;

/**
 * Selection model for lists.
 */
public class ListSelectionModel
{
   /**
    * Single item selection mode.
    */
   public static final int SINGLE_SELECTION = 201;
   
   /**
    * Single interval selection mode.
    */
   public static final int SINGLE_INTERVAL_SELECTION = 202;
   
   /**
    * Multiple intervals selection mode.
    */
   public static final int MULTIPLE_INTERVAL_SELECTION = 203;
   
   /** List of listeners. */
   private ArrayList<SelectionListener> listeners = new ArrayList<SelectionListener>();

   /** Active selection mode. */
   private int mode = ListSelectionModel.SINGLE_SELECTION;

   /** Selected item index storage. */
   private TreeSet<Integer> selection = new TreeSet<Integer>();

   /**
    * Select specified range of items.
    * 
    * @param   from
    *          Start index in the range.
    * @param   to
    *          End index in the range.
    */
   public void select(int from, int to)
   {
      TreeSet<Integer> range = asSet(from, to);

      range.addAll(selection);

      setSelection(range);
   }

   /**
    * Add selection item listener.
    * 
    * @param   listener
    *          Listener to add.
    */
   public void add(SelectionListener listener)
   {
      listeners.add(listener);
   }

   /**
    * Clear model and remove selection.
    */
   public void clear()
   {
      if (!empty())
      {
         notifyListeners();

         selection.clear();
      }
   }

   /**
    * Check if selection is empty.
    * 
    * @return  <code>true</code> if selection is empty.
    */
   public boolean empty()
   {
      return selection.isEmpty();
   }

   /**
    * Get first selected index.
    * 
    * @return  first selected index or -1 if selection is empty.
    */
   public int first()
   {
      try
      {
         return selection.first().intValue();
      }
      catch (NoSuchElementException e)
      {
         return -1;
      }
   }

   /**
    * Get last selected index.
    * 
    * @return  last selected index or -1 if selection is empty.
    */
   public int last()
   {
      try
      {
         return selection.last().intValue();
      }
      catch (NoSuchElementException e)
      {
         return -1;
      }
   }

   /**
    * Get current selection mode.
    * 
    * @return  current selection mode.
    */
   public int mode()
   {
      return mode;
   }

   /**
    * Get all selected indices as array.
    * 
    * @return  selected indices.
    */
   public int[] selection()
   {
      int[] res = new int[selection.size()];
      int i = 0;
      
      for (Integer selected : selection)
      {
         res[i++] = selected.intValue();
      }
      
      return res;
   }

   /**
    * Deselect specified range of items.
    * 
    * @param   from
    *          Start index in the range.
    * @param   to
    *          End index in the range.
    */
   public void deselect(int from, int to)
   {
      TreeSet<Integer> range = asSet(from, to);

      TreeSet<Integer> newSelection = new TreeSet<Integer>(selection);
      newSelection.removeAll(range);

      setSelection(newSelection);
   }

   /**
    * Check if specified index is selected.
    * 
    * @param   index
    *          Index to check.
    *          
    * @return  <code>true</code> if index is in one of selected ranges. 
    */
   public boolean selected(int index)
   {
      return selection.contains(Integer.valueOf(index));
   }

   /**
    * Set selection to specified interval.
    * 
    * @param   from
    *          Start index in the range.
    * @param   to
    *          End index in the range.
    */
   public void set(int from, int to)
   {
      setSelection(asSet(from, to));
   }

   /**
    * Set selection mode.
    * 
    * @param   mode
    *          Selection mode.
    */
   public void setMode(int mode)
   {
      this.mode = mode;
   }

   /**
    * Trigger notification of listeners.
    */
   public void notifyListeners()
   {
      for (SelectionListener listener : listeners)
      {
         listener.changed();
      }
   }

   /**
    * Convert index range into set.
    *  
    * @param   from
    *          Initial index.
    * @param   to
    *          Final index.
    *          
    * @return  set containing integers from specified range. 
    */
   private TreeSet<Integer> asSet(int from, int to)
   {
      TreeSet<Integer> range = new TreeSet<Integer>();

      for (int i = Math.min(from, to); i <= Math.max(from, to); i++)
         range.add(Integer.valueOf(i));
      
      return range;
   }

   /**
    * Selection setting worker.
    * 
    * @param   newSel
    *          Source selection.
    */
   private void setSelection(TreeSet<Integer> newSel)
   {
      TreeSet<Integer> copy = new TreeSet<Integer>(selection);
      TreeSet<Integer> diff = new TreeSet<Integer>(newSel);

      diff.removeAll(selection);
      copy.removeAll(newSel);
      diff.addAll(copy);

      selection = newSel;

      if (!diff.isEmpty())
      {
         notifyListeners();
      }
   }
}