/*
** Module   : AbstractWidget.java
** Abstract : Base class for widgets.
**
** Copyright (c) 2010-2018, Golden Code Development Corporation.
**
** -#- -I- --Date-- ---------------------------------Description----------------------------------
** 001 SIY 20100629 Created initial version
** 002 SIY 20100804 Changes in imports, some refactorings.
** 003 SIY 20100822 Added base implementation of new interface methods config() and isFocusable().
** 004 SIY 20101122 Switching from KeyEvent to KeyInput.
** 005 SIY 20111115 Added physical widget location.
** 006 CA  20131121 The hidden/visible state changes must be collected when the widget's attribute
**                  is actually set.
** 007 EVL 20140617 Added method to check if the widget needs the cursor to be on.
** 008 CA  20140711 Enhanced font support. Added physicalDimension, in preparation for GUI support.
** 009 CA  20140805 Added generics to force the abstract widget classes to not use implementation-
**                  specific code: the abstract widgets will have access only to the base APIs
**                  provided by the OutputManager.
**                  Added mouse event processing.  Added APIs to work with physical units, 
**                  depending on the UI type.
** 010 MAG 20140912 Added insets member used on GUI only.
** 011 HC  20140911 GUI coordinate support - data types refactoring.
** 012 CA  20140926 Added shared widget configuration support, to allow GUI/ChUI concrete
**                  implementation for the same widget ID.  Refactored the widget configuration
**                  classes: all fields were made public; these classes need to be as dumb as
**                  possible, all logic related to setting/getting a certain field should be at
**                  the widget or at the caller. Refs #2254
** 013 MAG 20140917 Added parent width and height inheritance.
**                  Added mouse position translate.
** 014 CA  20141021 Changed the resize() API to specify the X and Y deltas, plus flags to indicate
**                  if the top-left corner needs to move on the X or Y axis.
**                  Added API to allow a widget to raise (or not) legacy mouse events, when a 
**                  mouse action was performed on it; by default all widgets can raise these events.
** 015 VIG 20141016 Added window() method to resolving widgets Window.
** 016 MAG 20141020 Added clipping rectangle computation.
** 017 HC  20141021 Implemented window move/position support.
** 018 HC  20141118 Added screenPhysicalLocation().
** 019 CA  20141119 Refactored drawing code to enable multiple OS windows, for GUI support.  Misc
**                  changes/fixes related to drawing on multiple windows.
** 020 HC  20141207 Implemented runtime support for widget coordinate attributes.
** 021 HC  20150111 Fixed widget position change in a frame.
** 022 HC  20150128 Added support for frame widget position changes induced by setting
**                  ROW/COL/X/Y attributes.
** 023 CA  20150212 Allow events to be captured and processed by the widgets without server-side
**                  or main event-loop notification.  This fixes problems in the EDITING and
**                  WAIT-FOR event loops, for i.e. COMBO-BOX and EDITOR widgets.
** 024 IAS 20150219 Added getVisiblePart() method
** 025 HC  20150323 Removed business logic from config classes.
** 026 CA  20150429 When creating a rectangle from a location and dimension, use the base units
**                  to determine the coordinates for current client mode.
** 027 VIG 20150330 Added prepParentLocation as extension point. Added pop-up menu key listener
**                  adding.
** 026 GES 20150507 Added a comment to clarify that AWT events usage is intentional.
** 027 HC  20150612 Implemented support for modal windows and GUI ALERT-BOX.
** 028 HC  20150626 Fixed an unhandled exception when showing a modal window.
** 029 HC  20150702 Improved parent-child relationship of top-level windows.
** 030 IAS 20150703 Added 'wasRealized' flag support, 'debug' method
** 031 EVL 20150619 Adding method to check if the widget needs LEAVE event on outside mouse press.
**                  Another method generates the LEAVE event and repost mouse to be sure it will
**                  be handled after LEAVE processing completed.
** 032 CA  20150722 Added postprocessConfig - required additional processing, as layout is 
**                  performed only on the first frame definition push.
** 033 VIG 20150809 Removed wrong condition check for posting LOST_EVENT. We must post it, even
**                  the current focused widget is the parent of this widget. Method requestFocus().
** 034 GES 20150710 Added getSelectedText() default implementation for clipboard support.
** 035 IAS 20150810 Added isTabStop method 
** 036 CA  20150905 Enhanced javadoc for mouseClicked.
** 037 HC  20150831 Widgets are automatically detached from their parents when destroyed.
** 038 CA  20150826 Changes for improved mouse support.
** 039 IAS 20150926 AUTO-RESIZE, MENU-MOUSE and MOVE-TO-TOP()/MOVE-TO-BOTTOM() support.
** 040 HC  20150927 Layout fixes when insets are involved. Fixed multi-window drawing issues
**                  caused by concurrent drawing access to the GuiDriver: all access to the
**                  driver is done exclusively, each thread taking and releasing ownership for
**                  the duration of the drawing operation or other GUI API invocation which
**                  requires access to the underlying physical window. Changes to allow
**                  multi-window focus management in GUI.
** 041 HC  20151013 Changes to extend GUI multi-window focus management and ACTIVE-WINDOW system
**                  handle processing to modal windows.
** 042 VIG 20151020 Activate MENUBAR for CHUI and set default focused menu item on F3.
**     HC  20151020 Fixed getTextWidth() and getTextHeight() to use actual top-level window.
** 043 CA  20151112 Made setHidden() public.
**     HC  20151112 Implemented THREE-D runtime support.
** 044 HC  20151213 Improved DIALOG-BOX window parenting logic and related changes.
** 045 VIG 20160120 Added navigable() method.
** 046 IAS 20160130 LOAD-MOUSE-POINTER support.
** 047 HC  20160304 Removed moveBy() method.
** 048 CA  20160327 Added support for LAST-EVENT:X and LAST-EVENT:Y attributes.
** 049 CA  20160219 Fixed clipRectangle: bottom/right coordinates were added one extra pixel.
** 050 VIG 20160209 Added isRMBEntrySupported() method, changed showPopupMenu() call, 
**                  removed popup menu location setting.
** 051 HC  20160421 A widget is focused only when it is in the active top-level window.
** 052 HC  20160511 Fixed COMBO-BOX focusing issues and related changes.
** 053 HC  20160607 GUI menus reworked to be displayed in dedicated top-level windows.
** 054 SVL 20160715 Updates to translate() which fix mouse hot spot.
** 055 EVL 20160725 Adding code to refresh editable frames list while setting widget to be
**                  enabled.
** 056 CA  20160915 Added unrealize() method.  Added TODOs repaint: unrealized widgets should 
**                  never have a paint event raised. 
** 057 SBI 20160914 Fixed a possible NPE.
** 058 SVL 20161014 Added clearWidget function.
** 059 HC  20170102 Performance improvement of hasFocus().
** 060 CA  20161101 Added support for mnemonic processing.
** 061 EVL 20170628 Adding register/deregister the direct manipulation handler.  Adding direct
**                  manipulation default processing.
** 062 HC  20170629 Implemented support of WINDOW:DISABLE-REDRAW attribute, the FWD extension to
**                  4GL.
** 063 CA  20171020 Added support for FRAMEs sent to named or unnamed streams, in GUI.
** 064 CA  20171023 In GUI, streamed frames don't affect any state of its UI counterpart.
** 065 HC  20171027 Extended DISABLE-REDRAW to all widget types. Previously only WINDOW widget
**                  was supported.
** 066 OM  20171027 Avoided NPE in mouseClicked().
** 067 CA  20171124 Avoid unnecessary repaints.  If font or location has changed, repaint.
** 068 CA  20171211 Fixed H067 - if widget location/dimension has changed, than post a PaintEvent
**                  directly on the queue, with the old widget boundary.
**     CA  20171218 Fixed LAST-EVENT:WIDGET-ENTER/LEAVE attributes cases, plus focus-related fixes.
** 069 CA  20180130 One more fix for H067/H068: when location is changed, post the paint on the 
**                  explicit screen location, don't rely on repaint() to compute it.
** 070 CA  20180201 Small optimization: moved the ChUI/GUI check from checkRectangle() to 
**                  isTotallyObscured(), as the bounds/ancestor computation is expensive.
**                  Improved performance for ancestor calculation, by caching the computed values.
**                  Minor performance improvement for parent() (avoid multiple calls).
**     CA  20180208 Improved performance for screenLocation and screenPhysical location - these 
**                  are cached via the LocationChangeAspect, and reset via reflectLocationChange.
** 071 CA  20180226 Made reflectAncestorChange public.
** 072 EVL 20180328 Fix for MODIFIED attribute reset while disabled widget enabling.
** 073 HC  20180420 Implemented popup menu resolution in the tree of widgets.
** 074 HC  20180503 Fixed FILL-IN stops responding to mouse clicks after its popup menu
**                  displayed.
** 075 CA  20180502 Fixed COLOR option, COLOR statement and other color-related issues.
**     CA  20180530 Fixed problems with BUTTON:MOVABLE and mouse manipulation.
**     HC  20180530 Fixed the widget stops responding to mouse clicks after its popup menu
**                  displayed. This is second attempt to resolve this issue.
** 076 SBI 20180713 Changed to remove registered font resolvers for a destroyed widget.
** 077 SVL 20180529 getTextWidth functions support explicit fonts.
** 078 HC  20180813 Implemented initial version of dynamic layout.
*/

/*
** This program is free software: you can redistribute it and/or modify
** it under the terms of the GNU Affero General Public License as
** published by the Free Software Foundation, either version 3 of the
** License, or (at your option) any later version.
**
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
** GNU Affero General Public License for more details.
**
** You may find a copy of the GNU Affero GPL version 3 at the following
** location: https://www.gnu.org/licenses/agpl-3.0.en.html
** 
** Additional terms under GNU Affero GPL version 3 section 7:
** 
**   Under Section 7 of the GNU Affero GPL version 3, the following additional
**   terms apply to the works covered under the License.  These additional terms
**   are non-permissive additional terms allowed under Section 7 of the GNU
**   Affero GPL version 3 and may not be removed by you.
** 
**   0. Attribution Requirement.
** 
**     You must preserve all legal notices or author attributions in the covered
**     work or Appropriate Legal Notices displayed by works containing the covered
**     work.  You may not remove from the covered work any author or developer
**     credit already included within the covered work.
** 
**   1. No License To Use Trademarks.
** 
**     This license does not grant any license or rights to use the trademarks
**     Golden Code, FWD, any Golden Code or FWD logo, or any other trademarks
**     of Golden Code Development Corporation. You are not authorized to use the
**     name Golden Code, FWD, or the names of any author or contributor, for
**     publicity purposes without written authorization.
** 
**   2. No Misrepresentation of Affiliation.
** 
**     You may not represent yourself as Golden Code Development Corporation or FWD.
** 
**     You may not represent yourself for publicity purposes as associated with
**     Golden Code Development Corporation, FWD, or any author or contributor to
**     the covered work, without written authorization.
** 
**   3. No Misrepresentation of Source or Origin.
** 
**     You may not represent the covered work as solely your work.  All modified
**     versions of the covered work must be marked in a reasonable way to make it
**     clear that the modified work is not originating from Golden Code Development
**     Corporation or FWD.  All modified versions must contain the notices of
**     attribution required in this license.
*/

package com.goldencode.p2j.ui.client.widget;

// this usage of AWT events is intentional, they are being used separately from the
// AWT/Swing event model itself; no other AWT/Swing usage is occurring in this class
import java.awt.event.MouseEvent;
import java.awt.event.MouseWheelEvent;
import java.lang.ref.*;
import java.util.*;
import java.util.logging.*;

import com.goldencode.p2j.ui.*;
import com.goldencode.p2j.ui.chui.*;
import com.goldencode.p2j.ui.client.*;
import com.goldencode.p2j.ui.client.Menu.*;
import com.goldencode.p2j.ui.client.driver.*;
import com.goldencode.p2j.ui.client.event.*;
import com.goldencode.p2j.ui.client.event.listener.*;
import com.goldencode.p2j.ui.client.gui.*;
import com.goldencode.p2j.ui.client.gui.driver.*;
import com.goldencode.p2j.ui.client.layout.*;
import com.goldencode.p2j.util.*;

/**
 * Base widget class. 
 */
public abstract class AbstractWidget<O extends OutputManager<?>>
implements Widget<O>
{
   /** Logger. */
   protected static final Logger LOG = LogHelper.getLogger(AbstractWidget.class.getName());

  /** Widget enabled flag. */
   private boolean enabled = true;
   
   /** Widget hidden flag. */
   private boolean hidden = false;
   
   /** Widget highlighted flag. */
   private boolean highlighted = false;
   
   /** Widget visible flag. This field is needed only for instances without config. */
   private boolean visible = true;
   
   /** Storage for focus listeners. */
   private ArrayList<FocusListener> focusListeners = new ArrayList<FocusListener>();
   
   /** Storage for key listeners. */
   private ArrayList<KeyListener> keyListeners = new ArrayList<KeyListener>();

   /** Reference to parent container. */
   private WeakReference<Container<O>> parent = null;
   
   /** The cached widget's window.  Gets reset by {@link #reflectAncestorChange()}. */
   private WeakReference<Window<O>> window = null;
   
   /** The previously computed ancestor cache.  Gets reset by {@link #reflectAncestorChange()}. */
   private Map<Class<?>, WeakReference<Container<?>>> ancestorCache = new IdentityHashMap<>();
   
   /** <code>true</code> if and only if {@link PopupKeyListener} was added to this widget. */ 
   private boolean popupKeyListnrAdded = false;

   /** List of listeners which require notifications about the state of this widget. */
   private List<WidgetStateListener> listeners = new ArrayList<>(); 
   
   /** Reference to screen. */
   private O screen;
   
   /** A set of mouse actions on which this widget can react. */
   private Set<Integer> mouseActions = null;
   
   /**
    * Widget origin (upper left corner) relative to container
    * expressed in character cell units.
    */
   protected Point location = new Point(0, 0);
   
   /**
    * Widget physical origin (upper left corner) relative to
    * container. expressed in units native to UI type.
    */
   protected NativePoint physicalLocation = new NativePoint(0, 0);
   
   /** Widget insets */
   protected Insets insets = Insets.ZERO_INSETS;
   
   /** Widget native insets */
   protected NativeInsets nativeInsets = NativeInsets.ZERO_INSETS;

   /** Custom mouse pointer */
   protected Optional<MousePtrWrapper> mousePtr = Optional.empty();
   
   /** Mouse hover event handler */
   private MouseHoverAction mouseHoverAction = null;
   
   /** Flag indicating the direct manipulation handler registration. */
   private boolean directManipulationActive = false;

   /** The cached value of {@link #screenLocation()}. */
   private Point screenLocation = null;
   
   /** The cached value of {@link #screenPhysicalLocation()}. */
   private NativePoint screenPhysicalLocation = null;
   
   /**
    * Constructor
    */
   @SuppressWarnings("unchecked")
   public AbstractWidget()
   {
      screen = (O) OutputManager.instance();

      ClientConfigManager mgr = ConfigManager.getInstance();
      this.registerWidgetStateListener(mgr);
   }
   
   /**
    * Copy the bg/fg colors from the source config to the target config.
    * 
    * @param    source
    *           The source config.
    * @param    target
    *           The target config.
    */
   protected static void copyColors(BaseConfig source, BaseConfig target)
   {
      target.bgcolor    = source.bgcolor;
      target.fgcolor    = source.fgcolor;
      target.bgcolorRgb = source.bgcolorRgb;
      target.fgcolorRgb = source.fgcolorRgb;
   }
   
   /**
    * Widget ID.
    * 
    * @return  see above.
    */
   @Override
   public WidgetId getId()
   {
      WidgetConfig c = config();
      return c != null ? c.id : null;
   }

   /**
    * Get access to {@link OutputManager} instance.
    * 
    * @return  reference to {@link OutputManager}.
    */
   @Override
   public O screen() 
   {
      return screen;
   }

   /**
    * Add focus listener.
    * 
    * @param   listener
    *          Listener to add.
    */
   @Override
   public void addFocusListener(FocusListener listener)
   {
      focusListeners.add(listener);
   }

   /**
    * Add key listener.
    * 
    * @param   listener
    *          Listener to add.
    */
   @Override
   public void addKeyListener(KeyListener listener)
   {
      keyListeners.add(listener);
   }

   /**
    * Get ancestor window for this widget.
    * 
    * @return  ancestor window.
    */
   @SuppressWarnings("unchecked")
   @Override
   public TitledWindow<O> ancestor()
   {
      return parentOrSelf(TitledWindow.class);
   }
   
   /**
    * Get the ID of the ancestor window for this widget.
    * 
    * @return  ancestor window ID or <code>null</code> if no valid ancestor exists.
    */
   public WidgetId ancestorId()
   {
      Widget<?> ancestor = ancestor();
      if (ancestor == null)
      {
         return null;
      }
      
      return ancestor.getId();
   }

   /**
    * Get ancestor Window for this widget. This method must be called
    * only for widgets that are already attached to Window.
    * 
    * @return  ancestor Window.
    * 
    * @throws  IllegalStateException
    *          When this widget is not attached to window.
    */
   @Override
   public Window<O> window()
   {
      Window<O> w = null;
      
      if (window != null)
      {
         w = window.get();
         if (w != null)
         {
            return w;
         }
      }
      
      Widget<O> parent = parent();
      
      if (parent == null)
      {
         int wid = UiUtils.getWidgetIdAsInt(this);
         throw new IllegalStateException("Widget " + name() + " " + wid
                                         + " is not attached to a Window instance.");
      }
      
      w = parent.window();
      
      window = w == null ? null : new WeakReference(w);
      
      return w;
   }
   
   /**
    * Get ancestor TopLevelWindow for this widget. This method must be called
    * only for widgets that are already attached to a TopLevelWindow.
    * 
    * @return ancestor TopLevelWindow.
    * 
    * @throws    IllegalStateException
    *            When this widget is not attached to a top-level window.
    */
   public TopLevelWindow<O> topLevelWindow()
   {
      @SuppressWarnings("unchecked")
      TopLevelWindow<O> wnd = parentOrSelf(TopLevelWindow.class);
         
      if (!(wnd instanceof TopLevelWindow))
      {
         int wid = UiUtils.getWidgetIdAsInt(this);
         throw new IllegalStateException("Widget " + name() + " " + wid
               + " is not attached to a TopLevelWindow instance.");
      }
      
      return wnd;
   }
   
   /**
    * Get widget bounding rectangle.
    * 
    * @return  bounding rectangle.
    */
   @Override
   public Rectangle bounds()
   {
      Rectangle result = new Rectangle(location(), dimension(), screen().coordinates().baseUnits());
      
      return result;
   }

   /**
    * Get widget insets.
    * 
    * @return  widget insets.
    */
   @Override
   public Insets insets()
   {
      return insets;
   }

   /**
    * Get widget dimension.
    * 
    * @return  widget dimension.
    */
   @Override
   public abstract Dimension dimension();

   /**
    * Get widget physical dimension, expressed in units native to UI type.
    *  
    * @return  widget dimension.
    */
   @Override
   public NativeDimension physicalDimension()
   {
      return screen.coordinates().dimensionToNative(dimension());
   }

   /**
    * Check if widget is displayed.
    * <p>
    * Only widget which does not delegate to the parent is {@link TitledWindow}, with the 
    * exception of {@link MenuItemGuiImpl}.  For this case, we avoid delegating and up the tree 
    * and just check the window parent for common case.  Other cases (like {@link MenuItemGuiImpl} 
    * will still rely on the parent delegation.
    * 
    * @return  <code>true</code> if widget is displayed.
    */
   @Override
   public boolean isDisplayed()
   {
      TitledWindow<?> wnd = parent(TitledWindow.class);
      return wnd != null && wnd.isDisplayed();
   }

   /**
    * The default implementation delegates to WidgetConfig if it has one, otherwise return false. 
    * 
    * @return  <code>true</code> if config().realized is true.
    */
   @Override
   public boolean isRealized()
   {
      WidgetConfig widgetConfig = config();
      
      return widgetConfig != null && widgetConfig.realized;
   }

   /**
    * Draw widget.
    */
   @Override
   public abstract void draw();

   /**
    * Draw caret.
    */
   @Override
   public void drawCaret()
   {
   }

   /**
    * Check if widget is enabled.
    * 
    * @return  <code>true</code> if widget is enabled.
    */
   @Override
   public boolean isEnabled()
   {
      return enabled;
   }

   /**
    * Check if widget is focus traversable.
    * 
    * @return  <code>true</code> if widget is focus traversable.
    */
   @Override
   public boolean focusTraversable()
   {
      return (enabled && visible && !hidden);
   }

   /**
    * Returns <code>true</code> only if this widget is navigable.
    *
    * @return  <code>true</code> if widget is navigable.
    */
   @Override
   public boolean navigable()
   {
      return isEnabled();
   }
   
   /**
    * Force widget visibility change without associated repainting.
    * 
    * @param   visibility
    *          New visibility state.
    */
   @Override
   public void forceVisibility(boolean visibility)
   {
      _setVisible(visibility); // direct access
   }

   /**
    * Check if widget has focus.
    * 
    * @return  <code>true</code> if widget has focus.
    */
   @Override
   public boolean hasFocus()
   {
      TopLevelWindow<O> tlw = parentOrSelf(TopLevelWindow.class);
      if (tlw == null || tlw != WindowManager.getFocusWindow())
      {
         return false;
      }

      return tlw.currentFocus(true) == this;
   }

   /**
    * The method returns true if the widget is set as current in its parent container.
    *
    * @return  see above.
    */
   public boolean isCurrent()
   {
      Container<O> parent = parent();
      return parent != null && parent.currentFocus(false) == this;
   }

   /**
    * Get widget height.
    * 
    * @return  widget height.
    */
   @Override
   public abstract double height();

   /**
    * Get current hidden state.
    * 
    * @return   hidden state.
    */
   @Override
   public boolean hidden()
   {
      return hidden;
   }

   /**
    * Hide widget.
    */
   @Override
   public void hide()
   {
      if (isVisible())
      {
         _setVisible(false); // direct access
         repaint();
      }
   }

   /**
    * Get current highlighted state.
    * 
    * @return   highlighted state.
    */
   @Override
   public boolean highlighted()
   {
      return highlighted;
   }

   /**
    * Get widget location relative to container.
    * <p>
    * The method always returns a valid location. If a component
    * has not been assigned location explicitly, the method returns
    * the default location 0,0.
    * 
    * @return  relative widget location.
    */
   @Override
   public Point location()
   {
      BaseConfig bc = UiUtils.getConfig(this, BaseConfig.class);
      
      if (bc == null)
      {
         double col = location.x;
         double row = location.y;
         
         if (col == BaseConfig.INV_COORD)
         {
            col = 0;
         }

         if (row == BaseConfig.INV_COORD)
         {
            row = 0;
         }
           
         return new Point(col, row);
      }
      
      return ConfigHelper.getEffectiveLocation(bc);
   }

   /**
    * Get widget physical location relative to container, expressed in units native to UI type.
    * <p>
    * The method always returns a valid location. If a component has not been assigned location
    * explicitly, the method returns the default location 0,0.
    * <p>
    * The value returned by the method is in native (also called physical) units. Native units 
    * should not be used by the layout logic but only by the output subsystems.
    * 
    * @return  relative widget location.
    */
   @Override
   public NativePoint physicalLocation()
   {
      BaseConfig bc = UiUtils.getConfig(this, BaseConfig.class);
      
      if (bc == null)
      {
         int x = physicalLocation.x;
         int y = physicalLocation.y;
         
         if (x == BaseConfig.INV_COORD)
         {
            x = 0;
         }

         if (y == BaseConfig.INV_COORD)
         {
            y = 0;
         }
           
         return new NativePoint(x, y);
      }
      
      if (screen().isChui())
      {
         // in chui the characters-native conversion is 1:1 
         Point p = ConfigHelper.getEffectiveLocation(bc);
         return new NativePoint((int) p.x, (int) p.y);
      }
      else
      {
         // in gui the pixels-native conversion is 1:1
         return ConfigHelper.getEffectivePixelLocation(bc);
      }
   }
   
   /**
    * Get widget minimum size.
    * 
    * @return  widget minimum size.
    */
   @Override
   public abstract Dimension minimumSize();

   /**
    * Get widget name.
    * 
    * @return  widget name.
    */
   @Override
   public String name()
   {
      WidgetConfig c = config();
      return c instanceof BaseConfig ? ((BaseConfig) c).name : "";
   }

   /**
    * Get widget parent.
    * 
    * @return  widget parent.
    */
   @Override
   public Container<O> parent()
   {
      return (parent == null) ? null : parent.get();
   }

   /**
    * Get widget parent of the specified type.
    * <p>
    * This method is final as it caches the intermediate value in {@link #ancestorCache}.
    * Any future sub-implementation must take this into account.
    *
    * @param   clazz
    *          The parent type to return.
    *
    * @return  widget parent.
    */
   @SuppressWarnings({ "unchecked", "rawtypes" })
   @Override
   public final <T extends Widget<O>> T parent(Class<T> clazz)
   {
      return parentObject(clazz);
   }

   /**
    * Get widget parent of the specified type.
    * <p>
    * This method is final as it caches the intermediate value in {@link #ancestorCache}.  
    * Any future sub-implementation must take this into account.
    * 
    * @param   clazz
    *          The parent type to return.
    * 
    * @return  widget parent.
    */
   @SuppressWarnings({ "unchecked", "rawtypes" })
   @Override
   public final <T> T parentObject(Class<T> clazz)
   {
      boolean known = ancestorCache.containsKey(clazz);
      WeakReference<Container<?>> parentRef = known ? ancestorCache.get(clazz) : null;
      
      Widget<?> parent = null;
      if (known)
      {
         if (parentRef != null)
         {
            parent = parentRef.get();
            if (parent != null)
            {
               return (T) parent;
            }
         }
         else
         {
            return null;
         }
      }

      parent = parent();
      while (parent != null && !(clazz.isAssignableFrom(parent.getClass())))
      {
         parent = parent.parent();
      }
      
      ancestorCache.put(clazz, parent == null ? null : new WeakReference(parent));
      
      return (T) parent;
   }

   /**
    * Get this or widget parent of the specified type.
    * 
    * @param   clazz
    *          The type to return.
    * 
    * @return  widget parent or this.
    */
   @Override
   @SuppressWarnings("unchecked")
   public <T extends Widget<O>> T parentOrSelf(Class<T> clazz)
   {
      return parentOrSelfObject(clazz);
   }

   /**
    * Get this or widget parent of the specified type.
    *
    * @param   clazz
    *          The type to return.
    *
    * @return  widget parent or this.
    */
   @Override
   @SuppressWarnings("unchecked")
   public <T> T parentOrSelfObject(Class<T> clazz)
   {
      return clazz.isAssignableFrom(getClass()) ? (T) this : parentObject(clazz);
   }

   /**
    * Process focus event.
    * 
    * @param   event
    *          Event to process.
    */
   @Override
   public void processFocusEvent(FocusEvent event)
   {
      for (FocusListener listener : focusListeners)
      {
         event.dispatch(listener);
      }
   }

   /**
    * Process key event.
    * 
    * @param   event
    *          Key event to process.
    */
   @Override
   public void processKeyEvent(KeyInput event)
   {
      processKeyListeners(event);
   }

   /**
    * Process key listeners.
    * 
    * @param    event
    *           Keyboard event.
    */
   @Override
   public void processKeyListeners(KeyInput event)
   {
      for (KeyListener listener : keyListeners)
      {
         event.dispatch(listener);

         if (event.isConsumed())
            break;
      }
   }

   /**
    * Process direct manipulation event.
    * 
    * @param   event
    *          Event to process.
    */
   @Override
   public void processDirectManipulationEvent(DirectManipulationEvent event)
   {
      int id = event.id();
      
      // process events depending on ID code
      if (id == Keyboard.SE_SELECTION || id == Keyboard.SE_DESELECTION)
      {
         boolean newSelected = (id == Keyboard.SE_SELECTION); 
         BaseConfig bc = (BaseConfig)config();
         // check if we need to sync numbers of selected widgets 
         if (bc.selected != newSelected)
         {
            // yes, need to sync parent container config
            Container<?> pCont = event.getParentContainer();
            if (pCont != null)
            {
               if (newSelected)
               {
                  ((PaneConfig)pCont.config()).numSelectedWidgets++;
               }
               else
               {
                  ((PaneConfig)pCont.config()).numSelectedWidgets--;
               }
            }
            bc.selected = newSelected;
         }
      }
      else if (id == Keyboard.SE_START_BOX_SELECTION)
      {
      }
      else if (id == Keyboard.SE_END_BOX_SELECTION)
      {
         // clean up and refresh the container
         repaint();
      }
      else if (id == Keyboard.SE_START_RESIZE)
      {
      }
      else if (id == Keyboard.SE_END_RESIZE)
      {
         // compute new widget size and location depending on resize direction
         NativePoint     pLocCurr = physicalLocation();
         NativeDimension pDimCurr = physicalDimension();
         switch (event.getResizeDirection())
         {
            case MouseDirectManipulation.RESIZE_DIRECTION_NW:
               pLocCurr.x += event.getXChange();
               pLocCurr.y += event.getYChange();
               pDimCurr.width  -= event.getXChange();
               pDimCurr.height -= event.getYChange();
               break;
            case MouseDirectManipulation.RESIZE_DIRECTION_N:
               pLocCurr.y += event.getYChange();
               pDimCurr.height -= event.getYChange();
               break;
            case MouseDirectManipulation.RESIZE_DIRECTION_NE:
               pLocCurr.y += event.getYChange();
               pDimCurr.width  += event.getXChange();
               pDimCurr.height -= event.getYChange();
               break;
            case MouseDirectManipulation.RESIZE_DIRECTION_E:
               pDimCurr.width  += event.getXChange();
               break;
            case MouseDirectManipulation.RESIZE_DIRECTION_SE:
               pDimCurr.width  += event.getXChange();
               pDimCurr.height += event.getYChange();
               break;
            case MouseDirectManipulation.RESIZE_DIRECTION_S:
               pDimCurr.height += event.getYChange();
               break;
            case MouseDirectManipulation.RESIZE_DIRECTION_SW:
               pLocCurr.x += event.getXChange();
               pDimCurr.width  -= event.getXChange();
               pDimCurr.height += event.getYChange();
               break;
            case MouseDirectManipulation.RESIZE_DIRECTION_W:
               pLocCurr.x += event.getXChange();
               pDimCurr.width  -= event.getXChange();
               break;
         }
         // set up new location and size
         setPhysicalLocation(pLocCurr.x, pLocCurr.y);
         setPhysicalDimensionInt(pDimCurr.width, pDimCurr.height);
      }
      else if (id == Keyboard.SE_START_MOVE)
      {
      }
      else if (id == Keyboard.SE_END_MOVE)
      {
         // change widget location according to XY shift values encoded into message
         NativePoint physLoc = physicalLocation();
         setPhysicalLocation(physLoc.x + event.getXChange(), physLoc.y + event.getYChange());

         // once the widget has moved, assume a non-automatic client-side location
         physLoc = physicalLocation();
         BaseConfig bc = (BaseConfig)config();
         bc.x = physLoc.x;
         bc.y = physLoc.y;
      }
      else if (id == Keyboard.SE_EMPTY_SELECTION)
      {
         // clean up container
         repaint();
      }
      else if (id == Keyboard.SE_DROP_FILE_NOTIFY)
      {
         // default processing for file list dropping is to call END-FILE-DROP method
         // moreover this is mandatory event NO-APPLY returned from trigger executed
         // because allocated resources must be freed before next DnD operation starts
         DragDropHelper.endFileDrop();
         // and repaint to commit possible widget changes
         repaint();
      }

      // repaint parent container to make the changes visible
      if (event.isRepaintRequired())
      {
         event.getParentContainer().repaint();
      }
   }

   /**
    * Request widget repainting.
    */
   @Override
   public void repaint()
   {
      // TODO: unrealized widgets must never be repainted.  code bellow needs to be enabled only
      // when all widgets which have configs (part of the window decorations or frames) have the
      // realized state activated properly
      if (!isDisplayed()/* || (config() != null && !isRealized())*/)
         return;

      repaint(bounds());
   }

   /**
    * Request widget repainting.
    * 
    * Only the given update rectangle will be repainted (note that it is
    * clipped by all parent containers). No event is posted if the widget
    * itself or any of the parent containers are hidden.
    * 
    * @param    ur
    *           The update rectangle (relative to the origin of its parent
    *           Container) which will be repainted. Note that it is clipped by
    *           all parent containers.
    */
   @Override
   public void repaint(Rectangle ur)
   {
      // TODO: unrealized widgets must never be repainted.  code bellow needs to be enabled only
      // when all widgets which have configs (part of the window decorations or frames) have the
      // realized state activated properly
      if (!isDisplayed() || ur.empty() /* || (config() != null && !isRealized())*/)
         return;

      // do not draw the widget, if it should not be repainted
      if (isRepaintDisabled())
      {
         return;
      }

      Container<O> ct = parent();

      // do not readjust coordinates if the root window is reached
      while (ct != null && !(ct instanceof TopLevelWindow))
      {
         if (!ct.isVisible())
         {
            return;
         }

         if (ct.clipping())
         {
            ur = ur.intersection(ct.getClipRect());

            if (ur == null || ur.empty())
            {
               return;
            }
         }
         
         Point parentLoc = prepParentLocation(ct);

         // insets must be translated the same way they are in 
         // screenLocation/screenPhysicalLocation so that we end up
         // with the correct screen location of the update rect
         if (ct instanceof BorderedPanelGuiImpl)
         {
            parentLoc = parentLoc.translate(ct.insets());
         }
         
         ur = ur.translate(parentLoc);
         
         ct = ct.parent();
      }

      EventManager.postEvent(new PaintEvent(this, ur));
   }

   /**
    * Request focus. This implementation ignores the {@code force} parameter as it always sets
    * focus unconditionally.
    *
    * @param   force
    *          When {@code true} the method will set focus on this widget unconditionally.
    */
   @Override
   public void requestFocus(boolean force)
   {
      TitledWindow<O> windowOwner = ancestor();
      if (windowOwner == null)
      {
         // closed
         return;
      }
      Widget<O> currentFocus = windowOwner.currentFocus();
      
      if (hasFocus())
         return;

      if (!ThinClient.getInstance().getFocusManager().isTempFocusChange())
      {
         if (currentFocus != null)
            EventManager.postEvent(new FocusEvent(currentFocus, this, EventType.FOCUS_LOST));

         EventManager.postEvent(new FocusEvent(this, currentFocus, EventType.FOCUS_GAINED));
      }

      parent().setFocus(this);
      repaint();
   }

   /**
    * Request screen synchronization.  
    */
   @Override
   public void requestSync()
   {
      EventManager.postEvent(new SyncEvent(this));
   }

   /**
    * Get absolute location of widget in screen coordinates.
    * 
    * @return  absolute widget location.
    */
   @Override
   public Point screenLocation()
   {
      Container<O> parent = parent();
      
      if (parent == null) {
         throw new IllegalStateException("widget is not in the container");
      }
      
      Point p = parent.screenLocation();
      
      // Account for insets being translated during BorderedPanelGuiImpl.draw().
      // We have to end up in the same location whether calling draw() on the top-level 
      // window or whether using screen location and calling draw() on the leaf widget.
      // TODO: Currently only BorderedPanelGuiImpl translates its origin during draw(),
      //       this should be a more generic feature however, implemented at AbstractContainer.
      //       Once insets draw translation is moved to AbstractContainer remove 
      //       the conditional.
      if (parent instanceof BorderedPanelGuiImpl)
      {
         p = p.translate(parent.insets());
      }
      
      return Coordinate.scale(p.translate(location()));
   }

   /**
    * Get absolute location of widget in screen coordinates.
    * The result is in native units.
    * 
    * @return  absolute widget location.
    */
   @Override
   public NativePoint screenPhysicalLocation()
   {
      Container<O> parent = parent();
      
      if (parent == null) 
      {
         throw new IllegalStateException("widget is not in the container");
      }

      CoordinatesConversion cc = screen().coordinates();
      NativePoint p = parent.screenPhysicalLocation();
      
      // Account for insets being translated during BorderedPanelGuiImpl.draw().
      // We have to end up in the same location whether calling draw() on the top-level 
      // window or whether using screen location and calling draw() on the leaf widget.
      // TODO: Currently only BorderedPanelGuiImpl translates its origin during draw(),
      //       this should be a more generic feature however, implemented at AbstractContainer.
      //       Once insets draw translation is moved to AbstractContainer remove 
      //       the conditional.
      if (parent instanceof BorderedPanelGuiImpl)
      {
         p = p.translate(cc.insetsToNative(parent.insets()));
      }
            
      return p.translate(physicalLocation());
   }
   
   /**
    * Get display location of widget in screen coordinates.
    * The result is in native units.
    * 
    * @return  display widget location.
    */
   @Override
   public NativePoint displayPhysicalLocation()
   {
      TopLevelWindow<?> window = topLevelWindow();
      if (window == null)
      {
         return screenPhysicalLocation();
      }
      
      // window physical location represents a location on the display, 
      // whereas screen physical location is the offset from the parent window
      return window.physicalLocation().translate(screenPhysicalLocation());
   }

   /**
    * Get the cached value of {@link #screenLocation()}.
    * 
    * @return   The cached {@link #screenLocation}.  <code>null</code> means no value is
    *           cached.
    */
   public final Point getCachedScreenLocation()
   {
      return screenLocation == null ? null : new Point(screenLocation);
   }

   /**
    * Get the cached value of {@link #screenPhysicalLocation()}.
    * 
    * @return   The cached {@link #screenPhysicalLocation}.  <code>null</code> means no value is
    *           cached.
    */
   public final NativePoint getCachedScreenPhysicalLocation()
   {
      return screenPhysicalLocation == null ? null : new NativePoint(screenPhysicalLocation);
   }
   
   /**
    * Set the cached value for {@link #screenLocation()}, in {@link #screenLocation}.
    * 
    * @param    p
    *           The computed location.
    */
   public final void setCachedScreenLocation(Point p)
   {
      screenLocation = new Point(p);
   }
   
   /**
    * Set the cached value for {@link #screenPhysicalLocation()}, in {@link #screenPhysicalLocation}.
    * 
    * @param    p
    *           The computed location.
    */
   public final void setCachedScreenPhysicalLocation(NativePoint p)
   {
      screenPhysicalLocation = new NativePoint(p);
   }
   
   /**
    * Check if either {@link #screenLocation} or {@link #screenPhysicalLocation} contains a cached
    * value.
    * 
    * @return    <code>true</code> if either of {@link #screenLocation} or 
    *            {@link #screenPhysicalLocation} is not-null.
    */
   public final boolean isScreenLocationCached()
   {
      return screenPhysicalLocation != null || screenLocation != null;
   }
   
   /**
    * Called when the widget's location changes. This allows the subclasses to react
    * on this, for example by notifying the driver.
    */
   public void reflectLocationChange()
   {
      screenLocation = null;
      screenPhysicalLocation = null;
   }

   /**
    * Called when any of the widget's ancestors gets re-assigned.  It will clear all cached values
    * ({@link #ancestorCache} and {@link #window}).
    */
   public void reflectAncestorChange()
   {
      ancestorCache.clear();
      window = null;
   }

  /**
    * Enable/disable widget.
    * 
    * @param   enabled
    */
   @Override
   public void setEnabled(boolean enabled)
   {
      // check if the status is changing
      boolean flagChanged = isEnabled() != enabled;
      
      // enabling widget resets MODIFIED attribute regardless of the current enabled state
      if (enabled)
      {
         WidgetConfig widCfg = config();
         // not all widgets have config
         if (widCfg != null)
         {
            ConfigHelper.setModified(widCfg, false);
         }
      }
      
      _setEnabled(enabled);
      
      // a kind of optimization to minimize unnecessary work
      if (flagChanged)
      {
         if (enabled && isFocusable())
         {
            screen().getRegistry().addEditableFrameForWidget(this);
         }
         else
         {
            screen().getRegistry().removeEditableFrameForWidget(this);
         }
      }

      if (flagChanged)
      {
         repaint();
      }
   }

   /**
    * Tracks the hidden state of the widget.
    * 
    * @param    hidden
    *           New hidden state.
    */
   @Override
   public void setHidden(boolean hidden)
   {
      _setHidden(hidden);
   }

   /**
    * Tracks the highlighted state of the widget.
    * 
    * @param    highlighted
    *           New highlighted state.
    */
   @Override
   public void setHighlighted(boolean highlighted)
   {
      this.highlighted = highlighted;
   }

   /**
    * Set widget relative location.
    * 
    * The x value corresponds to COLUMN while
    * y value corresponds to ROW 4GL attribute.
    * Note however that the values are expected to be
    * 0-based.
    * 
    * @param   x
    *          Widget X.
    * @param   y
    *          Widget Y.
    */
   @Override
   public void setLocation(double x, double y)
   {
      BaseConfig bc = UiUtils.getConfig(this, BaseConfig.class);
      
      if (bc == null)
      {
         // although little hackish this will nicely take advantage
         // of the config automated conversion 
         bc = new BaseConfig();
      }

      if (Double.compare(location.x, x) == 0 && Double.compare(location.y, y) == 0 &&
          Double.compare(bc.clientColumn, x) == 0 && Double.compare(bc.clientRow, y) == 0)
      {
         return;
      }
      
      if (isVisible())
         repaint(); 
      
      bc.clientColumn = x;
      bc.clientRow = y;
      
      // When this has associated config, physicalLocation/location is not effectively used. 
      // Still, keep it synchronized for clarity - for example this may prevent 
      // confusion during debugging.
      physicalLocation.x = bc.clientX;
      physicalLocation.y = bc.clientY;
      
      location.x = bc.clientColumn;
      location.y = bc.clientRow;
      
      if (isVisible())
         repaint();

      reflectLocationChange();
   }

   /**
    * Set widget relative location.
    * 
    * @param   origin
    *          Location to set.
    */
   @Override
   public void setLocation(Point origin)
   {
      setLocation(origin.x, origin.y);
   }

   /**
    * Set widget relative physical location.
    * 
    * @param   x
    *          Widget physical X.
    * @param   y
    *          Widget physical Y.
    */
   @Override
   public void setPhysicalLocation(int x, int y)
   {
      BaseConfig bc = UiUtils.getConfig(this, BaseConfig.class);
      
      if (bc == null)
      {
         // although little hackish this will nicely take advantage
         // of the config automated conversion 
         bc = new BaseConfig();
      }
      
      if (screen().isChui())
      {
         // in chui the characters-native conversion is 1:1
         bc.clientColumn = x;
         bc.clientRow = y;
      }
      else
      {
         // in gui the pixels-native conversion is 1:1
         bc.clientX = x;
         bc.clientY = y;
      }
      
      // When this has associated config, physicalLocation/location is not effectively used. 
      // Still, keep it synchronized for clarity - for example this may prevent 
      // confusion during debugging.
      physicalLocation.x = bc.clientX;
      physicalLocation.y = bc.clientY;
      
      location.x = bc.clientColumn;
      location.y = bc.clientRow;

      reflectLocationChange();
   }
   
   /**
    * Copy the widget state from the specified source widget to this instance.
    * 
    * @param    w
    *           The source widget which state needs to be copied.
    */
   protected void copyState(Widget<?> w)
   {
      // do not set parent here

      AbstractWidget<?> widget = (AbstractWidget<?>) w;
      screen = (O) widget.screen.managerInstance();
      
      if (config() instanceof BaseConfig)
      {
         BaseConfig bc = (BaseConfig) config();
         bc.widgetPlaced = ((BaseConfig) widget.config()).widgetPlaced;
      }
      
      enabled = widget.enabled;
      hidden = widget.hidden;
      highlighted = widget.highlighted;
      visible = widget.visible;
      focusListeners = widget.focusListeners;
      keyListeners = widget.keyListeners;
      popupKeyListnrAdded = widget.popupKeyListnrAdded;
      listeners = widget.listeners;
      mouseActions = widget.mouseActions;
      insets = widget.insets;
      nativeInsets = widget.nativeInsets;
      mousePtr = widget.mousePtr;
      mouseHoverAction = widget.mouseHoverAction;
      directManipulationActive = widget.directManipulationActive;

      // these need to be recomputed with ChUI location points...
      // location = widget.location;
      // physicalLocation = widget.physicalLocation;

      screen.getRegistry().addWidget((Widget) this);
   }

   /**
    * Set widget physical dimension.
    * 
    * @param   width
    *          Widget physical width.
    * @param   height
    *          Widget physical height.
    */
   private void setPhysicalDimensionInt(int width, int height)
   {
      BaseConfig bc = UiUtils.getConfig(this, BaseConfig.class);
      
      if (bc == null)
      {
         // widget must be realized to handle direct manipulation resize
         return;
      }
      
      if (screen().isChui())
      {
         // in chui the characters-native conversion is 1:1
         bc.clientWidthChars = width;
         bc.clientHeightChars = height;
         bc.widthChars = width;
         bc.heightChars = height;
      }
      else
      {
         // in gui the pixels-native conversion is 1:1
         bc.clientWidthPixels = width;
         bc.clientHeightPixels = height;
         bc.widthPixels = width;
         bc.heightPixels = height;
      }
      
      bc.fixedWidth = true;
      bc.fixedHeight = true;
   }
   
   /**
    * Determine the physical bounds, relative to screen, expressed in units native to UI type.
    * 
    * @return  See above.
    */
   @Override
   public NativeRectangle physicalBounds()
   {
      return new NativeRectangle(physicalLocation(), physicalDimension());
   }
   
   /**
    * Resize the widget to the specified dimensions.
    * 
    * @param    origin
    *           The top-left window corner new position.
    * @param    size
    *           The new window size.
    * @param    raise
    *           Flag indicating if the 4GL WINDOW-RESIZED event needs to be raised.
    */
   @Override
   public void resizeTo(NativePoint origin, NativeDimension size, boolean raise)
   {
      // no-op by default
   }
   
   /**
    * Set widget parent.
    * 
    * @param   container
    *          Parent to set.
    */
   @Override
   public final void setParent(Container<O> container)
   {
      parent = (container == null) ? 
               null : 
               new WeakReference<Container<O>>(container);
      
      reflectAncestorChange();
      reflectLocationChange();
   }

   /**
    * Hide/show widget.
    * 
    * @param   visible
    *          New visible state.
    */
   @Override
   public void setVisible(boolean visible)
   {
      if (visible)
         show();
      else
         hide();
   }

   /**
    * Show widget.
    */
   @Override
   public void show()
   {
      if (!isVisible())
      {
         _setVisible(true); // direct access
         repaint();
      }
   }

   /**
    * Check if widget is totally obscured.
    * 
    * @return  <code>true</code> if widget is totally obscured.
    */
   @Override
   public boolean isTotallyObscured()
   {
      // avoid bounds/ancestor logic in GUI
      
      if (!ThinClient.getInstance().isChui())
      {
         // in GUI, all windows use their own canvas.  In ChUI, the canvas/screen is shared,
         // so intersection needs to be checked.
         return false;
      }
      
      return WindowManager.checkRectangle(bounds(), ancestor());
   }

   /**
    * Check if widget is visible.
    * 
    * @return  <code>true</code> if widget is visible.
    */
   @Override
   public boolean isVisible()
   {
      return visible;
   }

   /**
    * Check if widget is visible.
    *
    * @param   recursive
    *          <code>false</code> behaves as {@link #isVisible()}, otherwise the method returns
    *          <code>true</code> only when the parent's call to <code>isVisible(true)</code> returns
    *          <code>true</code>.
    *
    * @return  <code>true</code> if widget is visible.
    */
   @Override
   public boolean isVisible(boolean recursive)
   {
      boolean thisVisible = isVisible();
      if (!thisVisible || !recursive)
      {
         return thisVisible;
      }

      Widget parent = parent();
      return parent == null ? thisVisible : parent.isVisible(true);
   }
   
   /**
    * Check if widget is part of a header.
    * 
    * @return  <code>true</code> if widget is part of header.
    */
   @Override
   public boolean isHeader()
   {
      return config().header;
   }

   /**
    * Check if widget needs in cursor to be on when having the input focus.  By default this
    * method always returns <code>true</code> so it needs to be overwritten if other return value
    * is required for particular widget implementation class.
    * 
    * @return  <code>true</code> as the default value.
    */
   @Override
   public boolean forceCursorOnFocus()
   {
      return true;
   }

   /**
    * Get widget width.
    * 
    * @return  widget width.
    */
   @Override
   public abstract double width();

   /**
    * Disable widget.
    */
   protected void disable()
   {
      _setEnabled(false);
   }

   /**
    * Enable widget.
    */
   protected void enable()
   {
      _setEnabled(true);
   }
   
   /**
    * Instructs the widget that its size may have changed. The widget updates
    * its state from the associated config instance. It is the responsibility
    * of the caller to call this method only after the config size is set with
    * the new values.
    * <p>
    * This implementation uses the supplied config instance to detect whether
    * size has changed. If <code>null</code> is passed in the method does
    * nothing.
    * 
    * @param    beforeUpdate
    *           {@link BaseConfig} reference or <code>null</code>.
    */
   protected <C extends WidgetConfig> void updateSizeWorker(C beforeUpdate)
   {
      if (beforeUpdate == null)
      {
         return;
      }
      autoResize(beforeUpdate);
      if (!(beforeUpdate instanceof BaseConfig))
      {
         return;
      }
      ControlConfig bc = UiUtils.getConfig(this, ControlConfig.class);
      
      // apply only to controls
      if (bc != null && ((BaseConfig)beforeUpdate).widthChars != bc.widthChars)
      {
         repaint();
      }
   }   

   /**
    * Invoke auto-resize logic if needed
    * 
    * @param    beforeUpdate
    *           {@link BaseConfig} reference or <code>null</code>.
    */
   protected <C extends WidgetConfig>  void autoResize(C beforeUpdate)
   {
      if (this instanceof AutoResizable<?>)
      {
         AutoResizable<C> ar = (AutoResizable<C>) this;
         if (ar.isAutoResize())
         {
            ar.updateSize(beforeUpdate);
         }
      }
   }
   
   /**
    * Process the given key, if this widget's mnemonic matches the key.
    * Widgets which can handle mnemonics will override this method.
    * 
    * @param    key
    *           The pressed mnemonic key code.
    *           
    * @return   <code>true</code> if this widget accepted this mnemonic key.  Not enabled or not
    *           visible widget will not accept it.   Returns <code>false</code> by default.
    */
   @Override
   public boolean processMnemonic(int key)
   {
      return false;
   }

   /**
    * Common code to process the specific mnemonic key by this widget.
    * 
    * @param    key
    *           The pressed mnemonic key code.
    *           
    * @return   <code>true</code> if this widget accepted this mnemonic key.  Not enabled or not
    *           visible widget will not accept it.
    */
   protected final boolean processMnemonicImpl(int key, MnemonicTrait m)
   {
      if (!isEnabled() || !isVisible())
      {
         return false;
      }
      
      String lbl = Keyboard.keyLabel(key);
      if (lbl != null && lbl.length() == 1)
      {
         // all mnemonics are case-insensitive...
         lbl = lbl.toUpperCase();
         
         key = Keyboard.keyCode(lbl);
      }
      
      Integer mnemCode = (Integer) m.getMnemonicKeyCode(false, false).orElse(null);
      if (mnemCode != null && mnemCode.equals(key))
      {
         return true;
      }

      mnemCode = (Integer) m.getMnemonicKeyCode(false, true).orElse(null);

      return (mnemCode != null && mnemCode.equals(key));
   }
   
   /**
    * Process event and dispatch it to specific method.
    * 
    * @param   evt
    *          Event to process.
    */
   @Override
   public void processEvent(Event evt)
   {
      if (evt instanceof KeyInput)
      {
         Widget<O> ancestor = ancestor();

         if (ancestor != null)
            ancestor.processKeyEvent((KeyInput) evt);
      }
      else if (evt instanceof FocusEvent)
         processFocusEvent((FocusEvent) evt);
   }

   /**
    * Get externalizable widget config which can be used to restore widget 
    * state after via network. 
    * 
    * Note that only very limited set of widgets actually have config.
    * By default widget has no config.
    * 
    * @param   <W>
    *          Config type.
    *          
    * @return  Reference to widget config.
    */
   @Override
   public <W extends WidgetConfig> W config()
   {
      return null;
   }
   
   /**
    * Checks whether this widget can receive focus generally.
    * 
    * @return  <code>true</code> if this widget can be focused under
    *          certain conditions. <code>false</code> if, and only if, this
    *          widget can never receive focus.    
    */
   @Override
   public boolean isFocusable()
   {
      return true;
   }
   
   /**
    * This API is responsible for resolving the font associated with this widget's title.  
    * This can be one of the default fonts, a special system font, or an entry in the font-table.
    * 
    * @return   See above.
    */
   public int resolveTitleFont()
   {
      throw new RuntimeException("Not implemented.");
   }
   
   /**
    * This API is responsible for resolving the widget's font.  This can be one of the default 
    * fonts, a special system font, or an entry in the font-table.
    * 
    * @return   See above.
    */
   public int resolveFont()
   {
      throw new RuntimeException("Not implemented.");
   }
   
   /**
    * Find the widget positioned just bellow the specified mouse position (in physical units).
    * 
    * @param    p
    *           The mouse physical location.
    * 
    * @return   A widget over which the mouse is positioned or <code>null</code> if none found.
    */
   @Override
   public Widget<O> findMouseSource(NativePoint p)
   {
      return isVisible() && this.physicalBounds().contains(p) ? this : null;
   }

   /**
    * Get a point to represent the LAST-EVENT:X and LAST-EVENT:Y coordinates.  For mouse portable
    * events, these represent:
    * <ul>
    *    <li>If the widget is attached to a frame, then is a point relative to the frame's 
    *    top-left corner (excluding title).
    *    </li>
    *    <li>If the widget is a window component, then is a point relative to the window's 
    *    workspace top-left corner
    *    </li>
    * </ul>
    * 
    * @param    point
    *           The point (relative to the window's top-left corner) for which the coordinates for
    *           the portable mouse events are needed.
    * @return   A point representing the coordinates of a mouse-portable event or 
    *           <code>null</code> if such an event can't be raised.
    */
   public NativePoint getPortableMouseEventCoordinates(NativePoint point)
   {
      Frame<O> frame = parentOrSelf(Frame.class);
      
      if (frame != null)
      {
         return frame.getPortableMouseEventCoordinates(point);
      }
      
      TopLevelWindow<O> window = parentOrSelf(TopLevelWindow.class);
      
      if (window != null)
      {
         return window.getPortableMouseEventCoordinates(point);
      }
      
      return null;
   }
   
   /**
    * Notification of a mouse dragged event occurred for this widget.  No-op by default.
    * 
    * @param    e
    *           The mouse event.
    */
   @Override
   public void mouseDragged(MouseEvent e)
   {
      // no-op
   }

   /**
    * Notification of a mouse moved event occurred for this widget.  No-op by default.
    * 
    * @param    e
    *           The mouse event.
    */
   @Override
   public void mouseMoved(MouseEvent e)
   {
      // no-op
   }
   
   /**
    * Notification of a mouse clicked event occurred for this widget.
    * <p>
    * In GUI mode, if the widget has a popup-menu attached, it will display it.
    * 
    * @param    e
    *           The mouse event.
    */
   @SuppressWarnings("unchecked")
   @Override
   public void mouseClicked(MouseEvent e)
   {
      // right button is clicked under GUI
      if (e.getButton() == getMenuMouse() && !ThinClient.getInstance().isChui() && isDisplayed())
      {
         showPopupMenu(e, null);
         e.consume();
      }
   }

   /**
    * Notification of a mouse pressed event occurred for this widget.  No-op by default.
    * 
    * @param    e
    *           The mouse event.
    */
   @Override
   public void mousePressed(MouseEvent e)
   {
      // no-op
   }

   /**
    * Notification of a mouse released event occurred for this widget.  No-op by default.
    * 
    * @param    e
    *           The mouse event.
    */
   @Override
   public void mouseReleased(MouseEvent e)
   {
      // no-op
   }

   /**
    * Notification of a mouse entered event occurred for this widget.  No-op by default.
    * 
    * @param    e
    *           The mouse event.
    */
   @Override
   public void mouseEntered(MouseEvent e)
   {
      // no-op
   }

   /**
    * Notification of a mouse exited event occurred for this widget.  No-op by default.
    * 
    * @param    e
    *           The mouse event.
    */
   @Override
   public void mouseExited(MouseEvent e)
   {
      // no-op
   }

   /**
    * Notification of a mouse wheel moved event occurred for this widget.  No-op by default.
    * 
    * @param    e
    *           The mouse event.
    */
   @Override
   public void mouseWheelMoved(MouseWheelEvent e)
   {
      // no-op
   }
   
   /**
    * Destroy this widget.  Notify all listeners about the widget being destroyed.
    */
   public void destroy()
   {
      for (WidgetStateListener l : listeners)
      {
         l.notifyWidgetDestroyed(this);
      }
      
      if (parent != null)
      {
         Container<O> ctr = parent.get();
         if (ctr != null)
            ctr.detach(this);
      }
      
      try
      {
         FontManager.removeFontResolver(window(), this);
      }
      catch(IllegalStateException ex)
      {
         // this widget is not attached to a window
      }
   }

   /**
    * Register a widget state listener for this widget.
    * 
    * @param    l
    *           The widget state listener.
    */
   public void registerWidgetStateListener(WidgetStateListener l)
   {
      if (l != null)
      {
         this.listeners.add(l);
      }
   }
   
   /**
    * Translate mouse position relative to this widget. Widget location is relative to container
    * on which is layout. The mouse position is relative to window content pane which is the mouse
    * listener. This function translate mouse position coordinates relative to the widget.
    *  
    * @param   evt
    *          Mouse event.
    *
    * @return  see above.
    */
   public NativePoint translate(MouseEvent evt)
   {
      NativePoint p = screenPhysicalLocation();

      // mouse position relative to widget position
      return new NativePoint(evt.getX() - p.x, evt.getY() - p.y);
   }

   /**
    * Gets the current state of this component's data.
    * 
    * @return   The state of this widget's data  (for example 
    *           <code>ScreenBuffer.UNINITIALIZED</code>).
    */
   @Override
   public byte getState()
   {
      return config().state;
   }

   /**
    * Gets the sub-rectangle which should be visible.
    * 
    * @return   The sub-rectangle which should be visible 
    */
   public Rectangle getVisiblePart()
   {
      return null;
   }

   /**
    * Stores the new state of the component's data. 
    * 
    * @param    state
    *           The new state (for example <code>ScreenBuffer.CHANGED</code>).
    */
   @Override
   public void setState(byte state)
   {
      if (state < ScreenBuffer.STATE_MIN || state > ScreenBuffer.STATE_MAX)
         return;

      WidgetConfig config = this.config();
      if (config != null)
      {
         config.state = (byte) state;
      }
   }

   /**
    * Build clipped rectangle for given coordinates.
    * 
    * @param   origin
    *          Physical location.
    * @param   dimension
    *          Physical dimension.
    * 
    * @return  see above.
    */
   public static NativeRectangle clipRectangle(NativePoint origin, NativeDimension dimension)
   {
      // corrections for scrolling; substract 1 pixel as the width/height is computed as 
      // right - left + 1 and bottom - top + 1
      int x = dimension.width  - (origin.x < 0 ? origin.x : 0) - 1;
      int y = dimension.height - (origin.y < 0 ? origin.y : 0) - 1;
      // clip rectangle
      return new NativeRectangle(0, 0, y, x);
   }

   /**
    * Returns the popup menu effective for this widget. This implementation only considers the @POPUP-MENU@
    * attribute. In native 4GL popup menus can be overriden, when widgets overlap and both define a popup
    * menu, the one higher in the z-order takes precedence.
    *
    * @return  Popup menu id or -1 if no menu is found.
    */
   public int resolvePopupMenuId()
   {
      WidgetConfig c = config();
      if (c != null && c.popupMenuId != -1)
      {
         return c.popupMenuId;
      }

      Container<O> parent = parent();
      return parent != null ? parent.resolvePopupMenuId() : -1;
   }
   
   /**
    * Build widget clipped rectangle.
    * 
    * @return  see above.
    */
   protected NativeRectangle clipRectangle()
   {
      NativePoint origin = physicalLocation();
      NativeDimension dimension = physicalDimension();
      return clipRectangle(origin, dimension);
   }

   /**
    * Returns true if the widget repaint is disabled. This method is needed to support the
    * WINDOW:DISABLE-REDRAW attribute.
    *
    * @return   See above.
    */
   public boolean isRepaintDisabled()
   {
      WidgetConfig c = config();
      if (c != null && c.disableRedraw)
      {
         return true;
      }
      else
      {
         Container parent = parent();
         return parent != null ? parent.isRepaintDisabled() : false;
      }
   }

   /**
    * Update the {@link #enabled} flag in the configuration and in the widget's state.
    * 
    * @param    enabled
    *           The widget's enabled state.
    */
   private void _setEnabled(boolean enabled)
   {
      this.enabled = enabled;
      if (config() != null)
      {
         config().enabled = enabled;
      }
      else
      {
         return;
      }
      
      if (!supportsCustomMousePointer() || 
            ((this instanceof Frame) && !config().realized))
      {
         return;
      }
      
      if (!enabled && mouseHoverAction == null)
      {
         return;
      }

      if (parent() != null || this instanceof Frame)
      {
         if (mouseHoverAction == null)
         {
            mouseHoverAction = createMouseHoverAction(); 
         }

         mouseHoverAction.gd().selectWindow(topLevelWindow().getId().asInt());
   
         if (enabled)
         {
            mouseHoverAction.register();
         }
         else
         {
            mouseHoverAction.deregister();
         }
         
         mouseHoverAction.gd().releaseWindow();
      }
   }
   
   /**
    * Create handler for processing mouse hover events
    * 
    * @return handler for processing mouse hover events
    */
   protected MouseHoverAction createMouseHoverAction()
   {
      return new MouseHoverAction(this);
   }

   /**
    * Update the {@link #visible} flag in the configuration and in the widget's state.
    * 
    * @param    visible
    *           The widget's visible state.
    */
   public void _setVisible(boolean visible)
   {
      this.visible = visible;
      WidgetConfig config = config();
      if (config != null)
      {
         config.visible = visible;
         if (visible && !config.realized)
         {
            config.realized = true;
            config.wasRealized = true;
         }
      }
   }
   
   /**
    * Update the {@link #hidden} flag in the configuration and in the widget's state.
    * 
    * @param    hidden
    *           The widget's hidden state.
    */
   public void _setHidden(boolean hidden)
   {
      this.hidden = hidden;
      if (config() != null)
      {
         config().hidden = hidden;
      }
   }

   /**
    * Check if the widget implementation can raise legacy, 4GL-style, mouse events.
    * 
    * @return   Always <code>true</code>.
    */
   @Override
   public boolean raiseLegacyMouseEvents()
   {
      return true;
   }
   
   /**
    * Determine the widget's window and synchronize output with physical terminal.
    * 
    * @see   OutputManager#sync(int)
    */
   public void sync()
   {
      int wid = topLevelWindow().getId().asInt();
      
      screen().sync(wid);
   }
   
   /**
    * Determine the widget's window and mark a rectangle as needing repainting. 
    * 
    * @param   rect
    *          rectangle to mark for repainting
    *          
    * @see     OutputManager#refresh(int, Rectangle)
    */
   public void refresh(Rectangle rect)
   {
      int wid = topLevelWindow().getId().asInt();
      
      screen().refresh(wid, rect);
   }
   
   /**
    * Determine the widget's window and clear specified area of the screen.
    * 
    * @param   origin
    *          Upper left corner of the area to clear.
    * @param   size
    *          Dimensions of the area.
    * @param   color
    *          Color.
    * 
    * @see     OutputManager#clear(int, NativePoint, NativeDimension, Color)
    */
   public void clear(Point origin, Dimension size, Color color)
   {
      int wid = topLevelWindow().getId().asInt();
      
      screen().clear(wid, origin, size, color);
   }
   
   /**
    * Determine the widget's window, get a copy of the current screen bitmap and reset clipping 
    * entirely. All screen positions become available for output. The previous state can be
    * restored using setBitmap() method.
    * 
    * @return  a copy of the current screen bitmap.
    * 
    * @see     OutputManager#getAndResetBitmap(int)
    */
   public ScreenBitmap getAndResetBitmap()
   {
      int wid = topLevelWindow().getId().asInt();

      return screen().getAndResetBitmap(wid);
   }
   
   /**
    * Determine the widget's window and reset the screen bitmap.
    * 
    * @see     OutputManager#resetBitmap(int)
    */
   public void resetBitmap()
   {
      int wid = topLevelWindow().getId().asInt();

      screen().resetBitmap(wid);
   }

   /**
    * Determine the widget's window, get a copy of the current screen bitmap and clip output 
    * further so it is guaranteed not to go outside of the provided rectangle.
    * 
    * @param   rect
    *          The extra clipping rectangle. <code>null</code> has no effect
    *          on clipping.
    * 
    * @return  a copy of the current screen bitmap.
    * 
    * @see     OutputManager#getAndSetBitmap(int, Rectangle)
    */
   public ScreenBitmap getAndSetBitmap(Rectangle rect)
   {
      int wid = topLevelWindow().getId().asInt();

      return screen().getAndSetBitmap(wid, rect);
   }

   /**
    * Determine the widget's window and get a copy of the current screen bitmap.
    * 
    * @return  a copy of the current screen bitmap.
    * 
    * @see     OutputManager#getBitmapCopy(int)
    */
   public ScreenBitmap getBitmapCopy()
   {
      int wid = topLevelWindow().getId().asInt();

      return screen().getBitmapCopy(wid);
   }
   
   /**
    * Determine the widget's window and restore the bitmap to a previous state.
    * 
    * @param   bitmap
    *          The bitmap instance to set as the output primitive's bitmap.
    * 
    * @see     OutputManager#setBitmap(int, ScreenBitmap)
    */
   public void setBitmap(ScreenBitmap bitmap)
   {
      int wid = topLevelWindow().getId().asInt();

      screen().setBitmap(wid, bitmap);
   }

   /**
    * Determine the widget's window and add a new clipping rectangle to the bitmap. If the 
    * rectangle is null, fill the entire bitmap with <code>true</code> values. Else, if the bitmap
    * was just reset, fill it with <code>false</code> values and set the clipping region to true.
    * 
    * @param   rect
    *          The clipping rectangle to be added.
    * 
    * @see     OutputManager#clipWith(int, Rectangle)
    */
   public void clipWith(Rectangle rect)
   {
      int wid = topLevelWindow().getId().asInt();

      screen().clipWith(wid, rect);
   }
   
   /**
    * Determine the widget's window and get the number of rows on the screen.
    * 
    * @return  The number of rows (lines of characters).
    */
   public double screenHeight()
   {
      int wid = topLevelWindow().getId().asInt();

      return screen.screenHeight(wid);
   }
   
   /**
    * Compute the text width using {@link GuiDriver#getTextWidth}.
    * 
    * @param    txt
    *           The text.
    * @param    font
    *           The font index in the font table.
    *           
    * @return   The text width.
    * 
    * @see      AbstractWidget#getTextWidth(String, int, boolean)
    */
   public int getTextWidthNative(String txt, int font)
   {
      return getTextWidth(txt, font, true);
   }

   /**
    * Compute the text width using {@link GuiDriver#getTextWidth}.
    *
    * @param    txt
    *           The text.
    * @param    font
    *           The font details.
    *
    * @return   The text width.
    *
    * @see      AbstractWidget#getTextWidth(String, FontDetails, boolean)
    */
   public int getTextWidthNative(String txt, FontDetails font)
   {
      return getTextWidth(txt, font, true);
   }
   
   /**
    * Compute the text width using {@link FontManager#getTextWidth}.
    * 
    * @param    txt
    *           The text.
    * @param    font
    *           The font index in the font table.
    *           
    * @return   The text width.
    * 
    * @see      AbstractWidget#getTextWidth(String, int, boolean)
    */
   public int getTextWidth(String txt, int font)
   {
      return getTextWidth(txt, font, false);
   }

   /**
    * Compute the text width using {@link FontManager#getTextWidth}.
    *
    * @param    txt
    *           The text.
    * @param    font
    *           The font details.
    *
    * @return   The text width.
    *
    * @see      AbstractWidget#getTextWidth(String, int, boolean)
    */
   public int getTextWidth(String txt, FontDetails font)
   {
      return getTextWidth(txt, font, false);
   }
   
   /**
    * Compute the text height using {@link GuiDriver#getTextHeight}.
    * 
    * @param    txt
    *           The text.
    * @param    font
    *           The font index in the font table.
    *           
    * @return   The text height.
    * 
    * @see      AbstractWidget#getTextHeight(String, int, boolean)
    */
   public int getTextHeightNative(String txt, int font)
   {
      return getTextHeight(txt, font, true);
   }
   
   /**
    * Compute the text height using {@link FontManager#getTextHeight}.
    * 
    * @param    txt
    *           The text.
    * @param    font
    *           The font index in the font table.
    *           
    * @return   The text height.
    * 
    * @see      AbstractWidget#getTextHeight(String, int, boolean)
    */
   public int getTextHeight(String txt, int font)
   {
      return getTextHeight(txt, font, false);
   }
   
   /**
    * Determine the widget's window and get the number of columns on the screen.
    *
    * @return  The number of columns (in characters).
    */
   public double screenWidth()
   {
      int wid = topLevelWindow().getId().asInt();

      return screen.screenWidth(wid);
   }

   /**
    * Determine the widget's window and get the screen dimensions.
    * 
    * @return screen dimension.
    */
   public Dimension screenDimension()
   {
      int wid = topLevelWindow().getId().asInt();

      return screen.screenDimension(wid);
   }
   
   /**
    * Check if the current widget can manage non-global system actions on its own, and not via the
    * enclosing event loop.
    * 
    * @return   <code>true</code> if the widget has its own event management.
    *           <code>false</code> by default.
    */
   public boolean manageSystemActions()
   {
      return false;
   }

   /**
    * Obtain any selected text in the widget for copying to the clipboard.
    *
    * @return   The selected text or <code>null</code> if there is no selected text or if the
    *           widget type does not support clipboard operations.
    */
   @Override
   public String getSelectedText()
   {
      return null;
   }
   
   /**
    * Get the default event associated for this widget and the given key.
    * 
    * @param    key
    *           The current key.
    * @param    inEditingBlock
    *           <code>true</code> if EDITING mode is active.
    *           
    * @return   The action code for the widget's default event.
    */
   public int getDefaultEvent(KeyInput key, boolean inEditingBlock)
   {
      return -1;
   }

   /**
    * Process the action associated with the given key in an independent drawing bracket.
    * <p>
    * This must be called only in the context of widget-level event processing.
    * 
    * @param    key
    *           The key to be processed.
    */
   public void processSystemKey(final KeyInput key)
   {
      if (!manageSystemActions())
      {
         return;
      }
      
      // let the widget process this key
      ThinClient.getInstance().independentEventDrawingBracket(this, new Runnable()
      {
         @Override
         public void run()
         {
            // activate MENUBAR for CHUI and set default focused menu item.
            if (key.keyCode() == Keyboard.keyCode("F3") && ThinClient.getInstance().isChui())
            {
               if (ancestor() instanceof Window)
               {
                  Window<O> wnd = (Window<O>) ancestor();
                  if (wnd.getMenubar() != null && wnd.getMenubar().isVisible())
                  {
                     Widget<?> currentFocus = UiUtils.getCurrentFocus();
                     if (!(currentFocus instanceof MenuItem) && !(currentFocus instanceof SubMenu))
                     {
                        wnd.getMenubar().setFocus(wnd.getMenubar().defaultFocus());
                        wnd.repaint();
                     }
                  }
               }
            }
            
            processKeyEvent(key);
         }
      });
   }
   
   /**
    * Check if the given event needs to be ignored, as it can't be treated by the widget.
    * 
    * @param    action
    *           The event code.
    *           
    * @return   See above.
    */
   public boolean ignoreAction(int action)
   {
      return EventDefinition.isEventUniversal(action) ||
             action == Keyboard.KA_HELP               ||
             action == Keyboard.KA_TAB                ||
             action == Keyboard.SE_NEXT_FRAME         ||
             action == Keyboard.SE_PREV_FRAME
             ;
   }

   /**
    * The method is called after the configuration associated with 
    * the implementor changes. The parameter points to a reference
    * representing the original configuration state. This reference
    * can be used to detect what configuration fields changed and so
    * optimize any processing related to the configuration change.
    * 
    * @param    beforeUpdate
    *           Config reference capturing the config state before 
    *           the owner's config was modified.
    */
   public <C extends WidgetConfig> void afterConfigUpdateBase(C beforeUpdate)
   {
      C c = config();
      
      if (c == null)
      {
         return;
      }
      
      setVisible(c.visible);
      setHidden(c.hidden);
      if (c.enabled != isEnabled())
      {
         setEnabled(c.enabled);
      }
      
      updateSizeWorker(beforeUpdate);

      //  c.popupMenuId != beforeUpdate.popupMenuId don't work here 
      if (c.popupMenuId != -1 && !popupKeyListnrAdded)
      {
         addKeyListener(new Menu.PopupKeyListener());
         popupKeyListnrAdded = true;
      }

      // register/deregister direct manipulation handlers
      // should work only for GUI code and not for space/skip widgets
      if (parent != null && isDirectManipulationCapable())
      {
         if (isDirectManipulationRequired() && !directManipulationActive)
         {
            activateDirectManipulation();
         }
         else if (!isDirectManipulationRequired() && directManipulationActive)
         {
            deactivateDirectManipulation();
         }
      }

      Container<O> parent;
      if (c instanceof BaseConfig)
      { 
         BaseConfig bc = (BaseConfig) c;

         if (beforeUpdate.realized       &&
             (parent = parent()) != null &&
             (c.visible || beforeUpdate.visible))
         {
            BaseConfig beforeConfig = (BaseConfig) beforeUpdate;
            
            Point beforeLoc = ConfigHelper.getEffectiveLocation(beforeConfig);
            Dimension beforeDim = ConfigHelper.getEffectiveDimension(beforeConfig);
            
            // current (new) boundary must always use the explicit column/row or width/height, if set
            Point loc = location();
            if (bc.column != BaseConfig.INV_COORD)
            {
               loc.x = bc.column;
            }
            if (bc.row != BaseConfig.INV_COORD)
            {
               loc.y = bc.row;
            }
            Dimension dim = dimension();
            if (bc.widthChars != BaseConfig.INV_COORD)
            {
               dim.width = bc.widthChars;
            }
            if (bc.heightChars != BaseConfig.INV_COORD)
            {
               dim.height = bc.heightChars;
            }
            
            if (!loc.equals(beforeLoc) || !dim.equals(beforeDim))
            {
               Point unit = screen().coordinates().baseUnits();
               Rectangle bounds = new Rectangle(beforeLoc, beforeDim, unit); 
               bounds = bounds.translate(parent.screenLocation());
               // this event is posted directly to the queue
               EventManager.postEvent(new PaintEvent(this, bounds));
               
               // repaint on new location, too
               bounds = new Rectangle(loc, dim, unit);
               bounds = bounds.translate(parent.screenLocation());
               // this event is posted directly to the queue
               EventManager.postEvent(new PaintEvent(this, bounds));
            }
            
            if (bc.font != beforeConfig.font)
            {
               repaint();
            }
         }
         
         if (bc.locationChanged)
         {
            bc.locationChanged = false;
            
            reflectLocationChange();
         }
         
         BaseConfig bcBefore = (BaseConfig) beforeUpdate;
         
         if (bcBefore.bgcolor != bc.bgcolor || 
             bcBefore.fgcolor != bc.fgcolor ||
             bcBefore.bgcolorRgb != bc.bgcolorRgb ||
             bcBefore.fgcolorRgb != bc.fgcolorRgb ||
             (bcBefore.dcolor != null && !bcBefore.dcolor.equals(bc.dcolor)) ||
             (bcBefore.pfcolor != null && !bcBefore.pfcolor.equals(bc.pfcolor)))
         {
            repaint();
         }
      }
      
      if (!c.disableRedraw && c.disableRedraw != beforeUpdate.disableRedraw)
      {
         AbstractContainer<O> parentContainer = parent(AbstractContainer.class);
         if (parentContainer != null)
         {
            parentContainer.triggerContainerStateRedrawEnabled(this);
         }

         repaint();
      }
   }
   
   /**
    * Activates direct manipulation engine for this widget.
    */
   protected void activateDirectManipulation()
   {
      ((GuiDriver)screen().getDriver()).registerDirectManipulation(this);
      directManipulationActive = true;
   }
   
   /**
    * Deactivates direct manipulation engine for this widget.
    */
   protected void deactivateDirectManipulation()
   {
      MouseHandler.deregisterWidget(getId().asInt());
      directManipulationActive = false;
   }
   
   /**
    * Checks if direct manipulation feature can be used for this widget.  Only GUI mode can be
    * used with the feature and not with all widget types.
    * 
    * @return   <code>TRUE</code> if direct manipulation is available,
    *           <code>FALSE</code> otherwise. 
    */
   protected boolean isDirectManipulationCapable()
   {
      return isDirectManipulable() && !ThinClient.getInstance().isChui();
   }
   
   /**
    * Checks if direct manipulation feature must be installed or removed in GUI mode.  If config
    * is upadting in a way that stops using direct manipulation it must be removed from system.
    * Different widgets can use different criteria to make this decision.
    * 
    * @return   <code>TRUE</code> if direct manipulation is need to be installed,
    *           <code>FALSE</code> if direct manipulation is need to be removed.
    */
   protected boolean isDirectManipulationRequired()
   {
      WidgetConfig cfg = config();
      return cfg.selectable || (cfg instanceof BaseConfig && ((BaseConfig) cfg).movable);
   }
   
   /**
    * Utility method for debugging, can be used to include information from the server-side
    * to the client-side log
    * 
    * @param message
    *       String parameter
    */
   public void debug(String message)
   {
      LOG.log(Level.INFO,  "{0};  #{1}: {2}", 
            new Object[] {this.getClass().getSimpleName(), getId(), message});
   }

   /**
    * Check the value of the TAB-STOP attribute
    * 
    * @return   the value of the attribute or <code>true</code> if not applicable 
    */
   public boolean isTabStop()
   {
      BaseConfig bc = UiUtils.getConfig(this, BaseConfig.class);
      return bc == null ? true : bc.tabStop;
   }

   /**
    * Additional processing required when widget configuration is updated AFTER the frame layout
    * has been performed.
    */
   @Override
   public void postprocessConfig()
   {
      WidgetConfig wcfg = config();
      if (wcfg instanceof BaseConfig)
      {
         BaseConfig bc = (BaseConfig) wcfg;
         
         if (bc.column != BaseConfig.INV_COORD)
         {
            bc.clientColumn = bc.column;
         }
         
         if (bc.row != BaseConfig.INV_COORD)
         {
            bc.clientRow = bc.row;
         }
      }
   }

   /**
    * Moves this widget to the bottom of other widgets in the z-ordering.
    * 
    * @return  <code>true</code> if operation was successful and
    *          <code>false</code> otherwise.
    */
   @Override
   public boolean moveToBottom()
   {
      Container<O> parent = parent();

      if (parent == null)
         return false;
      
      return parent.moveToBottomInClass(this);
   }

   /**
    * Moves this widget to the top of other widgets in the z-ordering.
    * 
    * @return  <code>true</code> if operation was successful and
    *          <code>false</code> otherwise.
    */
   @Override
   public boolean moveToTop()
   {
      Container<O> parent = parent();

      if (parent == null)
         return false;
      
      return parent.moveToTopInClass(this);
   }

   
   /**
    * Get the Z-Order class
    * 
    * @return Z-Order class of the widget
    */
   @Override
   public ZOrderClass getZOrderClass()
   {
      return ZOrderClass.WIDGET;
   }

   /**
    * Get the value of the MENU-MOUSE attribute
    * 
    * @return the value of the attribute
    */
   @Override
   public int getMenuMouse()
   {
      WidgetConfig config = config();
      
      return config == null || config.menuMouse == null ? 
             MouseEvent.BUTTON3 : config.menuMouse.intValue(); 
   }

   /**
    * Get the mouse actions on which this widget can respond.
    * 
    * @return   The {@link #mouseActions()}.
    */
   @Override
   public Set<Integer> getMouseActions()
   {
      if (mouseActions == null)
      {
         int[] ma = mouseActions();
         if (ma == null)
         {
            mouseActions = Collections.EMPTY_SET;
         }
         else
         {
            Set<Integer> set = new HashSet<Integer>(Utils.primitiveToIntegerList(ma));
            mouseActions = Collections.unmodifiableSet(set);
         }
      }

      return mouseActions;
   }
   
   /**
    * If the method returns <code>true</code> the widget should be represented with 3D visuals.
    * <p>
    * This implementation forwards the call to {@link #isThreeD()} of the parent frame. If
    * no parent frame exists for this widget the method returns <code>false</code>.   
    * 
    * @return   see above.
    */
   @Override
   public boolean isThreeD()
   {
      Widget<?> frame = UiUtils.locateFrame(this);
      return frame != null && frame.isThreeD();
   }
   
   
   /**
    * Returns a string representation of the object.
    * 
    * @return   A string representation of the object.
    */
   @Override
   public String toString()
   {
      String widgetType = getClass().getSimpleName();
      return widgetType + " [id=" + Utils.nonNull(getId(), "<null>") + "]";
   }

   /**
    * Set the custom mouse pointer
    * 
    * @param ptr
    *        the custom mouse pointer
    */
   @Override
   public void setMousePointer(MousePtrWrapper ptr)
   {
      mousePtr = Optional.ofNullable(ptr);
      ThinClient.propagate(topLevelWindow().getId().asInt(), getId().asInt(), ptr);
      ThinClient.getInstance().refreshHover();
   }
   
   /**
    * Get the Loaded custom mouse pointer if it is set
    * 
    * @return the loaded custom mouse pointer if it is set  
    *        
    */
   @Override
   public Optional<MousePtrWrapper> getLoadedMousePointer()
   {
      return mousePtr; 
   }

   /**
    * Get the current custom mouse pointer if it is set
    * 
    * @return the current custom mouse pointer if it is set  
    *        
    */
   @Override
   public Optional<MousePtrWrapper> getCustomMousePointer()
   {
      // TODO: override for Browse & BrowseColumn
      return mousePtr.isPresent() ? mousePtr : 
         enclosingFrame().map(Frame::getCustomMousePointer).orElse(Optional.empty());
   }

   /**
    * Get the current effective custom mouse pointer
    * 
    * @return the current effective custom mouse pointer  
    *        
    */
   @Override
   public MousePtrWrapper getEffectiveMousePointer()
   {
      return getCustomMousePointer().orElse(
                  defaultMousePtr().orElse(MousePtrWrapper.DEFAULT)
             );
   }

   /**
    * Get the widget's actual bounds
    * 
    * @return widget's actual bounds
    */
   @Override
   public NativeRectangle getActualBounds()
   {
      NativeDimension dim = physicalDimension();
      return new NativeRectangle(screenPhysicalLocation(), dim);
   }
   
   /**
    * Check if this widget gains ENTRY event when right/middle mouse button is clicked.
    *
    * @return  <code>true</code> if this widget gains ENTRY event
    *          when right/middle mouse button is clicked.
    */
   public boolean isRMBEntrySupported()
   {
      return true;
   }

   /**
    * Clear widget state on CLEAR FRAME call for the parent frame.
    *
    * @return <code>true</code> on success. <code>false</code> if CLEAR cannot be applied to this
    *         widget, e.g. if it is static text.
    */
   @Override
   public boolean clearWidget()
   {
      if (!UiUtils.hasConfig(this))
      {
         return false;
      }

      ControlTextConfig cc = UiUtils.getConfig(this, ControlTextConfig.class);

      if (cc != null && cc.flagStatic)
      {
         return false;
      }

      setState(ScreenBuffer.UNINITIALIZED);

      // Future changes to re-enable the FILL-IN and EDITOR specific code should be
      // implemented in those specific widgets!

      // TODO (this text was copied from Frame.clearComponents)
      // Probably this is more  correct than handling of UNINITIALIZED in FillIn.draw(), but this
      // may require much more testing and regression prone.
      /*
        if (comp instanceof FillIn)
        ((FillIn) comp).setValue(chr);

        if (comp instanceof Editor)
        ((Editor) comp).clear();
      */

      return true;
   }
   
   /**
    * Sub-classes will override this to return a list of mouse actions on which the widget 
    * responds.  This is compatible with the AWT {@link MouseEvent} codes and P2J specific
    * {@link MouseEvt} actions.
    * <p>
    * If <code>null</code>, it means the widget does not respond to mouse events.
    * 
    * @return   See above.
    */
   protected int[] mouseActions()
   {
      return null;
   }
   
   /**
    * Trick, that lets to correct parent location in cases, when it needed,
    * for example for menus.
    * 
    * @param    parent
    *           Parent component.
    *           
    * @return   corrected parent location.
    */
   protected Point prepParentLocation(Widget<?> parent)
   {
      return parent.location();
   }

   /**
    * Check if the widget supports LOAD-MOUSE-POINTER() method 
    * 
    * @return <code>true</code> if the widget supports 
    *         LOAD-MOUSE-POINTER() method
    */
   @Override
   public boolean supportsCustomMousePointer()
   {
      return false;
   }

   /**
    * Get the enclosing FRAME if it exists
    * 
    * @return the enclosing FRAME if it exists  
    *        
    */
   @Override
   public  Optional<Frame<?>> enclosingFrame()
   {
      Widget<?> widget = (this instanceof Frame) ? parent() : this;
      return Optional.ofNullable((Frame<?>)UiUtils.locateFrame(widget));
   }

   /**
    * Unrealize this widget.
    */
   @Override
   public void unrealize()
   {
      WidgetConfig cfg = config();
      
      if (cfg == null)
      {
         return;
      }
      
      cfg.realized = false;
      cfg.wasRealized = false;
      cfg.visible = false;
      cfg.hidden = false;
      
      this.visible = false;
      this.hidden = false;
   }

   /**
    * The method shows the popup menu for this widget. The menu is resolved with the help of
    * {@link #resolvePopupMenuId()}.
    *
    * @param   e
    *          The related mouse event that caused a popup to be shown.
    * @param   dismissHandler
    *          When not null, the runnable instance is executed when the popup menu is dismissed, regardless
    *          whether an item is chosen or not.
    *
    * @return  Returns the popup menu shown or {@code null} if no popup is resolved for this widget.
    */
   protected Menu<O> showPopupMenu(MouseEvent e, Runnable dismissHandler)
   {
      // show popup menu if assigned.
      int menuId = resolvePopupMenuId();
      if (menuId != -1)
      {
         Menu<O> menu = (Menu<O>) screen().getRegistry().getComponent(menuId);
         if (menu != null)
         {
            menu.showPopup(this, new NativePoint(e.getX(), e.getY()), dismissHandler);
         }
         return menu;
      }

      return null;
   }

   /**
    * Compute the width of the given text, using the specified font in the widget's window.
    * If the <code>useNative</code> flag is set, then the width is computed directly from the
    * driver.  Otherwise, {@link FontManager#getTextWidth} is used to determine the width:
    * the custom text metrics are interrogated first via and, if not found, it fallbacks to using
    * the driver to determine the width.
    *
    * @param    txt
    *           The text.
    * @param    font
    *           Font details.
    * @param    useNative
    *           Flag indicating that the {@link GuiDriver#getTextWidth} needs to be used.
    *
    * @return   The text width.
    */
   private int getTextWidth(String txt, FontDetails font, boolean useNative)
   {
      return getTextWidth(txt, font, -1, useNative);
   }

   /**
    * Compute the width of the given text, using the specified font in the widget's window.
    * If the <code>useNative</code> flag is set, then the width is computed directly from the
    * driver.  Otherwise, {@link FontManager#getTextWidth} is used to determine the width:
    * the custom text metrics are interrogated first via and, if not found, it fallbacks to using
    * the driver to determine the width.
    *
    * @param    txt
    *           The text.
    * @param    font
    *           The font index in the font table.
    * @param    useNative
    *           Flag indicating that the {@link GuiDriver#getTextWidth} needs to be used.
    *
    * @return   The text width.
    */
   private int getTextWidth(String txt, int font, boolean useNative)
   {
      return getTextWidth(txt, null, font, useNative);
   }

   /**
    * Compute the width of the given text, using the specified font in the widget's window.
    * If the <code>useNative</code> flag is set, then the width is computed directly from the 
    * driver.  Otherwise, {@link FontManager#getTextWidth} is used to determine the width:
    * the custom text metrics are interrogated first via and, if not found, it fallbacks to using 
    * the driver to determine the width.
    *
    * @param    txt
    *           The text.
    * @param    font
    *           The font details. Leave it <code>null</code> to specify the font by
    *           <code>fontNum</code>.
    * @param    fontNum
    *           Font number. If <code>font</code> is not <code>null</code>, it is ignored.
    * @param    useNative
    *           Flag indicating that the {@link GuiDriver#getTextWidth} needs to be used.
    *           
    * @return   The text width.
    */
   private int getTextWidth(String txt, FontDetails font, int fontNum, boolean useNative)
   {
      TitledWindow<O> topWindow = ancestor();
      Window<O> window = null;
      if (topWindow == null)
      {
         // check if it's frame-id is set - if so, use that
         if (config().frameId != -1)
         {
            Widget<?> parent = ThinClient.getInstance().getWidget(config().frameId);
            topWindow = window = (Window<O>) parent.window();
         }
      }
      else if (topWindow instanceof TopLevelWindow)
      {
         window = topWindow.window();
      }
      else
      {
         window = window();
      }
      
      if (useNative)
      {
         GuiDriver gd = (GuiDriver) screen().getInstanceDriver();
         gd.selectWindow(UiUtils.getWidgetIdAsInt(topWindow));
         try
         {
            if (font != null)
            {
               return gd.getTextWidth(txt, font);
            }
            else
            {
               FontDetails fd = FontManager.resolveFont(window, fontNum, gd);
               return gd.getTextWidth(txt, fd);
            }
         }
         finally
         {
            gd.releaseWindow();
         }
      }
      else
      {
         ScreenDriver sd = screen().getInstanceDriver();
         if (font != null)
         {
            return FontManager.getTextWidth(txt, window, font, sd);
         }
         else
         {
            return FontManager.getTextWidth(txt, window, fontNum, sd);
         }
      }
   }

   /**
    * Compute the height of the given text, using the specified font in the widget's window.
    * If the <code>useNative</code> flag is set, then the height is computed directly from the 
    * driver.  Otherwise, {@link FontManager#getTextHeight} is used to determine the height:
    * the custom text metrics are interrogated first via and, if not found, it fallbacks to using 
    * the driver to determine the height.
    * 
    * @param    txt
    *           The text.
    * @param    font
    *           The font index in the font table.
    * @param    useNative
    *           Flag indicating that the {@link GuiDriver#getTextHeight} needs to be used.
    *           
    * @return   The text height.
    */
   private int getTextHeight(String txt, int font, boolean useNative)
   {
      TitledWindow<O> topWindow = ancestor();
      Window<O> window = null;
      if (topWindow == null)
      {
         // check if it's frame-id is set - if so, use that
         if (config().frameId != -1)
         {
            Widget<?> parent = ThinClient.getInstance().getWidget(config().frameId);
            topWindow = window = (Window<O>) parent.window();
         }
      }
      else if (topWindow instanceof TopLevelWindow)
      {
         window = topWindow.window();
      }
      else
      {
         window = window();
      }
      
      if (useNative)
      {
         GuiDriver gd = (GuiDriver) screen().getInstanceDriver();
         gd.selectWindow(UiUtils.getWidgetIdAsInt(topWindow));
         FontDetails fd = FontManager.resolveFont(window, font, gd);
         try
         {
            return gd.getTextHeight(txt, fd);
         }
         finally
         {
            gd.releaseWindow();
         }
      }
      else
      {
         return FontManager.getTextHeight(txt, window, font, screen().getInstanceDriver());
      }
   }
} 
