/*
** Module   : DefaultList.java
** Abstract : Default scrollable item list implementation.
**
** Copyright (c) 2010-2017, Golden Code Development Corporation.
**
** -#- -I- --Date-- ---------------------------------Description----------------------------------
** 001 SIY 20100629 Created initial version
** 002 SIY 20100804 Changes in imports, some refactorings.
** 003 SIY 20100822 Added javadoc.
** 004 SIY 20101122 Switching from KeyEvent to KeyInput.
** 003 CA  20140805 Added generics to force the abstract widget classes to not use implementation-
**                  specific code: the abstract widgets will have access only to the base APIs
**                  provided by the OutputManager.
**                  Moved ChUI-specific requestFocus impl to the ChUI impl classes.
** 004 HC  20140911 GUI coordinate support - data types refactoring.
** 005 MAG 20140929 Added ensureRowVisible with scrolling option.
** 006 HC  20141021 Implemented window move/position support.
** 007 IAS 20150202 Added index() method.
** 008 HC  20150831 Scrolling support overhaul.
** 009 GES 20150923 Javadoc fix.
** 010 VIG 20160420 Added isScrollPopupSupported implementation.
** 011 SVL 20160412 Javadoc fix.
*/
/*
** This program is free software: you can redistribute it and/or modify
** it under the terms of the GNU Affero General Public License as
** published by the Free Software Foundation, either version 3 of the
** License, or (at your option) any later version.
**
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
** GNU Affero General Public License for more details.
**
** You may find a copy of the GNU Affero GPL version 3 at the following
** location: https://www.gnu.org/licenses/agpl-3.0.en.html
** 
** Additional terms under GNU Affero GPL version 3 section 7:
** 
**   Under Section 7 of the GNU Affero GPL version 3, the following additional
**   terms apply to the works covered under the License.  These additional terms
**   are non-permissive additional terms allowed under Section 7 of the GNU
**   Affero GPL version 3 and may not be removed by you.
** 
**   0. Attribution Requirement.
** 
**     You must preserve all legal notices or author attributions in the covered
**     work or Appropriate Legal Notices displayed by works containing the covered
**     work.  You may not remove from the covered work any author or developer
**     credit already included within the covered work.
** 
**   1. No License To Use Trademarks.
** 
**     This license does not grant any license or rights to use the trademarks
**     Golden Code, FWD, any Golden Code or FWD logo, or any other trademarks
**     of Golden Code Development Corporation. You are not authorized to use the
**     name Golden Code, FWD, or the names of any author or contributor, for
**     publicity purposes without written authorization.
** 
**   2. No Misrepresentation of Affiliation.
** 
**     You may not represent yourself as Golden Code Development Corporation or FWD.
** 
**     You may not represent yourself for publicity purposes as associated with
**     Golden Code Development Corporation, FWD, or any author or contributor to
**     the covered work, without written authorization.
** 
**   3. No Misrepresentation of Source or Origin.
** 
**     You may not represent the covered work as solely your work.  All modified
**     versions of the covered work must be marked in a reasonable way to make it
**     clear that the modified work is not originating from Golden Code Development
**     Corporation or FWD.  All modified versions must contain the notices of
**     attribution required in this license.
*/

package com.goldencode.p2j.ui.client.widget;

import java.util.*;

import com.goldencode.p2j.ui.*;
import com.goldencode.p2j.ui.client.*;
import com.goldencode.p2j.ui.client.event.*;
import com.goldencode.p2j.ui.client.event.listener.*;
import com.goldencode.p2j.ui.client.model.*;

/**
 * Base class for scrollable list implementations.
 * 
 * @param <T> 
 */
public abstract class DefaultList<T, O extends OutputManager<?>>
extends AbstractWidget<O>
implements ScrollableWidget<O>,
           ScrollPublisher
{
   /** Enable item deselection flag. */
   private boolean allowDeselect = true;

   /** Number of columns in the list. */
   private double columns = 10;

   /** Current row in the list. */
   private int currentRow = 0;

   /** List model. */
   private ListModel<T> listModel;
   
   /** Storage for listeners. */
   private ArrayList<ScrollListener> scrollListeners = new ArrayList<ScrollListener>();

   /** List selection model. */
   private ListSelectionModel selectionModel = new ListSelectionModel();

   /** Number of visible rows in the list. */
   private int visibleRows = 5;

   /**
    * Constructor.
    * 
    * @param   model
    *          List model.
    */
   public DefaultList(ListModel<T> model)
   {
      listModel = model;
   }

   /**
    * Add scroll listener.
    *  
    * @param   listener
    *          Listener to add.
    */
   @Override
   public void addScrollListener(ScrollListener listener)
   {
      scrollListeners.add(listener);
   }
   
   /**
    * Add selection listener.
    *  
    * @param   listener
    *          Listener to add.
    */
   public void addSelectionListener(SelectionListener listener)
   {
      selectionModel.add(listener);
   }

   /**
    * Control removing of selection from current item.
    * 
    * @param    allow
    *           <code>true</code> allow deselection of current item.
    */
   public void allowDeselect(boolean allow)
   {
      allowDeselect = allow;
   }

   /**
    * Clear list.
    */
   public void clear()
   {
      selectionModel.clear();
   }

   /**
    * Get current row in the list.
    * 
    * @return  current row.
    */
   public int currentRow()
   {
      return currentRow;
   }

   /**
    * Deselect specified range of items.
    * 
    * @param   from
    *          Start index in the range.
    * @param   to
    *          End index in the range.
    */
   public void deselect(int from, int to)
   {
      selectionModel.deselect(from, to);
   }

   /**
    * Get widget dimension.
    * 
    * @return  widget dimension.
    */
   @Override
   public Dimension dimension()
   {
      return new Dimension(width(), height());
   }

   /**
    * Ensure that specified row is visible. Scroll list if necessary.
    * 
    * @param   row
    *          The row to process.
    */
   public void ensureRowVisible(int row)
   {
      ensureRowVisible(row, true);
   }

   /**
    * Ensure that specified row is visible. Scroll list if necessary.
    * 
    * @param   row
    *          The row to process.
    * @param   scroll
    *          Scrolling option.
    */
   public void ensureRowVisible(int row, boolean scroll)
   {
      if (!(parent() instanceof Viewport))
         return;
      
      int newRow = fixRow(row);
      int delta = newRow - currentRow;
      
      if (delta != 0)
      {
         shiftCurrentRow(delta, scroll);
      }
   }

   /**
    * Get first selected row index.
    * 
    * @return  first selected row index.
    */
   public int first()
   {
      return selectionModel.first();
   }
   
   /**
    * Get widget height.
    * 
    * @return  widget height.
    */
   @Override
   public double height()
   {
      return listModel.size();
   }

   /**
    * Get widget minimum size.
    * 
    * @return  widget minimum size.
    */
   @Override
   public Dimension minimumSize()
   {
      columns = 1;

      for (int i = 0; i < listModel.size(); i++)
      {
         String c = listModel.get(i).toString();
         
         if (c.length() > columns)
            columns = c.length();
      }

      return new Dimension(columns, visibleRows);
   }

   /**
    * Get underlying list model.
    * 
    * @return  list model.
    */
   public ListModel<T> model()
   {
      return listModel;
   }

   /**
    * Process key event.
    * 
    * @param   event
    *          Key event to process.
    */
   @Override
   public void processKeyEvent(KeyInput event)
   {
      super.processKeyEvent(event);

      if (event.isConsumed())
         return;

      int key = event.keyCode();

      if (focusChange(key))
         return;
      
      processKey(key);

      redraw();
   }

   /**
    * Process key event.
    * 
    * @param key
    *          key to process
    */
   protected void processKey(int key)
   {
      switch (key)
      {
         case Key.VK_CURSOR_DOWN :
            shiftCurrentRow(1, true);
            break;

         case Key.VK_PAGE_DOWN :
            shiftCurrentRow(visibleRows, true);
            break;

         case Key.VK_END :
            shiftCurrentRow(listModel.size(), true);
            break;

         case Key.VK_CURSOR_UP :
            shiftCurrentRow(-1, true);
            break;

         case Key.VK_PAGE_UP :
            shiftCurrentRow(-visibleRows, true);
            break;

         case Key.VK_HOME :
            shiftCurrentRow(-listModel.size(), true);
            break;
            
         case Key.VK_ENTER :
            selectItem();
            break;
      }
   }

   /**
    * Select row with specified index.
    * 
    * @param   index
    *          Item index.
    */
   public void select(int index)
   {
      selectionModel.set(index, index);
   }

   /**
    * Select specified range of items.
    * 
    * @param   from
    *          Start index in the range.
    * @param   to
    *          End index in the range.
    */
   public void select(int from, int to)
   {
      selectionModel.select(from, to);
   }

   /**
    * Check if item with specified index is selected.
    *  
    * @param   index
    *          Item index.
    *          
    * @return  <code>true</code> if item is selected.
    */
   public boolean selected(int index)
   {
      return selectionModel.selected(index);
   }

   /**
    * Get indices of all selected items.
    * 
    * @return  array of selected indices.
    */
   public int[] selection()
   {
      return selectionModel.selection();
   }

   /**
    * Get access to selection model.
    * 
    * @return  selection model.
    */
   public int selectionMode()
   {
      return selectionModel.mode();
   }

   public double getColumns()
   {
      return columns;
   }
   
   /**
    * Set visible list width.
    * 
    * @param   cols
    *          Width in columns.
    */
   public void setColumns(double cols)
   {
      columns = Coordinate.scale(cols);
   }
   
   /**
    * Set list selection mode.
    * 
    * @param   mode
    *          Selection mode to set.
    */
   public void setSelectionMode(int mode)
   {
      selectionModel.setMode(mode);
   }

   /**
    * Set number of visible rows.
    * 
    * @param   rows
    *          Number of rows to set.
    */
   public void setVisibleRows(int rows)
   {
      visibleRows = rows;
   }

   /**
    * Get value of first selected item.
    * 
    * @return  value of first selected item.
    */
   public T value()
   {
      int index = selectionModel.first();

      if (index == -1)
         return null;

      return listModel.get(index);
   }

   /**
    * Get index of first selected item.
    * 
    * @return index of first selected item.
    */
   public int index()
   {
      return selectionModel.first();
   }

   /**
    * Get number of visible rows.
    * 
    * @return  number of visible rows.
    */
   public int visibleRows()
   {
      return visibleRows;
   }

   /**
    * Get widget width.
    * 
    * @return  widget width.
    */
   @Override
   public double width()
   {
      return columns;
   }

   /**
    * Returns the scroll dimension of this widget. The scroll dimension is
    * the full size of the widget's content, it may be bigger than widget's
    * visible size.
    * <p>
    * The returned dimension doesn't need to be in the same units as the 
    * widget's physical size but it must be compatible with the result
    * of {@link #getVisibleDimension()} and {@link #getScrollStep()}.
    * 
    * @return   widget's scroll dimension
    */
   @Override
   public NativeDimension getScrollDimension()
   {
      return new NativeDimension(0, listModel.size());
   }

   /**
    * Returns the visible dimension of this widget. The visible dimension is
    * the visible portion of the widget's content (including scrollbars area).
    * <p>
    * The returned dimension doesn't need to be in the same units as the 
    * widget's physical size but it must be compatible with the result
    * of {@link #getScrollDimension()} and {@link #getScrollStep()}.
    *
    * @return   widget's visible dimension
    */
   @Override
   public NativeDimension getVisibleDimension()
   {
      return new NativeDimension(0, visibleRows());
   }
   
   /**
    * Returns the dimension of the base scroll step desired by the widget
    * in the X and Y directions.
    * <p>
    * The scroll step must be in the same units as the dimensions returned
    * by {@link #getScrollDimension()} and {@link #getVisibleDimension()}. 
    * 
    * @return   The dimension of the base scroll step.
    */
   @Override
   public NativeDimension getScrollStep()
   {
      return new NativeDimension(1, 1);
   }

   /**
    * Directs the widget to scroll to the specified X and Y positions.
    * It is the widget's responsibility to update its visuals accordingly.
    * Note that the position values are in the same units as the returned 
    * dimensions from {@link #getScrollDimension()}, {@link #getVisibleDimension()}
    * and {@link #getScrollStep()}. 
    * 
    * @param    newXPosition
    *           New X scroll position.
    * @param    newYPosition
    *           New Y scroll position.
    */
   @Override
   public void scroll(Integer newXPosition, Integer newYPosition)
   {
      currentRow = fixRow(newYPosition);      
      repaint();
   }
   
   /**
    * Check if this widget that supports scrolling also supports scroll popup.
    *
    * @return  <code>true</code> if this widget supports scroll popup.
    */
   public boolean isScrollPopupSupported()
   {
      return true;
   }

   /**
    * Shifts current row by specified delta.
    * 
    * @param   delta
    *          the number of list items to shift current row by
    * @param   scroll
    *          indicates whether list should be scrolled
    *          to the shift row
    */
   protected void shiftCurrentRow(int delta, boolean scroll)
   {
      int newCurrentRow = fixRow(currentRow + delta);
      
      int pos = 0;
      
      if (scroll)
      {
         pos = calculateScrollYPosition();
      }
      
      currentRow = newCurrentRow;
         
      if (scroll)
      {
         int newPos = calculateScrollYPosition();
         
         if (pos != newPos)
         {
            postScrollEvent(null, newPos);
         }
      }
      
      repaint();
   }
   
   /**
    * Calculate the scroll Y position based on the current state. 
    * 
    * @return   the new scroll Y position
    */
   protected int calculateScrollYPosition()
   {
      // this implementation makes the current row always the top row 
      return currentRow;
   }

   /**
    * Handle selection/deselection of current item.
    */
   protected void selectItem()
   {
      if (listModel.size() > 0)
      {
         if (selected(currentRow))
         {
            if (allowDeselect)
               deselect(currentRow, currentRow);
            else
               selectionModel.notifyListeners();
         }
         else
         {
            if (selectionMode() == ListSelectionModel.SINGLE_SELECTION)
               select(currentRow);
            else
               select(currentRow, currentRow);
         }
         repaint();
      }
   }
   
   /**
    * Redrawing worker.
    */
   protected void redraw()
   {
      if (parent() instanceof Viewport)
         return;
      
      draw();
      requestFocus();
      requestSync();
   }

   /**
    * Post scroll event with the specified row.
    * 
    * @param   row
    *          Scroll row.
    */
   protected void postScrollEvent(Integer column, Integer row)
   {
      for (ScrollListener l : this.scrollListeners)
      {
         l.handleScrollEvent(new ScrollEvent(this, column, row));
      }
   }

   /**
    * Fix row to make sure it fits into model bounds.
    * 
    * @param   row
    *          Row to fix.
    *          
    * @return  fixed row.
    */
   private int fixRow(int row)
   {
      if (row < 0 || listModel.size() < 1)
      {
         row = 0;
      }
      else if (row > listModel.size() - 1)
      {
         row = listModel.size() - 1;
      }
      return row;
   }

   /**
    * Handle keys which may trigger focus change.
    * 
    * @param   key
    *          Key to check.
    *          
    * @return  <code>true</code> if key was processed.
    */
   private boolean focusChange(int key)
   {
      if (key == Key.VK_TAB)
      {
         parent(Frame.class).getContentPane().nextFocus();
         return true;
      }

      if (key == Key.VK_BACK_TAB)
      {
         parent(Frame.class).getContentPane().prevFocus();
         return true;
      }

       return false;
   }
}