/*
** Module   : ScrollContainer.java
** Abstract : Base class for scrollable containers.
**
** Copyright (c) 2015-2017, Golden Code Development Corporation.
**
** -#- -I- --Date-- ---------------------------------Description---------------------------------
** 001 HC  20150831 Created initial version.
** 002 GES 20150923 Javadoc fix.
** 003 VIG 20160420 Added isScrollPopupSupported implementation.
*/
/*
** This program is free software: you can redistribute it and/or modify
** it under the terms of the GNU Affero General Public License as
** published by the Free Software Foundation, either version 3 of the
** License, or (at your option) any later version.
**
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
** GNU Affero General Public License for more details.
**
** You may find a copy of the GNU Affero GPL version 3 at the following
** location: https://www.gnu.org/licenses/agpl-3.0.en.html
** 
** Additional terms under GNU Affero GPL version 3 section 7:
** 
**   Under Section 7 of the GNU Affero GPL version 3, the following additional
**   terms apply to the works covered under the License.  These additional terms
**   are non-permissive additional terms allowed under Section 7 of the GNU
**   Affero GPL version 3 and may not be removed by you.
** 
**   0. Attribution Requirement.
** 
**     You must preserve all legal notices or author attributions in the covered
**     work or Appropriate Legal Notices displayed by works containing the covered
**     work.  You may not remove from the covered work any author or developer
**     credit already included within the covered work.
** 
**   1. No License To Use Trademarks.
** 
**     This license does not grant any license or rights to use the trademarks
**     Golden Code, FWD, any Golden Code or FWD logo, or any other trademarks
**     of Golden Code Development Corporation. You are not authorized to use the
**     name Golden Code, FWD, or the names of any author or contributor, for
**     publicity purposes without written authorization.
** 
**   2. No Misrepresentation of Affiliation.
** 
**     You may not represent yourself as Golden Code Development Corporation or FWD.
** 
**     You may not represent yourself for publicity purposes as associated with
**     Golden Code Development Corporation, FWD, or any author or contributor to
**     the covered work, without written authorization.
** 
**   3. No Misrepresentation of Source or Origin.
** 
**     You may not represent the covered work as solely your work.  All modified
**     versions of the covered work must be marked in a reasonable way to make it
**     clear that the modified work is not originating from Golden Code Development
**     Corporation or FWD.  All modified versions must contain the notices of
**     attribution required in this license.
*/

package com.goldencode.p2j.ui.client.widget;

import java.util.*;

import com.goldencode.p2j.ui.*;
import com.goldencode.p2j.ui.client.*;
import com.goldencode.p2j.ui.client.event.*;
import com.goldencode.p2j.ui.client.event.listener.*;

/**
 * Base class for scrollable containers.
 */
public abstract class ScrollContainer<O extends OutputManager<?>>
extends AbstractContainer<O>
implements WidgetWithConfig<BaseConfig>, 
           ScrollableWidget<O>,
           ScrollPublisher
{   
   /** The widget config. */
   protected BaseConfig config;
   
   /** Storage for listeners. */
   private ArrayList<ScrollListener> scrollListeners = new ArrayList<ScrollListener>();

   /**
    * Constructor.
    */
   public ScrollContainer()
   {      
      // do not register with config manager, no server synchronization required,
      config = new BaseConfig(); 
               
   }
   
   /**
    * Replacement API for the logic which should exist in the widget's c'tor, but the logic 
    * can't be executed as it depends on the widget being registered into the official registry.
    * <p>
    * This API is called after the widget is created and added to the registry.
    * 
    * @param    id
    *           The widget's ID.
    * @param    cfg
    *           The config used to initialize this widget.
    */
   @Override
   public void initialize(WidgetId id, BaseConfig cfg)
   {
      // no-op
   }

   /**
    * Get container configuration.
    * 
    * @return  see above. 
    */
   @SuppressWarnings("unchecked")
   public BaseConfig config()
   {
      return config;
   }
   
   /**
    * The method is called after the configuration associated with 
    * the implementor changes. The parameter points to a reference
    * representing the original configuration state. This reference
    * can be used to detect what configuration fields changed and so
    * optimize any processing related to the configuration change.
    * 
    * @param    beforeUpdate
    */
   @Override
   public void afterConfigUpdate(BaseConfig beforeUpdate)
   {
      super.afterConfigUpdateBase(beforeUpdate);
   }
   
   /**
    * Returns the viewport reference.
    * 
    * @return   see above
    */
   public Viewport<O> getViewport()
   {
      @SuppressWarnings("unchecked")
      Viewport<O> result = parent(Viewport.class);
      
      return result;
   }

   /**
    * Returns the dimension of the base scroll step desired by the widget
    * in the X and Y directions.
    * <p>
    * The scroll step must be in the same units as the dimensions returned
    * by {@link #getScrollDimension()} and {@link #getVisibleDimension()}. 
    * 
    * @return   The dimension of the base scroll step.
    */
   @Override
   public NativeDimension getScrollStep()
   {
      // set scroll step to 1 char by default
      return screen().coordinates().dimensionToNative(new Dimension(1, 1));
   }
   
   /**
    * Add scroll listener.
    *  
    * @param   listener
    *          Listener to add.
    */
   @Override
   public void addScrollListener(ScrollListener listener)
   {
      scrollListeners.add(listener);
   }
   
   /**
    * Check if this widget that supports scrolling also supports scroll popup.
    *
    * @return  <code>true</code> if this widget supports scroll popup.
    */
   public boolean isScrollPopupSupported()
   {
      return true;
   }

   /**
    * Post scroll event with the specified scroll positions.
    * 
    * @param   scrollX
    *          New X position.
    * @param   scrollY
    *          New Y position.
    */
   protected void postScrollEvent(Integer scrollX, Integer scrollY)
   {
      for (ScrollListener l : this.scrollListeners)
      {
         l.handleScrollEvent(new ScrollEvent(this, scrollX, scrollY));
      }
   }
}