/*
** Module   : AsyncRequestImpl.java
** Abstract : implementation of the async request resource.
**
** Copyright (c) 2013-2017, Golden Code Development Corporation.
**
** -#- -I- --Date-- ---------------------------------Description----------------------------------
** 001 CA  20130813 Created initial version.
** 002 CA  20130927 Resource type is computed from the annotation. Removed setUnknown(false) call.
** 003 VIG 20131126 Added LAST-EVENT:LABEL support for PROCEDURE-COMPLETE event
** 004 CA  20131220 Changed to support both appserver and web service async requests.
** 005 GES 20140116 Fixed some member references since the resource moved between classes.
** 006 OM  20160712 Fixed synchronization issue.
*/
/*
** This program is free software: you can redistribute it and/or modify
** it under the terms of the GNU Affero General Public License as
** published by the Free Software Foundation, either version 3 of the
** License, or (at your option) any later version.
**
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
** GNU Affero General Public License for more details.
**
** You may find a copy of the GNU Affero GPL version 3 at the following
** location: https://www.gnu.org/licenses/agpl-3.0.en.html
** 
** Additional terms under GNU Affero GPL version 3 section 7:
** 
**   Under Section 7 of the GNU Affero GPL version 3, the following additional
**   terms apply to the works covered under the License.  These additional terms
**   are non-permissive additional terms allowed under Section 7 of the GNU
**   Affero GPL version 3 and may not be removed by you.
** 
**   0. Attribution Requirement.
** 
**     You must preserve all legal notices or author attributions in the covered
**     work or Appropriate Legal Notices displayed by works containing the covered
**     work.  You may not remove from the covered work any author or developer
**     credit already included within the covered work.
** 
**   1. No License To Use Trademarks.
** 
**     This license does not grant any license or rights to use the trademarks
**     Golden Code, FWD, any Golden Code or FWD logo, or any other trademarks
**     of Golden Code Development Corporation. You are not authorized to use the
**     name Golden Code, FWD, or the names of any author or contributor, for
**     publicity purposes without written authorization.
** 
**   2. No Misrepresentation of Affiliation.
** 
**     You may not represent yourself as Golden Code Development Corporation or FWD.
** 
**     You may not represent yourself for publicity purposes as associated with
**     Golden Code Development Corporation, FWD, or any author or contributor to
**     the covered work, without written authorization.
** 
**   3. No Misrepresentation of Source or Origin.
** 
**     You may not represent the covered work as solely your work.  All modified
**     versions of the covered work must be marked in a reasonable way to make it
**     clear that the modified work is not originating from Golden Code Development
**     Corporation or FWD.  All modified versions must contain the notices of
**     attribution required in this license.
*/

package com.goldencode.p2j.util;

import java.util.*;

import com.goldencode.p2j.*;
import com.goldencode.p2j.main.*;
import com.goldencode.p2j.ui.*;
import com.goldencode.p2j.ui.client.event.*;
import com.goldencode.util.*;

/**
 * Implements the async request resource. Beside adding support for its attributes, the async
 * resource is also responsible of initiating the async call and raising the PROCEDURE-COMPLETE
 * event.
 * <p>
 * As some the mutable fields of this class can be modified by more than one thread, they are
 * marked as {@code volatile}.
 */
public class AsyncRequestImpl
extends HandleChain
implements AsyncRequest,
           Runnable
{
   /** A handle to the SERVER resource. */
   private final handle server = new handle();
   
   /** The name of the EVENT-PROCEDURE. */
   private final character eventProcedure = new character();
   
   /** The context of the EVENT-PROCEDURE. */
   private final handle eventProcedureContext = new handle();
   
   /** The proxy procedure, if the remote call is for an internal procedure. */
   private final handle persistentProcedure = new handle();
   
   /** The name of the target procedure. */
   private final character procedureName = new character();
   
   /** The procedure's arguments. */
   private final Object[] args;
   
   /** The modes for the procedure's arguments. */
   private String modes;
   
   /** The result from the remote call. */
   private RemoteInvocationResult result = null;
   
   /**
    * Flag indicating that the async call has completed (will be {@code false} if the call was 
    * cancelled.
    */
   private volatile boolean complete = false;
   
   /** Flag indicating that the async call has been cancelled. */
   private volatile boolean cancelled = false;
   
   /** Flag indicating that the call was terminated with a QUIT condition. */
   private volatile boolean quit = false;
   
   /** Flag indicating that the call was terminated with a STOP condition. */
   private volatile boolean stop = false;
   
   /** Flag indicating that the call was terminated with an ERROR condition. */
   private volatile boolean error = false;
   
   /** Flag indicating if the resource is valid. */
   private volatile boolean valid = true;
   
   /** The ID of the remote request being executed. */
   private volatile int requestId = -1;
   
   /**
    * List of listeners which need to receive notifications when the call is started or finished.
    */
   private List<AsyncRequestListener> listeners = new ArrayList<>();
   
   /** A worker indicating what will be invoked. */
   private Runnable request;
   
   /**
    * Create a new async request resource, by specifying the attributes related to the async
    * call definition.
    * 
    * @param    server
    *           The server used to invoke the remote call. May be a {@code SERVER} or the
    *           {@code SESSION} resource.
    * @param    request
    *           The request which will execute the procedure call.
    * @param    eventProcedure
    *           The event procedure name.
    * @param    eventProcedureContext
    *           The context of the event procedure.
    * @param    procedureName
    *           The name of the procedure to be invoked.
    * @param    persistentProcedure
    *           The handle to the proxy procedure to which the target procedure belongs. Will
    *           be unknown if this is not a remote call.
    * @param    modes
    *           The modes for the procedure's arguments.
    * @param    args
    *           The procedure's arguments.
    */
   public AsyncRequestImpl(handle    server,
                           Runnable  request,
                           character eventProcedure,
                           handle    eventProcedureContext,
                           character procedureName,
                           handle    persistentProcedure,
                           String    modes,
                           Object... args)
   {
      this.server.assign(server);
      this.request = request;
      this.eventProcedure.assign(eventProcedure);
      this.eventProcedureContext.assign(eventProcedureContext);
      this.persistentProcedure.assign(persistentProcedure);
      this.procedureName.assign(procedureName);
      this.modes = modes;
      // arguments need special processing; copies are made of all BDT instances; TABLE
      // parameters remain intact
      this.args = new Object[args.length];
      for (int i = 0; i < args.length; i++)
      {
         if (args[i] instanceof BaseDataType)
         {
            this.args[i] = ((BaseDataType) args[i]).duplicate();
         }
      }
   }
   
   /**
    * Get the parameter modes for this async request.
    * 
    * @return   See above.
    */
   public String getModes()
   {
      return this.modes;
   }
   
   /**
    * Get the arguments for this async requests.
    * 
    * @return   See above.
    */
   public Object[] getArgs()
   {
      return this.args;
   }
   
   /**
    * Returns the state of the ERROR attribute
    *
    * @return    Current state of the ERROR attribute.
    */
   public logical isError()
   {
      return !complete ? new logical() : new logical(error);
   }

   /**
    * Setter for ERROR attribute.
    *
    * @param    value
    *           The new value of ERROR attribute.
    */
   public void setError(boolean value)
   {
      readOnlyError("ERROR");
   }

   /**
    * Setter for ERROR attribute.
    *
    * @param    value
    *           The new value of ERROR attribute.
    */
   public void setError(logical value)
   {
      readOnlyError("ERROR");
   }

   /**
    * Reports if this object is valid for use (has not been deleted).
    *
    * @return   {@code true} if we are valid (can be used).
    */
   @Override
   public boolean valid()
   {
      return valid;
   }

   /**
    * Returns a handle to the app server in which the procedure is ran.
    * 
    * @return   See above
    */
   @Override
   public handle getServerHandle()
   {
      return new handle(server);
   }

   /**
    * Get the {@code CANCELLED} attribute of this async request.
    * 
    * @return   See above.
    */
   @Override
   public logical isCancelled()
   {
      return new logical(cancelled);
   }

   /**
    * Get the {@code CANCELLED} attribute of this async request, as a Java boolean.
    * 
    * @return   See above.
    */
   public boolean _isCancelled()
   {
      return cancelled;
   }

   /**
    * Get the {@code COMPLETE} attribute of this async request.
    * 
    * @return   See above.
    */
   @Override
   public logical isComplete()
   {
      return new logical(complete);
   }

   /**
    * Get the {@code COMPLETE} attribute of this async request, as a Java boolean.
    * 
    * @return   See above.
    */
   public boolean _isComplete()
   {
      return complete;
   }

   /**
    * Get the {@code EVENT-PROCEDURE} attribute of this async request.
    * 
    * @return   See above.
    */
   @Override
   public character getEventProcedure()
   {
      return new character(eventProcedure);
   }

   /**
    * Get the {@code EVENT-PROCEDURE-CONTEXT} attribute of this async request.
    * 
    * @return   See above.
    */
   @Override
   public handle getEventProcedureContext()
   {
      return new handle(eventProcedureContext);
   }

   /**
    * Get the {@code PERSISTENT-PROCEDURE} attribute of this async request.
    * 
    * @return   See above.
    */
   @Override
   public handle getPersistentProcedure()
   {
      return new handle(persistentProcedure);
   }

   /**
    * Get the {@code PROCEDURE-NAME} attribute of this async request.
    * 
    * @return   See above.
    */
   @Override
   public character getProcedureName()
   {
      return new character(procedureName);
   }

   /**
    * Get the {@code QUIT} attribute of this async request.
    * 
    * @return   See above.
    */
   @Override
   public logical isQuit()
   {
      return new logical(quit);
   }

   /**
    * Get the {@code STOP} attribute of this async request.
    * 
    * @return   See above.
    */
   @Override
   public logical isStop()
   {
      return new logical(stop);
   }

   /**
    * Invoke the event procedure for this async request, if used.
    */
   public void run()
   {
      // change the RETURN-VALUE as set by this call
      boolean nullResult = (result == null);
      if (!nullResult && result.setReturnValue())
      {
         Object res = result.getResult();
         ControlFlowOps.setReturnValue(res == null ? null : new character((BaseDataType) res));
      }

      // set the QUIT/STOP/ERROR flags and the details about any raised ERRORs
      Throwable cause = nullResult ? null : result.getError();
      while (cause != null)
      {
         if (cause instanceof QuitConditionException)
         {
            setQuit();
            break;
         }
         else if (cause instanceof StopConditionException)
         {
            setStop();
            break;
         }
         else if (cause instanceof ErrorConditionException)
         {
            setError();
            
            ErrorConditionException err = (ErrorConditionException) cause;
            if (err.getCause() instanceof NumberedException)
            {
               NumberedException ne = (NumberedException) err.getCause();
               ErrorManager.addError(ne.getNumber(), ne.getMessage(), true);
            }

            break;
         }

         cause = cause.getCause();
      }
      
      // no further processing is needed if no event procedure is set
      if (eventProcedure.isUnknown())
      {
         return;
      }

      List<Object> eventArgsList = new ArrayList<>();
      Object[] eventArgs = new Object[0];
      String eventModes = null;
      if (modes == null)
      {
         // get the resolved modes, if not explicitly passed at the call
         modes = nullResult ? null : result.getModes();
      }
      
      if (modes != null)
      {
         boolean failed = error || quit || stop || cancelled;
         
         for (int i = 0; i < modes.length(); i++)
         {
            char mode = modes.charAt(i);
            if (InternalEntry.isOutputMode(mode))
            {
               // if the request did not complete OK, pass an unknown, but only for BDTs
               eventArgsList.add(failed && args[i] instanceof BaseDataType ? new unknown() 
                                                                           : args[i]);
            }
         }
         
         eventModes = StringHelper.repeatChar(InternalEntry.INPUT_MODE, eventArgsList.size());
         eventArgs = eventArgsList.toArray();
      }
   
      // invoke the event procedure
      try
      {
         SelfManager.pushSelf(new handle(this));
         ControlFlowOps.invokeInWithMode(eventProcedure,
                                         eventProcedureContext,
                                         eventModes,
                                         eventArgs);
      }
      finally
      {
         SelfManager.popSelf();
      }
   }

   /**
    * Set the ID of this remote request.
    * 
    * @param    reqId
    *           The ID of the request.
    */
   void setRequestId(int reqId)
   {
      this.requestId = reqId;
   }

   /**
    * Called when the async call has finished normally (not with a cancel).
    */
   void complete()
   {
      this.complete = true;
      notifyListeners(false);
   }
   
   /**
    * Called when the async call was cancelled by the user.
    */
   void cancelled()
   {
      this.cancelled = true;
      notifyListeners(false);
   }

   /**
    * Mark this async call as having ended with a STOP condition.
    */
   void setStop()
   {
      this.stop = true;
   }

   /**
    * Mark this async call as having ended with a QUIT condition.
    */
   void setQuit()
   {
      this.quit = true;
   }

   /**
    * Mark this async call as having ended with a ERROR condition.
    */
   void setError()
   {
      this.error = true;
   }
   
   /**
    * Stop this async request by sending the STOP condition to the remote side. This will be
    * used only in cases when the async call is cancelled by the user.
    */
   void stop()
   {
      if (requestId == -1)
      {
         // the request was not yet started, prevent from starting it
         complete();
         setStop();
         return;
      }

      // send a STOP condition to the remote side
      ServerImpl s = (ServerImpl) server.get();
      s.getServerHelper().sendStop(requestId);

      // wait for it to complete
      while (!complete)
      {
         synchronized (this)
         {
            try
            {
               this.wait();
            }
            catch (InterruptedException e)
            {
               break;
            }
         }
      }
   }

   /**
    * Set the result from the {@link #request} execution.
    * 
    * @param    result
    *           The result.
    */
   void setResult(RemoteInvocationResult result)
   {
      this.result = result;
   }
   
   /**
    * Register a new listener for notifications about the state of the async call.
    * 
    * @param    listener
    *           A listener instance.
    */
   void registerListener(AsyncRequestListener listener)
   {
      listeners.add(listener);
   }
   
   /**
    * Notify all listeners that the async call is about to start.
    */
   void notifyStart()
   {
      notifyListeners(true);
   }

   /**
    * Execute the async request. This will either raise the PROCEDURE-COMPLETE event or will
    * invoke the EVENT-PROCEDURE immediately.
    * 
    * @param    event
    *           Flag indicating if the PROCEDURE-COMPLETE event will be raised.
    */
   void execute(boolean event)
   {
      // check if the request was not cancelled (requests which are removed from queue and are
      // pending to be executed are marked as complete plus the STOP attribute to true
      if (complete)
      {
         return;
      }

      // execute the remote async call
      try
      {
         request.run();
      }
      catch (Throwable t)
      {
         // treat any conditions
         Throwable cause = t;
         while (cause != null && !(cause instanceof ConditionException))
         {
            cause = t.getCause();
         }
         
         if (cause != null)
         {
            if (cause instanceof StopConditionException)
            {
               setStop();
            }
            else if (cause instanceof QuitConditionException)
            {
               setQuit();
            }
            else if (cause instanceof ErrorConditionException)
            {
               setError();
            }
         }
      }
      finally
      {
         if (!_isCancelled())
         {
            complete();
         }

         if (event)
         {
            KeyReader.setLabelWorker(Keyboard.PROCEDURE_COMPLETE);
            
            ServerEvent serverEvent = new ServerEvent(handle.resourceId(this),
                                                      Keyboard.PROCEDURE_COMPLETE,
                                                      this);
            LogicalTerminal.postServerEvent(serverEvent);
         }
         else
         {
            // execute the event now
            run();
         }
         
         // notify that the request has finished.
         synchronized (this)
         {
            this.notify();
         }
      }
   }
   
   /**
    * The async resource can be deleted only if it has been completed or cancelled.
    */
   @Override
   protected boolean resourceDelete()
   {
      if (!(complete || cancelled))
      {
         final String msg = "Asynchronous request  has not completed.  Cannot DELETE";
         ErrorManager.recordOrThrowError(8978, msg);
         return false;
      }
      
      this.valid = false;
      return true;
   }
   
   /**
    * As this resource supports the name attribute in R/W mode, return {@code false}
    * 
    * @return   always {@code false}
    */
   @Override
   protected boolean hasNameReadOnly()
   {
      return false;
   }
   
   /**
    * Notify the listeners about the invocation request state (start or finish).
    * 
    * @param    start
    *           {@code true} to indicate invocation start, {@code false} to notify for
    *           invocation finish.
    */
   private void notifyListeners(boolean start)
   {
      for (AsyncRequestListener listener : listeners)
      {
         if (start)
         {
            listener.notifyStart();
         }
         else
         {
            listener.notifyFinish();
         }
      }
   }
}
