/*
** Module   : CommonFileInfo.java
** Abstract : Define all APIs available via the FILE-INFO system handle.
**
** Copyright (c) 2013-2017, Golden Code Development Corporation.
**
** -#- -I- --Date-- -----------------------Description------------------------
** 001 CA  20130112 Created initial version.
** 002 CA  20130927 Added LegacyResource annotation, to determine resource type.
** 003 EVL 20131008 Adding legacy attributes and methods annotation.
** 004 GES 20141124 Javadoc fix.
** 005 EVL 20160224 Javadoc fixes to make compatible with Oracle Java 8 for Solaris 10.
** 006 SVL 20170126 FILE-NAME setters moved to CommonFile interface.
*/
/*
** This program is free software: you can redistribute it and/or modify
** it under the terms of the GNU Affero General Public License as
** published by the Free Software Foundation, either version 3 of the
** License, or (at your option) any later version.
**
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
** GNU Affero General Public License for more details.
**
** You may find a copy of the GNU Affero GPL version 3 at the following
** location: https://www.gnu.org/licenses/agpl-3.0.en.html
** 
** Additional terms under GNU Affero GPL version 3 section 7:
** 
**   Under Section 7 of the GNU Affero GPL version 3, the following additional
**   terms apply to the works covered under the License.  These additional terms
**   are non-permissive additional terms allowed under Section 7 of the GNU
**   Affero GPL version 3 and may not be removed by you.
** 
**   0. Attribution Requirement.
** 
**     You must preserve all legal notices or author attributions in the covered
**     work or Appropriate Legal Notices displayed by works containing the covered
**     work.  You may not remove from the covered work any author or developer
**     credit already included within the covered work.
** 
**   1. No License To Use Trademarks.
** 
**     This license does not grant any license or rights to use the trademarks
**     Golden Code, FWD, any Golden Code or FWD logo, or any other trademarks
**     of Golden Code Development Corporation. You are not authorized to use the
**     name Golden Code, FWD, or the names of any author or contributor, for
**     publicity purposes without written authorization.
** 
**   2. No Misrepresentation of Affiliation.
** 
**     You may not represent yourself as Golden Code Development Corporation or FWD.
** 
**     You may not represent yourself for publicity purposes as associated with
**     Golden Code Development Corporation, FWD, or any author or contributor to
**     the covered work, without written authorization.
** 
**   3. No Misrepresentation of Source or Origin.
** 
**     You may not represent the covered work as solely your work.  All modified
**     versions of the covered work must be marked in a reasonable way to make it
**     clear that the modified work is not originating from Golden Code Development
**     Corporation or FWD.  All modified versions must contain the notices of
**     attribution required in this license.
*/

package com.goldencode.p2j.util;

/**
 * Exposes the following methods or attributes from the FILE-INFO handle:
 * <table summary="">
 * <tr>
 *    <th>FILE-INFO attribute or method</th>
 *    <th>API</th>
 * </tr>
 * <tr>
 *    <td>FILE-CREATE-DATE</td>
 *    <td>{@link #fileInfoGetCreationDate()}</td>
 * </tr>
 * <tr>
 *    <td>FILE-CREATE-TIME</td>
 *    <td>{@link #fileInfoGetCreationTime()}</td>
 * </tr>
 * <tr>
 *    <td>FILE-MOD-DATE</td>
 *    <td>{@link #fileInfoGetModDate()}</td>
 * </tr>
 * <tr>
 *    <td>FILE-MOD-TIME</td>
 *    <td>{@link #fileInfoGetModTime()}</td>
 * </tr>
 * <tr>
 *    <td>FILE-NAME</td>
 *    <td>{@link #initFileInfo(character)}, {@link #initFileInfo(String)} and
 *        {@link #getFileName()}
 *    </td>
 * </tr>
 * <tr>
 *    <td>FILE-SIZE</td>
 *    <td>{@link #fileInfoGetSize()}</td>
 * </tr>
 * <tr>
 *    <td>FILE-TYPE</td>
 *    <td>{@link #fileInfoGetType()}</td>
 * </tr>
 * <tr>
 *    <td>FULL-PATHNAME</td>
 *    <td>{@link #fileInfoGetFullPath()}</td>
 * </tr>
 * <tr>
 *    <td>PATHNAME</td>
 *    <td>{@link #fileInfoGetPathName()}</td>
 * </tr>
 * </table>
 * This interface is used as a returned type by the 
 * {@link handle#unwrapFileInfo} API, which, depending on the 
 * handle's state, it either returns an {@link FileSystemOps#asHandle()} 
 * instance or a dynamic proxy using {@link handle#invalidAttrAccessProxy}. 
 */
@LegacyResource(resource = LegacyResource.PSEUDO_WIDGET)
public interface CommonFileInfo
extends CommonFile,
        CommonHandle
{
   /**
    * If the current file information data structures are initialized, return
    * the absolute or relative path name, otherwise return the
    * <code>unknown value</code>. If the specified file name was absolute,
    * then an absolute pathname will be returned.  Otherwise the relative
    * path that was found via {@link FileSystemOps#searchPath} will be 
    * returned.
    *
    * @return   The path name of the current file being inspected or the
    *           <code>unknown value</code> if the data structures were never
    *           initialized or if initialization failed.
    */
   @LegacyAttribute(name = "PATHNAME")
   public character fileInfoGetPathName();
   
   /**
    * If the current file information data structures are initialized, return
    * the absolute (canonicalized) path name, otherwise return the
    * <code>unknown value</code>.
    *
    * @return   The absolute path name of the current file being inspected or
    *           the <code>unknown value</code> if the data structures were
    *           never initialized or if initialization failed.
    */
   @LegacyAttribute(name = "FULL-PATHNAME")
   public character fileInfoGetFullPath();
   
   /**
    * If the current file information data structures are initialized, return
    * the file's size, otherwise return the <code>unknown value</code>.
    *
    * @return   The size in bytes of the current file being inspected or the
    *           <code>unknown value</code> if the data structures were never
    *           initialized or if initialization failed.
    */
   @LegacyAttribute(name = "FILE-SIZE")
   public integer fileInfoGetSize();
   
   /**
    * If the current file information data structures are initialized, return
    * a string with characters that specify attributes about the file,
    * otherwise return the <code>unknown value</code>.
    * <p>
    * One of the following mutually exclusive characters will be present:
    * <p>
    * <pre>
    * Character   Meaning
    * ---------   -----------------------------
    * D           directory
    * F           file
    * X           unknown
    * </pre>
    * <p>
    * <b>There is no support for "S" (UNIX special devices) or "M" (procedure
    * library members) at this time.</b>
    * <p>
    * One or more of the following characters will be present:
    * <p>
    * <pre>
    * Character   Meaning
    * ---------   -----------------------------
    * H           hidden
    * R           readable
    * W           writable
    * </pre>
    * <p>
    * <b>There is no support for "L" (UNIX symbolic links) or "P" (pipes)
    * at this time.</b>
    *
    * @return   The file attributes of the current file being inspected or
    *           the <code>unknown value</code> if the data structures were
    *           never initialized or if initialization failed.
    */
   @LegacyAttribute(name = "FILE-TYPE")
   public character fileInfoGetType();
   
   /**
    * If the current file information data structures are initialized, return
    * the file's last modification date, otherwise return the 
    * <code>unknown value</code>.
    *
    * @return   The last modification date of the current file being inspected
    *           or the <code>unknown value</code> if the data structures were
    *           never initialized or if initialization failed.
    */
   @LegacyAttribute(name = "FILE-MOD-DATE")
   public date fileInfoGetModDate();
   
   /**
    * If the current file information data structures are initialized, return
    * the file's last modification time (in seconds from midnight), otherwise 
    * return the <code>unknown value</code>.
    *
    * @return   The last modification time in seconds from midnight of the
    *           current file being inspected or the <code>unknown value</code>
    *           if the data structures were never initialized or if
    *           initialization failed.
    */
   @LegacyAttribute(name = "FILE-MOD-TIME")
   public integer fileInfoGetModTime();
   
   /**
    * If the current file information data structures are initialized, return
    * the file's creation date, otherwise return the 
    * <code>unknown value</code>.
    *
    * @return   Always <code>unknown value</code> at this time.
    */
   @LegacyAttribute(name = "FILE-CREATE-DATE")
   public date fileInfoGetCreationDate();
   
   /**
    * If the current file information data structures are initialized, return
    * the file's creation time (in seconds from midnight), otherwise 
    * return the <code>unknown value</code>.
    *
    * @return   Always returns <code>unknown value</code> at this time.
    */
   @LegacyAttribute(name = "FILE-CREATE-TIME")
   public integer fileInfoGetCreationTime();
}