/*
** Module   : CommonProcedure.java
** Abstract : Defines common APIs used by the SESSION and PROCEDURE handles.
**
** Copyright (c) 2013-2017, Golden Code Development Corporation.

**
** -#- -I- --Date-- --------------------------------Description-----------------------------------
** 001 CA  20130112 Created initial version.
** 002 CS  20130218 Added SERVER, REMOTE, PROXY and TRANSACTION attributes.
** 003 CA  20130221 Moved to PersistentProcedure the procedure-only attributes/methods.
** 004 EVL 20130911 Adding legacy attributes and methods annotation.
** 005 CA  20131208 Emit SEARCH-TARGET and SEARCH-SELF literals as integer.
** 006 GES 20141124 Javadoc fix.
** 007 EVL 20160224 Javadoc fixes to make compatible with Oracle Java 8 for Solaris 10.
*/
/*
** This program is free software: you can redistribute it and/or modify
** it under the terms of the GNU Affero General Public License as
** published by the Free Software Foundation, either version 3 of the
** License, or (at your option) any later version.
**
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
** GNU Affero General Public License for more details.
**
** You may find a copy of the GNU Affero GPL version 3 at the following
** location: https://www.gnu.org/licenses/agpl-3.0.en.html
** 
** Additional terms under GNU Affero GPL version 3 section 7:
** 
**   Under Section 7 of the GNU Affero GPL version 3, the following additional
**   terms apply to the works covered under the License.  These additional terms
**   are non-permissive additional terms allowed under Section 7 of the GNU
**   Affero GPL version 3 and may not be removed by you.
** 
**   0. Attribution Requirement.
** 
**     You must preserve all legal notices or author attributions in the covered
**     work or Appropriate Legal Notices displayed by works containing the covered
**     work.  You may not remove from the covered work any author or developer
**     credit already included within the covered work.
** 
**   1. No License To Use Trademarks.
** 
**     This license does not grant any license or rights to use the trademarks
**     Golden Code, FWD, any Golden Code or FWD logo, or any other trademarks
**     of Golden Code Development Corporation. You are not authorized to use the
**     name Golden Code, FWD, or the names of any author or contributor, for
**     publicity purposes without written authorization.
** 
**   2. No Misrepresentation of Affiliation.
** 
**     You may not represent yourself as Golden Code Development Corporation or FWD.
** 
**     You may not represent yourself for publicity purposes as associated with
**     Golden Code Development Corporation, FWD, or any author or contributor to
**     the covered work, without written authorization.
** 
**   3. No Misrepresentation of Source or Origin.
** 
**     You may not represent the covered work as solely your work.  All modified
**     versions of the covered work must be marked in a reasonable way to make it
**     clear that the modified work is not originating from Golden Code Development
**     Corporation or FWD.  All modified versions must contain the notices of
**     attribution required in this license.
*/

package com.goldencode.p2j.util;

import com.goldencode.p2j.util.ProcedureManager.SearchMode;

/**
 * Exposes the following methods or attributes, common for the SESSION and
 * PROCEDURE handles:
 * <table summary="">
 * <tr>
 *    <th>attribute or method</th>
 *    <th>API</th>
 * </tr>
 * <tr>
 *    <td>SUPER-PROCEDURES</td>
 *    <td>{@link #superProcedures()}</td>
 * </tr>
 * <tr>
 *    <td>ADD-SUPER-PROCEDURE</td>
 *    <td>addSuperProcedure APIs</td>
 * </tr>
 * <tr>
 *    <td>REMOVE-SUPER-PROCEDURE</td>
 *    <td>removeSuperProcedure APIs</td>
 * </tr>
 * </table>
 * This interface is used as a returned type by the 
 * {@link handle#unwrapCommonProcedure} API, which, depending on the 
 * handle's state and type, it either returns an 
 * {@link SessionUtils#asHandle()} instance (for SESSION handles) or an 
 * {@link ExternalProgramWrapper} instance (for PROCEDURE handles), or a 
 * dynamic proxy using {@link handle#invalidAttrAccessProxy}. 
 */
public interface CommonProcedure
{
   /**
    * Get the SUPER-PROCEDURES attribute for this SESSION or procedure handle.
    * 
    * @return   See above.
    */
   @LegacyAttribute(name = "SUPER-PROCEDURES")
   public character superProcedures();

   /**
    * Add the given handle as a super-procedure for the current handle.
    * 
    * @param    h
    *           The handle to be added as a super-procedure.
    *           
    * @return   <code>true</code> if <code>h</code> is a valid procedure
    *           handle.
    */
   @LegacyMethod(name = "ADD-SUPER-PROCEDURE")
   public logical addSuperProcedure(handle h);

   /**
    * Add the given handle as a super-procedure for the current handle, using
    * the explicit {@link SearchMode mode}.
    * 
    * @param    h
    *           The handle to be added as a super-procedure.
    * @param    mode
    *           The {@link SearchMode} to be set for the <code>h</code> 
    *           handle.
    *           
    * @return   <code>true</code> if <code>h</code> is a valid procedure
    *           handle.
    */
   @LegacyMethod(name = "ADD-SUPER-PROCEDURE")
   public logical addSuperProcedure(handle h, SearchMode mode);

   /**
    * Add the given handle as a super-procedure for the current handle, using
    * the search-mode set as a string (one of the <code>SEARCH-TARGET</code>
    * or <code>SEARCH-SELF</code> string literals).
    * 
    * @param    h
    *           The handle to be added as a super-procedure.
    * @param    mode
    *           The search-mode to be set for the <code>h</code> handle.
    *           
    * @return   <code>true</code> if <code>h</code> is a valid procedure
    *           handle.
    */
   @LegacyMethod(name = "ADD-SUPER-PROCEDURE")
   public logical addSuperProcedure(handle h, int mode);

   /**
    * Add the given handle as a super-procedure for the current handle, using
    * the search-mode set as a string (one of the <code>SEARCH-TARGET</code>
    * or <code>SEARCH-SELF</code> string literals).
    * 
    * @param    h
    *           The handle to be added as a super-procedure.
    * @param    mode
    *           The search-mode to be set for the <code>h</code> handle.
    *           
    * @return   <code>true</code> if <code>h</code> is a valid procedure
    *           handle.
    */
   @LegacyMethod(name = "ADD-SUPER-PROCEDURE")
   public logical addSuperProcedure(handle h, integer mode);

   /**
    * Remove the given handle from this handle's list of super-procedures.
    *  
    * @param    h
    *           The procedure handle to be removed.
    *           
    * @return   <code>true</code> if <code>h</code> is a valid procedure
    *           handle.
    */
   @LegacyMethod(name = "REMOVE-SUPER-PROCEDURE")
   public logical removeSuperProcedure(handle h);
   
   /**
    * Indicates if the procedure is ran remote.
    * 
    * @return   <code>true</code> if the procedure is run remote , <code>false</code> otherwise.
    */
   @LegacyAttribute(name = "REMOTE")
   public logical isRemote(); 
}