/*
** Module   : ConditionException.java
** Abstract : represents a Progress condition (ERROR, END, STOP or QUIT)
**
** Copyright (c) 2005-2017, Golden Code Development Corporation.
**
** -#- -I- --Date-- --JPRM-- ----------------Description-----------------
** 001 GES 20050811   @22110 Created initial version which represents a
**                           Progress condition (ERROR, END, STOP or QUIT).
** 002 GES 20060623   @27559 Support storing a key code as a cause.
** 003 IAS 20160829          fillStackTrace override
*/
/*
** This program is free software: you can redistribute it and/or modify
** it under the terms of the GNU Affero General Public License as
** published by the Free Software Foundation, either version 3 of the
** License, or (at your option) any later version.
**
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
** GNU Affero General Public License for more details.
**
** You may find a copy of the GNU Affero GPL version 3 at the following
** location: https://www.gnu.org/licenses/agpl-3.0.en.html
** 
** Additional terms under GNU Affero GPL version 3 section 7:
** 
**   Under Section 7 of the GNU Affero GPL version 3, the following additional
**   terms apply to the works covered under the License.  These additional terms
**   are non-permissive additional terms allowed under Section 7 of the GNU
**   Affero GPL version 3 and may not be removed by you.
** 
**   0. Attribution Requirement.
** 
**     You must preserve all legal notices or author attributions in the covered
**     work or Appropriate Legal Notices displayed by works containing the covered
**     work.  You may not remove from the covered work any author or developer
**     credit already included within the covered work.
** 
**   1. No License To Use Trademarks.
** 
**     This license does not grant any license or rights to use the trademarks
**     Golden Code, FWD, any Golden Code or FWD logo, or any other trademarks
**     of Golden Code Development Corporation. You are not authorized to use the
**     name Golden Code, FWD, or the names of any author or contributor, for
**     publicity purposes without written authorization.
** 
**   2. No Misrepresentation of Affiliation.
** 
**     You may not represent yourself as Golden Code Development Corporation or FWD.
** 
**     You may not represent yourself for publicity purposes as associated with
**     Golden Code Development Corporation, FWD, or any author or contributor to
**     the covered work, without written authorization.
** 
**   3. No Misrepresentation of Source or Origin.
** 
**     You may not represent the covered work as solely your work.  All modified
**     versions of the covered work must be marked in a reasonable way to make it
**     clear that the modified work is not originating from Golden Code Development
**     Corporation or FWD.  All modified versions must contain the notices of
**     attribution required in this license.
*/

package com.goldencode.p2j.util;

import java.util.logging.*;

import com.goldencode.p2j.util.ErrorManager;

/**
 * Represents a Progress condition (ERROR, END, STOP or QUIT).
 * <p>
 * This class is an unchecked exception because the generated code always
 * emits catch blocks regardless of whether the body of the <code>try</code>
 * block can generate these exceptions or not.  The Java compiler fails in 
 * this condition (we have too much exception processing). The other solution
 * to this is to exhaustively analyze all <code>try</code> bodies to
 * determine if any <code>catch</code> blocks must be removed.  This has no
 * runtime benefit and is such a huge/complex effort that it is not
 * warrented.  Instead we use unchecked exceptions.
 */
public class ConditionException
extends RuntimeException
{
   /** Logger */
   private static final Logger LOG = LogHelper.getLogger(ErrorManager.class.getName());
   
   /** Flag indicating whether we capture stack trace data */
   private static final boolean captureStack = LOG.isLoggable(Level.FINE);

   /** If caused by a key press, this stores the key code. */
   private int keycode = -1;
   
   /**
    * Constructor which accepts an error message.
    *
    * @param   message
    *          Error message.
    */
   public ConditionException(String message)
   {
      super(message);
   }
    
   /**
    * Constructor which accepts an error message and the key code for the key
    * press that caused this condition.
    *
    * @param   message
    *          Error message.
    * @param   keycode
    *          The key code of the key that caused this condition to be
    *          raised.
    */
   public ConditionException(String message, int keycode)
   {
      super(message);
      this.keycode = keycode;
   }
    
   /**
    * Constructor which accepts an error message and a root cause. This
    * variant is used when the error condition is detected as the result of
    * catching another exception.
    *
    * @param   message
    *          Error message.
    * @param   rootCause
    *          Exception which has been caught in the course of processing.
    */
   public ConditionException(String message, Throwable rootCause)
   {
      super(message, rootCause);
   }
    
   /**
    * Constructor which accepts and wrappers a root cause. This variant is
    * used when the error condition is detected as the result of catching
    * another exception, and the message or context of the root cause is
    * sufficient to describe the problem.
    *
    * @param   rootCause
    *          Exception which has been caught in the course of processing.
    */
   public ConditionException(Throwable rootCause)
   {
      super(rootCause);
   }
    
   /**
    * Returns the key code that was the cause of this condition.
    *
    * @return   The key code that caused this condition or -1 if a key press
    *           was not the cause of this condition.
    */
   public int getKeyCode()
   {
      return keycode;
   }

   /**
    * Fills in the execution stack trace. This method records within this
    * {@code Throwable} object information about the current state of
    * the stack frames for the current thread.
    *
    * <p>If the stack trace of this {@code Throwable} {@linkplain
    * Throwable#Throwable(String, Throwable, boolean, boolean) is not
    * writable}, calling this method has no effect.
    *
    * @return  a reference to this {@code Throwable} instance.
    * @see     java.lang.Throwable#printStackTrace()
    */
   public synchronized Throwable fillInStackTrace() 
   {
      if (captureStack)
      {
         return super.fillInStackTrace();
      }
      
      return this;
  }
}