/*
** Module   : ExportField.java
** Abstract : handle export data fields for I/O helpers
**
** Copyright (c) 2006-2017, Golden Code Development Corporation.
**
** -#- -I- --Date-- --JPRM-- ---------------------------Description----------------------------
** 001 GES 20060925   @29901 Created initial version to export data fields for I/O helpers.
** 002 SIY 20071001   @35319 Made class suitable for network transfers. Ref: #35314
** 003 SIY 20071003   @35350 Added default constructor. Header cleanup. Fixed NPE during
**                           serialization. Ref: #35347.
** 004 GES 20160530          Use lambdas instead of Resolvable instances.
*/ 
/*
** This program is free software: you can redistribute it and/or modify
** it under the terms of the GNU Affero General Public License as
** published by the Free Software Foundation, either version 3 of the
** License, or (at your option) any later version.
**
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
** GNU Affero General Public License for more details.
**
** You may find a copy of the GNU Affero GPL version 3 at the following
** location: https://www.gnu.org/licenses/agpl-3.0.en.html
** 
** Additional terms under GNU Affero GPL version 3 section 7:
** 
**   Under Section 7 of the GNU Affero GPL version 3, the following additional
**   terms apply to the works covered under the License.  These additional terms
**   are non-permissive additional terms allowed under Section 7 of the GNU
**   Affero GPL version 3 and may not be removed by you.
** 
**   0. Attribution Requirement.
** 
**     You must preserve all legal notices or author attributions in the covered
**     work or Appropriate Legal Notices displayed by works containing the covered
**     work.  You may not remove from the covered work any author or developer
**     credit already included within the covered work.
** 
**   1. No License To Use Trademarks.
** 
**     This license does not grant any license or rights to use the trademarks
**     Golden Code, FWD, any Golden Code or FWD logo, or any other trademarks
**     of Golden Code Development Corporation. You are not authorized to use the
**     name Golden Code, FWD, or the names of any author or contributor, for
**     publicity purposes without written authorization.
** 
**   2. No Misrepresentation of Affiliation.
** 
**     You may not represent yourself as Golden Code Development Corporation or FWD.
** 
**     You may not represent yourself for publicity purposes as associated with
**     Golden Code Development Corporation, FWD, or any author or contributor to
**     the covered work, without written authorization.
** 
**   3. No Misrepresentation of Source or Origin.
** 
**     You may not represent the covered work as solely your work.  All modified
**     versions of the covered work must be marked in a reasonable way to make it
**     clear that the modified work is not originating from Golden Code Development
**     Corporation or FWD.  All modified versions must contain the notices of
**     attribution required in this license.
*/

package com.goldencode.p2j.util;

import java.io.*;
import java.util.function.*;

/**
 * Handles export formatted output of data fields for I/O helpers. 
 */
public class ExportField
extends FieldEntry
{
   /** The data to output to the stream. */
   private Supplier<BaseDataType> data = null;

   /** Resolved variable value. */
   private BaseDataType savedVar = null;
   
   /**
    * Default constructor which is necessary to restores instance when it is transferred via
    * the network.
    */
   public ExportField()
   {
   }
   
   /**
    * Generate output formatted using the <code>toStringExport</code> method of the associated
    * type.
    *
    * @param    val
    *           The value to be output.
    */
   public ExportField(int val)
   {
      this(new integer(val));
   }
   
   /**
    * Generate output formatted using the <code>toStringExport</code> method of the associated
    * type.
    *
    * @param    val
    *           The value to be output.
    */
   public ExportField(double val)
   {
      this(new decimal(val));
   }
   
   /**
    * Generate output formatted using the <code>toStringExport</code> method of the associated
    * type.
    *
    * @param    val
    *           The value to be output.
    */
   public ExportField(boolean val)
   {
      this(new logical(val));
   }
   
   /**
    * Generate output formatted using the <code>toStringExport</code> method of the associated
    * type.
    *
    * @param    val
    *           The value to be output.
    */
   public ExportField(String val)
   {
      this(new character(val));
   }
   
   /**
    * Generate output formatted using the <code>toStringExport</code> method of the associated
    * type.
    *
    * @param    val
    *           The value to be output.
    */
   public ExportField(final BaseDataType val)
   {
      this.data = () -> val;
   }
   
   /**
    * Generate output formatted using the <code>toStringExport</code> method of the associated
    * type.  This form is used for field references.
    *
    * @param    val
    *           The value to be output.
    */
   public ExportField(final Resolvable val)
   {
      this.data = () -> val.resolve();
   }
   
   /**
    * Generate output formatted using the <code>toStringExport</code> method of the associated
    * type.  This form is used for complex expressions whose evaluation bmust be deferred.
    *
    * @param    val
    *           The function which supplies the value to be output.
    */
   public ExportField(Supplier<BaseDataType> val)
   {
      this.data = val;
   }
   
  /**
    * Renders the contained field definition to the given stream.
    *
    * @param    out
    *           The stream to which the field should be rendered.
    */
   public void output(Stream out)
   throws ErrorConditionException
   {
      BaseDataType bdt = data.get();
      
      if (bdt instanceof memptr)
      {
         out.writeBlock((memptr) bdt);
      }
      else
      {
         out.writeField(bdt);
      }
   }
   
   /**
    * Notify instance that it is about to be transferred so all resolvable data should be
    * resolved and prepared for sending via network.
    */
   public void resolveData()   
   {
      savedVar = data.get();
   }
   
   /**
    * Replacement for the default object reading method. The latest state is read from the
    * input source.
    *
    * @param    in
    *           The input source from which fields will be restored.
    *
    * @throws   IOException
    *           In case of I/O errors.
    * @throws   ClassNotFoundException
    *           If payload can't be instantiated.
    */
   public void readExternal(ObjectInput in)
   throws IOException,
          ClassNotFoundException
   {
      final BaseDataType data = (BaseDataType) in.readObject();
      
      this.data = () -> data;
   }

   /**
    * Replacement for the default object writing method. The latest state is written to the
    * output destination.
    *
    * @param    out
    *           The output destination to which fields will be saved.
    *
    * @throws   IOException
    *           In case of I/O errors.
    */
   public void writeExternal(ObjectOutput out)
   throws IOException
   {
      out.writeObject(savedVar);
   }
}