/*
** Module   : FileSpecList.java
** Abstract : creates a file list based on user defined specifications,
**            optional recursion and the current dir or a user-defined
**            start directory
**
** Copyright (c) 2005-2017, Golden Code Development Corporation.
**
** -#- -I- --Date-- -T- --JPRM-- ----------------Description-----------------
** 001 GES 20050125 ADD   @19403 Created initial version supporting wildcards,
**                               searches of the current dir or a user-defined
**                               directory and optional recursion.
** 002 GES 20050224 CHG   @19950 Added lexicographic sorting of the resulting
**                               file list.
** 003 ECF 20050315 CHG   @20341 Added methods to access the properties of the
**                               file list object: getStartingDirectory,
**                               getFileSpecification, and isRecursive.
** 004 GES 20050316 CHG   @20400 Save off original (unmodified) file spec for
**                               use in reporting or debug purposes.
** 005 GES 20051005 CHG   @22973 Changed listImple() to protected instead of 
**                               private in order to allow subclasses to
**                               override the list algorithm. Also made a
**                               protected default constructor.
** 006 GES 20060129 CHG   @24128 Moved common logic into a new base class and
**                               removed this class from the parent hierarchy
**                               of ExplicitFileList.  This makes for a
**                               cleaner set of interfaces.
*/
/*
** This program is free software: you can redistribute it and/or modify
** it under the terms of the GNU Affero General Public License as
** published by the Free Software Foundation, either version 3 of the
** License, or (at your option) any later version.
**
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
** GNU Affero General Public License for more details.
**
** You may find a copy of the GNU Affero GPL version 3 at the following
** location: https://www.gnu.org/licenses/agpl-3.0.en.html
** 
** Additional terms under GNU Affero GPL version 3 section 7:
** 
**   Under Section 7 of the GNU Affero GPL version 3, the following additional
**   terms apply to the works covered under the License.  These additional terms
**   are non-permissive additional terms allowed under Section 7 of the GNU
**   Affero GPL version 3 and may not be removed by you.
** 
**   0. Attribution Requirement.
** 
**     You must preserve all legal notices or author attributions in the covered
**     work or Appropriate Legal Notices displayed by works containing the covered
**     work.  You may not remove from the covered work any author or developer
**     credit already included within the covered work.
** 
**   1. No License To Use Trademarks.
** 
**     This license does not grant any license or rights to use the trademarks
**     Golden Code, FWD, any Golden Code or FWD logo, or any other trademarks
**     of Golden Code Development Corporation. You are not authorized to use the
**     name Golden Code, FWD, or the names of any author or contributor, for
**     publicity purposes without written authorization.
** 
**   2. No Misrepresentation of Affiliation.
** 
**     You may not represent yourself as Golden Code Development Corporation or FWD.
** 
**     You may not represent yourself for publicity purposes as associated with
**     Golden Code Development Corporation, FWD, or any author or contributor to
**     the covered work, without written authorization.
** 
**   3. No Misrepresentation of Source or Origin.
** 
**     You may not represent the covered work as solely your work.  All modified
**     versions of the covered work must be marked in a reasonable way to make it
**     clear that the modified work is not originating from Golden Code Development
**     Corporation or FWD.  All modified versions must contain the notices of
**     attribution required in this license.
*/

package com.goldencode.p2j.util;

import java.io.*;
import java.util.*;

/**
 * Creates a sorted list of files in the current directory or a user-defined
 * start directory that match a user-defined specification (including
 * wildcards). No directories will be returned in this list.  The list is
 * sorted lexicographically.
 * <p>
 * The user-defined file specification is a shell matching pattern which
 * can accept a modified regular expression where '*' matches any number of
 * any character, '?' matches any single character and '[ ]' is a single
 * character match based on the set of defined alternatives. As an example,
 * to match all files, &quot;*&quot; can be used.  Or to match all files
 * that end in '.c' or '.h' one could use &quot;*.[cChH]&quot;.  Any '.' 
 * in the pattern will be escaped such that it matches a '.' directly 
 * instead of being a single character wildcard match as in standard regular
 * expressions.
 * <p>
 * The caller may optionally ask for the list to be created recursively
 * (all files that match the specification in the start/current directory 
 * and all sub-directories). 
 *
 * @author    GES
 */
public class FileSpecList
extends FileList
{
   /** The relative or absolute path to the starting directory. */ 
   private File startDir = null;
   
   /** The regular expression for filtering the file list. */
   private String fileSpec = null;
   
   /** Original unmodified file specification. */
   private String origFileSpec = null;
   
   /** Specifies if matching files in subdirectories will be included. */ 
   private boolean recursion = false;   
   
   /**
    * Base constructor to handle most likely cases.  If a <code>null</code>
    * or a non-directory is passed for the starting directory, the current
    * directory will be used.  In addition, if the directory doesn't exist,
    * the current directory will be used. 
    * <p>
    * If the file specification is <code>null</code> or an empty string, the
    * file specification will match all filenames.
    *
    * @param    startDir
    *           The starting directory for the file listing or the current
    *           directory if <code>null</code> or if it does not exist.
    * @param    fileSpec
    *           A shell pattern used to filter the list of files.
    *           Defaults to &quot;*&quot; if <code>null</code>
    * @param    recursion
    *           Specifies if matching filenames in all subdirectories should
    *           be included.
    * @param    caseSens
    *           Sets the case-sensitivity of the sorting algorithm.
    */
   public FileSpecList(File    startDir,
                       String  fileSpec,
                       boolean recursion,
                       boolean caseSens)
   {
      super(caseSens);
      
      if (startDir != null && startDir.isDirectory() && startDir.exists())
      {
         this.startDir = startDir;
      }
      else
      {
         this.startDir = new File(".");
      }
      
      if (fileSpec != null && fileSpec.length() > 0) 
      {
         origFileSpec = fileSpec;
         this.fileSpec = convertToRegEx(fileSpec);
      }
      else
      {
         origFileSpec = "*";
         this.fileSpec = ".*";
      }
      
      this.recursion = recursion;
   }
   
   /**
    * Base constructor to handle most likely cases.  If a <code>null</code>
    * or a non-directory is passed for the starting directory, the current
    * directory will be used.  In addition, if the directory doesn't exist,
    * the current directory will be used. 
    * <p>
    * If the file specification is <code>null</code> or an empty string, the
    * file specification will match all filenames.
    *
    * @param    startDir
    *           The starting directory for the file listing or the current
    *           directory if <code>null</code> or if it does not exist.
    * @param    fileSpec
    *           A shell pattern used to filter the list of files.
    *           Defaults to &quot;*&quot; if <code>null</code>
    * @param    recursion
    *           Specifies if matching filenames in all subdirectories should
    *           be included.
    */
   public FileSpecList(File startDir, String fileSpec, boolean recursion)
   {
      this(startDir, fileSpec, recursion, false);
   }
   
   /**
    * Get the starting directory for the match.
    *
    * @return  Starting directory.
    */
   public File getStartDir()
   {
      return startDir;
   }
   
   /**
    * Get the file specification (modified to be a regular expression) used
    * for the match.
    *
    * @return  File specification.
    */
   public String getFileSpecAsRegEx()
   {
      return fileSpec;
   }
   
   /**
    * Get the original unmodified file specification used for the match.
    *
    * @return  File specification.
    */
   public String getFileSpec()
   {
      return origFileSpec;
   }
   
   /**
    * Indicates whether the match is applied recursively into subdirectories.
    *
    * @return  <code>true</code> if match is recursive, else
    *          <code>false</code>.
    */
   public boolean isRecursive()
   {
      return recursion;
   }
   
   /**
    * Core implementation of the listing algorithm which uses the
    * <code>File</code> class to make a complete list of the directory 
    * contents, then this list is used to build a filtered list of just
    * those files that match the user-defined regular expression. Note
    * that if recursion was previously requested (on construction) by the
    * caller, this method will recursively call itself for every directory
    * it encounters in the list.  Only files will be returned in the results
    * list and all results are added to the results list that is passed as
    * a parameter.
    *
    * @param    results
    *           The list to which all matching files must be appended.
    */
   protected void listImpl(ArrayList results)
   {
      listImpl(startDir, results);
   }
   
   /**
    * Core implementation of the listing algorithm which uses the
    * <code>File</code> class to make a complete list of the directory 
    * contents, then this list is used to build a filtered list of just
    * those files that match the user-defined regular expression. Note
    * that if recursion was previously requested (on construction) by the
    * caller, this method will recursively call itself for every directory
    * it encounters in the list.  Only files will be returned in the results
    * list and all results are added to the results list that is passed as
    * a parameter.
    *
    * @param    dir
    *           The directory in which to list all files.
    * @param    results
    *           The list to which all matching files must be appended.
    */
   protected void listImpl(File dir, ArrayList results)
   {
      // we don't use a FilenameFilter here because this would filter out
      // all directories that don't match the regular expression, but this
      // would not be correct, we need to see all directories and we will
      // manually filter the filenames down ourselves
      File[] rawResults  = dir.listFiles();
      
      // separate file and directory results
      for (int i = 0; i < rawResults.length; i++)
      {
         if (rawResults[i].isDirectory())
         {
            // we only care about directories if recursion is specified
            // otherwise we just skip it
            if (recursion)
            {
               // recursively call ourselves to process this subdirectory
               listImpl(rawResults[i], results);
            }
         }
         else
         {
            // only accept matches that meet the user's criteria (regular
            // expression)
            if (rawResults[i].getName().matches(fileSpec))
            {
               // add the file to the results list
               results.add(rawResults[i]);
            }
         }
      }
   }
   
   /**
    * Converts a shell-type matching "pattern" to a traditional regular 
    * expression.  This allows a simplified patterns to be expressed by
    * a user but the prebuilt regular expression processing to be used
    * for implementation.
    * <p>
    * The following conversions are done:
    * <pre>
    *    Shell Pattern           Regular Expression
    *    -------------           ------------------
    *    .                       \.
    *    ?                       .
    *    *                       .*
    * </pre>
    * <p>
    * This seems trivial, except that the backslash '\' character must be
    * handled properly since it needs to be doubled up for Java to recognise
    * one in a Java string and if two are needed in the regular expression
    * then one needs to double them up, twice.
    *
    * @param   pattern
    *          Shell matching pattern.    
    * @return  Converted regular expression.
    */
   private String convertToRegEx(String pattern)
   {
      String converted;
      
      // the first parameter of repaceAll() is itself a regular expression!
      // this means that the matching looks a bit weird because we have to 
      // insert '\\' to escape the special characters that have meaning
      // in regular expressions!
      
      // convert . to \. (the target string needs \\\\ so that \ actually
      // appears in the resulting regular expression!)
      converted = pattern.replaceAll("\\.", "\\\\.");
      
      // convert ? to .
      converted = converted.replaceAll("\\?", ".");
      
      // replace * with .*
      converted = converted.replaceAll("\\*", ".*");

      return converted;
   }
   
   /**
    * Provides a command line interface for an end user to drive and/or test
    * the this class.  The list of resulting filenames is displayed to
    * the user on <code>stdout</code>.
    * <p>
    * Syntax:
    * <pre>
    *    java FileSpecList &lt;directory&gt; &quot;file_spec_pattern&quot;
    *                  [recursion]
    * </pre>
    * Where:
    * <ul>
    *    <li> &lt;directory&gt; is the name of the directory
    *    <li> &quot;file_spec_pattern&quot; is the filename pattern
    *         as used in a shell which can use '*' and '?' as wildcards as
    *         well as a subset of normal regular expression syntax (this
    *         pattern MUST be enclosed in double quotes)
    *    <li> [recursion] is an optional argument that is specified as
    *         'r' or 'R' to request recursion of subdirectories
    * </ul>
    *
    * @param   args 
    *          List of command line arguments.
    */
   public static void main(String[] args)
   {
      String syntax = "Syntax: java FileSpecList <directory> " +
                       "\"file_spec_pattern\" [-R]";
      
      if (args.length < 2)
      {
         System.err.println(syntax);
         System.exit(-1);
      }
      
      File dir = new File(args[0]);
      
      boolean recursion = false; 
      if (args.length > 2 && args[2].equalsIgnoreCase("-r"))
      {
         recursion = true;
      }
      
      try
      {
         FileList flist  = new FileSpecList(dir, args[1], recursion);
         File     list[] = flist.list();
         
         for ( int i = 0; i < list.length; i++ )
         {
            System.out.println(list[i].toString());
         }
      }
      
      catch (Exception e)
      {
         e.printStackTrace();
      }
   }
}