/*
** Module   : FileSystem.java
** Abstract : an interface to expose file system functionality
**
** Copyright (c) 2006-2018, Golden Code Development Corporation.
**
** -#- -I- --Date-- --JPRM-- ----------------------------Description-----------------------------
** 001 GES 20060211   @24538 Created initial version, supporting the basic file
**                           info, copy/append, delete, rename, search, mkdir,
**                           drive list and system property query.
** 002 GES 20110712          Searching now honors the case-sensitivity of the
**                           legacy system.
** 003 SVL 20171109          Added getFileContents.
** 004 SBI 20180522          Added setLegacyFileSystemParameters.
**     ECF 20180609          Added {get|set}WorkingDirectory methods.
*/
/*
** This program is free software: you can redistribute it and/or modify
** it under the terms of the GNU Affero General Public License as
** published by the Free Software Foundation, either version 3 of the
** License, or (at your option) any later version.
**
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
** GNU Affero General Public License for more details.
**
** You may find a copy of the GNU Affero GPL version 3 at the following
** location: https://www.gnu.org/licenses/agpl-3.0.en.html
** 
** Additional terms under GNU Affero GPL version 3 section 7:
** 
**   Under Section 7 of the GNU Affero GPL version 3, the following additional
**   terms apply to the works covered under the License.  These additional terms
**   are non-permissive additional terms allowed under Section 7 of the GNU
**   Affero GPL version 3 and may not be removed by you.
** 
**   0. Attribution Requirement.
** 
**     You must preserve all legal notices or author attributions in the covered
**     work or Appropriate Legal Notices displayed by works containing the covered
**     work.  You may not remove from the covered work any author or developer
**     credit already included within the covered work.
** 
**   1. No License To Use Trademarks.
** 
**     This license does not grant any license or rights to use the trademarks
**     Golden Code, FWD, any Golden Code or FWD logo, or any other trademarks
**     of Golden Code Development Corporation. You are not authorized to use the
**     name Golden Code, FWD, or the names of any author or contributor, for
**     publicity purposes without written authorization.
** 
**   2. No Misrepresentation of Affiliation.
** 
**     You may not represent yourself as Golden Code Development Corporation or FWD.
** 
**     You may not represent yourself for publicity purposes as associated with
**     Golden Code Development Corporation, FWD, or any author or contributor to
**     the covered work, without written authorization.
** 
**   3. No Misrepresentation of Source or Origin.
** 
**     You may not represent the covered work as solely your work.  All modified
**     versions of the covered work must be marked in a reasonable way to make it
**     clear that the modified work is not originating from Golden Code Development
**     Corporation or FWD.  All modified versions must contain the notices of
**     attribution required in this license.
*/

package com.goldencode.p2j.util;

import java.io.*;
import java.util.*;

/**
 * Provides an API for file system query and manipulation. 
 */
public interface FileSystem
{
   /** The Progress code for "no error". */
   public static final int ERR_NO_ERROR = 0;
   
   /** The Progress code for "not owner". */
   public static final int ERR_NOT_OWNER = 1;
   
   /** The Progress code for "no such file or directory". */
   public static final int ERR_NO_SUCH_FILE_OR_DIR = 2;
   
   /** The Progress code for "interrupted system call". */
   public static final int ERR_INTERRUPTED = 3;
   
   /** The Progress code for "I/O error". */
   public static final int ERR_IO = 4;
   
   /** The Progress code for "bad file number". */
   public static final int ERR_BAD_FILE_NUM = 5;
   
   /** The Progress code for "no more processes". */
   public static final int ERR_NO_MORE_PROCESSES = 6;
   
   /** The Progress code for "not enough core memory". */
   public static final int ERR_NOT_ENOUGH_CORE = 7;
   
   /** The Progress code for "permission denied". */
   public static final int ERR_PERM_DENIED = 8;
   
   /** The Progress code for "bad address". */
   public static final int ERR_BAD_ADDRESS = 9;
   
   /** The Progress code for "file exists". */
   public static final int ERR_FILE_EXISTS = 10;
   
   /** The Progress code for "not such device". */
   public static final int ERR_NO_SUCH_DEVICE = 11;
   
   /** The Progress code for "not a directory". */
   public static final int ERR_NOT_DIR = 12;
   
   /** The Progress code for "is a directory". */
   public static final int ERR_IS_DIR = 13;
   
   /** The Progress code for "file table overflow". */
   public static final int ERR_FILE_TABLE_OVERFLOW = 14;
   
   /** The Progress code for "too many open files". */
   public static final int ERR_TOO_MANY_FILES = 15;
   
   /** The Progress code for "file too large". */
   public static final int ERR_FILE_TOO_LARGE = 16;
   
   /** The Progress code for "no space left on device". */
   public static final int ERR_NO_SPACE_ON_DEVICE = 17;
   
   /** The Progress code for "directory not empty". */
   public static final int ERR_DIR_NOT_EMPTY = 18;
   
   /** The Progress code for "unknown error". */
   public static final int ERR_UNKNOWN = 999;
   
   /** Regexp pattern for matching any path separators */
   public static final String ANY_PATH_SEPARATOR = "[\\\\/]+";

   /**
    * Access file system details regarding a specific file.
    *
    * @param    name
    *           The relative or absolute filename for which to obtain 
    *           information.
    *
    * @return   The file details if the file was found or <code>null</code>
    *           if the file was not found.
    */
   public FileInfo accessFileInfo(String name);
   
   /**
    * Get the comma-separated list of filesystem roots (drive letters on
    * platforms that support the concept, otherwise returns the equivalent
    * of the empty string).
    *
    * @return   The comma-separated list of drive letters or the empty string
    *           if the platform does not support the concept.
    */
   public String getRootList()
   throws IOException;
   
   /**
    * Gets the working directory.
    *
    * @return   The current working directory.
    */
   public String getWorkingDirectory();
   
   /**
    * Get the system property (user-defined properties can be specified
    * at the Java command line) associated with the passed key.
    *
    * @return   The system property found or <code>null</code> if the
    *           property does not exist.
    */
   public String getProperty(String key);
      
   /**
    * Get the file separator for the system.
    *
    * @return   The file separator.
    */
   public String getFileSeparator();
      
   /**
    * Get the path separator for the system.
    *
    * @return   The path separator.
    */
   public String getPathSeparator();
      
   /**
    * Creates each of the directories requested, if it doesn't exist and all
    * path segments up to the last segment do exist.  If the directory name
    * already exists in the filesystem (even if it is a file), this method
    * just returns (this is the Progress behavior).  If any error occurs, no
    * exception will be thrown, but no subsequent directories in the list
    * will be created.
    *
    * @param    dirs
    *           The list of directories to create.    
    *
    * @return   The error code (<code>ERR_NO_ERROR</code> if no problems
    *           occur).
    */
   public int mkdir(String[] dirs);
   
   /**
    * Renames (and moves if the path is different) the specified source file
    * to the target filename (and path).  If any error occurs, no
    * exception will be thrown.
    *
    * @param    source
    *           The name of the source file or directory.
    * @param    target                             
    *           The name of the target file or directory.  If this is a
    *           directory name for a directory that already exists and the
    *           <code>source</code> is a file, then the source file name will
    *           be used as the target file in the directory given. 
    *
    * @return   The error code (<code>ERR_NO_ERROR</code> if no problems
    *           occur).
    */
   public int rename(String source, String target);
   
   /**
    * Initializes the path for subsequent searches by {@link #searchPath}. 
    *
    * @param    paths
    *           The list of paths to search for files specified by relative
    *           filename.
    * @param    caseSens
    *           <code>true</code> if the legacy system had a case-sensitive
    *           file-system.
    */
   public void setSearchPath(String[] paths, boolean caseSens);
   
   /**
    * Search for the given file in each directory listed in the path
    * previously specified via {@link #setSearchPath}.  If the file system
    * object exists and is a file, then the filename is returned. If the path
    * contains an empty string or a '.', then the current directory is
    * searched, otherwise only the listed directories are searched.  
    * <p>
    * If the filename is specified as an absolute name, this specific 
    * filename is checked and no path searching is done.
    *
    * @param    filename
    *           The absolute or relative filename for which to search.
    *
    * @return   If the file is found in the current directory, the relative
    *           filename is returned. If the file is absolute or found in any
    *           other path, the full absolute filename is returned.  If the
    *           file is not found or the referenced file system object is not
    *           a file, <code>null</code> is returned.
    */
   public String searchPath(String filename);
   
   /**
    * Sets the working directory.
    *
    * @param    newWorkingDir
    *           A new working directory to set.
    */
   public void setWorkingDirectory(String newWorkingDir);
   
   /**
    * Copies (or appends) the specified source file to or over the target
    * file. The mode (copy/overwrite or append) is determined by the
    * <code>overwrite</code> parameter.  If any error occurs, no exception
    * will be thrown.
    *
    * @param    source
    *           The name of the source file.  Must NOT be a directory name.
    *           Must be an existing file.
    * @param    target
    *           The name of the target file.  If this is a directory name,
    *           the source file name will be used as the target file in
    *           the directory given. 
    * @param    overwrite
    *           If <code>true</code> this is a copy, otherwise it is an
    *           append.
    *
    * @return   The error code (<code>ERR_NO_ERROR</code> if no problems
    *           occur).
    */
   public int copy(String source, String target, boolean overwrite);
   
   /**
    * Deletes the specified file system entity, if necessary using recursion
    * to delete all contents (if it is a directory) first.  If any error 
    * occurs, no exception will be thrown.
    *
    * @param    target
    *           The target file or directory to delete.
    * @param    recursive
    *           Recursively decend through all subdirectories, deleting    
    *           the entire subtree.
    *
    * @return   The error code (<code>ERR_NO_ERROR</code> if no problems
    *           occur).
    */
   public int delete(String target, boolean recursive);   
   
   /**
    * Deletes each of the directories or files specified.  If a non-empty
    * directory name is specified, an error will occur unless the
    * <code>recursive</code> flag is <code>true</code>.  If any error occurs,
    * no exception will be thrown, the deletion process just continues. The
    * last error will be the one returned even if there were prior successful
    * deletes or other failures.
    *
    * @param    list
    *           The list of files and directories to delete.    
    * @param    recursive
    *           Recursively decend through all subdirectories, deleting    
    *           the entire subtree.
    *
    * @return   The error code (<code>ERR_NO_ERROR</code> if no problems
    *           occur).
    */
   public int delete(String[] list, boolean recursive);

   /**
    * Get the file contents from the client-side file system.
    *
    * @param  filename
    *         Name of the file in the client-side file system (relatively to the current
    *         directory on the client side).
    *
    * @return file contents as array of bytes.
    *
    * @throws java.io.IOException
    *         If file cannot not be found or read, or is too large.
    */
   public byte[] getFileContents(String filename)
   throws IOException;

   /**
    * Sets the legacy file system parameters to be used by the client.
    * 
    * @param    fileSeparator
    *           The legacy file separator
    * @param    caseSensitive
    *           The legacy case sensitive mode
    * @param    pathMap
    *           The legacy path map
    * @param    paths
    *           The list of paths to search for files specified by relative
    *           filename.
    * @param    areLegacyPathsNotResolved
    *           The true value of this parameter indicates that legacy path names must be resolved
    *           on the host system, and false value means that legacy path names are not required
    *           to be resolved.
    */
   public void setLegacyFileSystemParameters(String              fileSeparator,
                                             boolean             caseSensitive,
                                             Map<String, String> pathMap,
                                             String[]            paths,
                                             boolean             areLegacyPathsNotResolved);
}