/*
** Module   : LargeObjectOps.java
** Abstract : helper routines to process LOBs
**
** Copyright (c) 2013-2018, Golden Code Development Corporation.
**
** -#- -I- --Date-- --------------------------------Description---------------------------------
** 001 GES 20130516 First version implementing an early version of the COPY-LOB statement which
**                  is suitable for executing testcases but is not yet 100% complete.
** 002 EVK 20140206 Added method readFromFile. This is an early version of COPY-LOB statement
**                  which is suitable for executing testcases but is not yet 100% complete.
** 003 VIG 20140914 Added comment about behaviour in readFromFile method.
** 004 GES 20141002 Added simple implementation of copying between MEMPTR and LONGCHAR (either
**                  direction).  Also added alternate versions of readFromFile() that handle
**                  starting positions + length options.  All of these are only useful for simple
**                  testing, they are not 100% complete/compatible.
** 005 GES 20180409 Added basic implementation of LONGCHAR to FILE and FILE to LONGCHAR. Not
**                  100% compatible.
**     CA  20180514 Enhanced the COPY-LOP APIs which use binary data, to support start position,
**                  length and convert source/target code-page.
*/

/*
** This program is free software: you can redistribute it and/or modify
** it under the terms of the GNU Affero General Public License as
** published by the Free Software Foundation, either version 3 of the
** License, or (at your option) any later version.
**
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
** GNU Affero General Public License for more details.
**
** You may find a copy of the GNU Affero GPL version 3 at the following
** location: https://www.gnu.org/licenses/agpl-3.0.en.html
** 
** Additional terms under GNU Affero GPL version 3 section 7:
** 
**   Under Section 7 of the GNU Affero GPL version 3, the following additional
**   terms apply to the works covered under the License.  These additional terms
**   are non-permissive additional terms allowed under Section 7 of the GNU
**   Affero GPL version 3 and may not be removed by you.
** 
**   0. Attribution Requirement.
** 
**     You must preserve all legal notices or author attributions in the covered
**     work or Appropriate Legal Notices displayed by works containing the covered
**     work.  You may not remove from the covered work any author or developer
**     credit already included within the covered work.
** 
**   1. No License To Use Trademarks.
** 
**     This license does not grant any license or rights to use the trademarks
**     Golden Code, FWD, any Golden Code or FWD logo, or any other trademarks
**     of Golden Code Development Corporation. You are not authorized to use the
**     name Golden Code, FWD, or the names of any author or contributor, for
**     publicity purposes without written authorization.
** 
**   2. No Misrepresentation of Affiliation.
** 
**     You may not represent yourself as Golden Code Development Corporation or FWD.
** 
**     You may not represent yourself for publicity purposes as associated with
**     Golden Code Development Corporation, FWD, or any author or contributor to
**     the covered work, without written authorization.
** 
**   3. No Misrepresentation of Source or Origin.
** 
**     You may not represent the covered work as solely your work.  All modified
**     versions of the covered work must be marked in a reasonable way to make it
**     clear that the modified work is not originating from Golden Code Development
**     Corporation or FWD.  All modified versions must contain the notices of
**     attribution required in this license.
*/

package com.goldencode.p2j.util;

import java.io.*;

/**
 * Provides static utility methods for processing large objects (LOBs) in a Progress 4GL
 * compatible manner.
 * <p> 
 * At this time, the following Progress 4GL features are handled:
 * <p>
 * <pre>
 * Progress        From           To             Java Equivalent
 * -------------   ------------   ------------   ---------------------
 * COPY-LOB        MEMPTR         LONGCHAR       copy
 * COPY-LOB        LONGCHAR       MEMPTR         copy
 * COPY-LOB        File           MEMPTR         readFromFile
 * COPY-LOB        MEMPTR         File           writeToFile
 * COPY-LOB        File           LONGCHAR       readFromFile
 * COPY-LOB        LONGCHAR       File           writeToFile
 * </pre>
 */
public class LargeObjectOps
{
   /**
    * Private constructor to ensure instances are not created. All functionality of this class
    * is to be used via static methods.
    */
   private LargeObjectOps()
   {
   }
   
   /**
    * Copy the contents of one LOB to another.
    * <p>
    * No codepage conversion is implemented.  The entire contents will be copied.
    *
    * @param    src
    *           The object to copy from.
    * @param    target
    *           The object to copy to.
    */
   public static void copy(BinaryData src, longchar target)
   {
      copy(src, target, null, null);
   }
   
   /**
    * Copy the contents of one LOB to another.
    * <p>
    * No codepage conversion is implemented.  The entire contents will be copied.
    *
    * @param    src
    *           The object to copy from.
    * @param    target
    *           The object to copy to.
    * @param    sourceCodePage
    *           The source code page.  If <code>null</code> or unknown, it will be ignored.
    * @param    targetCodePage
    *           The target code page.  If <code>null</code> or unknown, it will be ignored.
    */
   public static void copy(BinaryData src,
                           longchar   target,
                           character  sourceCodePage,
                           character  targetCodePage)
   {
      // TODO: make this fully compatible, this is just a partially working stub right now
      
      // unknown/uninitialized handling should be here
      
      byte[] contents = src.getByteArray();
      
      if (raiseErrorOnNullChar(contents))
         return;
      
      boolean useSrcCp = sourceCodePage != null && !sourceCodePage.isUnknown();
      boolean useTarCp = targetCodePage != null && !targetCodePage.isUnknown();
      String srcCp = useSrcCp ? sourceCodePage.toStringMessage() : null;
      String tarCp = useTarCp ? targetCodePage.toStringMessage() : null;
      
      // TODO: code-page conversion needs to happen here
      target.assign(new String(contents));
   }

   /**
    * Copy the contents of one LOB to another.
    * <p>
    * No codepage conversion is implemented.  The entire contents will be copied.
    *
    * @param    src
    *           The object to copy from.
    * @param    startPos
    *           The start position.
    * @param    length
    *           The number of bytes to copy.
    * @param    target
    *           The object to copy to.
    */
   public static void copy(BinaryData src, long startPos, long length, longchar target)
   {
      copy(src, startPos, length, target, null, null);
   }

   /**
    * Copy the contents of one LOB to another.
    * <p>
    * No codepage conversion is implemented.  The entire contents will be copied.
    *
    * @param    src
    *           The object to copy from.
    * @param    startPos
    *           The start position.
    * @param    length
    *           The number of bytes to copy.
    * @param    target
    *           The object to copy to.
    * @param    sourceCodePage
    *           The source code page.  If <code>null</code> or unknown, it will be ignored.
    */
   public static void copy(BinaryData src, 
                           long       startPos,
                           long       length,
                           longchar   target,
                           character  sourceCodePage)
   {
      copy(src, startPos, length, target, sourceCodePage, null);
   }

   /**
    * Copy the contents of one LOB to another.
    * <p>
    * No codepage conversion is implemented.  The entire contents will be copied.
    *
    * @param    src
    *           The object to copy from.
    * @param    startPos
    *           The start position.
    * @param    length
    *           The number of bytes to copy.
    * @param    target
    *           The object to copy to.
    * @param    sourceCodePage
    *           The source code page.  If <code>null</code> or unknown, it will be ignored.
    * @param    targetCodePage
    *           The target code page.  If <code>null</code> or unknown, it will be ignored.
    */
   public static void copy(BinaryData src,
                           long       startPos,
                           long       length,
                           longchar   target,
                           character  sourceCodePage,
                           character  targetCodePage)
   {
      if (length < 0)
      {
         ErrorManager.recordOrThrowError(11344, 
                                         "Unknown or invalid length specified in COPY-LOB statement", 
                                         false);
         return;
      }
      
      if (startPos <= 0)
      {
         ErrorManager.recordOrThrowError(11332, 
                                         "Invalid offset specified in COPY-LOB statement", 
                                         false);
         return;
      }

      long len = src.visibleLength();
      if (startPos > len)
      {
         ErrorManager.recordOrThrowError(11334, 
                                         "Offset supplied for source large object is greater " +
                                         "than object size", 
                                         false);
         return;
      }
      
      if (startPos + length > len)
      {
         String msg = "Cannot copy large object starting at %d for %d bytes when object size is %d";
         ErrorManager.recordOrThrowError(11305,
                                         String.format(msg, startPos, length, len),
                                         false);
         return;
      }
      
      BinaryData bytes = src.getBytes(startPos, length);
      
      copy(bytes, target, sourceCodePage, targetCodePage);
   }

   /**
    * Copy the contents of one LOB to another.
    * <p>
    * No codepage conversion is implemented.  The entire contents will be copied.
    *
    * @param    src
    *           The object to copy from.
    * @param    startPos
    *           The start position.
    * @param    length
    *           The number of bytes to copy.
    * @param    target
    *           The object to copy to.
    */
   public static void copy(BinaryData src,
                           NumberType startPos,
                           NumberType length,
                           longchar target)
   {
      copy(src, startPos, length, target, null, null);
   }

   /**
    * Copy the contents of one LOB to another.
    * <p>
    * No codepage conversion is implemented.  The entire contents will be copied.
    *
    * @param    src
    *           The object to copy from.
    * @param    startPos
    *           The start position.
    * @param    length
    *           The number of bytes to copy.
    * @param    target
    *           The object to copy to.
    * @param    sourceCodePage
    *           The source code page.  If <code>null</code> or unknown, it will be ignored.
    */
   public static void copy(BinaryData src,
                           NumberType startPos,
                           NumberType length,
                           longchar   target,
                           character  sourceCodePage)
   {
      copy(src, startPos, length, target, sourceCodePage, null);
   }

   /**
    * Copy the contents of one LOB to another.
    * <p>
    * No codepage conversion is implemented.  The entire contents will be copied.
    *
    * @param    src
    *           The object to copy from.
    * @param    startPos
    *           The start position.
    * @param    length
    *           The number of bytes to copy.
    * @param    target
    *           The object to copy to.
    * @param    sourceCodePage
    *           The source code page.  If <code>null</code> or unknown, it will be ignored.
    * @param    targetCodePage
    *           The target code page.  If <code>null</code> or unknown, it will be ignored.
    */
   public static void copy(BinaryData src,
                           NumberType startPos,
                           NumberType length,
                           longchar   target,
                           character  sourceCodePage,
                           character  targetCodePage)
   {
      if (length.isUnknown())
      {
         ErrorManager.recordOrThrowError(11344, 
                                         "Unknown or invalid length specified in COPY-LOB statement", 
                                         false);
         return;
      }

      if (startPos.isUnknown())
      {
         ErrorManager.recordOrThrowError(11332, 
                                         "Invalid offset specified in COPY-LOB statement", 
                                         false);
         return;
      }

      copy(src, startPos.longValue(), length.longValue(), target, sourceCodePage, targetCodePage);
   }
   
   /**
    * Copy the contents of one LOB to another.
    * <p>
    * No codepage conversion is implemented.  The entire contents will be copied.
    *
    * @param    src
    *           The object to copy from.
    * @param    target
    *           The object to copy to.
    */
   public static void copy(longchar src, BinaryData target)
   {
      // TODO: make this fully compatible, this is just a partially working stub right now
      
      // unknown handling should be here
      
      String val = src.toStringMessage();
      
      // code-page conversion needs to happen here
      byte[] contents = val.getBytes();
      
      target.assign(contents);
   }

   /**
    * Read some contents of the file and write that to the given binary object.
    *
    * @param    fname
    *           The target filename.
    * @param    idx
    *           The 1-based offset into the file from which to start reading.
    * @param    len
    *           The number of bytes to read.
    * @param    data
    *           The buffer in which to write.
    */
   public static void readFromFile(String fname, int idx, int len, BinaryData data)
   {
      // TODO: this is a simple placeholder that really should be replaced with a
      //       single call to a client-side worker that can handle all features, but
      //       until the feature set is fully explored we are implementing the simple
      //       way
      
      if (fname == null || fname.length() == 0 || data == null || data.isUnknown())
         return;
      
      Stream file = StreamFactory.openFileStream(fname, false, false);

      try
      {
         // seek to the 0-based offset
         file.setPos(idx - 1);
         
         // read the data
         byte[] chunk = file.readChunk(len);
         
         // write into the buffer
         data.assign(chunk);
      }
      catch (EndConditionException | ErrorConditionException | IOException iioe)
      {
         throw new StopConditionException(iioe);
      }
      finally
      {
         file.close();
      }
   }
   
   /**
    * Read some contents of the file and write that to the given binary object.
    *
    * @param    fname
    *           The target filename.
    * @param    idx
    *           The 1-based offset into the file from which to start reading.
    * @param    len
    *           The number of bytes to read.
    * @param    data
    *           The buffer in which to write.
    */
   public static void readFromFile(String fname, NumberType idx, int len, BinaryData data)
   {
      if (idx == null || idx.isUnknown())
         return;
      
      readFromFile(fname, idx.intValue(), len, data);
   }
   
   /**
    * Read some contents of the file and write that to the given binary object.
    *
    * @param    fname
    *           The target filename.
    * @param    idx
    *           The 1-based offset into the file from which to start reading.
    * @param    len
    *           The number of bytes to read.
    * @param    data
    *           The buffer in which to write.
    */
   public static void readFromFile(String fname, int idx, NumberType len, BinaryData data)
   {
      if (len == null || len.isUnknown())
         return;
      
      readFromFile(fname, idx, len.intValue(), data);
   }
   
   /**
    * Read some contents of the file and write that to the given binary object.
    *
    * @param    fname
    *           The target filename.
    * @param    idx
    *           The 1-based offset into the file from which to start reading.
    * @param    len
    *           The number of bytes to read.
    * @param    data
    *           The buffer in which to write.
    */
   public static void readFromFile(String fname, NumberType idx, NumberType len, BinaryData data)
   {
      if (idx == null || idx.isUnknown() || len == null || len.isUnknown())
         return;
      
      readFromFile(fname, idx.intValue(), len.intValue(), data);
   }
   
   /**
    * Read some contents of the file and write that to the given binary object.
    *
    * @param    fname
    *           The target filename.
    * @param    idx
    *           The 1-based offset into the file from which to start reading.
    * @param    len
    *           The number of bytes to read.
    * @param    data
    *           The buffer in which to write.
    */
   public static void readFromFile(character fname, int idx, int len, BinaryData data)
   {
      if (fname == null || fname.isUnknown())
         return;
      
      readFromFile(fname.toStringMessage(), idx, len, data);
   }
   
   /**
    * Read some contents of the file and write that to the given binary object.
    *
    * @param    fname
    *           The target filename.
    * @param    idx
    *           The 1-based offset into the file from which to start reading.
    * @param    len
    *           The number of bytes to read.
    * @param    data
    *           The buffer in which to write.
    */
   public static void readFromFile(character fname, NumberType idx, int len, BinaryData data)
   {
      if (idx == null || idx.isUnknown())
         return;
      
      readFromFile(fname, idx.intValue(), len, data);
   }
   
   /**
    * Read some contents of the file and write that to the given binary object.
    *
    * @param    fname
    *           The target filename.
    * @param    idx
    *           The 1-based offset into the file from which to start reading.
    * @param    len
    *           The number of bytes to read.
    * @param    data
    *           The buffer in which to write.
    */
   public static void readFromFile(character fname, int idx, NumberType len, BinaryData data)
   {
      if (len == null || len.isUnknown())
         return;
      
      readFromFile(fname, idx, len.intValue(), data);
   }
   
   /**
    * Read some contents of the file and write that to the given binary object.
    *
    * @param    fname
    *           The target filename.
    * @param    idx
    *           The 1-based offset into the file from which to start reading.
    * @param    len
    *           The number of bytes to read.
    * @param    data
    *           The buffer in which to write.
    */
   public static void readFromFile(character  fname,
                                   NumberType idx,
                                   NumberType len,
                                   BinaryData data)
   {
      if (idx == null || idx.isUnknown() || len == null || len.isUnknown())
         return;
      
      readFromFile(fname, idx.intValue(), len.intValue(), data);
   }
   
   /**
    * Read the full contents of the file and write that to the given binary object.
    *
    * @param    fname
    *           The target filename.
    * @param    data
    *           The buffer in which to write.
    */
   public static void readFromFile(String fname, BinaryData data)
   {
// TODO: this is just a first pass, the real/full behavior should be implemented as following:
//      1. size(data) < size(file). For exapmle size(data) = 5 and size(file) = 23.
//      Progress reads all file and dynamically increases data size to needed value 
//      i.e. from 5 to 23 (in case of data).
//      Java impl only reads size(data) bytes from a file, i.e. 5
//      2. size(data) > size(file). For exapmle size(data) = 35 and size(file) = 23.
//      Both sides works equal except the one border case when the size of a file is 0.
//      In that case Progress works correctly without errors but Java implementation 
//      raises EndConditionException.
      
      if (data == null || data.isUnknown() || fname == null)
      {
         return;
      }

      Stream file = StreamFactory.openFileStream(fname, false, false);

      try
      {
         file.readBlock((memptr) data);
      }
      catch (EndConditionException | ErrorConditionException iioe)
      {
         throw new StopConditionException(iioe);
      }
      finally
      {
         file.close();
      }
   }

   /**
    * Read the full contents of the file and write that to the given binary object.
    *
    * @param    fname
    *           The target filename.
    * @param    data
    *           The buffer in which to write.
    */
   public static void readFromFile(character fname, BinaryData data)
   {
      if (fname == null || fname.isUnknown())
      {
         return;
      }
      readFromFile(fname.toStringMessage(), data);
   }

   /**
    * Write the full contents of the given binary object to the file specified by the given
    * filename.
    *
    * @param    data
    *           The data to write.
    * @param    fname
    *           The target filename.
    */
   public static void writeToFile(BinaryData data, character fname)
   {
      writeToFile(data, fname, false);
   }
   
   /**
    * Write the full contents of the given binary object to the file specified by the given
    * filename.
    *
    * @param    data
    *           The data to write.
    * @param    fname
    *           The target filename.
    * @param    append
    *           <code>true</code> if the contents should be appended to the file.
    */
   public static void writeToFile(BinaryData data, character fname, boolean append)
   {
      if (fname == null || fname.isUnknown())
      {
         return;
      }
      
      writeToFile(data, fname.toStringMessage(), append);
   }
   
   /**
    * Write the full contents of the given binary object to the file specified by the given
    * filename.
    *
    * @param    data
    *           The data to write.
    * @param    fname
    *           The target filename.
    */
   public static void writeToFile(BinaryData data, String fname)
   {
      writeToFile(data, fname, false);
   }
   
   /**
    * Write the full contents of the given binary object to the file specified by the given
    * filename.
    *
    * @param    data
    *           The data to write.
    * @param    fname
    *           The target filename.
    * @param    append
    *           <code>true</code> if the contents should be appended to the file.
    */
   public static void writeToFile(BinaryData data, String fname, boolean append)
   {
      // TODO: this is just a first pass, the real/full behavior needs to be determined
      
      if (data == null || data.isUnknown() || fname == null || fname.length() == 0)
      {
         return;
      }
      
      Stream file     = StreamFactory.openFileStream(fname, true, append);
      byte[] contents = data.getByteArray();
      
      try
      {
         // empty array should yield an empty file
         if (contents.length > 0)
         {
            file.write(contents);
         }
      }
      
      catch (InterruptedIOException iioe)
      {
         throw new StopConditionException(iioe);
      }
      
      catch (IOException ioe)
      {
         ErrorManager.showErrorAndAbend(140, Stream.PIPE_BROKEN_MESSAGE);
      }
      
      finally
      {
         file.close();
      }
   }
   
   /**
    * Read some contents of the file and write that to the given longchar.
    *
    * @param    fname
    *           The target filename.
    * @param    idx
    *           The 1-based offset into the file from which to start reading.
    * @param    len
    *           The number of bytes to read.
    * @param    data
    *           The buffer in which to write.
    */
   public static void readFromFile(String fname, int idx, int len, longchar data)
   {
      // TODO: this is a simple placeholder that really should be replaced with a
      //       single call to a client-side worker that can handle all features, but
      //       until the feature set is fully explored we are implementing the simple
      //       way
      
      if (fname == null || fname.length() == 0 || data == null || data.isUnknown())
         return;
      
      Stream file = StreamFactory.openFileStream(fname, false, false);

      try
      {
         // seek to the 0-based offset
         file.setPos(idx - 1);
         
         // read the data
         byte[] chunk = file.readChunk(len);
         
         if (raiseErrorOnNullChar(chunk))
            return;
         
         // code-page conversion needs to happen here
         data.assign(new String(chunk));
      }
      catch (EndConditionException | ErrorConditionException | IOException iioe)
      {
         throw new StopConditionException(iioe);
      }
      finally
      {
         file.close();
      }
   }
   
   /**
    * Read some contents of the file and write that to the given longchar.
    *
    * @param    fname
    *           The target filename.
    * @param    idx
    *           The 1-based offset into the file from which to start reading.
    * @param    len
    *           The number of bytes to read.
    * @param    data
    *           The buffer in which to write.
    */
   public static void readFromFile(String fname, NumberType idx, int len, longchar data)
   {
      if (idx == null || idx.isUnknown())
         return;
      
      readFromFile(fname, idx.intValue(), len, data);
   }
   
   /**
    * Read some contents of the file and write that to the given longchar.
    *
    * @param    fname
    *           The target filename.
    * @param    idx
    *           The 1-based offset into the file from which to start reading.
    * @param    len
    *           The number of bytes to read.
    * @param    data
    *           The buffer in which to write.
    */
   public static void readFromFile(String fname, int idx, NumberType len, longchar data)
   {
      if (len == null || len.isUnknown())
         return;
      
      readFromFile(fname, idx, len.intValue(), data);
   }
   
   /**
    * Read some contents of the file and write that to the given longchar.
    *
    * @param    fname
    *           The target filename.
    * @param    idx
    *           The 1-based offset into the file from which to start reading.
    * @param    len
    *           The number of bytes to read.
    * @param    data
    *           The buffer in which to write.
    */
   public static void readFromFile(String fname, NumberType idx, NumberType len, longchar data)
   {
      if (idx == null || idx.isUnknown() || len == null || len.isUnknown())
         return;
      
      readFromFile(fname, idx.intValue(), len.intValue(), data);
   }
   
   /**
    * Read some contents of the file and write that to the given longchar.
    *
    * @param    fname
    *           The target filename.
    * @param    idx
    *           The 1-based offset into the file from which to start reading.
    * @param    len
    *           The number of bytes to read.
    * @param    data
    *           The buffer in which to write.
    */
   public static void readFromFile(character fname, int idx, int len, longchar data)
   {
      if (fname == null || fname.isUnknown())
         return;
      
      readFromFile(fname.toStringMessage(), idx, len, data);
   }
   
   /**
    * Read some contents of the file and write that to the given longchar.
    *
    * @param    fname
    *           The target filename.
    * @param    idx
    *           The 1-based offset into the file from which to start reading.
    * @param    len
    *           The number of bytes to read.
    * @param    data
    *           The buffer in which to write.
    */
   public static void readFromFile(character fname, NumberType idx, int len, longchar data)
   {
      if (idx == null || idx.isUnknown())
         return;
      
      readFromFile(fname, idx.intValue(), len, data);
   }
   
   /**
    * Read some contents of the file and write that to the given longchar.
    *
    * @param    fname
    *           The target filename.
    * @param    idx
    *           The 1-based offset into the file from which to start reading.
    * @param    len
    *           The number of bytes to read.
    * @param    data
    *           The buffer in which to write.
    */
   public static void readFromFile(character fname, int idx, NumberType len, longchar data)
   {
      if (len == null || len.isUnknown())
         return;
      
      readFromFile(fname, idx, len.intValue(), data);
   }
   
   /**
    * Read some contents of the file and write that to the given longchar.
    *
    * @param    fname
    *           The target filename.
    * @param    idx
    *           The 1-based offset into the file from which to start reading.
    * @param    len
    *           The number of bytes to read.
    * @param    data
    *           The buffer in which to write.
    */
   public static void readFromFile(character  fname,
                                   NumberType idx,
                                   NumberType len,
                                   longchar   data)
   {
      if (idx == null || idx.isUnknown() || len == null || len.isUnknown())
         return;
      
      readFromFile(fname, idx.intValue(), len.intValue(), data);
   }
   
   /**
    * Read the full contents of the file and write that to the given longchar.
    *
    * @param    fname
    *           The target filename.
    * @param    data
    *           The buffer in which to write.
    */
   public static void readFromFile(String fname, longchar data)
   {
// TODO: this is just a first pass, the real/full behavior should be implemented as following:
//      1. size(data) < size(file). For exapmle size(data) = 5 and size(file) = 23.
//      Progress reads all file and dynamically increases data size to needed value 
//      i.e. from 5 to 23 (in case of data).
//      Java impl only reads size(data) bytes from a file, i.e. 5
//      2. size(data) > size(file). For exapmle size(data) = 35 and size(file) = 23.
//      Both sides works equal except the one border case when the size of a file is 0.
//      In that case Progress works correctly without errors but Java implementation 
//      raises EndConditionException.
      
      if (data == null || data.isUnknown() || fname == null)
      {
         return;
      }

      Stream file = StreamFactory.openFileStream(fname, false, false);

      try
      {
         int64 len = file.getLength();
         
         // read the data
         byte[] chunk = file.readChunk(len.intValue());

         
         if (raiseErrorOnNullChar(chunk))
            return;
         
         // code-page conversion needs to happen here
         data.assign(new String(chunk));
      }
      catch (EndConditionException | ErrorConditionException | IOException iioe)
      {
         throw new StopConditionException(iioe);
      }
      finally
      {
         file.close();
      }
   }

   /**
    * Read the full contents of the file and write that to the given longchar.
    *
    * @param    fname
    *           The target filename.
    * @param    data
    *           The buffer in which to write.
    */
   public static void readFromFile(character fname, longchar data)
   {
      if (fname == null || fname.isUnknown())
      {
         return;
      }
      readFromFile(fname.toStringMessage(), data);
   }

   /**
    * Write the full contents of the given longchar to the file specified by the given
    * filename.
    *
    * @param    data
    *           The data to write.
    * @param    fname
    *           The target filename.
    */
   public static void writeToFile(longchar data, character fname)
   {
      writeToFile(data, fname, false);
   }
   
   /**
    * Write the full contents of the given longchar to the file specified by the given
    * filename.
    *
    * @param    data
    *           The data to write.
    * @param    fname
    *           The target filename.
    * @param    append
    *           <code>true</code> if the contents should be appended to the file.
    */
   public static void writeToFile(longchar data, character fname, boolean append)
   {
      if (fname == null || fname.isUnknown())
      {
         return;
      }
      
      writeToFile(data, fname.toStringMessage(), append);
   }
   
   /**
    * Write the full contents of the given longchar to the file specified by the given
    * filename.
    *
    * @param    data
    *           The data to write.
    * @param    fname
    *           The target filename.
    */
   public static void writeToFile(longchar data, String fname)
   {
      writeToFile(data, fname, false);
   }
   
   /**
    * Write the full contents of the given longchar to the file specified by the given
    * filename.
    *
    * @param    data
    *           The data to write.
    * @param    fname
    *           The target filename.
    * @param    append
    *           <code>true</code> if the contents should be appended to the file.
    */
   public static void writeToFile(longchar data, String fname, boolean append)
   {
      // TODO: this is just a first pass, the real/full behavior needs to be determined
      
      if (data == null || data.isUnknown() || fname == null || fname.length() == 0)
      {
         return;
      }
      
      Stream file     = StreamFactory.openFileStream(fname, true, append);
      String contents = data.toStringMessage();
      
      try
      {
         // empty array should yield an empty file
         if (contents.length() > 0)
         {
            // TODO: handle charset conversion as needed
            file.write(contents);
         }
      }
      
      catch (InterruptedIOException iioe)
      {
         throw new StopConditionException(iioe);
      }
      
      catch (IOException ioe)
      {
         ErrorManager.showErrorAndAbend(140, Stream.PIPE_BROKEN_MESSAGE);
      }
      
      finally
      {
         file.close();
      }
   }
   
   /**
    * Raise a 12012 error when a <code>null</code> character is encountered.
    *
    * @param    data
    *           The bytes to check.
    *
    * @return   <code>true</code> if the error was raised. This may happen in a NO-ERROR
    *           case, which means the caller should not execute further.
    */
   private static boolean raiseErrorOnNullChar(byte[] data)
   {
      // TODO: the codepage name is hardcoded but should be resolved dynamically
      for (int i = 0; i < data.length; i++)
      {
         if (data[i] == 0)
         {
            String msg = "Invalid character data found in MEMPTR for codepage ISO8859-15";
            ErrorManager.recordOrThrowError(12012, msg);
            return true;
         }
      }
      
      return false;
   }
}