/*
** Module   : Launcher.java
** Abstract : process launching API
**
** Copyright (c) 2006-2017, Golden Code Development Corporation.
**
** -#- -I- --Date-- -T- --JPRM-- ----------------Description-----------------
** 001 GES 20060212 NEW   @24562 Provides 2 APIs: a launch that has its
**                               STDIO connected to the terminal and a launch
**                               which has STDIO redirected to and/or from
**                               streams that are accessible from the caller.
** 002 GES 20061020 CHG   @30577 Return lastkey from interactive form of
**                               launch since any pause that occurs during
**                               that time will update lastkey.
** 012 EVL 20171129              Adding support for NO-CONSOLE option. 
*/
/*
** This program is free software: you can redistribute it and/or modify
** it under the terms of the GNU Affero General Public License as
** published by the Free Software Foundation, either version 3 of the
** License, or (at your option) any later version.
**
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
** GNU Affero General Public License for more details.
**
** You may find a copy of the GNU Affero GPL version 3 at the following
** location: https://www.gnu.org/licenses/agpl-3.0.en.html
** 
** Additional terms under GNU Affero GPL version 3 section 7:
** 
**   Under Section 7 of the GNU Affero GPL version 3, the following additional
**   terms apply to the works covered under the License.  These additional terms
**   are non-permissive additional terms allowed under Section 7 of the GNU
**   Affero GPL version 3 and may not be removed by you.
** 
**   0. Attribution Requirement.
** 
**     You must preserve all legal notices or author attributions in the covered
**     work or Appropriate Legal Notices displayed by works containing the covered
**     work.  You may not remove from the covered work any author or developer
**     credit already included within the covered work.
** 
**   1. No License To Use Trademarks.
** 
**     This license does not grant any license or rights to use the trademarks
**     Golden Code, FWD, any Golden Code or FWD logo, or any other trademarks
**     of Golden Code Development Corporation. You are not authorized to use the
**     name Golden Code, FWD, or the names of any author or contributor, for
**     publicity purposes without written authorization.
** 
**   2. No Misrepresentation of Affiliation.
** 
**     You may not represent yourself as Golden Code Development Corporation or FWD.
** 
**     You may not represent yourself for publicity purposes as associated with
**     Golden Code Development Corporation, FWD, or any author or contributor to
**     the covered work, without written authorization.
** 
**   3. No Misrepresentation of Source or Origin.
** 
**     You may not represent the covered work as solely your work.  All modified
**     versions of the covered work must be marked in a reasonable way to make it
**     clear that the modified work is not originating from Golden Code Development
**     Corporation or FWD.  All modified versions must contain the notices of
**     attribution required in this license.
*/

package com.goldencode.p2j.util;

/**
 * Process launching API with 2 forms: a launch that has its STDIO connected
 * to the terminal and a launch which has STDIO redirected to and/or from
 * streams that are accessible from the caller.
 * <p>
 * This API is design to provide a clean mechanism for remote process
 * launching (launching that occurs on a remote network node).
 *
 * @author    GES
 */
public interface Launcher
{   
   /**
    * Launches the external executible specified by the <code>cmdlist</code>
    * taking into account the flags which specify synchronous or asynchronous
    * operation and whether a prompt is provided after the child process
    * terminates. The command is always executed as a child process of the 
    * shell which is a child process of the JVM process on the node in which
    * this is served. 
    * <p>
    * <code>STDIN</code>, <code>STDOUT</code> and <code>STDERR</code> are
    * all linked from the child process to the current JVM process using
    * a <code>StreamConnector</code> for each pipe.
    *
    * @param    cmdlist
    *           The command line to be executed, with the first element
    *           used as the program to run and all other array elements as
    *           the arguments.
    * @param    silent
    *           If <code>true</code> the default prompt is suppressed.
    *           Otherwise, <code>ThinClient.pause</code> is called when the
    *           child process exits.
    * @param    wait
    *           If <code>true</code>, synchronous operation is requested
    *           which waits for the termination of the child process on the
    *           current thread.  If <code>false</code>, a thread will be
    *           started to wait for the child process termination and to
    *           subsequently handle cleanup (asynchronous mode).
    * @param    console
    *           If <code>true</code>, the process is starting in a separate new console,
    *           <code>false</code> means using 4GL no-console option.
    *
    * @return   The key code for any key pressed during any pause or -1 if
    *           there was no pause.
    */
   public int launch(String[] cmdlist, boolean silent, boolean wait, boolean console);
      
   /**
    * Launches the external executible specified by the <code>cmdlist</code>
    * on an asynchronous basis. The command is always executed as a child
    * process of the shell which is a child process of the the JVM process on
    * the node in which this is served.
    * <p>
    * <code>STDIN</code>, <code>STDOUT</code> and <code>STDERR</code> are
    * all linked from the child process to the passed parameters if
    * those parameters are not <code>null</code>. Any <code>null</code>
    * pipes are closed immediately and those pipes connected with the
    * passed parameters must be closed by the caller.  Please see the
    * {@link ProcessStream} for more details.
    *
    * @param    cmdlist
    *           The command line to be executed, with the first element
    *           used as the program to run and all other array elements as
    *           the arguments.
    * @param    sout
    *           The id of a {@link ProcessStream} instance to be connected to  
    *           the child process' <code>STDOUT and STDERR</code> or
    *           -1 for no connection.  Used for reading.
    * @param    sin
    *           The id of a <code>ProcessStream</code> instance to be
    *           connected to the child process' <code>STDIN</code> or
    *           -1 for no connection.  Used for writing.
    */
   public void launch(String[] cmdlist, int sout, int sin);
}