/*
** Module   : LegacyLogOps.java
** Abstract : Factory to create and manage the LOG-MANAGER and DSLOG-MANAGER instances.
**
** Copyright (c) 2014-2017, Golden Code Development Corporation.
**
** -#- -I- --Date-- -----------------------Description------------------------
** 001 GES 20141029 Created initial version.
*/
/*
** This program is free software: you can redistribute it and/or modify
** it under the terms of the GNU Affero General Public License as
** published by the Free Software Foundation, either version 3 of the
** License, or (at your option) any later version.
**
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
** GNU Affero General Public License for more details.
**
** You may find a copy of the GNU Affero GPL version 3 at the following
** location: https://www.gnu.org/licenses/agpl-3.0.en.html
** 
** Additional terms under GNU Affero GPL version 3 section 7:
** 
**   Under Section 7 of the GNU Affero GPL version 3, the following additional
**   terms apply to the works covered under the License.  These additional terms
**   are non-permissive additional terms allowed under Section 7 of the GNU
**   Affero GPL version 3 and may not be removed by you.
** 
**   0. Attribution Requirement.
** 
**     You must preserve all legal notices or author attributions in the covered
**     work or Appropriate Legal Notices displayed by works containing the covered
**     work.  You may not remove from the covered work any author or developer
**     credit already included within the covered work.
** 
**   1. No License To Use Trademarks.
** 
**     This license does not grant any license or rights to use the trademarks
**     Golden Code, FWD, any Golden Code or FWD logo, or any other trademarks
**     of Golden Code Development Corporation. You are not authorized to use the
**     name Golden Code, FWD, or the names of any author or contributor, for
**     publicity purposes without written authorization.
** 
**   2. No Misrepresentation of Affiliation.
** 
**     You may not represent yourself as Golden Code Development Corporation or FWD.
** 
**     You may not represent yourself for publicity purposes as associated with
**     Golden Code Development Corporation, FWD, or any author or contributor to
**     the covered work, without written authorization.
** 
**   3. No Misrepresentation of Source or Origin.
** 
**     You may not represent the covered work as solely your work.  All modified
**     versions of the covered work must be marked in a reasonable way to make it
**     clear that the modified work is not originating from Golden Code Development
**     Corporation or FWD.  All modified versions must contain the notices of
**     attribution required in this license.
*/

package com.goldencode.p2j.util;

import com.goldencode.p2j.security.*;

/**
 * Factory to create and manage the LOG-MANAGER and DSLOG-MANAGER instances. These instances
 * are created per-context so that each session has its own set, since in the 4GL the logging
 * will be different per client.
 */
public class LegacyLogOps
{
   /** Stores this context's log manager instances. */
   private static final ContextLocal<WorkArea> local =
   new ContextLocal<WorkArea>()
   {
      protected WorkArea initialValue()
      {
         return new WorkArea();
      };
   };
   
   /**
    * Disables construction of instances of this object.  All access is via static methods.
    */
   private LegacyLogOps()
   {
   }
   
   /**
    * Returns the instance of the current session's <code>LOG-MANAGER</code> resource.
    * 
    * @return   The <code>LOG-MANAGER</code> instance.
    */
   public static LegacyLogManager logMgr()
   {
      WorkArea wa = local.get();
      
      LegacyLogManager lm = null;
      
      synchronized (wa)
      {
         if (wa.logMgr == null)
         {
            wa.logMgr = new LegacyLogManagerImpl(false);
         }
         
         lm = wa.logMgr;
      }
      
      return lm;
   }
   
   /**
    * Returns the instance of the current session's <code>LOG-MANAGER</code> system handle.
    * 
    * @return   The <code>LOG-MANAGER</code> system handle.
    */
   public static handle logMgrHandle()
   {
      return new handle(logMgr());
   }
   
   /**
    * Returns the instance of the current session's <code>DSLOG-MANAGER</code> resource.
    * 
    * @return   The <code>DSLOG-MANAGER</code> instance.
    */
   public static LegacyLogManager dsLogMgr()
   {
      WorkArea wa = local.get();
      
      LegacyLogManager lm = null;
      
      synchronized (wa)
      {
         if (wa.dsLogMgr == null)
         {
            wa.dsLogMgr = new LegacyLogManagerImpl(true);
         }
         
         lm = wa.dsLogMgr;
      }
      
      return lm;
   }
   
   /**
    * Returns the instance of the current session's <code>DSLOG-MANAGER</code> system handle.
    * 
    * @return   The <code>DSLOG-MANAGER</code> system handle.
    */
   public static handle dsLogMgrHandle()
   {
      return new handle(dsLogMgr());
   }
   
   /** 
    * Stores global data relating to the state of the current context.
    */
   private static class WorkArea
   {
      /** Equivalent of the LOG-MANAGER system handle for this session. */
      private LegacyLogManager logMgr = null;
      
      /** Equivalent of the DSLOG-MANAGER system handle for this session. */
      private LegacyLogManager dsLogMgr = null;
   }
}