/*
** Module   : OutputParameter.java
** Abstract : Data wrapper which associates with a variable or a DMO property
**
** Copyright (c) 2016-2017, Golden Code Development Corporation.
**
** -#- -I- --Date-- ---------------------------------Description---------------------------------
** 001 OM  20160114 Created initial version. 
*/
/*
** This program is free software: you can redistribute it and/or modify
** it under the terms of the GNU Affero General Public License as
** published by the Free Software Foundation, either version 3 of the
** License, or (at your option) any later version.
**
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
** GNU Affero General Public License for more details.
**
** You may find a copy of the GNU Affero GPL version 3 at the following
** location: https://www.gnu.org/licenses/agpl-3.0.en.html
** 
** Additional terms under GNU Affero GPL version 3 section 7:
** 
**   Under Section 7 of the GNU Affero GPL version 3, the following additional
**   terms apply to the works covered under the License.  These additional terms
**   are non-permissive additional terms allowed under Section 7 of the GNU
**   Affero GPL version 3 and may not be removed by you.
** 
**   0. Attribution Requirement.
** 
**     You must preserve all legal notices or author attributions in the covered
**     work or Appropriate Legal Notices displayed by works containing the covered
**     work.  You may not remove from the covered work any author or developer
**     credit already included within the covered work.
** 
**   1. No License To Use Trademarks.
** 
**     This license does not grant any license or rights to use the trademarks
**     Golden Code, FWD, any Golden Code or FWD logo, or any other trademarks
**     of Golden Code Development Corporation. You are not authorized to use the
**     name Golden Code, FWD, or the names of any author or contributor, for
**     publicity purposes without written authorization.
** 
**   2. No Misrepresentation of Affiliation.
** 
**     You may not represent yourself as Golden Code Development Corporation or FWD.
** 
**     You may not represent yourself for publicity purposes as associated with
**     Golden Code Development Corporation, FWD, or any author or contributor to
**     the covered work, without written authorization.
** 
**   3. No Misrepresentation of Source or Origin.
** 
**     You may not represent the covered work as solely your work.  All modified
**     versions of the covered work must be marked in a reasonable way to make it
**     clear that the modified work is not originating from Golden Code Development
**     Corporation or FWD.  All modified versions must contain the notices of
**     attribution required in this license.
*/

package com.goldencode.p2j.util;

import com.goldencode.p2j.persist.*;

/**
 * A utility class which associates a variable or DMO property with a data wrapper object. 
 * Static methods of this class are invoked by business logic to manage the hidden assignment of 
 * an INPUT-OUTPUT or an OUTPUT parameter to original variable or the database field with which
 * it is associated, upon return from a procedure or function.
 */
public final class OutputParameter
{
   /**
    * Creates variable wrapper and returns instance of T type. This form can be used only with 
    * OUTPUT parameters.
    *
    * @param   variable
    *          The variable to be wrapped.
    *
    * @return  instance of the T type.
    */
   public static <T extends BaseDataType> T wrap(T variable)
   {
      return wrap(variable, false);
   }
   
   /**
    * Creates wrapper and returns instance of T type. This form can be used only with both
    * OUTPUT and INPUT-OUTPUT parameters (depending on {@code io} parameter).
    *
    * @param   variable
    *          The variable to be wrapped.
    * @param   io
    *          Specifies an INPUT-OUTPUT parameter if {@code true}. In this case, the local
    *          variable in the called function will be initialized with the current value of the
    *          original variable. Otherwise (simple OUTPUT), the parameter is initialized to
    *          default value for respective datatype (follows the semantics of OUTPUT parameter
    *          passing from 4GL).
    *
    * @return  instance of the T type.
    */
   public static <T extends BaseDataType> T wrap(T variable, boolean io)
   {
      T wrapper = (T) (io ? variable.duplicate() : variable.instantiateDefault());
   
      new VariableAssigner(wrapper, variable);
      return wrapper;
   }
   
   /**
    * Creates wrapper and returns instance of T type.
    *
    * @param   dmo
    *          Proxy for record whose field is being accessed.
    * @param   property
    *          Name of dmo field.
    * @param   bdt
    *          The {@link BaseDataType} of the field. The returned value will have this datatype.
    *          This argument can be detected dynamically at runtime but, for performance reasons,
    *          it is computed and passed in at conversion time. 
    *
    * @return  instance of the T type.
    */
   public static <T extends BaseDataType> T wrap(DataModelObject dmo, String property, Class bdt)
   {
      return wrap(dmo, property, bdt, null);
   }
   
   /**
    * Creates wrapper and returns instance of T type.
    *
    * @param   dmo
    *          Proxy for record whose field is being accessed.
    * @param   property
    *          Name of dmo field.
    * @param   bdt
    *          The {@link BaseDataType} of the field. The returned value will have this datatype.
    *          This argument can be detected dynamically at runtime but, for performance reasons,
    *          it is computed and passed in at conversion time. 
    * @param   index
    *          Index to access, if property represents an extent field.
    *
    * @return  instance of the T type.
    */
   public static <T extends BaseDataType> T wrap(DataModelObject dmo, 
                                                 String property, 
                                                 Class bdt, 
                                                 int index)
   {
      return wrap(dmo, property, bdt, new integer(index));
   }
   
   /**
    * Creates wrapper and returns instance of T type.
    *
    * @param   dmo
    *          Proxy for record whose field is being accessed.
    * @param   property
    *          Name of dmo field.
    * @param   bdt
    *          The {@link BaseDataType} of the field. The returned value will have this datatype.
    *          This argument can be detected dynamically at runtime but, for performance reasons,
    *          it is computed and passed in at conversion time. 
    * @param   index
    *          Index to access, if property represents an extent field.
    *
    * @return  instance of the T type.
    */
   public static <T extends BaseDataType> T wrap(DataModelObject dmo, 
                                                 String property, 
                                                 Class bdt, 
                                                 NumberType index)
   {
      return wrap(dmo, property, bdt, index, false);
   }
   
   /**
    * Creates wrapper and returns instance of T type.
    *
    * @param   dmo
    *          Proxy for record whose field is being accessed.
    * @param   property
    *          Name of dmo field.
    * @param   bdt
    *          The {@link BaseDataType} of the field. The returned value will have this datatype.
    *          This argument can be detected dynamically at runtime but, for performance reasons,
    *          it is computed and passed in at conversion time. 
    * @param   io
    *          Specifies an INPUT-OUTPUT parameter if {@code true}. In this case, the local
    *          variable in the called function will be initialized with the current value of the
    *          original variable. Otherwise (simple OUTPUT), the parameter is initialized to
    *          default value for respective datatype (follows the semantics of OUTPUT parameter
    *          passing from 4GL).
    *
    * @return  instance of the T type.
    */
   public static <T extends BaseDataType> T wrap(DataModelObject dmo, 
                                                 String property, 
                                                 Class bdt, 
                                                 boolean io)
   {
      return wrap(dmo, property, bdt, null, io);
   }
   
   /**
    * Creates wrapper and returns instance of T type.
    *
    * @param   dmo
    *          Proxy for record whose field is being accessed.
    * @param   property
    *          Name of dmo field.
    * @param   bdt
    *          The {@link BaseDataType} of the field. The returned value will have this datatype.
    *          This argument can be detected dynamically at runtime but, for performance reasons,
    *          it is computed and passed in at conversion time. 
    * @param   index
    *          Index to access, if property represents an extent field.
    * @param   io
    *          Specifies an INPUT-OUTPUT parameter if {@code true}. In this case, the local
    *          variable in the called function will be initialized with the current value of the
    *          original variable. Otherwise (simple OUTPUT), the parameter is initialized to
    *          default value for respective datatype (follows the semantics of OUTPUT parameter
    *          passing from 4GL).
    *
    * @return  instance of the T type.
    */
   public static <T extends BaseDataType> T wrap(DataModelObject dmo, 
                                                 String property, 
                                                 Class bdt, 
                                                 int index, 
                                                 boolean io)
   {
      return wrap(dmo, property, bdt, new integer(index), io);
   }
   
   /**
    * Creates wrapper and returns instance of T type.
    *
    * @param   dmo
    *          Proxy for record whose field is being accessed.
    * @param   property
    *          Name of dmo field.
    * @param   bdt
    *          The {@link BaseDataType} of the field. The returned value will have this datatype.
    *          This argument can be detected dynamically at runtime but, for performance reasons,
    *          it is computed and passed in at conversion time. 
    * @param   index
    *          Index to access, if property represents an extent field.
    * @param   io
    *          Specifies an INPUT-OUTPUT parameter if {@code true}. In this case, the local
    *          variable in the called function will be initialized with the current value of the
    *          original variable. Otherwise (simple OUTPUT), the parameter is initialized to
    *          default value for respective datatype (follows the semantics of OUTPUT parameter
    *          passing from 4GL).
    *
    * @return  instance of the T type.
    */
   public static <T extends BaseDataType> T wrap(DataModelObject dmo, 
                                                 String property, 
                                                 Class bdt, 
                                                 NumberType index, 
                                                 boolean io)
   {
      T wrapper = null;
      try
      {
         wrapper = (T) bdt.newInstance();
      }
      catch (InstantiationException | IllegalAccessException e)
      {
         e.printStackTrace();
      }
      new FieldAssigner(wrapper, dmo, property, index, io);
      return wrapper;
   }
}