/*
** Module   : longchar.java
** Abstract : Progress 4GL compatible longchar object
**
** Copyright (c) 2012-2017, Golden Code Development Corporation.
**
** -#- -I- --Date-- --------------------------------Description--------------------------------
** 001 AIL 20121206 Created initial version to support for Progress LONGCHAR type.
** 002 GES 20130322 Added the BDT c'tor.
** 003 VIG 20131003 Overrided instantiateUnknown method to return longchar
** 004 MAG 20140602 Fixed the BDT c'tor and modified the assign(BDT,boolean) on _POLY morphing.
**                  Override compareTo() method for _POLY morphing.
** 005 GES 20141003 Added overridden method to disable assignments of values with embedded
**                  null characters.
** 006 CA  20151222 Added instantiateDefaultExtent.
** 007 GES 20171207 Removed forced version of assign().
*/
/*
** This program is free software: you can redistribute it and/or modify
** it under the terms of the GNU Affero General Public License as
** published by the Free Software Foundation, either version 3 of the
** License, or (at your option) any later version.
**
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
** GNU Affero General Public License for more details.
**
** You may find a copy of the GNU Affero GPL version 3 at the following
** location: https://www.gnu.org/licenses/agpl-3.0.en.html
** 
** Additional terms under GNU Affero GPL version 3 section 7:
** 
**   Under Section 7 of the GNU Affero GPL version 3, the following additional
**   terms apply to the works covered under the License.  These additional terms
**   are non-permissive additional terms allowed under Section 7 of the GNU
**   Affero GPL version 3 and may not be removed by you.
** 
**   0. Attribution Requirement.
** 
**     You must preserve all legal notices or author attributions in the covered
**     work or Appropriate Legal Notices displayed by works containing the covered
**     work.  You may not remove from the covered work any author or developer
**     credit already included within the covered work.
** 
**   1. No License To Use Trademarks.
** 
**     This license does not grant any license or rights to use the trademarks
**     Golden Code, FWD, any Golden Code or FWD logo, or any other trademarks
**     of Golden Code Development Corporation. You are not authorized to use the
**     name Golden Code, FWD, or the names of any author or contributor, for
**     publicity purposes without written authorization.
** 
**   2. No Misrepresentation of Affiliation.
** 
**     You may not represent yourself as Golden Code Development Corporation or FWD.
** 
**     You may not represent yourself for publicity purposes as associated with
**     Golden Code Development Corporation, FWD, or any author or contributor to
**     the covered work, without written authorization.
** 
**   3. No Misrepresentation of Source or Origin.
** 
**     You may not represent the covered work as solely your work.  All modified
**     versions of the covered work must be marked in a reasonable way to make it
**     clear that the modified work is not originating from Golden Code Development
**     Corporation or FWD.  All modified versions must contain the notices of
**     attribution required in this license.
*/

package com.goldencode.p2j.util;

/**
 * A class that represents a Progress 4GL compatible longchar object whose data is mutable.
 * <p>
 * Although Progress documentation states that the default value of any <code>longchar</code>
 * variable is the unknown value, the real behavior is that it gets initialized to the empty
 * string, same as character variables get initialized.
 */
public class longchar
extends Text
{
   /**
    * Default constructor, creates an instance that represents the unknown value.
    */
   public longchar()
   {
      super();
   }
   
   /**
    * This is a special c'tor which should be used only when converting the
    * value returned by a function or method with polymorphic return type into the
    * expected type (i.e. DYNAMIC-FUNCTION()).  In such cases, the 4GL does some
    * automatic type conversion (this is delegated to the superclass' constructor).
    * <p>
    * Only Text (character or longchar) BDT's can be converted to longchar.
    * 
    * @param    value
    *           The value to be used for this instance.
    */
   public longchar(BaseDataType value)
   {
      if (value == null || value.isUnknown())
      {
         setUnknown();
      }
      else
      {
         if (value instanceof Text)
         {
            assign(value);
         }
         else
         {
            incompatibleTypesOnConversion();
         }
      }
   }
   
   /**
    * Constructs an instance that has the <code>String</code> value passed
    * in and which <b>does not</b> represent the <code>unknown value</code>,
    * unless the <code>value</code> argument is <code>null</code>.
    *
    * @param    value
    *           The character value of the resulting instance.
    */
   public longchar(String value)
   {
      super(value);
   }
   
   /**
    * Constructs an instance that has the <code>String</code> value passed
    * in and which <b>does not</b> represent the <code>unknown value</code>
    * unless the <code>value</code> argument is <code>null</code>.
    *
    * @param    value
    *           The character value of the resulting instance.
    * @param    caseSens
    *           Specifies whether this instance should use case sensitive
    *           comparisons.
    */
   public longchar(String value, boolean caseSens)
   {
      super(value, caseSens);
   }
   
   /**
    * Constructs an instance that has a <code>char</code> value passed
    * in and which <b>does not</b> represent the <code>unknown value</code>. 
    *
    * @param    ch
    *           The character value of the resulting instance.
    */
   public longchar(char ch)
   {
      super(ch);
   }
   
   /**
    * Create a new instance, by copying the state of the passed {@link Text}.
    * 
    * @param    value
    *           The value which needs to be copied.
    */
   public longchar(Text value)
   {
      super(value, false);
   }

   /**
    * Create a new instance, by copying the state of the passed {@link Text}.
    * If the <code>output</code> parameter is set to <code>true</code>, it 
    * means that the referenced value is saved, as the passed instance will
    * be passed to a function call, and the state might need to be updated
    * on function exit. 
    * 
    * @param    value
    *           The value which needs to be copied.
    * @param    output
    *           Flag indicating if this instance is passed to as a reference
    *           for an output parameter.
    */
   public longchar(Text value, boolean output)
   {
      super(value, output);
   }

   /**
    * Report if existing instances that have embedded null characters are allowed to be
    * assigned to this instance.
    *
    * @return   Always <code>false</code>.
    */
   public boolean isEmbeddedNullAllowed()
   {
      return false;
   }

   /**
    * Compares this instance with the specified instance and returns a -1
    * if this instance is less than the specified, 0 if the two instances
    * are equal and 1 if this instance is greater than the specified
    * instance.  This is the implementation of the <code>Comparable</code>
    * interface.
    *
    * @param    obj
    *           The instance to compare against.
    *
    * @return   -1, 0 or 1 depending on if this instance is less than,
    *           equal to or greater (respectively) than the specified
    *           instance <code>obj</code>.
    */
   public int compareTo(Object obj)   
   {
      if (obj instanceof BaseDataType)
      {
         return super.compareTo(new longchar((BaseDataType) obj));         
      }

      return super.compareTo(obj);
   }   
   
   /**
    * Returns the code page of a the specified longchar variable. If the 
    * specified LONGCHAR is empty and the code page was not fixed using the 
    * FIX-CODEPAGE statement, must return the unknown value.
    * 
    * @param    value
    *           The value for which the codepage will be retrieved.
    *
    * @return   See above
    */
   public static character getCodePage(longchar value)
   {
      // TODO: IMPLEMENT THIS
      return null;
   }
   
   /**
    * Creates a new instance of the same type that represents the 
    * <code>unknown value</code>.
    *
    * @return   An instance that represents the <code>unknown value</code>.
    */
   public BaseDataType instantiateUnknown()   
   {
      return new longchar();
   }
   
   /**
    * Return <code>true</code> if the codepage of the given longchar variable 
    * is fixed.
    * 
    * @param    value
    *           The value for which the codepage will be retrieved.
    *           
    * @return   See above.
    */
   public static logical isCodePageFixed(longchar value)
   {
      // TODO: IMPLEMENT THIS
      return null;
   }
   
   /**
    * Return the default display format string for this type.
    *
    * @return   The default format string.
    */
   public String defaultFormatString()   
   {
      return null;
   }
   

   /**
    * Does the same as standard <code>clone()</code> method but returns an
    * instance of <code>BaseDataType</code> and doesn't throw the
    * <code>CloneNotSupportedException</code>.
    *
    * @return   A clone of this instance.  
    */
   public BaseDataType duplicate()
   {
      return new longchar(this);
   }

   /**
    * Creates a string representation of the instance data in a form that is
    * compatible with the <code>MESSAGE</code> language statement.  If the
    * instance represents the <code>unknown value</code>, a '?' will be
    * returned.
    *
    * @return   The 'message' formatted string.
    */
   public String toStringMessage()   
   {
      // TODO: IMPLEMENT THIS
      return getValue();
   }

   /**
    * Creates a string representation of the instance data using the 'export'
    * format.  For the character data type the entire string is returned
    * enclosed in a pair of double quotes.  If the instance represents the
    * <code>unknown value</code>, a '?' will be returned.  Any instances of
    * double quotes inside the string will be doubled up on output.
    *
    * @return   The 'export' formatted string.
    */
   public String toStringExport()   
   {
      // TODO: IMPLEMENT THIS
      return getValue();
   }

   /**
    * Creates a string representation of the instance data using the user-
    * specified Progress 4GL format string.
    *
    * @param    fmt
    *           The Progress 4GL format string for a character data type.
    *
    * @return   The formatted string.
    */
   public String toString(String fmt)
   {
      // TODO: IMPLEMENT THIS
      return null;
   }

   /**
    * Get the default initialization for an extent variable of this type.
    * 
    * @return   See above.
    */
   @Override
   protected BaseDataType instantiateDefaultExtent()
   {
      return new longchar("");
   }
}