/*
** Module   : AbstractModulesHandler.java
** Abstract : Delivers requested resources by loading them from the classpath
**
** Copyright (c) 2017, Golden Code Development Corporation.
**
** -#- -I- --Date-- ------------------------------Description----------------------------------
** 001 SBI 20170720 Generalized DojoToolkitHandler to be used to deliver js zipped modules.
*/
/*
** This program is free software: you can redistribute it and/or modify
** it under the terms of the GNU Affero General Public License as
** published by the Free Software Foundation, either version 3 of the
** License, or (at your option) any later version.
**
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
** GNU Affero General Public License for more details.
**
** You may find a copy of the GNU Affero GPL version 3 at the following
** location: https://www.gnu.org/licenses/agpl-3.0.en.html
** 
** Additional terms under GNU Affero GPL version 3 section 7:
** 
**   Under Section 7 of the GNU Affero GPL version 3, the following additional
**   terms apply to the works covered under the License.  These additional terms
**   are non-permissive additional terms allowed under Section 7 of the GNU
**   Affero GPL version 3 and may not be removed by you.
** 
**   0. Attribution Requirement.
** 
**     You must preserve all legal notices or author attributions in the covered
**     work or Appropriate Legal Notices displayed by works containing the covered
**     work.  You may not remove from the covered work any author or developer
**     credit already included within the covered work.
** 
**   1. No License To Use Trademarks.
** 
**     This license does not grant any license or rights to use the trademarks
**     Golden Code, FWD, any Golden Code or FWD logo, or any other trademarks
**     of Golden Code Development Corporation. You are not authorized to use the
**     name Golden Code, FWD, or the names of any author or contributor, for
**     publicity purposes without written authorization.
** 
**   2. No Misrepresentation of Affiliation.
** 
**     You may not represent yourself as Golden Code Development Corporation or FWD.
** 
**     You may not represent yourself for publicity purposes as associated with
**     Golden Code Development Corporation, FWD, or any author or contributor to
**     the covered work, without written authorization.
** 
**   3. No Misrepresentation of Source or Origin.
** 
**     You may not represent the covered work as solely your work.  All modified
**     versions of the covered work must be marked in a reasonable way to make it
**     clear that the modified work is not originating from Golden Code Development
**     Corporation or FWD.  All modified versions must contain the notices of
**     attribution required in this license.
*/


package com.goldencode.p2j.web;

import java.io.IOException;
import java.net.URL;
import java.net.URLConnection;

import javax.servlet.ServletException;
import javax.servlet.http.HttpServletRequest;
import javax.servlet.http.HttpServletResponse;

import org.eclipse.jetty.server.Request;


/**
 * This handler is responsible to deliver to the client resources stored in a zip file. It is
 * supposed that this zip file containing target resources is available for the context class
 * loader of the thread executing a current request.
 */
public abstract class AbstractModulesHandler extends AbstractHandlerCommons
{

   /**
    * Returns true if this handler is responsible for this request, otherwise false.
    * 
    * 
    * @param    target 
    *           The target of the request - either a URI or a name.
    * @param    base
    *           The base request.
    * @param    request
    *           The object or a wrapper of the request.
    * 
    * @return   true if this handler must respond on this request
    */
   public abstract boolean isModuleRequest(String             target,
                                           Request            base,
                                           HttpServletRequest request);
   
   /**
    * Maps the target of the request to its resource name.
    * 
    * @param    target
    *           The target of the request - either a URI or a name.
    * 
    * @return   The target resource name
    */
   public abstract String mapToResource(String target);
   
   /**
    * Return a requested module if found or empty response and
    * <code>HttpServletResponse.SC_NOT_FOUND</code> if the module was not found.
    *    
    * @param   target 
    *          The target of the request - either a URI or a name.
    * @param   base
    *          The base request.
    * @param   request
    *          The object or a wrapper of the request.
    * @param   response 
    *          The object or a wrapper of the response. 
    */
   public void handle(String              target, 
                      Request             base,
                      HttpServletRequest  request, 
                      HttpServletResponse response)
   throws IOException, 
          ServletException
   {
      if (!base.isHandled() && isModuleRequest(target, base, request))
      {
         // mark request as handled
         base.setHandled(true);
         deliverModule(target, response);
      }
   }
   
   /**
    * Deliver a module directly from zip file that MUST be on CLASSPATH. In order to work properly
    * this method MUST be synchronized.
    *
    * @param    target
    *           The target of the request - either a URI or a name.
    * @param    response 
    *           The details and state of the response.
    *           
    * @throws   IOException
    *           ServletException
    *           If any error conditions occur.
    */
   private synchronized void deliverModule(String target, HttpServletResponse response)
   throws IOException, 
          ServletException
   {
      String resource = mapToResource(target);

      // load resource
      URL url = Thread.currentThread().getContextClassLoader().getResource(resource);

      if (url != null)
      {
         URLConnection conn = url.openConnection();
         sendResponse(response, conn, resource);
      }
      else
      {
         // module not found
         response.setStatus(HttpServletResponse.SC_NOT_FOUND);
      }
   }
}