/*
** Module   : WebResourceHandler.java
** Abstract : A basic handler for resources served from class path jars.
**
** Copyright (c) 2018, Golden Code Development Corporation.
**
** -#- -I- --Date-- ------------------------------Description----------------------------------
** 001 CA  20180723 First version.
** 002 CA  20180815 Added more JavaScript libraries (d3, tabulator, fonts, material components).
*/
/*
** This program is free software: you can redistribute it and/or modify
** it under the terms of the GNU Affero General Public License as
** published by the Free Software Foundation, either version 3 of the
** License, or (at your option) any later version.
**
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
** GNU Affero General Public License for more details.
**
** You may find a copy of the GNU Affero GPL version 3 at the following
** location: https://www.gnu.org/licenses/agpl-3.0.en.html
** 
** Additional terms under GNU Affero GPL version 3 section 7:
** 
**   Under Section 7 of the GNU Affero GPL version 3, the following additional
**   terms apply to the works covered under the License.  These additional terms
**   are non-permissive additional terms allowed under Section 7 of the GNU
**   Affero GPL version 3 and may not be removed by you.
** 
**   0. Attribution Requirement.
** 
**     You must preserve all legal notices or author attributions in the covered
**     work or Appropriate Legal Notices displayed by works containing the covered
**     work.  You may not remove from the covered work any author or developer
**     credit already included within the covered work.
** 
**   1. No License To Use Trademarks.
** 
**     This license does not grant any license or rights to use the trademarks
**     Golden Code, FWD, any Golden Code or FWD logo, or any other trademarks
**     of Golden Code Development Corporation. You are not authorized to use the
**     name Golden Code, FWD, or the names of any author or contributor, for
**     publicity purposes without written authorization.
** 
**   2. No Misrepresentation of Affiliation.
** 
**     You may not represent yourself as Golden Code Development Corporation or FWD.
** 
**     You may not represent yourself for publicity purposes as associated with
**     Golden Code Development Corporation, FWD, or any author or contributor to
**     the covered work, without written authorization.
** 
**   3. No Misrepresentation of Source or Origin.
** 
**     You may not represent the covered work as solely your work.  All modified
**     versions of the covered work must be marked in a reasonable way to make it
**     clear that the modified work is not originating from Golden Code Development
**     Corporation or FWD.  All modified versions must contain the notices of
**     attribution required in this license.
*/

package com.goldencode.p2j.web;

import javax.servlet.http.*;

import org.eclipse.jetty.server.*;

/**
 * Jetty Handler which delivers web resource modules.
 * <p>
 * TODO:
 * IMPORTANT: Don't use logging anywhere inside this class. When used in a web client, this runs
 * in conjunction with an authentication plugin during session establishment with the P2J server
 * and using logging the application will deadlock.
*/
public class WebResourceHandler
extends AbstractModulesHandler
{
   /** The parent context path. */
   private final String basePath;

   /** The context path for this resource. */
   private final String webRoot;
   
   /** The root folder (as is found in the class path jar's) for where to serve the resources. */
   private final String resPath;
   
   /**
    * Create a new resource handler with an empty parent context path.
    * 
    * @param    webRoot
    *           The context path for this resource.
    * @param    resPath
    *           The root folder (as is found in the class path jar's) for where to serve the 
    *           resources.
    */
   public WebResourceHandler(String webRoot, String resPath)
   {
      this("", webRoot, resPath);
   }
   
   /**
    * Create a new resource handler with an empty parent context path.
    *
    * @param    basePath
    *           The parent context path.
    * @param    webRoot
    *           The context path for this resource.
    * @param    resPath
    *           The root folder (as is found in the class path jar's) for where to serve the 
    *           resources.
    */
   public WebResourceHandler(String basePath, String webRoot, String resPath)
   {
      if (!webRoot.endsWith("/"))
      {
         webRoot = webRoot + "/";
      }
      if (!resPath.endsWith("/"))
      {
         resPath = resPath + "/";
      }

      this.basePath = basePath;
      this.webRoot = webRoot;
      this.resPath = resPath;
   }
   
   /**
    * Get all the web resource handlers exposed by FWD, via custom paths.
    * 
    * @param    basePath
    *           The parent context path.  May be empty.
    *           
    * @return   The list of handlers.
    */
   public static Handler[] getWebResourceHandlers(String basePath)
   {
      // in build.gradle, add a WARNING for each zip artifact which is hard-coded here.
      
      // do not expose the "/" root resource folder - as this will allow access to any class path
      // file, which includes jar'ed classes, configurations, etc
      return new Handler[]
      {
         new DojoToolkitHandler(basePath),
         new WebResourceHandler(basePath, "/bootstrap", "bootstrap-3.3.7-dist"),
         new WebResourceHandler(basePath, "/d3", "d3-4.9.1"),
         new WebResourceHandler(basePath, "/fonts", "fonts"),
         new WebResourceHandler(basePath, "/jquery", "jquery-3.2.1"),
         new WebResourceHandler(basePath, "/jquery-ui", "jquery-ui-1.12.1.custom"),
         new WebResourceHandler(basePath, "/material-components-web", "material-components-web"),
         new WebResourceHandler(basePath, "/material-design-icons", "material-design-icons"),
         new WebResourceHandler(basePath, "/tabulator", "tabulator-master-2.12.0")
      };
   }
   
   /**
    * Returns true if it is a request for this module, otherwise false.
    * 
    * 
    * @param    target 
    *           The target of the request - either a URI or a name.
    * @param    base
    *           The base request.
    * @param    request
    *           The object or a wrapper of the request.
    * 
    * @return   true if it is a request for this module.
    */
   @Override
   public boolean isModuleRequest(String target, Request base, HttpServletRequest request)
   {
      // check if this is a request for this module
      
      return target.toLowerCase().startsWith(basePath + webRoot);
   }

   /**
    * Maps the target of the request to its resource name.
    * 
    * @param    target
    *           The target of the request - either a URI or a name.
    * 
    * @return   The target resource name
    */
   @Override
   public String mapToResource(String target)
   {
      return target.replace(basePath + webRoot, resPath);
   }
}