/*
** Module   : Driver.java
** Abstract : thread which processes trees obtained from a queue of tree names 
**
** Copyright (c) 2009-2017, Golden Code Development Corporation.
**
** -#- -I- --Date-- --JPRM-- ----------------Description-----------------
** 001 GES 20090525   @42533 Thread which processes trees obtained from a queue
**                           of tree names.
*/
/*
** This program is free software: you can redistribute it and/or modify
** it under the terms of the GNU Affero General Public License as
** published by the Free Software Foundation, either version 3 of the
** License, or (at your option) any later version.
**
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
** GNU Affero General Public License for more details.
**
** You may find a copy of the GNU Affero GPL version 3 at the following
** location: https://www.gnu.org/licenses/agpl-3.0.en.html
** 
** Additional terms under GNU Affero GPL version 3 section 7:
** 
**   Under Section 7 of the GNU Affero GPL version 3, the following additional
**   terms apply to the works covered under the License.  These additional terms
**   are non-permissive additional terms allowed under Section 7 of the GNU
**   Affero GPL version 3 and may not be removed by you.
** 
**   0. Attribution Requirement.
** 
**     You must preserve all legal notices or author attributions in the covered
**     work or Appropriate Legal Notices displayed by works containing the covered
**     work.  You may not remove from the covered work any author or developer
**     credit already included within the covered work.
** 
**   1. No License To Use Trademarks.
** 
**     This license does not grant any license or rights to use the trademarks
**     Golden Code, FWD, any Golden Code or FWD logo, or any other trademarks
**     of Golden Code Development Corporation. You are not authorized to use the
**     name Golden Code, FWD, or the names of any author or contributor, for
**     publicity purposes without written authorization.
** 
**   2. No Misrepresentation of Affiliation.
** 
**     You may not represent yourself as Golden Code Development Corporation or FWD.
** 
**     You may not represent yourself for publicity purposes as associated with
**     Golden Code Development Corporation, FWD, or any author or contributor to
**     the covered work, without written authorization.
** 
**   3. No Misrepresentation of Source or Origin.
** 
**     You may not represent the covered work as solely your work.  All modified
**     versions of the covered work must be marked in a reasonable way to make it
**     clear that the modified work is not originating from Golden Code Development
**     Corporation or FWD.  All modified versions must contain the notices of
**     attribution required in this license.
*/

package com.goldencode.trpl; 
 
import java.util.*;
import java.util.concurrent.*;
import com.goldencode.ast.*;

/**
 * Implements the core logic for a tree processing thread.
 */
public class Driver
extends Thread
{
   /** Queue from which to obtain the next tree name to process. */
   private final ConcurrentLinkedQueue<String> queue;
   
   /** Callbacks to execute based on the events of the tree walk. */ 
   private final Pipeline pipeline;
   
   /** State shared by all callbacks (across multiple different trees). */
   private final BaseState base;
   
   /** Mechanism to notify the main engine thread of completion. */
   private final CountDownLatch latch;
   
   /** Provides core AST services. */
   private final AstManager mgr;
   
   /**
    * Construct a driver thread instance.
    *
    * @param    queue
    *           Queue from which to obtain the next tree name to process.
    * @param    pipeline
    *           Callbacks to execute based on the events of the tree walk.
    * @param    base
    *           Common state shared with all callbacks.
    * @param    latch
    *           Mechanism to notify the main engine thread of completion.
    */
   public Driver(ConcurrentLinkedQueue<String> queue,
                 Pipeline                      pipeline,
                 BaseState                     base,
                 CountDownLatch                latch)
   {
      this.queue    = queue;
      this.pipeline = pipeline;
      this.base     = base;
      this.latch    = latch;
      this.mgr      = base.mgr;
   }
   
   /**
    * Execute the core logic of the thread. This is where tree processing
    * occurs.
    */
   public void run()
   {
      // duplicate the walker instances to make a unique copy for this
      // thread
      List<WalkElement> code = pipeline.getWalkers();
      
      try
      {
         String tname = queue.poll();
         
         while (tname != null)
         {
            long id = mgr.getTreeId(tname);
            
            // check the tree's validity
            if (id != AstManager.INVALID_ID)
            {
               try
               {
                  // lock the tree
                  mgr.lockTree(id);
                  
                  // load the tree
                  Aast      here     = mgr.loadTree(tname);
                  WalkState state    = null;
                  Resolver  resolver = null;
                  
                  // TODO: remove this
                  System.out.println(here.getFilename());
               
                  // iterate through each walker in the pipeline
                  for (WalkElement elem : code)
                  {
                     Iterator iter = null;
                     
                     // obtain the iterator for traversal in the view case
                     if (elem.getMode() == InputMode.VIEW)
                     {
                        if (resolver == null)
                        {
                           // the resolver is only null for the first
                           // element, in such a case there cannot be any
                           // view
                           String err = "First walk element may not use " +
                                        "the VIEW input mode.";
                           throw new IllegalStateException(err);
                        }
                        
                        iter = resolver.view();
                     }
                     
                     // prepare the initial walk state for this element
                     state    = prepareState(resolver, here);
                     resolver = state.resolver;
                     
                     TreeWalker walker = elem.getWalker();
                     
                     // invoke the pre hook
                     walker.pre(state);
                     
                     // obtain the iterator for traversal in the tree case
                     if (elem.getMode() == InputMode.TREE)
                     {
                        // create the walk listener for this walker
                        AstWalkListener listener =
                           new WalkListener(base, resolver, walker);
                        
                        int     depth = elem.getMaxDepth();
                        boolean honor = !elem.isWalkHidden();
                           
                        iter = here.iterator(depth, listener, honor);
                     }
                     
                     // drive the main walk (in view mode, there are no
                     // ascent/descent/nextChild hooks; in tree mode those
                     // hooks will be invoked indirectly by the listener)
                     while (iter.hasNext())
                     {
                        Aast next = (Aast) iter.next();
                        Aast copy = resolver.getAst(next.getId());
                        
                        // update state for this specific node and set it
                        // into the walk listener
                        WalkState specific =
                           new WalkState(base, resolver, next, copy);
      
                        // invoke the main visit hook
                        walker.walk(specific);
                     }
                        
                     // invoke the post hook
                     walker.post(state);
                     
                     // save off previous state values before we iterate
                     here     = state.copy;
                     resolver = state.resolver;
                  }
               
                  // persist changes if needed
                  if (pipeline.isAutoSave() && !pipeline.isReadOnly())
                  {
                     mgr.saveTree(state.copy, tname, false);
                     
                     // TODO: get the list of loaded ASTs and save each one
                  }
               }
               
               finally
               {
                  // unlock the tree
                  mgr.unlockTree(id);
               }
            }
            else
            {
               // TODO: long this
               System.out.printf("Invalid tree name %s!\n", tname);
            }
            
            // read the next tree name
            tname = queue.poll();
         }
      }
      
      catch (Throwable t)
      {
         // TODO: log this
         t.printStackTrace();
      }
      
      finally
      {
         latch.countDown();
      }
   }
   
   /**
    * Setup the resolver and walk state for the walk of a tree at the
    * beginning of a walk element in the pipeline. In each such element,
    * the editable tree from the previous element is the new source tree.
    * Then a duplicate is made of the source tree as the editable tree. If
    * the pipeline is read-only, then the source and editable trees will be
    * the same. Either way, the resolver must be configured to have the
    * ID to node cache pre-loaded. Finally, the walk state must be
    * instantiated and returned.
    *
    * @param    resolver
    *           The resolver used in the previous walk element or 
    *           <code>null</code> for the first walk element.
    * @param    here
    *           This is the root node for the read-only "source" tree.
    *
    * @return   The walk state for the first 
    */
   private WalkState prepareState(Resolver resolver, Aast here)
   {
      Aast copy = here;
      
      // implement read-only support if needed
      if (!pipeline.isReadOnly())
      {
         Map<Long, Aast> idMap = new HashMap<Long, Aast>();
         
         // make a deep copy of the source AST which will be the
         // editable version
         copy = here.duplicate(idMap);
         
         // pass the map of the editable nodes to populate the resolver
         resolver = new Resolver(idMap);
      }
      else
      {
         // re-use the previously setup resolver if it already exists
         if (resolver == null)
         {
            // map the read-only tree since there is no duplicate tree
            resolver = new Resolver(null);
            resolver.registerTree(here);
         }
         else
         {
            // reset the view but keep the rest of the node mapping cache
            resolver.reset();
         }
      }
      
      return new WalkState(base, resolver, here, copy);
   }
}