/*
** Module   : FibonacciCounter.java
** Abstract : A 64-bit counter which generates numbers of a Fibonacci series.
**
** Copyright (c) 2016-2017, Golden Code Development Corporation.
**
** -#- -I- --Date-- --------------------------------Description----------------------------------
** 001 ECF 20160820 Created initial version as a simple way to generate progressively larger
**                  integer values.
*/
/*
** This program is free software: you can redistribute it and/or modify
** it under the terms of the GNU Affero General Public License as
** published by the Free Software Foundation, either version 3 of the
** License, or (at your option) any later version.
**
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
** GNU Affero General Public License for more details.
**
** You may find a copy of the GNU Affero GPL version 3 at the following
** location: https://www.gnu.org/licenses/agpl-3.0.en.html
** 
** Additional terms under GNU Affero GPL version 3 section 7:
** 
**   Under Section 7 of the GNU Affero GPL version 3, the following additional
**   terms apply to the works covered under the License.  These additional terms
**   are non-permissive additional terms allowed under Section 7 of the GNU
**   Affero GPL version 3 and may not be removed by you.
** 
**   0. Attribution Requirement.
** 
**     You must preserve all legal notices or author attributions in the covered
**     work or Appropriate Legal Notices displayed by works containing the covered
**     work.  You may not remove from the covered work any author or developer
**     credit already included within the covered work.
** 
**   1. No License To Use Trademarks.
** 
**     This license does not grant any license or rights to use the trademarks
**     Golden Code, FWD, any Golden Code or FWD logo, or any other trademarks
**     of Golden Code Development Corporation. You are not authorized to use the
**     name Golden Code, FWD, or the names of any author or contributor, for
**     publicity purposes without written authorization.
** 
**   2. No Misrepresentation of Affiliation.
** 
**     You may not represent yourself as Golden Code Development Corporation or FWD.
** 
**     You may not represent yourself for publicity purposes as associated with
**     Golden Code Development Corporation, FWD, or any author or contributor to
**     the covered work, without written authorization.
** 
**   3. No Misrepresentation of Source or Origin.
** 
**     You may not represent the covered work as solely your work.  All modified
**     versions of the covered work must be marked in a reasonable way to make it
**     clear that the modified work is not originating from Golden Code Development
**     Corporation or FWD.  All modified versions must contain the notices of
**     attribution required in this license.
*/

package com.goldencode.util;

/**
 * A counter which provides the numbers of a Fibonacci sequence, starting with [1, 1] and
 * progressing indefinitely (practically limited by a 64-bit integer implementation -- values
 * higher than can fit into this data type will overflow and become meaningless).
 * <p>
 * An integer factor can be applied at construction, so that the numbers returned are a multiple
 * of the factor. The default factor is 1.
 */
public final class FibonacciCounter
{
   /** Factor to multiply with each number in the series */
   private final int factor;
   
   /** Previous number in the series */
   private long previous = 0L;
   
   /** Current number in the series */
   private long current = 1L;
   
   /**
    * Default constructor, which uses a factor of 1.
    */
   public FibonacciCounter()
   {
      this(1);
   }
   
   /**
    * Constructor.
    * 
    * @param   factor
    *          Factor to multiply with each number in the series. Does not change the series,
    *          only the number reported.
    */
   public FibonacciCounter(int factor)
   {
      this.factor = factor;
   }
   
   /**
    * Report the current number in the series, multiplied by the factor.
    * 
    * @return  Current number, multiplied by factor.
    */
   public long current()
   {
      return current * factor;
   }
   
   /**
    * Advance to the next number in the series, making it the current number, and report it
    * multiplied by the factor. Subsequent calls to {@link #current()} will report the same
    * number reported by this method, until this method is invoked again.
    * 
    * @return  Next number, multiplied by factor.
    */
   public long next()
   {
      long temp = current;
      current += previous;
      previous = temp;
      
      return current();
   }
   
   /**
    * Reset the series back to the starting number of 1.
    */
   public void reset()
   {
      previous = 0L;
      current = 1L;
   }
   
   /**
    * Syntax: {@code FibonacciCounter [count [factor]]}, where {@code count} is a positive
    * number of sequence elements to display and {@code factor} is the factor to apply to each
    * element.
    * 
    * @param   args
    *          See above.
    */
   public static void main(String[] args)
   {
      int count = 10;
      int factor = 1;
      
      if (args.length > 0)
      {
         count = Math.max(1, Integer.parseInt(args[0]));
      }
      
      if (args.length > 1)
      {
         factor = Integer.parseInt(args[1]);
      }
      
      FibonacciCounter fib = new FibonacciCounter(factor);
      System.out.print(fib.current());
      for (int i = 1; i < count; i++)
      {
         System.out.print(", " + fib.next());
      }
      
      System.out.println();
   }
}