/*
** Module   : Mapper.java
** Abstract : Bind XML nodes to arbitrary Java objects.
**
** Copyright (c) 2010-2017, Golden Code Development Corporation.
**
** -#- -I- --Date-- --JPRM-- ------------------Description-------------------
** 001 LMR 20110103          Created initial version; supports recursive
**                           composite structures and context sensitive type 
**                           mapping (tags that map to different Java types 
**                           depending on where in the XML tree they appear).
** 002 EVL 20160224          Javadoc fixes to make compatible with Oracle Java 8 for Solaris 10.
*/
/*
** This program is free software: you can redistribute it and/or modify
** it under the terms of the GNU Affero General Public License as
** published by the Free Software Foundation, either version 3 of the
** License, or (at your option) any later version.
**
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
** GNU Affero General Public License for more details.
**
** You may find a copy of the GNU Affero GPL version 3 at the following
** location: https://www.gnu.org/licenses/agpl-3.0.en.html
** 
** Additional terms under GNU Affero GPL version 3 section 7:
** 
**   Under Section 7 of the GNU Affero GPL version 3, the following additional
**   terms apply to the works covered under the License.  These additional terms
**   are non-permissive additional terms allowed under Section 7 of the GNU
**   Affero GPL version 3 and may not be removed by you.
** 
**   0. Attribution Requirement.
** 
**     You must preserve all legal notices or author attributions in the covered
**     work or Appropriate Legal Notices displayed by works containing the covered
**     work.  You may not remove from the covered work any author or developer
**     credit already included within the covered work.
** 
**   1. No License To Use Trademarks.
** 
**     This license does not grant any license or rights to use the trademarks
**     Golden Code, FWD, any Golden Code or FWD logo, or any other trademarks
**     of Golden Code Development Corporation. You are not authorized to use the
**     name Golden Code, FWD, or the names of any author or contributor, for
**     publicity purposes without written authorization.
** 
**   2. No Misrepresentation of Affiliation.
** 
**     You may not represent yourself as Golden Code Development Corporation or FWD.
** 
**     You may not represent yourself for publicity purposes as associated with
**     Golden Code Development Corporation, FWD, or any author or contributor to
**     the covered work, without written authorization.
** 
**   3. No Misrepresentation of Source or Origin.
** 
**     You may not represent the covered work as solely your work.  All modified
**     versions of the covered work must be marked in a reasonable way to make it
**     clear that the modified work is not originating from Golden Code Development
**     Corporation or FWD.  All modified versions must contain the notices of
**     attribution required in this license.
*/

package com.goldencode.util;

import java.lang.reflect.*;
import java.util.*;

import org.xml.sax.Attributes;

import com.goldencode.p2j.cfg.*;
import com.goldencode.util.GenericSAXHandler.*;

/**
 * Mappers are used by <code>GenericSAXHandler</code> for binding low level 
 * XML data to Java objects; this is the base class containing the main logic
 * responsible for that. Normally users of <code>GenericSAXHandler</code>
 * define a custom <code>Mapper</code>subclass.
 * <p>
 * The default behavior provided by <code>Mapper</code> base class is to 
 * attempt to relate XML attributes to setter methods in the Java class, on a
 * best effort basis, following the Java bean conventions. Attempts are made 
 * to convert <code>String</code> attribute values read from XML to the
 * respective properties' own types. In cases where this default approach does
 * not work, a custom property setter can be defined by <code>Mapper</code> 
 * subclasses, overriding the one in the base ({@link
 * #setProperty(Object, String, String)}).
 * <p>
 * As the nested structure of XML nodes reflect the composite nature of Java
 * objects, mappers also have the duty of controlling how objects that spring
 * from a certain sub-tree of the XML document are "consumed" by the parent
 * object, corresponding to its ancestor node in the XML tree. Unlike setting
 * properties, no assumptions can be made here regarding how an object is to
 * be integrated in its container, therefore the <code>Mapper</code> subclass
 * for the parent object would have to override method {@link #mergeSubObj(
 * Object, Object)}.
 * <p>
 * Recursive composite structures (types that contain instances of their own
 * type in the object structure) are supported by mappers by default. The same
 * mapper instance can be used simultaneously with any number of objects in
 * process of being constructed at a given time and subclasses can learn about
 * the context where they're operating by using the convenience methods 
 * provided by this <code>Mapper</code> base class ({@link #getXmlPath()}, 
 * {@link #getParentTag()}).
 * <p>
 * Mapping XML nodes to Java types is also sensitive to context. The same XML
 * tag can be mapped to different types, depending on where in the XML 
 * tree it appears. If a mapper is installed for a more specific context 
 * (e.g. "cfg/schema/namespace/dialect-specific/parameter") it takes 
 * precedence over the general purpose mapper installed for the simple tag
 * name (in this case "parameter").
 * <p>
 * Constraints can be added for data that is read from XML: for any element,
 * sets of mandatory and optional attributes as well as required and optional
 * child elements can be specified with {@link #setConstraints(String[], 
 * String[], String[], String[])}.
 * <p>
 * Check the documentation in <code>GenericSAXHandler</code> for more details
 * on how mappers are used.
 * <p>
 * This class is not synchronized.
 *  
 * @see GenericSAXHandler
 */
public class Mapper<T>
{
   /** The type of the Java object produced by this mapper. */
   private Class<T> cls;
   
   //constraints:
   
   /** 
    * Mandatory attributes for XML nodes handled by this mapper. If empty or
    * set to <code>null</code>, the constraint is disabled. By default this is
    * set to <code>null</code>.
    */ 
   Set<String> requiredAttrs = null;
   
   /** 
    * Optional attributes for XML nodes handled by this mapper. If this is
    * empty it means no optional attributes can exist, but if set to
    * <code>null</code> the constraint is disabled and any attributes are
    * allowed to occur. By default this is initialized to <code>null</code>.
    */ 
   Set<String> optionalAttrs = null;
   
   /**
    * Mandatory child elements for XML nodes handled by this mapper. If empty
    * or set to <code>null</code>, the constraint is disabled. By default this
    * is set to <code>null</code>.
    */ 
   Set<String> requiredChld = null;
   
   /** 
    * Optional child elements for XML nodes handled by this mapper. If this is
    * empty it means no optional child elements can exist, but if set to
    * <code>null</code> the constraint is disabled and any children are
    * allowed to occur. By default this is initialized to <code>null</code>.
    */ 
   Set<String> optionalChld = null;
   
   /** The path made of all XML elements (both handled and not handled by 
    * mappers) from root up to and including the current node, processed by 
    * this mapper. The path uses {@link GenericSAXHandler#XMLPATH_SEPARATOR}
    * as separator, defined to '/'. 
    * <p>
    * This information is not important for the logic in this mapper base 
    * class itself (it's only used in error reporting), but may be useful to
    * subclasses as context information.
    * 
    * @see Mapper#getParentTag()   
    */
   private String xmlPath = null;
   
   /**
    * Constructs a <code>Mapper</code> instance that is able to build and
    * populate Java objects of a given type.
    *   
    * @param  c 
    *         The Java type of the object constructed and populated by 
    *         this mapper. 
    *        
    * @throws IllegalArgumentException 
    *         If no class object is given.
    */
   public Mapper(Class<T> c)
   throws IllegalArgumentException
   {
      if (c == null)
      {
         throw new IllegalArgumentException("Null type given");
      }
      cls = c;
   }
   
   /**
    * Utility method for string capitalizing (first character to upper case). 
    * 
    * @param  s
    *         String to be capitalized.
    *        
    * @return The capitalized version of string <code>s</code>.
    */
   public static String cap1st(String s)
   {
      return s.substring(0, 1).toUpperCase() + s.substring(1);
   }
   
   /**
    * Adds constraints for the kind of XML nodes handled by this particular
    * mapper, in the form of mandatory and optional sets of attributes 
    * and/or child elements. The sets of mandatory and optional items have 
    * different meaning with respect to empty set and <code>null</code> 
    * value:
    * <p>
    * If the set of mandatory attributes/children is empty or <code>null
    * </code>, the respective constraint is disabled: there are no required
    * attributes or child elements, respectively. The meaning for the sets 
    * of optional items is different: <code>null</code> means the 
    * constraint is disabled (any elements can appear), while the empty set
    * indicates that there are no optional elements.
    * 
    * @param reqAttrs
    *        Array containing mandatory attribute names.
    * @param optAttrs
    *        Array containing optional attribute names.
    * @param reqChld
    *        Array containing mandatory child names.
    * @param optChld
    *        Array containing optional attribute names.
    */
   protected final void setConstraints(String[] reqAttrs, 
                                       String[] optAttrs, 
                                       String[] reqChld, 
                                       String[] optChld)
   {
      if (reqAttrs != null)
      {
         requiredAttrs = new HashSet<String>();
         requiredAttrs.addAll(Arrays.asList(reqAttrs));
      }
      if (optAttrs != null)
      {
         optionalAttrs = new HashSet<String>();
         optionalAttrs.addAll(Arrays.asList(optAttrs));
      }
      if (reqChld != null)
      {
         requiredChld = new HashSet<String>();
         requiredChld.addAll(Arrays.asList(reqChld));
      }
      if (optChld != null)
      {
         optionalChld = new HashSet<String>();
         optionalChld.addAll(Arrays.asList(optChld));
      }
   }

   /**
    * Checks if a child element can appear within the node processed by 
    * this mapper: if constraints are active (defined) and the child is 
    * neither mandatory nor optional, a <code>ConfigurationException</code>
    * is thrown. 
    *   
    * @param  chld
    *         The name of the XML child element that is being checked.
    *        
    * @throws ConfigurationException
    *         If the child is unexpected.
    */
   protected final void checkUnexpectedChild(String chld) 
   throws ConfigurationException
   {
      if ((requiredChld == null || (requiredChld != null && 
         !requiredChld.contains(chld))) && (optionalChld != null && 
         !optionalChld.contains(chld)))
      {
         throw new ConfigurationException("Encountered unexpected child: <"
               + chld + "> as: <" + getXmlPath() + ">");
      }
   }

   /**
    * Default factory method for creating the Java object that this mapper
    * will populate with data from XML. The default behavior is to attempt
    * to invoke the no-arg constructor in the given Java type in order to 
    * obtain a new instance. This behavior can be overridden in subclasses 
    * if a fancier object creation scheme is necessary.
    * 
    * @return A freshly built instance.
    * 
    * @throws IllegalStateException
    *         If the default (no-arg) constructor could not be invoked 
    *         successfully.  
    */
   protected T createPayloadObj()
   throws IllegalStateException
   {
      try
      {
         Constructor<T> c = cls.getDeclaredConstructor();
         c.setAccessible(true);
         return c.newInstance((Object[])null);
      }
      catch (Exception e)
      {
         throw new IllegalStateException("Factory could not create new " +
            "instance of type: " + cls.getName() + " by invoking the " +
            "default constructor", e);
      }
   }
   
   /**
    * Utility method for checking that the type of an object is (a subtype
    * of) a given class. An <code>IllegalArgumentException</code> is thrown
    * if this is not the case.
    * <p>
    * This method meant to be used in subclasses for validation of child 
    * (sub-objects) that are to be consumed by the object handled by this 
    * mapper.
    *  
    * @param  o
    *         The object to be checked.
    * @param  c
    *         The type that the object is to be checked against.
    *         
    * @throws IllegalArgumentException
    *         If the object is not an instance of the given type or any of
    *         its subtypes.
    */
   protected final void checkSubObjType(Object o, Class<?> c)
   throws IllegalArgumentException
   {
      if (!c.isAssignableFrom(o.getClass()))
      {
         throw new IllegalArgumentException("Bad sub-object type while " + 
            "processing node at <" + getXmlPath() + ">: " + 
            o.getClass().getName() + ", expected: " + c.getName() + 
            " (parent object java type: " + getType().getName() + ")");
      }
   }
   /**
    * Check that the given (child) object is of one of the supplied types.
    * The behavior is similar to {@link #checkSubObjType(Object, Class)}.
    * 
    * @param  o
    *         The object to be checked.
    * @param  clsTab
    *         The types that the object is to be checked against.
    */
   protected final void checkSubObjTypes(Object o, Class<?>[] clsTab)
   throws IllegalArgumentException
   {
      for (Class<?> c : clsTab)
      {
         if (c.isAssignableFrom(o.getClass()))
         {
            return;
         }
      }
      throw new IllegalArgumentException("Bad sub-object type while " +
         "processing node at <" + getXmlPath() + ">: " + 
         o.getClass().getName() + ", expected one of: " + clsTab + 
         " (parent object java type: " + getType().getName() + ")");
   }
   
   
   /**
    * By default we use the bean convention to fill in object properties.
    * Attempts will be made to convert the string representation of 
    * the property value to the actual Java type for that property. 
    * This method can be overriden in subclasses when custom/fancier 
    * setters are needed. 
    * 
    * @param  p  
    *         The target instance where the property will be set.  
    * @param  n
    *         The name of the property, following the Java bean convention.
    * @param  v
    *         The string representation of property's value, as it was read 
    *         from XML.  
    */
   protected void setProperty(T p, String n, String v)
   throws IllegalStateException
   {
      genSet(p, n, v);
   }
   
   /**
    * Applies a set of XML attributes as properties into the Java object 
    * produced by this mapper.
    *   
    * @param  attrs
    *         The set of attributes, as they were picked from the XML node
    *         by the SAX parser.
    *          
    * @throws ConfigurationException
    *         If unexpected attributes are encountered, as dictated by the 
    *         respective constraint, if active.
    */
   protected final void applyAttributes(Attributes attrs, T payloadObj)
   throws ConfigurationException,
          IllegalStateException
   {
      String k;
      String v;
      int n = attrs.getLength();
      for (int i = 0; i<n; i++)
      {
         k = attrs.getQName(i);
         v = attrs.getValue(i);
         if ((requiredAttrs == null || (requiredAttrs != null && 
            !requiredAttrs.contains(k))) && (optionalAttrs != null && 
            !optionalAttrs.contains(k)))
         {
            throw new ConfigurationException("Encountered unexpected " +
               "attribute: " + k + " of <" + getXmlPath() + ">.");
         }
         setProperty(payloadObj, k, v);
      }
   }
   
   /**
    * Gives the Java type of objects that this mapper is capable of 
    * building. 
    *   
    * @return  The Java type of the objects built by this mapper.
    */
   protected final Class<T> getType()
   {
      return cls;
   }
   
   /**
    * To be overridden in subclasses if the object built by this mapper has
    * a composite structure and needs to "consume" sub-objects. In most of the
    * cases this is the only piece of logic <code>Mapper</code> subclasses
    * must provide.
    *  
    * @param  p
    *         The current (parent) object that will consume the sub-object.  
    * @param  c
    *         The child (sub-object) that will be consumed by this mapper's
    *         object. 
    */
   protected void mergeSubObj(T p, Object c) 
   {
   }
   
   /**
    * Getter for the path to (and including) the XML node handled by this 
    * mapper. The path contains all XML tags along the way, including those 
    * that are unhandled, and uses {@link GenericSAXHandler#XMLPATH_SEPARATOR}
    * as a separator (defined to '/').
    *  
    * @return The XML path as a <code>String</code>.
    * 
    */
   protected final String getXmlPath()
   {
      return xmlPath;
   }

   /**
    * Provides the name of the parent XML tag. It is not used by the logic in
    * this <code>Mapper</code> base class, it's given here only as a
    * convenience method that may come handy as context information in 
    * subclasses for knowing the where we are in the XML tree / object graph.
    *  
    * @return The name of the parent XML tag, extracted from the current XML
    *         path, or <code>null</code> if no parent tag exists.
    *
    * @see Mapper#getXmlPath()
    */
   protected String getParentTag()
   {
      String path = getXmlPath();
      if (path == null)
      {
         return null;
      }
      int b2 = path.lastIndexOf(GenericSAXHandler.XMLPATH_SEPARATOR);
      if (b2 <= 0)
      {
         return null;
      }
      int b1 = path.lastIndexOf(GenericSAXHandler.XMLPATH_SEPARATOR, b2 - 1);
      if (b1 < 0)
      {
            b1 = -1;
      }
      return path.substring(b1 + 1, b2);
   }
   
   /**
    * Updates the path to (and including) the XML node handled by this mapper.
    * The path contains all XML tags along the way, including those that are 
    * unhandled, and uses {@link GenericSAXHandler#XMLPATH_SEPARATOR} as a 
    * separator (defined to '/').
    * 
    * @param path
    *        The XML path.
    */
   final void setXmlPath(String path)
   {
      xmlPath = path;
   }   
   
   /**
    * Creates generic <code>ActiveNode</code> containers, holding the properly
    * typed newly created Java object together with the correct <code>Mapper
    * </code> that is going to be used for populating it.
    *   
    * @param  mapperKey 
    *         The key that the mapper was installed for (usually an XML tag
    *         name, but can also be a complete path in the XML tree, using 
    *         {@link GenericSAXHandler#XMLPATH_SEPARATOR} as separator, 
    *         defined to '/').
    *         
    * @return An <code>ActiveNode</code> instance containing the mapper object
    *         together with a newly created payload object that will be 
    *         populated by the mapper.  
    */
   final ActiveNode<T> createActiveNode(String mapperKey)
   {
      T payload = createPayloadObj();
      return new ActiveNode<T>(this, payload, mapperKey);
   }
   
   /**
    * Generic property setter, following java bean conventions. Attempts 
    * will be made to convert the string representation of property's value
    * to the actual Java type of that property. 
    * 
    * @param  o 
    *         The target object where the property will be set.
    * @param  n 
    *         The name of the property to be set.
    * @param  v 
    *         The string representation of property's value.
    *  
    * @throws IllegalStateException
    *         If a suitable setter for the given property could not be 
    *         found or invoked successfully.
    */
   private void genSet(Object o, String n, String v)
   throws IllegalStateException
   {
      Class<?> c = o.getClass();
      Method[] methods = c.getDeclaredMethods();
      String setterName = "set" + cap1st(n);
      Method setter = null;
      for (Method m : methods)
      {
         if (m.getName().equals(setterName))
         {
            setter = m;
            break;
         }
      }
      if (setter == null)
      {
         throw new IllegalStateException("Could not find default setter " +
            "'" + setterName + "()' for property '" + n + "' in class " + 
            c.getName() + " (setProperty() may need to be overriden by " +
            "subclass in order to handle particular cases)");
      }
      Class<?>[] paramTypes = setter.getParameterTypes();
      
      if (setter.getParameterTypes().length != 1)
      {
         throw new IllegalStateException("Could not figure out setter " +
            "parameters for " + c.getName() + "." + setterName + 
            " (setProperty() may need to be overriden by subclass in order " +
            "to handle particular cases)");
      }
      setter.setAccessible(true);
      try
      {
         if (paramTypes[0].equals(String.class))
         {
            setter.invoke(o, new Object[]{v});
         }
         else
         {
            // add support for more types here:
            if (paramTypes[0].equals(Boolean.class))
            {
               setter.invoke(o, new Object[]{new Boolean(v)});
               return;
            }
            if (paramTypes[0].equals(Integer.class))
            {
               setter.invoke(o, new Integer(v));
               return;
            }
            if (paramTypes[0].equals(boolean.class))
            {
               setter.invoke(o, 
                             new Object[]{new Boolean(v).booleanValue()});
               return;
            }
            if (paramTypes[0].equals(int.class))
            {
               setter.invoke(o, new Integer(v).intValue());
               return;
            }
               throw new IllegalStateException(
                  "Parameter type not supported for setter " +
                  c.getName() + "." + setter.getName() + "(): " +
                  paramTypes[0].getName() +
                  " (custom behavior may be needed in subclass in order " +
                  "to fix this problem)");
         }
      }
      catch(InvocationTargetException e)
      {
         throw new IllegalStateException("Setter can't be invoked " +
            "generically: " + setter.getName() + " (setProperty() may " +
            "need to be overriden by subclass in order to handle " +
            "particular cases)");
      }
      catch(IllegalAccessException e)
      {
         throw new IllegalStateException("Setter not accessible: " + 
            setter.getName());
      }
   }
}
