/*
** Module   : get_text_metrics.c
** Abstract : a simple tool to capture the text metrics for given fonts defined in the font list
**            file with list of the text strings taken from the seperate file.
**
** Copyright (c) 2017, Golden Code Development Corporation.
**
** -#- -I- --Date-- --------------------------------Description---------------------------------
** 001 EVL 20170501 First version.
** 002 EVL 20170509 Adding taking into account the font anti-aliasing feature.  The default mode
**                  can be changed via command line option.
*/
/*
** This program is free software: you can redistribute it and/or modify
** it under the terms of the GNU Affero General Public License as
** published by the Free Software Foundation, either version 3 of the
** License, or (at your option) any later version.
**
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
** GNU Affero General Public License for more details.
**
** You may find a copy of the GNU Affero GPL version 3 at the following
** location: https://www.gnu.org/licenses/agpl-3.0.en.html
** 
** Additional terms under GNU Affero GPL version 3 section 7:
** 
**   Under Section 7 of the GNU Affero GPL version 3, the following additional
**   terms apply to the works covered under the License.  These additional terms
**   are non-permissive additional terms allowed under Section 7 of the GNU
**   Affero GPL version 3 and may not be removed by you.
** 
**   0. Attribution Requirement.
** 
**     You must preserve all legal notices or author attributions in the covered
**     work or Appropriate Legal Notices displayed by works containing the covered
**     work.  You may not remove from the covered work any author or developer
**     credit already included within the covered work.
** 
**   1. No License To Use Trademarks.
** 
**     This license does not grant any license or rights to use the trademarks
**     Golden Code, FWD, any Golden Code or FWD logo, or any other trademarks
**     of Golden Code Development Corporation. You are not authorized to use the
**     name Golden Code, FWD, or the names of any author or contributor, for
**     publicity purposes without written authorization.
** 
**   2. No Misrepresentation of Affiliation.
** 
**     You may not represent yourself as Golden Code Development Corporation or FWD.
** 
**     You may not represent yourself for publicity purposes as associated with
**     Golden Code Development Corporation, FWD, or any author or contributor to
**     the covered work, without written authorization.
** 
**   3. No Misrepresentation of Source or Origin.
** 
**     You may not represent the covered work as solely your work.  All modified
**     versions of the covered work must be marked in a reasonable way to make it
**     clear that the modified work is not originating from Golden Code Development
**     Corporation or FWD.  All modified versions must contain the notices of
**     attribution required in this license.
*/

// On Windows, you must install mingw from http://mingw-w64.sourceforge.net or
// from http://www.drangon.org/mingw. Take one of the full archives:
// mingw-w32-bin-i686-2013MMDD.7z or mingw-w64-bin-x86_64-2013MMDD.7z and unpack
// it to c:\.  Then include The c:\mingw32\bin\ or c:\mingw64\bin\ to the PATH
// variable, depending on the Windows architecture in use. We have not had good
// luck using the automatic installer for mingw.

// Then compile using this:
// gcc get_text_metrics.c -lgdi32 -o get_text_metrics.exe
                            
#include <conio.h>
#include <stdio.h>
#include <stdlib.h>
#include <windows.h>

int maxSize = 20;
BOOL useAntiAliasing = FALSE;
char fontFile[32] = "font-list.txt";
char textFile[32] = "lines-list.txt";
char xmlFileName[32] = "text-metrics.xml";

// global data cache
// not clear is LF_FACESIZE(32) size is enough to store the font name
typedef struct _tagLogFontInfo
{
   char name[LF_FACESIZE*2];
   HFONT hFnt;
   int size;
   BOOL bold;
   BOOL italic;
   BOOL under;
} LOGFONTINFO, *PLOGFONTINFO;

HDC hDC = NULL;
PLOGFONTINFO pLogFntArray = NULL;
int logFontArraySize = 0;

// inpit file font list definitions
#define MAX_FONT_LIST_SIZE 256
struct _fnt_params
{
   char name[LF_FACESIZE*2];
   int size;
   int bold;
   int italic;
   int under;
} fntList[MAX_FONT_LIST_SIZE];
int fntListSize = 0;

// three-state boolean logic constants for integer data type
#define INT_BOOL_UNDEF -1
#define INT_BOOL_TRUE  TRUE
#define INT_BOOL_FALSE FALSE

// the maximum line size for text strings file to be read
#define LINE_SIZE 4096

/**
 * The initialization routine. Displays the default values for initial options with ability to
 * change if necessary.
 */
void init()
{
   int newMaxSize = 0;
   char newString[32] = "";

   printf("\nMaximum font size to consider: %d, OK? (y/n) ", maxSize);
   if (getch() == 'n')
   {
      printf("\nEnter the maximum size for font to consider: ");
      scanf("%d", &newMaxSize);
      if (newMaxSize > 0)
      {
         maxSize = newMaxSize;
      }
   }
   else
   {
      printf("\n");
   }

   printf("Filename for the list of fonts to check: %s, OK? (y/n) ", fontFile);
   if (getch() == 'n')
   {
      printf("\nEnter the font list filename: ");
      scanf("%s", newString);
      if (strlen(newString) > 0)
      {
         strcpy(fontFile, newString);
      }
   }
   else
   {
      printf("\n");
   }
   
   printf("Filename for the list of strings to check: %s, OK? (y/n) ", textFile);
   if (getch() == 'n')
   {
      printf("\nEnter the string list filename: ");
      scanf("%s", newString);
      if (strlen(newString) > 0)
      {
         strcpy(textFile, newString);
      }
   }
   else
   {
      printf("\n");
   }
   
   // verify the initial settings
   printf("\nMaximum font size: %d\n", maxSize);
   printf("Font anti-aliasing is: %s\n", useAntiAliasing == TRUE ? "On" : "Off");
   printf("Font list: %s\n", fontFile);
   printf("String list: %s\n", textFile);
   // warn about line size limitation (2 chars are reserved for null terminator and new line)
   printf("The maximum line size for %s file is: %d chars\n", textFile, LINE_SIZE - 2);
   printf("Press any key to continue or \"a\" to abort.\n");
   if (getch() == 'a')
   {
      printf("Execution interrupted...\n");
      exit(-1);
   }
}

/**
 * The font list reading from input file. Displays the result of requested font options with
 * ability to stop execution if initial font configuration is wrong.
 */
void initFontList()
{
   FILE *ffnt;
   char fontLine[LINE_SIZE];
   // null terminated delimiters string
   char delim[3] = ",\n";

   // open font list file   
   ffnt = fopen(fontFile, "rt");
   if (ffnt == NULL)
   {
      printf("WARNING! Missing the font list file %s, application stoped.\n", fontFile);
      exit(-1);
   }

   // font list file opened, read it line by line
   while (fgets(fontLine, sizeof(fontLine), ffnt) != NULL && fntListSize < MAX_FONT_LIST_SIZE)
   {
      // init font style constants to undefined value
      fntList[fntListSize].bold   = INT_BOOL_UNDEF;
      fntList[fntListSize].italic = INT_BOOL_UNDEF;
      fntList[fntListSize].under  = INT_BOOL_UNDEF;
      char * fontEntry = strtok(fontLine, delim);
      if (fontEntry != NULL)
      {
         strcpy(fntList[fntListSize].name, fontEntry);
         fontEntry = strtok(NULL, delim);
         if (fontEntry != NULL)
         {
            sscanf(fontEntry, "%d", &fntList[fntListSize].size);
            if (fntList[fntListSize].size < 0)
            {
               fntList[fntListSize].size = 0;
            }
            // check if bold option specified
            fontEntry = strtok(NULL, delim);
            if (fontEntry != NULL)
            {
               if (strcmp(fontEntry, "true") == 0)
               {
                  fntList[fntListSize].bold = INT_BOOL_TRUE;
               }
               else if (strcmp(fontEntry, "false") == 0)
               {
                  fntList[fntListSize].bold = INT_BOOL_FALSE;
               }
               // check if italic option specified
               fontEntry = strtok(NULL, delim);
               if (fontEntry != NULL)
               {
                  if (strcmp(fontEntry, "true") == 0)
                  {
                     fntList[fntListSize].italic = INT_BOOL_TRUE;
                  }
                  else if (strcmp(fontEntry, "false") == 0)
                  {
                     fntList[fntListSize].italic = INT_BOOL_FALSE;
                  }
                  // check if underline option specified
                  fontEntry = strtok(NULL, delim);
                  if (fontEntry != NULL)
                  {
                     if (strcmp(fontEntry, "true") == 0)
                     {
                        fntList[fntListSize].under = INT_BOOL_TRUE;
                     }
                     else if (strcmp(fontEntry, "false") == 0)
                     {
                        fntList[fntListSize].under = INT_BOOL_FALSE;
                     }
                  }
               }
            }
         }
         fntListSize++;
         // protection from array out of range issue
         if (fntListSize == MAX_FONT_LIST_SIZE)
         {
            printf("WARNING! The font list size in file %s is too large,\n", fontFile);
            printf("         Only first %d items will be considered\n", MAX_FONT_LIST_SIZE);
         }
      }
   }
   fclose(ffnt);
   
   printf("\nFonts to be processed:\n");
   printf("\nName, size, bold, italic, underline\n");
   printf("-----------------------------------\n");
   int i;
   for (i = 0; i < fntListSize; i++)
   {
      printf("%s", fntList[i].name);
      if (fntList[i].size == 0)
      {
         printf(", 1 - %d", maxSize);
      }
      else
      {
         printf(", %d", fntList[i].size);
      }
      printf(", %s, %s, %s\n",
         fntList[i].bold == INT_BOOL_UNDEF
            ? "both" : fntList[i].bold == INT_BOOL_TRUE ? "true" : "false",
         fntList[i].italic == INT_BOOL_UNDEF
            ? "both" : fntList[i].italic == INT_BOOL_TRUE ? "true" : "false",
         fntList[i].under == INT_BOOL_UNDEF
            ? "both" : fntList[i].under == INT_BOOL_TRUE ? "true" : "false");
   }
   printf("-----------------------------------\n");
   printf("Press any key to start or \"a\" to abort.\n");
   if (getch() == 'a')
   {
      printf("Execution interrupted...\n");
      exit(-1);
   }
}

/**
 * The function to replace single character with given string with all occurences of this
 * character.
 *
 * @param   dstText  on input points to original sring to scan, on output points to the
 *                   modified final string.
 *
 * @param   xmlChar  character to be replaced.
 *
 * @param   replText text to substitute the character above.
 */
void replaceXMLChar(char * dstText, char xmlChar, const char * replText)
{
   char tmpLine[LINE_SIZE];
   char *xmlPtr = strchr(dstText, xmlChar);
   
   while (xmlPtr != NULL)
   {
      // build the replacer + rest of the sting
      strcpy(tmpLine, replText);
      // this append the rest of the string without char to be replaced
      strcat(tmpLine, &xmlPtr[1]);
      // overlap the original string beginning with the char to replace
      strcpy(xmlPtr, tmpLine);
      // find next match in the string after replacer
      xmlPtr = strchr(&xmlPtr[strlen(replText)], xmlChar);
   }
}

/**
 * The function calculates the text metrics(width and height of the line) for the given text
 * string value with specified font name, size and font attributes such as bold, italic and
 * underline.  The whole desktop device context is used. 
 *
 * @param   line     the text string to calculate the metrics.
 *
 * @param   fntName  the font name as character sequence.
 *
 * @param   hLogFnt  the handle of the logical font corresponding to the font name.
 *
 * @param   fntSize  font size in Windows points.
 *
 * @param   bBold    TRUE for bold attribute is on, FALSE for off.
 *
 * @param   bItalic  TRUE for italic attribute is on, FALSE for off.
 *
 * @param   bUnder   TRUE for underline attribute is on, FALSE for off.
 *
 * @param   wRes     pointer to store the width value of the calculated metrics.
 *
 * @param   hRes     pointer to store the height value of the calculated metrics.
 */
void getMetrics(const char * line, const char * fntName, HFONT hLogFnt, int fntSize,
                BOOL bBold, BOOL bItalic, BOOL bUnder, int *wRes, int *hRes)
{
   // and associate font with the current device context
   if (SelectObject(hDC, hLogFnt) == NULL)
   {
      printf("WARNING! Unable to SelectObject() for font and current DC.\n"\
             "Font name: %s, size: %d, bold: %d, italic: %d, underline: %d\n",
             fntName, fntSize, bBold, bItalic, bUnder);
      // just return, clean up will be performed later
      return;
   }

   // getting the text line metrics   
   SIZE sz = {0};
   BOOL rc = GetTextExtentPoint(hDC, line, strlen(line), &sz);
   if (rc)
   {
      // store the text X and Y dimensions
      *wRes = sz.cx;
      *hRes = sz.cy;
   }
   else
   {
      printf("WARNING! Failed to GetTextExtentPoint() for line %s.\n", line);
   }
}

/**
 * The function creates logical font handle for given font name, screen DPI, font size and
 * attributes. 
 *
 * @param   fntName   the font name as character sequence.
 *
 * @param   dpiScreen the device resolution in dots per inch.
 *
 * @param   fntSize   font size in Windows points.
 *
 * @param   bBold     TRUE for bold attribute is on, FALSE for off.
 *
 * @param   bItalic   TRUE for italic attribute is on, FALSE for off.
 *
 * @param   bUnder    TRUE for underline attribute is on, FALSE for off.
 *
 * @return   The logical font handle in case of success or NULL in case of failure.
 */
HFONT createLogicalFont(const char * fntName, int dpiScreen, int fntSize,
                        BOOL bBold, BOOL bItalic, BOOL bUnder)
{
   // building a logical font   
   LOGFONT logFnt = {0};
   logFnt.lfHeight = -MulDiv(fntSize, dpiScreen, 72);
   logFnt.lfWeight = (bBold) ? 700 : FW_DONTCARE;
   logFnt.lfItalic = bItalic;
   logFnt.lfUnderline = bUnder;
   logFnt.lfQuality = useAntiAliasing == TRUE ? ANTIALIASED_QUALITY : NONANTIALIASED_QUALITY; 
   strcpy(logFnt.lfFaceName, fntName);
   
   HFONT hFntRes = CreateFontIndirect(&logFnt);
   if (hFntRes == NULL)
   {
      printf("WARNING! Unable to CreateFontIndirect() for requested font: %s, size: %d,"\
             "bold: %d, italic: %d, underline: %d\n", fntName, fntSize, bBold, bItalic, bUnder);
   }
   
   return hFntRes;
}

/**
 * Creates logical font array to be used for metrics calculation from inital font list
 * provided inside the input file.
 *
 * @return   TRUE if operation successful or FALSE in case of failure or 0 size font array.
 */
BOOL initLogicalFonts()
{
   // getting display device for entire screen
   hDC = GetDC(NULL);
   if (hDC == NULL)
   {
      return FALSE;
   }
    
   // getting dpi value for Y direction   
   int dpi = GetDeviceCaps(hDC, LOGPIXELSY);
   if (dpi <= 0)
   {
      printf("WARNING! Unable to GetDeviceCaps().\n");
      // free the device context
      ReleaseDC(NULL, hDC);
      // prevent double release
      hDC = NULL;
      // and return
      return FALSE;
   }

   // create logical font array and initialize data inside
   int i,j,k,l,sz,szMin,szMax;
   // enumarate all fonts in the list
   // on the first pass we need to calculate total number of logical fonts to create 
   for (i = 0; i < fntListSize; i++)
   {
      // ensure to take into account possible explicit size
      if (fntList[i].size != 0)
      {
         szMin = fntList[i].size;
         szMax = fntList[i].size;
      }
      else
      {
         szMin = 1;
         szMax = maxSize;
      }
      // font size
      for (sz = szMin; sz <= szMax; sz++)
      {
         int boldMin = fntList[i].bold == INT_BOOL_UNDEF ? 0 : fntList[i].bold;
         int boldMax = fntList[i].bold == INT_BOOL_UNDEF ? 1 : fntList[i].bold;
         // bold
         for (j = boldMin; j <= boldMax; j++)
         {
            int italicMin = fntList[i].italic == INT_BOOL_UNDEF ? 0 : fntList[i].italic;
            int italicMax = fntList[i].italic == INT_BOOL_UNDEF ? 1 : fntList[i].italic;
            // italic
            for (k = italicMin; k <= italicMax; k++)
            {
               int underMin = fntList[i].under == INT_BOOL_UNDEF ? 0 : fntList[i].under;
               int underMax = fntList[i].under == INT_BOOL_UNDEF ? 1 : fntList[i].under;
               // underline
               for (l = underMin; l <= underMax; l++)
               {
                  // yet another logical font will be required
                  logFontArraySize++;
               }
            }
         }
      }
   }
   
   // now we know the fon numbers, allocate memory for font cache
   pLogFntArray = (PLOGFONTINFO)malloc(logFontArraySize * sizeof(LOGFONTINFO));
   if (pLogFntArray == NULL)
   {
      printf("WARNING! Memory allocation error for font array.\n");
      // free the device context
      ReleaseDC(NULL, hDC);
      // prevent double release
      hDC = NULL;
      // and return
      return FALSE;
   }
   
   // font array allocated, start filling it with logical fonts
   int currFntNumber = 0;
   for (i = 0; i < fntListSize; i++)
   {
      // ensure to take into account possible explicit size
      if (fntList[i].size != 0)
      {
         szMin = fntList[i].size;
         szMax = fntList[i].size;
      }
      else
      {
         szMin = 1;
         szMax = maxSize;
      }
      // font size
      for (sz = szMin; sz <= szMax; sz++)
      {
         int boldMin = fntList[i].bold == INT_BOOL_UNDEF ? 0 : fntList[i].bold;
         int boldMax = fntList[i].bold == INT_BOOL_UNDEF ? 1 : fntList[i].bold;
         // bold
         for (j = boldMin; j <= boldMax; j++)
         {
            int italicMin = fntList[i].italic == INT_BOOL_UNDEF ? 0 : fntList[i].italic;
            int italicMax = fntList[i].italic == INT_BOOL_UNDEF ? 1 : fntList[i].italic;
            // italic
            for (k = italicMin; k <= italicMax; k++)
            {
               int underMin = fntList[i].under == INT_BOOL_UNDEF ? 0 : fntList[i].under;
               int underMax = fntList[i].under == INT_BOOL_UNDEF ? 1 : fntList[i].under;
               // underline
               for (l = underMin; l <= underMax; l++)
               {
                  // try to create next logical font
                  BOOL isBold = (j == 1);
                  BOOL isItalic = (k == 1);
                  BOOL isUnder = (l == 1);
                  HFONT hFntLog = createLogicalFont(fntList[i].name, dpi, sz,
                                                    isBold, isItalic, isUnder);
                  
                  // logical font created, store it into the cache
                  if (hFntLog != NULL)
                  {
                     // save another logical font
                     pLogFntArray[currFntNumber].hFnt = hFntLog;
                     pLogFntArray[currFntNumber].size = sz;
                     pLogFntArray[currFntNumber].bold = isBold;
                     pLogFntArray[currFntNumber].italic = isItalic;
                     pLogFntArray[currFntNumber].under = isUnder;
                     strcpy(pLogFntArray[currFntNumber].name, fntList[i].name);
                     // next font entry recorded
                     currFntNumber++;
                  }
               }
            }
         }
      }
   }
   
   // check if all requested fonts were properly created and stored
   if (currFntNumber == 0)
   {
      printf("WARNING! Fatal error. Logical font array is empty.\n");
      logFontArraySize = currFntNumber;
      // and return
      return FALSE;
   }
   else if (currFntNumber < logFontArraySize)
   {
      printf("WARNING! Some logical fonts were not created, expect to have incomplete results.\n");
      logFontArraySize = currFntNumber;
   }
   
   // we are here, everything is OK
   return TRUE;
}

/**
 * Cleans up the resources dynamically allocated at program execution.
 */
void cleanUpGui()
{
   // first delete all logical fonts
   if (pLogFntArray != NULL)
   {
      int i;
      for (i = 0; i < logFontArraySize; i++)
      {
         if (pLogFntArray[i].hFnt != NULL)
         {
            // delete font
            DeleteObject(pLogFntArray[i].hFnt);
         }
      }
      // fonts deleted, now free the memory allocated for array itself
      free(pLogFntArray);
      pLogFntArray = NULL;
   }
   // free the device context if allocated
   if (hDC != NULL)
   {
      ReleaseDC(NULL, hDC);
   }
}

/**
 * The function records the text metrics data for the given line and font into output file
 * in XML based format. 
 *
 * @param   fo       the file object to use for data writing.
 *
 * @param   line     the text string to calculate the metrics.
 *
 * @param   fntName  the font name as character sequence.
 *
 * @param   hLogFnt  the handle of the logical font corresponding to the font name.
 *
 * @param   fntSize  font size in Windows points.
 *
 * @param   bBold    TRUE for bold attribute is on, FALSE for off.
 *
 * @param   bItalic  TRUE for italic attribute is on, FALSE for off.
 *
 * @param   bUnder   TRUE for underline attribute is on, FALSE for off.
 */
void recordMetrics(FILE * fo, const char * line, const char * fntName, HFONT hLogFnt,
                   int fntSize, BOOL bBold, BOOL bItalic, BOOL bUnder)
{
   int w = 0;
   int h = 0;
   char sBold[6];
   char sItalic[6];
   char sUnder[6];
   
   // getting the metrics for current conditions
   getMetrics(line, fntName, hLogFnt, fntSize, bBold, bItalic, bUnder, &w, &h);
   
   // convert boolean flags to strings
   strcpy(sBold, bBold ? "true" : "false");
   strcpy(sItalic, bItalic ? "true" : "false");
   strcpy(sUnder, bUnder ? "true" : "false");
   
   // send metrics to the output file
   fprintf(fo, "      <font name=\"%s\" size=\"%d\" bold=\"%s\" italic=\"%s\" underline=\"%s\""\
               " width=\"%d\" height=\"%d\"/>\n", fntName, fntSize, sBold, sItalic, sUnder, w, h);
}

/**
 * The main program entry point.
 *
 * @param   argc    the first option is flag to use anti-aliased mode.
 *
 * @param   argv    anti-aliasing means setting internal use anti-aliasing flag to true.
 *                  The default value is FALSE meaning do not use anti-aliasing for metrics
 *                  calculation.
 *
 * @return   The execution status, 0 for success, negative for failures.
 */
int main(int argc, char* argv[])
{
   FILE *fw, *fr;
   char textLine[LINE_SIZE];
   // reserve enough room in XML line for possible special char substitutions 
   char xmlTextLine[5*LINE_SIZE];
   
   // handle optional command line arguments
   if (argc > 1 && strcmp(argv[1], "anti-aliasing") == 0)
   {
      useAntiAliasing = TRUE;
   }
   
   // set up initials
   init();
   
   // set up font list to work with
   initFontList();
   
   // init logical fonts
   if (!initLogicalFonts())
   {
      // unable to continue, logical fonts init issues
      printf("WARNING! Unable to create required logical fonts, execution stopped.\n");
      cleanUpGui();
      exit(-1);
   }
   
   fr = fopen(textFile, "rt");
   if (fr == NULL)
   {
      printf("WARNING! Missing the input file %s, execution stopped.\n", textFile);
      exit(-1);
   }

   // open file to record text metrics   
   fw = fopen(xmlFileName, "wt");
   if (fw == NULL)
   {
      printf("WARNING! Unable to create output file %s for writing, execution stopped.\n",
             xmlFileName);
      // close already opened files
      fclose(fr);
      exit(-1);
   }
   
   // start generating text metrics file
   fprintf(fw, "<?xml version=\"1.0\" encoding=\"UTF-8\"?>\n");
   fprintf(fw, "<!DOCTYPE text-list SYSTEM \"text-metrics.dtd\">\n");
   fprintf(fw, "<text-list>\n");

   // reading text list line by line
   while (fgets(textLine, sizeof(textLine), fr) != NULL)
   {
       // remove extra line feed from input string string
       char *lf = strchr(textLine, '\n');
       if (lf != NULL)
       {
           *lf = 0;
       }

       // make a copy
       strcpy(xmlTextLine, textLine);
       
       // replace special XML chars with &xxx; sequence
       replaceXMLChar(xmlTextLine, '&', "&amp;");
       replaceXMLChar(xmlTextLine, '"', "&quot;");
       replaceXMLChar(xmlTextLine, '\'', "&apos;");
       replaceXMLChar(xmlTextLine, '<', "&lt;");
       replaceXMLChar(xmlTextLine, '>', "&gt;");
       
       // next line section in XML file
       fprintf(fw, "   <text value=\"%s\">\n", xmlTextLine);
       
       // processing the next text line here with the current font set
       int i;
       for (i = 0; i < logFontArraySize; i++)
       {
          recordMetrics(fw, textLine, pLogFntArray[i].name, pLogFntArray[i].hFnt,
                        pLogFntArray[i].size, pLogFntArray[i].bold, pLogFntArray[i].italic,
                        pLogFntArray[i].under);
       }
       
       // finish next line processing
       fprintf(fw, "   </text>\n");
   }
   
   // finalize XML file
   fprintf(fw, "</text-list>");
   // close opened files
   fclose(fr);
   fclose(fw);
   
   // clean up GUI related allocations
   cleanUpGui();
   return 0;
}
