/*
** Module   : shell_linux.c
** Abstract : Run LINUX/UNIX shell and control master PTY descriptor.
**
** Copyright (c) 2014-2017 Golden Code Development Corporation.
**
** -#- -I- --Date-- --------------------------------Description-----------------------------------
** 001 MAG 20140318 Created initial version.
** 002 GES 20150211 Added comments and return errno if the execvp() fails.
** 003 EVL 20160217 Adding Solaris 10 u8 customization for pty.h to compile.
** 004 EVL 20160227 Adding implementation for missed forkpty().
*/
/*
** This program is free software: you can redistribute it and/or modify
** it under the terms of the GNU Affero General Public License as
** published by the Free Software Foundation, either version 3 of the
** License, or (at your option) any later version.
**
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
** GNU Affero General Public License for more details.
**
** You may find a copy of the GNU Affero GPL version 3 at the following
** location: https://www.gnu.org/licenses/agpl-3.0.en.html
** 
** Additional terms under GNU Affero GPL version 3 section 7:
** 
**   Under Section 7 of the GNU Affero GPL version 3, the following additional
**   terms apply to the works covered under the License.  These additional terms
**   are non-permissive additional terms allowed under Section 7 of the GNU
**   Affero GPL version 3 and may not be removed by you.
** 
**   0. Attribution Requirement.
** 
**     You must preserve all legal notices or author attributions in the covered
**     work or Appropriate Legal Notices displayed by works containing the covered
**     work.  You may not remove from the covered work any author or developer
**     credit already included within the covered work.
** 
**   1. No License To Use Trademarks.
** 
**     This license does not grant any license or rights to use the trademarks
**     Golden Code, FWD, any Golden Code or FWD logo, or any other trademarks
**     of Golden Code Development Corporation. You are not authorized to use the
**     name Golden Code, FWD, or the names of any author or contributor, for
**     publicity purposes without written authorization.
** 
**   2. No Misrepresentation of Affiliation.
** 
**     You may not represent yourself as Golden Code Development Corporation or FWD.
** 
**     You may not represent yourself for publicity purposes as associated with
**     Golden Code Development Corporation, FWD, or any author or contributor to
**     the covered work, without written authorization.
** 
**   3. No Misrepresentation of Source or Origin.
** 
**     You may not represent the covered work as solely your work.  All modified
**     versions of the covered work must be marked in a reasonable way to make it
**     clear that the modified work is not originating from Golden Code Development
**     Corporation or FWD.  All modified versions must contain the notices of
**     attribution required in this license.
*/

#include <sys/types.h>
#include <sys/wait.h>
#include <stdio.h>
#include <errno.h>
#include <stdlib.h>
#include <unistd.h>
#ifdef SOLARIS
   #include <sys/stropts.h>
#else
   #include <pty.h>
#endif
#include <string.h>
#include <fcntl.h>

#ifdef SOLARIS
#define forkpty fork_pty

/**
 * Replacement for Linux forkpty() function missed in Solaris 10.  Only first parameter is used.
 * 
 * @param   amaster
 *          Pointer to master PTY file descriptor.
 * @param   unused1
 *          Not used for now.
 * @param   unused2
 *          Not used for now.
 * @param   unused3
 *          Not used for now.
 *
 * @return  On error, -1 is returned, and errno is set appropriately.
 */
int fork_pty (int *amaster, char *unused1, void *unused2, void *unused3) 
{
   // master and slave channels
   int master_channel, slave_channel;
   // slave channel name
   char * sname;
   // pid to return
   pid_t pid = -1;
   
   // opening master channel
   master_channel = open("/dev/ptmx", O_RDWR);
   // looking for possible errors in getting valid master channel while
   // change permission of the slave for given master and unlock it
   if (master_channel < 0 || grantpt(master_channel) < 0 || unlockpt(master_channel) < 0)
   {
      // close master if necessary
      if (master_channel >= 0)
      {
         close(master_channel);
      }
      return pid;
   }
   
   // get the slave channel name
   sname = ptsname(master_channel);
   if (sname != NULL)   
   {
      // open slave
      slave_channel = open(sname, O_RDWR);
   }
   
   // looking for possible errors while setting up slave channel 
   if (sname == NULL || slave_channel < 0 ||
       ioctl(slave_channel, I_PUSH, "ptem") < 0 ||
       ioctl(slave_channel, I_PUSH, "ldterm") < 0)
   {
      // close slave if necessary
      if (slave_channel >= 0)
      {
         close(slave_channel);
      }
      // close master
      close(master_channel);
      return pid;
   }
   
   // forking pty
   pid = fork();
   // checking the fork() return value
   if (pid == 0)
   {
      // child thread of execution
      close(master_channel);
      dup2(slave_channel, STDIN_FILENO);
      dup2(slave_channel, STDOUT_FILENO);
      dup2(slave_channel, STDERR_FILENO);
   }
   else if (pid > 0)
   {
      // parent thread of execution
      close(slave_channel);
   }
   else
   {
      // parent thread of execution, pid == -1, error
      // try to clean up as much as possible by closing opened channel 
      close(slave_channel);
      close(master_channel);
   }
      
   // store master channel to the caller if master is valid and opened
   if (amaster != NULL && pid > 0)
   {
      *amaster = master_channel;
   }
   
   return pid; 
}
#endif

/**
 * Use non blocking operations on master PTY descriptor.
 * 
 * @param   mfd
 *          Master PTY file descriptor.
 *
 * @return  On error, -1 is returned, and errno is set appropriately.
 */
int unblock(int mfd)
{
   int flags;
   
   // set non-blocking
   if ((flags = fcntl(mfd, F_GETFL, 0)) == -1)
   {
      flags = 0;
   }
   
   return fcntl(mfd, F_SETFL, flags | O_NONBLOCK); 
}

/**
 * Start a LINUX/UNIX shell session on a pseudo terminal. The master PTY file descriptor is used
 * to capture terminal INPUT and OUTOPUT streams.
 * 
 * @param   term
 *          A string to set terminal type environment.
 *          Example: "TERM=vt100"  
 * @param   cmdline
 *          Command line string.
 * @param   mfd_inp
 *          Pointer to a integer which store the master PTY descriptor.
 * @param   mfd_out
 *          Pointer to a integer which store the master PTY descriptor.
 * @param   overlay
 *          Overlapping mode. Not used on Linux always 1
 *
 * @return  On success return PID and set the master PTY descriptor.
 *          On error, -1 is returned, and errno is set appropriately.
 */
int os_shell(char* terminal, char** argv, int* mfd_inp, int *mfd_out, int overlay) 
{
   // PTY master descriptor
   int mfd = -1;
   
   // set TERM environment
   putenv(terminal);
   
   // create child process
   int pid = forkpty(&mfd, NULL, NULL, NULL);   
   
   if (pid == 0) 
   {
      // child process, will never return from this call unless there is an error
      execvp(*argv, argv);
      
      // rc will be -1 and the real problem is in errno (but who is going to see it?)
      return errno;
   }

   if (pid > 0)
   {
      // unblock
      unblock(mfd);      
   }
   
   // in Linux the INPUT and OUTPUT descriptors are the same
   *mfd_inp = mfd;
   *mfd_out = mfd;
   
   // parent
   return pid;
}

/**
 * Wait for a child process to exit.
 * 
 * @param   pid
 *          Child process pid.
 *
 * @return  Process status.
 *          On error, -1 is returned, and errno is set appropriately.
 */
int wait_pid(int pid)
{
   int status = -1;
   waitpid(pid, &status, 0);
   return status;
}

/**
 * Read from master PTY file descriptor.
 * 
 * @param   mfd
 *          Master PTY file descriptor.
 * @param   buffer
 *          The read buffer.
 * @param   size
 *          Maximum size to read.
 *
 * @return  Number of bytes read or -1 on error. 
 */
int read_pty(int mfd, char* buffer, int size)
{
   return read(mfd, buffer, size);
}

/**
 * Write to master PTY file.
 * 
 * @param   mfd
 *          Master PTY file descriptor.
 * @param   buffer
 *          Data to write.
 * @param   size
 *          Size of data to write.
 *
 * @return  Number of bytes written. 
 */
int write_pty(int mfd, void* buffer, int size)
{
   return write(mfd, buffer, (size_t) size);
}