/*
** Module   : ControlFlowOps.java
** Abstract : Progress compatible control flow helper methods
**
** Copyright (c) 2005-2018, Golden Code Development Corporation.
**
** -#- -I- --Date-- --JPRM-- -----------------------Description-----------------------------------
** 001 GES 20050616   @21506 Created initial version, supporting PAUSE.
** 002 SIY 20050815   @22152 Added invoke() method used by RUN VALUE().
** 003 GES 20050829   @22381 Added error condition processing.
** 004 GES 20060117   @23955 Removed PAUSE (it now resides in UI code).
** 005 GES 20060123   @24026 Added return-value support as context-local.
** 006 GES 20060314   @25063 Fixed exception processing in invoke().
** 007 ECF 20060329   @25259 Removed fillInStackTrace calls to exceptions
**                           caught and re-thrown in invoke(). These were
**                           masking the true origins of the exceptions.
** 008 GES 20060427   @25774 Safety code to allow wrapper classes to be
**                           passed as a program name.
** 009 GES 20061030   @30804 Better processing of condition exceptions.
** 010 GES 20070220   @32194 Added support for RUN IN.
** 011 GES 20080619   @38885 If a RUN VALUE (actually any RUN statement)
**                           cannot find the procedure to execute, a STOP
**                           condition is raised (after some error text
**                           has been displayed on the terminal), not an
**                           ERROR. This means that silent error mode does
**                           not apply, the exception is always thrown and
**                           the ERROR-STATUS handle is not updated.
** 012 GES 20080619   @39166 Fix for RUN VALUE of an internal procedure in
**                           compact mode.
** 013 SIY 20080804   @39280 Changed signature of simple invoke(). This
**                           enables correct calls to internal procedures
**                           without redundant instantiation of class and
**                           greatly simplifies compact mode.
**                           Some minor cleanups and potential NPE fix.
** 014 GES 20090422   @41906 Converted to standard string formatting.
** 015 GES 20101005          Added a stub (non-functional) for a RUN statement
**                           that references a PROCEDURE EXTERNAL definition.
**                           Added a stub for DELETE PROCEDURE.
** 016 CA  20130116          Major rework to provide support for IN handle and
**                           IN SUPER functions and procedures, plus RUN SUPER
**                           and SUPER statements.
** 017 CA  20130126          Added support for RELEASE EXTERNAL (converts to
**                           a ControlFlowOps.release(<library>) call, which
**                           is only a stub at this time).
** 018 CA  20130213          Fixed some cases of extent parameter validation.
** 019 CS  20130213          Added APIs supporting the ON server and ASYNC clauses for the RUN 
**                           statement.
** 020 CA  20130223          Added conversion support for the RUN ... SET hPort ON hWebServer stmt
**                           to invoke web services.
** 021 CA  20130315          Added support for cases when a WrappedResource is passed to a handle
**                           parameter.
** 022 OM  20130404          Fixed unknown parameter in function/procedure calls.
**                           Abort function/procedure call if errors occurred when converting 
**                           compatible parameters (ex: 32bit overflow from decimal/int64).
** 023 CA  20130529          Added support for RUN ... [PERSISTENT [handle]] ON SERVER server and
**                           RUN ... IN remoteProcHandle. 
** 024 CA  20130708          Fixed validation of arrays with non-BDT components (these will be 
**                           validated by attempting to instantiate a wrapper using a c'tor which
**                           accepts this array as a parameter).
** 025 CA  20130813          Added support for async procedure calls.
** 026 SVL 20131203          Changes caused by renaming of Nameable.get/setName to Nameable.name.
** 027 CA  20131116          Added support for the FUNCTION ... MAP TO clause.  Fixed runtime for 
**                           RUN SUPER and SUPER() calls, in search-target and search-self modes.
** 028 EVK 20131204          Made enum ArgValidationErrors public. Added method
**                           reachableInternalEntry() to check reachability of procedure or
**                           function in 4gl handle.
** 029 CA  20131220          Refactored the async invocation code and the remote invocation code
**                           to support both web service and appserver requests.
** 030 GES 20140116          Added support for native library calls.
** 031 CA  20140218          Fixed SOURCE- and TARGET-PROCEDURE when invoking external programs.
**                           Error messages must use the program name as it was used at the RUN
**                           statement. Internally, the absolute name must be used (i.e. when 
**                           invoking SourceNameMapper APIs).
** 032 ECF 20140613          Replaced renamed RecordBuffer method in getSignature.
** 033 MAG 20140626          Prevent fault messages to be displayed twice when invoke a web
**                           service.
** 034 VMN 20140702          Added method convertValue() for correct processing unknown value
**                           in invoke... methods.
** 035 HC  20140613          Improved extent parameter support and added logging.
** 036 SVL 20140709          Handle InvalidParameterConditionException in invokeImpl.
** 037 CA  20141106          Changed H035 - log only exceptions which are not propagated.
** 038 CA  20150504          A persistent procedure needs to be saved in its specified handle only
**                           if it completes properly.
** 039 ECF 20150527          Added a cache to InternalEntryCaller as a performance enhancement.
** 040 OM  20150624          Improved procedure parameters type checking for OUT and IN-OUT modes.
**                           Added implicit conversion for parameters.
** 041 OM  20150722          In calls to ControlFlowOps.runSuper(), the modes are unknown. 
** 042 CA  20150813          Added BlockManager.markPendingDynCall to initializeExternalProcedure.
**                           This is required so that the remote procedure call is not interpreted
**                           like a java-style method call, when a proxy procedure is used by an
**                           appserver.
** 043 ECF 20150815          Performance improvements in CacheKey c'tor and getSignature.
** 044 CA  20151024          Fixed assignment of DYNAMIC-FUNCTION ... NO-ERROR when the function 
**                           can not be found.
** 045 OM  20161012          Fixed variable parameter passing in OUTPUT & INPUT-OUTPUT mode.
** 046 CA  20151219          In case of ERROR conditions raised by shared resource management,
**                           if the caller has a NO-ERROR clause, then the condition must NOT be
**                           raised, but logged; also, the external program execution must stop,
**                           and return back to the caller immediately.
** 047 OM  20160212          Called cleanupPending() for various Managers on procedure
**                           instantiation failure.
** 048 CA  20160404          Enhanced the appserver invocations, to allow implicit or explicit 
**                           termination of requests being executing on a certain connection.
** 049 CA  20160418          Added infrastructure for message passing from converted code to a 
**                           customer-specific application or calls from the customer-specific 
**                           application into the converted code (i.e. invoke external programs,
**                           procedures or functions).  Can be used only when the P2J client runs 
**                           embedded in a customer-specific application.
** 050 OM  20160302          ExternalProgramResolver resolves .r code filenames and resolves them
**                           as the .p/.w source code procedure.
** 051 OM  20160527          SourceNam eMapper.convertName() changed to convertNameToClass().
** 052 HC  20160720          RUN, DYNAMIC-FUNCTION and function invocation must not show an error
**                           when the given target handle is invalid and NO-ERROR is specified.
** 053 CA  20160728          Fixed undoable support for data in external programs ran persistent.
** 054 HC  20160817          Fixed PUBLISH to resolve the target internal procedure in the
**                           subscribed procedure as well as in its SUPER procedures or SESSION
**                           procedures.
**                           Fixed RUN invocation logic to set up the correct TARGET-PROCEDURE
**                           when the resolved internal procedure is found in a SUPER procedure.
** 055 HC  20160824          Fixed a regression from the previous change, FUNCTION ... IN
**                           statement did not properly resolve the function from the super
**                           procedures.
** 056 HC  20160901          Fixed error handling of procedure handle validation in a RUN or
**                           function invocation, a regression introduced in H052.
** 057 ECF 20160912          Flattened invocation calling hierarchy. Although this is somewhat
**                           less maintainable due to argument list duplication across multiple
**                           methods, it avoids several layers of method calls in heavily executed
**                           code paths.
** 058 CA  20161020          Fixed case-insensitive resolution with .r (rcode) ext program names.
** 059 CA  20170328          Fixed a problem with parameter validation when invoked by a remote 
**                           non-P2J side (i.e. API invoked by an embedded client).
** 060 GES 20171206          Switch to ErrorManager.silent(). Added TODOs.
** 061 ECF 20171227          Performance improvement.
**     GES 20180102          Clarified error handling comments.
** 062 EVL 20180202          Filter pre/post steps for native library calls.
** 063 CA  20180510          Delegate lockWindowUpdate user32.dll calls to LT.disableRedraw.
**     CA  20180521          The IN SUPER function definition in the current file may provide a 
**                           return type different than its actual implementation - this affects 
**                           function calls, so that it must not fail if return types differ. 
*/

/*
** This program is free software: you can redistribute it and/or modify
** it under the terms of the GNU Affero General Public License as
** published by the Free Software Foundation, either version 3 of the
** License, or (at your option) any later version.
**
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
** GNU Affero General Public License for more details.
**
** You may find a copy of the GNU Affero GPL version 3 at the following
** location: https://www.gnu.org/licenses/agpl-3.0.en.html
** 
** Additional terms under GNU Affero GPL version 3 section 7:
** 
**   Under Section 7 of the GNU Affero GPL version 3, the following additional
**   terms apply to the works covered under the License.  These additional terms
**   are non-permissive additional terms allowed under Section 7 of the GNU
**   Affero GPL version 3 and may not be removed by you.
** 
**   0. Attribution Requirement.
** 
**     You must preserve all legal notices or author attributions in the covered
**     work or Appropriate Legal Notices displayed by works containing the covered
**     work.  You may not remove from the covered work any author or developer
**     credit already included within the covered work.
** 
**   1. No License To Use Trademarks.
** 
**     This license does not grant any license or rights to use the trademarks
**     Golden Code, FWD, any Golden Code or FWD logo, or any other trademarks
**     of Golden Code Development Corporation. You are not authorized to use the
**     name Golden Code, FWD, or the names of any author or contributor, for
**     publicity purposes without written authorization.
** 
**   2. No Misrepresentation of Affiliation.
** 
**     You may not represent yourself as Golden Code Development Corporation or FWD.
** 
**     You may not represent yourself for publicity purposes as associated with
**     Golden Code Development Corporation, FWD, or any author or contributor to
**     the covered work, without written authorization.
** 
**   3. No Misrepresentation of Source or Origin.
** 
**     You may not represent the covered work as solely your work.  All modified
**     versions of the covered work must be marked in a reasonable way to make it
**     clear that the modified work is not originating from Golden Code Development
**     Corporation or FWD.  All modified versions must contain the notices of
**     attribution required in this license.
*/

package com.goldencode.p2j.util;

import java.lang.reflect.*;
import java.util.*;
import java.util.concurrent.*;
import java.util.logging.*;
import com.goldencode.cache.*;
import com.goldencode.p2j.*;
import com.goldencode.p2j.library.*;
import com.goldencode.p2j.main.*;
import com.goldencode.p2j.net.*;
import com.goldencode.p2j.persist.*;
import com.goldencode.p2j.security.*;
import com.goldencode.p2j.security.SecurityManager;
import com.goldencode.p2j.ui.*;
import com.goldencode.p2j.util.InvocationRequestPayload.InvokeType;

/**
 * Progress compatible control-flow related static utility methods. Current
 * implementation contains support for the:
 * <p>
 * <pre>
 * Progress 4GL Feature                  Java Implementation
 * -----------------------------------   ------------------------------------
 * RUN [ &lt;procedure&gt; | VALUE() ]   invoke()
 * RUN ... IN &lt;handle&gt;             invokeIn()
 * RUN ... PERSIST                       invokePersistent()
 * RUN ... PERSIST SET &lt;handle&gt;    invokePersistentSet() 
 * RETURN-VALUE "global variable"        getReturnValue()/setReturnValue()
 * RUN SUPER                             runSuper()
 * simple function invocation            invokeFunction()
 * FUNCTION ... IN handle                invokeFunctionIn()
 * DYNAMIC-FUNCTION                      invokeDynamicFunction()
 * DYNAMIC-FUNCTION(... IN)              invokeDynamicFunctionIn()
 * SUPER()                               runSuper()
 * RELEASE EXTERNAL [PROCEDURE]          release()
 * </pre>
 */
public class ControlFlowOps
{
   /** Logger */
   private static final Logger LOG = LogHelper.getLogger(ControlFlowOps.class);

   /** Enum with the list of possible argument validation errors. */
   public enum ArgValidationErrors 
   {
      NO_ERROR,
      INCORRECT_COUNT,
      UNEXPECTED_PARAM,
      INCORRECT_TYPES,
      DYN_ARRAY_FOR_NON_ARRAY,
      NON_ARRAY_FOR_DYN_ARRAY,
      ARRAY_FOR_NON_ARRAY,
      NON_ARRAY_FOR_ARRAY,
      BAD_ARRAY_DIMENSION,
      DEFER_VALIDATION
   }
   
   /** Stores context-local state variables. */
   private static ContextContainer work = new ContextContainer();
   
   /**
    * Attempts to unlpad the specified library from memory (the RELEASE EXTERNAL statement).
    * This code just delegates the processing to the {@link NativeInvoker#release} method.
    * 
    * @param    libname
    *           The name of the library to attempt to unload.
    */
   public static void release(character libname)
   {
      if (libname == null)
      {
         // quick/silent out, there is no way to unload a library in this case
         return;
      }
      
      if (libname.isUnknown())
      {
         // strange case where the error is raised in normal mode (without any message
         // display) and in silent mode only the error flag is set (but normal details
         // are not recorded)
         ErrorManager.noRecordOrThrowError("Unknown cannot be given for a libname!");
      }
      else
      {
         release(libname.toStringMessage());
      }
   }
   
   /**
    * Attempts to unload the specified library from memory (the RELEASE EXTERNAL statement).
    * This code just delegates the processing to the {@link NativeInvoker#release} method.
    * 
    * @param    libname
    *           The name of the library to attempt to unload. If <code>null</code> or the
    *           empty string are passed, this is a no-operation.
    */
   public static void release(String libname)
   {
      if (libname == null || libname.length() == 0)
      {
         // quick/silent out, there is no way to unload a library in this case
         return;
      }
      
      NativeInvoker.release(libname);
   }
   
   /**
    * Gets the most recent RETURN-VALUE for the current user's session.
    *
    * @return   The RETURN-VALUE which defaults to the empty string if a
    *           the equivalent of a Progress 4GL RUN statement has executed
    *           but no explicit RETURN-VALUE was set.
    */
   public static character getReturnValue()
   {
      return work.obtain().returnValue;
   }
   
   /**
    * Sets the most recent RETURN-VALUE for the current user's session.
    *
    * @param    returnValue
    *           The RETURN-VALUE to set or <code>null</code> to assign the
    *           empty string.
    */
   public static void setReturnValue(character returnValue)
   {
      if (returnValue == null)
      {
         returnValue = new character("");
      }
      
      work.obtain().returnValue.assign(returnValue);
   }
   
   /**
    * Sets the most recent RETURN-VALUE for the current user's session.
    *
    * @param    returnValue
    *           The RETURN-VALUE to set or <code>null</code> to assign the
    *           empty string.
    */
   public static void setReturnValue(String returnValue)
   {
      setReturnValue(new character(returnValue == null ? "" : returnValue));
   }

   /**
    * Execute the RUN SUPER statement or SUPER function. Runtime decides on
    * whether this is a function or procedure, based on the 
    * {@link TransactionManager#nearestExecutingBlock() nearest block}.
    * 
    * @param    iename
    *           The internal-entry's legacy 4GL name.
    * @param    args
    *           The argument list.
    *           
    * @return   The return value from the block (or its nested blocks) which
    *           is the <code>unknown value</code> by default, on error or for
    *           procedures.
    */
   public static BaseDataType runSuper(String iename, Object... args)
   {
      BlockDefinition block = TransactionManager.nearestExecutingBlock();
      boolean function = (block.type == BlockType.FUNCTION);
      handle target = ProcedureManager.getSuperHandle(iename, function);
      if (target == null)
      {
         return new unknown();
      }

      character name = new character(iename);
      
      if (function)
      {
         return invokeFunctionImpl(name, target, false, true, null, args);
      }
      else
      {
         invokeInImpl(name, target, true, null, args);
         return new unknown();
      }
   }

   /**
    * Invoke the procedure with the given name (emitted for a standard RUN 
    * statement).
    * 
    * @param    name
    *           The legacy 4GL name for the procedure.
    * @param    modes
    *           A string representation of the modes of each parameter.
    * @param    args
    *           The procedure's arguments.
    */
   public static void invokeWithMode(String    name,
                                     String    modes, 
                                     Object... args) 
   {
      invokeWithMode(new character(name), modes, args);
   }

   /**
    * Invoke the procedure with the given name (emitted for a standard RUN 
    * statement).
    * 
    * @param    name
    *           The legacy 4GL name for the procedure.
    * @param    modes
    *           A string representation of the modes of each parameter.
    * @param    args
    *           The procedure's arguments.
    */
   public static void invokeWithMode(character name,
                                     String    modes,
                                     Object... args)
   {
      invokeImpl((handle) null, null, name, false, false, false, false, modes, args);
   }

   /**
    * Invoke the procedure with the given name (emitted for a standard RUN 
    * statement).
    * 
    * @param    name
    *           The legacy 4GL name for the procedure.
    * @param    args
    *           The procedure's arguments.
    */
   public static void invoke(String name, Object... args) 
   {
      invoke(new character(name), args);
   }

   /**
    * Invoke the procedure with the given name (emitted for a standard RUN 
    * statement).
    * 
    * @param    name
    *           The legacy 4GL name for the procedure.
    * @param    args
    *           The procedure's arguments.
    */
   public static void invoke(character name, Object... args)
   {
      invokeImpl((handle) null, null, name, false, false, false, false, null, args);
   }

   /**
    * Invoke the procedure with the given name (emitted for a standard RUN statement).
    * 
    * @param    name
    *           The legacy 4GL name for the procedure.
    * @param    portHandle
    *           A handle variable where to save the web service. May be null.
    * @param    remoteHandle
    *           The handle to the remote object used to run the procedure
    * @param    transactionDistinct
    *           This marks if "TRANSACTION DISTINCT" option is specified with the RUN ... ON stmt.
    * @param    modes
    *           A string representation of the modes of each parameter.
    * @param    args
    *           The procedure's arguments.
    */
   public static void invokeRemoteWithMode(String    name, 
                                           handle    portHandle,
                                           handle    remoteHandle,
                                           boolean   transactionDistinct,
                                           String    modes, 
                                           Object... args) 
   {
      invokeRemoteWithMode(new character(name), portHandle,
                           remoteHandle, transactionDistinct,
                           modes, args);
   }

   /**
    * Invoke the procedure with the given name (emitted for a standard RUN statement).
    * 
    * @param    name
    *           The legacy 4GL name for the procedure.
    * @param    portHandle
    *           A handle variable where to save the web service. May be null.
    * @param    remoteHandle
    *           The handle to the remote object used to run the procedure
    * @param    transactionDistinct
    *           This marks if "TRANSACTION DISTINCT" option is specified with the RUN ... ON stmt.
    * @param    modes
    *           A string representation of the modes of each parameter.
    * @param    args
    *           The procedure's arguments.
    */
   public static void invokeRemoteWithMode(character name, 
                                           handle    portHandle,
                                           handle    remoteHandle,
                                           boolean   transactionDistinct,
                                           String    modes, 
                                           Object... args) 
   {
      invokeRemoteWithModeImpl(null, name,
                               portHandle, remoteHandle,
                               transactionDistinct,
                               modes, args);
   }

   /**
    * Invoke the procedure with the given name (emitted for a standard RUN statement).
    * 
    * @param    asyncReq
    *           When not-null, an {@link AsyncRequestImpl} instance for the associated async
    *           request.
    * @param    name
    *           The legacy 4GL name for the procedure.
    * @param    portHandle
    *           A handle variable where to save the web service. May be null.
    * @param    remoteHandle
    *           The handle to the remote object used to run the procedure
    * @param    transactionDistinct
    *           This marks if "TRANSACTION DISTINCT" option is specified with the RUN ... ON stmt.
    * @param    modes
    *           A string representation of the modes of each parameter.
    * @param    args
    *           The procedure's arguments.
    */
   private static void invokeRemoteWithModeImpl(AsyncRequestImpl asyncReq,
                                                character        name, 
                                                handle           portHandle,
                                                handle           remoteHandle,
                                                boolean          transactionDistinct,
                                                String           modes, 
                                                Object...        args) 
   {
      if (portHandle != null)
      {
         // this is a RUN ... SET ... ON SERVER webService
         invokePersistentImpl(asyncReq, remoteHandle, name, null, portHandle,
                              transactionDistinct, modes, args);
      }
      else
      {
         // this is a RUN ... ON SERVER appServer
         invokeImpl(asyncReq, remoteHandle, null, name,
                    false, false, false,
                    transactionDistinct, modes, args);
      }
   }

   /**
    * Invoke the procedure with the given name on a remote server (emitted for a standard RUN ... 
    * ON ... statement).
    * 
    * @param    name
    *           The legacy 4GL name for the procedure.
    * @param    portHandle
    *           A handle variable where to save the web service. May be null.
    * @param    remoteHandle
    *           The handle to the remote object used to run the procedure
    * @param    transactionDistinct
    *           This marks if "TRANSACTION DISTINCT" option is specified with the RUN ... ON stmt.
    * @param    args
    *           The procedure's arguments.
    */
   public static void invokeRemote(String    name, 
                                   handle    portHandle,
                                   handle    remoteHandle, 
                                   boolean   transactionDistinct, 
                                   Object... args)
   {
      invokeRemote(new character(name), portHandle, remoteHandle, transactionDistinct, args);
   }

   /**
    * Invoke the procedure with the given name on a remote server  (emitted for a standard RUN ... 
    * ON ... statement).
    * 
    * @param    name
    *           The legacy 4GL name for the procedure.
    * @param    portHandle
    *           A handle variable where to save the web service. May be null.
    * @param    remoteHandle
    *           The handle to the remote object used to run the procedure
    * @param    transactionDistinct
    *           This marks if "TRANSACTION DISTINCT" option is specified with the RUN ... ON stmt.
    * @param    args
    *           The procedure's arguments.
    */
   public static void invokeRemote(character name,
                                   handle    portHandle,
                                   handle    remoteHandle, 
                                   boolean   transactionDistinct, 
                                   Object... args)
   {
      invokeRemoteWithMode(new character(name), portHandle,
                           remoteHandle, transactionDistinct,
                           null, args);
   }

   /**
    * Invoke the procedure with the given name on a remote server (emitted for a standard RUN ... 
    * ON ... ASYNC SET ... statement).
    * 
    * @param    name
    *           The legacy 4GL name for the procedure.
    * @param    portHandle
    *           A handle variable where to save the web service. May be null.
    * @param    remoteHandle
    *           The handle to the remote object used to run the procedure
    * @param    transactionDistinct
    *           This marks if "TRANSACTION DISTINCT" option is specified with the RUN ... ON stmt.
    * @param    asyncHandle
    *           The handle to the asynchronous request.
    * @param    eventProcName
    *           This represents the name of the internal procedure for the EVENT-PROCEDURE clause,
    *           null if this clause is not specified.
    * @param    inEventprocHandle
    *           The handle to a procedure in which context the specified EVENT-PROCEDURE internal
    *           procedure is found, null if EVENT-PROCEDURE clause is not specified.
    * @param    modes
    *           A string representation of the modes of each parameter.
    * @param    args
    *           The procedure's arguments.
    */
   public static void invokeRemoteAsyncWithMode(String    name, 
                                                handle    portHandle,
                                                handle    remoteHandle, 
                                                boolean   transactionDistinct,
                                                handle    asyncHandle,
                                                String    eventProcName,
                                                handle    inEventprocHandle,
                                                String    modes, 
                                                Object... args)
   {
      invokeRemoteAsyncWithMode(new character(name), portHandle, 
                                remoteHandle, transactionDistinct, 
                                asyncHandle, new character(eventProcName), inEventprocHandle, 
                                modes, args);
   }

   /**
    * Invoke the procedure on a remote server with the given name 
    * (emitted for a standard RUN ... ON ... ASYNC SET ...
    * 
    * @param    name
    *           The legacy 4GL name for the procedure.
    * @param    portHandle
    *           A handle variable where to save the web service. May be null.
    * @param    remoteHandle
    *           The handle to the remote object used to run the procedure
    * @param    transactionDistinct
    *           This marks if "TRANSACTION DISTINCT" option is specified with the RUN ... ON stmt
    * @param    asyncHandle
    *           The handle to the asynchronous request.
    * @param    eventProcName
    *           This represents the name of the internal procedure for the EVENT-PROCEDURE clause,
    *           null if this clause is not specified.
    * @param    inEventprocHandle
    *           The handle to a procedure in which context the specified EVENT-PROCEDURE internal
    *           procedure is found, null if EVENT-PROCEDURE clause is not specified.
    * @param    modes
    *           A string representation of the modes of each parameter.
    * @param    args
    *           The procedure's arguments.
    */
   public static void invokeRemoteAsyncWithMode(character name, 
                                                handle    portHandle,
                                                handle    remoteHandle, 
                                                boolean   transactionDistinct,
                                                handle    asyncHandle,
                                                String    eventProcName,
                                                handle    inEventprocHandle,
                                                String    modes, 
                                                Object... args)
   {
      invokeRemoteAsyncWithMode(name, portHandle, remoteHandle, transactionDistinct,
                                asyncHandle, new character(eventProcName), inEventprocHandle, 
                                modes, args);
   }

   /**
    * Invoke the procedure on a remote server with the given name 
    * (emitted for a standard RUN ... ON ... ASYNC SET ...
    * 
    * @param    name
    *           The legacy 4GL name for the procedure.
    * @param    portHandle
    *           A handle variable where to save the web service. May be null. May be null.
    * @param    remoteHandle
    *           The handle to the remote object used to run the procedure
    * @param    transactionDistinct
    *           This marks if "TRANSACTION DISTINCT" option is specified with the RUN ... ON stmt
    * @param    asyncHandle
    *           The handle to the asynchronous request.
    * @param    eventProcName
    *           This represents the name of the internal procedure for the EVENT-PROCEDURE clause,
    *           null if this clause is not specified.
    * @param    inEventprocHandle
    *           The handle to a procedure in which context the specified EVENT-PROCEDURE internal
    *           procedure is found, null if EVENT-PROCEDURE clause is not specified.
    * @param    modes
    *           A string representation of the modes of each parameter.
    * @param    args
    *           The procedure's arguments.
    */
   public static void invokeRemoteAsyncWithMode(String    name, 
                                                handle    portHandle,
                                                handle    remoteHandle, 
                                                boolean   transactionDistinct,
                                                handle    asyncHandle,
                                                character eventProcName,
                                                handle    inEventprocHandle,
                                                String    modes, 
                                                Object... args)
   {
      invokeRemoteAsyncWithMode(new character(name), portHandle,
                                remoteHandle, transactionDistinct,
                                asyncHandle, eventProcName, inEventprocHandle,
                                modes, args);
   }

   /**
    * Invoke the procedure on a remote server with the given name 
    * (emitted for a standard RUN ... ON ... ASYNC SET ...
    * 
    * @param    name
    *           The legacy 4GL name for the procedure.
    * @param    portHandle
    *           A handle variable where to save the web service. May be null.
    * @param    remoteHandle
    *           The handle to the remote object used to run the procedure
    * @param    transactionDistinct
    *           This marks if "TRANSACTION DISTINCT" option is specified with the RUN ... ON stmt
    * @param    asyncHandle
    *           The handle to the asynchronous request.
    * @param    eventProcName
    *           This represents the name of the internal procedure for the EVENT-PROCEDURE clause,
    *           null if this clause is not specified.
    * @param    inEventprocHandle
    *           The handle to a procedure in which context the specified EVENT-PROCEDURE internal
    *           procedure is found, null if EVENT-PROCEDURE clause is not specified.
    * @param    modes
    *           A string representation of the modes of each parameter.
    * @param    args
    *           The procedure's arguments.
    */
   public static void invokeRemoteAsyncWithMode(character name, 
                                                handle    portHandle,
                                                handle    remoteHandle, 
                                                boolean   transactionDistinct,
                                                handle    asyncHandle,
                                                character eventProcName,
                                                handle    inEventprocHandle,
                                                String    modes, 
                                                Object... args)
   {
      AsyncCall remoteRequest = new AsyncCall(remoteHandle, portHandle, name, new handle(),
                                              asyncHandle, eventProcName, inEventprocHandle,
                                              transactionDistinct,
                                              modes, args)
      {
         @Override
         public void run()
         {
            // use a copy here, to not allow the user to change the reference
            invokeRemoteWithModeImpl(asyncReq, pName,
                                     portHandle, hServer,
                                     transactionDistinct,
                                     asyncReq.getModes(), asyncReq.getArgs());
         }
      };

      invokeAsyncImpl(remoteRequest, remoteHandle, asyncHandle, new handle());
   }

   /**
    * Invoke the procedure on a remote server with the given name 
    * (emitted for a standard RUN ... ON ... ASYNC SET ...
    * 
    * @param    name
    *           The legacy 4GL name for the procedure.
    * @param    portHandle
    *           A handle variable where to save the web service. May be null.
    * @param    remoteHandle
    *           The handle to the remote object used to run the procedure
    * @param    transactionDistinct
    *           This marks if "TRANSACTION DISTINCT" option is specified with the RUN ... ON stmt
    * @param    asyncHandle
    *           The handle to the asynchronous request.
    * @param    eventProcName
    *           This represents the name of the internal procedure for the EVENT-PROCEDURE clause,
    *           null if this clause is not specified.
    * @param    inEventprocHandle
    *           The handle to a procedure in which context the specified EVENT-PROCEDURE internal
    *           procedure is found, null if EVENT-PROCEDURE clause is not specified.
    * @param    args
    *           The procedure's arguments.
    */
   public static void invokeRemoteAsync(String    name, 
                                        handle    portHandle,
                                        handle    remoteHandle, 
                                        boolean   transactionDistinct,
                                        handle    asyncHandle,
                                        String    eventProcName,
                                        handle    inEventprocHandle,
                                        Object... args)
   {
      invokeRemoteAsync(new character(name), portHandle, remoteHandle , transactionDistinct,
                        asyncHandle, new character(eventProcName), inEventprocHandle,
                        args);
   }

   /**
    * Invoke the procedure on a remote server with the given name 
    * (emitted for a standard RUN ... ON ... ASYNC SET ...
    * 
    * @param    name
    *           The legacy 4GL name for the procedure.
    * @param    portHandle
    *           A handle variable where to save the web service. May be null.
    * @param    remoteHandle
    *           The handle to the remote object used to run the procedure
    * @param    transactionDistinct
    *           This marks if "TRANSACTION DISTINCT" option is specified with the RUN ... ON stmt
    * @param    asyncHandle
    *           The handle to the asynchronous request.
    * @param    eventProcName
    *           This represents the name of the internal procedure for the EVENT-PROCEDURE clause,
    *           null if this clause is not specified.
    * @param    inEventprocHandle
    *           The handle to a procedure in which context the specified EVENT-PROCEDURE internal
    *           procedure is found, null if EVENT-PROCEDURE clause is not specified.
    * @param    args
    *           The procedure's arguments.
    */
   public static void invokeRemoteAsync(character name, 
                                        handle    portHandle,
                                        handle    remoteHandle, 
                                        boolean   transactionDistinct,
                                        handle    asyncHandle,
                                        String    eventProcName,
                                        handle    inEventprocHandle,
                                        Object... args)
   {
      invokeRemoteAsync(name, portHandle, remoteHandle, transactionDistinct,
                        asyncHandle, new character(eventProcName), inEventprocHandle,
                        args);
   }

   /**
    * Invoke the procedure on a remote server with the given name 
    * (emitted for a standard RUN ... ON ... ASYNC SET ...
    * 
    * @param    name
    *           The legacy 4GL name for the procedure.
    * @param    portHandle
    *           A handle variable where to save the web service. May be null.
    * @param    remoteHandle
    *           The handle to the remote object used to run the procedure
    * @param    transactionDistinct
    *           This marks if "TRANSACTION DISTINCT" option is specified with the RUN ... ON stmt
    * @param    asyncHandle
    *           The handle to the asynchronous request.
    * @param    eventProcName
    *           This represents the name of the internal procedure for the EVENT-PROCEDURE clause,
    *           null if this clause is not specified.
    * @param    inEventprocHandle
    *           The handle to a procedure in which context the specified EVENT-PROCEDURE internal
    *           procedure is found, null if EVENT-PROCEDURE clause is not specified.
    * @param    args
    *           The procedure's arguments.
    */
   public static void invokeRemoteAsync(String    name, 
                                        handle    portHandle,
                                        handle    remoteHandle, 
                                        boolean   transactionDistinct,
                                        handle    asyncHandle,
                                        character eventProcName,
                                        handle    inEventprocHandle,
                                        Object... args)
   {
      invokeRemoteAsync(new character(name), portHandle, remoteHandle, transactionDistinct,
                        asyncHandle, eventProcName, inEventprocHandle,
                        args);
   }

   /**
    * Invoke the procedure on a remote server with the given name 
    * (emitted for a standard RUN ... ON ... ASYNC SET ...
    * 
    * @param    name
    *           The legacy 4GL name for the procedure.
    * @param    portHandle
    *           A handle variable where to save the web service. May be null.
    * @param    remoteHandle
    *           The handle to the remote object used to run the procedure
    * @param    transactionDistinct
    *           This marks if "TRANSACTION DISTINCT" option is specified with the RUN ... ON stmt
    * @param    asyncHandle
    *           The handle to the asynchronous request.
    * @param    eventProcName
    *           This represents the name of the internal procedure for the EVENT-PROCEDURE clause,
    *           null if this clause is not specified.
    * @param    inEventprocHandle
    *           The handle to a procedure in which context the specified EVENT-PROCEDURE internal
    *           procedure is found, null if EVENT-PROCEDURE clause is not specified.
    * @param    args
    *           The procedure's arguments.
    */
   public static void invokeRemoteAsync(character name, 
                                        handle    portHandle,
                                        handle    remoteHandle, 
                                        boolean   transactionDistinct,
                                        handle    asyncHandle,
                                        character eventProcName,
                                        handle    inEventprocHandle,
                                        Object... args)
   {
      invokeRemoteAsyncWithMode(name, portHandle, remoteHandle, transactionDistinct,
                                asyncHandle, eventProcName, inEventprocHandle,
                                null, args);
   }

   /**
    * Invoke the procedure with the given name, defined in the given handle
    * (emitted for a standard RUN ... IN handle statement).
    * 
    * @param    name
    *           The legacy 4GL name for the procedure.
    * @param    h
    *           The handle to which this procedure belongs. When 
    *           <code>null</code>, defaults to THIS-PROCEDURE.
    * @param    modes
    *           A string representation of the modes of each parameter.
    * @param    args
    *           The procedure's arguments.
    */
   public static void invokeInWithMode(String    name,
                                       handle    h,
                                       String    modes,
                                       Object... args)
   {
      invokeInImpl(new character(name), h, false, modes, args);
   }

   /**
    * Invoke the procedure with the given name, defined in the given handle
    * (emitted for a standard RUN ... IN handle statement).
    * 
    * @param    name
    *           The legacy 4GL name for the procedure.
    * @param    h
    *           The handle to which this procedure belongs. When 
    *           <code>null</code>, defaults to THIS-PROCEDURE.
    * @param    modes
    *           A string representation of the modes of each parameter.
    * @param    args
    *           The procedure's arguments.
    */
   public static void invokeInWithMode(character name,
                                       handle    h,
                                       String    modes,
                                       Object... args)
   {
      invokeInImpl(name, h, false, modes, args);
   }

   /**
    * Invoke the procedure with the given name, defined in the given handle
    * (emitted for a standard RUN ... IN handle statement).
    * 
    * @param    name
    *           The legacy 4GL name for the procedure.
    * @param    h
    *           The handle to which this procedure belongs. When 
    *           <code>null</code>, defaults to THIS-PROCEDURE.
    * @param    args
    *           The procedure's arguments.
    */
   public static void invokeIn(String name, handle h, Object... args)
   {
      invokeInImpl(new character(name), h, false, null, args);
   }

   /**
    * Invoke the procedure with the given name, defined in the given handle
    * (emitted for a standard RUN ... IN handle statement).
    * 
    * @param    name
    *           The legacy 4GL name for the procedure.
    * @param    h
    *           The handle to which this procedure belongs. When 
    *           <code>null</code>, defaults to THIS-PROCEDURE.
    * @param    args
    *           The procedure's arguments.
    */
   public static void invokeIn(character name, handle h, Object... args)
   {
      invokeInImpl(name, h, false, null, args);
   }

   /**
    * Invoke the procedure with the given name, defined in the given handle
    * (emitted for a standard RUN ... IN handle statement).
    * 
    * @param    name
    *           The legacy 4GL name for the procedure.
    * @param    h
    *           The handle to which this procedure belongs. When 
    *           <code>null</code>, defaults to THIS-PROCEDURE.
    * @param    asyncHandle
    *           The handle to the asynchronous request.
    * @param    eventProcName
    *           This represents the name of the internal procedure for the EVENT-PROCEDURE clause,
    *           null if this clause is not specified.
    * @param    inEventprocHandle
    *           The handle to a procedure in which context the specified EVENT-PROCEDURE internal
    *           procedure is found, null if EVENT-PROCEDURE clause is not specified.
    * @param    modes
    *           A string representation of the modes of each parameter.
    * @param    args
    *           The procedure's arguments.
    */
   public static void invokeInAsyncWithMode(String    name,
                                            handle    h,
                                            handle    asyncHandle, 
                                            String    eventProcName,
                                            handle    inEventprocHandle,
                                            String    modes,
                                            Object... args)
   {
      invokeInAsyncWithMode(new character(name), h,
                            asyncHandle, new  character(eventProcName), inEventprocHandle,
                            modes, args);
   }

   /**
    * Invoke the procedure with the given name, defined in the given handle
    * (emitted for a standard RUN ... IN ASYNC.. handle statement).
    * 
    * @param    name
    *           The legacy 4GL name for the procedure.
    * @param    h
    *           The handle to which this procedure belongs. When 
    *           <code>null</code>, defaults to THIS-PROCEDURE.
    * @param    asyncHandle
    *           The handle to the asynchronous request.
    * @param    eventProcName
    *           This represents the name of the internal procedure for the EVENT-PROCEDURE clause,
    *           null if this clause is not specified.
    * @param    inEventprocHandle
    *           The handle to a procedure in which context the specified EVENT-PROCEDURE internal
    *           procedure is found, null if EVENT-PROCEDURE clause is not specified.
    * @param    modes
    *           A string representation of the modes of each parameter.
    * @param    args
    *           The procedure's arguments.
    */
   public static void invokeInAsyncWithMode(character name,
                                            handle    h,
                                            handle    asyncHandle, 
                                            String    eventProcName,
                                            handle    inEventprocHandle,
                                            String    modes,
                                            Object... args)
   {
      invokeInAsyncWithMode(name, h,
                            asyncHandle, new character(eventProcName), inEventprocHandle,
                            modes, args);
   }

   /**
    * Invoke the procedure with the given name, defined in the given handle
    * (emitted for a standard RUN ... IN ASYNC.. handle statement).
    * 
    * @param    name
    *           The legacy 4GL name for the procedure.
    * @param    h
    *           The handle to which this procedure belongs. When 
    *           <code>null</code>, defaults to THIS-PROCEDURE.
    * @param    asyncHandle
    *           The handle to the asynchronous request.
    * @param    eventProcName
    *           This represents the name of the internal procedure for the EVENT-PROCEDURE clause,
    *           null if this clause is not specified.
    * @param    inEventprocHandle
    *           The handle to a procedure in which context the specified EVENT-PROCEDURE internal
    *           procedure is found, null if EVENT-PROCEDURE clause is not specified.
    * @param    modes
    *           A string representation of the modes of each parameter.
    * @param    args
    *           The procedure's arguments.
    */
   public static void invokeInAsyncWithMode(String    name,
                                            handle    h,
                                            handle    asyncHandle, 
                                            character eventProcName,
                                            handle    inEventprocHandle,
                                            String    modes,
                                            Object... args)
   {
      invokeInAsyncWithMode(new character(name), h, asyncHandle, eventProcName, inEventprocHandle,
                            modes, args);
   }

   /**
    * Invoke the procedure with the given name, defined in the given handle
    * (emitted for a standard RUN ... IN handle statement).
    * 
    * @param    name
    *           The legacy 4GL name for the procedure.
    * @param    h
    *           The handle to which this procedure belongs. When 
    *           <code>null</code>, defaults to THIS-PROCEDURE.
    * @param    asyncHandle
    *           The handle to the asynchronous request.
    * @param    eventProcName
    *           This represents the name of the internal procedure for the EVENT-PROCEDURE clause,
    *           null if this clause is not specified.
    * @param    inEventprocHandle
    *           The handle to a procedure in which context the specified EVENT-PROCEDURE internal
    *           procedure is found, null if EVENT-PROCEDURE clause is not specified.
    * @param    modes
    *           A string representation of the modes of each parameter.
    * @param    args
    *           The procedure's arguments.
    */
   public static void invokeInAsyncWithMode(character name,
                                            handle    h,
                                            handle    asyncHandle, 
                                            character eventProcName,
                                            handle    inEventprocHandle,
                                            String    modes,
                                            Object... args)
   {
      ServerImpl server = null;
      if (h != null && h._isValid())
      {
         WrappedResource res = h.getResource();
         if (res instanceof PersistentProcedure)
         {
            handle hserver = ((PersistentProcedure) res).getServerHandle();
            
            if (!hserver.isUnknown())
            {
               server = (ServerImpl) hserver.getResource();
            }
         }
      }

      // if no server, default to SESSION
      handle hServer = (server != null ? new handle(server) : SessionUtils.asHandle());
      
      AsyncCall remoteRequest = new AsyncCall(hServer, null, name, h,
                                              asyncHandle, eventProcName, inEventprocHandle,
                                              false,
                                              modes, args)
      {
         @Override
         public void run()
         {
            // use a copy here, to not allow the user to change the reference
            invokeImpl(asyncReq, (handle) null, pHandle, pName, 
                       false, false, false, false,
                       asyncReq.getModes(), asyncReq.getArgs());
         }
      };

      invokeAsyncImpl(remoteRequest, hServer, asyncHandle, h);
   }
   
   /**
    * Invoke the procedure with the given name, defined in the given handle
    * (emitted for a standard RUN ... IN handle statement).
    * 
    * @param    name
    *           The legacy 4GL name for the procedure.
    * @param    h
    *           The handle to which this procedure belongs. When 
    *           <code>null</code>, defaults to THIS-PROCEDURE.
    * @param    asyncHandle
    *           The handle to the asynchronous request.
    * @param    eventProcName
    *           This represents the name of the internal procedure for the EVENT-PROCEDURE clause,
    *           null if this clause is not specified.
    * @param    inEventprocHandle
    *           The handle to a procedure in which context the specified EVENT-PROCEDURE internal
    *           procedure is found, null if EVENT-PROCEDURE clause is not specified.
    * @param    args
    *           The procedure's arguments.
    */
   public static void invokeInAsync(String    name,
                                    handle    h,
                                    handle    asyncHandle, 
                                    String    eventProcName,
                                    handle    inEventprocHandle,
                                    Object... args)
   {
      invokeInAsync(new character(name), h,
                    asyncHandle, new  character(eventProcName), inEventprocHandle, args);
   }

   /**
    * Invoke the procedure with the given name, defined in the given handle
    * (emitted for a standard RUN ... IN handle ASYNC..  statement).
    * 
    * @param    name
    *           The legacy 4GL name for the procedure.
    * @param    h
    *           The handle to which this procedure belongs. When 
    *           <code>null</code>, defaults to THIS-PROCEDURE.
    * @param    asyncHandle
    *           The handle to the asynchronous request.
    * @param    eventProcName
    *           This represents the name of the internal procedure for the EVENT-PROCEDURE clause,
    *           null if this clause is not specified.
    * @param    inEventprocHandle
    *           The handle to a procedure in which context the specified EVENT-PROCEDURE internal
    *           procedure is found, null if EVENT-PROCEDURE clause is not specified.
    * @param    args
    *           The procedure's arguments.
    */
   public static void invokeInAsync(character name,
                                    handle    h,
                                    handle    asyncHandle, 
                                    String    eventProcName,
                                    handle    inEventprocHandle,
                                    Object... args)
   {
      invokeInAsync(name, h, asyncHandle, new character(eventProcName), inEventprocHandle, args);
   }

   /**
    * Invoke the procedure with the given name, defined in the given handle
    * (emitted for a standard RUN ... IN handle ASYNC.. statement).
    * 
    * @param    name
    *           The legacy 4GL name for the procedure.
    * @param    h
    *           The handle to which this procedure belongs. When 
    *           <code>null</code>, defaults to THIS-PROCEDURE.
    * @param    asyncHandle
    *           The handle to the asynchronous request.
    * @param    eventProcName
    *           This represents the name of the internal procedure for the EVENT-PROCEDURE clause,
    *           null if this clause is not specified.
    * @param    inEventprocHandle
    *           The handle to a procedure in which context the specified EVENT-PROCEDURE internal
    *           procedure is found, null if EVENT-PROCEDURE clause is not specified.
    * @param    args
    *           The procedure's arguments.
    */
   public static void invokeInAsync(String    name,
                                    handle    h,
                                    handle    asyncHandle, 
                                    character eventProcName,
                                    handle    inEventprocHandle,
                                    Object... args)
   {
      invokeInAsync(new character(name), h, asyncHandle, eventProcName, inEventprocHandle, args);
   }

   /**
    * Invoke the procedure with the given name, defined in the given handle
    * (emitted for a standard RUN ... handle IN statement).
    * 
    * @param    name
    *           The legacy 4GL name for the procedure.
    * @param    h
    *           The handle to which this procedure belongs. When 
    *           <code>null</code>, defaults to THIS-PROCEDURE.
    * @param    asyncHandle
    *           The handle to the asynchronous request.
    * @param    eventProcName
    *           This represents the name of the internal procedure for the EVENT-PROCEDURE clause,
    *           null if this clause is not specified.
    * @param    inEventprocHandle
    *           The handle to a procedure in which context the specified EVENT-PROCEDURE internal
    *           procedure is found, null if EVENT-PROCEDURE clause is not specified.
    * @param    args
    *           The procedure's arguments.
    */
   public static void invokeInAsync(character name,
                                    handle    h,
                                    handle    asyncHandle, 
                                    character eventProcName,
                                    handle    inEventprocHandle,
                                    Object... args)
   {
      invokeInAsyncWithMode(name, h,
                            asyncHandle, eventProcName, inEventprocHandle,
                            null, args);
   }
   
   /**
    * Invoke the procedure with the given name, and set it as PERSISTENT
    * (emitted for a RUN ... PERSISTENT statement).
    * 
    * @param    name
    *           The legacy 4GL name for the procedure.
    * @param    modes
    *           A string representation of the modes of each parameter.
    * @param    args
    *           The procedure's arguments.
    */
   public static void invokePersistentWithMode(String    name,
                                               String    modes,
                                               Object... args)
   {
      invokePersistentWithMode(new character(name), modes, args);
   }

   /**
    * Invoke the procedure with the given name, and set it as PERSISTENT
    * (emitted for a RUN ... PERSISTENT statement).
    * 
    * @param    name
    *           The legacy 4GL name for the procedure.
    * @param    modes
    *           A string representation of the modes of each parameter.
    * @param    args
    *           The procedure's arguments.
    */
   public static void invokePersistentWithMode(character name,
                                               String    modes,
                                               Object... args)
   {
      invokePersistentImpl(null, name, null, null, false, modes, args);
   }

   /**
    * Invoke the procedure with the given name, and set it as PERSISTENT
    * (emitted for a RUN ... PERSISTENT statement).
    * 
    * @param    name
    *           The legacy 4GL name for the procedure.
    * @param    args
    *           The procedure's arguments.
    */
   public static void invokePersistent(String name, Object... args)
   {
      invokePersistent(new character(name), args);
   }
   
   /**
    * Invoke the procedure with the given name, and set it as PERSISTENT
    * (emitted for a RUN ... PERSISTENT ON ... statement).
    * 
    * @param    name
    *           The legacy 4GL name for the procedure.
    * @param    args
    *           The procedure's arguments.
    */
   public static void invokePersistent(character name, Object... args)
   {
      invokePersistentImpl(null, name, null, null, false, null, args);
   }

   /**
    * Invoke the procedure with the given name, and save its procedure handle
    * in the given handle (emitted for a RUN ... PERSISTENT set handle 
    * statement).
    * 
    * @param    name
    *           The legacy 4GL name for the procedure.
    * @param    h
    *           The handle to which the procedure reference must be saved.
    *           When <code>null</code>, just adds the procedure to the
    *           persistent procedure list.
    * @param    modes
    *           A string representation of the modes of each parameter.
    * @param    args
    *           The procedure's arguments.
    */
   public static void invokePersistentSetWithMode(String    name,
                                                  handle    h,
                                                  String    modes,
                                                  Object... args) 
   {
      invokePersistentSetWithMode(new character(name), h, modes, args);
   }

   /**
    * Invoke the procedure with the given name, and save its procedure handle
    * in the given handle (emitted for a RUN ... PERSISTENT set handle 
    * statement).
    * 
    * @param    name
    *           The legacy 4GL name for the procedure.
    * @param    h
    *           The handle to which the procedure reference must be saved.
    *           When <code>null</code>, just adds the procedure to the
    *           persistent procedure list.
    * @param    modes
    *           A string representation of the modes of each parameter.
    * @param    args
    *           The procedure's arguments.
    */
   public static void invokePersistentSetWithMode(character name,
                                                  handle    h,
                                                  String    modes,
                                                  Object... args)
   {
      invokePersistentImpl(null, name, h, null, false, modes, args);
   }

   /**
    * Invoke the procedure with the given name, and save its procedure handle
    * in the given handle (emitted for a RUN ... PERSISTENT set handle 
    * statement).
    * 
    * @param    name
    *           The legacy 4GL name for the procedure.
    * @param    h
    *           The handle to which the procedure reference must be saved.
    *           When <code>null</code>, just adds the procedure to the
    *           persistent procedure list.
    * @param    args
    *           The procedure's arguments.
    */
   public static void invokePersistentSet(String    name,
                                          handle    h,
                                          Object... args) 
   {
      invokePersistentSet(new character(name), h, args);
   }

   /**
    * Invoke the procedure with the given name, and save its procedure handle
    * in the given handle (emitted for a RUN ... PERSISTENT set handle 
    * statement).
    * 
    * @param    name
    *           The legacy 4GL name for the procedure.
    * @param    h
    *           The handle to which the procedure reference must be saved.
    *           When <code>null</code>, just adds the procedure to the
    *           persistent procedure list.
    * @param    args
    *           The procedure's arguments.
    */
   public static void invokePersistentSet(character name,
                                          handle    h,
                                          Object... args)
   {
      invokePersistentImpl(null, name, h, null, false, null, args);
   }
   
   /**
    * Invoke the procedure with the given name (emitted for a standard RUN 
    * statement).
    * 
    * @param    name
    *           The legacy 4GL name for the procedure.
    * @param    persistHandle
    *           The handle in which it will put the persistent external procedure.
    * @param    portHandle
    *           A handle variable where to save the web service. May be null.
    * @param    remoteHandle
    *           The handle to the remote object used to run the procedure
    * @param    transactionDistinct
    *           This marks if "TRANSACTION DISTINCT" option is specified with the RUN ... ON
    * @param    modes
    *           A string representation of the modes of each parameter.
    * @param    args
    *           The procedure's arguments.
    */
   public static void invokeRemotePersistentSetWithMode(String    name, 
                                                        handle    persistHandle,
                                                        handle    portHandle,
                                                        handle    remoteHandle,
                                                        boolean   transactionDistinct,
                                                        String    modes, 
                                                        Object... args) 
   {
      invokeRemotePersistentSetWithMode(new character(name), persistHandle, portHandle,
                                        remoteHandle, transactionDistinct,
                                        modes, args);
   }

   /**
    * Invoke the procedure with the given name (emitted for a standard RUN 
    * statement).
    * 
    * @param    name
    *           The legacy 4GL name for the procedure.
    * @param    persistHandle
    *           The handle in which it will put the persistent external procedure.
    * @param    portHandle
    *           A handle variable where to save the web service. May be null.
    * @param    remoteHandle
    *           The handle to the remote object used to run the procedure
    * @param    transactionDistinct
    *           This marks if "TRANSACTION DISTINCT" option is specified with the RUN ... ON stmt
    * @param    modes
    *           A string representation of the modes of each parameter.
    * @param    args
    *           The procedure's arguments.
    */
   public static void invokeRemotePersistentSetWithMode(character name, 
                                                        handle    persistHandle,
                                                        handle    portHandle,
                                                        handle    remoteHandle,
                                                        boolean   transactionDistinct,
                                                        String    modes, 
                                                        Object... args) 
   {
      invokePersistentImpl(remoteHandle, name,
                           persistHandle, portHandle,
                           transactionDistinct, modes, args);
   }

   /**
    * Invoke the procedure on a remote server with the given name, and set it as PERSISTENT
    * (emitted for a RUN ... PERSISTENT SET ON ... statement).
    * 
    * @param    name
    *           The legacy 4GL name for the procedure.
    * @param    persistHandle
    *           The handle in which it will put the persistent external procedure.
    * @param    portHandle
    *           A handle variable where to save the web service. May be null.
    * @param    remoteHandle
    *           The handle to the remote object used to run the procedure
    * @param    transactionDistinct
    *           This marks if "TRANSACTION DISTINCT" option is specified with the RUN ... ON stmt
    * @param    args
    *           The procedure's arguments.
    */
   public static void invokeRemotePersistentSet(String    name,
                                                handle    persistHandle,
                                                handle    portHandle,
                                                handle    remoteHandle,
                                                boolean   transactionDistinct,
                                                Object... args)
   {
      invokeRemotePersistentSet(new character(name), persistHandle, portHandle,
                                remoteHandle, transactionDistinct,
                                args);
   }

   /**
    * Invoke the procedure on a remote server with the given name, and set it as PERSISTENT
    * (emitted for a RUN ... PERSISTENT SET ON ... statement).
    * 
    * @param    name
    *           The legacy 4GL name for the procedure.
    * @param    persistHandle
    *           The handle in which it will put the persistent external procedure.
    * @param    portHandle
    *           A handle variable where to save the web service. May be null.
    * @param    remoteHandle
    *           The handle to the remote object used to run the procedure
    * @param    transactionDistinct
    *           This marks if "TRANSACTION DISTINCT" option is specified with the RUN ... ON stmt
    * @param    args
    *           The procedure's arguments.
    */
   public static void invokeRemotePersistentSet(character name,
                                                handle    persistHandle,
                                                handle    portHandle,
                                                handle    remoteHandle,
                                                boolean   transactionDistinct,
                                                Object... args)
   {
      invokeRemotePersistentSetWithMode(name, persistHandle, portHandle,
                                        remoteHandle, transactionDistinct,
                                        null, args);
   }

   /**
    * Invoke the procedure on a remote server with the given name, and set it as PERSISTENT
    * (emitted for a RUN ... PERSISTENT  ON ... statement).
    * 
    * @param    name
    *           The legacy 4GL name for the procedure.
    * @param    portHandle
    *           A handle variable where to save the web service. May be null.
    * @param    remoteHandle
    *           The handle to the remote object used to run the procedure
    * @param    transactionDistinct
    *           This marks if "TRANSACTION DISTINCT" option is specified with the RUN ... ON stmt
    * @param    args
    *           The procedure's arguments.
    */
   public static void invokeRemotePersistentWithMode(String    name,
                                                     handle    portHandle,
                                                     handle    remoteHandle,
                                                     boolean   transactionDistinct,
                                                     String    modes, 
                                                     Object... args)
   {
      invokeRemotePersistentWithMode(new character(name), portHandle,
                                     remoteHandle, transactionDistinct,
                                     modes, args);
   }

   /**
    * Invoke the procedure on a remote server with the given name, and set it as PERSISTENT
    * (emitted for a RUN ... PERSISTENT  ON ... statement).
    * 
    * @param    name
    *           The legacy 4GL name for the procedure.
    * @param    portHandle
    *           A handle variable where to save the web service. May be null.
    * @param    remoteHandle
    *           The handle to the remote object used to run the procedure
    * @param    transactionDistinct
    *           This marks if "TRANSACTION DISTINCT" option is specified with the RUN ... ON stmt
    * @param    args
    *           The procedure's arguments.
    */
   public static void invokeRemotePersistentWithMode(character name,
                                                     handle    portHandle,
                                                     handle    remoteHandle,
                                                     boolean   transactionDistinct,
                                                     String    modes, 
                                                     Object... args)
   {
      invokeRemotePersistentSetWithMode(name, null, portHandle,
                                        remoteHandle, transactionDistinct,
                                        modes, args);
   }

   /**
    * Invoke the procedure on a remote server with the given name, and set it as PERSISTENT
    * (emitted for a RUN ... PERSISTENT ON ... statement).
    * 
    * @param    name
    *           The legacy 4GL name for the procedure.
    * @param    portHandle
    *           A handle variable where to save the web service. May be null.
    * @param    remoteHandle
    *           The handle to the remote object used to run the procedure
    * @param    transactionDistinct
    *           This marks if "TRANSACTION DISTINCT" option is specified with the RUN ... ON stmt
    * @param    args
    *           The procedure's arguments.
    */
   public static void invokeRemotePersistent(String    name,
                                             handle    portHandle,
                                             handle    remoteHandle,
                                             boolean   transactionDistinct,
                                             Object... args)
   {
      invokeRemotePersistent(new character(name), portHandle,
                             remoteHandle, transactionDistinct,
                             args);
   }

   /**
    * Invoke the procedure on a remote server with the given name, and set it as PERSISTENT
    * (emitted for a RUN ... PERSISTENT ON ... statement).
    * 
    * @param    name
    *           The legacy 4GL name for the procedure.
    * @param    portHandle
    *           A handle variable where to save the web service. May be null.
    * @param    remoteHandle
    *           The handle to the remote object used to run the procedure
    * @param    transactionDistinct
    *           This marks if "TRANSACTION DISTINCT" option is specified with the RUN ... ON stmt
    * @param    args
    *           The procedure's arguments.
    */
   public static void invokeRemotePersistent(character name,
                                             handle    portHandle,
                                             handle    remoteHandle,
                                             boolean   transactionDistinct,
                                             Object... args)
   {
      invokeRemotePersistentWithMode(name, portHandle,
                                     remoteHandle, transactionDistinct,
                                     null, args);
   }
   
   /**
    * Invoke the procedure on a remote server with the given name 
    * (emitted for a standard RUN ... ON ... ASYNC SET ...
    * 
    * @param    name
    *           The legacy 4GL name for the procedure.
    * @param    portHandle
    *           A handle variable where to save the web service. May be null.
    * @param    remoteHandle
    *           The handle to the remote object used to run the procedure
    * @param    transactionDistinct
    *           This marks if "TRANSACTION DISTINCT" option is specified with the RUN ... ON stmt
    * @param    asyncHandle
    *           The handle to the asynchronous request.
    * @param    eventProcName
    *           This represents the name of the internal procedure for the EVENT-PROCEDURE clause,
    *           null if this clause is not specified.
    * @param    inEventprocHandle
    *           The handle to a procedure in which context the specified EVENT-PROCEDURE internal
    *           procedure is found, null if EVENT-PROCEDURE clause is not specified.
    * @param    modes
    *           A string representation of the modes of each parameter.
    * @param    args
    *           The procedure's arguments.
    */
   public static void invokeRemotePersistentAsyncWithMode(String    name, 
                                                          handle    portHandle,
                                                          handle    remoteHandle, 
                                                          boolean   transactionDistinct,
                                                          handle    asyncHandle,
                                                          String    eventProcName,
                                                          handle    inEventprocHandle,
                                                          String    modes, 
                                                          Object... args)
   {
      invokeRemotePersistentAsyncWithMode(new character(name), portHandle,
                                          remoteHandle, transactionDistinct, asyncHandle, 
                                          new character(eventProcName), inEventprocHandle,
                                          modes, args);
   }

   /**
    * Invoke the procedure on a remote server with the given name 
    * (emitted for a standard RUN ...PERSISTENT ON ... ASYNC SET ...
    * 
    * @param    name
    *           The legacy 4GL name for the procedure.
    * @param    portHandle
    *           A handle variable where to save the web service. May be null.
    * @param    remoteHandle
    *           The handle to the remote object used to run the procedure
    * @param    transactionDistinct
    *           This marks if "TRANSACTION DISTINCT" option is specified with the RUN ... ON stmt
    * @param    asyncHandle
    *           The handle to the asynchronous request.
    * @param    eventProcName
    *           This represents the name of the internal procedure for the EVENT-PROCEDURE clause,
    *           null if this clause is not specified.
    * @param    inEventprocHandle
    *           The handle to a procedure in which context the specified EVENT-PROCEDURE internal
    *           procedure is found, null if EVENT-PROCEDURE clause is not specified.
    * @param    modes
    *           A string representation of the modes of each parameter.
    * @param    args
    *           The procedure's arguments.
    */
   public static void invokeRemotePersistentAsyncWithMode(character name, 
                                                          handle    portHandle,
                                                          handle    remoteHandle, 
                                                          boolean   transactionDistinct,
                                                          handle    asyncHandle,
                                                          String    eventProcName,
                                                          handle    inEventprocHandle,
                                                          String    modes, 
                                                          Object... args)
   {
      invokeRemotePersistentAsyncWithMode(name, portHandle, 
                                          remoteHandle, transactionDistinct,
                                          asyncHandle, new character(eventProcName),
                                          inEventprocHandle,
                                          modes, args);
   }

   /**
    * Invoke the procedure on a remote server with the given name 
    * (emitted for a standard RUN ...PERSISTENT ON ... ASYNC SET ...
    * 
    * @param    name
    *           The legacy 4GL name for the procedure.
    * @param    portHandle
    *           A handle variable where to save the web service. May be null.
    * @param    remoteHandle
    *           The handle to the remote object used to run the procedure
    * @param    transactionDistinct
    *           This marks if "TRANSACTION DISTINCT" option is specified with the RUN ... ON stmt
    * @param    asyncHandle
    *           The handle to the asynchronous request.
    * @param    eventProcName
    *           This represents the name of the internal procedure for the EVENT-PROCEDURE clause,
    *           null if this clause is not specified.
    * @param    inEventprocHandle
    *           The handle to a procedure in which context the specified EVENT-PROCEDURE internal
    *           procedure is found, null if EVENT-PROCEDURE clause is not specified.
    * @param    modes
    *           A string representation of the modes of each parameter.
    * @param    args
    *           The procedure's arguments.
    */
   public static void invokeRemotePersistentAsyncWithMode(String    name, 
                                                          handle    portHandle,
                                                          handle    remoteHandle, 
                                                          boolean   transactionDistinct,
                                                          handle    asyncHandle,
                                                          character eventProcName,
                                                          handle    inEventprocHandle,
                                                          String    modes, 
                                                          Object... args)
   { 
      invokeRemotePersistentAsyncWithMode(new character(name), portHandle,
                                          remoteHandle, transactionDistinct,
                                          asyncHandle, eventProcName, inEventprocHandle,
                                          modes, args);
   }

   /**
    * Invoke the procedure on a remote server with the given name 
    * (emitted for a standard RUN ...PERSISTENT ON ... ASYNC SET ...
    * 
    * @param    name
    *           The legacy 4GL name for the procedure.
    * @param    portHandle
    *           A handle variable where to save the web service. May be null.
    * @param    remoteHandle
    *           The handle to the remote object used to run the procedure
    * @param    transactionDistinct
    *           This marks if "TRANSACTION DISTINCT" option is specified with the RUN ... ON stmt
    * @param    asyncHandle
    *           The handle to the asynchronous request.
    * @param    eventProcName
    *           This represents the name of the internal procedure for the EVENT-PROCEDURE clause,
    *           null if this clause is not specified.
    * @param    inEventprocHandle
    *           The handle to a procedure in which context the specified EVENT-PROCEDURE internal
    *           procedure is found, null if EVENT-PROCEDURE clause is not specified.
    * @param    modes
    *           A string representation of the modes of each parameter.
    * @param    args
    *           The procedure's arguments.
    */
   public static void invokeRemotePersistentAsyncWithMode(character name, 
                                                          handle    portHandle,
                                                          handle    remoteHandle, 
                                                          boolean   transactionDistinct,
                                                          handle    asyncHandle,
                                                          character eventProcName,
                                                          handle    inEventprocHandle,
                                                          String    modes, 
                                                          Object... args)
   {
      invokeRemotePersistentSetAsyncWithMode(name, null, portHandle,
                                             remoteHandle, transactionDistinct,
                                             asyncHandle, eventProcName, inEventprocHandle,
                                             modes, args);
   }

   /**
    * Invoke the procedure on a remote server with the given name 
    * (emitted for a standard RUN ...PERSISTENT ON ... ASYNC SET ...
    * 
    * @param    name
    *           The legacy 4GL name for the procedure.
    * @param    portHandle
    *           A handle variable where to save the web service. May be null.
    * @param    remoteHandle
    *           The handle to the remote object used to run the procedure
    * @param    transactionDistinct
    *           This marks if "TRANSACTION DISTINCT" option is specified with the RUN ... ON stmt
    * @param    asyncHandle
    *           The handle to the asynchronous request.
    * @param    eventProcName
    *           This represents the name of the internal procedure for the EVENT-PROCEDURE clause,
    *           null if this clause is not specified.
    * @param    inEventprocHandle
    *           The handle to a procedure in which context the specified EVENT-PROCEDURE internal
    *           procedure is found, null if EVENT-PROCEDURE clause is not specified.
    * @param    args
    *           The procedure's arguments.
    */
   public static void invokeRemotePersistentAsync(String    name, 
                                                  handle    portHandle,
                                                  handle    remoteHandle, 
                                                  boolean   transactionDistinct,
                                                  handle    asyncHandle,
                                                  String    eventProcName,
                                                  handle    inEventprocHandle,
                                                  Object... args)
   {
      invokeRemotePersistentAsync(new character(name), portHandle,
                                  remoteHandle, transactionDistinct,
                                  asyncHandle, new character(eventProcName), inEventprocHandle,
                                  args);
   }

   /**
    * Invoke the procedure on a remote server with the given name 
    * (emitted for a standard RUN ...PERSISTENT ON ... ASYNC SET ...
    * 
    * @param    name
    *           The legacy 4GL name for the procedure.
    * @param    portHandle
    *           A handle variable where to save the web service. May be null.
    * @param    remoteHandle
    *           The handle to the remote object used to run the procedure
    * @param    transactionDistinct
    *           This marks if "TRANSACTION DISTINCT" option is specified with the RUN ... ON stmt
    * @param    asyncHandle
    *           The handle to the asynchronous request.
    * @param    eventProcName
    *           This represents the name of the internal procedure for the EVENT-PROCEDURE clause,
    *           null if this clause is not specified.
    * @param    inEventprocHandle
    *           The handle to a procedure in which context the specified EVENT-PROCEDURE internal
    *           procedure is found, null if EVENT-PROCEDURE clause is not specified.
    * @param    args
    *           The procedure's arguments.
    */
   public static void invokeRemotePersistentAsync(character name, 
                                                  handle    portHandle,
                                                  handle    remoteHandle, 
                                                  boolean   transactionDistinct,
                                                  handle    asyncHandle,
                                                  String    eventProcName,
                                                  handle    inEventprocHandle,
                                                  Object... args)
   {
      invokeRemotePersistentAsync(name, portHandle,
                                  remoteHandle, transactionDistinct,
                                  asyncHandle, new character(eventProcName), inEventprocHandle,
                                  args);
   }
   
   /**
    * Invoke the procedure on a remote server with the given name 
    * (emitted for a standard RUN ... ON ... ASYNC SET ...
    * 
    * @param    name
    *           The legacy 4GL name for the procedure.
    * @param    portHandle
    *           A handle variable where to save the web service. May be null.
    * @param    remoteHandle
    *           The handle to the remote object used to run the procedure
    * @param    transactionDistinct
    *           This marks if "TRANSACTION DISTINCT" option is specified with the RUN ... ON stmt
    * @param    asyncHandle
    *           The handle to the asynchronous request.
    * @param    eventProcName
    *           This represents the name of the internal procedure for the EVENT-PROCEDURE clause,
    *           null if this clause is not specified.
    * @param    inEventprocHandle
    *           The handle to a procedure in which context the specified EVENT-PROCEDURE internal
    *           procedure is found, null if EVENT-PROCEDURE clause is not specified.
    * @param    args
    *           The procedure's arguments.
    */
   public static void invokeRemotePersistentAsync(String    name, 
                                                  handle    portHandle,
                                                  handle    remoteHandle, 
                                                  boolean   transactionDistinct,
                                                  handle    asyncHandle,
                                                  character eventProcName,
                                                  handle    inEventprocHandle,
                                                  Object... args)
   {
      invokeRemotePersistentAsync(new character(name), portHandle,
                                  remoteHandle, transactionDistinct,
                                  asyncHandle, eventProcName, inEventprocHandle, args);
   }

   /**
    * Invoke the procedure on a remote server with the given name 
    * (emitted for a standard RUN ...PERSISTENT ON ... ASYNC SET ...
    * 
    * @param    name
    *           The legacy 4GL name for the procedure.
    * @param    portHandle
    *           A handle variable where to save the web service. May be null.
    * @param    remoteHandle
    *           The handle to the remote object used to run the procedure
    * @param    transactionDistinct
    *           This marks if "TRANSACTION DISTINCT" option is specified with the RUN ... ON stmt
    * @param    asyncHandle
    *           The handle to the asynchronous request.
    * @param    eventProcName
    *           This represents the name of the internal procedure for the EVENT-PROCEDURE clause,
    *           null if this clause is not specified.
    * @param    inEventprocHandle
    *           The handle to a procedure in which context the specified EVENT-PROCEDURE internal
    *           procedure is found, null if EVENT-PROCEDURE clause is not specified.
    * @param    args
    *           The procedure's arguments.
    */
   public static void invokeRemotePersistentAsync(character name, 
                                                  handle    portHandle,
                                                  handle    remoteHandle, 
                                                  boolean   transactionDistinct,
                                                  handle    asyncHandle,
                                                  character eventProcName,
                                                  handle    inEventprocHandle,
                                                  Object... args)
   {
      invokeRemotePersistentSetAsync(name, null, portHandle,
                                     remoteHandle, transactionDistinct,
                                     asyncHandle, eventProcName, inEventprocHandle,
                                     args);
   }
   
   /**
    * Invoke the procedure on a remote server with the given name 
    * (emitted for a standard RUN ...PERSISTENT SET.. ON ... ASYNC SET ...
    * 
    * @param    name
    *           The legacy 4GL name for the procedure.
    * @param    persistHandle
    *           The handle in which it will put the persistent external procedure.
    * @param    portHandle
    *           A handle variable where to save the web service. May be null.
    * @param    remoteHandle
    *           The handle to the remote object used to run the procedure
    * @param    transactionDistinct
    *           This marks if "TRANSACTION DISTINCT" option is specified with the RUN ... ON stmt
    * @param    asyncHandle
    *           The handle to the asynchronous request.
    * @param    eventProcName
    *           This represents the name of the internal procedure for the EVENT-PROCEDURE clause,
    *           null if this clause is not specified.
    * @param    inEventprocHandle
    *           The handle to a procedure in which context the specified EVENT-PROCEDURE internal
    *           procedure is found, null if EVENT-PROCEDURE clause is not specified.
    * @param    modes
    *           A string representation of the modes of each parameter.
    * @param    args
    *           The procedure's arguments.
    */
   public static void invokeRemotePersistentSetAsyncWithMode(String    name,
                                                             handle    persistHandle,
                                                             handle    portHandle,
                                                             handle    remoteHandle, 
                                                             boolean   transactionDistinct,
                                                             handle    asyncHandle,
                                                             String    eventProcName,
                                                             handle    inEventprocHandle,
                                                             String    modes, 
                                                             Object... args)
   {
      invokeRemotePersistentSetAsyncWithMode(new character(name), persistHandle, portHandle,
                                             remoteHandle, transactionDistinct,
                                             asyncHandle, new character(eventProcName), 
                                             inEventprocHandle,
                                             modes, args);
   }

   /**
    * Invoke the procedure on a remote server with the given name 
    * (emitted for a standard RUN ...PERSISTENT SET.. ON ... ASYNC SET ...
    * 
    * @param    name
    *           The legacy 4GL name for the procedure.
    * @param    persistHandle
    *           The handle in which it will put the persistent external procedure.
    * @param    portHandle
    *           A handle variable where to save the web service. May be null.
    * @param    remoteHandle
    *           The handle to the remote object used to run the procedure
    * @param    transactionDistinct
    *           This marks if "TRANSACTION DISTINCT" option is specified with the RUN ... ON stmt
    * @param    asyncHandle
    *           The handle to the asynchronous request.
    * @param    eventProcName
    *           This represents the name of the internal procedure for the EVENT-PROCEDURE clause,
    *           null if this clause is not specified.
    * @param    inEventprocHandle
    *           The handle to a procedure in which context the specified EVENT-PROCEDURE internal
    *           procedure is found, null if EVENT-PROCEDURE clause is not specified.
    * @param    modes
    *           A string representation of the modes of each parameter.
    * @param    args
    *           The procedure's arguments.
    */
   public static void invokeRemotePersistentSetAsyncWithMode(character name,
                                                             handle    persistHandle,
                                                             handle    portHandle,
                                                             handle    remoteHandle, 
                                                             boolean   transactionDistinct,
                                                             handle    asyncHandle,
                                                             String    eventProcName,
                                                             handle    inEventprocHandle,
                                                             String    modes, 
                                                             Object... args)
   {
      invokeRemotePersistentSetAsyncWithMode(name, persistHandle, portHandle, 
                                             remoteHandle, transactionDistinct,
                                             asyncHandle, new character(eventProcName), 
                                             inEventprocHandle,
                                             modes, args);
   }

   /**
    * Invoke the procedure on a remote server with the given name 
    * (emitted for a standard RUN ...PERSISTENT SET.. ON ... ASYNC SET ...
    * 
    * @param    name
    *           The legacy 4GL name for the procedure.
    * @param    persistHandle
    *           The handle in which it will put the persistent external procedure.
    * @param    portHandle
    *           A handle variable where to save the web service. May be null.
    * @param    remoteHandle
    *           The handle to the remote object used to run the procedure
    * @param    transactionDistinct
    *           This marks if "TRANSACTION DISTINCT" option is specified with the RUN ... ON stmt
    * @param    asyncHandle
    *           The handle to the asynchronous request.
    * @param    eventProcName
    *           This represents the name of the internal procedure for the EVENT-PROCEDURE clause,
    *           null if this clause is not specified.
    * @param    inEventprocHandle
    *           The handle to a procedure in which context the specified EVENT-PROCEDURE internal
    *           procedure is found, null if EVENT-PROCEDURE clause is not specified.
    * @param    modes
    *           A string representation of the modes of each parameter.
    * @param    args
    *           The procedure's arguments.
    */
   public static void invokeRemotePersistentSetAsyncWithMode(String    name,
                                                             handle    persistHandle,
                                                             handle    portHandle,
                                                             handle    remoteHandle, 
                                                             boolean   transactionDistinct,
                                                             handle    asyncHandle,
                                                             character eventProcName,
                                                             handle    inEventprocHandle,
                                                             String    modes, 
                                                             Object... args)
   {
      invokeRemotePersistentSetAsyncWithMode(new character(name), persistHandle, portHandle,
                                             remoteHandle, transactionDistinct,
                                             asyncHandle, eventProcName, inEventprocHandle,
                                             modes, args);
   }

   /**
    * Invoke the procedure on a remote server with the given name 
    * (emitted for a standard RUN ...PERSISTENT SET.. ON ... ASYNC SET ...
    * 
    * @param    name
    *           The legacy 4GL name for the procedure.
    * @param    persistHandle
    *           The handle in which it will put the persistent external procedure.
    * @param    portHandle
    *           A handle variable where to save the web service. May be null.
    * @param    remoteHandle
    *           The handle to the remote object used to run the procedure
    * @param    transactionDistinct
    *           This marks if "TRANSACTION DISTINCT" option is specified with the RUN ... ON stmt
    * @param    asyncHandle
    *           The handle to the asynchronous request.
    * @param    eventProcName
    *           This represents the name of the internal procedure for the EVENT-PROCEDURE clause,
    *           null if this clause is not specified.
    * @param    inEventprocHandle
    *           The handle to a procedure in which context the specified EVENT-PROCEDURE internal
    *           procedure is found, null if EVENT-PROCEDURE clause is not specified.
    * @param    modes
    *           A string representation of the modes of each parameter.
    * @param    args
    *           The procedure's arguments.
    */
   public static void invokeRemotePersistentSetAsyncWithMode(character name, 
                                                             handle    persistHandle,
                                                             handle    portHandle,
                                                             handle    remoteHandle, 
                                                             boolean   transactionDistinct,
                                                             handle    asyncHandle,
                                                             character eventProcName,
                                                             handle    inEventprocHandle,
                                                             String    modes, 
                                                             Object... args)
   {
      boolean isServer = (remoteHandle.getResource() instanceof ServerImpl);
      if (!isServer)
      {
         // TODO: log this procedure call using a handle to non-server resource 
         return;
      }
      
      ServerImpl server = (ServerImpl) remoteHandle.getResource();
      if (server.isWebService())
      {
         final String err =
            "Unexpected Web service client error. Copy the following message and contact " +
            "technical support. Web service procedure not initialized in cwsSend (11461)";
         ErrorManager.recordOrThrowError(11461, err, false, true);
         return;
      }
      
      // do this only if the server is still connected; if not, it will fallback to non-async 
      // processing to produce the appropriate error
      final boolean obtainProxy = server._connected() && server.getServerHelper().isSessionFree();
      
      AsyncCall remoteRequest = new AsyncCall(remoteHandle, portHandle, name, new handle(),
                                              asyncHandle, eventProcName, inEventprocHandle,
                                              transactionDistinct,
                                              modes, args)
      {
         @Override
         public void run()
         {
            if (obtainProxy)
            {
               // if truly-async request, must run the "execute" method on the resolved proxy
               // if the proxy is invalid or can't map attributes or other error, then error 
               // messages are generated now
               initializeProxy(this);
            }
            else
            {
               // if not a server, go the normal way
               invokePersistentImpl(asyncReq, hServer, pName, pHandle, portHandle,
                                    transactionDistinct,
                                    asyncReq.getModes(), asyncReq.getArgs());
            }
         }
      };

      if (obtainProxy)
      {
         AppServerHelper helper = (AppServerHelper) server.getServerHelper();
         ProxyProcedureWrapper proxy = helper.obtainProxy(name);
         persistHandle.assign(proxy);
         remoteRequest.pHandle.assign(proxy);
      }
      
      invokeAsyncImpl(remoteRequest, remoteHandle, asyncHandle, new handle());
   }

   /**
    * Invoke the procedure on a remote server with the given name 
    * (emitted for a standard RUN ...PERSISTENT SET ON ... ASYNC SET ...
    * 
    * @param    name
    *           The legacy 4GL name for the procedure.
    * @param    persistHandle
    *           The handle in which it will put the persistent external procedure.
    * @param    portHandle
    *           A handle variable where to save the web service. May be null.
    * @param    remoteHandle
    *           The handle to the remote object used to run the procedure
    * @param    transactionDistinct
    *           This marks if "TRANSACTION DISTINCT" option is specified with the RUN ... ON stmt
    * @param    asyncHandle
    *           The handle to the asynchronous request.
    * @param    eventProcName
    *           This represents the name of the internal procedure for the EVENT-PROCEDURE clause,
    *           null if this clause is not specified.
    * @param    inEventprocHandle
    *           The handle to a procedure in which context the specified EVENT-PROCEDURE internal
    *           procedure is found, null if EVENT-PROCEDURE clause is not specified.
    * @param    args
    *           The procedure's arguments.
    */
   public static void invokeRemotePersistentSetAsync(String    name,
                                                     handle    persistHandle,
                                                     handle    portHandle,
                                                     handle    remoteHandle, 
                                                     boolean   transactionDistinct,
                                                     handle    asyncHandle,
                                                     String    eventProcName,
                                                     handle    inEventprocHandle,
                                                     Object... args)
   {
      invokeRemotePersistentSetAsync(new character(name), persistHandle, portHandle, 
                                     remoteHandle, transactionDistinct,
                                     asyncHandle, new character(eventProcName), inEventprocHandle,
                                     args);
   }

   /**
    * Invoke the procedure on a remote server with the given name 
    * (emitted for a standard RUN ...PERSISTENT SET ON ... ASYNC SET ...
    * 
    * @param    name
    *           The legacy 4GL name for the procedure.
    * @param    persistHandle
    *           The handle in which it will put the persistent external procedure.
    * @param    portHandle
    *           A handle variable where to save the web service. May be null.
    * @param    remoteHandle
    *           The handle to the remote object used to run the procedure
    * @param    transactionDistinct
    *           This marks if "TRANSACTION DISTINCT" option is specified with the RUN ... ON stmt
    * @param    asyncHandle
    *           The handle to the asynchronous request.
    * @param    eventProcName
    *           This represents the name of the internal procedure for the EVENT-PROCEDURE clause,
    *           null if this clause is not specified.
    * @param    inEventprocHandle
    *           The handle to a procedure in which context the specified EVENT-PROCEDURE internal
    *           procedure is found, null if EVENT-PROCEDURE clause is not specified.
    * @param    args
    *           The procedure's arguments.
    */
   public static void invokeRemotePersistentSetAsync(character name,
                                                     handle    persistHandle,
                                                     handle    portHandle,
                                                     handle    remoteHandle, 
                                                     boolean   transactionDistinct,
                                                     handle    asyncHandle,
                                                     String    eventProcName,
                                                     handle    inEventprocHandle,
                                                     Object... args)
   {
      invokeRemotePersistentSetAsync(name, persistHandle, portHandle,
                                     remoteHandle, transactionDistinct, asyncHandle, 
                                     new character(eventProcName), inEventprocHandle,
                                     args);
   }
   
   /**
    * Invoke the procedure on a remote server with the given name 
    * (emitted for a standard RUN ... ON ... ASYNC SET ...
    * 
    * @param    name
    *           The legacy 4GL name for the procedure.
    * @param    persistHandle
    *           The handle in which it will put the persistent external procedure.
    * @param    portHandle
    *           A handle variable where to save the web service. May be null.
    * @param    remoteHandle
    *           The handle to the remote object used to run the procedure
    * @param    transactionDistinct
    *           This marks if "TRANSACTION DISTINCT" option is specified with the RUN ... ON stmt
    * @param    asyncHandle
    *           The handle to the asynchronous request.
    * @param    eventProcName
    *           This represents the name of the internal procedure for the EVENT-PROCEDURE clause,
    *           null if this clause is not specified.
    * @param    inEventprocHandle
    *           The handle to a procedure in which context the specified EVENT-PROCEDURE internal
    *           procedure is found, null if EVENT-PROCEDURE clause is not specified.
    * @param    args
    *           The procedure's arguments.
    */
   public static void invokeRemotePersistentSetAsync(String    name, 
                                                     handle    persistHandle,
                                                     handle    portHandle,
                                                     handle    remoteHandle, 
                                                     boolean   transactionDistinct,
                                                     handle    asyncHandle,
                                                     character eventProcName,
                                                     handle    inEventprocHandle,
                                                     Object... args)
   {
      invokeRemotePersistentSetAsync(new character(name), persistHandle, portHandle,
                                     remoteHandle, transactionDistinct,
                                     asyncHandle, eventProcName, inEventprocHandle,
                                     args);
   }

   /**
    * Invoke the procedure on a remote server with the given name 
    * (emitted for a standard RUN ...PERSISTENT SET ON ... ASYNC SET ...
    * 
    * @param    name
    *           The legacy 4GL name for the procedure.
    * @param    persistHandle
    *           The handle in which it will put the persistent external procedure.
    * @param    portHandle
    *           A handle variable where to save the web service. May be null.
    * @param    remoteHandle
    *           The handle to the remote object used to run the procedure
    * @param    transactionDistinct
    *           This marks if "TRANSACTION DISTINCT" option is specified with the RUN ... ON stmt
    * @param    asyncHandle
    *           The handle to the asynchronous request.
    * @param    eventProcName
    *           This represents the name of the internal procedure for the EVENT-PROCEDURE clause,
    *           null if this clause is not specified.
    * @param    inEventprocHandle
    *           The handle to a procedure in which context the specified EVENT-PROCEDURE internal
    *           procedure is found, null if EVENT-PROCEDURE clause is not specified.
    * @param    args
    *           The procedure's arguments.
    */
   public static void invokeRemotePersistentSetAsync(character name, 
                                                     handle    persistHandle,
                                                     handle    portHandle,
                                                     handle    remoteHandle, 
                                                     boolean   transactionDistinct,
                                                     handle    asyncHandle,
                                                     character eventProcName,
                                                     handle    inEventprocHandle,
                                                     Object... args)
   {
      invokeRemotePersistentSetAsyncWithMode(name, persistHandle, portHandle,
                                             remoteHandle, transactionDistinct,
                                             asyncHandle, eventProcName, inEventprocHandle,
                                             null, args);
   }
   
   /**
    * Emitted for the DYNAMIC-FUNCTION statement.
    *
    * @param    cls
    *           Class.
    * @param    name
    *           The legacy 4GL name for the function.
    * @param    modes
    *           A string representation of the modes of each parameter.
    * @param    args
    *           The functions's arguments.
    *           
    * @return   The return value from the block (or its nested blocks) which
    *           is the <code>unknown value</code> by default or on error.
    */
   public static <T extends BaseDataType> T
   invokeDynamicFunctionWithMode(Class<T>  cls,
                                 String    name,
                                 String    modes,
                                 Object... args)
   {
      BaseDataType res = invokeFunctionImpl(new character(name), null, true, false, modes, args);
      return convertValue(cls, res);
   }

   /**
    * Emitted for the DYNAMIC-FUNCTION statement.
    *
    * @param    cls
    *           Class.
    * @param    name
    *           The legacy 4GL name for the function.
    * @param    modes
    *           A string representation of the modes of each parameter.
    * @param    args
    *           The functions's arguments.
    *           
    * @return   The return value from the block (or its nested blocks) which
    *           is the <code>unknown value</code> by default or on error.
    */
   public static <T extends BaseDataType> T
   invokeDynamicFunctionWithMode(Class<T>  cls,
                                 character name,
                                 String    modes,
                                 Object... args)
   {
      BaseDataType res = invokeFunctionImpl(name, null, true, false, modes, args);
      return convertValue(cls, res);
   }

   /**
    * Emitted for the DYNAMIC-FUNCTION(fname IN handle, ... ) statement.
    *
    * @param    cls
    *           Class.
    * @param    name
    *           The legacy 4GL name for the function.
    * @param    h
    *           The handle to which the function belongs. If <code>null</code>
    *           defaults to THIS-PROCEDURE.
    * @param    args
    *           The functions's arguments.
    *           
    * @return   The return value from the block (or its nested blocks) which
    *           is the <code>unknown value</code> by default or on error.
    */
   public static <T extends BaseDataType> T
   invokeDynamicFunctionInWithMode(Class<T>  cls,
                                   String    name,
                                   handle    h,
                                   String    modes,
                                   Object... args)
   {
      BaseDataType res = invokeFunctionImpl(new character(name), h, true, false, modes, args);
      return convertValue(cls, res);
   }
   
   /**
    * Emitted for the DYNAMIC-FUNCTION(fname IN handle, ... ) statement.
    *
    * @param    cls
    *           Class.
    * @param    name
    *           The legacy 4GL name for the function.
    * @param    h
    *           The handle to which the function belongs. If <code>null</code>
    *           defaults to THIS-PROCEDURE.
    * @param    args
    *           The functions's arguments.
    *           
    * @return   The return value from the block (or its nested blocks) which
    *           is the <code>unknown value</code> by default or on error.
    */
   public static <T extends BaseDataType> T
   invokeDynamicFunctionInWithMode(Class<T>  cls,
                                   character name,
                                   handle    h,
                                   String    modes,
                                   Object... args)
   {
      BaseDataType res = invokeFunctionImpl(name, h, true, false, modes, args);
      return convertValue(cls, res);
   }

   /**
    * Emitted for the DYNAMIC-FUNCTION statement.
    *
    * @param    cls
    *           Class.
    * @param    name
    *           The legacy 4GL name for the function.
    * @param    args
    *           The functions's arguments.
    *           
    * @return   The return value from the block (or its nested blocks) which
    *           is the <code>unknown value</code> by default or on error.
    */
   public static <T extends BaseDataType> T
   invokeDynamicFunction(Class<T>  cls,
                         String    name,
                         Object... args)
   {
      BaseDataType res = invokeFunctionImpl(new character(name), null, true, false, null, args);
      return convertValue(cls, res);
   }

   /**
    * Emitted for the DYNAMIC-FUNCTION statement.
    *
    * @param    cls
    *           Class.
    * @param    name
    *           The legacy 4GL name for the function.
    * @param    args
    *           The functions's arguments.
    *           
    * @return   The return value from the block (or its nested blocks) which
    *           is the <code>unknown value</code> by default or on error.
    */
   public static <T extends BaseDataType> T
   invokeDynamicFunction(Class<T>  cls,
                         character name,
                         Object... args)
   {
      BaseDataType res = invokeFunctionImpl(name, null, true, false, null, args);
      return convertValue(cls, res);
   }

   /**
    * Emitted for the DYNAMIC-FUNCTION(fname IN handle, ... ) statement.
    *
    * @param    cls
    *           Class.
    * @param    name
    *           The legacy 4GL name for the function.
    * @param    h
    *           The handle to which the function belongs. If <code>null</code>
    *           defaults to THIS-PROCEDURE.
    * @param    args
    *           The functions's arguments.
    *           
    * @return   The return value from the block (or its nested blocks) which
    *           is the <code>unknown value</code> by default or on error.
    */
   public static <T extends BaseDataType> T
   invokeDynamicFunctionIn(Class<T>  cls,
                           String    name,
                           handle    h,
                           Object... args)
   {
      BaseDataType res = invokeFunctionImpl(new character(name), h, true, false, null, args);
      return convertValue(cls, res);
   }
   
   /**
    * Emitted for the DYNAMIC-FUNCTION(fname IN handle, ... ) statement.
    *
    * @param    cls
    *           Class.
    * @param    name
    *           The legacy 4GL name for the function.
    * @param    h
    *           The handle to which the function belongs. If <code>null</code>
    *           defaults to THIS-PROCEDURE.
    * @param    args
    *           The functions's arguments.
    *           
    * @return   The return value from the block (or its nested blocks) which
    *           is the <code>unknown value</code> by default or on error.
    */
   public static <T extends BaseDataType> T
   invokeDynamicFunctionIn(Class<T>  cls,
                           character name,
                           handle    h,
                           Object... args)
   {
      BaseDataType res = invokeFunctionImpl(name, h, true, false, null, args);
      return convertValue(cls, res);
   }

   /**
    * Emitted for the DYNAMIC-FUNCTION statement.
    * 
    * @param    name
    *           The legacy 4GL name for the function.
    * @param    modes
    *           A string representation of the modes of each parameter.
    * @param    args
    *           The functions's arguments.
    *           
    * @return   The return value from the block (or its nested blocks) which
    *           is the <code>unknown value</code> by default or on error.
    */
   public static BaseDataType invokeDynamicFunctionWithMode(String    name,
                                                            String    modes,
                                                            Object... args)
   {
      return invokeFunctionImpl(new character(name), null, true, false, modes, args);
   }

   /**
    * Emitted for the DYNAMIC-FUNCTION statement.
    * 
    * @param    name
    *           The legacy 4GL name for the function.
    * @param    modes
    *           A string representation of the modes of each parameter.
    * @param    args
    *           The functions's arguments.
    *           
    * @return   The return value from the block (or its nested blocks) which
    *           is the <code>unknown value</code> by default or on error.
    */
   public static BaseDataType invokeDynamicFunctionWithMode(character name,
                                                            String    modes,
                                                            Object... args)
   {
      return invokeFunctionImpl(name, null, true, false, modes, args);
   }

   /**
    * Emitted for the DYNAMIC-FUNCTION(fname IN handle, ... ) statement.
    * 
    * @param    name
    *           The legacy 4GL name for the function.
    * @param    h
    *           The handle to which the function belongs. If <code>null</code>
    *           defaults to THIS-PROCEDURE.
    * @param    args
    *           The functions's arguments.
    *           
    * @return   The return value from the block (or its nested blocks) which
    *           is the <code>unknown value</code> by default or on error.
    */
   public static BaseDataType invokeDynamicFunctionInWithMode(String    name,
                                                              handle    h,
                                                              String    modes,
                                                              Object... args)
   {
      return invokeFunctionImpl(new character(name), h, true, false, modes, args);
   }
   
   /**
    * Emitted for the DYNAMIC-FUNCTION(fname IN handle, ... ) statement.
    * 
    * @param    name
    *           The legacy 4GL name for the function.
    * @param    h
    *           The handle to which the function belongs. If <code>null</code>
    *           defaults to THIS-PROCEDURE.
    * @param    args
    *           The functions's arguments.
    *           
    * @return   The return value from the block (or its nested blocks) which
    *           is the <code>unknown value</code> by default or on error.
    */
   public static BaseDataType invokeDynamicFunctionInWithMode(character name,
                                                              handle    h,
                                                              String    modes,
                                                              Object... args)
   {
      return invokeFunctionImpl(name, h, true, false, modes, args);
   }

   /**
    * Emitted for the DYNAMIC-FUNCTION statement.
    * 
    * @param    name
    *           The legacy 4GL name for the function.
    * @param    args
    *           The functions's arguments.
    *           
    * @return   The return value from the block (or its nested blocks) which
    *           is the <code>unknown value</code> by default or on error.
    */
   public static BaseDataType invokeDynamicFunction(String    name,
                                                    Object... args)
   {
      return invokeFunctionImpl(new character(name), null, true, false, null, args);
   }

   /**
    * Emitted for the DYNAMIC-FUNCTION statement.
    * 
    * @param    name
    *           The legacy 4GL name for the function.
    * @param    args
    *           The functions's arguments.
    *           
    * @return   The return value from the block (or its nested blocks) which
    *           is the <code>unknown value</code> by default or on error.
    */
   public static BaseDataType invokeDynamicFunction(character name,
                                                    Object... args)
   {
      return invokeFunctionImpl(name, null, true, false, null, args);
   }

   /**
    * Emitted for the DYNAMIC-FUNCTION(fname IN handle, ... ) statement.
    * 
    * @param    name
    *           The legacy 4GL name for the function.
    * @param    h
    *           The handle to which the function belongs. If <code>null</code>
    *           defaults to THIS-PROCEDURE.
    * @param    args
    *           The functions's arguments.
    *           
    * @return   The return value from the block (or its nested blocks) which
    *           is the <code>unknown value</code> by default or on error.
    */
   public static BaseDataType invokeDynamicFunctionIn(String    name,
                                                      handle    h,
                                                      Object... args)
   {
      return invokeFunctionImpl(new character(name), h, true, false, null, args);
   }
   
   /**
    * Emitted for the DYNAMIC-FUNCTION(fname IN handle, ... ) statement.
    * 
    * @param    name
    *           The legacy 4GL name for the function.
    * @param    h
    *           The handle to which the function belongs. If <code>null</code>
    *           defaults to THIS-PROCEDURE.
    * @param    args
    *           The functions's arguments.
    *           
    * @return   The return value from the block (or its nested blocks) which
    *           is the <code>unknown value</code> by default or on error.
    */
   public static BaseDataType invokeDynamicFunctionIn(character name,
                                                      handle    h,
                                                      Object... args)
   {
      return invokeFunctionImpl(name, h, true, false, null, args);
   }

   /**
    * Emitted for plain function calls.
    *
    * @param    cls
    *           Class.
    * @param    name
    *           The legacy 4GL name for the function.
    * @param    modes
    *           A string representation of the modes of each parameter.
    * @param    args
    *           The functions's arguments.
    *           
    * @return   The return value from the block (or its nested blocks) which
    *           is the <code>unknown value</code> by default or on error.
    */
   public static <T extends BaseDataType> T
   invokeFunctionWithMode(Class<T>  cls, String name, String modes, Object... args)
   {
      BaseDataType res = invokeFunctionImpl(new character(name), null, false, false, modes, args);
      return convertValue(cls, res);
   }

   /**
    * Emitted for plain function calls.
    *
    * @param    cls
    *           Class.
    * @param    name
    *           The legacy 4GL name for the function.
    * @param    modes
    *           A string representation of the modes of each parameter.
    * @param    args
    *           The functions's arguments.
    *           
    * @return   The return value from the block (or its nested blocks) which
    *           is the <code>unknown value</code> by default or on error.
    */
   public static <T extends BaseDataType> T
   invokeFunctionWithMode(Class<T>  cls, character name, String modes, Object... args)
   {
      BaseDataType res = invokeFunctionImpl(name, null, false, false, modes, args);
      return convertValue(cls, res);
   }

   /**
    * Emitted for the plain function calls, when they are defined as 
    * FUNCTION ... IN handle.
    *
    * @param    cls
    *           Class.
    * @param    name
    *           The legacy 4GL name for the function.
    * @param    h
    *           The handle to which the function belongs. If <code>null</code>
    *           defaults to THIS-PROCEDURE.
    * @param    modes
    *           A string representation of the modes of each parameter.
    * @param    args
    *           The functions's arguments.
    *           
    * @return   The return value from the block (or its nested blocks) which
    *           is the <code>unknown value</code> by default or on error.
    */
   public static <T extends BaseDataType> T
   invokeFunctionInWithMode(Class<T>  cls, String name, handle h, String modes, Object... args)
   {
      BaseDataType res = invokeFunctionImpl(new character(name), h, false, false, modes, args);
      return convertValue(cls, res);
   }
   
   /**
    * Emitted for the plain function calls, when they are defined as 
    * FUNCTION ... IN handle.
    *
    * @param    cls
    *           Class.
    * @param    name
    *           The legacy 4GL name for the function.
    * @param    h
    *           The handle to which the function belongs. If <code>null</code>
    *           defaults to THIS-PROCEDURE.
    * @param    modes
    *           A string representation of the modes of each parameter.
    * @param    args
    *           The functions's arguments.
    *           
    * @return   The return value from the block (or its nested blocks) which
    *           is the <code>unknown value</code> by default or on error.
    */
   public static <T extends BaseDataType> T
   invokeFunctionInWithMode(Class<T> cls, character name, handle h, String modes, Object... args)
   {
      BaseDataType res = invokeFunctionImpl(name, h, false, false, modes, args);
      return convertValue(cls, res);
   }

   /**
    * Emitted for plain function calls.
    *
    * @param    cls
    *           Class.
    * @param    name
    *           The legacy 4GL name for the function.
    * @param    args
    *           The functions's arguments.
    *           
    * @return   The return value from the block (or its nested blocks) which
    *           is the <code>unknown value</code> by default or on error.
    */
   public static <T extends BaseDataType> T invokeFunction(Class<T>  cls,
                                                           String    name,
                                                           Object... args)
   {
      BaseDataType res = invokeFunctionImpl(new character(name), null, false, false, null, args);
      return convertValue(cls, res);
   }

   /**
    * Emitted for plain function calls.
    *
    * @param    cls
    *           Class.
    * @param    name
    *           The legacy 4GL name for the function.
    * @param    args
    *           The functions's arguments.
    *           
    * @return   The return value from the block (or its nested blocks) which
    *           is the <code>unknown value</code> by default or on error.
    */
   public static <T extends BaseDataType> T invokeFunction(Class<T>  cls,
                                                           character name,
                                                           Object... args) 
   {
      BaseDataType res = invokeFunctionImpl(name, null, false, false, null, args);
      return convertValue(cls, res);
   }

   /**
    * Emitted for the plain function calls, when they are defined as 
    * FUNCTION ... IN handle.
    *
    * @param    cls
    *           Class.
    * @param    name
    *           The legacy 4GL name for the function.
    * @param    h
    *           The handle to which the function belongs. If <code>null</code>
    *           defaults to THIS-PROCEDURE.
    * @param    args
    *           The functions's arguments.
    *           
    * @return   The return value from the block (or its nested blocks) which
    *           is the <code>unknown value</code> by default or on error.
    */
   public static <T extends BaseDataType> T invokeFunctionIn(Class<T>  cls,
                                                             String    name,
                                                             handle    h, 
                                                             Object... args)
   {
      BaseDataType res = invokeFunctionImpl(new character(name), h, false, false, null, args);
      return convertValue(cls, res);
   }
   
   /**
    * Emitted for the plain function calls, when they are defined as 
    * FUNCTION ... IN handle.
    *
    * @param    cls
    *           Class.
    * @param    name
    *           The legacy 4GL name for the function.
    * @param    h
    *           The handle to which the function belongs. If <code>null</code>
    *           defaults to THIS-PROCEDURE.
    * @param    args
    *           The functions's arguments.
    *           
    * @return   The return value from the block (or its nested blocks) which
    *           is the <code>unknown value</code> by default or on error.
    */
   public static <T extends BaseDataType> T invokeFunctionIn(Class<T>  cls,
                                                             character name,
                                                             handle    h,
                                                             Object... args) 
   {
      BaseDataType res = invokeFunctionImpl(name, h, false, false, null, args);
      return convertValue(cls, res);
   }

   /**
    * Emitted for plain function calls.
    * 
    * @param    name
    *           The legacy 4GL name for the function.
    * @param    modes
    *           A string representation of the modes of each parameter.
    * @param    args
    *           The functions's arguments.
    *           
    * @return   The return value from the block (or its nested blocks) which
    *           is the <code>unknown value</code> by default or on error.
    */
   public static BaseDataType invokeFunctionWithMode(String    name,
                                                     String    modes,
                                                     Object... args)
   {
      return invokeFunctionImpl(new character(name), null, false, false, modes, args);
   }

   /**
    * Emitted for plain function calls.
    * 
    * @param    name
    *           The legacy 4GL name for the function.
    * @param    modes
    *           A string representation of the modes of each parameter.
    * @param    args
    *           The functions's arguments.
    *           
    * @return   The return value from the block (or its nested blocks) which
    *           is the <code>unknown value</code> by default or on error.
    */
   public static BaseDataType invokeFunctionWithMode(character name,
                                                     String    modes,
                                                     Object... args) 
   {
      return invokeFunctionImpl(name, null, false, false, modes, args);
   }

   /**
    * Emitted for the plain function calls, when they are defined as 
    * FUNCTION ... IN handle.
    * 
    * @param    name
    *           The legacy 4GL name for the function.
    * @param    h
    *           The handle to which the function belongs. If <code>null</code>
    *           defaults to THIS-PROCEDURE.
    * @param    modes
    *           A string representation of the modes of each parameter.
    * @param    args
    *           The functions's arguments.
    *           
    * @return   The return value from the block (or its nested blocks) which
    *           is the <code>unknown value</code> by default or on error.
    */
   public static BaseDataType invokeFunctionInWithMode(String    name,
                                                       handle    h,
                                                       String    modes,
                                                       Object... args)
   {
      return invokeFunctionImpl(new character(name), h, false, false, modes, args);
   }
   
   /**
    * Emitted for the plain function calls, when they are defined as 
    * FUNCTION ... IN handle.
    * 
    * @param    name
    *           The legacy 4GL name for the function.
    * @param    h
    *           The handle to which the function belongs. If <code>null</code>
    *           defaults to THIS-PROCEDURE.
    * @param    modes
    *           A string representation of the modes of each parameter.
    * @param    args
    *           The functions's arguments.
    *           
    * @return   The return value from the block (or its nested blocks) which
    *           is the <code>unknown value</code> by default or on error.
    */
   public static BaseDataType invokeFunctionInWithMode(character name,
                                                       handle    h,
                                                       String    modes,
                                                       Object... args) 
   {
      return invokeFunctionImpl(name, h, false, false, modes, args);
   }

   /**
    * Emitted for plain function calls.
    * 
    * @param    name
    *           The legacy 4GL name for the function.
    * @param    args
    *           The functions's arguments.
    *           
    * @return   The return value from the block (or its nested blocks) which
    *           is the <code>unknown value</code> by default or on error.
    */
   public static BaseDataType invokeFunction(String name, Object... args)
   {
      return invokeFunctionImpl(new character(name), null, false, false, null, args);
   }

   /**
    * Emitted for plain function calls.
    * 
    * @param    name
    *           The legacy 4GL name for the function.
    * @param    args
    *           The functions's arguments.
    *           
    * @return   The return value from the block (or its nested blocks) which
    *           is the <code>unknown value</code> by default or on error.
    */
   public static BaseDataType invokeFunction(character name, Object... args) 
   {
      return invokeFunctionImpl(name, null, false, false, null, args);
   }

   /**
    * Emitted for the plain function calls, when they are defined as 
    * FUNCTION ... IN handle.
    * 
    * @param    name
    *           The legacy 4GL name for the function.
    * @param    h
    *           The handle to which the function belongs. If <code>null</code>
    *           defaults to THIS-PROCEDURE.
    * @param    args
    *           The functions's arguments.
    *           
    * @return   The return value from the block (or its nested blocks) which
    *           is the <code>unknown value</code> by default or on error.
    */
   public static BaseDataType invokeFunctionIn(String name, handle h, Object... args)
   {
      return invokeFunctionImpl(new character(name), h, false, false, null, args);
   }
   
   /**
    * Emitted for the plain function calls, when they are defined as 
    * FUNCTION ... IN handle.
    * 
    * @param    name
    *           The legacy 4GL name for the function.
    * @param    h
    *           The handle to which the function belongs. If <code>null</code>
    *           defaults to THIS-PROCEDURE.
    * @param    args
    *           The functions's arguments.
    *           
    * @return   The return value from the block (or its nested blocks) which
    *           is the <code>unknown value</code> by default or on error.
    */
   public static BaseDataType invokeFunctionIn(character name, handle h, Object... args)
   {
      return invokeFunctionImpl(name, h, false, false, null, args);
   }

   /**
    * Check if an internal procedure or function is reachable and valid. This searches the
    * specified handle and all the super-procedures.
    *
    * @param    h
    *           A valid procedure handle.
    * @param    name
    *           The procedure or function name.
    * @param    function
    *           Flag indicating if the search is for a function or for an internal procedure.
    * @param    modes
    *           The parameter modes.
    * @param    validateModes
    *           Flag indicating if the parameter modes should be validated or not.
    * @param    args
    *           The passed arguments.
    *
    * @return   null if the internal entry can't be found; otherwise ArgValidationErrors instance
    *           with error status or {@link ArgValidationErrors#NO_ERROR} if error is not exists.
    */
   public static ArgValidationErrors reachableInternalEntry(handle    h,
                                                            String    name,
                                                            boolean   function,
                                                            String    modes,
                                                            boolean   validateModes,
                                                            Object... args)
   {
      WorkArea wa = work.get();
      InternalEntryCaller caller = null;
      for (Resolver r : wa.inHandleResolvers)
      {
         caller = r.resolve(h, name, function, false, null);

         if (caller != null)
         {
            break;
         }
      }

      if (caller == null)
      {
         // was not found
         return null;
      }
      
      ArgValidationErrors valid = caller.valid(function, modes, new int[2], args);
      if (valid != ArgValidationErrors.NO_ERROR)
      {
         return valid;
      }

      // validate the parameter modes
      String definedModes = caller.getParameterModes(function);
      if (validateModes && 
          (modes != null && !modes.equals(definedModes) ||
           (modes == null && definedModes != null)))
      {
         return ArgValidationErrors.INCORRECT_TYPES;
      }

      return valid;
   }

   /**
    * This is a special API which is used to delegate a call from converted 4GL code 
    * (via _P2J_REMOTE_CALL_ functions) to the remote side.  Usually, this is used when P2J client
    * runs in embedded mode, i.e. the P2J screens are embedded in another customer-specific 
    * application.  The call will be remoted into the customer-specific application.
    * <p>
    * This API is expected to block waiting for a response.
    * 
    * @param    content
    *           The request data incoming from 4GL.
    * 
    * @return   The response data from the customer-specific application, which is sent back to 
    *           the converted 4GL code.
    */
   public static character remoteCall(character content)
   {
      if (content == null)
      {
         content = new character();
      }
      
      UnformattedPayload payload = new UnformattedPayload();
      payload.payload = content.toStringMessage();
      UnformattedPayload response = LogicalTerminal.remoteCall(payload);
      
      return new character(response.payload);
   }
   
   /**
    * This is a special API which is used to delegate a call from converted 4GL code 
    * (via _P2J_REMOTE_CALL_ functions) to the remote side.  Usually, this is used when P2J client
    * runs in embedded mode, i.e. the P2J screens are embedded in another customer-specific 
    * application.  The call will be remoted into the customer-specific application.
    * <p>
    * This API is expected to block waiting for a response.
    * 
    * @param    content
    *           The request data incoming from 4GL.
    * 
    * @return   The response data from the customer-specific application, which is sent back to 
    *           the converted 4GL code.
    */
   public static character remoteCall(String content)
   {
      return remoteCall(new character(content));
   }
   
   /**
    * This API is used when P2J client runs in embedded mode (i.e. the P2J screens are embedded
    * into another customer-specific application).  It allows the customer-specific to execute
    * converted 4GL code.  Only external programs, procedures and functions can be invoked.  The
    * result of this invocation (and any additional OUTPUT parameters, errors, etc) will be 
    * returned back to the customer-specific application.
    * 
    * @param    payload
    *           The details about the request.
    *           
    * @return   The response for this request.
    */
   public static InvocationResponsePayload invoke(InvocationRequestPayload payload)
   {
      WorkArea wa = work.get();
      
      handle h;
      boolean persistent;

      switch (payload.type)
      {
         case EXTERNAL_PROGRAM:
         {
            persistent = payload.persistent;

            h = (persistent ? new handle() : null);
            break;
         }
         default:
         {
            persistent = false;

            long hid = payload.handle;
            h = (hid == 0 ? null : handle.fromResourceId(hid));
            
            if (h != null && h._isValid() && !ProcedureManager.isRemotePersistentProcedure(h.get()))
            {
               String msg = "Attempting to access a persistent procedure which was not obtained " +
                            "by the remote, customer-specific application!";
               if (LOG.isLoggable(Level.SEVERE))
               {
                  LOG.severe(msg);
               }

               throw new SilentUnwindException(new IllegalStateException(msg));
            }
            
            break;
         }
      }
      
      character cname = new character(payload.name);
      boolean function = payload.type == InvokeType.FUNCTION;
      MapArgumentResolver argResolver = new MapArgumentResolver(payload.arguments);

      BlockRunner block = new BlockRunner((BlockRunner br) ->
      {
         BaseDataType result = null;
         
         work.get().remoteInvoke = true;

         if (payload.type == InvokeType.EXTERNAL_PROGRAM)
         {
            invokeExternalProcedure(cname, persistent, h, false, null, null, argResolver);
            result = wa.returnValue;
            
            if (persistent && h != null && h._isValid())
            {
               ProcedureManager.registerRemotePersistentProcedure(h.get());
            }
         }
         else
         {
            result = invokeImpl(wa.inHandleResolvers, h, cname, function, 
                                   false, false, false, null, argResolver);
            
            if (payload.type == InvokeType.PROCEDURE)
            {
               // if it is a procedure, get the result from RETURN-VALUE
               result = wa.returnValue;
            }
         }
         
         return result;
      },
      () ->
      {
         // if the API was resolved and validated, the flag will be reset just before the entry
         // is executed (so any downstream calls are not seen as "remote"); if there are any 
         // errors, and the entry could not be executed, the flag needs to be reset. 
         // so, we ensure the flag is reset here, always.
         wa.remoteInvoke = false;
      }, payload.noError);
      
      block.execute();

      InvocationResponsePayload response = new InvocationResponsePayload();
      // the handle is sent always (so that the customer app knows in which handle the program
      // was ran... so it can delete it (maybe)
      response.handle = (h != null && h._isValid() ? h.getResourceId() : 0);
      response.returnValue = (block.result == null ? null : block.result.toStringMessage());
      response.error = block.error;
      
      if (block.error)
      {
         response.errorNumber  = block.errorNumber;
         response.errorMessage = block.errorMessage;
      }
      
      response.stop = block.stop;
      response.quit = block.quit;
      response.outputArguments = argResolver.getOutputArguments();
      
      return response;
   }
   
   /**
    * The implementation of all function and procedure calls. This disambiguates and implements
    * the following calls:
    * <ul>
    *    <li>if <code>h</code> is not null, then this is a RUN ... IN handle call.</li>
    *    <li>if the <code>exports</code> is not null, then this is a RUN ... ON SERVER call, so
    *        only external programs will be searched.  Note that this parameter will be not-null
    *        only when this call is executed on the remote, appserver, side.</li>
    *    <li>else, this is a RUN statement call with no special clauses.</li>
    * </ul>
    * 
    * @param    h
    *           The handle to which the procedure belongs. If <code>null</code>, defaults to
    *           THIS-PROCEDURE.
    * @param    name
    *           The legacy 4GL name for the function or procedure.
    * @param    function
    *           <code>true</code> if this is a function call.
    * @param    dynamicFunction
    *           <code>true</code> if this is a DYNAMIC-FUNCTION call.
    * @param    superCall
    *           <code>true</code> if this is a RUN SUPER or SUPER() call.
    * @param    transactionDistinct
    *           This marks if "TRANSACTION DISTINCT" option is specified with the RUN ... ON stmt.
    * @param    modes
    *           A string representation of the modes of each parameter.
    * @param    args
    *           The arguments.
    *           
    * @return   The return value from the block (or its nested blocks) which is the <code>unknown
    *           value</code> by default, on error or for procedures.
    */
   static BaseDataType invoke(handle    h,
                              character name,
                              boolean   function,
                              boolean   dynamicFunction,
                              boolean   superCall,
                              boolean   transactionDistinct,
                              String    modes,
                              Object... args)
   {
      // h will never be a proxy procedure handle at the time of this call. if it ends up a
      // remote procedure handle here, we have a problem in the runtime implementation
      if (ProcedureManager.isProxy(h))
      {
         throw new IllegalArgumentException("The supplied procedure handle is a proxy handle!");
      }
      
      WorkArea wa = work.obtain();
      List<Resolver> resolvers = null;
      if (h != null)
      {
         // The target of a "RUN proc" stmt or DYNAMIC-FUNCTION call ends on 
         // first find in: 
         // - THIS-PROCEDURE 
         // - THIS-PROCEDURE:super-procedures 
         // - SESSION-super-procedures
         resolvers = wa.inHandleResolvers;
      }
      else
      {
         // The target of a "RUN proc" stmt or DYNAMIC-FUNCTION call ends on 
         // first find in: 
         // - THIS-PROCEDURE 
         // - THIS-PROCEDURE:super-procedures 
         // - SESSION-super-procedures
         // - if an external program with the given name is found (for the 
         // "RUN proc" statement without "IN handle" clause)
         // - search distinguishes between procedures and functions with the 
         // same name. On first name match, will attempt to invoke it. 
         // - is not necessarily for the target procedure to have its header 
         // defined in the current procedure. Applies only to "RUN proc" 
         // statement. 
         // - "RUN proc0." and "proc0()" will end up calling different code 
         // (one is a super-proc, the other is a super-function). 
         // The body for these two may be in different super-procedures.
         resolvers = wa.resolvers;
      }

      return invokeImpl(resolvers,
                        h, 
                        name, 
                        function, 
                        dynamicFunction, 
                        superCall,
                        transactionDistinct,
                        modes,
                        new ArrayArgumentResolver(args));
   }

   /**
    * Resolve the external program with the specified name.
    * 
    * @param    name
    *           The legacy 4GL name for the function or procedure.
    * @param    exports
    *           The list of allowed external procedures to be invoked with a RUN ... ON SERVER
    *           statement. 
    * 
    * @return  The instantiated (not yet initialized) {@link ExternalProgramWrapper}.  
    */
   static ExternalProgramWrapper resolveExternalProcedure(String name, character exports)
   {
      InternalEntryCaller caller = work.obtain().externalResolver
                                                .resolve(null, name, false, false, exports, true);
      
      ExternalProgramWrapper extProg = null;
      
      if (caller != null)
      {
         extProg = new ExternalProgramWrapper(caller.getCallerInstance());
         
         ProcedureManager.addProcedure(new handle(extProg), name, true);
      }
      
      return extProg;
   }

   /**
    * Initialize the specified external procedure. This is done by calling the associated
    * <code>execute</code> method.
    * 
    * @param    extProg
    *           The associated external program.
    * @param    transactionDistinct
    *           This marks if "TRANSACTION DISTINCT" option is specified with the RUN ... ON stmt.
    * @param    modes
    *           A string representation of the modes of each parameter.
    * @param    args
    *           The arguments.
    */
   static void initializeExternalProcedure(ExternalProgramWrapper extProg,
                                           boolean                transactionDistinct,
                                           String                 modes,
                                           Object...              args)
   {
      InternalEntryCaller caller = new ExternalProcedureCaller(extProg.get());

      // resolve the name early, as at the time the invokeError is called the proc handle is gone
      String relName = ProcedureManager.getRelativeName(extProg.referent);

      Exception deferred = null;
      handle hProc = new handle(extProg);
      try
      {
         if (!validArguments(relName, caller, false, modes, args))
         {
            return;
         }
   
         boolean error = false;
         try
         {
            Object referent = extProg.get();
            
            Runnable push = () ->
            {
               ProcedureManager.pushCalleeInfo(false, referent, referent, 
                                               ProcedureManager.EXTERNAL_PROGRAM, relName,
                                               false, true);
            };
            caller.setPushWorker(push);
            
            caller.invoke(modes, args);
         }
         catch (Exception e)
         {
            // any exceptions here cause the persistent procedure to no longer be available
            error = true;
   
            throw e;
         }
         finally
         {
            if (error || ErrorManager.isPending())
            {
               ProcedureManager.removePersistentProcedure(hProc);
            }
         }
      }
      
      // allow Progress compatible exceptions to flow through
      catch (ConditionException | RetryUnwindException ce)
      {
         throw ce;
      }
      
      // handle reflection exceptions
      catch (NoSuchMethodException nsm)
      {
         invokeFailure(caller.getMethodName());
         return;
      }
      
      catch (InvocationTargetException exc)
      {
         // if the exception thrown was a P2J runtime type, we
         // must handle it normally, so extract the root cause
         // exception and rethrow it
         Throwable cause = exc.getCause();
         if (cause != null)
         {
            if (cause instanceof ConditionException)
            {
               throw (ConditionException) cause;
            }
            if (cause instanceof RetryUnwindException)
            {
               throw (RetryUnwindException) cause;
            }
            
            // exception is a runtime problem that we don't
            // handle here
            deferred = exc;
         }
      }
      
      // all other cases
      catch (Exception exp)
      {
         deferred = exp;
      }
      
      if (deferred != null)
      {
         invokeError(caller, relName, deferred);
      }
   }
   
   /**
    * The implementation of all RUN ... PERSISTENT calls or appserver calls for an external
    * procedure. If the <code>exports</code> parameter is not null, then the ON SERVER clause is
    * in effect; this parameter can be not-null only on the remote, appserver, side.
    * 
    * @param    name
    *           The legacy 4GL name for the function or procedure.
    * @param    persistent
    *           Flag indicating that this call needs to be persistent.
    * @param    h
    *           The handle where to save the persistent procedure.
    * @param    transactionDistinct
    *           This marks if "TRANSACTION DISTINCT" option is specified with the RUN ... ON stmt.
    * @param    exports
    *           The list of allowed external procedures to be invoked with a RUN ... ON SERVER
    *           statement. 
    * @param    modes
    *           A string representation of the modes of each parameter.
    * @param    args
    *           The arguments.
    */
   static void invokeExternalProcedure(character name,
                                       boolean   persistent,
                                       handle    h,
                                       boolean   transactionDistinct,
                                       character exports,
                                       String    modes,
                                       Object... args)
   {
      invokeExternalProcedure(name, 
                              persistent, 
                              h, 
                              transactionDistinct, 
                              exports, 
                              modes, 
                              new ArrayArgumentResolver(args));
   }
   
   /**
    * The implementation of all RUN ... PERSISTENT calls or appserver calls for an external
    * procedure. If the <code>exports</code> parameter is not null, then the ON SERVER clause is
    * in effect; this parameter can be not-null only on the remote, appserver, side.
    * 
    * @param    name
    *           The legacy 4GL name for the function or procedure.
    * @param    persistent
    *           Flag indicating that this call needs to be persistent.
    * @param    h
    *           The handle where to save the persistent procedure.
    * @param    transactionDistinct
    *           This marks if "TRANSACTION DISTINCT" option is specified with the RUN ... ON stmt.
    * @param    exports
    *           The list of allowed external procedures to be invoked with a RUN ... ON SERVER
    *           statement. 
    * @param    modes
    *           A string representation of the modes of each parameter.
    * @param    argResolver
    *           The argument resolver.
    */
   static void invokeExternalProcedure(character        name,
                                       boolean          persistent,
                                       handle           h,
                                       boolean          transactionDistinct,
                                       character        exports,
                                       String           modes,
                                       ArgumentResolver argResolver)
   {
      String pname = name.toStringMessage();
      
      InternalEntryCaller caller = null;
      Exception deferred = null;
      try
      {
         caller = work.obtain().externalResolver.resolve(null, pname, false, false, exports,
                                                         persistent);
         
         if (caller == null)
         {
            if (!ErrorManager.isPendingError())
            {
               // not a pending error, it means the program was not found
               invokeFailure(pname);
            }
            
            return;
         }
   
         Object[] args = argResolver.resolve(caller, pname, false);
         
         if (!validArguments(pname, caller, false, modes, args))
         {
            return;
         }
   
         Object instance = caller.getCallerInstance();
   
         boolean error = false;
         try
         {
            Runnable push = () ->
            {
               ProcedureManager.pushCalleeInfo(false, instance, instance,
                                               ProcedureManager.EXTERNAL_PROGRAM, pname, 
                                               false, persistent);
            };
            caller.setPushWorker(push);
            
            caller.invoke(modes, args);
         }
         catch (Exception e)
         {
            error = true;
   
            throw e;
         }
         finally
         {
            if (persistent)
            {
               if ((error || ErrorManager.isPending()))
               {
                  // any exceptions here cause the persistent procedure to no longer
                  // be available
                  ProcedureManager.removePersistentProcedure(h);
               }
               else if (h != null)
               {
                  // if the handle is not also passed as an OUTPUT argument, then assign it
                  boolean asArg = false;
                  String argModes = caller.getParameterModes(false);
                  int i = 0;
                  for (Object arg : args)
                  {
                     if (argModes == null || i >= argModes.length())
                     {
                        break;
                     }

                     char mode = argModes.charAt(i);
                     
                     if (arg == h && mode != 'I')
                     {
                        // not an input mode, set it to unknown
                        asArg = true;
                        break;
                     }

                     i++;
                  }

                  if (asArg)
                  {
                     h.setUnknown();
                  }
                  else
                  {
                     h.set(new ExternalProgramWrapper(instance));
                  }
               }
            }
         }
      }
      
      // allow Progress compatible exceptions to flow through
      catch (ConditionException | RetryUnwindException ce)
      {
         throw ce;
      }
      
      // handle reflection exceptions
      catch (NoSuchMethodException nsm)
      {
         invokeFailure(caller.getMethodName());
         return;
      }
      
      catch (InvocationTargetException exc)
      {
         // if the exception thrown was a P2J runtime type, we
         // must handle it normally, so extract the root cause
         // exception and rethrow it
         Throwable cause = exc.getCause();
         if (cause != null)
         {
            if (cause instanceof ConditionException)
            {
               throw (ConditionException) cause;
            }
            if (cause instanceof RetryUnwindException)
            {
               throw (RetryUnwindException) cause;
            }
            
            // exception is a runtime problem that we don't
            // handle here
            deferred = exc;
         }
      }
      
      // all other cases
      catch (Exception exp)
      {
         deferred = exp;
      }
      
      if (deferred != null)
      {
         invokeError(caller, pname, deferred);
      }
   }

   /**
    * Resolve the program handle of an internal procedure <code>name</code>.
    * <p>
    * The method will try to find the internal procedure <code>name</code> in program
    * <code>h</code>. If <code>name</code> is not found in <code>h</code>, super procedures of
    * <code>h</code> are searched. If <code>name</code> is not found in super procedures of
    * <code>h</code> <code>SESSION</code> super procedures are searched. If <code>name</code>
    * is not found, <code>null</code> is returned.
    *
    * @param    h
    *           The program handle.
    * @param    name
    *           The internal procedure name.
    *
    * @return   Program handle or <code>null</code> if not found.
    */
   static handle resolveProgramForProcedure(handle h, String name)
   {
      WorkArea wa = work.obtain();
      InternalEntryCaller caller = null;
      for (Resolver r : wa.inHandleResolvers)
      {
         caller = r.resolve(h, name, false, false, null);
         if (caller != null)
         {
            return caller.phandle;
         }
      }
      return null;
   }

   /**
    * Convert BaseDataType value to instance of given class. If value is <code>null</code>,
    * returns <code>null</code>. If value is <code>unknown value</code>, creates a new instance
    * of the given type that represents the <code>unknown value</code>, otherwise returns value
    * casted to given type.
    *
    * @param    cls
    *           Class.
    * @param    value
    *           Value.
    *
    * @return   If value is <code>null</code>, returns <code>null</code>. If value is
    *           <code>unknown value</code>, creates a new instance of the given type that
    *           represents the <code>unknown value</code>, otherwise returns value casted to
    *           given type.
    *
    * @throws   IllegalStateException
    *           If value is not compatible with given class.
    */
   private static <T extends BaseDataType> T convertValue(Class<T> cls, BaseDataType value)
   {
      if (value == null)
      {
         return null;
      }

      if (value instanceof unknown)
      {
         value = BaseDataType.generateUnknown(cls);
      }

      if (!value.getClass().isAssignableFrom(cls))
      {
         // IN SUPER may be defined with i.e. CHARACTER but actual definition may be LONGCHAR
         if (LOG.isLoggable(Level.WARNING))
         {
            LOG.warning("Value of type " + value.getClass().getName() +
                        " is not compatible with expected return type " + cls.getName());
         }
      }

      return (T) value;
   }

   /**
    * Execute an async procedure call.
    * 
    * @param    request
    *           The request which will execute the procedure call.
    * @param    hServer
    *           The server used to execute this async call. 
    * @param    asyncHandle
    *           The handle to the asynchronous request.     * @param    pHandle
    *           The handle to which the procedure belongs, if it is RUN ... IN handle statement.
    */
   private static void invokeAsyncImpl(AsyncCall request,
                                       handle    hServer,
                                       handle    asyncHandle,
                                       handle    pHandle)
   {
      // expose the async resource to the caller
      asyncHandle.assign(request.asyncReq);

      boolean isServer = (hServer.getResource() instanceof ServerImpl);
      ServerImpl server = isServer ? (ServerImpl) hServer.getResource() : null;

      // register for notifications the server and the procedure handle
      if (isServer)
      {
         request.asyncReq.registerListener(server);
      }
      // check the procedure handle
      if (pHandle._isValid() && pHandle.getResource() instanceof AsyncRequestListener)
      {
         request.asyncReq.registerListener((AsyncRequestListener) pHandle.getResource());
      }
      
      // notify all listeners an async call is in effect
      request.asyncReq.notifyStart();
      
      // perform invocation
      if (isServer && server._connected())
      {
         // let the server run this async request
         server.invokeAsync(request.asyncReq);
      }
      else
      {
         // invoke immediately
         request.asyncReq.execute(false);
      }
   }

   /**
    * The implementation for all function calls.
    * 
    * @param    name
    *           The legacy 4GL name for the function.
    * @param    h
    *           The handle to which the function belongs. If <code>null</code>
    *           defaults to THIS-PROCEDURE.
    * @param    dynamicFunction
    *           <code>true</code> if this is a DYNAMIC-FUNCTION call.
    * @param    superCall
    *           <code>true</code> if this is a SUPER() call.
    * @param    modes
    *           A string representation of the modes of each parameter.
    * @param    args
    *           The functions's arguments.
    *           
    * @return   The return value from the block (or its nested blocks) which
    *           is the <code>unknown value</code> by default or on error.
    */
   private static BaseDataType invokeFunctionImpl(character name,
                                                  handle    h,
                                                  boolean   dynamicFunction,
                                                  boolean   superCall,
                                                  String    modes,
                                                  Object... args) 
   {
      // TODO: add support for functions returning an extent
      return invokeImpl((handle) null, h, name, 
                        true, dynamicFunction, superCall,
                        false, modes, args);
   }

   /**
    * The implementation for RUN SUPER and RUN ... IN handle calls.
    * 
    * @param    name
    *           The legacy 4GL name for the procedure.
    * @param    h
    *           The handle to which the procedure belongs. If <code>null</code>
    *           defaults to THIS-PROCEDURE.
    * @param    superCall
    *           <code>true</code> if this is a RUN SUPER call.
    * @param    modes
    *           A string representation of the modes of each parameter.
    * @param    args
    *           The functions's arguments.
    */
   private static void invokeInImpl(character name, 
                                    handle    h,
                                    boolean   superCall,
                                    String    modes,
                                    Object... args)
   {
      invokeImpl((handle) null, h, name, false, false, superCall, false, modes, args);
   }
   
   
   /**
    * The implementation of function or procedure calls which do not use the
    * IN handle clause.
    * 
    * @param    server
    *           The target appserver, to which this call will be dispatched.
    *           Is <code>null</code> if the RUN statement doesn't have the ON SERVER clause.
    * @param    name
    *           The legacy 4GL name for the function or procedure.
    * @param    h
    *           The handle to which the procedure belongs. If <code>null</code> defaults to 
    *           THIS-PROCEDURE. If this is a proxy procedure, the call is dispatched to the
    *           procedure's server.
    * @param    function
    *           <code>true</code> if this is a function call.
    * @param    dynamicFunction
    *           <code>true</code> if this is a DYNAMIC-FUNCTION call.
    * @param    superCall
    *           <code>true</code> if this is a RUN SUPER or SUPER() call.
    * @param    transactionDistinct
    *           This marks if "TRANSACTION DISTINCT" option is specified with the RUN ... ON stmt.
    * @param    modes
    *           A string representation of the modes of each parameter.
    * @param    args
    *           The arguments.
    *           
    * @return   The return value from the block (or its nested blocks) which
    *           is the <code>unknown value</code> by default, on error or for
    *           procedures.
    */
   private static BaseDataType invokeImpl(handle    server,
                                          handle    h,
                                          character name,
                                          boolean   function,
                                          boolean   dynamicFunction,
                                          boolean   superCall,
                                          boolean   transactionDistinct,
                                          String    modes,
                                          Object... args)
   {
      return invokeImpl(null, server, h, name,
                        function, dynamicFunction, superCall, transactionDistinct,
                        modes, args);
   }

   /**
    * The implementation of function or procedure calls which do not use the
    * IN handle clause.
    * 
    * @param    asyncReq
    *           When not-null, an {@link AsyncRequestImpl} instance for the associated async
    *           request.
    * @param    server
    *           The target appserver, to which this call will be dispatched.
    *           Is <code>null</code> if the RUN statement doesn't have the ON SERVER clause.
    * @param    name
    *           The legacy 4GL name for the function or procedure.
    * @param    h
    *           The handle to which the procedure belongs. If <code>null</code> defaults to 
    *           THIS-PROCEDURE. If this is a proxy procedure, the call is dispatched to the
    *           procedure's server.
    * @param    function
    *           <code>true</code> if this is a function call.
    * @param    dynamicFunction
    *           <code>true</code> if this is a DYNAMIC-FUNCTION call.
    * @param    superCall
    *           <code>true</code> if this is a RUN SUPER or SUPER() call.
    * @param    transactionDistinct
    *           This marks if "TRANSACTION DISTINCT" option is specified with the RUN ... ON stmt.
    * @param    modes
    *           A string representation of the modes of each parameter.
    * @param    args
    *           The arguments.
    *           
    * @return   The return value from the block (or its nested blocks) which
    *           is the <code>unknown value</code> by default, on error or for
    *           procedures.
    */
   private static BaseDataType invokeImpl(AsyncRequestImpl asyncReq,
                                          handle           server,
                                          handle           h,
                                          character        name,
                                          boolean          function,
                                          boolean          dynamicFunction,
                                          boolean          superCall,
                                          boolean          transactionDistinct,
                                          String           modes,
                                          Object...        args)
   {
      // save the original received server instance.
      final handle origServer = server;

      if (server == null && h != null && h._isValid())
      {
         WrappedResource res = h.getResource();
         
         // get the server from the handle
         if (ProcedureManager.isProxy(h))
         {
            server = new handle(((ProxyProcedureWrapper) res).getServer());
         }
         else if (res instanceof PortTypeWrapper)
         {
            server = new handle(((PortTypeWrapper) res).getServer());
         }
      }
      
      // if the server is the SESSION handle, then this is a local call
      if (server == null || CompareOps._isEqual(server, SessionUtils.asHandle()))
      {
         // this is a normal call, invoke immediately
         return invoke(h, name, function, dynamicFunction, superCall, false, modes, args);
      }
      else
      {
         if (!validServer(server, h, name))
         {
            return new unknown();
         }

         ServerImpl s = (ServerImpl) server.get();
         boolean isWebService = s.isWebService();
         
         if (isWebService && origServer != null && h == null)
         {
            // the target portHandle must always be specified in case of 
            // RUN port-type ON SERVER h-web-server statements.
            // if is missing, the RUN statement is a no-op
            
            return new unknown();
         }

         BaseDataType result = new unknown();
         try
         {
            // if sync mode, no async requests must be running
            if (asyncReq == null && s.hasAsyncRequests())
            {
               return new unknown();
            }

            // invoke appServer
            ServerHelper helper = s.getServerHelper();
            // if ON SERVER is used (h is null), only external procedures can be invoked.
            // this is managed by the remote side.
            result = helper.invoke(0, asyncReq, name, h,
                                   function, dynamicFunction, superCall,
                                   transactionDistinct, modes, args);
            
            if (!function && asyncReq == null && result != null && !isWebService)
            {
               // TODO: the return-value is set for web-service requests only if the web-service 
               // is configured and exported via Progress.

               setReturnValue((character) result);
               result = new unknown();
            }
         }
         catch (ConditionException e)
         {
            if (!isWebService && e.getCause() instanceof NumberedException)
            {
               ErrorManager.recordOrThrowError((NumberedException) e.getCause());
            }
            else
            {
               throw e;
            }
         }

         return result;
      }
   }
   
   /**
    * The implementation of all function and procedure calls.
    * 
    * @param    resolvers
    *           The {@link Resolver} instances to use to resolve the internal
    *           entry to a Java method, based on the passed parameters.
    * @param    h
    *           The handle to which the procedure belongs. If <code>null</code>
    *           defaults to THIS-PROCEDURE.
    * @param    cname
    *           The legacy 4GL name for the function.
    * @param    function
    *           <code>true</code> if this is a function call.
    * @param    dynamicFunction
    *           <code>true</code> if this is a DYNAMIC-FUNCTION call.
    * @param    superCall
    *           <code>true</code> if this is a RUN SUPER or SUPER() call.
    * @param    transactionDistinct
    *           This marks if "TRANSACTION DISTINCT" option is specified with the RUN ... ON stmt.
    * @param    modes
    *           A string representation of the modes of each parameter.
    * @param    argResolver
    *           The argument resolver.
    *           
    * @return   The return value from the block (or its nested blocks) which
    *           is the <code>unknown value</code> by default, on error or for
    *           procedures.
    */
   private static BaseDataType invokeImpl(List<Resolver>   resolvers,
                                          handle           h,
                                          character        cname,
                                          boolean          function,
                                          boolean          dynamicFunction,
                                          boolean          superCall,
                                          boolean          transactionDistinct,
                                          String           modes,
                                          ArgumentResolver argResolver)
   {
      BaseDataType result = new unknown();

      String name = cname.toStringMessage();

      // handle validation is done first
      boolean explicit = h != null;
      if (!explicit)
      {
         h = ProcedureManager.thisProcedure();
      }
     
      // this is a special case only for functions. if the handle has been
      // deleted and it still contains the reference to the procedure, then
      // it defaults to this-procedure
      if (function && !h._isValid() && ProcedureManager.isProcedure(h))
      {
         h = ProcedureManager.thisProcedure();
      }

      if (!validHandle(h, name, explicit, function, dynamicFunction))
      {
         return result;
      }
   
      if (cname.isUnknown() || cname.toStringMessage().length() == 0)
      {
         String err = null;
         int errid = -1;

         if (function && dynamicFunction)
         {
            err = "Could not invoke dynamic user-defined function";
            errid = 5640;
         }
         
         if (err == null)
         {
            if (cname.isUnknown())
            {
               invokeFailure(name);
               return result;
            }
            else
            {
               err = "No PROGRESS program specified for RUN or COMPILE";
               errid = 467;
            }
         }
         
         ErrorManager.recordOrThrowError(errid, err);
         return result;
      }

      Exception deferred = null;
      InternalEntryCaller caller = null;
      try
      {
         for (Resolver r : resolvers)
         {
            caller = r.resolve(h, name, function, superCall, null);
      
            if (caller instanceof MapToNotFound)
            {
               // if a map-to was searched and not explicitly found, search in super-procedures
               name = caller.getInternalEntryName();
               caller = null;
            }

            if (caller != null)
            {
               break;
            }
            
            if (r instanceof ExternalProgramResolver && ErrorManager.isPendingError())
            {
               // instantiation failed (i.e. due to a SHARED problem), do nothing
               return result;
            }
         }
         
         if (caller == null)
         {
            String pname = ProcedureManager.getRelativeName(h.get());
            if (function)
            {
               // if the function is defined but could not be resolved, 
               // then show error
               if (dynamicFunction)
               {
                  final String msg = 
                     "User-defined function '%s' invoked  dynamically but could not be found";
                  if (SessionUtils._isRemote())
                  {
                     ErrorManager.recordOrThrowError(5639, String.format(msg, name), false);
                  }
                  else
                  {
                     ErrorManager.recordOrShowError(5639, String.format(msg, name), 
                                                    false, false, false);
                  }
               }
               else
               {
                  if (SessionUtils._isRemote())
                  {
                     final String msg =
                        "User-defined function '%s' invoked but could not be found";
                     ErrorManager.recordOrThrowError(6445, String.format(msg, name), false);
                  }
                  else
                  {
                     final String msg = "External user defined function '%s' is not defined in " +
                                        "the procedure context given";
                     ErrorManager.recordOrShowError(2779, String.format(msg, name),
                                                    false, false, false);
                  }
               }
               
               return result;
            }
            else if (explicit)
            {
               final String msg = "Procedure %s has no entry point for %s";
               ErrorManager.recordOrThrowError(6456, 
                                               String.format(msg, pname, name),
                                               false);
               return result;
            }

            invokeFailure(name);
            return result;
         }

         if (function)
         {
            // caller.phandle is null in cases of external programs
            handle callerHandle = caller != null ? caller.getPhandle() : null;

            if (callerHandle != null && !CompareOps._isEqual(callerHandle, h))
            {
               // TODO: this should recursively call invokeImpl(server, handle, ....), as the
               // new handle might be for a remote procedure

               // handle has switched, validate the new one. dynamic-function
               // needs to be false here.
               if (!validHandle(callerHandle, name, explicit, function, false))
               {
                  return result;
               }
            }

            handle funcProg = callerHandle != null ? callerHandle : h;

            // this recursive call is needed only for functions, as only they
            // can be set as IN handle.
            String progName = ProcedureManager.getAbsoluteName(funcProg.get());
            if (SourceNameMapper.isInHandle(progName, name, function))
            {
               try
               {
                  return invokeImpl(resolvers,
                                    funcProg,
                                    cname,
                                    function,
                                    dynamicFunction,
                                    superCall,
                                    transactionDistinct,
                                    modes,
                                    argResolver);
               }
               catch (StackOverflowError t)
               {
                  final String msg = "Giving up search for external user defined function " +
                                     "'%s' after 64 indirections";
                  String err = String.format(msg, name);
                  int num = 2757;

                  ErrorManager.displayError(num, err, false);
                  
                  return result;
               }
            }
         }
      
         Object[] args = argResolver.resolve(caller, name, function);
         
         if (!validArguments(name, caller, function, modes, args))
         {
            return result;
         }

         handle hTarget;
         boolean external = (caller instanceof ExternalProcedureCaller);

         final String iename;
         final String pname;
         if (external)
         {
            // we are invoking an external program. the target-procedure will be the 
            // to-be-invoked procedure.
            ExternalProgramWrapper extProg = new ExternalProgramWrapper(caller.callerInstance);
            hTarget = new handle(extProg);
            
            iename = ProcedureManager.EXTERNAL_PROGRAM;
            pname = name;
         }
         else
         {
            // use the original program handle as the target, consider
            // the case when the internal procedure is resolved from
            // the SUPER program
            hTarget = new handle(h);
            iename = name;
            pname = null;
         }

         Object callerInstance = caller.callerInstance;
         Runnable push = () ->
         {
            ProcedureManager.pushCalleeInfo(superCall, hTarget.get(), callerInstance, 
                                            iename, pname, function, false);
         };
         caller.setPushWorker(push);

         Object res = caller.invoke(modes, args);
         if (res == null)
         {
            if (!function)
            {
               result = new unknown();
            }
            else
            {
               final String msg = "Function call '%s' has returned a null value!";
               throw new NullPointerException(String.format(msg, name));
            }
         }
         else if (res instanceof BaseDataType)
         {
            result = (BaseDataType) res;
         }
         else
         {
            if (function)
            {
               final String spec = 
                  "Incompatible return value by function %s: [%s] of type [%s]";
               String err = String.format(spec, name,
                                          result.toString(), 
                                          result.getClass().toString());
               ErrorManager.displayError(err);

               throw new StopConditionException(err);
            }
         }
      }
      
      // allow Progress compatible exceptions to flow through
      catch (ConditionException | RetryUnwindException ce)
      {
         throw ce;
      }
      
      // handle reflection exceptions
      catch (NoSuchMethodException nsm)
      {
         // should never happen, unless name_map.xml is corrupt
         deferred = nsm;
      }
      
      catch (InvocationTargetException exc)
      {
         // if the exception thrown was a P2J runtime type, we
         // must handle it normally, so extract the root cause
         // exception and rethrow it
         Throwable cause = exc.getCause();
         while (cause != null)
         {
            if (cause instanceof ConditionException)
            {
               if (cause instanceof InvalidParameterConditionException)
               {
                  // TODO: it is not entirely clear how safe it is to directly instantiate this
                  //       error; the InvalidParameterConditionException can be raised (we think)
                  //       only when there is some temp-table parameter error (see 
                  //       TemporaryBuffer.handleTablesDoNotMatch); the location for the throw
                  //       suggests that the ERROR condition is raised regardless of NO-ERROR
                  //       mode
                  
                  // re-wrap NumberedException
                  throw new ErrorConditionException(cause.getCause());
               }
               else
               {
                  throw (ConditionException) cause;
               }
            }
            if (cause instanceof RetryUnwindException)
            {
               throw (RetryUnwindException) cause;
            }
            
            // exception is a runtime problem that we don't handle here
            if (deferred == null)
            {
               deferred = exc;
            }

            cause = cause.getCause();
         }
      }
      
      // all other cases
      catch (Exception exp)
      {
         deferred = exp;
      }
      
      if (deferred != null)
      {
         invokeError(caller, name, deferred);
      }
      
      return result;
   }

   /**
    * The implementation for the RUN ... PERSISTENT [SET handle] [ON SERVER server] statement. 
    * 
    * @param    server
    *           The server instance on which to run the procedure.  This will save the procedure
    *           in the remote persistent procedure list, not on the requester side.  May be null.
    * @param    name
    *           The legacy 4GL name for the procedure.
    * @param    persistHandle
    *           The handle to which the procedure reference must be saved.
    *           When <code>null</code> and not ran remotely, just adds the procedure to the
    *           persistent procedure list.
    * @param    portHandle
    *           A handle variable where to save the web service. May be null.
    * @param    transactionDistinct
    *           This marks if "TRANSACTION DISTINCT" option is specified with the RUN ... ON stmt.
    * @param    modes
    *           A string representation of the modes of each parameter.
    * @param    args
    *           The procedure's arguments.
    */
   private static void invokePersistentImpl(handle    server,
                                            character name,
                                            handle    persistHandle,
                                            handle    portHandle,
                                            boolean   transactionDistinct,
                                            String    modes,
                                            Object... args) 
   {
      invokePersistentImpl(null, server, name,
                           persistHandle, portHandle,
                           transactionDistinct,
                           modes, args);
   }
   
   /**
    * Initialize the proxy procedure associated with the async request.
    * 
    * @param    call
    *           The async request which currently holds a not-initialized proxy procedure.
    */
   private static void initializeProxy(AsyncCall call)
   {
      if (!validServer(call.hServer, call.pHandle, call.pName))
      {
         return;
      }

      ServerImpl s = (ServerImpl) call.hServer.get();
      if (s.isWebService())
      {
         throw new IllegalStateException(
                                "initializeProxy can be used only with an appserver connection!");
      }
      else
      {
         try
         {
            // execute the proxy
            AppServerHelper appServer = (AppServerHelper) s.getServerHelper();
            appServer.initializeProxy(call.asyncReq,
                                      (String) call.pHandle.get(), 
                                      call.transactionDistinct,
                                      call.asyncReq.getModes(),
                                      call.asyncReq.getArgs());
         }
         catch (ConditionException e)
         {
            if (e.getCause() instanceof NumberedException)
            {
               ErrorManager.recordOrThrowError((NumberedException) e.getCause());
            }
            else
            {
               throw e;
            }
         }
         finally
         {
            call.asyncReq.complete();
         }
      }
   }
   
   /**
    * The implementation for the RUN ... PERSISTENT [SET handle] [ON SERVER server] statement. 
    * 
    * @param    asyncReq
    *           When not-null, an {@link AsyncRequestImpl} instance for the associated async
    *           request.
    * @param    server
    *           The server instance on which to run the procedure.  This will save the procedure
    *           in the remote persistent procedure list, not on the requester side.  May be null.
    * @param    name
    *           The legacy 4GL name for the procedure.
    * @param    persistHandle
    *           The handle to which the procedure reference must be saved.
    *           When <code>null</code> and not ran remotely, just adds the procedure to the
    *           persistent procedure list.
    * @param    portHandle
    *           A handle variable where to save the web service. May be null.
    * @param    transactionDistinct
    *           This marks if "TRANSACTION DISTINCT" option is specified with the RUN ... ON stmt.
    * @param    modes
    *           A string representation of the modes of each parameter.
    * @param    args
    *           The procedure's arguments.
    */
   private static void invokePersistentImpl(AsyncRequestImpl asyncReq,
                                            handle           server,
                                            character        name,
                                            handle           persistHandle,
                                            handle           portHandle,
                                            boolean          transactionDistinct,
                                            String           modes,
                                            Object...        args) 
   {
      // this will always be an external procedure call
      if (name.isUnknown() || name.toStringMessage().length() == 0)
      {
         String err = "No PROGRESS program specified for RUN or COMPILE";
         ErrorManager.recordOrThrowError(467, err);
         return;
      }

      // if the server is the SESSION handle, then this is a local call
      if (server == null || CompareOps._isEqual(server, SessionUtils.asHandle()))
      {
         // this is a normal call, invoke immediately
         invokeExternalProcedure(name, true, persistHandle, false, null, modes, args);
      }
      else
      {
         // validate server first
         if (!validServer(server, null, name))
         {
            return;
         }

         ServerImpl s = (ServerImpl) server.get();

         if (s.isWebService())
         {
            if (portHandle == null)
            {
               final String err =
                  "The PERSISTENT keyword is invalid when used with a SERVER handle whose " +
                  "SUBTYPE is 'WEBSERVICE'";
               ErrorManager.recordOrThrowError(11481, err, false);
               return;
            }
            
            if (asyncReq != null)
            {
               final String err =
                  "Unexpected Web service client error. Copy the following message and contact " +
                  "technical support. Web service procedure not initialized in cwsSend";
               ErrorManager.recordOrThrowError(11461, err, false);
               return;
            }

            // this will establish a port connection
            WebServiceHelper helper = (WebServiceHelper) s.getServerHelper();
            PersistentProcedure portProcedure = helper.runPort(name);
            if (portProcedure != null)
            {
               // the port was ran, assign it
               portHandle.assign(portProcedure);
               
               // this is not a persistent procedure, and does not get registered with the 
               // ProcedureManager
            }
            
            return;
         }
         else
         {
            if (portHandle != null)
            {
               final String err = 
                  "'RUN ... ON SERVER server-handle' with the [SET handle] option is not valid " +
                  "without the PERSISTENT keyword unless the server is a Web service";
               ErrorManager.recordOrThrowError(14587, err, false);
               return;
            }

            // if sync mode, no async requests must be running
            if (asyncReq == null && s.hasAsyncRequests())
            {
               return;
            }

            ProxyProcedureWrapper proxyProcedure = null;
            handle h = persistHandle;

            AppServerHelper appServer = (AppServerHelper) s.getServerHelper();
            try
            {
               if (h == null)
               {
                  h = new handle();
               }
               character result = appServer.invokePersistent(0, asyncReq, name, h,
                                                             transactionDistinct,
                                                             modes, args);
               if (asyncReq == null)
               {
                  setReturnValue(result);
               }
               
               proxyProcedure = (ProxyProcedureWrapper) h.getResource();
               if (persistHandle != null)
               {
                  persistHandle.set(proxyProcedure);
               }
               proxyProcedure.setServer(s);
            }
            catch (ConditionException e)
            {
               if (e.getCause() instanceof NumberedException)
               {
                  ErrorManager.recordOrThrowError((NumberedException) e.getCause());
               }
               else
               {
                  throw e;
               }
            }

            // chain the procedure
            handle last = s.lastProcedure();
            if (last.isUnknown())
            {
               s.setFirstProcedure(proxyProcedure);
               s.setLastProcedure(proxyProcedure);
            }
            else
            {
               ProxyProcedureWrapper lr = (ProxyProcedureWrapper) last.getResource();
               lr.setNextSibling(proxyProcedure);
               proxyProcedure.setPrevSibling(lr);
               s.setLastProcedure(proxyProcedure);
            }
         }
      }
   }
   
   /**
    * Check if the caller can validate the arguments. If not, show an appropriate error message
    * and throw error condition.
    * <p>
    * Note that the content of the {@code args} array can be changed. An item will be changed if
    * any of following occurs:
    * <ul>
    *    <li>variables are replaced in O/U modes with local variables
    *    <li>handle access discrepancies
    *    <li>types of parameters do not match but they are compatible
    * </ul>
    * 
    * @param    caller
    *           The caller used to validate the arguments.
    * @param    function
    *           <code>true</code> if this is a function call.
    * @param    modes
    *           A string representation of the modes of each parameter.
    * @param    args
    *           The argument list.
    *           
    * @return   <code>true</code> if the arguments are valid.
    */
   private static boolean validArguments(String              name,
                                         InternalEntryCaller caller,
                                         boolean             function,
                                         String              modes,
                                         Object...           args)
   {
      int[] extentInfo = new int[2]; // [0] = argument extent, [1] = parameter (expected) extent
      ArgValidationErrors argsValid = caller.valid(function, modes, extentInfo, args); 
      
      // deferred validation needs no further processing here
      if (argsValid == ArgValidationErrors.DEFER_VALIDATION)
      {
         return true;
      }
      
      if (caller.iename != null)
      {
         // we have an internal entry, determine the relative name used to start this referent
         name = ProcedureManager.getRelativeName(caller.callerInstance);
      }
      
      if (argsValid == ArgValidationErrors.NO_ERROR)
      {
         // validate the parameter's modes
         if (modes != null)
         {
            String definedModes = caller.getParameterModes(function);
            if (!modes.equals(definedModes))
            {
               final String errp = 
                  "Mismatched parameters types passed to procedure %s %s";
               final String errf =
                  "Mismatched parameter INPUT/OUTPUT mode passed to routine %s %s";
               
               String err = function ? errf : errp;
               int errno = function ? 6349 : 3230;
               String msg = String.format(err, caller.getInternalEntryName(), name);
               ErrorManager.recordOrThrowError(errno, msg, false);
               return false;
            }
            
            // check is simple variables on OUTPUT and INPUT-OUTPUT mode. These modes requires an
            // intermediary variable to be used as actual value and to assign back on routine
            // return. If the variable is already an intermediary one (for a wrapper field), skip 
            // this step.
            for (int k = 0; k < modes.length(); k++)
            {
               char mode = modes.charAt(k);
               if (mode != 'I' &&                     // INPUT-OUTPUT or OUTPUT modes
                   args[k] instanceof BaseDataType)   // only variable or wrapped fields
               {
                  BaseDataType arg = (BaseDataType) args[k];
                  if (!AbstractParameter.isParameter(arg))
                  {
                     BaseDataType intParam = (mode == 'U') ? 
                           arg.duplicate() :         // INPUT-OUTPUT mode, duplicate value  
                           arg.instantiateDefault(); // OUTPUT mode, inited in routine, eventually
                     new VariableAssigner(intParam, arg);
                     
                     args[k] = intParam;
                  }
               }
               // otherwise mode is I
            }
         }
         
         return true;
      }
      
      String[] errs = null;
      int[] errIds = null;
      
      String callingProcedure = SessionUtils._isRemote() 
                                   ? "(Remote client)" 
                                   : ProcedureManager.getStackEntry(0);
      
      switch (argsValid)
      {
         case UNEXPECTED_PARAM:
            errIds = new int[] { 1005 };
            String msg = "Procedure %s passed parameters to %s, which did not expect any";
            msg = String.format(msg, callingProcedure, name);
            errs = new String[] { msg };
            break;
         case INCORRECT_COUNT:
            errIds = new int[] { 3234 };
            msg = "Mismatched number of parameters passed to routine %s %s";
            msg = String.format(msg, caller.getInternalEntryName(), name);
            errs = new String[] { msg };
            break;
         case INCORRECT_TYPES:
            errIds = new int[] { 5729, 2570 };
            msg = "Routine %s sent called routine %s %s mismatched parameters";
            msg = String.format(msg,
                                callingProcedure,
                                caller.getInternalEntryName(),
                                name);
            
            errs = new String[]
            {
               "Incompatible datatypes found during runtime conversion",
               msg
            };
            break;
         case ARRAY_FOR_NON_ARRAY:
         case NON_ARRAY_FOR_ARRAY:
         case BAD_ARRAY_DIMENSION:
            errIds = new int[] { 11428 };
            // TODO: msg doesn't end in DOT in 4GL
            msg = "Extent parameter dimension of %d from procedure %s is mismatched with " +
                  "sub-procedure %s %s parameter dimension of %d ...(328)";
            msg = String.format(msg,
                                extentInfo[0],
                                callingProcedure,
                                caller.getInternalEntryName(),
                                name,
                                extentInfo[1]);
            errs = new String[] { msg };
            break;
         case DYN_ARRAY_FOR_NON_ARRAY:
            errIds = new int[] { 12290 };
            msg = "Calling procedure %s cannot pass an indeterminate extent parameter if the " +
                  "sub-procedure %s %s parameter is a non-array parameter";
            msg = String.format(msg,
                                callingProcedure,
                                caller.getInternalEntryName(),
                                name);
            errs = new String[] { msg };
            break;
         case NON_ARRAY_FOR_DYN_ARRAY:
            errIds = new int[] { 15472 };
            msg = 
               "Routine %s passed a non-array parameter to an array parameter of routine %s %s";
            msg = String.format(msg,
                                callingProcedure,
                                caller.getInternalEntryName(),
                                name);
            errs = new String[] { msg };
            break;
      }
      
      ErrorManager.recordOrThrowError(errIds, errs, false);
      
      return false;
   }
   
   /**
    * Check if the handle is valid for function or procedure invocation.
    * If is invalid, show an appropriate message.
    * 
    * @param    h
    *           The handle to which the procedure belongs. If <code>null</code>
    *           defaults to THIS-PROCEDURE.
    * @param    name
    *           The legacy 4GL name for the function.
    * @param    explicit
    *           <code>true</code> if the passed handle was explicitly set (for
    *           RUN ... IN or FUNCTION ... IN handle cases.
    * @param    function
    *           <code>true</code> if this is a function call.
    * @param    dynamicFunction
    *           <code>true</code> if this is a DYNAMIC-FUNCTION call.
    *           
    * @return   <code>true</code> if the handle is valid for function or
    *           procedure invocation.
    */
   private static boolean validHandle(handle  h,
                                      String  name,
                                      boolean explicit,
                                      boolean function,
                                      boolean dynamicFunction)
   {
      if (function)
      {
         boolean isProcedure = ProcedureManager.isProcedure(h);
         if (h.isUnknown() || !isProcedure)
         {
            String err = null;
            int num = -1;

            if (!h.isUnknown() && !isProcedure)
            {
               final String msg =
                  "Could not locate external function '%s'. The handle to " +
                  "the procedure context is invalid";
               err = String.format(msg, name);
               num = 2777;
            }
            else if (explicit && dynamicFunction)
            {
               err = "DYNAMIC-FUNCTION function given invalid or Unknown context to run IN";
               num = 5705;
            }
            else
            {
               final String msg =
                  "Could not evaluate the expression describing the " +
                  "context of external function '%s'";
               err = String.format(msg, name);
               num = 2767;
            }

            ErrorManager.recordOrThrowError(num, err, false);
            return false;
         }
      }
      else
      {
         if (explicit && h.isUnknown())
         {
            String err = null;
            int num = -1;
            final String msg =
               "Could not evaluate procedure handle expression while trying to execute %s";
            err = String.format(msg, name);
            num = 2125;
             
            ErrorManager.recordOrThrowError(num, err, false);
            return false;
         }
         if (!ProcedureManager.hasReferent(h.get()))
         {
            String err = null;
            int num = -1;
            final String msg = "Invalid or inappropriate handle value given to RUN...IN " +
                               "statement. Procedure '%s':%d";
            num = 2128;
            
            handle pthis = ProcedureManager.thisProcedure();
            String thisName = ProcedureManager.getRelativeName(pthis.get());
            err = String.format(msg, thisName, 0);
            
            ErrorManager.recordOrThrowError(num, err, false);
            return false;
         }
      }
      
      return true;
   }
   
   /**
    * Check if the given handle is a valid, connected appserver.
    * 
    * @param    hServer
    *           The handle to validate as an appserver.
    * @param    hProc
    *           A procedure handle for IN handle cases.
    * @param    name
    *           The target 4GL legacy name.
    *           
    * @return   <code>true</code> if the given handle is valid.
    */
   private static boolean validServer(handle hServer, handle hProc, character name)
   {
      final String err2 =
         "Invalid or inappropriate server handle specified for RUN %s ... ON SERVER statement";
      final int code2 = 5453;

      if (hServer.isUnknown())
      {
         ErrorManager.recordOrThrowError(code2, err2, false);
         return false;
      }
      
      Object resource = hServer.get();
      if (!(resource instanceof ServerImpl))
      {
         final String err =
            "The handle for \"RUN ... ON SERVER\" must be either a SERVER handle or " +
            "a SESSION handle";
         ErrorManager.recordOrThrowError(8998, err, false);

         return false;
      }
      
      ServerImpl server = (ServerImpl) resource;
      if (!server._connected())
      {
         final String err1 = "SERVER  is not connected";
         final int code1 = 5451;

         int[] codes = null;
         String[] msgs = null;
         
         if (hProc != null && hProc._isValid() &&
             (ProcedureManager.isProxy(hProc) || hProc.get() instanceof PortTypeWrapper))
         {
            codes = new int[] { code1 };
            msgs = new String[] { err1 };
         }
         else
         {
            codes = new int[] { code1, code2 };
            msgs = new String[] { err1, String.format(err2, name.toStringMessage()) };
         }
         
         ErrorManager.recordOrThrowError(codes, msgs, false);
         
         return false;
      }
      
      return true;
   }
   
   /**
    * Common helper to display a standard error message and then throw a
    * STOP condition when the target procedure cannot be found. Silent
    * error mode has no effect on this since this is not an ERROR condition.
    * That means that this method never returns normally AND that the
    * ERROR-STATUS data is unchanged.
    *
    * @param    target    
    *           The procedure that cannot be found.
    */
   private static void invokeFailure(String target)
   throws StopConditionException
   {
      String spec = "\"%s\" was not found";
      String err  = String.format(spec, target);
      
      ErrorManager.displayError(293, err);
      throw new StopConditionException(err, new NumberedException(err, 293));
   }   
   
   /**
    * Throw a {@link RuntimeException}, as the function or procedure 
    * invocation has raised an exception.
    * 
    * @param    caller
    *           {@link InternalEntryCaller} caller instance used to invoke the
    *           function or procedure. May be null. 
    * @param    name
    *           The legacy 4GL name passed to the function or procedure call
    *           statement.
    * @param    deferred
    *           The deferred error.
    */
   private static void invokeError(InternalEntryCaller caller, 
                                   String name,
                                   Exception deferred)
   {
      if (caller != null)
      {
         throw new RuntimeException("invoke() of class "   +
                                    caller.getClassName()  +
                                    " and  method "        +
                                    caller.getMethodName() +
                                    " failed",
                                    deferred);
      }
      else
      {
         throw new RuntimeException("invoke() of program "   +
                                    name + " failed", deferred);
      }
   }
   
   /**
    * Override this native call to a FWD API call.
    * 
    * @param    libname
    *           The library name where this native procedure should be found.
    * @param    pname
    *           The legacy name of the external procedure from which this API is being called.
    * @param    iename
    *           The internal-entry name being called which is usually the same as the entry
    *           point name in the library except where an ordinal is being used.
    *           
    * @return   A non-null {@link InternalEntryCaller} instance, if this call can be overridden.
    */
   private static InternalEntryCaller overrideNativeCall(String libname, 
                                                         String pname, 
                                                         String iename)
   {
      // overrides lockWindowUpdate user32.dll calls to LogicalTerminal.disableRedraw
      
      if (!iename.equalsIgnoreCase("lockWindowUpdate") ||
          libname == null                              ||
          libname.toLowerCase().indexOf("user32") < 0)
      {
         return null;
      }
      
      InternalEntry ie = SourceNameMapper.getInternalEntry(pname, iename, false);
      boolean missingReturn = ie.getParameterListSize() == 1;
      if (!(ie instanceof NativeAPIEntry) ||
          ie.isFunction()                 ||
          (ie.getParameterListSize() != 2 && !missingReturn))
      {
         return null;
      }
   
      NativeAPIEntry calldef = (NativeAPIEntry) ie;
      String pmodes;
      
      Parameter pret = null;
      Parameter phwnd = null;
      
      if (missingReturn)
      {
         Parameter p0 = calldef.getParameter(0);

         if (p0.getMode().equalsIgnoreCase("INPUT"))
         {
            pmodes = "I";
            pret = null;
            phwnd = p0;
         }
         else
         {
            return null;
         }
      }
      else
      {
         Parameter p0 = calldef.getParameter(0);
         Parameter p1 = calldef.getParameter(1);
         if (p0.getMode().equalsIgnoreCase("INPUT") && p1.getMode().equalsIgnoreCase("RETURN"))
         {
            pmodes = "IO";
            pret = p1;
            phwnd = p0;
         }
         else if (p1.getMode().equalsIgnoreCase("INPUT") && p0.getMode().equalsIgnoreCase("RETURN"))
         {
            pmodes = "OI";
            pret = p0;
            phwnd = p1;
         }
         else
         {
            return null;
         }
      }
      
      // HWND data types are LONG, UNSIGNED-LONG or INT64
      switch (phwnd.getType())
      {
         case "LONG":
         case "UNSIGNED-LONG":
         case "INT64":
            break;
         default:
            return null;
      }
      // RETURN data types are BYTE, SHORT, UNSIGNED-SHORT, LONG, UNSIGNED-LONG or INT64
      if (!missingReturn)
      {
         switch (pret.getType())
         {
            case "BYTE":
            case "SHORT":
            case "UNSIGNED-SHORT":
            case "LONG":
            case "UNSIGNED-LONG":
            case "INT64":
               break;
            default:
               return null;
         }
      }
      
      return new NativeProcedureCaller(libname, pname, iename)
      {
         @Override
         Object invokeImpl(String modes, Object... param) throws Exception
         {
            if (missingReturn)
            {
               if (param.length == 1                       && 
                   (modes == null || pmodes.equals(modes)) && 
                   int64.class.isAssignableFrom(param[0].getClass()))
               {
                  int64 hwnd = (int64) param[0];
                  
                  LogicalTerminal.disableRedraw(new int64(hwnd));
                  return null;
               }
            }
            else
            {
               if (param.length == 2                                 && 
                   int64.class.isAssignableFrom(param[0].getClass()) &&
                   int64.class.isAssignableFrom(param[1].getClass()))
               {
                  int64 hwnd = null;
                  int64 ret = null;
                  
                  if (modes == null)
                  {
                     modes = pmodes;
                  }
                  
                  if (modes.equals("IO"))
                  {
                     hwnd = (int64) param[0];
                     ret = (int64) param[1];
                  }
                  else if (modes.equals("OI"))
                  {
                     ret = (int64) param[0];
                     hwnd = (int64) param[1];
                  }
                  
                  if (hwnd != null && ret != null)
                  {
                     LogicalTerminal.disableRedraw(new int64(hwnd), ret);
                     return null;
                  }
               }
            }

            ErrorManager.recordOrThrowError(0, "Could not invoke DISABLE-REDRAW!");

            return null;
         }
      };
   }

   /** 
    * Stores global data relating to the state of the current context.
    */
   private static class WorkArea
   {  
      /** Stores the most recent RETURN-VALUE for this user's session. */
      private character returnValue = new character("");

      /** 
       * List of {@link Resolver} instances used for plain RUN statements.
       */
      private final List<Resolver> resolvers;

      /**
       * List of {@link Resolver} instances used for RUN IN handle,
       * DYNAMIC-FUNCTION(... IN handle) or FUNCTION ... IN handle cases.
       */
      private final List<Resolver> inHandleResolvers;
      
      /**
       * {@link InternalResolver} instance which searches the internal-entries
       * for a match.
       */
      private final Resolver internalResolver;
      
      /**
       * {@link InternalProcSuperResolver} instance which searches the 
       * procedure's super-procedures for a match.
       */
      private final Resolver internalProcSuperResolver;
      
      /**
       * {@link SessionSuperResolver} instance which searches the sessions's 
       * super-procedures for a match.
       */
      private final Resolver sessionSuperResolver;
      
      /**
       * {@link ExternalProgramResolver} instance which searches the external
       * program list for a match.
       */
      private final Resolver externalResolver;
      
      /**
       * Flag indicating if this call is an invocation from a remote, customer-specific 
       * application.  When this flag is on, only external programs defined via the 
       * {@link RemoteEntryPointResource} ACLs can be invoked.  This flag needs to be reset BEFORE
       * the remote API is invoked.
       */
      private boolean remoteInvoke = false;

      private WorkArea()
      {
         internalResolver = new InternalResolver();
         internalProcSuperResolver = new InternalProcSuperResolver();
         sessionSuperResolver = new SessionSuperResolver();
         externalResolver = new ExternalProgramResolver();
         
         resolvers = new ArrayList<>();
         resolvers.add(internalResolver);
         resolvers.add(internalProcSuperResolver);
         resolvers.add(sessionSuperResolver);
         resolvers.add(externalResolver);

         inHandleResolvers = new ArrayList<>();
         inHandleResolvers.add(internalResolver);
         inHandleResolvers.add(internalProcSuperResolver);
         inHandleResolvers.add(sessionSuperResolver);
      }
   }
   
   /**
    * Simple container that stores and returns a context-local instance of
    * the global work area.
    */
   private static class ContextContainer
   extends ContextLocal<WorkArea>
   {
      /**
       * Obtains the context-local instance of the contained global work
       * area.
       *
       * @return   The work area associated with this context.
       */
      public WorkArea obtain() 
      {
         return this.get();
      }   
      
      /**
       * Initializes the work area, the first time it is requested within a
       * new context.
       *
       * @return   The newly instantiated work area.
       */
      protected synchronized WorkArea initialValue()
      {
         return new WorkArea();
      }   
   }

   /**
    * Invoke the configured internal entry.
    */
   private static class InternalEntryCaller
   {
      /** Cache of methods by type, name, class, and signature */
      private static final Cache<CacheKey, CacheValue> cache = new LFUAgingCache<>(8192);
      
      /** The external program instance where the internal entry's body is defined. */
      protected Object callerInstance;
      
      /** The java name of this internal entry. */
      protected String methodName;
      
      /** The legacy 4GL name of this internal entry. */
      protected String iename;
      
      /** The {@link Method} instance for this internal entry. */
      private Method mthd = null;
      
      /** The handle associated with the {@link #callerInstance}. */
      private handle phandle;
      
      /** Worker which performs the {@link ProcedureManager#pushCalleeInfo} call. */
      private Runnable pushWorker;
      
      /**
       * Basic c'tor.
       */
      public InternalEntryCaller()
      {
      }
      
      /**
       * Configure this instance to invoke the given internal entry.
       * 
       * @param    phandle
       *           The procedure handle where the internal entry's body is
       *           defined.
       * @param    methodName
       *           The resolved java name for this internal entry.
       * @param    iename
       *           The legacy 4GL name of this internal entry.
       */
      public InternalEntryCaller(handle phandle, String methodName, String iename)
      {
         this.phandle = new handle(phandle);
         this.callerInstance = phandle.get();
         this.methodName = methodName;
         this.iename = iename;
      }
      
      /**
       * Set the {@link #pushWorker}.
       * 
       * @param    push
       *           The code which performs the {@link ProcedureManager#pushCalleeInfo} call.
       */
      void setPushWorker(Runnable push)
      {
         this.pushWorker = push;
      }
      
      /**
       * Invoke the internal entry, using the {@link #mthd}, 
       * the {@link #callerInstance} and the passed arguments.
       * <p>
       * It performs an additional check, in case of remote invocation calls originating from 
       * external, customer-specific application (when {@link WorkArea#remoteInvoke} flag is set):
       * access to the external program must be allowed via a {@link RemoteEntryPointResource} ACL.
       * 
       * @param    modes
       *           An encoded string with the mode of each parameter specified as a
       *           character in the string.
       * @param    param
       *           The arguments for this Java method call.
       *           
       * @return   The returned value
       * 
       * @throws   Exception
       *           If any exceptions are encountered during method invocation.
       */
      final Object invoke(String modes, Object... param)
      throws Exception
      {
         WorkArea wa = work.get();
         if (wa.remoteInvoke)
         {
            wa.remoteInvoke = false;

            String pname = getExternalProgramName();
            
            SecurityManager sm = SecurityManager.getInstance();

            RemoteEntryPointResource repr = 
               (RemoteEntryPointResource) sm.getPluginInstance("remoteentrypoint");

            if (repr == null || !repr.isAllowed(pname))
            {
               String msg;
               if (repr == null)
               {
                  msg = "Remote program '%s' cannot be accessed " +
                        "because the RemoteEntryPointResource is NOT enabled.";
                  msg = LogHelper.generate(msg, pname);
               }
               else
               {
                  msg = "Access to remote program '%s' not allowed for user %s.";
                  msg = LogHelper.generate(msg, pname, sm.getUserId());
               }
               
               if (LOG.isLoggable(Level.SEVERE))
               {
                  LOG.logp(Level.SEVERE,
                           "ControlFlowOps$InternalEntryCaller.invoke",
                           "",
                           msg);
               }

               throw new SilentUnwindException(new IllegalStateException(msg));
            }
         }
         
         return invokeImpl(modes, param);
      }
      
      /**
       * Invoke the internal entry, using the {@link #mthd}, 
       * the {@link #callerInstance} and the passed arguments.
       *
       * @param    modes
       *           An encoded string with the mode of each parameter specified as a
       *           character in the string.
       * @param    param
       *           The arguments for this Java method call.
       *           
       * @return   The returned value
       * 
       * @throws   Exception
       *           If any exceptions are encountered during method invocation.
       */
      Object invokeImpl(String modes, Object... param)
      throws Exception
      {
         try
         {
            pushWorker.run();
            BlockManager.markPendingDynCall();
            
            mthd.setAccessible(true);
            return mthd.invoke(callerInstance, param);
         }
         finally
         {
            ProcedureManager.popCalleeInfo();
         }
      }
      
      /**
       * Get the signature, as defined by the parameters.
       * 
       * @param    param
       *           The arguments for this Java method call.
       *           
       * @return   An array with the type of each parameter.
       */
      Class<?>[] getSignature(Object... param)
      {
         Class<?>[] signature = new Class[param.length];
         
         for (int i = 0; i < signature.length; i++)
         {
            Class<?> type;
            Object p = param[i];
            if (p instanceof BufferReference)
            {
               // the only provided support is for buffer objects.
               Class<?> dmoBufIfc = RecordBuffer.getDMOBufInterfaceForObject(p);
               
               if (dmoBufIfc == null)
               {
                  throw new RuntimeException("Dynamic proxy not supported for parameter " + p);
               }
               
               type = dmoBufIfc;
            }
            else
            {
               type = p == null ? Object.class : p.getClass();
            }
            signature[i] = type;
         }
         
         return signature;
      }
      
      /**
       * Validate the given parameters against the method's real signature using the corresponding
       * parameter passing mode:
       * <ul>
       *    <li><strong>INPUT</strong>If the parameter's type is compatible with the real type
       *          (from the {@link #mthd method's} arguments), then cast it to it.
       *    <li><strong>OUTPUT</strong>If the real type is compatible with the parameter's type
       *          (from the {@link #mthd method's} arguments), the use a default value for the
       *          expected type. The compatibility assures the conversion will be possible at
       *          return time.
       *    <li><strong>INPUT-OUTPUT</strong>Both conversion must be available for this parameter
       *          to be validated. A cast at this moment will assure INPUT compatibility and a
       *          second cast when returning will also be needed.
       * </ul>
       * INPUT
       *
       * If all are valid, then {@link ArgValidationErrors#NO_ERROR} is returned. Else:
       * <ul>
       * <li>if the argument count does not match, then 
       *     {@link ArgValidationErrors#INCORRECT_COUNT} is returned.</li>
       * <li>if the types of the arguments do not match, then 
       *     {@link ArgValidationErrors#INCORRECT_TYPES} is returned.</li>
       * </ul>
       * 
       * @param   function
       *          <code>true</code> if this is a function call.
       * @param   modes
       *          The types of parameters: INPUT(I) / OUTPUT (O) / INPUT-OUTPUT (U).
       * @param   extentInfo
       *          An output parameter array that provides the caller with additional information
       *          about the extent sizes when they don't match. The first item is the size of the
       *          argument (provided extent), the second is the parameter (expected extent).
       * @param   param
       *          The arguments for this Java method call.
       *           
       * @return   One of the {@link ArgValidationErrors} constants.
       */
      ArgValidationErrors valid(boolean function, String modes, int[] extentInfo, Object... param)
      {
         Class<?>[] signature = getSignature(param); // required signature
         Class<?> cls = callerInstance.getClass();
         
         ArgValidationErrors rv;
         Method method = null;
         
         // check the cache first; we not only cache successful lookups, but also remember failed
         // lookups, so we don't repeat them (it is quite expensive to fail to find a method, as
         // it involves two exceptions being thrown, a search through all methods of the class,
         // and a good bit of fuzzy matching logic)
         
         CacheKey key = new CacheKey(function, cls, methodName, signature);
         CacheValue value = getCache(key); // TODO: rename [value] with a more appropriate name 
         
         if (value != null)
         {
            // only return immediately if we had an exact signature match, or if we failed to
            // find a method at all previously for this set of criteria; for a fuzzy match, we
            // still have parameter type massaging work to do below
            if (value.exactMatch)
            {
               if (value.method != null)
               {
                  mthd = value.method;
               }
               
               return value.returnValue;
            }
         }
         else
         {
            try
            {
               method = cls.getMethod(methodName, signature);
               mthd = method;
               rv = ArgValidationErrors.NO_ERROR;
               
               putCache(key, method, rv, true);
               
               return rv;
            }
            catch (Exception e)
            {
               method = null;
            }
            
            // fallback - check if the method is declared, but private
            try
            {
               method = cls.getDeclaredMethod(methodName, signature);
               mthd = method;
               rv = ArgValidationErrors.NO_ERROR;
               
               putCache(key, method, rv, true);
               
               return rv;
            }
            catch (Exception e)
            {
               method = null;
            }
         }
         
         // TODO: add support for the longchar-to-char narrowing and char-to-longchar widening
         
         final Class<?> AEP = AbstractExtentParameter.class;
         final String extProgName = ProcedureManager.getAbsoluteName(callerInstance);
         
         // go through all the methods (or the cached one only), and do some fuzzy parameter type
         // matching
         Method[] methods = (value != null)
                            ? new Method[] { value.method }
                            : cls.getDeclaredMethods();
         boolean correctNo = false;
         boolean hasParams = false;
         for (Method m : methods)
         {
            if (!m.getName().equals(methodName))
            {
               // if the name of the method does not match skip to next method
               continue;
            }
            Class<?>[] mtypes = m.getParameterTypes();
            hasParams = (mtypes.length > 0);
            
            if (mtypes.length != param.length)
            {
               // if parameter count does not match skip this method
               continue;
            }
            if (modes == null)
            {
               // if the call comes from ControlFlowOps.runSuper(), the modes are unknown
               modes = SourceNameMapper.getParameterModes(extProgName, iename, function);
            }
            
            correctNo = true;
            
            // try to cast each parameter to the expected type. If it works, then use this method
            boolean ok = true;
            Object[] newPars = new Object[param.length];
            for (int i = 0; i < param.length; i++)
            {
               newPars[i] = param[i];
               Class<?> pcls = param[i].getClass();
               Class<?> argType = signature[i];
               Class<?> expectedType = mtypes[i];
               
               // check if passing a non-array to an array or an array to non-array.
               // do not check arrays containing non-BDT components.
               if (value == null &&
                   ((argType.isArray() != expectedType.isArray() && 
                     (!argType.isArray() || 
                      BaseDataType.class.isAssignableFrom(argType.getComponentType()))) ||
                    AEP.isAssignableFrom(argType) != AEP.isAssignableFrom(expectedType)))
               {
                  boolean parExtent = argType.isArray() || AEP.isAssignableFrom(argType);
                  
                  int expectedExtent = SourceNameMapper.getExtentForParam(extProgName,
                                                                          iename,
                                                                          function,
                                                                          i);
                  int paramExtent = 0;
                  if (parExtent)
                  {
                     BaseDataType[] pvar = null;
                     if (AEP.isAssignableFrom(argType))
                     {
                        pvar = ((AbstractExtentParameter) newPars[i]).getVariable();
                     }
                     else
                     {
                        pvar = (BaseDataType[]) newPars[i];
                     }
                     
                     paramExtent = pvar.length;
                  }
                  
                  if (expectedExtent == 0)
                  {
                     // expected param is dynamic-extent
                     rv = parExtent
                          ? ArgValidationErrors.DYN_ARRAY_FOR_NON_ARRAY 
                          : ArgValidationErrors.NON_ARRAY_FOR_DYN_ARRAY;
                  }
                  else
                  {
                     // expected param is fixed-extent
                     rv = parExtent
                          ? ArgValidationErrors.ARRAY_FOR_NON_ARRAY
                          : ArgValidationErrors.NON_ARRAY_FOR_ARRAY;
                  }
                  
                  extentInfo[1] = parExtent ? 0 : expectedExtent;
                  extentInfo[0] = parExtent ? paramExtent : 0;
                  
                  // cache failed result as well, so we don't need to repeat validation
                  putCache(key, null, rv, true);
                  
                  return rv;
               }
               
               // provide compatibility between resources and handles
               if (WrappedResource.class.isAssignableFrom(argType) && 
                   handle.class.equals(expectedType))
               {
                  newPars[i] = new handle((WrappedResource) param[i]);
                  // TODO: check the OUTPUT mode case
                  continue;
               }
               
               if (!expectedType.isAssignableFrom(argType))
               {
                  try
                  {
                     // if the param's type is unknown class, construct the unknown for the
                     // expectedType
                     if (unknown.class.equals(argType))
                     {
                        // for some data types (date, datetime, datetime-tz, raw) the 
                        // unknown value is not created using default constructor
                        newPars[i] = BaseDataType.generateUnknown(expectedType);
                        // TODO: check the OUTPUT mode case
                     }
                     else
                     {
                        Class<?> toCompType;
                        Class<?> fromCompType;
                        boolean isArray = expectedType.isArray();
                        if (isArray)
                        {
                           // take into consideration the elements' type 
                           toCompType = expectedType.getComponentType();
                           fromCompType = pcls.getComponentType();
                        }
                        else 
                        {
                           toCompType = expectedType;
                           fromCompType = pcls;
                        }
                        Constructor<?> fwCtor;
                        Constructor<?> bwCtor = null;
                        // checking the parameter passing type: IN, OUT, IN-OUT
                        switch (modes.charAt(i))
                        {
                           case 'I':
                              try
                              {
                                 fwCtor = toCompType.getConstructor(fromCompType);
                              }
                              catch (NoSuchMethodException e)
                              {
                                 // on failure try the 'generic' conversion constructor
                                 fwCtor = toCompType.getConstructor(BaseDataType.class);
                              }
                              if (isArray)
                              {
                                 // copy the newPars[i] array through fwCtor morphism
                                 Object[] arrayArg = (Object[]) newPars[i];
                                 Object[] arrayParam = (Object[]) 
                                    Array.newInstance(toCompType, arrayArg.length);
                                 for (int k = 0; k < arrayArg.length; k++)
                                 {
                                    arrayParam[k] = fwCtor.newInstance(arrayArg[i]);
                                 }
                                 newPars[i] = arrayParam;
                              }
                              else
                              {
                                 newPars[i] = fwCtor.newInstance(param[i]);
                              }
                              break;
                           case 'O':
                              // check existence of the return cast:
                              bwCtor = fromCompType.getConstructor(toCompType);
                              // the output parameters aren't / don't need to be initiated
                              newPars[i] = BaseDataType.generateDefault(toCompType);
                              break;
                           case 'U':
                              // check existence of the return cast:
                              bwCtor = fromCompType.getConstructor(toCompType);
                              // but use as standard INPUT for now
                              fwCtor = toCompType.getConstructor(fromCompType);
                              newPars[i] = fwCtor.newInstance(param[i]);
                           break;
                        }
                        if (bwCtor != null)
                        {
                           // in the event the OUTPUT value must be converted back:
                           FieldAssigner.update(((BaseDataType) param[i]),
                                                ((BaseDataType) newPars[i]),
                                                bwCtor);
                        }
                     }
                     if (ErrorManager.isPending())
                     {
                        // if any errors occurred during parameter matching
                        // (eg.: INTEGER overflow) procedure cannot be called
                        ok = false;
                        break;
                     }
                  }
                  catch (Throwable t)
                  {
                     // was not able to convert this param
                     ok = false;
                     
                     // check if a condition was raised. if so, do not
                     // search for any other c'tors
                     boolean raisedCondition = false;
                     while (t != null)
                     {
                        if (t instanceof ConditionException)
                        {
                           raisedCondition = true;
                        }
                        t = t.getCause();
                     }
                     
                     if (raisedCondition)
                     {
                        break;
                     }
                  }
                  
                  // fallback, maybe there is a c'tor with a compatible type
                  if (!ok)
                  {
                     Constructor<?>[] ctors;
                     Constructor<?> bwCtor = null;
                     // checking the parameter passing type: IN, OUT, IN-OUT
                     switch (modes.charAt(i))
                     {
                        case 'I':
                           ctors = expectedType.getConstructors();
                           for (Constructor<?> ctor : ctors)
                           {
                              Class<?>[] ctypes = ctor.getParameterTypes();
                              if (ctypes.length == 1 && ctypes[0].isAssignableFrom(pcls))
                              {
                                 try
                                 {
                                    newPars[i] = ctor.newInstance(param[i]);
                                    ok = true;// found my c'tor
                                    break;
                                 }
                                 catch (Exception e)
                                 {
                                    // ignore, ok stays false
                                 }
                              }
                           }
                           break;
                        
                        case 'O':
                           // check existence of the return cast:
                           ctors = pcls.getConstructors();
                           for (Constructor<?> ctor : ctors)
                           {
                              Class<?>[] ctypes = ctor.getParameterTypes();
                              if (ctypes.length == 1 && 
                                  ctypes[0].isAssignableFrom(expectedType))
                              {
                                 try
                                 {
                                    bwCtor = ctor;
                                    newPars[i] = BaseDataType.generateDefault(expectedType);
                                    ok = true; // found my return cast c'tor
                                    break; // for each c'tor
                                 }
                                 catch (Exception e)
                                 {
                                    // ignore, ok stays false
                                 }
                              }
                           }
                           break;
                        
                        case 'U':
                           // check existence of the return cast:
                           ctors = pcls.getConstructors();
                           for (Constructor<?> ctor : ctors)
                           {
                              Class<?>[] ctypes = ctor.getParameterTypes();
                              if (ctypes.length == 1 &&
                                     ctypes[0].isAssignableFrom(expectedType))
                              {
                                 try
                                 {
                                    bwCtor = ctor;
                                    ok = true; // found my return cast c'tor
                                    break; // for each c'tor
                                 }
                                 catch (Exception e)
                                 {
                                    // ignore, ok stays false
                                 }
                              }
                           }
                           // but use as standard INPUT for now, if it exist
                           if (ok)
                           {
                              ok = false;
                              ctors = expectedType.getConstructors();
                              for (Constructor<?> ctor : ctors)
                              {
                                 Class<?>[] ctypes = ctor.getParameterTypes();
                                 if (ctypes.length == 1 && ctypes[0].isAssignableFrom(pcls))
                                 {
                                    try
                                    {
                                       newPars[i] = ctor.newInstance(param[i]);
                                       ok = true;// found my c'tor
                                       break;
                                    }
                                    catch (Exception e)
                                    {
                                       // ignore, ok stays false
                                    }
                                 }
                              }
                           }
                           if (!ok)
                           {
                              // if forward conversion fails, bwCtor is useless 
                              bwCtor = null;
                           }
                           break;
                     } // ~switch (mode)
                     if (bwCtor != null)
                     {
                        // in the event the OUTPUT value must be converted back:
                        FieldAssigner.update(((BaseDataType) param[i]),
                                             ((BaseDataType) newPars[i]),
                                             bwCtor);
                     }
                  }
                  
                  // if still wasn't found, then exit.
                  if (!ok)
                  {
                     break;
                  }
               }
               else if (AEP.isAssignableFrom(expectedType) &&
                        AEP.isAssignableFrom(argType))
               {
                  AbstractExtentParameter aep = (AbstractExtentParameter) newPars[i];
                  int argExt = aep.getVariable().length;
                  int paramExt = SourceNameMapper.getExtentForParam(
                                                           extProgName, iename, function, i);
                  
                  if (paramExt > 0 && paramExt != argExt)
                  {
                     // if paramExt is 0 it is dynamic extent so any extent is acceptable
                     extentInfo[0] = argExt;
                     extentInfo[1] = paramExt; 
                     return ArgValidationErrors.BAD_ARRAY_DIMENSION; // cache it?
                  }
                  
                  Class<? extends BaseDataType> argTypeE =
                     ((AbstractExtentParameter) param[i]).getVariable()[0].getClass();
                  Class<?> paramTypeE = 
                     SourceNameMapper.getTypeForParam(extProgName, iename, function, i);
                  
                  if (argTypeE != paramTypeE)
                  {
                     Constructor<?> fwCtor = null;
                     Constructor<?> bwCtor = null;
                     try
                     {
                        // checking the parameter passing type: IN, OUT, IN-OUT
                        switch (modes.charAt(i))
                        {
                           case 'I':
                              // this should not be normally reachable, 
                              // here we handle only wrapped EXTENT fields/variable 
                              throw new RuntimeException("Invalid execution path");
                           case 'U':
                              try
                              {
                                 fwCtor = paramTypeE.getConstructor(argTypeE);
                              }
                              catch (NoSuchMethodException e)
                              {
                                 // on failure try the 'generic' conversion constructor
                                 fwCtor = paramTypeE.getConstructor(BaseDataType.class);
                              }
                              // no break, fall through!
                           case 'O':
                              // check existence of the return cast:
                              try
                              {
                                 bwCtor = argTypeE.getConstructor(paramTypeE);
                              }
                              catch (NoSuchMethodException e)
                              {
                                 // on failure try the 'generic' conversion constructor
                                 bwCtor = argTypeE.getConstructor(BaseDataType.class);
                              }
                              aep.setReturnConversion(argTypeE, paramTypeE, fwCtor, bwCtor);
                              break;
                        }
                     }
                     catch (Throwable t)
                     {
                        // was not able to convert this param
                        ok = false;
                        
                        // check if a condition was raised. if so, do not search for any
                        // other c'tors
                        boolean raisedCondition = false;
                        while (t != null)
                        {
                           if (t instanceof ConditionException)
                           {
                              raisedCondition = true;
                           }
                           t = t.getCause();
                        }
                       
                        if (raisedCondition)
                        {
                           break;
                        }
                     }
                  }
                  // otherwise should be no problem, the same class of array elements
               }
            }
            
            if (ok)
            {
               // was able to cast the parameters to expected types
               System.arraycopy(newPars, 0, param, 0, param.length);
               mthd = m;
               rv = ArgValidationErrors.NO_ERROR;
               
               if (value == null)
               {
                  putCache(key, m, rv, false);
               }
               
               return rv;
            }
         } // ~ for each method
         
         // we failed fuzzy parameter matching; determine what type of error to report
         rv = !correctNo 
              ? (!hasParams ? ArgValidationErrors.UNEXPECTED_PARAM 
                            : ArgValidationErrors.INCORRECT_COUNT)
              : ArgValidationErrors.INCORRECT_TYPES;
         
         // cache failed result as well, so we don't need to repeat validation
         putCache(key, null, rv, true);
         
         return rv;
      }
      
      /**
       * Get the java method name for this internal entry.
       * 
       * @return   See above.
       */
      String getMethodName()
      {
         return methodName;
      }
      
      /**
       * Get the class name of the external program to which this internal
       * entry belongs.
       * 
       * @return   See above.
       */
      String getClassName()
      {
         return callerInstance.getClass().getName();
      }
      
      /**
       * Get the external program instance where this internal entry is 
       * defined.
       * 
       * @return   See above.
       */
      Object getCallerInstance()
      {
         return callerInstance;
      }
      
      /**
       * Get the legacy name of this internal entry.
       * 
       * @return   See above.
       */
      String getInternalEntryName()
      {
         return iename;
      }
      
      /**
       * Get the legacy name of the external program where the execution will take place.
       * 
       * @return   See above.
       */
      String getExternalProgramName()
      {
         return ProcedureManager.getAbsoluteName(callerInstance);
      }
      
      /**
       * Get the string representation of the parameter modes defined for this
       * internal entry.
       * 
       * @param    function
       *           <code>true</code> if this is a function call.
       * 
       * @return   See above.
       */
      String getParameterModes(boolean function)
      {
         String pname = ProcedureManager.getAbsoluteName(callerInstance);
         String iename = getInternalEntryName();
         return SourceNameMapper.getParameterModes(pname, iename, function);
      }
      
      /**
       * Get the parameters defined for this internal entry.
       * 
       * @param    function
       *           <code>true</code> if this is a function call.
       * 
       * @return   See above.
       */
      List<Parameter> getParameters(boolean function)
      {
         String pname = ProcedureManager.getAbsoluteName(callerInstance);
         String iename = getInternalEntryName();
         return SourceNameMapper.getParameters(pname, iename, function);
      }
      
      /**
       * Get the {@link #phandle} field.
       * 
       * @return   See above.
       */
      handle getPhandle()
      {
         return phandle;
      }
      
      /**
       * Get a cached value, if any, for the given key.
       * 
       * @param   key
       *          Cache key.
       * 
       * @return  Cached value, or <code>null</code>.
       */
      private CacheValue getCache(CacheKey key)
      {
         CacheValue value;
         
         synchronized (cache)
         {
            value = cache.get(key);
         }
         
         return value;
      }
      
      /**
       * Cache the specified information under the given cache key.
       * 
       * @param   key
       *          Cache key.
       * @param   method
       *          Method found matching a given name and signature, or {@code null} if no
       *          match was found.
       * @param   returnValue
       *          Enumerated return value for {@link #valid(boolean, String, int[], Object...)}
       *          method.
       * @param   exactMatch
       *          <code>true</code> to indicate the method signature stored in the key exactly
       *          matches that of the cached method; <code>false</code> to indicate one or more
       *          method arguments needs to be converted to a compatible type.
       */
      private void putCache(CacheKey key,
                            Method method,
                            ArgValidationErrors returnValue,
                            boolean exactMatch)
      {
         CacheValue value = new CacheValue(method, returnValue, exactMatch);
         
         synchronized (cache)
         {
            cache.put(key, value);
         }
      }
      
      /**
       * A cache key which encapsulates the name of a method, its signature and containing class,
       * and whether that method represents a converted Progress function or a procedure.
       */
      private static class CacheKey
      {
         /** <code>true</code> if method represents a Progress function */
         final boolean function;
         
         /** Method's enclosing class */
         final Class<?> callerClass;
         
         /** Name of the method */
         final String methodName;
         
         /** Method's signature types */
         final Class<?>[] signature;
         
         /**
          * Constructor.
          * 
          * @param   function
          *          <code>true</code> if method represents a Progress function.
          * @param   callerClass
          *          Method's enclosing class.
          * @param   methodName
          *          Name of the method.
          * @param   signature
          *          Method's signature types.
          */
         CacheKey(boolean function, Class<?> callerClass, String methodName, Class<?>[] signature)
         {
            this.function = function;
            this.callerClass = callerClass;
            this.methodName = methodName;
            this.signature = signature;
         }
         
         /**
          * Generate this object's hash code, based on its content, in a manner consistent with
          * {@link #equals(Object)}.
          * 
          * @return  Hash code.
          */
         @Override
         public int hashCode()
         {
            int result = 17;
            
            result = 37 * result + (function ? 0 : 1);
            result = 37 * result + callerClass.hashCode();
            result = 37 * result + methodName.hashCode();
            
            int len = signature.length;
            for (Class<?> aSignature : signature)
            {
               result = 37 * result + aSignature.hashCode();
            }
            
            return result;
         }
         
         /**
          * Indicate whether this object is equivalent with the given object, based on both of
          * their type and content, in a manner consistent with {@link #hashCode()}.
          * 
          * @return  <code>true</code> if this object is equivalent with the given object, else
          *          <code>false</code>.
          */
         @Override
         public boolean equals(Object o)
         {
            if (!(o instanceof CacheKey))
            {
               return false;
            }
            
            CacheKey that = (CacheKey) o;
            
            if (this.function != that.function)
            {
               return false;
            }
            
            if (!this.callerClass.equals(that.callerClass))
            {
               return false;
            }
            
            if (!this.methodName.equals(that.methodName))
            {
               return false;
            }
            
            int len = signature.length;
            if (len != that.signature.length)
            {
               return false;
            }
            
            for (int i = 0; i < len; i++)
            {
               if (!this.signature[i].equals(that.signature[i]))
               {
                  return false;
               }
            }
            
            return true;
         }
      }
      
      /**
       * A data structure for the cache's values.
       */
      private static class CacheValue
      {
         /** Method, if any, else <code>null</code> if none was found */
         final Method method;
         
         /** 
          * Enumerated return value for {@link 
          * InternalEntryCaller#valid(boolean, String, int[], Object...)}. 
          */
         final ArgValidationErrors returnValue;
         
         /** Flag indicating whether signature of caller's parameters exactly matches method */
         final boolean exactMatch;
         
         /**
          * Constructor.
          * 
          * @param   method
          *          Method, if any, else <code>null</code> if none was found.
          * @param   returnValue
          *          Enumerated return value for {@link
          *          InternalEntryCaller#valid(boolean, String, int[], Object...)}.
          * @param   exactMatch
          *          Flag indicating whether signature of caller's parameters exactly matches
          *          method.
          */
         CacheValue(Method method, ArgValidationErrors returnValue, boolean exactMatch)
         {
            this.method = method;
            this.returnValue = returnValue;
            this.exactMatch = exactMatch;
         }
      }
   }

   /**
    * This caller is returned when the FUNCTION ... MAP TO can not be resolved. It contains the 
    * function name from the last MAP TO clause.
    */
   private static class MapToNotFound
   extends InternalEntryCaller
   {
      public MapToNotFound(String iename)
      {
         this.iename = iename;
      }
   }
   
   /**
    * Implements native procedure invocation (an API call to a shared library).
    */
   private static class NativeProcedureCaller
   extends InternalEntryCaller
   {
      /** The library name where this native procedure belongs. */
      private String libname = null;
      
      /** The legacy name of this native procedure. */
      private String pname = null;
      
      /**
       * Configure this caller with the given data.
       * 
       * @param    libname
       *           The library name where this native procedure should be found.
       * @param    pname
       *           The legacy name of the external procedure from which this API is being called.
       * @param    iename
       *           The internal-entry name being called which is usually the same as the entry
       *           point name in the library except where an ordinal is being used.
       */
      public NativeProcedureCaller(String libname, String pname, String iename)
      {
         this.iename  = iename;
         this.libname = libname;
         this.pname   = pname;
      }
      
      /**
       * Call the native API with the arguments provided.
       *
       * @param    modes
       *           An encoded string with the mode of each parameter specified as a
       *           character in the string.
       * @param    args
       *           The array of procedure parameters.
       *
       * @return   Always <code>null</code> since procedures don't have a return value in the
       *           4GL.
       */
      @Override
      Object invokeImpl(String modes, Object... args) 
      throws Exception
      {
         boolean failed = false;
         OutputParameterAssigner opa = TransactionManager.deregisterOutputParameterAssigner();
         
         try
         {
            Class<?>[] signature = getSignature(args);
            NativeInvoker.invoke(this.libname, this.pname, this.iename, signature, modes, args);
            failed = ErrorManager.isPendingError();
         }
         catch (Exception exc)
         {
            failed = true;
            throw exc;
         }
         finally
         {
            if (opa != null)
            {
               if (failed)
               {
                  opa.abort();
               }
               else
               {
                  opa.processAssignments(pname, false);
               }
            }
         }
         return null;
      }
      
      /**
       * For native API calls, the validation is deferred to the actual time of the call since
       * the 4GL does other parameter checking before the more traditional type and mode
       * checking.  Rather than place that logic here, the decision was made to keep all of
       * that processing together in {@link NativeInvoker#invoke}.
       * 
       * @param   function
       *          <code>true</code> if this is a function call.
       * @param   modes
       *          The types of parameters: INPUT(I) / OUTPUT (O) / INPUT-OUTPUT (U).
       * @param   extentInfo
       *          An output parameter array that provides the caller with additional information
       *          about the extent sizes when they don't match. The first item is the size of the
       *          argument (provided extent), the second is the parameter (expected extent).
       * @param   param
       *          The arguments for this Java method call.
       *           
       * @return  Always returns {@link ArgValidationErrors#DEFER_VALIDATION}, which disables
       *          validation processing at this time.
       */
      @Override
      ArgValidationErrors valid(boolean function, String modes, int[] extentInfo, Object... param)
      {
         return ArgValidationErrors.DEFER_VALIDATION;
      }
      
      /**
       * Get the java method name for this internal entry.
       * 
       * @return   Always the same as the legacy native function name.  This is OK since
       *           this method is only used for error messages.
       */
      @Override
      String getMethodName()
      {
         return pname;
      }
      
      /**
       * Get the legacy name of the external program where the execution will take place.
       * 
       * @return   See above.
       */
      @Override
      String getExternalProgramName()
      {
         return pname;
      }
      
      /**
       * Get the class name of the external program to which this internal entry belongs.
       * 
       * @return   Always a nonsense class name.   This is OK since this method is only used
       *           for error messages.
       */
      @Override
      String getClassName()
      {
         return "NATIVE_LIBRARY_CALL";
      }
   }

   /**
    * Implements external program invocation.
    */
   private static class ExternalProcedureCaller
   extends InternalEntryCaller
   {
      /** Cache of classes for faster retrieval than {@code Class.forName} alone */
      private final static Map<String, Class<?>> classCache = new ConcurrentHashMap<>();
      
      /** The class name associated with the external program. */
      private String className;
      
      /**
       * Associate this caller with the specified external program instance.
       * 
       * @param    instance
       *           An instance of an external program.
       */
      public ExternalProcedureCaller(Object instance)
      {
         this.methodName = "execute";
         this.className = instance.getClass().getName();
         
         this.callerInstance = instance;
      }
      
      
      /**
       * Configure this caller with the given data.
       * 
       * @param    className
       *           The class name associated with the external program.
       * @param    persistent
       *           Flag indicating if this external program is being ran persistent.
       *           
       * @throws   Exception
       *           If the class can't be resolved. 
       */
      public ExternalProcedureCaller(String className, boolean persistent)
      throws Exception
      {
         this.methodName = "execute";
         this.className = className;
         
         Class<?> cls = classCache.get(className);
         if (cls == null)
         {
            cls = Class.forName(className);
            classCache.putIfAbsent(className, cls);
         }
         
         // activate collection of all undoable vars defined in this procedure; this flag will be
         // unset only when the external program starts.
         if (persistent)
         {
            TransactionManager.trackUndoables();
         }
         
         this.callerInstance = cls.newInstance();
      }

      @Override
      String getClassName()
      {
         return className;
      }

      @Override
      String getInternalEntryName()
      {
         return "";
      }
      
      @Override
      String getParameterModes(boolean function)
      {
         String pname = ProcedureManager.getAbsoluteName(callerInstance);
         return SourceNameMapper.getParameterModes(pname);
      }
      
      @Override
      handle getPhandle()
      {
         // For external procedures, the phandle field is always null.
         return null;
      }
   }

   /**
    * Abstract class. Provides APIs to determine the correct target of a 
    * function or procedure invocation statement. 
    */
   private static abstract class Resolver
   {
      /**
       * Based on the given data, determine the procedure handle to which the
       * internal entry belongs and its java method name (or the external
       * program to be invoked).
       * 
       * @param    phandle
       *           The procedure handle where this internal entry is defined.
       * @param    name
       *           The legacy name of the internal entry or external program.
       * @param    function
       *           <code>true</code> if need to search for functions. 
       *           <code>false</code> if need to search for procedures.
       * @param    superCall
       *           <code>true</code> if this is a RUN SUPER or SUPER() call.
       * @param    exports
       *           The list of allowed external procedures to be invoked with a RUN ... ON SERVER
       *           statement. 
       * @param    persistent
       *           Flag indicating if this external program is being ran persistent.
       *           
       * @return   An {@link InternalEntryCaller caller} or null if not found.
       */
      abstract InternalEntryCaller resolve(handle    phandle, 
                                           String    name,
                                           boolean   function,
                                           boolean   superCall,
                                           character exports,
                                           boolean   persistent);

      /**
       * Based on the given data, determine the procedure handle to which the
       * internal entry belongs and its java method name (or the external
       * program to be invoked).
       * 
       * @param    phandle
       *           The procedure handle where this internal entry is defined.
       * @param    name
       *           The legacy name of the internal entry or external program.
       * @param    function
       *           <code>true</code> if need to search for functions. 
       *           <code>false</code> if need to search for procedures.
       * @param    superCall
       *           <code>true</code> if this is a RUN SUPER or SUPER() call.
       * @param    exports
       *           The list of allowed external procedures to be invoked with a RUN ... ON SERVER
       *           statement. 
       *           
       * @return   An {@link InternalEntryCaller caller} or null if not found.
       */
      InternalEntryCaller resolve(handle    phandle, 
                                  String    name,
                                  boolean   function,
                                  boolean   superCall,
                                  character exports)
      {
         return resolve(phandle, name, function, superCall, exports, false);
      }
   }

   /** 
    * Search through the handle's defined internal-entries for a match.
    */
   private static class InternalResolver
   extends Resolver
   {
      @Override
      public InternalEntryCaller resolve(handle    phandle, 
                                         String    ieName, 
                                         boolean   function,
                                         boolean   superCall,
                                         character exports,
                                         boolean   persistent)
      {
         String pname = ProcedureManager.getAbsoluteName(phandle.get());

         String methodName = SourceNameMapper.getMethodName(pname, ieName, function);
         if (methodName == null)
         {
            return null;
         }

         if (SourceNameMapper.isPrivate(pname, ieName, function))
         {
            handle thisp = ProcedureManager.thisProcedure();
            if (!CompareOps.equals(thisp, phandle))
            {
               return null;
            }
         }

         InternalEntryCaller caller = null;
         
         String libname = SourceNameMapper.getLibraryName(pname, ieName);
         String mapTo   = (function ? SourceNameMapper.getMapTo(pname, ieName) : null);
         
         if (libname != null)
         {
            caller = overrideNativeCall(libname, pname, ieName);
            if (caller == null)
            {
               caller = new NativeProcedureCaller(libname, pname, ieName);
            }
         }
         else if (mapTo != null)
         {
            // go recursively
            InternalEntryCaller res = resolve(phandle, mapTo, function, superCall, exports, false);
            
            if (res == null)
            {
               res = new MapToNotFound(mapTo);
            }
            
            return res;
         }
         else if (!SourceNameMapper.isSuper(pname, ieName, function))
         {
            handle hcaller = phandle;
            
            if (SourceNameMapper.isInHandle(pname, ieName, function))
            {
               hcaller = ProcedureManager.getHandleForFunction(phandle, ieName);
            }
            
            caller = new InternalEntryCaller(hcaller, methodName, ieName);
         }
         /*
         else if (!superCall)
         {
            // IN SUPER functions and procedures are used only when this
            // call is not from a RUN SUPER or SUPER call.
            Resolver r = work.obtain().internalProcSuperResolver;
            return r.resolve(phandle, ieName, function, superCall);
         }*/
         
         return caller;
      }
   }

   /**
    * Search through the given super-procedure list for a match.
    */
   private static abstract class SuperResolver
   extends InternalResolver
   {
      private final Set<Object> used = new HashSet<>();
      
      protected InternalEntryCaller resolveSuperProc(List<handle> superProcs,
                                                     String       ieName,
                                                     boolean      function,
                                                     boolean      superCall,
                                                     character    exports)
      {
         // search is done backwards (last set super-procedure is searched
         // first)
         for (handle sphandle : superProcs)
         {
            Object referent = sphandle.get();
            
            if (used.contains(referent))
            {
               continue;
            }
            
            used.add(referent);
            try
            {
               // when looking for an internal-entry, only the direct super-
               // procedures are searched. no recursivity is done.
               InternalEntryCaller caller = super.resolve(sphandle,
                                                          ieName,
                                                          function,
                                                          superCall,
                                                          exports,
                                                          false);
               if (caller != null)
               {
                  return caller;
               }
            }
            finally
            {
               used.remove(referent);
            }
         }
         
         return null;
      }
   }

   /**
    * Search through the super-procedures of the given procedure handle for
    * a match.
    */
   private static class InternalProcSuperResolver
   extends SuperResolver
   {
      @Override
      public InternalEntryCaller resolve(handle    phandle,
                                         String    ieName,
                                         boolean   function,
                                         boolean   superCall,
                                         character exports,
                                         boolean   persistent)
      {
         List<handle> superProcs = ProcedureManager.getSuperProcedures(phandle.get());

         return resolveSuperProc(superProcs, ieName, function, superCall, exports);
      }
   }

   /**
    * Search through the session's super-procedures of the handle for a match.
    */
   private static class SessionSuperResolver
   extends SuperResolver
   {
      @Override
      public InternalEntryCaller resolve(handle    phandle,
                                         String    ieName,
                                         boolean   function,
                                         boolean   superCall,
                                         character exports,
                                         boolean   persistent)
      {
         List<handle> superProcs = ProcedureManager.getSuperProcedures();

         return resolveSuperProc(superProcs, ieName, function, superCall, exports);
      }
   }

   /**
    * Search through the available external programs for a match.
    */
   private static class ExternalProgramResolver
   extends Resolver
   {
      @Override
      public InternalEntryCaller resolve(handle    phandle,
                                         String    name,
                                         boolean   function,
                                         boolean   superCall,
                                         character exports,
                                         boolean   persistent)
      {
         if (function)
         {
            // for functions, don't search external procedures
            return null;
         }
         
         if (exports != null      &&
             !exports.isUnknown() &&
             !TextOps.matchesList(exports, name).booleanValue())
         {
            // in case this is call is via an appserver, only the exported procedures can be used
            return null;
         }
         
         name = SourceNameMapper.normalizeLegacyName(name);
         
         String className = SourceNameMapper.convertNameToClass(name);
         
         InternalEntryCaller caller = null;
         if (className != null)
         {
            // is a registered external procedure
            try
            {
               // set the 4GL ext prog being instantiated; this is required to expose the 4GL prog
               // name to any errors generated during instantiation (i.e. shared vars error msgs 
               // need to know the procedure name where they are defined, if an error happens)
               ProcedureManager.setInstantingExternalProgram(name);
               caller = new ExternalProcedureCaller(className, persistent);
            }
            catch (Throwable t)
            {
               // cleanup all pending (to be registered) shared resources
               TransactionManager.cleanupPending();
               SharedVariableManager.cleanupPending();
               LogicalTerminal.cleanupPending();
               ProcedureManager.cleanupPending();
               BufferManager.cleanupPending();
               
               Throwable cause = t;
               while (cause != null)
               {
                  // if the exception is a condition exception, let it propagate up the stack
                  if (cause instanceof ErrorConditionException)
                  {
                     // ERROR conditions raised during instantiation can be either logged
                     // (if NO-ERROR is in effect) or re-thrown.  Let the ErrorManager decide
                     // what to do; in any case, this returns null
                     ErrorConditionException ece = (ErrorConditionException) cause;
                     String msg = ece.getMessage();
                     int num = ece.getProgressErrorCode();
                     
                     ErrorManager.recordOrThrowError(num, msg, false);
                     return null;
                  }
                  else if (cause instanceof ConditionException)
                  {
                     throw (ConditionException) cause;
                  }
                  if (cause instanceof RetryUnwindException)
                  {
                     throw (RetryUnwindException) cause;
                  }
                  
                  cause = cause.getCause();
               }
               
               // if the exception was not propagated, log it, as something severe had happened.
               LOG.log(Level.SEVERE, "Unable to resolve external program.", t);
               
               return null;
            }
            finally
            {
               ProcedureManager.setInstantingExternalProgram(null);
            }
         }
         
         return caller;
      }
   }

   /**
    * An interface for which its implementation will resolve the arguments passed to a 
    * <code>ControlFlowOps.invoke</code> API.
    */
   private static interface ArgumentResolver
   {
      /**
       * Resolve the arguments.
       * 
       * @param    caller
       *           The resolved caller, to which the arguments will be passed.
       * @param    pname
       *           The caller's legacy name.
       * @param    function
       *           Flag indicating if this is a 4GL function call (when <code>true</code>).
       *           
       * @return   An array with the resolved arguments.
       */
      public Object[] resolve(InternalEntryCaller caller, String pname, boolean function);
   }
   
   /**
    * This implementation resolves arguments for converted 4GL statements (RUN, function calls,
    * DYNAMIC-FUNCTION).  It just passes back the received array.
    */
   private static class ArrayArgumentResolver
   implements ArgumentResolver
   {
      /** The arguments used for this 4GL-style call. */
      private final Object[] args;

      /**
       * Create a new instance and save the arguments.
       * 
       * @param    args
       *           The arguments for this 4GL-style call.
       */
      public ArrayArgumentResolver(Object[] args)
      {
         this.args = args;
      }

      /**
       * Resolve the arguments by returning back the {@link #args} received on instantiation.
       * 
       * @param    caller
       *           The resolved caller, to which the arguments will be passed.
       * @param    pname
       *           The caller's legacy name.
       * @param    function
       *           Flag indicating if this is a 4GL function call (when <code>true</code>).
       *           
       * @return   The initial {@link #args}.
       */
      @Override
      public Object[] resolve(InternalEntryCaller caller, String pname, boolean function)
      {
         return args;
      }
   }
   
   /**
    * This is a special implementation of resolving arguments received in a {@link Map}, where
    * the key is the legacy argument's name (case insensitive) and the value is a string-encoded
    * representation of this argument.
    * <p>
    * When resolving the arguments, they are matched against the caller's signature and an array
    * is constructed, so that the argument order matches the caller's signature.  If an argument
    * is not specified in the {@link #args map}, an unknown value is assumed for it.
    */
   private static class MapArgumentResolver
   implements ArgumentResolver
   {
      /**
       * A map with string representation of the 4GL supported types to their P2J implementation 
       * class.
       */
      private static final Map<String, Class<? extends BaseDataType>> SUPPORTED_TYPES;
      
      static
      {
         Map<String, Class<? extends BaseDataType>> types = new HashMap<>();
         types.put("RAW", raw.class);
         types.put("DATE", date.class);
         types.put("DATETIME", datetime.class);
         types.put("DATETIMETZ", datetimetz.class);
         types.put("LOGICAL", logical.class);
         types.put("DECIMAL", decimal.class);
         types.put("INT64", int64.class);
         types.put("INTEGER", integer.class);
         types.put("RECID", recid.class);
         types.put("ROWID", rowid.class);
         types.put("CHARACTER", character.class);
         types.put("LONGCHAR", longchar.class);
         
         SUPPORTED_TYPES = Collections.unmodifiableMap(types);
      }
      
      /** The argument map. */
      private final Map<String, String> args;

      /** The {@link #args} converted/casted to their associated P2J type. */
      private final Map<String, BaseDataType> resolvedArgs = new HashMap<>();

      /** A set of 4GL legacy parameter names, which are marked as OUTPUT or INPUT-OUTPUT. */
      private Set<String> outputArguments = new HashSet<>();
      
      /**
       * Create a new instance.
       * 
       * @param    args
       *           The map of arguments.
       */
      public MapArgumentResolver(Map<String, String> args)
      {
         this.args = new HashMap<>();
         for (String key : args.keySet())
         {
            this.args.put(key.toLowerCase(), args.get(key));
         }
      }

      /**
       * Resolve the arguments which will be passed to the caller.
       * 
       * @param    caller
       *           The resolved caller, to which the arguments will be passed.
       * @param    name
       *           The caller's legacy name.
       * @param    function
       *           Flag indicating if this is a 4GL function call (when <code>true</code>).
       *           
       * @return   The resolved arguments.
       * 
       * @throws   ErrorConditionException
       *           If the arguments can't be matched against the caller's signature (by number,
       *           name or type).
       */
      @Override
      public Object[] resolve(InternalEntryCaller caller, String name, boolean function)
      {
         List<Parameter> params = caller.getParameters(function);
         if (params == null || params.isEmpty())
         {
            // no parameters, check if the arg list is empty!
            if (!args.isEmpty())
            {
               // TODO: this has to do with the MSG_P2J_INVOKE incoming from the JS side, in
               //       embedded mode; this is a FWD feature with no 4GL counterpart, so we can
               //       change it to use recordOrThrowError(), as long as the remote JS side is
               //       properly informed of the error; this code was originally adapted from
               //       ControlFlowOps.validArguments 
               String msg = "Remote invoke passed parameters to %s, which did not expect any";
               msg = String.format(msg, name);
               throw new ErrorConditionException(1005, msg);
            }

            return new Object[0];
         }

         Object[] res = new Object[params.size()];
         Set<String> unused = new HashSet<>();
         unused.addAll(args.keySet());

         String modes = caller.getParameterModes(function);
         for (int i = 0; i < modes.length(); i++)
         {
            Parameter param = params.get(i);
            String pname = param.getLegacyName().toLowerCase();
            String ptype = param.getType();
            
            if (param.isExtent())
            {
               // TODO: this is not safe for silent error mode, ErrorManager.recordOrThrowError()
               //       should be used OR ErrorManager.throwError(NE, boolean) but generally we
               //       should not be instantiating this exception directly
               String msg = "Argument %s with type %s is defined EXTENT, which is not supported.";
               throw new ErrorConditionException(5729, String.format(msg, pname, ptype));
            }
            
            if (modes.charAt(i) != 'I')
            {
               outputArguments.add(pname);
            }
            
            Class<? extends BaseDataType> pclass = SUPPORTED_TYPES.get(ptype.toUpperCase());

            if (pclass == null)
            {
               // TODO: this is not safe for silent error mode, ErrorManager.recordOrThrowError()
               //       should be used OR ErrorManager.throwError(NE, boolean) but generally we
               //       should not be instantiating this exception directly
               String msg = "Argument %s has an unsupported type: %s";
               throw new ErrorConditionException(5729, String.format(msg, pname, ptype));
            }
            
            // support only BDT sub-classes (check ptype against a list of BDT sub-classes)
            
            String sval = args.get(pname);
            BaseDataType aval = null;
            if (sval == null)
            {
               // build a new instance, based on parameter's type
               aval = BaseDataType.generateUnknown(pclass);
            }
            else
            {
               Runnable error = () ->
               {
                  // raise error, could not convert!
                  String msg = "Remote called routine %s %s has mismatched parameters";
                  msg = String.format(msg, caller.getInternalEntryName(), name);

                  // TODO: this is not safe for silent error mode, recordOrThrowError()
                  //       should be used OR throwError(NE, boolean) but generally we
                  //       should not be instantiating this exception directly
                  throw new ErrorConditionException(5729, msg);
               };
               
               try
               {
                  // build a new instance using the BaseDataType(String) c'tor
                  aval = pclass.getConstructor(String.class).newInstance(sval);
               }
               catch (Exception expt)
               {
                  error.run();
               }
               
               if (ErrorManager.isPendingError())
               {
                  error.run();
               }
            }
            
            // this argument is used, save it
            unused.remove(pname);
            res[i] = aval;
            
            resolvedArgs.put(pname, aval);
         }
         
         if (!unused.isEmpty())
         {
            // TODO: this is not safe for silent error mode, ErrorManager.recordOrThrowError()
            //       should be used OR ErrorManager.throwError(NE, boolean) but generally we
            //       should not be instantiating this exception directly
            
            // some arguments were received which couldn't be matched - raise ERROR
            String msg = "Mismatched number of parameters passed to routine %s %s: %s";
            msg = String.format(msg, caller.getInternalEntryName(), name, unused.toString());
            throw new ErrorConditionException(3234, msg);
         }
         
         return res;
      }
      
      /**
       * Get a map with the OUTPUT or INPUT-OUTPUT arguments, with their state as after the caller
       * has been executed.  The keys will be the legacy 4GL names and the values will be a string
       * representation of each OUTPUT or INPUT-OUTPUT argument.
       * 
       * @return   See above.
       */
      public Map<String, String> getOutputArguments()
      {
         Map<String, String> res = new HashMap<>();
         for (String key : resolvedArgs.keySet())
         {
            if (outputArguments.contains(key))
            {
               res.put(key, resolvedArgs.get(key).toStringMessage());
            }
         }
         
         return res;
      }
   }
   
   /**
    * Helper class which also saves the {@link AsyncRequestImpl} instance, needed by the remote
    * request.
    */
   private static abstract class AsyncCall
   implements Runnable
   {
      /** The async request instance for this call. */
      public final AsyncRequestImpl asyncReq;

      /** The server used to execute this async call. */
      public final handle hServer;
      
      /** A handle variable where to save the web service. May be null. */
      public final handle portHandle;
      
      /** The target procedure name. */
      public final character pName;
      
      /** The handle to which the procedure belongs, if it is RUN ... IN handle statement. */
      public final handle pHandle;
      
      /** The handle to the asynchronous request. */
      public final handle asyncHandle;
      
      /** This represents the name of the internal procedure for the EVENT-PROCEDURE clause. */
      public final character eventProcName;
      
      /** The handle to a procedure in which context the EVENT-PROCEDURE is found. */
      public final handle inEventprocHandle;
      
      /**
       * This marks if "TRANSACTION DISTINCT" option is specified with the RUN ... ON statement.
       */
      public final boolean transactionDistinct;
      
      /**
       * 
       * @param    hServer
       *           The server used to execute this async call.
       * @param    portHandle
       *           A handle variable where to save the web service. May be null.
       * @param    pName
       *           The target procedure name.
       * @param    pHandle
       *           The handle to which the procedure belongs, if it is RUN ... IN handle 
       *           statement.
       * @param    asyncHandle
       *           The handle to the asynchronous request.
       * @param    eventProcName
       *           This represents the name of the internal procedure for the EVENT-PROCEDURE 
       *           clause, null if this clause is not specified.
       * @param    inEventprocHandle
       *           The handle to a procedure in which context the specified EVENT-PROCEDURE internal
       *           procedure is found, null if EVENT-PROCEDURE clause is not specified.
       * @param    transactionDistinct
       *           This marks if "TRANSACTION DISTINCT" option is specified with the RUN ... ON 
       *           statement.
       * @param    modes
       *           A string representation of the modes of each parameter.
       * @param    args
       *           The procedure's arguments.
      */
      public AsyncCall(handle     hServer,
                       handle     portHandle,
                       character  pName,
                       handle     pHandle,
                       handle     asyncHandle,
                       character  eventProcName,
                       handle     inEventprocHandle,
                       boolean    transactionDistinct,
                       String     modes,
                       Object...  args)
      {
         this.hServer = new handle(hServer);
         this.portHandle = portHandle == null ? null : new handle(portHandle);
         this.pName = new character(pName);
         this.pHandle = pHandle == null ? null : new handle(pHandle);
         this.asyncHandle = asyncHandle == null ? null : new handle(asyncHandle);
         this.eventProcName = new character(eventProcName);
         this.inEventprocHandle = new handle(inEventprocHandle);
         this.transactionDistinct = transactionDistinct;

         this.asyncReq = new AsyncRequestImpl(this.hServer,
                                              this,
                                              this.eventProcName,
                                              this.inEventprocHandle,
                                              this.pName,
                                              this.pHandle,
                                              modes,
                                              args);
      }
   }   
}